SystemOrganization addCategory: #'Qwaq-Objects'!
SystemOrganization addCategory: #'Qwaq-Objects-Document'!
SystemOrganization addCategory: #'Qwaq-Objects-General'!
SystemOrganization addCategory: #'Qwaq-Objects-Hacks'!
SystemOrganization addCategory: #'Qwaq-Objects-Jobs'!
SystemOrganization addCategory: #'Qwaq-Objects-Kernel'!
SystemOrganization addCategory: #'Qwaq-Objects-Logging'!
SystemOrganization addCategory: #'Qwaq-Objects-Manipulation'!
SystemOrganization addCategory: #'Qwaq-Objects-Motion'!
SystemOrganization addCategory: #'Qwaq-Objects-Physics'!
SystemOrganization addCategory: #'Qwaq-Objects-Places'!
SystemOrganization addCategory: #'Qwaq-Objects-Resources'!
SystemOrganization addCategory: #'Qwaq-Objects-Space'!
SystemOrganization addCategory: #'Qwaq-Objects-User'!
SystemOrganization addCategory: #'Qwaq-Objects-Widgets'!

TAvatarDriver subclass: #QAvatarDriver
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Objects-User'!

!QAvatarDriver class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QAvatarDriver methodsFor: 'driving' stamp: 'bgf 12/10/2009 14:14'!
drive

	"We override the default drive method
	 to intercept and handle a list of symbols in 'speedControl'.
	 The base TAvatarReplica expects speedControl to be a point in [-1,1]@[-1,1]
	 representing the mouse position within the viewing area.
	 QAvatar extends this to allow symbolic driving selectors for keyboard-based driving;
	 the symbols are driving methods, associated with keys via QDriveKeyMap."

	self sticky: false.

	avatar driveSymbols ifNotEmptyDo: [ : syms | 
		syms do: [ : sym | avatar perform: sym ].
	] ifEmpty: [
		super drive.
	].
	self window ifNotNilDo: [ : win | win updateLastEventTime ].
! !

!QAvatarDriver methodsFor: 'accessing' stamp: 'bgf 10/9/2007 23:41'!
stickyViewpoint: aStringOrScalarOrNil

	"We extend the TAvatarDriver method to accept
	 viewpoint symbolic names for sticky POVs.
	 The allowable names come from the UserPreferences viewpoint table
	 (so they match one the Viewpoint menu entries in the main UI and in preferences.)"

	aStringOrScalarOrNil isString ifTrue: [
		super stickyViewpoint: (QUserPreferences viewpointTable at: aStringOrScalarOrNil)
	] ifFalse: [
		super stickyViewpoint: aStringOrScalarOrNil
	]
! !

!QAvatarDriver methodsFor: 'as yet unclassified' stamp: 'bgf 12/10/2009 14:17'!
window
	"My TWindowedSession, or nil"
	^ avatar ifNotNil: [ avatar view ifNotNilDo: [ : view | view  window ]]! !

Object subclass: #BVHParser
	instanceVariableNames: 'motion jointRoot joints currentJoint channels fileName scale errors currentLine currentLineNumber currentTransform fromBlender switchYZ oddOffset rootName'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Objects-Motion'!

!BVHParser commentStamp: '<historical>' prior: 0!
This parser takes in bvh files. It also makes a TJoint structure out of it, which can be accesses later.

	fileName := FileDirectory pathFrom: { bvhFileName }.
	parser := BVHParser initializeWithFileName: fileName scale: scale.

	parser motion " returns the motion object "
	parser jointRoot " returns the joint structure "!

!BVHParser class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!BVHParser class methodsFor: 'instance creation' stamp: 'zamp 12/9/2005 09:14'!
initializeWithFileName: fileName
	^self initializeWithFileName: fileName scale: 1.0.! !

!BVHParser class methodsFor: 'instance creation' stamp: 'zamp 12/9/2005 09:14'!
initializeWithFileName: fileName scale: scl
	^self new initializeWithFileName: fileName scale: scl.! !

!BVHParser class methodsFor: 'instance creation' stamp: 'zamp 12/9/2005 09:14'!
initializeWithFileName: fileName scale: scl fromBlender: bool
	^self new initializeWithFileName: fileName scale: scl fromBlender: bool.! !

!BVHParser class methodsFor: 'instance creation' stamp: 'zamp 12/9/2005 09:14'!
initializeWithFileName: fileName scale: scl fromBlender: bool1 switchYZ: bool2
	^self new initializeWithFileName: fileName scale: scl fromBlender: bool1 switchYZ: bool2.! !

!BVHParser class methodsFor: 'instance creation' stamp: 'zamp 12/9/2005 09:14'!
initializeWithFileName: fileName scale: scl fromBlender: bool1 switchYZ: bool2 oddOffset: bool3
	^self new initializeWithFileName: fileName scale: scl fromBlender: bool1 switchYZ: bool2 oddOffset: bool3.! !

!BVHParser class methodsFor: 'instance creation' stamp: 'zamp 12/9/2005 09:14'!
new
	^ super new initialize! !

!BVHParser methodsFor: 'converting' stamp: 'hrs 9/12/2007 21:58'!
addBone: joint parent: parent skeleton: skeleton
	| frameBone jointName |

	jointName := joint objectName.
	jointName = 'Site' ifTrue: [jointName := parent name, 'End'].

	frameBone := TSkeletonBone new initializeWithName: jointName
			 initialTransform: joint initialTransform.
	frameBone index: (joints indexOf: joint) .

	(parent = nil)
		ifTrue: [ skeleton rootBone: frameBone ] "set the skeleton root bone"
		ifFalse: [ parent addChild: frameBone ].

	skeleton bones at: frameBone index put: frameBone.

	"Recursively call this method for the children."
	joint frameChildrenDo: [ :child | self addBone: child parent: frameBone skeleton: skeleton ].

! !

!BVHParser methodsFor: 'converting' stamp: 'hrs 8/27/2007 11:46'!
buildAnimations: skeleton
	
	| animations animation |
	animations := Dictionary new.

	"If a BVH file could have multiple motions, they would each be processed here."
	animation := motion makeSkeletalAnimation.
	animations at: animation name put: animation.

	skeleton addAnimations: animations.
	^ animations! !

!BVHParser methodsFor: 'converting' stamp: 'hrs 9/13/2007 10:12'!
buildSkeleton
	
	| skeleton rootBone|
	skeleton := TSkeleton new.
	skeleton bones: (Array new: joints size).

	rootBone := jointRoot frameChildren first .
	skeleton name: motion motionName.

	self addBone: rootBone parent: nil skeleton: skeleton.

	self buildAnimations: skeleton.

	"An assumption, but you can change it with #boneMap:."
	skeleton boneMap: TBoneMap fromSecondLife.

	^ skeleton! !

!BVHParser methodsFor: 'parse' stamp: 'zamp 5/5/2006 09:48'!
doChannels: args
	| channelCount error channel channelName channelDictionary jointChannels |

	error := false.
	(args size > 2) ifTrue: [
		channelCount := args at: 2.
		((channelCount + 2) <= args size) ifTrue: [
			1 to: channelCount do: [ :index |
				channel := TChannel new.

				channelName := (args at: (index + 2)).
				switchYZ ifTrue: [
					channelName = 'Zrotation' ifTrue: [
						channelName := 'Yrotation'
					]
					ifFalse: [
						channelName = 'Yrotation' ifTrue: [
							channelName := 'Zrotation'
						].
					].
				].

				channel channelName: channelName.
				channel jointName: currentJoint objectName.
				channels add: channel.

				" Store the saved motion data. "
				channelDictionary := motion channelsDictionary.
				jointChannels := channelDictionary at: (currentJoint objectName) ifAbsent: [
					jointChannels := OrderedCollection new.
					channelDictionary at: (currentJoint objectName) put: jointChannels.
					jointChannels
				].
				jointChannels add: channel.
			].
		]
 		ifFalse: [ error := true ].
	]
	ifFalse: [ error := true ].

	error ifTrue: [
		errors add: 'Invalid arguments for CHANNELS: line ', currentLineNumber, String cr.
	].

! !

!BVHParser methodsFor: 'parse' stamp: 'zamp 6/26/2006 13:24'!
doEndJoint: aStream
	currentJoint parent ifNotNil: [
		(currentJoint parent class = TJointRoot) ifFalse: [
			currentTransform translation: currentJoint parent initialTransform translation.
		].

		currentJoint := currentJoint parent.
	].
! !

!BVHParser methodsFor: 'parse' stamp: 'zamp 12/9/2005 09:14'!
doFile: aStream
	| args firstArg |

	" Get the next line from the stream and determine what to do. "
	currentLine := (self getNextLine: aStream) withBlanksTrimmed.
	args := currentLine findTokens: Character separators.

	(args isEmpty) ifFalse: [
		firstArg := args at: 1.
		firstArg caseOf: {
			[ #HIERARCHY ] -> [ self doHierarchy: aStream ].
			[ #MOTION ] -> [ self doMotion: aStream ].
		}
		otherwise: [
			" This part of the file is unknown. Show an error. "
			errors add: 'Invalid Format For BVH: line ', currentLineNumber, String cr.
		].
	].
! !

!BVHParser methodsFor: 'parse' stamp: 'zamp 1/20/2006 10:49'!
doFrameCount: aStream
	| args |

	currentLine := (self getNextLine: aStream) withBlanksTrimmed.
	args := currentLine findTokens: Character separators.

	((args at: 1) = 'Frames:') ifTrue: [
		motion frameCount: ((args at: 2) asNumber asFloat).
	]
	ifFalse: [
		errors add: 'Expected number of frames: line ', currentLineNumber, String cr.
	].
! !

!BVHParser methodsFor: 'parse' stamp: 'zamp 1/20/2006 10:49'!
doFrequency: aStream
	| args error |

	currentLine := (self getNextLine: aStream) withBlanksTrimmed.
	args := currentLine findTokens: Character separators.
	error := false.

	((args at: 1) = 'Frame') ifTrue: [
		((args at: 2) = 'Time:') ifTrue: [
			motion frequency: ((args at: 3) asNumber asFloat).
		] ifFalse: [ error := true ].
	] ifFalse: [ error := true ].

	error ifTrue: [
		errors add: 'Expected frequency: line ', currentLineNumber, String cr.
	].
! !

!BVHParser methodsFor: 'parse' stamp: 'das 2/15/2007 00:38'!
doHierarchy: aStream
	| args firstArg doingHierarchy |

	doingHierarchy := true.

	[ doingHierarchy ] whileTrue: [
		" Get the next line from the stream and determine what to do. "
		currentLine := (self getNextLine: aStream) withBlanksTrimmed.
		args := currentLine findTokens: Character separators.

		(args isEmpty) ifFalse: [
			firstArg := args at: 1.
			firstArg caseOf: {
				[ #ROOT ] -> [ self doStartJoint: aStream isRoot: true.].
				[ #JOINT ] -> [ self doStartJoint: aStream isRoot: false].
				[ #End ] -> [ self doStartJoint: aStream isRoot: false ].
				[ #} ] -> [ self doEndJoint: aStream ].
			}
			otherwise: [
				" The line didn't match. Put it back on the stream "
				" so that another method can handle it. "
				self putLineBackOnStream: aStream.
				doingHierarchy := false.
			].
		].
	].
! !

!BVHParser methodsFor: 'parse' stamp: 'ar 3/18/2007 20:19'!
doMotion: aStream
	| frame buffer line count value transformz |

	" Next should be the number or frames "
	self doFrameCount: aStream.

	" Next should be the frequency. "
	self doFrequency: aStream.

	buffer := Array new: channels size.
	frame := 1.

	transformz := channels collect:[:channel| | block |
		block := [:arg| arg].
		" The BVH export tool in Blender faces it in the wrong direction. "
		fromBlender ifTrue: [
			channel channelName = 'Xrotation' ifTrue: [
				block := [:arg| (arg + 90) negated].
			].
			channel channelName = 'Zrotation' ifTrue: [
				block := [:arg| arg negated].
			].
		].
		channel channelName = 'Xposition' ifTrue:[block := [:arg| arg * scale]].
		channel channelName = 'Yposition' ifTrue:[block := [:arg| arg * scale]].
		channel channelName = 'Zposition' ifTrue:[block := [:arg| arg * scale]].
		block].

	[ aStream atEnd ] whileFalse: [
		" Get the next line from the stream and determine what to do. "
		currentLineNumber := currentLineNumber + 1.
		line := aStream nextLine readStream.
		count := 0.
		[value := line nextFloat.
		value == nil] whileFalse:[buffer at: (count := count+1) put: value].

		" Add each line of channel data. "
		(count > 0) ifTrue: [
			(count = channels size) ifFalse: [^self error: 'Incorrect number of values'].
			1 to: count do: [ :index |
				(channels at: index) values add: ((transformz at: index) value: (buffer at: index)).
			].
		].

		frame := frame + 1.
	].

! !

!BVHParser methodsFor: 'parse' stamp: 'zamp 5/9/2006 17:07'!
doOffset: args
	| xPos yPos zPos offset newTranslation |

	(args size > 3) ifTrue: [
		xPos := ((args at: 2) asNumber asFloat) * scale.
		yPos := ((args at: 3) asNumber asFloat) * scale.
		zPos := ((args at: 4) asNumber asFloat) * scale.
		switchYZ ifTrue: [
			offset := Vector3 x: xPos y: zPos z: yPos.
		]
		ifFalse: [
			offset := Vector3 x: xPos y: yPos z: zPos.
		].
		newTranslation := offset.

		" Some of the BVH files had issues with the offset (I think). "
		oddOffset ifTrue: [
			(offset x > 0) ifTrue: [
				newTranslation x: (offset x -
						(currentJoint parent localToGlobal: currentJoint translation) x).
			].
		]
		ifFalse: [
			fromBlender ifTrue: [
				newTranslation := (currentJoint globalToLocal: newTranslation).
			].
		].

		currentTransform translation: newTranslation.
		currentJoint translation: currentTransform translation.
		currentJoint initialTransform: currentTransform copy.
	]
	ifFalse: [
		errors add: 'Invalid arguments for OFFSET: line ', currentLineNumber, String cr.
	].

! !

!BVHParser methodsFor: 'parse' stamp: 'das 2/14/2007 23:55'!
doStartJoint: aStream isRoot: isRoot
	| args firstArg doingJoint |

	" Create a new joint. "
	args := currentLine findTokens: Character separators.
	self newJoint: args isRoot: isRoot.
	doingJoint := true.

	" Check to see that the format is correct. Print out error if it isn't."
	currentLine := (self getNextLine: aStream) withBlanksTrimmed.
	(currentLine = '{') ifFalse: [
		errors add: 'Invalid Format for a JOINT: line: ', currentLineNumber, String cr.
		doingJoint := false.
	].

	[ doingJoint ] whileTrue: [
		" Get the next line from the stream and determine what to do. "
		currentLine := (self getNextLine: aStream) withBlanksTrimmed.
		args := currentLine findTokens: Character separators.

		(args isEmpty) ifFalse: [
			firstArg := args at: 1.
			firstArg caseOf: {
				[ #OFFSET ] -> [ self doOffset: args ].
				[ #CHANNELS ] -> [ self doChannels: args ].
			}
			otherwise: [
				" The line didn't match. Put it back on the stream "
				" so that another method can handle it. "
				self putLineBackOnStream: aStream.
				doingJoint := false.
			].
		].
	].
! !

!BVHParser methodsFor: 'parse' stamp: 'ar 3/18/2007 19:49'!
getNextLine: aStream
	currentLineNumber := currentLineNumber + 1.
	^aStream nextLine! !

!BVHParser methodsFor: 'initialize' stamp: 'jdm 8/21/2007 12:31'!
initialize
	currentLine := ''.
	currentLineNumber := 0.
	errors := OrderedCollection new.
	currentJoint := nil.
	currentTransform := Matrix4x4 new.
	fileName := nil.
	scale := 1.0.
	fromBlender := false.
	switchYZ := false.
	oddOffset := false.

	motion := TMotion new.
	jointRoot := TJointRoot new.
	joints := OrderedCollection new.
	channels := OrderedCollection new.

	jointRoot myScale: scale.
	
	self rootName: 'Root'.
! !

!BVHParser methodsFor: 'initialize' stamp: 'zamp 12/9/2005 09:14'!
initializeWithFileName: fName 
	^ self initializeWithFileName: fName scale: 1! !

!BVHParser methodsFor: 'initialize' stamp: 'zamp 12/9/2005 09:14'!
initializeWithFileName: fName scale: scl
	self initializeWithFileName: fName scale: scl fromBlender: false.! !

!BVHParser methodsFor: 'initialize' stamp: 'zamp 12/9/2005 09:14'!
initializeWithFileName: fName scale: scl fromBlender: bool
	self initializeWithFileName: fName scale: scl fromBlender: bool switchYZ: false.! !

!BVHParser methodsFor: 'initialize' stamp: 'zamp 12/9/2005 09:14'!
initializeWithFileName: fName scale: scl fromBlender: bool1 switchYZ: bool2
	self initializeWithFileName: fName scale: scl fromBlender: bool1 switchYZ: bool2 oddOffset: false
! !

!BVHParser methodsFor: 'initialize' stamp: 'zamp 6/5/2006 16:07'!
initializeWithFileName: fName scale: scl fromBlender: bool1 switchYZ: bool2 oddOffset: bool3
	scale := scl.
	jointRoot myScale: scl.
	fromBlender := bool1.
	switchYZ := bool2.
	oddOffset := bool3.

	self parseFileNamed: fName.
! !

!BVHParser methodsFor: 'accessing' stamp: 'zamp 5/5/2006 09:53'!
jointRoot
	^jointRoot! !

!BVHParser methodsFor: 'accessing' stamp: 'zamp 12/9/2005 09:14'!
motion
	^motion! !

!BVHParser methodsFor: 'parse' stamp: 'hrs 7/25/2007 10:00'!
newJoint: args isRoot: isRoot
	" Create a new empty and set it in the motion capture. If it has "
	" a parent, draw a line to the parent. "
	| joint name idx |

	(args size > 1) ifTrue: [
		" Make a new joint. "
		name := args at: 2.
		idx := name indexOf: $:.
		idx > 0 ifTrue: [
			name := name copyFrom: 1 to: (idx - 1).
		].
		(isRoot and: [rootName notNil]) ifTrue:[name := rootName.].
		joint := TJoint new initializeWithJointRoot: jointRoot objectName: name.
		joints add: joint.

		" This is the root of the motion. "
		((args at: 1) = 'ROOT') ifTrue: [ jointRoot addChild: joint ].

		" If there's a parent, add it to the parent's children. "
		currentJoint ifNotNil: [
			currentJoint addChild: joint.
		].

		" This joint is the current joint now. "
		currentJoint := joint.
	]
	ifFalse: [
		errors add: 'Invalid arguments for JOINT: line ', currentLineNumber, String cr.
	].

! !

!BVHParser methodsFor: 'parse' stamp: 'hrs 8/27/2007 10:39'!
parseFileNamed: aString 
	| aFile aStream |

	" Open the file and parse it. "
	fileName := aString.
	aFile := CrLfFileStream readOnlyFileNamed: fileName.
	aStream := aFile binary contents asString.
	self parseStream: aStream readStream.
	motion motionName: (aFile localName upTo: $.).  "The ReadStream no longer knows."! !

!BVHParser methodsFor: 'parse' stamp: 'zamp 6/7/2006 15:04'!
parseStream: aStream
	| motionName |

	" Parses through the file, displaying a progress bar. "
	'Parsing BVH File'
		displayProgressAt: Sensor cursorPoint
		from: 0 to: 100
		during: [ :bar |
			[aStream atEnd] whileFalse: [
				self doFile: aStream.
			].
		].
	aStream close.

	" Reference this motion as the file name. "
	motionName := (aStream localName) upTo: $..
	motion motionName: motionName.

	" Print out problems parsing the file and cleanup if necessary. "
	(errors size > 0) ifTrue: [
		self showErrors.
		self rollbackObjects.
	].! !

!BVHParser methodsFor: 'parse' stamp: 'zamp 6/7/2006 15:05'!
parseStream: aStream scale: aScale
	scale := aScale.
	self parseStream: aStream.! !

!BVHParser methodsFor: 'parse' stamp: 'zamp 12/9/2005 09:14'!
putLineBackOnStream: aStream
	aStream position: (aStream position - (String cr size + String lf size)).
	aStream backUpTo: String cr.
! !

!BVHParser methodsFor: 'error handling' stamp: 'zamp 5/9/2006 16:27'!
rollbackObjects
	" Cleanup any objects we have created. "
	jointRoot destroy.	
	joints do: [ :joint | joint destroy. ].
! !

!BVHParser methodsFor: 'initialize' stamp: 'hrs 7/25/2007 09:59'!
rootName: aStringOrNil
	"There is code that expects the root name to be 'Root', so that's what initialize does. But sometimes Poser gets rid of unnecesary elements (e.g., when you tell it to autoscale on export). When it does so, the 'hip' will be the root, and we will need to match up that channel with mesh named hip, so we'd better retain the names from the file.  By sending rootName: nil, you can tell this parser to just use the names from the file."
	rootName := aStringOrNil.! !

!BVHParser methodsFor: 'error handling' stamp: 'zamp 12/9/2005 09:14'!
showErrors
	| str |

	" These are the errors that have accumulated. "
	errors do: [ :error | Transcript show: error ].

	" Last and Final message. "
	str := 'Errors occurred in file: ', fileName, String cr.
	str := str, 'Motion not created!!', String cr.
	Transcript show: str.
! !

!Object class methodsFor: '*Qwaq-Objects' stamp: 'bgf 1/28/2010 11:40'!
genericName
	"Tool support: a short string identifying this object for a status-alert
	 where the specific name isn't needed, rather just the 'kind of thing.'
 	 Useful when reporting why a direct-click is not going to give a result,
	 By default, returns the class name, but with  leading pseudo-module cap stripped.
	 I.e. Cube for TCube,  Lozenge for QLozenge, etc. "
	| nm |
	nm := self name.
	nm size > 3 ifFalse: [ ^ nm ].
	^ nm second isUppercase
			ifFalse: [ nm  ]
			ifTrue: [ nm third isUppercase
						ifTrue: [ nm allButFirst: 2 ] 	
						ifFalse: [ nm allButFirst ] ]! !

!Object methodsFor: '*Qwaq-Objects' stamp: 'bgf 1/28/2010 11:37'!
genericName
	"Tool support: a short string identifying this object for a status-alert
	 where the specific name isn't needed, rather just the 'kind of thing.'"
	^ self class genericName! !

Object subclass: #QCache
	instanceVariableNames: 'prototypeCache'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Objects-Resources'!

!QCache commentStamp: '<historical>' prior: 0!
A cache table of TIslandCopier snapshots of prototype objects,
used to accelerate creation of the second and subsequent instances of TFrame assemblies.

This is a singleton-per-island class; most usage is via class methods;
state is kept in a single instance in any given TIsland.  See #current.

Gives two benefits: speeds creation of models requiring slow construction
(such as ASE meshes), and makes their creation independent of off-island 
disk resources (after the first instance of any given object.)

Access is usually through the class methods, which always operate
on an Island-local instance of the underlying prototype table.

Their are various forms for initial construction; see the various #make:... methods.
The simplest usage is:
	QCache make: QGotoButton

This takes a class, and uses 'new' to make the first instance.
This will always return a new instance of QGotoButton.
The first time it is done in an island, it will invoke 'QGotoButton new',
cache a snapshot of that result, and then return the new button.
Subsequently, it avoids the QGotoButton path, 
and simply re-internalizes the cached snapshot.

For constructions that are not simple new instances of classes,
use 
	QCache make: aSymbol constructor: aBlockResultingInANewObject.

The GotoButton example is equivalent to:
	QCache make: #QGotoButton constructor: [QGotoButton new])
It uses the class name as symbol, and  the simple 'new' block as constructor block.

The symbol must be unique to the constructor block;
it is used as the dictionary key for the cached snapshot.

There is support for XML persistence, if desired.
To persist the factory cache for a space,  use
	QCache persistInIslandXML
in the original world-creation process.

Persisting an island's QCache saves the cached binary prototype snapshots.
Doing this creates very large XML files that are opaque, since all the factory content 
appears as base64 encoded strings.

The choice is between xml ccompactness and file readability on the one hand, 
and speed and source-file indepedence on the other.

If the QCache is not saved in XML, then you must be aware that
the XML persistence data does not include cached prototypes. 
Hence underlying resources (ASE models, for instance)
must be available when the XML is internalized, and they will be reread as needed
if further instances are generated after loading from XML.
!

!QCache class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QCache class methodsFor: 'initializing' stamp: 'bgf 1/18/2007 17:43'!
clear

	"Empty the factory cache for the current island."
	self current clear.
! !

!QCache class methodsFor: 'initializing' stamp: 'bgf 8/21/2008 19:46'!
current

	"Return the caching factory for the current island."

	| island |

	island := Processor activeIsland.
	^ island at: #QCacheCurrent
	  ifAbsent: [
		QLogger logInfo: 'Creating QCache for ', island.
		self new
	  ].! !

!QCache class methodsFor: 'factory' stamp: 'bgf 1/18/2007 17:43'!
make: aClass

	"Like 'aClass new', but uses a cached Island stream for all but the first call in an island."
	^ self current make: (aClass name) factory: aClass selector: #new.
! !

!QCache class methodsFor: 'factory' stamp: 'bgf 1/18/2007 17:43'!
make: resourceName constructor: aBlock

	"Generate an instance of the resource named 'resourceName'.
       In a given island, the first instance is produced via 'aBlock value'.
	 Subsequent instances are produced by importing a saved snapshot of the first."

	^ self current make: resourceName constructor:aBlock
! !

!QCache class methodsFor: 'factory' stamp: 'bgf 1/18/2007 18:49'!
make: aResourceName factory: anObjectOrClass selector: aSelector

	"Generate an instance of the resource named 'aResourceName'.
       In a given island, the first instance is produced as 
       the value of 'anObjectOrClass aSelector'.
	 Subsequent instances are produced by importing a saved snapshot of the first."

	^ self current make: aResourceName factory: anObjectOrClass selector: aSelector
! !

!QCache class methodsFor: 'initializing' stamp: 'bgf 1/18/2007 18:50'!
persistInIslandXML

	"Initialization choice; register the QCache for inclusion in XML snapshots of this island;
	 this speeds recovery from XML files, but makes the files themselves
 	 bigger and opaque... the cache persists as huge blocks of base64-encoded text.
	 Call this once during initial setup of a TIsland."

	(Processor activeIsland prePersistList includes: QCache current) ifFalse:[
		Processor activeIsland prePersistList add: QCache current
	].
! !

!QCache class methodsFor: 'initializing' stamp: 'bgf 1/18/2007 17:43'!
reportAll

	"QwaqFactory reportAll.    
	 Report all QwaqFactory cache content in active TIslands."
	
	UIManager default
		edit: (String streamContents: [ :s | self reportAllOn:s])
		label: 'QwaqFactory content in active islands'.! !

!QCache class methodsFor: 'initializing' stamp: 'bgf 1/18/2007 18:04'!
reportAllOn: aStream

	"Report all QCache content in known TIslands."

	(TIsland allSubInstancesDo: [ :island |
		(island at:#QCacheCurrent ifAbsent: [nil]) ifNotNilDo: [ :qs|
			aStream nextPutAll: 'QCache in Island: ', island name; cr.
			qs reportCacheOn: aStream.
			aStream cr].
	]).
! !

!QCache methodsFor: 'factory' stamp: 'bgf 1/18/2007 17:43'!
at: symbolicName

	"Find an instance of the named resource; if not in the cache, then just return nil."

	| data result |

	data := prototypeCache at: symbolicName ifAbsent: [
		^ nil
	].

	result  := TIslandCopyImporter new importData: data. 
	^ result.
! !

!QCache methodsFor: 'private' stamp: 'bgf 1/18/2007 17:43'!
at: name put: cacheData

	"private - for persistence support; allows the internalizer to install the data
	 in 'QwaqFactory current' rather than itself (a transient QF created by the persistence engine.)"

	prototypeCache at: name put: cacheData! !

!QCache methodsFor: 'initialize' stamp: 'bgf 1/18/2007 17:43'!
clear
	prototypeCache := Dictionary new.! !

!QCache methodsFor: 'factory' stamp: 'bgf 1/18/2007 17:43'!
clear: symbolicName

	"Remove the named resource from the cache, if present."

	prototypeCache removeKey:symbolicName ifAbsent:[]! !

!QCache methodsFor: 'persist' stamp: 'bgf 1/18/2007 18:03'!
constructFrom: pStream

	| tmp |

	tmp := pStream readStringMap: 'prototypeCache' default: (Dictionary new).
	tmp keysAndValuesDo: [ :key :value |
		self at: key put: (value base64Decoded asByteArray).
	].
! !

!QCache methodsFor: 'initialize' stamp: 'bgf 3/10/2010 08:08'!
initialize
	
	|  island |
	"Reading a scene always uses regular constructors, not current,
	 so when rereading from persistence, we have to replace the default instance,
	 and make sure the replacement is the one listed for subsequent persisting."

	island := Processor activeIsland.
	(island at: #QCacheCurrent) ifNotNilDo: [ :old|
		"There was a default instance; we're going to overwrite it with the 
	       'new' version constructed by readScene; so unhook the old one.
		This case can only mean that we're internalizing a TISland,
		 since that's the code path that calls #new rather than #default" 
		(island class includesBehavior: TIsland) ifTrue: [
			(island prePersistList remove: old).
		].
	].
	island at: #QCacheCurrent put: self.

	"Normal initialization starts here."
	
	prototypeCache := Dictionary new.


! !

!QCache methodsFor: 'factory' stamp: 'bgf 1/18/2007 17:43'!
make: aClass

	"Generate an instance of the class.
       In a given island, the first instance of a given class
	 is the result of 'class new'.
	 Subsequent instances are produced by importing a saved snapshot of the first."

	^ self make: (aClass name) factory: aClass selector: #new.! !

!QCache methodsFor: 'factory' stamp: 'bgf 1/18/2007 17:43'!
make: symbolicName constructor: aBlock

	"Instantiate an instance of the named resource; use the cache if possible,
       otherwise use '[aFactory aSelector]' to generate a prototype and cache that."

	| data proto result |

	data := prototypeCache at: symbolicName ifAbsent: [
		proto := aBlock value.
		data := TIslandCopier new export: proto.
		prototypeCache at: symbolicName put: data.
		^ proto
	].

	result  := TIslandCopyImporter new importData: data. 
	^ result.! !

!QCache methodsFor: 'factory' stamp: 'bgf 1/18/2007 17:43'!
make: symbolicName factory: aFactory selector: aSelector

	"Instantiate an instance of the named resource; use the cache if possible,
       otherwise use '[aFactory aSelector]' to generate a prototype and cache that."

	| data proto result |

	data := prototypeCache at: symbolicName ifAbsent: [
		proto := aFactory perform: aSelector.
		data := TIslandCopier new export: proto.
		prototypeCache at: symbolicName put: data.
		^ proto
	].

	result  := TIslandCopyImporter new importData: data. 
	^ result.! !

!QCache methodsFor: 'persist' stamp: 'bgf 1/18/2007 17:43'!
persistOn: pStream

	| tmp |
	tmp := Dictionary new.

	"Base64 encode the snapshot data before writing them as XML strings."
	 prototypeCache keysAndValuesDo: [:key :value| 
		tmp at: key put: (value asString base64Encoded).

	].
	pStream storeStringMap: 'prototypeCache' value: tmp default: (Dictionary new).! !

!QCache methodsFor: 'factory' stamp: 'bgf 1/18/2007 17:43'!
reportCache

	"Report contents to the Transcript."

	Transcript cr; show: (String streamContents: [ :s | self reportCacheOn:s]).! !

!QCache methodsFor: 'factory' stamp: 'bgf 1/18/2007 17:43'!
reportCacheOn:aStream

	"List all the entries in the factory cache."

	| total |
	total := 0.
	aStream nextPutAll: 'Qwaq Factory Cache: '; cr.
	prototypeCache keys asArray sort do: [ :key |
		aStream nextPutAll: '    ', key, ' (', (prototypeCache at: key) size, ' bytes)'; cr.
		total := total + (prototypeCache at: key) size.
	].
	aStream nextPutAll: 'Total: ', total, ' bytes in factory cache.'; cr.! !

Object subclass: #QCachedFile
	instanceVariableNames: 'fileName binary lastModified checkInterval lastCheckTime contents'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Objects'!

!QCachedFile commentStamp: '<historical>' prior: 0!
QCachedFile is a simple cache which verifies whether its contents on disk has changed when clients request its contents.!

!QCachedFile class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QCachedFile class methodsFor: 'instance creation' stamp: 'ar 8/2/2008 21:09'!
fullName: fullName
	^self new fullName: fullName! !

!QCachedFile class methodsFor: 'instance creation' stamp: 'ar 3/18/2008 18:18'!
on: fileName
	^self new fileName: fileName! !

!QCachedFile methodsFor: 'accessing' stamp: 'ar 3/18/2008 18:07'!
ascii
	"Switch the contents of the file to be represented as ByteString (the default)"
	binary := false.
	contents ifNotNil:[contents := contents asString].
! !

!QCachedFile methodsFor: 'accessing' stamp: 'ar 3/18/2008 18:07'!
binary
	"Switch the contents of the file to be represented as ByteArray"
	binary := true.
	contents ifNotNil:[contents := contents asByteArray].
! !

!QCachedFile methodsFor: 'initialize' stamp: 'ar 3/18/2008 18:27'!
checkEvery: seconds
	"Convenience for users who'd rather specify seconds for the check interval"
	^self checkInterval: seconds * 1000! !

!QCachedFile methodsFor: 'accessing' stamp: 'ar 3/18/2008 18:21'!
checkInterval
	"The interval (in milliseconds) for checking whether the file has changed. 
	If nil, we will check every time when a request is made."
	^checkInterval! !

!QCachedFile methodsFor: 'accessing' stamp: 'ar 3/18/2008 18:26'!
checkInterval: msecs
	"The interval (in milliseconds) for checking whether the file has changed. 
	If nil, we will check every time when a request is made."
	checkInterval := msecs! !

!QCachedFile methodsFor: 'reading' stamp: 'ar 6/8/2009 18:05'!
contents
	"Answer the cached contents of the file. 
	The method will silently return nil if the file does not or no longer exists."
	| delta |
	checkInterval ifNotNil:[
		"Setting the checkInterval to nil forces us to check every time,
		but usually this is the common case so we instantly return 
		the cached result."
		delta := Time millisecondClockValue - lastCheckTime.
		(delta >= 0 and:[delta <= checkInterval]) ifTrue:[^contents].
	].
	^self updateContents! !

!QCachedFile methodsFor: 'accessing' stamp: 'ar 3/18/2008 17:56'!
fileName
	"The full path of the cached file"
	^fileName! !

!QCachedFile methodsFor: 'accessing' stamp: 'ar 4/23/2008 20:12'!
fileName: aString
	"The full path of the cached file"
	fileName := FileDirectory default pathFromURI: aString.! !

!QCachedFile methodsFor: 'initialize' stamp: 'ar 7/13/2009 20:48'!
forceBackup
	"Force a backup of the underlying file"
	| fd base actual file |
	fd := FileDirectory on: (FileDirectory dirPathFor: fileName).
	base := FileDirectory localNameFor: fileName.
	actual := fd nextNameFor: base extension: 'backup'.
	file := fd forceNewFileNamed: actual.
	[file binary; nextPutAll: contents asByteArray] ensure:[file close].
! !

!QCachedFile methodsFor: 'accessing' stamp: 'ar 8/2/2008 21:09'!
fullName
	"The full path of the cached file"
	^fileName! !

!QCachedFile methodsFor: 'accessing' stamp: 'ar 8/2/2008 21:09'!
fullName: aString
	"The full path of the cached file"
	fileName := aString.! !

!QCachedFile methodsFor: 'initialize' stamp: 'ar 4/24/2008 15:47'!
initialize
	"Initialize the receiver with a 30 secs check interval which should be good enough both for interactive use as well as under heavy load."
	fileName := ''.
	binary := false.
	lastModified := 0.
	lastCheckTime := -9999999.
	checkInterval := 30000.
	contents := nil.
! !

!QCachedFile methodsFor: 'reading' stamp: 'ar 7/13/2009 20:49'!
newContents: data
	"Save the data as the new contents for the cached file.
	Note: Consider using #forceBackup before doing this."
	| file |
	file := FileStream forceNewFileNamed: fileName.
	[file binary; nextPutAll: data asByteArray] ensure:[file close].
	"Ensure we actually do reload the contents from disk upon next access"
	lastModified := 0.
	lastCheckTime := -9999999.
! !

!QCachedFile methodsFor: 'reading' stamp: 'ar 6/8/2009 18:04'!
updateContents
	"Answer the cached contents of the file. 
	The method will silently return nil if the file does not or no longer exists."
	| entry buffer file |
	"Look up the directory entry for the file"
	FileDirectory splitName: fileName to:[:path :localName|
		"For some reason MacOS returns an entry for an empty file name so guard it"
		localName ifNotEmpty:[
			entry := (FileDirectory on: path) directoryEntryForName: localName.
		].
	].

	entry ifNil:[
		"The file doesn't exist, get out of here"
		lastCheckTime := Time millisecondClockValue.
		lastModified := 0.		
		^contents := nil
	].

	"Check whether the file has been modified since the last time we checked"
	entry modificationTime = lastModified ifTrue:[
		"The file hasn't been modified; get outta here"
		lastCheckTime := Time millisecondClockValue.
		^contents
	].

	"Read the contents of the file. I would like to factor this into a separate #readContents
	method but unfortunately AsyncFile doesn't know about #size, #atEnd etc. so we need
	to make use of the entry's file size which means we'd have to pass an argument and
	that makes that method completely pointless from a client-side perspective. So it's
	inline for now until such a time that #readContents could be usefully be called by
	some client code."
	buffer := ByteArray new: entry fileSize.
	[file := AsyncFile new open: fileName forWrite: false.
	file nextInto: buffer] ensure:[file close].
	contents := binary ifTrue:[buffer asByteArray] ifFalse:[buffer asString].

	lastCheckTime := Time millisecondClockValue.
	lastModified := entry modificationTime.

	^contents
! !

Object subclass: #QCapturedFrame
	instanceVariableNames: 'parentID childIndex transform surfaceID stickerIDs originalOID originalTranslation'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Objects-Manipulation'!

!QCapturedFrame commentStamp: '<historical>' prior: 0!
Private to QCapturedFrameList.
An entry describing the captured relationships of a single deleted or pasteboarded TFrame.  
Includes its parenting, its inbound and outbound attachments, etc.!

!QCapturedFrame class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QCapturedFrame class methodsFor: 'islands' stamp: 'bgf 2/3/2010 11:46'!
howToPassAsArgument
	"We use these for data-interchange - fetching lots from the island."
	^ #passByCopy:! !

!QCapturedFrame methodsFor: 'initialize' stamp: 'bgf 2/15/2010 20:20'!
capture: aTFrame forAvatar: anAvatar
	"Capture all the context we'll need in order to undo  a Delete.
	 (All but the persistence of the frame itself, handled by the captured-frame-list."

	| adjustZ origZ |
	originalOID := aTFrame objectID.
	aTFrame parent ifNotNilDo: [ : p | 
		(p isKindOf: QEditBox) ifTrue: [
			parentID := p parent objectID.
			childIndex := p parent indexOfChild: p
		] ifFalse: [
			parentID := p objectID.
			childIndex := p indexOfChild: aTFrame
		].
	] ifNil: [
		parentID := nil.
		childIndex := nil.
	].
	"Attachment relationships."
	surfaceID := aTFrame surfaceAttachment ifNotNil: [ : sa | sa objectID ].
	stickerIDs := (aTFrame stickers collect: [ : s | s objectID ]).
	transform := anAvatar relativeTransform: aTFrame.

	aTFrame carryWhenPasted ifTrue: [
		"The file-card case - or anything smallish that we automatically carry on paste.
	 	 Clamp the pasting distance."
		origZ := transform a34 min: -1.
		adjustZ := origZ max: (-8.0).
		transform translation: (transform translation * ((adjustZ / origZ) abs)).
	].
	originalTranslation := transform translation.
! !

!QCapturedFrame methodsFor: 'accessing' stamp: 'bgf 2/7/2010 01:01'!
childIndex
	"The index the frame appeared-at in its original parent's children list."
	^ childIndex! !

!QCapturedFrame methodsFor: 'accessing' stamp: 'bgf 2/3/2010 11:46'!
originalOID
	^ originalOID! !

!QCapturedFrame methodsFor: 'accessing' stamp: 'bgf 2/4/2010 13:14'!
originalX
	^ originalTranslation x! !

!QCapturedFrame methodsFor: 'accessing' stamp: 'bgf 2/4/2010 13:33'!
originalZ
	^ originalTranslation z! !

!QCapturedFrame methodsFor: 'accessing' stamp: 'bgf 2/7/2010 01:01'!
parentID
	"The objectID of the parent frame from which the captured frame was taken."
	^ parentID
! !

!QCapturedFrame methodsFor: 'accessing' stamp: 'bgf 2/14/2010 20:45'!
rayTransform
	"The ray transform from the capturing avatar to look at where I was cut from"
	^ (Matrix4x4 new)
		at: (transform translation negated)
		up: Vector3 yAxis! !

!QCapturedFrame methodsFor: 'accessing' stamp: 'bgf 2/7/2010 01:02'!
resetStagger
	"Return the destination transform (which is relative to the capturer's avatar
	 to its original values.   (See also staggerBy: forStep:.)"
	transform  
		translation: originalTranslation.! !

!QCapturedFrame methodsFor: 'accessing' stamp: 'bgf 2/5/2010 12:18'!
staggerBy: anXOffset forStep: signedStep
	"Shift the frames by a little bit, 
	 so a subsequent <paste> won't superimpose them."
	
	"The translations are relativeTransforms to the avatar replica; 	
	 so shifting X moves them side to side in the paster's view."
	transform  
		translationX: (originalTranslation x + (anXOffset ))
				y: (originalTranslation y)
				z: (originalTranslation z * (0.95 - (signedStep abs *0.01))).
 ! !

!QCapturedFrame methodsFor: 'accessing' stamp: 'bgf 2/7/2010 01:03'!
stickerIDs
	"The captured objectID's of other frames in the space that were attached-to this one."
	^ stickerIDs ifNil: [ #() ]! !

!QCapturedFrame methodsFor: 'accessing' stamp: 'bgf 2/7/2010 01:03'!
surfaceID
	"The objectID of whatver tframe this one was attached to; or nil."
	^ surfaceID! !

!QCapturedFrame methodsFor: 'accessing' stamp: 'bgf 2/7/2010 01:05'!
transform
	"The transform of the captured frame - relative to the capturing avatar.
	 We use the avatar-relative transform so that cut, drive-around, paste,
	 brings back a frame in the same place in the user's viewport... so you can see it
	 in the same viewing area as the original capture occupied when it was taken."
	^ transform! !

Object subclass: #QCapturedFrameList
	instanceVariableNames: 'captures viewerTransform xmlZip resourceZip averageX averageZ staggerStep serviceURL'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Objects-Manipulation'!

!QCapturedFrameList commentStamp: '<historical>' prior: 0!
A list of QCapturedFrame entries describing the recoverable state 
of a Delete, Copy, or Cut of the user's selectedFrames at some point in time.   

Also we record the viewerTransform - the pose of the avatar at the time the frames were captured.!

!QCapturedFrameList class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QCapturedFrameList class methodsFor: 'islands' stamp: 'bgf 2/3/2010 11:47'!
howToPassAsArgument
	"We use these for data-interchange - fetching lots from the island."
	^ #passByCopy:! !

!QCapturedFrameList class methodsFor: 'undo-redo' stamp: 'bgf 3/14/2010 18:05'!
undoForCreation: capList description: verb
	"An undoable-action that undoes creation of the given list of frames"
	^ QBlockAction 
			undoBlock: [ : avatar | avatar future redoDeleteFrames: capList ]
			redoBlock: [  : avatar | avatar future undoDeleteFrames: capList ]
			description:  verb
! !

!QCapturedFrameList class methodsFor: 'undo-redo' stamp: 'bgf 3/14/2010 18:06'!
undoForCut: capList
	"An undoable-action that undoes deletion of the given list of frames"
	^ QBlockAction
			undoBlock: [ : avatar | avatar future undoDeleteFrames: capList ]
			redoBlock: [ : avatar | avatar future redoDeleteFrames: capList ]
			description: 'Cut'
! !

!QCapturedFrameList class methodsFor: 'undo-redo' stamp: 'bgf 3/14/2010 18:06'!
undoForDeletion: capList
	"An undoable-action that undoes deletion of the given list of frames"
	^ QBlockAction
			undoBlock: [ : avatar | avatar future undoDeleteFrames: capList ]
			redoBlock: [ : avatar | avatar future redoDeleteFrames: capList ]
			description: 'Delete'
! !

!QCapturedFrameList methodsFor: 'initialize-release' stamp: 'ar 4/16/2011 17:40'!
capture: listOfFrames forAvatar: anAvatar 
	"Given a list of live in-scene TFrames... 
	 record everything we will need to undo the deletion of these frames
	 (which is superset of what we'll need to simply paste a copy of them into a space."
	| frames writer |
	viewerTransform := anAvatar localTransform clone.
	serviceURL := anAvatar root serviceURL clone.

	"Don't  capture children separately from their parents, since the parent will capture them."
	frames := listOfFrames asArray
				reject: [:f | listOfFrames
							anySatisfy: [:f1 | f hasAncestor: f1]].
	"Sort into forward order by parent-index, so that when pasting,
	we paste earlier sibs before later ones, so the later one's indices work"
	frames := frames sort: [:f1 :f2 | (f1 indexInParent ifNil: [0]) < (f2 indexInParent ifNil: [0])].
	captures := (frames asArray
				sort: [:f1 :f2 | (f1 indexInParent ifNil: [0]) < (f2 indexInParent ifNil: [0])])
				collect: [:f | QCapturedFrame new capture: f forAvatar: anAvatar;
						 yourself].

	averageX := (captures collect: [ : cap | cap originalX ]) average.
	averageZ := (captures collect: [ : cap | cap originalZ ]) average.
	staggerStep := 0.

	"Do one persist of the whole collection,
	 so recovery perserves ad-hoc interframe references efficiently."
	writer := (TSceneWriter new).
	writer initResourceList.
	writer preserveObjectIDs: true.
	xmlZip := writer storeSceneAsZip: (QCollection withAll: frames).

	"The resource manifest will be used when doing a 'paste' into a different forum."
	resourceZip := (TSceneWriter new) 
					storeSceneAsZip: (
						QCollection withAll: (
							(writer resourceList select: [ :r | r isBundled not ]) asSet asArray ) ).! !

!QCapturedFrameList methodsFor: 'drag-drop' stamp: 'bgf 2/14/2010 22:03'!
doPasteAsObjectDrop: frames forAvatar: anAvatar
	"If we've pasted a droppable object into a selection that can take it as a drop,
	 then carry out that drop.   This makes pasting equivalent to dropping for side-effects."

	| dt event |
	event := anAvatar lastSelectionEvent ifNil: [  ^ false ].
	event target ifNil: [ ^ false ].
	anAvatar singleSelectedFrame
		ifNil: [ ^ false ]
		ifNotNil: [ : sf | (sf isOrContains: (event target)) ifFalse: [ ^ false ]].

	"OK, the user's matches what was remembered from interactive clicking.
	 The possible logical drop target is the most-specific of 
		- whatever was clicked,
	 	- whatever logical component was selected by clicking that."

	dt :=  (event target isDropTarget: frames anyOne) 
			ifTrue: [ event target ]
			ifFalse: [ 
				(anAvatar singleSelectedFrame isDropTarget: frames anyOne) 
					ifTrue: [ anAvatar singleSelectedFrame ]
					ifFalse: [ nil ]].
	dt ifNil: [ ^ false ].
	(frames allSatisfy: [ : f | dt isDropTarget: f ]) ifFalse: [ ^ false ].

	"Drop is approved. So do it. 
	  Below, if the target consumes the drop (means takes the original into itself),
	  then we're done.  Otherwise, delete the incoming frame, since the drop-receiver has a copy."
	frames do: [ : f |
		 (dt consumeDrop: f avatar: anAvatar selection: event selection)
			ifFalse: [ 
				(dt isOrContains: f) 
					ifFalse: [ f future obliterateLater ] ].
	].
	^ true

! !

!QCapturedFrameList methodsFor: 'drag-drop' stamp: 'bgf 2/15/2010 20:04'!
finishPaste: frames asDropFor: anAvatar withStagger: signedStagger
	"Completion of frame paste.
	  We may do a logical content drop (e.g. into a display panel) if things match..."
	(self doPasteAsObjectDrop: frames forAvatar: anAvatar) ifTrue: [ ^ true ].

	"Otherwise we have the new frames themselves to finish placing."
	anAvatar selectFrames: frames.

	"The items may have spatial drag-adapters to manage arrival placement.
	 This can set up attachments, surface-contact, etc."
	(frames allSatisfy: [ : f | f parent = anAvatar root and: [ f dragAdapter notNil ]]) 
	 ifFalse: [ ^ false ].

	(frames anySatisfy: [ : f | f carryWhenPasted ]) ifTrue: [
		"Mostly for file cards. If it didn't go onto anything, then carry it. 
		 We don't want the auto-drop behavior below pasting cards, it is too unexpected,	
	 	 since cards mostly always turn into something else when dropped."
		frames do: [ : f | f carryWhenPasted ifTrue: [ anAvatar startCarrying: f ] ].
		^ true
	].

	frames with: captures do:  [ : f : capture |
		| event box depth pointer vector placed |
		"See also class-side QDragAdapter dragDrop:intoSpace:forEvent:"
		event := anAvatar fakePointerOverEvent: (f isGroundBound)
						 rayTransform: (
							capture rayTransform 
								addRotationAroundY: (signedStagger*2.3)).
									"signedStagger * 2.3 is the spacing angle in degrees."
		box := f positioningBox.
		depth := box extentZ.
		placed := false.
		"Check for contact with something."
		pointer := event makePointer.
		pointer pointerPickTree: (anAvatar root)
				excluding: frames, { anAvatar . anAvatar costume }.

		"Only do the drop if there is an obstacle to drop it in front of.
		 Sets up surface-attachments and drop-ins, and prevents being behind stuff.
		 (If no contact, f already has a placement avatar-relative to where it was copied from.)"
		pointer selection frame ifNotNil: [
			depth := ((pointer selection distance * 0.95) min: (box radius * 3) ) 
						max: (box extentZ * 0.55).
			vector :=  (anAvatar globalTransform * (event selection rayTransform)) 
						lookAt negated.

			f translation: (anAvatar translation) + (vector * depth).
			f allUpdateShape.
			placed := (f dragAdapter fakeDragForCreation: event).
		].
	].
	^ true

! !

!QCapturedFrameList methodsFor: 'accessing' stamp: 'bgf 2/24/2010 16:36'!
includesOID: oid
	^ captures anySatisfy: [ : c | c originalOID = oid ]! !

!QCapturedFrameList methodsFor: 'initialize-release' stamp: 'bgf 3/11/2010 16:33'!
pasteForAvatar: anAvatar intoParent: someFrame withDocumentRenames: pairwise

	"Paste the captured frames into the current space, parented in the given frame."
	| frames space renames sideStagger |

	space := anAvatar root ifNil: [ ^ #() ].

	"Part one - unpickle the raw captured frame list"
	frames := ((TSceneReader forSession:( anAvatar sessionID)) readSceneFromZip: xmlZip) items.

	"Fix up the shared names."
	renames := Dictionary newFromPairs: pairwise.
	frames do: [ : f |
		f allDocuments do: [ : doc |
			renames at: (doc sharedName) ifPresent: [ : newName | doc sharedName: newName ]
		]].

	"Restore transforms."
	frames with: captures do: [ : frame : capture |
		frame localTransform: ((anAvatar localTransform) * (capture transform)).
		].

	"Clear some annoying persisted properties..."
	frames do: [ :f | 
		f isPlacementLocked: false.
		anAvatar isEditingTemplate ifFalse: [ f markAsNotFromTemplate ].
		].

	"Restore parenting"
	frames with: captures do: [ : frame : capture |
		capture parentID ifNotNilDo: [ : pid |
			| idx | 
			idx := captures findFirst: [ : qfr | qfr originalOID = pid ].
			idx > 0 ifTrue: [
				(frames at: (idx)) addChild: frame atIndex: (capture childIndex).
			] ifFalse: [
				someFrame addPastedChild: frame
			]]].

	"Restore attachments"
	self restoreInternalAttachments: frames inSpace: space.

	"Followup mechanics"
	sideStagger := staggerStep odd 
						ifTrue: [ staggerStep-1 // 2]
						ifFalse: [ (staggerStep // 2) negated ].

	self finishPaste: frames asDropFor: anAvatar withStagger: sideStagger.

	"Capturing is futured to allow all the attachment relationships to be handled first."
	anAvatar future captureCreateFrames: frames description: 'Paste'.

	"Tell me all about it."
	frames size = 1 ifTrue: [
		anAvatar signal: #infoMessage with: 'Pasted one ', (frames first genericName)
	] ifFalse: [
		anAvatar signal: #infoMessage with: 'Pasted ', (frames size), ' items.'
	].! !

!QCapturedFrameList methodsFor: 'accessing' stamp: 'bgf 2/5/2010 10:13'!
resetStagger
	"Restore the captures to the unstaggered capture position."
	staggerStep := 0.
	captures do: [ : cap | cap resetStagger ].! !

!QCapturedFrameList methodsFor: 'accessing' stamp: 'bgf 2/7/2010 00:57'!
resourceList
	"Returns an array of QAbstractDocuments, TFileTextures, and TFrameThumbs,
	 all of which may represent server-side resource or document files
	 that will need to be copied if we reconstruct the capture in a forum other than its origin."
	resourceZip ifNotNil: [
		^ ((TSceneReader new) readSceneFromZip: (resourceZip)) items
	].
	^ #() ! !

!QCapturedFrameList methodsFor: 'initialize-release' stamp: 'bgf 3/15/2010 09:00'!
restoreForAvatar: anAvatar
	"Restore the capture as if it had been deleted and we want it exactly back.
	 This should only be done once, as it recovers the deleted thing's objectID's."

	| frames space reader |

	self resetStagger.
	space := anAvatar root ifNil: [ ^ #() ].

	"Unpickle the raw captured frame list - with restoration of object IDs "
	reader := TSceneReader forSession: anAvatar sessionID.
	reader preserveObjectIDs: true.
	reader preserveDocumentNames: true.
	frames := (reader readSceneFromZip: xmlZip) items.

	"Pass two - restore parenting"
	frames with: captures do: [ : frame : capture |
		capture parentID ifNotNilDo: [ : pid |
			(space island lookup: pid ifAbsent:[nil]) ifNotNilDo: [ : parent |
				(parent isKindOf: TFrame) ifTrue: [
					parent children size >= (capture childIndex-1) ifTrue: [
						parent addChild: frame atIndex: (capture childIndex) 
					] ifFalse: [
						parent addChild: frame
					].
				]]].
		frame parent ifNil: [
			QLogger logWarning: 'No parent to receive ', frame.
		]].

	"Restore attachments.  The rooted checks below help ensure that we're not recovering attachments
	 that were the leftover island-entries from a previous deletion."
	frames with: captures do: [ : frame : capture |
		capture surfaceID ifNotNilDo: [ : sid |
			(space island lookup: sid ifAbsent:[nil]) ifNotNilDo: [ : sfc |
				((sfc isKindOf: TFrame) and: [ sfc root notNil]) ifTrue: [
					space attach: frame to: sfc
				]]]].
	frames with: captures do: [ : frame : capture |
		capture stickerIDs do: [ : stickerID |
			(space island lookup: stickerID ifAbsent:[nil]) ifNotNilDo: [ : sticker |			
				((sticker isKindOf: TFrame) and: [ sticker root notNil]) ifTrue: [
					space attach: sticker to: frame
				]]]].

	"Need to know what to do about feedback."
	^ frames! !

!QCapturedFrameList methodsFor: 'private' stamp: 'bgf 2/3/2010 16:16'!
restoreInternalAttachments: frames inSpace: space 
	"Restore attachments from within the capture collection
	 (where both the sticker and surface are part of the capture)"

	frames with: captures do: [ : frame : capture |
		capture surfaceID ifNotNilDo: [ : sid |
			| idx | 
			idx := captures findFirst: [ : qfr | qfr originalOID = sid ].
			idx > 0 ifTrue: [
				space attach: frame to: (frames at: idx).
			]]].

	frames with: captures do: [ : frame : capture |
		capture stickerIDs do: [ : stickerID |
			| idx |
			idx := captures findFirst: [ : qfr | qfr originalOID = stickerID ].
			idx > 0 ifTrue: [
				space attach: (frames at: idx) to: frame
			]]].
! !

!QCapturedFrameList methodsFor: 'accessing' stamp: 'ar 4/16/2011 17:41'!
serviceURL
	"The Service URL identifying the forum from which this capture was taken."
	^ serviceURL! !

!QCapturedFrameList methodsFor: 'accessing' stamp: 'bgf 2/7/2010 00:58'!
staggerPositions
	"Shift the target paste-transform of each of the captured frames by a little bit, 
	 so that if multiple pastes are done from one source,
	 the resulting frames won't exactly superimpose in the space."
	| signed |
	staggerStep := staggerStep + 1.
	signed := staggerStep odd 
				ifTrue: [ staggerStep-1 // 2]
				ifFalse: [ (staggerStep // 2) negated ].
	captures do: [ : capture | 
		capture staggerBy: (averageX negated) + (signed * (averageZ * 0.05)) forStep: signed ]! !

Object subclass: #QChatMessage
	instanceVariableNames: 'message timestamp user place localDateCache'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Objects-Space'!

!QChatMessage commentStamp: '<historical>' prior: 0!
A single text-chat message (in a chat log.)
Cross-references a QChatUser with one text message and its time.
The timestamp is utcSeconds, as supplied by the submitting client at submission time.
!

!QChatMessage class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QChatMessage class methodsFor: 'island-args' stamp: 'bgf 8/27/2008 14:10'!
howToPassAsArgument
	"This allows us to retrieve and process the chat history with a local #get:
	 and not deal with large arrays of far refs.   Text might be bulky
	 but you can only type so fast..."
	^ #passByCopy:! !

!QChatMessage methodsFor: 'comparing' stamp: 'bgf 8/27/2008 12:08'!
= other
	^ (other class == self class)
		and: [ other timestamp = timestamp ]
		and: [ other user = user ]
		and: [ other message = message ]! !

!QChatMessage methodsFor: 'persist' stamp: 'hrs 2/8/2009 12:08'!
constructFrom: pStream

	"We deliberately use short tags - the elements are very repetitive."

	self user: (pStream readObject: 'user' default: nil).
	self message: (pStream readString: 'msg' default: '') base64Decoded.
	self timestamp: (pStream readNumber: 't' default: 0).
	self place: (pStream readString: 'loc' default: nil).
! !

!QChatMessage methodsFor: 'accessing' stamp: 'bgf 5/11/2010 12:20'!
fullMessage
	"The concatenated   message + place-string (if any) ...  the message body for chat-list display"

	message ifNil: [ ^ nil ].		"Not yet ready... marker for a message while being typed."

	place ifNotNil: [
		^ message, ' ', place 
	] ifNil: [
		^ message
	].! !

!QChatMessage methodsFor: 'accessing' stamp: 'das 8/5/2009 13:04'!
getChatMessage

	"return an XML form of the message. See also QChatLog>>getLog"
	^String tab, '<chatMessage>', String cr,
		String tab, String tab, '<user>', user asString, '</user>', String cr,
		String tab, String tab, '<message>' , message asString, '</message>', String cr,
		String tab, String tab, '<timestamp>', timestamp asString, '</timestamp>', String cr,
		String tab, String tab, '<place>', place asString, '</place>', String cr,
	String tab, '</chatMessage>'
! !

!QChatMessage methodsFor: 'comparing' stamp: 'bgf 8/27/2008 10:28'!
hash
	"To match ="
	user hash bitXor: (timestamp hash)! !

!QChatMessage methodsFor: 'private' stamp: 'bgf 3/8/2009 19:44'!
localDateLabel
	"For harness-side use only, NEVER via get or on-island !! 
	 This is a caching accessor for the chat display."
	^ localDateCache ifNil: [
		 ( localDateCache := (QChatLog localDateString: self) )
		].! !

!QChatMessage methodsFor: 'accessing' stamp: 'bgf 3/31/2007 22:39'!
message
	^ message! !

!QChatMessage methodsFor: 'accessing' stamp: 'bgf 3/31/2007 22:40'!
message: aString
	message := aString! !

!QChatMessage methodsFor: 'persist' stamp: 'hrs 2/8/2009 12:09'!
persistOn: pStream

	"We deliberately use short tags - the elements are very repetitive."
	pStream storeObject: 'user' value: user default: nil.
	pStream storeString: 'msg' value: (message ifNotNil:[message base64Encoded]) default: nil.
	pStream storeNumber: 't' value: timestamp default: 0.
	pStream storeString: 'loc' value: place default: nil.! !

!QChatMessage methodsFor: 'accessing' stamp: 'hrs 2/8/2009 12:07'!
place
	"A serializable, externally transmittable designation that can be used to bring a user back to the place of interest."
	^place! !

!QChatMessage methodsFor: 'accessing' stamp: 'hrs 2/8/2009 12:08'!
place: aString
	place := aString! !

!QChatMessage methodsFor: 'accessing' stamp: 'bgf 10/23/2007 11:53'!
timestamp
	"Time of the message, in UTC seconds, as originally determined by the issuing client."
	^ timestamp! !

!QChatMessage methodsFor: 'accessing' stamp: 'bgf 10/23/2007 11:53'!
timestamp: t

	"Time of the message, in UTC seconds, as originally determined by the issuing client."

	timestamp := t! !

!QChatMessage methodsFor: 'accessing' stamp: 'bgf 3/31/2007 23:33'!
user
	"The QChatUser record referring to this message's sender."
	^ user! !

!QChatMessage methodsFor: 'accessing' stamp: 'bgf 4/1/2007 01:10'!
user: aQChatUser
	"Inside the space, this is a QChatUser record."
	user := aQChatUser! !

Object subclass: #QChatUser
	instanceVariableNames: 'userName userLogin'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Objects-Space'!

!QChatUser commentStamp: '<historical>' prior: 0!
A user who has sent a chat message.
Used to reduce the overall object (and XML) volume of the chat log;
chat log entries hold references to these user records,
which in turn hold strings:
	userLogin is the login-string of the originating user
	userName is the screenName of the user at the time the message was emitted.
!

!QChatUser class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QChatUser class methodsFor: 'islands' stamp: 'bgf 8/27/2008 14:11'!
howToPassAsArgument
	"This allows us to retrieve and process the chat history with a local #get:
	 and not deal with large arrays of far refs.   Text might be bulky
	 but you can only type so fast..."
	^ #passByCopy:! !

!QChatUser methodsFor: 'comparing' stamp: 'bgf 8/28/2008 15:02'!
= other
	^ (other class == self class)
	 	and: [ other userLogin = userLogin ]! !

!QChatUser methodsFor: 'persist' stamp: 'bgf 8/28/2008 15:08'!
constructFrom: pStream
	"User ID is historical... no point changing the XML ..."
	self userLogin: (pStream readString: 'userID' default: userLogin).

	userLogin = (QChatLog oldSysUserID)
		ifTrue: [ userLogin := QChatLog sysUserID ].

	self userName: (pStream readString: 'userName' default: userName).
! !

!QChatUser methodsFor: 'comparing' stamp: 'bgf 8/28/2008 15:01'!
hash
	"Defined because = is defined"
	^ userLogin hash! !

!QChatUser methodsFor: 'accessing' stamp: 'bgf 8/28/2008 15:08'!
isSysUser
	^ self userLogin = QChatLog sysUserID! !

!QChatUser methodsFor: 'persist' stamp: 'bgf 8/28/2008 15:01'!
persistOn: pStream

	pStream storeString: 'userID' value: userLogin default: nil.
	pStream storeString: 'userName' value: userName default: nil.
! !

!QChatUser methodsFor: 'accessing' stamp: 'bgf 8/28/2008 15:02'!
userLogin
	^ userLogin! !

!QChatUser methodsFor: 'accessing' stamp: 'bgf 8/28/2008 18:17'!
userLogin: aString
	"This is the prime identity for a chat message's author - the forum login"
	userLogin := aString! !

!QChatUser methodsFor: 'accessing' stamp: 'bgf 3/31/2007 23:33'!
userName
	^ userName! !

!QChatUser methodsFor: 'accessing' stamp: 'bgf 3/31/2007 23:33'!
userName: aString
	userName := aString! !

Object subclass: #QColors
	instanceVariableNames: ''
	classVariableNames: 'Colors'
	poolDictionaries: ''
	category: 'Qwaq-Objects-Resources'!

!QColors commentStamp: '<historical>' prior: 0!
Colors for use in Qwaq spaces and widgets.
All my methods are class methods.
Each method returns a new instance VectorColor;
there is no need to 'copy' the results.

This class is stateless; there is no need to instantiate it.
In future, it may be expanded to be a per-island singleton,
to allow application-specific restyling.    The class methods
will continue to be the public accessors in that case.!

!QColors class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QColors class methodsFor: 'island-colors' stamp: 'bgf 11/27/2009 17:27'!
backdropColor
	"Color for empty display backgrounds (e.g. whiteboards, etc.)"
	^ self color: [  VectorColor r:1 g:1 b:1 ]
! !

!QColors class methodsFor: 'island-colors' stamp: 'bgf 11/27/2009 18:51'!
blackAlphaColor
	"Black basis for alpha-pass materials "
	^ self color: [  VectorColor r: 0 g: 0 b: 0 a: 0.999 ]! !

!QColors class methodsFor: 'island-colors' stamp: 'bgf 11/27/2009 21:15'!
blackColor
	^ self color: [ VectorColor r: 0 g:0 b:0  ]! !

!QColors class methodsFor: 'island' stamp: 'bgf 11/27/2009 20:27'!
color: makerBlock
	"Private!!  This lists the result of the given maker-block in the table 
	  under the name of the method that called this one.
	  See the VectorColor constant methods like 'gray5' .
	  The point is to make one instance of a given color per Island. "

	^ (self colorTableForThisIsland) 
		at: (thisContext sender method selector)  
		ifAbsentPut: makerBlock! !

!QColors class methodsFor: 'island' stamp: 'bgf 11/27/2009 20:26'!
colorTableForThisIsland
	^ Processor activeIsland at: #QColors ifAbsentPut: [ IdentityDictionary new ]! !

!QColors class methodsFor: 'island-colors' stamp: 'bgf 11/27/2009 17:27'!
confirmDeleteColor
	"To hilite: objects when menus ask whether really to delete them."
	^ self color: [ VectorColor r: 0.55 g: 0.52 b: 0.35 ]! !

!QColors class methodsFor: 'island-colors' stamp: 'bgf 11/27/2009 17:27'!
controlDisabledColor
	"Cool gray for controls that are disabled."
	^ self color: [  VectorColor r:0.35 g:0.35 b:0.4 ]! !

!QColors class methodsFor: 'island-colors' stamp: 'bgf 11/27/2009 17:28'!
controlDownColor

	"MouseDown hilite color for controls colored with 'controlNormalColor'"
	^ self color: [  VectorColor r:0.15 g:0.85 b:0.75 a: 0.4 ]
	! !

!QColors class methodsFor: 'island-colors' stamp: 'bgf 11/27/2009 17:28'!
controlGoColor
	"Bright warm green color to indicate activity or 'on' state of a control."
	^  self color: [ VectorColor r:0.2 g:0.85 b:0.1 ]! !

!QColors class methodsFor: 'island-colors' stamp: 'bgf 11/27/2009 17:28'!
controlHiliteColor

	"Used to draw attention to controls when actions have been enabled."
	^ self color: [  VectorColor r:0.6 g:0.6 b:0.3 a: 0.35 ]! !

!QColors class methodsFor: 'island-colors' stamp: 'bgf 11/27/2009 17:28'!
controlNormalColor
	
	"Color for simple elements (cubes, balls, etc) that serve as buttons or drag handles."
	^ self color: [  VectorColor r:0.1 g:0.36 b:0.65 ]! !

!QColors class methodsFor: 'island-colors' stamp: 'bgf 11/27/2009 17:28'!
controlOffColor
	"For a start-stop button when the subject is stopped
	 (so the action of the control is to start it).
	 Light dull greenish color.

	 (We use white for running, dull for stopped,
	 and the start-stop button uses arrow-vs-pausebars to clarify.)"

	^ self color: [  VectorColor r:0.6 g:0.75 b:0.65 ]! !

!QColors class methodsFor: 'island-colors' stamp: 'bgf 11/27/2009 17:28'!
controlOnColor
	"Bright warm yellow-green color to indicate activity or 'on' state of a control."
	^ self color: [  VectorColor r:0.3 g:0.9 b:0.25 ]! !

!QColors class methodsFor: 'island-colors' stamp: 'bgf 11/27/2009 17:28'!
controlOverColor

	"Mouseover hilite color for controls colored with 'controlNormalColor'"
	^ self color: [  VectorColor r: 0.55 g:0.85 b:0.75 a: 0.2 ]! !

!QColors class methodsFor: 'island-colors' stamp: 'bgf 11/27/2009 17:28'!
controlRunningColor
	"Bright sliver to indicate running state in a start-stop button.
	 (We use this for running, dull cool green for stopped,
	 and the start-stop button should use, for instance, arrow-vs-pausebar shapes to reinforce.)"
	^ self color: [  VectorColor r:0.9 g: 0.9 b: 0.95 ]! !

!QColors class methodsFor: 'island-colors' stamp: 'bgf 11/27/2009 17:53'!
defaultFogColor
	"The default base color for fog calculation."

	^ self color: [ VectorColor r: 0.85 g:0.85 b:0.88 ]! !

!QColors class methodsFor: 'island-colors' stamp: 'bgf 3/25/2010 09:49'!
dirtyAppOutlineColor
	^ self color: [ VectorColor r: 0.25 g: 0.35 b: 0.4 ]! !

!QColors class methodsFor: 'island-colors' stamp: 'bgf 11/27/2009 18:14'!
displaySignColor
	^ self color: [ VectorColor r: 0.51 g: 0.47 b: 0.42  ]! !

!QColors class methodsFor: 'island-colors' stamp: 'bgf 11/27/2009 17:28'!
dropTargetHiliteColor

	"Color used to indicate the current drop target during drag/drop interactions;
	this just modulates a bit, and blues a bit... some drop targets are large like walls
	and shouldn't have dramatic effects."

	^ self color: [  VectorColor r:0.3 g:0.45 b:0.65 a: 0.1 ]! !

!QColors class methodsFor: 'island-colors' stamp: 'bgf 11/27/2009 17:50'!
editboxCornerColor
	^ self color: [ VectorColor  r:0.1 g:0.43 b:0.36 a:0.35 ]! !

!QColors class methodsFor: 'island-colors' stamp: 'bgf 4/2/2010 17:44'!
editboxCrosshairColor
	^ self color: [ VectorColor r:0.45 g:0.7 b:0.5 a:0.95.]! !

!QColors class methodsFor: 'island-colors' stamp: 'bgf 11/27/2009 17:51'!
editboxEdgeColor
	^ self color: [ VectorColor r:0.1 g:0.45 b:0.36 a:0.3 ]! !

!QColors class methodsFor: 'island-colors' stamp: 'bgf 3/26/2010 17:23'!
editboxFaceColor
	^ self color: [ VectorColor r:0.05 g:0.55 b:0.4 a:0.15.]! !

!QColors class methodsFor: 'island-colors' stamp: 'bgf 11/27/2009 18:52'!
editboxGridColor
	^ self color: [ VectorColor r: 0.99 g: 0.96 b: 0.75 a: 0.21 ]! !

!QColors class methodsFor: 'island-colors' stamp: 'bgf 11/27/2009 17:52'!
editboxHiddenWireColor
	^ self color: [ VectorColor r: 0.25 g: 0.5 b:0.4 a: 0.45 ]! !

!QColors class methodsFor: 'island-colors' stamp: 'bgf 11/27/2009 18:53'!
editboxIndicatorColor
	^ self color: [ VectorColor r: 0.99 g:0.75 b:0.1 a:0.75 ]! !

!QColors class methodsFor: 'island-colors' stamp: 'bgf 11/27/2009 20:32'!
editboxShadowColor
	^ self color: [ VectorColor r:0.3 g: 1 b: 0.8 a: 0.175 ]! !

!QColors class methodsFor: 'island-colors' stamp: 'bgf 11/27/2009 17:52'!
editboxShadowDownColor
	^ self color: [ VectorColor r:0.6 g: 0.3 b: 0.05 a:0.8 ]! !

!QColors class methodsFor: 'island-colors' stamp: 'bgf 11/27/2009 17:53'!
editboxShadowOverColor
	^ self color: [ VectorColor r:0.9 g: 0.7 b: 0.2 a:0.75  ]! !

!QColors class methodsFor: 'island-colors' stamp: 'bgf 11/27/2009 17:53'!
editboxWireColor
	^ self color: [ VectorColor r: 0.1 g: 0.5 b:0.4 a:0.75  ]! !

!QColors class methodsFor: 'island-colors' stamp: 'bgf 11/27/2009 17:29'!
fixedDisplayColor
	"Blue like the blue of the fixed-display-material, and quite dark."
	^ self color: [  Color r: 0 g: 0.18 b: 0.27 ]! !

!QColors class methodsFor: 'island-colors' stamp: 'bgf 11/27/2009 17:29'!
generalDragColor

	"Hilite: for objects being dragged."
	^ self color: [  VectorColor r: 0.55 g: 0.65 b: 0.6 a: 0.3 ]! !

!QColors class methodsFor: 'island-colors' stamp: 'bgf 11/27/2009 17:29'!
generalOverColor

	"For mouseover feedback of draggable compound objects."
	^ self color: [  VectorColor r:0.85 g:0.85 b:0.75 a: 0.125 ]! !

!QColors class methodsFor: 'island-colors' stamp: 'bgf 11/27/2009 18:15'!
gray05
	^ self color: [ VectorColor r: 0.05 g:0.05 b:0.05 ]! !

!QColors class methodsFor: 'island-colors' stamp: 'bgf 11/27/2009 17:55'!
gray2
	^ self color: [ VectorColor r: 0.2 g:0.2 b:0.2 ]! !

!QColors class methodsFor: 'island-colors' stamp: 'bgf 11/27/2009 18:16'!
gray25
	^ self color: [ VectorColor r: 0.25 g:0.25 b:0.25 ]! !

!QColors class methodsFor: 'island-colors' stamp: 'bgf 11/27/2009 17:55'!
gray3
	^ self color: [ VectorColor r: 0.3 g:0.3 b:0.3 ]! !

!QColors class methodsFor: 'island-colors' stamp: 'bgf 11/27/2009 20:37'!
gray4
	^ self color: [ VectorColor r: 0.4 g:0.4 b:0.4 ]! !

!QColors class methodsFor: 'island-colors' stamp: 'bgf 11/27/2009 17:56'!
gray5
	^ self color: [ VectorColor r: 0.5 g:0.5 b:0.5 ]! !

!QColors class methodsFor: 'island-colors' stamp: 'bgf 11/27/2009 18:16'!
gray65
	^ self color: [ VectorColor r: 0.65 g:0.65 b:0.65 ]! !

!QColors class methodsFor: 'island-colors' stamp: 'bgf 11/27/2009 17:56'!
gray6Alpha
	^ self color: [ VectorColor r: 0.6 g:0.6 b:0.6 a: 0.999 ]! !

!QColors class methodsFor: 'island-colors' stamp: 'bgf 11/27/2009 18:18'!
gray9
	^ self color: [ VectorColor r: 0.9 g:0.9 b:0.9 ]! !

!QColors class methodsFor: 'island-colors' stamp: 'bgf 11/27/2009 18:24'!
gray99
	^ self color: [ VectorColor r: 0.99 g:0.99 b:0.99 ]! !

!QColors class methodsFor: 'island-colors' stamp: 'bgf 11/27/2009 18:12'!
groupOfNotesColor
	^ self color: [VectorColor r: 1.0 g: 0.99 b: 0.9 ]! !

!QColors class methodsFor: 'class initialization' stamp: 'bgf 9/6/2009 16:47'!
initialize
	" QColors initialize. "
	Colors := IdentityDictionary new.! !

!QColors class methodsFor: 'ui-colors' stamp: 'bgf 2/24/2008 23:44'!
innerPaneColor
	"The basic content color for floating panes."
	^ self paneColor lighter slightlyLighter! !

!QColors class methodsFor: 'list-coding' stamp: 'bgf 3/8/2009 21:01'!
listItemActiveColor
	"For list entries and menu labels colored to indicate that the item is active,
 	 e.g. the services in the forums list that the user has active right now."
	^ Colors at: #listItemActiveColor ifAbsentPut: [ Color r: 0.04 g: 0.25 b:0.06 ]! !

!QColors class methodsFor: 'list-coding' stamp: 'bgf 3/8/2009 21:02'!
listItemDefaultColor
	"For un-flagged list items in lists that use the other listItemXXXColors for flagged entries."
	^ Colors at: #listItemDefaultColor ifAbsentPut: [  Color gray: 0.12 ]! !

!QColors class methodsFor: 'list-coding' stamp: 'bgf 3/8/2009 21:02'!
listItemFavoriteColor
	"For list entries and menu labels colored to indicate that the item is a favorite,"
	^ Colors at: #listItemFavoriteColor ifAbsentPut: [  Color r: 0.1 g: 0.03 b: 0.24 ]! !

!QColors class methodsFor: 'list-coding' stamp: 'bgf 3/8/2009 21:02'!
listItemRecentColor
	"For list entries and menu labels colored to indicate that the item has recently been visited."
	^ Colors at: #listItemRecentColor ifAbsentPut: [  Color r: 0.25 g: 0.09 b: 0.06 ]! !

!QColors class methodsFor: 'list-coding' stamp: 'bgf 9/6/2009 16:46'!
listItemTemplateColor
	"For list entries and menu labels colored to indicate that the item is a template
	 for creating other things, rather than a usable entry in its own right."
	^ Colors at: #listItemTemplateColor ifAbsentPut: [  QColors qwaqGray darker ]! !

!QColors class methodsFor: 'ui-colors' stamp: 'bgf 8/16/2008 22:36'!
mainBorderColor
	"The border color for main pieces of the participant."
	^ Color r:0.55 g:0.548 b:0.54 
! !

!QColors class methodsFor: 'ui-colors' stamp: 'bgf 7/12/2009 14:07'!
menuBackgroundColor

	"Background for in-space TMenus"
	^ Color r: 0.98 g: 0.958 b: 0.95 alpha: 0.935
! !

!QColors class methodsFor: 'ui-colors' stamp: 'bgf 7/11/2009 02:01'!
menuBorderColor	
	"Outline for TMenus."
 	^ Color r: 0.45 g: 0.55 b: 0.7 alpha: 0.6! !

!QColors class methodsFor: 'island-colors' stamp: 'bgf 11/27/2009 18:19'!
menuButtonColor
	^ self color: [ VectorColor r: 0.8 g:0.81 b: 0.82 ]! !

!QColors class methodsFor: 'ui-colors' stamp: 'bgf 7/11/2009 01:50'!
menuSelectedColor	
	"Background for the selected item in TMenus."
 	^ CWidgetLibrary default selectionColor! !

!QColors class methodsFor: 'ui-colors' stamp: 'bgf 1/3/2009 19:32'!
paneBorderColor
	"The border color for floating panes."
	^ Color r:0.45 g:0.445 b:0.43 alpha: 0.45! !

!QColors class methodsFor: 'ui-colors' stamp: 'bgf 9/12/2009 13:09'!
paneColor
	"The basic color for tools and panels."
	^ (Color r: 0.917 g: 0.912 b: 0.908) ! !

!QColors class methodsFor: 'ui-colors' stamp: 'bgf 2/6/2009 01:41'!
paneDarkerColor
	"The color for dark-fill pane areas."
	^ self paneColor adjustSaturation: -0.005 brightness: -0.08! !

!QColors class methodsFor: 'island-colors' stamp: 'bgf 11/27/2009 17:31'!
paperColor
	"Color for paper-metaphor sheets."
	^ self color: [  VectorColor r:0.95 g:0.94 b:0.91 a:0.97 ]


! !

!QColors class methodsFor: 'policies' stamp: 'bgf 8/28/2008 00:15'!
paperColorForAvatarColor: anAvatarColor

	"Given an avatar color (usually from the standard auto-assigned choices,
	 or the personal preference choices,
	 manufacture a variation that is light enough to use as the background for black text,
	 but still have the same hue as the avatar."

	| color bri sat |
	anAvatarColor ifNil: [ ^ Color paleBuff ].
	color := anAvatarColor asColor clone.

	bri := color luminance.
	sat := color saturation.
	color := color adjustSaturation: ( (sat min: 0.65) * 0.7 ) negated
				   brightness: (1.0 - ((bri + sat/3.5) max: 0.98)) * 0.96.

	color luminance < 0.65 ifTrue: [
		color := color adjustBrightness: (1.0 - color luminance)*0.5
	].
	^ color

! !

!QColors class methodsFor: 'island-colors' stamp: 'bgf 11/27/2009 17:30'!
placemarkColor

	"Base color for placemarks and all kinds of goto arrows."
	^ self color: [  VectorColor r: 0.9 g: 0.7 b: 0.05 ]! !

!QColors class methodsFor: 'island-colors' stamp: 'bgf 11/27/2009 17:55'!
pureWhite
	^ self color: [ VectorColor r: 1 g: 1 b: 1 ]! !

!QColors class methodsFor: 'ui-colors' stamp: 'ar 4/16/2011 17:41'!
qwaqGray
	"Official OpenQwaq logo gray."
	^ Color fromHtmlColorRef: '455461'! !

!QColors class methodsFor: 'ui-colors' stamp: 'ar 4/16/2011 17:41'!
qwaqOrange
	"Official OpenQwaq logo orange."
	^ Color fromHtmlColorRef: 'F37033'! !

!QColors class methodsFor: 'island-colors' stamp: 'bgf 11/27/2009 18:09'!
redWeakHilite
	^ self color: [ VectorColor r: 1.0 g: 0.05 b: 0.05 a: 0.2 ]! !

!QColors class methodsFor: 'policies' stamp: 'bgf 7/19/2008 11:14'!
sameColor: v1 as: v2

	"An inexact color comparator that makes room for the inexact way 
	 we store vector colors in XML persistence.
	 Returns true if the two arguments (which must be colors or vector colors)
	 are the same color visually, even if they vary a bit in the low-order bits of values."

	v1 = v2 ifTrue: [ ^ true ].
	v1 ifNil: [ ^ false ].
	v2 ifNil: [ ^ false ].
	^ (v1 asVectorColor - (v2 asVectorColor)) squaredLength < 0.0001.

	"Note about calibration:
		((Color orange asVectorColor) - (Color orange slightlyLighter asVectorColor)) squaredLength
	 Produces  0.000110842575118484
	"
! !

!QColors class methodsFor: 'island-colors' stamp: 'bgf 11/27/2009 17:31'!
selectionHiliteColor
	"For indicated a selected state of a generic object."
	^ self color: [ TFrame selectionHiliteColor ]! !

!QColors class methodsFor: 'island-colors' stamp: 'bgf 11/27/2009 17:30'!
signTextBackground

	"Color for translucent sign background, for dark text."
	^ self color: [  VectorColor r:0.75 g: 0.67 b: 0.54 a: 0.8 ]! !

!QColors class methodsFor: 'island-colors' stamp: 'bgf 11/27/2009 17:31'!
signTextForeground

	"Color for the text in signage around the demo rooms."
	^ self color: [  VectorColor r: 0.15 g: 0.08 b: 0.02 a: 0.99 ]! !

!QColors class methodsFor: 'island-colors' stamp: 'bgf 11/27/2009 18:10'!
stickyNoteColor
	^ self color: [ VectorColor r: 0.99 g: 0.92 b: 0.55 ]! !

!QColors class methodsFor: 'policies' stamp: 'bgf 7/25/2009 03:26'!
textColorForAvatarColor: anAvatarColor

	"Given an avatar color (usually from the standard auto-assigned choices,
	 or the personal preference choices,
	 manufacture a variation that preserves the hue as best we can
	 but is readable as text on off-white backgrounds."

	| color bri sat |
	anAvatarColor ifNil: [ ^ Color gray:0.2 ].
	color := anAvatarColor asColor.

	bri := color luminance.
	sat := color saturation.
	color := color adjustSaturation: ( (sat atLeast: 0.2 atMost: 0.45) * 0.4) negated
				   brightness: (((bri * 0.3) + ((sat min: 0.7)*0.38)) min: 0.6 max: 0.02) negated.

	(color luminance > 0.32) ifTrue: [
		color := color adjustBrightness: (color luminance * 0.38) negated.
	].
	color luminance > 0.6 ifTrue: [
		"One final check so gray for very light avatars isn't too pale."
		color := color adjustBrightness: (color luminance * 0.6) negated.
	].

	^ color alpha: 1.0

! !

!QColors class methodsFor: 'island-colors' stamp: 'bgf 11/27/2009 18:11'!
topicCardColor
	^ self color: [ VectorColor r: 0.96 g: 0.96 b: 0.93 ]! !

!QColors class methodsFor: 'island-colors' stamp: 'bgf 11/27/2009 18:11'!
topicCardTitleColor
	^ self color: [ VectorColor r: 0.8 g: 0.88 b: 0.94 ]! !

!QColors class methodsFor: 'island-colors' stamp: 'bgf 11/27/2009 20:35'!
warmWhiteColor
	^ self color:  [  VectorColor r:0.99 g:0.985 b:0.97 ]! !

!QColors class methodsFor: 'island-colors' stamp: 'bgf 11/27/2009 17:49'!
whiteAlphaColor
	"White basis for alpha-pass materials "
	^ self color: [  VectorColor r: 1 g: 1 b: 1 a: 0.999 ]! !

!QColors class methodsFor: 'island-colors' stamp: 'bgf 11/27/2009 17:31'!
whiteOverColor
	"For color-neutral brightening mouseover feedback of draggable compound objects."
	^ self color: [  VectorColor r: 1 g: 1 b: 1 a: 0.25 ]! !

Object subclass: #QCompany
	instanceVariableNames: 'companyName tObjectID companyInformation parentCompany subsidiaryCompanies departments realms administrators'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Objects-User'!

!QCompany commentStamp: 'RJT 5/29/2007 23:10' prior: 0!
QCompany is not used yet.

The company can be in a chain of companies it could have a parentCompany or be the parent of a subsidiary.  At the company level users can further be defined by departments.   There can be one or more departments that place a user within a company.  Departments can have their own realms but this doesn't mean that a user will have access to that realm.  Departments also have their own departments.  Consider a Corporate Parent, with a Company that has a North American Department which has a sales East Cost department, you you are in the New York branch office, on the blue team. :)

Structure:
 companyName 		<String> -- Name of the company
 tObjectID 			<TObjectID> -- ID of company
 companyInformation <addesss, phone numbers ...> - information about the company
 parentCompany 	<QCompany> -- the parent of this company
 subsidiaryCompanies <Collection of QCompany> -- this company is the parent of these companies
 departments		<Collection of QDepartment> -- The departments that place a user with an organization
 realms				<Collection of Strings> -- The realms that belong to this organization
 administrators		<Collection of QAdministrator> -- The adminstrators allowed to add realms to this companies

Not currently used!

!QCompany class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QCompany class methodsFor: 'instance creation' stamp: 'RJT 5/23/2007 10:07'!
createForAdminstrator: aUser
	| aCompany |
	"return a new company and add aUser to the administrators, also add the company to the list of companies that this user administrates, this administrator may not be an employee of the company, for exampe this administrator may be a Qwaq Employee or some third party hired to administer users for this company"
	aCompany := self new
		administrators add: aUser;
		yourself.
	aUser administrates: aCompany.
	^aCompany.
	
	! !

!QCompany methodsFor: 'accessing' stamp: 'RJT 5/23/2007 10:04'!
administrators
	"Answer the value of administrators"

	^ administrators! !

!QCompany methodsFor: 'accessing' stamp: 'RJT 5/23/2007 10:04'!
administrators: anObject
	"Set the value of administrators"

	administrators := anObject! !

!QCompany methodsFor: 'accessing' stamp: 'RJT 5/23/2007 09:53'!
companyInformation
	"Answer the value of companyInformation"

	^ companyInformation! !

!QCompany methodsFor: 'accessing' stamp: 'RJT 5/23/2007 09:53'!
companyInformation: anObject
	"Set the value of companyInformation"

	companyInformation := anObject! !

!QCompany methodsFor: 'accessing' stamp: 'RJT 5/23/2007 09:53'!
companyName
	"Answer the value of companyName"

	^ companyName! !

!QCompany methodsFor: 'accessing' stamp: 'RJT 5/23/2007 09:53'!
companyName: anObject
	"Set the value of companyName"

	companyName := anObject! !

!QCompany methodsFor: 'accessing' stamp: 'RJT 5/23/2007 09:53'!
departments
	"Answer the value of departments"

	^ departments! !

!QCompany methodsFor: 'accessing' stamp: 'RJT 5/23/2007 09:53'!
departments: anObject
	"Set the value of departments"

	departments := anObject! !

!QCompany methodsFor: 'initialize-release' stamp: 'RJT 5/23/2007 10:05'!
initialize
	super initialize.
	self companyName: 'Enter a Company Name' copy.
	self tObjectID: (TObjectID new: 16).
	self companyInformation: nil.
	self parentCompany: nil.
	self subsidiaryCompanies: OrderedCollection new.
	self departments: OrderedCollection new.
	self realms: OrderedCollection new.
	self administrators: OrderedCollection new.! !

!QCompany methodsFor: 'accessing' stamp: 'RJT 5/23/2007 09:53'!
parentCompany
	"Answer the value of parentCompany"

	^ parentCompany! !

!QCompany methodsFor: 'accessing' stamp: 'RJT 5/23/2007 09:53'!
parentCompany: anObject
	"Set the value of parentCompany"

	parentCompany := anObject! !

!QCompany methodsFor: 'accessing' stamp: 'RJT 5/23/2007 09:53'!
realms
	"Answer the value of realms"

	^ realms! !

!QCompany methodsFor: 'accessing' stamp: 'RJT 5/23/2007 09:53'!
realms: anObject
	"Set the value of realms"

	realms := anObject! !

!QCompany methodsFor: 'accessing' stamp: 'RJT 5/23/2007 09:53'!
subsidiaryCompanies
	"Answer the value of subsidiaryCompanies"

	^ subsidiaryCompanies! !

!QCompany methodsFor: 'accessing' stamp: 'RJT 5/23/2007 09:53'!
subsidiaryCompanies: anObject
	"Set the value of subsidiaryCompanies"

	subsidiaryCompanies := anObject! !

!QCompany methodsFor: 'accessing' stamp: 'RJT 5/23/2007 09:53'!
tObjectID
	"Answer the value of tObjectID"

	^ tObjectID! !

!QCompany methodsFor: 'accessing' stamp: 'RJT 5/23/2007 09:53'!
tObjectID: anObject
	"Set the value of tObjectID"

	tObjectID := anObject! !

Object subclass: #QCustomsEntryData
	instanceVariableNames: 'name timestamp place fileCard internedPath'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Objects-Space'!

!QCustomsEntryData commentStamp: '<historical>' prior: 0!
A replicated item that was asynchronously imported.
It is passed by copy to the harness for easy manipulation.!

!QCustomsEntryData class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QCustomsEntryData class methodsFor: 'island-args' stamp: 'hrs 3/1/2009 11:16'!
howToPassAsArgument
	^ #passByCopy:! !

!QCustomsEntryData methodsFor: 'persist' stamp: 'hrs 3/1/2009 20:15'!
constructFrom: pStream
	name := pStream readString: 'name' default: nil.
	timestamp := pStream readNumber: 't' default: 0.
	fileCard := pStream readObject: 'card' default: nil.
	"The anticipated usage is that when a customs entry has a place specified, 
	the filecard will immediately be placed and the object removed from customs.
	Furthermore, entries will be processed as soon as someone handles the 
	second step of the asynchronous import (i.e., as soon as someone asks the
	server for new imports). Thus we would never persist an entry with a place 
	specified.  However, plans change.  In the meantime, knowing how to reconstruct
	an entry with a place specified will not hurt anything."
	place := pStream readString: 'place' default: ''.! !

!QCustomsEntryData methodsFor: 'accessing' stamp: 'jcg 6/2/2009 09:06'!
documentType
	^QDocumentType forSuffix: self fileExtension asUppercase! !

!QCustomsEntryData methodsFor: 'accessing' stamp: 'hrs 3/1/2009 11:48'!
fileCard
	"Postprocessing produces a filecard. Initially nil."
	^fileCard! !

!QCustomsEntryData methodsFor: 'accessing' stamp: 'hrs 3/1/2009 11:21'!
fileCard: aFileCard
	fileCard := aFileCard! !

!QCustomsEntryData methodsFor: 'accessing' stamp: 'hrs 3/1/2009 14:44'!
fileExtension
	"Answer the filename extension."
	^self name copyAfterLast: $..! !

!QCustomsEntryData methodsFor: 'accessing-names' stamp: 'hrs 3/1/2009 14:46'!
fileName
	"Answer just the document name, with extension (e.g., no usename)."
	^self name copyAfter: $/.! !

!QCustomsEntryData methodsFor: 'accessing-names' stamp: 'jcg 6/2/2009 10:17'!
internedPath
	"eg:   'services/JoshForum/sharedDocuments/jgargus@qwaq.com/foobar-1.png' "
	^internedPath! !

!QCustomsEntryData methodsFor: 'accessing-names' stamp: 'jcg 6/2/2009 10:18'!
internedPath: path
	"eg:   'services/JoshForum/sharedDocuments/jgargus@qwaq.com/foobar-1.png' "
	internedPath := path! !

!QCustomsEntryData methodsFor: 'accessing-names' stamp: 'jcg 6/2/2009 10:22'!
internedServicePath
	"Relative to current service."
	"eg:   'sharedDocuments/jgargus@qwaq.com/foobar-1.png' "
	| start end |
	start := internedPath indexOfSubCollection: 'sharedDocuments/'.
	end := internedPath size.
	^internedPath copyFrom: start to: end! !

!QCustomsEntryData methodsFor: 'accessing-names' stamp: 'jcg 6/2/2009 10:22'!
internedShortName
	"Just the filename"
	"eg:   'foobar-1.png' "
	^internedPath copyAfterLast: $/! !

!QCustomsEntryData methodsFor: 'accessing-names' stamp: 'hrs 3/1/2009 14:45'!
name
	"The full unique designation, e.g., user/fileName.fileExtension"
	^name! !

!QCustomsEntryData methodsFor: 'initialize-release' stamp: 'hrs 3/1/2009 20:16'!
name: aString time: anInteger place: aPlaceName
	name := aString.
	timestamp := anInteger.
	place := aPlaceName
	! !

!QCustomsEntryData methodsFor: 'persist' stamp: 'hrs 3/1/2009 20:15'!
persistOn: pStream
	pStream storeString: 'name' value: name default: nil.
	pStream storeNumber: 't' value: timestamp default: 0.
	pStream storeObject: 'card' value: fileCard default: nil.
	pStream storeString: 'place' value: place default: ''.! !

!QCustomsEntryData methodsFor: 'accessing' stamp: 'hrs 3/1/2009 11:04'!
place
	"The placeName to drop the card, as optionally specified when the file was uploaded."
	^place! !

!QCustomsEntryData methodsFor: 'accessing-names' stamp: 'jcg 6/2/2009 10:12'!
serviceFileName
	"Answer the filename on the server relative to this forum's service directory."
	"eg:   'jgargus@qwaq.com/foobar.png' "
	^self name! !

!QCustomsEntryData methodsFor: 'testing' stamp: 'hrs 3/12/2009 14:52'!
shortCircuitCustoms
	"Should we avoid sending the import to customs and just put a file card into the space?"
	^true "place isEmpty not"! !

!QCustomsEntryData methodsFor: 'accessing-names' stamp: 'jcg 6/2/2009 10:25'!
shortName
	"Just the filename"
	"eg:   'foobar.png' "
	^name copyAfterLast: $/! !

!QCustomsEntryData methodsFor: 'accessing' stamp: 'hrs 3/1/2009 11:00'!
timestamp
	"utcSeconds on server at time of import"
	^timestamp! !

!QCustomsEntryData methodsFor: 'accessing' stamp: 'hrs 3/1/2009 15:36'!
user
	^self name copyUpTo: $/! !

Object subclass: #QDefaultUIFormatter
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Objects-Space'!

!QDefaultUIFormatter commentStamp: '<historical>' prior: 0!
UI position readout Formatter for position data in the Qwaq demo spaces.
This is mostly just a test for the position formatter support;
it offsets the ground plane so that the readout shows zero when
the demo avatar is walking on the ground.
X and Z readouts delegate to the default QNavPanel implementation.

You use this formatter in a TSpace by implementing 

MySpace #uiFormatter
	^ #QDefaultUIFormatter!

!QDefaultUIFormatter class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QDefaultUIFormatter methodsFor: 'as yet unclassified' stamp: 'bgf 12/17/2006 15:02'!
xDisplayString: aFloat

	^ QPositionReadout xDisplayString: aFloat.! !

!QDefaultUIFormatter methodsFor: 'as yet unclassified' stamp: 'bgf 12/17/2006 15:03'!
yDisplayString: aFloat

	"In the conference spaces, the floor is at Y=0,
	 so the avatar reports itself at height 5 (QPlainAvatar heightAboveGround.)"

	^ QPositionReadout yDisplayString: (aFloat-5).! !

!QDefaultUIFormatter methodsFor: 'as yet unclassified' stamp: 'bgf 12/17/2006 15:03'!
zDisplayString: aFloat

	^ QPositionReadout zDisplayString: aFloat.! !

Object subclass: #QDocumentManager
	instanceVariableNames: 'queue docsToFetch thumbsToFetch cacheMgr owner fetchProcess fetching reportedMissing refetches uploadProgress mutex'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Objects-Document'!

!QDocumentManager commentStamp: '<historical>' prior: 0!
Replication of shared documents to participant local cache;
a manager wired into TIsland resource management 
(see QwaqCroquetSession #documentManager)

When a shared document has cacheLocally=true,
the resource system arranges to fetch a cache copy in the background.  
The documents are cached under the filename

	./TFileCacheManager directory/<SharedDocument SHA>.<sharedDocument cacheSuffix>

!

!QDocumentManager class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QDocumentManager class methodsFor: 'utilities' stamp: 'bgf 2/14/2009 21:40'!
incrementName: aFileName

	"Prepare a numbered successor to this filename, of the form
			foo-2.jpg       if the original was foo-1.jpg	"

	| base suffix dashIndex prevNum |
	(aFileName includes: $.) ifFalse: [
		base := aFileName.
		suffix := ''.
	] ifTrue: [
		base := aFileName copyUpToLast: $. .
		suffix := '.',  (aFileName copyAfterLast: $.).
	].

	dashIndex := base lastIndexOf: $- ifAbsent: [nil].
	dashIndex ifNil: [
		^ base, '-1', suffix.
	] ifNotNil: [
		(prevNum := base copyFrom: dashIndex+1 to: base size) isAllDigits ifTrue: [
			^ (base copyFrom: 1 to: dashIndex), (prevNum asNumber+1), suffix
		] ifFalse: [
			^ base, '-1', suffix.
		].
	].! !

!QDocumentManager methodsFor: 'cache' stamp: 'bgf 10/23/2008 03:46'!
asServiceManagerPath: aString
	"This is a hack:  Strip off everything until we hit 'services/' as the request path 
	for the service manager. If there is no 'services/' then insert one to make it look right."
	| path |
	path := aString.

	"Race condition avoidance on logout."
	owner serviceMgr ifNil: [ ^ path ].

	"Per-user storage has its own rules.."
	(path beginsWith: '/users/') ifTrue: [ ^ path ].

	"Bundled storage has its own rules"
	(TContentManager isBundledFileName: path) ifTrue: [
		^ 'system-resources/bundled/', path ].

	[path isEmpty] whileFalse:[
		(path beginsWith: 'services/') ifTrue:[^'/',path].
		path := path copyAfter: $/.
	].
	^'/services', (aString allButFirst: owner serviceMgr baseUrl size)! !

!QDocumentManager methodsFor: 'import-upload' stamp: 'bgf 3/14/2009 01:19'!
checkExistence: shortName path: svcPath policy: policy
	"Determine if the given file is present on the server.
	 If if is not, then simply return the shortName.
	 If it is, then the result depends on policy... 
		#error   - raise an error
		#clobber 	    - disregard, return shortName
		#makeUnique  - return unique new name based on shortname (no-clobber).
		#leaveOriginal - do nothing - return nil to signal that no file should be uploaded."
	| entries |

	(policy = #clobber) ifTrue: [ ^ shortName].
	[
		entries := owner serviceMgr listFiles: svcPath, '/', shortName.
	] on: Error do: [
		"Assume a listing error means there is no such containing directory... so no such file."
		entries := #()
	].
	entries isEmpty ifTrue: [  ^ shortName ].	"Unique!!"

	(policy == #makeUnique) ifTrue: [
		^ self checkExistence: (self class incrementName: shortName) path: svcPath policy: #makeUnique.
	].
	(policy == #leaveOriginal) ifTrue: [
		^ nil
	].

	"The remaining policy is #error..."
	self error: 'Shared file exists: ', svcPath,'/', shortName.! !

!QDocumentManager methodsFor: 'queue' stamp: 'bgf 11/8/2008 19:09'!
count: request

	"Increment the appropriate pending counter,
	 when this request is being enqueued.   MUST MATCH #discount: "

	( #(thumb preview prefetch) includes: request requestType)  ifTrue: [
		thumbsToFetch := (thumbsToFetch + 1) max: 0.
	] ifFalse: [
		docsToFetch := (docsToFetch + 1) max: 0.
	].
	
	request refetch ifTrue: [ 
		refetches := refetches + 1 max: 0.
	].
! !

!QDocumentManager methodsFor: 'accessing' stamp: 'bgf 6/13/2008 02:09'!
countToFetch

	"The user-relevant number of resource requests remaining to be retrieved."
	^ (queue size) + (fetching ifNil:[0] ifNotNil: [1]).
		! !

!QDocumentManager methodsFor: 'initialize-release' stamp: 'bgf 8/6/2009 01:53'!
destroy
	mutex ifNotNil:[
		mutex critical: [
			fetchProcess ifNotNil: [
				fetchProcess terminate.
				fetchProcess := nil.
			]].
		queue nextPut: nil	"Just to kick a wakeup." ].
	queue := nil.
	cacheMgr := nil.
	owner := nil.
	mutex := nil.! !

!QDocumentManager methodsFor: 'queue' stamp: 'bgf 8/30/2009 13:43'!
discardIslandResources: anIsland
	"When an island is abandoned, abandon all requests for it."
	queue flushAllSuchThat: [ : r | r requestor notNil and: [ r requestor island = anIsland ]].
	self discardPrefetches.! !

!QDocumentManager methodsFor: 'queue' stamp: 'bgf 12/6/2009 13:44'!
discardPrefetches
	"Discard all prefetches-in-progress, 
	 and abandon any fetches for requesting objects that are far-refs into dead islands.
	 Clears the queue after an abandoned login, or a quick log out."
	queue ifNotNil: [
		queue flushAllSuchThat: [ : r | r requestType = #prefetch ].
		queue flushAllSuchThat: [ : r | (r requestor isKindOf: TFarRef) and: [ r requestor isValidFarRef not ]].
	].! !

!QDocumentManager methodsFor: 'queue' stamp: 'bgf 6/14/2008 20:06'!
discount: request
	"When a fetch is complete, uncount it from pending"

	queue isEmpty ifTrue: [
		^ docsToFetch := thumbsToFetch := refetches := 0.		
			"Belt&braces safety, so if counts go outta whack, they'll self-heal."
	].
	( #(thumb preview) includes: request requestType)  ifTrue: [
		thumbsToFetch := (thumbsToFetch - 1) max: 0.
	] ifFalse: [
		docsToFetch := (docsToFetch - 1) max: 0.
	].
	request refetch ifTrue: [
		refetches := (refetches - 1) max: 0.
	].! !

!QDocumentManager methodsFor: 'user interface' stamp: 'bf 11/20/2008 13:09'!
displayProgressTitled: aString in: workBlock
	"Evaluate workBlock with the pre-set uploadProgress or a newly created ProgressBar. This ensures progress is displayed until all interested callers actually set uploadProgress"

	uploadProgress
		ifNotNil: [^workBlock value: uploadProgress].

	CProgressBar displayProgress: aString
		at: owner bounds center
		from: 0.0 to: 1.0 during: [:bar|
			self waitTick.
			workBlock value: bar].
	! !

!QDocumentManager methodsFor: 'cache' stamp: 'bgf 8/6/2009 01:44'!
doOneFetch: request
	"Private to the fetch process - fetch a given request."
	| serviceMgr path data msg short |
	"Fetching is set but not cleared, so we can use it in asynchronous status-line messages."

	mutex critical: [
		fetching := request.
	].
	serviceMgr := owner serviceMgr.
	path := fetching serviceManagerPath.
	short := path copyAfterLast: $/.
		
	[
		"Fetching is counted in count to fetch, hence the 1 below."
		(self countToFetch > 1) ifTrue: [owner flowControlFlag: true ].
		[ 
			"One last check to see if this was already fetched..."
			request refetch ifFalse: [
				data := (cacheMgr find: (request cacheFileName)) ].
			data ifNil: [
				self reportIntent: path shortName: short kind: (request requestType).
				data := serviceMgr get: path informing: [:value|
					owner flowControlFlag: true.
					fetching fetchProgress: value
				].
				data ifNil: [ self error: 'Empty fetch result.' ].
				fetching cacheFileName notNil ifTrue: [
					cacheMgr cache: data to: fetching cacheFileName.
				] ifFalse: [
					self error: 'Fetched but no cache file name.'.
				]].
		] on: Error do: [:ex|
			| nofile| 
			nofile := (ex description includesSubString: 'FileDoesNotExist').
			msg := nofile 
						ifTrue: [ 'Not found: ', path ]
						ifFalse: [ 'Cannot fetch: ', path, ': ', (ex description) ].
			
			(nofile and: [ reportedMissing includes: short ]) ifFalse: [
				self reportAbsence: path shortName: short message: msg.
			].
			data := nil.		
		].
		mutex critical: [ 
			fetching := nil.
		].

		"Inform interested parties about completion of this fetch
		 (listeners can observe the request, or the Document-Manager itself)"
		self signal: #fetchComplete with: request.
		request fetchComplete: data.
		self discount: request.

	]  ensure: [
		owner flowControlFlag: false.
	].! !

!QDocumentManager methodsFor: 'queue' stamp: 'bgf 3/25/2009 13:32'!
enqueueRequest: request

	"Queue this request, and also increment the appropriate pending counter."
	self count: request.
	[
		queue nextPut: request beforeFirst: [ : r | (r priority) > (request priority) ].
	] on: Error do: [ : ex | 
		QLogger logError: ex.
	 ].! !

!QDocumentManager methodsFor: 'cache' stamp: 'bgf 2/5/2008 08:36'!
fetchDocument: aSharedDocRef

	"Part of resource management  - add the given document
 	to the interest list of the manager; it will be retrieved from the server 
	in the background. 
	Answers the request being downloaded or nil if it's finished already;
	clients can use the request to listen for progress upates and 
	completion of the request."

	| request |

	"Cannot assume we're in a valid state to fetch..."
	aSharedDocRef isFarRef ifFalse: [^ nil ].
	aSharedDocRef isValidFarRef ifFalse: [ ^ nil ].

	request := QDocumentRequest new.
	request requestor: aSharedDocRef.
	request cacheFileName: (aSharedDocRef get:#cacheFileName).
	request downloadUrl: (aSharedDocRef get:#serviceURL).
	request serviceManagerPath:  (aSharedDocRef get:#asServiceManagerPath).
	request requestType: #document.
	^self fetchRequest: request.! !

!QDocumentManager methodsFor: 'cache' stamp: 'bgf 3/22/2007 11:38'!
fetchProcess

	"The Process running the background document sync thread,
	 or nil if it is not running."

	^fetchProcess! !

!QDocumentManager methodsFor: 'cache' stamp: 'bgf 3/1/2009 10:02'!
fetchQMSPreview: qmsApp

	"Request the screenshot preview for a QVNCApp (qmsApp).

	 When the result is received, calls qmsApp previewFormData: <thebinarydata>.
 	 That may happen synchronously if the cache already has the data,
	 or asynchronously if the resource must be fetched."

	| request fileName filePath url svcUrl svcName appRef |

	appRef := qmsApp app.
	fileName :=  (qmsApp previewFileName).
	fileName ifNil: [ 	"Previews not wanted for this app."
		^ self ].

	"Already in the cache?  Just use it."
	(cacheMgr find: fileName) ifNotNilDo: [ : data |
		qmsApp previewFormData: data.
		^ self
	].

	"Do we have a URL available?  (If not we may have reconnected, whatevah"   
	owner session ifNotNil: [
		(owner session findIslandUrl: appRef island) ifNotNilDo: [ : iurl |
			svcUrl := iurl copyUpToLast: $/.
			svcName := (svcUrl copyAfterLast:$/).
		]].
	svcName ifNil: [
		^ self
	].

	request := QDocumentRequest new.

	filePath := (QEmbeddedApp previewPath), fileName.
	url := TPostcard makeUrl: filePath relativeTo: svcUrl.

	request requestor: qmsApp.
	request cacheFileName: fileName.
	request downloadUrl: url.
	request serviceManagerPath: '/services/', svcName, '/', filePath.
	request requestType: #preview.

	(self fetchRequest: request) ifNotNilDo: [ :req |
		qmsApp startScript: #previewFormData: when: { req . #fetchComplete }.
	].! !

!QDocumentManager methodsFor: 'cache' stamp: 'bgf 12/3/2009 15:07'!
fetchRequest: docRequest

	"Part of resource management  - add the given document
 	to the interest list of the manager; it will be retrieved from the server 
	in the background.
	Answers the request being downloaded or nil if it's finished already;
	clients can use the request to listen for progress upates and 
	completion of the request."

	| name qurl pending |
	mutex ifNil: [ ^ nil ].		"Destroyed"
	docRequest refetch ifFalse: [
		name := docRequest cacheFileName.
		(name isNil or: [cacheMgr has: name]) ifTrue:[ ^nil ].
	].
	pending := queue detect:[:req| req = docRequest] ifNone:[nil].
	pending ifNotNil:[^pending]. "in queue already"
	mutex critical: [
		docRequest = fetching ifTrue:[^fetching]. "currently being fetched"	
	].

	qurl := docRequest downloadUrl.
	qurl ifNil:[^nil].

	self enqueueRequest: docRequest.
	self startFetchProcessIfNeeded.

	^docRequest! !

!QDocumentManager methodsFor: 'cache' stamp: 'bgf 2/8/2010 12:09'!
fetchThumb: thumbRef
	"Part of resource management  - add the given document
 	to the interest list of the manager; it will be retrieved from the server 
	in the background.
	Answers the request being downloaded or nil if it's finished already;
	clients can use the request to listen for progress upates and 
	completion of the request."
	| filename request requestUrl  sha |

	"Cannot assume we're in a valid state to fetch..."
	mutex ifNil: [ ^ nil ].		"Destroyed"
	thumbRef isValidRef ifFalse: [ ^ nil ].
	sha := (thumbRef get: #sha).

	"In queue or being gotten already?"
	mutex critical: [
		(fetching notNil and: [fetching cacheFileName = sha]) ifTrue: [ 
			request := fetching
		]].
	request ifNil: [
		request := (queue detect:[:req| req cacheFileName = sha] ifNone: [nil]).
	].

	request ifNotNil: [
		request requestType = #prefetch ifTrue: [
			"Take over this prefetch, so I'll be signaled back ASAP when complete."
			request requestType: #thumb;
					requestor: thumbRef.
			^ request].
			"If not a prefetch, this item has already been requested on the normal path."
		^ nil].


	filename := thumbRef get:#fileName.
	filename ifNil: [ ^ nil ].

	((TContentManager isBundledFileName: filename) or: [ filename beginsWith: '/users/']) ifTrue: [
		requestUrl := filename.
	] ifFalse: [
		|  islandUrl |
		islandUrl := owner session ifNotNil: [ owner session findIslandUrl: thumbRef island ].
		islandUrl ifNil: [ ^ nil ].
		requestUrl := TPostcard
						makeUrl: filename
						relativeTo: (islandUrl copyUpToLast: $/).
	].

	request := QDocumentRequest new.
	request requestor: thumbRef.
	request cacheFileName: sha.
	request downloadUrl: requestUrl.
	request serviceManagerPath:  (self asServiceManagerPath: requestUrl).
	request requestType: #thumb.

	^self fetchRequest: request! !

!QDocumentManager methodsFor: 'import-upload' stamp: 'bgf 1/7/2008 12:32'!
incrementName: aFileName

	"Prepare a numbered successor to this filename, of the form
			foo-2.jpg       if the original was foo-1.jpg	"

	| base suffix dashIndex prevNum |
	(aFileName includes: $.) ifFalse: [
		base := aFileName.
		suffix := ''.
	] ifTrue: [
		base := aFileName copyUpToLast: $. .
		suffix := '.',  (aFileName copyAfterLast: $.).
	].

	dashIndex := base lastIndexOf: $- ifAbsent: [nil].
	dashIndex ifNil: [
		^ base, '-1', suffix.
	] ifNotNil: [
		(prevNum := base copyFrom: dashIndex+1 to: base size) isAllDigits ifTrue: [
			^ (base copyFrom: 1 to: dashIndex), (prevNum asNumber+1), suffix
		] ifFalse: [
			^ base, '-1', suffix.
		].
	].
		! !

!QDocumentManager methodsFor: 'initialize-release' stamp: 'bgf 1/16/2009 17:26'!
initialize
	"The queue contains QSharedDocument far-refs."
	queue := SharedQueue new.
	docsToFetch := thumbsToFetch := refetches := 0.
	mutex := Mutex new.
	cacheMgr := TContentManager default cacheManager.
	reportedMissing := Set new.! !

!QDocumentManager methodsFor: 'accessing' stamp: 'bgf 3/20/2008 13:50'!
isAccessibleDocument:  aSharedDoc

	"Returns false if the document is known (without checking access control)
	 to be inaccessible, based on its file path alone.  I.e. detects the private
 	 storage paths of other users and returns false for such documents.

	 Used as a preflight check to avoid bursts of attempts to download
 	 other user's private inventory just because they've come into a forum.
	 Also prevents downloading when we don't know the underlying file path
 	 during network outages, etc."

	^ self isAccessiblePath: (aSharedDoc get:#asServiceManagerPath)! !

!QDocumentManager methodsFor: 'accessing' stamp: 'bgf 6/19/2009 14:17'!
isAccessiblePath:  sharedName

	"Returns false if the document is known (without asking the server)
	 to be inaccessible, based on its file path alone.  I.e. detects the private
 	 storage paths of other users and returns false for such documents.
	 This is a pre-flight check to test whether it's worth trying to fetch a file."

	sharedName ifNil: [ ^ false ].	"If we don't have a path, it cannot be fetched."

	"If the file is in some other user's private storage,
	 we assume we won't be able to fetch it.
	 This isn't quite exact, but it covers all the pathing policies we currently use."
	((sharedName beginsWith: '/users/') and: [
	 (sharedName includesSubString: '/private/')]) ifTrue: [
		^ (sharedName includesSubString: '/', (owner userLogin), '/private/') 
	].

	"Otherwise, we'll just ask the for the document to be sure;
	 at time of writing, documents not in users' private storage are accessible."
	^ true.

! !

!QDocumentManager methodsFor: 'accessing' stamp: 'bgf 2/13/2009 00:19'!
newTimeStamp: description
	"Convenience - produce a current-time timestamp to apply to a document."
	^ QTimeStamp new
		who: (owner userLogin) 
		what: description
		utc: (Time utcSeconds)! !

!QDocumentManager methodsFor: 'queue' stamp: 'bgf 9/2/2009 11:51'!
nextRequest
	"Blocking - return the next fetch request from the queue.
	 Note that discount: request comes later, after the request is done processing."

	"A bit wasteful, but simplifies every other part of the code...
	 sort the queue before pulling each new item.
	 (The order changes with various minor changes in attributes of requests,
	 so automating would be a bit cumbersome and full of on-whatever signalling chaos.) "
	queue sort: [ : r1 : r2 | (r1 priority) <= (r2 priority) ].
	^ queue nextTimeoutSeconds: 2
! !

!QDocumentManager methodsFor: 'queue' stamp: 'bgf 12/9/2009 15:22'!
noteNeededThumbs: thumbs
	"The given thumbs are needed for current display;
	 raise the priority of their fetch requests."

	thumbs do: [ : th |
		queue ifNil: [ ^ self ].
		(queue detect: [ : r | r requestor = th ] ifNone: [nil]) ifNotNil: [ : r | r isNeeded: true ].
	].! !

!QDocumentManager methodsFor: 'initialize-release' stamp: 'bgf 3/14/2009 01:37'!
owner: aParticipant
	"The participant UI - gives us access to session, current service manager, etc;
	 must be set before using the document manager."
	owner := aParticipant.
	[
		"Set up the defaullt pre-session cache manager to be able to read my local cache,
	 	 which is pathed to the login name of the service-session of the owner."
		cacheMgr := TContentManager default cacheManager.
		cacheMgr directory: (
			(FileDirectory cacheDirectory directoryNamed: (owner userLogin))
				assureExistence).
		
	] on: Error do: [ :ex|
		QLogger logError: ex from: self.
	].! !

!QDocumentManager methodsFor: 'cache' stamp: 'bgf 3/14/2009 01:39'!
prefetchResource: uri sha: sha service: qService
	"Schedule a download of the given resource (unless it is already cached).
	 This request can be made with only the uri and sha (i.e. before an in-island thumbnail exists).
	 The download will be scheduled even though we don't yet have a requestor.
	 See also fetchThumb:, which 'fixes' such requests to set the requestor,
	 so if the real thing expresses interest, it gets proper notification in fetch-complete.

	 On the other hand, if this fetch completes before it is needed,  
	 what is accomplished is that the resource is already on disk locally, 
	 ready to be used when asked for."

	| request |

	(cacheMgr has: sha) ifTrue: [ ^ nil ].
	(queue detect:[:req| req cacheFileName = sha] ifNone: [nil]) ifNotNil: [ ^ nil ].

	request := QDocumentRequest new.
	request requestor: nil.
	request cacheFileName: sha.
	request downloadUrl: (
			TPostcard makeUrl: uri 
					   relativeTo: (qService url)).
	request serviceManagerPath: (self asServiceManagerPath: request downloadUrl).
	request requestType: #prefetch.
	^self fetchRequest: request. ! !

!QDocumentManager methodsFor: 'accessing' stamp: 'ar 7/12/2011 12:57'!
privateDocsDir
	"Outer directory where private-copies of editable documents are kept. 
	 (i.e. not the cache - the user may change things in here.)
	 <Documents> / QwaqForums / <host> / <realm> / "
	[ 
		| dir rn |
		dir := FileDirectory on: (QFileDialog getFileLocation: 'documents').
		dir := dir directoryNamed: 'OpenQwaq'.
		dir := dir directoryNamed: (owner serviceHost).
		rn := (owner serviceMgr realm) ifNil: ['Default'].
		(rn first = $/) ifTrue: [ rn := rn allButFirst ].
		dir := dir directoryNamed: rn.
		dir assureExistence.
		^dir
	] on: Error do: [ : ex |
		| dir |
		QLogger logWarning: 'Cannot make private document area: ', ex description.
		dir := FileDirectory on: (QFileDialog getFileLocation: 'documents').
		dir assureExistence.
		^ dir
	].! !

!QDocumentManager methodsFor: 'accessing' stamp: 'bgf 3/9/2009 19:19'!
privateFileNameFor: document

	"Return the full file name for private local copy of the given QAbstractDocument (far-ref.)
	 This goes into the #privateDocDir,  in a subdirectory matching the services storage
	 (eg. forum / sharedDocuments / someone / foobar / file "
	| dir fileName suburi |
	dir := self privateDocsDir.

	(document get: #serviceName) ifNotNilDo: [ : sn |
		(sn first = $/) ifTrue: [ sn := sn allButFirst ].
		dir := dir directoryNamed: sn.
	] ifNil: [ 
		dir := dir directoryNamed: 'Default'.	
	].

	suburi := (document get: #sharedName) ifNil: [document get:#shortName].
	suburi ifNil: [ 
		self error: 'Cannot identify document for local path.'
	].

	(suburi first = $/) ifTrue: [ suburi := suburi allButFirst ].
	fileName := (dir pathFromURI: suburi).

	(FileDirectory forFileName: fileName) assureExistence.
	^ fileName! !

!QDocumentManager methodsFor: 'cache' stamp: 'bgf 3/1/2009 10:05'!
refetchThumb: thumbRef

	"Part of resource management  - add the given document
 	to the interest list of the manager; 
	it will be retrieved from the server  in the background
	whether or not is already present in the cache or content manager
 	This is the operation to force a fetch.

	Answers the request being downloaded or nil if it's finished already;
	clients can use the request to listen for progress upates and 
	completion of the request."

	| filename request requestUrl |
	owner session ifNil: [ ^nil ].
	request := QDocumentRequest new.
	request requestor: thumbRef.
	request cacheFileName: (thumbRef get:#sha).
	request refetch: true.
	filename := thumbRef get:#fileName.
	filename ifNil: [ ^ nil ].
	(filename beginsWith: '/users/') ifTrue: [
		requestUrl := filename.
	] ifFalse: [
		requestUrl := TPostcard makeUrl: filename
					relativeTo: ((owner session findIslandUrl: thumbRef island) copyUpToLast: $/).
	].
	requestUrl ifNil: [ ^ nil ].
	request downloadUrl: requestUrl.
	request serviceManagerPath:  (self asServiceManagerPath: requestUrl).
	request requestType: #thumb.
	^self fetchRequest: request.! !

!QDocumentManager methodsFor: 'private' stamp: 'bgf 2/20/2009 12:20'!
reportAbsence: path shortName: short message: msg
	"Log failure to fetch... unless its an app-preview thumb; 
	 logging those is too noisy
	 and we expect often to not find them."

	(reportedMissing includes: path) ifTrue: [ ^ self ].
	(path includesSubString: '/resources/screens/') ifTrue: [ ^ self ].

	QLogger logInfo: msg.
	reportedMissing add: path.

! !

!QDocumentManager methodsFor: 'private' stamp: 'bgf 7/9/2009 17:33'!
reportIntent: path shortName: short kind: kind
	"Log intent to fetch... unless its an app-preview thumb; 
	 logging those is too noisy
	 and we expect often to not find them."

	(reportedMissing includes: path) ifTrue: [ ^ self ].
	(path includesSubString: '/resources/screens/') ifTrue: [ ^ self ].

	QLogger logInfo: 'Get ', (kind ifNil: [ 'File']), ': ', short.	! !

!QDocumentManager methodsFor: 'accessing' stamp: 'bgf 6/14/2008 20:08'!
resourceLoadingStatus
	"The status string to display while loading resources; nil if nothing.
	 We include the content manager request-list in the tallies,
	 since texture load-from-disk is a notable part of the busy time at startup.
	 (So we help the user see why things are busy/slow at that time.)"

	|  message count elements |

	count := self countToFetch.
	(count <= 0) ifTrue:  [ ^ nil ].
	"This prevents retry bursts (usually app-preview refreshment)	
	  from being advertised to the status bar."
	count > refetches ifFalse: [ ^ nil ].
	
	"As long as we're not in an edge case, force the flow-control flag on
	 so long as we have stuff queued."
	(docsToFetch + thumbsToFetch > 1) ifTrue: [
		"Set the flag to enforce it, I know I'll clear it when the fetch completes."
		owner flowControlFlag: true.
	].
	elements := count - docsToFetch.		"Elements are all the graphical items."
	message := WriteStream on: (String new).
	message nextPutAll: 'Downloading ('.
	docsToFetch > 1 ifTrue: [
		message nextPutAll: '', docsToFetch, ' documents'
	] ifFalse: [
		docsToFetch > 0 ifTrue: [
			message nextPutAll: 'one document'
		].
	].
	elements > 0 ifTrue: [
		docsToFetch > 0 ifTrue: [ message nextPutAll: ', '].
		elements > 1 ifTrue: [	
			message nextPutAll: '', elements, ' resources pending)'.
		] ifFalse: [
			message nextPutAll: 'one resource pending)'
		].
	] ifFalse: [
		message nextPutAll: ' pending)'
	].
	^ message contents
! !

!QDocumentManager methodsFor: 'cache' stamp: 'bgf 12/3/2009 14:50'!
runFetchProcess
  "This is the script for the background document retrieval thread."
  [
	[true] whileTrue:[
		[
			owner exitTripped ifTrue: [ ^ self ].
			mutex ifNil: [ ^ self ].
			"Don't do this until we're somewhat logged in."
			owner session notNil ifTrue: [
				| next | 
				cacheMgr := owner session cacheManager.	"Shift from 'default' to the managed cache."
				owner flowControlFlag: false.
				next := self nextRequest.
				owner exitTripped ifTrue: [ ^ self ].
				owner session ifNil: [ ^ self ].
				next ifNotNil: [
					"Check again for presence in the cache,
					 in case a duplicate request was made while fetching."
					((next refetch) 
					    or: [(cacheMgr has: (next get:#cacheFileName)) not]) 
					  ifTrue: [
						self doOneFetch: next.
					  ] ifFalse: [
						self discount: next
					  ]].
				next := nil.

			] ifFalse: [
				(Delay forMilliseconds: 500) wait.	"Let login or reconnect complete."
			].
		] on: Error do: [ :ex |
			owner exitTripped ifTrue: [ ^ self ].
			owner flowControlFlag: false.
			QLogger logError: ex from: self.
			(Delay forSeconds: 5) wait.	"Don't degenerate into a flip-flop failure engine."
		].
	].
  ] ensure: [
	QLogger logInfo: 'Document manager loop ends.'.
  ].
! !

!QDocumentManager methodsFor: 'import-upload' stamp: 'bgf 3/14/2009 01:05'!
serverFileExists: fullPath 
	"Return true if the given fully-pathed file is present on the service-provider."
	| entries |
	[
		entries := owner serviceMgr listFiles: fullPath.
	] on: Error do: [
		"Assume a listing error means there is no such containing directory... so no such file."
		entries := #()
	].

	^ entries isEmpty not! !

!QDocumentManager methodsFor: 'import-upload' stamp: 'bgf 4/14/2010 17:50'!
serviceFileExists: filePath 
	"Return true if the given file name is present in './' in the service storage of the current service."

	^ self serverFileExists: '/services/', (owner currentForumName), '/', filePath
	! !

!QDocumentManager methodsFor: 'queue' stamp: 'bgf 12/3/2009 15:06'!
startFetchProcessIfNeeded

	"Private.  This done when fetching becomes needed. 
	 Separate method to avoid inheriting closure references to locals in the caller."
	mutex critical: [
		(fetchProcess isNil or: [fetchProcess isTerminated ]) ifTrue: [
			fetchProcess := [ self runFetchProcess ] newProcess.
			fetchProcess name: 'Document Fetch Loop'.
			fetchProcess priority: (Processor userBackgroundPriority+5).
			fetchProcess resume.
		]].
! !

!QDocumentManager methodsFor: 'queue' stamp: 'bgf 8/30/2009 13:36'!
unqueueSHA: sha forThumb: thumbRef
	"Not yet implemented... forget any queued request to fetch this item.
	 This is just here to log the possible activity... for information.
	 We'd have to sort out how to deal with the requester callbacks when canceling, etc."
	| request |
	request := (queue detect:[:req| req cacheFileName = sha] ifNone: [nil]).
	request ifNotNil: [ 
		QLogger logInfo: 'Content manager would unqueue ', (request downloadUrl).
	].! !

!QDocumentManager methodsFor: 'private' stamp: 'ar 4/16/2011 17:11'!
uploadData: bytesOrFile sha: sha fullPath: fullServicePath sharedName: sharedName friendlyName: friendly
	"Put the given data on the service manager using the given path.
	This will replace any existing file of the same name.
	A QSharedDocument for the resulting server-side file is returned.
	nil is returned if the upload does not succeed.
	Interactive - issues #inform calls if there are errors to report."
	| mgr response v frName |
	mgr := owner serviceMgr.
	[
		owner flowControlFlag: true.
		bytesOrFile size>10000 ifTrue:[ "only show progress bar for larger files"
			"strip Python [...] prefix from friendly name"
			frName := friendly.
			(friendly findString: '%5B')>=1 ifTrue:[
				v := friendly findString: '%5D'.
				v >=1 ifTrue:[ frName := friendly copyFrom: v+3 to: friendly size ].
			].
			self displayProgressTitled:
				'Uploading ', frName, ' (', (bytesOrFile size // 1024), ' kb) ...'
				in: [:bar | response := mgr
					put: bytesOrFile
					name: fullServicePath
					informing: bar].
		]ifFalse:[
			response := mgr
				put: bytesOrFile
				name: fullServicePath
				informing: nil
		].
	] ensure: [
		owner flowControlFlag: false.
	].
	response name = #ok ifFalse: [
		self inform: ('Upload to OpenQwaq host did not complete.\\' withCRs),
			(response
				ifNil: [ 'No response from server.\Possible network problem.' withCRs]
				ifNotNil: [ response contentString ]).
		^ nil
	] ifTrue: [
		| klass |
		klass := (QDocumentType forSharedName: sharedName) docClass.
		^ (klass new)
			sha: sha;
			sharedName: sharedName;
			ctime: (self newTimeStamp: 'Import');
			suffix: (sharedName copyAfterLast:$.);
			yourself.
	].! !

!QDocumentManager methodsFor: 'import-upload' stamp: 'bgf 4/14/2010 17:50'!
uploadDocumentData: bytesOrFile shortName: fileName collisions: aPolicySymbol owner: bOwner
	"Import the given user file as a shared document of the same basic name.
	 Behavior when it already exists on the server depends on the policySymbol:
		#clobber 		- overwrites
		#makeUnique	- generates a numbered variant.
		#leaveOriginal 	- does nothing; returns Document referring to the original.
	 A QSharedDocument for the resulting server-side file is returned.
	 Nil is returned if the upload does not succeed.
	 Interactive - issues #inform calls if there are errors to report.
	 The file goes in
		 <realm>/services/<forum>/sharedDocuments/<me>/aShortName-<unique>
	 (Built up using the ream/service path, 
	  and the sharedDocuments path for this user,
	  and then shortName is the basis of the uploaded name.)"

	| actualName doc servicePath sharedArea |

	servicePath :=  '/services/', owner currentForumName.
	bOwner ifTrue:[
		sharedArea := 'sharedDocuments/', (owner simplerLogin).
	]ifFalse:[
		sharedArea := 'sharedDocuments'.
	].
	actualName := self checkExistence: fileName 
				  		path: (servicePath, '/', sharedArea) 
				  		policy: aPolicySymbol.
	actualName ifNotNil: [
		| sha notifier |
		notifier := nil.
		"If the file is large, it may take a while to compute the SHA.  If this is the
		 case, pop up a notifier so that the user doesn't wonder what's happening."
		(bytesOrFile size > 10000000 "10MB") ifTrue: [
			notifier := CUserDialog new.
			notifier label: 'Preparing document for upload'.
			notifier message: 'This may take a few seconds...'.
			notifier addButton: 'OK' value: true.
			notifier messageText 
						allowEditing: false;
						hResizing: #shrinkWrap; 
						vResizing: #shrinkWrap.
			notifier openCenteredAt: owner bounds center.
			notifier waitTick; waitTick; waitTick.
		].
		sha := TContentManager generateSHA: bytesOrFile.
		notifier ifNotNil: [notifier close].
		doc := self uploadData: bytesOrFile
			sha: sha
			fullPath: (servicePath, '/', sharedArea, '/', actualName)
			sharedName: (sharedArea, '/', actualName)
			friendlyName: (fileName).
	].

	(doc notNil and: [ doc docType isCachedLocally] ) ifTrue: [
		"Push it directly into the download cache, to avoid re-download."
		cacheMgr cache: bytesOrFile to: (doc cacheFileName).
		doc cacheLocally: true.
	].
	^ doc! !

!QDocumentManager methodsFor: 'import-upload' stamp: 'das 7/30/2009 15:40'!
uploadDocumentFile: aFile collisions: aPolicySym
	"See uploadDocumentData:.
	 This is shorthand when starting from an open File."

	| fileName doc |
	fileName := aFile localName.
	[
		aFile binary.
		doc := self 
				uploadDocumentData: aFile 
				shortName: fileName 
				collisions: aPolicySym
				owner: true.
		^doc
	] ensure: [ aFile close ].

	! !

!QDocumentManager methodsFor: 'import-upload' stamp: 'bgf 4/14/2010 17:50'!
uploadResource: data shortName: shortName sha: shaOrName
	"Import the given data as a server-side resource file for the current forum.
	 Interactive - calls #inform if there are errors to report.
	 	data 	      - the resource bits to upload
		shortName  - source file name including suffix,
	 	shaOrName - the sha if known.  Nil to compute it here.
		 <<Optionally if shaOrName =  shortName, forces 
		      use of shortName in lieue of the sha-generated name.  
		      This is so we can pre-compute the sha of pickled meshes before zipping,
		      preserving identity regardless of the zip internal metadata. >>
	 Normally, resource files are given the name 
		sha.suffix
	 This method does not re-upload if the final name it computes
 	 is already present on the server. (In that case we trust the sha.)
	 Returns the sharedName for the resulting  resource.
	 nil is returned if the upload does not succeed.
	 The file goes in 
		<realm>/services/ <forum> /resources/  < (sha.suffix) or (aShortName) > "

	| servicePath sharedArea sha suffix sharedName |

	servicePath :=  '/services/', owner currentForumName.
	sharedArea := 'resources'.
	sha := shaOrName. 
	shaOrName = shortName ifTrue: [
		"Don't do  renaming ... do what we're told."
		(shortName includes: $/) ifFalse: [
			"If told to use a literal path, just do it.  
			 Happens when re-importing objects
			 that had old-style named pictures on them."
			sharedName := (sharedArea, '/', shortName).
		] ifTrue: [
			sharedName := shortName
		].
		(self serviceFileExists: sharedName) ifTrue: [ ^ sharedName ].
		sha ifNil: [ sha := TContentManager generateSHA: data ].

	] ifFalse: [
		"Normal case... upload using the sha."
		sha ifNil: [ sha := TContentManager generateSHA: data ].
		suffix := (shortName copyAfterLast: $.).
		suffix isEmpty ifFalse: [ suffix := '.', suffix ].
		sharedName := (sharedArea, '/',  sha, suffix).
		(self serviceFileExists: sharedName) ifTrue: [  ^ sharedName ].
	].
	QLogger logInfo: 'Uploading ', shortName, ' as ', sharedName.
	^ (self uploadData: data
			sha: sha
			fullPath: (servicePath, '/', sharedName)
			sharedName: sharedName
			friendlyName: shortName)
		ifNotNilDo: [ : doc | doc sharedName ]
		ifNil: [ nil ]
! !

!QDocumentManager methodsFor: 'user interface' stamp: 'bf 11/20/2008 13:24'!
useUploadProgress: progressBar during: workBlock
	"While workBlock is executing, report progress into progressBar"

	| previousProgress |
	previousProgress := uploadProgress.
	uploadProgress := progressBar.
	[workBlock value] ensure: [uploadProgress := previousProgress]! !

Object subclass: #QDocumentRequest
	instanceVariableNames: 'myEventMap downloadUrl cacheFileName serviceManagerPath requestType requestor refetch fetchedData isDone isNeeded'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Objects-Document'!

!QDocumentRequest commentStamp: '<historical>' prior: 0!
QDocumentRequest represents a request for downloading a shared document.

Instance variables:
	downloadUrl		<String> The download url for this request.
	cacheFileName	<String> Local cache file name.!

!QDocumentRequest class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QDocumentRequest methodsFor: 'comparing' stamp: 'ar 6/27/2007 14:49'!
= aDocRequest
	"Compare by URL"
	^aDocRequest class == self class 
		and:[self downloadUrl = aDocRequest downloadUrl
		"Also test for the cacheFileName because doc requests only create one file only"
		and:[self cacheFileName = aDocRequest cacheFileName]].! !

!QDocumentRequest methodsFor: 'accessing' stamp: 'ar 4/30/2007 17:22'!
cacheFileName
	"Answer the cache file name for this request"
	^cacheFileName! !

!QDocumentRequest methodsFor: 'accessing' stamp: 'ar 4/30/2007 17:22'!
cacheFileName: aString
	"Set the cache file name for this request"
	cacheFileName := aString.! !

!QDocumentRequest methodsFor: 'accessing' stamp: 'ar 4/30/2007 17:22'!
downloadUrl
	"Answer the download url for this request"
	^downloadUrl! !

!QDocumentRequest methodsFor: 'accessing' stamp: 'ar 4/30/2007 17:22'!
downloadUrl: aString
	"Set the download url for this request"
	downloadUrl := aString! !

!QDocumentRequest methodsFor: 'progress' stamp: 'bgf 3/16/2008 21:50'!
fetchComplete: data

	"Notify interested parties that we completed a fetch;
	 the signal receives the data array as argument.
	 (The data may be nil.)"

	fetchedData := data.
	isDone := true.
	self signal: #fetchComplete with: data.
! !

!QDocumentRequest methodsFor: 'progress' stamp: 'ar 5/28/2007 13:29'!
fetchProgress: amount
	"Notify interested parties that we made progress in the download (0 .. 1)"
	^self signal: #fetchProgress with: amount.
! !

!QDocumentRequest methodsFor: 'accessing' stamp: 'bgf 6/12/2008 16:50'!
fetchedData
	"ByteArray - Once the request is done, this is the data that was fetched.
	 If it is nil , the resource was not succesfully retrieved."
	^ fetchedData! !

!QDocumentRequest methodsFor: 'comparing' stamp: 'ar 4/30/2007 17:35'!
hash
	"Hash is implement because #= is implemented"
	^downloadUrl hash! !

!QDocumentRequest methodsFor: 'initialize-release' stamp: 'bgf 3/16/2008 21:49'!
initialize
	refetch := false.
	isDone := false.! !

!QDocumentRequest methodsFor: 'accessing' stamp: 'bgf 1/29/2008 15:31'!
isAppPreview
	"App-preview graphics - especially from older forums and inventory,
	 are frequent.  (When older forums are involved, we expect repeated checks.)
	 So we filter them out of the flow-control mechanisms and status-reports."

	^ downloadUrl 
		ifNotNil: [ downloadUrl includesSubString: QEmbeddedApp previewPath ]
		ifNil: [false].! !

!QDocumentRequest methodsFor: 'comparing' stamp: 'bgf 3/25/2009 14:43'!
isBundledContent
	"Is this request a fetch of presupplied, Qwaq-defined content?
	 Used to give bundled items (which are compact, and basic to space appearance)
	 higher priority during resource fetch."
	^ TContentManager isBundledFileName: (self downloadUrl).! !

!QDocumentRequest methodsFor: 'accessing' stamp: 'bgf 3/16/2008 22:10'!
isDone
	"Set to true when the response is received; 	 
 	 the fetchedData may or may not be non-nil, 
	 depending on whether the resource was found."

	^ isDone! !

!QDocumentRequest methodsFor: 'accessing' stamp: 'bgf 3/25/2009 14:42'!
isNeeded
	"Used by the session to mark requests that are wanted for current rendering."
	^ isNeeded == true! !

!QDocumentRequest methodsFor: 'accessing' stamp: 'bgf 3/25/2009 14:41'!
isNeeded: bool
	"Used by the session to mark requests that are wanted for current rendering."
	isNeeded := bool! !

!QDocumentRequest methodsFor: 'private' stamp: 'ar 5/28/2007 13:26'!
myEventMap
	"For #signal:ing events"
	^myEventMap! !

!QDocumentRequest methodsFor: 'private' stamp: 'ar 5/28/2007 13:27'!
myEventMap: aMap
	"For #signal:ing events"
	myEventMap := aMap! !

!QDocumentRequest methodsFor: 'comparing' stamp: 'bgf 3/25/2009 14:04'!
priority
	"For sorting the request queue."
	| prio |
	prio := requestType caseOf: {
		 [ #thumb ] -> [ self isBundledContent ifTrue: [1] ifFalse: [2] ] .
		 [ #document ] -> [ 4 ] .
		 [ #prefetch ] -> [ 7 ] .
		 [ #preview ] -> [ 8 ] .
	 } otherwise: [ 5 ].

	^ self isNeeded ifTrue: [ prio - 1 ] ifFalse: [ prio ].! !

!QDocumentRequest methodsFor: 'accessing' stamp: 'bgf 1/24/2008 19:00'!
refetch
	"If true, fetch the resource even if already present in local cache"
	^ refetch! !

!QDocumentRequest methodsFor: 'accessing' stamp: 'bgf 1/24/2008 19:00'!
refetch: yesNo
	"If true, fetch the resource even if already present in local cache"
	refetch := yesNo! !

!QDocumentRequest methodsFor: 'accessing' stamp: 'ar 5/1/2007 10:51'!
requestType
	"Answer the request type (document, texture etc)"
	^requestType! !

!QDocumentRequest methodsFor: 'accessing' stamp: 'ar 5/1/2007 10:51'!
requestType: aSymbol
	"Set the request type (document, texture etc)"
	requestType := aSymbol.! !

!QDocumentRequest methodsFor: 'accessing' stamp: 'ar 5/1/2007 10:52'!
requestor
	"Answer the requestor for this request"
	^requestor! !

!QDocumentRequest methodsFor: 'accessing' stamp: 'ar 5/1/2007 10:53'!
requestor: anObject
	"Set the requestor for this request"
	requestor := anObject! !

!QDocumentRequest methodsFor: 'accessing' stamp: 'ar 4/30/2007 17:28'!
serviceManagerPath
	"Answer the service manager path for the request"
	^serviceManagerPath! !

!QDocumentRequest methodsFor: 'accessing' stamp: 'ar 4/30/2007 17:29'!
serviceManagerPath: aString
	"Set the service manager path for the request"
	serviceManagerPath := aString! !

!QDocumentRequest methodsFor: 'accessing' stamp: 'bgf 6/12/2008 22:36'!
shortName
	"A short display name for the progress bar."
	^ self isAppPreview ifTrue: [
		'Image'
	] ifFalse: [
		 (downloadUrl copyAfterLast: $/)
	]! !

Object subclass: #QDocumentType
	instanceVariableNames: 'suffixes appFactoryClassName description isCachedLocally docClass'
	classVariableNames: 'DocumentTypes'
	poolDictionaries: ''
	category: 'Qwaq-Objects-Document'!

!QDocumentType commentStamp: '<historical>' prior: 0!
An entry in QSharedDocuments documentTypes,
defining the handling of a particular kind of document 
(a file, with an understanding of how it is used by an application.)

!

!QDocumentType class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QDocumentType class methodsFor: 'classification' stamp: 'ar 4/16/2011 17:11'!
documentTypes
	"The master table for DocumentTypes, our registry of import-able document file types.
	 (Each suffix mentioned here will mean that kind of file can be dropped into a Forum;
	 see QWindowedSession dropDocument:...)

	 Changes should be reflected, since the participant initializer calls
	 	QDocumentType initialize
	 so that updates here get reflected in DocumentTypes - the main lookup table.

	 Each entry is a list of at least three items:
	 1) The array of file suffixes that match this document type,
	 2) the class to which we send newForDocument:extent: to create a TEmbeddedApp to open the document.
	 3) and a user-visible description word for this kind of document.

	 The remaining items are optional.
	 Fourth: whether the doc is replicated to local caches; defaults to false.
	 Fifth: the name of the class of QAbstractDocument that handles this type,
			defaults to QSharedDocument..
	
	 (Suffixes	AppClass	Description 	[cached]<false>  [Doc-Class]<QSharedDocument>)"

	^#(	
		(('mpg' 'mpeg' 'mp2')	QEmbeddedMpegPlayer	'MPEG Movie' true )
		(('mp4' 'm4v')			QStreamingVideoApp 	'MPEG-4 Movie' false QStreamingVideoDocument)
		(('mri')			QRecordedSessionApp 	'Recorded Session' false QRecordedSessionDocument)
		(('txt' 'text') 	 	QEmbeddedTextApp		'Text' 	false 	QTextDocument )

		"Images don't have application-type; display-sheet handles them directly.
		 We could add a QEmbeddedImageViewerApp, but it would be a bit inefficient
	       (and add more 'start-stop' confusions.)" 
		(('png' 'jpg' 'jpeg' 'gif' 'bmp')	nil		'Image'	true		QImageDocument )

		"QSZ is a dummy suffix, allows slide-show app to be a registered docType."
		(('qsz')			QEmbeddedSlidePlayer	'Slide Show'	false 	QSlideShowDocument)

		"More dummy suffixes."
		(('qvnc')			QPersonalVNCApp	'VNC Connection'	false )
		(('qwid')			QPersonalShareApp	'Personal Share'	false )

		"(('ics')			QRemoteSunbirdApp		'iCalendar')"

		(('doc')			QRemoteOpenOfficeApp	'Word Doc')
		(('ppt')			QRemoteOpenOfficeApp	'Presentation')
		(('xls')			QRemoteOpenOfficeApp	'Spreadsheet')
		(('docx')			QRemoteOpenOfficeApp	'Word Doc (XML)')
		(('pptx')			QRemoteOpenOfficeApp	'Presentation (XML)')
		(('xlsx')			QRemoteOpenOfficeApp	'Spreadsheet (XML)')

		(('rtf')			QRemoteOpenOfficeApp	'Rich Text')

		(('sdw')			QRemoteOpenOfficeApp	'Writer')
		(('sxw')			QRemoteOpenOfficeApp	'Writer')
		(('sdd')			QRemoteOpenOfficeApp	'Presentation')
		(('sdp')			QRemoteOpenOfficeApp	'Presentation')
		(('sxi')			QRemoteOpenOfficeApp	'Presentation')
		(('sda')			QRemoteOpenOfficeApp	'Drawing')
		(('sxd')			QRemoteOpenOfficeApp	'Drawing')
		(('sdc') 			QRemoteOpenOfficeApp	'Calc')
		(('sxc') 			QRemoteOpenOfficeApp	'Calc')
		(('sds')			QRemoteOpenOfficeApp	'Chart')

		(('odc')			QRemoteOpenOfficeApp	'Chart')
		(('odg')			QRemoteOpenOfficeApp	'Drawing')
		(('odi')			QRemoteOpenOfficeApp	'Image')
		(('odp')			QRemoteOpenOfficeApp	'Presentation')
		(('ods')			QRemoteOpenOfficeApp	'Spreadsheet')
		(('odt')			QRemoteOpenOfficeApp	'Open Text')

		"Reference for open office file types:
			 http://framework.openoffice.org/documentation/mimetypes/mimetypes.html"

		(('pdf')			QRemoteFirefox			'PDF'  false)

		(('zip')			nil						'Zip Archive')
		(('qpz')			PyPanelApp				'OpenQwaq Python App' false		PyPanelDocCS)
		(('tpz')			PyPanelApp				'OpenQwaq Python App' false		PyPanelDocCS)
		(('qpd')			PyPanelApp				'OpenQwaq Python Doc' false		PyPanelDocCS)
		(('tpd')			PyPanelApp				'OpenQwaq Python Doc' false		PyPanelDocCS)

		"Sampled Sounds."
		(('wav' 'aiff' 'aif') 	nil					'Sound File'  true)
	)

	"----- Inkscape registrations...
		  brad suppressed these (Jan 2009) since Inkscape is not in our production system,
		  and especially because people try to drop SVG's thinking they're for Whiteboard...

		We list here only those formats for Inkscape that:
		   	we do not handle as an image,
			claim to have been tested on Inkscape, and 
			which do not require additional software to be installed.
		For example, .ggr can be handled if Gimp is installed, ps/eps 
		can be handled if pstoedit and Skencil are installed, and dxf can 
		be handled if dxf2svg is installed.

			(('svg' 'svgz')	QRemoteInkscapeApp	'Vector Graphics')
			(('cur' 'ico' 'xpm' 'tga' 'targa' 'pbm' 'pgm' 'pnm' 'ppm')
							QRemoteInkscapeApp	'Raster Graphics')
	----"



! !

!QDocumentType class methodsFor: 'private' stamp: 'bgf 1/29/2009 22:15'!
forDocTypeEntry: aDocTypeEntry
	"Instances are fixed to be the set built by the QSharedDocument class initializer."
	^ self basicNew forDocTypeEntry: aDocTypeEntry! !

!QDocumentType class methodsFor: 'classification' stamp: 'bgf 3/13/2009 11:49'!
forFile: aFile
	"The registered QDocumentType for the given file
	 (based on the suffix of its name)
	 the result is nil if the suffix is not recognized.
	 This adds detection of certain specialized .zips
		- we process zips of images as slide shows.
	 See also #documentTypes"

	^ self forFile: aFile suffix: (aFile name copyAfterLast: $.)! !

!QDocumentType class methodsFor: 'classification' stamp: 'bgf 3/13/2009 11:57'!
forFile: aFile suffix: suffix
	"The registered QDocumentType for the given file of the given suffix;
	 the result is nil if the suffix is not recognized.
	 This adds detection of certain specialized .zips
		- we process zips of images as slide shows.
	 See also #documentTypes"
	[
		(suffix asLowercase = 'zip') ifTrue: [
			(QSlideShowDocument isSlideShowZip: aFile) ifTrue: [ ^ self forSuffix: 'qsz' ].
		].
		^ self forSuffix: suffix
	] ensure:  [ aFile reset ]! !

!QDocumentType class methodsFor: 'private' stamp: 'bgf 3/13/2009 12:41'!
forSharedName: aNameString
	"Always non-nil, the document type for the given file or uri path string."

	| sfx | 
	sfx := (aNameString copyAfterLast: $.).

	^ (QDocumentType forSuffix: sfx)
		ifNil: [ QDocumentType forDocTypeEntry: {{sfx} . nil . 'File'} ]! !

!QDocumentType class methodsFor: 'classification' stamp: 'bgf 3/5/2009 01:34'!
forSuffix: suffix

	"The registered QDocumentType for the given file suffix;
	 the result is nil if the suffix is not recognized.
	 See also #documentTypes"

	^ DocumentTypes detect: [ :type | type suffixes includes: (suffix ifNil: ['???']) asLowercase ] ifNone: [nil].
! !

!QDocumentType class methodsFor: 'class initialization' stamp: 'ar 4/6/2009 18:40'!
initialize
	"QDocumentType initialize.
	 (Called from participant init-misc-bits, 
	 so that the table always matches the static input data.)"

	DocumentTypes := 
		self documentTypes collect: [ :entry |
			QDocumentType forDocTypeEntry: entry  ].
! !

!QDocumentType methodsFor: 'accessing' stamp: 'bgf 4/14/2010 10:30'!
appFactoryClassName
	"Name of the class that serves as factory to make a TEmbeddedApp to view this document.
	 The framework sends
			newForDocument:extent:  
	 when it needs to create a TEmbeddedApp viewing/editing this document."

	^ appFactoryClassName! !

!QDocumentType methodsFor: 'accessing' stamp: 'bgf 1/29/2009 23:54'!
description	
	"Short phrase like 'Word Document' describing the content."
	^ description! !

!QDocumentType methodsFor: 'accessing' stamp: 'bgf 3/13/2009 03:18'!
docClass
	"Return the type of QSharedDocument to represent this docType.
	 Should not return nil; the fallback is a plain QSharedDocument."

	^ docClass ifNil: [ QSharedDocument ]! !

!QDocumentType methodsFor: 'private' stamp: 'bgf 4/14/2010 10:30'!
forDocTypeEntry: array

	"Initialize from the static description array."
	suffixes := array first.
	appFactoryClassName := array second.
	description := array third.
	isCachedLocally := (array size > 3) ifTrue: [ array fourth ] ifFalse: [ false ].	
	docClass := (array size > 4) ifTrue: [ (Smalltalk at: (array fifth) ifAbsent: [ nil ] )] 
							      ifFalse: [ QSharedDocument ].! !

!QDocumentType methodsFor: 'accessing' stamp: 'bgf 1/29/2009 22:09'!
isCachedLocally
	^ isCachedLocally! !

!QDocumentType methodsFor: 'testing' stamp: 'hrs 6/16/2009 15:55'!
isImage
	^ self keySuffix = 'png'! !

!QDocumentType methodsFor: 'testing' stamp: 'jcg 2/4/2009 23:47'!
isMP4
	"Is this a MPEG-4 shared-document?"
	^ suffixes includes: 'mp4'! !

!QDocumentType methodsFor: 'testing' stamp: 'bgf 3/19/2010 14:09'!
isMicrosoftXML
	"Is this one of the importable but not saveable MS XML doc formats?
	 Open Office can read these but not write them as such."
	^ #(docx xlsx pptx) includes: (self keySuffix)! !

!QDocumentType methodsFor: 'testing' stamp: 'bgf 2/16/2010 13:56'!
isMpegTwo
	" Is this a ratty old Mpeg-2 file? "
	^ suffixes includes: 'mpg'! !

!QDocumentType methodsFor: 'testing' stamp: 'bgf 1/30/2009 01:44'!
isPDF
	^ self keySuffix = 'pdf'! !

!QDocumentType methodsFor: 'testing' stamp: 'bgf 9/11/2009 23:07'!
isQPZ
	"Is this a Python-application shared-document?"
	^ (suffixes includes: 'qpz') or:[ suffixes includes: 'tpz' ].! !

!QDocumentType methodsFor: 'testing' stamp: 'bgf 1/29/2009 23:44'!
isSampledSound
	"Is this a Sound-file document?"
	^ suffixes includes: 'wav'! !

!QDocumentType methodsFor: 'accessing' stamp: 'bgf 1/30/2009 01:24'!
keySuffix
	"The first suffix in my suffix list... can be used (for example)
	 as the identifier for building case statements about types."
	^ suffixes ifEmpty: [nil] ifNotEmpty: [suffixes first]! !

!QDocumentType methodsFor: 'customs' stamp: 'jcg 5/29/2009 16:08'!
needsFullDownloadToCreateFileCard
	"Be conservative until we know better.  In other words, error on the side of excessive downloads rather than incorrect behavior."
	docClass needsFullDownloadToCreateFileCard ifFalse: [^false].
	^self isCachedLocally! !

!QDocumentType methodsFor: 'printing' stamp: 'bgf 1/29/2009 22:22'!
printOn: aStream
	"Mostly for ease of inspection."
	aStream nextPutAll: 'DocType('.
	aStream nextPutAll: self description.
	aStream nextPutAll: ', '.
	self suffixes printOn: aStream.
	aStream nextPut: $).! !

!QDocumentType methodsFor: 'accessing' stamp: 'bgf 1/29/2009 22:08'!
suffixes
	^ suffixes! !

Object subclass: #QEventLoop2
	instanceVariableNames: 'mutex stopped process priority stopPromise internalEvents externalEvents turnTime clock lastClockTime'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Objects-Kernel'!

!QEventLoop2 commentStamp: 'jcg 4/6/2010 00:55' prior: 0!
QEventLoop2 is adapted from the MIT-licensed Squeak EventLoop and ProcessEventLoop classes, which was based on the original Hedgehog-Croquet message-loop.

QEventLoop2 is typically subclassed to add domain-specific functionality.  The core event-processing code is typically not overridden.

_______________
What does it do?

QEventLoop2 utilizes the standard #future/#future: methods in Object to schedule "eventual sends" (we use the term from the E programming language).  Instead of using the default scheduler, the EventLoop is itself responsible for scheduling and execution of these message-sends.

An important distinction is made between "internal sends" and "external sends".  External sends are those originating "outside the event-loop", and internal sends are those scheduled as a result of execution "within the event loop" (see "What does it not do?" below for discussion of what it means to be "inside" vs. "outside" of the event-loop).

Two queues are provided: a linear, thread-safe SharedQueue for external sends, and a non-thread-safe heap for internal sends; in the latter, messages are sorted based on their scheduled execution-time.  Messages are processed one "turn" at a time.  The meaning of a "turn" is defined below, and provides some useful guarantees about the order in which messages are executed.  Our turns are very similar to the turns defined by the E language, but extended to support messages scheduled to execute after a specified delay; we'll call these"deferred eventual sends" or "deferred sends".  A non-deferred send is equivalent to a deferred-send with a delay of zero.  The distinction of "deferred" vs. "non-deferred" is not applicable to external sends, which are always immediately read to execute as soon as they're enqueued.

For simplicity, we will first address the case where the internal messages are all non-deferred sends.

A turn consists of:
	1) obtain a message from the external queue (if available)
	2) obtain a timestamp for the turn (possibly computed from the external message)
	3) process internal messages before the timestamp
	4) process the message from 1), if any
	5) again, process internal messages before the timestamp
	
NOTE: this documentation is unfinished, but it's not too difficult.  If you understand the two methods #enqueueMessage: and #processTurn:atTime:, you know everything you need to about turn-processing.
	
Under our simplifying "non-deferred" assumption    XXXXX: UNFINISHED
	
Internal messages are scheduled only as the result of #future messages sent to the event-loop while a turn is being processed (see method #isInTurn).  This is the reason that step 5) is necessary... after step 3) is completed there are no internal messages to execute (we just processed them all), but the message processed in step 4) might result in additional internal messages being scheduled.       XXXXX: UNFINISHED 

 A single Process is created that loops, blocking on the external-queue until a message is available.  We don't block forever, since we still want to process deferred internal messages once their scheduled time rolls around.  The interesting method is #processTurn.

__________________
What does it not do?

EventLoop provides no thread to dequeue and execute scheduled messages.  ProcessEventLoop does, and will meet most needs.  However, if your use-case requires more precise control, you may need to implement your own subclass.  It's difficult to provide further guidance here, without knowing anything about use-cases that may arise.

#isInTurn is used to differentiate between internal and external sends.  It is therefore crucial for subclasses to implement it "sensibly".  For ProcessEventLoop, one Process loops and processes messages, so it is clear how to implement #isInTurn... if you're in that Process the answer is true, and false otherwise.


XXXXX: UNFINISHED?  What else is missing?!

!QEventLoop2 class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QEventLoop2 methodsFor: 'assertion' stamp: 'jcg 4/6/2010 15:37'!
assertIsInTurn
	| receiverClass selector |
	self isInTurn ifTrue: [^self "dandy!!"].
	receiverClass := thisContext sender receiver class.
	selector := thisContext sender selector.
	self error: 'must call ', receiverClass, '>>', selector, ' from within event-loop'.
	! !

!QEventLoop2 methodsFor: 'private' stamp: 'jcg 5/12/2010 09:35'!
clockTime
	"Subclasses may override, but most will not need to."
	| old msg |
	old := lastClockTime.
	lastClockTime := clock elapsed.
	lastClockTime >= old ifTrue: [^lastClockTime "no problemo"].

	"This is not good... it means something is wonky with the VM clock."
	[	
		"Use Eliot's VM time-logging to print out the last values reported by the VM."
		| primLog millis micros block |
		"The first time gets the log-data that we will use.  At the same time,
		 stop logging to avoid race condition that will mess up the data while
		 we're retrieving it."
		primLog := Time getClockLogsAndContinueLoggingIf: false.
		"Now that we have the data, it's safe to start logging again."
		Time getClockLogsAndContinueLoggingIf: true.

		millis := primLog second. 
		micros := primLog first.
		block := [:values :stream |
			1 to: values size by: 5 do: [:v |
				stream cr; tab; tab.
				0 to: 4 do: [:off |
					v+off <= values size ifTrue: [
						(values at: v+off) printOn: stream. 
						stream nextPutAll: '   '.
					]
				].
			].
			stream nextPut: $).
		].
		msg := String streamContents: [:str |
			str nextPutAll: 'Logging times recorded by VM'.
			str cr; tab; nextPutAll: 'Last millisecond times:     #('.
			block value: millis value: str.
			str cr; tab; nextPutAll: 'Last microsecond times:      #('.
			block value: micros value: str.
		].
		QLogger logWarning: msg.
	] on: Error do: [:err | err pass. QLogger logWarning: 'failed to log microsecond clock times'].

	"Cause a bug report to be sent.  We handle the error right here (instead of letting the event-loop hander catch it), so that we don't drop any events on the floor."	
	msg := 'stopwatch time went backward!!   (', old, 'ms to ', lastClockTime, 'ms)'.
	[self error: msg] on: Error do: [:err | self handleError: err].

	"Adjust clock time to the present, so that we can try to carry on."
	clock elapsed: (lastClockTime := old).

	^lastClockTime

! !

!QEventLoop2 methodsFor: 'testing' stamp: 'jcg 6/10/2010 18:53'!
currentlyInEventLoop
	"For compatibility with QEventLoop, which we are moving away from."
	^self isInTurn! !

!QEventLoop2 methodsFor: 'initialize' stamp: 'jcg 4/6/2010 00:56'!
defaultProcessPriority
	"The default Process priority that the event-loop will run at.  Subclasses may override, and the priority can be set later anyway."
	^Processor userSchedulingPriority! !

!QEventLoop2 methodsFor: 'initialize' stamp: 'jcg 8/3/2010 17:24'!
destroy
	"DO NOT OVERRIDE... subclasses should instead override #privateDestroy."
	| promise |
	self isInTurn ifFalse: [^self future destroy].
	self privateDestroy.
	self signal: #destroyed.
	promise := self stopEventLoop.
	
	"So we'll explode if anyone tries to use us after we're #destroyed."
	mutex := nil.

	"In case anyone wants to wait for this to finish."
	^promise! !

!QEventLoop2 methodsFor: 'private' stamp: 'jcg 4/4/2010 23:40'!
enqueueMessage: msg "an EventLoopMessage" at: deltaMSecs
	self isInTurn
		"If we're in the event-loop, schedule an internal message at the appropriate time."
		ifTrue: 
			[deltaMSecs < 0 ifTrue: [self error: 'deltaMSecs must be >= 0'].
			 msg scheduledTime: turnTime + deltaMSecs.
			 internalEvents add: msg]
		"Outside the event-loop, there is no meaningful scheduled time...
		 just enqueue the message on the external queue."
		ifFalse: 
			[deltaMSecs = 0 ifFalse: [self error: 'deltaMSecs must be 0 outside event-loop'].
			 msg scheduledTime: nil. "no meaningful value."
			 externalEvents nextPut: msg].
		! !

!QEventLoop2 methodsFor: 'private' stamp: 'jcg 4/6/2010 00:53'!
execute: msg "an EventLoopMessage or nil"
	msg ifNil: [^self].
	turnTime := msg scheduledTime.
	^msg value! !

!QEventLoop2 methodsFor: 'future messages' stamp: 'jcg 4/4/2010 14:23'!
futureDo: aSelector at: deltaMSecs args: args
	"Enqueue a message on the event-loop.  The compiler has been modified to generate this message when the #future message appears in the source code."
	self futureDo: aSelector receiver: self at: deltaMSecs args: args.
	^nil
! !

!QEventLoop2 methodsFor: 'future messages' stamp: 'jcg 6/11/2010 11:55'!
futureDo: aSelector receiver: receiver args: args
	"Called by implementors of #futureDo::at:args: who want to enqueue a message on the event-queue.  This extra indirection helps other objects enqueue messages on this event-loop."
	^ self futureDo: aSelector receiver: receiver at: 0 args: args! !

!QEventLoop2 methodsFor: 'future messages' stamp: 'jcg 4/6/2010 01:06'!
futureDo: aSelector receiver: receiver at: deltaMSecs args: args
	"Called by implementors of #futureDo::at:args: who want to enqueue a message on the event-queue.  This extra indirection helps other objects enqueue messages on this event-loop."
	| msg |
	msg := QEventLoopMessageSend
						receiver: receiver
						selector: aSelector
						arguments: args.
						
	self enqueueMessage: msg at: deltaMSecs.
		
	"The compiler has determined that the result isn't used,
	 so don't bother creating a Promise."
	^nil
	! !

!QEventLoop2 methodsFor: 'future messages' stamp: 'jcg 4/4/2010 14:23'!
futureSend: aSelector at: deltaMSecs args: args
	"Enqueue a message on the event-loop.  The compiler has been modified to generate this message when the #future message appears in the source code.  Answers a promise that will be resolved with the result of performing the message."
	^self futureSend: aSelector receiver: self at: deltaMSecs args: args.
! !

!QEventLoop2 methodsFor: 'future messages' stamp: 'jcg 6/11/2010 11:55'!
futureSend: aSelector receiver: receiver args: args
	"Called by implementors of #futureSend:at:args: who want to enqueue a message on the event-queue.  This extra indirection helps other objects enqueue messages on this event-loop."
	^self futureSend: aSelector receiver: receiver at: 0 args: args! !

!QEventLoop2 methodsFor: 'future messages' stamp: 'jcg 4/6/2010 01:06'!
futureSend: aSelector receiver: receiver at: deltaMSecs args: args
	"Called by implementors of #futureSend:at:args: who want to enqueue a message on the event-queue.  This extra indirection helps other objects enqueue messages on this event-loop."
	| msg promise |
	msg := QEventLoopMessageSend
						receiver: receiver
						selector: aSelector
						arguments: args.
	msg promise: (promise := TPromise new).
	
	self enqueueMessage: msg at: deltaMSecs.
	
	"Answer the Promise that will be resolved when the message is evaluated."
	^promise
	! !

!QEventLoop2 methodsFor: 'private' stamp: 'jcg 6/11/2010 11:17'!
handleError: err
	"Subclasses may override this default implementation, which logs the error and causes the event-loop to stop."
	QLogger logWarning: 'Event-loop error: ', err description.
	QLogger logError: err.

	stopped := true.  "so that event-loop stops... see #runEventLoop"
	
	false ifTrue: [ "for debugging during development"
		(err signalerContext stackOfSize: 50) 
			inspectWithLabel: 'EVENT-LOOP ERROR: ', err description;
			halt.
	].


	! !

!QEventLoop2 methodsFor: 'initialize' stamp: 'jcg 6/25/2010 00:59'!
initialize
	super initialize.

	mutex := Mutex new.
	priority := self defaultProcessPriority.  "can set at any time"
	"We are not started until #startEventLoop is called."
	stopPromise := TPromise new fulfill: true.

	externalEvents := SharedQueue new.
	internalEvents := QEventLoopMessageQueue new.

	clock := QStopwatch new.
	clock start.
	lastClockTime := clock elapsed. "i.e. zero"

	stopped := true.
	self startEventLoop.! !

!QEventLoop2 methodsFor: 'private' stamp: 'jcg 4/6/2010 00:46'!
internalMessageAtOrBefore: nextTurnTime
	internalEvents ifEmpty: [ ^nil ].
	^ (internalEvents first scheduledTime <= nextTurnTime)
		ifTrue: [internalEvents removeFirst]
		ifFalse: [nil]
! !

!QEventLoop2 methodsFor: 'testing' stamp: 'jcg 6/10/2010 17:38'!
isInTurn
	"Answer true if the event-loop is currently processing a turn, and false otherwise.  This method has two use-cases.  The first is relevant to task-specific code, and the second is for internal use by the event-loop framework.
		- a task-specific subclass has public methods that should only be invoked from 
		  within the event-loop.  For example, if I want MyEventLoop>>foo to be
		  synchronized by the event-loop, then the first line of #foo might read...
			self isInTurn ifFalse: [^self future foo]
		  ... or ...
			self isInTurn ifFalse: [self error: 'called #foo out-of-turn']
		- for the event-loop to decide whether a #future send should be scheduled on
		  the internal or external event-queue.  If #isInTurn is true, then messages will
		  be scheduled on the internal queue, otherwise the external queue.
	"
	^process == Processor activeProcess
! !

!QEventLoop2 methodsFor: 'logging' stamp: 'jcg 4/6/2010 14:26'!
log: aString
	"Placeholder logging method.  Subclasses may override."
	QLogger logInfo: aString
! !

!QEventLoop2 methodsFor: 'initialize' stamp: 'jcg 4/6/2010 00:59'!
priority: newPriority
	"Set the priority that the event-loop Process runs at."
	mutex critical: [
		priority := newPriority.
		process ifNotNilDo: [:proc | proc priority: priority].
	].! !

!QEventLoop2 methodsFor: 'private' stamp: 'jcg 4/22/2010 23:41'!
privateDestroy
	"Override if you need to clean up."! !

!QEventLoop2 methodsFor: 'private' stamp: 'jcg 6/11/2010 11:11'!
privateStopEventLoop
	"Override if you need to clean up."! !

!QEventLoop2 methodsFor: 'private' stamp: 'jcg 5/5/2010 12:17'!
processTurn
	| timeout nextEvent |
	
	"Figure out how long we should wait for an external-event, if there isn't one 
	 already available  Don't wait more than 100msecs, because someone might 
	 have sent #stopEventLoop, and we don't want them to wait for too long."
	timeout := internalEvents isEmpty 
		ifTrue: [100]
		ifFalse: [ (internalEvents first scheduledTime - self clockTime) min: 100 ].
	
	"Obtain an external message to process.  If we don't get one within the specified 
	 timeout, that's OK... we don't want internal/deferred messages to languish after 
	 their scheduled time has passed."	
	nextEvent := externalEvents nextTimeoutMSecs: timeout.

	"Process the turn.  Freshen the clock time because we don't know how long we 
	 were waiting for an external event to become available."
	self processTurn: nextEvent atTime: self clockTime.
		
	
	! !

!QEventLoop2 methodsFor: 'private' stamp: 'jcg 5/5/2010 10:48'!
processTurn: externalEventOrNil atTime: nextTurnTime
	"A 'turn' consists of 3 steps:
		- process all internally-scheduled events at/before the 'turnTime'		
		- process 'externalEventOrNil' (if it is not nil)
		- process all internally-scheduled events at/before the 'turnTime'."
	"IMPORTANT: subclass-implementors must ensure that at most one turn is active at a time for a given event-loop."
	
	"Update the time that will be used throughout this turn.  For example, the scheduled
	 time for any newly-created internal events will be computed by adding the event's
	 delta to this turn-time."
	(turnTime notNil and: [nextTurnTime < turnTime]) ifTrue: [
		"There is now an additional assertion in #clockTime to see if the VM clock is
		 screwing us.  If we hit this error instead of that one, something is wrong with
		 the code in this event-loop or its subclass (Josh's fault, not Eliot's)."
		self error: 'event-loop time went backward!!'
	].
	turnTime := nextTurnTime.
	
	"Process internal events.  May result in the creation of new internal events."
	[self internalMessageAtOrBefore: nextTurnTime] 
		whileNotNilDo: [:event | self execute: event].
	"Process the external event  (if any).  May result in the creation of new internal events."
	externalEventOrNil ifNotNil:
		[externalEventOrNil scheduledTime: nextTurnTime.
		 self execute: externalEventOrNil].
	"Process remaining internal events, if any were created by the external event."
	[self internalMessageAtOrBefore: nextTurnTime] 
		whileNotNilDo: [:event | self execute: event].
! !

!QEventLoop2 methodsFor: 'private' stamp: 'jcg 6/11/2010 11:07'!
runEventLoop
	"Process messages in a loop, until someone tells us to stop."
	[	[stopped] whileFalse: [  [self processTurn]
											on: Error 
											do: [:err | self handleError: err]].
	] ifCurtailed: [
		"Support debugging by auto-restarting the event-loop
		 when you click 'Abandon' in the debugger."
		QLogger logWarning: 'restarting event-loop after process was curtailed'.
		stopPromise fulfill: true.
		self stopEventLoop; startEventLoop. 
	].

	"Last-minute cleanups."
	process := nil.
	self privateStopEventLoop.
	"So interested parties can tell that we've stopped."
	stopPromise fulfill: true.

! !

!QEventLoop2 methodsFor: 'running' stamp: 'jcg 6/10/2010 18:46'!
startEventLoop
	"Set up a looping Process to process internal and external events.  If event-loop is already running, first stop the existing Process."
	mutex critical: [
		self stopEventLoop wait.
		stopPromise := TPromise new.
		process := [self perform: #runEventLoop] newProcess.
		process priority: priority.
		process name: 'Event-loop for ', self printString.
		"Here we go..."
		stopped := false.
		process resume.
	]! !

!QEventLoop2 methodsFor: 'running' stamp: 'jcg 6/10/2010 17:36'!
stopEventLoop
	"Answer a Promise that will resolve to true when the event-loop is stopped.  Finishes processing the current turn before stopping."
	mutex critical: [
		stopped := true.
		^stopPromise
	].
! !

QEventLoop2 subclass: #QNetworkElement
	instanceVariableNames: 'streamManager publications subscriptions'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Objects-Kernel'!

!QNetworkElement commentStamp: '<historical>' prior: 0!
Provides a framework for implementing client-server interactions.  This is inspired by QPhoneUser and its subclasses; they should be migrated to use this as a base-class. !

QNetworkElement subclass: #QNetworkClient
	instanceVariableNames: 'controlStream features'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Objects-Kernel'!

!QNetworkClient commentStamp: 'jcg 4/22/2010 16:15' prior: 0!
Provides a basis for creating clients in a client-server interaction.  Supports sending requests and commands up to the server.  When it connects to a server, it requests a feature-list that can be used to determine what interactions are appropriate for the versions of the server and client code.!

!QNetworkClient class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QNetworkClient class methodsFor: 'instance creation' stamp: 'jcg 8/5/2009 01:02'!
newWithControlStream: streamID mediaController: controller
	^self newWithControlStream: streamID streamManager: controller mediaPlugin! !

!QNetworkClient class methodsFor: 'instance creation' stamp: 'jcg 8/5/2009 01:01'!
newWithControlStream: streamID streamManager: streamManager
	| client |
	client := self new.
	client streamManager: streamManager.
	client future subscribeToControlStream: streamID.
	^client! !

!QNetworkClient methodsFor: 'private' stamp: 'jcg 3/16/2009 09:29'!
controlStreamDestroyed
	"May be overridden to take additional action (i.e. resubscribe)."
	self signal: #controlStreamDestroyed with: controlStream with: self
! !

!QNetworkClient methodsFor: 'accessing' stamp: 'jcg 2/16/2009 16:29'!
controlStreamID
	^controlStream ifNotNil: [controlStream streamID]! !

!QNetworkClient methodsFor: 'private' stamp: 'jcg 4/22/2010 16:17'!
handleFeatureList: node "anXMLNode"
	"This just obtains the features, if available, and logs a warning otherwise.  On success, stash the raw XML feature-description in the 'features' iVar... subclasses are free to interpret the XML as they see fit."
	(node name = #command and: [(node attributeAt: 'type') = 'features']) ifFalse: [
		QLogger logWarning: 'expected <command type="features"> node, not: ', node.
		features := nil.
		^self
	].
	features := node.! !

!QNetworkClient methodsFor: 'testing' stamp: 'jcg 8/5/2009 16:33'!
isConnected
	^controlStream
		ifNil: [false]
		ifNotNilDo: [:cs | cs isConnected]! !

!QNetworkClient methodsFor: 'requests' stamp: 'jcg 4/22/2010 16:04'!
requestFeatureList
	"Send a message upstream to obtain the list of features supported by the server."
	self currentlyInEventLoop ifFalse: [^self future requestFeatureList].
	^self sendRequest: 'features' xmlBlock: [:empty | ].
	! !

!QNetworkClient methodsFor: 'private' stamp: 'jcg 3/12/2009 15:42'!
sendCommand: cmdName xmlBlock: writerBlock
	"Send command over 'controlStream'."
	self sendCommand: cmdName via: controlStream xmlBlock: writerBlock! !

!QNetworkClient methodsFor: 'private' stamp: 'jcg 3/12/2009 17:01'!
sendRequest: cmdName xmlBlock: writerBlock
	"Generate an XML request, and send it out over the network.  Answer a promise that will be resolved when the request is answered (resolves to a parsed XML node), or when an error occurs (resolves to nil)."
	^self sendRequest: cmdName via: controlStream xmlBlock: writerBlock expectXmlResponse: true
! !

!QNetworkClient methodsFor: 'private' stamp: 'jcg 3/12/2009 17:01'!
sendRequest: cmdName xmlBlock: writerBlock expectXmlResponse: expectXml
	"Generate an XML request, and send it out over the network.  Answer a promise that will be resolved when the request is answered (resolves to nil on error).  Most of the time, we expect the response to be in XML, but sometimes we want a binary response... in that case, 'expectXml' should be false."
	^self sendRequest: cmdName via: controlStream xmlBlock: writerBlock expectXmlResponse: expectXml
! !

!QNetworkClient methodsFor: 'initialize' stamp: 'jcg 6/22/2010 19:07'!
subscribeToControlStream: streamID

	QLogger logInfo: 'subscribing to control-stream: ', streamID shortHex.

	controlStream := self 
						subscribe: streamID 
						in: QControlStreamInput 
						out: QControlStreamOutput
						commandHandler: [:dgram | self handleCommand: dgram]
						dataHandler: nil.

	"Block until we've established the control-stream."
	(controlStream connectWithTimeout: 120000) wait.

	"XXXX: Need something more general... provide a hook for subclasses to decide what to do when we have connection problems."
	controlStream isConnected ifFalse: [
		QLogger logWarning: 'failed to connect to control-stream: ', streamID shortHex.
		controlStream destroy. 
		controlStream := nil.
		^self
	].

	self future requestFeatureList whenFulfilled: [:result |
		"The command-handling machinery ensures that this occurs in the event-loop."
		self handleFeatureList: result.
	].! !

!QNetworkClient methodsFor: 'private' stamp: 'jcg 3/16/2009 09:30'!
subscriptionDestroyed: subscription
	super subscriptionDestroyed: subscription.
	subscription == controlStream ifTrue: [self controlStreamDestroyed].! !

!QNetworkElement class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QNetworkElement methodsFor: 'private' stamp: 'jcg 3/1/2009 10:44'!
destroyAllPublications
	"Make a copy of the dictionary, because #destroy will result in it being removed from the dictionary (via a signal-handler)."
	publications copy do: [:pub | pub destroy]! !

!QNetworkElement methodsFor: 'private' stamp: 'jcg 3/1/2009 10:44'!
destroyAllSubscriptions
	"Make a copy of the dictionary, because #destroy will result in it being removed from the dictionary (via a signal-handler)."
	subscriptions copy do: [:sub | sub destroy]! !

!QNetworkElement methodsFor: 'private' stamp: 'jcg 1/13/2009 23:16'!
generateCommand: cmdName xmlBlock: xmlWriterBlock
	"Generate an XML command."
	^String streamContents: 
		[:s |
			| writer |
			writer := XMLWriter on: s.
			writer 
				startTag: 'command'; 
				attribute: 'type' value: cmdName;
				endTag.
			xmlWriterBlock value: writer.
			writer endTag: 'command'.
		].! !

!QNetworkElement methodsFor: 'network' stamp: 'jcg 1/23/2009 14:26'!
handleCommand: dgram
	"Called by the control-stream (a QControlStreamInput/Output... see #handleCommand:)."
	self currentlyInEventLoop ifFalse: [
		"We expect to be in our event-loop... see eg: #subscribeTo/publishControlStream:, and the way that the command-handler is wrapped in #subscribe/publish:in:out:commandHandler:dataHandler:"
		self error: 'we should be in our event-loop'.
	].
	^self handleCommandDatagram: dgram

! !

!QNetworkElement methodsFor: 'private' stamp: 'jcg 1/22/2009 14:07'!
handleCommandDatagram: dgram
	"Called from within our event-loop.  Handles parsing and error-response; the responsibility actually processing the message is delegated to subclasses (via #handleParsedCommand:)."
	^ [self handleParsedCommand: (self parseCommandDatagram: dgram)]
			on: Error 
			do: [:err |
				"Log the error, and generate an XML request
				(this might have been a request, not a command)."
				QLogger logError: 'failed to handle command datagram: ', err description.
				self generateCommand: 'user-error' xmlBlock: [:writer |
					writer tag: 'request-failure' pcData: err description, ' (in handler)'.
				].
			]! !

!QNetworkElement methodsFor: 'subclass responsibilities' stamp: 'jcg 1/22/2009 14:06'!
handleParsedCommand: node "an XML node"
	"A command datagram was received from the control-stream and enqueued for processing on our event-loop, when it was parsed into an XML node, and passed here for processing by app-specific logic."
	self subclassResponsibility! !

!QNetworkElement methodsFor: 'initialize' stamp: 'jcg 1/21/2009 13:59'!
initialize
	super initialize.
	publications := Dictionary new.
	subscriptions := Dictionary new.! !

!QNetworkElement methodsFor: 'private' stamp: 'jcg 1/22/2009 14:43'!
parseCommandBytes: bytes
	"Parse the bytes into an XML-node, and do some minimal sanity checking."
	| doc node |
	doc := XMLDOMParser parseDocumentFrom: bytes asString readStream.
	node := doc elements first.
	node name == #command ifFalse: 
		[self error: '#parseCommandBytes:  expected "command" tag'].
	(node attributeAt: 'type') ifNil:
		[self error: '#parseCommandBytes:  no command type specified'].
	^node! !

!QNetworkElement methodsFor: 'private' stamp: 'jcg 1/22/2009 14:40'!
parseCommandDatagram: dgram
	"We're only interested in the datagram's payload.  Parse it, and answer the resulting XML-node."
	^self parseCommandBytes: dgram data! !

!QNetworkElement methodsFor: 'initialize' stamp: 'jcg 8/23/2009 15:34'!
privateDestroy
	"Intended to be overridden by subclasses."
	self destroyAllSubscriptions.
	self destroyAllPublications.! !

!QNetworkElement methodsFor: 'private' stamp: 'jcg 3/1/2009 02:35'!
publicationDestroyed: publication
	"Subclasses may EXTEND to specify how to react when a publication is destroyed."
	publications removeKey: publication streamID.! !

!QNetworkElement methodsFor: 'publish/subscribe' stamp: 'ar 5/26/2009 17:20'!
publish: publishID in: streamInputClass out: streamOutputClass commandHandler: commandHandler dataHandler: dataHandler
	| pub |
	pub := QControlStreamPublication 
				publishID: publishID streamManager: streamManager 
				inputClass: streamInputClass outputClass: streamOutputClass.

	"Ensure that handler-blocks execute in our event-loop.  This eases the burden on implementors of network apps, since they don't need to explicitly put a #future message in their handlers.  We need the #yourself for the command-handler so that the compiler generates a 'futureSend' instead of a 'futureDo'."
	dataHandler ifNotNilDo: [:h |
		pub dataHandler: [:dgram | self future evaluate: h with: dgram].
	].
	commandHandler ifNotNilDo: [:h |
		pub commandHandler: [:dgram | (self future evaluate: h with: dgram) yourself].
	].

	"Add to dictionary, first ensuring that we don't have one with a matching ID."
	publications 
		at: pub streamID 
		ifPresent: [self error: 'already published a stream with ID: ', pub streamID hex].
	publications at: pub streamID put: pub.

	"Register for signals."
	self runScript: [:p | self future publicationDestroyed: p] when: {pub. #destroyed}.

	^pub! !

!QNetworkElement methodsFor: 'private' stamp: 'jcg 5/7/2010 12:01'!
resolvePromise: promise with: value
	"Ensure that the promise is resolved within the event-loop."
	[
		self currentlyInEventLoop ifFalse: [
			self future resolvePromise: promise with: value.
			^self "so compiler generates #futureDo:, not #futureSend:"
		].
		promise fulfill: value
	] on: Error do: [:err | 
		QLogger logWarning: 'could not resolve promise (full error-stack follows...)'.
		QLogger logError:  err.
	]! !

!QNetworkElement methodsFor: 'private' stamp: 'jcg 3/12/2009 15:47'!
sendCommand: cmdName via: pubOrSub  xmlBlock: writerBlock
	"Generate an XML request, and send it out over the network."
	| cmd |
	[
		pubOrSub ifNil: [QStreamUnavailable signal: 'pub/sub is nil'].
		cmd := self generateCommand: cmdName xmlBlock: writerBlock.
		pubOrSub sendCommand: cmd.
	] on: Error do: [:err |
		(err isKindOf: QStreamUnavailable)
			ifTrue: [QLogger logWarning: 'no stream to send command: ', cmdName]
			ifFalse: [QLogger logError: 'failed to send command: ', cmdName, ' (', err, ')'].
	].
! !

!QNetworkElement methodsFor: 'private' stamp: 'jcg 6/16/2010 18:52'!
sendRequest: cmdName via: pubOrSub xmlBlock: writerBlock expectXmlResponse: expectXml
	"Generate an XML request, and send it out over the network.  Answer a promise that will be resolved when the request is answered (resolves to nil on error).  Most of the time, we expect the response to be in XML, but sometimes we want a binary response... in that case, 'expectXml' should be false."
	| promise cmd |
	promise := TPromise new.
	[	cmd := self generateCommand: cmdName xmlBlock: writerBlock.
		(pubOrSub sendRequest: cmd) whenFulfilled: [:resultBytes | 
			| result |
			result := expectXml
							ifTrue: [self parseCommandBytes: resultBytes]
							ifFalse: [resultBytes].
			"Ensure that the response-processing occurs in the event-loop." 
			self resolvePromise: promise with: result
		].
	] on: Error do: [:err |
		(err isKindOf: QStreamUnavailable)
			ifTrue: [QLogger logWarning: 'no stream to send request: ', cmdName]
			ifFalse: [QLogger logError: 'failed to send request: ', cmdName, ' (', err, ')'].
		promise fulfill: nil
	].
	^ promise! !

!QNetworkElement methodsFor: 'accessing' stamp: 'ar 5/26/2009 17:17'!
streamManager
	"The stream manager for publish/subscribe operations"
	^streamManager! !

!QNetworkElement methodsFor: 'accessing' stamp: 'ar 5/26/2009 17:17'!
streamManager: aMediaStreamManager
	"The stream manager for publish/subscribe operations"
	streamManager := aMediaStreamManager! !

!QNetworkElement methodsFor: 'publish/subscribe' stamp: 'ar 5/26/2009 17:21'!
subscribe: streamID in: streamInputClass out: streamOutputClass commandHandler: commandHandler dataHandler: dataHandler
	| sub |
	sub := QControlStreamSubscription
				streamID: streamID streamManager: streamManager
				inputClass: streamInputClass outputClass: streamOutputClass.

	"Ensure that handler-blocks execute in our event-loop.  This eases the burden on implementors of network apps, since they don't need to explicitly put a #future message in their handlers.  We need the #yourself for the command-handler so that the compiler generates a 'futureSend' instead of a 'futureDo'."
	dataHandler ifNotNilDo: [:h |
		sub dataHandler: [:dgram | self future evaluate: h with: dgram].
	].
	commandHandler ifNotNilDo: [:h |
		sub commandHandler: [:dgram | (self future evaluate: h with: dgram) yourself].
	].

	"Add to dictionary, first ensuring that we don't have one with a matching ID."
	subscriptions 
		at: streamID 
		ifPresent: [self error: 'already subscribed to a stream with ID: ', streamID hex].
	subscriptions at: streamID put: sub.

	"Register for signals."
	self runScript: [:s | self future subscriptionDestroyed: s] when: {sub. #destroyed}.

	^sub! !

!QNetworkElement methodsFor: 'private' stamp: 'jcg 3/1/2009 02:34'!
subscriptionDestroyed: subscription
	"Subclasses may EXTEND to specify how to react when a subscription is destroyed."
	subscriptions removeKey: subscription streamID.! !

!QNetworkElement methodsFor: 'private' stamp: 'jcg 3/27/2009 14:39'!
xmlBoolean: node for: elemName
	"Answer a Boolean created from the pcData of the specified xml-element."
	^Boolean readFrom: (node contentStringAt: elemName ifAbsent: ['false'])! !

!QNetworkElement methodsFor: 'private' stamp: 'jcg 1/14/2009 01:26'!
xmlOID: node for: elemName
	"Answer a TObjectID created from the pcData of the specified xml-element."
	^TObjectID readHexFrom: (node contentStringAt: elemName)! !

QNetworkElement subclass: #QNetworkServer
	instanceVariableNames: 'controlStream controlStreamReady controlStreamTimeout'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Objects-Kernel'!

!QNetworkServer class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QNetworkServer class methodsFor: 'instance creation' stamp: 'ar 5/26/2009 17:01'!
newWithControlStreamID: publishID streamManager: streamManager
	| server |
	server := self new.
	server streamManager: streamManager.
	server future publishControlStream: publishID.
	^server! !

!QNetworkServer class methodsFor: 'instance creation' stamp: 'ar 5/26/2009 17:19'!
newWithPublishedControlStream: publishID mediaController: controller
	| server |
	server := self new.
	server streamManager: controller mediaPlugin.
	server future publishControlStream: publishID.
	^server! !

!QNetworkServer methodsFor: 'private' stamp: 'jcg 3/16/2009 09:28'!
controlStreamDestroyed
	"May be overridden to take additional action (i.e. republish)."
	self signal: #controlStreamDestroyed with: controlStream with: self
! !

!QNetworkServer methodsFor: 'initialize' stamp: 'crl 2/11/2010 15:18'!
controlStreamFilter: aBlockClosure
	controlStream ifNotNilDo: [:cs | cs stream stream closeFilter: aBlockClosure].! !

!QNetworkServer methodsFor: 'accessing' stamp: 'jcg 2/16/2009 17:19'!
controlStreamID
	^controlStream ifNotNil: [controlStream streamID]
! !

!QNetworkServer methodsFor: 'private' stamp: 'jcg 8/14/2009 00:56'!
controlStreamTimeout
	"How long should we wait before timing out the control-stream?  See #handleCloseRequest"
	^controlStreamTimeout ifNil: [120 "2 minutes"]! !

!QNetworkServer methodsFor: 'initialize' stamp: 'jcg 8/14/2009 01:05'!
controlStreamTimeout: timeout
	controlStreamTimeout := timeout asString.
	controlStream ifNotNilDo: [:cs | cs closeTimeout: self controlStreamTimeout].! !

!QNetworkServer methodsFor: 'private' stamp: 'jcg 3/1/2009 02:18'!
futureCloseRequest
	"Nobody's been connected to the server for a while.  What should we do?  Depends on how subclasses implement #handleCloseRequest."
	self future handleCloseRequest! !

!QNetworkServer methodsFor: 'private' stamp: 'jcg 3/1/2009 11:44'!
handleCloseRequest
	"By default, we destroy the server when nobody has been using it for a while.  Subclasses are encouraged to override this method if necessary."
	QLogger logInfo: 'destroying server after noticing that the control-stream closed'.
	self destroy! !

!QNetworkServer methodsFor: 'subclass responsibilities' stamp: 'jcg 4/22/2010 16:03'!
handleParsedCommand: node "an XML node"
	"The only thing that we know how to do is generate an XML list of our features.  Different types of servers may use entirely unrelated sets of features... whoever connects to this server is expected to understand what the features mean."
	| commandType |
	commandType := (node attributeAt: 'type').
	^commandType caseOf: {
		['features'] -> [
			"Answer the request."
			self generateCommand: 'features' xmlBlock: [:xml |
				self writeFeatureListingXMLOn: xml.
			].
		].
	} otherwise: [
		"Print out a more useful error than a call-stack."
		| msg |
		msg := 'unexpected command type: ', commandType.
		QLogger logError: msg details: node printString from: self.
		self generateCommand: 'user-error' xmlBlock: [:writer |
			writer tag: 'request-failure' pcData: msg
		].
	].
				! !

!QNetworkServer methodsFor: 'testing' stamp: 'jcg 8/5/2009 16:33'!
isConnected
	^controlStream
		ifNil: [false]
		ifNotNilDo: [:cs | cs isConnected]! !

!QNetworkServer methodsFor: 'private' stamp: 'jcg 3/16/2009 09:27'!
publicationDestroyed: publication
	super publicationDestroyed: publication.
	publication == controlStream ifTrue: [self controlStreamDestroyed].! !

!QNetworkServer methodsFor: 'initialize' stamp: 'jcg 11/23/2009 13:19'!
publishControlStream: publishID
	controlStream := self 
						publish: publishID 
						in: QControlStreamInput 
						out: QControlStreamOutput
						commandHandler: [:dgram | self handleCommand: dgram]
						dataHandler: nil.
	controlStream closeTimeout: self controlStreamTimeout.
	controlStream streamName: self class name, '-ControlStream'.
	self runScript: #futureCloseRequest when: {controlStream. #sendCloseRequest}.


	"Block until we've established the control-stream."
	(controlStream connectWithTimeout: 120000) wait.

	"XXXX: Need something more general... provide a hook for subclasses to decide what to do when we have connection problems."
	controlStream isConnected ifFalse: [
		controlStream destroy. 
		controlStream := nil
	].! !

!QNetworkServer methodsFor: 'private' stamp: 'jcg 3/12/2009 15:42'!
sendCommand: cmdName xmlBlock: writerBlock
	"Send command over 'controlStream'."
	self sendCommand: cmdName via: controlStream xmlBlock: writerBlock! !

!QNetworkServer methodsFor: 'private' stamp: 'jcg 4/22/2010 01:09'!
writeFeatureListingXMLOn: xml
	"By default, no features are listed.  Subclasses should override if necessary."
! !

Object subclass: #QForce
	instanceVariableNames: 'time force direction on'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Objects-Physics'!

!QForce class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QForce methodsFor: 'compute' stamp: 'das 8/9/2006 21:09'!
addForce: masses

	self compute.
	masses do:[:m | m addForce: force].
	! !

!QForce methodsFor: 'compute' stamp: 'das 8/9/2006 21:09'!
addForceNorm: masses

	self compute.
	masses do:[:m | m addForce: force norm: direction.].
	! !

!QForce methodsFor: 'compute' stamp: 'das 8/9/2006 20:53'!
compute

	! !

!QForce methodsFor: 'accessing' stamp: 'das 8/9/2006 20:51'!
direction

	^ direction.! !

!QForce methodsFor: 'accessing' stamp: 'das 8/9/2006 20:51'!
force 

	^ force.! !

!QForce methodsFor: 'accessing' stamp: 'das 8/9/2006 21:00'!
force: f 

	force := f.
	direction := force normalized.! !

!QForce methodsFor: 'accessing' stamp: 'das 8/9/2006 20:56'!
hasNorm

	^ false.! !

!QForce methodsFor: 'accessing' stamp: 'das 8/9/2006 20:52'!
initialize

	time := 0.0.
	on := false.! !

!QForce methodsFor: 'accessing' stamp: 'das 8/9/2006 20:52'!
on

	^ on.! !

!QForce methodsFor: 'accessing' stamp: 'das 8/9/2006 20:52'!
on: bool

	on := bool.! !

QForce subclass: #QWind
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Objects-Physics'!

!QWind commentStamp: 'das 8/10/2006 09:45' prior: 0!
QWind

This is a kind of QForce field that generates a vector that sweeps along a cone.!

!QWind class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QWind methodsFor: 'compute' stamp: 'bgf 2/22/2007 15:20'!
compute

	| windPeriod windPhase twoPi scl |
	scl := 2.5.
	windPeriod := 15.0.
	windPhase := 0.9.
	twoPi := 2 * Float pi.
	force :=scl * (Vector3 x: 0.5 * (( time * twoPi/(windPeriod * windPhase)) cos) 
		y: 0.5 * ((time * twoPi/windPeriod) sin)
		z: -4.0).
	direction := force normalized.

	"Varying the 'time' is an easy way of making adjacent objects
	 not appear overly synchronized.  Without this, for instance,
	 two adjacent ribbons flap identically."
	time := time + 0.04 + ((-0.005 to: 0.005) atRandom).
! !

!QWind methodsFor: 'compute' stamp: 'das 8/9/2006 21:00'!
hasNorm

	^ true.! !

Object subclass: #QForms
	instanceVariableNames: 'forms'
	classVariableNames: 'BundledFormsByFullName BundledFormsByName'
	poolDictionaries: ''
	category: 'Qwaq-Objects-Resources'!

!QForms commentStamp: '<historical>' prior: 0!
Form import manager.
Manages hardcoded Form graphics for Qwaq applications.

Also maintains the tables for looking up bundled texture forms
(see QTextures find:extension:ifAbsent:)
which are held class-side, and initialized at image-sealing time.

A per-island singleton.   Access is normally through the class methods,
which delegate to 'current' - the instance for the current island.

Forms are mostly used only in the non-replicated harness.
Use QTextures to manage texture graphics in Croquet spaces.

During initialization, directories of graphics can be imported;
this causes the directory to be scanned, and all the images it contains
to be listed by their short names... a sort of namespace import.
For example, the application startup for the participant does this:
	 QForms importDirectory: 'icons/Qwaq'
to import its custom sounds.

The UI construction code can subsequently use 
	Forms find: 'icon.png' 
to get the Form from that graphic file.
!

!QForms class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QForms class methodsFor: 'bundling-forms' stamp: 'bgf 10/23/2008 23:12'!
addFormFromManifest: xmlNode

	"Add a form record to BundledFormsTables from the xml manifest."

	| uri ext color sha extent clr record |
	uri := xmlNode attributeAt: 'uri'.
	uri ifNil: [ ^ QLogger logWarning: 'Disregarding form manifest entry: no uri.'].
	ext := xmlNode attributeAt: 'extent'.
	ext ifNil: [ ^ QLogger logWarning: 'Disregarding form manifest entry: no extent.'].
	clr := xmlNode attributeAt: 'color'.
	clr ifNil: [ ^ QLogger logWarning: 'Disregarding form manifest entry: no color.'].
	sha := xmlNode attributeAt: 'sha'.
	sha ifNil: [ ^ QLogger logWarning: 'Disregarding form manifest entry: no sha.'].

	[ 
		| stream | stream := (ReadStream on: ext).
		extent := (Integer readFrom: stream) @ (Integer readFrom: (stream next; yourself)).
	] on: Error do: [
		^ QLogger logWarning: 'Disregarding form manifest entry: extent is not a point'.
	].
	color := Color fromHtmlColorRef: clr.

	record := { uri . extent . sha . color }.
	BundledFormsByName at: (uri copyAfterLast: $/) put: record.
	BundledFormsByFullName at: uri put: record.
! !

!QForms class methodsFor: 'class variables' stamp: 'bgf 10/19/2008 21:32'!
bundledFormsByFullName
	"The dictionary of form info records keyed by the full-uri of the bundled Form
	 (the full URI is the one starting 'Content/') "

	^ BundledFormsByFullName! !

!QForms class methodsFor: 'bundling-forms' stamp: 'bgf 10/23/2008 22:57'!
captureBundledImagesFrom: dirName

	"Walk the named directory and its subdirectories, finding all available texture images,
	 and preload their information into the BundledForm tables.
	 This is done once during sealed-image creation, to make a standalone image."

	(FileDirectory default directoryNamed: (FileDirectory pathFromURI: dirName) )
	 withAllSubdirectoriesCollect: [ : dir |
		(dir fullName includesSubString: '.svn') ifFalse: [
			self captureBundledImagesFromDir: dir 
		].
	 ].
		! !

!QForms class methodsFor: 'bundling-forms' stamp: 'bgf 10/23/2008 22:57'!
captureBundledImagesFromDir: directory

	"(Worker method - see generateBundledFormTable)
	 For each file in the given directory, 
	 if it is an image file, record its description in the BundledForms... tables."
	
	| frontSize defName image record f fPath |

	frontSize := FileDirectory default fullName size + 1.

	directory fileNames do: [ : n | 
		defName := (directory fullName), (directory slash), n.
		[ 
			image := ImageReadWriter formFromFileNamed: defName.
			fPath := (defName allButFirst: frontSize).
			record := { (fPath replaceAll: directory pathNameDelimiter with: $/). 
						(image extent) .  
						((SecureHashAlgorithm new 
							hashStream: (f := directory readOnlyFileNamed: n) binary) radix:36) .
						TTexture averageColor: image }.
			f close.
			BundledFormsByName at: n put: record.
			BundledFormsByFullName at: record first put: record.
		] on: Error do: [ :ex | ]  
	].
! !

!QForms class methodsFor: 'accessing' stamp: 'bgf 3/4/2009 18:29'!
cleanForm: form
	"Force texture transparent parents into the pure-empty flavor required by warp-blt."
	| clean |
	clean := Form extent: form extent depth: 32.
	form displayOn: clean at: 0@0 rule: Form blend.
	^ clean! !

!QForms class methodsFor: 'instance creation' stamp: 'bgf 1/18/2007 12:30'!
clear

	"Discard the texture manager for the current island."

	| island |

	island := Processor activeIsland.
	island at: #QFormsCurrent put: nil.
! !

!QForms class methodsFor: 'instance creation' stamp: 'bgf 8/21/2008 13:59'!
current

	"Return the texture manager for the current island."

	| island instance |

	island := Processor activeIsland.
	instance := island at: #QFormsCurrent ifAbsent: [nil].
	instance ifNil: [
		QLogger logInfo: 'Creating QForms for ', island.
		instance := self new.	"Initialize puts the instance in the island."
	  ].
	^instance! !

!QForms class methodsFor: 'accessing' stamp: 'bgf 1/18/2007 11:45'!
find: fileNameOrURI

	"See the instance method for details;
	 this is the most common sound lookup method."
	^ self current find: fileNameOrURI! !

!QForms class methodsFor: 'class variables' stamp: 'bgf 10/19/2008 23:14'!
findBundledInfo: imgFileName

	"Look up the texture information for a bundled image;
	 the result is a 3-element array 
		{ 'fileURI (string)'  .  extent(Point) .  SHA (string)  . averageColor }
	"
	^ BundledFormsByFullName at: imgFileName 
		ifAbsent: [
			BundledFormsByName at: imgFileName 
				ifAbsent: [ nil ]].! !

!QForms class methodsFor: 'bundling-forms' stamp: 'bgf 1/1/2009 12:27'!
generateBundledFormTable

	" QForms generateBundledFormTable;
			   writeBundledFormsTable
	  "

	"Call this to process the Content directory tree, 
	 reading all its textures and recording them as bundled forms entries.

	 The ./Content/ subdirectories of the application workarea
	 must be populated before this is called.

	 This would normally be followed by 
		QForms writeBundledFormsTable 
	 which outputs the processed result to the xml manifest for inclusion in the product."

	BundledFormsByName := Dictionary new.
	BundledFormsByFullName := Dictionary new.

	self captureBundledImagesFrom: 'Content/Textures'.
	self captureBundledImagesFrom: 'Content/Miramar'.
	self captureBundledImagesFrom: 'Content/Widgets/textures'.
	self captureBundledImagesFrom: 'Content/Qwaq/Environment'.

	self captureBundledImagesFrom: 'Content/Qwaq/Models'.
	self captureBundledImagesFrom: 'Content/Qwaq/Textures'.
	self captureBundledImagesFrom: 'Content/Qwaq/TexturePalette/'.

	QLogger logInfo: 'Captured information for ', (BundledFormsByName size), ' textures.'! !

!QForms class methodsFor: 'accessing' stamp: 'bgf 1/18/2007 11:45'!
importDirectory:aDirectoryURI

	"Usually called at application startup and/or TIsland initialization time;
	 add the given directory to the set ofsound-fie directories that are 
	 imported.  The sounds from these directories are subsequently available
	 using QSounds #find: and its variants, using just the file name (no path)
	 or using the full URI relative to the default directory."

	self current importDirectory: aDirectoryURI! !

!QForms class methodsFor: 'accessing' stamp: 'bgf 10/20/2008 12:01'!
initialize

	"Dictionarys for bundled image data.
	 These are images for which we can create TFileTextures without local disk access
	 until the content manager needs to load the image.
	 These must be populated via 
		QForms setupBundledFormTables.
	 (That is a long job, and requires fresh SVN Content, so is not done here.)"

	BundledFormsByName := Dictionary new.
	BundledFormsByFullName := Dictionary new.


! !

!QForms class methodsFor: 'qwaq logo' stamp: 'ar 4/16/2011 17:54'!
openqwaqLogoClear42
	"A 42-pixel height OpenQwaq logo with clear background."
	^ self current find: 'openqwaq-logo-clear-42.png'! !

!QForms class methodsFor: 'qwaq logo' stamp: 'ar 4/16/2011 17:54'!
openqwaqLogoClear48
	"A 48-pixel height OpenQwaq logo with clear background."
	^ self current find: 'openqwaq-logo-clear-48.png'! !

!QForms class methodsFor: 'qwaq logo' stamp: 'ar 4/16/2011 17:54'!
openqwaqLogoClearCBForm
	"A 25-hieght transparent-background OpenQwaq logo."
	^ self current find: 'openqwaq-logo-clear-25.png'! !

!QForms class methodsFor: 'qwaq logo' stamp: 'ar 4/16/2011 17:55'!
openqwaqLogoGrey42
	"A 42-pixel height Light-on-Dark OpenQwaq logo with grey background."
	^ self current find: 'openqwaq-logo-grey-42.png'! !

!QForms class methodsFor: 'bundling-forms' stamp: 'bgf 10/23/2008 22:50'!
readBundledFormsTable

	"Read the default bundled-form manifest into the bundled forms table.
	 Done at startup to prepare for using predefined textures."

	self readBundledFormsTable: 'Objects/bundled-forms.xml'.
! !

!QForms class methodsFor: 'bundling-forms' stamp: 'bgf 3/25/2009 16:16'!
readBundledFormsTable: bundledFormsURI
	| doc |
	[
		doc := XMLDOMParser parseDocumentFrom: 
			((FileDirectory default readOnlyFileNamed: 
				(FileDirectory default pathFromURI: bundledFormsURI))
					binary contentsOfEntireFile asString readStream).		"!!!! &*%^%$*&*("

		(doc elementAt: 'bundled-forms')
		  ifNotNilDo: [ : forms |
			"If we have a valid new table, discard the old."
			BundledFormsByName := Dictionary new.
			BundledFormsByFullName := Dictionary new.

			forms elementsDo: [ :node |		
				self addFormFromManifest: node
			].
			QLogger logInfo: 'Manifest describes ', (BundledFormsByName size), ' bundled forms'.
		] ifNil: [
			QLogger logWarning: 'Bundled forms manifest not found in ', bundledFormsURI.
		].
	] on: Error do: [ : ex |
		^ QLogger logError: 'Cannot read forms manifest: ', ex description.
	].
	! !

!QForms class methodsFor: 'bundling-forms' stamp: 'bgf 10/23/2008 22:33'!
writeBundledFormsTable

	" QForms writeBundledFormsTable
	  Writes ./Objects/bundled-forms.xml,  where we record all the bundled Content textures. "

	self writeBundledFormsTable: 'Objects/bundled-forms.xml'.
! !

!QForms class methodsFor: 'bundling-forms' stamp: 'bgf 10/23/2008 23:12'!
writeBundledFormsTable: fileUri

	" QForms writeBundledFormsTable: 'Objects/bundled-forms.xml'.

	  Normally done after #generateBundledFormsTable, 
	  when processing our bundled content and registering the forms for use in packaged apps.

	  Writes  the xml manifest of all the forms recorded in BundledFormsByName,
	  which gives QTextures find:  the ability to generate textures for our bundled content
	  without having the image files present on disk. "

	| out |
	out := FileDirectory default forceNewFileNamed: (FileDirectory default pathFromURI: fileUri).
	out ascii.

	out nextPutAll: '<!!-- Automatically generated by QForms writeBundledFormsTable -->'; crlf; crlf.
	out nextPutAll: '<bundled-forms>'; crlf.
	BundledFormsByFullName keys asArray sort do: [ : path |
		| record | 	
		record := BundledFormsByFullName at: path.

		out nextPutAll: '<form uri="';
			nextPutAll: path;
			nextPutAll: '" extent="';
			nextPutAll: (record second x asString);
			nextPutAll: '@'; 
			nextPutAll: (record second y asString);
			nextPutAll: '" color="';
			nextPutAll: (record fourth asHtmlColorRef);
			nextPutAll: '" sha="';
			nextPutAll: (record third);
			nextPutAll: '"/>'; 
			crlf.
	].
	out nextPutAll: '</bundled-forms>'; crlf.
	out close.

	Transcript cr; show: 'Wrote ', BundledFormsByFullName size, ' entries to ', fileUri.! !

!QForms methodsFor: 'accessing' stamp: 'bgf 3/4/2009 18:39'!
cleanForm: aForm
	^ self class cleanForm: aForm! !

!QForms methodsFor: 'accessing' stamp: 'bgf 3/4/2009 21:57'!
fileCardForm: backingName
	"Return the image for the standard backing-card texture whose name is given.
	 ( See <forums> ./icons/Qwaq/Cards/... for the names ) "

	^ self find: backingName ifAbsent: [
		(backingName = 'DocumentCard.png') ifFalse: [
			self fileCardForm: 'DocumentCard.png'.
		] ifTrue: [
			"Out of fallbacks"
				^ nil
			]].! !

!QForms methodsFor: 'accessing' stamp: 'bgf 8/21/2008 21:03'!
find: filename

	"Find a preloaded Form with the given file name;
	 this may be the simple file name (relative to some preloaded directory),
	 or the path concatenation of the file name and the preloaded directory path.

	Forms are mostly used in the non-replicated harness;
	use QTextures to manage TFileTextures for in-file graphics."

	^ forms at: filename 
			 ifAbsent: [  QLogger logWarning: 'Unknown graphic: ', filename. 
						nil ].! !

!QForms methodsFor: 'accessing' stamp: 'bgf 1/18/2007 11:49'!
find: filename ifAbsent: aBlock

	"Find an imported Form with the given graphic file name;
	 this may be the simple file name (relative to some imported directory),
	 or the path concatenation of the file name and the imported directory path."

	^forms at: filename ifAbsent: aBlock! !

!QForms methodsFor: 'initialize-release' stamp: 'bgf 1/8/2010 17:38'!
import: fileName from: aDirPath

	"Import a single named image from the given directory."
	
	| form maybeSlash relPath absPath stream |

	maybeSlash := (aDirPath endsWith: '/') ifTrue: [''] ifFalse: ['/'].

	relPath := (aDirPath, maybeSlash, fileName).
	absPath := FileDirectory default pathFromURI:relPath.

	[
		stream := (FileDirectory default readOnlyFileNamed: absPath).
		[
			form := Form fromBinaryStream: stream.
			forms at: relPath put: form.
			forms at: fileName put: form.
		] ensure: [
			stream close.
		].

	] on: Error do: [ :ex| 
		QLogger logWarning: 'Skipping non-image ', relPath, ' (', ex description, ')'.
		form := nil.
		ex return.
	].

	^ form! !

!QForms methodsFor: 'initialize-release' stamp: 'bgf 3/25/2009 16:07'!
importDirectory: aPathName
	| form maybeSlash relPath absPath stream |
	"Preload a collection of Forms from the given local directory.
	 Note that this feature is used only on the harness-side,
	 and these directory preloads are not persisted."

	maybeSlash := (aPathName endsWith: '/') ifTrue: [''] ifFalse: ['/'].

	(FileDirectory on: (FileDirectory default pathFromURI:aPathName ))
	  fileNames
         select: [ :name | ((name beginsWith: '.') or: [name endsWith: 'Thumbs.db']) not ] 
	    thenDo: [ :name |  
		 [
			relPath := (aPathName, maybeSlash, name).
			absPath := FileDirectory default pathFromURI:relPath.

			stream :=  (FileDirectory default readOnlyFileNamed: absPath)
						binary contentsOfEntireFile readStream.
			form := Form fromBinaryStream: stream.
			(aPathName includesSubString: '/Cards') ifTrue: [
				"Have to filter the photoshop-versus-squeak jeebies out of these.
	 			 But doing this to other items (such as the Qwaq logos) grays them out oddly,
				 hence the '/Cards' thingie."
				form := self cleanForm: (form).
			].
			forms at: name put: form.
			forms at: relPath put: form.

		] on: Error do: [ :ex | 
			QLogger logWarning: 'Skipping non-image ', relPath, ' (', ex description, ')'.
			ex return.
		].
	].! !

!QForms methodsFor: 'initialize-release' stamp: 'bgf 1/18/2007 12:29'!
initialize
	
	"We are a singleton-per-Island class; use QTextures #current rather than new."

	|  island |

	super initialize.
	forms := Dictionary new.

	island := Processor activeIsland.
	island at: #QFormsCurrent put: self.

	"Forms used in all our Croquet spaces."
	self import: 'farAvatar.png' from: 'icons/Qwaq'.
! !

Object subclass: #QForumSessionRecord
	instanceVariableNames: 'urlString clientVersion islandTime leaderSID flags'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Objects-User'!

!QForumSessionRecord commentStamp: '<historical>' prior: 0!
An entry in the presence manager forum-session table.

Has no behaviors; 
this is an exchange record to facilitate sharing information 
between the QPresenceManager TIsland and the client harness.!

!QForumSessionRecord class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QForumSessionRecord class methodsFor: 'islands' stamp: 'bgf 6/20/2009 10:02'!
howToPassAsArgument
	"This is a data interchange record, containing only strings and integers."
	^ #passByCopy:! !

!QForumSessionRecord methodsFor: 'accessing' stamp: 'bgf 6/20/2009 10:04'!
clientVersion
	"The forums version string of the client used to open the forum."
	^ clientVersion! !

!QForumSessionRecord methodsFor: 'accessing' stamp: 'bgf 6/20/2009 10:07'!
clientVersion: aVersionString
	"The forums version string of the client used to open the forum."
	clientVersion := aVersionString! !

!QForumSessionRecord methodsFor: 'accessing' stamp: 'bgf 8/8/2009 01:53'!
flag: bitMask
	"The value of a single flag"
	^ flags anyMask: (bitMask)! !

!QForumSessionRecord methodsFor: 'accessing' stamp: 'bgf 8/8/2009 01:54'!
flag: mask set: bool
	"Set an individual bit-flag (or several)"
	flags := (bool ifTrue: [ flags bitOr: mask ] ifFalse: [ flags bitClear: mask ])! !

!QForumSessionRecord methodsFor: 'accessing' stamp: 'bgf 6/21/2009 18:26'!
flags
	"Under client control, a word of boolean flags for the forum-session."
	^ flags ifNil: [ 0 ]! !

!QForumSessionRecord methodsFor: 'accessing' stamp: 'bgf 6/21/2009 18:26'!
flags: anInteger
	"Under client control, an 	word of boolean flags for the forum-session.
	 Uninterpreted inside the presence manager, for more version stability."
	flags = anInteger ifFalse: [
		flags := anInteger
	].
	! !

!QForumSessionRecord methodsFor: 'initialize-release' stamp: 'bgf 8/8/2009 02:07'!
initialize
	flags := 0.! !

!QForumSessionRecord methodsFor: 'accessing' stamp: 'bgf 8/8/2009 20:20'!
isPrivateTextBlocked
	"Meeting control to to block private chat is advertised here."
	^ (flags anyMask: (QPresenceManager privateTextBlockFlagMask)) ! !

!QForumSessionRecord methodsFor: 'accessing' stamp: 'bgf 8/8/2009 20:55'!
isPrivateTextBlocked: bool
	"Meeting control to to block private chat is advertised here."
	self flag: QPresenceManager privateTextBlockFlagMask set: bool! !

!QForumSessionRecord methodsFor: 'accessing' stamp: 'bgf 6/20/2009 10:40'!
islandTime
	"The island time seconds of most recent assertion that this forum is still active."
	^ islandTime! !

!QForumSessionRecord methodsFor: 'accessing' stamp: 'bgf 6/20/2009 10:41'!
islandTime: islandHeartbeatSeconds
	"The island time seconds of most recent assertion that this forum is still active."
	islandTime := islandHeartbeatSeconds! !

!QForumSessionRecord methodsFor: 'accessing' stamp: 'bgf 6/20/2009 16:17'!
leaderSID
	"If a formally led meeting-session is in progress, this is the login of its leader.
	 Non-nil values denote a conducted-meeting in progress."
	^ leaderSID! !

!QForumSessionRecord methodsFor: 'accessing' stamp: 'bgf 6/20/2009 16:16'!
leaderSID: aSessionID
	"If a formally led meeting-session is in progress, this is the session-ID of its leader.
	 Non-nil values denote a conducted-meeting in progress."
	leaderSID := aSessionID! !

!QForumSessionRecord methodsFor: 'accessing' stamp: 'bgf 6/20/2009 10:04'!
urlString
	"The canonical url of the forum I denote."
	^ urlString! !

!QForumSessionRecord methodsFor: 'accessing' stamp: 'bgf 6/20/2009 10:08'!
urlString: aForumURLString
	"The canonical url of the forum I denote."
	urlString := aForumURLString! !

Object subclass: #QJoshHacks
	instanceVariableNames: ''
	classVariableNames: 'LastNetVidID ReceiveButIgnoreVideo'
	poolDictionaries: ''
	category: 'Qwaq-Objects-Hacks'!

!QJoshHacks commentStamp: '<historical>' prior: 0!
What the class-name implies.!

!QJoshHacks class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QJoshHacks class methodsFor: 'testing' stamp: 'jcg 7/23/2009 16:29'!
enabled
	^false! !

!QJoshHacks class methodsFor: 'accessing' stamp: 'jcg 7/23/2009 18:04'!
lastNetVidID
	^LastNetVidID! !

!QJoshHacks class methodsFor: 'accessing' stamp: 'jcg 7/23/2009 18:04'!
lastNetVidID: streamID
	LastNetVidID := streamID! !

!QJoshHacks class methodsFor: 'testing' stamp: 'jcg 7/28/2010 14:05'!
receiveButIgnoreVideo
	"Answer true if NetVidClients should always receive video from the network (even if nobody is looking at the in-world video-pane), but never decode it."
	"Used for Augmentum testing... want to separate network-effects of video/pshare from the load associated with decoding it and uploading the OpenGL texture."
	^ ReceiveButIgnoreVideo == true! !

!QJoshHacks class methodsFor: 'testing' stamp: 'jcg 7/28/2010 14:05'!
receiveButIgnoreVideo: trueOrFalse
	"When set to true,  NetVidClients will always receive video from the network (even if nobody is looking at the in-world video-pane), but never decode it."
	"Used for Augmentum testing... want to separate network-effects of video/pshare from the load associated with decoding it and uploading the OpenGL texture."
	ReceiveButIgnoreVideo := trueOrFalse! !

Object subclass: #QLogEntry
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Objects-Logging'!

!QLogEntry commentStamp: '<historical>' prior: 0!
A structured log entry.!

QLogEntry subclass: #QActivityLogEntry
	instanceVariableNames: 'date host action user organization extra'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Objects-Logging'!

!QActivityLogEntry commentStamp: '<historical>' prior: 0!
Describes an entry in the activity log.!

!QActivityLogEntry class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QActivityLogEntry methodsFor: 'accessing' stamp: 'ar 8/12/2008 11:18'!
action
	"The name of the activity"
	^action! !

!QActivityLogEntry methodsFor: 'accessing' stamp: 'ar 8/12/2008 11:18'!
action: aString
	"The name of the activity"
	action := aString! !

!QActivityLogEntry methodsFor: 'converting' stamp: 'bgf 8/22/2008 16:11'!
asString
	"Convert the receiver into human readable form (for transcript use etc)"
	^String streamContents:[:s|
		s nextPut: $[.
		date asLocal printHMSOn: s.
		s nextPutAll: '] '. 
		s nextPutAll: action,', ', user,', ', organization,', ', extra.
	].! !

!QActivityLogEntry methodsFor: 'converting' stamp: 'ar 8/13/2008 00:48'!
asXml
	"Convert the receiver to XML"
	| writer |
	^String streamContents:[:s|
		writer := XMLWriter on: s.
		writer startTag: 'Activity'; endTag.
		writer 
			tag: 'Date' pcData: date asString;
			tag: 'Action' pcData: action;
			tag: 'User' pcData: user;
			tag: 'Organization' pcData: organization;
			tag: 'Host' pcData: host;
			tag: 'Extra' pcData: extra.
		writer endTag: 'Activity'.
	].! !

!QActivityLogEntry methodsFor: 'accessing' stamp: 'ar 8/12/2008 11:18'!
date
	"The UTC DateAndTime of the log entry"
	^date! !

!QActivityLogEntry methodsFor: 'accessing' stamp: 'ar 8/13/2008 00:51'!
date: aDateAndTime
	"The UTC DateAndTime of the log entry"
	date := aDateAndTime asUTC.! !

!QActivityLogEntry methodsFor: 'accessing' stamp: 'ar 8/12/2008 11:20'!
extra
	"The extra information associated with the activity"
	^extra! !

!QActivityLogEntry methodsFor: 'accessing' stamp: 'ar 8/12/2008 11:20'!
extra: aString
	"The extra information associated with the activity"
	extra := aString! !

!QActivityLogEntry methodsFor: 'initialize' stamp: 'ar 8/13/2008 00:51'!
fromXml: xmlNode
	"Initialize the receiver from XML"
	date := (xmlNode contentStringAt: 'Date') asDateAndTime asUTC.
	host := xmlNode contentStringAt: 'Host'.
	action := xmlNode contentStringAt: 'Action'.
	user := xmlNode contentStringAt: 'User'.
	organization := xmlNode contentStringAt: 'Organization'.
	extra := xmlNode contentStringAt: 'Extra'.! !

!QActivityLogEntry methodsFor: 'accessing' stamp: 'ar 8/12/2008 11:18'!
host
	"The host on which the activity occurred"
	^host! !

!QActivityLogEntry methodsFor: 'accessing' stamp: 'ar 8/12/2008 11:18'!
host: aString
	"The host on which the activity occurred"
	host := aString! !

!QActivityLogEntry methodsFor: 'testing' stamp: 'ar 8/12/2008 11:33'!
isActivityLogEntry
	^true! !

!QActivityLogEntry methodsFor: 'accessing' stamp: 'ar 8/12/2008 11:19'!
organization
	"The organization in which the activity occurred"
	^organization! !

!QActivityLogEntry methodsFor: 'accessing' stamp: 'ar 8/12/2008 11:19'!
organization: aString
	"The organization in which the activity occurred"
	organization := aString! !

!QActivityLogEntry methodsFor: 'accessing' stamp: 'ar 8/12/2008 11:19'!
user
	"The user executing the activity"
	^user! !

!QActivityLogEntry methodsFor: 'accessing' stamp: 'ar 8/12/2008 11:19'!
user: aString
	"The user executing the activity"
	user := aString! !

QLogEntry subclass: #QEventLogEntry
	instanceVariableNames: 'type date source description details'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Objects-Logging'!

!QEventLogEntry commentStamp: '<historical>' prior: 0!
Describes an entry in the event log (used for info, warnings, errors etc).!

!QEventLogEntry class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QEventLogEntry methodsFor: 'converting' stamp: 'bgf 8/22/2008 16:39'!
asString
	"Convert the receiver into human readable form (for transcript use etc).
       A tab-terminated column after the date contains markers 'W' for warning
	 and 'E!!' for error, so they stand out a bit in the client transcript."

	^String streamContents:[:s|
		s nextPut: $[.
		date asLocal printHMSOn: s.
		s nextPutAll: '] '. 
		type caseOf: {
			[#Warning] -> [ s nextPutAll: 'W' ].
			[#Error]     -> [ s nextPutAll: 'E!!' ].
			} otherwise: [ s nextPutAll: ' '  ].
		s tab.
		s nextPutAll: source.
		s nextPutAll: ': '.
		s nextPutAll: description.
		details ifNotEmpty:[s cr; nextPutAll: details].
	].! !

!QEventLogEntry methodsFor: 'converting' stamp: 'ar 8/13/2008 00:50'!
asXml
	"Convert the receiver to XML"
	| writer |
	^String streamContents:[:s|
		writer := XMLWriter on: s.
		writer startTag: type; endTag.
		writer 
			tag: 'Date' pcData: date asString;
			tag: 'Source' pcData: source;
			tag: 'Description' pcData: description;
			tag: 'Details' pcData: details.
		writer endTag: type.
	].! !

!QEventLogEntry methodsFor: 'accessing' stamp: 'ar 8/12/2008 11:22'!
date
	"The UTC DateAndTime when the event occurred"
	^date! !

!QEventLogEntry methodsFor: 'accessing' stamp: 'ar 8/13/2008 00:51'!
date: aDateAndTime
	"The UTC DateAndTime when the event occurred"
	date := aDateAndTime asUTC.! !

!QEventLogEntry methodsFor: 'accessing' stamp: 'ar 8/12/2008 11:24'!
description
	"The description of the event"
	^description! !

!QEventLogEntry methodsFor: 'accessing' stamp: 'ar 8/12/2008 11:24'!
description: aString
	"The description of the event"
	description := aString.! !

!QEventLogEntry methodsFor: 'accessing' stamp: 'ar 8/12/2008 11:37'!
details
	"Optional details assocated with the event"
	^details! !

!QEventLogEntry methodsFor: 'accessing' stamp: 'ar 8/12/2008 11:37'!
details: aString
	"Optional details assocated with the event"
	details := aString.! !

!QEventLogEntry methodsFor: 'initialize' stamp: 'ar 8/13/2008 00:51'!
fromXml: xmlNode
	"Initialize the receiver from XML"
	type := xmlNode name.
	date := (xmlNode contentStringAt: 'Date') asDateAndTime asUTC.
	source := xmlNode contentStringAt: 'Source'.
	description := xmlNode contentStringAt: 'Description'.
	details := xmlNode contentStringAt: 'Details'.! !

!QEventLogEntry methodsFor: 'testing' stamp: 'ar 8/12/2008 11:32'!
isEventLogEntry
	^true! !

!QEventLogEntry methodsFor: 'accessing' stamp: 'ar 8/12/2008 11:23'!
source
	"The source of the event"
	^source! !

!QEventLogEntry methodsFor: 'accessing' stamp: 'ar 8/12/2008 11:24'!
source: aString
	"The source of the event"
	source := aString! !

!QEventLogEntry methodsFor: 'accessing' stamp: 'ar 8/12/2008 11:23'!
type
	"The type of the event (info, error, warning etc)"
	^type! !

!QEventLogEntry methodsFor: 'accessing' stamp: 'ar 8/12/2008 11:23'!
type: aString
	"The type of the event (info, error, warning etc)"
	type := aString! !

!QLogEntry class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QLogEntry methodsFor: 'converting' stamp: 'ar 8/12/2008 12:08'!
asString
	"Convert the receiver into human readable form (for transcript use etc)"
	^self subclassResponsibility! !

!QLogEntry methodsFor: 'converting' stamp: 'ar 8/12/2008 12:09'!
asXml
	"Convert the receiver to XML"
	^self subclassResponsibility! !

!QLogEntry methodsFor: 'initialize' stamp: 'ar 8/12/2008 12:08'!
fromXml: xmlNode
	"Initialize the receiver from XML"
	^self subclassResponsibility! !

!QLogEntry methodsFor: 'testing' stamp: 'ar 8/12/2008 11:32'!
isActivityLogEntry
	^false! !

!QLogEntry methodsFor: 'testing' stamp: 'ar 8/12/2008 11:32'!
isEventLogEntry
	^false! !

Object subclass: #QMS
	instanceVariableNames: 'massesInit masses springsInit springs forceFields doCompute counter'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Objects-Physics'!

!QMS commentStamp: '<historical>' prior: 0!
QMS
Qwaq Mass Spring simulation
This class works with the QMass, QSpring, and QForce objects.
A QMass is a point mass. If its mass is set to 0.0 it is considered to be infinitely heavy. It is initialized with a mass of 1.0, which is usually pretty good.
A QSpring is a relationship between two masses. The relative initial locations of the masses to each other define the rest length of the spring - no restoring forces. There is a default stiffness of -800.

masses := OrderedCollection new.
m1 := QMass new. m1 location: 0@0@0. 
m1 mass: 0.0 "infinity"
masses add: m1.
m2 := QMass new. m2 location: 1@1@1.
m2 frame: tCube. "attach a TFrame to the mass. As the mass moves, the TFrame will follow"
masses add: m2.
springs := OrderedCollection new.
s1 := QSpring new.
s1 mass: m1 mass: m2.
springs add: s1.

qms := QMS new masses: masses springs: springs.
force := QForce new. "default is a simple gravitational force"
force on: true. "turn on the force field"
qms addForce: force.

"once this is all set up, we can run the simulation by calling:"

qms compute. 


!

!QMS class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QMS methodsFor: 'initialize' stamp: 'das 9/8/2006 17:44'!
addForce: ff

	forceFields add: ff.
	self start. "only compute if we actually have a force"! !

!QMS methodsFor: 'compute' stamp: 'bgf 8/15/2007 17:40'!
compute

	"Single computation step"
	self computeNorms.
	self computeAll.
	self computeAll.

! !

!QMS methodsFor: 'compute' stamp: 'das 6/20/2007 13:15'!
computeAll

	| energy |
	counter = 0 ifTrue:[ ^ self stop].
	forceFields size = 0 ifTrue:[counter := counter - 1].
	forceFields do:[:f | f on ifTrue:[
		f hasNorm ifTrue:[ f addForceNorm: masses ] ifFalse:[ f addForce:masses. ]. ].
	].
	energy := 0.0.
	springs do:[ :s | s compute. ]. 
	masses do:[ :m | energy := energy + m compute. ].
	energy <0.01 ifTrue:[ counter := counter - 1].
! !

!QMS methodsFor: 'compute' stamp: 'das 8/9/2006 20:19'!
computeNorms! !

!QMS methodsFor: 'accessing' stamp: 'das 8/10/2006 09:46'!
deltaTime: dt

	masses do:[ :m | m deltaTime: dt].! !

!QMS methodsFor: 'compute' stamp: 'das 8/9/2006 20:22'!
doCompute: bool

	doCompute:= bool.! !

!QMS methodsFor: 'accessing' stamp: 'das 8/9/2006 21:36'!
forceFields

	^ forceFields.! !

!QMS methodsFor: 'accessing' stamp: 'das 8/9/2006 21:36'!
masses

	^ masses.! !

!QMS methodsFor: 'initialize' stamp: 'das 8/9/2006 21:31'!
masses: m springs: s

	massesInit := m.
	springsInit := s.
	forceFields := OrderedCollection new.
	doCompute := false.
	self reset.

	! !

!QMS methodsFor: 'initialize' stamp: 'das 8/10/2006 11:38'!
removeForce: ff

	forceFields remove: ff.! !

!QMS methodsFor: 'initialize' stamp: 'das 9/8/2006 17:42'!
reset

	masses := massesInit copy.
	masses do:[:m | m reset]. "reposition the frame attached to the mass (if any)."
	springs := springsInit copy.
	counter := 0.


	! !

!QMS methodsFor: 'accessing' stamp: 'das 6/20/2007 13:04'!
restart

	counter := 2000.
	doCompute := true.! !

!QMS methodsFor: 'accessing' stamp: 'das 8/9/2006 21:36'!
springs

	^ springs.! !

!QMS methodsFor: 'stepping' stamp: 'das 6/20/2007 12:50'!
start

	counter := 2000.
	doCompute ifFalse:[
		doCompute := true.
		self step.
	].! !

!QMS methodsFor: 'stepping' stamp: 'das 6/20/2007 12:56'!
step

	doCompute ifTrue:[
		self compute.
		(self future:200) step.
	].! !

!QMS methodsFor: 'stepping' stamp: 'das 9/8/2006 17:58'!
stop

	doCompute := false.
! !

Object subclass: #QMass
	instanceVariableNames: 'location iLocation velocity mass force rForce fCount deltaTime uv normal frame'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Objects-Physics'!

!QMass commentStamp: 'das 8/10/2006 09:26' prior: 0!
This class is part of the QMS suite of objects. A QMass includes information about the mass, location, and summed forces of a mass. It also computes the masses trajectory.

A QMass with mass = 0.0 is considered to be an infinite mass.

The interesting method here is #compute.

DAS!

!QMass class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QMass methodsFor: 'accessing' stamp: 'das 7/25/2006 12:59'!
addForce: f
	
	force := force + f.
! !

!QMass methodsFor: 'accessing' stamp: 'das 8/9/2006 20:45'!
addForce: f norm: n
	
	force := force + (f*(n dot: normal)abs).
! !

!QMass methodsFor: 'compute' stamp: 'das 6/20/2007 12:45'!
compute

	| accel |
	mass ~= 0.0 ifTrue:[
		force := force+(iLocation - location *0.001).
		accel := force/mass.
		velocity := velocity + (0.90 *accel * deltaTime) *0.99. "acceleration increases velocity"
		location := location + (velocity * deltaTime). "change location"
		frame ifNotNil:[frame translation: location. ].
	].
	force := 0.0@0.0@0.0.
	^ velocity squaredLength.

! !

!QMass methodsFor: 'accessing' stamp: 'das 8/10/2006 09:23'!
deltaTime: dt

	deltaTime := dt.! !

!QMass methodsFor: 'accessing' stamp: 'das 7/26/2006 11:20'!
force

	^ force.! !

!QMass methodsFor: 'accessing' stamp: 'das 7/25/2006 12:59'!
force: f

	force := f.! !

!QMass methodsFor: 'accessing' stamp: 'das 8/18/2006 15:32'!
frame: fr

	frame:= fr.
	frame ifNotNil:[frame translation: location. ].! !

!QMass methodsFor: 'accessing' stamp: 'das 7/25/2006 13:00'!
iLocation

	^ iLocation.
! !

!QMass methodsFor: 'accessing' stamp: 'das 7/25/2006 13:00'!
iLocation: loc

	iLocation  := loc.
! !

!QMass methodsFor: 'accessing' stamp: 'das 7/25/2006 13:00'!
location

	^  location.! !

!QMass methodsFor: 'accessing' stamp: 'das 8/18/2006 15:21'!
location: loc

	location := loc.
	iLocation := loc.
	velocity := 0@0@0.
	force := 0@0@0.
	deltaTime ifNil:[deltaTime := 0.02.].
	frame ifNotNil:[frame translation: location].
	rForce := 0@0@0.
	fCount := 1.! !

!QMass methodsFor: 'accessing' stamp: 'das 7/25/2006 13:39'!
mass

	^ mass.! !

!QMass methodsFor: 'accessing' stamp: 'das 7/25/2006 13:01'!
mass: m

	mass := m.! !

!QMass methodsFor: 'accessing' stamp: 'das 7/25/2006 13:01'!
normal

	^ normal! !

!QMass methodsFor: 'accessing' stamp: 'das 7/25/2006 13:01'!
normal: n

	normal := n.! !

!QMass methodsFor: 'compute' stamp: 'das 8/9/2006 20:28'!
reset
		
	frame ifNotNil:[frame translation: location. ].
! !

!QMass methodsFor: 'compute' stamp: 'das 7/25/2006 13:02'!
restore

	| v l |
	location ~= iLocation ifTrue:[
		"move the point slowly towards the iLocation"
		v := (iLocation - location).
		l := v length.
		l <= 0.02 ifTrue:[location := iLocation]
		ifFalse:[ location := location + (0.1*v/l).].
		^ false] ifFalse:[^ true].
! !

!QMass methodsFor: 'accessing' stamp: 'das 7/25/2006 13:01'!
uv

	^ uv.! !

!QMass methodsFor: 'accessing' stamp: 'das 7/25/2006 13:01'!
uv: tuv

	uv := tuv.! !

Object subclass: #QMaterials
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Objects-Resources'!

!QMaterials commentStamp: '<historical>' prior: 0!
Materials for use in Qwaq spaces and widgets.
All my methods are class methods.
Each method returns a new instance of TMaterial;
there is no need to 'copy' the results.

This class is stateless; there is no need to instantiate it.
In future, it may be expanded to be a per-island singleton,
to allow application-specific restyling.    The class methods
will continue to be the public accessors in that case.!

!QMaterials class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QMaterials class methodsFor: 'materials' stamp: 'bgf 5/29/2009 17:11'!
abstractBodyMaterial

	"Material for abstracted avatar parts."

	^ (TMaterial new)
		ambientColor: #(0.42 0.42 0.33);
		diffuseColor: #(0.28 0.28 0.21 1.0);
		opacity: 1.0;
		yourself
! !

!QMaterials class methodsFor: 'materials' stamp: 'bgf 7/28/2010 13:27'!
aluminumMaterial

	"Material for brushed aluminum look."
	| material |
	material := TMaterial new.
	material
			ambientColor: #(0.44 0.44 0.45) asVectorColor;
			diffuseColor: #(0.24 0.23 0.26) asVectorColor;
			specularColor: #(0.42 0.43 0.45) asVectorColor;
			emissiveColor: #(0.04 0.05 0.03) asVectorColor;
			texture: (QTextures find: 'aluminum.jpg').
	^ material.
! !

!QMaterials class methodsFor: 'materials' stamp: 'bgf 1/23/2007 13:28'!
blackMaterial

	"Material for unimportant neutral dark parts."
	| material |
	material := TMaterial new.
	material
			ambientColor: #(0.05 0.05 0.04 1) asVectorColor;
			diffuseColor: #(0.04 0.04 0.04 1) asVectorColor;
			specularColor: #(0 0 0.05 1) asVectorColor.
	^ material.
! !

!QMaterials class methodsFor: 'materials' stamp: 'bgf 8/13/2010 11:51'!
brainstormBoardMaterial

	"Material for dark grayish cork with a gradient."
	| material |
	material := TMaterial new.
	material
			ambientColor: #(0.7 0.7 0.7) asVectorColor;
			diffuseColor: #(0.7 0.7 0.7) asVectorColor;
			specularColor: #(0.1 0.1 0.1) asVectorColor;
			texture: (QTextures find: 'Content/Qwaq/TexturePalette/Cork-Wall.jpg');
			objectName: 'Brainstorming Cork'.
	^ material.
! !

!QMaterials class methodsFor: 'materials' stamp: 'bgf 12/31/2006 10:56'!
clearMaterial

	"Invisible material, for things which
	 need to be rendered and pickable but not seen."

	| |
	^ (TMaterial new)
		transparency:0.0;
		yourself! !

!QMaterials class methodsFor: 'materials' stamp: 'bgf 1/21/2007 14:33'!
concreteMaterial
	"Concrete as in cement."

	^ (TMaterial new)
		texture: (QTextures find: 'concrete.gif');
		ambientColor: #(0.4 0.4 0.4  1) asVectorColor;
		specularColor: #(0.7 0.7 0.7 1) asVectorColor;
		diffuseColor: #(0.7 0.7 0.7 1) asVectorColor;
		specularColor: #(0.3 0.3 0.3 0.5) asVectorColor;
		shininess: 0.1;
		yourself.
! !

!QMaterials class methodsFor: 'materials' stamp: 'bgf 3/7/2008 21:01'!
controlDisabledMaterial

	"Material precolored to indicate an interactive control in an 'on' state (green)."

	^ (TMaterial new)
		ambientColor: (QColors controlDisabledColor/1.6);
		diffuseColor: (QColors controlDisabledColor/1.2);
		specularColor: (QColors controlDisabledColor/4);
		emissiveColor: (QColors controlDisabledColor/5);
		opacity: 0.99;
		yourself! !

!QMaterials class methodsFor: 'materials' stamp: 'bgf 3/7/2008 20:50'!
controlMaterial

	"Material precolored to indicate an interactive control."

	| |
	^ (TMaterial new)
		ambientColor: (QColors controlNormalColor)/1.5;
		diffuseColor: (QColors controlNormalColor);
		specularColor: (QColors controlNormalColor)/6;
		emissiveColor: (QColors controlNormalColor)/5;
		opacity: 0.9;
		yourself
! !

!QMaterials class methodsFor: 'materials' stamp: 'bgf 5/30/2008 23:49'!
controlOffMaterial

	"Material precolored to indicate a start-stop button when the subject is stopped
	 (so the action of the control is to start it)."

	^ (TMaterial new)
		ambientColor: (QColors controlOffColor/1.6);
		diffuseColor: (QColors controlOffColor/1.2);
		specularColor: (QColors controlOffColor/6);
		emissiveColor: (QColors controlOffColor/5);
		opacity: 0.99;
		yourself! !

!QMaterials class methodsFor: 'materials' stamp: 'bgf 3/7/2008 20:48'!
controlOnMaterial

	"Material precolored to indicate an interactive control in an 'on' state (green)."

	^ (TMaterial new)
		ambientColor: (QColors controlOnColor/1.6);
		diffuseColor: (QColors controlOnColor/1.2);
		specularColor: (QColors controlOnColor/6);
		emissiveColor: (QColors controlOnColor/5);
		opacity: 0.99;
		yourself! !

!QMaterials class methodsFor: 'materials' stamp: 'bgf 8/21/2009 00:59'!
darkGreyMaterial

	"Material for unimportant neutral black parts."
	| material |
	material := TMaterial new.
	material
			ambientColor: #(0.15 0.15 0.125 1) asVectorColor;
			diffuseColor: #(0.2 0.195 0.18 1) asVectorColor;
			specularColor: #(0.2 0.2 0.15 1) asVectorColor.
	^ material.
! !

!QMaterials class methodsFor: 'materials' stamp: 'bgf 8/4/2009 15:46'!
darkWoodMaterial 
	"A red-brown woodgrain."

	^ (TMaterial new)
		texture: (QTextures find: 'smoothWood.jpg');
		ambientColor: #(0.24 0.22 0.18  1) asVectorColor;
		diffuseColor: #(0.48 0.46 0.4 1) asVectorColor;
		specularColor: #(0.2 0.14 0.1) asVectorColor;
		shininess: 0.6;
		yourself.
! !

!QMaterials class methodsFor: 'materials' stamp: 'bgf 1/1/2009 12:39'!
expandButtonMaterial
	"Material showing the 'expand-display-button' texture (for buttons that lead to floating displays)."
	^ QMaterials fullWhiteAlphaMaterial
			opacity: 0.95;
			texture: (QTextures expandDisplayButton);
			yourself.! !

!QMaterials class methodsFor: 'materials' stamp: 'bgf 8/29/2009 16:05'!
feltMaterial
	"Red-beige felt with a slight light-on-top gradient."
	| material |
	material := TMaterial new.
	material
			ambientColor: #(0.96 0.96 0.95) asVectorColor;
			diffuseColor: #(0.64 0.6 0.64 ) asVectorColor;
			emissiveColor: #(0.1 0.1 0.1) asVectorColor;
			texture: (QTextures find: 'Content/Qwaq/Textures/noteboardBeige.jpg');	
			cullFace: false.
	^ material.
! !

!QMaterials class methodsFor: 'materials' stamp: 'bgf 8/13/2010 11:51'!
fixedDisplayMaterial
	"Material for the stylized blue-cube frame for fixed-displays 
	 (eg. in the welcome forum.) "
	^(TMaterial new) 
		ambientColor: #(0.45 0.42 0.45) asVectorColor;
		diffuseColor: #(0.95 0.90 0.96) asVectorColor;
		emissiveColor: #(0.05 0.05 0.05 1) asVectorColor;
	 	specularColor: #(0 0 0 1)  asVectorColor;
		texture: (QTextures fixedDisplayBackground);
		objectName: 'Fixed Display';
		yourself.! !

!QMaterials class methodsFor: 'materials' stamp: 'jcg 12/20/2007 02:19'!
flameMaterial 
	"An orange-brown woodgrain."

	^ (TMaterial new)
		texture: (QTextures find:'flame2.png');
		transparency: 0.3;
		emissiveColor: #(1 1 1 1) asVectorColor;
		ambientColor: #(0 0 0 0.9) asVectorColor;
		diffuseColor: #(0 0 0 0.9) asVectorColor;
		specularColor: #(0 0 0 1) asVectorColor;
		shininess: 0.5;
		yourself.	

	! !

!QMaterials class methodsFor: 'materials' stamp: 'bgf 8/29/2009 16:54'!
flatFeltMaterial
	"Red-beige felt with no gradient... can be tiled."
	| material |
	material := TMaterial new.
	material
			ambientColor: #(0.96 0.96 0.95) asVectorColor;
			diffuseColor: #(0.64 0.6 0.64 ) asVectorColor;
			emissiveColor: #(0.1 0.1 0.1) asVectorColor;
			texture: (QTextures find: 'Content/Qwaq/TexturePalette/ground002.jpg');
			cullFace: false.
	^ material.
! !

!QMaterials class methodsFor: 'materials' stamp: 'bgf 11/27/2009 21:15'!
flatGrayMaterial
	"Opaque light-affected black.."

	^ (TMaterial new)				
		ambientColor: (QColors gray25); 
		diffuseColor: (QColors gray5);
		emissiveColor: (QColors blackColor);
		yourself

	! !

!QMaterials class methodsFor: 'materials' stamp: 'bgf 11/27/2009 18:16'!
flatWhiteAlphaMaterial
	"Flat white material, with tiny amount of transparency to ensure rendering in the alpha pass."
	^ (TMaterial new)				
		ambientColor: (QColors gray65); 
		diffuseColor: (QColors whiteAlphaColor);
		emissiveColor: (QColors gray05);
		yourself
	! !

!QMaterials class methodsFor: 'materials' stamp: 'bgf 11/27/2009 18:17'!
flatWhiteMaterial
	"Opaque light-affected white.
	 Good basis for most textured environment surfaces."

	^ (TMaterial new)				
		ambientColor: (QColors gray65); 
		diffuseColor: (QColors pureWhite);
		emissiveColor: (QColors gray05);
		yourself

	! !

!QMaterials class methodsFor: 'materials' stamp: 'bgf 11/27/2009 18:18'!
fullWhiteAlphaMaterial
	"Bright white material, with a touch of emissive;
	 microscopically transparent to force render in the alpha pass."
	^(TMaterial new) 
		ambientColor: (QColors gray9);
		diffuseColor: (QColors whiteAlphaColor);
		emissiveColor: (QColors gray2);
		shininess: 0.1;
		yourself
	! !

!QMaterials class methodsFor: 'materials' stamp: 'bgf 11/27/2009 18:18'!
fullWhiteMaterial
	"Bright white opaque material, with a touch of emissive."

	^(TMaterial new) 
		ambientColor: (QColors gray9);
		diffuseColor: (QColors pureWhite);
		emissiveColor: (QColors gray2);
		shininess: 0.1;
		yourself! !

!QMaterials class methodsFor: 'materials' stamp: 'bgf 5/29/2009 13:53'!
gateFootMaterial
	"The sill plate of a q-gate"
	^ (TMaterial new)
		ambientColor: #(0.24 0.24 0.25) asVectorColor;
		specularColor: #(0.0 0.0 0.0 1) asVectorColor;
		diffuseColor: #(0.55 0.55 0.56 1) asVectorColor;
		shininess: 0.0;
		yourself.
! !

!QMaterials class methodsFor: 'materials' stamp: 'bgf 2/17/2008 19:49'!
grayIndicatorLightMaterial

	"A material with only emissive color, nominal transparency,
	 whose initial color is gray."
	| material |
	material := TMaterial new.
	material
			ambientColor: #(0 0 0 0.99) asVectorColor;
			diffuseColor: #(0 0 0 0.99) asVectorColor;
			specularColor: #(0 0 0) asVectorColor;
			emissiveColor: #(0.5 0.5 0.5) asVectorColor.
	^ material.
! !

!QMaterials class methodsFor: 'materials' stamp: 'bgf 9/5/2008 19:13'!
greyMaterial

	"Material for unimportant neutral parts."
	| material |
	material := TMaterial new.
	material
			ambientColor: #(0.35 0.33 0.3 1) asVectorColor;
			diffuseColor: #(0.45 0.43 0.4 1) asVectorColor;
			specularColor: #(0.05 0.03 0.01 1) asVectorColor.
	^ material.
! !

!QMaterials class methodsFor: 'materials' stamp: 'bgf 9/21/2007 12:36'!
groundMaterial: textureOrFileName

	"A warm texture-underlay material with warm spectral aspect
       and low ambient level."

	| material |
	material := TMaterial new.
	material ambientColor: #(0.5 0.45 0.35 1) asVectorColor.
	material diffuseColor: #(0.75 0.75 0.6 1) asVectorColor.
	material specularColor: #(0.25 0.35 0.2 1)  asVectorColor.
	material shininess: 0.08.

	(textureOrFileName isKindOf: TTexture) ifFalse: [
		material texture: (QTextures find: textureOrFileName).
	] ifTrue: [
		material texture: textureOrFileName.
	].		
	^ material! !

!QMaterials class methodsFor: 'materials' stamp: 'bgf 8/6/2009 19:47'!
lightWoodMaterial
	"Pale woodgrain material."

	^(TMaterial new)
		texture: (QTextures find:'woodCedar.jpg');
		uvAngle: 45.0;
		ambientColor: #(0.5 0.4 0.4  1) asVectorColor;
		diffuseColor: #(0.75 0.74 0.62 1) asVectorColor;
		specularColor: #(0.25 0.24 0.2 1) asVectorColor;
		shininess: 0.35;
		yourself
	! !

!QMaterials class methodsFor: 'materials' stamp: 'bgf 8/13/2010 11:53'!
matteMaterial

	"Light-insensitive material for making translucent
	 background mattes, for 3-D podiums and the like."

	^(TMaterial new) 
		opacity: 0.99;
		ambientColor: #(0 0 0 0) asVectorColor;
		diffuseColor: #(0 0 0 0.7) asVectorColor;
		emissiveColor: #(0.65 0.67 0.69 0.99) asVectorColor;
	 	specularColor: #(0 0 0 1)  asVectorColor;
		shininess: 0.0;
		objectName: 'Pedestal Matte';
		yourself
	! !

!QMaterials class methodsFor: 'materials' stamp: 'bgf 10/3/2007 17:21'!
mediumDarkGreyMaterial

	"Material for unimportant neutral black parts."
	| material |
	material := TMaterial new.
	material
			ambientColor: #(0.27 0.26 0.25 1) asVectorColor;
			diffuseColor: #(0.37 0.36 0.35 1) asVectorColor.
	^ material.
! !

!QMaterials class methodsFor: 'materials' stamp: 'bgf 8/4/2009 15:49'!
mediumDarkWoodMaterial 
	"A red-brown woodgrain."

	^ (TMaterial new)
		texture: (QTextures find: 'smoothWood.jpg');
		ambientColor: #(0.28 0.26 0.21  1) asVectorColor;
		diffuseColor: #(0.57 0.56 0.45 1) asVectorColor;
		specularColor: #(0.15 0.14 0.12) asVectorColor;
		shininess: 0.6;
		yourself.
! !

!QMaterials class methodsFor: 'materials' stamp: 'bgf 8/4/2009 15:49'!
mediumWoodMaterial 
	"An orange-brown woodgrain."

	^ (TMaterial new)
		texture: (QTextures find:'smoothWood.jpg');
		ambientColor: #(0.38 0.37 0.35  1) asVectorColor;
		diffuseColor: #(0.58 0.57 0.42 1) asVectorColor;
		specularColor: #(0.2 0.2 0.15 1) asVectorColor;
		shininess: 0.5;
		yourself.	

	! !

!QMaterials class methodsFor: 'materials' stamp: 'bgf 11/27/2009 18:19'!
menuButtonRectangleMaterial
	"Material for the rounded-rectangle menu-button."
	^ (TMaterial new)
		texture: (QTextures findAlpha: 'Content/Qwaq/Textures/MenuButton2.png');
		diffuseColor: (QColors menuButtonColor);
		fullBright: true;
		premultiplied: true;
		opacity: 0.995;
		uvScale: 0.99@0.99;
		uvOffset: (0.006@-0.006);
		yourself
! !

!QMaterials class methodsFor: 'materials' stamp: 'bgf 11/5/2007 11:11'!
mirrorGlassMaterial

	"Material for the dark cloudy-edged glass over mirrors."

	^ (TMaterial new)
		texture: QTextures portalGlass;
		transparency: 0.3;
		ambientColor: #(0.1 0.12 0.15 0.3);
		diffuseColor: #(0 0 0 0.3);
		cullFace: false;
		yourself! !

!QMaterials class methodsFor: 'rectangles' stamp: 'bgf 8/29/2009 11:36'!
newRadialShadow
	"A rectangle bearing a black circle fading outward to transparent."
	| shadow |
	shadow := TSimpleRectangle new.
	shadow
		objectName: 'shadow';
		shouldPersist: false;
		pickable: false.
	(shadow material) 	
		texture: (QTextures find: 'RadialShadow.png');
		color: (VectorColor r:1 g: 1 b: 1 a: 0.6).
	shadow
		checkAlpha.
	^ shadow
! !

!QMaterials class methodsFor: 'rectangles' stamp: 'bgf 8/29/2009 11:47'!
newRectangularShadow
	"A bordered rectangle whose border is a shadow-fade to the edge.
	(To fit a shadow under a cube, make the rectangle the same x-z extent,
	 and set its borderWidth: to control how far out the shadow extends.) "
	| shadow |
	shadow := TBorderRectangle new.
	shadow		
		objectName: 'shadow';
		shouldPersist: false;
		pickable: false.
	(shadow material)
		color: (VectorColor r: 0 g: 0 b: 0 a: 0.3).
	(shadow borderMaterial) 	
		texture: (QTextures find: 'RadialShadow.png');
		color: (VectorColor r:1 g: 1 b: 1 a:0.5).
	shadow
		checkAlpha.
	^ shadow
! !

!QMaterials class methodsFor: 'materials' stamp: 'bgf 8/13/2010 11:52'!
noteBoardMaterial

	"Material for warm mid-light cork with a gradient."
	| material |
	material := TMaterial new.
	material
			ambientColor: #(0.64 0.78 0.86) asVectorColor;
			diffuseColor: #(0.64 0.56 0.32 ) asVectorColor;
			specularColor: #(0.06 0.06 0.06) asVectorColor;
			texture: (QTextures find: 'Content/Qwaq/TexturePalette/Cork-Board.jpg');	
			objectName: 'Noteboard Cork';
			cullFace: false.
	^ material.
! !

!QMaterials class methodsFor: 'materials' stamp: 'bgf 1/21/2007 14:23'!
paperMaterial
	"Paper - with a touch of emissive color to keep it readable in bad light"

	| |
	^ (TMaterial new)
		color: (QColors paperColor);
		emissiveColor: (QColors paperColor/8.0);
		yourself! !

!QMaterials class methodsFor: 'materials' stamp: 'bgf 8/13/2010 11:53'!
placemarkMaterial

	"Material precolored to indicate an interactive control, translucent."

	^ (TMaterial new)
		ambientColor: (QColors placemarkColor/1.5);
		diffuseColor: ((QColors placemarkColor/1.5) alpha:0.8);
		emissiveColor: (QColors placemarkColor/3);
		objectName: 'Placemark';
		yourself! !

!QMaterials class methodsFor: 'materials' stamp: 'bgf 11/5/2007 11:09'!
portalGlassMaterial

	"Material for the cloudy-edged class over portals and other space openings."

	^ (TMaterial new)
		texture: QTextures portalGlass;
		transparency: 0.3;
		ambientColor: #(0.7 0.75 0.78 0.4);
		diffuseColor: #(0.5 0.54 0.55 0.3);
		cullFace: false;
		yourself! !

!QMaterials class methodsFor: 'materials' stamp: 'bgf 8/18/2009 18:13'!
qwaqGrayMaterial
	"Official Qwaq logo gray."
	^ (TMaterial new)	
		diffuseColor: (QColors qwaqGray asVectorColor);
		ambientColor: ((QColors qwaqGray * 0.667) alpha: 1); 
		yourself! !

!QMaterials class methodsFor: 'materials' stamp: 'bgf 8/18/2009 18:13'!
qwaqOrangeMaterial
	"Official Qwaq logo orange."
	^ (TMaterial new)	
		diffuseColor: (QColors qwaqOrange asVectorColor);
		ambientColor: ((QColors qwaqOrange * 0.667) alpha: 1); 
		yourself! !

!QMaterials class methodsFor: 'materials' stamp: 'bgf 8/6/2009 22:32'!
redFabricMaterial
	"The warm red fabric in the Training Center furniture."
	| material |
	material := TMaterial new.
	material
			ambientColor: #(0.33 0.13 0.01 1) asVectorColor;
			diffuseColor: #(0.58 0.25 0.04 1) asVectorColor;
			specularColor: #(0.32 0.06 0.02 1) asVectorColor.
	^ material.
! !

!QMaterials class methodsFor: 'materials' stamp: 'bgf 9/5/2008 19:07'!
resizeCornerBlackMaterial
	"Material for a lower-right-corner resizer attachment for 2-D rectangles."
	^ (TMaterial new)
		texture: (QTextures find: 'Content/Qwaq/Textures/ResizeCornerLRBlack.png');
		ambientColor: (VectorColor r: 0.7 g:0.68 b: 0.61);
		diffuseColor: (VectorColor r: 0.7 g:0.68 b: 0.61);
		opacity: 1;
		yourself
! !

!QMaterials class methodsFor: 'materials' stamp: 'bgf 8/21/2009 00:58'!
resizeCornerDPMaterial
	"Material for the rounded-rectangle DisplayPanel resizer."
	^ (TMaterial new)
		texture: (QTextures findAlpha: 'Content/Qwaq/Textures/ResizeCornerDP.png');
		ambientColor: (VectorColor r: 0.7 g:0.68 b: 0.61);
		diffuseColor: (VectorColor r: 0.5 g:0.48 b: 0.41);
		opacity: 0.995;
		yourself
! !

!QMaterials class methodsFor: 'materials' stamp: 'bgf 7/12/2009 17:47'!
resizeCornerMaterial
	"Material for a lower-right-corner resizer attachment for 2-D rectangles."
	^ (TMaterial new)
		texture: (QTextures find: 'Content/Qwaq/Textures/ResizeCornerLR.png');
		ambientColor: (Color white) asVectorColor;
		diffuseColor: (Color white) asVectorColor;
		opacity: 0.7;
		yourself
! !

!QMaterials class methodsFor: 'materials' stamp: 'bgf 8/13/2010 11:51'!
soundChipMaterial
	"The material for QSoundFrames, which are card-size lozenges when visible.
	 The texture is the familiar three-curves sound indicator, blue on pale-blue.
	 (So they're a bit like in-space icons for sounds.)"
	
	| material |
	material := TMaterial new.
	material
			ambientColor: #(0.6 0.6 0.62) asVectorColor;
			diffuseColor: #(0.95 0.95 0.97 0.99) asVectorColor;
			texture: (QTextures find: 'Sound.png');
			objectName: 'Sound Chip'.
	^ material.
! !

!QMaterials class methodsFor: 'materials' stamp: 'bgf 11/27/2009 18:05'!
translucentEditorMaterial
	"Material precolored to indicate an interactive editing control, translucent."
	| c |
	c := QColors editboxCornerColor.

	^ (TMaterial new)
		ambientColor: (c/1.3 alpha:0.7);
		diffuseColor: (c/1.2 alpha: 0.7);
		specularColor: (c/5);
		emissiveColor: (c/3);
		cullFace: false;
		yourself! !

Object subclass: #QMotion
	instanceVariableNames: 'frequency size motion translation'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Objects-Motion'!

!QMotion commentStamp: 'das 12/3/2006 23:13' prior: 0!
TMotion is a motion file designed to use with QCharacter.
motion is a list of lists of transforms. Each list of transform is applied across all of the main transform frames of the character as it changes position.
size is the total number of elements in the motion list - that is, the number of different poses the character moves through.
frequency is the time in milliseconds for each pose.!

!QMotion class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QMotion methodsFor: 'accessing' stamp: 'das 12/3/2006 17:29'!
frequency
	
	^ frequency.! !

!QMotion methodsFor: 'accessing' stamp: 'das 12/3/2006 17:29'!
motion

	^ motion.
! !

!QMotion methodsFor: 'initialize' stamp: 'das 12/3/2006 17:37'!
motion: m frequency: freq size: sz

	motion := m.
	frequency := freq.
	size := sz.! !

!QMotion methodsFor: 'initialize' stamp: 'das 2/19/2007 08:33'!
motion: m translation: t frequency: freq size: sz

	motion := m.
	translation := t.
	frequency := freq/1000.0. "convert to milliseconds"
	size := sz asInteger.! !

!QMotion methodsFor: 'accessing' stamp: 'das 12/3/2006 17:39'!
size

	^ size.! !

!QMotion methodsFor: 'accessing' stamp: 'das 2/15/2007 10:01'!
translation

	^ translation.! !

Object subclass: #QMotionCaptures
	instanceVariableNames: 'motions'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Objects-Resources'!

!QMotionCaptures commentStamp: '<historical>' prior: 0!
Per-island singleton cache of BVH mocaps.
The items in this cache are QBVHParser instances;
each holds a QMotion (the motion sequence) and a TJointRoot (the skeleton geometry.)!

!QMotionCaptures class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QMotionCaptures class methodsFor: 'instance creation' stamp: 'bgf 3/18/2007 12:20'!
clear

	"Discard the bvh cache for the current island."

	| island |

	island := Processor activeIsland.
	island at: #QMotionsCurrent put: nil.
! !

!QMotionCaptures class methodsFor: 'instance creation' stamp: 'bgf 8/21/2008 13:59'!
current

	"Return the bvh cache for the current island."

	| island instance |

	island := Processor activeIsland.
	instance := island at: #QMotionsCurrent ifAbsent: [nil].
	instance ifNil: [
		QLogger logInfo: 'Creating QMotions for ', island.
		instance := self new.	"Initialize puts the instance in the island."
	  ].
	^instance! !

!QMotionCaptures class methodsFor: 'accessing' stamp: 'bgf 3/18/2007 12:21'!
find: fileNameOrURI

	"Return a chached bvh motion, by filename or uri."
	^ self current find: fileNameOrURI! !

!QMotionCaptures class methodsFor: 'motions' stamp: 'bgf 3/18/2007 23:16'!
manIdle
	"The basic male standing-around-fidgeting loop."

	" stances.bvh"
	^ self current import: 'stances-coarse.bvh' from: 'Content/Motion/Male' atScale:0.095.! !

!QMotionCaptures class methodsFor: 'motions' stamp: 'bgf 3/18/2007 23:11'!
manStandingStill
	"Male stationary."

	" stances.bvh"
	^ self find: 'stand.bvh'! !

!QMotionCaptures class methodsFor: 'motions' stamp: 'bgf 3/18/2007 12:21'!
manWalk
	"The basic male walk loop."
	^ self find: 'male_walkLoop.bvh'! !

!QMotionCaptures methodsFor: 'accessing' stamp: 'bgf 8/21/2008 20:22'!
find: filename

	"Find a preloaded BVH with the given file name;
	 this may be the simple file name (relative to some preloaded directory),
	 or the path concatenation of the file name and the preloaded directory path."

	^ motions at: filename 
			 ifAbsent: [  QLogger logWarning: 'Unknown BVH: ', filename. 
						nil ].! !

!QMotionCaptures methodsFor: 'accessing' stamp: 'bgf 3/18/2007 12:22'!
find: filename ifAbsent: aBlock

	"Find an imported BVH capture with the given file name;
	 this may be the simple file name,
	 or the path concatenation of the file name and the image-relative directory path."

	^motions at: filename ifAbsent: aBlock! !

!QMotionCaptures methodsFor: 'initialize-release' stamp: 'bgf 8/21/2008 20:22'!
import: fileName from: aDirPath atScale: aScale

	"Import a single bvh file from the given directory."
	
	| maybeSlash relPath absPath bvh |

	maybeSlash := (aDirPath endsWith: '/') ifTrue: [''] ifFalse: ['/'].

	relPath := (aDirPath, maybeSlash, fileName).
	absPath := FileDirectory default pathFromURI:relPath.

	"Don't load motions twice."
	motions at: relPath ifPresent: [:motion| ^ motion].

	[
		bvh := BVHParser initializeWithFileName: absPath scale: aScale.

		motions at: fileName put: bvh.
		motions at: relPath put: bvh.

	] on: Error do: [ :ex| 
		QLogger logWarning: 'Skipping BVH ', relPath, ' (', ex description, ')'.
		bvh := nil.
		ex return.
	].

	^ bvh! !

!QMotionCaptures methodsFor: 'initialize-release' stamp: 'bgf 3/18/2007 21:08'!
initialize
	
	"We are a singleton-per-Island class; use QMotionCaptures #current rather than new."

	|  island |

	super initialize.
	motions := Dictionary new.

	island := Processor activeIsland.
	island at: #QMotionsCurrent put: self.

	"Motions used by the generic avatar (and the tree fort characters)."
	self import: 'stand.bvh' from: 'Content/Motion/Male' atScale:0.095.
	self import: 'male_walkLoop.bvh' from:  'Content/Motion/Male' atScale:0.095.
! !

Object subclass: #QPatchManager
	instanceVariableNames: 'owner serviceMgr'
	classVariableNames: 'RestartRequired UpdateSemaphore'
	poolDictionaries: ''
	category: 'Qwaq-Objects-Kernel'!

!QPatchManager commentStamp: '<historical>' prior: 0!
Handles retrieval, caching, and early startup-time loading of software patches.

Patches are numbered ChangeSet fileouts,
such as
	0001SamplePatch.cs
whose name must always start with a number
(and always have first digit zero!!)
Patches are always applied in numerical order.

Patches are posted on the server in the
system-resources/updates/ directory.

They are advertised in an XML file 
		<ServiceProvider>/system-resources/updates/patches.xml
which contains a list of the patch files for each patched Forums version.
(See #checkForPatchesFromMasterFile:).

What patching means
-------------------------------
The patch update process does not use image saving.
It simply fetches and applies software patches 
(in the form of Smalltalk changeset fileouts)
at first login to a service provider.

This imposes limitations on what can be patched:
one cannot add instance variables to the participant UI 
or anything created at bootstrap/startup for instance,
since that code is already complete before the patches are applied.

There could be version-skew problems if updates are present
on one service/forums host, but not on a different one.
Users who touch a service host that has patches will update to that patch level,
while users who see only their own local hosts won't see the patches.
So if a user jumps from forums.qwaq.com (where a patch was loaded)
to a private QFES host (without the patch), and attempts to join
other users who haven't got the patch, then they will be barred
until they restart forums.
!

!QPatchManager class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QPatchManager class methodsFor: 'instance creation' stamp: 'bgf 7/5/2008 13:35'!
forOwner: owner service: serviceManager

	"Create a patch manager.  The service manager may be nil,
	 but in that case, only #applyCachedPatches can be used.

	There is no long-term state stored in the patch manager,
	so instances can be created at need.   
      This should be done 'one at a time' however
	 (eg. one for prestart,  a different one later for checking new patches)
       so the update semaphore hackery doesn't trip over itself."

	UpdateSemaphore := Semaphore new.
	^(self new)
		owner: owner serviceManager: serviceManager;
		yourself! !

!QPatchManager class methodsFor: 'private' stamp: 'bgf 7/5/2008 13:38'!
patchCacheDirectory

	"The FileDirectory where patch update files for the current version 
	 are cached after downloading.
	 E.g. Update-Qwaq-Forums-1.1.80
	(hyphenated for easier handling in the shell when browsing/debugging)."

	| dirName |
	dirName  := ((QwaqVersion current version) 
					copyReplaceAll: ' ' with: '-')
					copyReplaceAll: ':' with: '-'.	 "Just in case..."

	^ (FileDirectory prefsDirectory 
			directoryNamed: ('Update-', dirName)) 
			assureExistence! !

!QPatchManager class methodsFor: 'accessing' stamp: 'bobw 11/30/2009 10:38'!
patchesFilename

	^'patches.xml'! !

!QPatchManager class methodsFor: 'accessing' stamp: 'bgf 7/5/2008 15:03'!
restartRequired
	"For use in patch updates; a do-it in an update.cs of the form
		 QUpdateManager restartRequired: true
	 will induce a restart after that patch is applied.
	 NOT to be used lightly, as the effect is very sloppy;
	 since the patcher doesn't save image updates, 
	 restart happens during first login of each new application-start."

	^ RestartRequired ! !

!QPatchManager class methodsFor: 'accessing' stamp: 'bgf 7/5/2008 15:03'!
restartRequired: yesNo
	"For use in patch updates; a do-it in an update.cs of the form
		 QUpdateManager restartRequired: true
	 will induce a restart after that patch is applied.
	 NOT to be used lightly, as the effect is very sloppy;
	 since the patcher doesn't save image updates, 
	 restart happens during first login of each new application-start."

	RestartRequired := yesNo! !

!QPatchManager class methodsFor: 'accessing' stamp: 'bobw 11/30/2009 10:34'!
updatesPath

	"This is the front of the path we use in service-manager file gets to
       find patches.xml and the patch.cs files themselves."

	^ 'system-resources/updates/'! !

!QPatchManager methodsFor: 'unused' stamp: 'ar 4/16/2011 17:14'!
applyCachedPatches

	"(Unused.)
	 Find all patches we've cached on this machine for QwaqVersion current version,
	 that are also listed in the given list of updates,
	 and load them if they are not already counted in the QwaqVersion update-level.
	 Returns the number of patches applied.

	 This method reloads patch files that were previously cached 
	 by the method cacheCodeUpdates:names: 

	 Currently not used, because doing this automatically (although a nicer experience)
	 would create version mismash for mixed-users who run QFES but also 
	 use their clients to connect to the hosted service."

	| patchDir files contents highest archive file filePattern |

	[ patchDir := self class patchCacheDirectory ] on: Error do: [
		"Do the directory-assure safely and return if it fails.
	       We don't want the 'give up horribly' fallback below to make startup impossible
		 just because the user can't see into their preferences tree."
		QLogger logWarning: 'Cannot create patch cache directory.'.
		^ 0
	].

	filePattern := '0*.csz'. 	"See cacheCodeUpdates:names: .. this is the only kind of file we consider."
	highest := QwaqVersion current highestUpdate.

	[
		files := patchDir fileNames 
					select: [ : fName | 
						(filePattern match: fName) 
						and: [ fName initialIntegerOrNil > highest ] ].	"Per match, it starts with '0'"
		files := files sort.
		contents := files collect: [ :fName |
			archive := ZipArchive new.
			archive readFrom:  (file := (patchDir readOnlyFileNamed: fName)).
			[ (archive memberNamed: 'source') contents ] 
				ensure: [ file close ].
		].
		self installCodeUpdates: contents names: files notifying: nil.
		^ contents size

	] on: Error do: [ :ex |

		"In the event of any problem whatever:
	       Dump and remove all the cached updates so we'll restart clean,
	       refetching any still-listed updates from the service provider."

		[ patchDir deleteLocalFiles ] on: Error do: [].
		
		"We're kinda wedged, since we don't know what code state we're in.
	       So just give up; after purging the patch directory, the next start should be clean."

		self inform: ('Cannot resume from cache due to an internal error.\({1})\\Please restart OpenQwaq.' 
					withCRs format: {(ex description)}).

		owner isSealedApplication ifTrue: [ Smalltalk quitPrimitive ].
	].

	^ 0
! !

!QPatchManager methodsFor: 'patching' stamp: 'bgf 7/5/2008 16:52'!
cacheCodeUpdates: listOfUpdates names: listOfFilenames

	"Given a list of update-stream content (strings),
       and a matching list of the names of the files they came from,
	 cache all the updates into our local patchCacheDirectory (for this version).

	 Cache each update into a file of the given name - 
	 with a twist: the cached updates are stored as ZIP archives
	 with a single member called 'source'.
	 Partly to obscure the contents, partly to give some integrity-protection.
	 We add a 'z' to the name of each update to reflect this."       
 
	| updateFile file archive |
	
	CProgressBar displayProgress:  'Caching Updates' 
				  at: owner bounds center
				  from: 0 to: listOfUpdates size
				  during:[:bar|

		listOfUpdates with: listOfFilenames do: [ :update :filePath |
			updateFile := (filePath findTokens: '/') last.
			[
				archive := ZipArchive new.
				archive addDeflateString: update as: 'source'.
				file := (self class patchCacheDirectory) forceNewFileNamed: (updateFile, 'z').
				archive writeTo: file.
				file close.
			] on: Error do: [ :ex |
				"This error gets us out of the loop and back to my caller."
				self error: 'Cannot cache update file:\' withCRs, (ex description)
			].
			bar value: (listOfFilenames indexOf: filePath).
		].
	].! !

!QPatchManager methodsFor: 'public' stamp: 'bgf 8/21/2008 15:58'!
checkAndApplyPatches

	"Public.
	 Ask the service manager for the patch list to apply to this version.
	 Fetch any patches we haven't cached,
	 and apply the whole required set.

	 Returns the number of patches applied.
	 If -1 is returned, patches were applied that require restart,
	 so the caller should do nothing further.

	 TODO: we should deal somehow with the 'too-many-patches' problem
	 if a user switches back from hosted to private server in mid-session."

	| updates count |

	updates := self checkForPatches.
	count := updates size.
	count > 0 ifFalse: [
		QLogger logInfo: 'No updates to apply.'.
	] ifTrue: [
		owner statusInfo: 'Applying updates.'.
		self class restartRequired: false.
		self gatherAndInstallCodeUpdates: updates.
		"If any of the patches set my restartRequired flag, now is the time to restart.
	       (This is extremely ugly, and only to be used in dire circumstances,
	 	 since it forces a restart at every relaunch of the client.)"
		self class restartRequired ifTrue: [
			self inform:
'Updates have been applied that require a restart.
Click OK to proceed.'.

			WorldState addDeferredUIMessage:[
					(owner projectMorph window ifNil:[owner projectMorph]) delete. "this kills us"
					owner class launchRestart: false. "this restarts the app"
				].
			^ -1
		].
	].
	^ count! !

!QPatchManager methodsFor: 'patching' stamp: 'bobw 11/30/2009 10:38'!
checkForPatches

	"Return a list of  changeSet file names to fetch and apply to this image,
	 fetched from the service manager.

       In the result, each entry is a uri-path 
	 relative to the organization updates directory."
	
	^ self checkForPatchesFromMasterFile: (self class updatesPath), self class patchesFilename! !

!QPatchManager methodsFor: 'patching' stamp: 'bgf 1/22/2009 12:34'!
checkForPatchesFromMasterFile: patchesMasterFile

	"Check for updates using the given service manager.
	 Retrieves the indicated updateFile,  
       which is expected to be an XML file of the form

	 <update version='Qwaq Forums 1.0.36'>   ... but use doublequotes in there
		<patch>0001TestUpdate.cs</patch>
		<patch>0002SecondUpdate.cs</patch>
	</update>

	  <update version='Qwaq Forums 1.0.37'>
		<patch>0001FirstUpdateFor37.cs</patch>
		<patch>0002AnotherFor37.cs</patch>
	</update>

	The method returns the list of patches (changeset file names)
	that are currently not loaded in this image,
 	according to the QwaqVersion current update level.
	The names are URI paths relative to the '#updatesPath' 

	This is an XML-based variant of Utilities parseListContents:,
	which is used in the base-system auto-update support.
	That method relies on the text payload being strictly Squeak-line-ended,
	so its format is not really practical for multi-system use
	(ya can't even use VI to edit those files for hellsakes.) "

	| names updatesForMe version highest num data xdoc |

	data := [serviceMgr getFile: patchesMasterFile] 
			on: Error do:[:ex| 
				QLogger logWarning: 'No patch list found (', (ex description), ')'.
				ex return: nil].
	data ifNil:[^#()   "Bail out - no patches file " ].

	xdoc := [  XMLDOMParser parseDocumentFrom: (ReadStream on: data asString) ] 
			on: Error do: [ : ex |
				QLogger logWarning: 'Cannot read ', patchesMasterFile, ':',  (ex description).
				^ #()		"Bail - cannot read patches file " 
			].

	version := QwaqVersion current version.
	highest := QwaqVersion current highestUpdate.

	"Pull out the list of updates that are listed for this version."
	updatesForMe := 
		xdoc elements detect: [ : elem | 
				(elem name = 'update') 
				and: [ (elem attributeAt: 'version' ifAbsent: [nil]) = version ] ]
			 ifNone: [ ^#() "Bail - nothing to update" ].
	names := (updatesForMe elements 
			select: [ : e | e name = 'patch' ])
			collect: [ : e | e contentString withBlanksTrimmed ].

	"From that list, select the updates that haven't yet been applied.
	 (Elsewhere, as updates are applied,
	 QwaqVersion current highestUpdate is incremented.)"

	^ names select: [ : each |
			num := (each findTokens: '/') last initialIntegerOrNil.
			num notNil and: [ num > highest ]].
		! !

!QPatchManager methodsFor: 'patching' stamp: 'bgf 8/21/2008 20:25'!
gatherAndInstallCodeUpdates: updateList 

	"Download the updates from the given list,
	 and then apply them directly into the image.

	 If anything goes wrong during the downloading, 
	 we bail out of the entire process.
       (We trust that if we get the complete files from the service provider,
	  then it's OK to apply them as code updates.)"

	| updates |

	updates := self gatherCodeUpdates: updateList.
	updates ifNil: [ ^ self ].

	self installCodeUpdates: updates names: updateList.

	"Finally, cache them for next time around."
	[ self cacheCodeUpdates: updates names: updateList ] 
	  on: Error do: [ : ex |
		QLogger logError: ex from: self. 
	].

	^ updateList size.! !

!QPatchManager methodsFor: 'patching' stamp: 'bobw 11/30/2009 10:34'!
gatherCodeUpdates: updateList 

	"Assemble the update contents for the given list of patch file names,
	 either by reading them from local cache if present,
	 or by fetching from the service provider. "

	| updates data patchDir archive file |

	[ patchDir := self class patchCacheDirectory ] on: Error do: [
		"Do the directory-assure safely outside the generic failure handler below."
		QLogger logWarning: 'Cannot create patch cache.'.
		patchDir := nil.
	].

	CProgressBar displayProgress:  'Updating' 
				  at: owner bounds center 	
				  from: 0.0 to: 1.0 
				  during:[:bar|
		updates := updateList collect:[:filePath|

			data := nil.
			filePath := filePath withBlanksTrimmed.

			"If there is one in the cache, just use it."
			(patchDir notNil and: [ patchDir fileExists: (filePath, 'z') ])
			 ifTrue: [
				[
					file := patchDir readOnlyFileNamed: (filePath, 'z').
					[ archive := ZipArchive new readFrom: (file := patchDir readOnlyFileNamed: (filePath, 'z')).
					 data := (archive memberNamed: 'source') contents.
					] ensure: [ file close ].
				] on: Error do: [
					"An error in reading/unzipping the cached update. 
					 Since all these files are fetchable, we play it safe: 
	 				 scrub the cache so all subsequent files are refetched cleanly."
					[ patchDir deleteLocalFiles ] on: Error do: [].
					data := nil.
				]].

			data ifNil: [
				"No cached version could be read.  Fetch it."
				ProgressNotification signal: '' extra: 'Downloading ', filePath.
				data := [serviceMgr get: (self class updatesPath),filePath informing: bar] 
						on: Error do:[:ex| 
							 self inform: 'Cannot download update file:', (String cr, String tab), 
									filePath, (String cr), 
									ex description.
							^ nil ].
				data ifNil: [ 
					self inform: 'Cannot download update file:', (String cr, String tab), filePath.
					^ nil ].
			].
			data].
	].

	^ updates! !

!QPatchManager methodsFor: 'patching' stamp: 'bgf 7/5/2008 14:28'!
installCodeUpdates: listOfUpdates names: listOfFilenames

	"Given a list of update-stream contents  (ie the .cs file contents),
       and a matching list of the names of the files they came from,
	 load all the updates, reporting progress."       
 
	CProgressBar displayProgress:  'Installing Updates' 
				  at: owner bounds center
				  from: 0 to: listOfUpdates size
				  during:[:bar|

			self installCodeUpdates: listOfUpdates
				 names: listOfFilenames 
				 notifying: bar.
	].

! !

!QPatchManager methodsFor: 'patching' stamp: 'bgf 8/21/2008 16:00'!
installCodeUpdates: listOfUpdates names: listOfFilenames notifying: bar 

	"Given a list of update-stream contents  (ie the .cs file contents),
       and a matching list of the names of the files they came from,
	 load all the updates.   The loading is done asynchronously
	 from the World's deferred messages system, to avoid various entanglements
	 with the business of changing the code that is driving the updates."       
 
	| updateName updateStream |
	UpdateSemaphore ifNil:[UpdateSemaphore := Semaphore new].

	listOfUpdates with: listOfFilenames do: [ :update :filePath |
		updateName := (filePath findTokens: '/') last sansPeriodSuffix.

		QLogger logInfo: 'Installing patch: ', updateName.
		updateStream := (RWBinaryOrTextStream with: update) reset.
		(ChangeSet named: updateName) 
			ifNotNilDo:[ :cs | ChangeSet removeChangeSet: cs].

		WorldState addDeferredUIMessage:[
				[ChangeSet newChangesFromStream: updateStream named: updateName]
					ensure:[UpdateSemaphore signal].
			].		
		UpdateSemaphore wait.

		(ChangeSet named: updateName) ifNotNilDo:[:cs| ChangeSet removeChangeSet: cs].

		"And mark the current version as having been updated with this patch."
		QwaqVersion current registerUpdate: updateName initialIntegerOrNil.
		bar ifNotNil: [bar value: (listOfFilenames indexOf: filePath) ].
	].
! !

!QPatchManager methodsFor: 'initialize-release' stamp: 'bgf 7/5/2008 12:15'!
owner: aParticipantUI  serviceManager: aServiceMgr

	"The service manager may be nil, 
	 in which case the only service available is 'applyCachedPatches'."

	owner := aParticipantUI.
	serviceMgr := aServiceMgr.! !

Object subclass: #QPlace
	instanceVariableNames: 'occupants category label stance destinationID stanceSelector cameraSelector suggestedPOV isSticky posture hasApp seatingRange'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Objects-Places'!

!QPlace commentStamp: '<historical>' prior: 0!
An avatar destination descriptor for use in the Places... list,
and for <Tab> support in driving,
and generally for defining Go-To Button destinations.

This class is PASS-BY-COPY for island/harness exchanges:
calling or returning QPlaces never involves far-ref operations.
(=/hash are defined so that comparisons remain meaningful across copies.)

The intended use-pattern is for the harness to treat places
only as information handles (annotated names for destinations.)

Places are only 'resolved' 
(used as destinations, by conversion to geometry)  by operations on-island.
No frame far-refs are included in the QPlace itself.

This ensures that place-array manipulation is fast and free of locking,
necessary since the Places lists  and operations are 'hot' and 
frequently refreshed en-masse by the client UI.

A place is defined by:
	1 Occupants			- (OPTIONAL) integer count of avatars at the place, or nil
	2 Category			- for grouping in the places menu
	3 Label				- user-visible name for this place
	4 DestinationFrame	- (ID of) frame defining this place
	5 Selector			- selector for a method on the destination frame;
						  the method takes an avatar replica and returns a global transform for it.
	6 SuggestedPOV		- camera point-of-view to apply when the user goes-to this place.
	7 Posture			- An avatar posture Standing Sitting Hovering

Items 6 and 7 are optional, and considered to be unspecified (not governed by the place) if nil.

For instance, the QPlace for a display panel has
	Category 	- Panels
	Label	  	- panel title
	Dest-Frame - (the panel)
	Selector		- destinationTransformFor: 
	POV		- First Person

	When a user selects this Place from the places menu, or <tab> reaches it,
	their avatar is sent to the transform result  (panel destinationTransformFor: avReplica),
	and the user's viewpoint switches to First Person.

Objects that are natural named destinations
(e.g. QPlacemark, QDisplayPanel)
define a method 
	asPlace
which produces their Place description.  

Arrays of Places are returned from QSpaces by 'landmarksFor: anAvatar'
and used by the participant UI to assemble the client Places menu;
this same landmark list determines the <tab> stops for the space.

For most spaces, the landmarksFor: array is built using a policy:
	Any TFrame that is to be advertised as a named Place
	should return true from #isListedLandmark, 
	and should implement asPlace to return its QPlace description.





!

!QPlace class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QPlace class methodsFor: 'instance creation' stamp: 'bgf 2/24/2009 02:54'!
category: catLabel label: aString frame: aFrame 
	"Make a non-sticky landmark."
	
	^ (self basicNew)
		category: catLabel
		label: aString
		frame: aFrame
		selector: #destinationStanceFor: 
		pov: nil
		posture: nil! !

!QPlace class methodsFor: 'instance creation' stamp: 'bgf 2/24/2009 02:55'!
category: catLabel label: aString frame: aFrame posture: pSym

	"Make a non-sticky landmark with a posture."
	^ (self basicNew)
		category: catLabel
		label: aString
		frame: aFrame
		selector: #destinationStanceFor: 
		pov: nil
		posture: pSym! !

!QPlace class methodsFor: 'instance creation' stamp: 'bgf 2/24/2009 02:55'!
category: catLabel label: aString frame: aFrame pov: aPOVorNil

	"The POV, if non-nil, makes the goto sticky, and adjusts the users' POV.
	 It must be one of the standard POV names from <QUserPreferences viewpointTable>"
	^ (self basicNew)
		category: catLabel
		label: aString
		frame: aFrame
		selector: #destinationStanceFor: 
		pov: aPOVorNil
		posture: nil
! !

!QPlace class methodsFor: 'instance creation' stamp: 'bgf 2/24/2009 02:56'!
category: catLabel label: aString frame: aFrame pov: aPOVorNil posture: pSym

	"The POV, if non-nil, makes the goto sticky, and adjusts the users' POV.
	 It must be one of the standard POV names from <QUserPreferences viewpointTable>"

	^ (self basicNew)
		category: catLabel
		label: aString
		frame: aFrame
		selector: #destinationStanceFor: 
		pov: aPOVorNil
		posture: pSym
! !

!QPlace class methodsFor: 'instance creation' stamp: 'bgf 2/24/2009 02:56'!
category: catLabel label: aString frame: aFrame selector: sel pov: aPOVorNil

	"The POV, if non-nil, makes the goto sticky, and adjusts the users' POV.
	 It must be one of the standard POV names from <QUserPreferences viewpointTable>"

	^ (self basicNew)
		category: catLabel
		label: aString
		frame: aFrame
		selector: sel
		pov: aPOVorNil
		posture: nil

! !

!QPlace class methodsFor: 'instance creation' stamp: 'bgf 2/24/2009 02:56'!
category: catLabel label: aString frame: aFrame selector: sel pov: aPOVorNil posture: pSym

	"The POV, if non-nil, makes the goto sticky, and adjusts the users' POV.
	 It must be one of the standard POV names from <QUserPreferences viewpointTable>"

	^ (self basicNew)
		category: catLabel
		label: aString
		frame: aFrame
		selector: sel
		pov: aPOVorNil
		posture: pSym

! !

!QPlace class methodsFor: 'instance creation' stamp: 'bgf 2/24/2009 09:20'!
category: catLabel label: aString frame: aFrame selector: sel sticky: bSticky

	"Create the specified landmark for a places menu.
	 If sticky is true, the landmark is a first-person POV destination;
	 if false, the landmark is not a sticky destination."

	^ (self basicNew)
		category: catLabel
		label: aString
		frame: aFrame
		selector: sel
		pov:  ((bSticky == true) ifTrue: [ 'First Person'] ifFalse: [ nil ])
		posture: nil
		! !

!QPlace class methodsFor: 'instance creation' stamp: 'bgf 2/24/2009 09:30'!
category: catLabel label: aString frame: aFrame selector: sel sticky: bSticky posture: pSym

	"Create the specified landmark for a places menu.
	 If sticky is true, the landmark is a first-person POV destination;
	 if false, the landmark is not a sticky destination."

	^ (self basicNew)
		category: catLabel
		label: aString
		frame: aFrame
		selector: sel 
		pov:  ((bSticky == true) ifTrue: [ 'First Person'] ifFalse: [ nil ])
		posture: pSym! !

!QPlace class methodsFor: 'instance creation' stamp: 'bgf 2/24/2009 02:57'!
category: catLabel label: aString frame: aFrame sticky: bSticky

	"Create the specified landmark for a places menu.
	 If sticky is true, the landmark is a first-person POV destination;
	 if false, the landmark is not a sticky destination."

	^ (self basicNew)
		category: catLabel
		label: aString
		frame: aFrame
		selector: #destinationStanceFor: 
		pov:  ((bSticky == true) ifTrue: [ 'First Person'] ifFalse: [ nil ])
		posture: nil! !

!QPlace class methodsFor: 'instance creation' stamp: 'bgf 2/24/2009 02:58'!
category: catLabel label: aString frame: aFrame sticky: bSticky posture: pSym

	"Create the specified landmark for a places menu.
	 If sticky is true, the landmark is a first-person POV destination;
	 if false, the landmark is not a sticky destination."
	^ (self basicNew)
		category: catLabel
		label: aString
		frame: aFrame
		selector: #destinationStanceFor: 
		pov:  ((bSticky == true) ifTrue: [ 'First Person'] ifFalse: [ nil ])
		posture: pSym! !

!QPlace class methodsFor: 'instance creation' stamp: 'bgf 2/24/2009 02:58'!
category: aCategory label: aLabel here: hereVec there: thereVec
	"Creates a static place, with no reference to a destination frame."
	^ (self new)
		category: aCategory;
		label: aLabel;
		here: hereVec there: thereVec;
		yourself
! !

!QPlace class methodsFor: 'instance creation' stamp: 'bgf 4/23/2009 15:02'!
category: aCategory label: aLabel stance: placeStance
	"Creates a static place, with no reference to a destination frame."
	^ (self new)
		category: aCategory;
		label: aLabel;
		stance: placeStance;
		yourself
! !

!QPlace class methodsFor: 'islands' stamp: 'bgf 8/8/2008 23:31'!
howToPassAsArgument
	"Copy QPlaces - to avoid far-refs (and their locks) in places-list and goto management."
	^#passByCopy:! !

!QPlace methodsFor: 'comparing' stamp: 'bgf 7/9/2009 13:59'!
= aPlace
	"Places are clone-copied across island boundaries,
	 so we base equality on the key attributes:"
	^ (self == aPlace) 
	   or: [
		(self class == aPlace class) 
		and: [ destinationID = aPlace destinationID  	"Same target"
			and: [ stanceSelector = aPlace stanceSelector ]]		"Same transform-operation"
		and: [ stance = aPlace stance ]				"Same base stance"
		and: [ label = aPlace rawLabel  				"Same user presentation"
			    and: [ category = aPlace category ]]			"Same sort position"
	      and: [ hasApp = aPlace hasApp ]]! !

!QPlace methodsFor: 'private' stamp: 'bgf 6/28/2009 10:05'!
avatarToFollow
	"Private to on-island operations!!
	 If I am the beside-me place for an avatar,
	 return that avatar."
	stanceSelector = #besideMeStanceFor: ifTrue: [
		self destinationFrame ifNotNilDo: [ : df |
			(df isKindOf: QAvatarReplica)
				ifTrue: [ ^ df ].
		]].
	^ nil! !

!QPlace methodsFor: 'accessing' stamp: 'bgf 4/23/2009 17:08'!
cameraSelector
	"A method selector to apply to the destination frame to position a user's camera.
	 Uses the same signature as destinationStanceFor: , etc. "
	^ cameraSelector! !

!QPlace methodsFor: 'accessing' stamp: 'bgf 4/23/2009 17:08'!
cameraSelector: aSelector
	"A method selector to apply to the destination frame to position a user's camera.
	 Uses the same signature as destinationStanceFor: , etc. "
	cameraSelector := aSelector! !

!QPlace methodsFor: 'resolving' stamp: 'bgf 4/23/2009 17:09'!
cameraStanceFor: anAvatarReplica
	"Produce the camera-control stance for this avatar to see as if at this place.
	 This gives the same stance as destinationStanceFor:, unless
	 a specific cameraSelector has been supplied (uncommon)"

	self isStatic ifTrue: [ ^ stance ].

	cameraSelector ifNil: [ ^ self destinationStanceFor: anAvatarReplica ].

	self destinationFrame
		ifNil: [ ^ stance ]
		ifNotNilDo: [ : df |
			df root ifNil: [ ^ nil ].
			cameraSelector isUnary ifTrue: [
				^ df perform: cameraSelector
			] ifFalse: [
				^ df perform: cameraSelector with: anAvatarReplica
			]].! !

!QPlace methodsFor: 'accessing' stamp: 'bgf 2/24/2009 02:05'!
category

	"String: friendly name of the group of locations I belong to.
	 Used to organize the places menu.
	 All placemarks belong to 'Named Places'.
	 Display panels belong to 'Panels'.
	 Other objects define their own group names.

	 In the places menu, all entries with the same category are grouped,
	 and each group is set apart from other groups using separators.
	 If a category has a lot of entries, it may be displayed in a cascading submenu,
	 and in that case, the category appears as the label of the submenu."

	^ category
	! !

!QPlace methodsFor: 'accessing' stamp: 'bgf 2/24/2009 02:18'!
category: aCategory
	"String: friendly name of the group of locations I belong to.
	 Used to organize the places menu.
	 All placemarks belong to 'Named Places'.
	 Display panels belong to 'Panels'.
	 Other objects define their own group names.

	 In the places menu, all entries with the same category are grouped,
	 and each group is set apart from other groups using separators.
	 If a category has a lot of entries, it may be displayed in a cascading submenu,
	 and in that case, the category appears as the label of the submenu."

	category := aCategory
	! !

!QPlace methodsFor: 'initialize-release' stamp: 'bgf 4/23/2009 16:53'!
category: catLabel label: aString frame: frame selector: sel pov: povSymbol posture: pSym

	"This is an alternate initializer ... cover all defaults here."
	occupants := 0.
	category := catLabel.
	label := aString.
	destinationID := (frame get: #objectID).
	stanceSelector := sel ifNil: [ #destinationStanceFor: ].
	suggestedPOV := povSymbol.
	posture := pSym.
	hasApp := false.
	! !

!QPlace methodsFor: 'resolving' stamp: 'bgf 2/24/2009 02:37'!
destinationFrame
	"TFrame defining this destination - retrievable only within the space."
	self isStatic ifTrue: [ ^ nil ].

	(Processor activeIsland isKindOf: TIsland) ifFalse: [self error: 'Not a t-island'].
	(destinationID isKindOf: TObjectID) ifFalse: [self error: 'Not an object id'].
	^ (Processor activeIsland lookup: (destinationID) ifAbsent: [ nil ])! !

!QPlace methodsFor: 'private' stamp: 'bgf 3/5/2010 13:50'!
destinationID
	"The ObjectID of the destinationFrame, if the Place is not static."
	^ destinationID! !

!QPlace methodsFor: 'resolving' stamp: 'bgf 2/24/2009 02:37'!
destinationID: anObjectID
	destinationID := anObjectID! !

!QPlace methodsFor: 'resolving' stamp: 'bgf 4/23/2009 17:10'!
destinationStanceFor: anAvatarReplica
	"Produce the transform for this avatar to be at this destination;
	 the actual result is computed by the destinationFrame using my #selector.
	 NOTE: usable only within the TIsland of the destination frame;
	 errors will result if invoked on a QPlace outside the island.
 	 (QPlaces can move freely from island to harness and back,
	 back are only resolved to actual destination transforms inside the island."

	self isStatic ifTrue: [
		^ stance
	].
	self destinationFrame
		ifNil: [ ^ stance ]
		ifNotNilDo: [ : df |
			df root ifNil: [ ^ nil ].
			self stanceSelector isUnary ifTrue: [
				^ df perform: self stanceSelector
			] ifFalse: [
				^ df perform: self stanceSelector with: anAvatarReplica
			]].! !

!QPlace methodsFor: 'accessing' stamp: 'bgf 2/24/2009 09:42'!
hasApp
	"Optional indicator that the place's destination frame
	 supports    #showFloatingAppPane:   to open a floating embedded-app window for the user."
	^ hasApp ifNil: [ false ]! !

!QPlace methodsFor: 'accessing' stamp: 'bgf 2/24/2009 09:42'!
hasApp: bool
	"Optional indicator that the place's destination frame
	 supports    #showFloatingAppPane:   to open a floating embedded-app window for the user."
	hasApp := bool! !

!QPlace methodsFor: 'testing' stamp: 'bgf 4/24/2009 12:00'!
hasCameraStance
	"Do I have a camera stance distinct from the place (avatar) stance ? "
	^ self isStatic not 
	   and: [ cameraSelector notNil ] 
	   and: [ cameraSelector ~~ stanceSelector ]! !

!QPlace methodsFor: 'comparing' stamp: 'bgf 4/23/2009 14:50'!
hash
	"Hash is defined because = is defined."
	^ ((destinationID hash) bitXor: (label hash)) bitXor: (stance hash).
	! !

!QPlace methodsFor: 'accessing' stamp: 'bgf 4/23/2009 16:58'!
here
	"Overloaded fields... since here/there and destinationFrame/selector are exclusive.
	 Should be called only when self isStatic. "
	^ stance ifNotNil: [ stance here ]! !

!QPlace methodsFor: 'accessing' stamp: 'bgf 4/23/2009 14:50'!
here: hereVector there: thereVector
	"Overloaded fields... since here/there and destinationID/selector are exclusive.
	 Assigning these vectors makes this place answer isStatic true"
	stance := Stance here: hereVector there: thereVector! !

!QPlace methodsFor: 'initialize-release' stamp: 'bgf 4/23/2009 16:58'!
initialize

	"The other initialize-release method
		category:label:etc:etc...
	 is an alternate initializer ... cover all defaults there as well as here."

	occupants := 0.
	stanceSelector := #destinationStanceFor:.
	hasApp := false.! !

!QPlace methodsFor: 'testing' stamp: 'bgf 8/2/2009 12:31'!
isFollowMePlace
	"True if this is the followme place for being-with some user."
	^ stanceSelector = #besideMeStanceFor:! !

!QPlace methodsFor: 'testing' stamp: 'ar 3/16/2009 22:56'!
isNamedPlace
	"Helpful to abstract from the concrete naming of 'named places'"
	^category = 'Named Places'! !

!QPlace methodsFor: 'resolving' stamp: 'bgf 5/5/2009 21:01'!
isNear: anAvatar
	"Is the avatar near enough to this place to merit checking its entry in the Places list?
	 This is usable ONLY within the TIsland of the destination"
	^  (self cameraStanceFor: anAvatar) 
		ifNotNilDo: [ : destStance |
			anAvatar isNearStance: destStance	
		] ifNil: [ 
			false
		]

! !

!QPlace methodsFor: 'testing' stamp: 'bgf 8/9/2009 14:36'!
isSeat
	^ posture == #Sitting! !

!QPlace methodsFor: 'accessing' stamp: 'bgf 4/23/2009 14:56'!
isStatic
	"Static places hold only a Stance,  instead of a destination frame ID and selector."
	^ destinationID isNil
	 
	   ! !

!QPlace methodsFor: 'accessing' stamp: 'bgf 3/4/2009 11:46'!
isSticky
	"If true, the avatar will be marked sticky when jumped to this location,
	 regardless of suggested-pov.  This accessor gets the override flag itself,
	 see also #sticky"
	^ isSticky ! !

!QPlace methodsFor: 'accessing' stamp: 'bgf 3/4/2009 11:38'!
isSticky: flag
	"If true, the avatar will be marked sticky when jumped to this location,
	regardless of suggested-pov.  "
	isSticky := flag! !

!QPlace methodsFor: 'accessing' stamp: 'bgf 2/24/2009 02:06'!
label
	"The user-friendly name of this location, suitable for use as a user-visible list entry."
	| str hint |

	"We may encode a sorting hint inside <& &> at the end of the label."
	str := label.
	str ifNil: [ ^ str ].
	hint := str indexOfSubCollection: '<&'.
	hint > 0 ifTrue: [
		^ str first: (hint-1).
	].
	^str! !

!QPlace methodsFor: 'accessing' stamp: 'bgf 2/24/2009 02:19'!
label: aString
	"The user-friendly name of this location, suitable for use as a user-visible list entry."
	"We may encode a sorting hint inside <& &> at the end of the label."
	label := aString! !

!QPlace methodsFor: 'accessing' stamp: 'bgf 2/24/2009 02:06'!
occupants
	"Optional - usually unused; the count of people at this place.
	 Heavily used in one case - places list updates.
	 See QAvatarReplica currentLandmarksWithCensus "
	^occupants ifNil: [ 0 ]
	! !

!QPlace methodsFor: 'accessing' stamp: 'bgf 2/24/2009 02:06'!
occupants: anInteger
	"Optional - usually unused; the count of people at this place.
	 Heavily used in one case - places list updates.
	 See QAvatarReplica currentLandmarksWithCensus "
	occupants := anInteger! !

!QPlace methodsFor: 'accessing' stamp: 'bgf 2/24/2009 02:07'!
posture
	"The appropriate avatar - posture symbol for this place,
	  or nil if none is mandated."
	^ posture! !

!QPlace methodsFor: 'accessing' stamp: 'bgf 2/24/2009 02:19'!
posture: aSymbol
	"The appropriate avatar - posture symbol for this place,
	  or nil if none is mandated."
	posture := aSymbol! !

!QPlace methodsFor: 'printing' stamp: 'bgf 2/24/2009 02:11'!
printOn: stream
	stream nextPutAll: 'QPlace('.
	label printOn: stream.
	stream nextPut: $).! !

!QPlace methodsFor: 'private' stamp: 'bgf 2/24/2009 03:20'!
rawLabel
	"The whole label string, including embedded sorting hint."
	^ label! !

!QPlace methodsFor: 'accessing' stamp: 'bgf 4/22/2009 15:20'!
seatingRange
	"If zero, no auto-seating within the destination.
	 If > zero, then this is a margin added to the destination-frame bounding box,
	 to define a total volume in which seats are detected."
	^ seatingRange ifNil: [ 0 ]! !

!QPlace methodsFor: 'accessing' stamp: 'bgf 4/22/2009 15:20'!
seatingRange: scalar
	"Is this place one where arrivals are seated in available seats?
	 The scalar, if nonzero, is a margin applied the boundingBox of the destination	
	 to define a box in space... that box is searched for seating positions when avatars come here,
	 so anyone coming-to this destination would land in any available seat whose mass
	 is within (self boundingBox plus seatingRange) "
	seatingRange := scalar! !

!QPlace methodsFor: 'accessing' stamp: 'bgf 5/27/2009 22:48'!
seatingVolume
	"The global TBox defined by my bounding-box plus my seating range
	 (or nil if my seating range is zero, which disables seat-assignment.)"
	self seatingRange > 0 ifFalse: [ ^ nil ].
	self destinationFrame ifNotNilDo: [ : df |
		| box |
		box := df boundingBoxUnscaled transformedBy: (df globalTransform).
		^ TBox min: (box min - self seatingRange) 
				max: (box max + self seatingRange).	
	].
	^ nil
! !

!QPlace methodsFor: 'accessing' stamp: 'bgf 2/24/2009 02:07'!
sortingHint
	| str end begin |
	"User may encode a sorting hint inside <& &> at the end of the label."
	str := label.
	str ifNil: [ ^ nil ].
	begin := str indexOfSubCollection: '<&'.
	begin > 0 ifTrue: [
		end := str indexOfSubCollection: '&>' startingAt: (begin+2).
		end = 0 ifFalse: [
			^ str copyFrom: (begin+2) to: end-1.
		].
	].
	^ ''! !

!QPlace methodsFor: 'accessing' stamp: 'bgf 12/19/2008 14:28'!
sortingLabel

	"The full label, including sort hint, for ordering this label in a places list."
	^ (self sortingHint), (self label).
! !

!QPlace methodsFor: 'accessing' stamp: 'bgf 4/23/2009 14:48'!
stance
	"If a fixed stance is asserted for this place, return it.
	 This is not usually the same as the destination stance."
	^ stance! !

!QPlace methodsFor: 'accessing' stamp: 'bgf 4/23/2009 16:36'!
stance: aStance
	"The static stance associated with this place. 
	 (This is 'the' place if I have no destination frame,  but may be set regardless of that
	 for other purposes; e.g. to note the main focal point for a gathering-destination, etc.)"
	stance := aStance.
	! !

!QPlace methodsFor: 'private' stamp: 'bgf 4/23/2009 17:14'!
stanceSelector
	"Single-argument (taking an avatar) or unary selector 
	 to apply to #destinationFrame to get a destination stance."
	^ stanceSelector ifNil: [ #destinationStanceFor: ]

! !

!QPlace methodsFor: 'private' stamp: 'bgf 4/23/2009 16:53'!
stanceSelector: aSymbol
	"Single-argument selector to apply to #destinationFrame to get a destination stance."
	stanceSelector := aSymbol! !

!QPlace methodsFor: 'accessing' stamp: 'bgf 3/4/2009 11:34'!
sticky
	"If true, the avatar will be marked sticky when jumped to this location,
	 which zooms to suggested point-of-view, and disables falling until the user drives away."
	^ isSticky ifNil: [ suggestedPOV notNil ]! !

!QPlace methodsFor: 'accessing' stamp: 'bgf 2/24/2009 02:07'!
suggestedPOV
	"A string; nil or one of the values from QUserPreferences viewpointTable.
	 If non-nil, the avatar will be marked sticky when jumped to this location,
	 and will have the camera POV adjusted to this value."
	^ suggestedPOV! !

!QPlace methodsFor: 'accessing' stamp: 'bgf 2/24/2009 02:19'!
suggestedPOV: aPOVSymbol
	"A string; nil or one of the values from QUserPreferences viewpointTable.
	 If non-nil, the avatar will be marked sticky when jumped to this location,
	 and will have the camera POV adjusted to this value."
	suggestedPOV := aPOVSymbol! !

!QPlace methodsFor: 'accessing' stamp: 'bgf 4/23/2009 16:57'!
there
	"Overloaded fields... since here/there and destinationFrame/selector are exclusive.
	 Only call this if the Place is static."
	^ stance ifNotNil: [ stance there ]! !

!QPlace methodsFor: 'accessing' stamp: 'bgf 1/4/2009 21:05'!
trimmedLabel
	"Prepare a display-appropriate label from this landmark's assigned label."
	| lbl |
	lbl := self label.
	lbl ifNil: [ ^ 'Unnamed' ].			"just a safety catch"
	lbl := lbl withBlanksCondensed.
	(lbl includes: Character cr) ifTrue: [ lbl := lbl copyUpTo: (Character cr) ].

	lbl := lbl truncateWithElipsisTo: 64.
	^ lbl! !

Object subclass: #QPresenceRecord
	instanceVariableNames: 'userLogin sessionID clientTime islandTime notice sessionQuality urlString otherURLs flags publicName'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Objects-User'!

!QPresenceRecord commentStamp: '<historical>' prior: 0!
A datum describing a user's live forum presence, or last known forum presence.

Has no behaviors; 
this is an exchange format to facilitate sharing information 
between the QPresenceManager TIsland and the client harness.

Fields: 
	login		- (string) 	user login name
	urlString		- (string) 	forum URL where this user is/was present
	clientTime	- (integer) 	utcSeconds timestamp (from the client pov) when recorded
	notice		- (string)	optional 'away' status notice for live user listings.
	islandTime	- (integer) 	internal record-creation time in PresenceManager private seconds.	
	otherURLs	- (string[]) 	names of forums logged in (other than the one I'm in right now)

The representation is lean for economy,
and to allow passByCopy: between island and harness.

Arrays of of these are fetched by the client harness frequently.
The Org-Services island is replicated across all organization clients
(at login, before the forum list / users list are seen) 
and so the island and its content must be small and simple.

All data is copied between TIsland and harness on each fetch.  
No far-refs or complex data are used in the state, just strings and scalars.
!

!QPresenceRecord class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QPresenceRecord class methodsFor: 'instance creation' stamp: 'hrs 8/9/2009 18:09'!
fromXML: aString for: aUserAccountName
	"Answer a presence record from the contents of 
	users/<accountName>/public/locator.xml."
	| parsed location utcSeconds |
	[
		parsed := XMLDOMParser parseDocumentFrom: aString readStream.
		location := (parsed elementAt: 'location' ifAbsent: [
						"Pre 3.0 clients used locationURL, without pcData escaping."
						parsed elementAt: 'locationURL' ifAbsent: []
						])
			ifNil: ['Unknown']
			ifNotNilDo: [:node | node contentString].
		utcSeconds := (parsed elementAt: 'clientUTC' ifAbsent: [
						"Pre 3.0 clients used localTime, but of course, it's really UTC"
						parsed elementAt: 'localTime' ifAbsent: []
				])
			ifNil: []
			ifNotNilDo: [:node | Number readFrom: node contentString].
		^QPresenceRecord login: aUserAccountName url: location clientTime: utcSeconds.
	] on: Error do: [:ex | 
		QLogger logWarning: 'Cannot locate ', aUserAccountName, ': ', ex description.
		^QPresenceRecord login: aUserAccountName url: 'Unknown'.
	].! !

!QPresenceRecord class methodsFor: 'islands' stamp: 'bgf 6/20/2009 09:50'!
howToPassAsArgument
	^ #passByCopy:! !

!QPresenceRecord class methodsFor: 'instance creation' stamp: 'bgf 6/20/2009 09:48'!
login: aName url: aURLString 
	"A presence-record showing that the given user was last attributed this location,
	 at an unknown time... usually used for 'Unknown' placeholder records in harness-side lists."
	^ (self new) 
		userLogin: aName;
		urlString: aURLString;
		yourself! !

!QPresenceRecord class methodsFor: 'instance creation' stamp: 'bgf 6/20/2009 09:49'!
login: aName url: aURLString clientTime: utcSeconds
	"A presence record (off-island, harness side) showing the user in this forum at this user-harness time."
	^ (self new) 
		userLogin: aName;
		urlString: aURLString;
		clientTime: utcSeconds;
		yourself! !

!QPresenceRecord class methodsFor: 'instance creation' stamp: 'bgf 6/20/2009 09:49'!
session: aSessionID login: aName url: aURLString clientTime: utcSeconds islandTime: islandSeconds 
	"A presence-island record showing the user in the given forum at the given client harness time;
	 the islandSeconds time is internal to the presence island and is used for stale-timeout aging."
	^ (self new) 
		sessionID: aSessionID;
		userLogin: aName;
		urlString: aURLString;
		clientTime: utcSeconds;
		islandTime: islandSeconds;
		yourself! !

!QPresenceRecord methodsFor: 'accessing' stamp: 'bgf 8/8/2009 02:21'!
allURLs
	urlString ifNotNil: [
		^ (Array with: urlString), (otherURLs ifNil: [ #() ])
	].
	^ #()! !

!QPresenceRecord methodsFor: 'accessing' stamp: 'bgf 6/20/2009 09:42'!
clientTime
	"Integer - client time in UTC  when the presence was affirmed."
	^ clientTime
	! !

!QPresenceRecord methodsFor: 'accessing' stamp: 'bgf 6/20/2009 09:45'!
clientTime: utcSeconds
	"Integer - client time in UTC  when the presence was affirmed."
	clientTime := utcSeconds
	! !

!QPresenceRecord methodsFor: 'private' stamp: 'bgf 8/8/2009 01:54'!
flag: bitMask
	"The value of a single flag"
	^ flags anyMask: (bitMask)! !

!QPresenceRecord methodsFor: 'private' stamp: 'bgf 8/8/2009 01:42'!
flag: mask set: bool
	"Set an individual bit-flag (or several)"
	flags := (bool ifTrue: [ flags bitOr: mask ] ifFalse: [ flags bitClear: mask ])! !

!QPresenceRecord methodsFor: 'private' stamp: 'bgf 8/8/2009 02:59'!
hasURL: aForumURL
	"Is this user session reporting use of the given forum ?
	  May be the primary, or another forum the user has open."
	^ urlString = aForumURL 
	   or: [ otherURLs notNil and: [ otherURLs includes: aForumURL ]]
		! !

!QPresenceRecord methodsFor: 'initialize-release' stamp: 'bgf 8/8/2009 02:07'!
initialize
	flags := 0.
	sessionQuality := 3.! !

!QPresenceRecord methodsFor: 'accessing' stamp: 'bgf 6/21/2009 10:12'!
isInLobby
	^ self urlString = (QOrgServices lobbyURL)! !

!QPresenceRecord methodsFor: 'accessing' stamp: 'bgf 8/8/2009 01:44'!
isMeetingLeader
	^ (flags anyMask: (QPresenceManager meetingLeaderFlagMask)) ! !

!QPresenceRecord methodsFor: 'accessing' stamp: 'bgf 8/8/2009 01:43'!
isMeetingLeader: bool
	self flag: (QPresenceManager meetingLeaderFlagMask) set: (bool)! !

!QPresenceRecord methodsFor: 'accessing' stamp: 'bgf 8/8/2009 01:49'!
isPrivateTextBlocked
	"Personal preference to block private chat is advertised here."
	^ (flags anyMask: (QPresenceManager privateTextBlockFlagMask)) ! !

!QPresenceRecord methodsFor: 'accessing' stamp: 'bgf 8/8/2009 01:49'!
isPrivateTextBlocked: bool
	"Personal preference to block private chat is advertised here."
	self flag: (QPresenceManager privateTextBlockFlagMask) set: (bool)! !

!QPresenceRecord methodsFor: 'accessing' stamp: 'bgf 6/20/2009 09:42'!
islandTime
	"Integer - internal to the presence manager island.
	 Island time of creation or affirmation of this entry."
	^ islandTime
	! !

!QPresenceRecord methodsFor: 'accessing' stamp: 'bgf 6/20/2009 09:42'!
islandTime: seconds
	"Integer - internal to the presence manager island.
	 Island time of creation or affirmation of this entry."
	islandTime := seconds
	! !

!QPresenceRecord methodsFor: 'accessing' stamp: 'bgf 6/20/2009 09:57'!
login
	"String - the service login of the user (alias for my #userLogin)"
	^ userLogin! !

!QPresenceRecord methodsFor: 'accessing' stamp: 'bgf 6/20/2009 09:43'!
notice
	"Possibly a brief status notice (e.g. 'Away') - nil when active."
	^ notice! !

!QPresenceRecord methodsFor: 'accessing' stamp: 'bgf 6/20/2009 09:42'!
notice: aString
	"Possibly a brief status notice (e.g. 'Away') - nil when active."
	notice := aString! !

!QPresenceRecord methodsFor: 'accessing' stamp: 'bgf 8/8/2009 02:20'!
otherURLs
	^ otherURLs ! !

!QPresenceRecord methodsFor: 'accessing' stamp: 'bgf 8/8/2009 02:19'!
otherURLs: anArrayOfStrings
	otherURLs := anArrayOfStrings asArray! !

!QPresenceRecord methodsFor: 'accessing' stamp: 'bgf 8/8/2009 20:32'!
publicName
	^ publicName! !

!QPresenceRecord methodsFor: 'accessing' stamp: 'bgf 8/8/2009 20:33'!
publicName: pName
	publicName := pName! !

!QPresenceRecord methodsFor: 'accessing' stamp: 'bgf 6/20/2009 09:43'!
sessionID
	"The primary key of a live presence record - the user's generated Croquet session ID."
	^ sessionID! !

!QPresenceRecord methodsFor: 'accessing' stamp: 'bgf 6/20/2009 09:45'!
sessionID: aSessionID
	"The primary key of a live presence record - the user's generated Croquet session ID."
	sessionID := aSessionID! !

!QPresenceRecord methodsFor: 'accessing' stamp: 'bgf 6/20/2009 09:43'!
sessionQuality
	"A scalar indicating this user's session goodness. A positive integer from 0 (no information), 1 (poor) to 5 (best);
	 we use discrete integers to reduce the signal-issuing and handling load from minor floating-point changes."
	^sessionQuality! !

!QPresenceRecord methodsFor: 'accessing' stamp: 'bgf 6/20/2009 09:43'!
sessionQuality: qos
	"A scalar indicating this user's session goodness. A positive integer from 0 (no information), 1 (poor) to 5 (best);
	 we use discrete integers to reduce the signal-issuing and handling load from minor floating-point changes."
	sessionQuality := qos! !

!QPresenceRecord methodsFor: 'accessing' stamp: 'bgf 6/20/2009 09:43'!
urlString
	"String - url of the service/forum where the user was present at time of record affirmation.
	 In the harness, this may be 'Unknown' to indicate absence (for uniformity in lists of records)."
	^ urlString
	! !

!QPresenceRecord methodsFor: 'accessing' stamp: 'bgf 6/20/2009 09:43'!
urlString: aUrlString
	"String - url of the service/forum where the user was present at time of record affirmation.
	 In the harness, this may be 'Unknown' to indicate absence (for uniformity in lists of records)."
	urlString := aUrlString
	! !

!QPresenceRecord methodsFor: 'accessing' stamp: 'bgf 6/20/2009 09:44'!
userLogin
	"String - the service login of the user"
	^ userLogin! !

!QPresenceRecord methodsFor: 'accessing' stamp: 'bgf 6/20/2009 09:44'!
userLogin: aLoginName
	"String - the service login of the user"
	userLogin := aLoginName! !

Object subclass: #QPropsFile
	instanceVariableNames: 'cachedFile cachedData cachedContents'
	classVariableNames: 'Default'
	poolDictionaries: ''
	category: 'Qwaq-Objects'!

!QPropsFile commentStamp: '<historical>' prior: 0!
QPropsFile is a simple file for processing properties files. The properties files look like:

# This is a property file
Qwaq.SP.Active: true     # enable the service provider
Qwaq.SP.Label: This is a label for the SP # this will not be shown
Qwaq.SP.Port: 9991
!

!QPropsFile class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QPropsFile class methodsFor: 'accessing' stamp: 'ar 4/23/2008 10:18'!
default
	"Answers the default (system- or application-wide) properties file if any"
	^Default ifNil:[Default := self new] "an empty properties file"! !

!QPropsFile class methodsFor: 'accessing' stamp: 'ar 4/23/2008 10:16'!
default: aPropsFile
	"Sets the default (system- or application-wide) properties file if any"
	Default := aPropsFile! !

!QPropsFile class methodsFor: 'instance creation' stamp: 'ar 4/23/2008 10:11'!
on: aCachedFile
	^self new on: aCachedFile! !

!QPropsFile methodsFor: 'accessing' stamp: 'ar 4/23/2008 09:52'!
at: aKey
	"Answer the result at aKey or the empty string if not found."
	^self at: aKey ifAbsent:['']! !

!QPropsFile methodsFor: 'accessing' stamp: 'ar 4/23/2008 09:51'!
at: aKey ifAbsent: aBlock
	self updateContents.
	^cachedContents at: aKey ifAbsent: aBlock! !

!QPropsFile methodsFor: 'accessing' stamp: 'ar 6/16/2010 14:24'!
at: aKey put: aString
	| crlfSet delimSet line key data outStream inStream rawLine found |
	data := cachedFile updateContents. "ensure we have everything"
	(self at: aKey) = aString ifTrue:[^self].

	outStream := String new writeStream.
	inStream := (data ifNil:['']) readStream.

	crlfSet := String crlf asCharacterSet.
	delimSet := '#:' asCharacterSet.
	found := false.
	[inStream atEnd] whileFalse:[
		rawLine := (inStream upToAnyIn: crlfSet).
		(rawLine last = Character cr and:[inStream peek = Character lf]) 
			ifTrue:[inStream next].
		rawLine := rawLine withBlanksTrimmed.
		line := rawLine readStream.
		key := line upToAnyIn: delimSet.
		(key notEmpty and:[key last = $:]) ifTrue:[
			key := key allButLast withBlanksTrimmed.
			key = aKey ifTrue:[
				rawLine := key,': ', aString.
				found := true.
			].
		].
		outStream nextPutAll: rawLine, String crlf.
	].

	found ifFalse:[
		QLogger logWarning: 'New system property encountered: ', aKey.
		outStream nextPutAll: aKey,': ', aString, String crlf.
	].

	cachedFile newContents: outStream contents.
! !

!QPropsFile methodsFor: 'accessing' stamp: 'ar 4/23/2008 10:12'!
cachedContents
	"The actual dictionary storing the contents"
	^cachedContents! !

!QPropsFile methodsFor: 'accessing' stamp: 'ar 4/23/2008 10:12'!
cachedData
	"The cached data our properties come from"
	^cachedData! !

!QPropsFile methodsFor: 'accessing' stamp: 'ar 4/23/2008 10:11'!
cachedFile
	"The file we're using for the properties"
	^cachedFile! !

!QPropsFile methodsFor: 'accessing' stamp: 'ar 4/23/2008 10:11'!
fileName
	"The file name we're using for the properties"
	^cachedFile fileName! !

!QPropsFile methodsFor: 'initialize' stamp: 'ar 7/13/2009 20:43'!
forceBackup
	"Force a backup of the underlying file"
	cachedFile forceBackup.! !

!QPropsFile methodsFor: 'initialize' stamp: 'ar 4/23/2008 10:17'!
initialize
	cachedContents := Dictionary new.! !

!QPropsFile methodsFor: 'initialize' stamp: 'ar 7/13/2009 21:08'!
newContents: data
	"Replace all the data in the receiver with the new data.
	Note: Consider using #forceBackup before doing this."
	cachedFile newContents: data.
	cachedData := nil.! !

!QPropsFile methodsFor: 'initialize' stamp: 'ar 4/23/2008 10:10'!
on: aCachedFile
	cachedFile := aCachedFile.
	self updateContents.! !

!QPropsFile methodsFor: 'private' stamp: 'ar 6/16/2010 14:24'!
reloadContents: data
	"Unconditionally reload the contents from the data"
	| stream crlfSet delimSet line key value |
	cachedContents := Dictionary new.
	data ifNil:[^self]. "props file was removed"
	stream := data readStream.
	crlfSet := String crlf asCharacterSet.
	delimSet := '#:' asCharacterSet.
	[stream atEnd] whileFalse:[
		line := (stream upToAnyIn: crlfSet) withBlanksTrimmed readStream.
		key := line upToAnyIn: delimSet.
		(key notEmpty and:[key last = $:]) ifTrue:[
			key := key allButLast withBlanksTrimmed.
			line skipSeparators.
			value := (line upTo: $#) withBlanksTrimmed.
			cachedContents at: key put: value.
		].
	].! !

!QPropsFile methodsFor: 'private' stamp: 'ar 4/23/2008 10:16'!
updateContents
	"Update the contents from the cachedFile if necessary"
	(cachedFile == nil or:[cachedData = cachedFile contents]) ifTrue:[^self].
	cachedData := cachedFile contents.
	self reloadContents: cachedData.
! !

Object subclass: #QSafeProcessLoop
	instanceVariableNames: 'process priority state actionBlock handlerBlock'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Objects-Kernel'!

!QSafeProcessLoop commentStamp: '<historical>' prior: 0!
(c) 2007 Qwaq Inc.  Encapsulates the #start/stop/runLoop pattern that we've been using everywhere.  Avoids sending #terminate to the process. 

Here is an example of how to use it (these methods are defined in a class that need such a loop).

startLoop
	self stopLoop.
	loop := QSafeProcessLoop new.
	loop actionBlock: [self doAction].
	loop handerBlock: [:err | self handleError: err fromLoop: loop].
	loop start.

stopLoop
	loop ifNotNil: [
		loop stop.
		loop := nil.
	]

TODO:
- add the ability to tick at a fixed rate (either here or in a subclass)!

!QSafeProcessLoop class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QSafeProcessLoop methodsFor: 'initialize' stamp: 'jcg 7/9/2007 00:07'!
actionBlock: aBlock
	"The actionBlock performs the desired action each time through the loop."
	actionBlock := aBlock! !

!QSafeProcessLoop methodsFor: 'running' stamp: 'jcg 9/28/2007 17:16'!
avoidDeath
	"This is used by an error-handler that has successfully recovered from an error and wants the loop to keep running (as opposed to eg: logging the error, but still wanting the loop to die).  See #run."
	state == #dying ifTrue: [state := #running].! !

!QSafeProcessLoop methodsFor: 'running' stamp: 'jcg 7/9/2007 01:07'!
blockingStop
	"Wait for the loop to really die before returning."
	state == #new ifTrue: [^self "don't kill a new guy"].
	process == Processor activeProcess ifTrue: [
		"Don't need to wait; we're doing both the killing and the dying."
		state := #dead.
		^self
	].
	state == #dead ifTrue: [^self "already dead"].
	state == #running ifTrue: [state := #dying].
	[state == #dead] whileFalse: [(Delay forMilliseconds: 100) wait]! !

!QSafeProcessLoop methodsFor: 'initialize' stamp: 'jcg 7/9/2007 00:05'!
handlerBlock: aBlock
	"In case of error, the handlerBlock is evaluated with the Error."
	handlerBlock := aBlock! !

!QSafeProcessLoop methodsFor: 'initialize' stamp: 'jcg 7/9/2007 01:08'!
initialize
	super initialize.
	state := #new.
	priority := Processor activeProcess priority.! !

!QSafeProcessLoop methodsFor: 'testing' stamp: 'jcg 9/20/2007 21:06'!
isCurrentlyRunning
	"Is my process the one that Squeak is currently running?"
	^Processor activeProcess == process
! !

!QSafeProcessLoop methodsFor: 'testing' stamp: 'jcg 7/10/2007 01:49'!
isDead
	^ state == #dead! !

!QSafeProcessLoop methodsFor: 'testing' stamp: 'jcg 7/10/2007 01:50'!
isDeadOrDying
	^ state == #dead or: [state == #dying]! !

!QSafeProcessLoop methodsFor: 'testing' stamp: 'jcg 7/10/2007 01:50'!
isNew
	^ state == #new! !

!QSafeProcessLoop methodsFor: 'testing' stamp: 'jcg 7/10/2007 01:49'!
isRunning
	^ state == #running! !

!QSafeProcessLoop methodsFor: 'accessing' stamp: 'jcg 7/9/2007 00:35'!
priority
	^ priority ifNil: [priority := Processor activeProcess priority]! !

!QSafeProcessLoop methodsFor: 'initialize' stamp: 'jcg 7/9/2007 01:09'!
priority: newPriority
	priority := newPriority! !

!QSafeProcessLoop methodsFor: 'private' stamp: 'hrs 8/22/2008 15:27'!
run
	[state == #running] whileTrue: [
		[actionBlock value] 
			on: Error 
			do: [:err | 
				state := #dying. "unless the error handler intervenes, we will die."
				handlerBlock 
					ifNil: ["swallow error, but at least report it." QLogger logError: err.]
					ifNotNil: [handlerBlock value: err]. "up to the client to report or not."
			]
	].
	state := #dead.
! !

!QSafeProcessLoop methodsFor: 'running' stamp: 'jcg 7/9/2007 01:09'!
start
	state == #new ifFalse: [self error: 'safe process loop may only be run once'].
	state := #running.
	process := self fork: #run at: priority! !

!QSafeProcessLoop methodsFor: 'running' stamp: 'jcg 7/9/2007 01:00'!
stop
	state == #running ifFalse: [^self "nothing to do"].
	state := #dying.
! !

Object subclass: #QSounds
	instanceVariableNames: 'soundDirectories sounds'
	classVariableNames: 'Default'
	poolDictionaries: ''
	category: 'Qwaq-Objects-Resources'!

!QSounds commentStamp: '<historical>' prior: 0!
Sound file import manager.
Gives on-demand loading of prepackaged sounds,
and resource-file loading of dropped-in sounds.

A per-island singleton.   Access is normally through the class methods,
which delegate to 'current' - the instance for the current island.

During initialization, directories of sounds can be imported;
this causes the directory to be scanned, and all the sounds it contains
to be listed by their short names... a sort of namespace import.
For example, the world setup for TreeFort islands does this:
	 QSounds importDirectory: 'Content/Qwaq/Sounds/Treehouse'
to import its custom sounds.

The conferencing code can subsequently use 
	QSounds find: 'underwater.wav' 
to get the SampledSound for 'Content/Qwaq/Sounds/Treehouse/underwater.wav'

See QSoundEffects for a more efficient way to play the standard system sound effects.
For most standard sounds, there is no need to import them or refer 
directly to sampled sounds; instead, the named sound effects can be played by name.

!

!QSounds class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QSounds class methodsFor: 'preset effects' stamp: 'bgf 1/18/2007 10:52'!
acceptDrop
	"Positive feedback for a successful drop operation."

	^ self find: 'switch.aiff'! !

!QSounds class methodsFor: 'preset effects' stamp: 'bgf 3/11/2007 01:41'!
avatarEnter

	"Indicates arrival of an avatar in the active space."
	| result |
	result := self find: 'avatarEnter.wav'.
	result ifNotNil: [
		result initialVolume: 0.2.
	].
	^ result! !

!QSounds class methodsFor: 'preset effects' stamp: 'bgf 3/11/2007 01:59'!
avatarLeave

	"A sound played to indicate an avatar's departure."
	| result |
	result :=  self find: 'avatarLeave.wav'.
	result ifNotNil: [
		result initialVolume: 0.2.
	].
	^ result! !

!QSounds class methodsFor: 'preset effects' stamp: 'bgf 1/18/2007 10:54'!
beep
	^ self find: 'shortbeep.wav'! !

!QSounds class methodsFor: 'preset effects' stamp: 'bgf 1/14/2008 13:19'!
blip
	^ self find: 'blip.wav'! !

!QSounds class methodsFor: 'preset effects' stamp: 'bgf 3/11/2007 01:59'!
changeSpace

	"Played when the avatar crosses from one space to another. 
	 This is the #changeSpace sound effect in QSoundEffects."
	| result |
	result := self find: 'chime.wav'.
	result ifNotNil: [
		result initialVolume: 0.35.
	].
	^ result! !

!QSounds class methodsFor: 'preset effects' stamp: 'bgf 3/10/2007 21:47'!
chime
	"A gentle electronic chime."
	^ self find: 'chime.wav'! !

!QSounds class methodsFor: 'instance creation' stamp: 'bgf 1/29/2009 18:56'!
clear
	"Discard the sounds manager."
	Default := nil.! !

!QSounds class methodsFor: 'preset effects' stamp: 'bgf 1/18/2007 10:30'!
click

	"General purpose simple click."
	| result |
	result := self find: 'click.wav'.
	result ifNotNil: [
		result initialVolume: 0.15.
	].
	^ result! !

!QSounds class methodsFor: 'instance creation' stamp: 'bgf 1/29/2009 18:56'!
current
	"Return the sounds manager."

	^ Default ifNil: [
		Default := (self new)
	].! !

!QSounds class methodsFor: 'instance creation' stamp: 'bgf 1/10/2009 12:50'!
defaultDirectories
	"The sound effect directories I always load."
	^ OrderedCollection 
		with: 'Objects/Sounds'
		with: 'Objects/Sounds/Idle'! !

!QSounds class methodsFor: 'preset effects' stamp: 'bgf 1/18/2007 10:53'!
discardObject
	
	"Sort of a dull thud... used to indicate transitions into storage or the trash."
	^ self find: 'drum.aiff'! !

!QSounds class methodsFor: 'preset effects' stamp: 'bgf 1/18/2007 10:32'!
exit

	"A sound played to indicate an avatar's departure."
	| result |
	result :=  self find: 'hangup.wav'.
	result ifNotNil: [
		result initialVolume: 0.05.
	].
	^ result! !

!QSounds class methodsFor: 'accessing' stamp: 'bgf 2/14/2009 12:59'!
find: fileNameOrURI
	"See the instance method for details;
	 this is the most common sound lookup method.
	 Warns if the sound is not found. "
	^ (self current find: fileNameOrURI)
		ifNil: [ 
			QLogger logWarning: 'Unknown sound: ' , fileNameOrURI.
			nil ].! !

!QSounds class methodsFor: 'accessing' stamp: 'bgf 1/18/2007 10:24'!
importDirectory:aDirectoryURI

	"Usually called at application startup and/or TIsland initialization time;
	 add the given directory to the set ofsound-fie directories that are 
	 imported.  The sounds from these directories are subsequently available
	 using QSounds #find: and its variants, using just the file name (no path)
	 or using the full URI relative to the default directory."

	self current importDirectory: aDirectoryURI! !

!QSounds class methodsFor: 'preset effects' stamp: 'bgf 1/18/2007 10:33'!
iris
	^ self find: 'camera_click.aiff'
! !

!QSounds class methodsFor: 'preset effects' stamp: 'bgf 3/10/2007 21:46'!
laser
	"A short laser-beam zing"
	^ self find: 'laser.wav'! !

!QSounds class methodsFor: 'preset effects' stamp: 'bgf 1/18/2007 10:33'!
lift
	"Mechanical elevator-like sound."
	^ self find: 'lift.aiff'
! !

!QSounds class methodsFor: 'preset effects' stamp: 'bgf 7/15/2008 14:39'!
message

	"Arrival of a chat message."
	| result |
	result := self find: 'message.wav'.
	result ifNotNil: [
		result initialVolume: 0.02.
	].
	^ result! !

!QSounds class methodsFor: 'preset effects' stamp: 'bgf 1/18/2007 10:34'!
offline

	"Effect to indicate the end of any sort of connection or session."
	^ self find: 'hangup.wav'! !

!QSounds class methodsFor: 'preset effects' stamp: 'bgf 1/18/2007 10:34'!
online

	"To indicate the start of any sort of connection or session."
	^ self find: 'coin_flip.wav'
! !

!QSounds class methodsFor: 'debugging' stamp: 'jcg 10/7/2009 17:32'!
sine
	"Create a short sine-wave sound."
	| snd |
	snd := self sineBuffer.
	5 timesRepeat: [snd := snd, snd].
	^SampledSound samples: snd samplingRate: 16000

! !

!QSounds class methodsFor: 'debugging' stamp: 'jcg 10/7/2009 18:08'!
sineAAC
	"Create a short sine-wave sound, encoded as AAC.  When you encode the same sine-wave with our AAC encoder, the first 7 frames are different from one another.  After that, it stabilizes and outputs the following."
	^'APQVgdYQdYQNYQM/9wDjn+3Z4X/MHgAAAAABx3FcIIAAAAAAAAAAOO6/hYAAAAAAABv+hClp
aWlpaWlpaWlpaWlpaWlpaWlpaWlpaWlpaWlpaWlpaWlpaWlpaWlpaWlpaWlpaWlpaWlpaWlp
aWlpaWlpaWlpaWlpaWlpaWlpaWlpaWlpaWlpaWlpaWlpaWlpaWlpaWlpaWlpaWlpaWlpaWlp
aWlpaWlpaWlpaWlpaWlpaWlpaWlpaWlpaWlpaWlpaWlpaWlpaWlpaWlpaWlpaWlpaWlpaWlp
aWlpaWlpaWlpaWlpaWlpaWlpaWlpaWlpaWlpaWlpaWlpaWlpaWlpaWlpaWlpaWlpaWlpaWlp
aWlpaWlpaWlpaWlpaWlpaWlpaWlpaWlpaWlpaWlpaWlpaWlpaWlpaWlpaWlpd9aIUtLS0tLS
0tLS0tLS0tLS0tLS0tLS0tLS0tLS0tLS0tLS0tLS0tLS0tLS0tLS0tLS0tLS0tLS0tLS0tLS
0tLS0tLS0tLS0tLS0tLS0tLS0tLS0tLS0tLS0tLS0tLS0tLS0tLS0tLS0tLS0tLS0tLS0tLS
0tLS0tLS0tLS0tLS0tLS0tLS0tLS0tLS0tLS0tLS0tLS0tLS0tLS0tLS0tLS0tLS0tLS0tLS
0tLS0tLS0tLS0tLS0tLS0tLS8A==' base64Decoded asByteArray! !

!QSounds class methodsFor: 'debugging' stamp: 'jcg 10/7/2009 17:31'!
sineBuffer
	"Create a short sine-wave sound."
	| snd |
	snd := SoundBuffer newMonoSampleCount: 1024.
	1 to: 1024 do: [:i |
		snd at: i put: (8000.0 * (2 * Float pi * (i-1) / 16) sin) rounded.
	].
	^snd! !

!QSounds class methodsFor: 'preset effects' stamp: 'bgf 1/18/2007 10:38'!
spring
	
	"Sort of a deep fast sproing sound... used for spring-loaded objects that jump
	 open or closed during an interaction."

	^ self find: 'thunk.wav'! !

!QSounds class methodsFor: 'preset effects' stamp: 'bgf 1/18/2007 10:36'!
switch
	"Generic click of a mechanical switch."
	^ self find: 'switch.aiff'! !

!QSounds class methodsFor: 'preset effects' stamp: 'bgf 1/18/2007 10:36'!
tingle
	"Trekkie tingle... used for materializing things in space."

	^ self find: 'tingle.wav'! !

!QSounds methodsFor: 'initialize-release' stamp: 'bgf 1/10/2009 12:48'!
cleanupOldPrepersist
	"This is invoked if an island-load creates a QSounds...
	  which means that the loader found an old QSounds entry in the island pre-persist list."

	"Remove any old QSounds instances from pre-persist"
	self island prePersistList 
				select: [:item | item isKindOf: QSounds]
				thenDo: [ :defInstance | self island prePersistList remove: defInstance].

	"And excise it after the XML loader is done."
	self island future removeGlobal: #QSoundsCurrent.
! !

!QSounds methodsFor: 'persist' stamp: 'bgf 1/10/2009 12:16'!
constructFrom: pStream

	"See also #initialize; this class is no longer used on-island.."

	
! !

!QSounds methodsFor: 'accessing' stamp: 'bgf 2/14/2009 12:55'!
find: filename 

	"Find an imported SampledSound with the given sound file name.
	 This may be the simple file name (relative to some imported directory),
	 or the path concatenation of the file name and an imported directory path.
	 The find will fail if such a file (and/or URI) has not been imported -
	 these calls never go to disk; only the import... operations go to disk."

	^ self
		find: filename
		ifAbsent: [nil]
! !

!QSounds methodsFor: 'accessing' stamp: 'bgf 1/28/2007 22:04'!
find: filename ifAbsent: aBlock

	"Find an imported SampledSound with the given  file name;
	 this may be the simple file name (relative to some imported directory),
	 or the URI of the file name (including the imported directory)"

	| entry fname |

	entry := sounds at: filename ifAbsent: [
		^ aBlock value.
	].

	"In the table under the given filename we may find: 
		1) the sound - if already fetched
		2) a string (the long path) if filename is the short name
		3) nil, if filename is the long name, and this is the first request for the sound.

	If we find a string, it's an indirection, so we look it up.
	If we find nil, we haven't loaded the sound yet.
	We always have at most one indirection!!"
		
	(entry isString) ifTrue: [
		fname := entry.
		entry := sounds at: fname.
		entry ifNil: [
			"Fetch the sound and replace it in both simple and full name slots."
			(entry := self loadSound: fname) ifNotNil: [
				sounds at: fname put: entry.
				sounds at: filename put: entry.
			].
		] ifNotNil: [
			"We fetched the sound already under the long time;
	 		 overrwite the indirection entry under the short name."
			sounds at: filename put: entry.
		].
	].
	entry ifNil: [
		(entry := self loadSound: filename) ifNotNil:[
			sounds at: filename put: entry.
		].	
	].
	^ entry! !

!QSounds methodsFor: 'initialize-release' stamp: 'bgf 1/29/2009 17:07'!
findSoundResource: aSoundDocumentUri
	"Import a for sound from the document cache.
	 The local URI should refer to a sha-based cache file,
	 relying on the cacheLocally behavior of  'Sound File' documents in QSharedDocument. "
	
	^ (self find: aSoundDocumentUri) ifNil: [
		(self loadSound: aSoundDocumentUri) ifNotNilDo: [ : sound |
			sounds at: aSoundDocumentUri put: sound.
			sound
		]].
! !

!QSounds methodsFor: 'initialize-release' stamp: 'bgf 1/28/2007 21:30'!
import: fileName from: aDirPath

	"Import a single named sound file."
	
	| maybeSlash relPath sound |

	maybeSlash := (aDirPath endsWith: '/') ifTrue: [''] ifFalse: ['/'].
	relPath := (aDirPath, maybeSlash, fileName).
	(sound := self loadSound: relPath) ifNotNil: [
		sounds at: fileName put: sound.
		sounds at: relPath put: sound.
	].
	^ sound
	! !

!QSounds methodsFor: 'initialize-release' stamp: 'bgf 10/25/2008 23:40'!
importDirectory: origPath

	"Load all the .wav/.aiff files from the given directory;
	they can subsequently be retrieved using their simple file name."

	| fullname path |

	origPath ifNil: [ ^ self ].

	"The whole tree of these was relocated to get it out of 'Content/' 
	 as part of lazy-resource partitioning for textures.
	 So we need to up-convert the imports stashed in saved-forums, 
	 where I live in the pre-persist list."
	(origPath beginsWith: 'Content/Qwaq/Sounds/') ifTrue: [
		path := 'Objects/Sounds/', ((origPath allButFirst: ('Content/Qwaq/Sounds/')size )).
	] ifFalse: [
		path := origPath.
	].

	"We do not import the same directory twice; this mechanism isn't meant
	 for handling dynamic content, just prepackged directories of hardcoded textures."
	(soundDirectories includes: path) ifTrue: [
		^ self
	].
	QLogger logInfo: ('Importing sounds: ', path).

	"Remember the path we're preloading; we'll preload the same path
	 if a QwaqStyle is persisted with the island."
	((FileDirectory on: (FileDirectory default pathFromURI: path ))
	  fileNamesMatching: '*.aiff;*.wav')
	    do: [ :name |  
			fullname := path, '/', name.
			sounds at: name put: fullname.
			sounds at: fullname put: nil.
		].! !

!QSounds methodsFor: 'initialize-release' stamp: 'bgf 1/29/2009 18:58'!
initialize
	"We are a singleton class; use QSounds #current rather than new
	 Usually, it suffices to call the class methods."
	|  island |

	super initialize.

	"QSounds used to be set up as a per-island singeton in islands; 
	 this is no longer the case.  
	 So we want to remove ourself from all TIslands."
	island := Processor activeIsland.
	(island class includesBehavior: TIsland) ifTrue: [
		self future cleanupOldPrepersist.
	] ifFalse: [
		"Normal initialization."
		soundDirectories := OrderedCollection new.
		sounds := Dictionary new.
		self class defaultDirectories do: [ :dd|
			self importDirectory: dd
	]].! !

!QSounds methodsFor: 'private' stamp: 'bgf 1/29/2009 18:02'!
loadSound: relPath
	| sound |
	[
		sound := self loadSoundFromPath: (FileDirectory default pathFromURI:relPath)
	] on: Error do: [ :ex| 
		QLogger logWarning: 'Sound file is not a loadable sound: ', relPath, ' (', ex description, ')'.
		ex return.
	].
	^ sound
	! !

!QSounds methodsFor: 'private' stamp: 'bgf 1/29/2009 18:17'!
loadSoundFromPath: absPath

	"Raw sound reader..."
	| stream sound |
	(absPath endsWith: '.wav' caseSensitive: false) ifTrue: [
		stream := FileStream readOnlyFileNamed: absPath.
		sound := SampledSound fromWaveStream: stream.
		stream close.
	].

	((absPath endsWith: '.aiff' caseSensitive: false) 
	 or: [absPath endsWith: '.aif' caseSensitive: false]) ifTrue: [
		sound := SampledSound fromAIFFfileNamed: absPath. 
	].

	^ sound! !

!QSounds methodsFor: 'persist' stamp: 'bgf 1/10/2009 12:16'!
persistOn: pStream

	"See #initialize; this class is no longer held on-island."! !

Object subclass: #QSpring
	instanceVariableNames: 'mass1 mass2 rest stiffness'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Objects-Physics'!

!QSpring commentStamp: 'das 8/10/2006 09:29' prior: 0!
This is part of the QMS suite of classes. A QSpring is a relationship between two masses. The initial location of the two masses defines the rest length of the spring (no forces). Any changes to this cause the spring to attempt to reform to the original length. 

mass1
mass2 - the masses
rest - the rest length
stiffness - hookes constant k in f=-kx. stiffness must always be a negative value.
!

!QSpring class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QSpring methodsFor: 'compute' stamp: 'das 8/14/2006 19:07'!
compute
	
	self compute: stiffness. "arbitrary stiff spring"! !

!QSpring methodsFor: 'compute' stamp: 'das 8/14/2006 23:17'!
compute: k
	
	| v n |
	v := (mass1 location - mass2 location).
	n := v normalized * rest.
	v := (v-n)*k.
	mass1 addForce: v.
	mass2 addForce: v negated.
! !

!QSpring methodsFor: 'accessing' stamp: 'das 7/25/2006 13:05'!
mass1

	^ mass1.! !

!QSpring methodsFor: 'accessing' stamp: 'das 7/25/2006 13:05'!
mass2

	^ mass2.! !

!QSpring methodsFor: 'accessing' stamp: 'das 8/10/2006 09:25'!
mass: m1 mass: m2

	mass1 := m1.
	mass2 := m2.
	rest := (m1 location - m2 location) length.
	stiffness ifNil:[stiffness := -800].! !

!QSpring methodsFor: 'accessing' stamp: 'das 8/10/2006 09:25'!
stiffness: val

	stiffness := val.! !

Object subclass: #QStyles
	instanceVariableNames: 'textStyles'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Objects-Resources'!

!QStyles commentStamp: '<historical>' prior: 0!
TextStyles, Fonts, and TMenuStyle resources for Qwaq applications and spaces.

A per-island singleton.  Access is normally through the class methods,
which delegate to 'current' - the instance for the current island.

Maintains a dictionary of logical text styles named by their roles;
each of those styles has one font+size.

These logical text styles are the namespace of styles recognized
by QLabel and QLabel2D  (#textStyle:); they are safe and compact for use in replicated state.
(During updates, the QLabel objects look up the actual text styles
and use them to composite image forms; the TextStyle references aren't held in Croquet state.)

The basic system styles required by the Qwaq applications appear
in my class method #defaultStyles.

There are convenience class methods 'e.g. QStyles controlFont'
which can be used in harness-side code to return actual Fonts for the UI widgets.
Behind the scenes, those font operations also indirect through my logical style table,
so customizing the style table can restyle the Participant UI, as well as the Croquet content.

Customization can be done by overwriting the core system styles,
and/or by adding new named styles during application and/or Croquet space setup.

!

!QStyles class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QStyles class methodsFor: 'styles' stamp: 'bgf 1/18/2007 14:32'!
at: aTextStyleName putStyle: aTextStyle

	"Override or add a logically-named text style; see #find:"
	(self current) at: aTextStyleName putStyle: aTextStyle! !

!QStyles class methodsFor: 'fonts' stamp: 'bgf 7/11/2009 17:35'!
avatarFontStyle

	"the font used for an avatar name label, in the given size
	 (Avatar labels shift size with distance to counteract 2D-3D rendering effects)"

	^ self find: #AvatarLabelStyle! !

!QStyles class methodsFor: 't-menus' stamp: 'bgf 7/11/2009 01:57'!
avatarMenuStyle

	"TMenuStyle (not TextStyle!!) for menus the avatar editing regime."

	| result fore |

	fore := Color r: 0.7 g: 0.5 b: 0.2.
	result := TMenuStyle new.
	result itemFont: self tmenuFont.
	result separatorColor: ((fore) duller duller lighter alpha:0.4).
	result backgroundColor: (QColors menuBackgroundColor).
	result selectedItemColor:(fore alpha: 0.9).
	result borderColor: (fore darker alpha: 0.6).

	"We use the thematic color as the foreground for title and unavailables."
	result disabledTextColor: (fore blacker alpha:0.9).

	^ result.! !

!QStyles class methodsFor: 'initializing' stamp: 'bgf 1/18/2007 13:20'!
clear

	"Discard the current instance for this island."
	Processor activeIsland at: #QFontsCurrent put: nil.
! !

!QStyles class methodsFor: 'fonts' stamp: 'bgf 1/18/2007 14:23'!
controlFont

	"Default font for labelling controls (buttons, etc.)"
	^ (self current find: #ControlStyle) defaultFont! !

!QStyles class methodsFor: 'initializing' stamp: 'bgf 11/25/2007 14:20'!
current

	"Return the fonts manager for the current island."

	| island instance |

	island := Processor activeIsland.
	instance := island at: #QFontsCurrent ifAbsent: [nil].
	instance ifNil: [
		instance := self new.	"Initialize puts the instance in the island."
	  ].
	^instance! !

!QStyles class methodsFor: 'date utilities' stamp: 'bgf 3/1/2009 17:23'!
dateFormattedFromDateTime: dateTime
	| str  |
	"Returns a time + date string for use where consistent columns are needed."
	str := WriteStream on: String new.
	str
		nextPutAll: (dateTime hour asString padded: #left to: 2 with: $0);
		nextPut: $:;
		nextPutAll: (dateTime minute asString padded: #left to: 2 with: $0).
	dateTime printYMDOn: str withLeadingSpace:true.
	^ str contents
! !

!QStyles class methodsFor: 'date utilities' stamp: 'bgf 3/1/2009 17:46'!
dateFormattedFromUTC: seconds

	"Returns a time + date string for listed uses in the UI (always the same length)."

	^ self dateFormattedFromDateTime: (DateAndTime fromUTCSeconds: seconds).! !

!QStyles class methodsFor: 'date utilities' stamp: 'bgf 8/30/2008 13:04'!
dateStringFromDateTime: dateTime
	| str  |

	"Returns a time + date string for routine use in the UI, 
	 except we make it a wee bit of niceness for recent times
	 (we call today 'Today' rather than showing it as mm-dd-yy or whatever)"

	str := WriteStream on: String new.
	str
		nextPutAll: (dateTime hour asString padded: #left to: 2 with: $0);
		nextPut: $:;
		nextPutAll: (dateTime minute asString padded: #left to: 2 with: $0).

	(dateTime asDate = Date today) ifTrue: [
		str nextPutAll: ' Today'.
	] ifFalse: [
		(dateTime asDate = Date yesterday) ifTrue: [
			str nextPutAll: ' Yesterday'.
		] ifFalse:  [ 
			dateTime printYMDOn: str withLeadingSpace:true.
		]].

	^ str contents
! !

!QStyles class methodsFor: 'date utilities' stamp: 'bgf 7/17/2008 22:07'!
dateStringFromUTC: seconds

	"Returns a time + date string for routine use in the UI, 
	 except we make it a wee bit of niceness for recent times
	 (we call today 'Today' rather than showing it as mm-dd-yy or whatever)"

	^ self dateStringFromDateTime: (DateAndTime fromUTCSeconds: seconds).! !

!QStyles class methodsFor: 't-menus' stamp: 'bgf 7/11/2009 02:01'!
defaultMenuStyle

	"Default TMenuStyle (not TextStyle!!) for menus in qwaq widgets."

	| result |
	result := TMenuStyle new.
	result itemFont: self tmenuFont.
	result separatorColor: QColors menuBorderColor duller slightlyLighter.
	result backgroundColor: QColors menuBackgroundColor.
	result selectedItemColor: ((QColors menuSelectedColor) alpha: 0.9).
	result borderColor: QColors menuBorderColor.

	"We use the control-blue color as the foreground of the title."
	result disabledTextColor:(QColors menuSelectedColor adjustSaturation: -0.12 brightness:-0.3).

	^ result.! !

!QStyles class methodsFor: 'initializing' stamp: 'sn 5/7/2011 18:27'!
defaultStyles
	"These are the core system styles always preloaded at instance creation time."
	(TextConstants includesKey: #'LiberationMono') ifTrue: [
		^ self defaultStylesWithLM.
	].

	(TextConstants includesKey: #'Lucida Grande') ifTrue: [
		^ self defaultStylesWithLG.
	].


	^ Dictionary newFromPairs: {
		#AnnotationStyle . 	"QLabel2D" 
				(QStyles style: #BitstreamVeraSans size: 12) .

		#SmallControlStyle . 	"Participant controls."  
				(QStyles style:  #BitstreamVeraSans size: 9) .
		#ControlStyle . 		"Participant controls."  
				(QStyles style:  #BitstreamVeraSans size: 10) .
		#LargeControlStyle	. "Headline labels." 
				(QStyles style: #BitstreamVeraSans size: 13) .

		#SmallListStyle. 		"List items."
				(QStyles style:  #BitstreamVeraSans size: 9) .
		#ListStyle. 		"List items."
				(QStyles style: #BitstreamVeraSans size: 10).

		#SmallTextStyle .	"Bulk text"
				(QStyles style: #BitstreamVeraSans size: 9).
		#TextEditStyle .		"Various multiline text areas."
				(QStyles style: #BitstreamVeraSans size: 11).

		#LabelStyle	.		"QLabel" 		
				((QStyles style: #BitstreamVeraSans size: 16) centered; yourself) .

		"The font size controls the -quality- of a sign or note, not its size.
		  Smaller allows longer notes to stay within the 'true-size' range."
		#SignStyle	.		"QDisplaySign" 		
				((QStyles style: #BitstreamVeraSans size: 30) centered; yourself) .
		#NoteStyle	.		"QStickyNote" 		
				((QStyles style: #BitstreamVeraSans size: 13) yourself) .
		#CardStyle	.		"Topic cards, etc."
				((QStyles style: #BitstreamVeraSans size: 13) yourself) .

		#TMenuStyle	.	 "In-space menu rendering" 
				(QStyles style: #BitstreamVeraSans size: 16) .

		#AvatarLabelStyle .	"Variable-size labels above avatars."
				(TextStyle named: #BitstreamVeraSans)
	}.! !

!QStyles class methodsFor: 'initializing' stamp: 'bgf 7/12/2009 03:48'!
defaultStylesWithLG
	"These are the core system styles always preloaded at instance creation time,
	 if Lucida Grande is in the image."
	| style |
	style := #'Lucida Grande'.

	^ Dictionary newFromPairs: {
		#AnnotationStyle . 	"QLabel2D" 
				(QStyles style: style size: 12) .

		#SmallControlStyle . 	"Participant controls."  
				(QStyles style:  style size: 10) .
		#ControlStyle . 		"Participant controls."  
				(QStyles style:  style size: 11) .
		#LargeControlStyle	. "Headline labels." 
				(QStyles style: style size: 14) .

		#SmallListStyle. 		"List items."
				(QStyles style:  style size: 10) .
		#ListStyle. 		"List items."
				(QStyles style: style size: 11).

		#SmallTextStyle .	"Bulk text"
				(QStyles style: style size: 10).
		#TextEditStyle .		"Various multiline text areas."
				(QStyles style: style size: 11).

		#AvatarLabelStyle .	"Variable-size labels above avatars."
				(TextStyle named: style) .

		"The font size controls the -quality- of a sign or note, not its size.
		  Smaller allows longer notes to stay within the 'true-size' range."
		#LabelStyle	.		"QLabel" 		
				((QStyles style: style size: 14) centered; yourself) .
		#SignStyle	.		"QDisplaySign" 		
				((QStyles style: style size: 24) centered; yourself) .
		#NoteStyle	.		"QStickyNote" 		
				((QStyles style: style size: 12) yourself) .
		#CardStyle	.		"Topic cards, etc."
				((QStyles style: style size: 14) yourself) .

		#TMenuStyle	.	 "In-space menu rendering" 
				(QStyles style: style size: 14) .
	}.! !

!QStyles class methodsFor: 'initializing' stamp: 'sn 5/7/2011 18:19'!
defaultStylesWithLM
	"These are the core system styles always preloaded at instance creation time,
	 if Liberation Mono is in the image."
	| style |
	style := #'LiberationMono'.

	^ Dictionary newFromPairs: {
		#AnnotationStyle . 	"QLabel2D" 
				(QStyles style: style size: 12) .

		#SmallControlStyle . 	"Participant controls."  
				(QStyles style:  style size: 10) .
		#ControlStyle . 		"Participant controls."  
				(QStyles style:  style size: 11) .
		#LargeControlStyle	. "Headline labels." 
				(QStyles style: style size: 14) .

		#SmallListStyle. 		"List items."
				(QStyles style:  style size: 10) .
		#ListStyle. 		"List items."
				(QStyles style: style size: 11).

		#SmallTextStyle .	"Bulk text"
				(QStyles style: style size: 10).
		#TextEditStyle .		"Various multiline text areas."
				(QStyles style: style size: 11).

		#AvatarLabelStyle .	"Variable-size labels above avatars."
				(TextStyle named: style) .

		"The font size controls the -quality- of a sign or note, not its size.
		  Smaller allows longer notes to stay within the 'true-size' range."
		#LabelStyle	.		"QLabel" 		
				((QStyles style: style size: 14) centered; yourself) .
		#SignStyle	.		"QDisplaySign" 		
				((QStyles style: style size: 24) centered; yourself) .
		#NoteStyle	.		"QStickyNote" 		
				((QStyles style: style size: 12) yourself) .
		#CardStyle	.		"Topic cards, etc."
				((QStyles style: style size: 14) yourself) .

		#TMenuStyle	.	 "In-space menu rendering" 
				(QStyles style: style size: 14) .
	}.! !

!QStyles class methodsFor: 'c-menus' stamp: 'bgf 1/26/2008 17:21'!
delayedActionMenuDelay

	"The time in milliseconds to hold the mouse before showing a delayed-action menu;
	 long enough to distinguish from a leisurely click, short enough to feel responsive."
	^ 350! !

!QStyles class methodsFor: 'cursors' stamp: 'bgf 1/4/2009 15:43'!
drivingCursor
	"The mouse Cursor that indicates right-mouse driving."
	^ Cursor marker! !

!QStyles class methodsFor: 'styles' stamp: 'bgf 1/18/2007 14:23'!
find: aLogicalTextStyleName

	"Return the TextStyle associated with the given symbolic name."
	^ (self current) find: aLogicalTextStyleName! !

!QStyles class methodsFor: 'string utilities' stamp: 'bgf 9/2/2008 02:01'!
firstLinesFrom: string max: nChars

	"Take up to the first nChars of the given string, replacing newlines with CRs
	 and removing surrounding whitespace.... used for a one-line preview of a long string."

	^ ((((string ifNil: [ '']) withBlanksTrimmed)
			truncateTo: nChars + 4 )
				withBlanksTrimmed
					copyReplaceAll: (String cr) with: ' ')
						truncateWithElipsisTo: nChars

	! !

!QStyles class methodsFor: 'string utilities' stamp: 'bgf 8/23/2009 18:37'!
formatForAlert: string
	"Make the given running string usable in a 'status-alert' message,
	 by imposing carriage returns to break it into lines (Poor mans paragraphing).  
	 Since this is for the alert system, the returns are backslashes. "

	^ String streamContents: [ : strm |
		| run  cur |
		run := 0.
		string doWithIndex: [ : c : i |
			run := run + 1.
			(c = Character cr or: [ run > 36 and: [ c isSeparator]  ]) ifTrue: [
				strm nextPut: $\.
				run := 0.
			] ifFalse: [
				run > 54 ifTrue: [ 
					strm nextPut:$-.
					strm nextPut:$\.
					strm nextPut: c.
					run := 0
				] ifFalse: [
					strm nextPut: c.
				]]]].
! !

!QStyles class methodsFor: 'class initialization' stamp: 'bgf 2/18/2008 14:15'!
initialize

	"QStyles initialize."		
	"We use our own TrueType fonts; preload them into the TextConstants table
	 as fully populated TextStyles; this saves fuss later. "

	self clear. 	" Monticello boot-to-the-head class initialize comment "
	self preloadFonts.! !

!QStyles class methodsFor: 'c-menus' stamp: 'bgf 8/25/2008 10:22'!
installMenuFontInMenu: menu
	"This is a no-op - Tweak menus now default to our preferred font."

	"Install the MenuFont in the given menu or menu item."
	false ifTrue: [
		 menu ifNotNil: [ menu installFont: (self menuFont) ]
	].

! !

!QStyles class methodsFor: 'string utilities' stamp: 'bgf 8/6/2010 10:57'!
isNiceFileName: aString

	"Heuristic check to see if an arbitrary bit of string text is friendly to use
	 as a candidate simple file name when proposing to save something.
	Reject anything, for example, containing anything that might be a path marker
	or special character on any system.  Can play it very safe."
	
	^ aString size < 32 
	   and: [ 
		aString allSatisfy: [ : c |
			c isLetter or: [ c isDigit ]
			or: ['-_.' includes: c ]]].! !

!QStyles class methodsFor: 'fonts' stamp: 'bgf 1/18/2007 14:23'!
largeControlFont

	"Default font for large display labels like splash messages."
	^ (self current find:#LargeControlStyle) defaultFont! !

!QStyles class methodsFor: 'fonts' stamp: 'bgf 6/10/2009 14:52'!
listItemFont

	"For grouped-list items"
	^ (self current find: #ListStyle) defaultFont! !

!QStyles class methodsFor: 'class initialization' stamp: 'bgf 1/18/2007 13:15'!
loadTTDescriptor: fName from: aDirPath

	"(Helper method). Load a single named true type font from the given directory,
	 and return its descriptor."
	
	| maybeSlash relPath absPath fd |

	maybeSlash := (aDirPath endsWith: '/') ifTrue: [''] ifFalse: ['/'].
	relPath := (aDirPath, maybeSlash, fName).
	absPath := FileDirectory default pathFromURI:relPath.

	[
		fd := TTCFontDescription addFromTTFile: absPath.
		Transcript cr; show: 'Loaded TrueType font: ', relPath.

		^ fd first  "addFromTTFile returns an array - we know our files are single-font."

	] on: Error do: [ :ex| 
		Transcript cr; show: 'Cannot load TrueType font: ', relPath, ' (', ex description, ')'.
		ex return.
	].
	^ nil! !

!QStyles class methodsFor: 'class initialization' stamp: 'bgf 7/21/2007 17:46'!
loadTTFont: fileName asTextStyle: textStyleName

	"Load the named true-type font from Qwaq/Fonts,
	 Populate its registered style with a full spread of sizes,
	 so that app startup initializations don't trigger strike font generation."

	| fontArray ttf style |

	ttf := self loadTTDescriptor: fileName from: 'Content/Qwaq/Fonts'.
	ttf ifNil:[^nil].
	fontArray := #(10 11 12 13 14 15 16 24 30) collect:
		[:each|
			(TTCFont new) 
				ttcDescription: ttf;
				pointSize: each;
				yourself].

	style := TextStyle fontArray: fontArray.
	style defaultFontIndex: 3.	"12 as default point size"
	TextConstants at: textStyleName asSymbol put: style.! !

!QStyles class methodsFor: 'fonts' stamp: 'bgf 8/25/2008 10:34'!
menuFont
	"Default font for Tweak CMenus. Not for in-space menus!!
	 (For reference - in current code we do rely on the Tweak default CMenu font.)"
	^ CWidgetLibrary default menuFont! !

!QStyles class methodsFor: 't-menus' stamp: 'bgf 1/18/2007 13:15'!
menuStyle

	"Style for TMenuFrames.
	 Returns a TMenuStyle, not a text style!!"

	^ self defaultMenuStyle! !

!QStyles class methodsFor: 'class initialization' stamp: 'bgf 10/22/2008 15:48'!
preloadFonts

	"QStyles preloadFonts"
	"Preload all the fonts we're likely to use, generating font sizes
	 at all the point sizes we care about."

	#(7 8 9 10 11 12 13 14 15 16 24 30) do:[:fs|
		(TextStyle named: 'BitstreamVeraSans') addNewFontSize: fs.
	].

	#(9 10  13) do: [ :fs|
		(TextStyle named: 'BitstreamVeraSerif') addNewFontSize: fs.
	].

	false ifTrue: [
		self loadTTFont: 'CALVINN.TTF' asTextStyle: #CalvinandHobbes.
	].
		"The name has to match what's inside the TTF file
	 	 so we can use 'font familyName' later."
	
	"For future reference.
		self loadTTFont: 'lt_70001.ttf' asTextStyle: #OneStrokeScript.
	"! !

!QStyles class methodsFor: 'string utilities' stamp: 'bgf 8/24/2009 12:36'!
shorten: string to: nLetters
	"Shorten the string to the given number of letters,  at a word break if possible."
	string ifNil: [ ^ '' ].
	string size > nLetters ifFalse: [ ^ string ].

	nLetters to: (nLetters // 2 ) by: -1 do: [ : i |
		(string at: i) isSeparator ifTrue: [
			^ string first: (i-1)
			]].

	^ string truncateWithElipsisTo: nLetters! !

!QStyles class methodsFor: 'string utilities' stamp: 'bgf 3/30/2010 14:29'!
shorten: aString toLabelOfLength: nLetters
	"Shorten the string for use as a label,
	 to at most the given number of letters, cutting at a word break if possible."
	| str i white |
	aString ifNil: [ ^ '' ].

	"First, prepare the string to be a label.  Peel off leading whitespace"
	str := aString.
	i := 1.
	white := true.
	[ white and: [ i < str size ]] whileTrue: [
		(str at: i) isSeparator 
			ifFalse: [ white := false ]
			ifTrue: [ i := i + 1 ].
	].
	i > 1 ifTrue: [ str := str copyFrom: i to: (str size) ].

	"No carriage returns in labels."
	str := str copy replaceAll: (Character cr) with: $ .
	"If there's a period in mid-string, near the end of what we want, just cut before that."
	(nLetters min: (str size)) to: (nLetters // 2) by: -1 do: [ : ci |
		((str at: ci) = $.) ifTrue: [
			^ str first: (ci-1).	
			]].

	"Now trim to the desired length."
	^ self shorten: str to: nLetters! !

!QStyles class methodsFor: 'string utilities' stamp: 'bgf 12/18/2008 12:44'!
simpleFileName: pathOrURI

	"Given a pathed file name of any religion,
	 return the basic file name without path prefix or .suffix."

	| nm dot |
	nm :=  ((pathOrURI copyAfterLast: $/) ifEmpty: [pathOrURI]).
	nm := ((nm copyAfterLast: $\) ifEmpty: [nm]).
	"Remove the .jpg or whatever suffix."
	dot := nm lastIndexOf: $. .
	(dot > 1 and: [ dot between: (nm size - 5) and: nm size ]) ifTrue: [
		nm := nm copyFrom: 1 to: dot -1
	].
	^ nm! !

!QStyles class methodsFor: 'string utilities' stamp: 'bgf 4/29/2009 16:30'!
simplerLoginName: aUserLoginOrName
	"A utility to make shorter strings out of email-logins - for presentation,
	  and when using user-name  strings as part of filename suggestions."
	|nm|
	aUserLoginOrName ifNil: [ ^ nil ].

	nm := aUserLoginOrName.

	"Peel off the @domain if it looks like an email address."
	(nm includes: $@) ifTrue: [
		nm := (nm copyUpTo: $@) ifEmpty: [ nm ].
	].
	"Peel out obvious structure characters so it looks a name"
	nm := (((nm copyWithout: $:)	"now we know we have a copy"
			 replaceAll: '/' with: '-')
			  replaceAll: '\' with: '-')
			   replaceAll: $. with: $ .

	^ nm withBlanksTrimmed! !

!QStyles class methodsFor: 'fonts' stamp: 'bgf 6/10/2009 14:59'!
smallControlFont
	"For buttons that deliberately use smaller labeling"
	^ (self current find: #SmallControlStyle) defaultFont! !

!QStyles class methodsFor: 'fonts' stamp: 'bgf 6/10/2009 14:53'!
smallListItemFont

	"For grouped-list items"
	^ (self current find: #SmallListStyle) defaultFont! !

!QStyles class methodsFor: 'fonts' stamp: 'bgf 8/9/2007 02:00'!
smallTextFont

	"For fine print (transcript, license, etc.)"
	^ (self current find: #SmallTextStyle) defaultFont! !

!QStyles class methodsFor: 'styles' stamp: 'bgf 6/13/2009 14:53'!
style: systemStyleName size: pointSize

	"Initialization utility; produce a single-font-size TextStyle
	 based on the system text style of the given name.
 	 Use this to create single font styles for #at:putStyle:"

	(TextStyle named: systemStyleName) ifNotNilDo: [ : ts |
		^ TextStyle fontArray: {ts addNewFontSize:pointSize }
	].
	^ nil! !

!QStyles class methodsFor: 'fonts' stamp: 'bgf 7/15/2009 23:28'!
surveyQuestionFont
	"The survey questions have to fit in their one-line label thingies,
	 so force them to a maximum 11-point font regardless.
	 This allows more freedom with the other text..font logical names.
	 (The max size cap may need to be lower; here's the place for it in any case."
	| possible |
	possible := (self current find: #TextEditStyle) defaultFont.
	possible pointSize > 11 ifTrue: [
		^ (TextStyle named: possible familyName) fontOfPointSize: 11
	].
	^ possible! !

!QStyles class methodsFor: 'fonts' stamp: 'bgf 1/21/2007 10:00'!
textEditFont

	"Default font for multiline text displays, especially editable text."
	^ (self current find: #TextEditStyle) defaultFont! !

!QStyles class methodsFor: 'fonts' stamp: 'bgf 7/15/2009 23:27'!
textFont
	"For ordinary print (messages, etc)"
	^ (self current find: #TextEditStyle) defaultFont! !

!QStyles class methodsFor: 'fonts' stamp: 'bgf 12/23/2007 11:34'!
tmenuFont

	"Font for rendering in-space menus onto texture."
	^ (self current find:#TMenuStyle) defaultFont! !

!QStyles class methodsFor: 'date utilities' stamp: 'bgf 7/26/2008 20:38'!
tzDescription: tzOffsetMinutes

	"Squeak has no modern zoneinfo or time-zone name support
	 and only knows about 4 timezones.  So for room clocks,
	 we use tzOffset minutes in various places, 
	 and show it this way: GMT-08:30"

	| str dur |
	str := WriteStream on: (String new).
      str nextPutAll:  'GMT'.

	dur := Duration minutes: tzOffsetMinutes.
	dur negative ifTrue: [
		dur := dur negated.
		str nextPutAll: '-'.
	] ifFalse: [
		str nextPutAll: '+'.
	].
	dur hours < 10 ifTrue: [ str nextPut:$0 ].  dur hours printOn: str.
	str nextPut: $:.
	dur minutes < 10 ifTrue: [ str nextPut:$0 ].  dur minutes printOn: str.
	str nextPutAll: str.

	^ str contents! !

!QStyles class methodsFor: 'cursors' stamp: 'bgf 1/4/2009 15:43'!
warpPlaceCursor
	"The mouse Cursor that indicates when a click will be a go-to."
	^ Cursor webLink! !

!QStyles methodsFor: 'accessing' stamp: 'bgf 1/18/2007 13:28'!
at: aTextStyleName putStyle: aTextStyle

	"Override or add a named text style; see #findTextStyle"
	textStyles at: aTextStyleName put: aTextStyle! !

!QStyles methodsFor: 'persist' stamp: 'bgf 7/11/2009 00:54'!
constructFrom: pStream

	"To preserve the relationship between the island and its unique QwaqStyle,
       we replace the island global registered instance with the one being read from XML"

	| styleStrings styleName styleSize |

	"We don't exploit this, and it causes spurious retention 
	 when changing the defaults from test to test."
	true ifTrue: [ ^ self ].

	styleStrings := pStream readStringMap: 'textStyles' default: nil.
	styleStrings ifNotNil: [
		styleStrings keysAndValuesDo: [ :key :styleSummary |
			| style |
			"Ugh - just dig out the space-separated stylename (first) and size (second)."
			styleName := styleSummary substrings first.
			styleSize := styleSummary substrings second initialIntegerOrNil.
			style := (self class style: styleName size: styleSize).
			style ifNotNil: [
				self at: key putStyle: (self class style:styleName size: styleSize).
			] ifNil: [
				QLogger logWarning: 'Unknown font: ', styleName, ' for ', key.
				self at: key putStyle: (QStyles style: #BitstreamVeraSans size: 10).
			].
		].
	].! !

!QStyles methodsFor: 'accessing' stamp: 'bgf 1/18/2007 13:28'!
find: aTextStyleName

	"Find a predefined text style.
	QwaqStyle maintains a style dictionary using logical names.
	It is loaded with defaults for the Qwaq widget set.
	See #initialize for the default table of choices."

	^ textStyles at: aTextStyleName ifAbsent: [nil]
! !

!QStyles methodsFor: 'initialize' stamp: 'bgf 2/18/2007 09:58'!
initialize
	
	|  island |

	"We are a singleton-per-Island class.
	 Reading a scene always uses regular constructors, not current,
	 so when rereading from persistence, we have to replace the default instance,
	 and make sure the replacement is the one listed for subsequent persisting."

	island := Processor activeIsland.
	(island at: #QFontsCurrent) ifNotNilDo: [ :old|
		"There was a default instance; we're going to overwrite it with the 
	       'new' version constructed by readScene; so unhook the old one.
		 This case can only mean that we're internalizing a TISland,
		 since that's the code path that calls #new rather than #default." 
		(island class includesBehavior: TIsland) ifTrue: [
			island prePersistList 
				select: [:item | item isKindOf: QStyles]
				thenDo: [ :defInstance | island prePersistList remove: defInstance].
		]
	].

	self future prePersist.
	"The future avoids a corner problem. If the first contact with QStyles
	 is during reload of XML, then the island loader will re-add us to its prePersistList;
	 We don't want to be in the list twice.
	 But we can't check for that here, because the loader can't list us
	 until after we're constructed.  Instead, do it in the future."

	island at: #QFontsCurrent put: self.

	"Normal initialization starts here."
	textStyles := Dictionary new.
	self loadCoreStyles.! !

!QStyles methodsFor: 'initialize' stamp: 'bgf 1/18/2007 19:26'!
loadCoreStyles

	"Text styles for the Qwaq participant and widgets. 
	 Styles must be listed for the system defaults;
	 they can be overridden to customize an application."

	self class defaultStyles keysAndValuesDo: [ :key :value |
		textStyles at:key put: value.
	].
! !

!QStyles methodsFor: 'persist' stamp: 'bgf 7/11/2009 00:55'!
persistOn: pStream
	| styleStrings defaults |

	"Record the text style choices.   The PersistStream model doesn't give us
 	 a way to save TextStyles, so we reduce each entry to a string that
 	 is just the system-level style name for the font it uses, and the font size it uses.
	 (TODO: we also need the aligment here)"

	"We don't exploit this, and it causes spurious retention 
	 when changing the defaults from test to test."
	true ifTrue: [ ^ self ].

	defaults := self class defaultStyles.
	styleStrings := Dictionary new.
	textStyles keysAndValuesDo: [ :entryName :style | 
		"We have to do our own default filtering since the string map we're saving
		 is just made up.    This lets us save just the overridden and extra entries."
		(self style:style equals: (defaults at: entryName ifAbsent:[nil])) 
		 ifFalse: [
			styleStrings 
				at: entryName 
				put: (style defaultFont familyName, ' ', style defaultFont pointSize asInteger).
			].
	].

	styleStrings isEmpty ifFalse: [
		pStream storeStringMap: 'textStyles' value: styleStrings default: nil.
	].


! !

!QStyles methodsFor: 'initialize' stamp: 'bgf 2/18/2007 09:50'!
prePersist

	"If not already present, add to the prePersist list."
	Processor activeIsland prePersistList ifNotNilDo: [ :list|
		(list includes: self) ifFalse: [list add:self].
	].! !

!QStyles methodsFor: 'private' stamp: 'bgf 1/18/2007 19:42'!
style: ts1 equals: ts2

	"Compare two text styles for equivalence, relative to persistence needs."

	ts1 = ts2 ifTrue: [ ^true ].

	ts1 ifNil: [ ^ false ].
	ts2 ifNil: [ ^ false ].

	ts1 name = ts2 name ifFalse: [ ^ false ].
	ts1 defaultFont pointSize = ts2 defaultFont pointSize ifFalse: [ ^ false ].

	^ true! !

Object subclass: #QTextures
	instanceVariableNames: 'imports'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Objects-Resources'!

!QTextures commentStamp: '<historical>' prior: 0!
Texture import manager - allows directories of textures to be imported
once, and then loaded (as needed) using only their simple file names,
rather than the full texture URI.

A per-island singleton.   Normally only the class methods are used by client code.
(The class methods delegate to 'current' - the instance for the current island.)

During initialization, directories of textures can be imported.
Each directory is scanned, and all the textures it contains
are recorded by their short names  (like a namespace import).

For example, the world setup for Conferencing islands does this:
	 QTextures importDirectory: 'Content/Qwaq/Textures/Conference'
to import the Conferencing textures.

The conferencing code can subsequently use 
	QTextures find: 'randomUML.gif' 
to get the texture 'Content/Qwaq/Textures/Conference/randomUML.gif'

QTextures does lazy loading. At first, it only records
the short-name long-name association for each entry in each import directory.
Actual TFileTextures are created only when fetched with #find:

The core directory 'Content/Qwaq/Textures' is always imported first.

If an imported directory contains filenames that have already been 
imported in earlier imports, then the new ones replace the old ones
in the short-name lookup table.     This is how restyling can be done;
a custom space can load new versions of core textures.

For example, the Qwaq textures include Content/Qwaq/Textures/AppCard.png.
If application setup includes 'QTextures importDirectory: 'Content/Alternate',
and if that finds a texture such as 'Content/Alternate/AppCard.png',  
then (QTextures get:'AppCard.png') gives the alternate rather than the original.
(This example would change the background template that QFileCard 
uses for applications in the affected spaces.)

The full URI (e.g. Content/Qwaq/Textures/AppCard.png) will always
yield the specified texture.

The QTextures instance in any TIsland registers itself on the island prePersistList, 
so that reloading the island from XML recovers the same texture imports.
!

!QTextures class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QTextures class methodsFor: 'presets' stamp: 'bgf 9/22/2007 20:41'!
blankPanel

	"The background texture of display panels;
	 a neutral gray with edge bevel effects...
	 also a gradient shadow at the top where the title bar ends."

	^ self find: 'blankPanel.gif'! !

!QTextures class methodsFor: 'presets' stamp: 'bgf 7/25/2008 16:47'!
blankPanelSizable

	"The background texture of display panels;
	 a neutral gray with edge bevel effects...
	 also a gradient shadow at the top where the title bar ends."

	^ self find: 'blankPanelSizable.gif'! !

!QTextures class methodsFor: 'presets' stamp: 'bgf 9/22/2007 20:28'!
blankSheet

	"For empty display sheets; this is an almost neutral gray,
	 the same hue as blankPanel, without any edge effects."
	^ self find: 'blank.gif'! !

!QTextures class methodsFor: 'instance creation' stamp: 'bgf 1/17/2007 21:28'!
clear

	"Discard the texture manager for the current island."

	| island |

	island := Processor activeIsland.
	island at: #QTexturesCurrent put: nil.
! !

!QTextures class methodsFor: 'instance creation' stamp: 'bgf 11/25/2007 14:21'!
current

	"Return the texture manager for the current island."

	| island instance |

	island := Processor activeIsland.
	instance := island at: #QTexturesCurrent ifAbsent: [nil].
	instance ifNil: [
		instance := self new.	"Initialize puts the instance in the island."
	  ].
	^instance! !

!QTextures class methodsFor: 'presets' stamp: 'bgf 1/1/2009 12:49'!
expandDisplayButton
	"Specialized surface-texture for the 'floating-window' button of the fixed display."
	^ self findAlpha: 'Content/Qwaq/Textures/expand-display-button.png'! !

!QTextures class methodsFor: 'accessing' stamp: 'bgf 12/6/2007 14:13'!
find: fileNameOrURI

	"See the instance method for details;
	 this is the most general texture lookup method.

	When doing this from within a TIsland, only use texture names
	that are pre-imported by #importCoreTextures,
	and make sure to get the filename case correct!! "

	^ self current find: fileNameOrURI! !

!QTextures class methodsFor: 'accessing' stamp: 'bgf 1/18/2007 00:18'!
findAlpha: fileNameOrURI

	"Somewhat legacy - return the texture identified,
	 forcing its TTexture extension: to be #alpha.
	 Currently we import all textures that way, to avoid having to single out
	 all alpha-channeled textures."

	^ self current find: fileNameOrURI extension:#alpha! !

!QTextures class methodsFor: 'presets' stamp: 'bgf 1/1/2009 12:50'!
fixedDisplayBackground
	"The glowing blue box texture of the fixed display frame."
	^ self find: 'Content/Qwaq/Textures/FixedDisplayBackground.jpg'! !

!QTextures class methodsFor: 'accessing' stamp: 'bgf 1/27/2007 10:35'!
palette
	"A dictionary of the system-supplied palette textures."

	^ self current texturesFrom: 'Content/Qwaq/TexturePalette'
! !

!QTextures class methodsFor: 'presets' stamp: 'bgf 1/7/2008 09:22'!
portalGlass
	^ self findAlpha: 'Content/Textures/box.png'! !

!QTextures methodsFor: 'persist' stamp: 'bgf 10/19/2008 21:25'!
constructFrom: pStream

	"Legacy, for older forums."
	
! !

!QTextures methodsFor: 'accessing' stamp: 'bgf 10/21/2008 17:13'!
find: bundledTextureURI

	"Find an imported TFileTexture with the given texture file name.
	 This may be the simple file name (relative to some imported directory),
	 or the path concatenation of the file name and an imported directory path."

	bundledTextureURI ifNil: [ ^ nil ].
	^ self
		find: bundledTextureURI
		extension: (TTextures extensionForFile: bundledTextureURI)
		ifAbsent: [
			QLogger logWarning: 'Unknown Bundled Texture: ', bundledTextureURI.
			nil]
! !

!QTextures methodsFor: 'accessing' stamp: 'bgf 10/21/2008 17:13'!
find: bundledTextureURI extension: ext

	"Get a shared TTexture using the given imported texture file name,
	 return a version whose extension is the given symbol."

	^ self find: bundledTextureURI extension: ext ifAbsent: [ 
			QLogger logWarning: 'Unknown Bundled Texture: ', bundledTextureURI.
			nil ].! !

!QTextures methodsFor: 'accessing' stamp: 'bgf 10/21/2008 20:41'!
find: bundledTextureURI extension: ext ifAbsent: aBlock

	"Find a bundled TFileTexture with the given texture file name;
	 this may be the simple file name (relative to some bundled-form directory),
	 or the URI of the file name (starting with Content/.)

	 This is the bottleneck method for island-side reference to bundled textures.
	 (Bundled textures are presupplied, so they are not for instance included in fat-zips
	 when generating templates or mesh-resource.c3x files) "

	| texture record  |

	bundledTextureURI ifNil: [ ^ aBlock value ].

	^ imports at: bundledTextureURI ifAbsent: [
		record := QForms findBundledInfo: bundledTextureURI.
		record ifNotNil: [
			texture := (TFileTexture new)
						extent: (record second clone);
						fileName: (record first clone);
						extension: ext;
						bMipmap: true;
						shrinkFit: true;
						sha: (record third clone);
						color: (record fourth clone);
						yourself.

			"Stash it under both short and full names so all later lookups succeed."
			imports at: (record first clone) put: texture.
			imports at: (record first copyAfterLast: $/) put: texture.
			texture
		] ifNil: [
			(aBlock value)
		]].

! !

!QTextures methodsFor: 'initialize-release' stamp: 'bgf 10/20/2008 10:52'!
initialize
	"Note: We are a singleton-per-Island class; 
	 client code should use QTextures #current rather than new."
	|  island |
	super initialize.

	"Reading a scene always uses regular constructors, not current,
	 so when rereading from persistence, we have to replace the default instance,
	 and make sure the replacement is the one listed for subsequent persisting."
	island := Processor activeIsland.
	(island at: #QTexturesCurrent) ifNotNilDo: [ :old|
		"There was a default instance; we're going to overwrite it with the 
	       'new' version constructed by readScene; so unhook the old one.
		 This case can only mean that we're internalizing a TIsland,
		 since that's the code path that calls QTextures #new rather than #default." 
		(island class includesBehavior: TIsland) ifTrue: [
			island prePersistList 
				select: [:item | item isKindOf: QTextures]
				thenDo: [ :defInstance | island prePersistList remove: defInstance].
		]
	].
	"And as of version 1.2.20+, we are not a persistent item;
	 the old means of telling islands to pre-scan directories is obsolete."

	TTextures current bundler: self.		"I control bundled-texture loading."
	island at: #QTexturesCurrent put: self.
	imports := Dictionary new.
! !

!QTextures methodsFor: 'persist' stamp: 'bgf 10/19/2008 21:33'!
persistOn: pStream

	"Legacy"


! !

!QTextures methodsFor: 'initialize-release' stamp: 'bgf 2/18/2007 09:49'!
prePersist

	"If not already present, add to the prePersist list."
	Processor activeIsland prePersistList ifNotNilDo: [ :list|
		(list includes: self) ifFalse: [list add:self].
	].! !

!QTextures methodsFor: 'accessing' stamp: 'bgf 10/19/2008 21:32'!
texturesFrom: aDirectoryURI

	"A dictionary of the textures from within the given directory,
	 which should already have been preloaded with #importDirectory.
	 (May be imprecise, just does uri heuristics.)
	 This will load all those textures if they haven't been touched already."

	| result baseName skip |
	result := Dictionary new.

	skip := (aDirectoryURI endsWith:'/') ifTrue: [1] ifFalse: [2].
	QForms bundledFormsByFullName keysDo: [ :key  |
		(key beginsWith: aDirectoryURI) ifTrue: [
			baseName := key copyFrom: (aDirectoryURI size)+skip to: (key size).
			(baseName includes: $/) ifFalse: [
				result at: baseName put: (self find:key).
			].
		].
	].
	^ result.! !

Object subclass: #QUserRelationshipMap
	instanceVariableNames: 'userID tObjectID userName userSummary userRoles realm company companyDepartments emailAddress password manages alternateUserRelationshipMaps'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Objects-User'!

!QUserRelationshipMap commentStamp: 'RJT 5/22/2007 23:04' prior: 0!
Main comment stating the purpose of this class and relevant relationship to other classes.

Possible useful expressions for doIt or printIt.

Structure:
	userID 		<String> User id for logging in
	tObjectID 	<TObjectID> uniquely identify object for linking
	userName 	<String> Users real name
	userSummary <String> information about user
	userRoles 	<collection of QUserRoles> for provisioning permissions
	realm 		<String> Name of realm
	company 	<QCompany> Company that the user works for
	companyDepartments <QCompanyDepartment> Departments that place a user within a company
	emailAddress <String> Email to use for this map / realm
	password 	<String> currently hex string base64Encoded
	manages 	<collection of Companies/Departments/Users/Realms>
	alternateUserRelationshipMaps <collection of QUserRelationshipMap> alternate maps connected to this user for different realms

See accessors for more specific comments.

This class is used to read realm config files and create a object model that can be used to manage users.  The result of the class could be either to update realm config files directly or to produce a set of command xmlFiles that are used to further issolate the server from the config manager.!

!QUserRelationshipMap class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QUserRelationshipMap class methodsFor: 'services' stamp: 'RJT 5/22/2007 23:06'!
getUserMapsFrom: aRealm
	"read xml files to build maps"! !

!QUserRelationshipMap methodsFor: 'accessing' stamp: 'RJT 5/22/2007 22:19'!
alternateUserRelationshipMaps
	"aCollection of other maps that define this user.  A user could have multiple maps if they worked for multiple companies that were managed by a higher up parent company.  None of the data need be the same, in other words a user could have different user names and passwords for each company and even have different email addresses for each company.  But this does allow for data to be shared between faces of the same user, if desired.  Remember this relationship can only happen if there is a manager of multiple realms or companies above this user that addes realms or companies to this user.  Otherwise a user may have multiple disconnected maps."

	^ alternateUserRelationshipMaps! !

!QUserRelationshipMap methodsFor: 'accessing' stamp: 'RJT 5/22/2007 22:15'!
alternateUserRelationshipMaps: anObject
	"Set the value of alternateUserRelationshipMaps"

	alternateUserRelationshipMaps := anObject! !

!QUserRelationshipMap methodsFor: 'accessing' stamp: 'RJT 5/22/2007 22:20'!
company
	"Return the company that the user works for.  To find out the relationship of companies or the realms that a company supports follow the company parent or company realms"

	^ company! !

!QUserRelationshipMap methodsFor: 'accessing' stamp: 'RJT 5/22/2007 22:15'!
company: anObject
	"Set the value of company"

	company := anObject! !

!QUserRelationshipMap methodsFor: 'accessing' stamp: 'RJT 5/22/2007 22:24'!
companyDepartments
	"return the collection of company departments.  This can be one or more departments that places a user within an orgainization.  Departments can have their own realms but this doesn't mean that the user will have access to that realm.  Departments also have their own departments.  Consider a Corporate Parent, with a Company that has a North American Department which has a Sales East Cost department, but you are in the New York branch office, on the blue team. :)"

	^ companyDepartments! !

!QUserRelationshipMap methodsFor: 'accessing' stamp: 'RJT 5/22/2007 22:15'!
companyDepartments: anObject
	"Set the value of companyDepartments"

	companyDepartments := anObject! !

!QUserRelationshipMap methodsFor: 'accessing' stamp: 'RJT 5/22/2007 22:26'!
emailAddress
	"return the email address that this user uses for this User Relationship map.  The user may have a different email address for each realm.  But a user can only have one map and therefore one email per realm (unless they sign up twice with different maps."

	^ emailAddress! !

!QUserRelationshipMap methodsFor: 'accessing' stamp: 'RJT 5/22/2007 22:15'!
emailAddress: anObject
	"Set the value of emailAddress"

	emailAddress := anObject! !

!QUserRelationshipMap methodsFor: 'accessing' stamp: 'RJT 5/22/2007 22:29'!
manages
	"return a collection of objects that contain possible users.  This might be a company or department, it could be a realm, or a list of users.  The goal of this ivar is to provide the means of collecting up the realms that this user is allowed to manage and / or the users whos information this user can modify"

	^ manages! !

!QUserRelationshipMap methodsFor: 'accessing' stamp: 'RJT 5/22/2007 22:15'!
manages: anObject
	"Set the value of manages"

	manages := anObject! !

!QUserRelationshipMap methodsFor: 'accessing' stamp: 'RJT 5/22/2007 22:47'!
password
	"return a password which is currently a base64Encoded"

	^ password! !

!QUserRelationshipMap methodsFor: 'accessing' stamp: 'RJT 5/22/2007 22:15'!
password: anObject
	"Set the value of password"

	password := anObject! !

!QUserRelationshipMap methodsFor: 'accessing' stamp: 'RJT 5/22/2007 22:47'!
realm
	"Return the realm for this map"

	^ realm! !

!QUserRelationshipMap methodsFor: 'accessing' stamp: 'RJT 5/22/2007 22:15'!
realm: anObject
	"Set the value of realm"

	realm := anObject! !

!QUserRelationshipMap methodsFor: 'accessing' stamp: 'RJT 5/22/2007 22:49'!
tObjectID
	"Answer the value of tObjectID"

	^ tObjectID! !

!QUserRelationshipMap methodsFor: 'accessing' stamp: 'RJT 5/22/2007 22:49'!
tObjectID: anObject
	"Set the value of tObjectID"

	tObjectID := anObject! !

!QUserRelationshipMap methodsFor: 'accessing' stamp: 'RJT 5/22/2007 22:15'!
userID
	"Answer the value of userID"

	^ userID! !

!QUserRelationshipMap methodsFor: 'accessing' stamp: 'RJT 5/22/2007 22:15'!
userID: anObject
	"Set the value of userID"

	userID := anObject! !

!QUserRelationshipMap methodsFor: 'accessing' stamp: 'RJT 5/22/2007 22:15'!
userName
	"Answer the value of userName"

	^ userName! !

!QUserRelationshipMap methodsFor: 'accessing' stamp: 'RJT 5/22/2007 22:15'!
userName: anObject
	"Set the value of userName"

	userName := anObject! !

!QUserRelationshipMap methodsFor: 'accessing' stamp: 'RJT 5/22/2007 22:48'!
userRoles
	"answer a collection of roles for this user"

	^ userRoles! !

!QUserRelationshipMap methodsFor: 'accessing' stamp: 'RJT 5/22/2007 22:15'!
userRoles: anObject
	"Set the value of userRoles"

	userRoles := anObject! !

!QUserRelationshipMap methodsFor: 'accessing' stamp: 'RJT 5/22/2007 22:49'!
userSummary
	"Answer the value of userSummary"

	^ userSummary! !

!QUserRelationshipMap methodsFor: 'accessing' stamp: 'RJT 5/22/2007 22:49'!
userSummary: anObject
	"Set the value of userSummary"

	userSummary := anObject! !

Object subclass: #QUserRelationshipMapManager
	instanceVariableNames: 'administrator users'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Objects-User'!

!QUserRelationshipMapManager commentStamp: 'RJT 5/22/2007 23:13' prior: 0!
This class is used to retrieve a realms and users that are managed by an administrator.

Possible useful expressions for doIt or printIt.

Structure:
 adminstrator	<QUserRelationshipMap> the administrator of this Manager
 users		<Collection of QUserRelationshipMaps> the users that this administrator manages

!

!QUserRelationshipMapManager class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QUserRelationshipMapManager class methodsFor: 'instance creation' stamp: 'RJT 5/22/2007 23:14'!
forAdministrator: aRelationshipMap
	"create a new instance and fill it with the users that this user manages"! !

!QUserRelationshipMapManager methodsFor: 'accessing' stamp: 'RJT 5/22/2007 23:11'!
administrator
	"Answer the value of administrator"

	^ administrator! !

!QUserRelationshipMapManager methodsFor: 'accessing' stamp: 'RJT 5/22/2007 23:11'!
administrator: anObject
	"Set the value of administrator"

	administrator := anObject! !

!QUserRelationshipMapManager methodsFor: 'accessing' stamp: 'RJT 5/22/2007 23:13'!
users
	"Answer the value of users"

	^ users! !

!QUserRelationshipMapManager methodsFor: 'accessing' stamp: 'RJT 5/22/2007 23:13'!
users: anObject
	"Set the value of users"

	users := anObject! !

!TIsland methodsFor: '*Qwaq-Objects-Documents' stamp: 'bgf 3/5/2010 09:56'!
allAreRootedFrames: arrayOfOIDS
	"Undo support - verify that the given array of objectIDs
	 all identity TFrames that are rooted in their spaces."

	^ arrayOfOIDS allSatisfy: [ : oid | 
		| f | 
		f := (self lookup: oid ifAbsent: [ nil ]).
		f notNil and: [ f root notNil ]  ]! !

!TIsland methodsFor: '*Qwaq-Objects-Documents' stamp: 'bgf 4/29/2009 23:59'!
allManifestResources
	"Return the list of all registered resources that we'll include in the manifest.xml
 	 (each element must be of a class supporting 
		writeManifestEntry: resRef on: xmlWriter checkingDoneSet: doneShas)
	 This must return an Array (the harness-side callers expect that, not a far-ref to a collection.)"

	| doneSet |

	"The manifest is concerned with payload, not all the different aliased instances,
	 and so we prefilter to avoid redundant entries.  This also helps with documents,
	 were various QxDocument instances sometimes refer to the same underlying file-data
 	 from the POV of different holders."

	doneSet := Set new.
	^ (self documentResources, self contentResources)
		reject: [ : res | 
			(res isBundled)   			"Don't record prepackaged thumbs." 
			or: [ res fileName isNil ]	  	 "or ones we haven't URI's for"
			or: [ res sha notNil and: [
				  (doneSet includes: (res sha)) 
					ifTrue: [ true ]
					ifFalse: [ doneSet add: res sha.  false ] ] ]
			or: [ res fileName beginsWith: '/users/' ]
		].	
! !

!TIsland methodsFor: '*Qwaq-Objects-Documents' stamp: 'bgf 6/11/2009 23:07'!
appendToSummaryNote: aString
	"Add the given note as a new line at the end of my summary note,
	 or make it my summary note if I have none."
	self summaryNote: ( 
		(self summaryNote isEmptyOrNil) 
				ifTrue: [ aString ]
				ifFalse: [ self summaryNote, String cr, aString  ] )
! !

!TIsland methodsFor: '*Qwaq-Objects-Documents' stamp: 'bgf 6/12/2009 08:41'!
ctime
	"A QTimeStamp for creation of the forum itself, under client control.
	 The QSpace ctime is  an alias to this island property."
	^ self at: #QwaqCTime ifAbsentPut: QTimeStamp new ! !

!TIsland methodsFor: '*Qwaq-Objects-Documents' stamp: 'bgf 6/12/2009 09:18'!
ctime: aTimestamp
	"A QTimeStamp for creation, under client control.
	 The QSpace ctime is  an alias to this island property.
	 (The space is responsible for persisting it.)"
	self ctime copyFrom: aTimestamp! !

!TIsland methodsFor: '*Qwaq-Objects-Documents' stamp: 'bgf 6/12/2009 08:38'!
forumDocData
	"Returns the forum 'document metadata' as an array of values.
	 Paired with forumDocData: <dataArray> to allow save/restore
 	 of all forum data that is modified during a forum-save ...
	 so that save-copy can restore the original values after the save process.
	 Must return an array of immediate (copy-by-value) results."

	^ { (self summaryNote) . (self originalTemplate) . (self ctime) . (self mtime) }
! !

!TIsland methodsFor: '*Qwaq-Objects-Documents' stamp: 'bgf 6/12/2009 08:39'!
forumDocData: docDataArray
	"Set the forum 'document metadata' values from the given saved-value array.
	 Paired with forumDocData  to allow save/restore
 	 of all forum data that is modified during a forum-save ...
	 so that save-copy can restore the original values after the save process."
	self summaryNote: (docDataArray first).
	self originalTemplate:  (docDataArray second).
	self ctime: (docDataArray third).
	self mtime: (docDataArray fourth).

! !

!TIsland methodsFor: '*Qwaq-Objects-Documents' stamp: 'bgf 6/12/2009 08:42'!
mtime
	"A QTimeStamp for last-saved-modification of the forum itself, under client control.
	 The QSpace mtime is  an alias to this island property."
	^ self at: #QwaqMTime ifAbsentPut: QTimeStamp new ! !

!TIsland methodsFor: '*Qwaq-Objects-Documents' stamp: 'bgf 6/12/2009 09:18'!
mtime: aTimestamp
	"A QTimeStamp for last-modification (last save), under client control.
	 The QSpace mtime is  an alias to this island property.
	 (The space is responsible for persisting it.)"
	self mtime copyFrom: aTimestamp! !

!TIsland methodsFor: '*Qwaq-Objects-Documents' stamp: 'bgf 6/11/2009 21:05'!
originalTemplate
	"The name of the forum -template used to create me, if known. 
	 Under client / QSpace control."
	^ self at: #QwaqOriginalTemplate! !

!TIsland methodsFor: '*Qwaq-Objects-Documents' stamp: 'bgf 6/12/2009 09:18'!
originalTemplate: aString
	"The name of the forum -template used to create me, if known. 
	 The QSpace summaryNote is just an alias to this island property.
	 (The space is responsible for persisting it.)"
	self at: #QwaqOriginalTemplate put: aString
! !

!TIsland methodsFor: '*Qwaq-Objects-Documents' stamp: 'bgf 5/19/2010 10:51'!
printOn: aStream
	"Mostly here to make island logs decipherable by including the island service url"
	(self at: #ServiceURLString) ifNotNil: [  
		: url | aStream nextPutAll: (self class name); 
					  nextPut: $(; 
					  nextPutAll: ( url copyAfterLast: $/ );
					  nextPut: $).
	] ifNil: [
		super printOn: aStream.
	].! !

!TIsland methodsFor: '*Qwaq-Objects-Documents' stamp: 'bgf 6/11/2009 15:29'!
summaryNote
	"A descriptive string under client control.
	 The QSpace summaryNote is just an alias to this island property."
	^ (self at: #QwaqSummaryNote) ifNil: [ '' ]! !

!TIsland methodsFor: '*Qwaq-Objects-Documents' stamp: 'bgf 6/12/2009 09:17'!
summaryNote: aString
	"A descriptive string under client control.
	 The QSpace summaryNote is just an alias to this island property
	 (The space is responsible for persisting it.)"
	self at: #QwaqSummaryNote put: aString
! !

!TAvatarProxy methodsFor: '*Qwaq-Objects-User' stamp: 'bgf 4/5/2008 04:49'!
asPlace

	"So drive-shortcut go-to press on an avatar behaves like 'go to...' the avatar."
	^ avatar asPlace! !

TCylinder subclass: #QCylinderRing
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Objects-Manipulation'!

!QCylinderRing commentStamp: '<historical>' prior: 0!
A capless circular band, colored as a translucent control,
with 45-degree detent markings.  
Used by QCylinderPivot to give feedback during interactive rotations.

Rendered with both the interior and exterior of the band visible.!

!QCylinderRing class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QCylinderRing methodsFor: 'initialize' stamp: 'bgf 8/30/2007 13:25'!
initialize
	|  |

	super initialize.
	self material: ((TMaterial new)
			transparency: 0.999;
			texture: (QTextures findAlpha: 'rotationTube.png'); 
			emissiveColor: #(1.0 1.0 1.0 0.5);
			diffuseColor: #(1.0 1.0 1.0 0.999); 
			ambientColor: #(1.0 1.0 1.0 0.7);
			shininess: 0.2;
			cullFace: false;
			yourself).

	self	pickable: false;
		capped: false;
		stacks: 2;
		slices: 24.
! !

!QCylinderRing methodsFor: 'transforming' stamp: 'bgf 9/19/2006 18:17'!
radius: r
	super radius: r.
	self height: (r/3).! !

!QCylinderRing methodsFor: 'transforming' stamp: 'bgf 9/19/2006 18:17'!
translation: aVector

	"Present a center-based 'translation:' variant for parity with other Pivots."

	self translationX: (aVector x) y: (aVector y - (self height / 2)) z: (aVector z).! !

TAvatarUser subclass: #QAvatarUser
	instanceVariableNames: 'tweakHand userPreferences shiftSpeed baseSpeed lastUpdateTime lastTypeTime lastLocatorHeartbeat activities qos qosTime focalCameraTransform driveSymbols pasteboardScrap lastLocatorRecord'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Objects-User'!

!QAvatarUser commentStamp: '<historical>' prior: 0!
This is Qwaq's extended avatar controller.
It adds  Tweak necessities for some of the Qwaq participant controls,
support for full-avatar coloring when the laser color is changed,
and provides the bridge between harness-side user preferences
and the shared public userInfo kept in the replica.!

!QAvatarUser class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QAvatarUser class methodsFor: 'as yet unclassified' stamp: 'bgf 8/6/2009 12:00'!
currentStepIntervalForUsers: nUsers favored: bFavored moving: bCanMove
	"The smallest allowed interval for my avatar to send pulse messages.
	  We adjust upward from croquet basics when meetings get crowded,
	  and especially when a led meeting has movement restricted."
	| base |
	base := self minStepInterval.

	"This rule identifies people explicitly singled out in a meeting;
	 leader, or someone with current microphone in moderated restricted-chat.
	 Those few folk get higher rates."
	bFavored ifTrue: [		
		base := (base * (1.0 + (nUsers / 40.0))) truncated.
	] ifFalse: [
		"Ramp down the basic replicated pulse as population grows."
		base := (base * (1.0 + (nUsers/ 8.0))) truncated.
	].
	bCanMove ifFalse: [
		"If movement-locked, use a much lower pulse rate.  This especially helps all-follow-me."
		(base := base * 2.2) truncated
	].
	^ base! !

!QAvatarUser class methodsFor: 'defaults' stamp: 'bgf 8/6/2009 11:40'!
minStepInterval
	"Min milliseconds between drive step events - caps the
	rate at which faster machines flood the time stream with drive steps.
	We start at a lower basis than Croquet because we scale by participant counts;
	see QwaqParticipantUI avatarStepInterval."

	^ 50! !

!QAvatarUser methodsFor: 'events' stamp: 'bgf 1/6/2010 11:31'!
addToInventory: inventoryCardZip

	"<on inventoryChanged in replica>"
	self readyForEvents ifTrue: [
		view owner addToInventory: inventoryCardZip.
	].! !

!QAvatarUser methodsFor: 'portals' stamp: 'bgf 3/13/2010 09:07'!
addToNewSpace: aSpace
	"Extended protocol to inform the replica of the urlString to its current island."
	| url target oldXForm newReplica |

	focalCameraTransform := nil.
	oldXForm := self localTransform clone.
	self driver isDriving ifTrue: [
		"Pull the return-destination back away from portals
	       when we cross through by navigating...	
		 7 units back is enough for some context --- can usually see portal titles ---
	 	 Prevents the go-back button from landing us nose-to-glass against a portal."
		oldXForm translation: (oldXForm translation + (oldXForm lookAt * 7)).
	].

	(super addToNewSpace: aSpace) ifFalse: [ ^ false ].
	newReplica := replica.		"See base method... this is reconnect protection..."

	(view avatar = self) ifFalse: [ ^ false ].		"Edge case - quit during blocked transition."

	url := view session findIslandUrl: aSpace island.

	"Inform the history manager that we're changing spaces."
	target := (aSpace get:#defaultEntry) ifNil: [aSpace].

	(replica = newReplica and: [replica isValidFarRef] )  ifFalse: [ ^ false ].
	view owner undoRedoList reset.
	view owner historyPane 
			noteSpaceChange: (TPostcard urlString:
				(TPostcard makeUrl: (target get:#postcard) urlString relativeTo: url))
			toTransform: (self newTransform)
			fromTransform: (oldXForm).

	(replica = newReplica and: [replica isValidFarRef] )  ifFalse: [ ^ false ].

	pasteboardScrap ifNotNilDo: [ : capframes | replica future setCopiedFramesScrap: capframes ].

	(replica = newReplica and: [replica isValidFarRef] )  ifFalse: [ ^ false ].

	"Give the avatar its new island URL - so it can disambiguate internal postcards."
	replica future islandURL: (view session findIslandUrl: aSpace island).

	"Initialize the activities table to my current set so deltas work."
	activities ifNotEmpty: [
		(replica = newReplica and: [replica isValidFarRef] )  ifFalse: [ ^ false ].
		replica future setActivities: (activities asArray).
	].
	view videoSession ifNotNilDo: [ :vidSession | 
		(replica = newReplica and: [replica isValidFarRef] )  ifFalse: [ ^ false ].
		vidSession cameraManager updateAvatarWithVideo.
	].
	QLogger logInfo: 'Teleported: ', url.
	^ true
! !

!QAvatarUser methodsFor: 'accessing' stamp: 'das 3/11/2007 14:50'!
baseSpeed
	"the calibration setting for basic speed;
	 adjusted by the harness periodically to compensate for rendering rates."

	^ baseSpeed! !

!QAvatarUser methodsFor: 'accessing' stamp: 'das 3/11/2007 14:50'!
baseSpeed: aFactor

	"The calibration setting for basic speed; automatically adjusted to compensate for rendering speed."

	baseSpeed := aFactor.
! !

!QAvatarUser methodsFor: 'driving' stamp: 'bgf 7/28/2009 22:53'!
cameraInLockStep
	"Should the camera be kept exactly where specified, or slerped to follow its target position?
	 We extend this to ensure slerping is -on- when there is a focal override for the camera,	
	 so that there is smoother transition rather than abrupt jumping when it engages."

	^ focalCameraTransform isNil and: [ lockStep ]! !

!QAvatarUser methodsFor: 'events' stamp: 'bgf 11/21/2008 14:48'!
checkTypingActivity
	"If we've been typing, but haven't hit a key lately, stop #Typing.
	 Called on a regular ticker."	
	(activities includes: #Typing) ifTrue: [
		(Time millisecondsSince: lastTypeTime) > 1200 ifTrue: [
			self stopActivity: #Typing
		]].! !

!QAvatarUser methodsFor: 'editing' stamp: 'bgf 2/6/2010 21:54'!
copySelection: evt
	self readyForEvents ifTrue: [
		replica future copySelection: (view event).
	].
	! !

!QAvatarUser methodsFor: 'events' stamp: 'bgf 4/11/2008 23:35'!
costumeChanged

	"An in-space change to avatar replica costume."
	driver heights: ((replica get: #heights) ifNil: [ driver heights ])! !

!QAvatarUser methodsFor: 'landmarks' stamp: 'bgf 3/15/2008 17:24'!
currentLandmarks
	"Returns an array of QPlaces, 
	 representing the defined landmarks in the current space
 	 that are available to this avatar."

	^ replica get: #currentLandmarks.
! !

!QAvatarUser methodsFor: 'landmarks' stamp: 'bgf 8/8/2008 22:15'!
currentLandmarksWithCensus

	"Returns #currentLandmarks, except that the places listed
	 have their occupant-counts set by counting the in-space avatars in each place."

	^ replica get: #currentLandmarksWithCensus
! !

!QAvatarUser methodsFor: 'landmarks' stamp: 'jjl 8/5/2008 13:35'!
currentPlace

	"If my avatar is at a specific QPlace in the space,
	 returns that QPlace; else nil.
	 The avatar notes places at the start of gotos,
	 so this gives information 'early' - which allows gathering places
	 to assign slots appropriately, and allows fast-tabbing to go forward."

	self readyForEvents ifFalse:[ ^nil ].

	^replica get: #currentPlace
! !

!QAvatarUser methodsFor: 'editing' stamp: 'bgf 2/6/2010 21:54'!
cutSelection: evt
	self readyForEvents ifTrue: [
		replica future cutSelection: (view event).
	].
! !

!QAvatarUser methodsFor: 'editing' stamp: 'bgf 2/6/2010 21:54'!
deleteSelection: evt
	self readyForEvents ifTrue: [
		replica future deleteSelection: (view event).
	].
! !

!QAvatarUser methodsFor: 'initialize' stamp: 'bgf 3/21/2009 00:31'!
destroy
	"Mostly aids to garbage collection."
	self stopAllScripts.
	view := nil.
	replica := nil.
	super destroy.
! !

!QAvatarUser methodsFor: 'events' stamp: 'bgf 7/18/2009 10:41'!
doGestureForKeystroke: cEvent

	"Builtin simple gestures.
	 If the given keyStroke event maps to a builtin gesture, issue it.
 	 Gesture keystrokes... these should be chosen not to overlap with drive keys.

	 Returns true if a gesture was recognized and performed, false otherwise."

	cEvent keyCharacter caseOf: {
		[$ ] 			-> [replica future attentionGesture  "generic hi!! gesture"].
		[$y] 		-> [replica future sayYesGesture ].
		[$n] 		-> [replica future sayNoGesture ].
		[$b] 		-> [replica future smallBowGesture ].
		[$t]			-> [self ifAllowedToMove: [replica future togglePosture ]].	"Sit/Stand"
	} otherwise: [
		^ false
	].
	^ true	"If we didn't reach the otherwise clause, we did a gesture."! !

!QAvatarUser methodsFor: 'driving' stamp: 'hrs 8/6/2010 14:23'!
driveStep: force
	| t qual |
	"Extended to add location service heartbeat."
	super driveStep: force.

	"Heartbeat the presence service; this is a slow heartbeat,
	 that allows the presence service to time-out disconnected clients."

	view owner serviceSvc ifNil: [ ^ self ].
	qual := self sessionQuality.
	view owner usersPane mySessionQuality: qual.
	view owner orgServices ifNotNilDo: [ :os |
		(((t := Time totalSeconds) - lastLocatorHeartbeat) >= QPresenceManager heartbeatInterval ) 
		 ifTrue: [		
			| doXML |
			doXML := false. 
			"Or the following to persist periodically. Turned off for now because 
			TPages/SmartRooms/Mobile need a stable time of entry -- not periodically updating --
			which it gets from the locator.xml persistence. See https://bugs.qwaq.com/view.php?id=2594
			doXML := (t - lastLocatorRecord) >= QPresenceManager xmlUpdateInterval."
			os recordMyLocation: (view owner currentForumURL) 
				quality:qual 
				updateXML: doXML.
			lastLocatorHeartbeat := t.
			doXML ifTrue: [ lastLocatorRecord := t ].
		].
	].
! !

!QAvatarUser methodsFor: 'driving' stamp: 'bgf 5/24/2009 12:25'!
driveSymbols
	"The selector symbol or symbols representing the currently depressed driving keys;
	 these are supplied by the view pane as determined by its drive-key mappings.
	 The collection is empty when no keyboard driving is happening."
	^ driveSymbols! !

!QAvatarUser methodsFor: 'driving' stamp: 'bgf 5/24/2009 12:25'!
driveSymbols: symCollection
	"The selector symbol or symbols representing the currently depressed driving keys;
	 these are supplied by the view pane as determined by its drive-key mappings.
	 The collection is empty when no keyboard driving is happening."
	driveSymbols := symCollection ifNil: [ #() ] ifNotNil: [ symCollection clone ].! !

!QAvatarUser methodsFor: 'accessing' stamp: 'das 3/6/2007 15:44'!
driverClass

	^ QAvatarDriver.! !

!QAvatarUser methodsFor: 'accessing' stamp: 'bgf 7/28/2009 23:00'!
effectiveThirdPersonLag
	"The camera-catchup lag to use for the current moment's camera updates.
	 We adjust this to supply lag we know we want (for cinematic motion)."

	driver isDriving ifTrue: [
		| tp |
		tp := self currentThirdPerson.
		(tp between: 0.5 and: 2.5)
		 ifTrue: [
			^ tp sqrt min: 1.3 max: 1.0		"Less character-winking in close third person driving."
		]].

	^ (super effectiveThirdPersonLag) max: 2! !

!QAvatarUser methodsFor: 'driving' stamp: 'bgf 4/25/2009 22:25'!
focalCameraTransform
	"If appropriate, return the override camera transform 
	 to use right now for this user's main view.
	 The focal camera transform, if non-nil, is a Global transform (not avatar-relative)
	 since it defines an exact focal viewpoint onto a destination.)
	 The value is provided by the replica, which signals changes as needed."

	(self inTransit or: [ self isDriving ]) ifTrue: [ ^ nil ].
	^ focalCameraTransform
! !

!QAvatarUser methodsFor: 'landmarks' stamp: 'bgf 7/18/2009 11:10'!
goToPlace: aPlace
	"Send the replica to the given place."
	self ifAllowedToMove: [
		(aPlace label = 'Home' and: [ aPlace category = 'Named Places']) 
		ifTrue: [
			self goToSpaceHome.		"Allows miramar to define this action with using a QPlace,
									and prevents arbitary markers intercepting the <home> key." 
		] ifFalse: [
			replica future goToPlace: aPlace
		]].

! !

!QAvatarUser methodsFor: 'landmarks' stamp: 'bgf 6/22/2008 12:08'!
goToSpaceHome
	"Go to the home position of the space we're in.
	 Differs from the super version in that it uses a Place
	 so the avatar knows it is at home rather than just near home."

	| home |
	self readyForEvents ifFalse:[ ^ self ].
	home := replica get: #homePlace.
	home ifNotNil: [
		replica future goToPlace: home.	"Not self goToPlace: as that can call into here."
	] ifNil: [
		"Fallback - must be in a non-Q-Space..."
		super goToSpaceHome.
	].
! !

!QAvatarUser methodsFor: 'status' stamp: 'bgf 1/12/2009 17:26'!
graphicSessionQuality
	"A scalar from 1 (poor) to 5 (best) indicating health of render rates..."
	| frameTime |
	currentSpace isValidFarRef ifFalse: [ ^ 0 ].

	"Normally we want the network speed to be the significant measure,
	 so we only put this below 5 if things are quite impaired."
	frameTime := [ self view owner controlBar recentRenderTime ] ifError: [ :msg | 0 ].
	frameTime ifNil: [ ^ 3 ].	"Don't force it way down, but don't claim perfection."
	frameTime > 200 ifTrue: [ ^ 1 ].
	frameTime > 160 ifTrue: [ ^ 2 ].
	frameTime > 120 ifTrue: [ ^ 4 ].		"I.e. don't say 5 in this case."
	^ 5
! !

!QAvatarUser methodsFor: 'editing' stamp: 'bgf 3/5/2010 16:35'!
hasPastedFrames
	"Returns true if Ithis user has a Ctl-C or Ctl-X copied-buffer of frames available to paste."
	^ pasteboardScrap notNil! !

!QAvatarUser methodsFor: 'landmarks' stamp: 'bgf 7/3/2009 14:45'!
homePlace
	"A default 'Home' landmark mapped to the space home position."
	^ self readyForEvents ifTrue: [
		replica get: #homePlace.
	] ifFalse: [
		nil
	].! !

!QAvatarUser methodsFor: 'private events' stamp: 'bgf 7/18/2009 10:40'!
ifAllowedToMove: block
	"Filtered check before sending go-to commands from me into the replica,
	 to honor movement restrictions for meetings."
	self owner ifNotNil: [ : ui | ui ifAllowedToMove: block ]! !

!QAvatarUser methodsFor: 'status' stamp: 'jjl 9/15/2008 09:16'!
informUser: string

	view owner signal: #userNeedsInformation with: string! !

!QAvatarUser methodsFor: 'initialize' stamp: 'bgf 8/6/2010 07:42'!
initialize
	super initialize.
	shiftSpeed := #normal.
	baseSpeed := 1.0.
	lastUpdateTime := 0.
	lastLocatorHeartbeat := Time totalSeconds.
	lastLocatorRecord := lastLocatorHeartbeat.
	activities := IdentitySet new.
	lastTypeTime := 0.
	qosTime := 0.
	driveSymbols := #().

	"Reminder - these rely on replica signals being automatically re-raised on me."

	self runScript: #costumeChanged when: {self . #costumeChanged }.
	self runScript: #onFocalTransformChanged: when: { self . #focalTransformChanged }.

	self startScript: #informUser: when: { self . #informUser: }.
	self startScript: #suggestPOV:forPlace: when: { self . #suggestPOV:forPlace: }.

	self startScript: #onCopyFrames: when: { self . #copyFrames }.
	self startScript: #onPasteFrames: when: { self . #pasteFrames }.
	self startScript: #onVideoRectangleClicked when: { self . #videoRectangleClicked }.

	self startScript: #recordAction:class:arguments:  when: { self . #recordAction:class:arguments: }.! !

!QAvatarUser methodsFor: 'portals' stamp: 'bgf 1/6/2010 12:01'!
installCostumeAndWait: bWait

	"Seed the user info first, so the costume has name/color info off the bat."
	userPreferences setUserInfoRef: (replica get: #userInfo) inClient: (view owner).

	"The inherited base does most of the costume setup work."
	super installCostumeAndWait: bWait.

	self stopScript: #addToInventory:.
	self startScript: #addToInventory: when: { replica . #inventoryCard }.

	self stopScript: #onBadgeDropped:.
	self startScript: #onBadgeDropped: when: { replica . #badgeDropped }.! !

!QAvatarUser methodsFor: 'locating' stamp: 'bgf 3/20/2009 10:51'!
isAtApplicationPlace
	"True if the avatar is at a place that focuses on an embedded application."
	replica isValidFarRef ifFalse: [ ^ false ].
	^ (replica get: #currentPlace) ifNotNil: [ : place | place hasApp ] ifNil: [ false ]! !

!QAvatarUser methodsFor: 'driving' stamp: 'bgf 11/21/2008 15:09'!
isDriving
	^ driver isDriving! !

!QAvatarUser methodsFor: 'testing' stamp: 'bgf 11/12/2009 12:27'!
isEditingTemplate
	"In template (full-scene) editing mode?"
	^ [ view owner isEditingTemplate ] on: Error do: [ false ].! !

!QAvatarUser methodsFor: 'accessing' stamp: 'das 6/27/2007 17:39'!
isSticky

	^ driver sticky.! !

!QAvatarUser methodsFor: 'events' stamp: 'bgf 3/5/2010 17:39'!
keyDown: evt
	"If a key stroke reaches the avatar, it is because no frame is in focus
	 that takes keystrokes; in this case, if the key was a motion key,
	 release stickiness on the avatar so that motion keys can reach the driving system."
	| action ui |
	(view keySymbolFor: evt) ifNotNil: [
		view keyboardDriving: true.
	].
	(ui := (self owner) ifNil: [ ^ self]) canModifyForum
	 ifTrue: [
		action := CPlatform current standardEditingKeys lookupEvent: evt.
		action caseOf: {
			[#cutSelection] -> [ self cutSelection: evt ].
			[#copySelection] -> [ self copySelection: evt ].
			[#pasteSelection] -> [ self pasteSelection: evt ].
			[#undo]			-> [ self undo: evt ].
			[#redo]			-> [ self redo: evt ].
			[#backspace]		-> [ self deleteSelection: evt ].
			[#forwardDelete]	-> [ self deleteSelection: evt ].
			[#delete]			-> [ self deleteSelection: evt ].

			" Command-S  does save - but not <enter> or other implicit 'accept' keys" 
			[#accept]		-> [ (evt controlKeyPressed or: [ evt commandKeyPressed ]) 
								 ifTrue: [ ui saveForum ] ].
		} otherwise: [
			super keyDown: evt.
		].
		action ifNotNil: [ QActivityCapture action: #EditKey details: action ].
	] ifFalse: [
		super keyDown: evt.
	].
 ! !

!QAvatarUser methodsFor: 'events' stamp: 'bgf 8/7/2009 00:52'!
keyStroke: event

	"Key Strokes to the avatar user are delivered directly by the view
	 to support specialized driving commands."

	driver isDriving ifTrue: [
		"We don't do anything while the harness is driving us;
		 driving is managed in the harness, mostly through keyDown/Up events."
		^ self.
	].

	"Interpret scroll wheel keyStrokes; the modifiers vary by platform,
	 so we give only one interpretation: zoom in and out the viewpoint."

	(view keyStrokeSymbolFor: event) caseOf: {
		[#scrollUp] -> [ 
			event controlKeyPressed ifTrue:[
				self zoomIn.
				^ self.
			].
		].
		[#scrollDown] -> [
			event controlKeyPressed ifTrue:[ 
				self zoomOut.
				^ self.
			].
		].
	} otherwise: [].

	"Fallthrough - not a scroll-wheel remapped event."

	(self doGestureForKeystroke: event) ifFalse: [
		"Pseudo-driving keys that have no corresponding nav-panel controls,
	       and so are not part of the main drive-key mapping system. "
		event keyValue = 4 " END" ifTrue: [ 
			self ifAllowedToMove: [ replica future goToBirdsEye ].  
			^ self ].
		event keyCharacter = $	 "tab" ifTrue: [ 
			view owner goToNextPlace: event shiftPressed not.
			^self ].
		event keyCharacter = $? ifTrue: [
			view owner controlBar toggleAskToSpeak.
			^ self ].
		event keyCharacter = Character cr ifTrue: [
			self signal: #autoChat.
			^ self].
		view isReadonlyUser ifFalse: [
			super keyStroke: event.	
		].
	].! !

!QAvatarUser methodsFor: 'events' stamp: 'bgf 6/21/2008 00:31'!
keyUp: evt

	"For consistency - enforce the readonly guard against events into the space.
	 The avatar does this for all key-events to permit some control finesse."
	view isReadonlyUser ifFalse: [
		super keyUp: evt.
	].
! !

!QAvatarUser methodsFor: 'driving' stamp: 'bgf 7/22/2010 15:54'!
locateFrame: aFrame
	"Support for various 'Locate' menu commands'.
	 Bring the avatar to look at the given frame, and select it for emphasis."
	(aFrame isValidFarRef 
	 and: [ self readyForEvents ] 
	 and: [ aFrame island = replica island ]) ifTrue: [
			replica future locateFrame: aFrame
	].
		! !

!QAvatarUser methodsFor: 'driving' stamp: 'bgf 7/12/2009 17:00'!
lookDown
	"Drive key mapping."
	self pitch > -60 ifTrue: [
		self addPitch: (-2 * driver checkClutch)
	].! !

!QAvatarUser methodsFor: 'driving' stamp: 'bgf 5/8/2009 00:51'!
lookModeMouseDriving: bFlag
	"Do nothing; we use a different means (the QLookTask) for camera-steering."! !

!QAvatarUser methodsFor: 'driving' stamp: 'bgf 7/12/2009 17:00'!
lookUp
	"Drive key mapping."

	(self pitch < 60) ifTrue: [
		self addPitch: 2 * (driver checkClutch)
	].! !

!QAvatarUser methodsFor: 'private events' stamp: 'bgf 8/3/2009 13:16'!
minStepInterval
	"The minimum milliseconds between my pulse: messages."
	| ui  base |

	view ifNil: [ ^ 1000 ].	"Don't care."
	ui := view owner ifNil: [ ^ 1000 ].
	base := self class 
				currentStepIntervalForUsers: (ui usersPane nPeopleHere) 
				favored: ( ui isMeetingLeader 
						  or: [ ui meetingState hasLoginPermissionToSpeak: (ui userLogin) ])
				moving: (ui allowedToMove).

	"If we know we're having network lag, 
	 don't exacerbate the problem with clickety 'what's up messages."
	self heartbeatLag > 2  "Seconds"    ifTrue: [
		^ (base * 4) atLeast: 2000 atMost: 3000
	].
	^ base! !

!QAvatarUser methodsFor: 'driving' stamp: 'bgf 5/27/2007 23:25'!
moveBack
	"Drive key mapping."

	self moveBack: 1! !

!QAvatarUser methodsFor: 'driving' stamp: 'bgf 5/27/2007 23:25'!
moveDown
	"Drive key mapping.  There's usually gravity also, so go slower than usual."

	self moveDown: 0.5! !

!QAvatarUser methodsFor: 'driving' stamp: 'bgf 5/27/2007 23:25'!
moveForward
	"Drive key mapping."

	self moveForward: 1! !

!QAvatarUser methodsFor: 'driving' stamp: 'bgf 5/27/2007 23:24'!
moveLeft
	"Drive key mapping."

	self moveLeft:1! !

!QAvatarUser methodsFor: 'driving' stamp: 'bgf 5/27/2007 23:24'!
moveRight
	"Drive key mapping."

	self moveRight:1! !

!QAvatarUser methodsFor: 'driving' stamp: 'bgf 5/27/2007 23:24'!
moveUp
	"Drive key mapping... we have to go fast enough to overcome gravity."
	self moveUp: (1.2)! !

!QAvatarUser methodsFor: 'status' stamp: 'bgf 1/12/2009 17:27'!
networkDelayQuality
	"A scalar from 1 (poor) to 5 (best) indicating how free-of-lag 
	 this user's session appears to be."
	| lag |
	currentSpace isValidFarRef ifFalse: [ ^ 0 ].

	"We want the network jitter to be the dominant measure,
	 since it most accurately predicts voice quality except when other factors are horrible...
	 so we only put this below 5 if things are quite impaired."
	lag := [ self view owner controlBar recentLatency ] ifError: [ :msg | 0 ].
	lag ifNil: [ ^ 3 ].	"Don't force it way down, but don't claim perfection."

	lag > 360 ifTrue: [ ^ 1 ].
	lag > 390 ifTrue: [ ^ 2 ].	
	lag > 220 ifTrue: [ ^ 4 ].		"I.e. not 5 if this goes on."
	^ 5
! !

!QAvatarUser methodsFor: 'status' stamp: 'bgf 7/26/2009 11:44'!
networkJitterQuality
	"A scalar from 1 (poor) to 5 (best) indicating how responsive/up-to-date/good/happy
	 this user's Forums session appears to be.  
	 Advertised to all users (bars by the name in the people pane).
	 0 generally means no data (i.e. zero signal bars.)"
	| croquetRoundTripJitter samples stats |
	currentSpace isValidFarRef ifFalse: [ ^ 0 ].

	stats := (currentSpace controller jitterStats).
	stats ifNil: [ ^ 5 ]. "Local controller."
	samples := (stats subWindowOfDuration: 5000).
	samples ifNil: [ ^ 3 ].	"Don't force it way down, but don't claim perfection."
	samples size < 3 ifTrue: [ ^ 2 ].
	croquetRoundTripJitter := samples allButFirst average asFloat.

	"Ramp maps jitter-measure into an arbitrary health-bar level (0...5) 
	 Sound quality seems to collapse when jitter exceeds
	 the jitter-buffer margin (e.g. 320ms),
	 so crossing that level for general croquet-network jitter 
	 means the overall session can fairly be blamed for poor sound.  
	 (If sound is poor when overall jitter isn't so high, 
	 then we'd want to look for more specific causes.)"
	croquetRoundTripJitter < 30   ifTrue: [ ^ 5 ].  "Eggcellent"
	croquetRoundTripJitter < 60   ifTrue: [ ^ 4 ].
	croquetRoundTripJitter < 120 ifTrue: [ ^ 3 ].
	croquetRoundTripJitter < 240 ifTrue: [ ^ 2 ].
	croquetRoundTripJitter < 480 ifTrue: [ ^ 1 ]. "Very bad."
	^ 0	"Horrid."! !

!QAvatarUser methodsFor: 'events' stamp: 'bgf 11/24/2009 13:43'!
onBadgeDropped: aTexture
	
	"The avatar has signalled that a badge texture was dropped upon it.
	We need to record the image for re-use as the avatar moves among spaces."

	| form data sha |

	sha := aTexture get:#sha.
	data := TContentManager default cacheManager find: (sha value).
	form := [Form fromBinaryStream: (ReadStream on: data)] 
		on: Error do:[:ex|
			self signal: #alertMessage with: 'That image cannot be used as a badge.'.
			^ self
		].

	"For now, just do it.  We should confirm that this is acceptable,
	 by opening the My Avatar tool.
	 We keep these images small to avoid lags on space-changes."
	[  
		userPreferences badgePicture: (form ifNotNilDo: [:f | f scaledToSize:96@96]).
		view owner commitPreferencesInBackground.

	] on: Error do: [ :ex|
		self owner inform: 'A badge was dropped on your avatar,\but could not be saved:\\' withCRs ,
						    (ex description)
	].
! !

!QAvatarUser methodsFor: 'editing' stamp: 'bgf 2/7/2010 08:35'!
onCopyFrames: frameCaptureList
	"Callback when a frame is copied (or cut) in the scene.
	 The cfL is  a QCapturedFrameList,  which is pass-by-copy from the island."
	pasteboardScrap := frameCaptureList
! !

!QAvatarUser methodsFor: 'private events' stamp: 'bgf 6/14/2009 20:12'!
onFocalTransformChanged: xformOrNil
	"From the replica - a new relativeCameraTransform basis for a focal viewpoint."
	
	(self isDriving and: [ xformOrNil notNil]) ifTrue: [
		QLogger logInfo: 'FT received while driving.'
	].
	focalCameraTransform := xformOrNil
! !

!QAvatarUser methodsFor: 'private events' stamp: 'ar 4/16/2011 17:40'!
onPasteFrames: targetParent 
	"In response to pasting frames....
	 if we're pasting into different forums than the frames were copied from,
	 then ensure the resources are made available in the new site."
	| owner docRenames | 
	self readyForEvents ifFalse: [ ^ self ].
	(owner := view owner) ifNil: [ ^ self ].
	pasteboardScrap ifNil: [ ^ self ].
	docRenames := #().

	(view owner urlCacheAt: (pasteboardScrap serviceURL) ifAbsent: [ nil ]) 
	ifNotNil: [ : svc |
		svc = (view owner serviceSvc) ifFalse: [
			| reslist missed |
			reslist := pasteboardScrap resourceList.
			reslist isEmptyOrNil ifFalse: [
				reslist do: [ : res | res serviceURL: (svc url) ].
			 	owner mouseCursor: Cursor wait.
				[
					missed := view session
						ensureCopyResources: (reslist) 
						fromService: svc
						toService: (view owner serviceSvc).
					docRenames := view session
						ensureDocuments: reslist
						fromService: svc
						toService: svc.
				] ensure: [
					owner mouseCursor: nil.
				].
				missed isEmptyOrNil ifFalse: [
					view owner 
						inform: 'Pasted items require resources that could not be copied.\See the transcript for details.' withCRs.
					^ self
				].
		]]].

	(self readyForEvents and: [ replica island = targetParent island ]) 
	ifTrue: [
		replica future doPasteIntoParent: targetParent withDocumentRenames: docRenames.
	].! !

!QAvatarUser methodsFor: 'events' stamp: 'bgf 5/29/2008 18:40'!
onVideoRectangleClicked

	"  <on: videoRectangleClicked >  signaled by my replica,
	   i.e. my user has clicked on some video surface.  
	   Check and re-enable video if necessary since the 
	   disabled video replacement texture says 'click to enable' "

	view readyForEvents ifTrue: [
		view owner controlBar checkVideoSession.
	].! !

!QAvatarUser methodsFor: 'accessing' stamp: 'bgf 7/18/2009 10:23'!
owner
	"The participant UI I operate in."
	^ [ view window owner ] on: Error do: [ nil ]! !

!QAvatarUser methodsFor: 'editing' stamp: 'bgf 2/6/2010 21:54'!
pasteSelection: evt

	self readyForEvents ifTrue: [
		replica future pasteSelection: (view event).
	].

! !

!QAvatarUser methodsFor: 'private events' stamp: 'bgf 8/6/2009 12:00'!
pointerStepInterval
	"Min milliseconds between pointer-motion drive step events - 
	caps the rate at which faster machines flood the time stream with drive steps
	that involve only pointer motion, not avatar position or posture changes."

	view ifNil: [ ^ 500].	"Don't care."

	"We have ramping rules for population control; use them."
	^ self minStepInterval ! !

!QAvatarUser methodsFor: 'portals' stamp: 'bgf 6/18/2008 00:25'!
postContextMenu

	"We disallow all context menus for readonly users."
	view isReadonlyUser ifFalse: [
		super postContextMenu
	].! !

!QAvatarUser methodsFor: 'testing' stamp: 'bgf 3/23/2010 10:17'!
prepareRayTest

	"If at or in-transit to a Place, we're in a logical go to, and do not consider or allow portal-crossing.
	 We do the super-work first because it updates my portalRay and assumes a span since last check."
	^ super prepareRayTest and: [  (replica get: #currentPlace) isNil ]
! !

!QAvatarUser methodsFor: 'accessing' stamp: 'bgf 6/22/2009 18:07'!
publicName	
	"My user's public name string."
	^ userPreferences publicName! !

!QAvatarUser methodsFor: 'editing' stamp: 'bgf 3/25/2010 12:08'!
recordAction: actionMakerSelector class: klass arguments: arguments

	"Support for registering an undoable action that fired on-island.
	 On-island widgets send this message to the avatar-replica;
	 it is relayed and performed here.   The ActionMakerSelector must be implemented on klass,
	 and must yield a new QUndoableAction,  which we register in the user's undo list."
	
	| undoable |
	self owner ifNotNilDo: [ : ui |
		undoable := klass perform: actionMakerSelector withArguments: (arguments).
		undoable ifNotNil: [ 
			ui recordAction: undoable.
		].
	].! !

!QAvatarUser methodsFor: 'locating' stamp: 'bgf 6/21/2009 10:08'!
recordLocation

	"Triggered by a space change.  Tell the service manager which service I'm in.
	 This creates a record of my most recent forum entry.
	 The users / services panes use this information, but only after a coarse
	 service-list check that a user is indeed logged in
	 ... so there's no need to 'clear' the record on logout."

	view ifNil: [ ^ self ].
	view owner ifNil: [ ^ self ].
	view owner serviceSvc ifNil: [ ^ self ].
	view owner orgServices ifNotNilDo: [ : orgsvc |		
		orgsvc recordMyLocation: (view owner currentForumURL) 
			     quality: (self sessionQuality) 
			     updateXML: true.
	].
! !

!QAvatarUser methodsFor: 'editing' stamp: 'bgf 3/13/2010 11:12'!
redo: evt
	"Edit-action - the user has invoked 'Redo', probably by hitting command-shift-Z "
	self owner ifNotNil: [ : ui | 
		ui mouseCursor: Cursor wait.
		[  ui undoRedoList redoOneForClient: ui ] ensure: [ ui waitTick; mouseCursor: nil ]
	].
! !

!QAvatarUser methodsFor: 'driving' stamp: 'bgf 4/25/2009 11:48'!
relativeCameraTransform
	"The relative transform between my origin and my camera basis position
	 (i.e. before third-person zoom adjustments are applied.)
	 This is the identity matrix, except when I have an overriding focal transform."
	self focalCameraTransform ifNotNilDo: [ : cft |
		^ localTransform inverseByTranspose * cft
	].
	^ Matrix4x4 theIdentity! !

!QAvatarUser methodsFor: 'status' stamp: 'bgf 6/2/2009 17:20'!
resetActivities: symCollection
	"Replace #activities with the given collection,
	 passing along to the replica regardless of whether we think there is agreement.
	 Done after reconnect, teleport, etc, to ensure agreement."

	activities := symCollection asSet.
	self readyForEvents ifTrue: [
		replica future setActivities: activities asArray.
	].! !

!QAvatarUser methodsFor: 'driving' stamp: 'das 3/6/2007 18:41'!
resetLook
	"Drive key mapping."
	self pitch: 0.! !

!QAvatarUser methodsFor: 'driving' stamp: 'bgf 7/18/2009 10:43'!
resetPosition
	"Drive key mapping."
	self ifAllowedToMove: [
		self goToSpaceHome
	]! !

!QAvatarUser methodsFor: 'driving' stamp: 'bgf 7/27/2009 10:41'!
restoreThirdPersonAfter: stickyViewpoint
	"Restore the viewpoint I had just before being sticky-zoomed,
	 or my own thirdPerson preference, whichever is wider.
	 (I.e. honor the user preference unless they'd already taken a wider view;
	 it is more pleasant to be zoomed out than in when leaving a focus place)."
	| tp |
	tp := thirdPerson.  
	tp < 1.0 ifTrue: [ tp := 0.0 ].		"Repair slop at near-fp."

	self userPreferences povValue ifNotNilDo: [ : v |
		self thirdPerson: (v max: tp).
	] ifNil: [
		super restoreThirdPersonAfter: stickyViewpoint
	].
! !

!QAvatarUser methodsFor: 'events' stamp: 'bgf 1/4/2009 15:44'!
rightMouseMove: event
	[
		super rightMouseMove: event.
		"We're in regular driving; make sure the dot is up."
		view owner viewPanel mouseCursor: QStyles drivingCursor
	] 
	on: Error do: []! !

!QAvatarUser methodsFor: 'events' stamp: 'bgf 11/12/2009 12:27'!
rightMouseUp: event
	"Yellow/right mouse release (end of driving cycle, or just a click.)"

	view driveSymbols ifNil: [
		"Always stop this here, in case we're in a run-on drive from yellow-down leak-in.
	       The user reflexively clicks the yellow button to stop that, so make that work.
	   	 (The TAvatarUser doesn't consider a click-release a stop because that's for menus.)"
		self stopDriving.
		self lookModeMouseDriving: false.
	].
	[
		event type ~= #mouseUp ifTrue: [		
			"Just clear any mouse-down state.  (Sent for -leave-, etc.)"
			self stopDriving.
			view owner viewPanel mouseCursor: nil.
			pendingDriveRequestTime := nil.
			^ self 
		].

		"If it's only a click, check to see if we're in a warp situation;
		 if so, warp here instead of letting the default context menu appear,
		 for consistency - at > warp distance, we go-to- rather than interact."
		( pendingDriveRequestTime notNil and: [ self isEditingTemplate not ] )
		 ifTrue: [
			(replica get: #warpPlaceUnderMouse: with: view event) ifNotNilDo: [ :place |
				view owner viewPanel mouseCursor: nil.
				pendingDriveRequestTime := nil.
				self goToPlace: place.
				^ self
			].
		].
		view owner viewPanel mouseCursor: nil.

		"And go on to check for normal 'context menu' versus stop cases."
		super rightMouseUp: event.

	] on: Error do: [ :ex | 
		QLogger logWarning: 'Exception in right-up: ', ex
	].

! !

!QAvatarUser methodsFor: 'driving' stamp: 'bgf 5/27/2007 23:29'!
rotateLeft

	"Drive key mapping."
	self addYaw: (3 * self speed).! !

!QAvatarUser methodsFor: 'driving' stamp: 'jdm 8/22/2007 16:30'!
rotateRight
	"Drive key mapping."
	self addYaw: (-3 * self speed).! !

!QAvatarUser methodsFor: 'costume' stamp: 'bgf 8/4/2010 12:08'!
saveAvatarCostume

	"The user has requested that the current avatar (live in-space) be saved for re-use.
	 Returns true on success, false on failure.
	 (Failures are reported as inform:s here before returning false."
	| data |

	"First walk the in-space costume and copy resources to my per-user public store
	 so they'll become available for others in any space I visit. 
	 This happens first because it does in-place changes to the thumbs in space."

	(view session copyResourcesToUserForFrame: (replica get: #costume)) ifFalse: [ ^ false ].

	view owner statusInfo: 'Saving avatar...'; waitTick.

	"Now grab the persistent snapshot of the replica and save it upstream.
	 First we need to synchronize, as resource-repathing uses future sends 
       to update the costume texture file paths to point to my file space."
	(replica future yourself waitTimeoutSeconds: 30) ifFalse: [ 
		self inform: 'The avatar was not saved.\The operation timed out.\\Please try again.' withCRs. 
		^ false
	].
	replica isValidFarRef ifTrue: [
		data := (replica future costumeData) wait; value.
	].
	data ifNotNil: [
		userPreferences avatarCostumeData: data value.
		costumeXMLZip := data.
		^ true
	] ifNil: [
		self inform: 'The avatar was not saved.\There were communication problems.\\Please try again.' withCRs. 
		^ false.
	].! !

!QAvatarUser methodsFor: 'events' stamp: 'bgf 3/8/2008 22:23'!
saveInventory: inventoryZip

	"<on inventoryChanged in replica>"

	userPreferences inventoryXMLZip: inventoryZip! !

!QAvatarUser methodsFor: 'editing' stamp: 'bgf 3/5/2010 14:31'!
selectionSummary
	"String summarizing the selection briefly"
	^ replica ifNotNil: [ replica get: #selectionSummary ]! !

!QAvatarUser methodsFor: 'status' stamp: 'jcg 3/17/2009 08:54'!
sessionQuality
	"A scalar from 1 (poor) to 5 (best) indicating how responsive/up-to-date/good/happy
	 this user's Forums session appears to be.  
	 Advertised to all users (bars by the name in the people pane).
	 0 generally means no data (i.e. zero signal bars.)"

	"Take the worst measure.  
	 The graphic and network delay measures
	 are filtered so they only become 'worst' if there's a serious problem.
	 Usually, we want a reflection of connection jitter...
	 since that's the best predictor of voice quality.
	 But in extremis, the other problems need also to be indicated since they can
	 arise without much increase in jitter."

	"Taking the stats is expensive, do it only every so often."
	(qos notNil 
	 and: [ (Time coarseMillisecondsSince: qosTime) < 500 ]) 
	 ifTrue: [ ^ qos ].
	qosTime := Time coarseClockValue.
	^ qos := (self networkJitterQuality 
	    			min: (self graphicSessionQuality))
	    			min: (self networkDelayQuality)! !

!QAvatarUser methodsFor: 'status' stamp: 'bgf 4/26/2008 18:47'!
setActivities
	"Update the replica activities to match my current set;
 	 (e.g. done after reconnect to ensure we're in sync."
	self readyForEvents ifTrue: [
		replica future setActivities: activities asArray.
	].! !

!QAvatarUser methodsFor: 'status' stamp: 'bgf 4/26/2008 18:47'!
setActivities: symCollection
	"Replace #activities with the given collection.
	  (what matters is doing it in the replica; 	
 	  the user-side copy is here to economize future-sends."
	| symSet |
	symSet := symCollection asSet.	
	(activities = symSet) ifTrue: [ ^self ].
	activities := symSet.
	self readyForEvents ifTrue: [
		replica future setActivities: activities asArray.
	].! !

!QAvatarUser methodsFor: 'accessing' stamp: 'bgf 5/27/2007 22:20'!
shiftSpeed: highMedLow

	"A symbol - #high for fast, #low for slow."

	shiftSpeed := highMedLow
! !

!QAvatarUser methodsFor: 'accessing' stamp: 'bgf 7/12/2009 17:19'!
speed
	"Basic speed factor (ordinary value is 1.0) for a driving step.
	 Affects the clutch; call once per drive-step."
	^ (driver checkClutch)
	   *  (shiftSpeed caseOf: {
		[#high] -> [ 5 * baseSpeed ] .
		[#low] -> [ 0.3 * baseSpeed ]
		} otherwise: [ baseSpeed ] ) 
	
! !

!QAvatarUser methodsFor: 'status' stamp: 'bgf 4/26/2008 18:47'!
startActivity: aSymbol
	"Add the given activity to #activities.
	  (what matters is doing it in the replica; 	
 	  the user-side copy is here to economize future-sends."
	(activities includes: aSymbol) ifTrue: [ ^ self ].
	activities add: aSymbol.
	self readyForEvents ifTrue: [
		replica future startActivity: aSymbol.
	].! !

!QAvatarUser methodsFor: 'accessing' stamp: 'bgf 5/24/2009 12:01'!
startDriving

	super startDriving.
	view resetToNormalDriving.
	self stopActivity: #Typing.
	focalCameraTransform := nil.

	QActivityCapture countAction: (
		driveSymbols isEmptyOrNil 
				ifTrue: [ #StartDrivingMouse ]
				ifFalse: [ #StartDrivingKeyboard ] ).
! !

!QAvatarUser methodsFor: 'status' stamp: 'bgf 4/26/2008 18:47'!
stopActivity: aSymbol
	"Remove the given activity from #activities 
	  (what matters is removing it in the replica; 	
 	  the user-side copy is here to economize future-sends."
	(activities includes: aSymbol) ifFalse: [ ^ self ].
	activities remove: aSymbol.
	self readyForEvents ifTrue: [
		replica future stopActivity: aSymbol.
	].! !

!QAvatarUser methodsFor: 'accessing' stamp: 'bgf 7/28/2009 22:42'!
stopDriving

	driver isDriving ifTrue: [
		QActivityCapture countAction: #StopDriving.
	].
	self cameraInLockStep: false.
	super stopDriving.
	self stopActivity: #Typing.	"In case keys overlap."! !

!QAvatarUser methodsFor: 'accessing' stamp: 'bgf 11/14/2009 22:26'!
stopMouseDriving
	"If we are only mouse driving... stoppit."
	driveSymbols ifEmpty: [ self stopDriving ]! !

!QAvatarUser methodsFor: 'landmarks' stamp: 'bgf 3/19/2010 18:50'!
suggestPOV: aNameString forPlace: aPlace
	"Change the user's viewpoint to the given option.
	 The options are those in the participant viewpoint menu,
	 and in the QUserPreferences ViewpointTable.
	 These viewpoint changes will probably be optional 
	 (via a preference to disallow them) in the future,
	 hence the 'suggest' in the name."

	| pov |
	aNameString = #Preferred ifTrue: [
		pov := (self owner preferences pointOfView).
	] ifFalse: [
		pov := aNameString.
	].

	aPlace ifNotNil: [
		driver stickyViewpoint: pov
	] ifNil: [
		(view notNil and: [ view owner notNil ]) ifTrue: [
			view owner changeViewpoint: pov
		]
	].! !

!QAvatarUser methodsFor: 'editing' stamp: 'bgf 7/25/2010 15:27'!
texturePalette

	"Returns the array of TFileTextures that describe the textures 
	 to make available for selection in Material editing pane.
	 The list is partly derived from resources and documents already in this forum,
	 and partly derived from the QTextures palette (builtin suite of textures)"
	
	| textures suite |

	"Find all the routine resource file-textures in the island, 
	 and include them in the material editor textures list."

	textures := (self readyForEvents ifTrue: [ replica future texturePalette wait; value ] ifFalse: [ #() ]).
	suite := QTextures palette values asOrderedCollection.
	
	^ (suite,  textures) asArray 
	! !

!QAvatarUser methodsFor: 'driving' stamp: 'bgf 9/1/2009 15:39'!
turnTo: aPosition
	"Turn to face in the direction of the given global position (only in the x-z plane)"

	| eyePos bearing xform pyr |
	self readyForEvents ifFalse: [^ nil ].	
	aPosition ifNil: [ ^ nil ].

	driver posture = #Sitting ifTrue: [
		"Sitting implies look-mode, so simply look that way."
		pyr := self pitchYawRoll.
		eyePos := (localTransform * (self cameraViewpoint)) translation.
		bearing := (eyePos x @ eyePos z) 
					bearingToPoint: (aPosition x @ aPosition z).
		xform := Matrix4x4
					translation: (Vector3 zero) 
					pitchYawRoll: 0@ (bearing negated - (pyr y)) @ 0.
		self onFocalTransformChanged: (localTransform * xform ).
	] ifFalse: [
		"If we need motion, let the replica handle it in-space."
		replica future turnTo: aPosition
	].! !

!QAvatarUser methodsFor: 'accessing' stamp: 'ar 4/24/2006 16:26'!
tweakHand
	^tweakHand! !

!QAvatarUser methodsFor: 'accessing' stamp: 'ar 4/24/2006 16:26'!
tweakHand: aHand
	tweakHand := aHand! !

!QAvatarUser methodsFor: 'events' stamp: 'bgf 11/21/2008 15:10'!
typingActivity
	"The user hit a key on the keyboard."
	self isDriving ifFalse: [
		self startActivity: #Typing.
		lastTypeTime := Time millisecondClockValue.
	].! !

!QAvatarUser methodsFor: 'editing' stamp: 'bgf 3/18/2010 10:55'!
undo: evt
	"Edit-action - the user has invoked 'Undo', probably by hitting command-Z "
	self owner ifNotNil: [ : ui | 
		ui mouseCursor: Cursor wait.
		[  ui undoRedoList undoOneForClient: ui ] ensure: [ ui waitTick; mouseCursor: nil ]
	].
! !

!QAvatarUser methodsFor: 'accessing' stamp: 'das 3/12/2007 11:54'!
updateBaseSpeed

	| updateTime fps |
	updateTime := Processor activeIsland millisecondClockValue.
	(lastUpdateTime > 0) ifTrue: [
		
			"Calibrate baseSpeed adjustment assuming a default driving rate for 30 FPS."
			fps := (1000.0 / ((updateTime - lastUpdateTime) min:150 max: 25)).
			self baseSpeed: ((30.0 / fps) min: 3.0 max: 0.8).
	].
	lastUpdateTime := updateTime.
! !

!QAvatarUser methodsFor: 'accessing' stamp: 'bgf 10/26/2006 13:43'!
userInfo
	^ userPreferences userInfo! !

!QAvatarUser methodsFor: 'accessing' stamp: 'bgf 3/29/2007 17:00'!
userPreferences
	^ userPreferences! !

!QAvatarUser methodsFor: 'accessing' stamp: 'bgf 9/13/2007 21:09'!
userPreferences: prefs

	userPreferences := prefs.! !

!QAvatarUser methodsFor: 'private events' stamp: 'ar 1/8/2008 17:40'!
valueWithEvent: anEvent
	"more temporary evilness"
	Processor activeProcess hand: tweakHand.
	^self privateSignalEvent: (anEvent class new
		setSender: self selector: anEvent selector arguments: anEvent arguments
	)! !

Object subclass: #QLogger
	instanceVariableNames: 'queue process baseFileName mutex lastLogTimes'
	classVariableNames: 'ActivityLog Default DoPanelLog ErrorLog InfoLog LastTraceLog LogDirectory TraceLogs WarningLog'
	poolDictionaries: 'EventSensorConstants'
	category: 'Qwaq-Objects-Logging'!

!QLogger commentStamp: '<historical>' prior: 0!
(c) 2007 Qwaq Inc.  Logs to the specified file.!

!QLogger class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QLogger class methodsFor: 'default' stamp: 'ar 8/6/2008 16:46'!
activityLog
	"The log used for logging activity records"
	^ActivityLog ifNil:[self default].! !

!QLogger class methodsFor: 'default' stamp: 'ar 8/6/2008 12:31'!
activityLog: aLogger
	"The log used for logging activity records"
	ActivityLog := aLogger.! !

!QLogger class methodsFor: 'default' stamp: 'ar 8/6/2008 16:27'!
default
	Default ifNil: [
		"By default, we log to the Transcript."
		Default := QLoggerToTranscript new.
	].
	^Default! !

!QLogger class methodsFor: 'default' stamp: 'ar 8/6/2008 18:46'!
default: aLogger
	Default := aLogger! !

!QLogger class methodsFor: 'log api' stamp: 'das 1/5/2009 10:42'!
doPanelLog

	^DoPanelLog.! !

!QLogger class methodsFor: 'log api' stamp: 'das 1/5/2009 10:43'!
doPanelLog: bool

	DoPanelLog := bool.! !

!QLogger class methodsFor: 'log api' stamp: 'bgf 9/23/2009 14:39'!
drainLogger
	"Make an effort to get the log queue flushed to the current logging destination."
	^ self default drainLogger! !

!QLogger class methodsFor: 'default' stamp: 'ar 8/6/2008 16:46'!
errorLog
	"The log used for logging errors"
	^ErrorLog ifNil:[self default]! !

!QLogger class methodsFor: 'default' stamp: 'ar 8/6/2008 12:31'!
errorLog: aLogger
	"The log used for logging errors"
	ErrorLog := aLogger.! !

!QLogger class methodsFor: 'utilities' stamp: 'bgf 5/11/2008 16:25'!
evtStringEmbedded: evtBuf
	"Return string summarizing TEmbeddedApp event-buffer."

	^ evtBuf first caseOf: { 
		[#keyDown] -> [ 'key down ', evtBuf fifth ].
		[#keyUp] -> [ 'key up ', evtBuf fifth ].
		[#keystroke ] -> [ 'key stroke ', (evtBuf fifth asCharacter) ].
		[#mouseDown ] -> [ 'mouse down ', (evtBuf fifth), ' - ', (evtBuf third) ] .
		[#mouseUp ] -> [ 'mouse up ', (evtBuf fifth), ' - ', (evtBuf third) ] .
		[#mouseMove ] -> [ 'mouse move ', (evtBuf fifth), ' - ', (evtBuf third) ] .
		[#mouseOver ] -> [ 'mouse over ', (evtBuf fifth), ' - ', (evtBuf third) ] .
	} otherwise: [
		'unknown e-app event ', (evtBuf first)
	].
	! !

!QLogger class methodsFor: 'utilities' stamp: 'bgf 5/11/2008 16:00'!
evtStringMorphic: evtBuf
	"Return string summarizing a morphic evt buffer."
	^ evtBuf first caseOf: {
		[EventTypeMouse] -> [
			'mouse ', (evtBuf at:3), '@', (evtBuf at: 4), ' - ', ((evtBuf at: 5)).
			].
		[EventTypeKeyboard] -> [
			evtBuf fourth caseOf: {
				[EventKeyChar ]-> [ 'key stroke ', ((evtBuf third) asCharacter)] .
				[EventKeyDown] -> [ 'key down ', (evtBuf third) ] .
				[EventKeyUp] -> [ 'key up ', (evtBuf third) ].
			} otherwise: [ 'keyboard event' ]].

		} otherwise: [
			'Unknown event'
		].		! !

!QLogger class methodsFor: 'default' stamp: 'ar 8/6/2008 16:46'!
infoLog
	"The log used for logging extra information"
	^InfoLog ifNil:[self default]! !

!QLogger class methodsFor: 'default' stamp: 'ar 8/6/2008 12:32'!
infoLog: aLogger
	"The log used for logging extra information"
	InfoLog := aLogger! !

!QLogger class methodsFor: 'class initialization' stamp: 'jcg 2/11/2009 15:33'!
initialize
	"QLogger initialize."
	Smalltalk addToShutDownList: self.
	TLogger default: QLogger. "Use QLogger by default"
	DoPanelLog := false.
	self reset. ! !

!QLogger class methodsFor: 'logging' stamp: 'jcg 9/20/2007 23:06'!
log: aStringOrBlock
	QLogger default log: aStringOrBlock.
	^nil! !

!QLogger class methodsFor: 'logging' stamp: 'jcg 9/20/2007 23:06'!
log: aStringOrBlock from: anObject
	"See QLogger>>log:from:"
	QLogger default log: aStringOrBlock from: anObject.
	^nil! !

!QLogger class methodsFor: 'log api' stamp: 'ar 8/6/2008 17:47'!
logActivity: action user: userName organization: orgName host: hostName extra: extraString
	"Log to the default info log"
	self activityLog logActivity: action user: userName organization: orgName host: hostName extra: extraString! !

!QLogger class methodsFor: 'class initialization' stamp: 'ar 8/9/2008 12:10'!
logDirectory
	"Answer the default log directory"
	^LogDirectory ifNil:[FileDirectory logDirectory]! !

!QLogger class methodsFor: 'class initialization' stamp: 'ar 8/24/2007 15:09'!
logDirectory: aFileDirectory
	"Answer the default log directory"
	LogDirectory := aFileDirectory.
	LogDirectory ifNotNil:[LogDirectory assureExistence].! !

!QLogger class methodsFor: 'log api' stamp: 'ar 8/7/2008 16:34'!
logError: errorOrString
	"Log to the default error log"
	^self logError: errorOrString from: thisContext sender receiver! !

!QLogger class methodsFor: 'log api' stamp: 'ar 8/12/2008 11:32'!
logError: errorOrString details: debugInfo from: anObject
	"Log to the default error log"
	self errorLog logError: errorOrString details: debugInfo from: anObject.! !

!QLogger class methodsFor: 'log api' stamp: 'ar 8/7/2008 16:34'!
logError: errorOrString from: anObject
	"Log to the default error log"
	self errorLog logError: errorOrString from: anObject.! !

!QLogger class methodsFor: 'log api' stamp: 'ar 8/7/2008 16:34'!
logInfo: aString
	"Log to the default info log"
	^self logInfo: aString from: thisContext sender receiver! !

!QLogger class methodsFor: 'log api' stamp: 'jcg 2/11/2009 15:16'!
logInfo: aString every: msecs
	"Log to the default info log"
	| tag |
	tag := thisContext selector -> thisContext sender receiver.
	(self infoLog isReadyToLog: tag interval: msecs) ifFalse: [^self "wait longer"].
	^self logInfo: aString from: thisContext sender receiver! !

!QLogger class methodsFor: 'log api' stamp: 'ar 8/6/2008 16:45'!
logInfo: aString from: anObject
	"Log to the default info log"
	self infoLog logInfo: aString from: anObject.! !

!QLogger class methodsFor: 'log api' stamp: 'das 11/22/2009 16:07'!
logMethod: str
	"Log to the default info log"
	^self logInfo: ('>>',thisContext sender methodSelector asString, ' | ',str) from: thisContext sender receiver! !

!QLogger class methodsFor: 'default' stamp: 'ar 1/25/2008 12:06'!
logToFile: baseName
	"Create a new default logger using the specified baseName."
	self reset.
	Default := QLogger new baseName: baseName; yourself.
	^ Default! !

!QLogger class methodsFor: 'default' stamp: 'ar 8/6/2008 16:46'!
logToTranscript
	"Create a new default logger using the Transcript."
	self reset.
	Default := QLoggerToTranscript new.
	^ Default! !

!QLogger class methodsFor: 'log api' stamp: 'ar 8/7/2008 16:34'!
logWarning: aString
	"Log to the default warning log"
	^self logWarning: aString from: thisContext sender receiver! !

!QLogger class methodsFor: 'log api' stamp: 'jcg 2/11/2009 15:15'!
logWarning: aString every: msecs
	"Log to the default warning log"
	| tag |
	tag := thisContext selector -> thisContext sender receiver.
	(self warningLog isReadyToLog: tag interval: msecs) ifFalse: [^self "wait longer"].
	^self logWarning: aString from: thisContext sender receiver! !

!QLogger class methodsFor: 'log api' stamp: 'ar 8/7/2008 16:34'!
logWarning: aString from: anObject
	"Log to the default warning log"
	self warningLog logWarning: aString from: anObject.! !

!QLogger class methodsFor: 'log api' stamp: 'ar 8/12/2008 12:12'!
logXmlNode: node
	"Dispatch to the appropriate logger"
	| logger |
	node name caseOf:{
		[#Activity]	->	[logger := QLogger activityLog].
		[#Error]		->	[logger := QLogger errorLog].
		[#Warning]	->	[logger := QLogger warningLog].
		[#Info]		->	[logger := QLogger infoLog].
	} otherwise:[logger := QLogger default].
	^logger logXmlNode: node! !

!QLogger class methodsFor: 'log api' stamp: 'das 1/5/2009 15:55'!
panelLog

	| myMssg |
	DoPanelLog ifTrue:[
		myMssg := thisContext sender.
		self logInfo: myMssg asString from:  '[PanelLog]'.
	].! !

!QLogger class methodsFor: 'log api' stamp: 'das 1/5/2009 15:55'!
panelLog: mssg

	| myMssg |
	DoPanelLog ifTrue:[
		myMssg := thisContext sender.
		mssg ifNotNil:[myMssg := myMssg asString, ' - ', mssg].
		self logInfo: myMssg from:  '[PanelLog]'.
	].! !

!QLogger class methodsFor: 'utilities' stamp: 'ar 6/23/2009 13:35'!
printErrorStack: initialContext
	"Print an error stack for logging"
	| aContext cnt |
	aContext := initialContext.
	^String streamContents:[:strm|
		"Print the details first"
		cnt := 0.
		[aContext notNil and: [(cnt := cnt + 1) <= 5]] whileTrue:
			[aContext printDetails: strm.	"variable values"
			strm cr.
			aContext := aContext sender].
		strm nextPutAll:'--------------------------------------------------------------------'; cr.
		"Then the full stack"
		cnt := 0.
		[aContext notNil and: [(cnt := cnt + 1) < 500]] whileTrue:
			[aContext printOn: strm.
			strm cr.
			aContext := aContext sender].
	].
! !

!QLogger class methodsFor: 'default' stamp: 'ar 8/6/2008 12:34'!
reset
	"Drop any pending activity, close and discard the default log and the trace-log table."
	LastTraceLog := nil.
	TraceLogs ifNotNil: [ 
		TraceLogs do: [ : tlog | 
			tlog destroy.
		]].
	TraceLogs := nil.
	Default ifNotNilDo: [:logger |
		Default := nil.
		logger stopLoggerProcess
	].
	ErrorLog ifNotNil:[
		ErrorLog destroy.
		ErrorLog := nil.
	].
	WarningLog ifNotNil:[
		WarningLog destroy.
		WarningLog := nil.
	].
	InfoLog ifNotNil:[
		InfoLog destroy.
		InfoLog := nil.
	].
	ActivityLog ifNotNil:[
		ActivityLog destroy.
		ActivityLog := nil.
	].! !

!QLogger class methodsFor: 'utilities' stamp: 'ar 8/9/2008 22:38'!
setupLogs: propsFile prefix: prefix
	"Set up the logs from the given props file"
	| logs logName logDir |
	QLogger reset.
	logDir := propsFile at: prefix,'.LogDir'.
	logDir ifNotEmpty:["don't clobber the existing log dir unless it's specified"
		QLogger logDirectory: (FileDirectory on: (FileDirectory pathFromURI: logDir)).
	].
	logs := Dictionary new.
	logName := propsFile at: prefix,'.DefaultLog'.
	logName ifNotEmpty:[
		QLogger default: (logs at: logName ifAbsentPut:[self new baseName: logName]).
	].
	logName := propsFile at: prefix,'.ActivityLog'.
	logName ifNotEmpty:[
		QLogger activityLog: (logs at: logName ifAbsentPut:[self new baseName: logName]).
	].
	logName := propsFile at: prefix,'.ErrorLog'.
	logName ifNotEmpty:[
		QLogger errorLog: (logs at: logName ifAbsentPut:[self new baseName: logName]).
	].
	logName := propsFile at: prefix,'.WarningLog'.
	logName ifNotEmpty:[
		QLogger warningLog: (logs at: logName ifAbsentPut:[self new baseName: logName]).
	].
	logName := propsFile at: prefix,'.InfoLog'.
	logName ifNotEmpty:[
		QLogger infoLog: (logs at: logName ifAbsentPut:[self new baseName: logName]).
	].
! !

!QLogger class methodsFor: 'class initialization' stamp: 'jcg 8/14/2007 11:55'!
shutDown: quitting
	quitting ifTrue: [ self reset ].! !

!QLogger class methodsFor: 'logging' stamp: 'bgf 5/11/2008 15:54'!
trace: aTraceName log: aStringOrBlock

	"Append the given entry to the specified (symbolically-named) time trace log,
	 using the current millisecond-clock value as the timestamp.
		The log will be
			<default log directory> / <aTraceName>-Trace.log.
	 If aTraceName is nil, nothing is done."
	aTraceName ifNotNil: [
		self trace: aTraceName millis: (Time millisecondClockValue) log: aStringOrBlock
	]
! !

!QLogger class methodsFor: 'logging' stamp: 'bgf 5/7/2008 17:20'!
trace: aTraceName millis: millisOrNil log: aStringOrBlock

	"Append the given entry to the specified (symbolically-named) time trace log.
		The log will be
			<default log directory> / <aTraceName>-Trace.log.
	 If aTraceName is nil, nothing is done."

	aTraceName ifNotNil: [
		(self traceLog: aTraceName) 
			log: (('', (millisOrNil ifNil: [ Time millisecondClockValue ]) asStringWithCommas), 
			        ' ', 
				  (aStringOrBlock value)).
	].! !

!QLogger class methodsFor: 'log api' stamp: 'bgf 5/8/2008 14:41'!
traceLog: aSymName

	"The QLogger for time-tracing of the given trace 
	 (usually a particular embedded app, or maybe a voice channel, or somesuch.
	 The Logger keeps a table of TraceLogs by symbolic name,
	 so that disperse parts of the system can append timestamps...
	 See    trace:millis:log:  

	The log will be
			<default log directory> / <aTraceName>-Trace.log "


	| result |
	TraceLogs ifNil: [ TraceLogs := Dictionary new ].
	
	result :=  TraceLogs at: aSymName ifAbsentPut: [ 
			QTraceLogger new 
				baseName: (aSymName asString,  '-Trace');
				yourself].

	result ifNotNil: [ 
		LastTraceLog := result.
	].
	^ result! !

!QLogger class methodsFor: 'logging' stamp: 'bgf 5/8/2008 14:44'!
traceToLast: aStringOrBlock
	"Convenience ... log this message, with timestamp, to the last trace log that received a message."
	LastTraceLog ifNotNil: [ 
		LastTraceLog log: 
				((Time millisecondClockValue asStringWithCommas), 
			        ' ', 
				  (aStringOrBlock value)) ]! !

!QLogger class methodsFor: 'default' stamp: 'ar 8/6/2008 16:46'!
warningLog
	"The log used for logging errors"
	^WarningLog ifNil:[self default]! !

!QLogger class methodsFor: 'default' stamp: 'ar 8/6/2008 12:31'!
warningLog: aLogger
	"The log used for logging errors"
	WarningLog := aLogger! !

!QLogger class methodsFor: 'logging' stamp: 'jcg 8/14/2007 18:24'!
withLogStreamDo: aBlock
	"Provide a stream for the client to write the log message onto."
	QLogger default withLogStreamDo: aBlock! !

!QLogger methodsFor: 'private' stamp: 'ar 8/12/2008 11:52'!
appendEntry: e toFile: f

	"Inner workings of logging an entry;
	 write it to the supplied log file. "

	f nextPutAll: e asString; nextPutAll: String crlf.
! !

!QLogger methodsFor: 'accessing' stamp: 'ar 1/25/2008 12:14'!
baseName
	"The base portion of the log file name"
	^baseFileName! !

!QLogger methodsFor: 'accessing' stamp: 'ar 1/25/2008 12:14'!
baseName: aString
	"The base portion of the log file name"
	mutex critical:[baseFileName := aString].! !

!QLogger methodsFor: 'initialize' stamp: 'jcg 2/11/2009 14:41'!
clearLastLogTimes
	lastLogTimes := Dictionary new.! !

!QLogger methodsFor: 'initialize' stamp: 'bgf 11/7/2007 11:19'!
destroy
	self stopLoggerProcess.! !

!QLogger methodsFor: 'logging' stamp: 'bgf 9/23/2009 14:41'!
drainLogger
	"Make an effort to get the log queue flushed to the current logging destination."
	| prio |
	process ifNil: [ ^ self ].

	prio := process priority.
	process suspend.
	process priority: (Processor activeProcess priority + 5).	"So it will run before me."
	process resume.

	Processor yield.				"Let that high-p process have some cycles."
	(Delay forMilliseconds: 5) wait.	"More emphatic yield, in case it goes waitable while writing."

	process suspend.
	process priority: prio.			"As you were."
	process resume.! !

!QLogger methodsFor: 'private' stamp: 'ar 8/6/2008 12:15'!
errorDescriptionFor: errorOrString
	"Answer the description for an error or string"
	^errorOrString isString ifTrue:[errorOrString] ifFalse:[errorOrString description].
! !

!QLogger methodsFor: 'initialize' stamp: 'bgf 3/21/2009 21:16'!
initialize
	queue :=SharedQueue new.
	mutex := Mutex new.
	self clearLastLogTimes.
	self startLoggerProcess.! !

!QLogger methodsFor: 'logging' stamp: 'jcg 2/11/2009 15:12'!
isReadyToLog: timerTag interval: msecs
	"Someone wants to log... check if its been long enough since they last logged.  Answer true or false.  If true, also update the last-log-time to now."
	lastLogTimes at: timerTag ifPresent: [:time |
		(Time millisecondsSince: time) < msecs ifTrue: [^false "wait longer"]
	].
	lastLogTimes at: timerTag put: Time millisecondClockValue.
	^true
! !

!QLogger methodsFor: 'logging' stamp: 'bgf 8/22/2008 16:14'!
log: aStringOrBlock 

	"Omit the date if we're using the transcript."
	| time |
	baseFileName ifNotNil: [
		queue nextPut: 
			'[', DateAndTime now utcString, '] ', aStringOrBlock value.
	] ifNil: [
		"Formatted to align content with QEVentLogEntry asString..."
		time := String streamContents: [ : s | 
							s nextPut: $[.
							DateAndTime now printHMSOn: s.
							s nextPutAll: ']   '. ].

		queue nextPut:  (time, (String tab), (aStringOrBlock value)).
	].! !

!QLogger methodsFor: 'logging' stamp: 'jcg 8/13/2007 14:26'!
log: aStringOrBlock from: anObject
	| description |
	description := nil.
	(anObject respondsTo: #logName) ifTrue: [description := anObject logName].
	description ifNil: [
		description := anObject isBehavior 
								ifTrue: [anObject name] 
								ifFalse: [anObject class name].
	].
	self log: '[', description , ']: ', aStringOrBlock value.
	! !

!QLogger methodsFor: 'log api' stamp: 'ar 8/13/2008 00:49'!
logActivity: action user: userName organization: orgName host: hostName extra: extraString
	"Log an activity record"
	self logEntry:(
		(QActivityLogEntry new)
			date: DateAndTime now;
			action: action asString;
			user: userName asString;
			organization: orgName asString;
			host: hostName asString;
			extra: extraString asString;
			yourself
	).
! !

!QLogger methodsFor: 'log api' stamp: 'ar 8/12/2008 12:14'!
logEntry: aQLogEntryOrString
	"Log an already prepared log entry or string"
	queue nextPut: aQLogEntryOrString! !

!QLogger methodsFor: 'log api' stamp: 'ar 8/13/2008 00:49'!
logError: desc details: debugInfo from: anObject
	"Log an error with associated detailed information."
	self logEntry:(
		(QEventLogEntry new)
			type: #Error;
			date: DateAndTime now;
			source: (self logNameFor: anObject);
			description: desc asString;
			details: debugInfo asString;
			yourself
	).
! !

!QLogger methodsFor: 'log api' stamp: 'ar 6/23/2009 13:30'!
logError: error from: anObject
	"Utility method to log either string or Exception, e.g.,
		QLogger logError: 'This went wrong' from: self.
		[...] on: Error do:[:err| QLogger log: err from: self]. 
	"
	self logError: (self errorDescriptionFor: error)
		details: (self printError: error)
		from: anObject.
! !

!QLogger methodsFor: 'accessing' stamp: 'ar 1/25/2008 12:15'!
logFileName
	"Answer the actual (computed) log file name"
	^baseFileName, '.', Date today yyyymmdd,'.log'! !

!QLogger methodsFor: 'private' stamp: 'bgf 6/2/2008 22:32'!
logForAppending

	"Open, and move to end, of my log file,
	 returning the open File.   Return nil if I log to transcript."

	| file dir |
	"If no fileName is specified, we log to the Transcript..."
	baseFileName ifNil: [
		^ nil
	].

	"We have a fileName, so we log to a file."
	dir := LogDirectory ifNil:[FileDirectory default].
	file := [dir fileNamed: self logFileName] on: Error do:[:ex| ex return: nil].
	file ifNotNil:[
		file setToEnd.
		^ file
	].
	^ nil! !

!QLogger methodsFor: 'log api' stamp: 'ar 8/13/2008 00:49'!
logInfo: aString details: details from: anObject
	"Log a bit of information"
	self logEntry:(
		(QEventLogEntry new)
			type: #Info;
			date: DateAndTime now;
			source: (self logNameFor: anObject);
			description: aString asString;
			details: details asString;
			yourself
	).
! !

!QLogger methodsFor: 'log api' stamp: 'ar 8/12/2008 11:40'!
logInfo: aString from: anObject
	"Log a bit of information"
	^self logInfo: aString details: '' from: anObject! !

!QLogger methodsFor: 'private' stamp: 'ar 8/9/2008 22:20'!
logNameFor: anObject
	"Answers a printable name for the object"
	anObject isString ifTrue:[^anObject].
	(anObject respondsTo: #logName)
		ifTrue:[^anObject logName]
		ifFalse:[^anObject class name].
! !

!QLogger methodsFor: 'log api' stamp: 'ar 8/13/2008 00:49'!
logWarning: aString details: details from: anObject
	"Log a warning"
	self logEntry:(
		(QEventLogEntry new)
			type: #Warning;
			date: DateAndTime now;
			source: (self logNameFor: anObject);
			description: aString asString;
			details: details asString;
			yourself
	).
! !

!QLogger methodsFor: 'log api' stamp: 'ar 8/12/2008 11:40'!
logWarning: aString from: anObject
	"Log a warning"
	^self logWarning: aString details: '' from: anObject! !

!QLogger methodsFor: 'log api' stamp: 'ar 8/12/2008 12:03'!
logXmlNode: xmlNode
	"For convenience; if we have an XML node already convert it to a log entry and pass it through"
	| type entry |
	type := xmlNode name.
	(type == #Activity) 
		ifTrue:[entry := QActivityLogEntry new fromXml: xmlNode]
		ifFalse:[entry := QEventLogEntry new fromXml: xmlNode].
	self logEntry: entry.! !

!QLogger methodsFor: 'private' stamp: 'bgf 3/21/2009 21:17'!
mutex
	"Only for internal use.  Bulletproof agains existing QLoggers that don't have a mutex."
	^ mutex ifNil: [mutex := Mutex new]! !

!QLogger methodsFor: 'private' stamp: 'ar 6/23/2009 13:35'!
printError: error
	"Print an error stack for logging"
	| aContext |
	error isString 
		ifTrue:[aContext := thisContext sender home sender]
		ifFalse:[aContext := error signalerContext].
	^QLogger printErrorStack: aContext! !

!QLogger methodsFor: 'private' stamp: 'bgf 8/17/2008 11:26'!
runLoggerProcess
	"Pull entries of the queue and log them."

	[true] whileTrue:[
		| entry file |
		entry := queue next.	"I.e. wait for work"
		file := self logForAppending.
		file ifNil: [
			WorldState addDeferredUIMessage: [
				"Running from a demon thread, this is much safer than straight Transcript show."
				Transcript cr; show: entry.
			].
		] ifNotNil: [
			[  
				"Once we're writing, consume the queue before closing."
				self appendEntry: entry toFile: file.
				entry := queue nextOrNil.
				entry notNil
			] whileTrue.
			file close.
		].
	]! !

!QLogger methodsFor: 'initialize' stamp: 'bgf 5/11/2008 09:16'!
startLoggerProcess
	self startLoggerProcessAt: Processor userBackgroundPriority+1.
! !

!QLogger methodsFor: 'initialize' stamp: 'jcg 8/15/2007 15:46'!
startLoggerProcessAt: priority
	self mutex critical: [
		self stopLoggerProcess.
		process := self fork: #runLoggerProcess at: priority.
	]! !

!QLogger methodsFor: 'initialize' stamp: 'jcg 8/15/2007 15:46'!
stopLoggerProcess
	self mutex critical: [
		process ifNotNil:[
			process terminate.
			process := nil.
		].
	]! !

!QLogger methodsFor: 'logging' stamp: 'jcg 9/23/2007 14:45'!
withLogStreamDo: aBlock
	"Provide a stream for the client to write the log message onto."
	| stream |
	stream := String new writeStream.
	aBlock value: stream.
	self log: stream contents.
	^nil! !

!QLogger methodsFor: 'logging' stamp: 'ar 8/8/2008 10:47'!
xmlLog: xmlTag attributes: attrBlock contents: contentsBlock
	"Create a structured log entry using the given XML tag. 
	The argument block will be evaluated with an XMLWriter."
	self xmlLog: xmlTag attributes: attrBlock contents: contentsBlock do:[:xml| queue nextPut: xml].! !

!QLogger methodsFor: 'logging' stamp: 'ar 8/8/2008 10:47'!
xmlLog: xmlTag attributes: attrBlock contents: contentsBlock do: doBlock
	"Create a structured log entry using the given XML tag. 
	The argument block will be evaluated with an XMLWriter."
	| writer |
	writer := XMLWriter on: (String new: 100) writeStream.
	writer startTag: xmlTag.
	attrBlock ifNotNil:[attrBlock value: writer].
	writer endTag.
	contentsBlock ifNotNil:[contentsBlock value: writer].
	writer endTag: xmlTag.
	doBlock value: writer stream contents.! !

!QLogger methodsFor: 'logging' stamp: 'ar 8/6/2008 15:34'!
xmlLog: xmlTag contents: aBlock
	"Create a structured log entry using the given XML tag. 
	The argument block will be evaluated with an XMLWriter."
	self xmlLog: xmlTag attributes: nil contents: aBlock.
! !

QLogger subclass: #QLoggerToTranscript
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Objects-Logging'!

!QLoggerToTranscript commentStamp: '<historical>' prior: 0!
A logger that outputs to Transcript.!

!QLoggerToTranscript class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QLoggerToTranscript methodsFor: 'private' stamp: 'bgf 8/22/2008 16:41'!
appendEntry: e toFile: f

	WorldState addDeferredUIMessage: [
		"Running from a demon thread, this is much safer than straight Transcript show."
		f cr; nextPutAll: e asString; endEntry.
	].! !

!QLoggerToTranscript methodsFor: 'private' stamp: 'ar 8/6/2008 16:28'!
logForAppending
	^Transcript! !

!QLoggerToTranscript methodsFor: 'private' stamp: 'ar 1/13/2009 16:40'!
startLoggerProcess
	"For transcript logging, make the logger process be slightly higher than tweak script priority. Otherwise transcript logging in the client is depending on the client going effectively idle."
	self startLoggerProcessAt: Processor userScriptPriority+1.! !

QLogger subclass: #QRemoteLogger
	instanceVariableNames: 'logHost baseLog'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Objects-Logging'!

!QRemoteLogger commentStamp: '<historical>' prior: 0!
Remote logging facility, allowing routers and application servers to send logs to the main forums box.!

!QRemoteLogger class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QRemoteLogger class methodsFor: 'instance creation' stamp: 'ar 1/25/2008 12:06'!
baseName: baseName logHost: logHost
	^(self new) 
		baseName: baseName;
		logHost: logHost;
		yourself! !

!QRemoteLogger methodsFor: 'accessing' stamp: 'ar 8/12/2008 10:36'!
baseLog
	^baseLog! !

!QRemoteLogger methodsFor: 'accessing' stamp: 'ar 8/12/2008 10:36'!
baseLog: aLogger
	baseLog := aLogger! !

!QRemoteLogger methodsFor: 'logging' stamp: 'ar 8/11/2008 14:12'!
log: aString
	"Don't allow unstructured logging"
	^self logInfo: aString from: '<unstructured>'.! !

!QRemoteLogger methodsFor: 'logging' stamp: 'ar 8/11/2008 14:13'!
log: aStringOrBlock from: anObject
	"Don't allow unstructured logging"
	^self logInfo: aStringOrBlock value from: anObject.! !

!QRemoteLogger methodsFor: 'logging' stamp: 'ar 8/12/2008 11:59'!
logEntry: aQLogEntry
	"Log to the remote as well as the base log if provided"
	super logEntry: aQLogEntry.
	baseLog logEntry: aQLogEntry.! !

!QRemoteLogger methodsFor: 'accessing' stamp: 'ar 10/1/2007 21:32'!
logHost
	"The host to send the log entries to"
	^logHost! !

!QRemoteLogger methodsFor: 'accessing' stamp: 'ar 10/1/2007 21:32'!
logHost: aForumsHost
	"The host to send the log entries to"
	logHost := aForumsHost! !

!QRemoteLogger methodsFor: 'private' stamp: 'ar 8/12/2008 12:21'!
runLoggerProcess
	"Pull entries of the queue and log them."
	| resp |
	[true] whileTrue:[
		| entry |
		entry := queue next.	"I.e. wait for work"
		[resp := logHost logEntry: entry asXml.
		resp name == #ok 
			ifFalse:[QLogger logWarning: 'Logging failed: ', resp contentString].
		] on: Error do:[:ex| QLogger logError: ex].
	]! !

QLogger subclass: #QRevolvingLogger
	instanceVariableNames: 'fileSize count curIndex header'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Objects-Logging'!

!QRevolvingLogger commentStamp: '<historical>' prior: 0!
A Logger that creates a recycled series of numbered log files of limited size.

Logs run from 1 to count, each named:
	<baseName>.<curIndex>.log
When a log file size passes fileSize, 
the next log is reset to empty and becomes active.
When the nth log fills, we start back at .1.

If a header is supplied (usually as a reusable block)
its value is written:
  - as the first line in any newly cycled log
  - as the first new line when we re-begin logging after a hiatus.


!

!QRevolvingLogger class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QRevolvingLogger class methodsFor: 'instance creation' stamp: 'bgf 11/7/2007 18:19'!
baseName: aSimpleName count: anInteger size: aFileSize

	"Creating a revolving logger writing into files of the pattern
		<baseName>.<index>.log

	where <index> runs from 1 to anInteger,
	and the log ring steps to the next log when the current one crosses the given file size."

	^ (self new)
		baseName: aSimpleName;
		count: anInteger;
		fileSize: aFileSize;
		yourself! !

!QRevolvingLogger methodsFor: 'accessing' stamp: 'ar 1/25/2008 12:13'!
baseName: aSimpleName

	"The name portion for all the iterant file names.
	E.g. if 'activity' is given, the log names are activity.1.log, ... , activity.<count>.log "

	mutex critical: [
		baseFileName := aSimpleName.
		curIndex := nil.
	].! !

!QRevolvingLogger methodsFor: 'private' stamp: 'bgf 8/12/2008 21:52'!
begin
	"Part of (re)initialization... examine the directory,
	 to figure out where to start writing new log messages;
	 we try to pick up where a previous logging activity left off."

	| newest |
	newest := self mostRecent.
	newest ifNil: [
		curIndex := 1.
	] ifNotNil: [
		[
			curIndex := Number readFrom: 
				(newest copyFrom: (baseFileName size+2) to: (newest size - ('.log' size))).

		] on: Error do: [ : ex |
			Transcript cr; show: 'Logging setup problem: ', ex description.
			curIndex := 1.
		]
	].
	self checkRotation.
	header ifNotNil: [
		| log |
		log := super logForAppending.
		log ifNotNil: [
			log nextPutAll: (header value); crlf.
			log close.
		].
	].
! !

!QRevolvingLogger methodsFor: 'private' stamp: 'ar 1/25/2008 12:02'!
checkRotation

	"If my current target is too big, move to the next one.
	 Return true if the target changes."
	| entry |
	entry :=  self class logDirectory directoryEntryForName: self logFileName.
	entry ifNil: [ ^ false ].
	entry fileSize >= fileSize ifTrue: [
		self stepFile.
		^ true.
	].
	^ false
! !

!QRevolvingLogger methodsFor: 'accessing' stamp: 'bgf 11/7/2007 19:32'!
count: numFiles

	"The number of distinct files to create;  the files are named 
		<baseName>.1.log 
	 through
	     <baseName>.<numFiles>.log 
	"
	mutex critical: [
		count := numFiles.
		curIndex := nil.
	].! !

!QRevolvingLogger methodsFor: 'accessing' stamp: 'bgf 11/7/2007 19:34'!
fileSize: aThreshold
	"The threshold num-bytes at which we cap the size of single log file."
	mutex critical: [
		fileSize := aThreshold.
		curIndex := nil.
	].! !

!QRevolvingLogger methodsFor: 'private' stamp: 'ar 1/25/2008 12:14'!
filesYoungerThan: nMinutes

	"Full names of my log files (matching my name pattern)
	 that have been modified within the last nMinutes."

	| names now |
	now := Time totalSeconds // 60.

	names := (self class logDirectory) fileNamesMatching: (baseFileName, '.*.log').

	^ (((names 
		collect: [ :n | (self class logDirectory directoryEntryFor: n) ])
		select: [ :e | e notNil and: [ (now - (e modificationTime // 60)) <= nMinutes ]])

		asArray sort: [ :e1 :e2 | e1 modificationTime < e2 modificationTime ])
		collect: [:e| self class logDirectory fullNameFor:(e name) ]

! !

!QRevolvingLogger methodsFor: 'accessing' stamp: 'bgf 11/7/2007 22:21'!
header: aBlockOrValue
	"Evaluated and written as the first line each time we cycle to a new log,
	 and inserted as the first new line if we pick up an existing log in #begin."

	header := aBlockOrValue! !

!QRevolvingLogger methodsFor: 'accessing' stamp: 'ar 1/25/2008 12:16'!
logFileName
	"The actual (computed) log file name"
	^baseFileName, '.', curIndex, '.log'! !

!QRevolvingLogger methodsFor: 'private' stamp: 'bgf 8/12/2008 21:51'!
logForAppending

	"The log to which we should append the next entry
	 (returned in the form of an open File already positioned for writing."
	| log |
	curIndex ifNil: [ self begin ].
	(self checkRotation) ifTrue: [
		log := super logForAppending.
		log ifNotNil: [
			log nextPutAll: (header value); crlf.
		].
	].
	^ log  ifNil: [ super logForAppending ]! !

!QRevolvingLogger methodsFor: 'private' stamp: 'ar 1/25/2008 12:13'!
mostRecent
	"Local name of the most recently modified file matching my pattern;
	 used to pick up where we left off."

	| names |
	names := (self class logDirectory) fileNamesMatching: (baseFileName, '.*.log').
	^ names detectMax: [ :name | (self class logDirectory directoryEntryFor: name) 
										ifNil: [ 0 ]
										ifNotNilDo: [:entry | entry modificationTime ]]. ! !

!QRevolvingLogger methodsFor: 'private' stamp: 'ar 1/25/2008 12:04'!
stepFile
	"Go to the next file.  Should only be called after we've 'begun', 
	 as it assumes that curIndex is valid."

	curIndex := (curIndex \\ count) + 1.
	self class logDirectory deleteFileNamed: self logFileName.! !

QLogger subclass: #QTraceLogger
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Objects-Logging'!

!QTraceLogger commentStamp: '<historical>' prior: 0!
A customization of logger for supporting time-trace logs.

(The only difference is that we simplify the content to remove the date-mark,
since the time-trace log API supplies the millisecond clock value.)!

!QTraceLogger class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QTraceLogger methodsFor: 'logging' stamp: 'bgf 5/7/2008 14:12'!
log: aStringOrBlock 

	"no decoration, just the millisecond value supplied upstream"
	queue nextPut: aStringOrBlock value.
! !

TSpace subclass: #QSpace
	instanceVariableNames: 'avatarBox attachments garbage defaultEntry downPointer chatLog customs autoSaveJob soundRegions uiPreference uiStyle miramarUI welcomeDocument welcomeAppSheet collectedThemeElements'
	classVariableNames: 'AnglesTable'
	poolDictionaries: ''
	category: 'Qwaq-Objects-Space'!

!QSpace commentStamp: '<historical>' prior: 0!
TSpace extended with support for Qwaq toolkit.!

QSpace subclass: #QExteriorSpace
	instanceVariableNames: 'lights minuteOfDay compassRotation sky radius baseFogColor'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Objects-Space'!

!QExteriorSpace commentStamp: '<historical>' prior: 0!
A utility class for defining ordinary bounded outdoor spaces.
Provides lighting that can change with time of day.
Supports generated random terrain.

Provides easy creation of entry gates.  By default, the 'places' menu
lists all the QGates in the space, using their banner labels as menu entries.

The basic size of the space is controlled by #radius:, usually around 300.

To generate a space with default daynight lighting, and new fractal terrain,
use  
	QExteriorSpace new populate.

Supports  persistence efficiently; nothing but the control parameters is persisted.!

!QExteriorSpace class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QExteriorSpace class methodsFor: 'defaults' stamp: 'bgf 11/27/2009 18:06'!
defaultFogColor
	"The default base color for fog calculation."
	^ QColors defaultFogColor! !

!QExteriorSpace class methodsFor: 'defaults' stamp: 'bgf 2/1/2007 11:21'!
defaultRadius

	"The default basis for the size of generated terrain and the avatar box."
	^ 400! !

!QExteriorSpace methodsFor: 'menus' stamp: 'bgf 2/23/2008 02:14'!
actionVerbs
	"Selectors to use in the action menu; available to all visitors."

	^ (super actionVerbs),
	   {
		{#daytime . 'Daytime'} .
		{#nighttime . 'Nighttime' }
	   }.! !

!QExteriorSpace methodsFor: 'initialize' stamp: 'bgf 7/17/2007 23:51'!
addGate: aSymbol label: aString

	"Add a gate portal into the space, placing it on the ground.
	 The gate portal is registered under the given global symbolfor inbound linking).

	 Note that the result will have to be translated into a new position
	 if more than one gate is needed."

	| gate |
	gate := (QGate new)
			translation: (self adjustToGround:(5@0@-12));
			label: aString
			yourself.
	self addChild: gate.
	gate registerGlobal: aSymbol.
	^ gate! !

!QExteriorSpace methodsFor: 'initialize' stamp: 'jjl 8/5/2008 09:34'!
addLandmark: aString at: location facing: pyr sort: sortString

	"Add a place marker. If the location y is 0, adjust it to local ground height."
	| result |

	result := super addLandmark: aString at: location facing: pyr sort: sortString.
	location y = 0 ifTrue: [
		result translation: (self adjustToGround: location) + (0@(result extent y/2)@0).
	].
	"The next line is a hack to retrofit COI.  COI is an arbitrary distance (1) in front of the location."
	result centerOfInterest: (result translation + result lookAtLocal).
	^ result.
	! !

!QExteriorSpace methodsFor: 'initialize' stamp: 'bgf 2/23/2008 02:28'!
addSkyDome
	"If we add a sky, make it <the> sky."
	 ^ sky := super addSkyDome.! !

!QExteriorSpace methodsFor: 'initialize' stamp: 'bgf 3/3/2007 22:15'!
assuredDefaultEntry

	"Extended so the default entry for spaces with terrain
	 matches the ground height at the origin."

	^ defaultEntry ifNil: [  "Only adjust on creation - otherwise it would climb up each time."
		|de|
		de := super assuredDefaultEntry.
		de translation: (self adjustToGround: (de translation)).
		de
	].! !

!QExteriorSpace methodsFor: 'accessing' stamp: 'bgf 1/27/2007 19:29'!
baseFogColor

	"We vary the TSpace fog color according to time of day;
	 this is the basis for the calculated colors."

	^ baseFogColor! !

!QExteriorSpace methodsFor: 'accessing' stamp: 'bgf 1/27/2007 19:29'!
baseFogColor: aColor
	baseFogColor := aColor asVectorColor! !

!QExteriorSpace methodsFor: 'menus' stamp: 'jjl 8/5/2008 13:40'!
birdsEyeLandmark
	"The QPlace defining the Birds-Eye view of this space...
	 Always the first entry in my landmarksFor:"

	^ QPlace 
		category: 'Named Places' 
		label: 'Bird''s Eye' 
		frame: self
		selector: #birdsEyeStanceFor:
		pov: 'First Person'! !

!QExteriorSpace methodsFor: 'menus' stamp: 'jjl 8/5/2008 13:40'!
birdsEyeStanceFor: anAvatar

	| trans |
	trans := self birdsEyeTransformFor: anAvatar.
	^Stance here: trans translation there: (trans translation - trans lookAt)! !

!QExteriorSpace methodsFor: 'menus' stamp: 'bgf 3/15/2008 17:53'!
birdsEyeTransformFor: anAvatar

	"Default birds-eye view destination transform"

	^ Matrix4x4 new
		translation: -200@100@-200;
		pitchYawRoll: -15 @ -130 @ 0;
		yourself
! !

!QExteriorSpace methodsFor: 'lighting' stamp: 'bgf 1/27/2007 16:29'!
brightnessForMinuteOfDay: minute

	"Minute ranges from 0 to 1440.
       Result ranges from 0 to 1.  
	 Plotted, the cuve is flattish at 0 and 12, climbs and drops rapidly at 6 and 18"

	| base min |

	"daylight savings time - so transitions are 7:30 AM and PM."
	min := (minute - 90) max: 0.

	base := (720 - ((720 - min)) abs)/720.0.  "Linear 0-1 and back down, for hour [1..24]"

	"Want more rapid increase in the midrange, flat at dark and bright,
	 to simulate daybreak.  It would be nice to push sunset later."

	1 to: 3 do: [:i|
		base := (base - 0.5 * 90) degreeSin + 0.5 min: 1 max: 0.
	].

	^ base min: 0.99 max: 0.01! !

!QExteriorSpace methodsFor: 'accessing' stamp: 'bgf 1/27/2007 16:29'!
compassRotation

	"A hint to navigation displays: the rotation from this space's negative-z to user North."
	^ compassRotation! !

!QExteriorSpace methodsFor: 'accessing' stamp: 'bgf 1/27/2007 16:29'!
compassRotation: cr

	"A hint to navigation displays: the rotation from this space's negative-z to user North."
	compassRotation := cr.! !

!QExteriorSpace methodsFor: 'persist' stamp: 'bgf 2/4/2010 20:57'!
constructFrom: pStream

	super constructFrom: pStream.


	self baseFogColor: (pStream readColor: 'baseFogColor' default: self baseFogColor).

	radius := pStream readNumber: 'radius'  default: self radius.
	lights := pStream readObjectMap: 'lights' default: (Dictionary new).
	sky := pStream readObject: 'sky' default: sky.
	avatarBox := pStream readObject: 'avatarBox' default: (self avatarBox).
	garbage addAll: (pStream readSequence: 'garbage' default: #()).

	self setMinuteOfDay:(pStream readNumber: 'minuteOfDay' default: minuteOfDay).! !

!QExteriorSpace methodsFor: 'menus' stamp: 'bgf 1/27/2007 16:29'!
daytime
	"An action verb."

	self setMinuteOfDay: 600.! !

!QExteriorSpace methodsFor: 'accessing' stamp: 'das 3/12/2007 15:25'!
initialAvatarTranslation: trans

	"Position an avatar upon entry through a portal;
	 keep the avatar above ground."

	^ trans x @ (5.5 + (self groundHeightAt: trans))@ trans z
! !

!QExteriorSpace methodsFor: 'initialize' stamp: 'bgf 3/31/2007 10:18'!
initialize
	super initialize.

	self baseFogColor: self class defaultFogColor.

	minuteOfDay := 600. "((IslandTime current hour24)*60) + IslandTime current minute."
	compassRotation := 0.
	lights := Dictionary new.

	self radius: (self class defaultRadius).


! !

!QExteriorSpace methodsFor: 'lighting' stamp: 'bgf 9/20/2007 15:10'!
lightColorPrimary: warmColor secondary: coolColor fill: fillColor

	"Set the corner lights using the given color scheme."

	| light  warm cool fill |

	warm := warmColor asVectorColor.
	cool   := coolColor asVectorColor.
	fill     := fillColor asVectorColor.

	"warm in northwest "

	(light := (lights at: #primary ifAbsent:[nil])) ifNotNil: [ 
		light ambientColor: warm/1.5;
			diffuseColor: warm;
			specularColor: warm/4.
	].

	"cool light diagonally opposite"

	(light := (lights at: #secondary ifAbsent:[nil])) ifNotNil: [ 
		light ambientColor: cool / 1.5;
		 diffuseColor: cool;
		 specularColor: cool/5.
	].

	"fill at other corners"

	(light := (lights at: #fill ifAbsent:[nil])) ifNotNil: [
		light ambientColor: fill / 2;
		 diffuseColor: fill;
		 specularColor: fill/4.
	].
! !

!QExteriorSpace methodsFor: 'accessing' stamp: 'bgf 1/27/2007 16:47'!
lights
	"The directional lights that surround the space."
	^ lights! !

!QExteriorSpace methodsFor: 'menus' stamp: 'bgf 1/27/2007 16:29'!
nighttime
	"An action verb."

	self setMinuteOfDay: 1400.! !

!QExteriorSpace methodsFor: 'persist' stamp: 'bgf 2/4/2010 20:57'!
persistOn: pStream

	super persistOn: pStream.

	pStream storeNumber: 'radius' value: radius default: (self class defaultRadius).
	pStream storeColor: 'baseFogColor' value: self baseFogColor default: self class defaultFogColor.
	pStream storeObjectMap: 'lights' value: lights default: (Dictionary new).
	pStream storeObject: 'sky' value: sky default: nil.
	pStream storeObject: 'avatarBox' value: avatarBox default: nil.
	pStream storeSequence: 'garbage' value: garbage default: #().
	pStream storeNumber: 'minuteOfDay' value: minuteOfDay default: nil.! !

!QExteriorSpace methodsFor: 'lighting' stamp: 'bgf 1/2/2009 12:48'!
placeLights: r

"Place a stage-lighting spread of lights with lights at the corners of a square of the given radius.
  Load them into my target's 'lights' dictionary at #primary, #secondary, #fill1, and #fill2
  Does NOT set colors in the lights."

	| light  |

"Primary: southeast corner, high"

	light := TLight new.
	light type: #directional;
		translationX: r  y:r*1.1 z: r/3;
		pitchYawRoll: (115@-120@0);
		solid: false.
	self addChild: light.
	light objectName: 'primaryLight'.
	self lights at: #primary put: light.

"Secondary:  opposite, lower"

	light := TLight new.
	light type: #directional;
		 translationX: r negated y:r*0.8 z: r/3;
		 pitchYawRoll: (110@120@0);
		 solid: false.
	self addChild: light.
	light objectName: 'secondaryLight'.
	self lights at: #secondary put: light.

"Fill north, lower"

	light := TLight new.
	light type: #directional;
		 translationX: 0 y:r*0.7 z: r negated;
		 pitchYawRoll: (80@170@0);
		 solid: false.
	self addChild: light.
	light objectName: 'fillLight'.
	self lights at: #fill put: light.
! !

!QExteriorSpace methodsFor: 'initialize' stamp: 'bgf 2/23/2008 02:27'!
populate

	"Set up the default environment: skybox and lighting."

	self addSkyDome.
	self placeLights:(radius * 1.1).
	self setMinuteOfDay: 600.						"start in the morning"
	(self future: 1000) stepMinuteOfDay: 0.0167.		"start the clock."

	self homeTransform: ( (Matrix4x4 new) 
				translation: (self adjustToGround:(1@5.5@1)); yourself).! !

!QExteriorSpace methodsFor: 'accessing' stamp: 'bgf 1/27/2007 16:51'!
radius
	"Approximate extent of the space out from its origin."

	^ radius! !

!QExteriorSpace methodsFor: 'accessing' stamp: 'bgf 1/27/2007 16:53'!
radius: aScalar
	"Approximate extent of the space out from the origin."
	radius := aScalar.
	avatarBox ifNil: [
		avatarBox := TBox min:(aScalar@aScalar@aScalar)*-1.0 max: aScalar@aScalar@aScalar.
	] ifNotNil: [
		avatarBox  min:(aScalar@aScalar@aScalar)*-1.0 max: aScalar@aScalar@aScalar.
	].! !

!QExteriorSpace methodsFor: 'lighting' stamp: 'bgf 1/27/2007 18:36'!
setLightsForIndoorBrightness: aFloat

	"Set the lights for near-uniform indoor lighting;
       adjust the primary lightness to the given brightnes (0 to 2).
	1 is the brightness of 'indoor normal'"

	|  warm cool fill |

	warm := #(0.48 0.41 0.43) * (aFloat min: 2 max: 0).
	cool   := #(0.43 0.41 0.48) * (aFloat min: 2 max: 0).
	fill	   := #(0.32 0.4 0.32).

	self lightColorPrimary: warm secondary: cool fill: fill.! !

!QExteriorSpace methodsFor: 'lighting' stamp: 'bgf 1/27/2007 16:29'!
setLightsForIndoorNormal

	"Set the 4-way corner lights indoor regular lighting."

	self setLightsForIndoorBrightness: 1.0! !

!QExteriorSpace methodsFor: 'lighting' stamp: 'bgf 9/21/2007 12:25'!
setLightsForOutdoorBrightness: brightness

	"Set the lighting for an outdoor effect of the given 'brightness' (0-1).
	 We put a floor on the brightness (around half); there's always a full moon."

	|  warm cool fill |

	cool   := #(0.77 0.77 0.80).
	warm := #(0.38 0.38 0.35).
	fill   := #(0.1 0.1 0.1).

	warm := warm * (brightness min: 1.0 max: 0.5).
	(brightness > 0.3 & (brightness < 0.6)) ifTrue: [
		"bump red during daybreak and twilight."
		warm := warm + (#(0.2 0.03 0.0) *brightness).
	].
	cool := cool * (brightness min: 1.0 max: 0.5).
	fill := fill * (brightness min: 1.0 max: 0.4).

	self lightColorPrimary: cool secondary: warm fill: fill.! !

!QExteriorSpace methodsFor: 'lighting' stamp: 'bgf 2/3/2007 23:24'!
setMinuteOfDay: minute

	"Set the minute counter (0-1440),  and adjust atmospherics and lighting." 

	| dayness fog |

	minuteOfDay := minute.
	dayness := self brightnessForMinuteOfDay:minuteOfDay.	 "1.0 at noon, 0.0 at midnight."

	"Uses a pale illuminated atmosphere; more brighter fog at noon."

	fog := self baseFogColor.
	self fogOn: true;
		fogColor: (fog asColor * (0.4 +(dayness*0.6))) asVectorColor; 
		fogDensity: (0.00025 + (0.00025*dayness));
		fogStart: (120-(50*dayness));
		fogEnd: 300.

	self setLightsForOutdoorBrightness: dayness.
	sky ifNotNil: [
		sky brightness: (dayness min:1.0 max:0.4) fullBright:false.	
	].! !

!QExteriorSpace methodsFor: 'initialize' stamp: 'bgf 11/21/2007 20:53'!
simpleFractalTerrain: rad

	"Utility method  - generate and return 
	 a fractal landscape in square of the given radius. "
	| ground size heightScale smoothness seed |

	size := 2 raisedTo: 4.
	heightScale :=  0.13.
	smoothness :=  0.95.
	seed := Random new seed.

	ground := QFractalTerrain 
		initializeSize: size smoothness: smoothness heightScale: heightScale randomSeed: seed.
	ground genGrid;
			buildMesh;
			boundsDepth: 2;
		 	scale: rad;
			uvFactor: (rad/12.0)@(rad/15.0).
	ground material: (QMaterials groundMaterial:'lawn002.jpg').

	^ ground

! !

!QExteriorSpace methodsFor: 'accessing' stamp: 'bgf 2/5/2007 10:42'!
sky
	"If the space is populated using the default #populate, this is its SkyDome."
	^ sky! !

!QExteriorSpace methodsFor: 'step' stamp: 'bgf 1/27/2007 17:42'!
stepMinuteOfDay: increment

	"Adjust atmospherics and space colors for time of day." 

	| acceleration |

	acceleration := 1.0.  "1 is clock time; increase to accelerate time."

	self setMinuteOfDay: (minuteOfDay + increment) \\ 1440.
	
	(self future: (1000*60*5/acceleration)) stepMinuteOfDay:5.! !

!QExteriorSpace methodsFor: 'lighting' stamp: 'bgf 1/27/2007 16:29'!
toggleLight: aSymbol

	"Find the light in my lights dictionary; if present,
       flip it on or off (by adding/removing it as a frame child.)"

	| light |
	light := self lights at: aSymbol ifAbsent: [ ^nil ].
	(light hasAncestor: self) ifTrue: [
		self removeChild: light.
	] ifFalse: [ 
		self addChild: light.
	].
! !

!QExteriorSpace methodsFor: 'accessing' stamp: 'bgf 1/27/2007 16:29'!
uiFormatterClass 
	^#QDefaultUIFormatter.
! !

QSpace subclass: #QJobSpace
	instanceVariableNames: 'jobs activeJob lastListener'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Objects-Jobs'!

!QJobSpace commentStamp: '<historical>' prior: 0!
The Job Space is a special-purpose space
that manages the lifecycle of 'QParticipantJob' requests,
and provides an empty 'studio' space as a workspace for the active request.

Its purpose is to support operations that cannot be invoked
directly from within the replicated universe 
(such as rendering scenes to images.)

The class maintains a singleton instance in each TIsland it is used in.
It is normally not directly seen by client code;
it is a servant of the QParticipantJob implementation.!

!QJobSpace class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QJobSpace class methodsFor: 'instance creation' stamp: 'bgf 8/21/2008 14:01'!
current

	| island instance |
	island := Processor activeIsland.
	(island isKindOf:TIsland) ifFalse: [
		self error: 'Job spaces can only exist in TIslands.'.
	].

	^ island at: #QJobSpaceForIsland
	  ifAbsent: [
		QLogger logInfo:  'QJobSpace for ', island name.
		instance := self new.
		island at: #QJobSpaceForIsland put: instance.
		instance.
	  ].! !

!QJobSpace methodsFor: 'public' stamp: 'bgf 1/11/2007 16:40'!
addJob: job

	"Enter this job into the queue for processing;
	 a job should only appear in the queue once."

	jobs nextPut: job.
	activeJob ifNil: [
		self beginNextJob.
	].! !

!QJobSpace methodsFor: 'jobs' stamp: 'bgf 6/24/2007 19:14'!
beginNextJob

	"Private.  Launch the next job on the queue, if any."

	lastListener := nil.
	activeJob ifNotNil:[
		self error: 'JobSpace: Starting a job when one is already active ', activeJob jobNumber.
	].

	"Look for the first queued job that isn't done.
    	 A job may be both done and in the queue... 
	 this happens if we defer a job after its check timeout, 
	 but later the harness sends it a photo."

	[jobs isEmpty] whileFalse: [
		(jobs nextOrNil) ifNotNilDo:[ :job|
			job isDone ifFalse: [
				job begin ifTrue: [
					job countTry.
					activeJob := job.
					(self future: (job class timeout)) checkJob: job.
					^ self
				].
			].
		].
	].! !

!QJobSpace methodsFor: 'jobs' stamp: 'bgf 8/21/2008 20:21'!
checkJob: job

	"This is our state-advance future message;
	 a checkJob: job message is always pending in the future
	 when activeJob is nonNil.

	 We either have an active job,
	 or had an active job that ended in normal processing."

	(job = activeJob) ifTrue: [
		activeJob := nil.
	].

	job isDone ifFalse: [
		(job tries >= job class maxTries) ifTrue: [
			"If the most-recently dispatched job didn't get processed before its timeout,
			 remove the listener we sent it to from the listener table.
	 		 This is how we clean stale session-listeners from the island globals."
			lastListener ifNotNil: [
				QLogger logInfo: 'Stale job listener ', lastListener key,  ' (', job class name, ')'.
				self island future removeGlobal: lastListener key.
			].
			job notifyUnhandled.
			lastListener := nil.
			activeJob ifNil: [
				self beginNextJob.
			].
		] ifFalse: [
			job defer.
			self addJob: job.
		].
	] ifTrue: [
		lastListener := nil.
		activeJob ifNil: [
			self beginNextJob.
		].
	].! !

!QJobSpace methodsFor: 'persist' stamp: 'bgf 5/24/2007 17:51'!
constructFrom: pStream
	super constructFrom: pStream! !

!QJobSpace methodsFor: 'public' stamp: 'bgf 8/23/2007 10:17'!
deferJob: job for: msecs

	"Put this job at the back of the list, waiting at least the given period before attempting it;
	 the job will not be scheduled if it has past its retry count."
	
	| |
	job defer.
	job tries > job class maxTries ifFalse: [
		(self future: msecs) addJob: job.
		^ true.
	] ifTrue: [
		job finish.
		^ false.
	].
! !

!QJobSpace methodsFor: 'public' stamp: 'bgf 1/11/2007 01:13'!
garbage
	"TreeFort space protocol."
	^ OrderedCollection new! !

!QJobSpace methodsFor: 'initialize' stamp: 'bgf 1/8/2010 13:43'!
initialize

	super initialize.
	self unregisterResource.		"We don't want to persist."
	Processor activeIsland at: #QJobSpaceForIsland put: self.

	self color: (VectorColor r: 0.68 g: 0.7 b: 0.75 a:0.4).
	self placeLights.

	jobs := SharedQueue new.
	activeJob := nil.! !

!QJobSpace methodsFor: 'jobs' stamp: 'bgf 1/22/2008 02:09'!
jobDone: job

	"Callback from a job when finished."
	job finish.
	(job = activeJob) ifTrue: [
		activeJob := nil.
	].

	lastListener := nil.
	activeJob ifNil: [
		self beginNextJob.
	].! !

!QJobSpace methodsFor: 'private' stamp: 'bgf 6/24/2007 17:08'!
lastListener: aListener
	"The most recently signalled job listener; an association from key to value
	 in the island's global table.  Private to the implementation."
	lastListener := aListener! !

!QJobSpace methodsFor: 'persist' stamp: 'bgf 5/24/2007 18:01'!
persistOn: pStream
	"Save nothing - not even inherited aspects."
! !

!QJobSpace methodsFor: 'initialize' stamp: 'bgf 1/9/2010 12:19'!
placeLights
	"This is the lighting for Copy-To-Inventory previews of stored objects."
	|  width light |
	width := 400.0.

	"Cool sidelight in southeast ."	
	light := TLight new.
	light type: #directional;
		objectName: #northwestLight;
		 ambientColor: #(0.45 0.48 0.5 1);
		 diffuseColor: #(0.56 0.59 0.62 1);
		 specularColor: #(0.21 0.24 0.27 1);
		translationX: width*-1 y:2*width z: (width*0.75);
		pitchYawRoll: 120@155@0;
		solid: false.
	self addChild: light.

	"warm light  opposite."

	light := TLight new.
	light type: #directional;
		objectName: #southeastLight;
		 ambientColor: #(0.15 0.12 0.09 1);
		 diffuseColor: #(0.32 0.29 0.26 1);
		 specularColor: #(0.22 0.19 0.16 1);	
		translationX: width y:2*width z: 1.5;
		pitchYawRoll: 120@-90@0;
		solid: false.
	self addChild: light.

	! !

!QJobSpace methodsFor: 'initialize' stamp: 'hrs 3/14/2009 10:52'!
spawnAutoImportJob
	"Job spaces have nothing to import."
	^nil! !

!QJobSpace methodsFor: 'initialize' stamp: 'ar 6/3/2007 13:07'!
spawnAutoSaveJob
	"A job space never wants auto-save"
	^nil! !

!QSpace class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QSpace class methodsFor: 'properties' stamp: 'ar 4/16/2011 17:23'!
addPropertyRows: propertyDialog forFrame: spaceref
	"No super - we don't want any of the default frame properties here."

	"It isn't useful to 'Locate...' the space you're in, it just sends you to look at the origin"
	propertyDialog remove: (propertyDialog locateButton).
	
	"Should we respect walls?"
	propertyDialog addRow: (
		(QBooleanPropertyRow new)	
			titleString: ' Solid Obstacles Block Avatar Navigation';
			fetchBlock: [ :space| space get: #useObstacleDetection ];
			acceptBlock: [ : v :space | space future useObstacleDetection: v ];
			balloonHelp: 'If checked, avatars stop at eye-height solid objects.';
		yourself).

	"We co-opt placement locked, to mean 'home position' placement locked. "
	propertyDialog addRow: (
		(QBooleanPropertyRow new)	
			titleString: ' Lock Home Position';
			fetchBlock: [ :space| space get: #isPlacementLocked ];
			acceptBlock: [ : v :space | space future isPlacementLocked: v ];
			balloonHelp: 'Prevents changing the Home position.';
		yourself).


	"TODO - refactor avatarBox up from QConferenceRoom et al"
	"If we have an avatarBox, that's the main editable thing for a QSpace."
	
	propertyDialog addRow: (
		(QBoundingBoxPropertyRow new)
			titleString: 'Outer Limit of Avatar Movement';
			fetchBlock: [ :space| space get: #avatarBox ];
			acceptBlock: [ :v  :space | space future avatarBox: v ];
			balloonHelp: 'Avatars cannot drive outside this boundary box.';
		yourself).

	propertyDialog addRow: (
		(QTFramePropertyRow new)
			titleString: 'Welcome Document';
			nameSelector: #labelForBrowser;
			fetchBlock: [:space| (space get: #welcomeDocument) ];
			validateBlock: [ : frameRef :space : row | self validateWelcomeDoc: frameRef forRow: row ];
			acceptBlock: [: v  :space | space future welcomeDocFromProperties: v ];
			balloonHelp: 'Document with which to greet arrivals.  (Also in the Help menu.)';
		yourself).

	"Allow spaceSpeed calibration, so you can temper driving in very small or large volumes"
	propertyDialog addRow: (
		(QScalarPropertyRow new)
			titleString: 'Avatar Speed Adjustment';
			fetchBlock: [ :space| space get: #spaceSpeed ];
			acceptBlock: [ : v  :space  | space future spaceSpeed: ( v min: 10.0 max: 0.5) ];
			balloonHelp: 'Modulates the speed of keyboard-driven motion.' ;
		yourself).

	propertyDialog addRow: (
		(QTextPropertyRow new)
			titleString: 'Summary';
			fetchBlock: [ :space| space get: #summaryNote ];
			acceptBlock: [ : st  :space  | space future summaryNote: (st ifNil: [''] )];
			balloonHelp: 'Descriptive note for this Forum.' ;
		yourself).

	propertyDialog addRow: (
		(QStringPropertyRow new)
			titleString: 'Created';
			fetchBlock: [ :space| space get: #ctimeSummary ];
			acceptBlock: [ : v |];
			balloonHelp: 'Creation time of the Forum.';
		      inputEnabled: false;
			extraLong;
			yourself).

	propertyDialog addRow: (
		(QStringPropertyRow new)
			titleString: 'Template';
			fetchBlock: [ :space| (space get: #originalTemplate) ifNil: [''] ];
			acceptBlock: [ : v |];
			balloonHelp: 'Template used to make the Forum.';
		      inputEnabled: false;
			extraLong;
			yourself).! !

!QSpace class methodsFor: 'defaults' stamp: 'bgf 2/15/2010 23:11'!
defaultAvatarBox
	^ TBox origin: (-1000@-1000@-1000) corner: (1000@1000@1000)! !

!QSpace class methodsFor: 'private' stamp: 'bgf 5/1/2009 00:02'!
desktopDisplayV2ZipData
	"For in-island auto-upgrade of Desktop Display to Desktop Display (V2),
	 which adds
		atomic flag set to prevent deleting parts
		base made into a portable-object dragging grip so the display is movable.
	 The data is the same data as the packaged objects-menu template:
		 objects-menu / Forums / Display / Desktop Display.c3z  
	(builtin for speed, and so we aren't reaching out to the disk from on-island.)"

	|encoded|
	encoded := 'UEsDBBQAAAAIAEe8njoq7dwUzAcAANchAAAJABUAc2NlbmUueG1sVVQJAAPll/pJ5pf6SVV4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'  .

	^ encoded base64Decoded

"The magic string is produced via: 

	 Clipboard default clipboardText: 
		(FileStream readOnlyFileNamed: (FileDirectory pathFromURI: 'objects-menu/Forums/Display/Desktop Display.c3z'))
			binary contentsOfEntireFile asString base64Encoded printString.
	"! !

!QSpace class methodsFor: 'class initialization' stamp: 'bgf 7/30/2009 03:47'!
initialize
	AnglesTable := nil.! !

!QSpace class methodsFor: 'as yet unclassified' stamp: 'bgf 3/14/2010 17:53'!
space: spaceRef newHome: newxform oldHome: oldxform
	"Undoable record of home-transform change"

	^ QBlockAction
		undoBlock: [ spaceRef future homeTransform: (oldxform) ]
		redoBlock: [ spaceRef future homeTransform: (newxform) ]
		description: 'Home Position'
	        requiredFrames: { spaceRef }! !

!QSpace class methodsFor: 'as yet unclassified' stamp: 'bgf 7/30/2009 03:52'!
trigTableAtAngle: angle
	"Memo-izing cache of trig results for the freePlaceWith: for:  method.
	 Result is a Vector3, 
		x = cos angle negated .
		y = 0.
		z = sin angle negated. "

	^ (AnglesTable ifNil: [ (AnglesTable := Dictionary new) ])
		at: angle clone
		ifAbsentPut: [ Vector3 x: angle degreeCos negated  y: 0 z: angle degreeSin ].! !

!QSpace class methodsFor: 'properties' stamp: 'bgf 4/7/2009 23:05'!
validateWelcomeDoc: frame forRow: row

	"Property-row support - validator for the 'welcome application' frame-typed property."

	frame isValidFarRef ifFalse: [ ^ true ].	"Nil cases are OK"

	(((frame get:#class) includesBehavior: QFileCard)
	  and: [(frame get: #canUseAsForumWelcome) == true]) ifFalse: [
		self inform: 'Welcome Document must be an application document.'.
		^ false
	].

	^ true
! !

!QSpace methodsFor: 'protocols' stamp: 'bgf 2/23/2009 23:01'!
aboutToQuit
	"Client notification that it is in the logout process;
	 if warranted, fire an autosave asap."
	autoSaveJob ifNotNil: [
		autoSaveJob requestASave
	].! !

!QSpace methodsFor: 'editing' stamp: 'bgf 2/23/2008 02:13'!
actionVerbs
	"If the space has no sky dome, offer to add one."
	(self children anySatisfy: [ : c | c isKindOf: QSkyDome] ) ifFalse: [
		^ {{#addSkyDome . 'Add a Sky Dome' }}
	] ifTrue: [
		^ #()
	].! !

!QSpace methodsFor: 'initialize' stamp: 'bgf 2/8/2007 22:15'!
addLandmark: aString at: location facing: pyr

	"Add a place marker at the given position and orientation."
	^ self addLandmark: aString at: location facing: pyr sort: nil! !

!QSpace methodsFor: 'initialize' stamp: 'bgf 5/31/2009 13:30'!
addLandmark: aString at: location facing: pyr sort: sortString

	"Add a place marker at the given position and orientation, with the given menu sorting string."
	| result |
	result := QPlaceMarker new.
	result translation: location + (0@(result extent y/2)@0).
	result pitchYawRoll: pyr.
	result label: aString.
	result sortString: sortString.
	result isFromTemplate: true.
	self addChild:result.
	^ result.! !

!QSpace methodsFor: 'editing' stamp: 'bgf 7/1/2010 12:13'!
addModelInEditor: model inFrontOf: transform for: avatar doScale: doScale
	"Heuristically scale and place the given model into an editor 
	 (unless its class overrules use of editor)
	 for use from the given (transform) avatar viewing position."

	| ray avail distance scale size vertical groundY xform ey bb  onSupport |

	model allUpdateShape.
	bb := model boundingBox.

	doScale ifTrue: [
		"Don't let model start out smaller than a minimal boundingBox size."
		bb extentMax < 1 ifTrue: [
			model scale: 1 / (bb extentMax max: 0.01).
			bb := model boundingBox.
			]].

	"How much space is there out in front of the avatar?"
	ray := TRay new.
	ray localTransform: transform copy.
	ray globalTransform: (ray localTransform).
	ray excludeFrame: avatar.
	ray excludeFrame: avatar costume.
	ray pointerPickTree: self.
	onSupport := false.

	ray selectedPoint ifNotNil: [
		avail := ((ray selection globalPoint - transform translation) length).
		onSupport := (ray selectedNormal notNil and: [ ray selectedNormal y > 0.9 ]).
	] ifNil: [
		avail := 120.0.
	].

	onSupport ifTrue: [ 
		"When the pick hits a floor, place the object nearly at full distance,
	 	 so it stands visually as if dropped onto there from above.
	 	 But not quite the full distance, else it seems to far."
		distance := avail * 0.9 min: 75 max: 4.0
	] ifFalse: [
		"When the pick reaches a barrier, or nothing at all,
	 	 use the outer part of the available distance with the model,
		 and use the rest as distance from us to it."
		distance := (avail * 0.75) min: ((bb radius * 8) min: 75) max: 2.0.  "I.e. to translation point"
	].

	doScale 
		ifFalse: [scale := model scale] 
		ifTrue: [
			avail := (avail * 0.6) min: 60.0 max: 2.0.
			"Scale so the horizontals are within the far obstacle."
			size := (bb extent x max: (bb extent z)) max: 0.01.
			size < avail ifTrue: [
				scale := 1.0 + (avail/(size*16)).	"Enlarge gently"
				distance := (distance / scale) max: 2.0.  "and come closer."
			] ifFalse: [ 
				scale := avail/(size*1.5) min: 1.		"Shrink to fit."
			]].

	"Now we know where we want to place the object."
	xform :=((Matrix4x4 new translation:(transform lookAt*distance negated))*transform).
	"Filter the camera pitch/roll out of the placement - place things upright."
	xform pitchYawRoll: (0@(xform pitchYawRoll y)@0).

	"Find the ground."
	groundY := self groundHeightAt: xform translation.
	doScale ifTrue: [
		"Don't extend down past the ground, if possible."
		ey := bb extent y max: 0.01.
		groundY < xform translation y ifTrue: [
			scale := scale min: ((xform translation y - groundY) * 1.98 / ey)
		].

		"Now check upward limit.  We just make sure that the object,
	 	centered on our eyes, doesn't touch the overhead obstacle if any."
		ray resetTotal.
		ray isDownRay: false.
		ray localTransform: (
			((Matrix4x4 new) translation: xform translation)
				rotationAroundX:90;
				yourself).
		ray globalTransform: ray localTransform.
		ray pointerPickTree: self.
		ray selectedPoint ifNotNil: [
			vertical := 1.9 *  
					 ((ray selection globalPoint - ray globalTransform translation) length).
			scale := scale min: (vertical / ey).
		].
	].

	ray obliterate.
	model isFromTemplate: false;
		    isPlacementLocked: false;
		    doContextMenu: true.
	model localTransform: xform;
		    nominalScale: scale.
	self addChild: model.
	model allUpdateShape.

	(model isGroundBound or: [ onSupport ]) ifTrue: [
		"Reckoning against the ground, or if the object -wants- to be on the ground."
		| my |
		my := model translation y + (model boundingBox min y).
		my < (groundY + 2.5) ifTrue: [
			model translateIntoContactWith: (model translation x @ groundY @ model translation z) 
				    alongNormal: (0@1@0)
		].
	].
	"Except for inherently manipulable objects, new creations start in a resize box."
	(self getsEditBoxOnCreation: model) ifTrue: [
		avatar future editFrame: model
	].

	avatar future captureCreateFrames: { model } description: 'Add Object'.
! !

!QSpace methodsFor: 'editing' stamp: 'bgf 2/23/2008 02:31'!
addSkyDome
	"On the first day..."

	| result |
	result := (QSkyDome new) 
					style: #normal;
					yourself.
	self addChild: result.
	^ result ! !

!QSpace methodsFor: 'accessing' stamp: 'bgf 2/19/2008 21:32'!
addSoundRegion: aSoundRegion
	soundRegions add: aSoundRegion! !

!QSpace methodsFor: 'accessing' stamp: 'bgf 4/26/2009 11:46'!
adjustToGround: aPoint
	"If an x/y point, supply the ground height at that plan coordinate as y,
	 and return the resulting 3-point.   If an x-y-z vector, add the ground height
	 to its y component. Returns a 3-point."

	((aPoint isKindOf: Point)or: [aPoint size = 2]) ifTrue: [
		^ (aPoint x)@(self groundHeightAt: aPoint)@(aPoint y).
	] ifFalse: [
		^(aPoint x)@(aPoint y + (self groundHeightAt: aPoint))@(aPoint z).
	].! !

!QSpace methodsFor: 'accessing' stamp: 'bgf 2/19/2008 21:35'!
allSoundRegions
	"All QSoundRegions rooted in this space."
	^ soundRegions asArray		"asArray for copy semantics in far-ref retrieval"! !

!QSpace methodsFor: 'repair-upgrade' stamp: 'bgf 6/2/2009 17:37'!
applyTemplateFlagAggressively

	"This is aimed at marking most all of the neutral parts of project-forums,
	 receptions and libraries as part of the template."

	self children do:  [ : f |
		( ((f doContextMenu and: [ f pickable ]) not)
		 or: [ {TFrameThumb . TGeometry . TGroup  . TCube . TRectangle . 
			    QFixedDisplay . QFixedSurface .
			    QDisplaySign . QDocumentShelf .
			    QMirror . QDestination } includes: (f class) ] 
		) ifTrue: [
			f markAsFromTemplate
		] ].! !

!QSpace methodsFor: 'initialize' stamp: 'bgf 3/3/2007 22:08'!
assuredDefaultEntry

	"The default main entry frame - a tgroup at the origin, at height 0@5@0.
	 This places it at the eye height of the standard plain avatar,
	 so that normal portals present the ground plane properly aligned
	 when linked directly into the space (for service drag-drop linking of spaces.)

	Client code is responsible for registering this as #mainEntry,
 	using #registerDefaultMainEntry
 	Only one space per TIsland should be registered."

	defaultEntry ifNil: [
		defaultEntry := TGroup new.
		defaultEntry translation: 0@5@0;
				objectName: 'defaultMainEntry';
				visible: false;
				solid: false.
		self addChild: defaultEntry.
	].
	^defaultEntry.! !

!QSpace methodsFor: 'attaching' stamp: 'bgf 7/31/2009 01:59'!
attach: sticker to: surface
	"Create an attachment in the attachment manager.
	 Caller is responsible to ensure that both frames are persisted, component-level objects."
	attachments attach: sticker  to: surface.! !

!QSpace methodsFor: 'accessing' stamp: 'bgf 7/31/2009 00:10'!
attachments
	"My QFrameAttachments, the master table of movable-object attachments to other frames."
	^ attachments! !

!QSpace methodsFor: 'accessing' stamp: 'bgf 2/15/2010 23:13'!
avatarBox
	"If defined,  a TBox giving outer limits of avatar driving motion."
	^ avatarBox! !

!QSpace methodsFor: 'accessing' stamp: 'bgf 2/15/2010 23:14'!
avatarBox: aBox
	"TBox defining the limits of avatar driving."
	avatarBox := aBox! !

!QSpace methodsFor: 'editing' stamp: 'bgf 6/23/2009 09:33'!
avatarForLogin: userLogin
	"If any, the present avatar bearing the given user login "
	^ self allUsers detect: [ : avatar | avatar userLogin = userLogin ] ifNone: [ nil ]! !

!QSpace methodsFor: 'editing' stamp: 'bgf 2/12/2009 18:08'!
avatarForSession: sessID
	"If any, the present avatar bearing the given session ID "
	^ self allUsers detect: [ : avatar | avatar sessionID = sessID ] ifNone: [ nil ]! !

!QSpace methodsFor: 'accessing' stamp: 'bgf 4/1/2007 01:23'!
chatLog
	"The history of messages in this space."
	^ chatLog! !

!QSpace methodsFor: 'repair-upgrade' stamp: 'bgf 8/6/2010 15:05'!
checkForRepairCase: child
	"Load-time check for repair macros - called once per top-level space child.
	 Each method should
		- test the object very specifically to see if it applies, returning false if not.
	 	- if true, convert the object in some way that makes it fail the above test
	        (as well as carrying out the improvement or repair.)
		- return true if the object was changed, so that subsequent tests need not be applied
      The first case that returns true will end the checks,
	so we -deliberately- don't support piling up mutliple update conversions here."

	true caseOf: {
		[ self repairReceptionWall: child ] -> [].
		[ self repairLight: child ] -> [].
		[ self repairGround: child ] -> [] .
		[ self repairDoorAssembly: child ] -> [].
		[ self repairCeiling: child ] -> [].
		[ self repairBirdsEye: child] -> [].
		[ self repairDesktopDisplay: child ] -> [].
		[ self repairProjectWorkNiche: child ] -> [].
		[ self repairBalticChair: child ] -> [].
		[ self repairCaribbeanChair: child ] -> [].
		[ self repairPacificChair: child ] -> [].
		[ self repairFabricTaskChair: child ] -> [].
		[ self repairBalticRedBench: child ] -> [].
		[ self repairCaribbeanKireiBench: child ] -> [].
		[ self repairPacificWalnutBench: child ] -> [].
		[ self repairLibraryFabricBench: child ] -> [].
		[ self repairLibraryTaskCouch: child ] -> [].
		[ self repairLibraryTaskChair: child ] -> [].
		[ self repairConferenceTableDestination: child ] -> [].
		[ self repairReceptionSofa: child ] -> [].
		[ self repairUmbrellaTable: child ] -> [].
		[ self repairLibraryStack: child ] -> [].	 

	} otherwise: [].
! !

!QSpace methodsFor: 'protocols' stamp: 'bgf 2/15/2010 22:47'!
checkMotion: anAvatar from: fromVector delta: delta

	"Confine avatars to the driving limits, if defined."
	| collided |
	collided := super checkMotion: anAvatar from: fromVector delta: delta.
	self avatarBox ifNotNil: [ : box |
		^ box confinePoint: collided
	] ifNil: [
		^ collided
	].
! !

!QSpace methodsFor: 'accessing' stamp: 'bgf 7/8/2009 01:47'!
childrenForBrowser
	"So the inspector doesn't show avatars and their parts,
	 or other transient items that ought not be edited."
	^ self children reject: [ : f | 
			(f isInAvatar) 
			 or: [ f == defaultEntry ] 
			 or: [ f shouldPersist not ]].! !

!QSpace methodsFor: 'theme' stamp: 'jjl 1/9/2009 17:05'!
collectThemeElements: themeName

	collectedThemeElements := MThemeHolder new.
	collectedThemeElements objectName: themeName.

	frameChildren select:[ :child|
		self partOfTheme: child.
	] thenDo:[ :child|
		collectedThemeElements addChildLight: child.
	].

	^collectedThemeElements! !

!QSpace methodsFor: 'persist' stamp: 'bgf 7/31/2009 05:52'!
constructFrom: pStream
	| hp hr origName |

	"If we persisted a default entry, but the constructor has also made one,
	 we need to use the persisted one, but preserve any registrations made beforehand."
	defaultEntry ifNotNil: [
		origName := defaultEntry globalName.
	].

	"Now we've captured the original ME name, we can reconstruct children."
	super constructFrom: pStream.

	"Pre-1.1.48, the home transform was kept separately from defaultEntry's local transform;
	 let it prevail if present.   (Spaces saved with current code save the entry transform
	 as attributes of defaultEntry.)"
	(pStream readObject: 'defaultEntry' default: nil) ifNotNilDo: [ : de |
		defaultEntry ifNotNil: [
			| name |
			name := defaultEntry globalName.
			defaultEntry destroy.
			name ifNotNil: [de globalName: name].
			origName ifNotNil: [de globalName: origName].
		].
		defaultEntry := de.
	].
	"The super constructFrom: above discarded constructor-created children.
	 (And if defaultEntry was persisted/restored, then this is a harmless no-op."
	defaultEntry ifNotNil: [
		self addChild: defaultEntry.
	].

	"Pre 1.1.75, many templates made a 'Center' placemark with sort order 'a'
	 to be the first item in the places list.
	 This is now obsolete; we coordinate home position and entry position explicitly."
	((self 
			find: [:mark | mark isKindOf: QPlaceMarker]) 
			select: [ :m | m label = 'Center' and: [m sortString = 'a' ]])	
			do: [ : m | m destroy ].

	"Finally, make sure that an orphaned legacy frame
	 isn't clinging to the main entry slot (works around bug in 1.0.12)"
	(Processor activeIsland at: #mainEntry ifAbsent: [nil]) ifNotNilDo: [ :me |
		me root ifNil: [
			"If we were constructed to have a mainEntry, make sure we do."
			origName ifNotNil: [
				self assuredDefaultEntry globalName: origName.
			]]].

	chatLog constructLogFrom: pStream.

	hp := pStream readVector3: 'homePosition' 
					default: (self assuredDefaultEntry translation).
	hr := pStream readQuaternion: 'homeRotation' 
					default: (self assuredDefaultEntry localTransform asQuaternion).
	self homeTransform: 
			(hr asMatrix4x4 translation: hp).

	self welcomeDocument: (pStream readObject: 'welcomeDocument' default: nil).
	self summaryNote: (pStream readString: 'note' default: '').
	self originalTemplate: (pStream readString: 'originalTemplate' default: (self originalTemplate)).
	self ctime constructFieldsFrom: pStream prefix: 'ctime'.
	self mtime constructFieldsFrom: pStream prefix: 'mtime'.

	self customs: (pStream readObject: 'customs' default: customs).
	
	attachments constructContentsFrom: pStream.

	"Miscellaneous patches and updates to matching objects.
	 (This is when way we modernize things like furniture in forums
	 when simple mesh-content from an old template needs new behavior.
	 The future-send is to ensure that this happens after the full unpickling of the space.)"
	self future repairSpecialCases.
	self future fixAttachments.! !

!QSpace methodsFor: 'accessing' stamp: 'bgf 6/11/2009 17:17'!
ctime
	"A timestamp recording the creation-time of this forum."

	^ (self island ctime) ! !

!QSpace methodsFor: 'accessing' stamp: 'bgf 6/12/2009 09:20'!
ctime: aTimeStamp
	"A timestamp recording the creation-time of this forum."
	"This is an alias for a (Qwaq-specific) island property;
	 stored this way for easy editing in a forum space properties sheet,
	 while the save/load mechanisms can operate on islands.
	 The save system mirrors the value into the forum manifest
	 so it can be shown in the service chooser. "

	self island ctime: aTimeStamp! !

!QSpace methodsFor: 'accessing' stamp: 'bgf 6/11/2009 23:11'!
ctimeSummary
	"Gettable creation-time string summary of timestamp."
	"We use 'Not Recorded' rather than 'Unknown' simply 
	 to make it less troubling in the forum details panel."

	^ self ctime isDefined ifTrue: [ self ctime summary ]  ifFalse: ['(Not Recorded)']! !

!QSpace methodsFor: 'accessing' stamp: 'hrs 2/13/2009 14:27'!
customs
	^customs! !

!QSpace methodsFor: 'accessing' stamp: 'hrs 2/16/2009 12:34'!
customs: aQCustoms
	aQCustoms signalee: self.
	customs := aQCustoms.
! !

!QSpace methodsFor: 'accessing' stamp: 'hrs 2/16/2009 11:07'!
customsList
	"A list of resources that have been asynchronously imported."
	^self customs imports! !

!QSpace methodsFor: 'accessing' stamp: 'bgf 3/3/2007 22:10'!
defaultEntry
	"The default entry reference frame, if created, otherwise nil.
	 See #assuredDefaultEntry, which creates this frame on demand."

	^ defaultEntry! !

!QSpace methodsFor: 'attaching' stamp: 'bgf 7/31/2009 01:59'!
detach: sticker
	"Remove an attachment from the attachment manager."
	attachments attach: sticker to: nil! !

!QSpace methodsFor: 'accessing' stamp: 'bgf 1/15/2007 17:18'!
discard: aFrame

	"Dispose of the given TFrame.  
	 Wrap it and add it to the garbage list."

	| wrapper |

	(aFrame parent) ifNotNil: [
		aFrame parent removeChild:aFrame.
	].
	
	wrapper := QStorageWrapper new.
	"The size is loosely based on the standard trash can."
	wrapper enclose: aFrame within: (3@3@3).
	garbage add: wrapper.
! !

!QSpace methodsFor: 'accessing' stamp: 'bgf 12/26/2008 14:13'!
discardWelcomeApp

	"If the welcome-doc app-container has been set up, throw it away."
	welcomeAppSheet ifNotNil: [ 
		welcomeAppSheet destroy. 
		welcomeAppSheet := nil ].
	! !

!QSpace methodsFor: 'protocols' stamp: 'bgf 2/9/2010 18:52'!
dispose: aFrame
	"A frame is being destroyed.  Excise it from the attachment manager."
	super dispose: aFrame.
	(self stickersAttachedTo: aFrame)  do: [ : s | 
		(s root = self) ifTrue: [ self attach: s to: (aFrame surfaceAttachment) "... which may be nil" ]
		].
	self detach: aFrame.
! !

!QSpace methodsFor: 'accessing' stamp: 'bgf 4/26/2009 11:21'!
downRayAt: aPoint
	"Return the down-picking ray after aligning it to the position."
	| pt |

	((aPoint isKindOf: Point)or: [aPoint size = 2]) ifTrue: [
		pt := (aPoint x) @ 300 @(aPoint y).
	] ifFalse: [
		pt := aPoint.
	].

	downPointer resetTotal.
	downPointer translationX:pt x y: pt y z: pt z.
	downPointer globalTransform: (downPointer localTransform).

	^ downPointer

! !

!QSpace methodsFor: 'actions' stamp: 'hrs 4/6/2009 11:36'!
dropCard: aFileCard at: aPseudoPlaceName for: anAvatar
	"Drop aFileCard at aPseudoPlaceName, and answer whether or not we were successfull.

	aPseudoPlaceName is resolved first by an exact name match and then using QWebServices
	whitespace matching.

	If we can drop the card on the resolved target (see #dropCard:on:for:), we do.
	Otherwise we try the same way to drop the card on the Dropbox.
	Otherwise if we did resolve aPseudoPlaceName, we put aFileCard at that position.
	Only if these all fail do we answer false.

	The avatar is used for determining the available places and for ownership of open
	applications."

	"The current implementation of #dropboxFor: should never fail. 
	The freespace card position is futureproofing."

	| places target wild |
	places := anAvatar currentLandmarks.
	target := self findDropTargetNamed: aPseudoPlaceName 
				landmarks: places
				avatar: anAvatar 
				within: 4.
	target ifNil: [
		wild := QWebServices whitespaceMatcher: aPseudoPlaceName.
		target := self findDropTargetNamed: '*', wild, '*' 
					landmarks: places
					avatar: anAvatar
					within: 4.
	].
	
	(self dropCard: aFileCard on: target for: anAvatar) ifTrue: [^true].
	(self dropboxFor: anAvatar) ifNotNilDo: [:db |
		(self dropCard: aFileCard on: db for: anAvatar) ifTrue: [^true].
	].
	
	target ifNil: [^false].
	"Otherwise just put at target"
	"Don't loose it if you touch it. Cards start transient so we don't get multiple copies on drop."
	aFileCard isTransient: false. 
	aFileCard localTransform: 
		((target isKindOf: Stance) ifTrue: [target there] ifFalse: [target globalTransform]).
	aFileCard rotationAroundY: target yaw.
	self addChild: aFileCard.
	^true! !

!QSpace methodsFor: 'actions' stamp: 'hrs 4/1/2009 12:09'!
dropCard: aFileCard on: anObject for: anAvatar 
	"Drop aFileCard on anObject if possible and answer true, otherwise false.
	anObject might not be a TFrame, in which case we answer false."

	"This probably ought to evolve into something like
		^(anObject respondsTo: #acceptCard:fromUser:) 
			ifTrue: [anObject acceptCard: aFileCard fromUser: anAvatar]
			ifFalse: [false]"
			
	(anObject isKindOf: QDisplayPanel) ifTrue: [
		(QDisplaySheet canAcceptDrop: aFileCard) ifTrue: [
			anObject displayCard: aFileCard fromUser: anAvatar.
			^true
		].  
		(anObject historyRolodex insertCard: aFileCard) = aFileCard ifFalse: [
			aFileCard destroy. "It's a duplicate of one already there."
		].
		^true.
	].
	(anObject isKindOf: QDocumentShelf) ifTrue: [
		(anObject insertCard: aFileCard) = aFileCard ifFalse: [
			aFileCard destroy.  "It's a duplicate of one already there."
		].
		^true
	]. 
	^false! !

!QSpace methodsFor: 'accessing' stamp: 'hrs 3/11/2009 22:14'!
dropboxFor: anAvatar
	"Answer the designated drop box for the space, creating one if needed.
	This is the place where offline uploads or asynchronous imports go.

	We currently do not have different dropboxes for different users, but we could."
	
	| places frame home homeStance |
	places := anAvatar currentLandmarks.
	frame := self findDropboxFor: anAvatar among: places.
	frame ifNotNil: [^frame].
	
	"None found. Create one so that it is seen from home position."
	frame := QDisplayStand withSimpleFrame.
	home := places detect: [:lm | lm label = 'Home'].
	homeStance := home destinationStanceFor: anAvatar.
	frame rotationAroundY: homeStance yaw.
	"Don't trust homeStance #there. Construct our own a decent distance from #here"
	frame translation: (homeStance transform lookAt * -10 + homeStance here).
	frame standHeight: frame translation y - 3.08.
	self addChild: frame.
	frame beDropboxFor: anAvatar.
	^frame ! !

!QSpace methodsFor: 'editing' stamp: 'das 12/14/2006 19:28'!
editClass

	^QEditBox.! !

!QSpace methodsFor: 'menus' stamp: 'bgf 4/30/2009 09:30'!
findDocumentContainer: document

	"Inspector support for rudimentary find-in-space...
	 Locate a TFrame that holds the given shared document, if any.
	 Prefer finding running apps, then displays, finally anything.
	 A bit costly but is only run in response to deliberate command."

	self allFramesDo: [ : tframe |
		(tframe isEmbeddedApp and: [tframe document = document ])
		ifTrue: [
			^ tframe
		]].
	self allFramesDo: [ : tframe | 
		((tframe isKindOf: QDisplaySheet) and: [ tframe holdsDocument: document])
		ifTrue: [
			^ tframe
		]].
	(document isKindOf: QImageDocument) ifTrue: [	"Posters are always space-rooted."
		self frameChildrenDo: [ : tframe | 
			((tframe isKindOf: QPoster) and: [ tframe holdsDocument: document])
			ifTrue: [
				^ tframe
			]]].

	self allFramesDo: [ : tframe |
		 ((tframe respondsTo: #holdsDocument:)
		 and: [tframe isInAvatar not]
		 and: [tframe holdsDocument: document]) ifTrue: [
			^ tframe 
		].
	].
	^ nil! !

!QSpace methodsFor: 'accessing' stamp: 'hrs 4/29/2010 10:27'!
findDropTargetNamed: aPseudoPlaceName landmarks: places avatar: anAvatar within: aSizeFactorOrNil
	"Answer a frame or Stance that is heuristically as suitable as practical to have a file card
	dropped on it, or nil if none found.
 
	We resolve aPseudoPlaceName into a QPlace (among the specified places)
	or to the container of an existing file card with the designated name. 

	Once we find a match, we look for the nearest suitable target.
	We preferring panels to shelves, and shelves (rolodexes) to other frames.
	We prefer targets that are structurally close within the resolved object's hierarchy,
	but will settle for targets that are geometrically close.  aSizeFactorOrNil is as for 
	#findNearest:to:within:.

	This is on-island because some of the machinery depends on the Processor activeIsland, and
	because we don't really need a lot of intermediate results floating around on the network."

	|  place frame ourPosition |

	"First find the place, either by place label or destinationFrame objectName.
	The latter is useful for panels that have been explicitly named."
	place := places 
				detect: [:lm | aPseudoPlaceName match: lm label] 
				ifNone: [
					places 
						detect: [:lm | aPseudoPlaceName match: (lm destinationFrame ifNotNilDo: [:df | df objectName])]
						ifNone: []
					].
	frame := place ifNotNil: [place destinationFrame] ifNil: [ 
		"Look for an accessible file card that matches (e.g., not in inventory)"
		(self findFileCard: aPseudoPlaceName) ifNotNilDo: [:otherFileCard |
			"We prefer using it's container."
			(self findDocumentContainer: otherFileCard content) 
				ifNil: [frame := otherFileCard] 
				ifNotNilDo: [:container | frame := container].
		].
	].
	(frame isKindOf: QDestination) ifTrue: [frame := frame destination].
	frame ifNil: [^nil].   "Not found."

	"We've found some sort of match. 
	Look for it to include a panel or shelf that is structurally nearest.
	Invisible stuff is ignored (e.g., rolodexes in portal gates)."

	"Not much point in examining the structure of the whole space. 
	'Home' has the space as its destinationFrame, and docs might have it as container."
	frame = self ifFalse: [
		"First look for a display panel within frame..."
		frame allFramesDo: [:f | (f visible and: [f isKindOf: QDisplayPanel]) ifTrue: [^f]].
		"Fixed displays before shelves, when looking through stuctural containment"
		frame allFramesDo: [:f | (f visible and: [f isKindOf: QFixedDisplay]) ifTrue: [^f]].
		"Then a shelf/rolodex..."
		frame allFramesDo: [:f | (f visible and: [f isKindOf: QDocumentShelf]) ifTrue: [^f]].
	].

	"Still here? OK, look for the geometrically closest panel or shelf, with any parentage, 
	IFF it's close enough..."
	ourPosition := frame = self ifTrue: [
		"If the frame is the whole space, there's not a lot we can do..."
		place ifNil: [^nil] ifNotNil: [
			"... unless we have a place to work with. It will define a here and a there.
			If the 'there' was a meaningful position, it would better to use that rather than
			avatar's intended location (i.e., the 'here'). However, places that define a stance
			tend to have 'there' way the heck out there, mostly just defining a direction."
			(place destinationStanceFor: anAvatar) here
		]
	] ifFalse: [  "If frame is a real object other than the space, use it's position."
		frame globalPosition
	].

	"First panels and fixed displays..."
	(self findNearest: [:f| (f isKindOf: QDisplayPanel) or: [f isKindOf: QFixedDisplay]]
			to: ourPosition within: aSizeFactorOrNil) 
		ifNotNilDo: [:f | ^f].
	"Then shelves..."
	(self findNearest:  [:f| f isKindOf: QDocumentShelf] 
			to: ourPosition within: aSizeFactorOrNil) 
		ifNotNilDo: [:f | ^f].
	"Finally, just answer what we have, which may be a TFrame or a Stance."
	^frame
	
! !

!QSpace methodsFor: 'accessing' stamp: 'hrs 3/11/2009 19:33'!
findDropboxFor: anAvatar among: places
	"Answer the designated drop box among places in this space, or nil none found.
	This method is the definition of what is considered a Dropbox."
	^(places detect: [:lm | lm label = 'Dropbox'] ifNone: []) ifNotNilDo: [:place |
		place destinationFrame
	]
	! !

!QSpace methodsFor: 'accessing' stamp: 'hrs 2/6/2009 15:36'!
findFileCard: identifierString
	"Answer any file card in the space matching identifierString, else nil.
	Does not consider cards in inventory, and for multiple matches, will answer
	one already on a panel if possible."
	| cards |
	cards := self find: [:frame | 
				(frame isKindOf: QFileCard) and: 
				[frame isInInventory not] and:
				[identifierString match: frame identifier]
				].
	^cards detect: [:card | card isOpenInOwner]
		ifNone: [cards ifEmpty: [] ifNotEmpty: [cards first]].! !

!QSpace methodsFor: 'accessing' stamp: 'hrs 3/11/2009 21:49'!
findNearest: aPredicateBlock to: aPoint within: aFactorOrNil
	"Answer the nearest instance pasing aPredicateBlock, else nil.
	If aFactorOrNil is not nil, we only keep only those instances whose distance from aPoint
	is within the instance's bounding radius * aFactorOrNil."
	| frames distance |
	frames := OrderedCollection new.
	self allFramesDo: [:f | 
		(aPredicateBlock value: f) ifTrue: [
			distance := f globalPosition distanceTo: aPoint.
			(aFactorOrNil ifNil: [true] ifNotNil: [f boxRadius * aFactorOrNil >= distance ])
				ifTrue: [frames addLast: {f. distance}].
			].
	].
	frames ifEmpty: [^nil].
	frames := frames asArray sort: [:a :b | a second < b second].
	^frames first first
	
	
! !

!QSpace methodsFor: 'accessing' stamp: 'hrs 8/31/2009 17:38'!
findPoster: identifierString
	"Answer any poster card in the space matching identifierString, else nil."
	"posters are always space-rooted"
	self frameChildrenDo: [ : tframe | 
		((tframe isKindOf: QPoster) and: [identifierString match: tframe card identifier])
			ifTrue: [^ tframe]
	].
	^nil
! !

!QSpace methodsFor: 'menus' stamp: 'bgf 3/5/2009 12:57'!
findTextureContainer: aTFileTexture

	"Inspector support for rudimentary find-in-space...
	 Locate any TFrame that refers to the given file texture, and return the first we find.."

	self allFramesDo: [ : tframe |
		(tframe isInAvatar not 
		 and: [
		 (tframe fileTextures anySatisfy: [ : tf | tf = aTFileTexture ]) ]) ifTrue:  [
			^ tframe ].
	].
	^ nil! !

!QSpace methodsFor: 'attaching' stamp: 'bgf 7/31/2009 05:55'!
fixAttachments
	"After internalizing, give all surface-attaching frames
	 a chance to repair themselves from the pre-attachment-manager past."
	self children do: [ : f | 
		((attachments surfaceAttachmentOf: f) isNil 
		and: [f shouldPersist])
		ifTrue: [
			f setupAttachment
		]].! !

!QSpace methodsFor: 'places' stamp: 'bgf 8/19/2009 04:36'!
freePlaceWith: anAvatar for: askingAvatar
	"Find the tag-along site for askingAvatar to follow allong with anAvatar.
	 Returns a global position (V3)"
	| position angle distance useBounds offset spacing followers spot |

	"Magic number - based on typical avatar proportions."
	spacing := 2.1.
	followers := anAvatar followers.
	spot := followers indexOf: askingAvatar ifAbsent: [ followers size + 1 ].
	position := (anAvatar translation).
	
	"Assign position on the half-circle behind the leader.
	 We don't use the directly-behind spot, so shift the quarter-spots in a bit."
	spot < 5 ifTrue: [
		angle := #(170.0 10.0 122.0 58.0) at: ((spot - 1 \\ 4) + 1).
		distance := spacing.
	] ifFalse:[ 
		spot < 11 ifTrue: [	
			angle := #( 116.0 64.0  164.0 16.0 140.0 40.0 ) at: (spot - 5 \\ 6) + 1.
			distance := 2 * spacing.
		] ifFalse: [
			angle := #(101.0 79.0  139.0 41.0  120.0 60.0  158.0 22.0   ) at: (spot-11 \\ 8) +1.
			distance := (3 +  (spot - 11// 8) ) * spacing.
		].
	].

	offset := anAvatar orientation transformPoint: (self class trigTableAtAngle: angle) * distance.

	"Enforce motion limits only if the leader's position is within motion limits."
	useBounds := ((self checkMotion: anAvatar from: (position) delta:(0.05@0@0.1)) 
					distanceTo: (position + (0.05@0@0.1))) 
					< 0.02.
	^ useBounds 
		ifTrue: [ self checkMotion: anAvatar from: position delta: offset ]
		ifFalse: [ position + offset ].
		
		
	! !

!QSpace methodsFor: 'accessing' stamp: 'bgf 1/15/2007 17:28'!
garbage

	"List of (QStorageWrapper-encased) TFrame objects deleted from the space."

	^ garbage! !

!QSpace methodsFor: 'accessing' stamp: 'bgf 4/27/2007 20:51'!
gateForNewLink

	"Find a QGate to use to establish the far side of a new 2-way space link."
	| gates |
	gates := self find:[:fc|(fc isKindOf: QGate) and: [fc visible]].
	gates isEmpty ifTrue: [
		^ nil
	].

	"Find the first unlinked gate, if any are present."
	(gates detect: [:g | g postcardLink = nil] ifNone:[nil]) ifNotNilDo: [:g| ^ g].

	"By default, we don't offer to overwrite existing linkages."
	^ nil! !

!QSpace methodsFor: 'actions' stamp: 'hrs 8/16/2007 09:22'!
generate: nObjects maxExtent: max
	"Generate a bunch of junk for testing purposes, without creating a lot of promises.
	Usage: ExploreUI to activeSpace as a farRef, and do 
		self future generate: 1000 maxExtent: 50."
	|  range generator obj | 
	range := 0 to: max.
	generator := range class randomForPicking.
	nObjects timesRepeat: [
		obj := TCube new.
		obj translationX: (range atRandom: generator) 
			y:  (range atRandom: generator)
		 	z:  (range atRandom: generator).
		obj rotation: 45 aroundX: (range atRandom: generator) 
			y:  (range atRandom: generator) 
			z:  (range atRandom: generator).
		self addChild: obj.
		].! !

!QSpace methodsFor: 'protocols' stamp: 'bgf 1/5/2010 15:39'!
getsEditBoxOnCreation: aFrame
	 "Return true if a newly instantiated object should be wrapped in an edit box.
	 This should be changed to a policy check, when there's a policy;
	 for now it prevents the obvious oddities (e.g. an edit box around a card)."
	| klass |

	aFrame visible ifFalse: [ ^ true ].

	klass := aFrame class.
	{ QFileCard . QDraggable .  QPortableObject . 
	   QAnnotation . QDisplayPanel . QSoundFrame . MView  . QBarChart }  
	 do: [ : nec |
		(klass includesBehavior: nec) ifTrue: [ ^ false ]
	].
	^ true
! !

!QSpace methodsFor: 'accessing' stamp: 'bgf 8/6/2010 11:53'!
groundHeightAt: aPoint2or3
	"Backstop for general placement protocol; spaces with dedicated terrain or fixed ground 
	 should override this to return the y coordinate of the ground at the given point;
	 the point may a Point (in which case its y is space z) or a 3-vector."
	| terrains ray maxY |

	terrains := self terrainFrames.
	terrains isEmptyOrNil ifTrue: [ ^ 0 ].

	"First check from below the candidate point, if it has a y."
	( (aPoint2or3 isKindOf: Vector) and: [ aPoint2or3 size > 2 ]) ifTrue: [
		ray := self downRayAt: aPoint2or3.
		maxY := Float infinity negated.
		terrains do: [ : tf |
			ray pointerPickTree: tf.
			ray selectedPoint ifNotNil: [
				ray selection globalPoint ifNotNilDo: [ : gp | maxY := maxY max: gp y ].
			]].
	
		maxY = Float infinity negated ifFalse: [ ^ maxY ].
	]. 


	"If that fails, try from the air above."
	((aPoint2or3 isKindOf: Point) or: [aPoint2or3 size = 2]) ifTrue: [
		ray := self downRayAt: ( (aPoint2or3 x) @ 300 @(aPoint2or3 y) ).
	] ifFalse: [
		ray := self downRayAt: ( (aPoint2or3 x) @ (aPoint2or3 y + 24 ) @(aPoint2or3 z) ).
	].

	maxY := Float infinity negated.
	terrains do: [ : tf |
		ray pointerPickTree: tf.
		ray selectedPoint ifNotNil: [
			ray selection globalPoint ifNotNilDo: [ : gp | maxY := maxY max: gp y ].
		]].


	^ (maxY = Float infinity negated) ifTrue: [ 0 ] ifFalse: [ maxY ].! !

!QSpace methodsFor: 'accessing' stamp: 'bgf 3/13/2008 23:24'!
homeTransform

	"The entry transform for avatars teleporting to this space;
	 this is also the defaultEntry frame transform."

	^ self assuredDefaultEntry localTransform! !

!QSpace methodsFor: 'accessing' stamp: 'bgf 3/13/2008 23:25'!
homeTransform: aMat4x4

	"Set the transform at which in-teleporting avatars appear,
	 which is also the default linking site for portals opening into the space's #mainEntry
	 if defaultEntry has been registered as a global entry point."

	self assuredDefaultEntry localTransform: aMat4x4 clone! !

!QSpace methodsFor: 'initialize' stamp: 'bgf 7/1/2010 12:13'!
initialize
	super initialize.
	QStyles current.		"Induce initialization early so it happens in t-time, not later in get/render,"
	QTextures current.	"as QStyles-current does a dicey future-send to prePersist itself."
	garbage := OrderedCollection new.
	chatLog := QChatLog new.
	avatarBox := self class defaultAvatarBox.

	"It is very handy to have a prebuilt down-pointer."
	downPointer := TRay new.
	downPointer localTransform: (Matrix4x4 new rotationAroundX: -90).
	
	"Induce early retrieval - this pulls the initial bvh load
	 up early in startup, so it doesn't happen on-demand
	 during avatar setup.   (So it doesn't jar the 'space entry' 
	 sequence between render begin and avatar materialize.)"
	false ifTrue: [QMotionCaptures manWalk ].

	"spawn an auto-save job is one is requested for this space"
	autoSaveJob := self spawnAutoSaveJob.
	"Have someone check for new imports periodically."
	self spawnAutoImportJob.

	soundRegions := IdentitySet new.

	uiPreference := #Classic.
	self uiStyle: uiPreference.

	attachments := (QAttachmentManager new) space: self; yourself.
	self customs: QCustoms new
! !

!QSpace methodsFor: 'attaching' stamp: 'bgf 8/16/2009 03:27'!
isFrame: frame attachedOn: surfaceFrame
	"Is the given frame attached, directly or indirectly, to the given surface frame ? "
	^ attachments isFrame: frame attachedOn: surfaceFrame! !

!QSpace methodsFor: 'accessing' stamp: 'ar 4/16/2011 17:23'!
labelForBrowser	
	"The island name reflects the template from which this forum was drawn, e.g. 'Library',
	 so is a decent generic label for the scene as a whole.  Service names are unpredictably long."
	^ 'Forum Details (', ( self serviceName ifNil: [ Processor activeIsland name ]), ')'! !

!QSpace methodsFor: 'menus' stamp: 'bgf 6/21/2008 17:35'!
landmarksFor: anAvatar

	"Accumulate all the QPlaces (and possibly legacy landmarks) for the places menu."
	| collection |
	collection := {anAvatar homePlace},
			       (self markedPlacesFor: anAvatar), 
	   			 (self otherPlacesFor: anAvatar).
 
	^collection.
! !

!QSpace methodsFor: 'accessing' stamp: 'hrs 2/16/2009 11:08'!
lastCustomsEntryServerTimestamp
	"The server utcSeconds of the last entry in customs."
	^self customs lastEntryServerTimestamp! !

!QSpace methodsFor: 'accessing' stamp: 'bgf 3/18/2007 13:59'!
linkInboundPortal: postcardToFarEnd

	"Find a gate suitable for connecting back to the given inbound connector,
	 and set its link, returning the affected gate.
       Nil if no gate is found or set up."

	| gate |
	gate := self gateForNewLink ifNil: [ ^ nil ].
	gate postcardLink: postcardToFarEnd.
	gate openContents.

	^ gate! !

!QSpace methodsFor: 'menus' stamp: 'bgf 3/14/2008 15:45'!
markedPlacesFor: anAvatar

	"Landmarks for all the QPlacemarks in the space,
	 sorted according to the place mark sort order."

	^ ((((self 
			find: [:mark | mark isKindOf: self placeMarkerClass]) 
		     	sortBy: [:a :b | b sortString isNil or: [a sortString notNil and: [a sortString < b sortString]]])
	 	    	collect: [ :mark| mark asPlace] )
	  	asArray),
	    (self overheadLandmarksFor: anAvatar).

! !

!QSpace methodsFor: 'protocols' stamp: 'bgf 4/30/2008 16:10'!
miramarUI
	"Backstop so that QEditBox, etc, checks on root miramarUI don't MNU in non-miramar QSpace"
	^ false! !

!QSpace methodsFor: 'accessing' stamp: 'bgf 6/11/2009 17:17'!
mtime
	"A short string note (under user control) describing the forum."

	^ self island mtime! !

!QSpace methodsFor: 'accessing' stamp: 'bgf 6/12/2009 09:20'!
mtime: aTimeStamp
	"A timestamp recording the last modified-time of this forum."
	"This is an alias for a (Qwaq-specific) island property;
	 stored this way for easy editing in a forum space properties sheet,
	 while the save/load mechanisms can operate on islands.
	 The save system mirrors the value into the forum manifest
	 so it can be shown in the service chooser. "
	self island mtime: aTimeStamp! !

!QSpace methodsFor: 'accessing' stamp: 'bgf 6/11/2009 21:39'!
mtimeSummary
	"Gettable creation-time string summary of timestamp."
	^ self mtime isDefined ifTrue: [ self mtime summary ]  ifFalse: ['']! !

!QSpace methodsFor: 'initialize' stamp: 'hrs 9/30/2009 16:30'!
name: aTFrame uniquelyAmongClass: aClass
	"Changes the objectName of aTFrame if necessary to ensure that it is unique among all 
	subinstances of aClass within the space.
	This is used to define an initial stable destination name for Landmark Links"
	| id |
	(aTFrame objectName copyAfter: $-) asInteger ifNotNil: ["already numbered" ^nil].
	id := self  uniqueIntegerAmong: [:frame | frame isKindOf: aClass]
			checking: [:frame | (frame objectName copyAfter: $-) asInteger].
	id isZero ifFalse: [aTFrame objectName: aTFrame objectName, '-', id].
	! !

!QSpace methodsFor: 'actions' stamp: 'rmp 12/29/2008 15:21'!
openQMSForExtrapolationTest: url andAvatar: myAvatar
      | card panel box |
	card := QFileCard forURL: url.
	
	panel := QDisplayPanel withSimpleFrame.
	panel consumeDrop: card avatar: myAvatar selection: nil.

	box := self boundingBox: Matrix4x4 identity.

	panel extent: ((7.6@5.7) min: (box extent y)@(box extent y*0.75) max:(2.4@1.8)).
	panel localTransform: Matrix4x4 identity.
      panel translationX:1y:4z:1. 

	self addChild: panel.
^ panel
! !

!QSpace methodsFor: 'accessing' stamp: 'bgf 6/11/2009 21:05'!
originalTemplate
	"The name of the forum -template used to create me, if known. "
	^ self island originalTemplate! !

!QSpace methodsFor: 'accessing' stamp: 'bgf 6/12/2009 09:20'!
originalTemplate: aString
	"Name of the template used to make this forum, if known."
	"This is an alias for a (Qwaq-specific) island property;
	 stored this way for easy editing in a forum space properties sheet,
	 while the save/load mechanisms can operate on islands.
	 The save system mirrors the value into the forum manifest
	 so it can be shown in the service chooser. "

	self island originalTemplate: aString.! !

!QSpace methodsFor: 'menus' stamp: 'bgf 3/15/2008 18:00'!
otherPlacesFor: anAvatar

	"Places for all but the QPlacemarks in the space.
	 The result contains an entry for each frame that returns true from #isListedLandmark.
  	 All those must also implement asPlace, returning a QPlace.
	 (We do the QPlacemarks separately 
 	 because of the special menu sorts defined by  their sort-strings,
	 for nice tab tours in conference templates.)"

	^ ((self 
	  	find: [:mark | (mark isListedLandmark )])
	  	collect: [ : mark | mark asPlace ])
	  	asArray
! !

!QSpace methodsFor: 'places' stamp: 'bgf 6/28/2009 11:40'!
overheadLandmarksFor: anAvatar
	"Prepare an Array (Must be Array to be #get: - suitable)
	 of Landmarks for the overhead plan-style views of this space.
	 Default is a single comprehensive overview."
	| plan |
	plan := 
		QPlace 
			category: 'Named Places'
			label: 'Plan View'
			frame: self
			selector: #planViewStanceForAvatar:
			sticky: true.
	plan cameraSelector: #planViewCameraStanceForAvatar:.
	
	^ {plan}	! !

!QSpace methodsFor: 'theme' stamp: 'jjl 1/12/2009 10:11'!
partOfTheme: frame

	"This line lets anything be part of the theme if it can get the flag set."
	(frame isThemeElement) ifTrue: [ ^true ].

	"The goal is to eliminate anything from the theme that is constructed statically."
	frame class caseOf: {
	[ MCenterMarker ] -> [ ^false ].
	[ TLight ] -> [ ^false ].
	[ MGrid ] -> [ ^false ].
	[ MViewPuck ] -> [ ^false ].
	[ MThemeHolder ] -> [ ^false ].
	} otherwise:[ ].

	(frame isKindOf: QSkyDome) ifTrue:[ ^false ].
	(frame isKindOf: TAvatarProxy) ifTrue:[ ^false ].
	(frame isKindOf: TAvatarReplica) ifTrue:[ ^false ].
	(frame isKindOf: MFixedSurface) ifTrue:[ ^false ].
	(frame isKindOf: TGroup) ifTrue:[
		(frame objectName = 'defaultMainEntry') ifTrue:[ ^false ].
	].

	"If it can't provide a context menu, no go.  (Eliminates forum floors)"
	frame doContextMenu ifFalse:[ ^false ].

	^true! !

!QSpace methodsFor: 'persist' stamp: 'bgf 7/31/2009 03:21'!
persistOn: pStream
	super persistOn: pStream.

	"We persist defaultEntry by name because the space persists it as a child;
	 this just lets us recover the reference, and lets custom XML specify it."
	pStream storeObject: 'defaultEntry' value: defaultEntry default: nil.

	chatLog persistLogOn:pStream.
	pStream storeObject: 'welcomeDocument' value: welcomeDocument default: nil.

	pStream storeObject: 'customs' value: customs default: nil.
	pStream storeString: 'note' value: (self summaryNote) default: ('').
	pStream storeString: 'originalTemplate' value: (self originalTemplate) default: nil.
	self ctime persistFieldsOn: pStream prefix: 'ctime'.
	self mtime persistFieldsOn: pStream prefix: 'mtime'.

	"Persist attachments last so that the scene graph has already been persisted.
	 mostly just to avoid weird inside-out XML effects from persisting children before parents."
	attachments persistContentsOn: pStream.! !

!QSpace methodsFor: 'attaching' stamp: 'bgf 7/1/2010 12:13'!
pickSurfaceFor: frame alongNormal: frameRelativeNormal withinDistance: aDistance
	"Pretend we've moved the given frame:
 	 find a frame surface near it along the given axis from its origin,
	 and attach to that frame if found.
	 Used to set up attachments for unattached objects after loading,
	 or after moving them around with the 3D selection box. "
	| ray at up box  edge offset |

	at := (frame globalOrientation transformVector: frameRelativeNormal).
	up := (at y abs < 0.01) ifTrue: [  Vector3 x: 0 y: 1 z: 0 ] ifFalse: [ Vector3 x: 1 y: 0 z: 0 ].

	"See TFrame translateIntoContactWith:alongNormal:"
	box := (frame boundingBox transformedBy: (frame parent globalTransform))
				  transformedBy: (frame orientation).
	edge := Vector3
			x: (at x > 0 ifTrue: [box max x negated ] ifFalse: [box min x ]) 
			y: (at y > 0 ifTrue: [box max y  negated ] ifFalse: [box min y ]) 
			z: (at z > 0 ifTrue: [box max z  negated ] ifFalse: [box min z ]).
	offset :=  (edge * at).

	"Below, pull the ray in from my edge by (aDistance*0.3),  and extend it to (aDistance*1.5),
	 both so we know we're starting the pick from outside any candidates I'm very close to,
	 and to allow for picking numerical problems."
	ray := TRay new.
	ray localTransform: (
		Matrix4x4 identity
			translation: (frame globalPosition -  offset - (aDistance*0.3*at));
			at: at negated up: up;
			yourself ).
	ray globalTransform: (ray localTransform).
	ray excludeFrame: frame.
	ray minDistance: 0.
	ray maxDistance: aDistance*1.5.
	ray pointerPickTree: self.

	frame movedOverSelection: ray selection forEvent: nil! !

!QSpace methodsFor: 'actions' stamp: 'bgf 3/11/2010 16:34'!
placeInstantiatedObject: obj forAvatar: anAvatar 
	"Place the newly instantiated object 'in front' of the avatar."
	| pos box dim overSelection |

	QSoundEffects play:#tingle for:anAvatar.
	obj allUpdateShape.
	obj isFromTemplate: false; 
	       isPlacementLocked: false;
	       doContextMenu: true.

	"These limits bring things fairly close, to prevent losing stuff behind walls.
	 For things with drag adapters, there is less risk of that, but for plain meshes
	 we have to be conservative."

	obj rotationAroundY: anAvatar yaw.
	box := obj positioningBox.
	dim := (box extentX max: box extentY).
	pos := anAvatar translation - 
			(anAvatar lookAt * ( (dim max: 7) + ((box extentZ / 2.05)))).

	(obj isGroundBound | (dim > (anAvatar heightAboveGround*2))) ifTrue: [
		|ray|
		ray := self downRayAt:(pos x@(anAvatar globalPositionReal y + 1)@(pos z)).
		ray pointerPickTree: self.
		ray selection frame ifNotNil: [
			"Drop to floor, but not out of sight."
			pos y: ( (ray selection globalPoint y)  max: (pos y - (box extentY*0.8)) ).
		] ifNil: [
			pos y: pos y - (box extentY / 2.0) - 0.5.
		].
		overSelection := ray selection.		
	] ifFalse: [
		pos y: pos y -  (box extentY *0.5 ) - 0.5.	
	].
	obj translateIntoContactWith: pos alongNormal: (0@1@0).

	"If possible, have the drag adapter place the object,
	 so that creation works like positioning in most cases."
	(QDragAdapter dragDrop: obj 
				     intoSpace: self 
				     forEvent: (anAvatar fakePointerOverEvent: (obj isGroundBound))) 
	 ifFalse: [
		self addChild: obj.
		obj movedOverSelection: overSelection forEvent: nil.
	].
	anAvatar selectFrame: obj.
	"Future-send to create the edit box,  
	 so root: extensions and shapeChanges can settle first."
	(self getsEditBoxOnCreation: obj)  ifTrue: [ 
		anAvatar future editFrame: obj ].

	anAvatar future captureCreateFrames: { obj } description: 'Add Object'.
	^ obj.

! !

!QSpace methodsFor: 'initialize' stamp: 'jjl 1/24/2008 00:38'!
placeMarkerClass

	^ QPlaceMarker! !

!QSpace methodsFor: 'places' stamp: 'bgf 2/15/2010 22:49'!
planBounds

	"Attempt to find a good Plan Rectangle for this space."

	| minX maxX minZ maxZ box rect |

	"Shortcut - if an avatar range of motion is defined, use that, with a bit of context"
	self avatarBox ifNotNilDo: [ : abox |
		| c |
		rect := ((abox min x)@(abox min z) corner: (abox max x@abox max z)) .
			
		"Point centering calculations do '//' truncation, which we don't want."
		c := (rect topLeft + rect bottomRight / 2.0).
		"Grow the rectangle a fixed amount (14 units) to include the perspective of wall-tops
 		 in small spaces; in larger ones, just creates a bit of margin."
		^ Rectangle origin: (c - (rect extent/2.0) - 7)
						 extent: (rect extent  + 14.0).
	].

	minX := Float infinity.
	minZ := Float infinity.
	maxX := Float infinity negated.
	maxZ := Float infinity negated.

	self frameChildrenDo: [ : f |
		(f visible and: [f pickable] and: [f solid] 
			 and: [( f class includesBehavior: QPortableObject) not]
			 and: [f isInAvatar not])
		 ifTrue: [ 
			box := f boundingBox transformedBy: (f localTransform).
			minX := box min x min: minX.
			minZ := box min z min: minZ.
			maxX := box max x max: maxX.
			maxZ := box max z max: maxZ.
		].
	].

	rect := ((minX@ minZ) corner: (maxX @ maxZ)).
	"Bounding boxes are a bit loose, so we add less slop than for motion-bounds"
	^ Rectangle center: rect center  extent: (rect extent  * 1.01).! !

!QSpace methodsFor: 'places' stamp: 'bgf 6/28/2009 11:42'!
planViewCameraStanceForAvatar: avatar
	"The same as the destination stance, but a different selector so that 
	 my place treats me as having a camera-definition. 
	 (QPlace wants the cam-selector to differ from the stance-selector.)"
	^ self planViewStanceForAngle: avatar viewAngle aspect: avatar viewAspect.! !

!QSpace methodsFor: 'places' stamp: 'bgf 8/2/2009 04:42'!
planViewStanceForAngle: avAngle aspect: avAspect 
	
	"The transform at which to place a camera of the given view angle+apspect.
 	 to view the space in <Plan View> ... an overhead viewpoint."

	| box aspect center height wid gh zFudge xFudge |

	box := self planBounds.
	box area <= Float epsilon ifTrue: [ ^ nil ].

	"Rotate to the best match orientation (n-s or e-w) with the user's viewport.
	 The avatar viewAngle is the vertical viewing angle.
	 The fudge factors are used to make there not-exactly-above here,
	 which tells the stance how to decide where its 'up' should go. "
	zFudge := 0.0.
	xFudge := 0.0.
	(box isTall not = (avAspect >= 1)) ifFalse: [
		box := box rotateBy:#right centerAt: (box center).
		xFudge := 0.008 * (box width).	
	] ifTrue: [
		zFudge := -0.008 * (box height).	"Negative Z so plan north is up!!"
	].
		

	aspect := box width / box height.
	center := (box center x)@(box center y).
	wid := box height.
	gh := self groundHeightAt: center.

	height := gh + ((wid / 2.0) / ((avAngle / 2.0)) degreesToRadians tan) abs.

	"The angle+distance initially ensure that y fits; if the target is wider
	 than the user's view, adjust to get x to fit."
	aspect > avAspect ifTrue: [
		height := height * (aspect / avAspect).
	].

	^ Stance here: (center x) @(height min: 1500)@(center y)
			  there: (center x + xFudge) @ (gh) @ (center y + zFudge)
! !

!QSpace methodsFor: 'places' stamp: 'bgf 6/28/2009 12:27'!
planViewStanceForAvatar: avatar

	"For the stance we fix a consistent angle so that everyone's avatar is similar distance;
	 reduces the see-others effects."
	^ self planViewStanceForAngle: 45 aspect: avatar viewAspect.! !

!QSpace methodsFor: 'editing' stamp: 'bgf 6/23/2009 09:45'!
publicNameForLogin: userLogin
	"The screen name for the given login,  if available here
	 (i.e. if that user is present with their avatar carrying it)
	 otherwise, the simplerLoginName approximation. "

	^ (self avatarForLogin: userLogin)
		ifNil: [ QStyles simplerLoginName: userLogin ]
		ifNotNilDo: [ : avatar | avatar publicName ]! !

!QSpace methodsFor: 'initialize' stamp: 'bgf 3/3/2007 22:09'!
registerDefaultMainEntry

	"Create a default main entry frame (at 5 units above the origin
	 and register it in the island as the #mainEntry for this island.
	 Only one space per island should be registered this way."

	self assuredDefaultEntry registerGlobal:#mainEntry.
! !

!QSpace methodsFor: 'accessing' stamp: 'bgf 2/19/2008 21:33'!
removeSoundRegion: region
	soundRegions remove: region ifAbsent: []
! !

!QSpace methodsFor: 'repair-upgrade' stamp: 'bgf 6/3/2009 11:32'!
repairBalticChair: tframe
	| replace |

	(tframe isKindOf: TFrameThumb) ifFalse: [ ^ false ].
	(tframe sha = 'LR7G6SQQ6NEYQ7J7YPBS1B0TGQHSFXF') ifFalse: [ ^ false ].

	QLogger logInfo: 'Updating ', (tframe objectName).

	replace := TSceneReader new readScene:
		'<frameThumb id="1">
			<name>SwanChair</name>
			<scale>1.0</scale>
			<sha>ALCK1WOTY2OGC6SZK2TGGRUJKW2UP51</sha>
			<children id="2">
				<cube id="3">
					<name>RedSwanChair</name>
					<extent>2.7462 3.3213 2.6023</extent>
                        <location>0.0 1.66065 0.0</location>
				</cube>
			</children>
			<fileName>Content/Qwaq/Furniture/Baltic/WhiteRedSwanChair.c3z</fileName>
		</frameThumb>'.

	replace localTransform: (tframe localTransform clone).
	replace translation: (tframe translation + 
						(tframe localTransform transformVector: (4.95 @ 0 @ -4.5))).
	replace addRotationAroundY: 180.

	replace := (QSeat consistingOf: replace).
	replace seatOffset: (0@2@0).
	replace objectName: 'SwanChair'.
	replace buttonLabel: 'Conference Chair'.
	replace isAtomic: true.
	replace doContextMenu: false.
	self replaceChild: tframe with: replace.
	
	^ true

! !

!QSpace methodsFor: 'repair-upgrade' stamp: 'bgf 6/27/2009 12:37'!
repairBalticRedBench: tframe
	| replace |

	(tframe isKindOf: TFrameThumb) ifFalse: [ ^ false ].
	(tframe sha = '885JC6K482PYFX29WW6QNBN37W1WKEQ') ifFalse: [ ^ false ].

	QLogger logInfo: 'Updating ', (tframe objectName).

	replace := tframe wrapInNewGroup.
	replace objectName: 'Red-Bench'.

	"The positions are odd to compensate for the bench mesh internal offset."
	replace addChild: (
			(QSeat fakeSeat: 'Bench Seat')
				objectName: 'Supplied-Seat-1';
				translation: (-2.2 @ 1.7 @ -0.05);
				addRotationAroundY: 180.0;
				suggestedPOV: 'Near';		
				seatOffset: (0@1.0@1.0);
				yourself ).
	replace addChild: (
			(QSeat fakeSeat: 'Bench Seat')
				objectName: 'Supplied-Seat-2';
				translation: (1.9 @ 1.7 @ -0.05);
				addRotationAroundY: 180.0;
				suggestedPOV: 'Near';		
				seatOffset: (0@1.0@1.0);
				yourself ).
	replace isFromTemplate: true.
	replace doContextMenu: false.
	replace future transformChanged.

	^ true

! !

!QSpace methodsFor: 'repair-upgrade' stamp: 'bgf 6/4/2009 17:39'!
repairBirdsEye: frame
	"Mark from-template the place markers we deposit in templates."
	(frame class = QPlaceMarker 
	 and: [ (frame label = 'Bird''s Eye') or: [ frame label = 'Vista']]) ifTrue: [
		frame isFromTemplate: true.
		^ true
	].
	^ false! !

!QSpace methodsFor: 'repair-upgrade' stamp: 'bgf 6/3/2009 11:32'!
repairCaribbeanChair: tframe
	| replace |

	(tframe isKindOf: TFrameThumb) ifFalse: [ ^ false ].
	(tframe sha = 'KFAMKK2EFNYLNTQX3Z0DAMKI9EIHABQ') ifFalse: [ ^ false ].

	QLogger logInfo: 'Updating ', (tframe objectName).

	replace := TSceneReader new readScene:
		'<frameThumb id="1">
			<name>Swan-Chair</name>
			<scale>1.0</scale>
			<sha>3RW1EQRV1TONQN95E2GL4U9XFTB1WTP</sha>
			<children id="2">
				<cube id="3">
					<name>Swan-Chair-03</name>
					<extent>2.617 3.165 2.4798</extent>
					<location>0.0 1.5825 0.0</location>
				</cube>
			</children>
			<fileName>Content/Qwaq/Furniture/Caribbean/SwanChairMesh.c3z</fileName>
		</frameThumb>'.

	replace localTransform: (tframe localTransform clone).
	replace translation: (tframe translation + 
						(tframe localTransform transformVector: (5.0 @ 0 @ -4.5))).
	replace addRotationAroundY: 180.

	replace := (QSeat consistingOf: replace).
	replace objectName: 'SwanChair'.
	replace seatOffset: (0@2@0).
	replace buttonLabel: 'Conference Chair'.
	replace isAtomic: true.
	self replaceChild: tframe with: replace.
	replace doContextMenu: false.

	^ true

! !

!QSpace methodsFor: 'repair-upgrade' stamp: 'bgf 6/27/2009 12:24'!
repairCaribbeanKireiBench: tframe
	| replace |

	(tframe isKindOf: TFrameThumb) ifFalse: [ ^ false ].
	(tframe sha = 'S6T3UBYFVPVLJ8DF88VIXDOA7ETJQBW') ifFalse: [ ^ false ].

	QLogger logInfo: 'Updating ', (tframe objectName).

	replace := tframe wrapInNewGroup.
	replace objectName: 'Kirei-Bench'.

	"The positions are strange because the kirei bench mesh is badly self-offset."
	replace addChild: (
			(QSeat fakeSeat: 'Bench Seat')
				objectName: 'Supplied-Seat-1';
				translation: (8.9 @ 1.6 @ -6.9);
				addRotationAroundY: 180.0;
				suggestedPOV: 'Near';		
				seatOffset: (0@1.0@1.0);
				yourself ).
	replace addChild: (
			(QSeat fakeSeat: 'Bench Seat')
				objectName: 'Supplied-Seat-2';
				translation: (13.3 @ 1.6 @ -6.9);
				addRotationAroundY: 180.0;
				suggestedPOV: 'Near';		
				seatOffset: (0@1.0@1.0);
				yourself ).
	replace isFromTemplate: true.
	replace doContextMenu: false.
	replace future transformChanged.
	^ true

! !

!QSpace methodsFor: 'repair-upgrade' stamp: 'bgf 6/16/2009 08:50'!
repairCeiling: frame

	"Repair library, reception, and project forum ceilings 
	 to have no context menu, and be invisible from above."

	frame cullingNormal ifNotNil: [ ^ false ].

	((frame class = TFrameThumb 	"Library (V2)"
		and: [ frame objectName = 'ArcBeamedCeiling' 
			   or: [ frame objectName = 'SideCeiling' ] 
			   or: [ frame objectName = 'Main Ceiling']
			   or: [ frame objectName = 'BeamedCeiling']])
	or: [frame class = QWall "Library (V2)"
		and: [ frame objectName = 'BeamWall']]
	or: [frame class = TCube "Library (V2)"
		and: [ frame objectName = 'Partition' and: [ frame extent x truncated = 31]]]
	or: [frame class = TFrameThumb "Reception framework"
		and: [ frame objectName = 'AtriumOverhead'  ] ]
	or: [frame class = TGeometry 	"Project (Baltic/Caribbean/Pacific)"
		and: [ frame objectName = 'Roof' and: [ frame materialList size = 3] ] ]
	) ifTrue: [
		QLogger logInfo: 'Updating ceiling ', frame.
		frame cullingNormal: (Vector3 x:0 y:-1.0 z: 0).
		frame doContextMenu: false.
			
		^ true
	].
	^ false! !

!QSpace methodsFor: 'repair-upgrade' stamp: 'bgf 6/28/2009 03:54'!
repairConferenceTableDestination: tframe

	"Find any old conference-table destinations 
	 and give them the new seating-range property that allows seat assignment ."
	(tframe class == QDestination) ifFalse: [ ^ false ].
	(tframe children size > 1) ifFalse: [ ^ false ].
	(tframe objectName includesSubString: 'Conference') ifFalse: [ ^ false ].
	(tframe objectName includesSubString: 'Table') ifFalse: [ ^ false ].
	
	self setTableRange: tframe.

	"The library main table is too high for the adjusted chairs."
	(tframe globalPosition y - -5.939) abs < 0.001 ifTrue: [
		tframe translation: tframe translation - (0@0.26@0).
	].

	"The Baltic table also too high"
	(tframe globalPosition y - -0.78) abs < 0.001 ifTrue: [
		tframe translation: tframe translation - (0@0.25@0).
	].
	"The Cariibbean table also too high"
	(tframe globalPosition y - 1.067) abs < 0.001 ifTrue: [
		tframe translation: tframe translation - (0@0.25@0).
	].
	"The Pacific table also too high"
	(tframe globalPosition y - -0.347) abs < 0.001 ifTrue: [
		tframe translation: tframe translation - (0@0.25@0).
	].

	tframe isFromTemplate: true;
		     isPlacementLocked: true.
	^ true

! !

!QSpace methodsFor: 'repair-upgrade' stamp: 'bgf 4/30/2009 23:52'!
repairDesktopDisplay: tframe
	"Detect original 'Desktop Display'  
	 and replace it with Desktop Display (V2)',
 	 which is marked atomic and has a motion grip configured for the base."

	| replacement d1 d2 |

	"This recognizes the original desktop display."
	tframe class == TGroup ifFalse: [ ^ false ].
	tframe children size = 3 ifFalse: [ ^ false ].
	tframe objectName = 'Desktop Display' ifFalse: [ ^ false ].
	tframe children first class = TFrameThumb ifFalse: [ ^ false ].
	tframe children first sha = 'P7TS79YMYT97W5ZCKOT8YB5P3ZYUXBM' ifFalse: [ ^ false ].
	tframe children second class = TGeometry ifFalse: [ ^ false ].
	tframe children second objectName = 'bezel' ifFalse: [ ^ false ].
	tframe children third class = QDisplayPanel ifFalse: [ ^ false ].

	QLogger logInfo: 'Updating ', (tframe objectName).

	replacement := TSceneReader new readSceneFromZip: (self class desktopDisplayV2ZipData).
	d1 := tframe children third.
	d2 := replacement children third.

	d2 displayCard: d1 contentCard.
	d1 historyRolodex cards 
		reject: [ : cd | cd = d1 contentCard ]
		thenDo: [ : cd | d2 historyRolodex insertCard: (cd duplicate) ].

	replacement localTransform: (tframe localTransform clone).
	
	self replaceChild: tframe with: replacement.! !

!QSpace methodsFor: 'repair-upgrade' stamp: 'bgf 6/2/2009 17:16'!
repairDoorAssembly: tframe
	"These should be marked as template elements.
	 Making it atomic prevents getting menus on the nested parts."
	(tframe class = TGroup and: [ tframe objectName = 'WallWithDoubleDoors' ]) ifTrue: [
		tframe isFromTemplate: true.
		tframe doContextMenu: false.
		tframe isAtomic: true.
		^ true
	].
	^ false! !

!QSpace methodsFor: 'repair-upgrade' stamp: 'bgf 6/28/2009 03:55'!
repairFabricTaskChair: tframe
	| replace |

	(tframe isKindOf: TFrameThumb) ifFalse: [ ^ false ].
	((tframe sha = 'T6BTFUGPS9C8ORNO2RNDWDLLN56HUTJ')		"From templates"
	or: [ tframe sha = '7H8IY6NGU8ZFA9S8XN60QEHRY3XMY0J']) 	"Objects-menu FTC"
	ifFalse: [ ^ false ].

	QLogger logInfo: 'Updating ', (tframe objectName).

	replace := TSceneReader new readScene:
			'<frameThumb id="1">
			        <name>FabricTaskChair</name>
			        <scale>0.92</scale>
			        <sha>GODW326BLA7HJN6CPPA8J0BKTD4A9MB</sha>
			        <children id="2">
			                <cube id="3">
			                        <name>FabricTaskChair</name>
			                        <extent>1.81526 3.072 1.9777</extent>
			                        <location>0.0 1.536 0.0</location>            
						</cube>     </children>
			        <fileName>Content/Qwaq/Furniture/FabricTaskChairMesh.c3z</fileName>
			        <tFileTextures id="4">
			                <texture id="5" class="TFileTexture">
			                        <extent>256 192</extent>
			                        <fileName>Content/Qwaq/Textures/OrangeChairFabric.jpg</fileName>
			                        <sha>NNLAJR6HCFINQIA7I46TMOVTSF29A8R</sha>
			                        <extension>alpha</extension>
			                        <color>0.74194 0.25806 0.16129 1.0</color>           
						</texture> </tFileTextures> </frameThumb>'.

	replace localTransform: (tframe localTransform clone).
	replace := (QSeat consistingOf: replace).
	replace seatOffset: (0@1.85@0.14).
	replace suggestedPOV: 'Near'.
	replace objectName: 'FabricTaskChair'.
	replace buttonLabel: 'Task Chair'.
	replace isAtomic: true.
	self replaceChild: tframe with: replace.

	"One chair is conspicuously too low in Library V2"
	(replace translation y - 0.025) abs < 0.005 ifTrue: [
		replace translation: (replace translation + (0@0.22@0)).
	].
	
	^ true

! !

!QSpace methodsFor: 'repair-upgrade' stamp: 'bgf 6/2/2009 17:15'!
repairGround: tframe
	"These should be marked as template elements"
	((tframe class = QFixedSurface) and: [tframe objectName = 'ground']
	and: [ tframe children size = 1 ] and: [ tframe children first objectName = 'grass' ])
	ifTrue: [
		tframe isFromTemplate: true.
		tframe doContextMenu: false.
		tframe objectName: 'Ground'.
		^ true
	].
	^ false! !

!QSpace methodsFor: 'repair-upgrade' stamp: 'bgf 6/27/2009 12:06'!
repairLibraryFabricBench: tframe
	| replace |

	(tframe isKindOf: TFrameThumb) ifFalse: [ ^ false ].
	(tframe sha = '4AKP10JVHZUC01XC9O7TDC9GS574L2C') ifFalse: [ ^ false ].

	QLogger logInfo: 'Updating ', (tframe objectName).

	tframe addRotationAroundY: 180.
	tframe scale: 0.78; allUpdateShape.
	replace := tframe wrapInNewGroup.
	replace objectName: 'Fabric-Bench'.
	replace addChild: (
			(QSeat fakeSeat: 'Bench Seat')
				objectName: 'Supplied-Seat-1';
				translation: (-1.9 @ 1.75 @ 0.06);
				addRotationAroundY: 180.0;
				suggestedPOV: 'Near';		
				seatOffset: (0@1.0@0.6);
				yourself ).
	replace addChild: (
			(QSeat fakeSeat: 'Bench Seat')
				objectName: 'Supplied-Seat-2';
				translation: (1.9 @ 1.75 @ 0.06);
				addRotationAroundY: 180.0;
				suggestedPOV: 'Near';		
				seatOffset: (0@1.0@0.6);
				yourself ).
	replace doContextMenu: false;
			isFromTemplate: false.
	replace future transformChanged.
	^ true

! !

!QSpace methodsFor: 'repair-upgrade' stamp: 'bgf 8/29/2009 18:37'!
repairLibraryStack: frame

	| shelves stack |
	frame class = QDestination ifFalse: [ ^ false ].
	frame destination class = TGroup ifFalse: [ ^ false ].
	frame destination children ifEmpty: [ ^ false ].
	(frame destination children first objectName beginsWith: 'ShelfCarcasse') ifFalse: [ ^ false ].

	shelves := (frame destination findNonNested: [ : tf | tf class = QDocumentShelf ]).

	stack := (QShelfStack new)
				objectName: frame objectName;
				extent: (10@9@1.3);
				orientation: (frame orientation);
				nominalScale: (frame nominalScale);
				translation: (frame translation + (0@0.25@0));
				isFromTemplate: (frame isFromTemplate);
				isPlacementLocked: (frame isPlacementLocked);
				yourself.

	shelves doWithIndex: [ : shelf : i |
		shelf cards do: [ : cd |
			(stack shelves at: ((stack nShelves -i) max: 1)) insertCard: cd
		]].

	self addChild: stack.
	frame destroy.
	
	^ true ! !

!QSpace methodsFor: 'repair-upgrade' stamp: 'bgf 6/24/2009 18:50'!
repairLibraryTaskChair: tframe
	| replace |

	"Library (V2) Einstein room side-chair."

	(tframe isKindOf: TFrameThumb) ifFalse: [ ^ false ].
	(tframe sha = 'ASCH27GPR5RCEMK1LW3X8BT26H9K6T') ifFalse: [ ^ false ].

	QLogger logInfo: 'Updating ', (tframe objectName).

	replace := tframe wrapInNewGroup.
	replace objectName: 'Low-Chair'.
	tframe pickable: false.
	replace addChild: (
			(QSeat fakeSeat: 'Seat')
				objectName: 'Supplied-Seat';
				translation: (0.2 @ 4.7 @ -75.18);
				addRotationAroundX: -1;
				suggestedPOV: 'First Person';		
				seatOffset: (0.0 @ 1.6 @ -0.05);
				yourself ).

	replace doContextMenu: false; isFromTemplate: true.
	replace future transformChanged.

	^ true

! !

!QSpace methodsFor: 'repair-upgrade' stamp: 'bgf 6/24/2009 18:53'!
repairLibraryTaskCouch: tframe
	| replace |

	"Library (V2) Einstein room couch."

	(tframe isKindOf: TFrameThumb) ifFalse: [ ^ false ].
	(tframe sha = '170KG06765N47D2FPB26ZEKOZVG86BB') ifFalse: [ ^ false ].

	QLogger logInfo: 'Updating ', (tframe objectName).

	replace := tframe wrapInNewGroup.
	replace objectName: 'Low-Couch'.
	tframe pickable: false.

	"The position offsets compensate for the sofa mesh internal offset."
	replace addChild: (
			(QSeat fakeSeat: 'Couch Seat')
				objectName: 'Supplied-Seat-1';
				translation: (0.0 @ 5.85 @ -94.7);
				addRotationAroundX: -5;
				suggestedPOV: 'First Person';		
				seatOffset: (0.0 @ 1.7 @ -0.3);
				yourself ).
	replace addChild: (
			(QSeat fakeSeat: 'Couch Seat')
				objectName: 'Supplied-Seat-2';
				translation: (-3.0 @ 5.85 @ -94.7);
				addRotationAroundX: -5;
				suggestedPOV: 'First Person';		
				seatOffset: (0.0 @ 1.7 @ -0.3);
				yourself ).
	replace addChild: (
			(QSeat fakeSeat: 'Couch Seat')
				objectName: 'Supplied-Seat-3';
				translation: (-6.15 @ 5.85 @ -94.7);
				addRotationAroundX: -5;
				suggestedPOV: 'First Person';		
				seatOffset: (0.0 @ 1.7 @ -0.3);
				yourself ).
	replace doContextMenu: false; isFromTemplate: true.
	replace future transformChanged.
	^ true

! !

!QSpace methodsFor: 'repair-upgrade' stamp: 'bgf 6/2/2009 17:15'!
repairLight: tframe
	"Lights should be marked as template elements"
	| names |
	names := #(primaryLight secondaryLight fillLight northwestLight southeastLight southwestFill).

	((tframe class = TLight) and: [ names includes: (tframe objectName) ])
	ifTrue: [
		tframe isFromTemplate: true.
		^ true
	].
	^ false! !

!QSpace methodsFor: 'repair-upgrade' stamp: 'bgf 6/3/2009 11:33'!
repairPacificChair: tframe
	| replace |

	(tframe isKindOf: TFrameThumb) ifFalse: [ ^ false ].
	(tframe sha = 'CKZFJT1UMFFIAZICBQRSSPB8TY0LJXV') ifFalse: [ ^ false ].

	QLogger logInfo: 'Updating ', (tframe objectName).

	replace := TSceneReader new readScene:
		'<frameThumb id="1">
			<name>SwanChairOffwhite</name>
			<scale>1.0</scale>
			<sha>RPZSYZBQHMLAKVGWHVP5FXY4KBDZPVN</sha>
			<children id="2">
				<cube id="3">
                        <name>SwanChairOffwhite</name>
                        <extent>2.492 3.014 2.3615</extent>
                        <location>0.0 1.507 0.0</location>
				</cube>
			</children>
			<fileName>Content/Qwaq/Furniture/Pacific/SwanChairEbonyOffwhite.c3z</fileName>
		</frameThumb>'.

	replace localTransform: (tframe localTransform clone).
	replace translation: (tframe translation + 
						(tframe localTransform transformVector: (4.7 @ 0 @ -4.3))).
	replace addRotationAroundY: 180.

	replace := (QSeat consistingOf: replace).
	replace seatOffset: (0@2@0).
	replace objectName: 'SwanChair'.
	replace buttonLabel: 'Conference Chair'.
	replace isAtomic: true.
	self replaceChild: tframe with: replace.
	
	replace doContextMenu: false.

	^ true

! !

!QSpace methodsFor: 'repair-upgrade' stamp: 'bgf 6/27/2009 12:18'!
repairPacificWalnutBench: tframe
	| replace |

	(tframe isKindOf: TFrameThumb) ifFalse: [ ^ false ].
	(tframe sha = '2AIVSO9F94QBF4N6SGEC9QEWSTHXGB5') ifFalse: [ ^ false ].

	QLogger logInfo: 'Updating ', (tframe objectName).

	replace := tframe wrapInNewGroup.
	replace objectName: 'Walnut-Bench'.

	"The positions are strange because the bench mesh is badly self-offset."
	replace addChild: (
			(QSeat fakeSeat: 'Bench Seat')
				objectName: 'Supplied-Seat-1';
				translation: (9.5 @ 1.55 @ -7.1);
				addRotationAroundY: 180.0;
				suggestedPOV: 'Near';		
				seatOffset: (0@1.0@1.0);
				yourself ).
	replace addChild: (
			(QSeat fakeSeat: 'Bench Seat')
				objectName: 'Supplied-Seat-2';
				translation: (13.5 @ 1.55 @ -7.1);
				addRotationAroundY: 180.0;
				suggestedPOV: 'Near';		
				seatOffset: (0@1.0@1.0);
				yourself ).

	replace isFromTemplate: true.
	replace doContextMenu: false.
	replace future transformChanged.
	^ true

! !

!QSpace methodsFor: 'repair-upgrade' stamp: 'bgf 8/29/2009 17:15'!
repairProjectWorkNiche: child
	
	"In project forum work niches, replace the awkard noteboard-textures
	 with their builtin borders, using instead the noteboard flannels and a solid border."

	child class = QDestination ifFalse: [ ^ false ].
	child destination class = QDisplayNiche ifFalse: [ ^ false ].

	child destination noteboard ifNotNilDo: [ : board |
		board defaultMaterial texture ifNotNilDo: [ : txtr |
			txtr sha='A1VJ4Y3HZE0JXTX3Z23MCZUUWOV7ME0' ifTrue: [
				"Baltic Project Template"
				board material 
						texture: (QTextures find: 'noteboardGray.jpg');
						diffuseColor: (VectorColor r: 0.85 g: 0.5 b: 0.28 );
						ambientColor: (VectorColor r: 0.82 g: 0.79 b: 0.65  );
						emissiveColor: (VectorColor r: 0.1 g: 0.075 b: 0.03);
						uvScale: (0.5@1.0).
				board borderMaterial 
						texture: (QTextures find: 'greyFlannel.jpg');
						diffuseColor: (VectorColor r: 0.44 g: 0.45 b: 0.46  );
						ambientColor: (VectorColor r:0.44 g: 0.42 b: 0.36 ).
				board borderWidth: 0.16 ].
			txtr sha = '2GHLW4YWWU2XV1Y88FWC5E4QC148MBN' ifTrue: [
				"Pacific Project Template."
				board material 
						texture: (QTextures find: 'noteboardGray.jpg');
						uvScale: (0.5@1.0);
						diffuseColor: (VectorColor r: 1.0 g: 0.915 b: 0.20);
						ambientColor: (VectorColor r: 0.90 g: 0.8 b: 0.67);
						emissiveColor: (VectorColor r: 0.135 g: 0.12 b: 0.065).
				board borderMaterial 
						diffuseColor: (VectorColor r: 0.3 g: 0.3 b: 0.3);
						ambientColor: (VectorColor r: 0.25 g: 0.25 b: 0.25) ].
			txtr sha = 'D3PY1THJ4KJV9BQ5Q3JAHAJYYQZPQIH' ifTrue: [ 
				"Old default, or Caribean Project Template ..."
				board material: (QNoteBoard defaultMaterial).
				board material 
					emissiveColor: (VectorColor r: 0.2 g:0.08 b:0.05);
					uvScale: (0.5@1.0).
			].
			^true
		]].
	^ false	! !

!QSpace methodsFor: 'repair-upgrade' stamp: 'bgf 6/27/2009 13:15'!
repairReceptionSofa: tframe
	| replace |

	(tframe isKindOf: TFrameThumb) ifFalse: [ ^ false ].
	(tframe sha = '6ZS13A1R3BWXRCEJ8YBWXXBBV66Q9C8') ifFalse: [ ^ false ].

	QLogger logInfo: 'Updating ', (tframe objectName).


	replace := tframe wrapInNewGroup.
	replace objectName: 'Sand-Sofa'.
	replace addChild: (
			(QSeat fakeSeat: 'Sofa Seat')
				objectName: 'Supplied-Seat-1';
				translation: (-1.56 @ 2.4 @ 0.45);
				addRotationAroundY: 180.0;
				suggestedPOV: 'Near';		
				seatOffset: (0@1.5@0.5);
				yourself ).
	replace addChild: (
			(QSeat fakeSeat: 'Sofa Seat')
				objectName: 'Supplied-Seat-2';
				translation: (1.44 @ 2.4 @ 0.45);
				addRotationAroundY: 180.0;
				suggestedPOV: 'Near';		
				seatOffset: (0@1.5@0.5);
				yourself ).
	"So the seating pads are reachable, without the frame-thumb bounds 
	 getting in the way."
	tframe pickable: false.
	replace doContextMenu: false; isFromTemplate: true.
	^ true

! !

!QSpace methodsFor: 'repair-upgrade' stamp: 'bgf 8/6/2010 15:38'!
repairReceptionWall: tframe
	"Replace a broken geometry with a simple cube shape.  Mantis "
	
	| wall mat |
	((tframe class = QFixedSurface) 
	and: [tframe objectName = 'Fixed Surface (Atrium-Wall-Piece)']
	and: [ tframe children size = 1 ] 
	and: [ tframe children first isGeometry ]
	and: [ tframe children first objectName = 'Atrium-Wall-Piece' ])
	ifTrue: [
		(tframe nominalScale = 1.0 
		 and: [ (tframe translation - (Vector3 x: 17.378 y: -1.0772 z: -34.081 )) length < 0.1 ]
		 and: [ tframe children first materialList size > 5] ) 
		ifTrue: [
			wall := tframe children first.
			mat := (wall materialList at: 6).	
			
			tframe translation: (Vector3 x: -32.3948 y: 14.2453 z: -12.585).
			wall obliterate.
			wall := TCube new.
			wall material: mat.
			wall extent: (0.8@32@44).
			wall objectName: 'Wall'.
			tframe addChild: wall.
			
			(self find: [ : sign | sign class = QDisplaySign 
								and: [ sign contents beginsWith: 'Literature Rack' ]] ) 
			 ifNotNilDo: [ : find | 
				find size = 1 ifTrue: [
					self attach: find first to: nil.
					find first translation: (Vector3 x: -31.9502 y: 9.9346 z: -18.1213) 
					]]
		].
	
		^ true
	].
	^ false! !

!QSpace methodsFor: 'persist' stamp: 'bgf 4/21/2009 16:17'!
repairSpecialCases
	"Fish out any recognized special case objects 
	 that I have a macro-repair method for, and apply it.
	 This is done when unpickling spaces."

	self children clone do: [ : child |
		self checkForRepairCase: child.
	].! !

!QSpace methodsFor: 'repair-upgrade' stamp: 'bgf 5/31/2009 13:31'!
repairUmbrellaTable: tframe
	| replace |

	"Umbrella tables in various forums.  
       There are various specific appearances, all fortunately using the same mesh."

	(tframe isKindOf: TFrameThumb) ifFalse: [ ^ false ].

	((tframe sha = '1AR32QLMBZLIMTF6KYL0ALQHCLCPG2C') 	"Baltic red/black"
	or: [ tframe sha = 'OAR8AIB488FJZJ8UEDBUDE202DFY7IN' ]	"Caribbean/Pacific beige"
	or: [ tframe sha = 'HC0NWC28SV0X3PAD453SAEHCF2SK28X'])	"Library/Reception gray"
	ifFalse: [ ^ false ].

	QLogger logInfo: 'Updating ', (tframe objectName).

	replace := tframe wrapInNewGroup.
	replace objectName: 'Umbrella-Table'.
	tframe pickable: false.
	tframe scale: 0.97.
	replace addChild: (
			(QSeat fakeSeat: 'Patio Seat')
				objectName: 'Supplied-Seat-1';
				translation: (-2.34 @ 2.54 @ -0.8);
				addRotationAroundY: -92.0;
				suggestedPOV: 'Near';		
				seatOffset: (0@1.45@0.2);
				yourself ).
	replace addChild: (
			(QSeat fakeSeat: 'Patio Seat')
				objectName: 'Supplied-Seat-2';
				translation: (-0.68 @ 2.54 @ 2.41);
				addRotationAroundY: -2.2;
				suggestedPOV: 'Near';		
				seatOffset: (0@1.45@0.25);
				yourself ).

	replace isFromTemplate: true.
	replace doContextMenu: false.
	replace future transformChanged.
	
	^ true

! !

!QSpace methodsFor: 'theme' stamp: 'jjl 1/9/2009 17:41'!
restoreThemeElements

	collectedThemeElements := nil.
! !

!QSpace methodsFor: 'accessing' stamp: 'bgf 6/11/2009 16:33'!
serviceName
	"The service (forum) name from the Qwaq/Croquet service URL (//host/realm.../serviceName)
	 under which this ispace lives.  
	 May be nil if the client hasn't deposited it properly (relies on our client harness)"
	^ self serviceURL ifNotNilDo:[ :url | url copyAfterLast: $/]! !

!QSpace methodsFor: 'accessing' stamp: 'bgf 6/11/2009 16:33'!
serviceURL 
	"The forums URL I live in...
	 relies on our client depositing the island-property."
	^ (self island at: #ServiceURLString) ifNil: [
			QLogger logWarning: 'Early serviceURL (', (self summary), ')'.
			nil
	 	]! !

!QSpace methodsFor: 'accessing' stamp: 'bgf 3/13/2010 08:53'!
setHomeTransformFromUser: anAvatar
	"Set the transform at which in-teleporting avatars appear,
	 which is also the default linking site for portals opening into the space's #mainEntry
	 if defaultEntry has been registered as a global entry point."

	| newHome |
	newHome := anAvatar localTransform clone.
	anAvatar recordAction: #space:newHome:oldHome:
			 class: QSpace
			 arguments: { self . newHome . (self homeTransform clone) }.
	self assuredDefaultEntry localTransform: newHome.! !

!QSpace methodsFor: 'repair-upgrade' stamp: 'bgf 7/20/2009 17:42'!
setTableRange: qdest
	"Helper routine for on-the-fly forum updates.
 	 Set the sitting radius for a QDestination to enclose its bounding box plus some slop.
	 This gives the tables the behavior or automatically seating someone who goes to it,	
	 if there is a seat available."
	qdest seatingRange = 0 ifTrue: [
		qdest seatingRange: 3.5.
		QLogger logInfo: 'Seating Range set for ', (qdest objectName).
	].

! !

!QSpace methodsFor: 'accessing' stamp: 'bgf 2/19/2008 21:34'!
soundRegionContaining: aPoint

	"If the space contains a QSoundRegion that encloses this point,
	 return it; other wise return nil."

	^ soundRegions
		detect: [ : region| region containsPoint: aPoint ] 
		ifNone: [ nil ].

! !

!QSpace methodsFor: 'initialize' stamp: 'hrs 3/14/2009 10:52'!
spawnAutoImportJob
	"Have someone check periodically for imports to be released from customs.
	Override this with a no-op if you don't want that."
	^QAutoImportJob defaultForForumSpace: self! !

!QSpace methodsFor: 'initialize' stamp: 'ar 6/3/2007 13:06'!
spawnAutoSaveJob
	"Spawn an auto-save job if this space wants to be saved regularly.
	Different spaces have different tradeoffs so each type of space can define 
	its own. The default is *NO* auto-save at all."
	^nil! !

!QSpace methodsFor: 'attaching' stamp: 'bgf 7/31/2009 02:49'!
stickersAttachedTo: aSurface
	"All the frames considering themselves attached to that surface."
	^ attachments stickersAttachedTo: aSurface! !

!QSpace methodsFor: 'attaching' stamp: 'bgf 8/1/2009 02:11'!
stickersOn: surface
	"All the frames that are directly attached to the surface,
	 or to some other sticker that is on the surface."	
	^ attachments stickersOn: surface! !

!QSpace methodsFor: 'accessing' stamp: 'bgf 6/11/2009 15:26'!
summaryNote
	"A short string note (under user control) describing the forum."

	^ (self island summaryNote) ifNil: [ '' ]! !

!QSpace methodsFor: 'accessing' stamp: 'bgf 6/12/2009 09:21'!
summaryNote: aString
	"A short string note (under user control) describing the forum."

	"This is an alias for a (Qwaq-specific) island property;
	 stored this way for easy editing in a forum space properties sheet,
	 while the save/load mechanisms can operate on islands.
	 The save system mirrors the value into the forum manifest
	 so it can be shown in the service chooser. "
	self island summaryNote: aString.
! !

!QSpace methodsFor: 'attaching' stamp: 'bgf 7/31/2009 02:47'!
surfaceAttachmentOf: someFrame
	"Attachment manager query - what surface if any is the given frame attached to?"
	^ attachments surfaceAttachmentOf: someFrame! !

!QSpace methodsFor: 'accessing' stamp: 'bgf 7/29/2009 14:01'!
terrainFrames
	"Array of space-child frames to be used for ground height estimation (down-picking.)"
	^ self frameChildren select: [ : f |
		 (f isKindOf: QFloor) 
		 or: [ f pickable and: [ f solid ] and: [ f isInAvatar not ] ] ].
		
		"A little trickery there... we make a bunch of indoor 'terrain' objects	
	 	 like walls, etc, inert to context menus to simplify interactions"! !

!QSpace methodsFor: 'chat' stamp: 'bgf 8/29/2008 23:41'!
textMessage: txt fromAvatar: replicaOrNil atTime: utcSecs
	"Add a message to the space's main chat log;
	 if replica is nil, the message is attributed to the system-user 'Forums Host' "
	chatLog addMessage: txt fromAvatar: replicaOrNil atTime: utcSecs.! !

!QSpace methodsFor: 'chat' stamp: 'hrs 2/8/2009 12:17'!
textMessage: txt fromAvatar: replicaOrNil atTime: utcSecs place: place
	"Add a message to the space's main chat log;
	 if replica is nil, the message is attributed to the system-user 'Forums Host' "
	chatLog addMessage: txt fromAvatar: replicaOrNil atTime: utcSecs place: place.! !

!QSpace methodsFor: 'accessing' stamp: 'jjl 2/27/2008 17:04'!
uiPreference

	^uiPreference! !

!QSpace methodsFor: 'accessing' stamp: 'jjl 2/28/2008 17:18'!
uiStyle

	^uiStyle! !

!QSpace methodsFor: 'accessing' stamp: 'jjl 5/2/2008 14:00'!
uiStyle:style

	uiStyle := style.
	miramarUI := (uiStyle = #Miramar)! !

!QSpace methodsFor: 'initialize' stamp: 'hrs 2/6/2009 12:35'!
uniqueIntegerAmong: predicateBlock checking: readBlock
	"Answer an integer that can be used to generate a name that will be unique among all those
	space objects that pass predicateBlock. If readBlock is non-null, it is applied to each passing
	object to ensure that none answer the value produced.
	
	E.g., space uniqueIntegerOf: [:frame | frame isKindOf: TCube] 
			checking: [:cube | (cube objectName copyAfter: $-) asInteger]

	The method first determins one less than the number of objects passing predicateBlock.
	This number is incremented as necessary to ensure that readBlock produces no matching values.
	Therefore:
		- A value of zero indicates that there is only one unique object that passes predicateBlock.
		- Objects can store their results, which will remain unique as objects are added
		  and deleted. However, if a deleted object is re-added after other object have been added,
		 it will need its unique integer recomputed."
	| found counter |
	found := self find: predicateBlock.
	counter := found size - 1.
	readBlock ifNil: [^counter].
	[  
		(found noneSatisfy: [:f| (readBlock value: f) = counter]) ifTrue: [^counter].
		counter := counter + 1.
	] repeat.! !

!QSpace methodsFor: 'editing' stamp: 'bgf 8/9/2009 15:19'!
usersAtDestination: aFrame

	"Find all avatars (TAvatarReplica instances) within the given distance
	 of the given global position."

	^ self allUsers select: [ :avatar |
		(avatar isAtDestinationFrame: aFrame)
	].! !

!QSpace methodsFor: 'editing' stamp: 'bgf 3/16/2008 00:31'!
usersWithin: distance of: aVector3

	"Find all avatars (TAvatarReplica instances) within the given distance
	 of the given global position."

	^ self allUsers select: [ :avatar |
		(avatar translation distanceTo: aVector3) < distance
	].! !

!QSpace methodsFor: 'undo-redo' stamp: 'bgf 3/14/2010 16:13'!
welcome: cardPickle fromUser: anAvatar

	"Undo support for changing the welcome document."
	self welcomeDocument: (
			cardPickle ifNotNil: [ (TSceneReader new)
								preserveObjectIDs: true;
								readScene: cardPickle ] ) ! !

!QSpace methodsFor: 'editing' stamp: 'bgf 4/7/2009 22:59'!
welcomeDocFromProperties: fileCard
	"Property-sheet support.  
	 Verify and set the chosen welcome-document card, if it is valid."
	fileCard ifNil: [ 
		self welcomeDocument: nil.
		^ nil 	
	].

	"This is belt&braces, since the apply-validatar already tries to confirm,
	 but best to be sure before duplicating."
	((fileCard isKindOf: QFileCard) and: [ fileCard canUseAsForumWelcome ]) ifTrue: [
		^ self welcomeDocument: (fileCard duplicate).
	].

	QLogger logWarning: 'Welcome-Document is not a usable document type. '.
	self welcomeDocument: nil.
	^ nil! !

!QSpace methodsFor: 'accessing' stamp: 'bgf 12/26/2008 13:07'!
welcomeDocument
	"A QFileCard representing the welcome-document for this space."
	^ welcomeDocument! !

!QSpace methodsFor: 'accessing' stamp: 'bgf 4/29/2009 21:51'!
welcomeDocument: aQFileCard
	"A document-typed file card holding the welcome-document 
	 that should be presented to new arrivals..."
	welcomeDocument = aQFileCard ifFalse: [
		self discardWelcomeApp.
		welcomeDocument ifNotNil: [ welcomeDocument unregisterContent ].
		welcomeDocument := aQFileCard.
		welcomeDocument ifNotNil: [ welcomeDocument registerContent ].
	].
! !

!QSpace methodsFor: 'accessing' stamp: 'bgf 7/24/2009 23:15'!
welcomeDocumentAsApplicationForUser: anAvatar

	"For the welcome-app, the persistent data we hold is a document file-card.
	 To show it, we start an application in an unrooted display sheet (welcomeAppSheet) upon request."

	welcomeDocument ifNil: [ ^ nil ].
	welcomeAppSheet ifNotNil: [
		welcomeAppSheet embeddedApp
			ifNotNilDo: [ :eapp |  eapp isRunning ifTrue: [^ eapp] ]].
	
	self discardWelcomeApp.

	welcomeAppSheet := (QDisplaySheet new).
	self addChild: welcomeAppSheet.
	"We don't want the user messing with this... lock it down and hide it."
	welcomeAppSheet visible: false; pickable: false; doContextMenu: false; 
					 translation: 0@-100@0; rotationAroundX: 90;
					 isDropEnabled: false;
					 isPlacementLocked: true;
					 isFromTemplate: true;
					 objectName: 'Welcome Document Holder';
					 shouldPersist: false.	"Also keeps it out of the inspector."

	welcomeAppSheet displayCard: welcomeDocument fromUser: anAvatar.

	"Did it set up properly?"
	welcomeAppSheet embeddedApp ifNotNilDo: [ : app |
		app startRunning: (anAvatar sessionID).
		^  app
	] ifNil: [
		QLogger logWarning: 'Could not start application for welcome document.'.
		self discardWelcomeApp.
		^ nil
	].
! !

!QSpace methodsFor: 'accessing' stamp: 'bgf 4/29/2009 22:00'!
welcomeDocumentReference
	"The QAbstractDocument representing the welcome document for this space."
	^ welcomeDocument ifNotNil: [ welcomeDocument content ]! !

!TSpace methodsFor: '*Qwaq-Objects-Space' stamp: 'bgf 2/23/2009 22:43'!
aboutToQuit
	"Notification from the client that it is about to log out; backstop."! !

!TSpace methodsFor: '*Qwaq-Objects-Space' stamp: 'bgf 7/26/2009 13:28'!
bestSeatWithin: tbox nearest: avatarPosition
	"Spaces always seat the generic way; this is the starting point for seat-finding."
	^ self genericBestSeatWithin: tbox nearest: avatarPosition
	! !

!TSpace methodsFor: '*Qwaq-Objects-Space' stamp: 'bgf 6/28/2009 09:47'!
meetingState
	"The QMeetingState record where state of a meeting 
	  (the various leader-controlled aspects of a  forum session)
	  are shared across participants.
	 NOT appropriate for get: "
	^ self island at: #QwaqMeetingState
				 ifAbsentPut: [ QMeetingState new ]
! !

!TSpace methodsFor: '*Qwaq-Objects-Space' stamp: 'bgf 3/1/2008 21:01'!
soundRegionContaining: foo
	"Backstop for QSpace"
	^ nil! !

!TSpace methodsFor: '*Qwaq-Objects-Space' stamp: 'bgf 8/30/2008 00:01'!
textMessage: txt fromAvatar: replicaOrNil atTime: utcSecs
	"Backstop for in-space generated text-chat from the avatar replica;
	 TSpaces that are not QSpaces have no default chat log."! !

!TSpace methodsFor: '*Qwaq-Objects-Space' stamp: 'bgf 3/17/2008 02:51'!
uiStyle: aStyle
	"Backstop for Miramar extensions... set on space as part of policy installation;
	 needed right now so we can run the benchmark."! !

!Time class methodsFor: '*Qwaq-Objects' stamp: 'ar 8/9/2010 16:43'!
getClockLogsAndContinueLoggingIf: aBoolean
	"Answer a pair of Arrays of the most recent N samples of the mucrosecond and millisecond clocks.
	 These arrays have the most recent time first and the oldest time last.  Turn logging on or off depending
	 on the argument."
	"Time getClockLogsAndContinueLoggingIf: false"
	"Time getClockLogsAndContinueLoggingIf: true"
	"| r a b |
	 r := Time getClockLogsAndContinueLoggingIf: false.
	 a := r first.
	 b := r last.
	 (1 to: a size - 1) select: [:i| (a at: i) < (a at: i + 1) or: [(b at: i) < (b at: i + 1)]]"
	| them n usecs uidx msecs midx |
	them := self primitiveGetClockLogAddressesAndLogIf: aBoolean.
	n := them at: 1.
	usecs := Alien forPointer: (them at: 2).
	uidx := them at: 3. "this is the index of the next free entry in the usecs log 0-relative"
	msecs := Alien forPointer: (them at: 4).
	midx := them at: 5. "this is the index of the next free entry in the msecs log 0-relative"

	usecs := (1 to: n) collect:[:idx| usecs unsignedLongLongAt: idx - 1 * 8 + 1].
	usecs := (usecs copyFrom: uidx+2 to: n), (usecs copyFrom: 1 to: uidx+1).

	msecs := (1 to: n) collect:[:idx| msecs unsignedLongAt: idx - 1 * 4 + 1].
	msecs := (msecs copyFrom: midx+2 to: n), (msecs copyFrom: 1 to: midx+1).

	^{usecs reversed. msecs reversed} "msecs are unsigned 32-bit ints"! !

!TFrameThumb class methodsFor: '*Qwaq-Objects-Document' stamp: 'bgf 4/1/2010 19:10'!
existingDocumentFileName: thumbRef forClient: aClient
	"Document protocol - return the name of an existing local file holding this texture."
	| sha |
	sha := thumbRef get: #sha.
	sha ifNotNil: [ 
		(TContentManager default cacheManager findFileName: sha) 
			ifNotNilDo: [ : cached | ^ cached ]].
	QLogger logInfo: 'No cached model for sha ', sha.
	^ nil! !

!TFrameThumb class methodsFor: '*Qwaq-Objects-Document' stamp: 'bgf 4/1/2010 18:52'!
fetch: document toFileNamed: fileName forClient: owner
	"Simple delegation back to the generic."
	^ QSharedDocument fetch: document toFileNamed: fileName forClient: owner! !

!TFrameThumb class methodsFor: '*Qwaq-Objects-Document' stamp: 'bgf 8/1/2009 18:32'!
getContent: thumbRef  forClient: aUI
	"Document protocol - return the raw data bytes (the file content) for this texture.
	 These are the encoded zipped c3z bits"

	(TContentManager default cacheManager find: (thumbRef get: #sha)) 
		ifNotNilDo: [ : cached | ^ cached ].

	^nil! !

!TFrameThumb class methodsFor: '*Qwaq-Objects-Document' stamp: 'bgf 4/1/2010 18:51'!
saveDocument: thumbRef asFile: newName forClient: aUI
	"Document protocol.  See base definition in QAbstractDocument"
	^ QSharedDocument saveDocument: thumbRef asFile: newName forClient: aUI! !

!TFrameThumb class methodsFor: '*Qwaq-Objects-Document' stamp: 'bgf 2/12/2009 21:59'!
writeManifestEntry: thumbRef on: xmlWriter checkingDoneSet: doneShas

	"Write the Qwaq Forums manifest entry for the given TFrameThumb reference."
	| fName sha tag |
 	sha := thumbRef get: #sha.	
	(sha isNil or: [doneShas includes: sha]) ifFalse: [
		fName := thumbRef get:#fileName.
		tag := 'resource'.
		fName ifNotNil: [
			xmlWriter startTag: tag.
			xmlWriter attribute: 'type' value: (self resourceType).
			xmlWriter attribute: 'uri' value: fName.
			xmlWriter attribute: 'sha' value: sha.
			xmlWriter endEmptyTag: tag.
			doneShas add: sha.
		]].! !

!TFrameThumb methodsFor: '*Qwaq-Objects-Documents' stamp: 'bgf 2/7/2010 00:38'!
asServiceManagerPath
	"QServiceManager get/put argument we use to refer to this resource ."
	self isBundled ifTrue: [
		"Path support for moving bundled documents to the service provider."
		^ 'system-resources/bundled/', self sharedName
	].
	self isDocumentInForumStorage ifTrue: [
		^ '/services/', (self serviceName), '/', self sharedName.
	] ifFalse: [
		^ self sharedName
	].
	! !

!TFrameThumb methodsFor: '*Qwaq-Objects-Documents' stamp: 'bgf 4/1/2010 18:43'!
canBeSavedLocally
	"Abstract-Document protocol.
	 Indicates that doSaveAsForAvatar: is supported and appropriate for this instance."
	^ fileName notNil! !

!TFrameThumb methodsFor: '*Qwaq-Objects-Documents' stamp: 'bgf 4/1/2010 18:43'!
doSaveAsForAvatar: anAvatarReplica
	"Save the texture locally in the client of the current avatar.
	 Use of this is guarded by  # canBeSavedLocally  "
	self canBeSavedLocally ifTrue: [
		QSaveDocumentJob save: self forUser: anAvatarReplica
	] ifFalse: [
		anAvatarReplica signal: #alertMessage with: 'That image is not available to be saved.'.
	].! !

!TFrameThumb methodsFor: '*Qwaq-Objects-Documents' stamp: 'bgf 4/1/2010 18:47'!
fileChooserPattern
	^ {{ '3D-Model' .  '*.c3z' }}! !

!TFrameThumb methodsFor: '*Qwaq-Objects-Documents' stamp: 'bgf 4/1/2010 18:45'!
fileChooserPatternTweak
	^ {{ '3D-Model' . { '*.c3z' . '*.c3d' }}}! !

!TFrameThumb methodsFor: '*Qwaq-Objects-Documents' stamp: 'bgf 4/1/2010 18:49'!
fileSaveLabel
	^ 'Save 3D Model'! !

!TFrameThumb methodsFor: '*Qwaq-Objects-Documents' stamp: 'bgf 2/7/2010 00:39'!
isDocumentInForumStorage
	"Abstract-Document protocol.
	 True if the resource is stored in the service manager file space for this forum;
	 false (for instance) for bundled textures, or private documents in user inventory.
	 This is part of the filecard/document protocol for resource transfers."
	^ fileName notNil
	   and: [
		(fileName beginsWith: 'sharedDocuments/') or: [ fileName beginsWith: 'resources/' ]].
! !

!TFrameThumb methodsFor: '*Qwaq-Objects-Documents' stamp: 'bgf 2/6/2010 19:39'!
isDocumentInInventory
	"Abstract-Document protocol.
	 True if the resource is stored in the service manager file space for this forum;
	 false (for instance) for bundled textures, or private documents in user inventory.
	 This is part of the filecard/document protocol for resource transfers."
	^ fileName notNil
	   and: [
		(fileName beginsWith: '/users/') ]! !

!TFrameThumb methodsFor: '*Qwaq-Objects-Documents' stamp: 'bgf 2/7/2010 11:50'!
isRenamableContent
	"Document protocol; is this resource a user-modifiable application-document?"
	^ false! !

!TFrameThumb methodsFor: '*Qwaq-Objects-Documents' stamp: 'bgf 5/14/2010 14:06'!
isRepositoryDocument
	^ false! !

!TFrameThumb methodsFor: '*Qwaq-Objects-Documents' stamp: 'bgf 4/1/2010 18:56'!
saveComplete: cb
	"Backstop for save-copy protocols."! !

!TFrameThumb methodsFor: '*Qwaq-Objects-Documents' stamp: 'bgf 2/7/2010 12:05'!
serviceURL: aSvcUrlString
	"For polymorphism with file-textures and documents"! !

!TFrameThumb methodsFor: '*Qwaq-Objects-Documents' stamp: 'bgf 2/6/2010 19:39'!
sharedName
	"Document protocol.  
	 The service-manager path, relative to forum (or 'absolute' for bundled and user-inventory), 
	 where this item resides on the server."
	^ self fileName! !

!TFrameThumb methodsFor: '*Qwaq-Objects-Documents' stamp: 'bgf 4/1/2010 18:48'!
shortName
	^ self sharedName ifNotNil: [ : sn | sn copyAfterLast: $/ ] ifNil: [ 'model' ]! !

!TFrameThumb methodsFor: '*Qwaq-Objects-Documents' stamp: 'bgf 4/1/2010 18:58'!
suffix
	^ 'c3z'! !

CPlayer subclass: #QPlainAvatarDetails
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Objects-User'!

!QPlainAvatarDetails commentStamp: '<historical>' prior: 0!
Details pane for the simple avatar.
Allows picking or clearing the face and badge pictures.

Note - update signalling is not complete: 
changes made directly in Preferences or in the space
will not be picked up until a space change.!

QPlainAvatarDetails defineFields: '<?xml  version="1.0" ?>
<fields>
	<field toGet="chooser" toSet="chooser:" changeEvent="chooserChanged">chooser</field>
	<field toGet="photo" toSet="photo:" changeEvent="photoChanged">photo</field>
	<field toGet="badge" toSet="badge:" changeEvent="badgeChanged">badge</field>
	<field toGet="getPhoto" toSet="getPhoto:" changeEvent="getPhotoChanged">getPhoto</field>
	<field toGet="getBadge" toSet="getBadge:" changeEvent="getBadgeChanged">getBadge</field>
	<field toGet="clearBadge" toSet="clearBadge:" changeEvent="clearBadgeChanged">clearBadge</field>
	<field toGet="clearPhoto" toSet="clearPhoto:" changeEvent="clearPhotoChanged">clearPhoto</field>
	<field toGet="photoDirty" toSet="photoDirty:" changeEvent="photoDirtyChanged">photoDirty</field>
	<field toGet="badgeDirty" toSet="badgeDirty:" changeEvent="badgeDirtyChanged">badgeDirty</field>
</fields>'!

!QPlainAvatarDetails class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QPlainAvatarDetails methodsFor: 'accessing' stamp: 'bgf 9/16/2007 19:16'!
badge
	"Answer the badge of the receiver"
	<bewareOf: #badgeChanged>
	^self propertyValueAt: #badge! !

!QPlainAvatarDetails methodsFor: 'accessing' stamp: 'bgf 9/16/2007 19:16'!
badge: aValue
	"Modify the receiver's badge"
	^self propertyValueAt: #badge put: aValue with: #badgeChanged! !

!QPlainAvatarDetails methodsFor: 'accessing' stamp: 'bgf 9/17/2007 02:31'!
badgeDirty
	"Answer the badgeDirty of the receiver"
	<bewareOf: #badgeDirtyChanged>
	^self propertyValueAt: #badgeDirty! !

!QPlainAvatarDetails methodsFor: 'accessing' stamp: 'bgf 9/17/2007 02:31'!
badgeDirty: aValue
	"Modify the receiver's badgeDirty"
	^self propertyValueAt: #badgeDirty put: aValue with: #badgeDirtyChanged! !

!QPlainAvatarDetails methodsFor: 'accessing' stamp: 'bgf 9/17/2007 03:40'!
badgeHelp
	^ 'A picture that is displayed on your avatar''s chest.'! !

!QPlainAvatarDetails methodsFor: 'accessing' stamp: 'bgf 9/16/2007 19:10'!
chooser
	"Answer the chooser of the receiver"
	<bewareOf: #chooserChanged>
	^self propertyValueAt: #chooser! !

!QPlainAvatarDetails methodsFor: 'accessing' stamp: 'bgf 9/16/2007 19:10'!
chooser: aValue
	"Modify the receiver's chooser"
	^self propertyValueAt: #chooser put: aValue with: #chooserChanged! !

!QPlainAvatarDetails methodsFor: 'accessing' stamp: 'bgf 9/17/2007 02:31'!
clearBadge
	"Answer the clearBadge of the receiver"
	<bewareOf: #clearBadgeChanged>
	^self propertyValueAt: #clearBadge! !

!QPlainAvatarDetails methodsFor: 'accessing' stamp: 'bgf 9/17/2007 02:31'!
clearBadge: aValue
	"Modify the receiver's clearBadge"
	^self propertyValueAt: #clearBadge put: aValue with: #clearBadgeChanged! !

!QPlainAvatarDetails methodsFor: 'accessing' stamp: 'bgf 9/17/2007 02:31'!
clearPhoto
	"Answer the clearPhoto of the receiver"
	<bewareOf: #clearPhotoChanged>
	^self propertyValueAt: #clearPhoto! !

!QPlainAvatarDetails methodsFor: 'accessing' stamp: 'bgf 9/17/2007 02:31'!
clearPhoto: aValue
	"Modify the receiver's clearPhoto"
	^self propertyValueAt: #clearPhoto put: aValue with: #clearPhotoChanged! !

!QPlainAvatarDetails methodsFor: 'accessing' stamp: 'bgf 9/16/2007 19:16'!
getBadge
	"Answer the getBadge of the receiver"
	<bewareOf: #getBadgeChanged>
	^self propertyValueAt: #getBadge! !

!QPlainAvatarDetails methodsFor: 'accessing' stamp: 'bgf 9/16/2007 19:16'!
getBadge: aValue
	"Modify the receiver's getBadge"
	^self propertyValueAt: #getBadge put: aValue with: #getBadgeChanged! !

!QPlainAvatarDetails methodsFor: 'accessing' stamp: 'bgf 9/16/2007 19:16'!
getPhoto
	"Answer the getPhoto of the receiver"
	<bewareOf: #getPhotoChanged>
	^self propertyValueAt: #getPhoto! !

!QPlainAvatarDetails methodsFor: 'accessing' stamp: 'bgf 9/16/2007 19:16'!
getPhoto: aValue
	"Modify the receiver's getPhoto"
	^self propertyValueAt: #getPhoto put: aValue with: #getPhotoChanged! !

!QPlainAvatarDetails methodsFor: 'events' stamp: 'eem 3/19/2010 17:30'!
getPictureFromUser: prompt scaledToSize: aSize cropToSquare: doSquareCrop

	"Choose a picture from disk; return a Form or nil."

	| fullPath form size | 
	[
		fullPath := QFileDialog
					openFile: QFileDialog defaultImageSpec
					path: (QFileDialog getFileLocation: #pictures)
					label: prompt
	] on: Error do: [
		self inform: 'Could not open file chooser.'.
		^ nil
	].
	fullPath isEmptyOrNil ifTrue:[ ^ nil ].
	[
		form := ImageReadWriter formFromFileNamed: fullPath.
	] on: Error do: [
		self inform: 'That picture cannot be read.'.
		^ nil
	].
	form ifNil: [ ^ nil ].

	doSquareCrop ifTrue: [
		"Take a square out of the middle."
		size := form extent copy.
		size x > size y ifTrue: [
			form := form contentsOfArea: 
				(((size x - size y / 2) rounded @0) extent: (size y@size y)).
		].
		size x < size y ifTrue: [
			form := form contentsOfArea: 
				(0@((size y - size x / 2) rounded) extent: (size x@size x)).
		].
	].
	^ form scaledToSize: aSize
	! !

!QPlainAvatarDetails methodsFor: 'initialize' stamp: 'bgf 9/17/2007 02:32'!
initialize
	| |
	super initialize.

	photo := CPlayer new.
	getPhoto := CButton new.
	clearPhoto := CButton new.
	badge := CPlayer new.
	getBadge := CButton new.
	clearBadge := CButton new.

	photoDirty := false.
	badgeDirty := false.! !

!QPlainAvatarDetails methodsFor: 'events' stamp: 'bgf 8/5/2008 19:22'!
onApply

	"The chooser is about to save the avatar.
	 The photos are outside, in preferences..."

	[
		chooser owner preferences smallPicture: (photo graphicForm).
		chooser owner preferences badgePicture: (badge graphicForm).
		chooser owner preferences storeXml.
		badgeDirty := false.
		photoDirty := false.

	] on: Error do: [ :ex | 
		self inform: 'Your picture changes could not be saved:\\' withCRs,  (ex description)
	].! !

!QPlainAvatarDetails methodsFor: 'events' stamp: 'bgf 9/25/2007 19:44'!
onClearBadge
	<on: fire in: clearBadge>

	badge graphicForm: nil.
	chooser detailsDirty: ((badgeDirty := (chooser owner preferences badgePicture notNil)) or: [ photoDirty ]).
	chooser liveCostume future badge: nil.! !

!QPlainAvatarDetails methodsFor: 'events' stamp: 'bgf 9/25/2007 19:55'!
onClearPhoto
	<on: fire in: clearPhoto>

	"Back to Mr. Smiley   (Jack, not George)"

	photo graphicForm: QUserPreferences defaultPicture.
	chooser detailsDirty: ((photoDirty := (chooser owner preferences smallPicture notNil)) or: [ badgeDirty ]).
	chooser liveCostume future setFace: QUserPreferences defaultPictureTexture.
! !

!QPlainAvatarDetails methodsFor: 'events' stamp: 'bgf 9/17/2007 03:19'!
onClickBadge
	<on: mouseDown in: badge>
	self onGetBadge.! !

!QPlainAvatarDetails methodsFor: 'events' stamp: 'bgf 9/17/2007 08:51'!
onClickPhoto
	<on: mouseDown in: photo>
	self onGetPhoto.! !

!QPlainAvatarDetails methodsFor: 'events' stamp: 'bgf 8/5/2008 19:18'!
onGetBadge
	| form txtr ref |
	<on: fire in: getBadge>

	form := self getPictureFromUser: 'Choose Badge Picture' scaledToSize: 96@96 cropToSquare: false.
	form ifNotNil: [
		chooser mouseCursor: Cursor wait.
		QActivityCapture action: #PlainAvatarDetailsGetBadge.
		"A temporary tfiletexture for previewing..."
		[   txtr := chooser owner preferences
				putPublicPhoto: form asFile: '_tmp_badge_picture.png'.
			ref := TIslandCopier new
					import:(TIslandCopier new export: txtr) 
					to: chooser liveCostume island.
			chooser liveCostume future badge: ref.
			chooser detailsDirty: (badgeDirty := true).
			badge graphicForm: form; graphicFit: #scale.
		] on: Error do: [ :ex|
			chooser mouseCursor: nil.
			self inform: 'The picture could not be applied to your avatar:\\' withCRs, (ex description)
		].
		chooser mouseCursor: nil.
	].! !

!QPlainAvatarDetails methodsFor: 'events' stamp: 'bgf 8/5/2008 19:18'!
onGetPhoto
	| form txtr ref |
	<on: fire in: getPhoto>

	form := self getPictureFromUser: 'Choose Avatar Picture' scaledToSize: 96@96 cropToSquare: true.
	form ifNotNil: [
		chooser mouseCursor: Cursor wait.
		QActivityCapture action: #PlainAvatarDetailsGetPhoto.
		[
			"A temporary tfiletexture for previewing..."
			txtr := chooser owner preferences
				putPublicPhoto: form asFile: '_tmp_small_picture.png'.

			ref := TIslandCopier new
					import:(TIslandCopier new export: txtr) 
					to: chooser liveCostume island.
			chooser liveCostume future setFace: ref.
			chooser detailsDirty: (photoDirty := true).
			photo graphicForm: form; graphicFit: #scale.

		] on: Error do: [ :ex|
			chooser mouseCursor: nil.
			self inform: 'The picture could not be applied to your avatar:\\' withCRs, (ex description)
		].
		chooser mouseCursor: nil.
	].! !

!QPlainAvatarDetails methodsFor: 'events' stamp: 'bgf 9/17/2007 09:25'!
onPrefsBadgeChanged

	badge graphicForm: chooser owner preferences badgePicture.
	badge graphicForm ifNotNil: [
		badge graphicFit: #scale
	].

	badgeDirty := false.
	chooser detailsDirty: photoDirty.! !

!QPlainAvatarDetails methodsFor: 'events' stamp: 'bgf 9/17/2007 09:25'!
onPrefsPhotoChanged

	photo graphicForm: chooser owner preferences smallPicture.
	photo graphicForm ifNotNil: [
		photo graphicFit: #scale
	].

	photoDirty := false.
	chooser detailsDirty: badgeDirty.! !

!QPlainAvatarDetails methodsFor: 'events' stamp: 'bgf 9/17/2007 09:12'!
onReset
	"Called from the chooser - make the panel match the in-space costume."
	| live prefs |

	live := (chooser liveCostume).
	live ifNil: [ ^ self ].
	prefs := chooser owner preferences.
	"Get the preferences avatarCostume pictures and show their forms."

	self stopScript: #onPrefsPhotoChanged.
	self stopScript: #onPrefsBadgeChanged.

	photo graphicForm: (prefs smallPicture).
	photo graphicForm ifNotNil: [
		photo graphicFit: #scale.
	].

	badge graphicForm: (prefs badgePicture).
	badge graphicForm ifNotNil: [
		badge graphicFit: #scale.
	].

	photoDirty := false.
	badgeDirty := false.

	self startScript: #onPrefsBadgeChanged when: { prefs . #badgePictureChanged }.
	self startScript: #onPrefsPhotoChanged when: { prefs . #smallPictureChanged }.! !

!QPlainAvatarDetails methodsFor: 'events' stamp: 'bgf 12/30/2009 15:29'!
onUnready
	"When the container is taken out of session."

	self stopScript: #onPrefsPhotoChanged.
	self stopScript: #onPrefsBadgeChanged.! !

!QPlainAvatarDetails methodsFor: 'accessing' stamp: 'bgf 9/16/2007 19:16'!
photo
	"Answer the photo of the receiver"
	<bewareOf: #photoChanged>
	^self propertyValueAt: #photo! !

!QPlainAvatarDetails methodsFor: 'accessing' stamp: 'bgf 9/16/2007 19:16'!
photo: aValue
	"Modify the receiver's photo"
	^self propertyValueAt: #photo put: aValue with: #photoChanged! !

!QPlainAvatarDetails methodsFor: 'accessing' stamp: 'bgf 9/17/2007 02:31'!
photoDirty
	"Answer the photoDirty of the receiver"
	<bewareOf: #photoDirtyChanged>
	^self propertyValueAt: #photoDirty! !

!QPlainAvatarDetails methodsFor: 'accessing' stamp: 'bgf 9/17/2007 02:31'!
photoDirty: aValue
	"Modify the receiver's photoDirty"
	^self propertyValueAt: #photoDirty put: aValue with: #photoDirtyChanged! !

!QPlainAvatarDetails methodsFor: 'accessing' stamp: 'bgf 9/17/2007 02:16'!
pictureHelp
	^ 'A photograph (usually a portrait of you) that is displayed as the face of your avatar.'! !

!QPlainAvatarDetails methodsFor: 'initialize' stamp: 'bgf 9/17/2007 03:38'!
setupCostume

	|  label x1 x2 y1 |
	super setupCostume.
	
	self extent: 280@160.
	self balloonHelp: 'Choose pictures for the avatar.'.

	x1 := 140 - (72+15).
	x2 := 140 + 15.
	y1 := 10.

	label := CLabel on: 'My Picture:'.
	label font: QStyles controlFont;
		  hResizing: #shrinkWrap;
		  fullBounds;
		  topLeft: x1@y1;
		  balloonHelp: (self pictureHelp).
	self add: label.
	photo topLeft: (label topLeft x +1)@(label bottomLeft y + 5);	
		borderStyle: #simple;
		borderColor: (Color black);
		borderWidth: 1;
		height: 72;
		width: 72;
		balloonHelp: (self pictureHelp).
	self add: photo.
	getPhoto font: QStyles controlFont;
		label: 'Browse...';
		topLeft: (label topLeft x)@(photo bottomLeft y + 7);
		roundCorners:true;
		extent: 72@24;
		balloonHelp: (self pictureHelp).
	self add: getPhoto.
	clearPhoto font: QStyles controlFont;
		label: 'Clear';
		topLeft: getPhoto bottomLeft + (0@7);
		roundCorners:true;
		extent: 72@24;
		balloonHelp: 'Remove your picture'.
	self add: clearPhoto.

	label := CLabel on: 'My Badge:'.
	label font: QStyles controlFont;
		  hResizing: #shrinkWrap;
		  fullBounds;
		  topLeft: (x2)@y1;
		  balloonHelp: (self badgeHelp).
	self add: label.
	badge topLeft: (label topLeft x+1)@(photo topLeft y);
		borderStyle: #simple;
		borderColor: (Color black);
		borderWidth: 1;
		height: 72;
		width: 72;
		balloonHelp: (self badgeHelp).
	self add: badge.
	getBadge font: QStyles controlFont;
		label: 'Browse...';
		topLeft: (label topLeft x)@(badge bottomLeft y + 7);
		roundCorners:true;
		extent: 72@24;
		balloonHelp: (self badgeHelp).
	self add: getBadge.
	clearBadge font: QStyles controlFont;
		label: 'Clear';
		topLeft: getBadge bottomLeft + (0@7);
		roundCorners:true;
		extent: 72@24;
		balloonHelp: 'Remove your badge'.
	self add: clearBadge.
! !

!QPlainAvatarDetails methodsFor: 'events' stamp: 'bgf 9/16/2007 19:31'!
updatePreview

	"Update the live avatar replica to match the current selections,
	 and update the dirty flag to say whether or not the current selections
	 are different from the saved user-prefences avatar."

	| live dirty stored |
	
	dirty := false.

	"Only do the business when there is a live avatar in the real avatar proxy."
	live := (chooser liveCostume).
	live ifNil: [ ^ self ].
	stored := (chooser avatarCostume).
	stored class = chooser liveClass ifFalse: [ 
		dirty := true.
	].

	"Push the pictures into the space here..."

	chooser detailsDirty: dirty.! !

TFrame subclass: #QForceVector
	instanceVariableNames: 'force'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Objects-Physics'!

!QForceVector commentStamp: 'das 8/18/2006 14:27' prior: 0!
Used to visualize a force vector generated inside of QMS. Attached to the frame inside of QMass.!

!QForceVector class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QForceVector methodsFor: 'accessing' stamp: 'das 8/18/2006 15:27'!
force: f

	force := f/2.0.! !

!QForceVector methodsFor: 'initialize' stamp: 'das 8/16/2006 17:24'!
initialize

	super initialize.
	force :=0@0@0.! !

!QForceVector methodsFor: 'render' stamp: 'das 8/16/2006 17:21'!
render: ogl

	ogl glDisable: GLLighting;
	" This is used to render solid objects (if any). "
		glColor3fv: #(1.0 0 0)asFloatArray;
		glBegin: GLLines;
    
			glVertex3fv:#(0.0 0.0 0.0)asFloatArray;
    
			glVertex3fv:force asFloatArray;
     
		glEnd;
	glEnable: GLLighting.! !

TFrame subclass: #QRibbon
	instanceVariableNames: 'qms wind gravity mat wire xsize ysize linkLength linkWidth tackPoints doCompute stepTime countdown'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Objects-Physics'!

!QRibbon commentStamp: '<historical>' prior: 0!
A strip of ribbon that blows in the wind.
Fixed at along the origin end,  length extends along Z,
orientation is horizontal, wind blowing to positive Z.

The ribbon's overall dimensions are (extent x * linkWidth) by (extent y)*(linkLength).
The 'extent' counts cells in the mass-spring tapestry.
Wind and gravity are supplied within the ribbon implementation.

Created by copy&paste from QTapestry.
(Tapestry has hardcoded setup in a bridge shape,
various things for fading from one material to another,
cubes at every vertex, and suchlike.  QTapestry is also out of sync
with newer bound sphere and update handling..

Should be refactored into a common base tapestry - just the spring/mass grid and nodes -
with ribbon and bridge/flag as variations.!

!QRibbon class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QRibbon methodsFor: 'compute' stamp: 'das 11/18/2007 02:04'!
compute

	doCompute ifTrue:[
		self computeOneStep.
		wind on ifFalse:[
			countdown<0 ifTrue:[doCompute:= false]. 
			countdown := countdown - stepTime.
		].
		(self future: stepTime) compute.
	].
! !

!QRibbon methodsFor: 'compute' stamp: 'bgf 2/22/2007 12:12'!
computeNorms

	| index mn masses |
	masses := qms masses.
	1 to: xsize - 1 do:[:i |
		1 to: ysize - 1 do:[ :j |
			index := i-1*ysize+j.
			mn := masses at: index.
			mn normal: 
				(((masses at: index +1 ) location - mn location) cross: 
				((masses at: index + ysize)location - mn location))safelyNormalized.
			].
			index := i*ysize.
			mn := masses at: index.
			mn normal: 
				((mn location - (masses at: index - 1 ) location) cross: 
				((masses at: index + ysize)location - mn location))normalized.
		].
	masses size - xsize to: masses size do:[ :in | 
		(masses at: in) normal: (masses at: in - ysize)normal].
! !

!QRibbon methodsFor: 'compute' stamp: 'das 11/18/2007 01:53'!
computeOneStep

	self computeNorms.
	qms compute.
! !

!QRibbon methodsFor: 'render' stamp: 'bgf 2/22/2007 15:18'!
doRender: ogl wire: doWire

	| index m masses|

	ogl glDisable: GLCullFace.
	masses := qms masses.
	1 to: xsize-1 do:[:i |
		doWire 
			ifFalse:[ ogl glBegin: GLTriangleStrip.] 
			ifTrue:[ ogl glBegin: GLLineStrip.].
		1 to: ysize do:[:j |
			index := (i-1*ysize)+j.
			m := masses at: index.
			ogl glNormal3fv: m normal.
			ogl glTexCoord2f: m uv x with: m uv y.
			ogl glVertex3fv: m location.
			index := index +ysize.
			m := masses at: index.
			ogl glNormal3fv: m normal.
			ogl glTexCoord2f: m uv x with: m uv y.
			ogl glVertex3fv: m location.
		].
		ogl glEnd.
	].
	doWire ifTrue:[
		1 to: xsize do:[:j |
			ogl glBegin: GLLineStrip.
			1 to: ysize do:[:i |
				ogl glVertex3fv: (masses at:(j-1*ysize)+i) location.].
			ogl glEnd.
		].
	].

	ogl setCull.
! !

!QRibbon methodsFor: 'accessing' stamp: 'bgf 2/22/2007 12:12'!
extent

	^ xsize@ysize.
! !

!QRibbon methodsFor: 'accessing' stamp: 'bgf 2/22/2007 12:49'!
extent: ext

	xsize := ext x.
	ysize := ext y.
	self shapeChanged.! !

!QRibbon methodsFor: 'accessing' stamp: 'bgf 6/12/2009 21:48'!
gravity 
	"The QForce for gravity.  By default, almost straight down."
	^ gravity! !

!QRibbon methodsFor: 'accessing' stamp: 'bgf 2/22/2007 14:56'!
gravityOn
	^ gravity on
! !

!QRibbon methodsFor: 'accessing' stamp: 'bgf 2/22/2007 12:29'!
gravityOn: bool
	gravity on: bool
! !

!QRibbon methodsFor: 'accessing' stamp: 'bgf 2/22/2007 12:12'!
hasAlpha

	^ mat hasAlpha.! !

!QRibbon methodsFor: 'initialize' stamp: 'bgf 6/12/2009 20:56'!
initialize

	super initialize.

	stepTime := 50.
	xsize := 4.
	ysize := 20.
	linkLength := 1.0.
	linkWidth := 0.5.
	countdown := 0.
	mat := TMaterial new.
	mat	 diffuseColor: (VectorColor r:0.85 g:0.85 b:0.1 a:0.9);
		 ambientColor: (VectorColor r:0.6 g:0.6 b:0.1 a:0.9);
		 specularColor: (VectorColor r:0.1 g:0.1 b:0.1).

	wind := QWind new.
	gravity := QForce new.
	gravity force: -0.02 @ -1.5 @ 0.022.
	qms := QMS new.

	self windOn: true.
	self gravityOn: true.

	wire := false.
	doCompute := false.

	self shapeChanged.
	self future computeOneStep.! !

!QRibbon methodsFor: 'accessing' stamp: 'bgf 2/22/2007 12:55'!
linkLength
	^ linkLength! !

!QRibbon methodsFor: 'accessing' stamp: 'bgf 2/22/2007 14:54'!
linkLength: l
	"Length along the ribbon of a single mass-spring section."
	(linkLength ~= l) ifTrue: [
		linkLength := l.
		self shapeChanged.
	].! !

!QRibbon methodsFor: 'accessing' stamp: 'bgf 2/22/2007 13:30'!
linkWidth
	^ linkWidth! !

!QRibbon methodsFor: 'accessing' stamp: 'bgf 2/22/2007 14:54'!
linkWidth: w
	"Width across the ribbon of a single mass-spring section."
	(linkWidth ~= w) ifTrue: [
		linkWidth := w.
		self shapeChanged.
	].! !

!QRibbon methodsFor: 'accessing' stamp: 'bgf 2/22/2007 12:55'!
material
	^ mat! !

!QRibbon methodsFor: 'accessing' stamp: 'bgf 2/22/2007 12:12'!
material: m

	mat := m.! !

!QRibbon methodsFor: 'events' stamp: 'bgf 2/22/2007 12:12'!
pick: ptr

	^ true.! !

!QRibbon methodsFor: 'render' stamp: 'bgf 2/22/2007 15:17'!
render: ogl

	mat hasOpaque ifTrue:[
		mat enable: ogl.
		self doRender: ogl wire: wire.
		mat disable: ogl.].

! !

!QRibbon methodsFor: 'render' stamp: 'bgf 2/22/2007 15:17'!
renderAlpha: ogl

	mat hasAlpha ifTrue:[
		mat enable: ogl.
		self doRender: ogl wire: wire.
		mat disable: ogl.].
! !

!QRibbon methodsFor: 'events' stamp: 'das 6/20/2007 13:05'!
restart

	qms restart.! !

!QRibbon methodsFor: 'events' stamp: 'bgf 10/26/2008 17:11'!
start

	self doUpdateShape.
	doCompute ifFalse: [
		doCompute := true.
		self compute.
	].! !

!QRibbon methodsFor: 'accessing' stamp: 'bgf 2/22/2007 14:54'!
stepTime
	"Duration between steps; default 50.  Varying this keeps adjacent ribbons
	 from appearing too synchronized."
	^ stepTime! !

!QRibbon methodsFor: 'accessing' stamp: 'bgf 2/22/2007 14:55'!
stepTime:anInteger
	"Duration between steps; default 50.  Varying this keeps adjacent ribbons
	 from appearing too synchronized."
	stepTime := anInteger! !

!QRibbon methodsFor: 'events' stamp: 'bgf 2/22/2007 12:12'!
stop

	doCompute := false.
! !

!QRibbon methodsFor: 'initialize' stamp: 'bgf 2/22/2007 12:12'!
terminate

"	self root ~= self ifTrue:[
		self root stopAmbientSound.]."
	self removeSelf.! !

!QRibbon methodsFor: 'accessing' stamp: 'bgf 2/22/2007 12:24'!
texture

	^ mat texture.! !

!QRibbon methodsFor: 'accessing' stamp: 'bgf 2/22/2007 12:23'!
texture: txtr

	mat texture: txtr.
! !

!QRibbon methodsFor: 'accessing' stamp: 'das 11/24/2007 10:12'!
textureSHA

	^ mat ifNotNil:[mat textureSHA].
! !

!QRibbon methodsFor: 'initialize' stamp: 'bgf 3/17/2008 02:19'!
updateShape

	"Remake the tapestry to current dimensions."

	| m s masses springs node |

	self removeAll.
	masses := OrderedCollection new.
	springs := OrderedCollection new.
	tackPoints := OrderedCollection new.
	
	1 to: xsize do:[:i |
		1 to: ysize do:[:j |
			m := QMass new.
			m force: Vector3 zero.
			m location: ( Vector3 x: (i-1*linkWidth) y:0 z:(j-1*linkLength) ).
			m uv: (m location x/(xsize-1*linkWidth)) @  (1-(m location z/(ysize-1*linkLength))).
			m mass: 1.0 + (xsize *0.03).
			m normal: (Vector3 zAxis clone).

			"The nodes are invisible frames of reference, used for transforming."
			node := TGroup new.
			self addChild: node.
			m frame: node.
			masses add: m.
		].
	].

	1 to: xsize do:[:i |
		1 to: ysize do:[:j |
			j~=ysize ifTrue:[
				s := QSpring new.
				s mass: (masses at:i-1*ysize+j) mass: (masses at:i-1*ysize+j+1).
				springs add: s.
				].
			i~=xsize ifTrue:[
				s := QSpring new.
				s mass: (masses at:i-1*ysize+j) mass: (masses at:i*ysize+j).
				springs add: s.
				].
			(i ~= xsize and:[j~=ysize])ifTrue:[
				s := QSpring new.
				s mass: (masses at:i-1*ysize+j+1) mass: (masses at:i*ysize+j).
				springs add: s.	
				].
		]
	].
	
	"The fixed tack points - across the y=1 end of the ribbon."
	1 to: (ysize*(xsize-1)+1) by: ysize do:[ :i|
		tackPoints add: i.
	].
	tackPoints do:[:tp | (masses at: tp) mass: 0.0.].

	qms masses: masses springs: springs.
	qms addForce: wind.
	qms addForce: gravity.
	qms stop.	"We drive the QMS loop from here; don't let it step independently."
! !

!QRibbon methodsFor: 'accessing' stamp: 'bgf 6/12/2009 21:25'!
wind
	"My wind force.  By default, an oscillating wind mainly in Z."
	^ wind! !

!QRibbon methodsFor: 'accessing' stamp: 'bgf 6/12/2009 21:26'!
wind: aQForce
	"My wind force.  By default, an oscillating wind mainly in Z."
	wind := aQForce! !

!QRibbon methodsFor: 'accessing' stamp: 'bgf 2/22/2007 14:55'!
windOn
	"Is the ribbon getting blown around?"
	^ wind on! !

!QRibbon methodsFor: 'accessing' stamp: 'das 11/18/2007 02:03'!
windOn: bool
	wind on: bool.
	countdown := self stepTime*500.0.! !

!QRibbon methodsFor: 'accessing' stamp: 'bgf 2/22/2007 15:16'!
wire
	"Draw in wireframe?"
	^ wire! !

!QRibbon methodsFor: 'accessing' stamp: 'bgf 2/22/2007 12:12'!
wire: w

	wire := w.! !

TFrame subclass: #QTapestry
	instanceVariableNames: 'qms wind gravity mat matTxtr deltaTrans boundSphere time wire xsize ysize linkLength tackPoints topLeft topRight bottomLeft bottomRight doCompute'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Objects-Physics'!

!QTapestry class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QTapestry class methodsFor: 'defaults' stamp: 'bgf 11/10/2009 16:58'!
defaultEventMask
	^ EventKeyboard + EventPointerDown! !

!QTapestry methodsFor: 'accessing' stamp: 'das 7/25/2006 13:15'!
bottomLeft

	^ bottomLeft.! !

!QTapestry methodsFor: 'accessing' stamp: 'das 7/25/2006 13:15'!
bottomRight

	^ bottomRight.! !

!QTapestry methodsFor: 'compute' stamp: 'das 7/25/2006 13:18'!
boundSphere

	^ boundSphere.! !

!QTapestry methodsFor: 'accessing' stamp: 'das 7/25/2006 13:15'!
boundingBox
	
	^ boundSphere box.! !

!QTapestry methodsFor: 'accessing' stamp: 'das 7/25/2006 13:15'!
colorize: clr

	mat ambientColor: clr.
	mat diffuseColor: clr
! !

!QTapestry methodsFor: 'compute' stamp: 'das 10/12/2006 13:27'!
compute

	doCompute ifTrue:[
		self computeNorms.
		qms compute.
		self updateBounds.
	deltaTrans ~=0.0 ifTrue:[
		matTxtr transparency: matTxtr transparency + deltaTrans.
		matTxtr transparency > 1.0 ifTrue:[matTxtr transparency: 1.0. deltaTrans := 0.0].
		matTxtr transparency < 0.0 ifTrue:[matTxtr transparency: 0.0. deltaTrans := 0.0].
		mat transparency: 1.0-matTxtr transparency.
		].
matTxtr transparency ~=0.0 ifTrue:[bottomLeft restore].

		(self future: 50) compute.
	].
! !

!QTapestry methodsFor: 'compute' stamp: 'das 8/9/2006 21:38'!
computeNorms

	| index mn masses |
	masses := qms masses.
	1 to: xsize - 1 do:[:i |
		1 to: ysize - 1 do:[ :j |
			index := i-1*ysize+j.
			mn := masses at: index.
			mn normal: 
				(((masses at: index +1 ) location - mn location) cross: 
				((masses at: index + ysize)location - mn location))safelyNormalized.
			].
			index := i*ysize.
			mn := masses at: index.
			mn normal: 
				((mn location - (masses at: index - 1 ) location) cross: 
				((masses at: index + ysize)location - mn location))normalized.
		].
	masses size - xsize to: masses size do:[ :in | 
		(masses at: in) normal: (masses at: in - ysize)normal].
! !

!QTapestry methodsFor: 'render' stamp: 'das 8/9/2006 21:37'!
doRender: ogl wire: doWire

	| index m masses|

	ogl glDisable: GLCullFace.
	masses := qms masses.
	1 to: xsize-1 do:[:i |
	doWire ifFalse:[ ogl glBegin: GLTriangleStrip.] 
		ifTrue:[ ogl glBegin: GLLineStrip.].
		1 to: ysize do:[:j |
			index := (i-1*ysize)+j.
			m := masses at: index.
			ogl glNormal3fv: m normal.
			ogl glTexCoord2f: m uv x with: m uv y.
			ogl glVertex3fv: m location.
			index := index +ysize.
			m := masses at: index.
			ogl glNormal3fv: m normal.
			ogl glTexCoord2f: m uv x with: m uv y.
			ogl glVertex3fv: m location.

		].
		ogl glEnd.
	].
	doWire ifTrue:[
		1 to: xsize do:[:j |
			ogl glBegin: GLLineStrip.
			1 to: ysize do:[:i |
				ogl glVertex3fv: (masses at:(j-1*ysize)+i) location.].
			ogl glEnd.
		].
	].
"
This is to render the springs to see if these make any sense...they do now.
	springs do:[:s |
		ogl glBegin: GLLineStrip.
		ogl glVertex3fv: s mass1 location+(0@2@0.3).
		ogl glVertex3fv: s mass2 location+(0@2@0.3).
		ogl glEnd.
		].
"
"
	masses do:[:ms |
		ogl glBegin: GLLineStrip.
		ogl glVertex3fv: ms location.
		ogl glVertex3fv: ms location - ms normal.
		ogl glEnd.
		].

"

	ogl setCull.
! !

!QTapestry methodsFor: 'accessing' stamp: 'das 7/25/2006 13:15'!
extent

	^ xsize@ysize.
! !

!QTapestry methodsFor: 'accessing' stamp: 'das 7/25/2006 13:16'!
extent: ext

	xsize := ext x.
	ysize := ext y.
	self reset.! !

!QTapestry methodsFor: 'accessing' stamp: 'das 7/25/2006 13:16'!
hasAlpha

	^ mat hasAlpha.! !

!QTapestry methodsFor: 'initialize' stamp: 'das 8/9/2006 21:14'!
initialize

	super initialize.
	time := 0.0.
	xsize := 17.
	ysize := 9.
	linkLength := 1.0.
	mat := TMaterial new.
	mat ambientColor:#(1 1 1 1)asFloatArray.
	mat diffuseColor: #(1 1 1 1)asFloatArray.
	matTxtr := TMaterial new.
	matTxtr ambientColor:#(1 1 1 1)asFloatArray.
	matTxtr diffuseColor: #(1 1 1 1)asFloatArray.
	self reset.
	wire := true.
	doCompute := false.
! !

!QTapestry methodsFor: 'events' stamp: 'das 8/9/2006 21:16'!
keyStroke: event

	| c |
	c := event keyCharacter.

	c = $t ifTrue:[
		self wire: wire not.
		].

	c = $r ifTrue:[
		self reset.
		].

	c = $m ifTrue:[
	"	masses anyOne class openScriptor."
	].

	c = $s ifTrue:[
	"	springs anyOne class openScriptor."
	].

	c = $w ifTrue:[
		wind on: wind on not.
	].

	c = $g ifTrue:[
		gravity on: gravity on not.
	].

	c =$f ifTrue:[
		self release.
		].
	c =$D ifTrue:[
		self terminate.
		].

	c = $+ ifTrue:[
		gravity force: gravity force - (0@0.5@0).
	].

	c = $= ifTrue:[
		gravity force: gravity force - (0@0.5@0).
	].
	c = $- ifTrue:[
		gravity force: gravity force + (0@0.5@0).
	].! !

!QTapestry methodsFor: 'initialize' stamp: 'das 8/9/2006 21:38'!
log

qms masses do:[ :m | Transcript show: m location;cr].
! !

!QTapestry methodsFor: 'accessing' stamp: 'das 7/25/2006 13:16'!
material: m

	mat := m.! !

!QTapestry methodsFor: 'events' stamp: 'das 7/25/2006 13:20'!
pick: ptr

	^ true.! !

!QTapestry methodsFor: 'events' stamp: 'das 7/25/2006 13:21'!
pointerDown: event

	self reset.! !

!QTapestry methodsFor: 'initialize' stamp: 'das 8/9/2006 21:37'!
release

"release the tack points"
	tackPoints do:[:tp | (qms masses at: tp) mass: 1.0.].
	topLeft mass:0.0.
	bottomLeft mass:0.0.
	deltaTrans := 0.003.
"	self root ambientSoundPlaying ifFalse:[
		self root ambientSound: (self root loadAmbientSound:'totoro.mp3').
		self root startAmbientSound.
	]."! !

!QTapestry methodsFor: 'render' stamp: 'das 7/26/2006 13:24'!
render: ogl

mat ambientColor:#(0.3 0.3 0.5 1.0)asFloatArray.
	mat transparency = 1.0 ifTrue:[
		mat enable: ogl.
		self doRender: ogl wire: true.
		mat disable: ogl.].

	matTxtr transparency = 1.0 ifTrue:[
		matTxtr enable: ogl.
		self doRender: ogl wire: false.
		matTxtr disable: ogl.
		].

! !

!QTapestry methodsFor: 'render' stamp: 'das 7/26/2006 13:25'!
renderAlpha: ogl

	mat transparency ~= 1.0 & (mat transparency ~= 0.0) ifTrue:[
		mat enable: ogl.
		self doRender: ogl wire: true.
		mat disable: ogl.].

	matTxtr transparency ~= 1.0 ifTrue:[
		matTxtr enable: ogl.
		self doRender: ogl wire: false.
		matTxtr disable: ogl.
		].

! !

!QTapestry methodsFor: 'initialize' stamp: 'das 10/12/2006 13:28'!
reset

	| m s cube masses springs|
"	self root ~= self ifTrue:[
		self root stopAmbientSound.]."

	self removeAll.
	masses := OrderedCollection new.
	springs := OrderedCollection new.
	tackPoints := OrderedCollection new.
	wind := QWind new initialize.
	gravity := QForce new initialize.
	gravity force: -0.01 @ -2.0 @ 0.01.
	
	1 to: xsize do:[:i |
		1 to: ysize do:[:j |
			m := QMass new.
			m force: 0@0@0.
			m location: linkLength*((i-1)@0@(j-1)).
			m uv: ((m location x/(xsize-1))@(1-(m location z/(ysize-1))))/linkLength.
			j = 1 ifTrue:[m location: m location + (0@(linkLength negated)@linkLength)].
			j = ysize ifTrue:[m location: m location + (0@(linkLength negated)@linkLength negated)].
			m mass: 1.0.
			m normal: (0@0@1).
			cube := TCube new.
			cube extent: 0.1@0.1@0.1.
			self addChild: cube.
			m frame: cube.
			masses add: m.
		].
	].

	1 to: xsize do:[:i |
		1 to: ysize do:[:j |
			j~=ysize ifTrue:[
				s := QSpring new.
				s mass: (masses at:i-1*ysize+j) mass: (masses at:i-1*ysize+j+1).
				springs add: s.
				].
			i~=xsize ifTrue:[
				s := QSpring new.
				s mass: (masses at:i-1*ysize+j) mass: (masses at:i*ysize+j).
				springs add: s.
				].
			(i ~= xsize and:[j~=ysize])ifTrue:[
				s := QSpring new.
				s mass: (masses at:i-1*ysize+j+1) mass: (masses at:i*ysize+j).
				springs add: s.	
				].
		"	s := QSpring new.
			s mass: (masses at:i-1*xsize+j+1) mass: (masses at:i*xsize+j).
			springs add: s."
		]
	].

	topLeft := masses at:1.
	topRight := masses at: masses size -ysize+1.
	bottomLeft := masses at:ysize.
	bottomRight := masses at: masses size.
	bottomLeft iLocation:linkLength*(0@(ysize)@0).
"set the tack points"
	tackPoints add:1.
	tackPoints add: 2.
	tackPoints add: ysize.
	tackPoints add: ysize-1.
	tackPoints add: masses size-ysize+1.
	tackPoints add: masses size-ysize+2.
	tackPoints add: masses size - 1.
	tackPoints add: masses size.
	tackPoints do:[:tp | (masses at: tp) mass: 0.0.].

	qms := QMS new masses: masses springs: springs.
	qms addForce: wind.
	qms addForce: gravity.

	boundSphere := TBoundSphere new.
	boundSphere frame: self.
	self updateBounds.
 	matTxtr transparency: 0.0.
	mat transparency: 1.0.
	deltaTrans := 0.0.

! !

!QTapestry methodsFor: 'events' stamp: 'bgf 10/26/2008 17:11'!
start

	self doUpdateShape.
	doCompute := true.
	self compute.! !

!QTapestry methodsFor: 'events' stamp: 'das 7/25/2006 13:31'!
stop

	doCompute := false.
! !

!QTapestry methodsFor: 'initialize' stamp: 'das 8/9/2006 21:38'!
terminate

"	self root ~= self ifTrue:[
		self root stopAmbientSound.]."
	self removeSelf.! !

!QTapestry methodsFor: 'accessing' stamp: 'das 7/25/2006 13:16'!
texture

	^ matTxtr texture.! !

!QTapestry methodsFor: 'accessing' stamp: 'das 7/25/2006 13:16'!
texture: txtr

	matTxtr texture: txtr.
! !

!QTapestry methodsFor: 'accessing' stamp: 'das 7/25/2006 13:16'!
topLeft

	^ topLeft.! !

!QTapestry methodsFor: 'accessing' stamp: 'das 7/25/2006 13:16'!
topRight

	^ topRight.! !

!QTapestry methodsFor: 'accessing' stamp: 'das 7/25/2006 13:16'!
wire: w

	wire := w.! !

!TFrame class methodsFor: '*Qwaq-Objects' stamp: 'bgf 2/9/2010 17:11'!
addGroupMenuItems: tmenu mask: maskSet frames: frames event: cEvent
	"Protocol - add applicable context menu items when I am the common base class of the list of frames."

	"The things that can attach to note-collectors include annotations, displays, polls, etc...	
	 so TFrame is the common base class for the 'group the notes' operation."
	(maskSet includes: #groupNotes) ifTrue: [
		| sa |
		sa := frames anyOne surfaceAttachment.
		((sa isKindOf: QFixedSurface) 
		 and: [ sa isFlatSurface ]
		 and:[ frames allSatisfy: [ : f | (f isListedTopic) and: [ f surfaceAttachment = sa ] ] ] )
	        ifTrue: [
			tmenu add: 'Group' 
				    target: sa
				    action: #generateNoteGroup:event:
				    arguments: { frames clone . cEvent  }
		].
	] ifFalse: [
		"General case - if the set of children share a parent, offer ordinary Group"
		(maskSet includes: #group) ifTrue: [
			tmenu add: 'Group'
				target: (cEvent avatar)
				action: #makeGroup:event:
				arguments: { frames clone . cEvent }
		].
	].! !

!TFrame class methodsFor: '*Qwaq-Objects' stamp: 'bgf 3/19/2010 13:53'!
reparent: frame oldParent: p1 newParent: p2 oldIndex: i1 newIndex: i2
	"Undo support for the 'acceptDroppedFrame' reparenting action."
	^ QBlockAction
		undoBlock: [ frame future transferTo: p1 atIndex: i1 ]
		redoBlock: [ frame future transferTo: p2 atIndex: i2 ]
		description: 'Reparent'
		requiredFrames: { p1 . p2 . frame }		
! !

!TFrame methodsFor: '*Qwaq-Objects-Space' stamp: 'bgf 8/17/2009 23:55'!
addNote: aNote
	"Recursively inform the containing note collector under a stack."
	self surfaceAttachment ifNotNilDo: [ : sfc |
		sfc isNoteCollector ifTrue:[
			sfc addNote: aNote
		]].! !

!TFrame methodsFor: '*Qwaq-Objects-Space' stamp: 'bgf 2/6/2010 19:05'!
allDocuments
	"All the QAbstractDocuments held in this frame tree."
	^ (Array streamContents: [ : s |
		self allFramesDo: [ : f | s nextPutAll: (f documents)  ]
		])! !

!TFrame methodsFor: '*Qwaq-Objects-Space' stamp: 'bgf 3/18/2010 09:00'!
applyGraspState: stateVec
	"Protocol for being manipulated via a grasp adapter, with undo.
	  The grasp-adapter state vector is a local transform (for me), 
	  and a surface frame I should be attached to."

	self localTransform: (stateVec first).
	self root ifNotNil: [ : space | space attach: self to: (stateVec second) ].! !

!TFrame methodsFor: '*Qwaq-Objects-Space' stamp: 'bgf 3/18/2010 11:15'!
applyResizerState: stateVec
	"Protocol for resizing via QAbstractResizer, with undo.
	 The resizer state vector is a local transform, a surface to be attached to, and a 2D extent."

	self localTransform: (stateVec first).
	self root ifNotNil: [ : space | space attach: self to: (stateVec second) ].
	self extent: (stateVec third).	! !

!TFrame methodsFor: '*Qwaq-Objects-Space' stamp: 'bgf 10/25/2009 20:35'!
asPlace
	"Backstop for identifying go-to destination Places in response to arbitrary clicks..
	 find the enclosing frame that defines a place, if any."
	^ frameParent ifNotNil: [ frameParent asPlace ].! !

!TFrame methodsFor: '*Qwaq-Objects-Space' stamp: 'bgf 8/18/2009 00:18'!
attachedTo: surfaceObject
	"Attachment manager notification for subclass extension.
	 This object has been attached to the surface, and will follow its transform changes."

	self isListedTopic ifTrue: [	
		"Maintain note-collection stacks."
		| sfc |
		sfc := self surfaceAttachment.
		sfc = surfaceObject ifFalse: [
			sfc ifNotNil: [
				"Detach from the old"
				sfc isNoteCollector ifTrue: [ sfc removeNote: self ]].
			surfaceObject ifNotNil:[
				"Attach to the new."
				surfaceObject isNoteCollector ifTrue: [ surfaceObject addNote: self ]]
		]].
! !

!TFrame methodsFor: '*Qwaq-Objects-Space' stamp: 'bgf 8/4/2009 17:26'!
attachmentReceiver
	"When something is dragged over me as event selection or target frame,
	 what object (me or an ancestor) should be recorded as the attachment ?
	 We need a persistent component since we want the object to come back
	 on reload of the forum."
	| receiver |
	receiver := self component.
	[ receiver ] whileNotNil: [ 
		receiver shouldPersist ifFalse: [ 
			receiver := receiver parent.
			receiver ifNotNil: [ receiver := receiver component ].
		] ifTrue: [
			^ receiver
		]].

	^ nil
! !

!TFrame methodsFor: '*Qwaq-Objects-Space' stamp: 'bgf 7/26/2009 12:42'!
bestSeatWithin: tbox nearest: avatarPosition
	"Backstop."
	^ nil
! !

!TFrame methodsFor: '*Qwaq-Objects-Space' stamp: 'bgf 4/1/2010 09:11'!
canBeCopiedBy: anAvatar
	"Protocol - can the given user copy me into their inventory."
	^ self isInAvatar not! !

!TFrame methodsFor: '*Qwaq-Objects-Space' stamp: 'bgf 2/14/2010 15:03'!
carryWhenPasted 
	"Protocol - should this object be taken as a carry when pasted into the scene?
	 Do this only for things that have natural drop mechanics (filecards)"
	^ false! !

!TFrame methodsFor: '*Qwaq-Objects-Space' stamp: 'bgf 4/1/2010 08:43'!
contextMenuMaskForAvatar: avatar
	"Returns a set of standard symbols
	 to determine what builtin context menu commands are to be offered."
	| mask isPlace editProtect |

	(avatar notNil 
	 and: [ self doContextMenu or: [ avatar isEditingTemplate  ]]
	) ifFalse: [
		^ (self asPlace notNil and: [ avatar allowedToMove] )
			ifTrue: [ Set with: #approach ]
			ifFalse: [ Set new ].	"No system-supplied context menu at all."
	].

	"For documentation, these are all the standard symbols.  
	 We remove those not applicable."
	mask := #(group approach meetHere attachNote resize material properties copyToInventory delete) 
			asSet.

	isPlace := self asPlace notNil.
	editProtect := (avatar canModifyForum not) 
				  or: [ self fromTemplate and: [ avatar isEditingTemplate not ]].

	"The rules start here. "

	(self isListedTopic and: [self parent = self root]) ifTrue: [ mask add: #groupNotes ].

	(avatar allowedToMove and: [ isPlace ])
		ifFalse: [ mask remove: #approach ].

	"For clarity, don't offer attach note... to notes."
	(avatar isMeetingLeader and: [ isPlace ] and: [ self asPlace posture ~= #Sitting]) 
		ifFalse: [ mask remove: #meetHere ].

	self canTakeAnnotations ifFalse: [ mask remove: #attachNote ].
	editProtect ifTrue: [
		mask remove: #resize.
		mask remove: #material.
		mask remove: #properties.
		mask remove: #group.
		mask remove: #groupNotes ifAbsent: [].
	] ifFalse: [
		self canEditInBox ifFalse: [ mask remove: #resize  ].
		self canEditMaterial ifFalse: [ mask remove: #material ].
	].

	((self canBeCopiedBy: avatar) and: [ avatar canUseInventory ]) 
		ifFalse: [ mask remove: #copyToInventory ].

	(avatar canDeleteFrame: self) ifFalse: [ 
		mask remove: #delete.
		mask remove: #group ifAbsent: [].
	].

	^ mask
	! !

!TFrame methodsFor: '*Qwaq-Objects-Space' stamp: 'bgf 7/23/2010 13:51'!
defaultCameraStanceFor: anAvatar
	"For avatar approach.  
       A transform for an avatar to view my entirety,
 	 looking out at me from nearer the center of the space."
	| pos rad look |

	pos := self centeredGlobalTransform.
	rad := self boxRadius.

	"Pick a viewpoint inward of the center, looking out at it
 	 along its axis from the origin, not pitched or rolled."

	look := pos translation safelyNormalized negated.
	
	look y abs > 0.9 ifFalse: [
		look := (look x @ 0.05 @ look z).
	] ifTrue: [
		look := -0.25 @ 0.3 @ 0.25.
	].

	((look + pos translation) length > pos translation length) ifTrue: [
		look := look negated.
		look y < 0 ifTrue: [ look y: (look y negated). ].
	].

	^ Stance here: (pos translation + (look * (rad*2.5 min: 36 max: 6)))
			   there: (pos translation).

! !

!TFrame methodsFor: '*Qwaq-Objects-Space' stamp: 'bgf 7/23/2010 10:11'!
defaultDestinationStanceFor: avatar
	"For avatar approach.  
       A transform for an avatar to view my entirety,
 	 looking out at me from nearer the center of the space."

	^ QGotoButton groundAdjust: (self defaultCameraStanceFor: avatar) 
					forAvatar: avatar
	! !

!TFrame methodsFor: '*Qwaq-Objects-Space' stamp: 'jjl 8/5/2008 11:07'!
destinationStanceFor: anAvatar

	"For avatar #takeMeThere.  If this frame is contained within a frame that defines
	 a destination transform, return that transform. Otherwise return nil.
	 Components with natural front-viewing positions like display panels
	 should override this - then the takeMeThere right-click on anything 
	 within that frame will take the user  to the destination position."

	frameParent ifNil:[ ^nil ].

	^ frameParent destinationStanceFor: anAvatar
! !

!TFrame methodsFor: '*Qwaq-Objects-Space' stamp: 'bgf 12/25/2008 12:06'!
distanceIsPastWarpThreshold: distanceScalar forViewExtent: vextent

	"Return true if the given distance from me is past
	 the point at which warp-to  should be offered.
	 (I.e. beyond a certain distance, the client may offer a go-to 
 	 when you click on something far away, rather than interacting with it.)"

	^ ((self boundingBox extentMax max: 2.0) / (distanceScalar max: 1.0)) * (self warpThresholdAdjustment)
			< (100.0 / (vextent minDimension max: 64))! !

!TFrame methodsFor: '*Qwaq-Objects-Space' stamp: 'bgf 2/7/2010 00:25'!
documents
	"Protocol: all QAbstractDocuments held by this frame.  (Usually none, sometimes one, rarely more)"
	^ #()! !

!TFrame methodsFor: '*Qwaq-Objects-Space' stamp: 'bgf 8/16/2009 00:45'!
dragAdapter
	"The QDragAdapter holding this frame as its luggage,
	  when appropriate to use for initial positioning."
	^ nil! !

!TFrame methodsFor: '*Qwaq-Objects-Space' stamp: 'bgf 3/18/2010 11:20'!
graspState
	"QGraspAdapter support  -
 	 This is the captured state record that can be used to undo or redo one grasp-drag gesture.
	 See 'applyGraspState:', which consumes these.
	 Subclasses should extend the array, but not alter the first two elements,
	 which are assumed by the grasp-adapter."
	^ { self localTransform clone . self surfaceAttachment  }! !

!TFrame methodsFor: '*Qwaq-Objects-Space' stamp: 'jcg 10/12/2006 15:47'!
isGroundBound
	"Keep default Croquet behavior wrt ground-bound-ness."

	^ false! !

!TFrame methodsFor: '*Qwaq-Objects-Space' stamp: 'bgf 3/15/2008 17:57'!
isListedLandmark
	"Backstop for QSpace/QPlace trolling, when creating the places menu.
 	 If this is overridden to return true, 
	 the returning object must also implement
		asPlace
	 returning a QPlace describing the destination.
	 Any such place-frame will be listed in the Places menu of the browser
	 of any user in this space."

	^ false! !

!TFrame methodsFor: '*Qwaq-Objects-Space' stamp: 'bgf 8/1/2009 11:50'!
isListedTopic
	"Backstop for annotation trolling, when populating note-lists.
 	 If this is overridden to return true, 
	 the returning object must also implement asTopicRecord.
	 See QAnnotation/QStickyNote"

	^ false! !

!TFrame methodsFor: '*Qwaq-Objects-Space' stamp: 'bgf 8/6/2009 22:06'!
isSeat
	"For fast detection of QSeats for seat-finding protocols,
	 which troll the space on a regular basis."
	^ false! !

!TFrame methodsFor: '*Qwaq-Objects-Space' stamp: 'bgf 7/23/2010 10:19'!
locateMePlace
	^ (QPlace 
			category: 'Miscellaneous'
			label: 'Viewing ', (self labelForBrowser)
			frame: self
			selector: #defaultDestinationStanceFor:
			pov: 'First Person')
		cameraSelector: #defaultCameraStanceFor: ;
		yourself
			

	! !

!TFrame methodsFor: '*Qwaq-Objects-Space' stamp: 'bgf 4/8/2008 23:17'!
menuStyle
	"Customize all the TFrame menus for our blue/black-on-white styling;
	 do this en-masse rather than frame by frame."
	^ QStyles menuStyle! !

!TFrame methodsFor: '*Qwaq-Objects-Space' stamp: 'bgf 2/28/2010 09:23'!
movedOverSelection: aSelectionOrNil forEvent: anEvent
	"Backstop for an object-moved followup notification from something
	 that has been repositioned via a motion adapter.
	 The default implementation (if I am a persistent component)
	  is to create an attachment between me
 	  and the innermost persistent component around the target frame of the selection."
	| surface |

	self root ifNil: [ ^ self ].
	(self isComponent and: [ self shouldPersist ]) ifFalse: [ ^ self ].

	surface := aSelectionOrNil ifNotNil: [ aSelectionOrNil frame ifNil: [ aSelectionOrNil target ]].
	surface := surface ifNotNil: [ surface attachmentReceiver ].
	self root attach: self  to: surface.
! !

!TFrame methodsFor: '*Qwaq-Objects-Jobs' stamp: 'jjl 8/22/2008 12:59'!
photoBox

	^self boundingBox! !

!TFrame methodsFor: '*Qwaq-Objects-Jobs' stamp: 'bgf 8/1/2009 17:56'!
photographerClass
	"The job-class to use to take a photograph.
	 Must support the QPhotoJob class request API: 
		photograph: subjectFrame forUser: reqFrame recipient: anObject selector: aSelector.

	The default photo class takes the object to a vacant space and takes an oblique photo;
	 this is best for models and 3-D content.  "

	^ QStudioPhotoJob! !

!TFrame methodsFor: '*Qwaq-Objects-Space' stamp: 'bgf 9/13/2008 23:28'!
placeMenuActions: anAvatar

	"Protocol for adding actions to the Places context menu.
	 The result must be an array of arrays-of-2-elements.
	 The inner arrays are:
		 { 'label-string', #unary-action-selector: }
	 The selector gets an avatar-replica as argument.

	 Should be used very sparingly - no more than a couple of actions.

	 Usual warning: this is used in  GET: operations.
	 Tts implementations MUST BE SIDE-EFFECT-FREE"

	^ #()! !

!TFrame methodsFor: '*Qwaq-Objects-Space' stamp: 'bgf 12/17/2008 15:48'!
placeName
	"If I offer a place, what is its name?"
	^ self asPlace
			ifNotNilDo: [ : place | place label ]
			ifNil: [ self objectName ]! !

!TFrame methodsFor: '*Qwaq-Objects-User gestures' stamp: 'RJT 5/13/2010 14:00'!
playAnimation: name type: aType playType: aPlayType
	"atype = #skeletal or #vertex.  aPlayType = #cycle or #action
	When an avatar does not support animations do nothing.  "
	^self "NO OP"
! !

!TFrame methodsFor: '*Qwaq-Objects-Jobs' stamp: 'bgf 1/10/2010 12:18'!
portraitExtent
	"My front-on extent for photos taken with the QOnsitePhotoJob,
	 which does not take the frame out to an empty studio space."
	^ self destinationExtent! !

!TFrame methodsFor: '*Qwaq-Objects-Jobs' stamp: 'bgf 1/10/2010 12:19'!
portraitTransform
	"My front-on transform for photos taken with the QOnsitePhotoJob,
	 which does not take the frame out to an empty studio space."
	| camera ext |
	camera := (self globalTransform clone).
	ext := (self portraitExtent).
	camera translation: (
		camera translation
		+ (camera orientation transformVector: (0@0@ (ext maxDimension * 1.05)))).

	^ camera! !

!TFrame methodsFor: '*Qwaq-Objects-Space' stamp: 'bgf 8/18/2009 00:18'!
removeNote: aNote
	"Recursively inform the containing note collectors under a stack."
	self surfaceAttachment ifNotNilDo: [ : sfc |
		sfc isNoteCollector ifTrue:[
			sfc removeNote: aNote
		]].! !

!TFrame methodsFor: '*Qwaq-Objects-Space' stamp: 'bgf 8/20/2009 12:54'!
resizeBegin: event
	"Resize adapter notification to luggage. A resize interaction is beginning. 
	 See QAbstractResizer "
	! !

!TFrame methodsFor: '*Qwaq-Objects-Space' stamp: 'bgf 8/20/2009 12:54'!
resizeCanceled
	"Resize adapter protocol, informing that  a resize interaction has been aborted.
	 This is sent after extent is restored to end a canceled drag.
	 See QAbstractResizer "
	! !

!TFrame methodsFor: '*Qwaq-Objects-Space' stamp: 'bgf 8/20/2009 12:54'!
resizeComplete: event
	"Resize adapter notification to luggage. A resize interaction has been completed.
	 (The last extent-update from the resizer should be treated as the finished extent.)
	 See QAbstractResizer " ! !

!TFrame methodsFor: '*Qwaq-Objects-Space' stamp: 'bgf 3/19/2010 10:22'!
resizerState
	"QAbstractResizer support  -
 	 This is the captured state record that can be used to undo or redo one resizer gesture.
	 See 'applyResizerState:', which consumes these state-arrays.
	 Subclasses should extend the array, but not alter the first three elements,
	 which are assumed by the resizer."

	^ { self localTransform clone . self surfaceAttachment . self extent clone }! !

!TFrame methodsFor: '*Qwaq-Objects-Space' stamp: 'bgf 8/15/2009 23:24'!
rootAttachment
	"The surface at the root of my attached-to tree."
	^ self surfaceAttachment ifNotNilDo: [ : sa |
		sa rootAttachmentExcluding: {self}
	]! !

!TFrame methodsFor: '*Qwaq-Objects-Space' stamp: 'bgf 8/15/2009 23:23'!
rootAttachmentExcluding: frames
	"Private worker routine for rootAttachment"
	^ self surfaceAttachment ifNotNilDo: [ : sa |
		(frames includes: sa) ifFalse: [
			sa rootAttachmentExcluding: (frames copyWith: self)
		] ifTrue: [	
			sa
		]
	] ifNil: [ self ].! !

!TFrame methodsFor: '*Qwaq-Objects-Space' stamp: 'bgf 8/27/2009 20:36'!
setupAttachment
	"Fires after scene-loading (for repairing legacy scenes)
	 and after releasing an object from an editing box...
	 if appropriate, rediscover its attachment point to a very nearby surface."

	((self root = self parent) 
	 and: [ self isGroundBound ]
	 and: [ self dragAdapter notNil ] ) 
	 ifTrue: [
		self root pickSurfaceFor: self alongNormal: (0@-1@0) withinDistance: 0.5
	].! !

!TFrame methodsFor: '*Qwaq-Objects-Space' stamp: 'bgf 7/31/2009 02:47'!
surfaceAttachment
	"Attachment manager query - what surface if any am I attached to?"
	^ self root ifNotNilDo: [ : root | 
		root surfaceAttachmentOf: self
	].! !

!TFrame methodsFor: '*Qwaq-Objects-Space' stamp: 'bgf 12/25/2008 12:12'!
warpThresholdAdjustment
	"Protocol:
	 factor to use to to adjust the default warp-threshold distance,
	 beyond which a click on an object will bring you to it rather 
	 than interacting with it.
	 A value of 2.0, for example, would give a doubled 'warp threshold'."

	^ 1.0! !

TFrame subclass: #TJoint
	instanceVariableNames: 'jointRoot initialTransform trackingObjects currentRotationMatrix'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Objects-Motion'!

!TJoint commentStamp: 'das 12/4/2006 10:04' prior: 0!
A TJoint is a point which represents a joint value from the .bvh file. 


This was seriously hacked up by DAS. I am just using this as a simple importer and will replace it later with something that imports directly into the form I want.!

!TJoint class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!TJoint methodsFor: 'creation' stamp: 'zamp 5/5/2006 10:18'!
addTrackingObject: obj
	trackingObjects add: obj! !

!TJoint methodsFor: 'events' stamp: 'zamp 12/9/2005 09:16'!
boundingBox: trans
	"Answer the bounding box for the receiver and its children"
	| box |
	box := self frameBox.
	^box! !

!TJoint methodsFor: 'accessing' stamp: 'zamp 5/23/2006 16:08'!
currentRotationMatrix
	currentRotationMatrix ifNil: [
		(frameParent class = TJointRoot) ifTrue: [
			self currentRotationMatrix: self localTransform.
		]
		ifFalse:[
			self currentRotationMatrix: ((frameParent currentRotationMatrix) * self localTransform).
		].
	].
	^ currentRotationMatrix

					 
! !

!TJoint methodsFor: 'accessing' stamp: 'zamp 5/18/2006 14:08'!
currentRotationMatrix: matrix
	currentRotationMatrix := matrix! !

!TJoint methodsFor: 'accessing' stamp: 'zamp 12/9/2005 09:16'!
extent
	^(self radius @ self radius @ self radius)! !

!TJoint methodsFor: 'events' stamp: 'zamp 12/9/2005 09:16'!
frameBox
	"Answer the local bounding box of this frame"
	| box |
	box := TBox origin: (self extent negated) corner: (self extent).
	^box
! !

!TJoint methodsFor: 'accessing' stamp: 'zamp 12/9/2005 09:16'!
initialTransform
	^ initialTransform! !

!TJoint methodsFor: 'accessing' stamp: 'zamp 5/5/2006 10:05'!
initialTransform: trans
	initialTransform := trans! !

!TJoint methodsFor: 'initialize' stamp: 'jdm 8/21/2007 08:20'!
initialize
	super initialize.

	trackingObjects := OrderedCollection new.
	initialTransform := Matrix4x4 new.

	" For drawing. Fix later. "
	self extent:0.15@0.15@0.15.

! !

!TJoint methodsFor: 'initialize' stamp: 'das 8/31/2006 14:57'!
initializeWithJointRoot: aJointRoot

	jointRoot := aJointRoot.
! !

!TJoint methodsFor: 'initialize' stamp: 'zamp 6/16/2006 09:30'!
initializeWithJointRoot: aJointRoot objectName: jointName
	self initializeWithJointRoot: aJointRoot.
	self objectName: jointName.

! !

!TJoint methodsFor: 'accessing' stamp: 'zamp 5/5/2006 11:41'!
jointRoot
	^jointRoot! !

!TJoint methodsFor: 'accessing' stamp: 'zamp 5/5/2006 11:41'!
jointRoot: obj
	jointRoot := obj.! !

!TJoint methodsFor: 'accessing' stamp: 'jdm 8/21/2007 08:20'!
makePipes: aParent

	| v cyl len group mat |

	group :=TGroup new.
	group objectName: self objectName.
	group localTransform: self localTransform.
	group setTransformChanged.	   "because we need recalc before we're rooted."
	aParent addChild: group.
	v := frameParent globalPosition-self globalPosition.
	len := v length.
	len ~= 0 ifTrue:[
		frameParent class = TJoint ifTrue:[
			(v x = 0) & (v z = 0) ifTrue:[
				mat :=Matrix4x4 new.
				v y > 0 ifTrue:[mat addRotationAroundX:180].
			] ifFalse:[
				mat := Matrix4x4 new up:v negated at:(0@1@0).
			].
			cyl := TCylinder new.
			cyl radius:0.1.
			cyl height: len.
			cyl slices: 5; stacks: 2.
			cyl objectName:#pipe.
			cyl localTransform:mat.
			aParent addChild: cyl.
		].
	].
	frameChildren ifNotNil:[frameChildren do:[:fc | fc makePipes: group.].].
	! !

!TJoint methodsFor: 'accessing' stamp: 'bgf 3/18/2007 04:24'!
makeSkeletonIn: parent

	"Make a TGroup in the given parent ground
	 with node name and transform this joint of the bvh structure,
	 and continue recursively."

	| group |

	group :=TGroup new.
	group objectName: self objectName.
	group localTransform: self localTransform.
	parent addChild: group.
	frameChildren ifNotNil:[
		frameChildren
			select:[:fc|fc isKindOf: TJoint]
			thenDo:[:fc | fc makeSkeletonIn: group].
	].
	group setTransformChanged.
	^ group
	! !

!TJoint methodsFor: 'animation' stamp: 'jdm 8/21/2007 08:20'!
reset
	self orientation: Matrix4x4 new.
	self translation: initialTransform translation.

	self frameChildrenDo: [ :child | child class = TJoint ifTrue:[child reset] ].
! !

PromiseMessageSend subclass: #QEventLoopMessageSend
	instanceVariableNames: 'scheduledTime sequenceNumber'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Objects-Kernel'!

!QEventLoopMessageSend commentStamp: '<historical>' prior: 0!
EventLoopMessageSend is a MessageSend that is specialized for scheduling and execution within an EventLoop.  The iVars are used by QEventLoop2 to stash various values used by the event-loop machinery.  Not all are necessarily used for a given message-send.  For example, "scheduledTime" and "sequenceNumber" are only used for internal message-sends, and "promise" is only used for a EventLoopMessageSend created by a "futureSend", not a "futureDo".!

!QEventLoopMessageSend class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QEventLoopMessageSend methodsFor: 'accessing' stamp: 'jcg 4/4/2010 23:03'!
scheduledTime
	"Return the time (in milliseconds) that the receiver is scheduled to be executed"
	^scheduledTime! !

!QEventLoopMessageSend methodsFor: 'accessing' stamp: 'jcg 4/4/2010 23:03'!
scheduledTime: msecs
	"Set the time (in milliseconds) that the receiver is scheduled to be executed"
	scheduledTime := msecs! !

!QEventLoopMessageSend methodsFor: 'accessing' stamp: 'jcg 4/5/2010 16:08'!
sequenceNumber
	"Answer the sequence number of the alarm, which is used to preserve ordering for messages scheduled for the same time."
	^sequenceNumber ifNil: [0]! !

!QEventLoopMessageSend methodsFor: 'accessing' stamp: 'jcg 4/5/2010 16:08'!
sequenceNumber: positiveInteger
	"Set the sequence number of the alarm, which is used to preserve ordering for messages scheduled for the same time."
	sequenceNumber := positiveInteger! !

TObject subclass: #QAbstractDocument
	instanceVariableNames: 'objectID suffix isInInventory mtime ctime serviceURL'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Objects-Document'!

!QAbstractDocument commentStamp: '<historical>' prior: 0!
Base protocols for documents (typically files) used as the content of file cards,
and the editable content of application sessions.

A document is a reference to some content data,
together with the metadata that allow us to launch applications to view it,
and to describe and depict it for the user.

A document is a resource, and the harness manages document resources
in some ways.  In particular, some shared documents are marked 'cacheLocally',
and for those, the harness automatically downloads in the background.

Document resource registration is the responsibility of client code;
documents do -not- register themselves.   Usually a file-card registers the document
it is holding, and unregisters it when the file card is unrooted.
(I.e. documents are usually registered when they are reachable in space.)

This base level implementation provides for using a suffix (e.g. txt, mpg, doc)
as a key into the registry 
	QDocumentType forSuffix: suffix
which uses a table of registered file-types,
from which descriptive strings,  the player application class, 
and such doc-type properties are retrieved.

Most documents identify files that live outside the shared space,
often only on the server (for use in QMS applications.)
Some documents include their content inline (such as QTextDocuments),
using the forum itself for storage.  
!

!QAbstractDocument class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QAbstractDocument class methodsFor: 'inspector' stamp: 'bgf 1/28/2010 11:51'!
addPropertyRows: dialog forFrame: docref
	"Base document properties (nested within file card properties)
	 give a way to see the mtime/ctime ."

	dialog addRow: (
		(QStringPropertyRow new)
			titleString: 'Created';
			fetchBlock: [ :f | f get: #ctimeSummary ];
			acceptBlock: [ : v |];
			balloonHelp: 'Creation time of the document';
		      inputEnabled: false;
			extraLong;
			yourself).
	dialog addRow: (
		(QStringPropertyRow new)
			titleString: 'Modified';
			fetchBlock: [ :f | f get: #mtimeSummary ];
			acceptBlock: [ : v |];
			balloonHelp: 'Last-modified time of the document';
		      inputEnabled: false;
			extraLong;
			yourself).! !

!QAbstractDocument class methodsFor: 'classifying' stamp: 'bgf 12/3/2008 15:02'!
defaultSuffix
	^ nil! !

!QAbstractDocument class methodsFor: 'files' stamp: 'bgf 9/9/2009 00:23'!
existingDocumentFileName: aDocument forClient: aClient
	"If available, return the filename of an existing local file containing the document;
	 nil if there is no such known."
	^ nil ! !

!QAbstractDocument class methodsFor: 'files' stamp: 'bgf 9/9/2009 14:12'!
fetch: document toFileNamed: fileName forClient: aClient
	"Harness-side protocol: 
	  fetch the given document into the given local file name, if this can be done efficiently.   
	 Return true on completion if the file exists as required;
	 false if this retrieval is unsupported.   
	 Raise Errors to prevent faling back to using getContent:forClient:

	 This is a worker routine for the local-save main methods, especially 
		saveDocument: asFile: forClient: 
	 which try various methods including this one to retrieve the data. "

	^ false! !

!QAbstractDocument class methodsFor: 'files' stamp: 'bgf 3/13/2009 02:13'!
forLocalFile: file suffix: ext  sharedName: sharedName inClient: owner
	^ nil
	
	" A subclass responsibility "! !

!QAbstractDocument class methodsFor: 'files' stamp: 'ar 4/16/2010 14:30'!
fromRepository: repoInfo docType: type session: qSession.
	"Protocol: if the file is accepted as this document type,
	create a repository document for it.
	 Otherwise returns nil (if no Document is produced.)."

	^nil

! !

!QAbstractDocument class methodsFor: 'files' stamp: 'bgf 12/3/2008 20:53'!
getContent: document forClient:owner

	"Harness-side operation to acquire the document (file) content as a byte array."

	self subclassResponsibility! !

!QAbstractDocument class methodsFor: 'files' stamp: 'hrs 4/7/2009 20:26'!
mailDocument: docRef forClient: aUI
	"Harness-side protocol for mailing out a document; 
	 returns the name of the file, or nil on failure,
	 or raises errors for the caller to report."
	^nil! !

!QAbstractDocument class methodsFor: 'customs' stamp: 'jcg 5/29/2009 15:53'!
needsFullDownloadToCreateFileCard
	"Be conservative, until we know better."
	^true! !

!QAbstractDocument class methodsFor: 'files' stamp: 'hrs 4/7/2009 20:26'!
openDocumentWebPage: docRef forClient: aUI
	"Harness-side protocol for opening a document Web page; 
	 returns the name of the file, or nil on failure,
	 or raises errors for the caller to report."
	^nil! !

!QAbstractDocument class methodsFor: 'customs' stamp: 'jcg 6/2/2009 11:44'!
processCustomsEntryData: entryData customs: customsPane type: docType path: path
	"Subclasses should create a file-card for the entry, if possible."
	^nil! !

!QAbstractDocument class methodsFor: 'resources' stamp: 'bgf 12/21/2008 17:21'!
resourceType
	"Documents (especially shared documents) 
	 are registered as resources for two reasons:
	 1) some shared documents are marked cacheLocally, and are fetched as resource files
	   by the content manager.
	 2) the document resource set for a forum is used (e.g. by the inspector)
	   to allow finding documents from a list of names."

	^ #document! !

!QAbstractDocument class methodsFor: 'files' stamp: 'bgf 3/3/2009 03:35'!
saveDocument: document asFile: newName forClient:aUI
	"Harness-side protocol for saving an instance as a local file; 
	 returns the short local name of the resulting file, or nil on failure,
	 or raises errors for the caller to report."
	self subclassResponsibility! !

!QAbstractDocument class methodsFor: 'manifest' stamp: 'bgf 2/16/2009 10:27'!
terseDocumentPath: uri
	"General purpose reducer for names of shared documents"
	| result |
	result := uri.
	"For most documents we just peel off all leading paths."
	(result includes: $/) ifTrue: [
		result := result copyFrom: (result lastIndexOf: $/) + 1 to: (result size) 
	].		
	^ result! !

!QAbstractDocument class methodsFor: 'files' stamp: 'bgf 4/29/2009 22:17'!
uploadForImport: aFile docType: type session: qSession 
	"Protocol: if the file is accepted as this document type,
	 upload it to the server as shared document content
	 and return the Document handle for it.
	 Otherwise returns nil (if no Document is produced.)."
	self subclassResponsibility
! !

!QAbstractDocument class methodsFor: 'manifest' stamp: 'bgf 2/23/2009 11:20'!
writeManifestEntry: docRef on: xmlWriter checkingDoneSet: doneShas
	"Write the Qwaq Forums manifest entry for the given document reference,
	 using get: #manifestXML to get the literal manifest xml snippet from the document itself."
	|id|
	docRef isValidFarRef ifFalse: [ ^ self ].
	(id := docRef get: #hashableName) ifNil: [ ^ self ].

	(doneShas includes: id) ifFalse: [
		| snippet | 
		docRef isValidFarRef ifFalse: [ ^ self ].
		snippet := docRef get: #manifestXML.
		snippet ifNotNil: [
			xmlWriter flush.
			xmlWriter stream nextPutAll: snippet; crlf.
			doneShas add: id.
		]].! !

!QAbstractDocument class methodsFor: 'xml' stamp: 'hrs 3/30/2009 20:44'!
xmlCtimeValue: anXMLElement
	"Answer the ctimeValue of a QAbstractDocument subclass made from anXMLElement."
	^(anXMLElement contentStringAt: 'ctime' ifAbsent: ['0']) asNumber! !

!QAbstractDocument class methodsFor: 'xml' stamp: 'hrs 3/30/2009 20:44'!
xmlCtimeWhat: anXMLElement
	"Answer the ctime what of a QAbstractDocument subclass made from anXMLElement."
	^anXMLElement contentStringAt: 'ctime-what' ifAbsent: []! !

!QAbstractDocument class methodsFor: 'xml' stamp: 'hrs 3/30/2009 20:45'!
xmlCtimeWho: anXMLElement
	"Answer the ctime who of a QAbstractDocument subclass made from anXMLElement."
	^anXMLElement contentStringAt: 'ctime-who' ifAbsent: []! !

!QAbstractDocument class methodsFor: 'xml' stamp: 'hrs 3/30/2009 20:45'!
xmlMtimeValue: anXMLElement
	"Answer the mtimeValue of a QAbstractDocument subclass made from anXMLElement."
	^(anXMLElement contentStringAt: 'mtime' ifAbsent: ['0']) asNumber! !

!QAbstractDocument class methodsFor: 'xml' stamp: 'hrs 3/30/2009 20:45'!
xmlMtimeWhat: anXMLElement
	"Answer the mtime what of a QAbstractDocument subclass made from anXMLElement."
	^anXMLElement contentStringAt: 'mtime-what' ifAbsent: []! !

!QAbstractDocument class methodsFor: 'xml' stamp: 'hrs 3/30/2009 20:45'!
xmlMtimeWho: anXMLElement
	"Answer the mtime who of a QAbstractDocument subclass made from anXMLElement."
	^anXMLElement contentStringAt: 'mtime-who' ifAbsent: []! !

!QAbstractDocument class methodsFor: 'xml' stamp: 'hrs 6/12/2009 11:24'!
xmlSHA: anXMLElement
	"Answer the sha of the QAbstractDocument subclass made from anXMLElement."
	^anXMLElement contentStringAt: 'sha' ifAbsent: []! !

!QAbstractDocument class methodsFor: 'xml' stamp: 'hrs 3/30/2009 20:45'!
xmlSharedName: anXMLElement
	"Answer the sharedName of a QAbstractDocument subclass made from anXMLElement."
	^anXMLElement contentStringAt: 'sharedName' ifAbsent: []! !

!QAbstractDocument class methodsFor: 'xml' stamp: 'hrs 3/30/2009 21:36'!
xmlShortName: anXMLElement
	"Answer the shortName of a QAbstractDocument subclass made from anXMLElement."
	^anXMLElement contentStringAt: 'shortName' ifAbsent: [
		anXMLElement contentStringAt: 'title' ifAbsent: [
			(self xmlSharedName: anXMLElement) ifNotNil: [:sn | sn copyAfterLast: $/]
		]
	]! !

!QAbstractDocument methodsFor: 'comparing' stamp: 'bgf 2/14/2009 22:40'!
= other
	"Usually, two instances are considered equal if they represent the same document (file),
	 and so equality is based on the hashableName."
	^ (self == other) or: [
		(other class = self class)
		 and: [other hashableName = self hashableName]].! !

!QAbstractDocument methodsFor: 'accessing' stamp: 'bgf 12/4/2008 12:22'!
asLocalFileUri
	"A file URI, relative to FileDirectory default, or an absolute URI,
	 by which off-island players can refer to this document.
	 *** Usable only if my #isKeptLocally is true ****

	 This method should -not- guide replicated calculations;
	 it is for use in local get's (from embedded apps, or from file-openin jobs)
	 Its main purpose is to distinguish bundled from cached files
	 and return proper paths fo those cases."

	| cachePath nm |

	self isKeptLocally ifFalse: [
		QLogger logWarning: 'Non-local document asLocalFileURI.'.
		^ nil
		].
	
	"If predefined bundled content,  and -is- present in the bundle, just use it."
	self isBundled ifTrue: [
		nm := self sharedName.
		(nm notNil 
		    and: [ FileDirectory default fileExists: ((FileDirectory default pathFromURI: nm))])
	 	ifTrue: [
			^ nm
		]].

	"General case - the local file URI for kept-local content is its cache path."
	nm := self cacheFileName.
	nm ifNotNil: [
		cachePath := (TContentManager default cacheManager directoryForFile: nm) uri path.
		^ cachePath, nm
	].
	^ nil! !

!QAbstractDocument methodsFor: 'accessing' stamp: 'bgf 12/4/2008 12:23'!
cacheFileName

	"The filename (a simple name, always referring to a file in the session file-cache)
	 for local cached copies of this document;
	 if the document requires local caching."

	^ nil! !

!QAbstractDocument methodsFor: 'accessing' stamp: 'bgf 12/4/2008 12:23'!
cacheLocally
	"Boolean - if true, tells the system to maintain a locally cached copy of the document file.
	 If this is overridden, then any document for which it returns true must also return a #cacheFileName"
	^ false! !

!QAbstractDocument methodsFor: 'protocol' stamp: 'bgf 12/3/2008 14:26'!
canBeCopiedToInventory
	"Can users do 'Copy To Inventory...?' 
	 Assume yes (since the default case is self-contained internals)"
	^ true! !

!QAbstractDocument methodsFor: 'protocol' stamp: 'bgf 12/3/2008 13:48'!
canBeExecuted
	"Can this be run as a python app? No."
	^ false! !

!QAbstractDocument methodsFor: 'protocol' stamp: 'bgf 12/3/2008 21:00'!
canBeOpenedPrivately
	"Can users do 'Open Private Copy...?' to invoke a QOpenDocExternalJob ?
	 This is offered for most documents.
	 See QOpenDocExternalJob"

	^ true! !

!QAbstractDocument methodsFor: 'protocol' stamp: 'bgf 12/8/2008 07:13'!
canBeSavedLocally
	"Can users do 'Save Local Copy...?' 
	 Offer save-local for most shared documents.
	 (The document subclass should override if it does not support this.)"
	^ true! !

!QAbstractDocument methodsFor: 'protocol' stamp: 'ar 4/16/2011 17:40'!
canCopyContentsOnly
	"Can this document exist when only its contents is being copied?
	The problem arises for various documents that have context in their
	associated file card / document. A Python document for example, stores the 
	application that created the contents. A repository document stores
	the repository it exists in and has no actual content bits unless checked out.

	When copying contents only (no file cards / documents) between Forums
	this method can be used to determine whether such a copy operation is
	meaningful."

	^false! !

!QAbstractDocument methodsFor: 'testing' stamp: 'jcg 4/10/2009 11:16'!
canUseAsForumWelcome
	"Default is yes."
	^true! !

!QAbstractDocument methodsFor: 'persist' stamp: 'bgf 2/12/2009 20:27'!
constructFrom: pStream
	super constructFrom: pStream.
	self suffix: (pStream readString: 'cacheSuffix' default: suffix).

	ctime constructFieldsFrom: pStream prefix: 'ctime'.
	mtime constructFieldsFrom: pStream prefix: 'mtime'.
! !

!QAbstractDocument methodsFor: 'accessing' stamp: 'bgf 12/3/2008 14:02'!
contentSummary
	"Short string summarizing my document file name."
	
	self subclassResponsibility! !

!QAbstractDocument methodsFor: 'initialize-release' stamp: 'bgf 3/3/2009 22:14'!
copyFrom: aDocument
	"Copies the Q...Document - not necessarily the underlying document.
	 This restates the resource registration,
	 and should probably be called at the -end- of subclass extensions."

	suffix := aDocument suffix.
	"Since the timestamps are about payload changes, just copy them also;
	 system mechanisms like inventory-copying should update their mtimes when data moves."
	self ctime: aDocument ctime.
	self mtime: aDocument mtime.
! !

!QAbstractDocument methodsFor: 'accessing' stamp: 'bgf 3/2/2009 11:30'!
creationTime
	"The ctime if known, of this document (zero if unknown)."
	^ (ctime time)
		ifNil: [ 0 ]! !

!QAbstractDocument methodsFor: 'accessing' stamp: 'bgf 2/20/2009 11:22'!
ctime
	"The creation QTimeStamp for this document"
	^ ctime ! !

!QAbstractDocument methodsFor: 'accessing' stamp: 'bgf 2/11/2009 15:45'!
ctime: astamp
	"The creation QTimeStamp for this document"
	ctime copyFrom: astamp.
! !

!QAbstractDocument methodsFor: 'accessing' stamp: 'bgf 2/15/2009 01:40'!
ctime: utcSeconds who: loginString what: tagString
	"Future-sendable shorthand to update the created mark."
	ctime who: loginString what: tagString utc: utcSeconds.! !

!QAbstractDocument methodsFor: 'accessing' stamp: 'bgf 2/28/2009 13:54'!
ctimeString
	"Gettable creation-time string (just the time, not the author/action)"
	^ ctime timeString! !

!QAbstractDocument methodsFor: 'accessing' stamp: 'bgf 2/19/2009 17:29'!
ctimeSummary
	"Gettable creation-time string summary of timestamp."
	^ ctime summary! !

!QAbstractDocument methodsFor: 'accessing' stamp: 'bgf 2/28/2009 14:26'!
ctimeValue
	"The creation utcSeconds for this document"
	^ ctime time ifNil:[ 0 ].! !

!QAbstractDocument methodsFor: 'initialize-release' stamp: 'bgf 3/3/2009 23:41'!
destroy
	"When released (e.g. by a file card.)
	 Ought not be doing this anymore.  Documents do not impose a destroy responsibility.
	 Rather, they impose the requirement that the primary owner
	 arrange to registerResource/unregisterResource. "

	QLogger logWarning: 'Destroyed a document.\' withCRs, (thisContext shortStack).
	self unregisterResource.! !

!QAbstractDocument methodsFor: 'save' stamp: 'bgf 12/21/2008 12:18'!
doOpenForAvatar: anAvatarReplica

	"Open the document privately in an application on the user's computer.
	 Use of this is guarded by  # canBeOpenedPrivately "

	self canBeOpenedPrivately ifTrue: [
		QOpenDocExternalJob open: self forUser: anAvatarReplica
	] ifFalse: [
		anAvatarReplica signal: #alertMessage with: 'That document cannot be opened privately.'.
	].! !

!QAbstractDocument methodsFor: 'save' stamp: 'bgf 12/21/2008 12:19'!
doSaveAsForAvatar: anAvatarReplica
	"Save the document locally in the client of the current avatar.
	 Use of this is guarded by  # canBeSavedLocally  "

	self canBeSavedLocally ifTrue: [
		QSaveDocumentJob save: self forUser: anAvatarReplica
	] ifFalse: [
		anAvatarReplica signal: #alertMessage with: 'That document cannot be saved locally.'.
	].! !

!QAbstractDocument methodsFor: 'private' stamp: 'bgf 1/29/2009 23:53'!
docType
	"The QDocumentType describing the handlers for this document.

	 This must return a non-nil QDocumentType 
	  - so we return a transient uninteresting 'File' type if no registered type is found.
	 (Use QDocumentType forSuffix:  directly to find the registered types.)

	 It is important that the type is -not- held as an instance variable,
	 but only looked-up... otherwise we'd have island/global/harness referencing to sort out;
	 this way, the single DocumentType table in QDocumentType can be used anywhere."

	^ (QDocumentType forSuffix: suffix)
		ifNil: [  QDocumentType forDocTypeEntry: {{suffix} .	nil . 'File'}  ]! !

!QAbstractDocument methodsFor: 'registry properties' stamp: 'jcg 2/5/2009 10:38'!
embeddedAppClass

	"The TEmbeddedApp subclass
	 for the viewer/editor of this kind of document."

	^ Smalltalk at: self embeddedAppClassName ifAbsent: [ nil ]! !

!QAbstractDocument methodsFor: 'registry properties' stamp: 'bgf 4/14/2010 10:30'!
embeddedAppClassName
	"Name of the factory class to which we send 
		newForDocument:extent:  
	 to create an embedded-app to open this document."

	^ self docType appFactoryClassName! !

!QAbstractDocument methodsFor: 'registry properties' stamp: 'bgf 3/5/2009 15:34'!
fileCardBacking
	"Return the name of one of the Qwaq/Cards/<something.png> 
	 file-card icon graphics.   E.g. 'DocumentCard.png'   
	 A custom document may specify its card-backing name by overriding this method.
	 (Must return the name of a backing that is in the product package, and not nil)"

	self docType isPDF ifTrue: [
		"Use the document card rather than the firefox url card that QRemoteFirefox suggests."
		^ 'MultiShareDocCard.png' ].

	self embeddedAppClass ifNotNilDo: [ : app |
		app fileCardBacking ifNotNilDo: [ : appBacking | ^ appBacking ] ].

	self docType isPDF ifTrue: [ 'MultishareDocCard.png' ].
	self docType isSampledSound ifTrue: [ ^ 'SoundCard.png' ].
	self docType isMP4 ifTrue: [ 'MovieCard.png' ].
	self docType keySuffix = 'mpg' ifTrue: [ ^ 'MovieCard.png' ].
	self docType keySuffix = 'txt' ifTrue: [ ^ 'TextCard.png' ].
	self suffix = 'c3z' ifTrue: [ ^ 'ObjCard.png' ].

	^ 'DocumentCard.png'! !

!QAbstractDocument methodsFor: 'registry properties' stamp: 'bgf 3/5/2009 17:32'!
fileCardTexture
	"Return a TFileTexture for the outer body for QFileCards holding me.
	 Subclasses  supply #fileCardBacking (the short file name) rather than changing this."

	^ (QTextures find: 'Content/Qwaq/Textures/Cards2/',  (self fileCardBacking ifNil: [ 'DocumentCard.png'])).


! !

!QAbstractDocument methodsFor: 'save' stamp: 'bgf 12/3/2008 14:57'!
fileChooserPattern
	"When opening such documents from the QFileChoosers."
	^ { {self kindSummary . '*.', self suffixes first}}.! !

!QAbstractDocument methodsFor: 'save' stamp: 'bgf 12/3/2008 13:41'!
fileChooserPatternTweak
	"When opening such documents from the Tweak file choosers."

	^ { {self kindSummary . {'*.', self suffixes first }}}.! !

!QAbstractDocument methodsFor: 'registry properties' stamp: 'bgf 12/3/2008 13:41'!
fileModeSelector
	"Stream mode when writing the contents of the document; performed on a FileStream."
	^ #binary! !

!QAbstractDocument methodsFor: 'protocol' stamp: 'bgf 12/3/2008 13:48'!
fileName
	"For compatibility with TFileTexture, etc, for resource capture.
	 This is the server-side file uri for the resource, relative to the containing forum directory"
	^ nil! !

!QAbstractDocument methodsFor: 'save' stamp: 'bgf 12/3/2008 21:12'!
fileSaveLabel
	"Title for a save-as dialog"
	^ 'Save ', (self kindSummary)! !

!QAbstractDocument methodsFor: 'protocol' stamp: 'bgf 1/31/2009 19:08'!
fileTextures
	"Optional (most documents won't need this.)
	 If the document embeds any TFileTexture  dependencies,
	 list them here, to make them locatable in search tools."
	^ #()! !

!QAbstractDocument methodsFor: 'protocol' stamp: 'ar 4/16/2010 14:35'!
fullySupportsRepositories
	"Answers if the underlying document fully supports repository workflows,
	i.e., check-in, check-out, locking etc. Many of our current document types don't"

	^false
! !

!QAbstractDocument methodsFor: 'accessing' stamp: 'bgf 2/28/2009 10:51'!
get: aSymbol
	"For far-ref / instance interchangeability."
	^ self perform: aSymbol! !

!QAbstractDocument methodsFor: 'comparing' stamp: 'bgf 12/3/2008 13:41'!
hash
	"This is implemented because = is implemented."
	^ self hashableName hash

! !

!QAbstractDocument methodsFor: 'comparing' stamp: 'bgf 12/3/2008 23:01'!
hashableName
	"Name for comparison purposes.
	 In subclasses that represent off-island document files,
	 this can be based on that off-island file path+name."
	^ self objectID hex! !

!QAbstractDocument methodsFor: 'miramar' stamp: 'ar 2/11/2009 18:36'!
iconForm
	^nil
! !

!QAbstractDocument methodsFor: 'initialize-release' stamp: 'bgf 3/3/2009 22:14'!
initialize

	super initialize.
	mtime := QTimeStamp new.
	ctime := QTimeStamp new.

	"Shared documents need to be far-ref accessible when the far refs
	 are fetched in collections (the UI gets lists from the resource manager)
	 so they need to be registered..."
	objectID := TObjectID for: (self class name).
	Processor activeIsland register: self name: objectID.
	isInInventory := false.
! !

!QAbstractDocument methodsFor: 'accessing' stamp: 'bgf 12/4/2008 12:17'!
isBundled
	"Is this document stored in Qwaq prepackaged Content files?
	 Bundled files are either already in the package (in older packaging)
	 or are fetched from the special system-resources/bunlded/Content area.
	 (Only specially named content of certain subclasses is bundled.)"

	^ false! !

!QAbstractDocument methodsFor: 'accessing' stamp: 'bgf 12/21/2008 13:02'!
isInInventory
	"True if I am held by a file card in a user's inventory;
	 the avatar inventory and file-card systems maintain this flag."
	^ isInInventory! !

!QAbstractDocument methodsFor: 'accessing' stamp: 'bgf 3/3/2009 22:15'!
isInInventory: aBool
	"True if I am owned by a file card in a user's inventory;
	 the avatar inventory and file-card systems maintain this flag for me."
	isInInventory := aBool.
	isInInventory ifTrue: [
		self unregisterResource.
		].! !

!QAbstractDocument methodsFor: 'accessing' stamp: 'bgf 12/4/2008 12:18'!
isKeptLocally
	"Boolean - true if the document is cached locally,
	 or bundled with the software.
	 If this returns true, then self #asLocalFileUri returns the uri path to the file."

	^ self cacheLocally or: [ self isBundled ]! !

!QAbstractDocument methodsFor: 'launching' stamp: 'bgf 3/3/2009 18:51'!
isOpenInApplication
	"True if I am the document claimed by any currently-running embedded-app"
	^ (self island resourcesAt: QEmbeddedApp resourceType) 
		anySatisfy: [ : ea  | ea isRunning and: [ ea document == self ] ]! !

!QAbstractDocument methodsFor: 'accessing' stamp: 'bgf 7/30/2009 17:33'!
isPythonDoc
	"See PyPanelDocCS"
	^ false! !

!QAbstractDocument methodsFor: 'protocol' stamp: 'bgf 2/7/2010 12:54'!
isRenamableContent
	"Document protocol; is this resource a user-modifiable application-document?
	 (For items that return true, the server-side documents are renamed when content is pasted,
	 to create new instances of the documents. 
	 For items that return false, the presence of a resource of the given name is considered sufficient"
	^ self sharedName 
			ifNotNil: [ : sn | (sn beginsWith: 'resources/') not ]
			ifNil: [ false ]! !

!QAbstractDocument methodsFor: 'protocol' stamp: 'ar 4/15/2010 12:24'!
isRepositoryDocument
	"Repository support. Answers if this document is stored in a repository.
	If so, the document is expected to implement the QRepositoryDocument
	protocol for retrieving repository information."

	^false! !

!QAbstractDocument methodsFor: 'registry properties' stamp: 'bgf 1/29/2009 22:41'!
kindSummary

	"User-visible name of the type of this document."
	^ self docType description! !

!QAbstractDocument methodsFor: 'protocol' stamp: 'bgf 2/25/2009 09:28'!
labelForBrowser
	"For uniform inspector/property treatment."
	^ (self kindSummary), ': ', (self class terseDocumentPath: (self contentSummary) )! !

!QAbstractDocument methodsFor: 'accessing' stamp: 'bgf 3/13/2009 16:23'!
latestTimeString
	"Gettable most recent timestamp string."
	^ mtime >= ctime ifTrue: [ self mtimeString ] ifFalse: [ self ctimeString ].
	! !

!QAbstractDocument methodsFor: 'save' stamp: 'hrs 4/4/2009 15:39'!
mailFromAvatar: anAvatarReplica
	"Open a mail client for this doc on the machine of the current avatar. "

	self canBeSavedLocally ifTrue: [
		QMailDocumentJob mail: self fromUser: anAvatarReplica
	] ifFalse: [
		anAvatarReplica signal: #alertMessage with: 'That document cannot be mailed.'.
	].! !

!QAbstractDocument methodsFor: 'private' stamp: 'bgf 2/24/2009 20:14'!
manifestPersistString
	"Generate the persistString of the document, but elide content payload if any.
	 I.e. should write just the naming, timestamps and other such metadata.
	 Most kinds of QAbstractDocument hold only metadata (the doc files holding content)
	 but some, like Text and SlideShow hold open-ended amounts of data
 	 that would bloat the manifest."

	^ TSceneWriter new storeScene: self! !

!QAbstractDocument methodsFor: 'protocol' stamp: 'ar 4/19/2010 11:23'!
manifestXML

	"Must be get: safe.  Return the literal xml, e.g.
		<document attr='x' attr2='y' />
	 that is written to the forum manifest for this document."

	^ String streamContents: [ : str |
		| xmlWriter tag |
		xmlWriter := XMLWriter on: str.
		tag := #document.

		xmlWriter startTag: tag.	
		"Basic human-readable info so the manifest is not opaque when debugging"
		self fileName ifNotNilDo: [ : uri | 
				xmlWriter attribute: 'uri' value: uri ].
		self summary  ifNotNilDo: [ : summary | 
				xmlWriter attribute: 'summary' value: summary ].
		self sourceUrl  ifNotNilDo: [ : url | 
				xmlWriter attribute: 'source' value: url ].
		xmlWriter endTag.

		"For real use, we embed the persist-string of the document."
		xmlWriter pcData:  (self manifestPersistString zipped base64Encoded).
		xmlWriter endTag: tag.
		].! !

!QAbstractDocument methodsFor: 'accessing' stamp: 'bgf 3/2/2009 11:28'!
modificationTime
	"The most recent utc time in either creation or modified stamps"
	^ ((mtime isUndefined) ifTrue: [ ctime time ] ifFalse:[ mtime time ])
		ifNil: [ 0 ]! !

!QAbstractDocument methodsFor: 'accessing' stamp: 'bgf 2/11/2009 11:53'!
mtime
	"Most recent modification (QTimeStamp) known for this document."
	^ mtime! !

!QAbstractDocument methodsFor: 'accessing' stamp: 'bgf 2/12/2009 22:29'!
mtime: astamp
	"The most-recent modification QTimeStamp for this document"
	mtime copyFrom: astamp.
	ctime isUndefined ifTrue: [ ctime copyFrom: mtime ].! !

!QAbstractDocument methodsFor: 'accessing' stamp: 'bgf 2/15/2009 01:01'!
mtime: utcSeconds who: loginString what: tagString
	"Future-sendable shorthand to update the modified mark."
	mtime who: loginString what: tagString utc: utcSeconds.! !

!QAbstractDocument methodsFor: 'accessing' stamp: 'bgf 2/28/2009 13:53'!
mtimeString
	"Gettable last-modified string of time (without author/action)."
	^ mtime timeString! !

!QAbstractDocument methodsFor: 'accessing' stamp: 'bgf 2/19/2009 17:29'!
mtimeSummary
	"Gettable last-modified string summary of timestamp."
	^ mtime summary! !

!QAbstractDocument methodsFor: 'accessing' stamp: 'bgf 2/28/2009 14:26'!
mtimeValue
	"Most recent modification utcseconds known for this document."
	^ mtime time ifNil: [ 0 ]! !

!QAbstractDocument methodsFor: 'launching' stamp: 'jcg 2/5/2009 11:33'!
newEmbeddedAppForExtent: aRectExtent
	"Instantiate the embedded application class for this document;
	 we refer this to the EmbeddedApp class registered in #documentTypes."
	^(self embeddedAppClass) newForDocument: self extent: aRectExtent.! !

!QAbstractDocument methodsFor: 'accessing' stamp: 'bgf 12/3/2008 15:45'!
objectID
	^ objectID! !

!QAbstractDocument methodsFor: 'save' stamp: 'hrs 4/6/2009 22:09'!
openWebPageForAvatar: anAvatarReplica
	"Open a mail client for this doc on the machine of the current avatar. "

	self canBeSavedLocally ifTrue: [
		QOpenDocumentWebPageJob openWebPage: self fromUser: anAvatarReplica
	] ifFalse: [
		anAvatarReplica signal: #alertMessage with: 'That document does not have a Web page.'.
	].! !

!QAbstractDocument methodsFor: 'persist' stamp: 'bgf 2/5/2010 19:59'!
persistOn: pStream
	super persistOn: pStream.

	"Note: document persistence not should not write open outbound object references,
	 so that documents can safely be persisted in self-contained xml."
	ctime persistFieldsOn: pStream prefix: 'ctime'.
	mtime persistFieldsOn: pStream prefix: 'mtime'.

	pStream storeString: 'cacheSuffix' value: self suffix default: (self class defaultSuffix).

	"N.B. for subclassing:
       any subclass should persist at least one of these file/path name properties:
		 sharedName
		 shortName
		 title
	to satisfy the server-side RSS unpacker.   
	See QRSSItem forDoc:organization:forum:qrl: "

	self sha ifNotNil: [
		pStream addResource: self
	].! !

!QAbstractDocument methodsFor: 'copying' stamp: 'bgf 3/3/2009 22:15'!
postCopy
	"This is followup to a normal shallow copy. 
	 I am a new handle on the same underlying document data."
	objectID := TObjectID for: self.
	Processor activeIsland register: self name: objectID.
	! !

!QAbstractDocument methodsFor: 'island-copy' stamp: 'bgf 3/3/2009 22:15'!
postImportFromIslandCopier: importer
	super postImportFromIslandCopier: importer.
	"Shared documents need to be far-ref accessible when the far refs
	 are fetched in collections (the UI gets lists from the resource manager.)"
	objectID := TObjectID for: (self class name).
	Processor activeIsland register: self name: objectID.
! !

!QAbstractDocument methodsFor: 'protocol' stamp: 'bgf 3/5/2009 12:36'!
prepareForDropIntoSession: aSession serviceURL: aForumUrl
	"Specialized preflight opportunity.
	 This offline (harness-side) document is about to be island-copied 
	 into the active island of the given session.
	 (The document currently is thought to reside in some other forum.)
	 Should raise error if something prevents that from working."

! !

!QAbstractDocument methodsFor: 'resources' stamp: 'bgf 3/3/2009 21:23'!
registerResource
	"Register as a resource for island resource management.
	 There are many paths triggering this registration...
	 it is simplest to keep the operation idempotent. "
	
	self isInInventory ifTrue: [ 
		^ self island unregisterResource: self type: self resourceType ].

	"Doing it this way rather un/re prevents flip-flop update signals."
	(self island resourcesAt: (self resourceType)) 
		detect: [ : me | me == self ]
		ifNone: [
			self island registerResource: self type: self resourceType
		].! !

!QAbstractDocument methodsFor: 'resources' stamp: 'bgf 12/21/2008 12:05'!
resourceType
	^ self class resourceType! !

!QAbstractDocument methodsFor: 'protocol' stamp: 'bgf 3/3/2009 16:03'!
saveComplete: shortName
	"Callback from save-as (and save for private open) 
	 when the harness side operation completes.  
	 The shortName is that used on the client's machine."

	self signal: #localSaveComplete with: shortName! !

!QAbstractDocument methodsFor: 'accessing' stamp: 'bgf 3/1/2009 23:59'!
searchString
	"The summary string to search for name-matches for this document."
	^ String streamContents: [ : str |
			str nextPutAll: (self kindSummary); space;
			     nextPutAll: (self contentSummary); space.
			ctime who ifNotNilDo: [ : w | str nextPutAll: w; space ].
			mtime who ifNotNilDo: [ : w | str nextPutAll: w; space ].
			].
! !

!QAbstractDocument methodsFor: 'accessing' stamp: 'bgf 12/3/2008 13:41'!
serviceName
	"The service (forum) name from the Qwaq/Croquet service URL (//host/realm.../serviceName)
	 under which this document lives.
	 This will be nil for documents that are bundled (e.g. in the software, or available globally
	 from the system-resources of the server."

	self isBundled ifTrue: [ ^ nil ].		"Not in the storage of any particular service"

	^ self serviceURL ifNotNilDo:[ :url | url copyAfterLast: $/]! !

!QAbstractDocument methodsFor: 'accessing' stamp: 'bgf 1/10/2010 14:11'!
serviceURL

	"Return the Qwaq-URL for the service (forum) containing this document.
	 We usually get the service URL from the island containing me, 
	 since it is not really appropriate to stuff a full url string into every QSharedDocument.
	 May be nil, for documents not in service islands.
	 Relies on living in a Qwaq harness - see QwaqCroquetSession #addIsland:..."

	^ serviceURL ifNil: [ 
		self island at: #ServiceURLString
	 	].! !

!QAbstractDocument methodsFor: 'accessing' stamp: 'bgf 2/28/2009 11:57'!
serviceURL: aString

	"The Qwaq-URL for the service (forum) containing this document.
	 Note that on-island, we get the service URL from the island containing me, 
	 since it is not really appropriate to stuff a full url string into every QSharedDocument.
	 This setter is for harness-side use of shadow documents."

	serviceURL := aString.
	! !

!QAbstractDocument methodsFor: 'protocol' stamp: 'bgf 2/13/2009 11:09'!
sha
	"A sha hex string - necessary for documents that are cached locally "
	^ nil! !

!QAbstractDocument methodsFor: 'accessing' stamp: 'bgf 12/3/2008 14:28'!
sharedName
	"The URI path relative to the service's router-side storage root
	 under which this file is kept...
	 usually a relative path of the form 'sharedDocuments/[user]/baseFile.mpeg'

	 For documents in inventory, the name is /users/<user>/..... '
	 For special-case preloaded content (like bundled demo movies)
	 this will be a path into <default>/Content/..."
	^ nil! !

!QAbstractDocument methodsFor: 'accessing' stamp: 'bgf 12/3/2008 14:27'!
sharedName: aString

	"The path relative to the service's router-side storage root
	 under which this file is kept.
	 One of three cases can be expected:

	 - Relative name starting with 'sharedDocuments/...'   
	   refers to a document in this forum's sharedDocuments area.
	 
	-  Absolute name starting with /users/...  
	   refers   to a document in a user's subdirectory in realm/users.
	   These are in inventory - cannot (by construction) be opened in shared apps.

	- Special Case: paths starting with 'Content/' are BUNDLED;
	  these live in the software installation and are never downloaded
 	  or copied to cache.  Used for seeding the QwaqCroquetMovie mpegFile
	  in forum templates."

	self subclassResponsibility! !

!QAbstractDocument methodsFor: 'accessing' stamp: 'bgf 3/1/2009 10:10'!
shortName
	"The basic file name used for private copies.
	 Most subclasses should override this for their own naming convention."
	^ 'document-', (self objectID hex)
! !

!QAbstractDocument methodsFor: 'accessing' stamp: 'bgf 12/3/2008 13:41'!
shortName: aString
	"No-op; used to remember the most recent local copy save name,
	 but we don't bother to note that in true shared documents."
! !

!QAbstractDocument methodsFor: 'protocol' stamp: 'ar 4/16/2011 17:40'!
sourceUrl

	"For documents that have a 'source' outside of OpenQwaq, this method must
	return the document source. For example, for repository documents the
	sourceUrl is the url of the document in its repository. Documents that have
	no source associated with them, should return nil."

	^nil! !

!QAbstractDocument methodsFor: 'private' stamp: 'bgf 2/15/2009 15:13'!
space
	"Documents live  outside the scene graph, members of other objects,
	 but it is convenient to have access to the (presumed) containing space."
	^ self root ifNil: [
		((self island resourcesAt: (TSpace resourceType)) ifEmpty: [^nil])  first
	].! !

!QAbstractDocument methodsFor: 'accessing' stamp: 'bgf 12/3/2008 14:08'!
suffix
	"The basic type-determining suffix to be used
	 when making a local privately-named copy of this document.
	 Does not include the '.'  "
	^ suffix ifNil: [self suffixes first].! !

!QAbstractDocument methodsFor: 'accessing' stamp: 'bgf 12/3/2008 14:33'!
suffix: sfx
	"The basic type-determining suffix to be used
	 when making a local privately-named copy of this document.
	 Does not include the '.'  "
	suffix := sfx! !

!QAbstractDocument methodsFor: 'registry properties' stamp: 'bgf 1/29/2009 22:41'!
suffixes
	"The file name suffixes that match my type."

	^ self docType suffixes! !

!QAbstractDocument methodsFor: 'accessing' stamp: 'bgf 12/3/2008 13:41'!
summary

	"The short summary string (used in mouseovers) to describe this document."

	^ (self kindSummary), ': ', self contentSummary.
! !

!QAbstractDocument methodsFor: 'resources' stamp: 'bgf 12/21/2008 12:05'!
unregisterResource
	"Unregister the resource"
	self island unregisterResource: self type: self resourceType.! !

!QAbstractDocument methodsFor: 'private' stamp: 'bgf 2/28/2009 11:53'!
wait
	"No-op for far-ref equivalency."! !

QAbstractDocument subclass: #QExistingStreamingVideoDocument
	instanceVariableNames: 'streamID'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Objects-Hacks'!

!QExistingStreamingVideoDocument commentStamp: '<historical>' prior: 0!
Used to view an existing NetVid video, assuming that you know its streamID.!

!QExistingStreamingVideoDocument class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QExistingStreamingVideoDocument methodsFor: 'accessing' stamp: 'jcg 7/23/2009 16:19'!
contentSummary
	^'Hack to connect to existing NetVid stream: ', streamID! !

!QExistingStreamingVideoDocument methodsFor: 'accessing' stamp: 'jcg 7/23/2009 16:18'!
embeddedAppClass
	^QStreamingVideoApp! !

!QExistingStreamingVideoDocument methodsFor: 'accessing' stamp: 'jcg 7/23/2009 16:18'!
embeddedAppClassName
	^self embeddedAppClass name! !

!QExistingStreamingVideoDocument methodsFor: 'accessing' stamp: 'jcg 7/23/2009 16:07'!
isAutoPlay
	^false! !

!QExistingStreamingVideoDocument methodsFor: 'accessing' stamp: 'jcg 7/23/2009 16:07'!
isLooping
	^false! !

!QExistingStreamingVideoDocument methodsFor: 'launching' stamp: 'jcg 7/23/2009 16:10'!
newEmbeddedAppForExtent: aRectExtent
	| app |
	app := super newEmbeddedAppForExtent: aRectExtent.
	app streamID: streamID.
	^app! !

!QExistingStreamingVideoDocument methodsFor: 'accessing' stamp: 'jcg 7/23/2009 16:08'!
streamID: newStreamID
	streamID := newStreamID! !

!QExistingStreamingVideoDocument methodsFor: 'accessing' stamp: 'jcg 7/23/2009 16:06'!
volume
	^0! !

!QExistingStreamingVideoDocument methodsFor: 'accessing' stamp: 'jcg 7/23/2009 16:06'!
volume: ignore
! !

QAbstractDocument subclass: #QImageDocument
	instanceVariableNames: 'fileTexture sharedName'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Objects-Document'!

!QImageDocument class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QImageDocument class methodsFor: 'inspector' stamp: 'bgf 1/28/2010 14:02'!
addPropertyRows: dialog forFrame: docref
	"Shared document properties... add sharedName to the info,
	 since the basic labeling uses the shorter form."

	dialog addRow: (
		(QStringPropertyRow new)
			titleString: 'Image File';
			fetchBlock: [ : d | d get: #shortName ];
			acceptBlock: [ : v |];
			balloonHelp: 'File name of the imported image.';
		      inputEnabled: false;
			extraLong;
			yourself).
	dialog addRow: (
		(QStringPropertyRow new)
			titleString: 'Image Extent';
			fetchBlock: [ : d | d get: #extent ];
			acceptBlock: [ : v |];
			balloonHelp: 'Pixel extent of the imported image.';
		      inputEnabled: false;
			extraLong;
			yourself).
	super addPropertyRows: dialog forFrame: docref.! !

!QImageDocument class methodsFor: 'files' stamp: 'ar 4/16/2011 17:12'!
copy: docRef toActiveForum: aParticipant 
	"Delegate if possible... the TFT method does some nice redundancy checks
	 (images are immutable and sha-identified.)"
	(docRef get: #fileTexture) ifNotNil: [ : ft |
		ft isValidRef ifTrue: [
			TFileTexture copy: ft toActiveForum: aParticipant.
			docRef isValidFarRef ifTrue: [
				docRef future serviceURL: (aParticipant serviceSvc url).
				docRef future getSharedNameFromTexture.
			] ifFalse: [
				docRef isValidRef ifTrue: [ 
					docRef serviceURL: (aParticipant serviceSvc url).
					docRef getSharedNameFromTexture.
				]].
			^ self.
		]] ifNil: [
			QLogger logWarning: 'Image copy to Forum - empty document.'.
		].

	QSharedDocument copy: docRef toActiveForum: aParticipant ! !

!QImageDocument class methodsFor: 'files' stamp: 'bgf 4/30/2009 12:49'!
copy: docFarRef toUserInventory: aParticipant
	QSharedDocument copy: docFarRef toUserInventory: aParticipant.! !

!QImageDocument class methodsFor: 'files' stamp: 'bgf 9/9/2009 01:16'!
existingDocumentFileName: aDocument forClient: aClient
	"Get the name of the local cached file of the image, or nil.  Simple delegation."
	^ TFileTexture existingDocumentFileName: (aDocument get: #fileTexture) 
				   forClient: aClient! !

!QImageDocument class methodsFor: 'files' stamp: 'bgf 4/29/2009 14:56'!
forLocalFile: file suffix: ext  sharedName: sharedName inClient: owner
	| texture |
	"Import the texture to trap any errors (broken files etc)."
	texture := TContentManager default
		loadTexture: file name 
		mipmap: true 
		shrinkFit: true 
		extension: (TTextures extensionForFile: file name).
	texture ifNil: [
		self error:'Cannot import that file type as an image.'.
	].
	texture fileName: sharedName.

	^QImageDocument forTexture: texture

! !

!QImageDocument class methodsFor: 'instance creation' stamp: 'bgf 4/29/2009 13:59'!
forTexture: tfileTexture
	^ (self new) 
		fileTexture: tfileTexture;
		yourself! !

!QImageDocument class methodsFor: 'files' stamp: 'bgf 4/30/2009 20:12'!
getContent: document forClient:owner
	"Get the image byte-array.  Simple delegation..."
	^ TFileTexture getContent: (document get: #fileTexture) forClient: owner! !

!QImageDocument class methodsFor: 'files' stamp: 'bgf 4/29/2009 13:28'!
mailDocument: docRef forClient: aUI
	"We can't attach docs on Mac yet, so just send a QRL to the doc."
	^aUI mailDocumentLink: docRef! !

!QImageDocument class methodsFor: 'files' stamp: 'bgf 4/29/2009 13:28'!
openDocumentWebPage: docRef forClient: aUI
	"Open the Web page for the document."
	^aUI openQrlForDocument: docRef! !

!QImageDocument class methodsFor: 'customs' stamp: 'jcg 6/2/2009 12:43'!
processCustomsEntryData: entryData customs: customsPane type: docType path: path
	"We need the whole file; bounce the responsibility back to the 'customsPane'."
	^customsPane
		postProcessEntryData: entryData
		fromFile: path
		sharedName: entryData internedServicePath
		extension: entryData fileExtension asUppercase! !

!QImageDocument class methodsFor: 'files' stamp: 'bgf 9/9/2009 01:17'!
saveDocument: docRef asFile: newName forClient: aUI
	"Simple delegation; push the work down to the texture,
	  whose class implements shortcuts through the cache."
	^ QSharedDocument 
		saveDocument: (docRef get: #fileTexture) 
		asFile: newName 
		forClient: aUI! !

!QImageDocument class methodsFor: 'files' stamp: 'bgf 4/29/2009 13:26'!
uploadForImport: aFile docType: type session: session
	| texture |
	texture  := session importImageAsTexture: aFile shrinkFit: true.
	^ (self new)
		fileTexture: texture;
		yourself! !

!QImageDocument methodsFor: 'accessing' stamp: 'bgf 4/30/2009 10:42'!
asLocalFileUri
	"Support prootocol for file transfer management."
	^ fileTexture ifNotNil: [ fileTexture asLocalFileUri ]! !

!QImageDocument methodsFor: 'save' stamp: 'bgf 4/30/2009 10:40'!
asServiceManagerPath
	"Support prootocol for file transfer management."
	^ fileTexture ifNotNil: [ fileTexture asServiceManagerPath ]! !

!QImageDocument methodsFor: 'accessing' stamp: 'bgf 4/30/2009 10:42'!
cacheFileName
	"The filename for local cached copies of this document;
	 if the document requires local caching."
	^ fileTexture ifNotNil: [ fileTexture cacheFileName ]! !

!QImageDocument methodsFor: 'accessing' stamp: 'bgf 7/9/2009 17:32'!
cacheLocally
	"The image is cached locally, but this is handled by the texture resource,
	 and the cache handling of the doc-resource should not conflict with that,
	 so we don't say true here."
	^ false! !

!QImageDocument methodsFor: 'accessing' stamp: 'bgf 4/30/2009 17:52'!
cacheLocally: bool
	"No-op; textures are always cache-locally"! !

!QImageDocument methodsFor: 'protocol' stamp: 'ar 5/18/2010 17:06'!
canCopyContentsOnly
	"Image files can be copied"

	^true! !

!QImageDocument methodsFor: 'persist' stamp: 'bgf 4/30/2009 10:36'!
constructFrom: pStream
	super constructFrom: pStream.

	"The sharedName stored at this level is a fallback/safety measure in case
	 there is no well-formed file texture. It is mostly saved so the QRSS unpacker
	 on the server side finds a proper shared name for an image document."
	sharedName := pStream readString: 'sharedName' default: self sharedName.
	
	self fileTexture: (pStream readObject: 'image' default: nil).! !

!QImageDocument methodsFor: 'accessing' stamp: 'bgf 4/30/2009 10:43'!
contentSummary
	^ fileTexture ifNotNil: [ fileTexture shortName ] ifNil: [ 'empty' ]! !

!QImageDocument methodsFor: 'initialize-release' stamp: 'bgf 4/30/2009 10:43'!
copyFrom: aDocument
	super copyFrom: aDocument.
	"Defeat the careful ways the texture system protects against surplus copies,
	 since this is usually done to make a fresh instance to receive a modified file path
	 when copying to/from inventory.  Don't want to clobber the original."
 	fileTexture ifNotNil: [fileTexture destroy].
	fileTexture := TFileTexture new.
	TSceneReader new 
		read: fileTexture 
		from: (aDocument fileTexture persistString).! !

!QImageDocument methodsFor: 'save' stamp: 'bgf 4/29/2009 14:57'!
doOpenForAvatar: anAvatarReplica
	^ fileTexture doOpenForAvatar: anAvatarReplica! !

!QImageDocument methodsFor: 'save' stamp: 'bgf 4/29/2009 14:57'!
doSaveAsForAvatar: avatar
	^ fileTexture doSaveAsForAvatar: avatar! !

!QImageDocument methodsFor: 'accessing' stamp: 'bgf 4/30/2009 11:08'!
extent
	"For property dialog support."
	^ fileTexture ifNotNil: [ fileTexture extent ] ifNil: [ 0@0 ]! !

!QImageDocument methodsFor: 'registry properties' stamp: 'bgf 4/29/2009 12:03'!
fileCardBacking
	^ 'ImageCard.png'! !

!QImageDocument methodsFor: 'save' stamp: 'bgf 4/30/2009 20:02'!
fileChooserPattern
	^ fileTexture 
		ifNotNil: [ fileTexture fileChooserPattern ]
		ifNil: [ super fileChooserPattern ]! !

!QImageDocument methodsFor: 'protocol' stamp: 'bgf 4/30/2009 10:44'!
fileName
	"For compatibility with TFileTexture, etc, for resource capture.
	 This is the server-side file uri for the resource, relative to the containing forum directory"
	^  fileTexture ifNotNil: [fileTexture fileName]! !

!QImageDocument methodsFor: 'save' stamp: 'bgf 4/30/2009 20:03'!
fileSaveLabel
	"Abstract-Document protocol for local Save-As.
	  When saving a texture, the title of the file dialog."
	^ 'Save Image'
! !

!QImageDocument methodsFor: 'accessing' stamp: 'bgf 4/30/2009 10:37'!
fileTexture
	"The TFileTexture I stand for.  
	 (The whole purpose of QImageDocument is to present a filetexture as a document.)"
	^ fileTexture! !

!QImageDocument methodsFor: 'initialize-release' stamp: 'bgf 4/30/2009 00:45'!
fileTexture: aTFileTexture
	fileTexture := aTFileTexture.
	fileTexture ifNotNil: [  
		suffix := ((fileTexture fileName ifNil: ['']) copyAfterLast: $.).
		suffix size = 0 ifTrue: [ suffix := nil ].
		serviceURL ifNotNil: [ fileTexture serviceURL: serviceURL ].
	].
! !

!QImageDocument methodsFor: 'protocol' stamp: 'bgf 4/30/2009 10:44'!
fileTextures
	^  fileTexture ifNotNil: [ {fileTexture} ] ifNil: [ #() ]! !

!QImageDocument methodsFor: 'protocol' stamp: 'bgf 5/31/2009 20:38'!
getSharedNameFromTexture
	"Callback support for document copying...  update the shared name,
	 as the system has given the texture a new one."
	self sharedName: (self fileTexture sharedName).! !

!QImageDocument methodsFor: 'comparing' stamp: 'bgf 4/29/2009 18:52'!
hashableName
	"Name for comparison purposes"
	^ (self sha ifNil:[ '' ]), self sharedName! !

!QImageDocument methodsFor: 'protocol' stamp: 'bgf 4/30/2009 00:13'!
identifier
	^ fileTexture
		ifNotNil: [ fileTexture identifier ]
		ifNil: [ 
			self sharedName ifNotNilDo: [ :sn | sn copyAfter: $/ ]]! !

!QImageDocument methodsFor: 'accessing' stamp: 'bgf 4/30/2009 10:45'!
isBundled
	^  fileTexture ifNotNil: [fileTexture isBundled] ifNil: [ false ]! !

!QImageDocument methodsFor: 'save' stamp: 'bgf 4/30/2009 10:45'!
isDocumentInForum: forumURL
	^  fileTexture ifNotNil: [fileTexture isDocumentInForum: forumURL] ifNil: [false]
! !

!QImageDocument methodsFor: 'save' stamp: 'bgf 4/30/2009 10:45'!
isDocumentInForumStorage
	^  fileTexture ifNotNil: [fileTexture isDocumentInForumStorage] ifNil: [false]! !

!QImageDocument methodsFor: 'save' stamp: 'bgf 4/30/2009 10:46'!
isDocumentInInventory
	^ fileTexture ifNotNil: [fileTexture isDocumentInInventory] ifNil: [false]
	! !

!QImageDocument methodsFor: 'accessing' stamp: 'bgf 4/30/2009 10:38'!
isKeptLocally
	^ fileTexture ifNotNil: [ fileTexture isKeptLocally ] ifNil: [ false ]! !

!QImageDocument methodsFor: 'persist' stamp: 'bgf 4/30/2009 10:05'!
persistOn: pStream
	super persistOn: pStream.
	pStream storeObject: 'image' value: fileTexture default: nil.
	"For the server-side xml-based manifest unpacking"
	pStream storeString: 'sharedName' value: (self sharedName) default: nil.! !

!QImageDocument methodsFor: 'accessing' stamp: 'bgf 4/30/2009 20:09'!
serviceURL
	^ fileTexture ifNotNil: [ fileTexture serviceURL ] ifNil: [ super serviceURL ].! !

!QImageDocument methodsFor: 'accessing' stamp: 'bgf 4/30/2009 00:50'!
serviceURL: aSvcUrl	
	super serviceURL: aSvcUrl.

	fileTexture ifNotNil: [ fileTexture serviceURL: aSvcUrl ].! !

!QImageDocument methodsFor: 'protocol' stamp: 'bgf 4/30/2009 10:46'!
sha
	^ fileTexture ifNotNil: [fileTexture sha]! !

!QImageDocument methodsFor: 'protocol' stamp: 'bgf 4/29/2009 13:34'!
sha: sha
	fileTexture ifNotNil: [
		fileTexture sha = sha ifFalse: [
			self error: 'Inconsistent sha in image import.'.
		]].! !

!QImageDocument methodsFor: 'accessing' stamp: 'bgf 4/30/2009 10:39'!
sharedName
	"The service-relative path to this image file.
	 (This is the same as the texture fileName.)"
	^ fileTexture ifNil: [ sharedName ] ifNotNil: [ fileTexture sharedName ]! !

!QImageDocument methodsFor: 'protocol' stamp: 'bgf 4/30/2009 12:52'!
sharedName: newSharedName
	"Document protocol.  
	 The service-manager path, relative to forum (or 'absolute' for bundled and user-inventory), 
	 where this item resides on the server."
	self isBundled ifFalse: [
		sharedName := newSharedName.
		fileTexture ifNotNil: [
			fileTexture sharedName: newSharedName
		]].! !

!QImageDocument methodsFor: 'accessing' stamp: 'bgf 4/30/2009 20:04'!
shortName
	"The basic file name, used as a prompt when saving locally."
	fileTexture ifNotNil: [ ^ fileTexture shortName ].
	^ self sharedName 
			ifNotNilDo: [ : sn | sn copyAfterLast:$/ ]
			ifNil: [ 'image' ]
! !

QAbstractDocument subclass: #QSharedDocument
	instanceVariableNames: 'sha sharedName cacheLocally repoInfo'
	classVariableNames: 'DocumentTypes'
	poolDictionaries: ''
	category: 'Qwaq-Objects-Document'!

!QSharedDocument commentStamp: '<historical>' prior: 0!
Handle on a file that has been uploaded to the shared documents area
of a particular Qwaq(Forums) service, for access by all participants in a service.

The content of a shared document is an off-island resource,
typically used by local client players (i.e. for MPEG)
or for served applications (e.g. Open Office via QMS).

A shared document is displayed by an off-island application (often a remote QMS app)
that varies by the document type.   See #newEmbeddedAppForExtent:,
which creates a QEmbeddedApp of the appropriate class for the document.

The document file may or may not replicated to local caches,
depending on the type of file and how it is handled.
The 'cacheLocally' property determines whether or not the file
is copied to participant client-local caches.

When local caching is used, it is generally done to meet the needs of
the displaying application (and to smooth out startup their startup times.)
For example, the MPEG viewer runs locally for each participant
and requires an on-disk file to work from.
By contrast, QMS applications work on the server-side file,
and no local caching of the document is needed.
(Local caching is a bit expensive since it means each visitor
incurs a download of the file.)

Currently local-caching is automated to work properly with static files
(such as those MPG's) and will -NOT- track document editing.

Instance variables are documented in their accessors.!

!QSharedDocument class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QSharedDocument class methodsFor: 'inspector' stamp: 'bgf 1/28/2010 11:50'!
addPropertyRows: dialog forFrame: sdref
	"Shared document properties... add sharedName to the info,
	 since the basic labeling uses the shorter form."

	dialog addRow: (
		(QStringPropertyRow new)
			titleString: 'File';
			fetchBlock: [ : dr | dr  get: #sharedName ];
			acceptBlock: [ : v |];
			balloonHelp: 'Shared name of file.';
		      inputEnabled: false;
			extraLong;
			yourself).
	super addPropertyRows: dialog forFrame: sdref.! !

!QSharedDocument class methodsFor: 'job support' stamp: 'bgf 5/14/2010 14:04'!
copy: docRef toActiveForum: aParticipant

	"Job support protocol: transfer into a forum from inventory."

	| short newName response path me sharedName |

	((docRef get: #isRepositoryDocument) = true) ifTrue:[
		"Don't copy the contents (there is none) for a repository document"
		sharedName := 'sharedDocuments/', (docRef get: #repoUrlAsPath).
	] ifFalse:[
		me := aParticipant simplerLogin.
		short := docRef get:#shortName.
		path := 'services/', (aParticipant currentForumName), '/sharedDocuments/', me.
		QLogger logInfo: 'Copying ', short, ' to ', path.
		"Record the action using only the file type suffix, not the full name."
		QActivityCapture action: #CopyFromInventory details: (QActivityCapture fileDetails: short).
	
		newName := aParticipant documentManager 
					checkExistence: short 
					path: path
					policy: #makeUnique.
		response := aParticipant serviceMgr 
					copyFiles: (docRef get:#asServiceManagerPath)
					to: (path, '/', newName).
		response name == #ok ifFalse: [ 
			^self error: response contentString.
		].
		sharedName := 'sharedDocuments/', me, '/', newName
	].
	docRef isValidFarRef ifTrue: [
		docRef future serviceURL: (aParticipant currentForumURL).
		docRef future sharedName: sharedName.
	] ifFalse: [
		docRef isValidRef ifTrue: [ 
			docRef serviceURL: (aParticipant currentForumURL).
			docRef sharedName: sharedName.
		].
	].


! !

!QSharedDocument class methodsFor: 'job support' stamp: 'ar 4/16/2010 11:25'!
copy: docFarRef toUserInventory: aParticipant

	"Job support protocol: copy the documentfile into this UI user's private storage."
	|  short newName response path me |
	me := aParticipant serviceMgr user.
	path := '/users/', me, '/private/documents'.
	(docFarRef get: #isRepositoryDocument) ifTrue:[
		"Don't copy the contents (there is none) for a repository document"
		newName := docFarRef get: #repoUrlAsPath.
	] ifFalse:[
		short := docFarRef get:#shortName.
		QLogger logInfo: 'Copying ', short, ' to ', path.
		"Record the action using only the file type suffix, not the full name."
		QActivityCapture action: #CopyToInventory details: (QActivityCapture fileDetails: short).

		newName := aParticipant documentManager 
					checkExistence: short 
					path: path
					policy: #makeUnique.
		response := aParticipant serviceMgr 
				copyFiles: (docFarRef get:#asServiceManagerPath)
				to: (path, '/', newName).
		response name == #ok ifFalse: [ 
			self error: (response contentString).
		].
	].
	docFarRef future sharedName: (path, '/', newName).

! !

!QSharedDocument class methodsFor: 'files' stamp: 'ar 4/16/2010 14:46'!
fetch: document toFileNamed: fileName forClient: owner

	| path file extra |
	document isValidRef ifFalse: [ ^ false ].
	(path := (document get: #asServiceManagerPath)) ifNil: [ ^ false ].

	[[
		file := FileDirectory default forceNewFileNamed: fileName.
		file binary.
		CProgressBar displayProgress: 'Retrieving document ...' 
			at: owner bounds center from: 0.0 to: 1.0 
			during: [:bar|
				[owner serviceMgr get: path intoStream: file informing: bar] on: Error do:[:ex|
					"File wasn't on the server. But if the document is in a repository,
					we might be able to fetch it from there"
					(document get: #isRepositoryDocument) ifTrue:[
						self repoFetch: document toFile: file forClient: owner informing: bar.
					] ifFalse:[ex pass].
				].
			].
		file close.
		QLogger logInfo: 'Document direct to file: ', fileName.
		^  true.
		
	] on: CProgressCancel do: [ : pc |   
		[ 
			file ifNotNil: [ 
				file close.  
				FileDirectory default deleteFileNamed: fileName
			]
		] on: Error do: [ : exc |   ].
		pc pass.

	]] on: Error do: [ : ex |
		extra := ''.
		[ 
			file ifNotNil: [ 
				file close.  
				FileDirectory default deleteFileNamed: fileName
			]
		] on: Error do: [ : exc | extra := exc description ].

		QLogger logInfo: 'Cannot direct-disk fetch ', path, ': ', ex description, ' ', extra.
		^ false
	].

		! !

!QSharedDocument class methodsFor: 'files' stamp: 'bgf 3/13/2009 12:43'!
forLocalFile: file suffix: ext  sharedName: sharedName inClient: owner
	| data document |

	"Create a shared document describing the content of the given local file,
	 assuming that the document also lives on the server as a file of the given shared-name."
	data := file binary; contents.
	document := 
		(self new)
			sha: (TContentManager generateSHA: data);
			sharedName: sharedName;
			suffix: ext;
			yourself.

	"Since we have the data, stow in our cache if we want it there."
	(QDocumentType forSuffix: ext) isCachedLocally 
		ifTrue: [
			TContentManager default cacheManager 
					cache: data to: (document cacheFileName).
			document cacheLocally: true.
		].

	^document! !

!QSharedDocument class methodsFor: 'files' stamp: 'ar 4/16/2010 14:30'!
fromRepository: repoInfo docType: type session: qSession.

	"Protocol: if the file is accepted as this document type,
	create a repository document for it.
	 Otherwise returns nil (if no Document is produced.)."

	^self new repoInfo: repoInfo
! !

!QSharedDocument class methodsFor: 'files' stamp: 'ar 4/16/2010 15:14'!
getContent: document forClient:owner
	"Harness-side operation to acquire the document (file) content as a byte array."

	| cached fname file serviceMgr path data bundled |
	bundled := (document get:#isBundled).
	cached := (document get:#cacheLocally).

	"If prepackaged (service nil) or cached, try to give them the local copy."
	(bundled == true | (cached ==true)) ifTrue: [
		fname := (document get:#asLocalFileUri).
		[
			file :=  FileDirectory default readOnlyFileNamed: 
					(FileDirectory default pathFromURI: fname).
			file binary.
			^ file contentsOfEntireFile.
		] on: Error do: [
			[ file close ] on: Error do: [].
			bundled ifTrue: [
				QLogger logWarning: 'Cannot open bundled ', fname.
			]]].

	"Fallthrough - if we failed to open the cached file (or if not cached) fetch it explicitly."
	serviceMgr := owner serviceMgr.
	path := (document get:#asServiceManagerPath).
	owner statusStatus: 'Downloading ', (document get:#shortName).
	owner flowControlFlag: true.
	[[ ^ 
		CProgressBar displayProgress: 'Retrieving data ...' 
			at: owner bounds center from: 0.0 to: 1.0 
			during: [:bar|
				[data := serviceMgr get: path informing: bar] on: Error do:[:ex|
					"File wasn't on the server. But if the document is in a repository,
					we might be able to fetch it from there"
					(document get: #isRepositoryDocument) ifTrue:[
						data := ByteArray streamContents:[:s|
							self repoFetch: document toFile: s forClient: owner informing: bar.
						].
					] ifFalse:[ex pass].
				].
			].
	] on: Error do: [ :exc|
		| msg |
		data := nil.
		owner flowControlFlag:false.
		(exc messageText beginsWith: 'FileDoesNotExist') ifTrue: [
			msg := 'The original could not be found.'
		] ifFalse: [
			msg := exc messageText.
		].
		self error: ('Cannot download ',((document get:#shortName) ifNil: ['file']), ':\\', msg, '\') withCRs.
	]] ensure: [
		owner flowControlFlag:false.
	].
	^ data! !

!QSharedDocument class methodsFor: 'files' stamp: 'hrs 4/6/2009 09:57'!
mailDocument: docRef forClient: aUI
	"We can't attach docs on Mac yet, so just send a QRL to the doc."
	^aUI mailDocumentLink: docRef! !

!QSharedDocument class methodsFor: 'files' stamp: 'hrs 4/6/2009 21:27'!
openDocumentWebPage: docRef forClient: aUI
	"Open the Web page for the document."
	^aUI openQrlForDocument: docRef! !

!QSharedDocument class methodsFor: 'customs' stamp: 'jcg 6/4/2009 17:42'!
processCustomsEntryData: entryData customs: customsPane type: docType path: path
	"We attempt to get the SHA from the server instead of downloading the whole file."
	| isZip sha |
	
	".ZIP files are a special case... we always need  to look inside, 
	  so force a full download by setting the SHA to nil."
	isZip := entryData fileExtension asUppercase = 'ZIP'.
	isZip ifTrue: [QLogger logInfo: 'Customs is handling .ZIP file specially'].
	(isZip or: [docType isCachedLocally])
		ifFalse: [sha := customsPane getFileSHA: path]
		ifTrue: [sha := nil].

	sha 
		ifNil: [
			"Couldn't get the SHA (probably because SP version doesn't support it).
			 Fall back to downloading the whole file."
			^customsPane
				postProcessEntryData: entryData
				fromFile: path
				sharedName: entryData internedServicePath
				extension: entryData fileExtension asUppercase
		]
		ifNotNil: [
			"Success!!  We obtained the SHA; use it to initialize a new document,
			 and the document to create a new file-card."
			| doc |
			QLogger logInfo: 'Customs avoided full file-download; just got the SHA'.
			doc := self new
							sha: sha;
							sharedName: entryData internedServicePath;
							suffix: entryData fileExtension asUppercase;
							ctime: (QTimeStamp new
										who: entryData user
										what: 'Upload'
										utc: entryData timestamp).
			^QFileCard forDocument: doc
		].

! !

!QSharedDocument class methodsFor: 'files' stamp: 'ar 4/16/2010 15:15'!
repoFetch: docRef toFile: file forClient: owner informing: bar
	"Fetch a document from its repository."

	| repoUrl repoType repository resp repoPath |
	repoUrl := docRef get: #repoUrl.
	repoType := docRef get: #repoType.
	repoPath := docRef get: #repoPath.
	repository := (QDocumentRepository cache)
		at: repoUrl 
		ifAbsentPut:[QDocumentRepository type: repoType url: repoUrl owner: owner].
	repository ifNil:[^self error: 'Unknown repository: ', repoType].

	repository critical:[
		[resp := repository download: repoPath.
		resp isSuccess ifFalse:[^self error: resp status].
		resp streamFrom: resp contentStream to: file size: resp contentLength 
			progress:[:total :amt| total ifNotNil:[(bar value: amt / (total max: 1) asFloat)]].
		] ensure:[repository close].
	].
! !

!QSharedDocument class methodsFor: 'files' stamp: 'bgf 9/17/2009 18:53'!
saveDocument: document asFile: newName forClient:aUI
	"Harness-side protocol for saving an instance as a local file; 
	 returns the short name of the resulting file, or nil on failure
	 (inform: is issued on failure.)"
	| klass fname file data  |

	"If the given name doesn't end in appropriate suffix, supply it."
	fname := newName.
	(newName endsWith: '.', (document get:#suffix) caseSensitive:false) ifFalse: [
		fname := newName,'.', (document get:#suffix).
	].

	"Mac OS doesn't automatically categorize .mp4 with the appropriate OS-filetype.
	 Name these with m4v instead.  Prevents the Snow Leopard finder from deeming them 'text'.
	 We treat mp4 and m4v the same, but other places assume mp4 as the preferred."
	((fname endsWith: '.mp4') and: [ Smalltalk platformName = 'Mac OS' ]) ifTrue: [
		fname := (fname copyUpToLast: $.),  '.m4v'.
	].

	"This is a generic method; the specific class of the underlying document
	 may have protocols for the three cases below."
	klass := (document get: #class) ifNil: [ self ].

	"Case one: check for existing cached copy"
	(klass existingDocumentFileName: document forClient: aUI) ifNotNilDo: [ : ename | 
		ename = fname ifFalse: [
			FileDirectory default
				copyFileWithoutOverwriteConfirmationNamed: ename 
				toFileNamed: fname.
		].
		"This looks silly but gives confirmation that the copy really worked,
	 	 and gives us the local file name to return."
		(FileDirectory default readOnlyFileNamed: fname) ifNotNilDo: [ : created |
			created close.
			^ created localName
		]].

	"Case two: try for direct fetch-creation."
	(klass fetch: document toFileNamed: fname forClient: aUI) ifTrue: [ 
		"As above, silly but gives confirmation that the operation really worked,
	 	 and gives us the local file name to return."
		(FileDirectory default readOnlyFileNamed: fname) ifNotNilDo: [ : created |
			created close.
			^ created localName
		]].

	"Fallback case three - fetch the content bytes and write them directly. 
	 (This is the most general way, relying on old-fashioned service-maneger fetch-to-memory.)"
	data := self getContent: document forClient: aUI.
	data ifNotNil: [
		file := FileStream forceNewFileNamed: fname.
		file binary.
		[file nextPutAll: data] ensure: [file close].
		^ file localName
	] ifNil: [
		self error: 'The document is not available.\It may not have been fetched yet.'  withCRs
	].
	^ nil
! !

!QSharedDocument class methodsFor: 'files' stamp: 'bgf 3/14/2009 00:02'!
uploadForImport: aFile docType: type session: qSession 

	"If the file is recognized as a SharedDocument document type,
	 upload it and return the a SharedDocument handle for it.
	 Otherwise returns nil (if no QSharedDocument is produced.)"

	^ (qSession documentManager)
			uploadDocumentFile: aFile 
			collisions: #makeUnique
	! !

!QSharedDocument methodsFor: 'accessing' stamp: 'das 12/9/2008 12:00'!
asPyPanelDoc

	^ PyPanelDocCS new copyFrom: self; yourself.
! !

!QSharedDocument methodsFor: 'accessing' stamp: 'bgf 8/21/2008 20:42'!
asServerLocalFile

	"The prefixed path string that is used when launching a remote application,
 	 to identify this document file.  Prepends service:  to the sharedName of the document.

	 See XVNCAppProxy mountArgumentURL:, where the strings are decoded."
	
	"When the name doesn't match expectations, we return it un-prefixed,
	 so the server-side converters don't fire and the plain string shows up
	 as (unusable, but reported and logged) application argument.
	 We don't want to raise errors or trigger MNU's in the replicated world."

	self isBundled ifTrue: [ 
		QLogger logWarning:
					 'Cannot refer to bundled file as server-side file: ', sharedName.
		^ sharedName
	].

	sharedName ifNil: [ ^ nil].
	(sharedName beginsWith: 'sharedDocuments/')  ifFalse: [
		"Not a proper shared document."
		Transcript cr;
			show: ('Cannot use private document:\ ', sharedName, 
				    '\ in a remote application.') withCRs.
		^ sharedName
	].
		
	^ 'service:', sharedName.

! !

!QSharedDocument methodsFor: 'accessing' stamp: 'bgf 3/3/2009 02:41'!
asServiceManagerPath
	"QServiceManager get/put argument we use to refer to this document ."
	self isBundled ifTrue: [
		"Path support for moving bundled documents to the service provider.
	       E.g.	 Content/Qwaq/Movices/QwaqCroquetSmall.mpg"
		^ 'system-resources/bundled/', self sharedName
	].
	self isDocumentInForumStorage ifTrue: [
		^ '/services/', (self serviceName), '/', self sharedName.
	] ifFalse: [
		^ self sharedName
	].
	! !

!QSharedDocument methodsFor: 'accessing' stamp: 'bgf 12/3/2008 14:10'!
cacheFileName

	"The filename for local cached copies of this document;
	 if the document requires local caching."

	sha ifNil: [ ^ nil ].
	^ (sha, '.', suffix).! !

!QSharedDocument methodsFor: 'accessing' stamp: 'bgf 3/21/2007 17:16'!
cacheLocally
	"Boolean - if true, tells the system to maintain a locally cached copy of the document file."
	^ cacheLocally! !

!QSharedDocument methodsFor: 'accessing' stamp: 'bgf 3/3/2009 22:17'!
cacheLocally: bool
	"Boolean - if true, tells the system to maintain a locally cached copy of the document file."
	cacheLocally := bool.
	self island updateResource: self type: self class resourceType.
! !

!QSharedDocument methodsFor: 'protocol' stamp: 'bgf 8/13/2008 09:59'!
canBeCopiedToInventory
	"Can users do 'Copy To Inventory...?' 
	 Offer save-local for most shared documents..."
	^ sharedName notNil! !

!QSharedDocument methodsFor: 'protocol' stamp: 'bgf 2/28/2009 22:19'!
canBeOpenedPrivately
	"Can users do 'Open Private Copy...?' 
	 Offered by default for real documents"
	^ sharedName notNil and: [ self docType isQPZ not ]. ! !

!QSharedDocument methodsFor: 'protocol' stamp: 'bgf 8/13/2008 09:55'!
canBeSavedLocally
	"Can users do 'Save Local Copy...?' 
	 Offer save-local for most shared documents..."
	^ sharedName notNil! !

!QSharedDocument methodsFor: 'protocol' stamp: 'ar 5/18/2010 17:13'!
canCopyContentsOnly
	"Repository documents require additional context (the repository info)"

	^self isRepositoryDocument not
! !

!QSharedDocument methodsFor: 'persist' stamp: 'ar 4/15/2010 12:54'!
constructFrom: pStream
	super constructFrom: pStream.

	"Read repoInfo first as this will set sharedName 
	but we want sharedName from the persist stream"
	self readRepoInfoFrom: pStream.

	sha := pStream readString: 'sha' default: sha.
	self sharedName: (pStream readString: 'sharedName' default: sharedName).
	self cacheLocally: (pStream readBool: 'cacheLocally' default: cacheLocally).! !

!QSharedDocument methodsFor: 'accessing' stamp: 'das 9/10/2009 17:52'!
contentSummary
	"Short string summarizing my document file name."
	| partial |
	^ self isDocumentInForumStorage ifTrue: [
		partial := (self sharedName copyAfter:$/).	"Elide 'sharedDocuments/'"
		(partial includesSubString:'qpz') ifTrue:[partial := (partial copyUpToLast:$.).]. "Elide suffix"
		(partial includesSubString:'tpz') ifTrue:[partial := (partial copyUpToLast:$.).]. "Elide suffix"
		((partial indexOf: $/ ifAbsent: [0]) = 33 and: [
		  ('abcdef0123456789' includesAllOf: (partial first:32)) ])
		ifTrue: [
			"Aha - 32-bit hex path - must be an OID-as-hex path"
			partial copyAfter: $/
		] ifFalse: [
			partial
		]
	] ifFalse: [
		self sharedName ifNotNilDo: [ :sn | sn copyAfterLast:$/ ] ifNil: [ '' ]   "Elide the storage path."
	]
! !

!QSharedDocument methodsFor: 'initialize-release' stamp: 'ar 4/15/2010 12:49'!
copyFrom: aDocument
	"Copies the QSharedDocument - not the underlying document."

	sha := aDocument sha clone.
	sharedName := aDocument sharedName clone.
	cacheLocally := aDocument cacheLocally.
	repoInfo := aDocument repoInfo copy.
	
	"At tail, so resource registration is done once at the end."
	super copyFrom: aDocument.
! !

!QSharedDocument methodsFor: 'save' stamp: 'bgf 3/28/2007 12:27'!
doSaveAsForAvatar: anAvatarReplica

	"Save the document locally in the client of the current avatar."
	QSaveDocumentJob save: self forUser: anAvatarReplica! !

!QSharedDocument methodsFor: 'registry properties' stamp: 'ar 5/3/2010 11:53'!
fileCardBacking
	"Use the Sharepoint picture for repository-backed documents"

	self isRepositoryDocument ifTrue:[^'SharePointDocCard.png'].
	^super fileCardBacking! !

!QSharedDocument methodsFor: 'registry properties' stamp: 'bgf 1/2/2008 18:30'!
fileModeSelector
	"Stream mode when writing the contents of the document; performed on a FileStream."
	^ #binary! !

!QSharedDocument methodsFor: 'protocol' stamp: 'bgf 8/2/2008 12:52'!
fileName
	"For compatibility with TFileTexture, etc, for resource capture.
	 This is the server-side file uri for the resource, relative to the containing forum directory"
	^ self sharedName! !

!QSharedDocument methodsFor: 'repository' stamp: 'ar 4/16/2010 14:37'!
fullySupportsRepositories
	"Repositories are fully supported for TMS applications"

	^#(
		QRemoteOpenOfficeApp 
		QRemoteFirefox
	) includes: self docType appFactoryClassName! !

!QSharedDocument methodsFor: 'comparing' stamp: 'bgf 4/23/2007 13:39'!
hashableName
	"Name for comparison purposes"
	^ (sha ifNil: [ '' ]), self sharedName! !

!QSharedDocument methodsFor: 'accessing' stamp: 'hrs 4/7/2009 16:45'!
identifier
	"See QFileCard>>identifier"
	^(self sharedName ifNil:[''] ifNotNilDo: [:n| n copyAfter:$/.	"Elide 'sharedDocuments/'"])
		ifEmpty: [self contentSummary].
! !

!QSharedDocument methodsFor: 'initialize-release' stamp: 'bgf 12/3/2008 14:19'!
initialize
	super initialize.

	"The SHA will be generated when content is first assigned."
	sha := nil.
	cacheLocally := false.
	! !

!QSharedDocument methodsFor: 'accessing' stamp: 'bgf 1/5/2010 14:13'!
isBundled
	"Trick to recognize Qwaq prepackaged Content files,
	  which are never downloaded or cached."
	^ self sharedName ifNotNilDo: [ : sn |
		(sn beginsWith: 'Content/') or: [ sn beginsWith: 'objects-menu/' ]
	] ifNil: [ 
		false 
	]

! !

!QSharedDocument methodsFor: 'save' stamp: 'bgf 3/5/2009 21:38'!
isDocumentInForum: forumURL
	"Is the document already available as a resource for this forum (stored in the forum's storage) ?"
	^ self isBundled 
		or: [
			(self isDocumentInInventory not)
			and: [ self serviceURL = forumURL ]].! !

!QSharedDocument methodsFor: 'save' stamp: 'bgf 12/7/2008 13:49'!
isDocumentInForumStorage

	"Is the document in the shared area of a forum's server-side storage?"
	^ self sharedName notNil and: [self sharedName beginsWith: 'sharedDocuments/'].! !

!QSharedDocument methodsFor: 'save' stamp: 'bgf 12/7/2008 13:49'!
isDocumentInInventory
	"Is the document in a user's inventory storage? (as opposed to forum shared-doc storage)"
	^ self sharedName notNil and: [self sharedName beginsWith: '/users/'].! !

!QSharedDocument methodsFor: 'repository' stamp: 'ar 4/15/2010 12:24'!
isRepositoryDocument
	"Repository support. Answers if this document is stored in a repository.
	If so, the document is expected to implement the QRepositoryDocument
	protocol for retrieving repository information."

	^repoInfo notNil! !

!QSharedDocument methodsFor: 'protocol' stamp: 'ar 4/16/2010 11:49'!
kindSummary
	"Prefix the kind of repository if we have one"

	^(self isRepositoryDocument 
		ifTrue:[self repoType asString capitalized, ' ']
		ifFalse:['']), super kindSummary! !

!QSharedDocument methodsFor: 'persist' stamp: 'ar 4/15/2010 12:53'!
persistOn: pStream
	super persistOn: pStream.

	pStream storeString: 'sha' value: sha default:nil.
	pStream storeString: 'sharedName' value: sharedName default: nil.
	pStream storeBool: 'cacheLocally' value: cacheLocally default: false.

	self storeRepoInfoOn: pStream.! !

!QSharedDocument methodsFor: 'repository' stamp: 'ar 4/15/2010 12:52'!
readRepoInfoFrom: pStream
	"Read repository info from a persist stream"

	| repoType repoUrl repoPath |
	repoType := pStream readString: 'repoType' default: self repoType.
	repoUrl := pStream readString: 'repoUrl' default: self repoType.
	repoPath:= pStream readString: 'repoPath' default: self repoType.

	repoType ifNotNil:[
		self repoType: repoType url: repoUrl path: repoPath.
	].! !

!QSharedDocument methodsFor: 'repository' stamp: 'ar 4/15/2010 12:50'!
repoInfo
	"The repostiry information associated with the document"

	^repoInfo! !

!QSharedDocument methodsFor: 'repository' stamp: 'ar 4/16/2010 14:31'!
repoInfo: aQRepositoryInfo
	"The respository info for this document"

	repoInfo := aQRepositoryInfo! !

!QSharedDocument methodsFor: 'repository' stamp: 'ar 4/15/2010 12:25'!
repoPath
	"The document name and path inside the repository.
	Depending on repository type, the string may include a path.
	For example, a valid SharePoint document path is 'TeleDocs/sample.pdf'"

	^repoInfo ifNotNil:[repoInfo repoPath]! !

!QSharedDocument methodsFor: 'repository' stamp: 'ar 4/15/2010 12:25'!
repoType
	"The type of repository. For a list of all known repositories check 
		QDocumentRepository allSubclasses collect: [:cls| cls repoType].
	"

	^repoInfo ifNotNil:[repoInfo repoType]! !

!QSharedDocument methodsFor: 'repository' stamp: 'ar 4/16/2010 11:25'!
repoType: repoType url: repoUrl path: repoPath
	"Initializes the document from repo url and doc path."
	repoInfo := QRepositoryInfo new.
	repoInfo repoType: repoType.
	repoInfo repoUrl: repoUrl.
	repoInfo repoPath: repoPath.

	"Construct the shared name. The sharedName is created by
	using the repoUrl and repoPath under sharedDocuments/
	for example:
		repoUrl: http://win2008ad3.qwaq.com/DocTest
		repoPath: TeleDocs/sample.pdf
		sharedName: sharedDocuments/win2008ad3.qwaq.com/DocTest/TeleDocs/sample.pdf
	"
	sharedName := 'sharedDocuments/', self repoUrlAsPath
! !

!QSharedDocument methodsFor: 'repository' stamp: 'ar 4/15/2010 12:25'!
repoUrl
	"The URL associated with the document repository. 
	Depending on repository type, the url string may include a path as well.
	For example, a valid SharePoint url is 'http://win2008ad3.qwaq.com/DocTest'"

	^repoInfo ifNotNil:[repoInfo repoUrl]! !

!QSharedDocument methodsFor: 'repository' stamp: 'ar 4/16/2010 11:24'!
repoUrlAsPath
	"Answers the repository url as a path that can be used for storing the document."

	| path |
	path := self repoUrl  ifNil:[^nil].
	path := path copyFrom: (path findString: '://')+3 to: path size.
	(path beginsWith: '/') ifTrue:[path := path allButFirst].
	(path endsWith: '/') ifTrue:[path := path allButLast].
	^path, '/', self repoPath! !

!QSharedDocument methodsFor: 'repository' stamp: 'ar 4/16/2010 12:21'!
requiresCheckin
	"Answer if the document needs to be checked back in after use.
	For now we assume that for documents in Open Office."

	^self embeddedAppClassName = #QRemoteOpenOfficeApp
! !

!QSharedDocument methodsFor: 'repository' stamp: 'ar 4/16/2010 12:24'!
requiresLocking
	"Answer if the document needs to be locked when in use.
	For now, we assume that all documents requiring check-in
	should be locked during usage."

	^self requiresCheckin! !

!QSharedDocument methodsFor: 'accessing' stamp: 'bgf 4/30/2009 00:06'!
sha
	"The secure-hash identifying this document. 
	 (Possibly nil for some sorts of non-cached docs.)"
	^ sha! !

!QSharedDocument methodsFor: 'accessing' stamp: 'bgf 3/22/2007 15:40'!
sha: anSHA

	"The SHA identifies the cached local copy of this document;
 	 it is computed at upload (and eventually may be updated when a document is saved.)"
	sha := anSHA.
	self cacheLocally: self cacheLocally "kick the resource registration"! !

!QSharedDocument methodsFor: 'accessing' stamp: 'bgf 7/3/2007 23:11'!
sharedName
	"The URI path relative to the service's router-side storage root
	 under which this file is kept...
	 usually a relative path of the form 'sharedDocuments/[user]/baseFile.mpeg'

	 For documents in inventory, the name is /users/<user>/..... '
	 For special-case preloaded content (like bundled demo movies)
	 this will be a path into <default>/Content/..."
	^ sharedName! !

!QSharedDocument methodsFor: 'accessing' stamp: 'bgf 7/3/2007 22:39'!
sharedName: aString

	"The path relative to the service's router-side storage root
	 under which this file is kept.
	 One of three cases can be expected:

	 - Relative name starting with 'sharedDocuments/...'   
	   refers to a document in this forum's sharedDocuments area.
	 
	-  Absolute name starting with /users/...  
	   refers   to a document in a user's subdirectory in realm/users.
	   These are in inventory - cannot (by construction) be opened in shared apps.

	- Special Case: paths starting with 'Content/' are BUNDLED;
	  these live in the software installation and are never downloaded
 	  or copied to cache.  Used for seeding the QwaqCroquetMovie mpegFile
	  in forum templates."

	sharedName := aString! !

!QSharedDocument methodsFor: 'accessing' stamp: 'bgf 2/15/2009 02:04'!
shortName
	"The basic file name, used as a prompt when saving locally."
	^ self sharedName ifNotNilDo: [ : sn | sn copyAfterLast:$/ ]
! !

!QSharedDocument methodsFor: 'protocol' stamp: 'ar 4/16/2011 17:45'!
sourceUrl

	"For documents that have a 'source' outside of OpenQwaq, this method must
	return the document source. For example, for repository documents the
	sourceUrl is the url of the document in its repository. Documents that have
	no source associated with them, should return nil."

	^repoInfo ifNotNil:[self repoUrl, '/', self repoPath]! !

!QSharedDocument methodsFor: 'repository' stamp: 'ar 4/15/2010 12:53'!
storeRepoInfoOn: pStream
	"Store repository info from a persist stream"
	repoInfo ifNotNil:[
		pStream storeString: 'repoType' value: self repoType default: nil.
		pStream storeString: 'repoUrl' value: self repoUrl default: nil.
		pStream storeString: 'repoPath' value: self repoPath default: nil.
	].! !

QSharedDocument subclass: #QTextDocument
	instanceVariableNames: 'text'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Objects-Document'!

!QTextDocument commentStamp: '<historical>' prior: 0!
A text document is presented as a shared document,
but with its content also available immediately in the island.
(The shared-document backing store is updated
by application save, but is only used for offline examination.)!

!QTextDocument class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QTextDocument class methodsFor: 'inspector' stamp: 'bgf 1/28/2010 11:50'!
addPropertyRows: dialog forFrame: tdref
	"A means to change the text document name."
	dialog addRow: (
		(QStringPropertyRow new)
			titleString: 'Name';
			fetchBlock: [ : ref  | ref get: #shortName ];
			acceptBlock: [ : v  : ref | ref future shortName: v ];
			balloonHelp: 'Name of the document file.';
			extraLong;
			yourself).! !

!QTextDocument class methodsFor: 'instance creation' stamp: 'bgf 12/3/2008 15:11'!
defaultSuffix
	^ 'txt'! !

!QTextDocument class methodsFor: 'files' stamp: 'bgf 3/13/2009 03:23'!
forLocalFile: file suffix: ext  sharedName: sharedName inClient: owner

	"Create a text document describing the content of the given local file,
	 assuming that the document also lives on the server as a file of the given shared-name."
	| data |
	file ascii.
	file converter: Latin1TextConverter new.
	file wantsLineEndConversion: true.
	data := file contents.
	^ QTextDocument new
		shortName: file localName;
		text: data;
		yourself.! !

!QTextDocument class methodsFor: 'instance creation' stamp: 'bgf 2/19/2007 13:44'!
forString: aString
	^ (self new)
		 text: aString;
		  yourself! !

!QTextDocument class methodsFor: 'files' stamp: 'bgf 3/1/2009 14:51'!
getContent: document forClient: aUI

	"Shared document protocol - harness-side request for the file content of this document.
	 If this is a fat in-island document, we have the text right here..."
	| |
	document isValidFarRef ifTrue: [
		(document get: #text) ifNotNilDo: [ : txt | ^ txt ]
	].

	"Fallback to finding the content from the service provider."
	(super getContent: document forClient: aUI) ifNotNil: [ : data |
		^ (MultiByteBinaryOrTextStream on: (String new: (data size))) 	
			ascii;
			converter: (Latin1TextConverter new);
			nextPutAll: data;
			reset;
			contentsOfEntireFile
		].

	^ nil

		
		! !

!QTextDocument class methodsFor: 'files' stamp: 'hrs 4/6/2009 09:58'!
mailDocument: docRef forClient: aUI
	"Populates the users mail client with the content text."
	^aUI mailDocument: docRef class: self! !

!QTextDocument class methodsFor: 'files' stamp: 'hrs 4/6/2009 21:28'!
openDocumentWebPage: docRef forClient: aUI
	"Open the Web page for the document."
	^aUI openQrlForDocument: docRef! !

!QTextDocument class methodsFor: 'customs' stamp: 'jcg 6/2/2009 12:44'!
processCustomsEntryData: entryData customs: customsPane type: docType path: path
	"We need the whole file; bounce the responsibility back to the 'customsPane'."
	^customsPane
		postProcessEntryData: entryData
		fromFile: path
		sharedName: entryData internedServicePath
		extension: entryData fileExtension asUppercase! !

!QTextDocument class methodsFor: 'services' stamp: 'bgf 2/28/2009 21:02'!
saveDocument: document asFile: newName forClient:owner

	"Harness-side protocol for saving an instance as a local file; 
	 returns the short name of the resulting file, or nil on failure
	 (inform: is issued on failure.)"

	| fname file text |
	fname := newName.

	"Usability trick until the new file dialog smoke clears - 
	 if the result doesn't end in .txt, supply it."
	(newName endsWith: '.txt' caseSensitive:false) ifFalse: [
		fname := newName,'.txt'.
	].
	text := self getContent: document forClient: owner.
	text ifNil: [ ^ nil ].
		
	file := CrLfFileStream forceNewFileNamed: fname.
	file converter: Latin1TextConverter new.
	file ascii.
	
	[file nextPutAll:text ] ensure: [file close].

	^ file localName
! !

!QTextDocument class methodsFor: 'files' stamp: 'bgf 11/17/2009 18:20'!
uploadForImport: aFile docType: type session: qSession 
	"The documents  primary existence is within the island.
	 Its content gets refreshed as part of app-save later.
	 So for full visibility, we both upload and island-ize the content."
	| docname  file text doc |

	doc := super uploadForImport: aFile docType: type session: qSession.
	docname := aFile name.
	aFile close. "Windows has rules, must close the file before reopening as CrLF."
	file :=  FileStream readOnlyFileNamed: docname.
	file ifNil: [self error: 'Cannot open text file ', docname].
	[
		file converter: Latin1TextConverter new.
		file wantsLineEndConversion: true.
		text := file contents.
	] ensure: [
		file close.
	].
	doc text: text.
	^ doc.! !

!QTextDocument methodsFor: 'protocol' stamp: 'bgf 2/28/2009 16:59'!
canBeOpenedPrivately
	^ text notNil or: [ sharedName notNil ]! !

!QTextDocument methodsFor: 'persist' stamp: 'bgf 3/3/2010 15:29'!
constructFrom: pStream
	super constructFrom: pStream.
	self text: (pStream readString: 'text' default: nil).

	sharedName ifNil: [
		"Up-convert older text documents (pre 1.2.48)"
		| short | 
		short := pStream readString: 'shortName' default: 'Text Document.txt'.
		sharedName := 
			self makeSharedNameFrom: 'sharedDocuments/text/', short.
	] ifNotNil: [
		"Ensure I don't name-collide with any other registered text-document,
		 since I know I'm a different instance.  This allows duplicateScene: to work properly."
		pStream preserveDocumentNames ifFalse: [
			| sn |
			sn := sharedName.
			sharedName := nil.		"Hide myself from the registry scan..."
			sharedName := (self makeSharedNameFrom: sn).
		].
	].
! !

!QTextDocument methodsFor: 'accessing' stamp: 'bgf 3/3/2009 03:41'!
contentSummary
	| result |
	"Elide the 'text/' we use in place of creator-name."
	result := (super contentSummary) ifNil: [ '' ].
	(result beginsWith: 'text/') ifTrue: [ result := result copyAfter: $/ ].
	^ result.! !

!QTextDocument methodsFor: 'copying' stamp: 'bgf 2/15/2009 00:05'!
copyFrom: aTextDocument

	super copyFrom: aTextDocument.
	text := aTextDocument text.
! !

!QTextDocument methodsFor: 'registry properties' stamp: 'bgf 3/5/2009 14:34'!
fileCardBacking
	^ 'TextCard.png'! !

!QTextDocument methodsFor: 'accessing' stamp: 'bgf 1/2/2008 18:29'!
fileModeSelector
	"Either binary or ascii ... how to prepare a file when writing the contents."
	^ #ascii! !

!QTextDocument methodsFor: 'initialize-release' stamp: 'bgf 2/14/2009 23:33'!
initialize
	suffix := 'txt'.
	super initialize.

! !

!QTextDocument methodsFor: 'initialize-release' stamp: 'bgf 2/15/2009 02:37'!
makeSharedNameFrom: subPath
	| docs result |
	"Subpath is the shared-document path, eg. 'sharedDocuments/text/foobar.txt' .
	 Returns a unique-ified variant (by splicing in a number)
	 that is different from that of any registered document in this island."

	"If this is already my shared name, keep it."
	(self sharedName notNil and: [ self sharedName endsWith: '/', subPath ]) ifTrue: [
		^ self sharedName ].
	result := subPath.
	(self island isKindOf: TIsland) ifTrue: [
		docs := self island documentResources.
		[ docs anySatisfy: [ : doc | doc sharedName = result ]  ] 
		 whileTrue: [
			result := QDocumentManager incrementName: result
		 ]].
	^ result
! !

!QTextDocument methodsFor: 'persist' stamp: 'bgf 2/24/2009 19:29'!
manifestPersistString
	"Generate the persistString of the document, but eliding content-payload.
	 I.e. should write just the metadata."
	| save |
	save := text.	
	[
		text := nil.
		^ super manifestPersistString.
	] ensure: [
		text := save.
	].
! !

!QTextDocument methodsFor: 'persist' stamp: 'bgf 2/14/2009 23:16'!
persistOn: pStream
	super persistOn: pStream.
	pStream storeString: 'text' value: (text ifNotNil: [text asString]) default: nil.

	"Down-convert so older versions can read (pre 1.2.48)"
	pStream storeString: 'shortName' value: (self shortName) default: nil.
! !

!QTextDocument methodsFor: 'island-copy' stamp: 'bgf 2/15/2009 02:08'!
postImportFromIslandCopier: importer
	| orig |

	"This is how we assign new names to documents inbound from the harness,
	 for client drops of the Text Document c3x, mainly."

	orig := self sharedName.
	sharedName := nil.		"So its-already-mine doesn't kick in"
	self sharedName: (self makeSharedNameFrom: orig).

	super postImportFromIslandCopier: importer.! !

!QTextDocument methodsFor: 'protocol' stamp: 'bgf 3/5/2009 12:35'!
prepareForDropIntoSession: aSession serviceURL: aForumUrl
	"Preparing to drop an offline document.
	 Offline text documents (from manifest) have no text inside. 	
	 Online ones contain their text. So, like, get the text and put it in."

	self text: (self class getContent: self forClient: (aSession owner)).! !

!QTextDocument methodsFor: 'accessing' stamp: 'bgf 3/9/2009 02:30'!
saveComplete: shortName

	"Relay from document-save to document-rename.
	 Matches the legacy behavior where this was the only way to rename;
	 we'll leave it in place until document Properties are better exposed through menus."
	shortName ifNotNil: [
		self shortName: shortName ].! !

!QTextDocument methodsFor: 'accessing' stamp: 'bgf 7/31/2009 01:43'!
shortName: aName
	"The most recent filename the user associated with this document;
	 This is used for information within the space;
	 it serves as a hint for saving outside the space.
	 For text documents, we renamed the shared resource when this is changed."
	| candidate |

	(aName = self shortName) ifTrue: [ ^ self ].

	(self sharedName isNil or: [ self shortName isEmpty ]) ifTrue: [
		candidate :=  'sharedDocuments/text/', aName.
	] ifFalse: [
		candidate := (self sharedName copyFrom: 1 to: (self sharedName size - self shortName size)), 
							aName
	].
	self sharedName: (self makeSharedNameFrom: candidate).
	self signal: #shortNameChanged.		"For display sheets"
	self signal: #frameChanged with: self.	"For generic property/tool update"

! !

!QTextDocument methodsFor: 'accessing' stamp: 'bgf 2/19/2007 19:04'!
text
	"The String containing the content of this document."
	^ text! !

!QTextDocument methodsFor: 'accessing' stamp: 'bgf 12/3/2008 21:02'!
text: aString

	(aString ~= text) ifTrue: [
		text := aString.
		self signal: #textChanged with: text.
	].! !

TObject subclass: #QAttachmentManager
	instanceVariableNames: 'space attachments'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Objects-Manipulation'!

!QAttachmentManager commentStamp: '<historical>' prior: 0!
Manages relative-position attachments among TFrames.  

Attachments are situations where one frame (the sticker) 
is positioned relative to another frame (the surface)
such that when the surface moves the sticker follows.

A given sticker can be attached to at most one surface at a time.

In the manipulation system,
an attachment is created when a moveable object
(something manipulated by a DragAdapter with isMovable=true)
is dragged over a surface for movement (without a logical-drop at the end.)

Objects can establish attachments via
	self root attach: self to: surface

(In general, public access to the attachment manager should go through
the space methods in category 'attaching').

Once attached, the sticker is transformed (via localTransform: updates)
when the attachment surface is transformed.

Attachments are a service provided by the space with notifications.
When an object is attached to a surface (or detached) 
by a drag&drop operation,
or by any api call to root attach:to:   or root detach:
it receives the message:
	attachedTo: surfaceOrNil
and can update internally to reflect this.
!

!QAttachmentManager class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QAttachmentManager methodsFor: 'accessing' stamp: 'bgf 6/30/2010 11:26'!
attach: sticker to: surface
	"Create a transform-following attachment making sticker follow surface.
	 This is the basic entry point to make attachments.
	 The sticker is informed of the relationship change via the protocol
		sticker attachedTo: surface.

	 BOTH TFrames should be explicit parts of the persistent scene!!  "

	| loopCheck |
	loopCheck := surface ifNotNil: [ surface surfaceAttachment ].
	[ loopCheck ] whileNotNil: [
		(loopCheck isOrContains: sticker) ifTrue: [ 
				QLogger logWarning: 'Rejecting circular attachment ', sticker, ' to ', surface.
				^ nil.
		].
		loopCheck := loopCheck surfaceAttachment.
	].
		
	(attachments includesKey: sticker) ifTrue: [
		(attachments at: sticker) manager: self; surface: surface.
	] ifFalse: [
		surface ifNotNil: [
			| record |
			record := QAttachmentRecord forSticker: sticker surface: surface.
			attachments at: sticker put: record.
			record manager: self.
		].
	].
! !

!QAttachmentManager methodsFor: 'private' stamp: 'bgf 7/31/2009 00:56'!
attachment: arecord referenceDestroyed: frame
	"Notice from an attachment record that the target surface has been destroyed;	
 	 remove the attachment record and destroy it."
	(attachments at: (arecord sticker) ifAbsent:[ nil]) = arecord ifTrue: [
		attachments removeKey: arecord sticker.
	].
	arecord destroy.! !

!QAttachmentManager methodsFor: 'persist' stamp: 'bgf 8/1/2009 16:03'!
constructContentsFrom: pStream
	| records |
	records := (pStream readSequence: 'attachments' default: #()).
	records do: [ :record |
		attachments at: (record sticker) put: record.
		"Defer activating the attachments.  
		 This lets their transforms be up-to-date before we start following."
		record future manager: self.
	].! !

!QAttachmentManager methodsFor: 'accessing' stamp: 'bgf 7/31/2009 03:08'!
detach: aStickerFrame
	"If the given frame is attached, undo that."
	self attach: aStickerFrame to: nil! !

!QAttachmentManager methodsFor: 'initialize' stamp: 'bgf 7/30/2009 23:09'!
initialize
	super initialize.
	attachments := Dictionary new.! !

!QAttachmentManager methodsFor: 'accessing' stamp: 'bgf 8/16/2009 03:16'!
isFrame: frame attachedOn: surface
	"Is the given frame attached, directly or indirectly, to the given surface ? "
	^ (self surfaceAttachmentOf: frame)
		ifNotNilDo: [ : sfc |
			sfc = surface or: [ self isFrame: sfc attachedOn: surface ] 
		] ifNil: [ 
			false ]! !

!QAttachmentManager methodsFor: 'persist' stamp: 'bgf 8/16/2009 03:00'!
persistContentsOn: pStream
	| tmp |
	"A bit of a safety check so we don't persist any frames that are in the midst of slow deletion, etc;
	 only persist attachments between rooted stickers and persistent surfaces."

	tmp := attachments select: [ : a | a sticker root = space
								  and: [ a sticker shouldPersist ]
								  and: [ a surface notNil ]
								  and: [ a surface root = space ]
								  and: [ a surface shouldPersist ]].

	pStream storeSequence: 'attachments' value: (tmp values) default: #().
! !

!QAttachmentManager methodsFor: 'persist' stamp: 'bgf 8/1/2009 14:49'!
persistNotes: arrayOfNotes for: rootSurface on: pStream as: tagName
	| tmp |

	"For capture of the notes attached to a note collector when persisting it
	 (especially for isolating persistence for inventory or template.c3z work."

	tmp := attachments select: [ : a | 
			(arrayOfNotes includes: (a sticker))					  
			and: [ (a surface = rootSurface) or: [ arrayOfNotes includes: (a surface)] ]].

	pStream storeSequence: tagName  value: (tmp values) default: #().
! !

!QAttachmentManager methodsFor: 'persist' stamp: 'bgf 8/1/2009 16:53'!
recoverNotesFrom: attachmentArray  for: rootSurface
	"For capture of the notes attached to a note collector when persisting it
	 (especially for isolating persistence for inventory or template.c3z work.
	 Finds any notes in the array of attachment records that are not yet rooted,
	 and for each of those,
	 		adds it to the space,
			applies the stored relative-transform to it
			activates the attachment in the space."
	| adds |
	adds := OrderedCollection new.
	attachmentArray do: [ : record |
		record sticker root ifNil: [
			self space addChild: (record sticker).
			adds add: record.
		].		
	].
	adds do: [ : record |
		"Boot the sticker into its new position in this new stack..."
		record surfaceFrameChange: (record surface).
		"And activate the attachment ..."
		attachments at: (record sticker) put: record.
		record manager: self.
	].! !

!QAttachmentManager methodsFor: 'accessing' stamp: 'bgf 7/31/2009 02:43'!
removeAttachment: a
	"If a's sticker  is attached to anything, undo that attachment, removing the attachment-record."
	 
	(attachments at: (a sticker) ifAbsent:[ nil]) = a ifTrue: [
		attachments removeKey: a sticker.
	].
	a destroy.! !

!QAttachmentManager methodsFor: 'accessing' stamp: 'bgf 7/31/2009 00:34'!
space 
	"The QSpace I manage attachments for."
	^ space ifNil: [ (self island resourcesAt: TSpace resourceType) first ]
	! !

!QAttachmentManager methodsFor: 'initialize' stamp: 'bgf 7/31/2009 02:43'!
space: aSpace	
	"We remember this here simply as a backpointer, since I'm not a TFrame in the scene graph."
	space := aSpace.
	! !

!QAttachmentManager methodsFor: 'accessing' stamp: 'bgf 5/3/2010 17:08'!
stickersAttachedTo: aSurface
	"All the frames directly  attached to the surface."
	^ ((((attachments select: [ : a | a surface = aSurface ]) 
		values 
			collect: [ : a | a sticker ])
			asArray)
				sortBy: [ : s1 : s2 | s1 hash < s2 hash ])
					asArray

		"The sort encourages consistently ordered results across clients,
	 	 which is not guaranteed by the underlying set representations inside Dictionary."! !

!QAttachmentManager methodsFor: 'accessing' stamp: 'bgf 5/3/2010 17:14'!
stickersOn: aSurface
	"All the frames that are directly attached to the surface,
	 or to some other sticker that is on the surface."	

	^ (( ((self stickersAttachedTo: aSurface)  asOrderedCollection)
		inject: (OrderedCollection new)
		into: [  : allStickers : sticker |
				(allStickers includes: sticker) ifFalse: [ allStickers addLast: sticker ].
				(self stickersOn: sticker) do: [ : os | 
					(allStickers includes: os) ifFalse: [  allStickers addLast: os ].
				].
				allStickers
			   ]) asArray
				sortBy: [ : s1 : s2 | s1 hash < s2 hash ]) 
					asArray

		"The sort encourages consistently ordered results across clients,
	 	 which is not guaranteed by the underlying set representations inside Dictionary."
! !

!QAttachmentManager methodsFor: 'accessing' stamp: 'bgf 7/31/2009 02:47'!
surfaceAttachmentOf: someFrame
	"Attachment manager query - what surface if any is the given frame attached to?"
	^ attachments at: someFrame ifPresent: [ : a | a surface ]! !

TObject subclass: #QAttachmentRecord
	instanceVariableNames: 'sticker surface transform manager'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Objects-Manipulation'!

!QAttachmentRecord commentStamp: '<historical>' prior: 0!
Record of a relative-position attachment from one TFrames to another.

The attached frame is the 'sticker'.
The surface frame to which it is bound is the 'surface'.
 
Attachments are situations where one frame (the sticker) 
is positioned relative to another frame (the surface)
such that when the surface moves the sticker follows.

See QAttachmentManager / QSpace for the methods to manage attachments;
this class is private implementation for that service.
!

!QAttachmentRecord class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QAttachmentRecord class methodsFor: 'as yet unclassified' stamp: 'bgf 7/30/2009 23:07'!
forSticker: frame1 surface: frame2
	^ self basicNew forSticker: frame1 surface: frame2! !

!QAttachmentRecord methodsFor: 'persist' stamp: 'bgf 8/1/2009 14:30'!
constructFrom: pStream
	
	sticker := (pStream readObject: 'sticker' default: nil).
	surface := (pStream readObject: 'on' default: nil).
	transform :=  pStream readMatrix4x4: 'relation' default: transform.! !

!QAttachmentRecord methodsFor: 'initialize' stamp: 'bgf 7/31/2009 14:44'!
destroy
	sticker := nil.
	manager := nil.
	self stopAllScripts.
! !

!QAttachmentRecord methodsFor: 'initialize' stamp: 'bgf 8/1/2009 14:18'!
forSticker: frame1 surface: frame2
	self initialize.
	sticker := frame1.
	surface := frame2.
! !

!QAttachmentRecord methodsFor: 'initialize' stamp: 'bgf 8/1/2009 16:41'!
manager: attachManager
	"Install the record as a managed attachment.
	 This sets up the listeners to remove the attachment when its parts are deleted."
	manager := attachManager.
	manager ifNotNil: [ 
		| sfc |
		self startScript: #stickerDestroyed: when: { sticker . #destroyed }.
		self startScript: #stickerDestroyed: when: { sticker . #deleted }.

		"Complete the relationship."
		sfc := surface.
		surface := nil.
		self surface: sfc.
	].! !

!QAttachmentRecord methodsFor: 'persist' stamp: 'bgf 8/1/2009 14:30'!
persistOn: pStream

	pStream storeObject: 'sticker' value: sticker.
	pStream storeObject: 'on' value: surface default: nil.
	pStream storeMatrix4x4: 'relation' value: transform default: nil
! !

!QAttachmentRecord methodsFor: 'accessing' stamp: 'bgf 7/31/2009 03:13'!
sticker
	"The frame that I attach to the surface... I move the sticker when the sticker moves."
	^ sticker! !

!QAttachmentRecord methodsFor: 'private' stamp: 'bgf 7/31/2009 14:43'!
stickerDestroyed: sticker
	"When the sticker goes, the attachment goes also;
	 the following-relationship is released and the attachment record is removed from the manager."
	manager ifNotNil: [
		manager removeAttachment: self
	]! !

!QAttachmentRecord methodsFor: 'accessing' stamp: 'bgf 7/30/2009 23:26'!
surface
	^ surface! !

!QAttachmentRecord methodsFor: 'accessing' stamp: 'bgf 8/29/2009 12:50'!
surface: aTFrame
	"Update my surface attachment.
	 This is the workhorse method of the attachment manager. "
	
	(sticker isNil or: [manager isNil]) ifTrue: [
		QLogger logWarning: 'Surface update to stale attachment: ', aTFrame.
		^ self ].	"Destroyed."

	"Allow no loops."
	aTFrame ifNotNil: [
	 	(aTFrame isOrContains: sticker) ifTrue: [ 
			QLogger logWarning: 'Rejecting attachment to child: ', sticker.
			^ self surface: nil ].
		(sticker isOrContains: aTFrame) ifTrue: [ 
			QLogger logWarning: 'Rejecting attachment to parent: ', sticker.
			^ self surface: nil ].
	].
	(manager surfaceAttachmentOf: aTFrame) ifNotNilDo: [ : sfc |
		(sfc isOrContains: sticker) ifTrue: [
			QLogger logWarning: 'Detaching ', aTFrame, ' to avoid loop with ', sfc.
			manager detach: aTFrame
		]].
 
	surface = aTFrame ifFalse: [
		self stopScript: #surfaceFrameChange:.
		sticker attachedTo: aTFrame.
		surface := aTFrame.
		surface ifNotNil: [
			self startScript: #surfaceFrameChange: when: { surface . #frameChanged }.
			self startScript: #surfaceDestroyed: when: { surface . #destroyed }.
			].
		].

	surface ifNotNil: [
		"Only use the relative xform if the surface is unscaled, since it is rubbish under scaling."
		surface cumulativeScale = 1 ifTrue: [
			transform := surface relativeTransform: sticker
		] ifFalse: [
			transform := nil
		]].

	surface ifNil: [
		manager removeAttachment: self ]
! !

!QAttachmentRecord methodsFor: 'private' stamp: 'bgf 7/31/2009 03:00'!
surfaceDestroyed: aFrame
	"When the surface goes, the attachment goes also;
	 the following-relationship is released and the attachment record is removed from the manager."
	aFrame = surface ifTrue: [
		self surface: nil
	].! !

!QAttachmentRecord methodsFor: 'private' stamp: 'bgf 8/29/2009 12:47'!
surfaceFrameChange: referenceFrame
	"Adapt to changes in the reference frame."
	sticker ifNotNil: [
		(referenceFrame = surface and: [ transform notNil ]) ifTrue: [
			sticker localTransform: (surface  globalTransform * transform).
		]].! !

TObject subclass: #QButtonHandler
	instanceVariableNames: 'body target action overColor downColor fireColor readyToFire label'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Objects-General'!

!QButtonHandler commentStamp: '<historical>' prior: 0!
A very simple button-behavior to turn any TFrame into a pushbutton
that triggers an action when clicked.

A QButtonHandler is not a TFrame, and has no appearance of its own.
It is an accessory that adds button behaviors to any existing frame.

QButtonHandler supplies mouse over/down/up hiliting, and firing behavior.
A mouse down-up sequence within it causes the firing of its action.

Hiliting feedback does not alter the body's materials or coloring;
visual feedback is all done through transient hilite.

The button must be set up with a target and action (selector) in order to fire.
The target is any object in the button's island.
The action is a method selector of zero or one argument.

When the action is fired, if it takes an argument,
the CroquetEvent leading to the action is passed.
This is usually a pointerUp event.  From the event,
one can determine the avatar (hence the user session)
responsible for the action.!

!QButtonHandler class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QButtonHandler class methodsFor: 'instance creation' stamp: 'bgf 1/19/2007 16:53'!
body: aTFrame target: anObject action: aSelector

	"Wrap a handler around aTFrame, turning it into a button;
	 when the button fires,  'anObject aSelector' is performed." 

	^ (self new) 
		body: aTFrame;
		target: anObject;
		action: aSelector;
		yourself.! !

!QButtonHandler class methodsFor: 'instance creation' stamp: 'bgf 1/19/2007 16:08'!
body: aTFrame target: anObject action: aSelector label: aMessageString

	"Wrap a handler around aTFrame, turning it into a button;
	 when the button fires,  'anObject aSelector' is performed.
	 The label is used for mouse-over help information (in the status line.)" 

	^ (self new) 
		body: aTFrame;
		target: anObject;
		action: aSelector;
		label: aMessageString;
		yourself.! !

!QButtonHandler methodsFor: 'accessing' stamp: 'bgf 1/19/2007 16:25'!
action
	"The selector performed on the target when the button fires."
	^action! !

!QButtonHandler methodsFor: 'accessing' stamp: 'bgf 8/5/2006 18:05'!
action: aSelector

	"The action is a method selector that will be sent to my target.
	  The selector may take 0 or 1 arguments.  If it takes one argument,
	  then it will be performed with the CroquetEvent that triggered the action
	  (usually a pointer up event.)"

	action := aSelector asSymbol! !

!QButtonHandler methodsFor: 'accessing' stamp: 'bgf 1/19/2007 16:25'!
body
	"The body is the TFrame to which I add pushbutton behavior."
	^body! !

!QButtonHandler methodsFor: 'accessing' stamp: 'bgf 1/19/2010 16:09'!
body: aTFrame

	"The body is the TFrame to which I add pushbutton behavior."
	body := aTFrame.
	{#pointerEnter . #pointerOver . #pointerDown .  #pointerUp . #pointerLeave } 
	do: [ :sel |
		self stopScript: (sel,':')asSymbol.
		body ifNotNil: [
			self runScript: (sel,':')asSymbol when: { body . sel }
		]
	].
	body ifNotNil: [ body doesClickSelect: false ]! !

!QButtonHandler methodsFor: 'accessing' stamp: 'bgf 8/5/2006 19:51'!
canFire
	"True if the button is configured with a usable target+action."
	^ (target notNil and: [ action notNil ]) and:[target respondsTo: action].! !

!QButtonHandler methodsFor: 'accessing' stamp: 'bgf 1/19/2007 16:26'!
downColor
	"The button's children are hilit with this color when the mouse is pressed on any part of it.
	 The default is QwaqStyle controlDownColor."
	^ downColor! !

!QButtonHandler methodsFor: 'accessing' stamp: 'bgf 8/18/2009 02:21'!
downColor: dc
	"The body is hilit with this color when the mouse is pressed on any part of it.
	 The default is QwaqStyle controlDownColor."
	downColor := dc ifNotNil: [ dc asVectorColor ]! !

!QButtonHandler methodsFor: 'accessing' stamp: 'bgf 1/19/2007 16:26'!
fireColor
	"When fired, the button is hilit with this color for a brief period.
	 The default is Color green."
	^ fireColor! !

!QButtonHandler methodsFor: 'accessing' stamp: 'bgf 1/19/2007 17:19'!
fireColor: fc
	"When fired, the body is hilit with this color for a brief period.
	 The default is Color green."
	fireColor := fc asVectorColor! !

!QButtonHandler methodsFor: 'initialize' stamp: 'bgf 1/21/2007 13:23'!
initialize

	super initialize.
	readyToFire := false.
	downColor := QColors controlDownColor.
	overColor := QColors controlOverColor.
	fireColor := Color green.
! !

!QButtonHandler methodsFor: 'accessing' stamp: 'bgf 1/19/2007 14:30'!
label
	"Mouseover status label"
	^ label! !

!QButtonHandler methodsFor: 'accessing' stamp: 'bgf 12/20/2007 23:32'!
label: aString
	"Mouse-over floating hint (#floatMessage to the client)"
	label := aString! !

!QButtonHandler methodsFor: 'accessing' stamp: 'bgf 1/19/2007 16:26'!
overColor
	"The button's children are hilit with this color when the mouse is over any part of it.
	 The default is QwaqStyle controlOverColor."
	^ overColor! !

!QButtonHandler methodsFor: 'accessing' stamp: 'bgf 1/19/2007 17:19'!
overColor: oc
	"The body is hilit with this color when the mouse is over any part of it.
	 The default is QwaqStyle controlOverColor."
	overColor := oc asVectorColor! !

!QButtonHandler methodsFor: 'events' stamp: 'bgf 8/20/2009 21:08'!
pointerDown: event

	"We don't fire on down; rather, we fire on up.
	 readyToFire is false if we're not fireable, and is set false when the button leaves,
	 so it is only true when we're in the midst of a click on a fireable button."

	readyToFire := self canFire.
	readyToFire ifTrue: [
		body hilite: downColor.
	].
	(event nrClicks = 2) ifTrue: [
		body signal: #doubleClick with: event.
		readyToFire := false.
	].! !

!QButtonHandler methodsFor: 'events' stamp: 'bgf 1/28/2008 21:35'!
pointerEnter: event

	self pointerOver: event! !

!QButtonHandler methodsFor: 'events' stamp: 'bgf 1/19/2007 15:51'!
pointerLeave: event
	
	readyToFire := false.
	body unhilite.! !

!QButtonHandler methodsFor: 'events' stamp: 'bgf 11/13/2008 12:43'!
pointerOver: event
	event avatar ifNil: [ ^ self ].
	readyToFire := false.
	self canFire ifTrue: [
		body hilite: overColor.
	].
	label ifNotNil: [
		event avatar showFloatMessage: label forEvent: event.
	].! !

!QButtonHandler methodsFor: 'events' stamp: 'bgf 8/20/2009 21:35'!
pointerUp: event
	
	readyToFire ifFalse: [
		body unhilite.
		^ self.
	].

	fireColor ifNotNil: [
		body hilite: fireColor for: 300.
	] ifNil: [
		body future unhilite.
	].

	(action numArgs = 0) ifTrue: [
		target future perform: action.
	] ifFalse: [
		target future perform: action with: event.
	].

	readyToFire := false.  "Get ready for the next mousedown."! !

!QButtonHandler methodsFor: 'initialize' stamp: 'bgf 1/19/2007 16:28'!
release

	"Stop handling events for the body."
	self body: nil.! !

!QButtonHandler methodsFor: 'accessing' stamp: 'bgf 1/19/2007 16:25'!
target
	"This is the object that will be sent the 'action' message when I fire."
	^target! !

!QButtonHandler methodsFor: 'accessing' stamp: 'bgf 8/5/2006 18:05'!
target: t
	"This is the object that will be sent the 'action' message when I fire."
	target := t.! !

TObject subclass: #QChatLog
	instanceVariableNames: 'messages msgsToSave users'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Objects-Space'!

!QChatLog commentStamp: '<historical>' prior: 0!
The chat log is a persistent collection of all chat messages,
ordered by submission date.   This is an island-ready object;
the original use was for the text-chat occurring in a space.

Chat logs are far-ref registered; however, all messages and user records
held in a chat log are pass-by-copy, so fetching the text of a chat log 
for UI display is fairly fast, and using that data is unencumbered by far ref indirection.

The common use pattern for island/off-island use is to create a 'shadow' 
chat log off-island, use that to drive the display of an editor,
and use signal observation to reflect changes in the on-island master.

See for instance getRecentMessags: and initializeForMessages:
which support creation of such shadows.!

!QChatLog class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QChatLog class methodsFor: 'time' stamp: 'bgf 10/23/2007 11:25'!
clientUtcTime
	"Timestamp generator - named to reinforce that is a client-side harness responsibility
	 to generate the stamp."
	^ (DateAndTime now asUTCSeconds)! !

!QChatLog class methodsFor: 'instance creation' stamp: 'bgf 8/27/2008 10:28'!
fromMessages: arrayOfMessages

	^ (self basicNew)
		initializeForMessages: arrayOfMessages! !

!QChatLog class methodsFor: 'time' stamp: 'bgf 8/31/2008 14:37'!
localDateString: aChatMessage

	"The local date string for the timestamp of the given message, if it appears valid; otherwise nil.
	 Older chat logs contain unusable island-time timestamps, which we try not to interpret."

	| date |
	aChatMessage timestamp ifNil: [ ^ nil ].
	(aChatMessage timestamp > 0) ifTrue: [
		date := DateAndTime fromUTCSeconds: aChatMessage timestamp.
		(date year > 2006) ifTrue: [
			^ QStyles dateStringFromUTC: (aChatMessage timestamp).
		].
	].

	^nil! !

!QChatLog class methodsFor: 'defaults' stamp: 'ar 4/16/2011 17:41'!
maxMessages
	"The largest number of old messages that will be persisted.
	 We cap these to prevent infinite growth of the forum XML,
	 the chat message records can easily outbulk the rest of the scene in forums,
	 especially with all the automatic chat records we insert."

	^ 400! !

!QChatLog class methodsFor: 'defaults' stamp: 'bgf 8/28/2008 17:28'!
oldSysUserID
	"Pre-coal Magic userID (no collisions with OIds or email-logins) 
	  of the system-monitor pseudo-user.  This is legacy... used to recognize during internalization
	  for conversion to the newer sysUserID.   
	  (Part of converting chat from userID-oid-identity to userLogin as the identity.) "

	^ 'Forum Host'! !

!QChatLog class methodsFor: 'defaults' stamp: 'bgf 8/28/2008 17:29'!
sysUserID
	"Magic user-login value (no collisions with old user-ID oids or with email-logins) 
	 that identifies system-generated chat messages."

	^ '<*Forum Host*>'! !

!QChatLog class methodsFor: 'defaults' stamp: 'ar 4/16/2011 17:09'!
sysUserName
	"Pseudo screen-name of the system-monitor pseudo-user."
	^ 'Forum Host'! !

!QChatLog methodsFor: 'accessing' stamp: 'hrs 2/9/2009 08:20'!
addCopyMessage: chatMsg
	"Add a copy of the given message to my message list;
	 this for the local off-island copy used by the QChatPane,
	 which copies the original in, to ensure that we have such a user in our user list."
	self addMessage: (chatMsg message)
		fromUser: (self userForLogin: chatMsg user userLogin 
					    screenName: chatMsg user userName)
		atTime: chatMsg timestamp
		place: chatMsg place! !

!QChatLog methodsFor: 'accessing' stamp: 'bgf 8/29/2008 16:44'!
addMessage: txt fromAvatar: anAvatar atTime: utcSeconds
	"Handle a single text message event ."

	| user msg |

	"Arbitrary cap on message size for hack protection."
	msg := (txt ifNil:['']) contractTo:1000.
	user := self userForAvatar: anAvatar.
	self addMessage: msg fromUser: user atTime: utcSeconds.! !

!QChatLog methodsFor: 'accessing' stamp: 'hrs 2/8/2009 12:11'!
addMessage: txt fromAvatar: anAvatar atTime: utcSeconds place: place
	"Handle a single text message event ."

	| user msg |

	"Arbitrary cap on message size for hack protection."
	msg := (txt ifNil:['']) contractTo:1000.
	user := self userForAvatar: anAvatar.
	self addMessage: msg fromUser: user atTime: utcSeconds place: place.! !

!QChatLog methodsFor: 'accessing' stamp: 'bgf 8/29/2008 16:43'!
addMessage: txt fromAvatarRef: anAvatar atTime: t

	"Handle a single text message event - this is the CLIENT-SIDE version,
	 for the local copy used by the QChatPane."

	| user msg |

	"Arbitrary hack-protection - no huge messages."
	msg := txt ifNotNil: [txt contractTo:1000].
	user := self userForAvatarRef: anAvatar.
	self addMessage: msg fromUser: user atTime: t.! !

!QChatLog methodsFor: 'accessing' stamp: 'bgf 9/2/2008 02:39'!
addMessage: txt fromUser: cuser atTime: utcSeconds
	"Handle a single text message event ."
	"The client side version uses a nil msg for a user-is-typing ... indicator.
	 We allow at most one such nil-text message per user,
	 and update its timestamp.  Also, when a user types a real message,
	 the nil message is reused for the final text, and moved to the bottom."
	| pendingMsg |
	pendingMsg := messages detect: [ :m | m user = cuser and: [ m message isNil ]] ifNone: [nil].
	pendingMsg ifNotNil: [ 
		pendingMsg message: txt.
		txt ifNotNil: [
			messages remove: pendingMsg.
			messages add: pendingMsg.
		].
		pendingMsg timestamp: utcSeconds.
		self signal: #messageAdded with: pendingMsg

	] ifNil: [
		messages add: 
			((QChatMessage new)
				message: txt;
				user: (self userForUser: cuser);
				timestamp: utcSeconds;
				yourself).
		self signal: #messageAdded with: messages last
	].! !

!QChatLog methodsFor: 'accessing' stamp: 'hrs 2/8/2009 12:10'!
addMessage: txt fromUser: cuser atTime: utcSeconds place: place
	"Handle a single text message event ."
	"The client side version uses a nil msg for a user-is-typing ... indicator.
	 We allow at most one such nil-text message per user,
	 and update its timestamp.  Also, when a user types a real message,
	 the nil message is reused for the final text, and moved to the bottom."
	| pendingMsg |
	pendingMsg := messages detect: [ :m | m user = cuser and: [ m message isNil ]] ifNone: [nil].
	pendingMsg ifNotNil: [ 
		pendingMsg message: txt.
		txt ifNotNil: [
			messages remove: pendingMsg.
			messages add: pendingMsg.
		].
		pendingMsg timestamp: utcSeconds.
		self signal: #messageAdded with: pendingMsg

	] ifNil: [
		messages add: 
			((QChatMessage new)
				message: txt;
				user: (self userForUser: cuser);
				timestamp: utcSeconds;
				place: place
				yourself).
		self signal: #messageAdded with: messages last
	].! !

!QChatLog methodsFor: 'private' stamp: 'hrs 2/9/2009 10:50'!
addPlace: msg for: converter to: aStream
	"On-island stuff like us and our messages do not know about QRLs (and even if they did, we
	wouldn't really want to repeatedly copy the same string to lots of messages). And so
	if we want to prefix message place names with a qrl, we have to be told what prefix to use."

	msg place ifNotNilDo: [:place | 
		aStream nextPut: $ . 
		converter ifNotNil: [aStream nextPutAll: (converter value: place)]
			ifNil: [aStream nextPutAll: place].
		].! !

!QChatLog methodsFor: 'persist' stamp: 'bgf 4/1/2007 13:10'!
constructFrom: pStream

	"This is just to remind me that while QSpace persists its log inline,
	 the XML form is also used to readScene/storeScene for client transfers."
	self constructLogFrom: pStream.
	! !

!QChatLog methodsFor: 'persist' stamp: 'bgf 8/15/2007 16:31'!
constructLogFrom: pStream

	"Designed to write all this into the containing object's pStream without so much nesting,
	 and so we don't wire a QChatLog object into QSpace XML;
	 what is persisted there are an object map and a sequence of message entries,
 	 which could be interpreted differently by a future restructure."
	users := pStream readObjectMap: 'chatUsers' default: users.
	messages addAll: (pStream readSequence: 'chatMessages' default:#()).

	"Don't hang on to stale user entries from old unoptimized chat logs."
	self gcUsers.

! !

!QChatLog methodsFor: 'accessing' stamp: 'bgf 9/2/2008 02:41'!
gcUsers
	"Discard all QChatUsers that don't have messages in the transcript."
	| used |
	used := Set new.
	messages do: [  :msg| used add: msg user ].
	users associations copy do: [ :assoc| 
		(used includes: (assoc value)) ifFalse: [ users removeKey: assoc key ].
	].

	"And make sure all the message-users come from the dictionary."
	messages do: [ :msg | 
		msg user: (self userForUser: msg user)
	].! !

!QChatLog methodsFor: 'accessing' stamp: 'das 8/5/2009 13:37'!
getLog

	"return an XML form of the chat log. This is specifically for Python apps."

	^ String streamContents:[:stream |
		stream nextPutAll: '<chatLog>', String cr.
		messages do:[:m |
			stream nextPutAll: m getChatMessage, String cr.
		].
		stream nextPutAll: '</chatLog>'.
	].! !

!QChatLog methodsFor: 'accessing' stamp: 'bgf 8/29/2008 17:01'!
getRecentMessages: n

	"Return the n most recent chat messages 
	 (which will in turn contain references to the pertinent users.

	 Reminder - however this code changes  it must remain side-effect free on the island,	
	 suitable for get: retrieval."

	^ (messages last: (n min: messages size)) asArray! !

!QChatLog methodsFor: 'initialize-release' stamp: 'bgf 8/29/2008 22:47'!
initialize

	super initialize.
	Processor activeIsland register: self.	
	self registerGlobal: #ChatTranscript.
	
	msgsToSave := self class maxMessages.
	messages := OrderedCollection new.
	users := Dictionary new.! !

!QChatLog methodsFor: 'initialize-release' stamp: 'bgf 8/28/2008 14:59'!
initializeForMessages: arrayOfMessages

	super initialize.
	messages := arrayOfMessages asOrderedCollection.
	msgsToSave := self class maxMessages.
	users := Dictionary new.
	messages do: [ : msg |
		users at: (msg user userLogin) put: (msg user).
	].! !

!QChatLog methodsFor: 'accessing' stamp: 'bgf 4/1/2007 01:18'!
messages

	"The ordered collection of messages."
	^ messages! !

!QChatLog methodsFor: 'persist' stamp: 'bgf 8/15/2007 16:24'!
persistLogOn: pStream
	"Designed to write all this into the containing object's pStream without so much nesting,
	 and so we don't wire a QChatLog object into QSpace XML;
	 what is persisted there are an object map and a sequence of message entries,
 	 which could be interpreted differently by a future restructure."
	| msgs |

	self gcUsers.
	pStream storeObjectMap: 'chatUsers' value: users default: #().

	msgs := (msgsToSave <=0) 
			ifTrue: [ messages ] 
			ifFalse: [messages copyLast: 
					  (msgsToSave min: messages size) ].
	pStream storeSequence: 'chatMessages' value: msgs default: #().
	msgsToSave := self class maxMessages.	"Clear the cutoff marker, which is a per-call trick."
! !

!QChatLog methodsFor: 'persist' stamp: 'bgf 8/15/2007 16:24'!
persistOn: pStream

	"This is just to remind me that while QSpace persists its log inline,
	 the XML form is also used to readScene/storeScene for client transfers."

	self persistLogOn:pStream! !

!QChatLog methodsFor: 'accessing' stamp: 'bgf 8/31/2008 16:42'!
realMessageCount
	"How many messages from real users 
		(rather than system-announcement arrival/deparature ones)
	 do I contain?"

	^ messages count: [ : m | m user isSysUser not ]! !

!QChatLog methodsFor: 'accessing' stamp: 'bgf 7/4/2007 17:08'!
recentLogAsXmlScene: n
	"Return the XML for the last n messages."

	"We have to stash a temporary cutoff value since we use the whole scene-writer
	 infrastructure to store the user-message cross-references...
	 (so there's no path to pass the cutoff into the storeScene:... framework.)
	 The msgsToSave value is used in low-level persistOn:,
	 just for this one operation."
	msgsToSave := n.
	[
		^ TSceneWriter new storeScene: self.
	] ensure: [ 
		msgsToSave := self class maxMessages.
	].
! !

!QChatLog methodsFor: 'initialize-release' stamp: 'bgf 8/29/2008 23:04'!
reset
	"Empty the chat log.   Call this via QSpace resetChat to ensure proper signaling."
	msgsToSave := self class maxMessages.
	messages := OrderedCollection new.
	users := Dictionary new.

	self signal: #resetChat.! !

!QChatLog methodsFor: 'accessing' stamp: 'bgf 6/22/2009 17:58'!
userForAvatar: anAvatar
	"Find or generate the QChatUser record for this avatar-replica.  (If nil, returns the sys-user)"
	| login sn |

	login := anAvatar 
		ifNotNil: [ (anAvatar userLogin) ifNil: [ anAvatar sessionHex ] ].
	sn := anAvatar 
		ifNotNil: [ (anAvatar publicName) ].

	^ self userForLogin: login screenName: sn.
! !

!QChatLog methodsFor: 'accessing' stamp: 'bgf 6/22/2009 17:58'!
userForAvatarRef: anAvatarRef
	"Find or generate the QChatUser record for this avatar.  (If nil, returns the sys-user)"
	| login sn |

	login := anAvatarRef isValidFarRef
		ifTrue: [ (anAvatarRef get:#userLogin) ifNil: [ anAvatarRef get:#sessionHex ] ].
	sn := anAvatarRef isValidFarRef
		ifTrue: [ (anAvatarRef get:#publicName) ].

	^ self userForLogin: login screenName: sn.
! !

!QChatLog methodsFor: 'accessing' stamp: 'bgf 8/29/2008 16:34'!
userForLogin: login screenName: sn

	"Find or generate the QChatUser record for the given login name;
	 if the login name is nil, return the sys-user"
	| user |
	login ifNotNil: [
		user := users at: login ifAbsent: [
				users at: login put: (QChatUser new userLogin: login; yourself).
			].
		sn ifNotNil: [
			user userName: (sn copy).
		].
	] ifNil: [
		user := users at: (self class sysUserID) ifAbsent: [
					users at: (self class sysUserID) 
						    put: ((QChatUser new userLogin: (self class sysUserID))
									userName: (self class sysUserName);
									yourself).
				].
	].
	^ user! !

!QChatLog methodsFor: 'accessing' stamp: 'das 6/23/2009 21:45'!
userForUser: cChatUser

	"Find or generate my own listed QChatUser record for the given login name;
	 if the login name is nil, return the sys-user"
	"If it is a Python chat, then cChatUser is a string"
	cChatUser isString ifTrue:[^ self userForLogin:cChatUser screenName:cChatUser].
	^ self userForLogin: cChatUser userLogin screenName: cChatUser userName! !

!QChatLog methodsFor: 'accessing' stamp: 'hrs 3/10/2009 11:42'!
writeTranscriptOn: aStream

	"Write a simple ascii transcript of the text messages.
	 Uses simple cr line-endings; stream should convert to local convention."

	"If we change the placeConverter arg to show QRLs, it will need guarding by hasQRLSupport."
	self writeTranscriptOn: aStream placeConverter: nil.! !

!QChatLog methodsFor: 'accessing' stamp: 'hrs 2/9/2009 09:47'!
writeTranscriptOn: aStream placeConverter: unaryBlock

	"Write a simple ascii transcript of the text messages.
	 Uses simple cr line-endings; stream should convert to local convention.
	
       The unaryBlock is nil or a block that can converts a message place into a full, bullet-proofed url 
	 so that external applications can have a link to bring the user to the place of interest."
	| date lastWasHost |

	lastWasHost := true.
	messages do: [ :msg |

		date := QChatLog localDateString: msg.
		msg user isSysUser ifTrue: [
			"e.g., Arrival/departure line."

			lastWasHost ifFalse: [ aStream cr ].
			aStream nextPutAll: (msg message).
			self addPlace: msg for: unaryBlock to: aStream.
			date ifNotNil: [
				aStream nextPutAll: ' (', date, ')'.
			].
			aStream cr.
			lastWasHost := true.

		] ifFalse: [
			aStream cr.
			aStream nextPutAll: 'Comment By '.
			aStream nextPutAll: (msg user userName ifNil: ['Unknown User']).
			date ifNotNil: [
				aStream nextPutAll: ' (', date, ')'.
			].
			aStream cr.
			aStream nextPutAll: (msg message ifNil: ['']).
			self addPlace: msg for: unaryBlock to: aStream.
			aStream cr.
			lastWasHost := false.
		].
	].! !

TObject subclass: #QClickHandler
	instanceVariableNames: 'body readyToFire'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Objects-General'!

!QClickHandler commentStamp: '<historical>' prior: 0!
A very simple adapter to watch for pointerDown:/pointerMove:/pointerUp:,
and convert pointerUp: to a pointerClick signal when there is no intervening move.

Also filters double-clicks and raises doubleClick on the body (the target) when they happen.!

!QClickHandler class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QClickHandler class methodsFor: 'instance creation' stamp: 'bgf 8/1/2009 21:51'!
forFrame: aTFrame 
	"Install a handler in aTFrame, 
	 so it receives pointerClick when clicked."
	^ (self new) 
		body: aTFrame;
		yourself.! !

!QClickHandler methodsFor: 'accessing' stamp: 'bgf 8/1/2009 21:48'!
body
	"The body is the TFrame to which I add pushbutton behavior."
	^body! !

!QClickHandler methodsFor: 'accessing' stamp: 'bgf 8/1/2009 22:04'!
body: aTFrame
	"The body is the TFrame to which I add click detection behavior."
	body := aTFrame.
	{#pointerOver . #pointerDown  . #pointerUp . #pointerLeave } 
	do: [ :sel |
		self stopScript: (sel,':')asSymbol.
		body ifNotNil: [
			self runScript: (sel,':')asSymbol when: { body . sel }
		]
	].! !

!QClickHandler methodsFor: 'events' stamp: 'bgf 8/23/2009 15:52'!
isPotentialClick
	"True in the time period just after a pointer down,
	  before time has reached past the recognition interval for a simple click."
	^ readyToFire notNil 
	   and: [
		(self island time - readyToFire) abs < 280 ]! !

!QClickHandler methodsFor: 'events' stamp: 'bgf 8/23/2009 15:37'!
pointerDown: event
	"We don't fire on down; rather, we fire on up.
	 readyToFire is false if we're not fireable, and is set false when the button leaves,
	 so it is only true when we're in the midst of a click on a fireable button.
	 Double-clicks are filtered, 
	 and raised separately by signaling doubleClick (arg the event) on the target"
		
	(event nrClicks = 2) ifTrue: [
		body signal: #doubleClick with: event.
		readyToFire := nil.
	] ifFalse: [
		readyToFire := self island time.
	].! !

!QClickHandler methodsFor: 'events' stamp: 'bgf 8/1/2009 22:02'!
pointerLeave: event
	readyToFire := nil.! !

!QClickHandler methodsFor: 'events' stamp: 'bgf 8/1/2009 22:03'!
pointerOver: event
	readyToFire := nil.
! !

!QClickHandler methodsFor: 'events' stamp: 'bgf 1/5/2010 14:46'!
pointerUp: event
	"If still inside the time threshold, fire a click."
	(event avatar notNil and: [ self isPotentialClick ] )  
	 ifTrue: [
		body signal: #pointerClick with: event.
	].
	readyToFire := nil.
! !

!QClickHandler methodsFor: 'initialize' stamp: 'bgf 8/1/2009 21:51'!
release
	"Stop handling events for the body."
	self body: nil.! !

TObject subclass: #QCollection
	instanceVariableNames: 'items'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Objects-General'!

!QCollection commentStamp: '<historical>' prior: 0!
A Q-Collection is a persistable "ordered set " of TObjects :
an ordered collection, without duplicates,
of objects that can be stored/recovered in TPersist streams.

Signals self itemsChanged whenever the item-collection changes.

The items accessor always returns a copy;
use my methods 'add:', 'remove:', removeAll, to change the collection.
!

!QCollection class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QCollection class methodsFor: 'instance creation' stamp: 'bgf 7/20/2008 16:02'!
empty
	"Alias for readability; make a new empty QCollection.
	 Useful when specifying a default in persistOn: implementations."
	^ self new! !

!QCollection class methodsFor: 'as yet unclassified' stamp: 'bgf 1/13/2010 14:05'!
withAll: aCollection
	^ (self empty)
		addAll: (aCollection);
		yourself! !

!QCollection methodsFor: 'comparing' stamp: 'bgf 7/20/2008 15:59'!
= other
	"Do a collection-equality check for equals."
	^ (other class == self class)
	    and: [ other items hasEqualElements: (self items) ].! !

!QCollection methodsFor: 'accessing' stamp: 'bgf 2/15/2010 03:11'!
add: anItem
	"Add the item if absent.  Returns boolean, true if the item is new and was added."
	(items identityIncludes: anItem) ifFalse: [
		items add: anItem.
		self signal: #itemsChanged.
		^ true
	] ifTrue: [
		^ false
	].! !

!QCollection methodsFor: 'accessing' stamp: 'bgf 2/15/2010 03:11'!
addAll: newItems
	"Add the items if absent."
	| changed |

	changed := false.
	newItems do: [ : anItem | 
		(items identityIncludes: anItem) ifFalse: [
			items add: anItem.
			changed := true.
		].
	].
	changed ifTrue: [ self signal: #itemsChanged ].! !

!QCollection methodsFor: 'proxy' stamp: 'bgf 9/2/2008 00:44'!
apply: selector
	"Return the result of sending the given selector to each item,
	 as an Array, for far-ref convenience.
	 (Like collect:, except using a selector symbol, so the operation can be future-sent.)"

	^ (items collect: [ : item | item perform: selector ]) asArray! !

!QCollection methodsFor: 'accessing' stamp: 'bgf 12/5/2008 15:32'!
at: index
	^ items at: index! !

!QCollection methodsFor: 'persist' stamp: 'bgf 7/20/2008 16:01'!
constructFrom: pStream
	"My purpose in life: persist the items."
	(pStream readSequence: 'items' default: #()) do: [ : item | self add: item ].
	! !

!QCollection methodsFor: 'accessing' stamp: 'bgf 12/4/2008 23:29'!
count
	"The number of things in the collection. Alias for size."
	^ items size! !

!QCollection methodsFor: 'proxy' stamp: 'bgf 7/20/2008 12:35'!
do: aBlock
	"Convenience.  For other collection operations like select: , use 'qc items select:, etc...
	 My items will always return an Array."
	^ items do: aBlock! !

!QCollection methodsFor: 'comparing' stamp: 'bgf 7/20/2008 13:18'!
hash
	"Hash is defined because = is defined"
	^ self class hash bitXor: (items hash)! !

!QCollection methodsFor: 'initialize' stamp: 'bgf 7/20/2008 11:00'!
initialize
	items := OrderedCollection new.! !

!QCollection methodsFor: 'accessing' stamp: 'bgf 7/20/2008 12:36'!
items
	"The current item collection... as an array for ease of far-ref access.
	 And not to be modified directly (won't affect me.)"
	^ items asArray! !

!QCollection methodsFor: 'persist' stamp: 'bgf 7/20/2008 18:54'!
persistOn: pStream
	"My purpose in life: persist the items."
	pStream storeSequence: 'items' value: items default: (OrderedCollection new).! !

!QCollection methodsFor: 'printing' stamp: 'bgf 7/20/2008 12:12'!
printOn: aStream

	"For easier finding in inspectors of allInstances, mostly...
	 print the number of elements as part of the default printOn"
	super printOn: aStream.
	aStream nextPutAll: ' - '.
	items size printOn: aStream.
	! !

!QCollection methodsFor: 'accessing' stamp: 'bgf 2/15/2010 03:12'!
remove: aNote
	"Remove the item if present.  Returns true if a change was made."
	(items identityIncludes: aNote) ifTrue: [
		items remove: aNote.
		self signal: #itemsChanged.
		^ true
	].
	^ false! !

!QCollection methodsFor: 'accessing' stamp: 'bgf 7/20/2008 11:05'!
removeAll
	"Empty the collection."
	items ifNotEmpty: [
		items := OrderedCollection new.
		self signal: #itemsChanged
	].! !

!QCollection methodsFor: 'accessing' stamp: 'bgf 12/4/2008 23:28'!
size
	"The number of things in the collection."
	^ items size! !

TObject subclass: #QCustoms
	instanceVariableNames: 'imports lastEntryServerTimestamp signalee'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Objects-Space'!

!QCustoms commentStamp: 'hrs 2/15/2009 18:31' prior: 0!
A holding pen for items that have been asynchronously imported to the forum (i.e., not by an interactive user dragging them in, but by someone or something sending or uploading them to the server whilst maybe no one was in the forum).

Entries start out as strings (just pointers on the server to existing files), but get replaced with filecards when someone wants to do something with them.  The list is replicated in the space so that:
        The list of entries remains consistent as different people add and remove items.
        Conversion of a string to a filecard only has to happen once.
        Identity of underlying documents is preserved (between users and between any in-space 3D rolodexes of the same data).
!

!QCustoms class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QCustoms methodsFor: 'actions' stamp: 'hrs 3/1/2009 14:23'!
add: aQCustomsEntryData
	"Add an import to customs. serverTimestamp was the time on the server that it was added there."
	| aString serverTimestamp |
	aString := aQCustomsEntryData name.
	self release: aString.  "New entry replaces any previous imports."
	imports addLast: aQCustomsEntryData.
	serverTimestamp := aQCustomsEntryData timestamp.
	lastEntryServerTimestamp := serverTimestamp.
	signalee signal: #newImport with: aQCustomsEntryData. ! !

!QCustoms methodsFor: 'persist' stamp: 'hrs 3/1/2009 20:16'!
constructFrom: pStream
	| new |
	super constructFrom: pStream.
	imports := (pStream readSequence: 'imports' default: imports) asOrderedCollection.

	"backward compatability:"
	(imports anySatisfy: [:any | any isKindOf: Collection]) ifTrue: [
		imports := imports collect: [:each | 
			(each isKindOf: Collection) ifTrue: [
				new := QCustomsEntryData new name: each first time: each second place: ''.
				each size = 3 ifTrue: [new fileCard: each third].
				new
			] ifFalse: [each]
		]
	].
	
	lastEntryServerTimestamp := (pStream readNumber: 'time' default: 0).! !

!QCustoms methodsFor: 'accessing' stamp: 'hrs 2/16/2009 18:16'!
imports
	"A list of resources that have been asynchronously imported.
	The value may be strings or filecards. See #postProcess:."
	^imports asArray  "pass by copy to harness"! !

!QCustoms methodsFor: 'initialize' stamp: 'hrs 2/16/2009 18:28'!
initialize
	super initialize.
	imports := OrderedCollection new.! !

!QCustoms methodsFor: 'accessing' stamp: 'hrs 3/3/2009 08:59'!
lastEntryServerTimestamp
	"For the import we know about, answer the utcSeconds of the server at the time it was 
	added there. This is stored separate from any data within imports, because those can
	be deleted by a user and we still want to keep track of the last timestamp."
	^lastEntryServerTimestamp ifNil: [0]! !

!QCustoms methodsFor: 'accessing' stamp: 'hrs 3/2/2009 11:11'!
lastEntryServerTimestamp: utcSeconds
	lastEntryServerTimestamp := utcSeconds! !

!QCustoms methodsFor: 'persist' stamp: 'hrs 2/16/2009 13:45'!
persistOn: pStream
	super persistOn: pStream.
	pStream storeSequence: 'imports' value: imports default: nil.
	pStream storeNumber: 'time' value: lastEntryServerTimestamp default: 0.! !

!QCustoms methodsFor: 'actions' stamp: 'hrs 3/1/2009 11:22'!
release: name
	"entry is removed from customs."

	"lastEntryServerTimestamp is uneffected."

	|  isRemoved |
	isRemoved := false.
	imports removeAllSuchThat: [:data | 
		name = data name ifTrue:[isRemoved := true] ifFalse: [false].
		].
	isRemoved ifTrue: [signalee signal: #releasedImport with: name].! !

!QCustoms methodsFor: 'actions' stamp: 'hrs 3/1/2009 11:19'!
replace: aString with: aFileCard
	"Update an unprocessed name use the given filecard."
	| index old |
	index := 0.
	imports detect: [:entry | index := index + 1. entry name = aString].
	old := imports at: index.
	old fileCard: aFileCard.
	signalee signal: #processedImport with: aString with: aFileCard.! !

!QCustoms methodsFor: 'accessing' stamp: 'hrs 2/16/2009 12:29'!
signalee: aTObject
	"Tells us what well-known place signals should be sent.
	For example, we're not a rooted TFrame and so we don't know about the space."
	signalee := aTObject! !

TObject subclass: #QEventLoop
	instanceVariableNames: 'mutex eventQueue eventLoop stopPromise isDebugging myScripts myEventMap'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Objects-Kernel'!

!QEventLoop commentStamp: '<historical>' prior: 0!
TO-DO: get rid of the 'stopPromise' mechanism.  Instead, destroy the loop immediately in #destroy, and curtail any promises that haven't been processed.!

!QEventLoop class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QEventLoop methodsFor: 'testing' stamp: 'jcg 3/16/2009 10:38'!
currentlyInEventLoop
	"Answer true if we're currently executing in our event-loop, and false if we're executing in some other Process."
	^ eventLoop ifNil: [false] ifNotNilDo: [:loop | loop isCurrentlyRunning]! !

!QEventLoop methodsFor: 'initialize' stamp: 'crl 4/9/2010 14:24'!
debug
	isDebugging := true! !

!QEventLoop methodsFor: 'initialize' stamp: 'jcg 10/28/2009 17:27'!
destroy
	eventLoop ifNil: [^self "already destroyed"].
	eventLoop isDeadOrDying ifTrue: [
		"Unexpected... nothing we can do but log a warning."
 		^QLogger logWarning: 'invalid event-loop in #destroy'.
	].
	stopPromise ifNotNil: [
		"Already in the midst of being destroyed; don't do it twice."
		^self
	].

	"Ensure that we destroy ourself within the event-loop."
	self currentlyInEventLoop ifFalse: [
		"Block until destruction is complete."
		(self future destroy waitTimeoutSeconds: 5) ifFalse: [
			"The destruction will proceed, but we're not waiting for it any longer.
			 If this happens, it's likely that there's a deadlock, which is the reason
			 that we log a warning."
			QLogger logWarning: 'destruction time exceeds 5 seconds (deadlocked?)'
		].
		^self
	].

	"Actually do the destruction."
	self 
		privateDestroy; 	"Intended to be extended by subclasses."
		stopEventLoop;
		signal: #destroyed with: self.
! !

!QEventLoop methodsFor: 'private' stamp: 'crl 4/9/2010 14:17'!
enqueueEvent: msgSend
	isDebugging ifTrue: [QLogger logInfo: thisContext sender sender shortStack].
	eventQueue ifNotNilDo: [:q | q nextPut: msgSend].
! !

!QEventLoop methodsFor: 'initialize' stamp: 'ar 4/24/2009 12:51'!
eventLoopPriority
	"Answer the priority at which the event loop should run"
	^Processor lowIOPriority - 1. "by default not quite as high as network I/O"! !

!QEventLoop methodsFor: 'enqueuing' stamp: 'jcg 1/12/2009 23:32'!
futureDo: aSelector at: deltaMSecs args: args
	"Enqueue a message on the event queue.  The compiler has been modified to generate this message when the #future message appears in the source code."
	deltaMSecs = 0 ifFalse:[^self error: 'deltaMSecs must be zero'].
	self futureDo: aSelector receiver: self args: args.
	^nil
! !

!QEventLoop methodsFor: 'enqueuing' stamp: 'jcg 1/12/2009 23:36'!
futureDo: aSelector receiver: receiver args: args
	"Called by implementors of #futureSend:at:args: who want to enqueue a message on the event-queue.  This extra indirection helps other objects enqueue messages on our event loop."
	self enqueueEvent: (MessageSend 
									receiver: receiver 
									selector: aSelector 
									arguments: args).
	^nil
! !

!QEventLoop methodsFor: 'enqueuing' stamp: 'jcg 1/13/2009 01:43'!
futureSend: aSelector at: deltaMSecs args: args
	"Enqueue a message on the event queue.  The compiler has been modified to generate this message when the #future message appears in the source code.  Answers a promise that will be resolved with the result of performing the message."
	deltaMSecs = 0 ifFalse:[^self error: 'deltaMSecs must be zero'].
	^self futureSend: aSelector receiver: self args: args.
! !

!QEventLoop methodsFor: 'enqueuing' stamp: 'jcg 1/12/2009 23:35'!
futureSend: aSelector receiver: receiver args: args
	"Called by implementors of #futureSend:at:args: who want to enqueue a message on the event-queue.  This extra indirection helps other objects enqueue messages on our event loop."
	| promise |
	promise := TPromise new.
	self enqueueEvent: (PromiseMessageSend
									promise: promise
									receiver: receiver 
									selector: aSelector 
									arguments: args).
	^promise
! !

!QEventLoop methodsFor: 'private' stamp: 'jcg 1/27/2009 19:21'!
handleEventLoopError: err
	"XXXXX: need some sensible error handling."
	QLogger logError: err! !

!QEventLoop methodsFor: 'initialize' stamp: 'bgf 3/21/2009 21:13'!
initialize
	super initialize.
	mutex := Mutex new.
	isDebugging := false.
	self startEventLoop.! !

!QEventLoop methodsFor: 'testing' stamp: 'jcg 6/10/2010 17:10'!
isInTurn
	"For compatibility with QEventLoop2, which we are transitioning over to."
	^self currentlyInEventLoop! !

!QEventLoop methodsFor: 'private' stamp: 'bgf 4/2/2009 18:22'!
myEventMap
	"Adapter so this can participate efficiently in event signalling like CObject"
	^myEventMap! !

!QEventLoop methodsFor: 'private' stamp: 'bgf 4/2/2009 18:22'!
myEventMap: aMap
	"Adapter so this can participate efficiently in event signalling like CObject"
	myEventMap := aMap! !

!QEventLoop methodsFor: 'private' stamp: 'bgf 4/2/2009 18:22'!
myScripts
	"Adapter so this can participate efficiently in event signalling like CObject"
	^ myScripts ifNil: [#() ]! !

!QEventLoop methodsFor: 'private' stamp: 'bgf 4/2/2009 18:22'!
myScripts: scripts
	"Adapter so this can participate efficiently in event signalling like CObject"
	^ myScripts := scripts! !

!QEventLoop methodsFor: 'initialize' stamp: 'jcg 8/23/2009 15:30'!
privateDestroy
	"Subclasses should override this, not #destroy."! !

!QEventLoop methodsFor: 'private' stamp: 'jcg 1/13/2009 12:50'!
privateStopEventLoop
	"#stopEventLoop causes this to be called from within the event-loop.  Since we are in the event-loop, no subsequent events will be processed."
	mutex critical: [
		eventLoop stop.
		eventLoop := eventQueue := stopPromise := nil.
	]! !

!QEventLoop methodsFor: 'private' stamp: 'jcg 7/9/2010 21:37'!
processEventLoopMessage: msg
	| result desc promise |

	"Debugging."
	isDebugging ifTrue: [
		desc := msg selector printString, '  by: ', msg receiver.
		msg arguments size > 0 ifTrue: [desc := desc, '  with: ', msg arguments].
		QLogger logInfo: 'START PROCESSING MESSAGE: ', desc
	].

	"Process the next event, and resolve the promise (if any)."
	promise := msg promise.
	[	result := msg receiver perform: msg selector withArguments: msg arguments.
		promise ifNotNil: [:p | p fulfill: result].
	] ifCurtailed: [promise ifNotNil: [:p | p break: 'event-loop process terminated']].

	"Debugging."
	isDebugging ifTrue: [QLogger logInfo: 'END PROCESSING MESSAGE: ', desc].
! !

!QEventLoop methodsFor: 'initialize' stamp: 'ar 4/24/2009 12:51'!
startEventLoop
	self stopEventLoop wait. "wait until event-loop has finished stopping before restarting it"
	mutex critical: [
		"It's possible that someone else started the event-loop after it was stopped.
		 If so, there's nothing to do here."
		eventQueue ifNotNil: [^self].
		eventQueue := SharedQueue new.
		eventLoop := QSafeProcessLoop new.
		eventLoop priority: self eventLoopPriority.
		eventLoop actionBlock: [self processEventLoopMessage: eventQueue next].
		eventLoop handlerBlock: [:err | self handleEventLoopError: err].
		eventLoop start.
	]! !

!QEventLoop methodsFor: 'initialize' stamp: 'jcg 5/7/2010 11:59'!
stopEventLoop
	mutex critical: [
		"If we're stopped or trying to stop, exit early."
		eventQueue ifNil: [^TPromise new fulfill: self "already stopped"].
		stopPromise ifNotNil: [^stopPromise "already in the process of stopping"].
		
		"This promise will be resolved once the event-queue is fully stopped."
		^stopPromise := self future privateStopEventLoop
	]! !

TObject subclass: #QGraspAdapter
	instanceVariableNames: 'luggage grip graspPt graspDistance cameraNorm isEnabled isBeingDragged preDragState downColor overColor isCancelledDrag'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Objects-Manipulation'!

!QGraspAdapter commentStamp: '<historical>' prior: 0!
iAn abstract manager of pointer-based drag state for a TFrame.
Translates mouse events into a drag protocol.

QGraspAdapter manipulates two TFrames:

	Its 'lugguge' - a TFrame (the graspable itself, or its parent, some further ancestor)
	which represents the 'whole' object being manipulated.  
	Movement translations are reckoned aganst, and applied to, the luggage.

	Its 'grip' a TFrame (part of the luggage, or the luggage itself) 
	with which the user interacts to perform the grasping of the luggage.

GraspAdapters can be enabled and disabled freely; multiple adapters
can be attached to the same luggage and grip, so long as only
one adapter is enabled for a given grip at a given time.

At the start of a drag operation (pointer-down), 
the following instance variables are captured:

	preDragTransform	- the luggage transform before dragging; used for snap-back.
	cameraNorm		- the initial pick normal, for use in later TRay pickPlane: operations.
	graspPt				- the point (in my reference) at which the pointer first hit me.
	graspDistance		- the (global) distance from viewer to graspPt.

No drag protocol is implemented here.  The 'grip' should implement
	#beginDrag:
	#drag:   (Required.)
	#endDrag:
	#cancelDrag:

In particular, #drag:event should move (translate) the luggage as drag events occur,
and #endDrag: should make final disposition (e.g. do the drop).

Graspable arranges for those methods to be called properly,
and for the graspPt, graspDistance, and other state to be set and cleared properly.
!

QGraspAdapter subclass: #QAbstractPivot
	instanceVariableNames: 'graspRadius offset rotationStep ring'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Objects-Manipulation'!

!QAbstractPivot commentStamp: '<historical>' prior: 0!
Abstract base for pivot grips which afford rotation of their luggage
about the Y axis, using a grip that the user "drags" through space.

The pivot rides as an attachment on its luggage and grip frames.
It attaches mouse handlers on the grip,
and it operates on the transform of the luggage to carry out rotations.
It can be suspended and reactivated through #disable and #enable.

When the pivot has non-zero rotationStep, it rotates by discrete steps;
the rotations are delivered to the luggage via #rotationAroundY:

When the pivot has zero rotation, it sends every mouse motion
to the luggage, using #addRotationAroundY: 

Both the luggage and grips are expected to be TFrames;
the luggage assembly must be a frame tree ancestor of the grip
(the grip must be a part of the frame it moves.)
The grip can be the same frame as the luggage.

This abstract class provides no picking geometry.
Different concrete classes implement different picking geometries;
there are various ways to interpret the raw 2-D dragging information
as handle movement in the 3-D space.
This class provides pointer-handling setup,
so that subclasses need only implement the core drag: operations.

An optional 'ring' TFrame (usually a simple translucent primitive)
is added to the space, centered on the pivot axis,
for the duration of a pivot pick interaction.

The 'rotationRing' property determines the granularity of free rotations;
it is a scalar (number of degrees.)  Setting it zero gives free rotation,
without the option of rotation-by-large steps ('Shift-drag' behavior).
!

!QAbstractPivot class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QAbstractPivot class methodsFor: 'instance creation' stamp: 'bgf 9/19/2006 17:48'!
forLuggage: aFrame grip: aSubframe
	"Make the given subframe behave as a rotation grip for the given frame."

	^ (self new) luggage: aFrame; grip: aSubframe; yourself.! !

!QAbstractPivot class methodsFor: 'undo-redo' stamp: 'bgf 3/18/2010 16:46'!
verb
	"The label for this pivot's action, for undo-redo registration."
	^ 'Turn'! !

!QAbstractPivot methodsFor: 'drag' stamp: 'jdm 8/16/2007 16:48'!
beginDrag: event 

	"Mouse down... prepare to rotate.  Saves the initial graspRadius:
	 the distance from the luggage axis origin to the grasp pt.
	 Positions the feedback ring appropriately (centered on the rotation plane)
	 and adds it to the space." 

	| pt |

	pt := (self luggage relativeTransform: grip) transformPoint: graspPt.
	pt := (pt - offset).

	graspRadius :=((pt x * pt x) + (pt z * pt z)) sqrt.

	ring ifNotNil: [
		self luggage addChild: ring.
		ring radius: graspRadius;
			translation: (offset x) @ (pt y) @ (offset z);
			yaw: (self luggage yaw negated).
		ring releaseToRoot.
	].

	^ true! !

!QAbstractPivot methodsFor: 'drag' stamp: 'bgf 9/21/2006 11:42'!
cancelDrag: event

	"Abandoned drag operation.  Remove the feedback ring."
	super cancelDrag: event.
	ring ifNotNil: [
		ring root removeChild: ring.
	].

! !

!QAbstractPivot methodsFor: 'events' stamp: 'bgf 10/21/2006 09:55'!
clearHow

	self isEnabled ifTrue: [
		grip root signal: #statusMessage with: ''
	].! !

!QAbstractPivot methodsFor: 'drag' stamp: 'bgf 9/21/2006 11:42'!
doDrag: event
	
	"We take over full control of the drag protocol."
	self drag: event.! !

!QAbstractPivot methodsFor: 'drag' stamp: 'bgf 9/19/2006 17:49'!
drag: event

	"Mouse motion handler when interactive dragging is taking place...
	 rotate the luggage as indicated by the mouse position change."
	self subclassResponsibility.! !

!QAbstractPivot methodsFor: 'drag' stamp: 'bgf 3/27/2010 08:54'!
endDrag: event withUndo: bUndo

	"Successful conclusion to a rotation operation.  Remove the feedback ring."
	ring ifNotNil: [
		ring root removeChild: ring.
	].

	super endDrag: event withUndo: bUndo


! !

!QAbstractPivot methodsFor: 'accessing' stamp: 'bgf 10/21/2006 09:56'!
grip: aTFrame

	"The grip is the TFrame that the user manipulates to rotate the luggage."
	super grip: aTFrame.

	self stopScript: #sayHow.
	self stopScript: #clearHow.
	self runScript: #sayHow when: { aTFrame . #pointerEnter }.
	self runScript: #clearHow when: {aTFrame . #pointerLeave }.! !

!QAbstractPivot methodsFor: 'initialize' stamp: 'bgf 9/21/2006 11:42'!
initialize
	super initialize.

	rotationStep := 2.5.		"degrees"
	offset := 0@0@0.


! !

!QAbstractPivot methodsFor: 'initialize' stamp: 'bgf 1/25/2007 09:17'!
noVisualRing

	"Don't use the feedback ring (subclasses supply their feedback rings
	 by default - this allows the client to override the subclass's indicator."
	ring := nil.! !

!QAbstractPivot methodsFor: 'accessing' stamp: 'bgf 9/21/2006 11:44'!
offset

	"The offset from the luggage frame's origin to the desired rotation axis; 
	 the coordinates are in the luggage's frame of reference."

	^ offset! !

!QAbstractPivot methodsFor: 'accessing' stamp: 'bgf 9/19/2006 17:50'!
offset: aPoint

	"The offset from the luggage frame's origin to the desired rotation axis; 
	 the coordinates are in the luggage's frame of reference."

	(aPoint size = 2) ifTrue: [
		offset := (aPoint x)@0@(aPoint y).
	] ifFalse: [
		offset := (aPoint x)@0@(aPoint z).
	].! !

!QAbstractPivot methodsFor: 'accessing' stamp: 'bgf 9/19/2006 17:47'!
rotationStep

	"How many degrees rotation in a single increment?"
	^ rotationStep! !

!QAbstractPivot methodsFor: 'accessing' stamp: 'bgf 9/19/2006 17:47'!
rotationStep: degrees

	"How many degrees rotation in a single increment?"
	rotationStep := degrees.! !

!QAbstractPivot methodsFor: 'events' stamp: 'bgf 1/25/2007 11:14'!
sayHow

	| msg |
	self isEnabled ifTrue: [
		msg := (rotationStep > 0)
					ifTrue: [ 'Drag to rotate.  Press Shift to rotate in steps.'] 
					ifFalse: [ 'Drag to rotate.'].
		grip root signal: #statusMessage with: msg.
	].! !

QAbstractPivot subclass: #QCylinderPivot
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Objects-Manipulation'!

!QCylinderPivot commentStamp: '<historical>' prior: 0!
See QAbstractPivot for overview.

Affords rotation of its luggage (a TFrame assembly) around its origin axis
through pointer manipulation of a grip (aTFrame within the assembly).
The mouse  uses #pickCylinder: operations on a cylinder 
surrounding the axis of rotation, whose radius is determined by the initial
mouse-down position of the rotation drag.

A visible feedback tube is presented during the rotation,
which helps the user understand how mouse motion is related to rotation.

Offset is not (yet) supported for this pivot; rotation is always around
the x=0, z=0 axis of the luggage.
!

!QCylinderPivot class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QCylinderPivot class methodsFor: 'instance creation' stamp: 'bgf 9/14/2006 18:12'!
forLuggage: aFrame grip: aSubframe
	"Make the given subframe behave as a rotation grip for the given frame."

	^ (self new) luggage: aFrame; grip: aSubframe; yourself.! !

!QCylinderPivot methodsFor: 'drag' stamp: 'jdm 8/16/2007 16:52'!
drag: event

	"Rotate the luggage, according to how the grasp point of the grip is moved.
	 Adapted from TWindow#rotateWindow:"

	| pointer spinTo spinFrom delta deltaLoc theta yaw |

	pointer := event makePointer.
	(pointer frame: self luggage 
			pickRightCylinderFrom: (Vector3 x: 0 y: 100 z: 0)
			to: (Vector3 x: 0 y: -100 z: 0) 
			radius: graspRadius
	) ifTrue: [
		spinTo := pointer selectedPoint copy.
	] ifFalse: [
		"Not in the cylinder... reckon an approximate spin"
		delta := 2*graspRadius.
		(pointer rayPointer row1 dot: (pointer rayPosition))  > 0 ifFalse:[  
			delta := delta negated.
		].
		deltaLoc := pointer rayPointer row1 * delta.
		(pointer frame: self luggage
				pickRightCylinderFrom: (Vector3 x: deltaLoc x y: 100.0 z: deltaLoc z)
				to: (Vector3 x: deltaLoc x y: -100.0 z: deltaLoc z) 
				radius: graspRadius
		) ifFalse:[
			^ false.
		 ].
		spinTo := pointer selectedPoint - deltaLoc.
		spinTo x: spinTo x negated  y: 0 z: spinTo z negated.
	 ].

	spinFrom := (self luggage relativeTransform: grip) transformPoint: graspPt.
	spinFrom y: 0; normalize.
	spinTo y: 0; normalize.

	theta := ((spinFrom z * spinTo x) - (spinTo z * spinFrom x)).
	theta < -1.0 ifTrue:[theta:= -1.0].
	theta > 1.0 ifTrue:[theta := 1.0].
	theta := theta arcSin radiansToDegrees/2.0.

	"Enforce rotation in discrete steps unless free rotation is requested."
	rotationStep > 0 ifTrue: [
		(event shiftPressed) ifFalse: [ 
			yaw := (self luggage yaw + theta) roundTo: self rotationStep.
		] ifTrue: [
			yaw := 	(self luggage yaw + theta) detentBy: 45.0 atMultiplesOf: 45.0 snap: true.
		].
		self luggage rotationAroundY: yaw.
	] ifFalse: [
		"Under free rotation, we use the add callback!!"
		self luggage addRotationAroundY: theta.
	].
	

! !

!QCylinderPivot methodsFor: 'initialize' stamp: 'bgf 1/2/2009 12:45'!
initialize
	super initialize.

	"A hollow cylinder indicating the pick cylinder for rotation."
	ring := QCylinderRing new.
	ring solid: false.
! !

QAbstractPivot subclass: #QPlanePivot
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Objects-Manipulation'!

!QPlanePivot commentStamp: '<historical>' prior: 0!
Allows rotation of its luggage (a TFrame assembly) around a specified vertical axis,
by pointer manipulation of a grip (aTFrame within the assembly). 
The interaction uses #pickPlane:  on a plane perpendicular to the axis of rotation.
A visible feedback ring is presented during the rotation.

The plane-picking interaction is especially useful for large objects
where the rotation radius is long.

As an example of the use of the offset axis,
a doorknob is an offset pivot grip whose axis is at
the opposite edge of the door; to express this, one would
create a pivot grip whose grip was the doorknob,
and whose offset was the x-z translation from the origin 
of the door's frame of reference to the vertical line through the hinges.

The axis of rotation is specified as an offset from the luggage's 0,0,0 origin;
see #offset:.    The same luggage can have multiple plane pivots,
each with their own axis of rotation.
A common combination is a pair of grips at either side of a flat panel,
each using the other's position as its own axis... so moving each
end rotates the panel about the other end.
!

!QPlanePivot class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QPlanePivot class methodsFor: 'instance creation' stamp: 'bgf 9/18/2006 14:27'!
forLuggage: aFrame grip: aSubframe
	"Make the given subframe behave as a rotation grip for the given frame."

	^ (self new) luggage: aFrame; grip: aSubframe; yourself.! !

!QPlanePivot methodsFor: 'drag' stamp: 'jdm 8/16/2007 16:55'!
drag: event

	"Rotate the luggage, according to how the grasp point of the grip is moved.

	 We pretend we're in a horizontal plane, and calculate where along the
	 radius of rotation would be closest to the pick point."

	| pointer spinTo spinFrom theta shiftDir shiftLen shift graspDir toDir |

	pointer := event makePointer.

	spinFrom := ((self luggage relativeTransform: grip) transformPoint: graspPt).

	(pointer frame: self luggage
		     pickPlane: spinFrom
		     normal: (0@1@0)
	) ifTrue: [
		spinTo := pointer selectedPoint copy.
	] ifFalse: [
		^ false.
	].

	graspDir := (offset x @ offset z) bearingToPoint: (spinFrom x @ spinFrom z).
	toDir := (offset x @ offset z) bearingToPoint: (spinTo x @ spinTo z).

	theta := (graspDir - toDir).

	"Enforce rotation in small  steps unless hard steps are selected (shift)."

	(event shiftPressed) ifFalse: [ 
		theta  := ((self luggage yaw + theta) roundTo: self rotationStep) 
					- self luggage yaw.
	] ifTrue: [
		theta := ((self luggage yaw + theta) detentBy: 45.0 atMultiplesOf: 45.0 snap: true) 
					- self luggage yaw.
	].
	
	"Theta is the rotation desired around the offset pivot.
	  This needs to be translated into a rotation + translation
	  that brings the luggage into the desired orientation."

" Transcript cr; show: 'Offset = ', offset, ' graspDir = ', graspDir, ' toDir = ', toDir, ' theta = ', theta.  "

	(graspDir > 180) ifTrue: [
		shiftDir := 90.0 - (self luggage yaw + (theta/2)).
	] ifFalse: [
		shiftDir := -90 - (self luggage yaw + (theta/2)).
	].
	shiftLen := 2.0 * (offset length) * ((theta/2)degreeSin).

	shift := (Point new) setR: shiftLen degrees: shiftDir.

	"This shift is the vector (in the y=0 plane)
	  along which we must translate the luggage."

	self luggage translation: (self luggage translation + ((shift x)@0@(shift y))).
	self luggage rotationAroundY: (self luggage yaw + theta).
! !

!QPlanePivot methodsFor: 'initialize' stamp: 'bgf 1/2/2009 12:46'!
initialize
	super initialize.

	ring := QPlaneRing new.
	ring solid: false.
! !

QGraspAdapter subclass: #QAbstractResizer
	instanceVariableNames: 'dragOffset transient xsign ysign'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Objects-General'!

!QAbstractResizer commentStamp: '<historical>' prior: 0!
A generic adapter shell for resizing one or two dimensions in a plane.

Supply luggage and a grip when creating;
the concrete subclass determines how resize deltas are supplied to the target.!

!QAbstractResizer class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QAbstractResizer class methodsFor: 'instance creation' stamp: 'bgf 8/4/2009 16:34'!
forLuggage: luggageFrame grip: handle
	"Convenience method:
	 Installs a length-resize adapter in the luggage frame.
	 The resizer (result) should be held in the luggage as an ivar (to keep it reachable.)"
	|  resizer |
	resizer := (self new) 
		luggage: luggageFrame;	
		grip: handle;
		yourself.
	^ resizer
	! !

!QAbstractResizer class methodsFor: 'as yet unclassified' stamp: 'bgf 3/18/2010 16:47'!
undoDrag: luggage toState: resizerState
	"Support for undo/redo: restore the luggage posture to the given state."

	luggage ifNil: [ ^ nil ].
	luggage root ifNil:[ ^ nil  ].
	luggage applyResizerState: resizerState.
! !

!QAbstractResizer class methodsFor: 'undo-redo' stamp: 'bgf 3/18/2010 16:46'!
verb
	"The label for this adapters's action, for undo-redo registration."
	^ 'Resize'! !

!QAbstractResizer methodsFor: 'drag' stamp: 'bgf 3/15/2010 13:54'!
beginDrag: event 	
	"Beginning a drag.  
	 Using the graspPt, we derive luggageGraspPt:
	 the same point, but as an offset from the luggage's center,
	 and held in luggage frame coordinates.
	 This is used later to translate the luggage relative to a surface point."
	| luggageGraspPt |

	luggage resizeBegin: event.
	luggageGraspPt := graspPt.
	grip ~= luggage ifTrue: [
		luggageGraspPt := (luggage relativeTransform: grip) transformPoint: graspPt.
		grip visible: true.		"Purely to work around Croquet's insert of a leave before a mouse-down."
	].
	dragOffset := ((luggageGraspPt x * xsign) @ (luggageGraspPt y * ysign)) 
				- ( luggage extent asPoint / 2.0).
	
	preDragState := luggage resizerState.

	^ true
! !

!QAbstractResizer methodsFor: 'drag' stamp: 'bgf 3/15/2010 13:54'!
clearDragState
	super clearDragState.
	dragOffset := nil.! !

!QAbstractResizer methodsFor: 'accessing' stamp: 'bgf 8/4/2009 16:34'!
disable
	super disable.	
! !

!QAbstractResizer methodsFor: 'drag' stamp: 'bgf 3/23/2010 11:25'!
doDrag: event
	"Drag the corner"
	|  pointer pt old de |
	grip root ifNil: [
		^ self  "Guard against some transitional situations."
	].
	dragOffset ifNil: [ 		
		dragOffset := 0@0.
	].

	pointer := event makePointer.
	(pointer frame: grip pickPlane: graspPt normal: (0@0@-1) ) ifFalse: [
		"If pickplane fails, we can't reckon the drag position."
		^ self.
	].
	pt := pointer selectedPoint.
	event avatar laserShow: (grip globalTransform transformPoint: pt).
	grip ~= luggage ifTrue: [
		pt :=  (luggage relativeTransform: grip) transformPoint: pt.
	].

	old := luggage extent asPoint clone.
	de := (pt x * xsign @ (pt y * ysign)) - dragOffset  - (old  *0.5).

	self update: pt delta: de oldExtent: old


! !

!QAbstractResizer methodsFor: 'drag' stamp: 'bgf 3/18/2010 16:50'!
dragState
	"The current posture and extent of the luggage, as a state array for undo-redo support."
	^ luggage resizerState
! !

!QAbstractResizer methodsFor: 'accessing' stamp: 'bgf 8/4/2009 16:34'!
enable
	super enable.
	grip ifNotNil: [ grip visible: (transient not) ].! !

!QAbstractResizer methodsFor: 'drag' stamp: 'bgf 3/27/2010 08:55'!
endDrag: event withUndo: bUndo
	super endDrag: event withUndo: bUndo.
	dragOffset := nil.
	luggage resizeComplete: event.! !

!QAbstractResizer methodsFor: 'initialize' stamp: 'bgf 8/4/2009 17:06'!
initialize
	transient := false.
	xsign := 1.
	ysign := 1.
	super initialize.
! !

!QAbstractResizer methodsFor: 'events' stamp: 'bgf 8/4/2009 16:34'!
pointerDown: event
	super pointerDown: event.
	"Below a workaround for Croquet's insert of a leave mouse event before any down event..."
	(self isEnabled & transient and: [ grip notNil] ) ifTrue: [ grip visible: true ].! !

!QAbstractResizer methodsFor: 'drag' stamp: 'bgf 3/15/2010 13:55'!
snapBack
	"restore original place and size"
	isBeingDragged ifTrue: [
		(preDragState  notNil) ifTrue: [
			luggage applyResizerState: preDragState
		].
	].! !

!QAbstractResizer methodsFor: 'accessing' stamp: 'bgf 8/4/2009 16:34'!
transient
	^ transient! !

!QAbstractResizer methodsFor: 'accessing' stamp: 'bgf 8/4/2009 16:34'!
transient: bool
	"Transient means the owner is making the grip visible during mouseovers."
	transient := bool.
	(grip notNil and: [ grip ~= luggage ]) ifTrue: [ grip visible: (transient not) ].! !

!QAbstractResizer methodsFor: 'protocol' stamp: 'bgf 8/19/2009 15:11'!
update: cornerPt delta: de oldExtent: old

	"During resizing dragging, this is called for each motion.
	 cornerPt is the selection event point transformed into current luggage coordinates.
	 de is the delta-extent between this and the previous iteration,
	 and old is the extent from which that delta is reckoned."

	self subclassResponsibility! !

QAbstractResizer subclass: #QCorner2DResizer
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Objects-General'!

!QCorner2DResizer commentStamp: '<historical>' prior: 0!
A resizing adapter for adjusting the extent of a 2-D object.
The grip would ordinarily be a resizing-corner - a small rectangle,
typically in the lower-right corner of a resizable sheet.

The luggage can implement
		luggage dragCorner: #southeast delta: (de x @ (de y negated)).
and if that method is available it will be used for updates on the fly.

The lugguge must conform to 
	extent: aPoint 
	extent 
such that 
	aPoint = (luggage extent: aPoint;  extent) 	
is always true.


The luggage must implement
	minExtent 
returning a Point.


!

!QCorner2DResizer class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QCorner2DResizer class methodsFor: 'instance creation' stamp: 'bgf 8/27/2009 17:59'!
forLuggage: luggageFrame

	"Convenience method:
	 Installs a resize corner rectange in the luggage frame,
	 binding a resize-adapter to it and the luggage.
	 The resizer (result) should be held in the luggage as an ivar (to keep it reachable.)

	 See also 'updateShape' which should be used from the updateShape of 
	 a normal rectangular sheet to update the grip position."

	| cornerRect resizer |
	cornerRect := (TSimpleRectangle new)
		material: (QMaterials resizeCornerMaterial);
		shouldPersist: false;
		yourself.
	luggageFrame addChild: cornerRect.

	resizer := (self new) 
		luggage: luggageFrame;	
		grip: cornerRect;
		yourself.

	^ resizer
	! !

!QCorner2DResizer methodsFor: 'accessing' stamp: 'bgf 9/5/2008 15:26'!
disable
	super disable.	
	grip ifNotNil: [ grip visible: false ].! !

!QCorner2DResizer methodsFor: 'accessing' stamp: 'bgf 8/16/2009 22:07'!
enableClickOnly
	super enableClickOnly.
	grip ifNotNil: [ grip visible: false ].! !

!QCorner2DResizer methodsFor: 'accessing' stamp: 'bgf 8/16/2009 22:04'!
grip: aFrame
	super grip: aFrame.
	aFrame ifNotNil: [	
		self isEnabled ifFalse: [ grip visible: false ].
	].! !

!QCorner2DResizer methodsFor: 'initialize' stamp: 'bgf 8/4/2009 17:13'!
initialize
	super initialize.
	ysign := -1.			"Lower-right corner, so y grows down."! !

!QCorner2DResizer methodsFor: 'drag' stamp: 'bgf 3/15/2010 13:55'!
update: cornerPt delta: de  oldExtent: old
	"Use QPanelHandler/QDisplayPanel corner-drag protocol if available,
	 since that allows smart resizing of (for instance) displayStands."
	(luggage respondsTo: #drag:corner:delta:) ifTrue: [
		luggage drag: cornerPt corner: #southeast delta: (de x @ (de y negated)).
	] ifFalse: [
		| newSize dt |
		newSize := (old + de) max: ( [ luggage minExtent ] on: Error do: [ grip extent * 2 ] ).
		"Move the luggage half the delta, so the resize is a proper corner-drag."
		luggage extent: newSize.
		newSize := luggage extent asPoint. 	  "So motion is based on the actual extent."
		dt := (newSize - (preDragState third)) *0.5.
		dt := (dt x @ dt y negated).
		luggage
		        translation: ((preDragState first translation) 
						+ (luggage orientation transformPoint: (dt @ 0.01))).
						"The z fudge factor fights numeric creep away from the pick plane."
	].

! !

!QCorner2DResizer methodsFor: 'initialize' stamp: 'bgf 8/23/2009 20:41'!
updateShape
	"Fit my grip into the standard place at the corner of the luggage.
	 This can be called by the updateShape of luggage that is self-centered;
	 or the luggage can place the grip where it wants it."

	| ext gext z |
	ext := luggage extent.
	gext := (ext asPoint minDimension * 0.12) @ (ext asPoint minDimension * 0.12).

	gext := gext max: 0.12@0.12.
	gext := gext min: (ext * 0.275).
	gext := gext min: (0.32@0.32).

	ext isArray ifTrue: [
		z := (ext z * 0.5) + 0.005.
	] ifFalse:[
		z := 0.005
	].
	grip
		extent: gext;
		translation: ((ext x * 0.5) - (gext x * 0.501)) 
					@ ((ext y * -0.5) + (gext y * 0.499)) 
					@ z.! !

QAbstractResizer subclass: #QHeightResizer
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Objects-General'!

!QHeightResizer commentStamp: '<historical>' prior: 0!
A 2D-resizer that operates on the vertical (y) dimension of its luggage.
Updates during resize are reported via
	luggage dragHeightDelta:  (change in y of grasp point)
or if that method is not supported,
	then the extent is modulated!

!QHeightResizer class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QHeightResizer methodsFor: 'drag' stamp: 'bgf 8/23/2009 21:38'!
update: cornerPt delta: de oldExtent: old

	(luggage respondsTo: #dragHeightDelta:) ifTrue: [
		luggage dragHeightDelta: (de y).
	] ifFalse: [
		| newSize dt |
		newSize := (old + (0@ de y)) max: ( [ luggage minExtent ] on: Error do: [ 0.1@0.1 ] ).
		"Move the luggage half the delta, so the resize is a proper end-drag."
		luggage extent: newSize.
		newSize := luggage extent asPoint.	"So motion is based on the actual extent."
		dt := (newSize - old) / 2.0.
		dt := (0 @ dt y negated).
		luggage 
			translation: ((luggage translation) 
						+ (luggage orientation transformPoint: (dt @ 0 )));
			doUpdateShape.
	].

! !

QAbstractResizer subclass: #QLengthResizer
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Objects-General'!

!QLengthResizer commentStamp: '<historical>' prior: 0!
A 2D-resizer that operates on the horizontal (x) dimension of its luggage.

Updates during resize are reported via
	luggage dragLengthDelta:  (change in y of grasp point)
or if that method is not supported,
	then the extent is modulated!

!QLengthResizer class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QLengthResizer methodsFor: 'drag' stamp: 'bgf 8/23/2009 21:38'!
update: cornerPt delta: de oldExtent: old

	(luggage respondsTo: #dragLengthDelta:) ifTrue: [
		luggage dragLengthDelta: (de x).
	] ifFalse: [
		| newSize dt |
		newSize := (old + (de x @ 0)) max: ( [ luggage minExtent ] on: Error do: [ 0.1@0.1 ] ).
		"Move the luggage half the delta, so the resize is a proper end-drag."
		luggage extent: newSize.
		newSize := luggage extent asPoint.	"So motion is based on the actual extent."
		dt := (newSize - old) / 2.0.
		dt := (dt x @ dt y negated).
		luggage 
			translation: ((luggage translation) 
						+ (luggage orientation transformPoint: (dt @ 0 )));
			doUpdateShape.
	].

! !

QGraspAdapter subclass: #QDragAdapter
	instanceVariableNames: 'isDeletable currDropTarget isMovable isGroundBound isAttractedToSurfaces isDroppable luggageGraspPt contactOffset attractionRange surfacePolicy openSpacePolicy excludeAvatars'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Objects-Manipulation'!

!QDragAdapter commentStamp: '<historical>' prior: 0!
A grasp adapter extended to implement a spatial drag/drop protocol.

When attaching a drag adapter to a luggage frame,
the luggage frame should implement: 

	#payload - return the payload object (default is self.)

The luggage may also implement:
	#savePosition - used at the end of a movement drag (when no drop into a target occurs)
                             to indicate to the luggage that its current position
				     is a new user-assiged position.

Drop targetting decisions are made by testing for two methods
on frames encountered during the interactive drag process:

	isDropTarget: draggedFrame
		(called to control interactive feedback during drag operations, 
		 must return boolean, true if the receiver can handle a drop of draggedFrame)

	consumeDrop: draggedFrame avatar: aTAvatarReplica selection: aTXSel
		(called to carry out a drop operation, must return boolean,
		 true if the receiver assumes ownership of draggedFrame)

The consumeDrop true/false result controls whether or not I should put back the payload
after a drag/drop operation.   If consumeDrop returns true, 
I should assume the payload has been reparented into the drop target.  
If false,  I should restoremyself to the state I was in before the drag.

Public configuration attributes:

	isEnabled:	 if set false, makes the object ignore keyboard and mouse (drag) operations
	isDeletable:  can the object be destroyed by pressing the delete key?

	isMovable:	should dragging reposition the object when no drop target receives the final drop?
	isDroppable:  should dragging engage the drop target protocol?
					if false, only spatial movement and drop-on-space are possible.

	isAttractedToSurfaces:	should dragging favor positions near existing surfaces?
	isGroundBound:			should dragging strongly favor placement above horizontal surfaces?

The last two  properties control how elastic the 'outward' (away from the viewer) 
motion of the object is during drag.   When surface or ground attraction is on,
the object will move quite far from the viewer when the drag-ray hits 
a surface behind the object.   These only apply when the object is movable.
!

!QDragAdapter class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QDragAdapter class methodsFor: 'utilities' stamp: 'bgf 8/21/2009 02:04'!
dragDrop: payload intoSpace: space forEvent: event
	"Place the payload by pretending to drag and drop it using the given event's 
	 ray-stance to guide placement.  Returns true if a placement was made, false if not."
	| box vector depth pointer |

	"This only works for objects that have exposed a primary drag adapter."
	payload dragAdapter ifNil: [ ^ false ].

	payload localTransform: Matrix4x4 new.
	payload rotationAroundY: (event avatar yaw).
	box := payload positioningBox.
	depth := box extentZ.

	"Check for forward obstructions.   Start the drop in front of them if we find any."
	pointer := event makePointer.
	pointer pointerPickTree: space
			excluding: { payload . event avatar . event avatar costume }.
	pointer selection frame ifNotNil: [
		depth := ((pointer selection distance * 0.95) min: (box radius * 4) ) 
					max: (box extentZ * 0.55)
	] ifNil: [
		"No forward obstructions... place the object somewhat near, just in case."
		depth := ((box radius) * 3) max: 6
	].
	vector :=  (event cameraTransform * (event selection rayTransform)) 
					lookAt negated.
	vector := vector * (depth).
	vector := (event cameraTransform translation + vector).

	payload translation: vector.
	space addChild: payload.
	payload allUpdateShape.

	^ payload dragAdapter fakeDragForCreation: event! !

!QDragAdapter class methodsFor: 'utilities' stamp: 'bgf 3/26/2010 11:36'!
drop: payload inSpace: space forEvent: event pointer: pointer
	"Place a new object (payload) in space to complete an interactive drop.
	 Positions the payload in the line of the event's ray,  in front of solid obstacles, 
	 but far enough away to be seen by the avatar.

	 The pointer should carry global pickTree: result for the event.
	 The payload is added to the space as the last step.
	 For implementations of the drag method #dropInSpace:pointer:"

	| xform frame pt normal vector box dv dp onSurface |

	"If the object has a native drag-adapter let it do the positioning for us."
	(self dragDrop: payload intoSpace: space forEvent: event) ifTrue: [ 
		event avatar  future captureCreateFrames: { payload } description: 'Add Object'.
		^ self ].

	"Use common code for handling objects dropped onto the ground."
	onSurface := (event selection ifNotNilDo: [ : e | e normal ifNotNilDo: [ :n | n y > 0.9 ] ifNil: [ false ]]).
	onSurface ifTrue: [
		^ space addModelInEditor: payload 
				inFrontOf: (event cameraTransform * (event selection rayTransform))
			 	for: (event avatar)
			 	doScale: false
	].

	payload localTransform: Matrix4x4 new.
	payload rotationAroundY: (event avatar yaw).
	box := payload positioningBox.

	"Devise a global position for the drop in space... along the drop ray,
 	 far enough from the avatar to see the entire dropped object..."
	vector :=  (event cameraTransform * (event selection rayTransform)) 
					lookAt negated.
	vector := vector * ((box radius * 2) + 9).
	vector := (event cameraTransform translation + vector).
	vector := vector - (box center).

	"If I'm dropped in front of a surface, translate into an appropriate
	 physical relationship, so long as the surface it isn't too far away."
	(pointer selection notNil 
	 & pointer selection frame notNil &	
	    pointer selection normal notNil)  
	 ifTrue: [
		frame := pointer selection frame.
		pt := pointer selection globalPoint.
		dv := vector distanceTo: event avatar globalPosition.		"Spatial best distance".
		dp := pt distanceTo: event avatar globalPosition.			"Distance to surface contact."
		dp > (dv *3) ifTrue: [
			"TODO - check ground-bound,
					 and use a higher tolerance for ground drops than for walls,
					 in the way the interactive drag code does."
			"The surface contact is too far away to make visual sense; just drop in space."
			payload translation: vector.

		] ifFalse: [
			"Use the surface contact."
			(dp < dv) ifTrue: [
				"The object will be too close - it will crowd into the avatar.  
				  So just magically shrink it to fit."
				false ifTrue: [payload scale: (dp/dv).	].
				"XXX - brad - suppressing this for now because it's more important to preserve the saved size of stored objects.  There is a problem for objects with small radius but large mesh coordinates (i.e. saved off-center), which completely confuses the calculations."
			].
			"Place the object on the surface."
			xform := frame globalTransform.
			normal := (xform transformVector: (pointer selection normal)) normalized.
			payload translateIntoContactWith: pt alongNormal: normal.
		].
	] ifFalse: [
		"Just drop it into space in the direction of the drop-site."
		payload translation: vector.
	].

	space addChild: payload.

	event avatar future captureCreateFrames: { payload } description: 'Add Object'.
! !

!QDragAdapter class methodsFor: 'instance creation' stamp: 'bgf 9/14/2006 19:35'!
forLuggage: aFrame grip: aSubframe
	"Make the given subframe behave as a rotation grip for the given frame."

	^ (self new) luggage: aFrame; grip: aSubframe; yourself.! !

!QDragAdapter methodsFor: 'accessing' stamp: 'bgf 1/19/2008 10:55'!
attractionRange

	"The factor by which I can move away from the initial grasp distance
	 in order to align to a surface.  
	 If non-nil, implies #isAttractedToSurfaces. 
	 (isAttractedToSurfaces gives the effect of attractionRange 1.6) "

	^ attractionRange! !

!QDragAdapter methodsFor: 'accessing' stamp: 'bgf 1/19/2008 10:34'!
attractionRange: aScalar
	"The factor by which I can move away from the initial grasp distance
	 in order to align to a surface.  If non-nil, overrides isAttractedToSurfaces;
	 if very large, I leap to most surfaces as a sticky attachment."
	attractionRange := aScalar! !

!QDragAdapter methodsFor: 'drag' stamp: 'bgf 8/16/2009 02:11'!
beginDrag: event 	
	"Beginning a drag.  
	 Using the graspPt, we derive luggageGraspPt:
	 the same point, but as an offset from the luggage's bounding box center,
	 and held in luggage frame coordinates.
	 This is used later to translate the luggage relative to a surface point."

	luggageGraspPt := graspPt.
	grip ~= luggage ifTrue: [
		luggageGraspPt := (luggage relativeTransform: grip) transformPoint: graspPt.
	].

	"We're saving this to prepare for TFrame #translateIntoContactWith:...,
	 which works relative to the positioning box."
	luggageGraspPt := luggageGraspPt - (luggage positioningBox center).

	"Default behavior - forward to the grip if wanted."
	^ super beginDrag: event ! !

!QDragAdapter methodsFor: 'drag' stamp: 'bgf 9/21/2006 12:11'!
clearDragState

	"Abandon and clear all state related to in-process drag activity."
	super clearDragState.
	self hiliteDropTarget:nil.
	luggageGraspPt := nil.! !

!QDragAdapter methodsFor: 'accessing' stamp: 'bgf 9/14/2006 12:56'!
contactOffset

	"When dropped into 'contact' with a surface, 
	 how far should I be offset from that surface?  
	 Default is zero, for actual contact.  
	 Useful for objects that need to be positioned adjacent to things but not touching them."

	^ contactOffset ifNil: [0] ifNotNil: [ contactOffset ].! !

!QDragAdapter methodsFor: 'accessing' stamp: 'bgf 9/14/2006 12:56'!
contactOffset: aScalar

	"When dropped into 'contact' with a surface, 
	 how far should I be offset from that surface?  
	 Default is zero, for actual contact.  
	 Useful for objects that need to be positioned adjacent to things but not touching them."

	contactOffset := aScalar.! !

!QDragAdapter methodsFor: 'accessing' stamp: 'bgf 9/14/2006 13:29'!
currDropTarget
	"The most recently identified drop target; see #identifyDropTarget"
	^ currDropTarget! !

!QDragAdapter methodsFor: 'drag' stamp: 'bgf 3/23/2010 09:44'!
doDrag: event

	"The per-event drag master routine.  For each pointer motion,
	 check for a picked frame in the line of the pointer ray
	 and if there is one, drag relative to it;
	 otherwise, drag (translate the luggage) in empty space."

	|  pointer excludes |

	grip root ifNil: [
		^ self  "Guard against some transitional situations."
	].

	"Is there a drop target indicated by the pointer?"
	pointer := event makePointer.
	
	excludes := self excludePickListForEvent: event.

	(pointer pointerPickTree: grip root excluding: excludes) ifTrue: [
		event avatar laserShow: (pointer selection globalPoint).
		self identifyDropTarget: (pointer selection frame).
		(self dragInFront: event pointer: pointer) ifTrue: [
			"There is a drop target, and we have been pulled up to be in front of it."
			^ self.
		]
	].

	"We're not over a drop target.  Just move spatially."
	pointer := event makePointer.
	(pointer frame: grip pickPlane: graspPt normal: cameraNorm) ifFalse:[
		"If pickplane fails, we can't reckon the drag position."
		^ self.
	].

	self dragInSpace: event pointer: pointer ! !

!QDragAdapter methodsFor: 'drag' stamp: 'bgf 8/16/2009 02:05'!
dragInFront: event pointer: pointer

	"If there's a surface under the pointer, find the ray intersection,
       to see if I should be positioned towards it.  
	 Return true if I find and apply a translation, false otherwise."

	| frame normal pt slop offset |

	graspPt ifNil: [ ^ false ].
	graspDistance ifNil: [ ^ false ].

	"Some drop sites allow drag/drop interaction but do not affect the spatial 
	 position of the dragged object... this allows large objects to be dragged without
	 pulling them 'in front' of the target."
	(currDropTarget respondsTo: #spatiallyTransparentToDrag) ifTrue: [
		currDropTarget spatiallyTransparentToDrag ifTrue: [ ^ false ].
	].
	pointer selection ifNil: [
		^ false.
	].
 	frame := pointer selection frame.
	normal := pointer selection normal.
	pt := pointer selection point.

	"Can we orient relative to this frame?"
	((frame notNil & normal notNil) and: [grip root isOrContains:frame parent]) ifTrue:  [
		"Yes, we can use this frame pick;
	       transform the surface contact point and normal to global."
		pt :=  frame globalTransform transformPoint: pt.
		normal := frame globalTransform transformVector: normal.
		normal := normal normalized.
	] ifFalse: [
		^ false.		"no usable surface under the drop point"
	].

	"We have a new position relative to a surface. 
	 Next, we determine how far away (how far from the user) it is,
	 and reject the surface-contact point if it's too far away.
       We allow ground-bound objects to slide very far on the ground."
	slop := self effectiveAttractionRange.
	((pointer selection distance) > (slop* graspDistance)) ifTrue: [
		"Using this point would push the object too far away.
		 Only allow this when sliding a groundBound object along the ground."
		(self isGroundBound not or: [(normal dot: (0@1@0)) < 0.8]) ifTrue: [
			"Fall back to regular dragging (ignoring the surface.)"
			^ false.
		].
	].

	"Finally, we determine how to move the luggage;
	 So far, 'pt' is a surface position for the handle's graspPt,
	 but we have to translate the luggage as a whole.
	 We record luggageGraspPt in #beginDrag:point:
	 so we can use it as an offset here, to derive a new target pt."
	offset := (self luggage globalTransform orientation transformPoint: luggageGraspPt).

	"But... keep the target point on the target plane:"
	offset := offset - (normal * (normal dot: offset)).
	pt := pt - offset.

	"Apply the contact offset (adjusts our distance from the target surface)"
	pt := pt  + (normal * self contactOffset).

	"pt is now the contact target point for the luggageGraspPt,
	 in global coordinates.  Normal is the normal to the target surface, in global orientation.
	 Shift the contact point and normal into the luggage parent's reference frame
	 and use them to translate the luggage."
	pt := self luggage parent globalToLocal: pt.
	normal := self luggage parent inverseGlobalTransform transformVector: normal.

	surfacePolicy caseOf: {
		[#upright ] -> [ 
			"Upright follows surfaces except on flats, where we do turn-to-face."
			normal y abs < 0.95
				ifTrue: [ self orientToSurface: normal  forRay: pointer ]
				ifFalse: [ self luggage turnToFaceNormal: (pointer lookAt y:0) ] ].
		[#uprightOnly ] -> [
			"Upright-only omits the turn-to-face aspect;  we recover a 0-pitch posture on flats."
			normal y abs < 0.95 
				ifTrue: [  self orientToSurface: normal  forRay: pointer ] 
				ifFalse: [ self luggage pitch: 0 ] ].
		[#viewer] -> [self luggage turnToFaceNormal: (pointer lookAt y:0)].
		[#surface] -> [self orientToSurface: normal forRay: pointer].
	} otherwise:[].
	self luggage translateIntoContactWith: pt alongNormal: normal.

	^ true.
	! !

!QDragAdapter methodsFor: 'drag' stamp: 'bgf 8/16/2009 02:09'!
dragInFrontOf: selection graspDistance: aDistance

	"If there's a frame in the selection, position myself visibly in front of it.
	 Return true if I find and apply a translation, false otherwise.
	 This is a variant on normal drag-drop, meant for use from the harness;
	 it does not rely on any previous drag-setup state being set in the adapter."

	| frame normal pt slop offset |

 	frame := selection frame.
	normal := selection normal.
	pt := selection point.

	"Can we orient relative to this selection ?"
	(frame notNil & normal notNil) ifTrue:  [

		"Yes, we can use this frame pick;
	       transform the surface contact point and normal to global."
		pt :=  frame globalTransform transformPoint: pt.
		normal := frame globalTransform transformVector: normal.
		normal := normal normalized.

	] ifFalse: [
		^ false.		"no usable surface under the drop point"
	].

	"We have a new position relative to a surface. 
	 Next, we determine how far away (how far from the user) it is,
	 and reject the surface-contact point if it's too far away.
       We allow ground-bound objects to slide very far on the ground."

	slop := self effectiveAttractionRange.

	((selection distance) > (slop* aDistance)) ifTrue: [
	
		"Using this point would push the object too far away.
		 Only allow this when sliding a groundBound object along the ground."
		(self isGroundBound not or: [(normal dot: (0@1@0)) < 0.8]) ifTrue: [
			"Fall back to regular dragging (ignoring the surface.)"
			^ false.
		].
	].

	"Finally, we determine how to move the luggage;
	 So far, 'pt' is a surface position for the grip,
	 but we have to translate the luggage as a whole."
	offset := (self luggage globalOrientation transformPoint: 0@0@0).

	"But... keep the target point on the target plane:"
	offset := offset - (offset * normal).
	pt := pt - offset.

	"Apply the contact offset (adjusts our distance from the target surface)"
	pt := pt  + (normal * self contactOffset).

	"pt is now the contact target point for the luggageGraspPt,
	 in global coordinates.  Normal is the normal to the target surface, in global orientation.
	 Shift the contact point and normal into the luggage parent's reference frame
	 and use them to translate the luggage."

	pt := self luggage parent globalToLocal: pt.
	normal := self luggage parent inverseGlobalTransform transformVector: normal.

	self luggage translateIntoContactWith: pt alongNormal: normal.

	^ true.
	! !

!QDragAdapter methodsFor: 'drag' stamp: 'bgf 8/3/2008 16:41'!
dragInSpace: event pointer: pointer

	"Drag substep: move through space following the pointer.
       This step is called once we've determined that there isn't a colliding surface to follow."

	| delta  newTranslation slop distance xform pt globalTrans |

	graspPt ifNil: [ ^ self ].
	graspDistance ifNil: [ ^ self ].

	delta := graspPt - pointer selectedPoint.
	newTranslation := (grip translation - (grip orientation transformPoint: delta)).

	"Bound the movement  to [originalDistance/slop, originalDistance*slop].   
	 We have to do this here (when no surface position is being tracked)
	 to recover objects that have slid far along the ground or pulled close to a surface
	 in a prior drag step."

	slop := (self effectiveAttractionRange + 3.0)/4.0.

	xform := event selection cameraTransform.
	globalTrans := grip parent  localToGlobal: newTranslation.

	(distance := xform translation distanceTo: globalTrans) > (slop*graspDistance) ifTrue: [
		pt := xform inverse transformPoint: globalTrans.
		pt := pt * (slop*graspDistance/distance).
		newTranslation  := grip parent globalToLocal: (xform transformPoint: pt).
	] ifFalse: [
		(distance * slop <  graspDistance) ifTrue:	[
			pt := xform inverse transformPoint: globalTrans.
			pt := pt * graspDistance/(slop*distance).
			newTranslation := grip parent globalToLocal: (xform transformPoint: pt).
	]].

	delta := (newTranslation - grip translation).

	"Apply the bounded movement to the luggage."

	luggage ~= grip ifTrue: [
		delta := (luggage parent relativeTransform: (grip parent)) transformVector: delta.
	].

	openSpacePolicy caseOf: {
		[#viewer] -> [self luggage turnToFaceNormal: (pointer lookAt y:0)].
	} otherwise:[].
	self luggage translation: (self luggage translation + delta).

! !

!QDragAdapter methodsFor: 'drag' stamp: 'bgf 8/30/2007 20:53'!
drop: event

	"Drag release: if there is a drop target indicated by the event,
	 then send myself to it as a drop payload.   Returns true 
	 if a drop target was detected and a drop sent; false otherwise."

	| pointer excludes |

	self hiliteDropTarget:nil.
	event selection ifNil: [ 
		^ false
	].
	grip root ifNil: [	
		^false.
	].

	"Determine the drop target, by running a pickTree;
      this relies on my being transparent to pick during the drag operation."
	pointer := event makePointer.
	excludes := self excludePickListForEvent: event.
	(pointer pointerPickTree: grip root  excluding: excludes) ifTrue: [
		self identifyDropTarget: pointer selection frame.
		currDropTarget ifNotNil: [
			self dropOn: currDropTarget avatar:(event avatar) selection: (pointer selection).
			^ true.
		].
	].

	"Fallthrough... no drop target... drop into space."
	self dropInSpace: event pointer: pointer.
	 ^ false.! !

!QDragAdapter methodsFor: 'drag' stamp: 'bgf 2/28/2010 09:23'!
dropInSpace: event pointer: pointer
	"Protocol method for drop, when a drop gesture is made and no drop target
 	 is identified.   Allows the grip to intercept this operation.
	 Otherwise,  if I'm not movable, retake my initial position.
	 Returns true if the grip intercepts and deals with the drop,
	 false if this drop is a no-op."
	(grip respondsTo: #dropInSpace:pointer:) ifTrue: [
		^ grip dropInSpace: event pointer: pointer.
	].

	isMovable ifFalse: [
		self snapBack.
	] ifTrue: [
		"If my luggage remembers a home position, update it."
		 (self luggage respondsTo: #savePosition) ifTrue: [
			self luggage savePosition.
		].
		(self luggage) movedOverSelection: (pointer selection) forEvent: event
	].
	^ false! !

!QDragAdapter methodsFor: 'drag' stamp: 'bgf 5/20/2008 11:21'!
dropOn: aTarget avatar: avatar selection: aTXSel

	"Offer my payload to the target as a drop delivery;
	if the target does not assume ownership of the payload, 
	snap back to starting transformation."

	| result |

	"Optional notification protocol (used for activity recording, not functionality.)"
	(self payload respondsTo: #willBeDroppedOn:byUser:) ifTrue: [
		self payload willBeDroppedOn: aTarget byUser: avatar
	].
	"Delegate to the grip if it handles this protocol."
	result := (grip respondsTo:#dropOn:avatar:selection:) ifTrue: [
				grip dropOn:aTarget avatar: avatar selection: aTXSel.
			] ifFalse: [
				aTarget consumeDrop: (self payload) avatar: avatar selection: aTXSel.
			].

	"If the recipient didn't assume control, return to the drag origination point."
	result ifFalse: [
		self snapBack.
	].
	^ result.
! !

!QDragAdapter methodsFor: 'accessing' stamp: 'bgf 1/19/2008 10:33'!
effectiveAttractionRange

	"The factor by which I can move away from the initial grasp distance
	 in order to align to a surface. "

	^ attractionRange ifNil: [
		self isAttractedToSurfaces ifTrue: [1.6] ifFalse: [1.0].
	].! !

!QDragAdapter methodsFor: 'drag' stamp: 'bgf 3/27/2010 08:55'!
endDrag: event withUndo: bUndo

	"Normal conclusion of a drag operation."
	(event avatar notNil and:[self drop: event]) ifTrue: [
		(grip respondsTo: #endDrop:) ifTrue: [
			grip endDrop: event.
		].
	] ifFalse: [
		super endDrag: event withUndo: bUndo.
	].! !

!QDragAdapter methodsFor: 'accessing' stamp: 'bgf 4/30/2008 21:13'!
excludeAvatars
	"Boolean - should all avatar costumes be excluded from my drag-over candidate picking?"
	^ excludeAvatars! !

!QDragAdapter methodsFor: 'accessing' stamp: 'bgf 4/30/2008 21:25'!
excludeAvatars: yesNo
	"Boolean - should all avatars be excluded from my drag-over candidate picking?"
	excludeAvatars := yesNo! !

!QDragAdapter methodsFor: 'drag' stamp: 'bgf 7/19/2008 16:07'!
excludePickListForEvent: event

	| camDist |

	"Should the drag/drop pick process include the avatar and costume?
	 We use heuristics:
	 -	If the drag is ground-bound (i.e. furniture) then don't involve any avatars. 
	 -    If excludeAvatars is true, involve no avatars.
	 -	Otherwise, if the camera-avatar distance (third-person setting)
	 	is close enough that the avatar is grayed-out and only partially seen,
		  then don't involve my own avatar.
	 - 	Otherwise, if far and not ground-bound, then consider my avatar a potential target
	      (This is used for file-cards to allow dropping a badge on myself, for instance.)

	This allows most drags to proceed independent of my avatar, while leaving
	a path to drop content onto the avatar from a sufficient third-person distance."
	 
	(grip isNil or: [ grip root isNil ]) 
		ifTrue: [ ^ { self luggage } ].

	(self excludeAvatars or: [self isGroundBound]) ifTrue: [
		| allAvatars |
		allAvatars := grip root allUsers.
		^ allAvatars,
		   (allAvatars collect: [ : av | av costume ]),
		   { self luggage }.
	].

	camDist := (event selection cameraTransform translation 
			     - event avatar translation) 
			length.

	"Heuristic - if we're 7' back we can see enough of the costume to be looking at it, not around it..."
	(camDist < 7) ifTrue: [
		^ {self luggage . event avatar costume}	"Exclude costume, but not inventory."
	].

	^ {self luggage}	! !

!QDragAdapter methodsFor: 'drag' stamp: 'bgf 1/21/2007 13:51'!
hiliteDropTarget: aTarget

	"Save the given target as currDropTarget, and hilite it if nonNil.
	 Unhilite the previous saved target if necessary."

	(currDropTarget = aTarget) ifFalse: [
		currDropTarget ifNotNil: [
			currDropTarget unhilite.
		].
		currDropTarget := aTarget.
		currDropTarget ifNotNil: [
			currDropTarget hilite: (QColors dropTargetHiliteColor).
		].
	].! !

!QDragAdapter methodsFor: 'drag' stamp: 'bgf 12/10/2006 12:54'!
identifyDropTarget: selectedFrame

	"Find the nearest ancestor of SelectedFrame returns true for #isDropTarget:,
	  hilite it using hiliteDropTarget (which saves it in currDropTarget) and return it.
	  Returns nil if there is no drop target in the ancestry of the selected frame."

	| target |

	self isDroppable ifFalse: [
		self hiliteDropTarget: nil.
		^ nil.
	].

	target := selectedFrame.
	[target notNil] whileTrue: [
		(target isDropTarget: (self payload)) ifTrue: [
			self hiliteDropTarget: target.
			^ target
		] ifFalse: [
			target := target parent.
		].
	].
	self hiliteDropTarget: nil.
	^ nil
! !

!QDragAdapter methodsFor: 'initialize' stamp: 'bgf 4/30/2008 21:13'!
initialize
	"By default, we are enabled, droppable, and deletable, but not movable."
	super initialize.

	self isDroppable: true.
	self isDeletable: true.
	self isMovable: false.
	self isAttractedToSurfaces:false.
	self isGroundBound: false.	
	self openSpacePolicy: #never.
	self surfacePolicy: #never.
	self excludeAvatars: false.! !

!QDragAdapter methodsFor: 'accessing' stamp: 'bgf 9/14/2006 12:56'!
isAttractedToSurfaces

	"If true, dragging me around favors being adjacent to objects near the destination;
	 the drag will push and pull the object into adjacency with other objects in the space"

	^ isAttractedToSurfaces! !

!QDragAdapter methodsFor: 'accessing' stamp: 'bgf 2/10/2007 21:26'!
isAttractedToSurfaces: aBool

	"If true, dragging me around favors being adjacent to objects near the destination;
	 the drag will push and pull the object into adjacency with other objects in the space,
	 so long as the motion doesn't take me more than 1.5 times my original distance 
	 from the avatar.   

	This is overridden if attractionRange: is nonNil."

	isAttractedToSurfaces := aBool! !

!QDragAdapter methodsFor: 'accessing' stamp: 'bgf 9/14/2006 12:56'!
isDeletable
	"Can the object be destroyed by '<Delete>'?"
	^ isDeletable! !

!QDragAdapter methodsFor: 'accessing' stamp: 'bgf 9/14/2006 12:56'!
isDeletable: aBoolean
	"Can the object be destroyed by <Delete>?"
	isDeletable := aBoolean! !

!QDragAdapter methodsFor: 'accessing' stamp: 'bgf 9/14/2006 12:56'!
isDroppable

	"If true, I can participate in logical drops on payloads;
	 if false, I can only be moved around... isMovable should be true in that case."

	^ isDroppable ! !

!QDragAdapter methodsFor: 'accessing' stamp: 'bgf 9/14/2006 12:56'!
isDroppable: aBoolean

	"If true, I can participate in logical drops on payloads;
	 if false, I can only be moved around... isMovable should be true in that case."

	isDroppable := aBoolean! !

!QDragAdapter methodsFor: 'accessing' stamp: 'bgf 9/14/2006 12:56'!
isGroundBound

	"If true, then I tend to prefer resting places on horizontal surfaces,
	 rather than floating, and I'll move away along the X axis when
	 the dragging ray hits a horizontal surface at any distance.
	 This is an unconstrained form of 'isAttractedToSurfaces', 
	 but only for horizontal surfaces.   Interactively, it lets me
	 slide objects around on the floor or ground, regardless of distance."

	^ isGroundBound! !

!QDragAdapter methodsFor: 'accessing' stamp: 'bgf 9/14/2006 12:56'!
isGroundBound:aBool

	"If true, then I tend to prefer resting places on horizontal surfaces,
	 rather than floating, and I'll move away along ray direction when
	 the pointer ray hits a horizontal surface at any distance.
	 This is an unconstrained form of 'isAttractedToSurfaces', 
	 but only for horizontal surfaces.   Interactively, it lets me
	 slide objects around on the floor or ground, regardless of distance.
	 Default is false."

	isGroundBound := aBool.! !

!QDragAdapter methodsFor: 'accessing' stamp: 'bgf 9/14/2006 12:56'!
isMovable
	"When dropped into a place in space where there is no receiving DropTarget,
	 do I take up the new position?   If isMovable is false, then I'll snap back
	 to my original pre-drag position.  If true, I'll stay where I've been dragged.

	 Default is false.   

	 This defines a basic user-level distinction... non-movable draggables 
	 can be dragged as part of logical operations, but not for physical positioning,
	 while movable objects behave more like independent physical entities."

	^ isMovable! !

!QDragAdapter methodsFor: 'accessing' stamp: 'bgf 9/14/2006 12:56'!
isMovable: aBoolean
	
	"If true, then I remain where I'm placed when dragged
	 (unless I'm dropped into a logical drop target that does a copy operation.)
	 If false, then I snap back to my pre-drag position at the end of any drag
	 (unless there is a drop target, and it 'consumes' me by supressing snap-back.)"

	isMovable := aBoolean! !

!QDragAdapter methodsFor: 'events' stamp: 'bgf 10/2/2007 23:39'!
keyStroke: event

	"Some default behavior - supply 'Delete'"

	self isEnabled ifFalse: [
		^ self
	].

	luggage ifNil: [ ^ self ].

	(event keyValue = 127 "Delete"  | event keyValue = 8 "Backspace") ifTrue: [

		self isDeletable ifTrue: [

			self cancelDrag: event.

			"ToDo - we should automatically install a garbage buffer
		       in every TIsland as we come to need, rather than relying on the Space."
			(luggage root respondsTo:#discard:) ifTrue:[
				self luggage root future discard: self luggage.
			] ifFalse: [
				self luggage future destroy.
			].
			luggage := nil.
		].

	] ifFalse: [

		super keyStroke: event.
	].! !

!QDragAdapter methodsFor: 'accessing' stamp: 'bgf 8/28/2007 12:22'!
openSpacePolicy
	"One of:
		#viewer		- orient to face towards the user who is dragging the object
		#none			- do not rotate the object.
	Controls how the object is oriented in the X-Z plane when dragging through open space,
	when no surface is behind the drag location influencing the position."
	
	^ openSpacePolicy! !

!QDragAdapter methodsFor: 'accessing' stamp: 'bgf 8/28/2007 12:19'!
openSpacePolicy: aSymbol
	"One of:
		#viewer
		#none
	Controls how the object is oriented when dragging through open space,
	when no surface is behind the drag location influencing the position."
	
	openSpacePolicy := aSymbol! !

!QDragAdapter methodsFor: 'drag' stamp: 'jdm 9/26/2007 15:42'!
orientToSurface: normal forRay: aRay

	"Simple case - not a horizontal surface."
	| up side m |
	(normal y abs < 0.95) ifTrue: [
		^ self luggage turnToFaceNormal: normal.
	].

	"For a horizontal surface, we want to orient out from it,
	 but also revolve on the new plane to suit the avatar;
	 imagine a sign or note... we want to be able to read it."
	up := aRay lookAt * (normal y sign negated).
	side := (normal cross: up) normalized negated.
	up := (side cross: normal) normalized negated.
	m := Matrix4x4 new.
	m a11: side x.
	m a21: side y.
	m a31: side z.
	m a12: up x.
	m a22: up y.
	m a32: up z.
	m a13: normal x.
	m a23: normal y.
	m a33: normal z.
	m translation: luggage translation.
	luggage localTransform: m.
! !

!QDragAdapter methodsFor: 'accessing' stamp: 'bgf 12/10/2006 13:10'!
payload
	"Implementing subclasses may override this to return the payload
	 that a drop should deliver to the #acceptDrop: recipient.
	 By default, the draggable presents its luggage's payload, or failing that, the luggage itself."

	^ (luggage respondsTo: #payload) ifTrue: [luggage payload] ifFalse: [luggage].
! !

!QDragAdapter methodsFor: 'persist' stamp: 'bgf 2/10/2007 13:40'!
persistOn: pStream

	"Draft code - drag adapter is not yet persisted independently."
	super persistOn:pStream.
	pStream storeBool: 'isDeletable' value: self isDeletable default: true.
	pStream storeBool: 'isMovable' value: self isMovable default: false.
	pStream storeBool: 'isGroundBound' value: self isGroundBound default: false.
	pStream storeBool: 'isAttractedToSurfaces' value: self isAttractedToSurfaces default: false.
	pStream storeNumber: 'attractionRange' value: attractionRange default: nil.
	pStream storeBool: 'isDroppable' value: self isDroppable default: true.
	pStream storeNumber: 'contactOffset' value: self contactOffset default: 0.! !

!QDragAdapter methodsFor: 'accessing' stamp: 'bgf 8/28/2007 12:21'!
surfacePolicy

	"One of 
		#viewer		- orient to face the ray from the view who is dragging the object
		#surface		- orient to face out along the surface normal of the drag
		#none			- do not adjust the orientation

	Controls how the object is oriented in the X-Z plane when dragging in front
	of a surface that influences its position."

	^ surfacePolicy! !

!QDragAdapter methodsFor: 'accessing' stamp: 'bgf 8/15/2009 09:45'!
surfacePolicy: aSymbol
	"One of 
		#viewer		- orient to face the ray from the view who is dragging the object
		#surface		- orient to face out along the surface normal of the drag
		#upright 	- orient to non-horizontal surfaces, face the viewer uprightly on horizontal planes.
		#none		- do not adjust the orientation

	Controls how the object is oriented in the X-Z plane when dragging in front
	of a surface that influences its position."

	surfacePolicy := aSymbol! !

QDragAdapter subclass: #QMotionAdapter
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Objects-Manipulation'!

!QMotionAdapter commentStamp: '<historical>' prior: 0!
A simple adapter that can be attached to a compound object,
in order to give the user a simple drag-manipulation handle
that moves the compound object. 

Interactively, a motion grip is movable, but not droppable;
dragging an object by a grip moves it through space,
but does not check drop targets or allow logical drops.

Visually, the default grip is a simple disk, 
oriented so the flat base is in the X-Y plane.
You can use #mesh:aFrame to make any part of the overall assembly
into the visual handle; it will be adopted as a child of the grip.

Use #luggage:aFrame, where aFrame is an ancestor
(container) of the handle, to bind the grip
to the container frame.
When the user drags the grip, the whole 'luggage' frame moves.

The luggage must be a frame tree ancestor of the grip
(the grip must be part of the thing it moves.)

Grips are not droppable (you cannot do logical drops into containers);
their only behavior is motion around the space.  

By default, isAttractedToSurfaces is true, but isGroundBound is false.
!

!QMotionAdapter class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QMotionAdapter class methodsFor: 'instance creation' stamp: 'bgf 9/14/2006 18:12'!
forLuggage: aFrame grip: aSubframe
	"Make the given subframe behave as a rotation grip for the given frame."

	^ (self new) luggage: aFrame; grip: aSubframe; yourself.! !

!QMotionAdapter class methodsFor: 'undo-redo' stamp: 'bgf 3/18/2010 16:46'!
verb
	"The label for this adapters's action, for undo-redo registration."
	^ 'Move'! !

!QMotionAdapter methodsFor: 'initialize' stamp: 'bgf 9/14/2006 13:56'!
initialize
	super initialize.
	self isMovable: true.
	self isAttractedToSurfaces: true.
	self isDroppable: false.



! !

!QGraspAdapter class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QGraspAdapter class methodsFor: 'defaults' stamp: 'bgf 1/21/2007 13:17'!
defaultDownColor
	^ QColors generalDragColor! !

!QGraspAdapter class methodsFor: 'defaults' stamp: 'bgf 1/21/2007 13:17'!
defaultOverColor
	^ QColors generalOverColor! !

!QGraspAdapter class methodsFor: 'undo-redo' stamp: 'bgf 3/18/2010 08:57'!
undoDrag: luggage toState: dragAdapterState
	"Implementation of undo/redo: restore the luggate posture to the given state."

	luggage ifNil: [ ^ nil ].
	luggage root ifNil:[ ^ nil  ].
	luggage applyGraspState: dragAdapterState.
! !

!QGraspAdapter class methodsFor: 'undo-redo' stamp: 'bgf 3/22/2010 09:22'!
undoableDrag: luggage oldState: oState newState: nState
	"Create the undoable record of a drag interaction.
	  The state records are created with   
		 #postDragState and #preDragState
	 on the drag-adapter."
	^ QBlockAction 
		undoBlock: [ : avatar |
			avatar future selectFrame: luggage.
			luggage island future 
				in: (self)  perform: #undoDrag:toState:   withArguments: { luggage . oState } ]
		redoBlock: [ : avatar |
			avatar future selectFrame: luggage.
			luggage island future
				in: (self)  perform: #undoDrag:toState:   withArguments: { luggage . nState }]
		description: (self verb)
		requiredFrames:  { luggage } ! !

!QGraspAdapter class methodsFor: 'undo-redo' stamp: 'bgf 3/2/2010 09:42'!
verb
	"For user-visible messages... terse generic verb for an interaction with this adapter."
	^ 'Drag'! !

!QGraspAdapter methodsFor: 'events' stamp: 'bgf 3/18/2010 16:48'!
announceForUndo: event
	"If the most recent undo is more than a no-op,
	 then announce it to the undo-mechanism."
	| pre post |

	event avatar ifNil: [ ^ self ].
	pre := self preDragState.
	post := self dragState.

	(pre isNil or: [ pre hasEqualElements: post ]) 
	 ifFalse: [
		event avatar
			recordAction:  #undoableDrag:oldState:newState:
	   		class: (self class) 
			arguments: { luggage. pre . post }.
	].! !

!QGraspAdapter methodsFor: 'drag' stamp: 'bgf 6/2/2007 14:05'!
beginDrag: event
	"Optional protocol method: notify the grip that a drag activity is beginning.
	The grip can reject the operation by returning false."
	(grip respondsTo: #beginDrag:) ifTrue: [
		^ grip beginDrag:event 
	].
	^ true! !

!QGraspAdapter methodsFor: 'accessing' stamp: 'bgf 9/13/2006 17:49'!
cameraNorm
	"During a drag, the initial pick normal (relative to the grip.)"
	^ cameraNorm! !

!QGraspAdapter methodsFor: 'drag' stamp: 'bgf 3/1/2010 09:03'!
cancelDrag: event
	
	"Terminate a drag without any permanent changes (jump back)"
	self snapBack.
	self clearDragState.

	isCancelledDrag := true.		"Suppresses  pointerClick:"

	((grip ~= self) & grip respondsTo: #cancelDrag:) ifTrue: [
		grip cancelDrag: event.
	].! !

!QGraspAdapter methodsFor: 'drag' stamp: 'bgf 3/15/2010 13:50'!
clearDragState

	"Clear state that arises during drag processing."

	grip ifNotNil: [
		grip unhilite.
		grip checkAlpha.
	].

	graspPt := nil.
	graspDistance := nil.
	isBeingDragged := false.
	isCancelledDrag := false.
	preDragState := nil.
! !

!QGraspAdapter methodsFor: 'accessing' stamp: 'bgf 9/15/2006 01:07'!
disable
	"Turn off sensitivity to drag/drop interactions and keyboard controls."

	isEnabled := false.
	self clearDragState.
	grip ifNotNil: [grip unhilite]. "Just in case."
	self stopEventScripts.
! !

!QGraspAdapter methodsFor: 'drag' stamp: 'bgf 9/16/2006 11:50'!
doDrag: event

	"The event is a pointer move event; the subclass 
	 should implement its basic drag motion behavior here.
	 The default implementation delegates drag: to the grip."

	grip ~= self ifTrue: [
		grip drag: event.
	].! !

!QGraspAdapter methodsFor: 'accessing' stamp: 'bgf 9/13/2006 13:10'!
downColor
	"The color to use to hilite while dragging."
	^ downColor! !

!QGraspAdapter methodsFor: 'accessing' stamp: 'bgf 8/18/2009 02:21'!
downColor: aColor
	"The color to use to hilite while dragging."
	downColor := aColor ifNotNil: [ aColor asVectorColor ]! !

!QGraspAdapter methodsFor: 'drag' stamp: 'bgf 3/18/2010 16:50'!
dragState
	"The current posture of the luggage, as a state array for undo-redo support."
	^ luggage graspState! !

!QGraspAdapter methodsFor: 'accessing' stamp: 'bgf 2/19/2007 23:04'!
enable
	"turn on sensitivity to drag/drop operations and keyboard controls"
	isEnabled := true.	
	self stopEventScripts.
	self startEventScriptsFor: grip.! !

!QGraspAdapter methodsFor: 'accessing' stamp: 'bgf 7/25/2008 00:30'!
enableClickOnly
	"Turn on sensitivity to pointerClick:, without enabling general drag interaction."
	isEnabled := false.
	
	self stopEventScripts.

	self runScript: #pointerDown: when: {grip . #pointerDown}.
	self runScript: #pointerMove: when: {grip . #pointerMove}.
	self runScript: #pointerUp: when: {grip . #pointerUp}.! !

!QGraspAdapter methodsFor: 'drag' stamp: 'bgf 3/27/2010 08:53'!
endDrag:event

	"Conclude the current drag operation. The event is the pointerUp.
	 The default implementation simply notifies the luggage that a user has changed the position."

	self endDrag: event withUndo: (self isMovable)! !

!QGraspAdapter methodsFor: 'drag' stamp: 'bgf 3/27/2010 08:53'!
endDrag:event withUndo: bUndo

	"Conclude the current drag operation. The event is the pointerUp.
	 The default implementation simply notifies the luggage that a user has changed the position."

	(luggage respondsTo: #savePosition) ifTrue: [
		luggage savePosition.
	].

	(grip respondsTo: #endDrag:) ifTrue:  [
		grip endDrag: event.
	].

	bUndo ifTrue: [ self announceForUndo: event ].! !

!QGraspAdapter methodsFor: 'events' stamp: 'bgf 8/18/2009 14:04'!
fakeDragForCreation: event 
	"Stage a no-op drag-motion sequence using the given single event;
	 this is done to trick the adapter into posing the object as if it had been moved into place.
	The event is expected to a be a valid Croquet mouse (over) event.
	Returns true if the placement was done, false if not."

	| pointer |

	(grip isNil | luggage isNil) ifTrue: [ ^ false ].
	pointer := event makePointer.
	pointer pointerPickTree: grip root excluding: { event avatar . event avatar costume }.
	pointer selection ifNil: [ ^ false ].
	pointer selection frame ifNil: [ ^ false ].

	"The engine relies on an object-relative pick;
	 if we didn't get one, fall back to heuristic face-the-drop placement.
	 This will mostly happen with porous objects whose center is unoccupied."
	(luggage isOrContains: (pointer selection frame)) ifFalse: [ ^ false ].
	event selection: (pointer selection).

	self  fakePointerDownForCreation: event.
	self  fakePointerMoveForCreation: event.
	self  fakePointerUpForCreation: event.
	
	^ true! !

!QGraspAdapter methodsFor: 'events' stamp: 'bgf 3/15/2010 13:51'!
fakePointerDownForCreation: event

	"We capture the pick vector and the exact coordinate
        where we're grasping the object, so we can drag it accurately."

	isCancelledDrag := false.
	preDragState := luggage graspState.
	cameraNorm := grip inverseGlobalOrientation transformPoint: 
			event selection cameraTransform lookAt negated.
	graspPt := event selection point.
	(event selection frame ~= grip) ifTrue: [
		graspPt :=
			(grip relativeTransform: event selection frame) transformPoint: graspPt.
	].
	graspDistance := event distance.
! !

!QGraspAdapter methodsFor: 'events' stamp: 'bgf 8/16/2009 01:14'!
fakePointerMoveForCreation: event
	isBeingDragged ifFalse: [
		graspPt ifNotNil: [
			isBeingDragged := (self beginDrag: event).
		].
	].
	self doDrag: event.! !

!QGraspAdapter methodsFor: 'events' stamp: 'bgf 3/27/2010 08:54'!
fakePointerUpForCreation: event

	self endDrag: event withUndo: false.
	self clearDragState.
! !

!QGraspAdapter methodsFor: 'accessing' stamp: 'bgf 9/13/2006 17:48'!
graspDistance
	"During a drag, the avatar-object initial distance."
	^ graspDistance! !

!QGraspAdapter methodsFor: 'accessing' stamp: 'bgf 9/13/2006 17:48'!
graspPt
	"During a drag, the original contact point, in the grip's coordinates."
	^ graspPt! !

!QGraspAdapter methodsFor: 'accessing' stamp: 'bgf 9/15/2006 01:05'!
grip
	"The frame being used as a manipulation handle by this adapter."
	^ grip! !

!QGraspAdapter methodsFor: 'accessing' stamp: 'bgf 9/14/2006 12:32'!
grip: aTFrame

	| enabled |
	enabled := self isEnabled.
	grip := aTFrame.
	self disable.
	enabled ifTrue: [ self enable ].! !

!QGraspAdapter methodsFor: 'initialize' stamp: 'bgf 8/28/2007 23:03'!
initialize
	super initialize.

	downColor := self class defaultDownColor.
	overColor := self class defaultOverColor.
	isBeingDragged := false.
	isCancelledDrag := false.
	luggage := nil.
	grip := nil.

	self enable.! !

!QGraspAdapter methodsFor: 'accessing' stamp: 'bgf 9/13/2006 13:10'!
isBeingDragged
	^ isBeingDragged! !

!QGraspAdapter methodsFor: 'accessing' stamp: 'bgf 9/13/2006 13:10'!
isEnabled 
	"Am I currently sensitive to mouse drag/drop operations?"
	^ isEnabled! !

!QGraspAdapter methodsFor: 'accessing' stamp: 'bgf 12/4/2006 14:14'!
isEnabled: aBool

	aBool ifTrue: [self enable] ifFalse: [self disable].! !

!QGraspAdapter methodsFor: 'accessing' stamp: 'bgf 3/8/2010 07:36'!
isMovable
	"Return true if this object can be spatially dragged to new positions 
	 (as opposed to logically dragged for semantic drag-drop operations
	 like opening a document into an application panel.)"

	^ true

	"For override by subclasses."! !

!QGraspAdapter methodsFor: 'events' stamp: 'bgf 9/13/2006 13:10'!
keyStroke: event

	"ESC cancels dragging."

	isBeingDragged ifTrue: [
		event keyValue = 27 "escape" ifTrue: [
			^ self cancelDrag:event.
		].
	].

! !

!QGraspAdapter methodsFor: 'accessing' stamp: 'bgf 9/13/2006 17:20'!
luggage
	"The object I represent a handle for... what is being moved when I'm dragged;
	 this should be myself, my frame parent, or some containing ancestor of mine."

	^ luggage! !

!QGraspAdapter methodsFor: 'accessing' stamp: 'bgf 9/13/2006 17:20'!
luggage: aTFrame

	"The object I represent a handle for... what is being moved when I'm dragged;
	 this must be a frame ancestor of all grips"

	luggage := aTFrame! !

!QGraspAdapter methodsFor: 'accessing' stamp: 'bgf 9/13/2006 13:10'!
overColor
	"The color to use to hilite on pointer enter."
	^ overColor! !

!QGraspAdapter methodsFor: 'accessing' stamp: 'bgf 12/19/2007 23:07'!
overColor: aColor
	"The color to use to hilite on pointer enter."
	overColor := aColor ifNotNil: [ aColor asVectorColor ]! !

!QGraspAdapter methodsFor: 'persist' stamp: 'bgf 12/31/2006 11:24'!
persistOn: pStream

	"Draft code... grasp adapters are not yet persisted independently."
	super persistOn:pStream.
	pStream storeObject: 'luggage' value: self luggage default:nil.
	pStream storeObject: 'grip' value: self grip default: nil.
	pStream storeBool: 'isEnabled' value: self isEnabled default: true.
	pStream storeColor: 'downColor' 
			value: self downColor 
			default: (self class defaultDownColor).
	pStream storeColor: 'overColor' 
			value: self overColor 
			default: (self class defaultOverColor).! !

!QGraspAdapter methodsFor: 'events' stamp: 'bgf 3/15/2010 13:59'!
pointerDown: ev
	"NOTE - some changes here will need reflection in #fakePointerDownForCreation:"
	self isEnabled ifFalse: [
		isBeingDragged := false.
		^ self ].

	"Dodge some observed problems getting events late, after destruction."
	(grip isNil | luggage isNil) ifTrue: [ ^ self ].
	luggage root ifNil: [ ^ self ].
	ev selection frame ifNil: [ ^ self ].

	"We capture the pick vector and the exact coordinate
       where we're grasping the object, so we can drag it accurately."

	isBeingDragged ifFalse: [

		isCancelledDrag := false.
		preDragState := luggage graspState.
		cameraNorm := grip inverseGlobalOrientation transformPoint: 
			ev selection cameraTransform lookAt negated.
		graspPt := ev selection point.
		(ev selection frame ~= grip) ifTrue: [
			graspPt :=
				(grip relativeTransform: ev selection frame) transformPoint: graspPt.
		].
		graspDistance := ev distance.
	].! !

!QGraspAdapter methodsFor: 'events' stamp: 'bgf 12/19/2007 22:51'!
pointerEnter: ev

	self isEnabled ifFalse: [
		^ self
	].
	isBeingDragged ifFalse: [
		self overColor ifNotNilDo: [ : color | grip hilite: color ]
	].! !

!QGraspAdapter methodsFor: 'events' stamp: 'bgf 9/30/2006 21:43'!
pointerLeave: ev

	isBeingDragged ifFalse: [
		grip unhilite.
	].! !

!QGraspAdapter methodsFor: 'events' stamp: 'bgf 8/16/2009 15:50'!
pointerMove: ev
	"NOTE - some changes here will need reflection in #fakePointerMoveForCreation:"

	self isEnabled ifFalse: [
		isBeingDragged := true.	"So click-only handling works"
		^ self
	].

	"In case we didn't get a distance with the initial down"
	graspDistance ifNil: [
		graspDistance := ev distance
	].
	isBeingDragged ifFalse: [
		graspPt ifNotNil: [
			isBeingDragged := (self beginDrag: ev).
			isBeingDragged ifTrue: [
				grip hilite: self downColor.
				grip checkAlpha.
			].
		].
	] ifTrue: [
		self doDrag: ev.
	].! !

!QGraspAdapter methodsFor: 'events' stamp: 'bgf 3/10/2010 13:54'!
pointerUp: ev
	"NOTE - some changes here will need reflection in #fakePointerUpForCreation:"
	| click |
	click := isBeingDragged not & (isCancelledDrag not).
	self isEnabled ifTrue: [		
		grip unhilite.
		isBeingDragged ifTrue:[
			self endDrag:ev.
		].
		self clearDragState.
	] ifFalse: [
		isBeingDragged := false.
	].
	click ifTrue: [
		"Click can be handled when partly disabled... see enableClickOnly"
		isBeingDragged := false.
		ev avatar ifNotNil: [		"I.e. don't convert implicit timeout up-events to clicks."
			ev nrClicks = 2 ifTrue: [
				grip signal: #doubleClick with: ev.
			] ifFalse:[ 
				(grip respondsTo:#pointerClick:) ifTrue: [
					grip pointerClick: ev ].
				grip signal: #pointerClick with: ev
			].
		].
	].! !

!QGraspAdapter methodsFor: 'drag' stamp: 'bgf 3/15/2010 13:51'!
preDragState
	^ preDragState
! !

!QGraspAdapter methodsFor: 'drag' stamp: 'bgf 3/15/2010 14:01'!
snapBack
	"Recover the original position of the luggage before the current drag operation."
	isBeingDragged ifTrue: [
		(preDragState  notNil) ifTrue: [
			luggage applyGraspState: preDragState
		].
	].! !

!QGraspAdapter methodsFor: 'events' stamp: 'bgf 9/13/2006 16:44'!
startEventScriptsFor: aTFrame

	self runScript: #pointerEnter: when: {aTFrame . #pointerEnter}.
	self runScript: #pointerDown: when: {aTFrame . #pointerDown}.
	self runScript: #pointerMove: when: {aTFrame . #pointerMove}.
	self runScript: #pointerUp: when: {aTFrame . #pointerUp}.
	self runScript: #pointerLeave: when: {aTFrame . #pointerLeave}.
	self runScript: #keyStroke: when: {aTFrame . #keyStroke}.! !

!QGraspAdapter methodsFor: 'events' stamp: 'bgf 9/13/2006 13:25'!
stopEventScripts

	self stopScript: #pointerEnter:.
	self stopScript: #pointerDown:.
	self stopScript: #pointerMove:.
	self stopScript: #pointerUp:.
	self stopScript: #pointerLeave:.
	self stopScript: #keyStroke:.! !

!QGraspAdapter methodsFor: 'accessing' stamp: 'bgf 5/26/2007 20:28'!
toggleEnabled
	self isEnabled: (self isEnabled not)! !

TObject subclass: #QJobHandler
	instanceVariableNames: 'owner spaces listenerName listenerRefs'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Objects-Jobs'!

!QJobHandler commentStamp: '<historical>' prior: 0!
Maintains signal listeners for all known job classes,
and relays job request signals to the job class for processing.

This serves as the signal-based glue that hears requests from the replicated space,
and invokes the handling code in the client context (outside the replicated space.)
!

!QJobHandler class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QJobHandler class methodsFor: 'class variables' stamp: 'bgf 3/11/2007 22:35'!
jobClasses 

	"The set of all job classes;
	 each job class is registered to handle the signal defined by its own trigger
	 (which is its own class name)."

	^QParticipantJob allSubclasses
		! !

!QJobHandler methodsFor: 'initialize-release' stamp: 'bgf 8/21/2008 14:24'!
assuredListener: anIsland

	"For each island, each participant has a signal listener
       registered in its island globals - it is signalled to request jobs
	 that must be handled by the particular user's harness,
	 like file-save jobs.

	 Note that this method may return nil if the island far-ref is invalidated by disconnect,
	 or if the listener creation times out."

	| listener ln |

	ln := self listenerName. 
	ln ifNil: [ self error: 'Premature job handler setup - no owner/session' ].

	listener := anIsland future at: ln.
	(listener waitTimeoutSeconds: (Socket standardTimeout)) ifFalse: [
		listener := nil.	
	].
	anIsland isValidFarRef ifFalse: [ ^ nil ].	"Disconnected..."
	(listener isNil or: [listener value isNil]) ifTrue: [
		"Might be a revoked listener from a previous timeout."
		listener := listenerRefs at: anIsland value ifAbsent: [nil].
		listener ifNil:[
			QLogger logInfo: 'Making job listener for ', (anIsland get:#name).
			listener := anIsland future new: QSessionJobListener.
			listener wait.
		].
		((anIsland future at: ln put: listener) waitTimeoutSeconds: (Socket standardTimeout)) ifFalse: [
			QLogger logWarning:'Job listener abandoned (island did not resolve): ', anIsland.
			^ nil 
		].
		listenerRefs at: anIsland value put: listener value.
	].
	^ listener! !

!QJobHandler methodsFor: 'initialize-release' stamp: 'jcg 8/30/2008 00:19'!
currentSpace: curSpace

	"Current space has changed;
	 register for triggers from the new space and avatar.

	 Note - when we add partial logout (closing active spaces w/out full logout)
	 the job handler will need to be informed.)"

	| stale |

	(owner isNil or: [curSpace isValidFarRef not]) ifTrue: [
		spaces := Set new.
		self stopScript: #gotSignal:job:.
		^ self 
	].
	curSpace wait.

	stale := spaces reject: [ :s | s isValidFarRef ].
	stale ifNotEmpty: [	
		"Rebuild the whole signal handling state."
		spaces add: curSpace.
		self resetAllListeners.

	] ifEmpty: [
		"Just add this to the set of spaces we're listening within"
		(spaces includes: curSpace) ifFalse: [
			spaces add: curSpace.
			self setupSpace: curSpace.
		].
	].
! !

!QJobHandler methodsFor: 'initialize-release' stamp: 'bgf 6/30/2007 19:36'!
destroy

	"Extract my listeners, stop all scripts, forget all spaces."

	listenerName ifNotNil: [
		spaces do: [ :spc|
			[ spc island future removeGlobal: listenerName ] on: Error do: []
		].
	].
	spaces := Set new.
	listenerRefs := Dictionary new.
	listenerName := nil.
	self myScripts do:[:script| script terminateScript].
	self myScripts: nil.
	owner := nil.

! !

!QJobHandler methodsFor: 'initialize-release' stamp: 'bgf 3/23/2009 17:30'!
discardIsland: anIslandRef
	"Drop the references to this island"
	listenerRefs removeKey: anIslandRef value ifAbsent: [].
	self resetAllListeners.! !

!QJobHandler methodsFor: 'initialize-release' stamp: 'bgf 4/7/2009 13:40'!
doResetAllListeners
	
	"Rebuild the whole signal handling state."
	spaces := spaces select: [ :s | s isValidFarRef ].
	spaces do: [ :spc |
		spc isValidFarRef ifTrue: [ self setupSpace: spc ].
	]
	! !

!QJobHandler methodsFor: 'job handling' stamp: 'bgf 12/20/2009 10:41'!
gotSignal:aSymbol job: aJobRef

	"A job-request signal has been received; relay it to the job class for handling."

	aJobRef isValidFarRef ifFalse: [
		QLogger logInfo: 'Ignoring ', aSymbol, ' for stale job'.
		^ self
	].
	aJobRef whenResolved: [
		aJobRef isValidFarRef ifFalse: [
			QLogger logInfo: 'Discarding ', aSymbol, ' for stale job'.
			^ self
		].
		(aJobRef get:#class) ifNotNilDo:[ :klass | klass startJob: aJobRef forClient: owner ].
	].! !

!QJobHandler methodsFor: 'initialize-release' stamp: 'bgf 6/30/2007 17:31'!
initialize
	spaces := Set new.
	listenerRefs := Dictionary new.! !

!QJobHandler methodsFor: 'accessing' stamp: 'bgf 6/24/2007 15:14'!
listenerName
	|  |
	listenerName ifNil: [ 
		owner ifNil: [ ^ nil ].
		listenerName := ((owner session userID hex), '-JobListener') asSymbol
	].
	^ listenerName! !

!QJobHandler methodsFor: 'events' stamp: 'bgf 8/23/2007 12:37'!
onSpaceChange
	"Update the listeners for requests from the owner's active space."
	owner exitTripped ifFalse: [
		self currentSpace: owner activeSpace 
	]! !

!QJobHandler methodsFor: 'events' stamp: 'bgf 8/23/2007 10:53'!
onUnhandledJob: aParticipantJobRef

	"A job that was requested went unhandled; the job class may want to notify."
	aParticipantJobRef isValidFarRef ifTrue: [
		(aParticipantJobRef get:#class) notifyUnhandled: owner.
	]! !

!QJobHandler methodsFor: 'initialize-release' stamp: 'bgf 3/11/2007 23:24'!
owner: aParticipantUI

	"The main participant UI for which we are handling jobs;
	 gives us access to the mainView, the session, etc."

	owner := aParticipantUI.
	self startScript: #onSpaceChange when: {owner . #activeSpaceChanged}.! !

!QJobHandler methodsFor: 'initialize-release' stamp: 'bgf 4/7/2009 13:37'!
resetAllListeners
	
	"Rebuild the whole job-processor signal handling state for the current set 
	 of connected spaces."

	self stopScript: #gotSignal:job:.
	self stopScript: #onUnhandledJob:.

	"The main work involves resolving far refs which may be in transition...
	 background so it is stoppable and so it does not cause reconnect-recovery delays."
	self stopScript: #doResetAllListeners.
	self startScript: #doResetAllListeners.! !

!QJobHandler methodsFor: 'events' stamp: 'jcg 8/30/2008 00:18'!
setupSpace: spc

	"For  each job class, there is one triggering signal,
	 raised on the space or user-specific listener, to request processing.
	 Here we set up signal handlers for those signals, for the given space's island."

	| listener |

	"This is all asynchronous, so we watch for invalidated far refs (disconnects)"
	spc isValidFarRef ifFalse: [ ^ self ].
	listener := self assuredListener: spc island.
	listener isValidFarRef ifFalse: [ ^ self ].

	self startScript: #onUnhandledJob: when: {spc island . #unhandledJob}.
	self class jobClasses
		select: [:jc | jc isEnabledForClient: owner] 
		thenDo: [ :each |
			self startScript: #gotSignal:job: withArguments: {each trigger}
				when: {spc . each trigger}.
			self startScript: #gotSignal:job: withArguments: {each trigger}
				when: {listener . each trigger}.
	].! !

TObject subclass: #QMeetingState
	instanceVariableNames: 'objectID leaderLogin audienceIsMuted audienceStayPut isEditBlocked isPrivateTextBlocked allowedToSpeak wantsToSpeak meetingPlace'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Objects-Space'!

!QMeetingState commentStamp: '<historical>' prior: 0!
An on-island one-per-space singleton 
to carry the various state flags and properties of a meeting-session.
!

!QMeetingState class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QMeetingState methodsFor: 'accessing' stamp: 'bgf 7/25/2009 21:12'!
addAllowedToSpeak: login
	(allowedToSpeak includes: login) ifFalse: [
		allowedToSpeak add: login.
		self removeWantsToSpeak: login.
		self signal: #addAllowedToSpeak with: login
	].! !

!QMeetingState methodsFor: 'accessing' stamp: 'bgf 6/30/2009 23:57'!
addWantsToSpeak: login
	(wantsToSpeak includes: login) ifFalse: [
		wantsToSpeak add: login.
		self signal: #addWantsToSpeak with: login
	].! !

!QMeetingState methodsFor: 'accessing' stamp: 'bgf 7/1/2009 00:00'!
allowedToSpeak
	"Set of user logins, of those currently permitted to speak;
	 only applicable when audienceIsMuted."
	^ allowedToSpeak asArray
! !

!QMeetingState methodsFor: 'accessing' stamp: 'bgf 6/25/2009 19:18'!
audienceIsMuted
	"If true, the leader wants all participants muted except those she un-mutes."
	^ audienceIsMuted ifNil: [ false ]! !

!QMeetingState methodsFor: 'accessing' stamp: 'bgf 7/17/2009 14:00'!
audienceIsMuted: bMuteAudience
	"If true, the leader wants all participants muted except those she un-mutes."
	audienceIsMuted = bMuteAudience ifFalse: [
		audienceIsMuted := bMuteAudience.
		audienceIsMuted ifFalse: [
			allowedToSpeak do: [ : user |
				self signal: #removeAllowedToSpeak with: user  
			].
			allowedToSpeak := Set new.
		].
		self signal: #audienceIsMutedChanged with: bMuteAudience
	]! !

!QMeetingState methodsFor: 'accessing' stamp: 'bgf 6/26/2009 22:07'!
audienceStayPut
	"If true, the audience stays where the leader sends them (no free navigation.)"
	^ audienceStayPut ifNil: [false]! !

!QMeetingState methodsFor: 'accessing' stamp: 'bgf 6/26/2009 22:26'!
audienceStayPut: bStayPut
	"If true, the leader wants all participants to stay where he last placed them.."
	audienceStayPut = bStayPut ifFalse: [
		audienceStayPut := bStayPut.
		self signal: #audienceStayPutChanged with: bStayPut
	]! !

!QMeetingState methodsFor: 'accessing' stamp: 'bgf 7/5/2009 11:14'!
clearSpeakerLists
	| hadSome |

	hadSome := allowedToSpeak notEmpty or: [ wantsToSpeak notEmpty ].
	hadSome ifTrue: [
		allowedToSpeak := Set new.
		wantsToSpeak := Set new.
		self signal: #speakerListsCleared.
	].! !

!QMeetingState methodsFor: 'initialize-release' stamp: 'bgf 6/30/2009 23:58'!
initialize
	super initialize.

	"Register me in the island's nameMap for immediate get-ability."
	Processor activeIsland register:self name: (objectID := TObjectID new).

	allowedToSpeak := Set new.
	wantsToSpeak := Set new.! !

!QMeetingState methodsFor: 'accessing' stamp: 'bgf 8/18/2009 23:51'!
isAllFollowMe
	^ meetingPlace notNil and: [ meetingPlace isFollowMePlace ]! !

!QMeetingState methodsFor: 'accessing' stamp: 'bgf 6/26/2009 23:49'!
isEditBlocked
	"If true, the leader wants all participants prevented from dropping in content."
	^ isEditBlocked ifNil: [ false ]! !

!QMeetingState methodsFor: 'accessing' stamp: 'bgf 6/26/2009 23:50'!
isEditBlocked: bBlocked
	"If true, the leader wants all participants prevented from dropping in content."
	isEditBlocked = bBlocked ifFalse: [
		isEditBlocked := bBlocked.
		self signal: #isEditBlockedChanged with: bBlocked
	]! !

!QMeetingState methodsFor: 'accessing' stamp: 'bgf 8/8/2009 00:48'!
isPrivateTextBlocked
	"If true, the leader wants all participants prevented from dropping in content."
	^ isPrivateTextBlocked ifNil: [ false ]! !

!QMeetingState methodsFor: 'accessing' stamp: 'bgf 8/8/2009 00:50'!
isPrivateTextBlocked: bBlocked
	"If true, the leader wants all participants prevented from doing private chat."
	isPrivateTextBlocked = bBlocked ifFalse: [
		isPrivateTextBlocked := bBlocked.
		self signal: #isPrivateTextBlockedChanged with: bBlocked
	]! !

!QMeetingState methodsFor: 'accessing' stamp: 'bgf 6/25/2009 17:41'!
leaderLogin
	"Nil when there is no leader"
	^ leaderLogin! !

!QMeetingState methodsFor: 'accessing' stamp: 'bgf 8/26/2009 03:28'!
leaderLogin: stringOrNil
	| old |
	leaderLogin = stringOrNil ifFalse: [
		old := leaderLogin.
		leaderLogin := stringOrNil.	
		self signal: #leaderLoginChanged with: stringOrNil.
		"No matter how leadership has changed, if there is no leader, 
		 release all the leader-bound control flags."
		leaderLogin 
			ifNil: [ self leaderResigned ]
			ifNotNil:  [ self isAllFollowMe ifTrue: [ self meetingPlace: nil ] ].
		old 
			ifNotNil: [
				self removeWantsToSpeak: old; 
					removeAllowedToSpeak: old ]
	].! !

!QMeetingState methodsFor: 'initialize-release' stamp: 'bgf 5/6/2010 11:36'!
leaderResigned
	"Reset to natural posture when the leader ceases to be the leader."
	self isEditBlocked: false.
	self isPrivateTextBlocked: false.
	self audienceIsMuted: false.
	self audienceStayPut: false.
	self isAllFollowMe ifTrue: [ self meetingPlace: nil ].
	self clearSpeakerLists.
! !

!QMeetingState methodsFor: 'accessing' stamp: 'bgf 7/4/2009 15:04'!
meetingPlace
	"If not nil, the place to which all participants should be sent (or returned after navigating.)"
	^ meetingPlace ! !

!QMeetingState methodsFor: 'accessing' stamp: 'bgf 8/12/2009 21:19'!
meetingPlace: aPlace
	"If not nil, the place to which all participants should be sent (or returned after navigating.)"
	| oldMeetingPlace |

	meetingPlace = aPlace ifFalse: [
		oldMeetingPlace := meetingPlace.
		meetingPlace := aPlace.
		self signal: #meetingPlaceChanged with: aPlace.

		"If this was follow-the-leader, release everyone."
		self space ifNotNilDo: [ : space |
			(oldMeetingPlace notNil
			  and: [ oldMeetingPlace avatarToFollow notNil ])
			ifTrue: [
				space allUsers do: [ : avatar | 
					(avatar isAtPlace: oldMeetingPlace) ifTrue: [ 
						avatar clearCurrentPlace.
					]]]]].! !

!QMeetingState methodsFor: 'accessing' stamp: 'bgf 6/25/2009 15:55'!
objectID
	^ objectID! !

!QMeetingState methodsFor: 'accessing' stamp: 'bgf 8/22/2009 15:28'!
registerLeader: loginString
	QLogger logInfo: 'Setting ', (loginString ifNil: ['no-one']), ' as meeting leader.'.
	leaderLogin = loginString ifFalse: [
		self leaderLogin: loginString.
		loginString ifNotNil: [
			"Clear leader-control flags if I'm leader."
			self removeWantsToSpeak: loginString.
			self removeAllowedToSpeak: loginString.
		].
	].! !

!QMeetingState methodsFor: 'accessing' stamp: 'bgf 8/22/2009 15:27'!
releaseLeader: loginString
	"The given login has gone away or released their leadership status."
	loginString ifNotNil: [
		leaderLogin = loginString ifTrue: [
			QLogger logInfo: '', loginString, ' is no longer leading the meeting.'.
			self leaderLogin: nil
		]].! !

!QMeetingState methodsFor: 'accessing' stamp: 'bgf 7/1/2009 02:35'!
removeAllowedToSpeak: login
	(allowedToSpeak includes: login) ifTrue: [
		allowedToSpeak remove: login.
		self signal: #removeAllowedToSpeak with: login
	].! !

!QMeetingState methodsFor: 'accessing' stamp: 'bgf 7/2/2009 14:36'!
removeWantsToSpeak: login
	"The given user no longer wants to speak "
	(wantsToSpeak includes: login) ifTrue: [
		wantsToSpeak remove: login.
		self signal: #removeWantsToSpeak with: login
	].! !

!QMeetingState methodsFor: 'accessing' stamp: 'bgf 6/25/2009 15:51'!
space
	"The TSpace I govern"
	^ (self island resourcesAt: TSpace resourceType) first! !

!QMeetingState methodsFor: 'accessing' stamp: 'bgf 7/1/2009 00:00'!
wantsToSpeak
	"Logins of users who wish to speak."
	^ wantsToSpeak asArray! !

TObject subclass: #QMouseoverHiliter
	instanceVariableNames: 'luggage overColor isEnabled statusString floaterString'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Objects-General'!

!QMouseoverHiliter commentStamp: '<historical>' prior: 0!
A simple accessory to provide mouse-over hiliting for any TFrame.

To use, simply create (and keep a reference to) an instance:
	hiliter := QMouseoverHiliter forFrame: aTFrame.
this installs mouse over hiliting in aTFrame.
The hilite color can be  set:
	hiliter overColor: aColor

Hiliting can be suspended and resumed 
	hiliter disable
	hiliter enable

The hiliter itself is not a TFrame; it installs runScripts
as event handlers on the tframe it is created for.





!

!QMouseoverHiliter class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QMouseoverHiliter class methodsFor: 'instance creation' stamp: 'bgf 10/1/2006 19:32'!
forFrame: aTFrame
	^ (self new) luggage: aTFrame; enable; yourself! !

!QMouseoverHiliter methodsFor: 'accessing' stamp: 'bgf 10/1/2006 19:28'!
disable
	"Turn off sensitivity to drag/drop interactions and keyboard controls."

	isEnabled := false.
	luggage ifNotNil: [ 
		luggage unhilite.
		self stopEventScripts.
	].! !

!QMouseoverHiliter methodsFor: 'accessing' stamp: 'bgf 2/19/2007 23:04'!
enable
	"turn on sensitivity to drag/drop operations and keyboard controls"
	isEnabled := true.	
	self stopEventScripts.
	self startEventScriptsFor: luggage.! !

!QMouseoverHiliter methodsFor: 'accessing' stamp: 'bgf 12/20/2007 13:37'!
floaterString
	"The string (or block) to signal balloon help during mouseover"
	^ floaterString! !

!QMouseoverHiliter methodsFor: 'accessing' stamp: 'bgf 12/20/2007 14:18'!
floaterString: aString
	"The string to signal balloon help during mouseover"
	floaterString := aString! !

!QMouseoverHiliter methodsFor: 'initialize' stamp: 'bgf 1/21/2007 13:17'!
initialize
	super initialize.

	overColor := QColors controlOverColor.
	luggage := nil.
	self disable.! !

!QMouseoverHiliter methodsFor: 'accessing' stamp: 'bgf 10/1/2006 19:19'!
isEnabled 
	"Am I currently sensitive to mouse drag/drop operations?"
	^ (isEnabled & luggage notNil)! !

!QMouseoverHiliter methodsFor: 'accessing' stamp: 'bgf 10/1/2006 19:13'!
luggage
	"The object I represent a handle for... what is being moved when I'm dragged;
	 this should be myself, my frame parent, or some containing ancestor of mine."

	^ luggage! !

!QMouseoverHiliter methodsFor: 'accessing' stamp: 'bgf 10/1/2006 19:25'!
luggage: aTFrame

	"The object I hilite."
	luggage := aTFrame.
! !

!QMouseoverHiliter methodsFor: 'accessing' stamp: 'bgf 10/1/2006 19:13'!
overColor
	"The color to use to hilite on pointer enter."
	^ overColor! !

!QMouseoverHiliter methodsFor: 'accessing' stamp: 'bgf 10/1/2006 19:13'!
overColor: aColor
	"The color to use to hilite on pointer enter."
	overColor := aColor asVectorColor! !

!QMouseoverHiliter methodsFor: 'events' stamp: 'bgf 7/21/2009 13:36'!
pointerDown: ev

	self isEnabled ifTrue: [
		luggage future unhilite.
	].! !

!QMouseoverHiliter methodsFor: 'events' stamp: 'bgf 11/13/2008 13:13'!
pointerEnter: ev

	ev avatar ifNil: [ ^ self ].
	self isEnabled ifTrue: [
		luggage hilite: self overColor.
		statusString ifNotNil: [
			ev avatar future signal: #infoMessage with: statusString.
		].
		floaterString ifNotNil: [
			ev avatar future showFloatMessage: floaterString forEvent: ev.
		].
	].
! !

!QMouseoverHiliter methodsFor: 'events' stamp: 'bgf 10/1/2006 19:17'!
pointerLeave: ev

	self isEnabled ifTrue: [
		luggage unhilite.
	].! !

!QMouseoverHiliter methodsFor: 'events' stamp: 'bgf 11/13/2008 13:13'!
pointerOver: ev

	ev avatar ifNil: [ ^ self ].
	self isEnabled ifTrue: [
		floaterString ifNotNil: [
			ev avatar future showFloatMessage: floaterString forEvent: ev.
		].
	].
! !

!QMouseoverHiliter methodsFor: 'events' stamp: 'bgf 7/21/2009 13:33'!
startEventScriptsFor: aTFrame

	self runScript: #pointerEnter: when: {aTFrame . #pointerEnter}.
	self runScript: #pointerLeave: when: {aTFrame . #pointerLeave}.
	self runScript: #pointerOver: when: {aTFrame . #pointerOver}.
	self runScript: #pointerDown: when: {aTFrame . #pointerDown}.! !

!QMouseoverHiliter methodsFor: 'accessing' stamp: 'bgf 12/20/2007 14:18'!
statusString
	"The string to signal to the status line during mouseover"
	^ statusString! !

!QMouseoverHiliter methodsFor: 'accessing' stamp: 'bgf 12/20/2007 14:18'!
statusString: aString

	statusString := aString! !

!QMouseoverHiliter methodsFor: 'events' stamp: 'bgf 10/1/2006 19:17'!
stopEventScripts

	self stopScript: #pointerEnter:.
	self stopScript: #pointerLeave:.! !

TObject subclass: #QOrgServices
	instanceVariableNames: 'owner controller orgRecord presenceManager lobbySessionID errorTripped'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Objects-User'!

!QOrgServices commentStamp: '<historical>' prior: 0!
I am the client-side API for an organization-wide system island
and the services it contains.

An OrgServices session starts when a service manager is connected during login.
It is used throughout a logged-in forums session to observe
the location of other participants.     Using a shared island for this
allows us to update (for instance) the Users... location information
using signal handlers, rather than by polling the service manager.

The OrgServices object does the listening to signals in the service island,
and re-raises them on itself... this allows the UI to register for events
from the OrgServices object, rather than individual island-side objects.
This makes it possible to handle reconnects locally, without updating all
the various event subscribers in UI tools.   
!

!QOrgServices class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QOrgServices class methodsFor: 'instance creation' stamp: 'bgf 10/21/2007 11:11'!
forClient: aClient serviceManager: aSvcManager

	"Initialize and connect to the org-services island 
	 for the organization that the service manager is pointing to.
	 May raise errors if there is no support for org-services here,
	 or if the communication setup fails."

	^ self new initializeForClient: aClient serviceManager: aSvcManager! !

!QOrgServices class methodsFor: 'locating' stamp: 'bgf 6/21/2009 10:11'!
lobbyURL
	"The psuedo url used throughout location services to record presence in the 'Lobby',
	 i.e. in the service chooser view of the organization."
	^ '/lobby'! !

!QOrgServices class methodsFor: 'locating' stamp: 'hrs 8/9/2009 18:10'!
requestUserLocation: aUserAccountName fromManager: svcManager

	"Ask the service manager for the location record
	 where the given (offline) user last reported themselves.
	 This fetches the locator.xml file deposited in #reportLocation.
	 The record is { url, epochUTCSeconds } where the time may be nil if not recorded;
	 the URL will be 'Unknown' if no location record is found."

	| xmlData |
	[
		xmlData := svcManager getFile: '/users/', aUserAccountName, '/public/locator.xml'.
		^QPresenceRecord fromXML: xmlData asString for: aUserAccountName.
	] on: Error do: [ :ex |
		^ QPresenceRecord login: aUserAccountName url: 'Unknown'.
	].

! !

!QOrgServices class methodsFor: 'locating' stamp: 'hrs 8/9/2009 18:05'!
sendLocation: urlString forUser: aLogin toManager: serviceMgr

	"Update the location record on the server (for #offlineLocationRecord:)"

	| utcSeconds xmlData |
	utcSeconds := DateAndTime now asUTCSeconds.
	[
		"We don't compress - this is a tiny datum."
		xmlData := String streamContents: [:s |
			(XMLWriter on: s) 
				tag: 'location' pcData: urlString;
				tag: 'clientUTC' pcData: utcSeconds printString.
		].
	serviceMgr put: xmlData
				  	name: '/users/', aLogin, '/public/locator.xml'
					backup: false
				 	informing: nil. 
	] on: Error do: [ :ex |
		QLogger logWarning: 'Could not record location on server:\ 'withCRs, ex description.
	]! !

!QOrgServices class methodsFor: 'locating' stamp: 'hrs 8/21/2009 13:24'!
sendOrg: orgRecord forUser: aLogin toManager: serviceMgr

	"Update the org record on the server."

	| utcSeconds xmlData |
	utcSeconds := DateAndTime now asUTCSeconds.
	[
		"We don't compress - this is a tiny datum."
		xmlData := String streamContents: [:s |
			(XMLWriter on: s) 
				tag: 'org' pcData: orgRecord name;
				tag: 'clientUTC' pcData: utcSeconds printString.
		].
	serviceMgr put: xmlData
				  	name: '/users/', aLogin, '/public/lastOrg.xml'
					backup: false
				 	informing: nil. 
	] on: Error do: [ :ex |
		QLogger logWarning: 'Could not record org on server:\ 'withCRs, ex description.
	]! !

!QOrgServices methodsFor: 'meeting' stamp: 'bgf 3/23/2010 17:24'!
assignLeadership: sessionID
	"Become the leader of the current forum.
	 (Returns true/success or false/fail)"
	owner ifNil: [
		QLogger logWarning: 'beMeetingLeader, but os has no owner.'.
		^ false
	].
	owner readyForEvents ifFalse: [
		 owner statusAlert: 'Services not ready. Cannot set leader.'.
		^ false
	].
	presenceManager isValidFarRef ifFalse: [
		owner statusAlert: 'Not connected. Cannot set leader.'.
		^ false
	].

	presenceManager future
			assignLeader: (sessionID)
			toForum: (owner serviceSvc url).
.
	^ true! !

!QOrgServices methodsFor: 'initialize-release' stamp: 'bgf 3/31/2009 13:43'!
attachIsland: islandRef
	"Acquire our service references from the service island,
	 and register our event listeners."
	| checkManager |
	QLogger logInfo: 'Org services island registration. ', (islandRef).

	presenceManager := nil.

	"Get and resolve the presence manager.  Keep it in a temp,
	 in case (during the wait below) some other script tries to do a query."
	checkManager := islandRef future at: #QPresenceManager.
	(checkManager waitTimeoutSeconds: 45) ifFalse: [
		controller destroy.
		controller := nil.
		presenceManager := nil.
		self error: 'Organization services setup timed out.'.
	].

	QLogger logInfo: 'Assigning presence manager ', 
						(checkManager ifNotNil: [ checkManager hash ]).
	presenceManager := checkManager.

	"Register for service island signals that we relay for client-side consumption..."
	self signalMethodsToRelay do: [ :sigSelector |
		self stopScript: sigSelector.
		self startScript: sigSelector when: { presenceManager . sigSelector }.
	].

	"Catch reconnections. The participant UI discards the old island after reconnection."
	self stopScript: #islandDiscarded.
	self startScript: #islandDiscarded when: { controller island . #discard }.

	QLogger logInfo: 'Org services island attached ', 
						(presenceManager ifNotNil: [ presenceManager hash]).
! !

!QOrgServices methodsFor: 'meeting' stamp: 'bgf 6/28/2009 17:22'!
beMeetingLeader: bLeader
	"Become the leader of the current forum.
	 (Returns true/success or false/fail)"
	owner ifNil: [
		QLogger logWarning: 'beMeetingLeader, but os has no owner.'.
		^ false
	].
	owner readyForEvents ifFalse: [
		bLeader ifTrue: [ owner statusAlert: 'Services not ready. Cannot become leader.'].
		^ false
	].
	presenceManager isValidFarRef ifFalse: [
		bLeader ifTrue: [ owner statusAlert: 'Not connected. Cannot become leader.' ].
		^ false
	].
	bLeader ifTrue: [
		presenceManager future 
			assignLeader: (owner sessionID) 
			toForum: (owner serviceSvc url).
	] ifFalse: [
		presenceManager future
			retractLeader: (owner sessionID)
	].
	^ true! !

!QOrgServices methodsFor: 'locating' stamp: 'bgf 11/3/2009 17:05'!
clearPresence
	"Remove my presence-manager record, leaving behind only a last-location xml record."
	self recordMyLocation: nil quality: 0 updateXML: true.

	"And refresh the session ID in case I reconnect."
	lobbySessionID := TObjectID new.! !

!QOrgServices methodsFor: 'accessing' stamp: 'bgf 10/20/2007 19:49'!
controller

	"My island controller.  QOrgServices is a standalone router island,
	 not a persistent island - so we manage our own controller connection."
	^ controller! !

!QOrgServices methodsFor: 'locating' stamp: 'bgf 8/10/2009 02:11'!
currentUserFlags
	"The presence record flags I should specify in the presence updates of this user.
	 Constructed from the QPresenceManager class-side ...FlagMask constants."	
	| result |
	result := 0.
	owner preferences ifNotNilDo: [ : prefs |
		prefs doPrivateText ifFalse: [ 
			result := result bitOr: (QPresenceManager privateTextBlockFlagMask).
		]].
	^ result! !

!QOrgServices methodsFor: 'initialize-release' stamp: 'bgf 3/31/2009 13:33'!
destroy
	"Go away."
	QLogger logInfo: 'Org services destroy.'.
	self clearPresence.
	presenceManager ifNotNil: [
		QLogger logInfo: 'Detaching presence manager ', (presenceManager hash).
		self signalMethodsToRelay do: [  : sm | self stopScript: sm ].
	].
	self stopScript: #islandDiscarded.
	controller ifNotNil: [
		controller destroy. 
		QLogger logInfo: 'Org services controller discarded.'.
		controller := nil.
	].
	QLogger logInfo: 'Org services discarded.'.
	owner := nil.
	presenceManager := nil.
! !

!QOrgServices methodsFor: 'accessing' stamp: 'ar 11/24/2009 14:30'!
errorTripped
	"Has the client indicated that an error situation has arisen?
	When true, it prevents attempts to reconnect after the org services controller got nuked."
	^ errorTripped! !

!QOrgServices methodsFor: 'accessing' stamp: 'ar 11/24/2009 14:30'!
errorTripped: bError
	"Has the client indicated that an error situation has arisen?
	When true, it prevents attempts to reconnect after the org services controller got nuked."
	errorTripped := bError! !

!QOrgServices methodsFor: 'events' stamp: 'bgf 6/21/2009 16:39'!
forum: urlString flags: flagWord
	"Presence manager signal - change in the session-control flags for a given forum
	 (There are two subscribable signals, one including the previous value.)" 
	self signal: #forum:flags: with: urlString with: flagWord.! !

!QOrgServices methodsFor: 'events' stamp: 'bgf 6/21/2009 16:38'!
forum: urlString flags: flagWord oldFlags: oldFlagWord
	"Presence manager signal - change in the session-control flags for a given forum
	 (There are two subscribable signals, one including the previous value.)" 
	self signal: #forum:flags:oldFlags: with: urlString with: flagWord with: oldFlagWord.
	self signal: #forum:flags: with: urlString with: flagWord! !

!QOrgServices methodsFor: 'events' stamp: 'bgf 6/20/2009 21:52'!
forum: aForumURL leader: aPresenceRecord

	self signal: #forum:leader: with: aForumURL with: aPresenceRecord! !

!QOrgServices methodsFor: 'events' stamp: 'bgf 6/20/2009 21:53'!
forum: aForumURL version: aVersionString
	"Presence manager signal - new forum activation, or deactivation.
	 (For activation, the activator's client version is given.
	  For deactivation, it is nil.)"
	self signal: #forum:version: with: aForumURL with: aVersionString! !

!QOrgServices methodsFor: 'locating' stamp: 'bgf 6/21/2009 11:10'!
forumSessionForURL: forumURL

	"Return the presence record of the leader of this forum at this time."
	presenceManager isValidFarRef ifFalse: [ ^ nil ].

	^ (presenceManager get: #forumSessionForURL: with: forumURL) 
! !

!QOrgServices methodsFor: 'private' stamp: 'bgf 10/9/2008 18:46'!
forumSignal: aSymbol forForum: aForumURL fromUser: senderLogin argument: aStringOrArray

	"Receipt of a forumSignal from the presence-manager on-island.
	 Raise the signal on myself so harness-side listeners can receive them."

	owner serviceMgr ifNil: [ ^ self ].

	self signal: aSymbol with: aForumURL with: senderLogin with: aStringOrArray
	
! !

!QOrgServices methodsFor: 'initialize-release' stamp: 'ar 11/24/2009 14:31'!
initialize
	"Initialize the receiver"
	super initialize.
	errorTripped := false.! !

!QOrgServices methodsFor: 'initialize-release' stamp: 'bgf 7/22/2009 00:39'!
initializeForClient: aClient serviceManager: serviceMgr
	"Connect to the Org Services island using the client's  host/user/password.
	 This involves connecting and synchronizing (or bootstrapping)
 	 the org services island manually using the lower-level service manager router support.

	 This method deals with basic router-controller setup and acquisition.
	 The sub-methods
		#initializeNewIsland:    
		#attachIsland 
	 are where new service objects should be populated in the island,
 	 and attached for event listening when connecting to an island."

	QLogger logInfo: 'Starting org services connection.'.
	owner := aClient.
	orgRecord := serviceMgr currentOrg.
	lobbySessionID := TObjectID new.

	"Try it twice."
	[ self initializeForServiceManager: serviceMgr ]
	on: Error do: [ : ex |
		QLogger logWarning: 'Retrying org services: ', ex.
		Delay forMilliseconds: 500.
		[ 
			self initializeForServiceManager: serviceMgr.
		] on: Error do: [ : ex1 |
			self error: 'Cannot connect to org services. ',  (ex1 description).		
		].
	].	


! !

!QOrgServices methodsFor: 'initialize-release' stamp: 'ar 1/2/2010 17:33'!
initializeForServiceManager: serviceMgr
	"Connect to the Org Services island using the client's  host/user/password.
	 This involves connecting and synchronizing (or bootstrapping)
 	 the org services island manually using the lower-level service manager router support.

	 This method deals with basic router-controller setup and acquisition.
	 The sub-methods
		#initializeNewIsland:    
		#attachIsland 
	 are where new service objects should be populated in the island,
 	 and attached for event listening when connecting to an island."

	|  island name |

	"The island+router is identified with a name that includes the version...
	 so users from different environments won't prevent each other logging in.
	 They will, however, not be able to see each other as 'active' ".
	name := self versionedIslandName.

	owner statusInfo: 'Contacting organization..'.
	[
		QLogger logInfo: 'Org services login ', ([
			controller := serviceMgr loginOrgServices: name version: self versionedIslandName.
		  ] timeToRun), ' ms'.

	] on: Error do: [ :ex|
		QLogger logWarning: 'Org services finds no router - ', ex description.
	].

	controller ifNil: [
		self error: 'No organization services for ', orgRecord name.
	].

	QLogger logInfo: 'Org services join ', ([ 
		island := self joinAndSynchronize
		] timeToRun), ' ms'.
	owner statusInfo: 'Starting location service...'.

	"If we fall through with no island, we're the first man in."
	island ifNil: [
		controller island: (island := TIsland new).
		self initializeNewIsland: island.
	].
	self attachIsland: island.
! !

!QOrgServices methodsFor: 'initialize-release' stamp: 'bgf 6/3/2009 09:32'!
initializeNewIsland: islandRef

	"This is where the (transient) org services island is populated by first-man-in."
	islandRef future name: 'Org Services'.
	islandRef future new: QPresenceManager.! !

!QOrgServices methodsFor: 'initialize-release' stamp: 'bgf 1/6/2008 14:59'!
isConnected
	"Is the org services island currently connected?"

	^ controller notNil and:[ controller status = #ready ].! !

!QOrgServices methodsFor: 'accessing' stamp: 'bgf 7/23/2009 15:18'!
isForum: forumURL okForVersion: versionString
	"Support for pre-login version compatibility check.
	 Returns true (ok)  unless we're pretty sure there is a different version in use."

	^ (self forumSessionForURL: forumURL) ifNil: [ true ]
		ifNotNilDo: [ : fs | fs clientVersion notNil and: [ fs clientVersion = versionString] ]
! !

!QOrgServices methodsFor: 'locating' stamp: 'bgf 6/8/2010 11:17'!
isLogin: userLogin inForum: aURL
	"Is the user logged in to the given forum?"

	^ (self presencesOfLogin: (userLogin)) 
		ifNotNilDo: [ : presences | ^ presences anySatisfy: [ : pr | pr allURLs includes: aURL ]]  
		ifNil: [ false ]! !

!QOrgServices methodsFor: 'initialize-release' stamp: 'bgf 3/31/2009 13:39'!
islandDiscarded

	"Handle a reconnect event... reinitialize from the island.
	 We don't need to inform anyone else, 
	 because we do all the island-signal listening locally
	 and re-signal in the UI process."

	QLogger logInfo: 'Org services island - discarding: ', 
					  (presenceManager ifNotNil: [ presenceManager hash]).
	(controller notNil and: [controller island notNil])  ifTrue: [
		[
			QLogger logInfo: 'Org services reconnect: attaching new island.'.
			self attachIsland: controller island.
		] on: Error do: [ :ex |
			owner statusAlert: 'Organization services not available.'.
			QLogger logInfo: 'Reconnect: island reattached failed: ', ex.
		]
	] ifFalse: [
		QLogger logInfo: 'Org services island discard - no replacement.'
	].! !

!QOrgServices methodsFor: 'initialize-release' stamp: 'bgf 2/26/2009 01:30'!
joinAndSynchronize

	"Part of island-controller setup.   We have a controller - join it,
 	 and acquire its island.   Raises an error if this fails.

	 Returns the island if the service already had members;
	 otherwise returns nil ... in which case the caller should build the island."

	| p retry island syncData dots |

	owner statusInfo: 'Contacting organization...'.

	p := controller join.	"receive messages"
	p wait. "until joined"
	controller status = #ready ifFalse: [
		self error: 'Organization services unavailable.'.
	].

	owner statusInfo: 'Joining organization services..'.
	"Ask for our participant ID repeatedly because we may have moved into spot #1."
	retry := 0.
	island := nil.
	[island isNil and: [(controller getParticipantID wait) > 1]] 
	 whileTrue:[
		"Try requesting a sync snapshot"

		QLogger logInfo: 'Org services sync ', ([
			p := controller sync.
			[(Delay forMilliseconds: 200) wait. p isResolved] whileFalse:[
				owner statusInfo: 'Joining organization services...'.
			].
		   ] timeToRun), ' ms'.

		syncData := p result.
		syncData ifNil:["No sync from peer (yet) -> wait"
			retry := retry + 1.
			dots := '.....'.
			dots := dots copyFrom: 1 to: retry \\ dots size+1.
			owner statusInfo: 'Joining organization services..', dots.
			(Delay forMilliseconds:500) wait.
		] ifNotNil:[  "Sync successful; install result"
			owner statusInfo: 'Starting organization services...'.
			controller install: syncData.
			island := controller island.
		].
	].

	controller status = #ready ifFalse: [
		self error: 'Organization services unavailable.'.
	].
	owner statusInfo: 'Starting location service...'.
	controller setVersionInfo: (self versionedIslandName).
	controller heartbeat: 500.	"Slow heartbeat for this non-simulation island."
	controller beServer.

	^ island		"May be nil, but only if we're the first man in, so it's our job to make one."
! !

!QOrgServices methodsFor: 'accessing' stamp: 'bgf 2/6/2009 16:58'!
lobbySessionID
	^ lobbySessionID! !

!QOrgServices methodsFor: 'meeting' stamp: 'bgf 6/25/2009 01:58'!
meetingLeaderSessionID

	"Return the sessionID of the current leader of the current forum, if any."
	
	owner ifNil: [ ^ nil ].
	owner currentForumURL ifNil: [ ^ nil ].
	presenceManager isValidFarRef ifFalse: [ ^ nil ].
	^ presenceManager get: #forumLeaderSID: with: (owner currentForumURL).! !

!QOrgServices methodsFor: 'robots' stamp: 'bgf 3/20/2008 00:50'!
notRobot: aSessionID
	"Signal from presence manager; relay.  The given userLogin is no longer available as a robot."
	self signal: #notRobot: with: aSessionID.
! !

!QOrgServices methodsFor: 'locating' stamp: 'hrs 8/21/2009 14:34'!
noteLobbyEntry
	"Records (on server as lastOrg.xml) the org and time of this entry to the lobby, either on the way in, 
	a service chooser dropdown change, or a logout to the service chooser.

	Some of the above cases do not make noteLobbyPresenceWithQuality send a locator.xml.

	With lobby entry noted, we can tell whether an online user was really last in the forum specified by 
	locator.xml, by whether locator clientUTC >= lastOrg clientUTC."
	self class sendOrg: orgRecord forUser: owner preferences userLogin toManager: owner serviceMgr.! !

!QOrgServices methodsFor: 'locating' stamp: 'bgf 6/21/2009 10:13'!
noteLobbyPresenceWithQuality: qos
	"Tell the presence manager that I am in the lobby"
   	self recordMyLocation: (QOrgServices lobbyURL) quality: qos updateXML: false! !

!QOrgServices methodsFor: 'accessing' stamp: 'bgf 2/13/2009 22:31'!
numUsersLoggedIn
	"Approximate... number of users with active presence records."
	presenceManager isValidFarRef ifTrue: [
		^ (presenceManager get: #numUsersLoggedIn) ifNil: [ 1 "me" ]
	] ifFalse: [
		^ 0
	].! !

!QOrgServices methodsFor: 'locating' stamp: 'bgf 6/18/2008 22:05'!
offlineLocationRecord: aUserName

	"If the service manager has a locator.xml file showing the last location
	 of the given (offline) user, then fetch it, and return it as a locator record.   
	 The record is a QPresenceRecord.
	 For a modicum of privacy, this service is only available to administrators,
	 so that casual users can't check on the offline activity of others."

	owner isAdministrator ifFalse: [
		^ QPresenceRecord login: aUserName url: 'Unknown'.
	].
	owner serviceMgr ifNotNilDo: [ : mgr |
		^ self class requestUserLocation: aUserName fromManager: mgr
	].
	^ nil! !

!QOrgServices methodsFor: 'accessing' stamp: 'ar 7/14/2008 14:47'!
orgRecord
	^orgRecord! !

!QOrgServices methodsFor: 'accessing' stamp: 'bgf 10/20/2007 00:59'!
presenceManager
	"Far ref to the organization presence manager."
	^ presenceManager! !

!QOrgServices methodsFor: 'locating' stamp: 'bgf 8/8/2009 02:50'!
presenceOfLogin: aUserLogin

	"If the given user has a presence record in the manager, return it.   
	 The record is a QPresenceRecord."

	presenceManager isValidFarRef ifTrue: [
		^ presenceManager syncSend: #presenceOfLogin: withArguments: {aUserLogin}
	].
	^ nil! !

!QOrgServices methodsFor: 'locating' stamp: 'jcg 8/30/2008 00:18'!
presenceOfSession: aSessionID

	"If the given user has a presence record in the manager, return it.   
	 The record is a QPresenceRecord."

	presenceManager isValidFarRef ifTrue: [
		^ presenceManager syncSend: #presenceOfSession: withArguments: {aSessionID}
	].
	^ nil! !

!QOrgServices methodsFor: 'locating' stamp: 'bgf 8/8/2009 02:57'!
presencesOfLogin: aUserLogin
	"Return the array of all (separate-session) current presence records of the given login."
	presenceManager isValidFarRef ifTrue: [
		^ presenceManager syncSend: #presencesOfLogin: withArguments: {aUserLogin}
	].
	^ nil! !

!QOrgServices methodsFor: 'initialize-release' stamp: 'ar 8/31/2010 10:31'!
reconnect
	"Reconnect OrgServices."

	controller ifNotNil:[
		controller destroy.
		controller := nil.
	].

	[self initializeForServiceManager: owner serviceMgr] 
		on: Error do:[:ex|
			QLogger logWarning: ex description.
			^false].
	^true! !

!QOrgServices methodsFor: 'locating' stamp: 'bgf 8/11/2009 11:50'!
recordMyLocation: urlString quality: qos updateXML: doServer
	"Tell the presence manager that I am now at the given URL;
       the urlString may be nil to indicate logout."
	| localTime |
	owner ifNil: [ ^ self ].
	localTime := DateAndTime now asUTCSeconds.
	presenceManager isValidFarRef ifTrue: [
		| otherActiveUrls |
		otherActiveUrls := (
			urlString 
				ifNil: [ #() ] 
				ifNotNil: [ (owner activeServices collect: [ : svc | svc url ]) copyWithout: urlString ]).
		presenceManager future 
			presence: {
				(owner sessionID) .
				(owner userLogin) .
				urlString .
				localTime .
				qos .
				(self currentUserFlags) .
				(owner robotMode) .
				otherActiveUrls .
				QwaqVersion current infoString .
				(owner preferences
						ifNotNilDo: [ : prefs | prefs publicName ]
						ifNil: [ QStyles simplerLoginName: (owner userLogin) ] )
			}.
	] ifFalse: [
		QLogger logWarning: 'Location update, no presence manager: ', presenceManager.
	].

	"Update the locator record on the server (for #offlineLocationRecord:)"
	(doServer and: [urlString notNil]) ifTrue: [
		self class sendLocation: urlString 
				  forUser: (owner preferences userLogin)
				  toManager: owner serviceMgr.
	].
! !

!QOrgServices methodsFor: 'robots' stamp: 'bgf 3/20/2008 04:49'!
robot: aSessionID action: aSymbol arguments: anArray

	"Signal from presence manager; relay."
	self signal: #robot:action:arguments: with: aSessionID with: aSymbol with: anArray
! !

!QOrgServices methodsFor: 'robots' stamp: 'bgf 3/20/2008 00:51'!
robot: aSessionID mode: aSymbol

	"Signal from presence manager; relay.
	 A mode symbol of nil indicates that the user is available but not active as a robot;
	 other symbols are operating modes (passed through the org-service literally.)"

	self signal: #robot:mode: with: aSessionID with: aSymbol.
! !

!QOrgServices methodsFor: 'robots' stamp: 'jcg 8/30/2008 00:17'!
robot: aUserLogin takeAction: aSymbol withArguments: anArray

	"Basic signaling; ask the given robot to do the given action.
	Best to keep the arguments be (island-marshalling-wise) such as symbols."
	
	presenceManager isValidFarRef ifTrue: [
		presenceManager future robot: aUserLogin takeAction: aSymbol withArguments: anArray
	].! !

!QOrgServices methodsFor: 'robots' stamp: 'jcg 8/30/2008 00:17'!
robotModeOfUser: aSessionID

	aSessionID ifNil: [ ^ nil ].

	"Determine what robot mode symbol the current user is currently operating under, if any."
	presenceManager isValidFarRef ifTrue: [
		^ presenceManager syncSend: #robotModeOfUser: withArguments: {aSessionID}
	].
	^ nil! !

!QOrgServices methodsFor: 'signaling' stamp: 'bgf 2/26/2009 12:45'!
sendSignal: aSymbol forForum: aUrl 

	"Broadcast a client-defined signal related to a named forum.
	 All listening QOrgSrvices will raise the given signal (aSymbol)
 	 with arguments { forumName, senderLogin}."

	presenceManager isValidFarRef ifTrue: [
		presenceManager future 
			sig: aSymbol 
			forum: aUrl 
			user: (owner userLogin) 
			arg: nil.
	].! !

!QOrgServices methodsFor: 'signaling' stamp: 'bgf 2/26/2009 12:45'!
sendSignal: aSymbol forForum: aUrl argument: aStringOrArray

	"Broadcast a client-defined signal related to a named forum.
	 All listening QOrgSrvices will raise the given signal (aSymbol)
 	 with arguments { forumName, senderLogin, aStringOrArray}."

	presenceManager isValidFarRef ifTrue: [
		presenceManager future 
			sig: aSymbol 
			forum: aUrl 
			user: (owner userLogin) 
			arg: aStringOrArray.
	].! !

!QOrgServices methodsFor: 'signaling' stamp: 'bgf 2/26/2009 12:45'!
sendSignal: aSymbol forUser: aLogin argument: aStringOrArray

	"Issue a client-defined user-to-user signal.
	 The named user's QOrgSrvices will raise the given signal (aSymbol),
 	 with arguments {my login, aStringOrArray}."

	presenceManager isValidFarRef ifTrue: [
		presenceManager future 
			sig: aSymbol 
			user: aLogin 
			from: (owner userLogin) 
			arg: aStringOrArray.
	].! !

!QOrgServices methodsFor: 'private' stamp: 'bgf 8/8/2009 01:58'!
setForum: urlString flag: flagMask value: bValue
	"Private - set or clear one flag in the forum-session for the given url."

	owner ifNil: [ ^ nil ].
	presenceManager isValidFarRef ifFalse: [ ^ nil ].
	presenceManager future forum: urlString flag: flagMask value: bValue.

! !

!QOrgServices methodsFor: 'meeting' stamp: 'bgf 8/8/2009 02:15'!
setMeetingPrivateTextBlocked: bValue
	"In the current session, set the private-text-blocked as indicated.
	 This is done as information for off-forum clients
	  who cannot see the full meeting state of a forum."

	owner ifNil: [ ^ nil ].
	owner currentForumURL ifNil: [ ^ nil ].
	self setForum: (owner currentForumURL) 
		flag: (QPresenceManager privateTextBlockFlagMask) 
		value: bValue! !

!QOrgServices methodsFor: 'robots' stamp: 'hrs 4/21/2010 11:43'!
setUser: aSessionID isPrivateTextBlocked: aBoolean

	"Assign the given user the given isPrivateTextChat:."

	presenceManager isValidFarRef ifTrue: [
		(aSessionID = owner sessionID) ifFalse: [
			QLogger logWarning: 'Attempt to set isPrivateTextChat for other session: ', aSessionID.
			^ nil
		].
		presenceManager future user: aSessionID isPrivateTextBlocked: aBoolean.
		^ aBoolean
	].
	^ nil! !

!QOrgServices methodsFor: 'robots' stamp: 'jcg 8/30/2008 00:17'!
setUser: aSessionID notice: aString

	"Assign the given user the given notice; nil indicates full active presence."

	presenceManager isValidFarRef ifTrue: [
		(aSessionID = owner sessionID) ifFalse: [
			QLogger logWarning: 'Attempt to set notice for other session: ', aSessionID.
			^ nil
		].
		presenceManager future user: aSessionID notice: aString.
		^ aString
	].
	^ nil! !

!QOrgServices methodsFor: 'robots' stamp: 'jcg 8/30/2008 00:17'!
setUser: sessionID robotMode: aSymbol

	"Assign the given user the given robotic mode;
	 nil would indicate the user login is available for automaton duty but not active.
	 The given user login must be available for robotics,
	 or must be my own login."

	presenceManager isValidFarRef ifTrue: [
		((sessionID = owner sessionID) or: [self userIsRobot: sessionID]) ifFalse: [
			aSymbol ifNotNil: [
				QLogger logWarning: 'Attempt to set robot mode for non-robot: ', sessionID.
			].
			^ nil
		].
		presenceManager future robot: sessionID mode: aSymbol.
		^ aSymbol
	].
	^ nil! !

!QOrgServices methodsFor: 'private' stamp: 'hrs 4/21/2010 10:49'!
signalMethodsToRelay
	"These are all the presence-manager signal/methods we relay to the harness."
	^ #(user:location:time:  
		 forum:leader:
		 forum:version:
		 forum:flags:oldFlags:
		 notRobot:  
		 robot:mode:  
		 robot:action:arguments:  
		 user:notice: 
		 user:publicName:
		 user:quality:
		 user:isPrivateTextBlocked:
		 userSignal:forUser:fromUser:argument:
		 forumSignal:forForum:fromUser:argument: ) ! !

!QOrgServices methodsFor: 'robots' stamp: 'jcg 8/30/2008 00:17'!
stopBeingRobot: aSessionID

	"Retract the advertisement of the my user login as robotic."

	aSessionID = (owner sessionID) ifFalse: [
		QLogger logWarning: 'Cannot retract another user from robotic status!! ', aSessionID.
		^self
	].

	presenceManager isValidFarRef ifTrue: [
		presenceManager future stopBeingRobot: aSessionID.
	].! !

!QOrgServices methodsFor: 'events' stamp: 'hrs 4/21/2010 10:49'!
user: aSessionID isPrivateTextBlocked: aBoolean

	"Presence manager signal - updated isPrivateTextBlocked: quality for a logged-in user."
	self signal: #user:isPrivateTextBlocked: with: aSessionID with: aBoolean! !

!QOrgServices methodsFor: 'events' stamp: 'bgf 3/20/2008 00:43'!
user: aSessionID location: aURLString time: anEpochSecondsTime

	"Presence manager signal - new forum location for a logged-in user."
	self signal: #user:location:time: with: aSessionID with: aURLString with: anEpochSecondsTime! !

!QOrgServices methodsFor: 'events' stamp: 'bgf 3/20/2008 00:41'!
user: aSessionID notice: aString

	"Presence manager signal - new notice string for a logged-in user."
	self signal: #user:notice: with: aSessionID with: aString! !

!QOrgServices methodsFor: 'events' stamp: 'bgf 8/8/2009 22:01'!
user: aSessionID publicName: pName

	"Presence manager signal - updated session quality for a logged-in user."
	self signal: #user:publicName: with: aSessionID with: pName! !

!QOrgServices methodsFor: 'events' stamp: 'bgf 12/15/2008 16:56'!
user: aSessionID quality: sessionQuality

	"Presence manager signal - updated session quality for a logged-in user."
	self signal: #user:quality: with: aSessionID with: sessionQuality! !

!QOrgServices methodsFor: 'robots' stamp: 'jcg 8/30/2008 00:17'!
userIsRobot: aSessionID

	"Boolean - Is the user recorded as available for robotic automation?"

	presenceManager isValidFarRef ifTrue: [
		^ presenceManager syncSend: #userIsRobot: withArguments: {aSessionID}
	].
	^ false! !

!QOrgServices methodsFor: 'locating' stamp: 'jcg 8/30/2008 00:17'!
userNamesInService: aServiceURL

	"Given a service URL, show all users currently located in that service.
	 Returns an Array of Strings (no far-ref entanglements.) "

	presenceManager isValidFarRef ifTrue: [
		^ (presenceManager syncSend: #userNamesInService: withArguments: {aServiceURL}) ifNil: [#()]
	].
	^#()! !

!QOrgServices methodsFor: 'private' stamp: 'bgf 2/6/2009 16:18'!
userSignal: aSymbol forUser: aLogin fromUser: senderLogin argument: aStringOrArray

	"Receipt of userSignal from the island....
	 If I'm the recipient, raise it here.
	 The named user's QOrgSrvices will be signaled with
 	 the given signal, with arguments {my login, aStringOrArray}."

	"Note - for now we only attend to these once we're fully logged in to a forum.
	 Will need changing (probably with some more guards elsewhere for logout-time
	 if anyone wants to use userSignal:   for pre-login communication."

	(self = owner orgServices) ifFalse: [ ^ self ].
	(aLogin = (owner userLogin)) ifTrue: [
		[ 
			self signal: aSymbol with: senderLogin with: aStringOrArray
		] on: Error do: [ :ex|
			QLogger logError: ex from: self.
		].
	].! !

!QOrgServices methodsFor: 'version' stamp: 'ar 7/14/2008 14:45'!
versionedIslandName

	"Very important... if the class shape or behavior of any part of the content 
	 of the OrgServices island changes, then this string must also change."

	^ self versionedIslandNameForOrganization: orgRecord uid.! !

!QOrgServices methodsFor: 'version' stamp: 'bgf 8/24/2010 16:27'!
versionedIslandNameForOrganization: orgName

	"Very important... if the class shape or behavior of any part of the content 
	 of the OrgServices island changes, then this string must also change.
	 We do this separately from the master Qwaq version so that users from
	 different versions can be aware of each other."

	^ orgName, '-OrgServices-2.5-', ([Smalltalk vmParameterAt: 41] on: Error do: [ 1 ]).
									"Image binary format."

	"1.0.0 was first distribution."
	"1.0.1 adds presence heartbeat (third element in presence location record)"
	"1.0.2 adds robot table, notRobot: and robot:mode: signalling."
	"1.0.3 adds robot:action:arguments"
	"1.0.4 uses QPresenceRecord rather than a plain array."
	"1.0.5 adds inter-user signalling"
	"1.0.6 uses sessionID rather that user login as the primary presence table key."
	"1.0.7 was forced by a change to TIsland state for object-logging."
	"1.0.8 fixes robot retraction so that aged-out sessions don't linger in the presence robot table"

	"1.1 introduces forum-scope signaling, and the use of vm image format as part of the version number."
	"1.2 reflects Oct-2008 additions to TIsland class format (sets of frames for updateShape/etc)"
	"1.3 adds sessionQuality to the presence record."
	"1.4 switches to terse message names."
	"1.5 corresponds to the switch to using integer-coded method names for common messages in TMessageEncoder"

	"1.6 adds  forum session records, full tracking of user-logins into 'extra' forums,  
	 	           forum-leader negotiation, and flag words in presence records."
	"1.7 adds public name into the individual user presence records"
	"1.8 reflects adding version string fields into TIsland"
	"1.9 reflects a sweep of unused-ivar removals from TIsland, etc."
	"2.0 provides load balancing for org services (change in router type)"
	"2.1 updated list of compactly-encoded names in TMessageEncoder"
	"2.2 more compactly-encoded names"
	"2.3 switch to Squeak 4.1 changed class Heap"
	"2.4 switch symbols to deterministic hashing."
	"2.5 Island gains a mutex."! !

TObject subclass: #QPanelHandler
	instanceVariableNames: 'assembly extent maxExtent minExtent menuLabel rotGrip motionGrip rotationGripOnDemand motionGripOnDemand pivotGripOnDemand eastPivotGrip westPivotGrip downPoint resizeFrameOnDemand resizeFrame vAlign hAlign inset'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Objects-Manipulation'!

!QPanelHandler commentStamp: '<historical>' prior: 0!
DRAFT framework class for a flat panel adjustment handler.

Dispatches user interactions for rotating, resizing, and moving a flat panel.
Maintains the various grips (rotation, motion, resizeFrame) 
that map those behaviors onto visible subframes of the panel assembly.

The assembly is the "flat" tframe content that is managed by the handler.
It should be resizable in x and y, with z being a nominal thickness.  
The assembly must support extent: aPoint.

Supports resizing by dragging edges using a built-in QFlatFrame,
or any user-supplied QPanelFrame.
See class #forAssembly: and #forAssembly:withFrame.

The grips used for rotation and moving must
be supplied by the client code; otherwise rotation and motion are not available.

The supplied rotation and motion grips should be appropriate
subframes of the assembly (see QMotionGrip, QAbstractPivot).

At most one mode of adjustment at a time is active; 
a menu offers the selection of mode.

The menu is built-in, but client code must supply a tframe 
(possibly just a button-like cube) that user clicks to summon the menu.
(QDisplayPanel has an example, a simple cube in the ULH corner of the frame.)

Main instance variables:
	assembly: the panel; usually a TGroup containing all the other TFrames involved.
	resizeFrame: the supplied QPanelFrame offering resize control
!

!QPanelHandler class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QPanelHandler class methodsFor: 'instance creation' stamp: 'bgf 1/2/2007 17:21'!
forAssembly: aTFrame
	"A default panel handler with simple rectilinear frame,
	 wrapped around the given assembly."

	^ (self new) withSimpleFrame; assembly: aTFrame; yourself.! !

!QPanelHandler class methodsFor: 'instance creation' stamp: 'bgf 1/2/2007 17:56'!
forAssembly: aTFrame withFrame: aPanelFrame

	"A default panel handler with the given panel frame,
	 wrapped around the given assembly."

	^ (self new) withFrame:aPanelFrame; assembly: aTFrame; yourself.! !

!QPanelHandler methodsFor: 'menu' stamp: 'bgf 9/1/2009 11:14'!
addMenuActions: aMenu forEvent: event

	motionGrip ifNotNil: [
		aMenu addItemSpec: {'Allow Moving' . #toggleMove } target: self.
	].
	resizeFrame ifNotNil: [
		aMenu addItemSpec: {'Allow Resizing' . #toggleResize}  target: self.
	].
	(rotGrip notNil | eastPivotGrip notNil | westPivotGrip notNil) ifTrue: [
		aMenu addItemSpec: {'Allow Turning' .  #toggleRotate} target: self .
	].
	self checkMenu: aMenu.! !

!QPanelHandler methodsFor: 'private' stamp: 'bgf 3/15/2010 10:57'!
applyStateVector: vec
	"Undo/redo support"
	self extent: (vec first).
	assembly translation: (vec second).
	(assembly respondsTo: #standHeight) ifTrue: [
		assembly standHeight: (vec third)
	].
! !

!QPanelHandler methodsFor: 'accessing' stamp: 'bgf 9/15/2006 00:25'!
assembly
	"The outermost TFrame for which this is the manipulator."
	^ assembly! !

!QPanelHandler methodsFor: 'accessing' stamp: 'ar 5/17/2007 14:26'!
assembly: aTFrame

	"The assembly is the outermost group that is moved by the motion grip;
	  its center is also moved to preserve absolute position when the panel is resized."

	assembly := aTFrame.
	self runScript: #signalPointerEnter: when: { assembly. #pointerEnter }.
	self runScript: #signalPointerLeave: when: { assembly. #pointerLeave }.
	self runScript: #getKeyboardControl: when: { assembly. #getKeyboardControl.}.
	self runScript: #loseKeyboardControl when: { assembly. #loseKeyboardControl.}.
	self runScript: #signalRemoveUser: when: { assembly. #removeUser.}.
	rotGrip ifNotNil: [ rotGrip luggage: self assembly ].
	motionGrip ifNotNil: [ motionGrip luggage: self assembly ].

	self menuAccess: aTFrame.
! !

!QPanelHandler methodsFor: 'accessing' stamp: 'bgf 12/21/2008 09:04'!
broadcastColor: col
	"Relays the avatar color of an active user back to the embedded-app,
	 so that harness-side floating displays can reflect that activity."
	assembly ifNotNil:[
		assembly broadcastColor: col.
	].
! !

!QPanelHandler methodsFor: 'menu' stamp: 'bgf 1/23/2007 10:02'!
canMove
	"Is motion enabled?"
	^ motionGrip notNil and: [motionGrip isEnabled]! !

!QPanelHandler methodsFor: 'menu' stamp: 'bgf 1/23/2007 10:02'!
canResize
	"Is the resize frame enabled?"
	^ resizeFrame notNil and: [resizeFrame isEnabled]! !

!QPanelHandler methodsFor: 'menu' stamp: 'bgf 1/23/2007 10:02'!
canRotate
	"Is the rotation grip or pivot enabled?"
	rotGrip ifNotNil: [ ^ rotGrip isEnabled].
	eastPivotGrip ifNotNil: [ ^ eastPivotGrip isEnabled].
	westPivotGrip ifNotNil: [ ^ westPivotGrip isEnabled].		
	^false! !

!QPanelHandler methodsFor: 'menu' stamp: 'bgf 1/9/2009 22:32'!
cancelMenu: event
	"Dismiss any previous menu for this avatar, 
	 and turn off all my hot manipulation modes."
	(event avatar notNil and: [ event avatar cancelMenu ]) ifTrue: [
		assembly ifNotNil: [
			self disableMove.
			self disableRotate.
			self disableResize.
			assembly unhilite.
		].
		^ true
	].
	^ false.! !

!QPanelHandler methodsFor: 'menu' stamp: 'bgf 4/8/2008 21:40'!
checkMenu:aTMenu

	"Set the checkmarks in the menu - uses selector comparison."
	aTMenu menuItems do: [ :item| 
		| selector |
		selector := item action ifNotNilDo:[:a|a selector].
		selector caseOf: {
			[#toggleMove]-> [item checked: (self canMove)] .
			[#toggleRotate]-> [item checked: (self canRotate)] .
			[#toggleResize]-> [item checked: (self canResize)] .
		} otherwise:[].
	].
! !

!QPanelHandler methodsFor: 'accessing' stamp: 'bgf 9/20/2006 23:59'!
contactOffset: aScalar
	"Applied to the motion grip; see QDragAdapter #contactOffset."

	motionGrip ifNotNil: [
		motionGrip contactOffset: aScalar.
	].! !

!QPanelHandler methodsFor: 'menu' stamp: 'bgf 7/25/2008 01:02'!
disableMove

	"Deactivate the motion grip (for the time being),
	 except for its click behaviors (often bound to panel go-to shortcuts.)"

	assembly ifNil: [ ^ self ].
	assembly unhilite.
	motionGrip ifNotNil: [
		motionGrip enableClickOnly.
		motionGripOnDemand ifTrue: [
			self assembly removeChild: motionGrip grip.
		].
	].
! !

!QPanelHandler methodsFor: 'menu' stamp: 'bgf 9/6/2008 00:38'!
disableResize

	"Deactivate the motion grip."
	assembly ifNil: [ ^ self ].
	assembly unhilite.
	resizeFrame ifNotNil: [
		resizeFrame disable.
		resizeFrame unhilite.
		resizeFrameOnDemand ifTrue: [
			self assembly removeChild: resizeFrame.
		].
	].
	assembly shapeChanged.
! !

!QPanelHandler methodsFor: 'menu' stamp: 'bgf 4/23/2007 22:29'!
disableRotate

	"Deactivate the rotation grip and/or pivot grips."
	assembly ifNil: [ ^ self ].
	assembly unhilite.
	rotGrip ifNotNil: [ 
		rotGrip disable.
		rotGrip unhilite.
		rotationGripOnDemand ifTrue: [
			self assembly removeChild: rotGrip grip.
		].
	].
	
	eastPivotGrip ifNotNil: [ 
		eastPivotGrip disable.
		pivotGripOnDemand ifTrue: [
			self assembly removeChild: eastPivotGrip grip.
		].
	].
	westPivotGrip ifNotNil: [ 
		westPivotGrip disable.
		pivotGripOnDemand ifTrue: [
			self assembly removeChild: westPivotGrip grip.
		].
	].! !

!QPanelHandler methodsFor: 'accessing' stamp: 'bgf 3/15/2010 11:35'!
doUpdateShape
	"For capturing state at end of drag."
	assembly doUpdateShape.! !

!QPanelHandler methodsFor: 'events' stamp: 'bgf 3/15/2010 10:50'!
drag: cornerPt corner: compassPt delta: rawDelta
	"An interactive resize from the frame."
	| delta sign offset |

	sign := compassPt caseOf: {
		[#northeast] -> [ 1@1].
		[#northwest] -> [ -1@1].
		[#southeast] -> [ 1@-1].
		[#southwest] -> [ -1@-1].
	}.
	delta := (rawDelta x @ rawDelta y) * sign.		"Note - delta is a point!!"

	"For fixed aspect ratios, we consider the center pinned rather than the opposite edge.
	 So we need to double the delta so that the resulting resize moves this edge
	 the appropriate distance, to keep up with the pointer."
	resizeFrame aspect ifNotNil: [
		delta := delta * 2.
	].
	delta := delta max: ((extent - minExtent) negated).
	delta := delta min: (maxExtent - extent).

	self extent: (extent + delta).
	resizeFrame aspect ifNil: [
		"Panels are center-aligned in space, so to keep the opposite edge still,
	 	 we have to translate the whole thing by 1/2 the distance we resized by."
		offset := vAlign caseOf: {
			[#center] -> [delta/2 *sign].
			[#top] -> [(delta x /2 @ delta y) *sign].
			[#bottom] -> [(delta x /2 @ 0) *sign].
		} otherwise: [ 0@0 ].

		assembly translation:
			(assembly translation) +
			(assembly orientation transformPoint: offset@0).

		"For display stands (with legs), preserve the y position of the foot
		 by adjusting the stand height to counteract the change in size."
		(assembly respondsTo: #standHeight:) ifTrue: [	
			vAlign caseOf: {
				[#center] -> [ 
					(compassPt includesSubString: 'north') ifFalse: [
						assembly standHeight: assembly standHeight - delta y.
					]].
				[#top] -> [assembly standHeight: (assembly standHeight - delta y)].
				[#bottom] -> [assembly standHeight: (assembly standHeight + delta y)].
			} otherwise: [].
		].
	].
! !

!QPanelHandler methodsFor: 'accessing' stamp: 'bgf 8/16/2009 00:47'!
dragAdapter
	^ motionGrip! !

!QPanelHandler methodsFor: 'menu' stamp: 'bgf 4/23/2007 22:29'!
enableMove

	"Configure for interactive movement, disabling other modes."

	assembly ifNil: [ ^ self ].

	motionGrip ifNotNil: [
		motionGrip enable.
		motionGripOnDemand ifTrue: [
			self assembly addChild: motionGrip grip.
		].
	].
! !

!QPanelHandler methodsFor: 'menu' stamp: 'bgf 6/1/2009 17:03'!
enableResize

	| canMove |
	"Configure for interactive resize, disabling rotation."
	canMove := self canMove.
	self disableRotate.
	resizeFrame ifNotNil: [
		resizeFrameOnDemand ifTrue: [
			assembly addChild: resizeFrame.
		].
		resizeFrame enable.
	].
	assembly shapeChanged.

	"See QFlatFrame enable - we have to re-enable the top bar sensitivity."
	canMove ifTrue: [ self enableMove ].
	self shapeChanged.

! !

!QPanelHandler methodsFor: 'menu' stamp: 'bgf 3/4/2007 21:34'!
enableRotate

	"Configure for interactive rotation."

	rotGrip ifNotNil: [ 
		rotGrip enable.
		rotGrip grip hilite: (QColors controlHiliteColor).
		rotationGripOnDemand ifTrue: [
			self assembly addChild: rotGrip grip.
		].
	].
	
	eastPivotGrip ifNotNil: [
		eastPivotGrip enable.
		eastPivotGrip grip hilite: (QColors controlHiliteColor).
		pivotGripOnDemand ifTrue: [
			self assembly addChild: eastPivotGrip grip.
		].
	].

	westPivotGrip ifNotNil: [
		westPivotGrip enable.
		westPivotGrip grip hilite: (QColors controlHiliteColor).
		pivotGripOnDemand ifTrue: [
			self assembly addChild: westPivotGrip grip.
		].
	].

	(eastPivotGrip notNil & westPivotGrip notNil)  ifTrue: [
		eastPivotGrip offset: ((self assembly relativeTransform: westPivotGrip grip) translation).
		westPivotGrip offset: ((self assembly relativeTransform: eastPivotGrip grip) translation).
	].! !

!QPanelHandler methodsFor: 'accessing' stamp: 'bgf 10/28/2006 23:53'!
extent

	"The x-y extent of the panel."
	^ extent

! !

!QPanelHandler methodsFor: 'accessing' stamp: 'bgf 1/2/2007 17:33'!
extent: newExtent

	"Set the current extent to the given rectangle, 
	 after enforcing the minExtent and maxExtent bounds."

	| bounded |
	bounded := newExtent min: maxExtent max:minExtent.
	extent := bounded.
	self shapeChanged.

	(assembly ~= self and: [assembly respondsTo:#extent:]) ifTrue: [
		assembly extent: bounded.
	].

! !

!QPanelHandler methodsFor: 'accessing' stamp: 'bgf 3/4/2007 19:24'!
frame

	"The panel frame that offers resizing control for this panel, if any is present."
	^ resizeFrame
! !

!QPanelHandler methodsFor: 'accessing' stamp: 'bgf 1/2/2007 16:26'!
frame: aFlatFrame

	"The flat frame that offers resizing control for this panel, if any is present."
	resizeFrame := aFlatFrame.
	resizeFrame assembly: self.
	self shapeChanged.
! !

!QPanelHandler methodsFor: 'events' stamp: 'bgf 5/28/2007 09:10'!
getKeyboardControl: avatar

	self signal: #getKeyboardControl with: avatar.
! !

!QPanelHandler methodsFor: 'accessing' stamp: 'bgf 9/20/2006 23:34'!
hAlign: aSymbol
	
	"One of #top, #bottom, #center: indicating where the origin x=0 appears
	 in the frame of the assembly.
       Default is center, for assemblies where x=0 is the visual center.
	 (Necessary so that when I resize the assembly, I know how to adjust position accordingly)"

	hAlign := aSymbol.! !

!QPanelHandler methodsFor: 'initialize' stamp: 'bgf 8/6/2009 23:49'!
initialize
	super initialize.
	vAlign := #center.
	hAlign := #center.
	inset := 0.
	extent := (QDisplayPanel defaultExtent).
	minExtent := 0@0.
	maxExtent := (Float infinity)@(Float infinity).
	resizeFrameOnDemand := true.
	rotationGripOnDemand := false.
	motionGripOnDemand := false.

! !

!QPanelHandler methodsFor: 'accessing' stamp: 'bgf 9/20/2006 23:58'!
inset

	"The offset from any edge of the assembly to the corresponding inner edge of the resize frame."

	^ inset! !

!QPanelHandler methodsFor: 'accessing' stamp: 'bgf 9/20/2006 18:27'!
inset: anInset

	"How far in from the extent (of the assembly) do the inside edges
 	 of the resizeFrame appear?  If zero (the default), the resize frame
	 wraps the exterior of the assembly. If inset = thickness,
       the outer edge of the resize frame coincides with the outer edge of the assembly."

	inset := anInset! !

!QPanelHandler methodsFor: 'events' stamp: 'bgf 12/21/2008 20:25'!
isEmpty
	"Is the panel I'm handling empty of content?
	 This is particular to display panels."
	^ (assembly respondsTo: #isEmpty) and: [ assembly isEmpty ]! !

!QPanelHandler methodsFor: 'events' stamp: 'das 4/30/2007 14:56'!
loseKeyboardControl

	self signal: #loseKeyboardControl.
! !

!QPanelHandler methodsFor: 'accessing' stamp: 'bgf 10/28/2006 17:00'!
maxExtent
	^ maxExtent

! !

!QPanelHandler methodsFor: 'accessing' stamp: 'bgf 10/28/2006 17:14'!
maxExtent:  xY

	"A point defining the largest extent that the resize frame will apply to the assembly."
	maxExtent := xY asPoint.
	self extent: (extent min: maxExtent).! !

!QPanelHandler methodsFor: 'menu' stamp: 'bgf 1/31/2007 18:05'!
menuAccess: aTFrame

	"Make the given frame serve as the menu-access button for this panel."
	{#pointerDown . #pointerUp} do: [ :event |
		self stopScript: (event,':') asSymbol.
		(aTFrame notNil) ifTrue: [
			self runScript: (event , ':') asSymbol when: { aTFrame . event }.
		].
	].


! !

!QPanelHandler methodsFor: 'accessing' stamp: 'bgf 9/20/2006 23:57'!
menuLabel

	"The term used to refer to the panel in user-visible menus."
	^ menuLabel ifNil: [ 'Panel' ] ifNotNil: [ menuLabel ].! !

!QPanelHandler methodsFor: 'accessing' stamp: 'bgf 9/20/2006 23:44'!
menuLabel: aString
	"Specify the name to be used to refer to the assembly in menu labels;
 	 the default is 'Panel'."

	menuLabel := aString.! !

!QPanelHandler methodsFor: 'menu' stamp: 'bgf 9/14/2007 20:37'!
menuTitle
	^ menuLabel ifNil: ['Panel']! !

!QPanelHandler methodsFor: 'accessing' stamp: 'bgf 10/28/2006 17:01'!
minExtent
	^ minExtent! !

!QPanelHandler methodsFor: 'accessing' stamp: 'bgf 10/28/2006 17:14'!
minExtent:  xY

	"A point defining the least extent that the resize frame will apply to the assembly."
	minExtent := xY asPoint.
	self extent: (extent max: minExtent).
! !

!QPanelHandler methodsFor: 'accessing' stamp: 'bgf 10/1/2006 17:59'!
motionAdapter

	"Return the QMotionAdapter that governs the 'Move' behaviors of the panel;
	 this can be used to configure for drop versus move."
	^ motionGrip! !

!QPanelHandler methodsFor: 'accessing' stamp: 'bgf 8/28/2007 22:10'!
motionGrip: aFrame onDemand: aBool

	"Make the given frame (which must be a subframe of the luggage)
	 behave as the movement grip for the panel.
	 This behavior is in effect when the user has chosen Move from the menu.

	 If onDemand is true, the grip will be removed from the luggage when not in use;
       it must be a direct frame child of the luggage in that case, so I know where to re-install it."

	motionGripOnDemand := aBool.
	motionGrip ifNil: [
		motionGrip := QMotionAdapter new.
		motionGrip
			luggage: self assembly;
			isGroundBound: true;
			isAttractedToSurfaces: true;
			attractionRange: 5.0;	" Factor based on original distance "
			openSpacePolicy: #viewer;
			surfacePolicy: #surface.
	].

	motionGrip grip: aFrame.
	self disableMove.! !

!QPanelHandler methodsFor: 'accessing' stamp: 'bgf 9/20/2006 23:44'!
noMotion

	"Permanently disallow moving, removing it as an option from the menu."
	motionGrip ifNotNil: [
		motionGrip disable.
		motionGrip := nil.
	].! !

!QPanelHandler methodsFor: 'accessing' stamp: 'bgf 9/20/2006 23:43'!
noRotation

	"Permanently disallow rotation, removing it as an option from the menu."
	self disableRotate.
	eastPivotGrip := nil.
	westPivotGrip := nil.
	rotGrip := nil.
! !

!QPanelHandler methodsFor: 'accessing' stamp: 'bgf 1/6/2007 04:04'!
pinEdge: compassPt

	"Compass pt is one of #north, #south, #east, #west.
	 Fix the corresponding resize edge bar so that it cannot be dragged during resize.
	 The geometry of the assembly is expected to have its origin at a pnned edge
	 if the opposite edge is not pinned.   (For instance, if the south edge is pinned,
	 and the north is free, then the assembly will be treated as having y=0 
	 aligned with its south edge (rather than its center.)

	 DOES NOT AFFECT ROTATION OR PANEL MOVEMENT."

	| |
	
	resizeFrame pinEdge: compassPt.! !

!QPanelHandler methodsFor: 'accessing' stamp: 'bgf 1/6/2007 04:04'!
pinEdges: compassPoints

	"Fix the indicated edges so they cannot be dragged during resize;
	 DOES NOT AFFECT ROTATION OR PANEL MOVEMENT."

	self resizeFrame pinEdges: compassPoints.! !

!QPanelHandler methodsFor: 'accessing' stamp: 'bgf 9/20/2006 23:36'!
pivotGripEast: aFrame west: aFrame2 onDemand: aBool

	"Make the given frames (which must be subrames of the luggage)
	 behave as paired pivot grips (each pivots about the axis of the other.)
	 This pivot is enabled when the user chooses Rotate from the menu.

	 If onDemand is true, the grips will be removed from the luggage when not in use;
       they must be a direct frame child of the luggage in that case."

	pivotGripOnDemand := aBool.

	eastPivotGrip ifNil: [
		eastPivotGrip := QPlanePivot new.
		eastPivotGrip luggage: (self assembly);
			rotationStep: 1.
	].

	westPivotGrip ifNil: [ 
		westPivotGrip := QPlanePivot new.
		westPivotGrip luggage: (self assembly); 
			rotationStep: 1.
	].

	eastPivotGrip grip: aFrame.
	westPivotGrip grip: aFrame2.
	self disableRotate.
! !

!QPanelHandler methodsFor: 'events' stamp: 'bgf 9/19/2006 10:55'!
pointerDown: event
	"Pointer down in the assembly - prepare for possible menuing."
	downPoint := event point copy.! !

!QPanelHandler methodsFor: 'events' stamp: 'bgf 12/9/2006 11:32'!
pointerUp: event
	"Pointer up in the assembly - menu if this was a simple click"

	event avatar ifNil: [	
		"Must have an avatar to post a menu (no cross-space menus)"
		^ self
	].
	((downPoint - event selection point) squaredLength < 0.01) ifTrue: [
		self showMenu: event.
	].! !

!QPanelHandler methodsFor: 'menu' stamp: 'bgf 12/20/2007 21:37'!
reallyDelete: event

	"Destroy the assembly - action triggered after confirmation of the menu Remove command"
	assembly ifNotNil: [
		assembly reallyDelete: event.
		assembly := nil.
	].
! !

!QPanelHandler methodsFor: 'accessing' stamp: 'bgf 9/20/2006 23:35'!
resizeFrame

	"The QFlatFrame that offers resizing control for this panel."
	^ resizeFrame
! !

!QPanelHandler methodsFor: 'accessing' stamp: 'bgf 9/20/2006 23:35'!
resizeFrameOnDemand: aBool

	"If true, the resize frame is visible only when enabled; if false; always visible."

	resizeFrameOnDemand := aBool.
	resizeFrameOnDemand ifFalse: [
		self assembly addChild: resizeFrame.
	].! !

!QPanelHandler methodsFor: 'accessing' stamp: 'bgf 9/19/2006 17:44'!
rotationGrip: aFrame onDemand: aBool

	"Make the given frame (which must be a subframe of the luggage)
	 behave as the grip in a QCylinderPivot for the luggage.
	 The pivot is enabled when the user chooses Rotate from the menu.

	 If onDemand is true, the grip will be removed from the luggage when not in use;
       it must be a direct frame child of the luggage in that case."

	rotationGripOnDemand := aBool.

	rotGrip ifNil: [
		rotGrip := QCylinderPivot new.
		rotGrip luggage: (self assembly);
			rotationStep: 1.
	].

	rotGrip grip: aFrame.
	self disableRotate.
! !

!QPanelHandler methodsFor: 'accessing' stamp: 'bgf 1/21/2007 16:26'!
scale: aFactor

	minExtent := (minExtent * aFactor).
	maxExtent := (maxExtent * aFactor).
	self thickness: (self thickness * aFactor).
	self inset: (self inset * aFactor).
	self extent: (extent * aFactor).! !

!QPanelHandler methodsFor: 'initialize' stamp: 'bgf 1/5/2007 21:03'!
shapeChanged

	"Layout update - recompute frame geometry after an extent change."
	| x y |

	resizeFrame ifNotNil: [
		resizeFrame extent: (extent - (inset@inset * 2)).

		(vAlign ~= #center) | (hAlign ~= #center) ifTrue: [
			x := hAlign caseOf: {
				[#left] -> [ extent x / 2] .
				[#right] -> [ extent x / -2] .
			} otherwise: [
				0
			].
			y := vAlign caseOf: {
				[#top] -> [ extent y / -2] .
				[#bottom] -> [ extent y / 2] .
			} otherwise: [
				0
			].
			resizeFrame translationX: x y: y z: 0.
		].
	].! !

!QPanelHandler methodsFor: 'menu' stamp: 'bgf 9/1/2009 11:16'!
showMenu:event
	"The menu offers the editing modes for the panel."
	| lbl menu |
	event avatar ifNil: [ ^ self ].
	(event avatar cancelMenu) ifTrue: [ ^ self ].
	assembly ifNil: [ ^ self ].

	"We use a new menu each time, for multi-use."
	lbl := self menuTitle.
	menu := TMenuFrame new.
	menu hideLabelFrame.
	menu alignment: #topLeft.

	(menu addItemSpec: {lbl . #cancelMenu: }  target: self) disabled:true.
	menu addSeparator.

	(assembly respondsTo: #addMenuActions:forEvent:) ifTrue: [
		assembly addMenuActions: menu forEvent: event
	] ifFalse: [
		self addMenuActions: menu forEvent: event.
		menu addSeparator.
		menu addItemSpec: {('Delete ', lbl). #confirmDelete:} target: assembly.
	].
	assembly postMenu: menu forEvent: event.! !

!QPanelHandler methodsFor: 'private' stamp: 'jcg 2/10/2007 20:21'!
signalPointerEnter: event
	"Pass the event on."
	self signal: #pointerEnter with: event.! !

!QPanelHandler methodsFor: 'private' stamp: 'jcg 2/10/2007 20:22'!
signalPointerLeave: event
	"Pass the event on."
	self signal: #pointerLeave with: event.! !

!QPanelHandler methodsFor: 'private' stamp: 'ar 5/17/2007 14:27'!
signalRemoveUser: userID
	"Pass the event on."
	self signal: #removeUser with: userID.! !

!QPanelHandler methodsFor: 'private' stamp: 'bgf 3/15/2010 10:55'!
stateVector
	"Undo/redo support"
	(assembly respondsTo: #standHeight) ifTrue: [
		^ { self extent clone .  assembly translation clone .  (assembly standHeight) }
	] ifFalse: [
		^ { self extent clone .  assembly translation clone }
	].
! !

!QPanelHandler methodsFor: 'accessing' stamp: 'bgf 8/3/2009 10:27'!
thickness
	^ resizeFrame ifNotNil: [
		resizeFrame thickness
	] ifNil: [
		QPanelFrame defaultThickness
	]! !

!QPanelHandler methodsFor: 'accessing' stamp: 'bgf 8/3/2009 10:28'!
thickness: newThickness
	"The thickness of resize frame bars."
	resizeFrame ifNotNil:[ 
		resizeFrame thickness: newThickness.
	].
	self shapeChanged.! !

!QPanelHandler methodsFor: 'menu' stamp: 'bgf 1/23/2007 10:03'!
toggleMove
	self canMove ifTrue: [self disableMove] ifFalse: [self enableMove].! !

!QPanelHandler methodsFor: 'menu' stamp: 'bgf 1/23/2007 10:03'!
toggleResize
	self canResize ifTrue: [self disableResize] ifFalse: [self enableResize].! !

!QPanelHandler methodsFor: 'menu' stamp: 'bgf 1/23/2007 10:00'!
toggleRotate
	self canRotate  ifTrue: [self disableRotate] ifFalse: [self enableRotate].! !

!QPanelHandler methodsFor: 'accessing' stamp: 'bgf 9/20/2006 23:34'!
vAlign: aSymbol

	"One of #top, #bottom, #center: indicating where the origin y=0 appears
	 in the frame of the assembly.
       Default is center, for assemblies where y=0 is the visual center.
	 (Necessary so that when I resize the assembly, I know how to adjust position accordingly)"

	vAlign := aSymbol.! !

!QPanelHandler methodsFor: 'initialize' stamp: 'bgf 1/2/2007 18:41'!
withFrame: aPanelFrame

	"Initialize the panel handler to use the given panel frame as its visual component."

	resizeFrame ifNotNil: [
		resizeFrame assembly: nil.
		resizeFrame destroy.
	].

	resizeFrame := aPanelFrame.
	resizeFrame ifNotNil: [
		resizeFrame assembly: self.
		resizeFrame disable.
	].! !

!QPanelHandler methodsFor: 'initialize' stamp: 'bgf 7/23/2007 10:10'!
withSimpleFrame

	"Created but not initially visible - default rectilinear resize frame."
	resizeFrame := QBoxFrame new.
	resizeFrame 
			assembly: self;
			frameMaterial: (QMaterials translucentEditorMaterial);
			disable.
! !

TObject subclass: #QParticipantJob
	instanceVariableNames: 'jobNumber tries requester jobSpace isUserSpecific done'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Objects-Jobs'!

!QParticipantJob commentStamp: '<historical>' prior: 0!
An abstract class representing a background operation 
that requires roundtrip signalling to a participant harness.

Supports operations that cannot be invoked
directly inside the replicated universe (such as rendering scenes to images,
opening dialogs for the user, etc.)

This class (together with QJobSpace) defines a processing regime
so that participant jobs are performed serially,
and are retried later if the harness services they require are not available.

The 'jobSpace' instance variable is the current island's QJobSpace;
this is an empty neutrally-lit space in which rendering setups may be staged.

A concrete implementation MAY extend these instance methods:

	#begin - Issue the appropriate signal (to #self avatarToSignal) to request service.
			 Should return true if it initiates such a request, false if not.
	  		 If #begin returns false, the job will be forgotten.  
			 If later retry is desired, the job should send
					jobSpace defer:self for: <milliseconds>
			  just before returning false.

	#defer - Put the job into in an inactive state (called when task has begun but not completed),
			 and remove all working state from the job space.

	#finish - release all resources,
 			 and remove working state from the job space.
			 Finish may be called before the harness-side processing is complete;
	 		 it merely indicates that job scheduling has been dealt with.

	#isDone - should return true after '#finish' has been called, and/or after 
	  		the processing done by the harness is finished and the results have been received.

On the class side, you must implement

	#startJob: aJobFarRef forOwner: aParticipantUI
			- Does the work: carries out the harness-side non-replicated job,
			  returning any results by far-ref future calls back to the job instance.

It is required that the startJob:forOwner: implementation 
	job future detach	
(or some in-island extension that calls detach, or jobDone: )
as soon as the job system is free to move to subsequent jobs.

A job enters itself into the queue for processing by calling 'jobSpace addJob: self';
this is usually done only once, after full initialization of the job instace.
!

QParticipantJob subclass: #QAutoActionJob
	instanceVariableNames: 'seconds lastSaved minSeconds pending'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Objects-Jobs'!

!QAutoActionJob commentStamp: 'hrs 3/14/2009 12:27' prior: 0!
A mechanism for creating job that fires off periodically with volunteer answering.

The method names come from its original purpose as QAutoSaveJob, which is now a nearly empty subclass of this.!

!QAutoActionJob class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QAutoActionJob class methodsFor: 'instance creation' stamp: 'bgf 6/24/2007 15:21'!
duration: seconds minDuration: minSeconds space: aSpace
	^self new duration: seconds minDuration: minSeconds space: aSpace! !

!QAutoActionJob class methodsFor: 'instance creation' stamp: 'bgf 6/24/2007 18:04'!
duration: seconds space: aSpace
	"By default, the minDuration is 30 seconds (unless the duration is smaller.)"
	^self new duration: seconds minDuration: (((seconds/2)truncated) min: 30) space: aSpace! !

!QAutoActionJob class methodsFor: 'defaults' stamp: 'bgf 8/23/2007 10:49'!
isEnabledForClient: aParticipantUI
	"A preference controls whether this user participates in autosave.
	 The participant UI arranges for job handler updates when the preference changes;
	 the job handler tests this method before installing its trigger listener."
	^ (aParticipantUI preferences doAutoSave)! !

!QAutoActionJob class methodsFor: 'job handling' stamp: 'hrs 3/14/2009 10:30'!
jobActionTrigger
	"What do we signal to the participant when we start our action (e.g., to do the work)?"
	self subclassResponsibility.! !

!QAutoActionJob class methodsFor: 'job handling' stamp: 'hrs 3/14/2009 10:08'!
jobCompletedTrigger
	"What do we signal when done (e.g., to reset counters and such)?"
	self subclassResponsibility.! !

!QAutoActionJob class methodsFor: 'job handling' stamp: 'bgf 1/27/2008 11:30'!
maxTries
	"The number of job retries to allow before abandoning a job.
	 Since there may be no clients in the space configured to do auto-save,
	 we allow several tries with a short-time so the queue isn't blocked waiting for a volunteer.
	 The confirmSaveBegin protocol ensures that overlapping handlers won't all do a save."
	^ 60! !

!QAutoActionJob class methodsFor: 'job handling' stamp: 'bgf 3/15/2009 22:19'!
notifyUnhandled: anOwner

	"An auto-save event went unanswered;
	 If this client is one of the 'non-cooperators' who has auto save disabled, remind them."

	"If yer not allowed to save, you needn't see this."
	anOwner canModifyForum ifFalse: [ ^ self ].

	anOwner preferences doAutoSave ifFalse: [
		anOwner statusInfo: 'Auto-save is disabled in your preferences.'.
	] ifTrue: [
		QLogger logInfo: 'Disregarded ', self jobActionTrigger, ' event (preference is enabled.)'
	].! !

!QAutoActionJob class methodsFor: 'job handling' stamp: 'bgf 3/10/2010 12:43'!
startJob: job forClient: aParticipantUI
	"Handle the autosave request, unless someone else got there first."
	| resultRef nm |
	job isValidRef ifFalse: [ ^ self ].

	resultRef  := (job future confirmSaveBegin).
	nm := resultRef refName.
	resultRef whenResolved: [ : yesNo |
		(yesNo == true) ifTrue: [
			"Note we don't check readyForEvents here, 
	 		 since auto-saves need to fire around logouts, etc."
			aParticipantUI exitTripped ifFalse: [
				aParticipantUI signal: self jobActionTrigger with: (job get:#requester).
			]
		].
		"Have to use to the promise name directly.
	 	 (The far ref sometimes has a new myName after the resolve completes.)"
		resultRef isValidFarRef ifTrue: [ resultRef island future discardName: nm ].
	].
! !

!QAutoActionJob class methodsFor: 'job handling' stamp: 'bgf 1/27/2008 11:30'!
timeout
	"These use short timeouts and retries so this common job doesn't block the queue."
	^ 500! !

!QAutoActionJob methodsFor: 'initialize' stamp: 'bgf 6/4/2007 14:35'!
autoSaveLoop
	"Perform my auto-save loop"

	((Processor activeIsland totalSeconds - lastSaved) >= seconds) ifTrue: [
		self requestASave.
	].

	"Check each 1/10 of the interval, in milliseconds, so we're fairly prompt...
	 off-schedule saves can push the basic clock around."
	(self future: seconds*100) autoSaveLoop.! !

!QAutoActionJob methodsFor: 'job queue' stamp: 'bgf 1/27/2008 11:50'!
begin
	"Signal through the avatar to the participant, to do the actual work."
	| listener |

	listener := self listenerToSignal ifNil:[
		"if we don't have a potential volunteer avatar,
		 just bag it - no one has autosave turned on."
		self notifyUnhandled.
		^false
	].
	listener signal: self class trigger with: self.

	"We expect these to fail sometimes (when all save prefs are off)
	 so we do not want to invalidate the listener when they do.
	 When a client has autosave disabled, their listener does not listen to as jobs,
	 but listens normally to all others.) "

	jobSpace lastListener: nil.
	^true
! !

!QAutoActionJob methodsFor: 'initialize' stamp: 'ar 6/3/2007 13:01'!
cancelLoop
	"Cancel my auto-save loop"
	Processor activeIsland unscheduleMessageMatching:[:msg|
		msg receiver == self and:[msg selector == #autoSaveLoop].
	].! !

!QAutoActionJob methodsFor: 'job queue' stamp: 'bgf 1/27/2008 12:03'!
confirmSaveBegin
	"Callback from the participant UI to indicate that the save has begun. 
	 This declares the job finished.
	 It returns true if the sender should do the actual save, false otherwise,
	 to prevent overlapping accepters from doing the save in tandem."
	
	self detach.
	pending ifTrue: [			"This caller gets the baton to do the actual save."
		pending := false.
		lastSaved := Processor activeIsland totalSeconds.
		^ true
	].
	^ false! !

!QAutoActionJob methodsFor: 'initialize' stamp: 'hrs 3/14/2009 10:09'!
duration: secs minDuration: minSecs space: aSpace
	"Fire off this auto-save job every n seconds for the given space"

	done := true. "must be done for autoSaveLoop"
	lastSaved := Processor activeIsland totalSeconds.

	self seconds: secs.
	self minSeconds: minSecs.
	self requester: aSpace.

	"Whenever any save completes, reset the counter."	
	self startScript: #onIslandSaved when: { aSpace island . self class jobCompletedTrigger }.

	(self future: seconds*1000) autoSaveLoop.! !

!QAutoActionJob methodsFor: 'initialize' stamp: 'bgf 1/27/2008 11:25'!
initialize
	super initialize.
	seconds := 600.
	minSeconds := 60.	
	pending := false.! !

!QAutoActionJob methodsFor: 'accessing' stamp: 'bgf 6/24/2007 15:20'!
minSeconds
	"The least duration between saves (when induced by space jumps, for example)."
	^ minSeconds! !

!QAutoActionJob methodsFor: 'accessing' stamp: 'bgf 6/24/2007 15:20'!
minSeconds: minSecs
	"The least duration between saves (when induced by space jumps, for example)."
	minSeconds := minSecs! !

!QAutoActionJob methodsFor: 'accessing' stamp: 'bgf 1/27/2008 11:52'!
onIslandSaved

	"We detected a #saved event from the containing island."
	pending := false.	"We don't need to save again automatically, 	
						 so no harness needs to get a yes from confirmSaveBegin"

	lastSaved := Processor activeIsland totalSeconds.! !

!QAutoActionJob methodsFor: 'accessing' stamp: 'bgf 1/27/2008 11:58'!
requestASave

	"If the most recent save was not too recent, request another."

	((Processor activeIsland totalSeconds - lastSaved) > minSeconds)
	 ifTrue: [
		self isDone ifTrue:[
			"Schedule me again if this job is not currently in progress"
			done := false.
			pending := true.			"Pending is the second-phase commit baton;
									 if several harnesses try parallel invocations
									 due to timeout/retry, only one will receive the baton."
			jobSpace addJob: self.
		].
	].! !

!QAutoActionJob methodsFor: 'accessing' stamp: 'ar 6/3/2007 12:57'!
seconds
	"Answer the number of seconds between auto-save attempts"
	^seconds! !

!QAutoActionJob methodsFor: 'accessing' stamp: 'ar 6/3/2007 12:57'!
seconds: aNumber
	"Set the number of seconds between auto-save attempts"
	seconds := aNumber! !

QAutoActionJob subclass: #QAutoImportJob
	instanceVariableNames: 'avatar'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Objects-Jobs'!

!QAutoImportJob commentStamp: 'hrs 3/14/2009 12:28' prior: 0!
Asks a volunteer to see if there are offline imports on the server to be processed.!

!QAutoImportJob class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QAutoImportJob class methodsFor: 'instance creation' stamp: 'hrs 3/30/2009 08:25'!
defaultForForumSpace: aSpace

	"The default forum auto-import configuration:
	 check for imports in customs every at duration seconds."

	^ self duration: 12 minDuration: 5 space: aSpace.! !

!QAutoImportJob class methodsFor: 'defaults' stamp: 'hrs 3/31/2009 09:11'!
isEnabledForClient: aParticipantUI
	"A preference controls whether this user participates in autosave.
	 The job handler tests this method before installing its trigger listener."
	^true! !

!QAutoImportJob class methodsFor: 'job handling' stamp: 'hrs 3/14/2009 10:31'!
jobActionTrigger
	"What do we signal to the participant when we start our action (e.g., to do the work)?"
	^#processCustoms! !

!QAutoImportJob class methodsFor: 'job handling' stamp: 'hrs 3/14/2009 10:10'!
jobCompletedTrigger
	^#customsProcessed! !

!QAutoImportJob class methodsFor: 'job handling' stamp: 'hrs 4/9/2009 10:41'!
notifyUnhandled: anOwner
	"Checking for imports is mandatory in the sense that no one can opt out, so there's no need to notify about people opting out.

	On the other hand it is optional in sense that if there is no one in the forum to observe new imports, we don't care that no one 
	is checking for them or that no one has checked for a given number of retries.
	
	Either way, we don't need to tell folks about unhandled messages."! !

!QAutoImportJob class methodsFor: 'job handling' stamp: 'bgf 3/10/2010 12:42'!
startJob: job forClient: aParticipantUI
	"Handle the processing request, unless someone else got there first."
	| avatar resultRef nm |

	"Not if exiting, or reconnecting, or anything suspect.. only volunteer when happy."
	aParticipantUI readyForEvents ifFalse: [ ^ self ].
	aParticipantUI avatar isDelayed ifTrue: [ ^ self ].

	avatar := aParticipantUI avatar replica.
	(avatar isValidRef and: [avatar island = job island]) ifFalse: [ "eg., teleporting" ^self ].

	aParticipantUI readyForEvents ifFalse: [ ^ self ].

	resultRef  := (job future confirmSaveBegin: avatar).
	nm := resultRef refName.
	resultRef whenResolved: [ : yesNo |
		(yesNo == true) ifTrue: [
			aParticipantUI readyForEvents ifTrue: [ 
				aParticipantUI avatar isDelayed ifFalse: [
					aParticipantUI signal: self jobActionTrigger with: (job get:#requester).
				].
			].
		].

		"Have to use the promise name directly.
	 	 (The far ref sometimes has a new myName after the resolve completes, 
		  if other true-result things are pending.)."
		resultRef isValidFarRef ifTrue: [ resultRef island future discardName: nm ].
	].
! !

!QAutoImportJob methodsFor: 'job queue' stamp: 'hrs 4/24/2009 10:38'!
confirmSaveBegin: aReplica
	"There are two different flags:

	pending is managed by the superclass and means that this round of the job is up for grabs.
	It is true when a request is made and lasts until some participant says they've got the baton.

	avatar is managed by us and means that this round has not yet completed, so that a new
 	round is not started. It is non-nil once someone gets the baton and stays non-nil until the job
	is finished or the partipant has left.
	"
	
	^self confirmSaveBegin ifTrue: [
		(avatar notNil and: [avatar root isNil]) ifTrue: [avatar := nil].  "Don't wait if they've left."
		avatar ifNotNil: [
			false
		] ifNil: [
			avatar := aReplica.
			true
		]
	] ifFalse: [false].
! !

!QAutoImportJob methodsFor: 'accessing' stamp: 'hrs 4/24/2009 10:28'!
onIslandSaved
	super onIslandSaved.
	avatar := nil.! !

QAutoActionJob subclass: #QAutoSaveJob
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Objects-Jobs'!

!QAutoSaveJob commentStamp: 'hrs 3/14/2009 12:26' prior: 0!
Requests auto-saves for the requester space.

Two techniques are supported and can be mixed.
1)  Start a regular timed autosave via
		QAutoSaveJob duration: intervalSeconds space: requestor
     This will generate auto-saves at the given interval (seconds)
	so long as the space is occupied by an avatar.

2) Add extra triggers via
		myASJob startScript: #requestASave when: { object . signal }
	For instance, when: { theSpace . #avatarLeave }.

	Or: 
		myASJob startScript: #requestASlowSave when: { object . signal }
	which requests the save a couple of seconds in the future.
	
This will add extra auto-save triggering events.  
The extra saves can increase the number of autosaves produced...
within reason; no save request is generated unless it is more
minDuration seconds since the last save was performed.!

!QAutoSaveJob class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QAutoSaveJob class methodsFor: 'instance creation' stamp: 'hrs 3/13/2009 20:46'!
defaultForForumSpace: aSpace

	"The default forum auto-save configuration:
	 save every 10 minutes, 
	 and when avatars leave,
	 and when signalled an intention to log out."

	^ (self duration: 600 minDuration: 45 space: aSpace)
		startScript: #requestASlowSave when: { aSpace . #avatarLeave };
		yourself.! !

!QAutoSaveJob class methodsFor: 'job handling' stamp: 'hrs 3/14/2009 10:30'!
jobActionTrigger
	"What do we signal to the participant when we start our action (e.g., to do the work)?"
	^#autoSave! !

!QAutoSaveJob class methodsFor: 'job handling' stamp: 'hrs 3/14/2009 10:09'!
jobCompletedTrigger
	"What do we signal when done (e.g., to reset counters and such)?"
	^#saved! !

!QAutoSaveJob methodsFor: 'accessing' stamp: 'hrs 3/13/2009 20:54'!
requestASlowSave

	"Request a save after a brief delay;
	 useful when the triggering action is future-intensive, like avatar transitions."
	(self future:2000) requestASave
 ! !

QParticipantJob subclass: #QConfirmJob
	instanceVariableNames: 'message yesLabel noLabel arguments targetFrame selector'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Objects-Jobs'!

!QConfirmJob commentStamp: '<historical>' prior: 0!
Support for avatar replica confirm: utilities.

The job runs a confirmation (yes/no) dialog for the requesting user
and sends the response back into the island.

!

!QConfirmJob class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QConfirmJob class methodsFor: 'instance creation' stamp: 'bgf 11/17/2009 15:55'!
confirm: shortMsgString yesLabel: yesLabel selector: yes arguments: args noLabel: noLabel  target: targetFrame forAvatar: anAvatar

	"Initiate a harness-side dialog;
	 see QAvatarReplica 
		confirm: yesLabel: selector: arguments: noLabel:  target:  "

	(self new)
		requester: anAvatar;
		message: shortMsgString;
		yesLabel: yesLabel;	
		selector: yes;
		arguments: args;
		noLabel: noLabel;
		targetFrame: targetFrame;
		confirm
! !

!QConfirmJob class methodsFor: 'defaults' stamp: 'bgf 11/17/2009 15:15'!
maxTries

	"The number of job retries to allow before abandoning a job.
	 Very small, since text editors shouldn't come up long after asking."
	^ 0! !

!QConfirmJob class methodsFor: 'job handling' stamp: 'bgf 11/17/2009 16:10'!
startJob: confirmJob forClient: aUI
	"Get confirmation via dialog and send it back to the job relay."
	| spec result |

	confirmJob future detach.

	"If we don't have the view initialized yet, just do nothing.
	 We have this luxury since no means do-nothing (no callback) in the island."

	aUI readyForEvents ifFalse: [ ^ self ].
	(confirmJob waitTimeoutSeconds: 15) ifFalse: [ ^ self ].
	spec := (confirmJob get: #all3Labels).		"Array of strings."
	spec ifNil: [ ^ self ].
	aUI readyForEvents ifFalse: [ ^ self ].

	"Run the dialog modally so there's no driving around or getting two."
	aUI viewPanel disable.	
	[
		result := CDialog 
				confirm: (spec first) 
				yes: (spec second ifNil: ['Yes'] ) 
				no: (spec third ifNil: ['No']) 
				label: 'Confirm'  
				centerAt: (aUI hand position).
	] ensure: [
		aUI viewPanel ifNotNilDo: [ :panel | panel enable ].
	].

	aUI readyForEvents ifTrue: [
		"Don't take action if we're not in there anymore."
		confirmJob island = (aUI activeSpace island) ifTrue: [
			confirmJob future confirmResponse: (result == true).
		].
	].! !

!QConfirmJob class methodsFor: 'defaults' stamp: 'bgf 11/17/2009 15:15'!
timeout
	"A normal short timeout, since the editor should appear in conjunction with the gesture
	 that requests it."
	^ 3000! !

!QConfirmJob methodsFor: 'accessing' stamp: 'bgf 11/17/2009 15:41'!
all3Labels
	^ { message . yesLabel . noLabel }! !

!QConfirmJob methodsFor: 'accessing' stamp: 'bgf 11/17/2009 16:00'!
arguments: argArray
	"These must remain on-island, anything could be here.
	 (These are the original-tframe-supplied callback arguments 
	 passed back to the yes selector when implementing  a yes result.)"
	arguments := argArray! !

!QConfirmJob methodsFor: 'initialize-release' stamp: 'bgf 11/17/2009 15:17'!
confirm
	"This is the job-initiator"
	jobSpace addJob:self.! !

!QConfirmJob methodsFor: 'job queue' stamp: 'bgf 11/17/2009 16:04'!
confirmResponse: yesNo

	"Sent from the confirmation (harness-side) when complete.
	 This happens -after- the job has been cleared from the job system!!"

	targetFrame root ifNil: [
		"This is safety against having 'delete' and such overlap some other operation."
		QLogger logInfo: 'Confirmation ', (selector), ' dropped; target unrooted.'.
		self island future obliterate: self.
		^ self.
	].

	yesNo ifTrue: [
		(targetFrame respondsTo: selector) ifTrue: [
			| args |
			args := arguments ifNil: [ #() ].
			selector numArgs = args size ifTrue: [
				targetFrame perform: selector withArguments: args
			] ifFalse: [
				targetFrame perform: selector withArguments: (args copyWith:  requester ).
			]
		] ifFalse: [
			QLogger logWarning: 'Unhandled selector in confirm: ', selector from: targetFrame 
		].

	] ifFalse: [
		"No problem - the action just goes away."
	].

	self island future obliterate: self! !

!QConfirmJob methodsFor: 'initialize-release' stamp: 'bgf 11/17/2009 15:15'!
initialize
	super initialize.
	self isUserSpecific: true.
! !

!QConfirmJob methodsFor: 'accessing' stamp: 'bgf 11/17/2009 15:24'!
message: aString
	message := aString! !

!QConfirmJob methodsFor: 'accessing' stamp: 'bgf 11/17/2009 15:24'!
noLabel: aString
	noLabel := aString! !

!QConfirmJob methodsFor: 'accessing' stamp: 'bgf 11/17/2009 15:22'!
selector: aYesSelector
	"The callback selector to invoke on targetFrame: if the user confirms."
	selector := aYesSelector! !

!QConfirmJob methodsFor: 'accessing' stamp: 'bgf 11/17/2009 15:22'!
targetFrame: aFrame
	targetFrame := aFrame! !

!QConfirmJob methodsFor: 'accessing' stamp: 'bgf 11/17/2009 15:23'!
yesLabel
	^ yesLabel! !

!QConfirmJob methodsFor: 'accessing' stamp: 'bgf 11/17/2009 15:23'!
yesLabel: aString
	yesLabel := aString! !

QParticipantJob subclass: #QCopyDocumentJob
	instanceVariableNames: 'document destination'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Objects-Document'!

!QCopyDocumentJob commentStamp: '<historical>' prior: 0!
Copies documents between (private) user inventory 
and (public) forum sharedDocuments   areas  on the server.
As with all Jobs, use starts with the class methods under 'instance creation'.

The 'document' may be anything that supports these protocol methods:

	instance isDocumentInForumStorage
	instance isDocumentInInventory

	class copy: (instance-far-ref) toUserInventory: (participant-ui)
	class copy: (instance-far-ref) toActiveForum: (participant-ui)

There are examples in QSharedDocument and TFileTexture.

The completion of the transfer (whether successful or not)
is indicated with a signal raised on the document (#transferComplete).

!

!QCopyDocumentJob class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QCopyDocumentJob class methodsFor: 'job handling' stamp: 'ar 4/16/2011 17:10'!
copy: aCopyJob toForum: aParticipant

	"Implementation callback: copy the job's document into the current forum shared area."
	| document klass |

	document := aCopyJob get: #document.
	document ifNil: [ ^ self ].
	aCopyJob future detach.	"Don't let it try again after we've engaged."

	document wait.
	klass := (document get: #class).
	(klass respondsTo: #copy:toActiveForum:) ifTrue: [
		aParticipant statusInfo: 'Copying from Inventory...'.

		[ 
			klass copy: document toActiveForum: aParticipant 
		] on: Error do: [ : exc |
			aParticipant inform: ('Cannot copy ', (document get: #shortName),  ' to the Forum:\\', exc description) withCRs.
		].
	] ifFalse: [
		"Unsupported.  (The class of the copy-ee must support the protocol.)"
		aParticipant inform: 
			('Cannot copy content to the Forum.\\An internal error was detected.') withCRs.	
	].
	aCopyJob future transferComplete.
! !

!QCopyDocumentJob class methodsFor: 'job handling' stamp: 'bgf 2/7/2010 12:45'!
copy: aCopyJob toUser: aParticipant
	"Implementation callback: copy the job document into this UI user's private storage."
	| document klass |

	document := aCopyJob get: #document.
	document ifNil: [ ^ self ].
	aCopyJob future detach.	"Don't let it try again after we've engaged."

	"We use a class-side protocol."
	document wait.
	klass := (document get: #class).
	(klass respondsTo: #copy:toUserInventory:) ifTrue: [
		aParticipant statusInfo: 'Copying to Inventory...'. 
		[
			klass copy: document toUserInventory: aParticipant.
		] on: Error do: [ : exc |
			aParticipant inform: ('Cannot copy ', (document get: #shortName),  ' to inventory:\\', 
								(exc description)) withCRs.
		].

	] ifFalse: [
		"Unsupported.  (The class of the copy-ee must support the protocol.)"
		aParticipant inform: 
			('Cannot copy content:\\An internal error was detected.') withCRs.
	].
	aCopyJob future transferComplete.
! !

!QCopyDocumentJob class methodsFor: 'instance creation' stamp: 'bgf 12/7/2008 11:42'!
copyToForum: aDocument forUser: anAvatar

	"Job-requesting entry point (callable from TIsland).
	 Request a background harness-side job to copy the given document
	 into the shared storage of the forum the user currently occupies:

			..../sharedDocuments/ <user-name> / document-file-short-name

	The document must be of a class that supports
		copy: <instance-far-ref> toActiveForum: <participantUI>.
	 That class method will be invoked (in the harness-side processing ocean)
	 to perform the actual work.

	At the end of all that, the document receives
	 	future transferComplete
	and in that method it simply signals #transferComplete.

	The copy may fail; if so, the user is informed (somewhat tersely)
	and the shared document paths are -not- updated,
	so the document continues to refer to the source location.
	transferComplete is signalled after such a failure, 
	so in-space code awaiting the signal has a chance finish."

	(self new)
		requester: anAvatar;
		document: aDocument;
		saveToForum! !

!QCopyDocumentJob class methodsFor: 'instance creation' stamp: 'bgf 12/7/2008 11:42'!
copyToInventory: aDocument forUser: anAvatar

	"Job-requesting entry point (callable from TIsland).
 	 Request a background harness-side job to copy the given document
	 into the given user's inventory storage:
			/users/<name>/public/documents/short-name.

	The document must be of a class that supports
		copy: <instance-far-ref> toUserInventory: <participantUI>.
	 That class method will be invoked (in the harness-side processing ocean)
	 to perform the actual work.

	At the end of all that, the document receives
	 	future transferComplete
	and in that method it simply signals #transferComplete.

	The copy may fail; if so, the user is informed (somewhat tersely)
	and the shared document paths are -not- updated.
	transferComplete -is- signalled after such a failure (and the doc still has old paths.)"

	(self new)
		requester: anAvatar;
		document: aDocument;
		saveToInventory		! !

!QCopyDocumentJob class methodsFor: 'defaults' stamp: 'bgf 6/2/2007 01:54'!
maxTries

	"The number of job retries to allow before abandoning a job."
	^ 2! !

!QCopyDocumentJob class methodsFor: 'job handling' stamp: 'bgf 12/8/2008 10:05'!
startJob: copyJob forClient:aUI

	"Harness-side initiation - this is the beginning of the job execution."
	copyJob future detach.
	copyJob wait.

	aUI readyForEvents ifFalse: [ ^ self ].
	aUI mouseCursor: (Cursor wait); waitTick.
	[
		(copyJob get:#destination) caseOf: {
			[#user] -> [self copy: copyJob toUser: aUI].
			[#forum] -> [self copy: copyJob toForum: aUI].
		} otherwise: [
			aUI inform: ('An internal error occurred.\Your document was not transferred.') 
					withCRs.
			copyJob future transferComplete.
		].
	] ensure: [
		aUI mouseCursor: nil.
	].! !

!QCopyDocumentJob class methodsFor: 'defaults' stamp: 'bgf 6/2/2007 02:36'!
timeout
	"A normal timeout.  There are few retries for this job, so the timeout 
	 is long enough to ensure that the first request is caught (unless some heavy load is going on)."
	^ 2000! !

!QCopyDocumentJob methodsFor: 'accessing' stamp: 'bgf 5/28/2007 18:59'!
destination
	"Where to copy the document?
		#user 		- the current user's personal storage
		#forum	 	- this forum's shared documents directory.
	"
	^ destination! !

!QCopyDocumentJob methodsFor: 'accessing' stamp: 'bgf 12/7/2008 11:40'!
document
	"The document may be anything that supports the class-side protocols
	 for copying... see my class-side job-handling methods,
	 where the document's class methods are examined."
	^ document! !

!QCopyDocumentJob methodsFor: 'accessing' stamp: 'bgf 12/7/2008 11:40'!
document: aDocument
	"The document may be anything that supports the class-side protocols
	 for copying... see my class-side job-handling methods,
	 where the document's class methods are examined."
	document := aDocument! !

!QCopyDocumentJob methodsFor: 'initialize-release' stamp: 'bgf 6/24/2007 15:48'!
initialize
	super initialize.
	self isUserSpecific: true.
! !

!QCopyDocumentJob methodsFor: 'initialize-release' stamp: 'bgf 5/28/2007 22:09'!
saveToForum
	"Called after ininitializing to submit the job for processing,
	 with the destination path being assumed to be:
		user's current forum / sharedDocuments / userName / short-file-name
	"
	destination := #forum.
	jobSpace addJob:self.! !

!QCopyDocumentJob methodsFor: 'initialize-release' stamp: 'bgf 5/28/2007 22:10'!
saveToInventory 
	"Called after ininitializing to submit the job for processing,
	 with the destination path being assumed to be:
		/users/ user-name / public / documents / short-file-name

	And it should be the case that the document originally lives elsewhere,
	usually in forum storage."
	
	destination := #user.
	jobSpace addJob: self.! !

!QCopyDocumentJob methodsFor: 'initialize-release' stamp: 'bgf 12/7/2008 11:45'!
transferComplete

	"This signal is raised whether or not the server-side copying was 'successful'...
	 it merely indicates that we're done.  The user will have been informed if there was a problem;
	 in such cases,  there will be a missing document or resource underlying the island-side reference."

	document ifNotNil: [ document signal: #transferComplete ]! !

QParticipantJob subclass: #QDirectPhotoJob
	instanceVariableNames: 'frame transform recipient selector'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Objects-Jobs'!

!QDirectPhotoJob commentStamp: '<historical>' prior: 0!
Uses the job-space service to request a rendering of a TFrame into a texture,
on behalf of an avatar replica (or other TFrame) inside a space.

QPhotoJobs are created and used from within TSpaces.

Usage (again, this is done within a TSpace, on its island) is:
	QPhotoJob takePhoto: subjectTFrame 
			    forUser: anAvatar 
		            recipient:anObject 
			    selector: aSelector

This schedules a background request; when the photo is done 
and copied into a texture, it sent to anObject by peforming aSelector.
(aSelector should be a 1-argument selector expecting a TTexture.)

The photo is taken  full frontal, in the native setting of the object (including surroundings.)

Rendering is not directly available as a service inside replicated spaces;
rendering is done by the participant harness; the replicated spaces have
no access to the open-gl rendering machinery.   

This service works through a #QPhotoJob signal,
which relays through the QJobHandler of some harness.
In response, the class-side startJob:forClient: method
arranges to have the participant render the job space in a TSnapshot,
packages it as a TTexture, and sends it back into the island.

Some cautions: 

* 'anAvatar' should be an avatar replica in an active space, 
   which that avatar's client is rendering presently.  The mechanism 
   requires a harness and opengl Participant to take the picture.

* 'anAvatar' can actually be any rooted TFrame;
   if it is not an avatar, then the space is searched for avatars,
   one is picked at random, and that avatar is signalled to take the picture;
   so a random participant's client is enlisted to take such pictures.

* If there are no active avatar replicas in any space in the TIsland
   (in which case there is no rendering participant that we can signal)
   then the photo request is deferred indefinitely until one arrives.

!

!QDirectPhotoJob class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QDirectPhotoJob class methodsFor: 'accessing' stamp: 'bgf 8/1/2009 17:01'!
blankForm

	"Creates the form into which the resulting photograph will be composited;
	 this is also the initial form used for the texture when first created,
	 until the rendered image is received back from the participant-UI."

	^ (Form extent: (self photoSize) depth:32)
		fillColor: (Color gray alpha:0.1);
		yourself.! !

!QDirectPhotoJob class methodsFor: 'job handling' stamp: 'bgf 8/1/2009 17:05'!
photoCanceled:snap session: session

	self stopScript: #photoCanceled:session:.
	self stopScript: #photoComplete:forJob:session:.

	"Forget a pending photo request."
	session snapshots remove: snap ifAbsent:[].! !

!QDirectPhotoJob class methodsFor: 'job handling' stamp: 'bgf 11/19/2009 14:05'!
photoComplete:snap forJob: job session: session

	|  txtr tmpFileName dir file tmp data |

	"Deliver the result of a photograph snapshot, sending a texture back to the job.
	 We do this by generating a temporary file with the form in it,
	 and loading that through TContentManager.  
	 This creates a TFileTexture so it benefits from SHA caching.    
	 XXXX Future work in persistence of textures will probably change this."

	self stopScript: #photoCanceled:session:.
	job future snapshotTaken.

	session snapshots remove: snap 
				ifAbsent: [QLogger logWarning: 'Photo-job TSnapshot missing.'].

	"The job has a UUID already - use it as the basis of a file name."
	tmpFileName := (job get:#jobNumber) hex, '.png'.
	[
		"Make a PNG file from the snapshot."
		tmp := QFileDialog getFileLocation:#temp.
		dir := FileDirectory default directoryNamed: tmp.
		file := (dir forceNewFileNamed: tmpFileName).
		file binary.
		(PNGReadWriter on: file) nextPutImage: (snap form).
		file close.
		txtr := TTextures find: tmp, (FileDirectory slash), tmpFileName.
	
		"Upload the encoded image for re-access later (they are small photos.)"
		file := dir readOnlyFileNamed: tmpFileName.
		file binary.
		data := file contentsOfEntireFile.
		
		QLogger logInfo: 'Uploading batch job photo'.
		session owner preferences atServerName: 'public/photos/batch/', tmpFileName 
						 putData: data.
		"Finally adjust the texture so it refers to the server-side document."
		txtr fileName: (session owner preferences 
						 serviceFilePath: 'public/photos/batch/', tmpFileName).

	] ifError: [:msg |
		"Fallback to using a transient form texture."
		QLogger logWarning: 'Failed to generate photo file (', msg, ')'.
		txtr := TTexture new 
				initializeSHA: (TContentManager generateSHA: (TObjectID for: 'QPhotoJob'))
				form: snap form
				mipmap: true 
				shrinkFit: true 
				extension: #alpha
				static: true.
	].

	txtr := TIslandCopier new 
				import:(TIslandCopier new export:txtr) 
				to: job island.

	job future photoComplete:txtr.


! !

!QDirectPhotoJob class methodsFor: 'defaults' stamp: 'bgf 8/1/2009 17:01'!
photoSize
	"The pixel dimensions of the Form for photos taken here;
	  since the bits are future-sent to all participants, we keep it small." 
	^ 256@192! !

!QDirectPhotoJob class methodsFor: 'instance creation' stamp: 'bgf 8/1/2009 17:01'!
photograph: subjectFrame forUser: reqFrame recipient: anObject selector: aSelector
	
	"Queue a photo request of the given subjectFrame
	 on behalf of the given requester (usually an avatar).
	 When the photo is returned, the given selector will be performed on the recipient object."

	^(self new) 
		frame: subjectFrame;
		requester: reqFrame;
		recipient: anObject;
		selector: aSelector;
		takePhoto.
			! !

!QDirectPhotoJob class methodsFor: 'defaults' stamp: 'bgf 8/1/2009 17:01'!
retries
	^ 5! !

!QDirectPhotoJob class methodsFor: 'job handling' stamp: 'bgf 8/1/2009 17:35'!
startJob: photoJob forClient: aUI

	"The harness-side worker routine to actually carry out a photojob request."
	| snap |

	"If we don't have openGL yet, just do nothing;
	 the request will time out and retry later."
	aUI session ogl ifNotNil: [

		snap := (TSnapshot new)
				space: (photoJob get:#space) transform: (photoJob get:#transform);
				form: self blankForm;
				update:true;
				yourself.

		aUI session snapshots add: snap.

		self startScript: #photoCanceled:session:
			withArguments: {snap . aUI session}
			when: {photoJob . #canceled}.

		self startScript: #photoComplete:forJob:session:
			withArguments: {snap . photoJob . aUI session} 
			when: {snap . #formChanged}.
	].! !

!QDirectPhotoJob class methodsFor: 'defaults' stamp: 'bgf 8/1/2009 17:01'!
timeout
	"It can take the harness a while to respond, schedule the snapshot,
	 and generate the snapshot."
	^ 15000! !

!QDirectPhotoJob methodsFor: 'job queue' stamp: 'bgf 8/5/2009 11:05'!
begin
	"Set up the jobSpace for the given photo,
	 and request the processing of the picture;
	 note that this removes the subject from its original space
	 until the job is done."

	| listener |
	listener := (self listenerToSignal) ifNil: [
		"No-one is listening... wait longer to recheck."
		(jobSpace deferJob: self for: 30000) ifFalse: [
			self notifyUnhandled.
		].
		^ false
	].
	QLogger logInfo: 'New photo job ', self jobNumber, ' to ', listener.

	"Position the camera, unless this is preset."
	transform ifNil: [
		transform :=(frame portraitTransform)
	].

	"See QwaqParticipantUI onPhotoRequest"
	listener signal: (self class trigger) with: self.
	^true

! !

!QDirectPhotoJob methodsFor: 'job queue' stamp: 'bgf 8/1/2009 17:03'!
defer
	"Put this job on the back burner;
	 restore the frame to its original situation."

	self signal:#canceled.

! !

!QDirectPhotoJob methodsFor: 'accessing' stamp: 'bgf 8/1/2009 17:01'!
frame
	"The subject frame that we're going to take a picture of."
	^ frame! !

!QDirectPhotoJob methodsFor: 'accessing' stamp: 'bgf 8/1/2009 17:01'!
frame: aTFrame
	frame := aTFrame.
! !

!QDirectPhotoJob methodsFor: 'initialize-release' stamp: 'bgf 8/1/2009 17:01'!
initialize
	super initialize.

! !

!QDirectPhotoJob methodsFor: 'job queue' stamp: 'bgf 8/1/2009 17:02'!
photoComplete: aTexture
	"Calllback from the harness when the current photo is generated."
	QLogger logInfo: 'Got photo ',  jobNumber, ' for ', recipient.
	recipient ifNotNil: [
		recipient perform: selector with: aTexture.
		recipient := nil.
	].
! !

!QDirectPhotoJob methodsFor: 'accessing' stamp: 'bgf 8/1/2009 17:01'!
recipient 
	"the object that will be called back when the photo is ready."
	^ recipient! !

!QDirectPhotoJob methodsFor: 'accessing' stamp: 'bgf 8/1/2009 17:01'!
recipient: anObject
	"the object that will be called back when the photo is ready."
	recipient := anObject! !

!QDirectPhotoJob methodsFor: 'accessing' stamp: 'bgf 8/1/2009 17:01'!
selector
	"The selector that will be performed on the recipient to return the photo;
	 must be a single-position method selector; it will be invoked with a TTexture as argument."
	^ selector! !

!QDirectPhotoJob methodsFor: 'accessing' stamp: 'bgf 8/1/2009 17:01'!
selector: aSelector
	"The selector that will be performed on the recipient to return the photo;
	 must be a single-position method selector; it will be invoked with a TTexture as argument."
	selector := aSelector! !

!QDirectPhotoJob methodsFor: 'initialize-release' stamp: 'bgf 8/1/2009 17:01'!
snapshotTaken
	"The harness has delivered a snapshot to the handler;
	 we're done with the scheduling system, 
	 and can restore the object to its original posture."

	self finish.
	jobSpace jobDone: self.! !

!QDirectPhotoJob methodsFor: 'accessing' stamp: 'bgf 8/1/2009 17:06'!
space
	^ frame ifNotNil: [ frame root ]! !

!QDirectPhotoJob methodsFor: 'initialize-release' stamp: 'bgf 8/1/2009 17:01'!
takePhoto
	"Schedule the photo session; assumes we've been properly initialized."

	"We use future here so that the mechanism can be used more easily
	 during construction of container TFrames, which often are not 'rooted'
	 in space until after the basic internal construction finishes."
	jobSpace future addJob: self.
! !

!QDirectPhotoJob methodsFor: 'accessing' stamp: 'bgf 8/1/2009 17:01'!
transform

	"The camera position for the picture;
	 this is set up by #begin, and read by the participant UI."

	^transform! !

!QDirectPhotoJob methodsFor: 'accessing' stamp: 'bgf 8/1/2009 17:01'!
transform: aTransform
	"The transform from which to take the picture."
	transform := aTransform! !

QDirectPhotoJob subclass: #QStudioPhotoJob
	instanceVariableNames: 'saveParent saveTransform saveIndex'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Objects-Jobs'!

!QStudioPhotoJob commentStamp: '<historical>' prior: 0!
Uses the job-space service to request a rendering of a TFrame into a texture,
on behalf of an avatar replica (or other TFrame) inside a space.

The photo is taken at a slight (35-degree) angle from full frontal,
in the well-lit limbo defined by QJobSpace.

Some cautions: 

* The subject tframe is "borrowed":  it is reparented into the studio 
   for the duration of the photo-taking process.
!

!QStudioPhotoJob class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QStudioPhotoJob class methodsFor: 'defaults' stamp: 'bgf 7/20/2007 09:34'!
retries
	^ 5! !

!QStudioPhotoJob class methodsFor: 'defaults' stamp: 'bgf 7/20/2007 09:29'!
timeout
	"It can take the harness a while to respond, schedule the snapshot,
	 and generate the snapshot."
	^ 15000! !

!QStudioPhotoJob methodsFor: 'job queue' stamp: 'bgf 2/13/2010 10:50'!
begin

	"Set up the jobSpace for the given photo,
	 and request the processing of the picture;
	 note that this removes the subject from its original space
	 until the job is done."

	| camera radius listener |

	listener := (self listenerToSignal) ifNil: [
		"No-one is listening... wait longer to recheck."
		(jobSpace deferJob: self for: 30000) ifFalse: [
			self notifyUnhandled.
		].
		^ false
	].
	QLogger logInfo: 'New photo job ', self jobNumber, ' to ', listener.

	saveTransform ifNil: [
		"Grab the situation before for the first runthrough."
		saveParent := frame parent.
		saveTransform := frame localTransform copy.
		saveIndex := (saveParent ifNotNil: [ saveParent indexOfChild: frame ]).
	].

	"Pose the subject."
	QLogger logInfo: 'Job space claiming child: ', frame.
	jobSpace island = frame island ifFalse: [ self error: 'Island mismatch!!'].

	jobSpace addChild: frame.
	frame localTransform: (Matrix4x4 new).
	frame translation: (frame photoBox center negated).  "Visual center at 0,0"
	frame allUpdateShape.

	"Position the camera, unless this is preset."
	transform ifNil: [
		| box |
		camera := Matrix4x4 new.
		box := frame photoBox.
		radius := (box extentY  *1.9) max: ((box extentX max: (box extentZ))*1.2).
		camera translation: (radius * 0.6)@0@(radius * 0.8).	"3-4-5 triangle"
		camera rotationAroundY: 39.
		self transform: camera.
	].

	"See QwaqParticipantUI onPhotoRequest"
	listener signal: (self class trigger) with: self.
	^true

! !

!QStudioPhotoJob methodsFor: 'job queue' stamp: 'bgf 8/21/2008 20:27'!
defer
	"Put this job on the back burner;
	 restore the frame to its original situation."

	self signal:#canceled.

	QLogger logInfo: 'Deferring photo job.'.
	saveParent ifNotNil: [
		saveParent 
			ifNotNilDo: [:p| p addChild: frame].
		saveTransform
			ifNotNilDo: [:t| frame localTransform: saveTransform].
	].
! !

!QStudioPhotoJob methodsFor: 'job queue' stamp: 'bgf 10/19/2009 20:47'!
finish

	"Restore the frame to its original situation."
	super finish.

	saveParent ifNotNilDo: [:p| 
		saveIndex ifNotNil: [
			[  p addChild: frame atIndex: saveIndex ] 
			   on: Error do: [ : ex |
				QLogger logWarning: 'Possible reordering during photo job.'.
				p addChild: frame 
			].
		] ifNil: [
			p addChild: frame
		]].
	saveTransform
		ifNotNilDo: [:t| frame localTransform: saveTransform].

	saveParent := nil.
	saveTransform := nil.
	frame := nil.! !

!QStudioPhotoJob methodsFor: 'initialize-release' stamp: 'bgf 6/24/2007 15:55'!
initialize
	super initialize.

! !

!QStudioPhotoJob methodsFor: 'accessing' stamp: 'bgf 8/1/2009 17:59'!
space
	"Unlike my base class, I take the photo in the JobSpace rather than the frame's home space."
	^ self jobSpace! !

QParticipantJob subclass: #QMailDocumentJob
	instanceVariableNames: 'document'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Objects-Document'!

!QMailDocumentJob commentStamp: '<historical>' prior: 0!
Mails a document (or a link to the document) to the current People list.!

!QMailDocumentJob class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QMailDocumentJob class methodsFor: 'instance creation' stamp: 'hrs 4/4/2009 15:39'!
mail: aDocument fromUser: anAvatar
	"Initiate a harness-side mail client for the give doc."

	(self new) 
		requester: anAvatar;
		document: aDocument;
		mailDocument.! !

!QMailDocumentJob class methodsFor: 'defaults' stamp: 'hrs 4/4/2009 15:21'!
maxTries

	"The number of job retries to allow before abandoning a job.
	 This is interactive, so only one shot."
	^ 0! !

!QMailDocumentJob class methodsFor: 'job handling' stamp: 'hrs 4/7/2009 21:24'!
startJob: job forClient:aUI
	"Harness-side document save handler; see QSaveDocumentJob."
	| docName docRef |

	job future detach.
	job wait.

	docRef := job get:#document.
	docRef isString ifTrue: ["e.g., a url" ^aUI mailDocumentLink: docRef].
	docRef wait.	
	[	docName := (docRef get:#class) mailDocument: docRef forClient: aUI.
		docName ifNotNil: [
			aUI statusAlert: 'Mail composition window opened'.
		] ifNil: [
			aUI statusAlert: 'Cannot mail document.'.
		].
	] on: Error do: [ :exc|
		self inform: ('Could not mail document.\\', exc messageText) withCRs.
	].
! !

!QMailDocumentJob class methodsFor: 'defaults' stamp: 'hrs 4/4/2009 15:21'!
timeout
	"A short timeout, since the editor should appear in conjunction with the gesture
	 that requests it.  (But long enough to deal with slow native dialog startup)"
	^ 5000! !

!QMailDocumentJob methodsFor: 'accessing' stamp: 'hrs 4/4/2009 15:21'!
document
	^ document! !

!QMailDocumentJob methodsFor: 'accessing' stamp: 'hrs 4/4/2009 15:21'!
document: aDocument
	document := aDocument! !

!QMailDocumentJob methodsFor: 'initialize-release' stamp: 'hrs 4/4/2009 15:21'!
initialize
	super initialize.
	self isUserSpecific: true.
! !

!QMailDocumentJob methodsFor: 'initialize-release' stamp: 'hrs 4/4/2009 15:28'!
mailDocument
	jobSpace addJob:self.! !

QParticipantJob subclass: #QOpenDocumentWebPageJob
	instanceVariableNames: 'document'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Objects-Document'!

!QOpenDocumentWebPageJob commentStamp: 'hrs 4/6/2009 21:08' prior: 0!
Open a document Web Page in the user's default browser.!

!QOpenDocumentWebPageJob class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QOpenDocumentWebPageJob class methodsFor: 'defaults' stamp: 'hrs 4/6/2009 20:56'!
maxTries

	"The number of job retries to allow before abandoning a job.
	 This is interactive, so only one shot."
	^ 0! !

!QOpenDocumentWebPageJob class methodsFor: 'instance creation' stamp: 'hrs 4/6/2009 20:58'!
openWebPage: aDocument fromUser: anAvatar
	"Initiate a harness-side web browser for the given doc."

	(self new) 
		requester: anAvatar;
		document: aDocument;
		openDocumentWebPage.! !

!QOpenDocumentWebPageJob class methodsFor: 'job handling' stamp: 'hrs 4/7/2009 20:55'!
startJob: job forClient:aUI
	"Harness-side document save handler; see QSaveDocumentJob."
	| docName docRef |
	job future detach.
	job wait.

	docRef := job get:#document.
	docRef wait.	
	[	docName := (docRef get:#class) openDocumentWebPage: docRef forClient: aUI.
		docName ifNotNil: [
			aUI statusAlert: 'Web browser opened'.
		] ifNil: [
			aUI statusAlert: 'The document does not have a Web page.'.
		].
	] on: Error do: [ :exc| 
		^self inform: ('Could not open document Web Page.\\', exc messageText) withCRs.
	].

! !

!QOpenDocumentWebPageJob class methodsFor: 'defaults' stamp: 'hrs 4/6/2009 20:56'!
timeout
	"A short timeout, since the editor should appear in conjunction with the gesture
	 that requests it.  (But long enough to deal with slow native dialog startup)"
	^ 5000! !

!QOpenDocumentWebPageJob methodsFor: 'accessing' stamp: 'hrs 4/6/2009 20:56'!
document
	^ document! !

!QOpenDocumentWebPageJob methodsFor: 'accessing' stamp: 'hrs 4/6/2009 20:56'!
document: aDocument
	document := aDocument! !

!QOpenDocumentWebPageJob methodsFor: 'initialize-release' stamp: 'hrs 4/6/2009 20:56'!
initialize
	super initialize.
	self isUserSpecific: true.
! !

!QOpenDocumentWebPageJob methodsFor: 'initialize-release' stamp: 'hrs 4/6/2009 20:56'!
openDocumentWebPage
	jobSpace addJob:self.! !

!QParticipantJob class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QParticipantJob class methodsFor: 'defaults' stamp: 'bgf 8/23/2007 10:51'!
isEnabledForClient: aParticipantUI
	"For subclass override - should we handle jobs of this kind, in this participant.?
	 (If this changes dynamically, a call to the harness's QJobHandler #resetListeners
	 is needed to carry out the change,  so the participant UI will need to be involved.)"
	^ true! !

!QParticipantJob class methodsFor: 'defaults' stamp: 'bgf 1/16/2007 09:45'!
maxTries

	"The number of job retries to allow before abandoning a job;
	 a job class can override this class method."
	^ 30! !

!QParticipantJob class methodsFor: 'job handling' stamp: 'bgf 8/21/2008 14:01'!
notifyUnhandled: aClientUI

	"For class-specific override... a job request has been issued,
	 but no handler+listener was found to carry it out."
	QLogger logInfo: '', self class name, ' signal was not handled'.! !

!QParticipantJob class methodsFor: 'job handling' stamp: 'bgf 1/22/2008 02:56'!
startJob: job forClient: aParticipantUI
	"Handle a job request.
	 The processing that ensues should call detach, or jobDone:, as soon as possible."
	self subclassResponsibility! !

!QParticipantJob class methodsFor: 'defaults' stamp: 'bgf 1/16/2007 10:37'!
timeout

	"The number ms to wait before assuming the job hasn't been executed."
	^ 5000! !

!QParticipantJob class methodsFor: 'job handling' stamp: 'bgf 3/12/2007 00:09'!
trigger
	"The event with which the job is requested;
	 always signalled with one argument - the job instance."
	^ self name! !

!QParticipantJob methodsFor: 'job queue' stamp: 'bgf 8/23/2007 10:32'!
begin

	"Initiate the participant request.
	 Return true if the signal was emitted and an answer expected; false otherwise.
	 The default implementation signals only the requester harness if the requester is an avatar,
	 or signals any available harness at random if the requestor is any other in-space object.
	 If no valid listener is identified, the job is abandoned.

	 Protocol requirement: return true if the trigger signal is emitted, 
	 false otherwise (i.e. if no valid listener was detected.) "

	| listener |
	listener := self listenerToSignal.
	listener ifNotNil: [
		listener signal:(self class trigger) with: self.
		^ true.
	] ifNil: [
		self notifyUnhandled.
		^ false
	].
! !

!QParticipantJob methodsFor: 'job queue' stamp: 'bgf 1/11/2007 14:28'!
countTry
	tries := tries + 1.! !

!QParticipantJob methodsFor: 'job queue' stamp: 'bgf 6/24/2007 15:33'!
defer
	"If anything must be cleaned up when a job is requested and then deferred,
	 do it here."! !

!QParticipantJob methodsFor: 'job queue' stamp: 'bgf 1/22/2008 02:33'!
detach

	"The job has been launched in the harness, 
       and will proceed in the background.  
	 Other jobs can be processed."

	done := true.
	jobSpace jobDone: self.! !

!QParticipantJob methodsFor: 'job queue' stamp: 'bgf 6/24/2007 15:45'!
finish
	"The scheduler calls this when it is abandoning a job 
	 (usually due to the retry limit being exceeded;
	 after this has been called, isDone should return true."

	done := true! !

!QParticipantJob methodsFor: 'initialize-release' stamp: 'bgf 6/24/2007 15:44'!
initialize
	super initialize.

	done := false.
	isUserSpecific := false.
	Processor activeIsland register: self name: self jobNumber.
	jobSpace := QJobSpace current.
	tries := 0.! !

!QParticipantJob methodsFor: 'job queue' stamp: 'bgf 6/24/2007 15:45'!
isDone
	"Return true if the job has received its response from the participant harness,
	 or if #finish has been called on the job."
	^ done! !

!QParticipantJob methodsFor: 'accessing' stamp: 'bgf 6/24/2007 15:26'!
isUserSpecific

	"Jobs are user-specific if they must be requested by an avatar
 	an executed by that avatar's session harness."

	^ isUserSpecific! !

!QParticipantJob methodsFor: 'accessing' stamp: 'bgf 6/24/2007 15:26'!
isUserSpecific: bool

	"Jobs are user-specific if they must be requested by an avatar
 	an executed by that avatar's session harness."

	isUserSpecific := true! !

!QParticipantJob methodsFor: 'accessing' stamp: 'bgf 1/11/2007 14:47'!
jobNumber
	^ (jobNumber ifNil: [jobNumber := TObjectID for: 'Participant Job']).! !

!QParticipantJob methodsFor: 'accessing' stamp: 'bgf 1/11/2007 14:40'!
jobSpace
	"The Job Space in which this Job is managed."
	^ jobSpace! !

!QParticipantJob methodsFor: 'job queue' stamp: 'bgf 8/21/2008 20:25'!
listenerToSignal

	| candidates result |

	"Determine the job listener to signal to initiate the participant-side process.
	 If the requester is an avatar, then we signal the corresponding JobListener.
	 If the requester is not an avatar, then we pick a listener at random from the island."

	jobSpace lastListener: nil.

	(requester isKindOf: TAvatarReplica) ifTrue: [
		^ self island at: ((requester userID hex, '-JobListener') asSymbol).
	].

	self isUserSpecific ifTrue: [
		QLogger logWarning: 'User-Specific ', self class name, ' - requester is not an avatar.'.
		^ nil
	].

	candidates := Set new.
	self island globals associationsDo: [ :a |
		(a key isString and:[a key endsWith: '-JobListener']) ifTrue: [
			candidates add: a.
		]
	].
		
	result := candidates ifNotEmpty: [ candidates atRandom ] ifEmpty: [ nil ].
	result ifNotNil: [	
		jobSpace lastListener: result.
		^ result value
	].
	^ nil
! !

!QParticipantJob methodsFor: 'job queue' stamp: 'bgf 8/23/2007 12:43'!
notifyUnhandled
	Processor activeIsland signal: #unhandledJob with: self.
	done := true.! !

!QParticipantJob methodsFor: 'accessing' stamp: 'bgf 1/11/2007 14:33'!
requester
	"The requester is ideally a TAvatarReplica of some sort, but optionally may be
	 any frame rooted in a space that has active participants."
	^requester! !

!QParticipantJob methodsFor: 'accessing' stamp: 'bgf 6/24/2007 15:30'!
requester: anAvatarOrFrame

	"The requester is the frame or user-avatar on whose behalf
	 the job is required.  For non-user-specific jobs, the requester may be any object in the island.

	 If the job is user-specific, the requester must be an avatar,
	 because the avatar's identity allows us to select that user's harness for the job. "

	requester := anAvatarOrFrame.
	self isUserSpecific ifTrue: [
		(requester isKindOf: TAvatarReplica) ifFalse: [
			self error: 'User-specific job requester must be an avatar.'.
		].
	].! !

!QParticipantJob methodsFor: 'accessing' stamp: 'bgf 1/11/2007 14:46'!
tries
	"The number of times the 'begin' harness signal has been sent."
	^ tries! !

QParticipantJob subclass: #QSaveDocumentJob
	instanceVariableNames: 'document'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Objects-Document'!

!QSaveDocumentJob commentStamp: '<historical>' prior: 0!
Saves a QTextDocument for a particular user,
via the ParticipantUI's saveDocumentRequest handler.

The final step of saving a local copy relies on the document's
class method #saveDocument:aDocFarRef asClientFile:aFileName;
this allows different document subclasses to acquire
and write their shared content as they see fit.!

QSaveDocumentJob subclass: #QCreateStoredObjectJob
	instanceVariableNames: 'recipient'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Objects-Document'!

!QCreateStoredObjectJob commentStamp: '<historical>' prior: 0!
Applicable only to QSharedDocuments (or QTextDocuments) that 
contain XML stored scenes.  (See QFileCard class #forTFrame: user:).

Instantiates an instance of the scene in the island of the requestor,
and returns that instance in the objectCreated: callback.

Usage begins with the class method
	instantiate: doc recipient: aCallbackRecipient forUser: anAvatar

!

!QCreateStoredObjectJob class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QCreateStoredObjectJob class methodsFor: 'instance creation' stamp: 'bgf 6/1/2007 23:07'!
instantiate: doc recipient: aCallbackRecipient forUser: anAvatar

	"This is the in-space constructor used to initiate an object-creation request.
	 doc is the scene document, recipient will receive #objectCreated: theNewObject
	 after the scene document is instantiated, and anAvatar is the avatar replica for whom the action is carried out."

	(self new) 
		requester: anAvatar;
		recipient: aCallbackRecipient;
		document: doc;
		createObject.! !

!QCreateStoredObjectJob class methodsFor: 'job handling' stamp: 'bgf 12/29/2009 22:09'!
startJob: job forClient: aUI

	"Main inner work routine:
	 Fetch the persistent-xml data from the document,
	 and send it to the scene reader.
	 Then import the resuling object into the target space."
	| document data object archive |

	job future detach.		"Done with the scheduling system now."
	document := job get:#document.
	document wait.
	aUI statusInfo: 'Processing saved objects...'.
	aUI mouseCursor: Cursor wait.
	[
		"Get the data... it may be in-island textdoc (legacy case) or in a shared document."
		data := [ ((document get:#class) getContent: document forClient: aUI) ] 
			      on: Error 
			      do: [ : ex | QLogger logWarning: ( 'Retrieval failed: ', ex description ).
				       nil ].
		data ifNil: [
			self inform: 'Cannot retrieve stored scene data.'.
			job future objectCreated: nil.
			^ self
		].

		"Instantiate new objects from the data, which we expect to be a stored scene."
		[[
			archive := ZipArchive new.
			((document get:#sharedName) endsWith: 'c3z') ifTrue: [
				archive readFrom: data readStream.
			] ifFalse: [
				"Legacy - assume it is plain xml text"
				archive addDeflateString: data as: 'scene.xml'.
			].
			aUI statusInfo: 'Creating new instance...'; waitTick.

			object := aUI session
				importPersistedScene: archive 
				named: (document get:#shortName)
				toIsland: (job island).

			object ifNotNil: [
				aUI statusInfo: 'Inserting object into scene...'; waitTick.
			].
			job future objectCreated: object.
		] on: CProgressCancel do: [
			QLogger logInfo: 'Canceled stored-object creation'.
			job future objectCreated: nil.
			^ false
		]
		] on: Error do: [ :ex|
			aUI inform: ('Cannot create that object:\\', ex description) withCRs.
			QLogger logError: ex.
			job future objectCreated: nil.
		].
	] ensure: [ aUI mouseCursor: nil ].! !

!QCreateStoredObjectJob class methodsFor: 'defaults' stamp: 'bgf 6/2/2007 02:35'!
timeout
	"A normal timeout.  There are no retries for this job, so the timeout 
	 is long enough to ensure that the first request is caught (unless some heavy load is going on)."
	^ 2000! !

!QCreateStoredObjectJob methodsFor: 'initialize-release' stamp: 'bgf 6/1/2007 22:47'!
createObject
	"fire off the job"
	jobSpace addJob:self.! !

!QCreateStoredObjectJob methodsFor: 'initialize-release' stamp: 'bgf 6/24/2007 15:34'!
initialize
	super initialize.
	self isUserSpecific: true.! !

!QCreateStoredObjectJob methodsFor: 'job queue' stamp: 'bgf 6/5/2009 22:46'!
objectCreated: anObject
	"Callback from the harness-side; deliver the created object to the recipient."

	((anObject isNil not) and:[ anObject isRealMView ]) ifTrue:[
		requester miramarUI ifFalse:[
			requester informUser: 'Cannot drop this object into this space.'.
			recipient ifNotNil: [ recipient objectCreated: nil ].
			^self
		].
	].

	anObject ifNotNil: [
		anObject isFromTemplate: false;
				  isPlacementLocked: false
	].
	recipient ifNotNil: [ recipient objectCreated: anObject ].! !

!QCreateStoredObjectJob methodsFor: 'job queue' stamp: 'bgf 6/1/2007 22:54'!
recipient
	^ recipient! !

!QCreateStoredObjectJob methodsFor: 'job queue' stamp: 'bgf 6/2/2007 01:03'!
recipient: aCallbackRecipient

	"The in-island object that receives the constructed object.
	 The job calls
		recipient objectCreated: anObject
	 to complete its work and deliver the object."

	recipient := aCallbackRecipient! !

QSaveDocumentJob subclass: #QInstallPython
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Objects-Document'!

!QInstallPython commentStamp: 'das 12/18/2008 15:04' prior: 0!
Installing Python Apps

This is done via drag and drop. It will either generate a card or install directly into a panel.

At this point, it is a simple shared document (via subclass PyPanelDocCS)- the Python app has not been actually installed to run. To install, the user needs to either click in the panel, or select the card's menu to install the Python app.

When a user clicks on a document holding an uninstalled app, the system attempts to run it via the PyPanelApp#startRunning: method. Of the Python shared doc has never been installed, we know that we need to query the user if he would like to install it. We depend on the values of lastEvent to determine who is attempting to install the app. If this value is nil, such as it is when we first start up, this attempt to install is ignored. Otherwise a menu is generated with "Install Python Application" as the option. If selected, then PyPanelApp#doInstallPythonApp is sent.

#doInstallPythonApp will set the ivar pyInstalledUserID, set the #onPyAppInstalled method to be triggered when the document signals #pyAppInstalled, and then sends the PyPanelDocCS#doInstallPython: method. This uses the QInstallPython class to retrieve the Python application as a shared doc. In the QInstallPython class#startJob:forClient: method, we create a confirmation dialog where the user can continue to retrieve the document or exit. If the install is allowed to continue, then the QInstallPython class#doOpenPrivately:forClient: is sent. This both installs the Python code in the QwaqPython folder in the Qwaq cache, and generates a C3X card for the objects menu in the Installed Applications folder, also in the cache.

Once complete, the document is future sent #pyAppInstalled, which in turn signals #pyAppInstalled which in turn triggers the PyPanelApp#onPyAppInstalled method described above. Since we know that app has been installed, we now attempt to run the applicatio via the same PyPanelApp#startRunning: that first triggered this entire sequence, only this time we have an installed Python application.

When the Python application has been loaded by a user and then the forum has been saved, if a new user then joins and attempts to launch the application by clicking in it, he will get an error in the PyPanelWorldCS>>#installApp. At this point, if the shared document includes the original code, we will send the PyPanelApp>>#doInstallPythonApp and process just as above.!

!QInstallPython class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QInstallPython class methodsFor: 'as yet unclassified' stamp: 'das 12/20/2009 15:23'!
doNotice: document

	"Remind the user of a little fine print..."

	| msg result userSource |
	msg := 'The file ', ((document get:#tpzName)copyAfterLast:$/), '\will be copied to your Python applications and will\be available for execution once this is complete.\\Warning:\This Python application may contain code that\can harm your system.\'withCRs.
	userSource := document get: #userSource.
	userSource ifNotNil:[msg := (msg, '\This file was uploaded by:\', userSource) withCRs.].

	result := CDialog confirm: msg label: 'Installing Python Application'.
	result := result ifNil: [ false ].

	^ result
! !

!QInstallPython class methodsFor: 'as yet unclassified' stamp: 'ar 4/16/2011 17:13'!
doOpenPrivateCopy: document forClient: aUI

	"Get a local copy of the Python application file if we don't already have one,
	 and create a menu item for it so that it can be launched via the client."

	| data cacheDir pyDir unzipper menuDir menuFile menuString initFile appName userSource sha tpzName cacheLocally |
	document wait.
	[[
		data := (document get:#class) getCode: document forClient: aUI.
		appName := (document get: #tpzName)copyAfterLast:$/.
		userSource := document get: #userSource.
		sha := document get: #sha.
		tpzName := document get: #tpzName.
		cacheLocally := document get: #cacheLocally.
		(appName includes: $.) ifTrue:[appName  := appName copyUpToLast: $..].
		(appName includes: $-) ifTrue:[appName  := appName copyUpToLast: $-.].
		data ifNil: [
			self inform: 'Cannot install Python Application.'.
			^ self 
		].
		cacheDir := FileDirectory cacheDirectory.
		cacheDir assureExistenceOfPath: 'OpenQwaqPythonUserApps'.
		cacheDir assureExistenceOfPath: 'Installed Applications'.
		pyDir := cacheDir directoryNamed: 'OpenQwaqPythonUserApps'.
		(pyDir fileExists:'__init__.py') ifFalse:[ "we need an empty __init__.py file in this directory"
			initFile := pyDir forceNewFileNamed:'__init__.py'.
			initFile close.
		].
		unzipper := ZipArchive new.
		unzipper readFrom: (ReadStream on:data).
		unzipper extractAllTo: pyDir informing: nil overwrite: true.
		menuDir := cacheDir directoryNamed:'Installed Applications'.
		menuFile := menuDir forceNewFileNamed: appName,'.c3x'. "(document get: #initialName),'.c3x'."
		menuString  := 
'<group id="1" class="QFileCard">
	<name>', appName, '</name>
	<kind>panelApp</kind>
	<content>
		<object id="2" class="PyPanelDocCS">
			<appName>OpenQwaqPythonUserApps.', appName, '</appName>'.
			sha ifNotNil:[menuString := menuString,'
			<sha>',sha, '</sha>'.].
			tpzName ifNotNil:[menuString := menuString,'
			<tpzName>',tpzName, '</tpzName>'.].
			cacheLocally ifNotNil:[menuString := menuString,'
			<cacheLocally>',cacheLocally, '</cacheLocally>'.].
			"docName ifNotNil:[menuString := menuString,'
			<docName>',docName, '</docName>'.]."
			userSource ifNotNil:[menuString := menuString,'
			<userSource>', userSource, '</userSource>'.].
			menuString := menuString,'
		</object>
	</content>
</group>'.
	
	[menuFile nextPutAll: menuString] ensure: [menuFile close].

	"Get this new application into the content catalog."
	aUI catalogPane addCardFromFile: (menuFile localName) 
				     directory: menuDir 
				     category: 'Applications (Installed)'.

	document future pyAppInstalled.
"		
		fileName := 'Python',  FileDirectory slash,  (document get:#shortName).
		file := [cacheDir forceNewFileNamed:fileName] 
			on: Error do:[:ex| 
				QLogger logError: ex from: self.
				nil].
		file ifNil:[
			self inform: 'Cannot install Python Application'. ^self].
		data := (document get:#class) getContent: document forClient: aUI.
		data ifNil: [
			self inform: 'Cannot install Python Application.'.
			^ self 
		]."
		"file := FileStream newFileNamed: fileName."

"		file perform: (document get:#fileModeSelector).
		[file nextPutAll: data] ensure: [file close]."

	 ] on: CProgressCancel do: [
		QLogger logInfo: 'Canceled Python install.'.
		^self
	 ]
	] on: Error do: [ :exc|
		self inform: ('Cannot install the Python application.\\', exc description) withCRs.
		QLogger logError: exc from: self.
		^ self
	].
	
	QActivityCapture action: #InstallPythonApplication details: (QActivityCapture fileDetails:  (document get:#shortName)).

	"aUI shellOpen: fileName documentDescription: 'document' applicationDescription: 'application'."! !

!QInstallPython class methodsFor: 'installing' stamp: 'ar 4/16/2011 17:43'!
installQPZ: data in: modulePath
	"Install the app under the given module path (usually just 'OpenQwaqPythonApps' or 'OpenQwaqPythonUserApps')"
	| pyDir initFile unzipper |
	pyDir := FileDirectory on: (FileDirectory cacheDirectory pathFromURI: modulePath).
	pyDir assureExistence.
	unzipper := ZipArchive new.
	unzipper readFrom: (ReadStream on:data).
	unzipper extractAllTo: pyDir informing: nil overwrite: true.
	pyDir := FileDirectory cacheDirectory.
	 "we need an empty __init__.py file in all directories"
	(modulePath findTokens: '/') do:[:dirName|
		pyDir := pyDir directoryNamed: dirName.
		(pyDir fileExists:'__init__.py') ifFalse:[
			initFile := pyDir forceNewFileNamed:'__init__.py'.
			initFile close.
		].
	].! !

!QInstallPython class methodsFor: 'as yet unclassified' stamp: 'das 9/8/2008 19:11'!
open: sharedDocument forUser: anAvatar

	"This is the in-space constructor used to initiate a private-app open request."
	(self new) 
		requester: anAvatar;
		document: sharedDocument;
		openDocument.! !

!QInstallPython class methodsFor: 'as yet unclassified' stamp: 'das 9/8/2008 19:19'!
startJob: job forClient: aUI

	"Get a local copy of the file if we don't already have one,
	 and send it an OS-level open request."
	| document |
	job future detach.
	document := job get:#document.

	"A little fine print..."
	(self doNotice: document) ifFalse: [ ^ self ].
	^ self doOpenPrivateCopy: document forClient: aUI.
! !

!QInstallPython methodsFor: 'initialize' stamp: 'das 9/8/2008 19:14'!
initialize
	super initialize.
	self isUserSpecific: true.! !

!QInstallPython methodsFor: 'initialize' stamp: 'das 9/8/2008 19:14'!
openDocument
	"fire off the job"
	jobSpace addJob:self.! !

QSaveDocumentJob subclass: #QOpenDocExternalJob
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Objects-Document'!

!QOpenDocExternalJob commentStamp: '<historical>' prior: 0!
Opens a document privately on the user's desktop
in the default native application for its file type,
using the platform's shellOpen command.

Applied to QSharedDocuments, QTextDocuments,
and also to URL strings (see QFileCard #openLocally:).!

!QOpenDocExternalJob class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QOpenDocExternalJob class methodsFor: 'job handling' stamp: 'ar 4/16/2011 17:13'!
doNotice: document

	"Remind the user of a little fine print..."

	| msg result |
	msg := '', (document get:#shortName), '\will be copied to your desktop and opened.\\Changes you make to your copy\will not appear in the Forum\until you import it back.\' withCRs.

	result := CDialog confirm: msg label: 'Opening Private Copy'.
	result := result ifNil: [ false ].

	^ result
! !

!QOpenDocExternalJob class methodsFor: 'job handling' stamp: 'bgf 9/11/2009 23:01'!
doOpenPrivateCopy: document forClient: aUI

	"Get a private (semi-hidden, in our private doc-tree) local copy of the file,
	 and send it a shell-open request to run it in the user's normal app."

	| fileName |
	document wait.
	[[
		fileName := QOpenDocExternalJob localPrivateCopy: document forClient: aUI
	] on: CProgressCancel do: [
		QLogger logInfo: 'Canceled open private copy.'.
		^ self 
	]
	] on: Error do: [ :exc|
		self inform: ('Cannot make local copy of document file.\\', exc description) withCRs.
		QLogger logError: exc from: self.
		^ self
	].
	
	fileName ifNotNil: [
		QActivityCapture action: #OpenPrivateCopy details: (QActivityCapture fileDetails: fileName).
		self markFileReadOnly: fileName.
		aUI shellOpen: fileName documentDescription: 'document' applicationDescription: 'application'.
	].! !

!QOpenDocExternalJob class methodsFor: 'utilities' stamp: 'bgf 9/11/2009 23:01'!
localPrivateCopy: document forClient: aClient
	"Get a private (semi-hidden, in our private doc-tree) local copy of the file;
	 return its full URI local-fileName if created.
	 Returns nil on failure  if an error has already been shown to the user;
	 otherwise always returns the filename or raises error."

	| fileName shortName |
	
	(document get: #isKeptLocally) ifTrue: [
	 	fileName := FileDirectory pathFromURI: (document get:#asLocalFileUri).
		(FileDirectory default fileExists: fileName) ifFalse: [
			fileName := nil.
		]].
	fileName ifNil: [
		fileName := aClient documentManager privateFileNameFor: document.
		(FileStream isAFileNamed: fileName) ifTrue: [
			QLogger logInfo: 'Making read-write: ', fileName.		
						"For diagnosis of occasional file-delete problem."
			self markFileReadWrite: fileName.	"So we can overrwrite on windows."
			].
		(FileDirectory default directoryExists: fileName) ifTrue: [
			Smalltalk platformName ~= 'Win32' ifTrue: [
				self markFileReadWrite: fileName.	"So we can remove slides on Mac."
			]].
		(shortName := ((document get:#class) 
						saveDocument: document asFile: fileName forClient: aClient )
					 ) ifNil: [
						^ nil 	"Callee did the inform."
					].
		fileName := (fileName copyUpToLast: (FileDirectory pathNameDelimiter)), 
						FileDirectory slash,
						shortName.
	].
	
	^ fileName! !

!QOpenDocExternalJob class methodsFor: 'job handling' stamp: 'bgf 5/7/2007 21:40'!
makeUniqueNewName: fileName

	"If the given fileName exists as a file, then generate a unique name
	 by appending '-[ i ]'   to the base name, picking 'i' to make the name a new filename."

	| stem suffix result i |
	stem  := fileName copyUpToLast:$..
	suffix := (fileName includes: $.) ifTrue: [fileName copyAfterLast:$.] ifFalse: [ nil ].

	result := fileName.
	i := 0.

	[FileDirectory default fileExists: result] whileTrue: [
		result := stem, '-', (i := i+1) printString, (suffix ifNil: [''] ifNotNil:['.', suffix]).	
	].

	^ result.! !

!QOpenDocExternalJob class methodsFor: 'private' stamp: 'bgf 3/9/2009 23:03'!
markFileReadOnly: fileName
	"Set platform-specific readonly state on the given file.
	 (We do this to remind the user that their edits are their own problem.)"
	| platformClass |
	[
		platformClass := CPlatform current class.
		platformClass == CMacOSXPlatform ifTrue: [
			| result | 		
			result := (MacOSShell new 
						system: ('/bin/chmod a-w ', (MacOSShell escapeFileName: fileName))).
			(result = 0 ) ifFalse: [
				QLogger logWarning: ' ', (MacOSShell escapeFileName: fileName), ' chmod: ', result..
			]].
		platformClass == CWin32Platform ifTrue: [
			Win32File setReadOnly: fileName.
			].

	] on: Error do: [:ex | 	
		QLogger logWarning: 'Chmod ', fileName, ' - ', (ex description).
	].
! !

!QOpenDocExternalJob class methodsFor: 'private' stamp: 'bgf 3/9/2009 22:50'!
markFileReadWrite: fileName
	"Clear any readonly state we might have set.
	 We need this on Windows mostly, where forceNewFileNamed fails if the file exists readonly"
	| platformClass |
	[
		platformClass := CPlatform current class.
		platformClass == CMacOSXPlatform ifTrue: [
			| result | 		
			result := (MacOSShell new 
						system: ('/bin/chmod u+w ', (MacOSShell escapeFileName: fileName))).
			(result = 0 ) ifFalse: [
				QLogger logWarning: ' ', (MacOSShell escapeFileName: fileName), ' chmod: ', result..
			]].
		platformClass == CWin32Platform ifTrue: [
			Win32File setReadWrite: fileName.
			].

	] on: Error do: [:ex | 	
		QLogger logWarning: 'Chmod ', fileName, ' - ', (ex description).
	].
! !

!QOpenDocExternalJob class methodsFor: 'instance creation' stamp: 'bgf 4/17/2007 23:06'!
open: sharedDocument forUser: anAvatar

	"This is the in-space constructor used to initiate a private-app open request."
	(self new) 
		requester: anAvatar;
		document: sharedDocument;
		openDocument.! !

!QOpenDocExternalJob class methodsFor: 'private' stamp: 'ar 4/16/2011 17:14'!
privateDocsDir
	"Directory where open-private documents are kept. "
	| dir |
	[ 
		dir := FileDirectory on: (QFileDialog getFileLocation: 'documents').
		dir := dir directoryNamed: 'OpenQwaq'.
		dir := dir directoryNamed: 'Private'.
		dir assureExistence.
		^dir
	] on: Error do: [ : ex |
		QLogger logWarning: 'Cannot make private document area: ', ex description.
		dir := FileDirectory on: (QFileDialog getFileLocation: 'documents').
		dir assureExistence.
		^ dir
	].! !

!QOpenDocExternalJob class methodsFor: 'job handling' stamp: 'bgf 3/1/2009 10:16'!
startJob: job forClient: aUI

	"Get a local copy of the file if we don't already have one,
	 and send it an OS-level open request."
	| document |
	job future detach.
	document := job get:#document.
	"If it's a string, assume it's a URL and just open it immediately."
	document isString ifTrue: [
		QActivityCapture action: #OpenPrivateCopy details: 'url'.
		^ aUI shellOpen: document documentDescription: 'URL' 
				applicationDescription: 'Web browser'.
	].

	^ self doOpenPrivateCopy: document forClient: aUI.
! !

!QOpenDocExternalJob class methodsFor: 'defaults' stamp: 'bgf 10/4/2007 09:31'!
timeout
	"A medium timeout; this is expected to be a bit sluggish."
	^ 5000! !

!QOpenDocExternalJob methodsFor: 'initialize-release' stamp: 'bgf 6/24/2007 15:35'!
initialize
	super initialize.
	self isUserSpecific: true.! !

!QOpenDocExternalJob methodsFor: 'initialize-release' stamp: 'bgf 4/17/2007 19:00'!
openDocument
	"fire off the job"
	jobSpace addJob:self.! !

!QSaveDocumentJob class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QSaveDocumentJob class methodsFor: 'job handling' stamp: 'bgf 2/12/2010 20:56'!
askFileName: document 
	"Peel off the old <> wraps - they make windows fart."
	| shortName |
	shortName := (document get:#shortName) ifNil: ['document'].
	shortName := (shortName copyWithout:$<) copyWithout:$>.

	^ self askFileName: document shortName: shortName.
	! !

!QSaveDocumentJob class methodsFor: 'job handling' stamp: 'bgf 2/12/2010 20:55'!
askFileName: document shortName: shortName

	|  dialog newName |

	[ "Try the new native dialog..."
		dialog := QFileDialog new.
		dialog filters: (document get:#fileChooserPattern).
		" dialog fileName: shortName. - makes the native dialog fail !!"
		dialog filePath: (QFileDialog getFileLocation: 'desktop'), 
				FileDirectory slash, 
				shortName.
		dialog label: (document get:#fileSaveLabel).
		newName := dialog saveFileAs.

		"Todo - remember the directory path as the user's most-recent save path."

	] on: Error do: [
		dialog := CFileSaveDialog new.
		dialog fileSpec: (document get:#fileChooserPatternTweak).
		dialog fileName: shortName.
		newName  := dialog openLabel: (document get: #fileSaveLabel).
	].

	^ newName! !

!QSaveDocumentJob class methodsFor: 'defaults' stamp: 'bgf 3/28/2007 12:12'!
maxTries

	"The number of job retries to allow before abandoning a job.
	 This is interactive, so only one shot."
	^ 0! !

!QSaveDocumentJob class methodsFor: 'instance creation' stamp: 'bgf 2/19/2007 19:11'!
save:aDocument forUser:anAvatar 
	"Initiate a harness-side text editor to modify the given text;
	 the result, when complete is returned to the given recipient as the argument to the given selector."

	(self new) 
		requester: anAvatar;
		document: aDocument;
		saveDocument.! !

!QSaveDocumentJob class methodsFor: 'job handling' stamp: 'bgf 9/9/2009 01:11'!
startJob: saveJob forClient:aUI
	"Harness-side document save handler; see QSaveDocumentJob."
	| newName document |

	saveJob future detach.
	saveJob wait.

	document := saveJob get:#document.
	document wait.
	newName := self askFileName: document.
	newName ifNil:[
		aUI statusStatus: 'Cancelled save.'.
		^ self
	].
	
	QActivityCapture action: #SaveLocalCopy details: (QActivityCapture fileDetails: newName).

	[[
		newName := (document get:#class)
						saveDocument:document asFile: newName forClient:aUI.
		newName ifNotNil: [
			saveJob future documentSaved: newName.
			aUI statusInfo: 'Saved ', newName.
		].
		] on: CProgressCancel do: [
			QLogger logInfo: 'Canceled document local save.'.
		]
	] on: Error do: [ :exc|
		self inform: ('Could not save document.\\', exc description) withCRs.
	].
! !

!QSaveDocumentJob class methodsFor: 'defaults' stamp: 'bgf 3/28/2007 12:11'!
timeout
	"A short timeout, since the editor should appear in conjunction with the gesture
	 that requests it.  (But long enough to deal with slow native dialog startup)"
	^ 5000! !

!QSaveDocumentJob methodsFor: 'accessing' stamp: 'bgf 2/19/2007 18:43'!
document
	^ document! !

!QSaveDocumentJob methodsFor: 'accessing' stamp: 'bgf 2/19/2007 18:43'!
document: aDocument
	document := aDocument! !

!QSaveDocumentJob methodsFor: 'job queue' stamp: 'bgf 3/3/2009 05:20'!
documentSaved: fileName
	"Callback from the participant UI to indicate that the save is complete;
	 the filename is the new short filename used."

	document saveComplete: fileName.
! !

!QSaveDocumentJob methodsFor: 'initialize-release' stamp: 'bgf 6/24/2007 15:46'!
initialize
	super initialize.
	self isUserSpecific: true.
! !

!QSaveDocumentJob methodsFor: 'initialize-release' stamp: 'bgf 2/19/2007 18:42'!
saveDocument
	jobSpace addJob:self.! !

QSaveDocumentJob subclass: #QStoreObjectJob
	instanceVariableNames: 'recipient original resourceList toInventory sceneZip'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Objects-Document'!

!QStoreObjectJob commentStamp: '<historical>' prior: 0!
Harness-side implementation of Copy-To-Inventory for TFrames.
Captures the persistence representation of a TFrame,
together with its required resources, in a .c3z file that is uploaded to the user's inventory.

The partner of QCreateStoredObjectJob, which does the reverse,
fetching a stored object c3z from the server and inserting it into the space.
!

!QStoreObjectJob class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QStoreObjectJob class methodsFor: 'private' stamp: 'bgf 4/14/2010 17:51'!
putSceneAsDocument: zip toInventory: bInventory inClient: owner
	"Upload the fat c3z zipped scene as to resources,
	 returning the shared document that describes where we put it."

	| data path sha sharedName docmgr |

	"We have a fat .c3z, which gets uploaded to shared storage."
	data := ByteArray streamContents:[:s| zip writeTo: s].
	sha := TContentManager generateSHA: data.
	docmgr := owner documentManager ifNil: [ ^ nil ].

	bInventory ifTrue: [
		path :='/users/', owner serviceMgr user, '/private/resources/', sha, '.c3z'.
		sharedName := path.
	] ifFalse: [
		sharedName := 'resources/', sha, '.c3z'.
		path := '/services', (owner currentForumName), '/', sharedName.
	]. 
	(docmgr serverFileExists: path) ifTrue: [  
		"Since the name is based on the sha, just reuse it."
		^ (QSharedDocument new)
			sha: (TContentManager generateSHA: data);
			sharedName: path;
			ctime: (docmgr newTimeStamp: 'Import');
			suffix: 'c3z';
			yourself.

	] ifFalse:[
		QLogger logInfo: 'Storing object ', path.
		^ docmgr uploadData: data
			sha: (TContentManager generateSHA: data)
			fullPath: path
			sharedName: sharedName
			friendlyName: 'Stored Object'.
	].! !

!QStoreObjectJob class methodsFor: 'job handling' stamp: 'bgf 1/8/2008 21:14'!
saveConfirm: mName missingResources: set owner: aUI

	"Given a stored-object name,
	 and a list of missing resource names,
	 confirm that the save-to-inventory should proceed;
	 really a way of letting the user know that something went wrong."

	| msg array count cursor |

	(set isNil or: [set isEmpty]) ifTrue: [ aUI waitTick. ^ true ].	"Nothing missing... implicit OK"
	cursor := aUI mouseCursor.
	array :=  set asArray. 		"So we can get the first 10"
	count := array size.

	"Fussy because we don't apply withCRs to the filenames themselves.."
	msg := 'Missing resources in ', mName,
			('\\The following resources were not found:' withCRs).
	(array first: (count min:10)) do: [ :res| 
		msg := msg, ('\   ' withCRs), (res first), ': ', (res second)
	].
	array size > 10 ifTrue: [ msg := msg,  ('\   ...' withCRs)].
	msg := msg, ('\\Save template anyway?' withCRs).

	[
		aUI  mouseCursor: nil.
		^ self confirm: msg
	] ensure: [
		aUI mouseCursor: cursor.
		aUI waitTick.		"Clear the confirm dialog after ok or cancel."
	].! !

!QStoreObjectJob class methodsFor: 'job handling' stamp: 'bgf 8/1/2009 18:41'!
saveFrame: tframe to: path for: ui
	"Make a fat c3z of the given TFrame, and save it as a local file."

	| writer xml file archive data absent |
	writer := ui activeSpace island future new: TSceneWriter.
	writer future initResourceList.
	xml := writer future storeScene: tframe.
	xml wait.

	ui readyForEvents ifFalse: [ ^ self ].
	xml value isEmptyOrNil ifTrue: [ 
		self inform: 'Cannot copy that object.'.
		writer discard.
		^ self.
	].
	archive := ZipArchive new.
	archive addDeflateString: xml value as: 'tframe.xml'.
	(writer get: #resourceListSize) > 0 ifTrue:[
		absent := QStoreObjectJob saveResources: (writer get:#resourceList) asZip: archive forClient: ui.
		"Really ought not happen unless what's in the space is already missing content,
 		 but helps distinguish where problems arise.  
		 If there missing resources, give the user a veto."
		(QStoreObjectJob saveConfirm: (tframe get: #labelForBrowser) 
						 missingResources: absent owner: ui) ifFalse: [
			writer discard.
			^ self
		].
	].
	writer discard.
	xml discard.
	file := FileDirectory default forceNewFileNamed: path.
	data := ByteArray streamContents:[:s| archive writeTo: s].
	[file nextPutAll: data ] ensure: [file close].

	ui statusInfo: 'Saved ', (tframe get:#objectName).
! !

!QStoreObjectJob class methodsFor: 'job handling' stamp: 'bgf 2/5/2010 20:38'!
saveResources: resList asZip: archive forClient: aUI

	"Save the data that backs the given list of resources (far-refs to textures and models)
	 as separate named entities in the given ZIP archive.
	 The data comes from the active session cache... 
	 we assume it's already in play in the harness.
	 Note there is a workalike of this routine in QInspector."

	| absentList doneSet fname data |
	absentList := OrderedCollection new.
	doneSet := Set new.

	resList value select: [:resRef|
		resRef wait.
		(resRef get:#isBundled) not   "Don't upload prepackaged resources." 
	] thenDo:[:resRef|
		fname := resRef get: #fileName.
		
		"The same textures may be referenced multiple times, but we only write them once"
		(fname isNil or: [doneSet includes: fname]) ifFalse:[
			doneSet add: fname.
			data := (resRef get:#class) getContent: resRef forClient: aUI.
			data ifNil:["Texture not present; add it to the list"
				absentList add: { (resRef get: #resourceType) .  fname}.
			] ifNotNil:[
				"Don't compress the texture again"
				archive addString: data asByteArray as: fname.
			]
		].
	].

	^ absentList.
! !

!QStoreObjectJob class methodsFor: 'job handling' stamp: 'bgf 8/2/2009 03:40'!
startJob: job forClient: aUI
	"Main inner work routine:
	 Fetch the persistent-xml data from the document,
	 and send it to the scene reader.
	 Then import the resuling object into the target space."

	| original resourceList toInventory archive absent fileName data doc |
	[
		job future detach.		"Done with the scheduling system now."

		original := job get:#original.
		data := job get:#sceneZip.
		resourceList := job get:#resourceList.
		toInventory := job get:#toInventory.
		original wait.
		"The other parameters are immediate rather than far-ref based (bytearray, array, bool)"

		aUI statusInfo: 'Storing objects...'.
		aUI mouseCursor: Cursor wait.
	
		fileName := ((original get:#objectName) ifNil: [ (original get:#class) name ]), '.c3z'.

		"We have  Zip archive prepared in-space... 
	       we start with that, and add in all the resource members."
		archive := ZipArchive new.
		archive readFrom: data readStream.
		resourceList ifNotNil: [
			absent := self saveResources: resourceList asZip: archive forClient: aUI.
			"Really ought not happen unless what's in the space has missing pieces,
 			 but helps distinguish where problems arise."
			(self saveConfirm: fileName missingResources: absent owner: aUI) ifFalse: [
				job future objectStored: nil.
				^ self
			].
		].
		doc := self putSceneAsDocument: archive toInventory: toInventory inClient: aUI.
		"Finally, send the resulting doc-handle back into the space...
		 it will become the content of the file card that requested this operation."
		doc notNil ifTrue: [
			QActivityCapture action: #SaveObjectToInventory 
							 details: (QActivityCapture fileDetails: doc shortName).
			doc := TIslandCopier new import:(TIslandCopier new export: doc) 
								  to: job island.
		].
	] on: Error do: [ :ex|
		aUI inform: ('Cannot store that object:\\', ex description) withCRs.
		QLogger logError: ex.
		doc := nil.
	].
	[ "Second error block is just to protect against unexpected loss of space connections, etc.
	   Doc will be the uploaded .c3z handle (or nil, if there were any errors in the above.)".
		job future objectStored: doc.
	] on: Error do: [].

	aUI mouseCursor: nil.! !

!QStoreObjectJob class methodsFor: 'instance creation' stamp: 'bgf 1/8/2008 21:11'!
storeObject: original recipient: aCallbackRecipient forUser: anAvatar

	"This is the in-space constructor used to initiate an object-creation request.
	 doc is the scene document, recipient will receive #objectStored: aDocument
	 after the scene document is created as a handle on the uploaded data. 
	 anAvatar is the avatar replica for whom the action is carried out."

	(self new) 
		requester: anAvatar;
		recipient: aCallbackRecipient;
		original: original;
		toInventory: true;
		storeObject.! !

!QStoreObjectJob class methodsFor: 'defaults' stamp: 'bgf 1/8/2008 20:11'!
timeout
	"A longish timeout.  There are no retries for this job, so the timeout 
	 is long enough to ensure that the first request is caught (unless some heavy load is going on)."
	^ 4000! !

!QStoreObjectJob methodsFor: 'initialize-release' stamp: 'bgf 1/8/2008 20:00'!
initialize
	super initialize.
	self isUserSpecific: true.! !

!QStoreObjectJob methodsFor: 'job queue' stamp: 'bgf 1/22/2008 02:28'!
objectStored: aSharedDocument

	"Completion of the job... 
	 hand the recipient back a shared-document handle on the uploaded c3z."
	recipient objectStored: aSharedDocument
! !

!QStoreObjectJob methodsFor: 'accessing' stamp: 'bgf 1/8/2008 20:07'!
original
	"The in-island object that is to be captured for inventory."
	^ original! !

!QStoreObjectJob methodsFor: 'accessing' stamp: 'bgf 1/8/2008 20:07'!
original: aTFrame
	"The in-island object that is to be captured for inventory."
	original := aTFrame! !

!QStoreObjectJob methodsFor: 'accessing' stamp: 'bgf 1/8/2008 20:00'!
recipient
	^ recipient! !

!QStoreObjectJob methodsFor: 'accessing' stamp: 'bgf 1/8/2008 20:08'!
recipient: aCallbackRecipient
	"The in-island object that receives the uploaded shared-document handle.
	 The job calls
		recipient objectStored: aSharedDocument
	 to complete its work and deliver the object."
	recipient := aCallbackRecipient! !

!QStoreObjectJob methodsFor: 'accessing' stamp: 'bgf 1/8/2008 20:25'!
resourceList
	"The resource array accrued when storing the original as  a scene."
	^ resourceList ! !

!QStoreObjectJob methodsFor: 'accessing' stamp: 'bgf 1/8/2008 20:24'!
resourceList: aCollection
	"The resource list accrued when storing the original as  a scene."
	resourceList := aCollection asArray.! !

!QStoreObjectJob methodsFor: 'accessing' stamp: 'bgf 1/8/2008 21:56'!
sceneZip
	"The XML scene containing the basic persistence of <original>"
	^ sceneZip! !

!QStoreObjectJob methodsFor: 'accessing' stamp: 'bgf 1/8/2008 21:57'!
sceneZip: aSceneZip
	"The XML scene containing the basic persistence of <original>"
	sceneZip := aSceneZip! !

!QStoreObjectJob methodsFor: 'initialize-release' stamp: 'bgf 1/8/2008 22:22'!
storeObject
	"Launch for the on-island case: 
	 fire off the job, after doing some on-island prework:
	 we capture the scene XML and the computed resources list here,
	 so that the harness-side job need only get:# those results."

	| writer |
	writer := TSceneWriter new.
	writer initResourceList.
	self sceneZip: (writer storeSceneAsZip: original).
	self resourceList: (writer resourceList).

	jobSpace addJob:self.! !

!QStoreObjectJob methodsFor: 'accessing' stamp: 'bgf 1/8/2008 20:40'!
toInventory
	"If true, store to this user's inventory; if false, to active forum resource storage."
	^ toInventory! !

!QStoreObjectJob methodsFor: 'accessing' stamp: 'bgf 1/8/2008 20:40'!
toInventory: aBool
	"If true, store to this user's inventory; if false, to active forum resource storage."
	toInventory := aBool! !

QParticipantJob subclass: #QTextEditJob
	instanceVariableNames: 'title text recipient selector prevText prevTime extraData'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Objects-Jobs'!

!QTextEditJob commentStamp: '<historical>' prior: 0!
Allow editing of a piece of plain text for a particular avatar.

Usage is something like:
		QTextEditJob 
				edit: contents 			"The original text"
				title: 'Edit Label' 
				forUser: anEvent avatar "Must be an avatar"
				recipient: aFrame 
				selector: #contentsFromEditor:.  "Method on recipient"

This triggers a pop-up text edit panel in the participant.

Subsequently, if the user edits and applies the text, 
the recipient (aFrame) will receive it as a future message
using the given selector.   The text may be changed and updated
several times.

Todo: as yet, there is no detection of redundant editors;
this probably needs to be done at the participant level.

Note that the job service is only used to open the editor,
at which point the job is 'finished' and out of the batch service's domain.
The job is done, from the scheduling point of view, when we get #editorLaunched.
The Job itself survives, however, as the context by which the editor
sends results back to the space.    This is done because a text editor can
be open for a long time, and we can have several editors,
and none of them need impede the scheduling of unrelated tasks.


!

!QTextEditJob class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QTextEditJob class methodsFor: 'instance creation' stamp: 'bgf 3/29/2010 10:48'!
edit: text title: title forUser:anAvatar recipient: aFrame selector: aTernarySelector
	"Initiate a harness-side text editor to modify the given text;
	 the result, when complete is returned to the given recipient as the argument to the given selector.
	 If selector is nil, the text pane is not editable."

	(self new) 
		requester: anAvatar;
		text: text;
		title: title;
		recipient: aFrame;
		selector: aTernarySelector;
		extraData: nil;
		editText.! !

!QTextEditJob class methodsFor: 'instance creation' stamp: 'bgf 3/29/2010 10:32'!
edit: text title: title forUser:anAvatar recipient: aFrame selector: aTernarySelector extraData: extraData
	"Initiate a harness-side text editor to modify the given text;
	 the result, when complete is returned to the given recipient as the argument to the given selector.
	 If selector is nil, the text pane is not editable."

	(self new) 
		requester: anAvatar;
		text: text;
		title: title;
		recipient: aFrame;
		selector: aTernarySelector;
		extraData: extraData;
		editText.! !

!QTextEditJob class methodsFor: 'defaults' stamp: 'bgf 3/28/2007 12:13'!
maxTries

	"The number of job retries to allow before abandoning a job.
	 Very small, since text editors shouldn't come up long after asking."
	^ 0! !

!QTextEditJob class methodsFor: 'undo-redo' stamp: 'bgf 3/29/2010 10:44'!
reset: recipient selector: sel text: txt utc: utcSecs avatar: avatar extraData: extra

	| t |
	t := utcSecs ifNil: [ Time utcSeconds ].

	sel numArgs caseOf: {
		[4] -> [ recipient future perform: sel with: txt with: t with: avatar with: extra ].
		[3] -> [ recipient future perform: sel with: txt with: t with: avatar ].
		[2] -> [ recipient future perform: sel with: txt with: t ].
		[1] -> [ recipient future perform: sel with: txt  ]
	} otherwise: [
		QLogger logError: 'Text Editor callback selector must take 1-3 arguments.'
	].
	
	recipient future signal: #textJobReset with: txt! !

!QTextEditJob class methodsFor: 'job handling' stamp: 'bgf 8/31/2008 21:17'!
startJob: textJob forClient: aUI

	"Open a simple text editor for the user,
	 in response to a request from within the active space."

	| editor recipient klass |

	textJob future detach.

	"If we don't have the view initialized yet, just do nothing.."
	aUI mainView ifNil: [ ^ self ].

	textJob wait.
	recipient := (textJob get:#recipient).

	"Some recipient classes may use their own editor pane implementation
	 (see further below - such panes have to implement all the basic setup accessors we use.)"
	klass := (recipient get:#editToolClass) ifNil: [ QTextEditPane ].

	"If we've already got an editor on this guy, don't do another."
	(aUI elements detect: [ : e | 
			(e isKindOf: klass) 
			  and: [(e job get:#recipient) = recipient]] ifNone: [nil])
		ifNotNilDo: [ :e | 
				e expand.
				^ self ].

	editor := ((klass new) 
			owner: aUI;
			expand;
			job: textJob;
			yourself).

	editor openIn: aUI at: (aUI hand position + (-100@20)).	"So we can see what we clicked to get here."
	editor claimFocus.
! !

!QTextEditJob class methodsFor: 'defaults' stamp: 'bgf 10/4/2007 09:31'!
timeout
	"A normal short timeout, since the editor should appear in conjunction with the gesture
	 that requests it."
	^ 3000! !

!QTextEditJob class methodsFor: 'undo-redo' stamp: 'bgf 3/29/2010 10:35'!
undoAction: recipient selector: sel oldState: oldState newState: newState

	^ QBlockAction
		undoBlock: [ : avatar | 
			QTextEditJob reset: recipient selector: sel text: (oldState first) utc: (oldState second) avatar: avatar extraData: oldState third ]
		redoBlock: [ : avatar | 
			QTextEditJob reset: recipient selector: sel text: (newState first) utc: (newState second) avatar: avatar extraData: oldState third ]
		description: 'Edit Text'
		requiredFrames: { recipient }! !

!QTextEditJob methodsFor: 'initialize-release' stamp: 'bgf 1/16/2007 09:34'!
editText
	jobSpace addJob:self.! !

!QTextEditJob methodsFor: 'accessing' stamp: 'bgf 3/29/2010 10:33'!
extraData
	^ extraData! !

!QTextEditJob methodsFor: 'accessing' stamp: 'bgf 3/29/2010 10:49'!
extraData: data

	"Caller-supplied support data, will be passed back unchanged to edit-selector as fourth arg.
	 Should be given simple scalars, strings, or array-of-simple data. (E.g. by island pass-by-copy.)."

	extraData := data! !

!QTextEditJob methodsFor: 'initialize-release' stamp: 'bgf 6/24/2007 15:46'!
initialize
	super initialize.
	self isUserSpecific: true.
! !

!QTextEditJob methodsFor: 'accessing' stamp: 'bgf 1/16/2007 09:32'!
recipient
	^ recipient! !

!QTextEditJob methodsFor: 'accessing' stamp: 'bgf 1/16/2007 09:32'!
recipient: aFrame
	recipient := aFrame! !

!QTextEditJob methodsFor: 'accessing' stamp: 'bgf 1/16/2007 09:32'!
selector
	^ selector! !

!QTextEditJob methodsFor: 'accessing' stamp: 'bgf 3/29/2010 10:33'!
selector: aSelector
	"The callback selector to invoke on recipient 
	 to send back the result of the editing activity.
	 The first argument will always be the new text.
	 If binary, the second argument is the client UTC-seconds of the change.
	 If there is one, the third argument the avatar.
	 If there is one, the fourth argument is the extraData."
	selector := aSelector! !

!QTextEditJob methodsFor: 'accessing' stamp: 'bgf 1/16/2007 09:33'!
text
	^ text! !

!QTextEditJob methodsFor: 'accessing' stamp: 'bgf 3/9/2010 07:41'!
text: aString
	text := aString asString.
	prevText := text.! !

!QTextEditJob methodsFor: 'job queue' stamp: 'bgf 3/29/2010 10:36'!
textChanged: aString atClientTime: utcSeconds byUser: anAvatar

	"Callback from the text editor when editing is complete;
	 this happens -after- the job has been cleared from the job system!!"

	selector ifNotNil: [
		prevText = aString ifFalse: [
			anAvatar 
				recordAction: #undoAction:selector:oldState:newState:
				class: QTextEditJob
				arguments: { recipient .  
							selector .  
							{ prevText . prevTime . extraData} . 
							{ aString . utcSeconds . extraData } }.
			]].

	selector ifNotNil: [
		selector numArgs caseOf: {
			[4] -> [ recipient perform: selector with: aString with: utcSeconds with: anAvatar with: extraData ].
			[3] -> [ recipient perform: selector with: aString with: utcSeconds with: anAvatar ].
			[2] -> [ recipient perform: selector with: aString with: utcSeconds ].
			[1] -> [ recipient perform: selector with: aString ]
		} otherwise: [
			QLogger logError: 'Text Editor callback selector must take 1-3 arguments.'
		].
	] ifNil: [
		"No problem - this is just the completion of a viewing job."
	].
	prevText := aString clone.
	prevTime := utcSeconds.! !

!QTextEditJob methodsFor: 'accessing' stamp: 'bgf 1/16/2007 10:54'!
title
	^ title! !

!QTextEditJob methodsFor: 'accessing' stamp: 'bgf 1/16/2007 10:54'!
title: aString
	title := aString! !

TObject subclass: #QPasswordHolder
	instanceVariableNames: 'value'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Objects-General'!

!QPasswordHolder class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QPasswordHolder class methodsFor: 'instance creation' stamp: 'ar 5/6/2009 15:00'!
with: aString
	^self new value: aString! !

!QPasswordHolder methodsFor: 'accessing' stamp: 'ar 5/6/2009 14:59'!
value
	"The actual password"
	^value base64Decoded! !

!QPasswordHolder methodsFor: 'accessing' stamp: 'ar 6/17/2009 14:44'!
value: aString
	"The actual password"
	value := aString asString base64Encoded! !

TObject subclass: #QPreferenceSet
	instanceVariableNames: 'preferenceDictionary signalledObject'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Objects-User'!

!QPreferenceSet commentStamp: 'hrs 8/27/2008 12:46' prior: 0!
A cache of preferences from a single source (e.g., server, per-machine local, per-user local).

Not complete. E.g., not integrated with QUserPreference. But at least it's a place to start hanging things.!

QPreferenceSet subclass: #QMachinePreferences
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Objects-User'!

!QMachinePreferences commentStamp: 'hrs 8/27/2008 14:48' prior: 0!
Preferences that apply only to the local host machine.
They are stored locally.!

!QMachinePreferences class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QMachinePreferences methodsFor: 'change protocol' stamp: 'hrs 8/27/2008 15:56'!
dirty
	"Write out the full set."
	^self atLocalName: self persistenceName, '.xml' putData: self persistString.! !

!QMachinePreferences methodsFor: 'initialize-release' stamp: 'hrs 8/27/2008 16:09'!
load
	"Load from disk. Must be after initialization because clients may change things such as #signalledObject:."
	(self getLocalXmlString: self persistenceName) ifNotNilDo: [:string | 
		self constructFromString: string.
		].! !

!QPreferenceSet class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QPreferenceSet class methodsFor: 'objects from disk' stamp: 'hrs 8/27/2008 15:44'!
xmlDirectory
	"Directory where user preference xml files are kept."
	^FileDirectory prefsDirectory! !

!QPreferenceSet methodsFor: 'persist' stamp: 'hrs 8/27/2008 15:39'!
atLocalName: fileName putData: ascii

	"Write ascii data as a local file for this user.
	 Consumes errors and logs them;
	 returns boolean (true for success.)"

	| file |
	[
		self class xmlDirectory assureExistence.
		file := self class xmlDirectory forceNewFileNamed: fileName.
		file ascii.
		file nextPutAll: ascii.
		file close.	
		^ true	
	] on: Error do:[:ex| 
		QLogger logWarning: 'Cannot locally save ', file , ' preferences: ', ex description.
		^ false
	].
! !

!QPreferenceSet methodsFor: 'persist' stamp: 'hrs 1/27/2009 08:54'!
constructFrom: aStream
	"Deserialize from aStream. Default populates all preferences without interleaving signals, 
	and then sends #signalChanges once when done.  This way signal handlers can rely on all the other
	preferences already being defined."
	| node |
	node := (XMLDOMParser on: aStream) nextEntity.
	node name = self persistenceName ifFalse: [self error: aStream asString, ' is not a ', self persistenceName].
	node elements do: [:prefNode |
		preferenceDictionary at: prefNode name put: prefNode contentString.
		].
	self signalChanges.
	! !

!QPreferenceSet methodsFor: 'persist' stamp: 'hrs 8/27/2008 15:17'!
constructFromString: aString
	"Deserialize from previously pickled string.  Ex: QPreferenceSet new constructFromString: anotherSet persistString"
	self constructFrom: (ReadStream on: aString)
	! !

!QPreferenceSet methodsFor: 'change protocol' stamp: 'hrs 8/27/2008 14:46'!
dirty
	"Some preference has changed. Overridable by subclasses, typically to save all preferences. Default does nothing."
! !

!QPreferenceSet methodsFor: 'error handling' stamp: 'hrs 8/27/2008 14:41'!
doesNotUnderstand: aMessage
	"Access the value from the dictionary. In this version, getters always answer nil if the value is not present, not dNU."
	aMessage arguments
		ifEmpty: [^self getPreference: aMessage]
		ifNotEmpty: [
			aMessage arguments size = 1 ifTrue: [^self setPreference: aMessage]
			].
	^super doesNotUnderstand: aMessage.! !

!QPreferenceSet methodsFor: 'persist' stamp: 'hrs 8/27/2008 15:33'!
getLocalXmlString: unitName

	"Return the given preferences data unit (usually xml text),
	 from the local preferences files on this computer.  
	The unitName is used without a prefix, but .xml and .xml.zip are both tried."
	| file |

	file := [self class xmlDirectory
				readOnlyFileNamed: unitName, '.xml.zip'.
			] on: Error do: [
				[self class xmlDirectory 
					readOnlyFileNamed:unitName, '.xml']
				on: Error do:  [ 
					QLogger logWarning: 'No ', unitName, ' preferences found locally.'.
					^ nil 
				].
			].
	file size = 0 ifTrue: [	"Common transient failure case."
		QLogger logInfo: 'Empty ', unitName, ' preferences found locally.'.
		file close.
		^ nil
	].

	"Read it in."
	[	[
			(file localName endsWith: 'zip') ifTrue: [
				file binary.
				^ (self class zipToXml: (file contentsOfEntireFile) unitName: unitName)
			] ifFalse: [
				file ascii.
				^ file contentsOfEntireFile
			].
	  	] ensure: [ file close ].

	] on: Error do: [ :exc|
		QLogger logWarning: 'Cannot locally read ', unitName, ' preferences: ', exc description.
		^ nil
	].! !

!QPreferenceSet methodsFor: 'error handling' stamp: 'hrs 8/27/2008 14:29'!
getPreference: aMessage
	"Read the value from the dictionary.  Overridable by subclasses."
	^preferenceDictionary at: aMessage selector ifPresent: [:val | val].! !

!QPreferenceSet methodsFor: 'initialize-release' stamp: 'hrs 8/27/2008 14:43'!
initialize
	super initialize.
	preferenceDictionary := Dictionary new.
	signalledObject := self.! !

!QPreferenceSet methodsFor: 'persist' stamp: 'hrs 8/27/2008 15:58'!
persistOn: aStream
	"Serialize self in way from which our data can be recovered. Default is XML."
	| writer |
	writer := XMLWriterIndenting on: aStream.
	writer startTag: self persistenceName; endTag.
		preferenceDictionary keysAndValuesDo: [:key :val |
			writer startTag: key; endTag; pcData: val asString; endTag: key.
			].
	writer endTag: self persistenceName.! !

!QPreferenceSet methodsFor: 'persist' stamp: 'hrs 8/27/2008 15:00'!
persistString
	"Answer a string suitable for pickling us in a file so that we can be recovered."
	^String streamContents: [:s | self persistOn: s].! !

!QPreferenceSet methodsFor: 'accessing' stamp: 'hrs 8/27/2008 14:57'!
persistenceName
	"Name of the top-level XML tag for default method of #persistOn:."
	^self class asString.! !

!QPreferenceSet methodsFor: 'error handling' stamp: 'hrs 8/27/2008 14:45'!
setPreference: aMessage
	"Set the value from the dictionary.  Overridable by subclasses.
	This version
		changes the value;
		sends signalChange;
		sends #dirty....
	...but only if the value has actually changed."
	| arg getterName |
	arg := aMessage arguments first.
	getterName := aMessage selector allButLast.
	preferenceDictionary at: getterName ifPresent: [:old | old = arg ifTrue: [^self]].
	preferenceDictionary at: getterName put: arg.
	self signalledObject signal: (getterName, 'Changed') asSymbol with: arg.
	self dirty.
! !

!QPreferenceSet methodsFor: 'change protocol' stamp: 'hrs 8/27/2008 14:34'!
signalChange: getterName with: newValue
	"Signal that newValue has been assigned for getterName.  Overridable by subclasses."
	self signalledObject signal: (getterName, 'Changed') asSymbol with: newValue.

! !

!QPreferenceSet methodsFor: 'change protocol' stamp: 'hrs 8/27/2008 15:14'!
signalChanges
	"Signal all preference changes. Used when first loading."
	preferenceDictionary keysAndValuesDo: [:key :val |
		self signalChange: key with: val.
		].! !

!QPreferenceSet methodsFor: 'accessing' stamp: 'hrs 8/27/2008 16:11'!
signalledObject
	"The object that should receive signals from #signalChange:with:."
	^signalledObject! !

!QPreferenceSet methodsFor: 'accessing' stamp: 'hrs 8/27/2008 16:12'!
signalledObject: anObject
	"Change the #signalledObject. For example, clients of QUserPreferences may be set 
	up to listen for changes in that, regardless of whether the changes are stored
	locally, on the server, etc."
	signalledObject := anObject! !

TObject subclass: #QPresenceManager
	instanceVariableNames: 'locationTable forumTable robots'
	classVariableNames: 'MeetingLeaderFlagMask PrivateTextBlockFlagMask'
	poolDictionaries: ''
	category: 'Qwaq-Objects-User'!

!QPresenceManager commentStamp: '<historical>' prior: 0!
One QPresenceManager exists in the shared organization-services island,
and keeps track of the current live location of connected users,
and keeps track of what forums are active.

The queries, such as,
	userLocation: aUserLogin
	usersInService: aServiceURL
return simple arrays of pass-by-copy data, so there are no far-ref entaglements
except for the one presence-manager far-ref held by a client's QOrgServices object.

When a user updates his location (via #user:location:localTime:, 
I signal myself (using that selector as signal name) passing the same data;
clients can handle that signal to keep track of participants movements among services.

The robots dictionary maps user-logins to 'robotic mode' symbols;
this is used to offer and detect robotic clients.   
Presence of a userLogin as a key indicates that the user is available for control.
The value under the key is the current demon suite in use by that user;
The value may be nil, in which case the user is available but idle.
(The  symbol set is open;  they are not interpreted here.)

No attempt is made to restrict or secure robotic control;
this is simply an open registry and signalling mechanism.

The active-forum table is maintained obliquely through user-heartbeats
(since various users may have various forums opens in overlapping sets,
and no one client can authoritatively tell me that a forum is not active...
each user heartbeat can only update what forums it has active at that time.
So instead, we use heartbeat-aging to decide when to make a forum inactive...
I.e. if no-one claims to be using a forum for long enough, that forum is declared inactive.

The user-table also has a heartbeat regime, but individual users
can also declare themselves 'out' with message, so user-entry heartbeats
protect against unexpected session loss.   
!

!QPresenceManager class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QPresenceManager class methodsFor: 'defaults' stamp: 'bgf 11/6/2007 15:48'!
heartbeatInterval

	"Interval at which the client sends user:location:localTime: to affirm presence.
	 See #recordIsCurrent:, which uses a multiple of this value when checking for timeout."

	^ 5  "Seconds - rate at which client future-sends into presence service."! !

!QPresenceManager class methodsFor: 'class initialization' stamp: 'bgf 8/8/2009 01:37'!
initialize
	" QPresenceManager initialize "

	"Bit flags for presence and session record flags"
	MeetingLeaderFlagMask := 1.
	PrivateTextBlockFlagMask := 2.! !

!QPresenceManager class methodsFor: 'class variables' stamp: 'bgf 8/8/2009 01:36'!
meetingLeaderFlagMask
	^ MeetingLeaderFlagMask! !

!QPresenceManager class methodsFor: 'class variables' stamp: 'bgf 8/8/2009 01:36'!
privateTextBlockFlagMask
	^ PrivateTextBlockFlagMask! !

!QPresenceManager class methodsFor: 'defaults' stamp: 'bgf 8/6/2010 07:44'!
xmlUpdateInterval

	"Interval at which the client sends xml locator record updates to the server,
	 to increment the last-online time record-keeping."
	
	^ 299  "So one locator.xml update about every five minutes"! !

!QPresenceManager methodsFor: 'leadership' stamp: 'ar 4/16/2011 17:14'!
assignLeader: aSessionID toForum: aForumURL
	"The given user (identified by sessionID) is to assume meeting-leadership for the given forum. "
	| presence forum |

	forum := self forumSessionForURL: aForumURL.
	forum ifNil: [
		QLogger logWarning: 'Leader claim for inactive Forum: ', aForumURL.
		^ nil
		].
	forum leaderSID = aSessionID ifTrue: [
		QLogger logWarning: 'Leadership reclaimed by leader of ', aForumURL.
		^ nil
		].
	aSessionID ifNotNil: [
		presence := self presenceOfSession: aSessionID.
		presence ifNil: [
			QLogger logWarning: 'Leader claim by unknown session ', aSessionID.
			^ nil
		]].

	"You can only lead the forum you're in. 
	 The client is responsible to transfer leadership when teleporting.
	 Here, we retract leadership in other forums when assuming it in a new one.
	 (A location-presence change also clears leadership, so this is a bit belt&braces.) "
	forum leaderSID ifNotNilDo: [ : sid | 
		self retractLeader: sid forum: forum.
	].
	self retractLeader: aSessionID.

	presence ifNotNil: [
		presence isMeetingLeader: true.
		forum leaderSID: aSessionID.
		QLogger logInfo: '', aForumURL, ' lead is now ', (QStyles simplerLoginName: presence userLogin).
		self signal: #forum:leader:
			with: aForumURL
			with: presence.
	] ifNil: [
		self signal: #forum:leader:
			with: aForumURL
			with: nil.
	].
	
		! !

!QPresenceManager methodsFor: 'private' stamp: 'bgf 8/8/2009 02:02'!
checkAging
	"Regular check for timed-out location records."
	locationTable associations shallowCopy do: [ : assoc |
		(self presenceRecordIsCurrent: assoc value) ifFalse: [
			"Expired -- cancel the record (by setting its location to nil) "
			self session: assoc key 
				user: assoc value userLogin 
				location: nil 
				localTime: 0 
				quality: 0
				flags: 0
				robotMode: nil.
		]].

	"If we're maintaining any forum sessions that no-one has heartbeat, they die too..."
	forumTable associations shallowCopy do: [ : assoc |
		(self forumRecordIsCurrent: assoc value) ifFalse: [
			self timeoutForum: assoc value.
		]].

	"Finally, ensure that any leadership claims are satisfied by living users."
	forumTable values do: [ : record |
		record leaderSID ifNotNilDo: [ : leader |
			(locationTable 
				anySatisfy: [ : presence | 
					(presence sessionID = leader) and: [ presence urlString = record urlString] ])
				ifFalse: [ self retractLeader: leader forum: record ].
			]].

	(self future: 2000) checkAging! !

!QPresenceManager methodsFor: 'forum-sessions' stamp: 'bgf 8/8/2009 01:56'!
forum: urlString flag: aBitMask value: aBoolean
	"Accessor for the control-flags of a given forum session.
	 Deliberately not interpreted inside the presence manager to allow longer version-stability;
	 the org-services client determine the meaning of bits in the flag."
	| forum oldValue |	
	forum := self forumSessionForURL: urlString.	
	forum ifNil: [
		QLogger logWarning: 'Flags set in inactive forum: ', urlString.
		^ nil
	].
	oldValue := forum flag: aBitMask.
	oldValue = aBoolean ifFalse: [
		| oldWord |
		oldWord := forum flags.
		forum flag: aBitMask set: aBoolean.
		self signal: #forum:flags:oldFlags:  with: urlString with: (forum flags) with: (oldWord).
	].! !

!QPresenceManager methodsFor: 'forum-sessions' stamp: 'bgf 6/21/2009 18:27'!
forum: urlString flags: anInteger
	"Accessor for the control-flags of a given forum session.
	 Deliberately not interpreted inside the presence manager to allow longer version-stability;
	 the org-services client determine the meaning of bits in the flag."
	| forum oldValue |	
	forum := self forumSessionForURL: urlString.	
	forum ifNil: [
		QLogger logWarning: 'Flags set in inactive forum: ', urlString.
		^ nil
	].

	oldValue := forum flags.
	oldValue = anInteger ifFalse: [
		forum flags: anInteger.
		self signal: #forum:flags:oldFlags:  with: urlString with: anInteger with: oldValue.
	].! !

!QPresenceManager methodsFor: 'forum-sessions' stamp: 'bgf 6/21/2009 11:08'!
forum: urlString version: clientVersionString
	"Hearbeat indicating current activity in the given forum"
	| forum hbSeconds |

	hbSeconds := self island totalSeconds. 	"Timestamp for aging."
	forum := self forumSessionForURL: urlString.

	(forum notNil and: [ forum clientVersion = clientVersionString ]) ifTrue: [
		"No change in status, just record the heartbeat time."
		forum islandTime: hbSeconds.
	] ifFalse: [	
		clientVersionString ifNotNil: [
			forumTable at: urlString put:
				(forum := 
					(QForumSessionRecord) new
							urlString: urlString;
							clientVersion: clientVersionString;
							islandTime: hbSeconds;
						yourself).
			self signal: #forum:version:
				with: urlString 
				with: clientVersionString
		]].

	^ forum
! !

!QPresenceManager methodsFor: 'leadership' stamp: 'bgf 6/21/2009 11:08'!
forumLeaderSID: urlString
	"Return the sessionID of the leader of the given forum,
	 if that forum is in a session with a leader.  Else nil."
	(self forumSessionForURL: urlString) ifNotNilDo: [ : forum |
		^ forum leaderSID 
	].
	^ nil
! !

!QPresenceManager methodsFor: 'forum-sessions' stamp: 'bgf 8/8/2009 03:01'!
forumRecordIsCurrent: aForumRecord
	"Aging support.
	 Are any users still claiming to be in this forum? "

	^ locationTable values anySatisfy: [ : pr | pr hasURL: aForumRecord urlString ]
! !

!QPresenceManager methodsFor: 'forum-sessions' stamp: 'bgf 6/21/2009 12:40'!
forumSessionForURL: aForumURL
	"Return the current forum-session record for the specified forum, if any.
	 The result is a QForumSessionRecord, or nil if the forum is not active."

	^ forumTable at: aForumURL ifAbsent: [ nil ]! !

!QPresenceManager methodsFor: 'initialize' stamp: 'bgf 6/20/2009 11:30'!
initialize

	super initialize.
	self registerGlobal: #QPresenceManager.
	locationTable := Dictionary new.
	forumTable := Dictionary new.
	robots := Dictionary new.

	(self future: 5000) checkAging.! !

!QPresenceManager methodsFor: 'private' stamp: 'bgf 8/8/2009 20:52'!
knownURLString: aURLString
	"Pseudo-interning of url strings to save total memory footprint.
	 if the given URL already appears as a session url, then return the original."

	(forumTable detect: [ : sr | sr urlString = aURLString ] ifNone: [nil] ) 
		ifNotNilDo: [ :  sr | ^ sr urlString ].

	"First appearance."
	^ aURLString! !

!QPresenceManager methodsFor: 'accessing' stamp: 'bgf 2/13/2009 22:17'!
numUsersLoggedIn
	"Approximate... number of users with active presence records."
	^ locationTable size! !

!QPresenceManager methodsFor: 'locating' stamp: 'bgf 8/8/2009 22:02'!
presence: pArray
	"This is the basic heartbeat all clients send to org services to keep their presence alive."

	| curForum otherForums clientVersion pName rec |

	curForum := self knownURLString: (pArray at: 3).			"The forum I claim to be in now."
	otherForums := (((pArray at: 8) ifNil: [#()]) 				"Other forums I have active in my client."
					 collect: [ : forum | (self knownURLString: forum) ]) 
					 asArray.	
	clientVersion := (pArray at: 9).		"Will be nil to indicate logout."
	pName := (pArray at:10).

	"Touch all the forums the user has activated.
	 Clients in the lobby have none."
	(curForum notNil and: [ clientVersion notNil ]) ifTrue: [
		curForum = (QOrgServices lobbyURL) ifFalse: [
			self forum: curForum version: clientVersion.
			otherForums do: [ : url |
				self forum: url version: clientVersion
			]]].

	"Reassert my presence record."
	rec := self session: (pArray at: 1)
			user: (pArray at: 2)
			location: curForum 
			localTime: (pArray at: 4) 
			quality: (pArray at: 5)
			flags: (pArray at: 6)
			robotMode: (pArray at: 7).

	rec ifNotNil: [
		rec otherURLs: otherForums.
		rec publicName = pName ifFalse: [
			rec publicName: pName.
			self signal: #user:publicName: with: (pArray at: 1) with: pName.
		 ].
	].
! !

!QPresenceManager methodsFor: 'locating' stamp: 'bgf 7/25/2009 23:56'!
presenceOfLogin: aUserLogin

	"Return the current presence record for the given user, if any.
	 The record is a QPresenceRecord.
	 The clientTime is UTC-seconds, as reported by the user client.
	 If there is more than one entry, favor returning one that is in a forum rather than the lobby,
	 for better client-side information about where I already am."

	^ (locationTable select: [ :each| each userLogin = aUserLogin]) 
		ifNotEmptyDo: [ : entries | 
			(entries 
				detect: [ : kv | kv value urlString ~= QOrgServices lobbyURL] 
				 ifNone: [  entries associations first ]
			) value
		] ifEmpty: [ ^ nil ].
! !

!QPresenceManager methodsFor: 'locating' stamp: 'bgf 6/21/2009 10:59'!
presenceOfSession: aSessionID
	"Return the presence record for the user (identifed by user sessionID), if any.
	 The result is a QPresenceRecord, or nil."

	^ locationTable at: aSessionID ifAbsent: [ nil ]! !

!QPresenceManager methodsFor: 'locating' stamp: 'bgf 6/21/2009 09:44'!
presenceRecordIsCurrent: aLocationRecord
	"Has this record been refreshed within the expected heartbeat interval?"
	"Coordinated with replica in-space timeouts so that the OrgServices records die last."
	| recordAge |
	recordAge := (self island totalSeconds - aLocationRecord islandTime).
	recordAge <=  (TAvatarReplica visibleTimeout//1000 + 3)  ifTrue: [
		"Not yet expired.
		 But, if older than it ought to be, indicate the quality as low.  (Hearbeat will restore it.)"
		(aLocationRecord isInLobby not 
		 and: [ recordAge >= (3 * QPresenceManager heartbeatInterval)]) 
		 ifTrue: [
			aLocationRecord sessionQuality: (aLocationRecord sessionQuality min: 3).
			self signal: #user:quality: 
				with: (aLocationRecord sessionID) 
				with: (aLocationRecord sessionQuality).
		].
		^ true
	].

	^	false		"Past the sell-by-date."! !

!QPresenceManager methodsFor: 'locating' stamp: 'bgf 8/8/2009 02:54'!
presencesOfLogin: aUserLogin
	"Return all current presence records for the given user.
	 We do not support generally the re-use of logins for many sessions,
	 but the presence manager keeps such records.
	 The one important use is warning about tromping yourself from one machine
	 when already logged in from another."

	^ (locationTable select: [ :each| each userLogin = aUserLogin]) 
		asArray
! !

!QPresenceManager methodsFor: 'leadership' stamp: 'bgf 6/20/2009 21:11'!
retractLeader: leaderSID
	"Take away the leadership status of the given user-SID in any forums it holds."
	forumTable values do: [ : forum | 
		self retractLeader: leaderSID forum: forum ].

	(self presenceOfSession: leaderSID) ifNotNilDo: [ : presence | 
		presence isMeetingLeader: false ].! !

!QPresenceManager methodsFor: 'leadership' stamp: 'ar 4/16/2011 17:14'!
retractLeader: leaderSID forum: forumRecord
	"If the given user (SID) holds leadership for the given forum, retract it."
	| presence |
	forumRecord leaderSID = leaderSID ifFalse: [
		^ self
	].
	forumRecord leaderSID: nil.
	presence := (self presenceOfSession: leaderSID).
	presence ifNotNil: [
		presence urlString = forumRecord urlString ifTrue: [
			presence isMeetingLeader: false.
		].
		QLogger logInfo: 'Forum ', forumRecord urlString, ' leader cleared.'.
		self signal: #forum:leader: 
			with: (forumRecord urlString) 
			with: nil.
	].! !

!QPresenceManager methodsFor: 'robots' stamp: 'bgf 3/20/2008 00:51'!
robot: aSessionID mode: aSymbol
	"Mark the given user as running the given robotic mode."

	self robot: aSessionID mode: aSymbol silent: false! !

!QPresenceManager methodsFor: 'robots' stamp: 'bgf 3/20/2008 03:37'!
robot: aSessionID mode: aSymbol silent: bSilent
	"Mark the given user as running the given robotic mode.
	 The change is signaled unless bSilent is true"

	(self presenceOfSession: aSessionID) ifNotNilDo: [ : entry |
		aSymbol ifNil: [
			self stopBeingRobot: aSessionID silent: bSilent.
		] ifNotNil: [
			robots at: (entry sessionID) put: aSymbol.
			bSilent ifFalse: [
				self signal: #robot:mode: with: aSessionID with: aSymbol.
			].
		].
	] ifNil: [
		self stopBeingRobot: aSessionID silent: true
	].! !

!QPresenceManager methodsFor: 'robots' stamp: 'bgf 8/21/2008 20:28'!
robot: aSessionID takeAction: aSymbol withArguments: anArray

	"If the given user is a robot, issue an action signal for it."
	(self userIsRobot: aSessionID) ifTrue: [
		self signal: #robot:action:arguments: withArguments: {aSessionID . aSymbol . anArray }
	] ifFalse: [
		QLogger logWarning: 'Action received for non-robot: ', aSessionID.
	].! !

!QPresenceManager methodsFor: 'robots' stamp: 'bgf 3/20/2008 03:37'!
robotModeOfUser: aSessionID
	"The current mode symbol set for this robot, or nil if none."
	(self presenceOfSession: aSessionID) ifNotNilDo: [ : entry |
		^ robots at: (aSessionID) ifAbsent: [nil]
	].
	^ nil
! !

!QPresenceManager methodsFor: 'locating' stamp: 'bgf 8/8/2009 02:23'!
session: aSessionID user: aName location: aURL localTime: epochSeconds quality: qos flags: word robotMode: aSymbolOrNil
	"Record the location of the given user at the given time.
	 Emit an update signal if the location has changed."

	| presRecord hbSeconds |
	hbSeconds := self island totalSeconds. 	"Timestamp for aging."
	presRecord := self presenceOfSession: aSessionID.

	(presRecord notNil and: [presRecord urlString = aURL]) ifTrue: [
		"No change in location, just record the update time."
		presRecord clientTime: epochSeconds.
		presRecord islandTime: hbSeconds.
	] ifFalse: [	
		presRecord ifNotNil: [ 			
			self retractLeader: aSessionID
		].
		aURL ifNotNil: [
			locationTable at: aSessionID put:
				(presRecord := 
					QPresenceRecord 
							session: aSessionID
							login: aName 
							url: aURL asString 
							clientTime: epochSeconds
							islandTime: hbSeconds).
		] ifNil: [
			locationTable removeKey: aSessionID ifAbsent: [].
			self stopBeingRobot: aSessionID.
			presRecord := nil.
		].
		self signal: #user:location:time: 
			with: aSessionID 
			with: (aURL ifNotNil: [aURL asString])
			with: epochSeconds
	].

	(presRecord notNil and: [presRecord sessionQuality ~= qos ]) ifTrue: [	
		presRecord sessionQuality: qos.
		self signal: #user:quality: with: aSessionID with: qos
		].

	"Robot mode if needed - this keeps the robot mode fresh after timeout/reconnect,
	 but we do not signal this as a change requiring response/notification
	 (to break the circularity if a client receives a change instruction
	 between a emission of a locator-update and receipt of the signals arising from the update.)"
	aURL ifNotNil: [
		((self robotModeOfUser: aSessionID) = aSymbolOrNil) ifFalse: [
			self robot: aSessionID mode: aSymbolOrNil silent: true
		]].

	^ presRecord! !

!QPresenceManager methodsFor: 'signaling' stamp: 'bgf 2/26/2009 12:38'!
sig: aSymbol forum: url user: senderLogin arg: otherData

	"Raise this signal on myself,
	 so that org services can relay it to downstream listeners."

	self signal: #forumSignal:forForum:fromUser:argument: 
		with: aSymbol 
		with: url 
		with: senderLogin
		with: otherData.! !

!QPresenceManager methodsFor: 'signaling' stamp: 'bgf 2/26/2009 12:38'!
sig: aSymbol user: login1 from: login2 arg: userData

	"Signal the island, so that org services can pick it up and relay to the given user."

	self signal: #userSignal:forUser:fromUser:argument: with: aSymbol with: login1 with: login2 with: userData.! !

!QPresenceManager methodsFor: 'robots' stamp: 'bgf 3/20/2008 01:05'!
stopBeingRobot: aSessionID
	"Remove this user from those available for robotic automation
	 (if it is listed as available.)"
	self stopBeingRobot: aSessionID silent: false.! !

!QPresenceManager methodsFor: 'robots' stamp: 'bgf 8/18/2008 20:46'!
stopBeingRobot: aSessionID silent: bSilent
	"Remove this user from those available for robotic automation
	 (if it is listed as available.)"


	robots removeKey: aSessionID  ifAbsent: [ ^ self ].

	"Fallthrough means it was listed...."
	bSilent ifFalse: [
		self signal: #notRobot: with: aSessionID.
	].! !

!QPresenceManager methodsFor: 'private' stamp: 'bgf 6/21/2009 16:45'!
timeoutForum: forumRecord
	"There have been no clients reporting use of this forum for the aging period;
	 mark it as inactive and remove its session-record."

	forumTable removeKey: (forumRecord urlString) ifAbsent: [ ^ self ].

	forumRecord leaderSID ifNotNilDo: [ : leaderSID |
		"Just to be sure that (e.g.) lobby clients can be aware that there is no led-session here,
	 	 probably unneeded since the forum-idle state could cover that,
	 	 but for completeness...."
		self retractLeader: leaderSID forum: forumRecord
	].

	self signal: #forum:version:
		with: forumRecord urlString
		with: nil! !

!QPresenceManager methodsFor: 'locating' stamp: 'hrs 4/21/2010 11:23'!
user: aSessionID isPrivateTextBlocked: aBoolean

	"Change the isPrivateTextBlocked: for the given user;
	 these are entirely under client control.
	 If this is actually a change, then signal #user:isPrivateTextBlocked:  "

	| record |
	record := self presenceOfSession: aSessionID.
	record ifNotNil: [
		record islandTime: (self island totalSeconds).
		(record isPrivateTextBlocked ~= aBoolean) ifTrue: [
			record isPrivateTextBlocked: aBoolean.
			self signal: #user:isPrivateTextBlocked: with: aSessionID with: aBoolean.
		].
	].! !

!QPresenceManager methodsFor: 'locating' stamp: 'bgf 3/20/2008 04:41'!
user: aSessionID notice: aNoticeString

	"Change the notice for the given user;
	 these are entirely under client control.
	 If this is actually a change, then signal #user:notice:  "

	| record |
	record := self presenceOfSession: aSessionID.
	record ifNotNil: [
		record islandTime: (self island totalSeconds).
		(record notice ~= aNoticeString) ifTrue: [
			record notice: aNoticeString.
			self signal: #user:notice: with: aSessionID with: aNoticeString.
		].
	].! !

!QPresenceManager methodsFor: 'robots' stamp: 'bgf 3/20/2008 00:49'!
userIsRobot: aSessionID
	"Has this user marked itself available for automatic operation?"
	^ (self robotModeOfUser: aSessionID) notNil! !

!QPresenceManager methodsFor: 'locating' stamp: 'bgf 3/20/2008 12:43'!
userNamesInService: aServiceURL

	"Given a service URL, show all users currently located in that service.
	 Returns an Array of Strings, each entry a service-user login string
	 (no far-ref entanglements.) "

	^ ((locationTable associationsSelect: [ :a | a value urlString = aServiceURL ])
		values collect: [ : each | each userLogin ])
		asArray! !

TObject subclass: #QSessionJobListener
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Objects-Jobs'!

!QSessionJobListener commentStamp: '<historical>' prior: 0!
A proxy for a job-handler (a user session) in replicated islands.
The only role of this object is to recieve signals; it is stateless.
The class exists so that the listener can self-register on creation,
so signal-relaying always works from the get-go.!

!QSessionJobListener class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QSessionJobListener methodsFor: 'initialize-release' stamp: 'bgf 6/30/2007 17:44'!
initialize

	Processor activeIsland register: self name: (TObjectID for: 'Job Listener').! !

TObject subclass: #QTimeStamp
	instanceVariableNames: 'time what who'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Objects-Document'!

!QTimeStamp commentStamp: '<historical>' prior: 0!
Records a time for some user-relevant occurrence.
Timestamps comprise 
	time = a time in UTC seconds  (hence always come from some user-client perspective)
	who = a user-login 	(string identifying who) 
	what = a description (short string describing the action that occurred.)

A timestamp can be undefined (isUndefined),
meaning it contains no information.

(This is so that client classes can use timestamps created in initialize as instance variables,
and the check on whether there is timestamp information
can consistently be
	timestamp isDefined 
rather than timestamp notNil and: [ timestamp time > 0 ]   or whatever.)


!

!QTimeStamp class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QTimeStamp class methodsFor: 'instance creation' stamp: 'bgf 6/11/2009 19:55'!
fromXML: xmlElement prefix: pfx
	"Unpacks a time stamp from attributes, the converse of 
		writeXMLAsAttributesOn: xmlWriter prefix: pfx 
	 Returns an entirely undefined timestamp if any expected item is missing."

	^ (self new)
		who: (xmlElement attributeAt: pfx,'-who' ifAbsent: [ ^ self new ] )
		what: (xmlElement attributeAt: pfx, '-what' ifAbsent: [ ^ self new ] )
		utc: (xmlElement attributeAt: pfx ifAbsent: [ ^ self new ] ) asNumber! !

!QTimeStamp class methodsFor: 'islands' stamp: 'bgf 2/28/2009 14:25'!
howToPassAsArgument
	^ #passByCopy:! !

!QTimeStamp methodsFor: 'comparing' stamp: 'bgf 2/28/2009 14:17'!
< aTimeStamp

	^ self timeOrZero < (aTimeStamp timeOrZero)! !

!QTimeStamp methodsFor: 'comparing' stamp: 'bgf 2/28/2009 14:17'!
<= aTimeStamp

	^ self timeOrZero <= aTimeStamp timeOrZero! !

!QTimeStamp methodsFor: 'comparing' stamp: 'bgf 2/12/2009 21:17'!
= another
	^ (another class = self class)
	    and: [ time = another time ]
	    and: [ who = another who ]
	    and: [ what = another what ]! !

!QTimeStamp methodsFor: 'comparing' stamp: 'bgf 2/28/2009 14:17'!
>  aTimeStamp

	^ self timeOrZero > aTimeStamp timeOrZero! !

!QTimeStamp methodsFor: 'comparing' stamp: 'bgf 2/28/2009 14:17'!
>=  aTimeStamp

	^ self timeOrZero >= aTimeStamp timeOrZero! !

!QTimeStamp methodsFor: 'persist' stamp: 'bgf 2/12/2009 22:02'!
constructFieldsFrom: pStream prefix: prefix

	"If prefix is nil, expect the fields as 'time', 'who', 'what'.
	 If non-nil (eg. 'ctime') expect them as 'ctime', 'ctime-who', 'ctime-what'   "

	time := pStream readNumber: (prefix ifNil: ['time']) default: time.
	who := pStream readString: ((prefix ifNil: [''] ifNotNil: [ prefix, '-']),'who')  default: who.
	what := pStream readString: ((prefix ifNil: [''] ifNotNil: [ prefix, '-']), 'what')  default: what.
! !

!QTimeStamp methodsFor: 'persist' stamp: 'bgf 2/12/2009 20:22'!
constructFrom: pStream
	super constructFrom: pStream.
	self constructFieldsFrom: pStream prefix: nil.
! !

!QTimeStamp methodsFor: 'copying' stamp: 'bgf 2/12/2009 22:09'!
copyFrom: another
	"A minor extension so we can use copyFrom: ts to update instance-variable QTimeStamps
	 as the accessor implementation, without surplus guards.
	 I.e.
	    timestamp copyFrom: nil 
	 turns timestamp into one that answers true to isUndefined"

	another ifNil: [ 
		self initialize
	] ifNotNil: [
		super copyFrom: another
	].! !

!QTimeStamp methodsFor: 'comparing' stamp: 'bgf 2/12/2009 21:09'!
hash
	"Hash is reimplemented because = is implemented."
	^ time hash bitXor: (who hash)! !

!QTimeStamp methodsFor: 'initialize' stamp: 'bgf 2/12/2009 21:09'!
initialize
	"Leaving the QTimeStamp in default state makes it undefined."
	time := nil.
	what := nil.
	who := nil.! !

!QTimeStamp methodsFor: 'testing' stamp: 'bgf 2/12/2009 22:03'!
isDefined
	"Support for holding QTimeStamps as lifetime-scoped ivars
	 that may or may not hold useful timestamps."
	^ time notNil and: [ time > 0 ]! !

!QTimeStamp methodsFor: 'testing' stamp: 'bgf 2/12/2009 22:03'!
isUndefined
	"Support for holding QTimeStamps as lifetime-scoped ivars
	 that may or may not hold useful timestamps."
	^ time isNil or: [ time <= 0 ]! !

!QTimeStamp methodsFor: 'persist' stamp: 'bgf 2/13/2009 11:11'!
persistFieldsOn: pStream prefix: prefix
	"If prefix is nil, write the fields as 'time', 'who', 'what'.
	 If non-nil (eg. 'ctime', write them as 'ctime', 'ctime-who', 'ctime-what'.
	 This is broken out so that an object persist its timestamps inline, in its own persistence,
	 for economy in the XML size."

	pStream storeNumber: (prefix ifNil: ['time']) value:  time default: nil.
	pStream storeString: ((prefix ifNil: [''] ifNotNil: [ prefix, '-']), 'who') value: who default: nil.
	pStream storeString: ((prefix ifNil: [''] ifNotNil: [ prefix, '-']), 'what') value: what default: nil.! !

!QTimeStamp methodsFor: 'persist' stamp: 'bgf 2/12/2009 20:24'!
persistOn: pStream
	super persistOn: pStream.
	self persistFieldsOn: pStream prefix: nil.
	! !

!QTimeStamp methodsFor: 'accessing' stamp: 'bgf 6/11/2009 18:36'!
summary
	"A string reporting the timestamp, using fixed-length dates."
	self isDefined ifTrue: [
		^ String streamContents: [ : str | 
			str nextPutAll: (QForumVersionItem mtimeString: time).
			who ifNotNil: [ str space:3; nextPutAll: (QStyles simplerLoginName: who).
			what ifNotNil: [ str nextPutAll: '   ('; nextPutAll: what; nextPut: $)  ].
		]].
	] ifFalse: [
		^ 'Unknown'
	].! !

!QTimeStamp methodsFor: 'accessing' stamp: 'bgf 2/10/2009 21:38'!
time
	"Seconds UTC... the timestamp of this action."
	^ time! !

!QTimeStamp methodsFor: 'private' stamp: 'bgf 2/12/2009 22:15'!
time: utcSeconds
	"Seconds UTC... the timestamp itself."
	time := utcSeconds! !

!QTimeStamp methodsFor: 'private' stamp: 'bgf 2/28/2009 14:16'!
timeOrZero	
	^ time ifNil: [ 0 ]! !

!QTimeStamp methodsFor: 'accessing' stamp: 'bgf 2/28/2009 13:52'!
timeString
	"A string reporting the time itself in user-friendly form."
	self isDefined ifTrue: [
		^ QStyles dateStringFromUTC: time
	] ifFalse: [
		^ 'Unknown'
	].! !

!QTimeStamp methodsFor: 'accessing' stamp: 'bgf 2/12/2009 21:11'!
what
	"A very short description of whatever event I timestamp"
	^ what! !

!QTimeStamp methodsFor: 'private' stamp: 'bgf 2/12/2009 22:01'!
what: aString
	"A short descriptive tag for the occurrence"
	what := aString! !

!QTimeStamp methodsFor: 'accessing' stamp: 'bgf 2/12/2009 22:00'!
who
	"The user login string of the principal responsible for the occurrence"
	^ who! !

!QTimeStamp methodsFor: 'private' stamp: 'bgf 2/12/2009 22:00'!
who: aUserLogin
	"The user login string of the principal responsible for the occurrence"
	who := aUserLogin! !

!QTimeStamp methodsFor: 'initialize' stamp: 'bgf 2/12/2009 22:14'!
who: lstring what: descripString utc: utcSeconds
	"Update the timestamp to these particulars.
	 Changes to timestamps should always be done this way,
	 not by modifying individual fields piecemeal,
	 so we can observe such changes transactionally."
	^ self
		time: utcSeconds;
		who: lstring;
		what: descripString;
		yourself! !

!QTimeStamp methodsFor: 'private' stamp: 'bgf 2/22/2009 20:04'!
writeXMLAsAttributesOn: xmlWriter prefix: pfx 
	"Semi-private utility to write the timestamp as an  XML attribute triple
	 (for inclusion in a manifest entry)."
	self isDefined ifTrue: [
		xmlWriter
			attribute: pfx value: (time printString);
			attribute: pfx, '-who' value: (who ifNil: ['?']);
			attribute: pfx, '-what' value: (what ifNil: ['?']).
	].! !

TObject subclass: #QUserInfo
	instanceVariableNames: 'screenName smallPicture publicEmail personalSummary avatarColor badgePicture capabilities features userLogin tzOffset'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Objects-User'!

!QUserInfo commentStamp: '<historical>' prior: 0!
UserInfo is a block of publically visible (replicated) information
that describes a participant (user) to other participants.

QUserInfo contains a subset of the information managed by QUserPreferences,
and its contents are supplied and updated only from its user's QUserPreferences.
(Ideally,  A QUserInfo should only be modifiable from its user's session;
such protection is not in place at the moment.)

Fields:
	screenName - the user's public name (defaults to her login name)
	avatarColor	- a VectorColor, to use for avatar and laser coloring
	personalSummary - optional text the user provides about herself
	publicEmail	- an email address string, also optional.
	smallPicture	- a TTexture containing a small user-supplied image

Note that some fields (avatarColor, smallPicture) are in different types
than in the master QUserPreferences:
	avatarColor - a Croquet vector color (the master keeps a symbolic color name)
	smallPicture - a TTexture (the master keeps a Form)
!

!QUserInfo class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QUserInfo methodsFor: 'accessing' stamp: 'bgf 10/26/2006 14:49'!
avatarColor
	"A color to use when colorizing the user's avatar."
	^ avatarColor! !

!QUserInfo methodsFor: 'accessing' stamp: 'bgf 10/26/2006 19:37'!
avatarColor: aColor
	"The basic color with which to colorize the avatar."

	avatarColor := aColor.
	self signal:#avatarColorChanged! !

!QUserInfo methodsFor: 'accessing' stamp: 'bgf 4/3/2007 22:44'!
badgePicture
	"A small texture to be used (for instance) as the torso graphic for this user."
	^ badgePicture! !

!QUserInfo methodsFor: 'accessing' stamp: 'bgf 4/22/2008 08:22'!
badgePicture: aTexture
	"A small image to be used (for instance) as the icon graphic for this user."

	badgePicture == aTexture ifFalse: [
		badgePicture := aTexture.
		self signal:#badgePictureChanged
	].! !

!QUserInfo methodsFor: 'accessing' stamp: 'bgf 6/10/2009 11:50'!
capabilities
	"Set of symbols - the capability keys true for this user at this time"
	^ capabilities! !

!QUserInfo methodsFor: 'accessing' stamp: 'bgf 6/10/2009 11:50'!
capabilities: anArray
	"Set of symbols - the capability keys true for this user at this time"
	capabilities := anArray asSet.! !

!QUserInfo methodsFor: 'accessing' stamp: 'bgf 6/10/2009 11:49'!
features
	"Set of symbols - the feature-entitlement keys true for this user at this time"
	^ features! !

!QUserInfo methodsFor: 'accessing' stamp: 'bgf 6/10/2009 11:49'!
features: anArray
	"Set of symbols - the feature-entitlement keys true for this user at this time"
	features := anArray asSet! !

!QUserInfo methodsFor: 'initialize' stamp: 'bgf 6/10/2009 11:08'!
initialize

	super initialize.
	features := Set new.
	capabilities := Set new.

	"We have to register manually since the outside world retrieves us with 'get'."
	Processor activeIsland register: self name: (TObjectID for:'QUserInfo').! !

!QUserInfo methodsFor: 'initialize' stamp: 'bgf 6/10/2009 11:10'!
login: ulogin summary: sum email: email screenName: sn avColor: avColor face: faceTexture badge: badgeTexture capabilities: capSyms features: featureSyms tzOffset: minutes
	"Batch-initalize a userInfo (sent by the user preferences at space-entry.)"
	self userLogin: ulogin.
	self personalSummary: sum.
	self publicEmail: email.
	self screenName: sn.
	self avatarColor: avColor.
	self tzOffset: minutes.
	self capabilities: capSyms.
	self features: featureSyms.

	faceTexture ifNotNil: [ 
		[ self smallPicture: faceTexture ] on: Error do: [ :ex|
			QLogger logWarning: 'Error initializing face: ', (ex description).
		]].
	badgeTexture ifNotNil: [
		[ self badgePicture: badgeTexture ] on: Error do: [ :ex|
			QLogger logWarning: 'Error initializing badge: ', (ex description).
		]].! !

!QUserInfo methodsFor: 'initialize' stamp: 'bgf 4/22/2008 08:24'!
obliterate
	"User info gets a lot of future traffic, and holds bulky data."
	badgePicture ifNotNil: [ 
		badgePicture destroy.
		badgePicture := nil
	].
	smallPicture ifNotNil: [ 
		smallPicture destroy.
		smallPicture := nil
	].
	Processor activeIsland obliterate: self.

! !

!QUserInfo methodsFor: 'accessing' stamp: 'bgf 10/25/2006 16:08'!
personalSummary
	"A small text block, provided by the user as an introduction."
	^ personalSummary! !

!QUserInfo methodsFor: 'accessing' stamp: 'bgf 10/26/2006 19:37'!
personalSummary: aString
	"A small text block, provided by the user as an introduction."

	(personalSummary ~= aString) ifTrue: [
		personalSummary := aString.
		self signal: #personalSummaryChanged
	].! !

!QUserInfo methodsFor: 'accessing' stamp: 'bgf 10/25/2006 16:07'!
publicEmail
	"May be nil; if present, the user's Email address."
	^ publicEmail! !

!QUserInfo methodsFor: 'accessing' stamp: 'bgf 10/26/2006 19:37'!
publicEmail: aString
	"May be nil; if present, the user's Email address."

	(publicEmail ~= aString) ifTrue: [
		publicEmail := aString.
		self signal: #publicEmailChanged.
	].! !

!QUserInfo methodsFor: 'accessing' stamp: 'bgf 6/22/2009 17:48'!
publicName
	"What to put in user-name displays?
	 If the user has supplied a screen name, use that,
	 otherwise if we know the login, use a tidier form of that.
	 (Should only used for display and labeling, never for comparisons across time.)"
	^ QStyles simplerLoginName: (
				(screenName notNil and: [screenName size > 0]) 
					ifTrue: [ screenName ] 
					ifFalse: [ userLogin ] ).! !

!QUserInfo methodsFor: 'accessing' stamp: 'bgf 6/22/2009 18:00'!
screenName
	"The name the user has entered for themselves. If supplied, this becomes the publicName.
	 (If this is nil, we strip email details out of the login and use that as publicName.)"
	^ screenName! !

!QUserInfo methodsFor: 'accessing' stamp: 'bgf 10/26/2006 19:37'!
screenName: aString
	"What to call the user on-screen"

	(screenName ~= aString) ifTrue: [
		screenName := aString.
		self signal: #screenNameChanged
	].! !

!QUserInfo methodsFor: 'initialize' stamp: 'bgf 9/16/2007 17:34'!
signalChanges
	"Signal all the costume-related values to trigger reset in a new costume."
	self signal: #avatarColorChanged.
	self signal: #badgePictureChanged.
	self signal: #screenNameChanged.
	self signal: #smallPictureChanged.! !

!QUserInfo methodsFor: 'accessing' stamp: 'bgf 10/25/2006 16:06'!
smallPicture
	"A small texture to be used (for instance) as the icon graphic for this user."
	^ smallPicture! !

!QUserInfo methodsFor: 'accessing' stamp: 'bgf 4/22/2008 08:21'!
smallPicture: aTexture
	"A small image to be used (for instance) as the icon graphic for this user."

	smallPicture == aTexture ifFalse: [
		smallPicture := aTexture.
		self signal:#smallPictureChanged
	].! !

!QUserInfo methodsFor: 'accessing' stamp: 'bgf 7/26/2008 20:18'!
tzOffset
	"My local time zone's offset in minutes from GMT (as returned by Time primTimezoneOffset)."
	^ tzOffset! !

!QUserInfo methodsFor: 'accessing' stamp: 'bgf 7/26/2008 20:18'!
tzOffset: minutes
	"My local time zone's offset in minutes from GMT (as returned by Time primTimezoneOffset)."
	tzOffset := minutes! !

!QUserInfo methodsFor: 'accessing' stamp: 'bgf 5/28/2007 19:26'!
userLogin
	"This is the user's login name in the Qwaq Service."
	^ userLogin
! !

!QUserInfo methodsFor: 'accessing' stamp: 'bgf 5/28/2007 19:26'!
userLogin: aString
	userLogin := aString.
	self signal: #userLoginChanged
! !

TObject subclass: #QUserPreferences
	instanceVariableNames: 'userLogin userInfo avatarColor avatarCostume customFigures personalSummary publicEmail screenName smallPicture smallPictureTexture doNavPanel doSoundEffects pressToTalk useOpenAL pointOfView showUserNames viewAngle badgePicture badgePictureTexture owner inventoryXMLZip avatarXMLZip doAntialiasing doQMSAutoPause showMyName idleFramesPerSecond forumCacheSize forumCacheAge doPrivateText doAutoSave enableExperimentalFeatures videoQuality videoDisabled favoriteForums favoriteUsers showMy3DPointer toolStateXML doActivityReports machinePreferences showOnlySpeakers homeOrganization'
	classVariableNames: 'BadgeFile ColorTable PictureFile SmileyForm ViewAngleTable ViewpointTable'
	poolDictionaries: ''
	category: 'Qwaq-Objects-User'!

!QUserPreferences commentStamp: '<historical>' prior: 0!
A Qwaq Croquet user's preferences and settings.

QUserPreferences are currently stored as zipped XML
on the service-provider storage of the user's organization.

NOTE: when you add a new preference, you probably want to take the following steps:
	- set the default value in #initialize
	- add accessors (of course).. have the setter signal when the value changes
	- enable XML import/export (add it to #readXmlFrom: and #storeXmlOn:)
	- add it to #signalChanges

User description fields (copied via QUserInfo to the avatar replica):
	avatarColor		- an entry from ColorTable, the preferred avatar coloring.
	personalSummary - a few lines of text
	screenName 	- the user name presented in shared spaces 
	smallPicture  	- a small color Form depicting me.

Preferences (private to the participant harness):
	doNavPanel 	- should the nav panel appear in the UI?
	doSoundEffects	- ambient sound effects?
	pointOfView		- an entry from ViewpointTable, my preferred camera viewpoint!

!QUserPreferences class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QUserPreferences class methodsFor: 'accessing' stamp: 'bgf 11/11/2006 15:36'!
colorTable

	"the dictionary of named colors for avatar color preference"
	
	^ ColorTable  ifNil: [self initialize. ColorTable]! !

!QUserPreferences class methodsFor: 'defaults' stamp: 'bgf 11/17/2009 17:20'!
defaultPicture
	"the default user face"
	| tmp copy |
	^ SmileyForm ifNil: [
		SmileyForm := [
				"Overcome photoshop's odd treatment of the PNG" 
				tmp := (Form fromFileNamed: (FileDirectory pathFromURI: self defaultPictureIconFile)).
				copy := Form extent: (tmp extent) depth: 32.
				tmp displayOn: copy at: 0@0 rule: Form blend.
				copy

			] ifError: [:msg | 
				Transcript cr; show: 'QUserPreferences: ', self defaultPictureIconFile, ' not found.
Update your workspace area from SVN and reinitialize the class.'.
				 nil]
		].! !

!QUserPreferences class methodsFor: 'defaults' stamp: 'bgf 10/22/2008 17:59'!
defaultPictureIconFile
	^ 'icons/Qwaq/small-smile.png'! !

!QUserPreferences class methodsFor: 'defaults' stamp: 'bgf 10/22/2008 18:00'!
defaultPictureTexture
	"A TFileTexture using the smiley-face"
	^ QTextures find: (self defaultPictureTextureURI)! !

!QUserPreferences class methodsFor: 'defaults' stamp: 'bgf 10/22/2008 17:59'!
defaultPictureTextureURI
	^ 'Content/Qwaq/Textures/small-smile.png'! !

!QUserPreferences class methodsFor: 'accessing' stamp: 'bgf 8/28/2008 13:43'!
faceFilePath: userLogin
	"The service provider path to the face picture for the given user login
	 (which can be set in preferences or in the plain avatar details editor)"

	^ '/users/', userLogin, '/public/photos/preferences/', PictureFile! !

!QUserPreferences class methodsFor: 'objects from disk' stamp: 'bgf 3/22/2009 18:25'!
forUser: loginName inClient: anOwner

	"Return the QUserPreferences for the given user,
	 reading its record from XML if we find one.
	 The core preferences are kept in the record named 'croquet'"

	| result xml |
	result := QUserPreferences new.
	result userLogin: loginName.
	result owner: anOwner.
	xml := result getXmlTree: 'croquet'.
	xml ifNil: [
		"Fall back to the organization defaults, if provided."
		loginName = 'default' ifFalse: [
			xml :=  QUserPreferences orgDefaultXml: 'croquet' forOwner: anOwner 
		].
	].
	result initializeMachinePreferences.

	xml ifNotNil: [
		[
			result readXmlFrom: xml.
		] on: Error do: [ :ex|
			QLogger logWarning: 'Cannot read preference record: ', ex description.
		].
	] ifNil: [
		anOwner setNewUserPreferences: result
	].

	"Pre-cache my own avatar appearance items."
	result smallPicture ifNotNilDo: [ :face| 
			face = self defaultPicture ifFalse: [ 
				anOwner faceCache cacheFace: face forLogin: loginName
			]].
	anOwner faceCache cacheColor: result avatarColorValue forLogin: loginName.

	"Prefetch this once so there's a chance to retry before entering the space."
	result inventoryXMLZip.

	^ result! !

!QUserPreferences class methodsFor: 'class initialization' stamp: 'bgf 8/27/2008 23:59'!
initialize

	"QUserPreferences initialize."		

	super initialize.

	"Avatar color scheme choices."
	ColorTable := Dictionary newFromPairs: {
		 '(Random)' . 	nil .
		 'Aqua'		.	#(0.1 0.75 0.75) asVectorColor.
		 'Black'		. 	#(0.07 0.07 0.07) asVectorColor.
		 'Blue'		.	#(0.11 0.22 0.95) asVectorColor.
		 'Brown'	.	#(0.54 0.28 0.05) asVectorColor .
		 'Denim'		. 	#(0.5 0.6 0.96) asVectorColor .
		 'Green'		.	#(0.1 0.9 0.2) asVectorColor.
		 'Gray' 		. 	#(0.5 0.5 0.5) asVectorColor.
		 'Orange'	.	#(0.95 0.45 0.05) asVectorColor.
		 'Purple'	.	#(0.81 0.08 0.84) asVectorColor .
		 'Red' 		.	#(0.99 0.1 0.1) asVectorColor.
		 'White'		.	#(0.99 0.99 0.99) asVectorColor.
		 'Yellow'	.	#(0.95 0.85 0.08) asVectorColor.
		}.

	"Perspective choices."
	ViewpointTable := Dictionary newFromPairs: {
		'First Person'	.	0 .
		'Near'			.	2.0 .
		'Normal'		.	3.0 .
		'Far'			.	4.0 .
	}.

	SmileyForm := nil.
	self defaultPicture.

	"Camera viewing angle choices.
       These are a bit close together, so that avatars with different settings
 	 do not fall too far apart when at full-view distance from things like panels.
	 (Wide-views mean you stand closer, and might occlude someone with normal view.)"
	ViewAngleTable := Dictionary newFromPairs: {
		'Normal' 	. 	45.0 .
		'Wider'		.	48.0 .
		'Widest'		.	52.0 .
	}.

	BadgeFile := '_badge_picture.png'.
	PictureFile := '_small_picture.png'.! !

!QUserPreferences class methodsFor: 'objects from disk' stamp: 'bgf 2/7/2009 10:30'!
localForUser: loginName inClient: anOwner
	"A transient (use it and discard it) Preferences object, 
	 for use -only- for locally-stored records
	 (such as window shape records.)"

	| result |
	result := QUserPreferences new.
	result userLogin: loginName.
	^ result! !

!QUserPreferences class methodsFor: 'objects from disk' stamp: 'bgf 8/18/2007 17:23'!
orgDefaultXml: unitName forOwner: ownerUI

	"Return the default preferences xml unit (e.g. croquet, or avatar) from the organization, if any."
	| temp |
	temp := QUserPreferences new.
	temp userLogin: 'default'.
	temp owner: ownerUI.
	^ temp getXmlTree: unitName! !

!QUserPreferences class methodsFor: 'objects from disk' stamp: 'bgf 11/21/2007 15:26'!
orgDefaultXmlZip: unitName forOwner: ownerUI

	"Return the default preferences xml unit (e.g. croquet, or avatar) from the organization, if any."
	| temp |
	temp := QUserPreferences new.
	temp userLogin: 'default'.
	temp owner: ownerUI.
	^ temp getXmlZip: unitName! !

!QUserPreferences class methodsFor: 'utilities' stamp: 'bgf 12/22/2007 11:26'!
resetUser: userName forOrganization: svcManager

	"Forcibly remove preferences from
	 the server-side persistence for the given user.
	 May raise exceptions if the service-provider operation complains."

	| paths response |
	paths := {
		'/users/', userName, '/*.xml.zip' .	"inventory, avatar, croquet"
		'/users/', userName, '/*.backup'	.	"save some space"
		'/users/', userName, '/public' .		"pictures"
		'/users/', userName, '/private' .		"inventory documents"
	}.

	paths do: [ : path |
		response := svcManager deleteFiles: path forReal: true.
		response name == #ok ifFalse: [
			self error: response contentString.
		]
	].
		
! !

!QUserPreferences class methodsFor: 'accessing' stamp: 'bgf 11/11/2006 15:35'!
viewAngleTable
	^ ViewAngleTable ifNil: [ self initialize. ViewAngleTable ]! !

!QUserPreferences class methodsFor: 'accessing' stamp: 'bgf 11/11/2006 15:36'!
viewpointTable

	"the dictionary of named viewpoints for viewpoint preference"
	
	^ ViewpointTable ifNil: [self initialize. ViewpointTable]! !

!QUserPreferences class methodsFor: 'objects from disk' stamp: 'ar 3/23/2007 15:30'!
xmlDirectory
	"Directory where user preference xml files are kept."
	^FileDirectory prefsDirectory! !

!QUserPreferences class methodsFor: 'utilities' stamp: 'hrs 8/4/2008 21:00'!
xmlToZip: xmlData unitName: unitName

	"Convert the xml string to a zip-archive xml byte array 
	 suitable for upload to the server."

	^self xmlToZip: xmlData unitName: unitName resources: #()! !

!QUserPreferences class methodsFor: 'utilities' stamp: 'hrs 8/4/2008 21:00'!
xmlToZip: xmlData unitName: unitName resources: resourcesCollection

	"Convert the xml string to a zip-archive xml byte array 
	 suitable for upload to the server.  If resourcesCollection is not empty, 
	the elements' data is added to form a fat zip."

	| archive resourceData |
	archive := ZipArchive new.
	archive addDeflateString: xmlData as: unitName, '.xml'.
	resourcesCollection do: [:res |
		resourceData := TContentManager default cacheManager find: res sha.
		"The resourceData is already compressed."
		archive addString: resourceData as: res fileName.
		].
	^ ByteArray streamContents: [ :s | archive writeTo: s ].! !

!QUserPreferences class methodsFor: 'utilities' stamp: 'bgf 9/11/2007 11:12'!
zipToXml: data unitName: unitName

	"Given the zipped-xml data from the server, unpack the xml unit."
	| archive |
	archive := ZipArchive new.
	archive readFrom: data readStream.
	^ (archive memberNamed: unitName, '.xml') contents
	! !

!QUserPreferences methodsFor: 'accessing' stamp: 'hrs 9/5/2008 17:29'!
acceleratedAnimation
	"Answers the GPU acceleration technique (e.g., shader capability) that the user wants tried for such things as animated meshes, dynamic-effects materials, and so forth.  

This preference is what the user has requested, and is deliberately decoupled:

	* from changes over time.  It curently recognizes:
		#bestEffort - Software can try anything.
		#none - no GPU accelerated animation. (Can always be accomodated.)
nil is not an acceptable value, because there's too much chance of confusion between nil, 'nil' and #nil (and if this were stored here rather than in machinePreferences, #readXmlFrom: is not going to perform #acceleratedAnimation: if the stored value is nil).

	* from runtime interrogation.  Information about hardware capabilities is maintained separately, as in TAnimationAccelerationManager hardwareAnimationType, and the runtime behavior may be a combination of such separate state and this preference. A failure on trying may result in the system using a lesser value than what was requested, but will not generally change this preference.  (Even VM crash recovery changes TAnimationAccelerationManager hardwareAnimationType, but not this.)

	* from the UI. For example, QwaqPrefsPane can project the above values to just a boolean checkbox, while at the same time reflecting various interrogated values in the label for that checkbox."
	^machinePreferences acceleratedAnimation ifNil: [#bestEffort].! !

!QUserPreferences methodsFor: 'accessing' stamp: 'hrs 8/27/2008 16:49'!
acceleratedAnimation: aSymbolOrString
	"What kind of accelerated animation (e.g., shaders) does the user or system request?"
	machinePreferences acceleratedAnimation: aSymbolOrString! !

!QUserPreferences methodsFor: 'private' stamp: 'bgf 8/21/2008 20:37'!
atLocalName: fileName putData: ascii

	"Write ascii data as a local file for this user.
	 Consumes errors and transcripts them;
	 returns boolean (true for success.)"

	| file |
	[
		self class xmlDirectory assureExistence.
		file := self class xmlDirectory forceNewFileNamed: fileName.
		file ascii.
		file nextPutAll: ascii.
		file close.	
		^ true	
	] on: Error do:[:ex| 
		QLogger logWarning: 'Cannot locally save ', file , ' preferences: ', ex description.
		^ false
	].
! !

!QUserPreferences methodsFor: 'storage' stamp: 'bgf 5/9/2008 10:45'!
atName: unitName putXmlData: dataString withBackup: yesNo

	"Store the given preferences to the user's service-storage."

	| zip |
	zip := (self class xmlToZip: dataString unitName: unitName).
	self atServerName: unitName putXmlZip: zip withBackup: yesNo
! !

!QUserPreferences methodsFor: 'private' stamp: 'bgf 8/21/2008 20:37'!
atServerName: fileName putData: bytes

	"Per-user data storage on the service manager.
	 Write the data using the given fileName (a relative uri path is OK)
	 for this user. Consumes errors and transcripts them;
	 returns a boolean - true for success."

     | path response |
	[
		path := self serviceFilePath: fileName.
		response := owner serviceMgr put: bytes name: path informing: nil.
		response name = #ok ifFalse: [
			QLogger logWarning: 'Cannot upload ', path, ' preferences: ', 
							response contentString.
			^ false
		].
		^ true
	] on: Error do:[:ex| 
		QLogger logWarning: 'Cannot upload ', fileName , ' preferences: ', ex description.
		^ false
	].
! !

!QUserPreferences methodsFor: 'private' stamp: 'bgf 11/24/2009 14:09'!
atServerName: unitName putXmlZip: zipData withBackup: yesNo

	"Preferences storage on the service manager.
	  Write the data as the preferences unit of the given name
	 for this user. Consumes errors and transcripts them;
	 returns a boolean - true for success.ss"

     | path response |
	[
		path := '/users/', self serviceLogin, '/', unitName, '.xml.zip'.
		response := owner serviceMgr 
							put: zipData
							name: path 
							backup: yesNo
							informing: nil.
		response name = #ok ifFalse: [
			QLogger logWarning: 'Cannot upload ', unitName, ' preferences: ', 
								response contentString.
			^ false
		].
		^ true
	] on: Error do:[:ex| 
		QLogger logWarning: 'Cannot upload ', unitName , ' preferences: ', ex description.
		owner ifNotNil: [ 
			owner statusAlert: 'Settings were not saved.\Possible connection problem.'
		].
		^ false
	].
! !

!QUserPreferences methodsFor: 'accessing' stamp: 'bgf 5/29/2010 11:13'!
audioPlayer
	"The user's choice of audio player (output, speaker) device on this machine."
	| def |
	^machinePreferences audioPlayer ifNil: [
		[
			def := SoundPlayer defaultPlayer.  
		] on: Error do: [ 
			def := nil.	"Prim failure - expected on Linux."
		].
		def isEmptyOrNil ifTrue:  [
			def := QAudioDeviceManager reasonableDefaultDeviceFrom: QAudioDeviceManager playerDevices.
			].
		def
		]
! !

!QUserPreferences methodsFor: 'accessing' stamp: 'hrs 10/27/2008 12:28'!
audioPlayer: aSymbolOrString
	"Change the user's choice for audio player (speaker, output) on this machine."
	machinePreferences audioPlayer: aSymbolOrString! !

!QUserPreferences methodsFor: 'accessing' stamp: 'hrs 1/12/2010 11:27'!
audioPlayers
	"The available audio player (speaker, output) devices on this machine."
	^machinePreferences audioPlayers ifNil: ['']! !

!QUserPreferences methodsFor: 'accessing' stamp: 'hrs 10/27/2008 16:33'!
audioPlayers: aString
	"The available audio player (speaker, output) devices on this machine."
	^machinePreferences audioPlayers: aString! !

!QUserPreferences methodsFor: 'accessing' stamp: 'bgf 5/29/2010 11:13'!
audioRecorder
	"The user's choice of audio recorder (microphone, input) device on this machine."
	| def |
	^machinePreferences audioRecorder ifNil: [
		[
			def := SoundRecorder defaultRecorder.  
		] on: Error do: [
			def := nil.		"Prim failure - expected on Linux."
		].
		def isEmptyOrNil ifTrue:  [
			def := QAudioDeviceManager reasonableDefaultDeviceFrom: QAudioDeviceManager recorderDevices.
			].
		def
		].! !

!QUserPreferences methodsFor: 'accessing' stamp: 'hrs 10/27/2008 13:57'!
audioRecorder: aSymbolOrString
	"Change the user's choice for audio recorder (microphone, input) on this machine."
	machinePreferences audioRecorder: aSymbolOrString! !

!QUserPreferences methodsFor: 'accessing' stamp: 'hrs 1/12/2010 11:48'!
audioRecorders
	"The available audio recorder (microphone, input) devices on this machine."
	^machinePreferences audioRecorders ifNil: ['']! !

!QUserPreferences methodsFor: 'accessing' stamp: 'hrs 10/27/2008 16:32'!
audioRecorders: aString
	"The audio recorder (microphone, input) devices available on this machine."
	machinePreferences audioRecorders: aString! !

!QUserPreferences methodsFor: 'accessing' stamp: 'bgf 10/26/2006 13:48'!
avatarColor
	"the name of a dictionary color"
	^ avatarColor! !

!QUserPreferences methodsFor: 'accessing' stamp: 'bgf 2/28/2009 14:52'!
avatarColor: aColorName
	"Set the name of the avatar color (scheme) I use."

	avatarColor := aColorName.
	userInfo isValidFarRef ifTrue: [
		userInfo future avatarColor: (self class colorTable at: aColorName ifAbsent:[nil]).
	].

	self signal: #avatarColorChanged with: avatarColor! !

!QUserPreferences methodsFor: 'accessing' stamp: 'bgf 11/13/2006 14:22'!
avatarColorValue
	"The Color named by my 'avatarColor' preference."

	^ self class colorTable at: (avatarColor) ifAbsent: [nil]! !

!QUserPreferences methodsFor: 'accessing' stamp: 'bgf 8/21/2008 20:37'!
avatarCostume
	"The persistent avatar costume used as model for this user.
	 Loaded lazily, and kept in its own XML unit (name-avatar.xml)."

	| scene |
	avatarCostume ifNil: [
		scene :=  self avatarXMLZip.
		scene ifNil: [
			"Fall back to the organization default, if provided."
			userLogin = 'default' ifFalse: [
				scene := QUserPreferences orgDefaultXmlZip:'avatar' forOwner: owner
			].
		].
	].

	scene ifNotNil: [[  
			avatarCostume := (TSceneReader new readSceneFromZip: scene sceneName: 'avatar.xml') 
		]  on: Error do: [ :ex|
			QLogger logWarning: 'Cannot load avatar: ', ex description.
		]].
	^ avatarCostume
! !

!QUserPreferences methodsFor: 'accessing' stamp: 'bgf 11/21/2007 18:27'!
avatarCostume: anAvatarModel
	"Given an actual instance of the costume,
	 persist it and save to preferences storage,
	 and update my working-model avatarCostume to match it."

	self avatarCostumeData: (self class xmlToZip: (anAvatarModel persistString) unitName:'avatar').

! !

!QUserPreferences methodsFor: 'accessing' stamp: 'bgf 11/21/2007 18:26'!
avatarCostumeData: avatarSceneZip
	"The persistent avatar costume used as model for this user;
	 save to preferences storage, 
	and make my working-copy avatarCostume an instance of it."

	avatarCostume := TSceneReader new readSceneFromZip: avatarSceneZip sceneName: 'avatar.xml'.
	avatarCostume ifNotNil: [
		self avatarXMLZip: avatarSceneZip
	].
	self signal: #avatarCostumeChanged

	! !

!QUserPreferences methodsFor: 'accessing' stamp: 'bgf 11/21/2007 17:11'!
avatarXMLZip
	
	"The user's avatar is stored in a separate zipped-XML unit
	 so it can be saved and fetched separately.
	 We cache that data so it needn't be fetched at each space transition."

	^ avatarXMLZip ifNil: [ avatarXMLZip := self getXmlZip: 'avatar' ]! !

!QUserPreferences methodsFor: 'accessing' stamp: 'bgf 5/9/2008 11:11'!
avatarXMLZip: zippedAvatarScene
	
	"The user's inventory is stored in a separate XML unit
	 so it can be saved and fetched separately.
	 We cache that data so it needn't be fetched at each space transition."

	zippedAvatarScene = avatarXMLZip ifFalse: [
		self atServerName: 'avatar' putXmlZip: zippedAvatarScene withBackup: true.
		avatarXMLZip := zippedAvatarScene.
	].! !

!QUserPreferences methodsFor: 'accessing' stamp: 'bgf 4/3/2007 22:38'!
badgePicture
	"A form containing a user-supplied personal icon."
	^  badgePicture! !

!QUserPreferences methodsFor: 'accessing' stamp: 'bgf 2/28/2009 14:52'!
badgePicture: aForm
	"A form containing a user-supplied personal icon.
	 This is the main public accessor - setting it also creates
	 and sets the small picture texture.
	 WILL signal error if there are problems uploading."

	| ref |
	badgePicture := aForm.
	badgePicture ifNotNil: [
		badgePictureTexture := self putPublicPhoto: badgePicture asFile: BadgeFile.
	] ifNil: [
		badgePicture := nil.
		badgePictureTexture := nil.
	].

	"Send the texture into the space if we're live"
	userInfo isValidFarRef ifTrue: [
		badgePictureTexture ifNotNil: [
			ref := TIslandCopier new
					import:(TIslandCopier new export: badgePictureTexture) 
					to: userInfo island.
		] ifNil: [ 
			ref := nil
		].
		userInfo future badgePicture: ref.
	].	
	self signal: #badgePictureChanged.! !

!QUserPreferences methodsFor: 'accessing' stamp: 'bgf 2/8/2010 16:12'!
clearOrganizationInfo
	"Notification that the user is changing the working organization.
	 Some organizations are on servers with different file-storage."
	inventoryXMLZip := nil.
	avatarCostume := nil.
	avatarXMLZip := nil.! !

!QUserPreferences methodsFor: 'accessing' stamp: 'bgf 3/1/2009 09:49'!
customFigures
	"Answer the group of custom AnimatableFigures that this user has obtained."
	^customFigures ifNil: [
		customFigures := (self getXmlZip: 'custom-figures')
							ifNotNilDo: [:scene | 
								TSceneReader new readFatSceneFromZip: scene 
									sceneName: 'custom-figures.xml'
									documentManager: self owner documentManager.
								]
							ifNil: [TGroup new]
		]! !

!QUserPreferences methodsFor: 'accessing' stamp: 'hrs 8/4/2008 21:05'!
customFigures: object
	"Persist the custom AnimatableFigures that this user has obtained (as a fat zip)."
	| zipData writer persistString |
	writer := TSceneWriter new initResourceList. "Unlike #persistString, we init so that #unbundledResources works."
	persistString := writer storeScene: object.

	zipData := self class xmlToZip: persistString unitName:  'custom-figures' resources: writer unbundledResources.

	"No need to backup, because it only grows (unless we change it at the server, in which case we don't want 
	folks to go back and get stuff we've removed)."
	self atServerName: 'custom-figures' putXmlZip: zipData withBackup: false.! !

!QUserPreferences methodsFor: 'accessing' stamp: 'bgf 5/25/2008 13:08'!
doActivityReports
	"Does the user permit activity statistics reporting?"
	^ doActivityReports! !

!QUserPreferences methodsFor: 'accessing' stamp: 'bgf 5/25/2008 13:09'!
doActivityReports: aBool
	"Does the user permit activity statistics reportinging?"
	doActivityReports := aBool.
	self signal: #doActivityReportsChanged.! !

!QUserPreferences methodsFor: 'accessing' stamp: 'bgf 5/7/2007 15:11'!
doAntialiasing
	"Does the user want AA?"
	^ doAntialiasing! !

!QUserPreferences methodsFor: 'accessing' stamp: 'bgf 5/7/2007 16:22'!
doAntialiasing: aBool
	"Does the user want AA?"
	(doAntialiasing ~= aBool) ifTrue: [
		doAntialiasing := aBool.
		self signal: #doAntialiasingChanged.
	].! !

!QUserPreferences methodsFor: 'accessing' stamp: 'bgf 8/23/2007 09:33'!
doAutoSave
	"Does theuser want auto-save enabled?"
	^ doAutoSave! !

!QUserPreferences methodsFor: 'accessing' stamp: 'bgf 8/23/2007 09:33'!
doAutoSave: aBool
	"Does the user want auto-save enabled?"
	doAutoSave := aBool.
	self signal: #doAutoSaveChanged.! !

!QUserPreferences methodsFor: 'accessing' stamp: 'bgf 10/26/2006 16:44'!
doNavPanel
	"Does the user want the nav panel?"
	^ doNavPanel! !

!QUserPreferences methodsFor: 'accessing' stamp: 'bgf 10/26/2006 16:44'!
doNavPanel: aBool
	"Does the user want the nav panel?"
	doNavPanel := aBool.
	self signal: #doNavPanelChanged.! !

!QUserPreferences methodsFor: 'accessing' stamp: 'bgf 8/8/2009 21:15'!
doPrivateText
	"Does the user want private text?"
	^ doPrivateText! !

!QUserPreferences methodsFor: 'accessing' stamp: 'bgf 8/8/2009 21:16'!
doPrivateText: aBool
	"Does the user want private text?"
	doPrivateText := aBool.
	self signal: #doPrivateTextChanged.! !

!QUserPreferences methodsFor: 'accessing' stamp: 'jcg 6/24/2007 14:31'!
doQMSAutoPause
	"Does the user want QMS apps to pause when they're not being looked at?"
	^ doQMSAutoPause! !

!QUserPreferences methodsFor: 'accessing' stamp: 'jcg 6/24/2007 14:53'!
doQMSAutoPause: aBool
	"Does the user want AA?"
	(doQMSAutoPause ~= aBool) ifTrue: [
		doQMSAutoPause := aBool.
		self signal: #doQMSAutoPauseChanged with: aBool.
	].! !

!QUserPreferences methodsFor: 'accessing' stamp: 'bgf 10/26/2006 20:04'!
doSoundEffects
	"Does the user want sound effects?"
	^ doSoundEffects! !

!QUserPreferences methodsFor: 'accessing' stamp: 'bgf 10/26/2006 19:58'!
doSoundEffects: aBool
	"Does the user want sound effects?"
	doSoundEffects := aBool.
	self signal: #doSoundEffectsChanged.! !

!QUserPreferences methodsFor: 'sharepoint' stamp: 'ar 4/14/2010 11:45'!
documentRepositories
	"Document repositories are stored in the user's preferences as repositories.xml.zip."

	^self getServerXmlData: 'repositories'! !

!QUserPreferences methodsFor: 'sharepoint' stamp: 'ar 4/14/2010 11:13'!
documentRepositories: xmlBlob
	"Document repositories are stored in the user's preferences as repositories.xml.zip."

	^self atName: 'repositories' putXmlData: xmlBlob withBackup: false! !

!QUserPreferences methodsFor: 'accessing' stamp: 'jcg 9/5/2007 11:09'!
enableExperimentalFeatures
	"At any given time, we may have some 'experimental' features that Greg might want to demo, but which aren't ready for general consumption.  This preference controls whether they are accessible."
	^ enableExperimentalFeatures! !

!QUserPreferences methodsFor: 'accessing' stamp: 'jcg 9/5/2007 12:06'!
enableExperimentalFeatures: aBool
	"At any given time, we may have some 'experimental' features that Greg might want to demo, but which aren't ready for general consumption.  This preference controls whether they are accessible."
	enableExperimentalFeatures = aBool ifFalse: [
		enableExperimentalFeatures := aBool.
		self signal: #enableExperimentalFeaturesChanged.
	].
! !

!QUserPreferences methodsFor: 'favorites' stamp: 'bgf 1/31/2008 14:23'!
favoriteForums
	"Return the user's list of favorite forums,
	 fetching it from the server if necessary.
	 The result is a set of url strings."

	| ffXML |
	favoriteForums ifNil: [
		favoriteForums := Set new.
		ffXML := self getXmlTree: 'favorite-forums'.
		ffXML ifNotNil: [
			ffXML elementsDo: [ :e |
				e name == #url ifTrue: [
					favoriteForums add: e contentString
				].
			].
		].
	].
	^ favoriteForums! !

!QUserPreferences methodsFor: 'favorites' stamp: 'bgf 5/9/2008 10:45'!
favoriteForums: aSetOfUrls

	"Update the user's list of favorite forums, saving the new list on the server."

	| xml |

	favoriteForums := aSetOfUrls asSet.
	xml := (WriteStream on: (String new: 256)).
	xml nextPutAll: '<forums>'; crlf.
	favoriteForums do: [ : url|
		xml nextPutAll: '<url>'.
		xml nextPutAll: url.
		xml nextPutAll: '</url>'.
		xml crlf.
	].
	xml nextPutAll: '</forums>'; crlf.

	self atName: 'favorite-forums' putXmlData: (xml contents) withBackup: false.
	! !

!QUserPreferences methodsFor: 'favorites' stamp: 'bgf 1/31/2008 21:02'!
favoriteUsers
	"Return the user's list of favorite users,
	 fetching it from the server if necessary.
	 The result is a set of login name strings."

	| ffXML |
	favoriteUsers ifNil: [
		favoriteUsers := Set new.
		ffXML := self getXmlTree: 'favorite-users'.
		ffXML ifNotNil: [
			ffXML elementsDo: [ :e |
				e name == #name ifTrue: [
					favoriteUsers add: e contentString.
				].
			].
		].
	].
	^ favoriteUsers! !

!QUserPreferences methodsFor: 'favorites' stamp: 'bgf 5/9/2008 10:45'!
favoriteUsers: aSetOfNames

	"Update the user's list of favorite people, saving the new list on the server."

	| xml |

	favoriteUsers := aSetOfNames asSet.
	xml := (WriteStream on: (String new: 256)).
	xml nextPutAll: '<users>'; crlf.
	favoriteUsers do: [ : name |
		xml nextPutAll: '<name>'.
		xml nextPutAll: name.
		xml nextPutAll: '</name>'.
		xml crlf.
	].
	xml nextPutAll: '</users>'; crlf.

	self atName: 'favorite-users' putXmlData: (xml contents) withBackup: false
	! !

!QUserPreferences methodsFor: 'accessing' stamp: 'bgf 8/15/2007 22:59'!
forumCacheAge
	"Number of minutes to wait after last departure before removing a forum from the active cache."

	^ forumCacheAge! !

!QUserPreferences methodsFor: 'accessing' stamp: 'bgf 8/15/2007 22:59'!
forumCacheAge: anInteger

	forumCacheAge := (anInteger asInteger) max: 0.
	self signal: #forumCacheAgeChanged.! !

!QUserPreferences methodsFor: 'accessing' stamp: 'bgf 8/15/2007 22:57'!
forumCacheSize
	"Number of active forums to cache (older forums are turfed out.)
	 0 indicates unlimited size."
	^ forumCacheSize! !

!QUserPreferences methodsFor: 'accessing' stamp: 'bgf 8/15/2007 22:57'!
forumCacheSize: anInteger

	forumCacheSize := (anInteger asInteger) max: 0.
	self signal: #forumCacheSizeChanged.! !

!QUserPreferences methodsFor: 'private' stamp: 'bgf 8/21/2008 20:36'!
getDataAtServerName: relName

	"Per-user preferences data storage on the service manager.
	 Read the data using the given fileName (a relative uri path is OK)
	 for this user. Consumes errors and transcripts them; returns data or nil."

     | path |
	[
		path := self serviceFilePath: relName.
		^ owner serviceMgr get: path informing: nil.
		
	] on: Error do:[:ex| 
		QLogger logWarning: 'Found no ', relName , ' preferences. ', 	
			 ((ex messageText includesSubString: 'FileDoesNotExist') 
				ifFalse: [ex messageText]
				ifTrue: ['']).
		^ nil
	].
! !

!QUserPreferences methodsFor: 'private' stamp: 'bgf 8/21/2008 20:36'!
getLocalXmlData: unitName

	"Return the given preferences data unit (usually xml text) for this user,
	 from the local preferences files on this computer."
	| file |

	file := [self class xmlDirectory
				readOnlyFileNamed: userLogin, '-', unitName, '.xml.zip'.
			] on: Error do: [
				[self class xmlDirectory 
					readOnlyFileNamed: userLogin, '-', unitName, '.xml']
				on: Error do:  [ 
					QLogger logWarning: 'No ', unitName, ' preferences found locally.'.
					^ nil 
				].
			].
	file size = 0 ifTrue: [	"Common transient failure case."
		Transcript cr; show: 'Empty ', unitName, ' preferences found locally.'.
		file close.
		^ nil
	].

	"Read it in."
	[	[
			(file localName endsWith: 'zip') ifTrue: [
				file binary.
				^ (self class zipToXml: (file contentsOfEntireFile) unitName: unitName)
			] ifFalse: [
				file ascii.
				^ file contentsOfEntireFile
			].
	  	] ensure: [ file close ].

	] on: Error do: [ :exc|
		QLogger logWarning: 'Cannot locally read ', unitName, ' preferences: ', exc description.
		^ nil
	].! !

!QUserPreferences methodsFor: 'private' stamp: 'bgf 8/21/2008 20:36'!
getPublicPhotoFromFile: basefile

	"Put the given image as a texture file on the server in my public photo storage,
	 and return a TFileTexture referring to it."

	| data |
	[
		data := self getDataAtServerName: 'public/photos/preferences/', basefile.
		data ifNotNil: [
			^ (ImageReadWriter formFromStream: (ReadStream on: data)) 
		] ifNil: [ 
			^ nil
		].
	] on: Error do: [ :ex |
		QLogger logWarning: 'Cannot fetch ', basefile, ': ', ex description.
		^ nil
	]! !

!QUserPreferences methodsFor: 'private' stamp: 'bgf 8/21/2008 20:35'!
getServerXmlData: unitName

	"Preference retrieval from the service manager.
	Return the given preferences data unit (usually xml text) for this user; 
	nil if absent or retrieval fails."

     | path data |
	[

		path := '/users/', self serviceLogin, '/', unitName, '.xml.zip'.
		data :=  (owner serviceMgr get: path  informing: nil).
		data ifNotNil: [
			^ self class zipToXml: data unitName: unitName.
		] ifNil: [
			path := '/users/', self serviceLogin, '/', unitName, '.xml'.
			data := (owner serviceMgr get: path  informing: nil).
			data ifNil: [
					QLogger logWarning: 'Got empty ', unitName, ' prefs from server.'.
					^ nil			
				].
			^ data asString
		].

	] on: Error do:[:ex| 
		QLogger logWarning: 'Found no ', unitName , ' preferences. ', 	
					 ((ex messageText includesSubString: 'FileDoesNotExist') 
						ifFalse: [ex messageText]
						ifTrue: ['']).
		^ nil
	].
! !

!QUserPreferences methodsFor: 'private' stamp: 'bgf 3/22/2009 18:33'!
getServerXmlZip: unitName
	"Preference retrieval from the service manager.
	Return the given preferences data zip file for this user;
	Return nil if absent; raise error for other failures (e.g. timeout.)"
     | path data |
	[
		path := '/users/', self serviceLogin, '/', unitName, '.xml.zip'.
		data :=  (owner serviceMgr get: path  informing: nil).
		data ifNotNil: [
			^ data		"Got the .zip"
		].
		path := '/users/', self serviceLogin, '/', unitName, '.xml'.
		data := (owner serviceMgr get: path  informing: nil).
		data ifNil: [
			QLogger logWarning: 'No ', unitName, ' prefs from server.'.
			^ nil			
		].
		
		"Got an older non-zipped file."
		^ self class xmlToZip: data unitName: unitName

	] on: Error do:[:ex| 
		(ex description includesSubstring: 'FileDoesNotExist' caseSensitive: false) ifTrue: [
			QLogger logInfo: 'No ', unitName , ' preferences (no such file.)'.
			^ nil
		] ifFalse: [
			ex pass
		].
	].
! !

!QUserPreferences methodsFor: 'storage' stamp: 'bgf 8/21/2008 20:34'!
getXmlTree: unitName

	"Retrieve and parse the preferences xml unit of the given name, 
	 consuming errors and return nil on any problem."
	| data node |
	data := self getServerXmlData: unitName.
	data ifNil: [ ^ nil ].
	[
		node := XMLDOMParser parseDocumentFrom: (ReadStream on: data).
		^ node elements first
	] on: Error do: [
		QLogger logWarning:'', (self userLogin), '-', unitName, ' is not a preferences record.'.
		^ nil
	]! !

!QUserPreferences methodsFor: 'storage' stamp: 'bgf 3/22/2009 18:31'!
getXmlZip: unitName

	"Get the preferences XML of the given name -
	 try to get it from the service manager.  
	 Returns nil on any sort of problem (or if there is no such file.)"
	[
		^ self getServerXmlZip: unitName.
	] on: Error do:[:ex| 
		QLogger logWarning: 'Error fetching ', unitName , ' prefs (', ex messageText, ')'.
		^ nil
	].
! !

!QUserPreferences methodsFor: 'accessing' stamp: 'bgf 11/20/2009 20:54'!
homeOrganization
	"This is where a click-through login will take you."
	^ homeOrganization! !

!QUserPreferences methodsFor: 'accessing' stamp: 'bgf 11/20/2009 20:54'!
homeOrganization: aString
	"This is where a click-through login will take you."
	| ho |
	ho := (aString isEmptyOrNil ifTrue: [ '' ] ifFalse: [ aString ]).
	ho = homeOrganization ifFalse: [
		homeOrganization := ho.
		self signal: #homeOrganizationChanged.
	].
	! !

!QUserPreferences methodsFor: 'accessing' stamp: 'das 5/23/2007 16:31'!
idleFramesPerSecond

	^ idleFramesPerSecond.! !

!QUserPreferences methodsFor: 'accessing' stamp: 'das 5/23/2007 16:32'!
idleFramesPerSecond: ifps

	idleFramesPerSecond := ifps.
	self signal: #idleFramesPerSecondChanged.! !

!QUserPreferences methodsFor: 'initialize-release' stamp: 'bgf 8/28/2009 08:20'!
initialize
	super initialize.

	"Defaults (the ones that aren't derived from the user identity.)"
	doNavPanel := true.
	doActivityReports := true.
	doAutoSave := true.
	doPrivateText := true.
	doSoundEffects := true.
	doAntialiasing := true.
	doQMSAutoPause := false.
	forumCacheSize := 3.
	forumCacheAge := 5.
	pressToTalk := false.
	useOpenAL := true.
	pointOfView := 'First Person'.
	viewAngle := 'Normal'.
	showUserNames := true.
	showOnlySpeakers := false.
	showMyName := true.
	showMy3DPointer := true.
	idleFramesPerSecond := 5.
	smallPictureTexture := self class defaultPictureTexture.
	smallPicture := self class defaultPicture.
	enableExperimentalFeatures := false.
	videoQuality := #Normal.
	videoDisabled := false.! !

!QUserPreferences methodsFor: 'initialize-release' stamp: 'hrs 9/28/2009 18:38'!
initializeMachinePreferences
	"Initialize any per-machine preferences."
	machinePreferences := QMachinePreferences new signalledObject: self.
	[
		machinePreferences load.
	] on: Error do: [:err |
		QLogger logWarning: 'Cannot read machine preference record: ', err description.
	].! !

!QUserPreferences methodsFor: 'accessing' stamp: 'bgf 3/22/2009 18:19'!
inventoryXMLZip
	"The user's inventory is stored in a separate XML unit
	 so it can be saved and fetched separately.
	 Caches the data so it needn't be fetched at each space transition;
	 caches no-such-file results also so refetches aren't triggered by teleports."

	(inventoryXMLZip = #None) ifTrue: [ 
		^ nil		"We've cached that we don't have one."
	].

	^ inventoryXMLZip ifNil: [ 
		[
			inventoryXMLZip := self getServerXmlZip: 'inventory-v2'.  "1.1.38 +"
			inventoryXMLZip ifNil: [
				inventoryXMLZip := self getServerXmlZip: 'inventory'    "Pre 1.1.38" 
			].
			inventoryXMLZip ifNil: [ 
				inventoryXMLZip := #None.		"So we don't keep re-fetching on teleport"
				nil
			].	
		] on: Error do: [ :ex|
			QLogger logWarning: 'Cannot fetch inventory: ', ex description.
			nil
		]].
			
	
! !

!QUserPreferences methodsFor: 'accessing' stamp: 'bgf 5/9/2008 11:11'!
inventoryXMLZip: zippedInventoryScene
	
	"The user's inventory is stored in a separate XML unit
	 so it can be saved and fetched separately.
	 We cache that data so it needn't be fetched at each space transition.
	 Inventory is only stored on the server (no local copy) because
	 it generally refers to server-stored documents, images, and the like."

	zippedInventoryScene = inventoryXMLZip  ifFalse: [
		self atServerName: 'inventory-v2' putXmlZip: zippedInventoryScene withBackup: true.
		inventoryXMLZip := zippedInventoryScene
	].! !

!QUserPreferences methodsFor: 'accessing' stamp: 'hrs 4/27/2009 11:38'!
machinePreferences
	"Answer the machine-specific preferences."
	^machinePreferences! !

!QUserPreferences methodsFor: 'accessing' stamp: 'bgf 4/11/2007 23:01'!
owner
	"the client UI"
	^ owner! !

!QUserPreferences methodsFor: 'accessing' stamp: 'bgf 4/11/2007 23:02'!
owner: aParticipantUI
	"the client UI"
	owner := aParticipantUI! !

!QUserPreferences methodsFor: 'accessing' stamp: 'bgf 10/26/2006 13:54'!
personalSummary
	^ personalSummary! !

!QUserPreferences methodsFor: 'accessing' stamp: 'bgf 2/28/2009 14:52'!
personalSummary: aString
	"Set the screen name in my user info."
		"Set the screen name in my user info."
	personalSummary := aString.
	userInfo isValidFarRef ifTrue: [
		userInfo future personalSummary: aString.
	].! !

!QUserPreferences methodsFor: 'accessing' stamp: 'bgf 10/26/2006 20:50'!
pointOfView
	"My preferred avatar viewpoint Ia name from class viewpointTable)"

	^ pointOfView! !

!QUserPreferences methodsFor: 'accessing' stamp: 'bgf 10/26/2006 20:56'!
pointOfView: aViewpointName
	"Set my preferred avatar viewpoint.
	 The argument should be one of the keys in my class viewpointTable."

	pointOfView := aViewpointName.
	self signal: #pointOfViewChanged.
	
	! !

!QUserPreferences methodsFor: 'accessing' stamp: 'bgf 10/26/2006 20:54'!
povValue
	"My preferred avatar point of view: its numeric value."
	^ self class viewpointTable at: pointOfView ifAbsent: [1.5].
	! !

!QUserPreferences methodsFor: 'accessing' stamp: 'bgf 1/20/2008 14:01'!
pressToTalk
	"Does the user want pressToTalk?"
	^ pressToTalk! !

!QUserPreferences methodsFor: 'accessing' stamp: 'bgf 1/20/2008 14:02'!
pressToTalk: yesNo
	"Does the user want pressToTalk?"
	(yesNo ~= pressToTalk) ifTrue: [
		pressToTalk := yesNo.
		self signal: #pressToTalkChanged.
	].! !

!QUserPreferences methodsFor: 'accessing' stamp: 'bgf 10/26/2006 23:10'!
publicEmail
	"An email address for the user, visible in replicated space."
	^ publicEmail! !

!QUserPreferences methodsFor: 'accessing' stamp: 'bgf 2/28/2009 14:53'!
publicEmail: aString
	"An email address for the user, visible in replicated space."
	publicEmail := aString.
	userInfo isValidFarRef ifTrue: [
		userInfo future publicEmail: aString.
	].! !

!QUserPreferences methodsFor: 'accessing' stamp: 'bgf 8/9/2009 11:20'!
publicName
	"For client-side convenience; see userInfo publicName."
	^ screenName isEmptyOrNil ifTrue: [
		QStyles simplerLoginName: (userLogin) 
	   ] ifFalse: [
		screenName
	   ]! !

!QUserPreferences methodsFor: 'private' stamp: 'bgf 8/5/2008 19:30'!
putPublicPhoto: form asFile: basefile

	"Put the given image as a texture file on the server in my public photo storage,
	 and return a TFileTexture referring to it.
	 WILL raise ERROR for conversion or upload problems. "

	| tmp dir file txtr data tmpName |
	[
		"Make a PNG file from the (usually resized) snapshot.
		 We use PNG in case there's alpha,
	 	 although other formats would probably be a tiny bit more efficient."
		tmpName := (TObjectID new asUIDString), '.png'.
		tmp := QFileDialog getFileLocation:#temp.
		dir := FileDirectory default directoryNamed: tmp.
		file := (dir forceNewFileNamed: tmpName).
		file binary.
		(PNGReadWriter on: file) nextPutImage:  form.
		file close.
		txtr := TTextures find: tmp, (FileDirectory slash), tmpName.
	
		"Upload the encoded image for re-access later (they are small photos.)"
		file := dir readOnlyFileNamed: tmpName.
		file binary.
		data := file contentsOfEntireFile.
		
		dir deleteFileNamed: tmpName.

		(self atServerName: 'public/photos/preferences/', basefile putData: data) 
		 ifFalse: [
			self error: 'Upload to service host failed.'.
		].

		"Finally adjust the texture so it refers to the server-side document."
		txtr fileName: (self  serviceFilePath: 'public/photos/preferences/', basefile).
		^ txtr

	] on: Error do: [ :ex |
		self error: 'Could not save picture as resource:\' withCRs, (ex description).
	]! !

!QUserPreferences methodsFor: 'read/write' stamp: 'jcg 6/24/2010 14:55'!
readXmlFrom: xmlNode

	"Internalize this user preferences record from the given XML"
	| attr |

	xmlNode name ~= (self xmlTag) ifTrue: [
		self error: 'Not a user preferences record.'.
		^ self
	].

	"String fields"
	{#avatarColor . #homeOrganization .  #personalSummary . #publicEmail . #screenName . #pointOfView . #viewAngle . #videoQuality . #acceleratedAnimation} 
	do: [ :field|
		attr := xmlNode elementAt: field.
		attr ifNotNil: [
			self perform: (field, ':') asSymbol with: attr contentString.
		].
	].

	"boolean fields"
	{#doSoundEffects . #doNavPanel . 
	  #showUserNames . #showOnlySpeakers . #showMyName . 
	  #showMy3DPointer .  #doAntialiasing . #doAutoSave . 
	  #doPrivateText . #doQMSAutoPause . 
	  #enableExperimentalFeatures . #pressToTalk. 
	  #videoDisabled . #doActivityReports} 
	 do: [:field |		
		attr := xmlNode elementAt: field.
		attr ifNotNil: [
			self perform: (field, ':')asSymbol 
				with: ((attr attributeAt: 'value') = 'true').
		].
	].
	self useOpenAL: true. "OpenAL is always enabled now."
	
	"numeric fields"
	{#idleFramesPerSecond . #forumCacheAge . #forumCacheSize} do: [ :field|
		attr := xmlNode elementAt: field.
		attr ifNotNil: [
			self perform: (field, ':')asSymbol 
				with: ((attr attributeAt: 'value') asNumber).
		].
	].

	"my photo as a file texture"
	attr := xmlNode elementAt: 'smallPictureTexture'.
	attr ifNotNil: [
		smallPictureTexture := TSceneReader new readScene: attr contentString.
		"Also snag the image, for the preferences editor, and to make sure its valid."
		(self getPublicPhotoFromFile: PictureFile) ifNotNilDo: [:form|
			smallPicture := form.
		] ifNil: [
			"This prevents us re-using the 'local' record's picture-spec
	 		 to specify a picture that hasn't been sent to the current service."
			smallPictureTexture := self class defaultPictureTexture.
			smallPicture := self class defaultPicture.
		].
	] ifNil: [
		"Older format - my photo as an encoded png."
		attr := xmlNode elementAt: 'smallPicture'.
		attr ifNotNil: [
			"Note that this will trigger an upload of the badge picture,
	  		 so that when we save the preferences, the new file texture form
	 		 will have a referent."

			[
				self smallPicture: (PNGReadWriter on: (
					attr contentString base64Decoded asByteArray readStream)) nextImage.
			] on: Error do: [ : ex |
				QLogger logError: ex from: self. 
			].
		].
	].

	"my badge as a file texture"
	attr := xmlNode elementAt: 'badgePictureTexture'.
	attr ifNotNil: [
		badgePictureTexture := TSceneReader new readScene: attr contentString.
		"Also snag the image, for the preferences editor, and to ensure it is valid"
		(self getPublicPhotoFromFile: BadgeFile) ifNotNilDo: [:form|
			badgePicture := form.
		] ifNil: [
			badgePicture := nil.
			badgePictureTexture := nil.
		].
	] ifNil: [
		"Older format - old badge encoded inline in preferences"
		attr := xmlNode elementAt: 'badgePicture'.
		attr ifNotNil: [
			"Note that this will trigger an upload of the badge picture,
	  		 so that when we save the preferences, the new file texture form
	 		 will have a referent."
			[
				self badgePicture: (PNGReadWriter on: (
					attr contentString base64Decoded asByteArray readStream)) nextImage.
			] on: Error do: [ : ex |
				QLogger logError: ex from: self. 
			].
		].
	].! !

!QUserPreferences methodsFor: 'private' stamp: 'bgf 9/10/2008 22:19'!
removePublicPhoto: basefile

	"Delete the file from the server so it doesn't show up subsequent fetches. "
	[
		owner serviceMgr 
				deleteFiles: (self  serviceFilePath: 'public/photos/preferences/', basefile).

	] on: Error do: [ :ex |
		QLogger logInfo: 'Could not remove picture file (', ex description, ')'.
	]! !

!QUserPreferences methodsFor: 'ui-state' stamp: 'bgf 8/21/2008 20:34'!
restoreToolStates: toolsDictionary

	"Update the given tools to the saved tool states; 
	 true if saved states were found and applied.
	 All the meat is in QFloatingPane #restoreState: "

	|  node |

	toolStateXML ifNil: [
		toolStateXML := self getServerXmlData: 'tool-states'.
	].
	toolStateXML ifNil: [  ^ false].
	[
		node := XMLDOMParser parseDocumentFrom: (ReadStream on: toolStateXML).
		node := node elements first.
	] on: Error do: [ :ex|
		QLogger logWarning: '', (self userLogin), '-tool-states ignored: ', ex description.
		^ false
	].

	node elementsDo: [ :e |
		[ toolsDictionary at: e name ifPresent: [ :tool |
			tool restoreState: e.
		]] on: Error do: [ :ex|
			QLogger logWarning: 'Cannot apply tool state ', e name, ': ', ex description.
			^ false.
		].
	].
	^ true
	
! !

!QUserPreferences methodsFor: 'ui-state' stamp: 'bgf 3/27/2009 14:30'!
restoreWindow: recordName shape: theUI default: rectangle
	"If there's a size record for the window, apply it via #restoreShape: rect;
	 if not, apply the default if notnil. "
	|  x y w h node windowXml |
	[
		(windowXml := self getLocalXmlData: recordName) ifNil: [ 
			rectangle ifNotNil: [ theUI restoreShape: rectangle ].
			^ false 
		].
		node := XMLDOMParser parseDocumentFrom: (ReadStream on: windowXml).
		(ReadStream on: (node elements first contentString))
		 ifNotNilDo: [ :stream|
			[stream peek isAlphaNumeric or: [ stream peek = $-]] whileFalse: [stream skip: 1].
			x := Integer readFrom: stream.
			[stream peek isAlphaNumeric or: [ stream peek = $-]] whileFalse: [stream skip: 1].
			y := Integer readFrom: stream.
			[stream peek isAlphaNumeric or: [ stream peek = $-]] whileFalse: [stream skip: 1].
			w := Integer readFrom: stream.
			[stream peek isAlphaNumeric or: [ stream peek = $-]] whileFalse: [stream skip: 1].
			h := Integer readFrom: stream.	
			theUI restoreShape: (x@y extent: w@h).
			^  true.
		] ifNil:[
			rectangle ifNotNil: [ theUI restoreShape: rectangle ].
		].
	] on: Error do: [ :ex|
		QLogger logWarning: 'No saved window (', ex description, ')'.
		rectangle ifNotNil: [ theUI restoreShape: rectangle ].
	].
	^ false

	
! !

!QUserPreferences methodsFor: 'ui-state' stamp: 'bgf 5/9/2008 10:45'!
saveToolStates: toolsDictionary

	"Update the user's saved tool states on the server.
 	 All the meat is in QFloatingPane #stateToRestore; this just
	 assembles it all into a single XML document and writes it"

	| xml |

	xml := (WriteStream on: (String new: 256)).
	xml nextPutAll: '<tools>'; crlf.
	toolsDictionary keysAndValuesDo: [ :name  : player |
		player stateToRestore ifNotNilDo: [ : playerXML |
			xml nextPutAll: '<', name, '>'.
			xml nextPutAll:  playerXML.
			xml nextPutAll: '</', name, '>'.
		].
		xml crlf.
	].
	xml nextPutAll: '</tools>'; crlf.
	self atName: 'tool-states' putXmlData: (xml contents) withBackup: false.
	toolStateXML := xml contents.
	! !

!QUserPreferences methodsFor: 'ui-state' stamp: 'bgf 2/8/2009 12:02'!
saveWindow: recordName shape: rect
	"Record the bounds of the named window - LOCALLY on this machine.
	 (Unlike inner tools, we scope full-window memory to the hardware,
	 since different boxes have different monitors, etc."
	| xml |

	rect ifNotNil: [
		xml := (WriteStream on: (String new: 256)).
		xml nextPutAll: '<window>'; crlf.
		xml nextPutAll:  '', rect left, ' ', rect top, ' ', rect width, ' ', rect height.
		xml crlf.
		xml nextPutAll: '</window>'; crlf.
		self atLocalName: (userLogin, '-', recordName, '.xml') putData: (xml contents).
	].
	! !

!QUserPreferences methodsFor: 'accessing' stamp: 'bgf 10/26/2006 23:10'!
screenName
	"The name by which the user is known in replicated space."
	^  screenName! !

!QUserPreferences methodsFor: 'accessing' stamp: 'bgf 2/28/2009 14:53'!
screenName: aString

	"The name by which the user is known in replicated space."
	screenName := aString.
	userInfo isValidFarRef ifTrue: [
		userInfo future screenName: aString.
	].! !

!QUserPreferences methodsFor: 'storage' stamp: 'bgf 5/10/2007 19:42'!
serviceFilePath: fileName
	"The service manager get: path to use 
	 to deal with a file that is uploaded to personal preferences storage."

	^'/users/', self serviceLogin, '/', fileName! !

!QUserPreferences methodsFor: 'accessing' stamp: 'bgf 6/27/2007 13:16'!
serviceLogin

	"The login name applicable for the current service manager.
	 We defer to the login name held in the client's current service manager...
	 looking forward to the time when we'll be able span organizations in a session.
	 But when ul is 'default' it means we're looking for fallback records, so use that literally."

	userLogin = 'default' ifTrue: [ ^ userLogin ].
	^ owner serviceMgr user ifNil: [ 
		self  userLogin
	]! !

!QUserPreferences methodsFor: 'initialize-release' stamp: 'bgf 6/10/2009 11:23'!
setUserInfoRef: aFarRef inClient: aParticipant
	"Update my reference to the current in-space user info record,
	 and update the user information with my current information;
	 this is done as part of moving the avatar into a new space."
	| faceRef badgeRef |
	userInfo := aFarRef.

	aFarRef isValidFarRef ifFalse:[ ^ self ].

	"Inject textures for my face and badge, if any.  Note that these are simple TFileTextures,
	 so the injection is quite cheap ... just reference to a server-side texture file by name,
	 with all the sha-cache optimizations in play."
	smallPictureTexture ifNotNil: [
		faceRef := TIslandCopier new 
				import:(TIslandCopier new export:smallPictureTexture) 
				to: userInfo island.
	].
	badgePictureTexture ifNotNil: [
		badgeRef := TIslandCopier new 
				import:(TIslandCopier new export:badgePictureTexture) 
				to: userInfo island.
	].

	"Pass the configuration.   We have an omnibus future-message for this
	 to avoid a lot of separate future message sends and their overhead."
	userInfo future
		login: self userLogin
		summary: self personalSummary 
		email: self publicEmail 
		screenName: self screenName 
		avColor: self avatarColorValue
		face: faceRef
		badge: badgeRef
		capabilities: (aParticipant capabilities asArray)
		features: (aParticipant currentFeatures asArray)
		tzOffset: (Time primTimezoneOffset)

! !

!QUserPreferences methodsFor: 'accessing' stamp: 'bgf 2/21/2008 21:37'!
showMy3DPointer

	"Whether or not to display my 3D bobble-cursor to me."

	^ showMy3DPointer! !

!QUserPreferences methodsFor: 'accessing' stamp: 'bgf 2/21/2008 21:35'!
showMy3DPointer: aBool

	"Whether or not to display my 3D bobble-cursor to me."

	(aBool ~= showMy3DPointer) ifTrue: [
		showMy3DPointer := aBool.
		self signal: #showMy3DPointerChanged.
	].! !

!QUserPreferences methodsFor: 'accessing' stamp: 'bgf 5/20/2007 11:39'!
showMyName

	"Whether or not to display my own name above my own avatar."

	^ showMyName ifNil: [ false ]! !

!QUserPreferences methodsFor: 'accessing' stamp: 'bgf 5/20/2007 10:50'!
showMyName: aBool

	"Whether or not to display my own name above my own avatar."

	(aBool ~= showMyName) ifTrue: [
		showMyName := aBool.
		self signal: #showMyNameChanged.
	].! !

!QUserPreferences methodsFor: 'accessing' stamp: 'bgf 7/23/2009 19:22'!
showOnlySpeakers
	"If true, name tags are only shown for people who are speaking."
	^ showOnlySpeakers! !

!QUserPreferences methodsFor: 'accessing' stamp: 'bgf 7/23/2009 19:23'!
showOnlySpeakers: aBool
	"If true, name tags are only shown for people who are speaking."
	(aBool ~= showOnlySpeakers) ifTrue: [
		showOnlySpeakers := aBool.
		self signal: #showOnlySpeakersChanged.
	].! !

!QUserPreferences methodsFor: 'accessing' stamp: 'bgf 5/20/2007 11:57'!
showUserNames

	"Whether or not to display other users' names above their avatars."

	^ showUserNames! !

!QUserPreferences methodsFor: 'accessing' stamp: 'bgf 5/20/2007 11:57'!
showUserNames: aBool

	"Whether or not to display other users' names above their avatars."

	(aBool ~= showUserNames) ifTrue: [
		showUserNames := aBool.
		self signal: #showUserNamesChanged.
	].! !

!QUserPreferences methodsFor: 'initialize-release' stamp: 'bgf 11/20/2009 19:50'!
signalChanges

	"Signal all preferences changes so upstream observers can get in sync
	 through their normal mechanisms; used when the client first loads prefs."
	self signal: #avatarColorChanged.
	self signal: #avatarCostumeChanged.
	self signal: #badgePictureChanged.
	self signal: #doAutoSaveChanged.
	self signal: #doNavPanelChanged.
	self signal: #doActivityReportsChanged.
	self signal: #doPrivateTextChanged.
	self signal: #doSoundEffectsChanged.
	self signal: #homeOrganizationChanged.
	self signal: #pressToTalkChanged.
	self signal: #personalSummaryChanged.
	self signal: #doAntialiasingChanged.
	self signal: #videoDisabledChanged.
	self signal: #doQMSAutoPauseChanged with: doQMSAutoPause.
	self signal: #forumCacheSizeChanged.
	self signal: #forumCacheAgeChanged.
	self signal: #pointOfViewChanged.
	self signal: #publicEmailChanged.
	self signal: #screenNameChanged.
	self signal: #showUserNamesChanged.	
	self signal: #showOnlySpeakersChanged.
	self signal: #showMyNameChanged.
	self signal: #showMy3DPointerChanged.
	self signal: #smallPictureChanged.
	self signal: #useOpenALChanged with: useOpenAL.
	self signal: #enableExperimentalFeaturesChanged.
	self signal: #viewAngleChanged.
	self signal: #videoQualityChanged.
! !

!QUserPreferences methodsFor: 'accessing' stamp: 'bgf 10/26/2006 23:09'!
smallPicture
	"A form containing a user-supplied personal icon."
	^  smallPicture! !

!QUserPreferences methodsFor: 'accessing' stamp: 'bgf 2/28/2009 14:45'!
smallPicture: aForm
	"A form containing a user-supplied personal icon.
	 This is the main public accessor - setting it also creates
	 and sets the small picture texture.  
	 WILL signal error if there are problems uploading."

	| ref |
	smallPicture := aForm.
	(smallPicture notNil and: [smallPicture ~= self class defaultPicture]) ifTrue: [
		smallPictureTexture := self putPublicPhoto: smallPicture asFile: PictureFile.
	] ifFalse: [
		smallPicture := self class defaultPicture.
		smallPictureTexture := self class defaultPictureTexture.
		self removePublicPhoto: PictureFile.
	].
	"Update my own faces cache immediately."
	smallPicture ifNotNilDo: [ :face| 
		face = self class defaultPicture ifFalse: [ 
			owner faceCache cacheFace: face forLogin: userLogin
		]].
	"Send the texture.."
	userInfo isValidFarRef ifTrue: [
		smallPictureTexture ifNotNil: [
			ref := TIslandCopier new
				 import:(TIslandCopier new export:smallPictureTexture) 
				 to: userInfo island.
		] ifNil: [
			ref := nil.
		].
		userInfo future smallPicture: ref.
	].
	owner orgServices ifNotNilDo: [ : os | 
		os sendSignal: #smallPictureChanged forForum: '*' argument: nil.
			"Includes my login as the sender, second argument."
		].

	self signal: #smallPictureChanged.
! !

!QUserPreferences methodsFor: 'read/write' stamp: 'bgf 11/24/2009 13:55'!
storeXml

	"Store the preferences record; this method consumes errors internally."
	| writer |

	writer := XMLWriter on: (WriteStream on: (String new:5000)).
	self storeXmlOn: writer.
	self atName: 'croquet' putXmlData: writer stream contents withBackup: false.
	owner serviceMgr ifNotNil: [
		QLogger logInfo: 'Saved to ', (owner serviceMgr host).
	].			! !

!QUserPreferences methodsFor: 'read/write' stamp: 'bgf 11/20/2009 19:49'!
storeXmlOn: xmlWriter

	"Write this UserPreferences on the given XmlWriter stream."
	| value |
	xmlWriter startTag: self xmlTag.
		xmlWriter attribute: 'version' value: '1.0'.
	xmlWriter endTag.
	xmlWriter stream cr.

	"String fields"

	{#avatarColor . #homeOrganization . #personalSummary . #publicEmail . #screenName .#pointOfView . #viewAngle . #videoQuality} 
	do: [ :field|
		(value := self perform: field) ifNotNil: [
			xmlWriter startTag: field; endTag.
			xmlWriter pcData: (value).
			xmlWriter endTag: field.
			xmlWriter stream cr.
		].
	].

	"Boolean fields."
	{#doSoundEffects . #doNavPanel . 
	   #showUserNames . #showOnlySpeakers . #showMyName .   
	   #showMy3DPointer . 
	  #useOpenAL . #doAntialiasing . #doAutoSave . #doPrivateText .
	  #doQMSAutoPause. #enableExperimentalFeatures . #pressToTalk. 
	  #videoDisabled . #doActivityReports } 
	 do: [:field |
		(value := self perform: field) ifNotNil: [
			xmlWriter startTag: field.
			xmlWriter attribute: 'value' value: value asString.
			xmlWriter endEmptyTag: field.
			xmlWriter stream cr.
		].
	].

	"Numeric fields."
	{#idleFramesPerSecond . #forumCacheAge . #forumCacheSize } do: [:field |
		(value := self perform: field) ifNotNil: [
			xmlWriter startTag: field.
			xmlWriter attribute: 'value' value: value asString.
			xmlWriter endEmptyTag: field.
			xmlWriter stream cr.
		].
	].

	"Save my photo... just nest the persisted form of the texture."
	(smallPictureTexture notNil and: [smallPictureTexture isKindOf: TFileTexture]) ifTrue: [
		(smallPictureTexture = self class defaultPictureTexture) ifFalse: [
			xmlWriter stream cr.
			xmlWriter startTag: 'smallPictureTexture'; endTag;
				   pcData: (TSceneWriter new storeScene: smallPictureTexture);
				   endTag: 'smallPictureTexture'.
			xmlWriter stream cr.
		].
	].

	"Save my badge... just nest the persisted form of the texture."
	(badgePictureTexture notNil and: [badgePictureTexture isKindOf: TFileTexture]) ifTrue: [
		xmlWriter stream cr.
		xmlWriter startTag: 'badgePictureTexture'; endTag;
				   pcData: (TSceneWriter new storeScene: badgePictureTexture);
				   endTag: 'badgePictureTexture'.
		xmlWriter stream cr.
	].
	xmlWriter endTag: self xmlTag.! !

!QUserPreferences methodsFor: 'accessing' stamp: 'jcg 2/15/2007 17:44'!
useOpenAL
	"Does the user want sound effects to be played via OpenAL?"
	^ useOpenAL! !

!QUserPreferences methodsFor: 'accessing' stamp: 'jcg 6/24/2010 14:56'!
useOpenAL: aBool
	"Does the user want sound effects to be played via OpenAL?"
	aBool == true ifFalse: [self error: 'OpenAL is always enabled now'].
	useOpenAL := aBool.
	self signal: #useOpenALChanged with: aBool.! !

!QUserPreferences methodsFor: 'accessing' stamp: 'bgf 10/26/2006 21:28'!
userInfo
	"The current in-space userInfo for this user; always a TFarRef"
	^ userInfo! !

!QUserPreferences methodsFor: 'accessing' stamp: 'bgf 10/26/2006 13:53'!
userLogin
	"This is the non-shared basic identity key for this user."
	^ userLogin! !

!QUserPreferences methodsFor: 'accessing' stamp: 'bgf 2/7/2009 10:18'!
userLogin: login
	"This is the user's account login name,
	 held because it forms the name of various paths to shared storage."
	userLogin := login! !

!QUserPreferences methodsFor: 'accessing' stamp: 'jcg 5/15/2008 00:08'!
videoDisabled
	"Globally disable video?"
	^videoDisabled! !

!QUserPreferences methodsFor: 'accessing' stamp: 'jcg 5/15/2008 00:07'!
videoDisabled: boolean
	"Globally disable video?"
	videoDisabled ~= boolean ifTrue: [
		videoDisabled := boolean.
		self signal: #videoDisabledChanged.
	].! !

!QUserPreferences methodsFor: 'accessing' stamp: 'bgf 9/19/2007 11:29'!
videoQuality
	"Video manager quality choice."
	^ videoQuality! !

!QUserPreferences methodsFor: 'accessing' stamp: 'bgf 9/19/2007 11:29'!
videoQuality: aSymbol
	"Video manager quality choice."
	videoQuality ~= aSymbol ifTrue: [
		videoQuality := aSymbol.
		self signal: #videoQualityChanged.
	].! !

!QUserPreferences methodsFor: 'accessing' stamp: 'bgf 11/11/2006 16:45'!
viewAngle
	"My choice of camera view angle (a key in my class viewAngleTable)"

	^ viewAngle! !

!QUserPreferences methodsFor: 'accessing' stamp: 'bgf 11/11/2006 15:38'!
viewAngle: aViewAngleName
	"Set my preferred camera view angle.
	 The argument should be one of the keys in my class viewAngleTable."

	viewAngle := aViewAngleName.
	self signal: #viewAngleChanged.
	
	! !

!QUserPreferences methodsFor: 'accessing' stamp: 'bgf 11/11/2006 15:39'!
viewAngleValue

	"The numeric value (in degrees) of my choice of view angle."

	^ self class viewAngleTable at: viewAngle ifAbsent:[45.0]
	
	! !

!QUserPreferences methodsFor: 'read/write' stamp: 'bgf 10/26/2006 12:18'!
xmlTag
	^ 'user-preferences'! !

TObject subclass: #TChannel
	instanceVariableNames: 'channelName jointName values'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Objects-Motion'!

!TChannel commentStamp: 'das 12/4/2006 10:04' prior: 0!
The .bvh files load data into categories for things like translationX, translationY, translationZ, rotationX, rotationY, rotationZ, scaleX, scaleY, scaleZ. This is an object to hold those values. It is loaded from the BVH parser.

This was seriously hacked up by DAS. I am just using this as a simple importer and will replace it later with something that imports directly into the form I want.!

!TChannel class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!TChannel methodsFor: 'accessing' stamp: 'zamp 12/9/2005 09:15'!
channelName
	^channelName! !

!TChannel methodsFor: 'accessing' stamp: 'zamp 12/9/2005 09:15'!
channelName: aName
	channelName := aName! !

!TChannel methodsFor: 'initialize' stamp: 'zamp 12/9/2005 09:15'!
initialize
	super initialize.
	channelName := ''.
	values := OrderedCollection new.! !

!TChannel methodsFor: 'accessing' stamp: 'zamp 1/20/2006 09:08'!
jointName
	^jointName! !

!TChannel methodsFor: 'accessing' stamp: 'zamp 1/20/2006 09:08'!
jointName: aName
	jointName := aName! !

!TChannel methodsFor: 'accessing' stamp: 'zamp 12/9/2005 09:15'!
values
	^values! !

TObject subclass: #TMotion
	instanceVariableNames: 'motionName channelsDictionary frameCount frequency continuousMotion streamFrames translateMotion doNotReset'
	classVariableNames: 'DictionaryOfMotions'
	poolDictionaries: ''
	category: 'Qwaq-Objects-Motion'!

!TMotion commentStamp: 'das 12/4/2006 10:04' prior: 0!
When a .bvh file is loaded, it is put into a TMotion object. This object stores things like the frequency, the frame count, and the channels. The motions can be played with a TMotionPlayer.


This was seriously hacked up by DAS. I am just using this as a simple importer and will replace it later with something that imports directly into the form I want.!

!TMotion class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!TMotion methodsFor: 'accessing' stamp: 'zamp 1/20/2006 11:24'!
channelsDictionary
	^channelsDictionary! !

!TMotion methodsFor: 'accessing' stamp: 'zamp 1/20/2006 11:24'!
channelsDictionary: dict
	channelsDictionary := dict! !

!TMotion methodsFor: 'accessing' stamp: 'zamp 1/20/2006 13:47'!
continuousMotion
	^ continuousMotion! !

!TMotion methodsFor: 'accessing' stamp: 'zamp 1/20/2006 13:47'!
continuousMotion: bool
	continuousMotion := bool! !

!TMotion methodsFor: 'accessing' stamp: 'zamp 1/24/2006 09:36'!
doNotReset
	^doNotReset! !

!TMotion methodsFor: 'accessing' stamp: 'zamp 1/24/2006 09:36'!
doNotReset: bool
	doNotReset := bool! !

!TMotion methodsFor: 'accessing' stamp: 'zamp 1/20/2006 10:16'!
frameCount
	^frameCount! !

!TMotion methodsFor: 'accessing' stamp: 'zamp 1/20/2006 10:15'!
frameCount: count
	frameCount := count! !

!TMotion methodsFor: 'accessing' stamp: 'zamp 1/20/2006 10:16'!
frequency
	^frequency! !

!TMotion methodsFor: 'accessing' stamp: 'zamp 1/20/2006 10:16'!
frequency: freq
	frequency := freq! !

!TMotion methodsFor: 'initialize' stamp: 'zamp 5/5/2006 11:17'!
initialize
	super initialize.

	channelsDictionary := Dictionary new.

	continuousMotion := true.
	streamFrames := true.
	translateMotion := true.
	doNotReset := false.
! !

!TMotion methodsFor: 'converting' stamp: 'bgf 1/3/2008 15:55'!
makeSkeletalAnimation
	| keyFrames matrices matrix translation val scale initialTranslations |
	keyFrames := OrderedCollection new: frameCount asInteger.
	"A joint can have a translation in each key frame:
	     for the root, this means moving around in place;
	     for others, it would mean stretching limbs, which I've never seen.
       Sometimes, a BVH will have a non-zero initialTranslation (for the root).
	 I'm not sure what that is supposed to express, but doesn't make sense to me,
	 and it causes our animatins to leap somewhere when they start animating.
       So I'm zeroing that out, and storing only the relative translation to that first frame."
      initialTranslations := Dictionary new.
	1 to: frameCount do: [:frameIndex |
		matrices := Dictionary new.
		channelsDictionary keysAndValuesDo: [:jointName :channels |
			matrix := Matrix4x4 identity.
			translation := Vector3 zero.
			scale := Vector3 x: 1 y: 1 z: 1.
			channels do: [:channel |
				val := channel values at: frameIndex.
				channel channelName caseOf: {
					[ #Xposition ] -> [translation x: val ].
					[ #Yposition ] -> [translation y: val ].
					[ #Zposition ] -> [translation z: val ].
					[ #Xrotation ] -> [ matrix addRotationAroundX: val. ].
					[ #Yrotation ] -> [ matrix addRotationAroundY: val. ].
					[ #Zrotation ] -> [ matrix addRotationAroundZ: val. ].
					[ #Xscale ] -> [scale x: val.].
					[ #Yscale ] -> [scale y: val.].
					[ #Zscale ] -> [scale z: val.].
					}.
				].
			translation := translation - (initialTranslations at: jointName ifAbsentPut: [translation]).
			matrix translation: translation * scale.
			matrices at: jointName put: matrix.
			].
		keyFrames add: (TSkeletonKeyFrame new time: frameIndex - 1* frequency; 
							boneMatrices: matrices).
		].
	^TSkeletonAnimation new name: motionName; 
			keyframes: keyFrames;
			length: keyFrames size * frequency.! !

!TMotion methodsFor: 'accessing' stamp: 'zamp 1/20/2006 10:20'!
motionName
	^motionName! !

!TMotion methodsFor: 'accessing' stamp: 'zamp 1/20/2006 10:20'!
motionName: obj
	motionName := obj! !

!TMotion methodsFor: 'accessing' stamp: 'zamp 6/2/2006 11:45'!
name
	^motionName! !

!TMotion methodsFor: 'accessing' stamp: 'zamp 1/20/2006 13:43'!
streamFrames
	^streamFrames! !

!TMotion methodsFor: 'accessing' stamp: 'zamp 1/20/2006 13:42'!
streamFrames: bool
	streamFrames := bool! !

!TMotion methodsFor: 'accessing' stamp: 'zamp 1/20/2006 13:42'!
translateMotion
	^ translateMotion! !

!TMotion methodsFor: 'accessing' stamp: 'zamp 1/20/2006 13:42'!
translateMotion: bool
	translateMotion := bool! !

!Island methodsFor: '*Qwaq-Objects-Documents' stamp: 'bgf 4/29/2009 21:28'!
documentResources
	"All shared-documents registered as resources,
	 but only unique instances.  (Resource registration is identity based,
	 and the same document - particular an image document -
	 may appear in several contexts.  
	 This accessor returns only one of such repeatedly-registered documents, using equality."
	^(self resourcesAt: (QSharedDocument resourceType)) asSet asArray! !

!Island methodsFor: '*Qwaq-Objects-Documents' stamp: 'bgf 7/19/2010 11:47'!
imageDocumentResources
	"All QImageDocuments registered as resources,
	 but only unique instances."
	^((self resourcesAt: (QSharedDocument resourceType)) asSet 
		select: [ : doc | doc isKindOf: QImageDocument ])
			asArray! !

TFractalTerrain subclass: #QFractalTerrain
	instanceVariableNames: 'randomSeed uvFactor mesh'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Objects-Space'!

!QFractalTerrain commentStamp: '<historical>' prior: 0!
Fractal terrain with persistence support;
all that is held in the XML are the generation parameters.!

!QFractalTerrain class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QFractalTerrain class methodsFor: 'defaults' stamp: 'bgf 3/21/2009 12:51'!
defaultDoContextMenu
	"Fewer unintended menus."
	^ false! !

!QFractalTerrain class methodsFor: 'defaults' stamp: 'bgf 6/4/2009 16:12'!
defaultIsFromTemplate
	"Reduces xml volume since these are most always terrain."
	^ true! !

!QFractalTerrain methodsFor: 'mesh building' stamp: 'bgf 1/26/2007 23:32'!
buildMesh

	^ mesh := super buildMesh.! !

!QFractalTerrain methodsFor: 'testing' stamp: 'bgf 2/18/2009 21:30'!
canEditMaterial
	^ mesh notNil! !

!QFractalTerrain methodsFor: 'persist' stamp: 'bgf 1/26/2007 23:48'!
constructFrom: pStream
	super constructFrom: pStream.

	texture := pStream readObject: 'texture' default: nil.
	self initializeSize: (pStream readNumber: 'size' default: 2)
		smoothness: (pStream readNumber: 'smoothness' default: 1.0)
		heightScale: (pStream readNumber: 'heightScale' default: 1.0)
		randomSeed: (pStream readNumber: 'randomSeed' default: (Random new seed)).

	self genGrid.
	self buildMesh.

	self scale: (pStream readNumber: 'scale' default: scale).

	(pStream readObject: 'material' default: nil) ifNotNilDo: [ :mat| mesh material: mat ].

	self uvFactor: (pStream readPoint: 'uvFactor' default: uvFactor).
! !

!QFractalTerrain methodsFor: 'initialize' stamp: 'bgf 1/27/2007 21:31'!
initializeSize: sz smoothness: smooth heightScale: hscale randomSeed: seed

	"Intercept to save the seed; the other attributes are already instance variables."

	uvFactor := uvFactor ifNil: [1@1]. 
	randomSeed := seed.
	^ super initializeSize: sz smoothness: smooth heightScale: hscale randomSeed: seed! !

!QFractalTerrain methodsFor: 'persist' stamp: 'bgf 1/26/2007 23:54'!
persistOn: pStream
	super persistOn: pStream.

	pStream storeObject: 'texture' value: texture default:nil.
	pStream storeNumber: 'size' value: size.
	pStream storeNumber: 'smoothness' value: smoothness default: 1.0.
	pStream storeNumber: 'heightScale' value: heightScale default: 1.0.
	pStream storeNumber: 'randomSeed' value: randomSeed.
	pStream storePoint: 'uvFactor' value: uvFactor.
	pStream storeNumber: 'scale' value: scale default: 1.0.

	mesh ifNotNil: [
		pStream storeObject: 'material' value: mesh material default: nil.
	].! !

!QFractalTerrain methodsFor: 'accessing' stamp: 'bgf 1/26/2007 23:36'!
uvFactor
	^ uvFactor! !

!QFractalTerrain methodsFor: 'accessing' stamp: 'bgf 1/26/2007 23:33'!
uvFactor: aPoint

	uvFactor := aPoint.
	mesh ifNotNil: [
		mesh texCoords do: [ :v | 	
			v x: v x * uvFactor x.
			v y: v y * uvFactor y.
		].
	].! !

TAvatarReplica subclass: #QAvatarReplica
	instanceVariableNames: 'activities userInfo selectedPoint selectedFrame avatarHeight speaker inventory islandURL arrowCursor shadow shadowTranslation shadowExtent label notice lastSpeech birthTime hasVideoStream lastViewportChange currentPlace currentPlaceStance currentPlaceVia besideMePlace currentFocalTransform lastMovedTime isRecordingSession isBroadcastingSession followPlaceTime standin pasteboardScrap deletionScrap editBox camera texturePalette'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Objects-User'!

!QAvatarReplica commentStamp: '<historical>' prior: 0!
This is Qwaq's extended avatar replica.
It adds 
	- support for keyboard-based driving (category 'driving')
	- some support for some keystroke commands
	- support for multicostume dynamic switching for Meredith,
	- support for QUserInfo personal information and settings,
	- imposter billboard rendering for faraway avatars.
	- handling of space landmarks and tabbing
	- a rolodex to serve as a wallet for personal content
	- an #activities set-of-symbols that allow human-user-status to be depicted in the avatar/costume!

!QAvatarReplica class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QAvatarReplica class methodsFor: 'undo-redo' stamp: 'bgf 3/15/2010 17:34'!
attachNote: aNote toFrame: sFrame capture: noteCapture kind: aString
	"Undoable note-creation"

	^ QBlockAction
		undoBlock: [ : avatar | aNote future obliterate ]
		redoBlock: [ : avatar | avatar future undoDeleteFrames: noteCapture]
		description: aString
		requiredFrames: { sFrame }! !

!QAvatarReplica class methodsFor: 'undo-redo' stamp: 'bgf 8/20/2010 14:07'!
dropMaterialOntoFrame: target oldState: matStateOld  newState: matStateNew
	"Private support for the undo of a drop of a material.
	 The oldState and newState are from #captureState
	 on the target frame's primary material."
	
	^ QBlockAction
		undoBlock: [ : avatar | target future dropMaterialState: matStateOld ]
		redoBlock: [ : avatar | target future dropMaterialState: matStateNew ]
		description: 'Paint'
		requiredFrames: { target }
		! !

!QAvatarReplica methodsFor: 'status' stamp: 'bgf 4/22/2008 10:34'!
activities
	"An array (for far-ref access convenience)
	 of symbols describing the high-level activities 
	 that the avatar's human user is currently engaged in.
	 Possibilities include
	 	#Away  		(exclusive; when present, this is the only activity)
		#Speaking		(not yet implemented... user is sending microphone traffic)
		#Idle			(no mouse/keyboard/voice activity from the user in > one minute)
		#Typing		(currently means 'in app-input keyboard mode', will become more dynamic)
		#Driving		(foot on the gas - right-mouse or driving-key is down)

	Note that whenever this set changes, self is signaled:
	I.e. each addition includes:
		self signal: #activityStarted with: newActivitySymbol
	and each removal includes:
		self signal: #activityStopped with: oldActivitySymbol.

	The actual content of the actvities set is under exclusive control of the client harness,
	and should only be altered via QAvatarUser's methods."

	^ activities asArray! !

!QAvatarReplica methodsFor: 'useful' stamp: 'bgf 3/15/2010 17:32'!
addAnnotation: klass toFrame: frame selection: selection
	"Place a sticky-note at the given attachment point on the given frame."
	| note sFrame |

	(klass includesBehavior: QAnnotation) ifFalse: [
		QLogger logWarning: 'Cannot use ', (klass name), ' as an annotation.'.
		^ nil
	].

	"Unpack the T-Selection, in case it came from the harness."
	frame ifNotNil: [
		selection frame ifNil: [
			selection frame: (frame getOwnedObject: selection frameID)
		]].
	sFrame := selection frame.
	sFrame isNil ifTrue: [ 
		self signal: #infoMessage with: 'There is nothing there to put a note on.'.
		^ nil
	].
	frame canTakeAnnotations ifFalse: [
		self future signal: #alertMessage with: 'You cannot put a note there.'.
		^ nil
	].
	(selection point isNil or: [selection normal isNil]) ifTrue: [
		self future signal: #alertMessage with: 'You cannot put a note there.'.
		^ nil
	].

	"Create the note."
	note := klass new.
	note objectName: (note genericName).
	self root addChild: note.
	note creator: self.

	"We don't specify a last-modified time here;
	 rather, we get the note editor going right away,
	 and it gives the last-modified time when it is applied."

	note attachToFrame: sFrame atPoint: selection point normal: selection normal.

	self recordAction: #attachNote:toFrame:capture:kind:
		class: (QAvatarReplica)
		arguments: { note . sFrame .  (self captureFrames: { note } ) . klass genericName }.

	note editByUser: self atCreation: true.
	^ note! !

!QAvatarReplica methodsFor: 'useful' stamp: 'bgf 3/11/2010 10:34'!
addCardToInventory: aFileCard

	"Take the argument into my inventory as a new card;
	 this will make the appropriate kind of card depending on the content."
	| newCard |

	inventory ifNil: [ ^ self ].
	(inventory hasContent: aFileCard content) ifTrue: [
		^ self signal: #infoMessage with: 'Inventory already includes ', aFileCard summary.
	].
	newCard := inventory insertDroppedCard: aFileCard fromUser: self.

	QActivityCapture action: #AddToInventory
				       details: (newCard activityDetails) 
				       forSession: (self sessionHex).
	^ newCard! !

!QAvatarReplica methodsFor: 'useful' stamp: 'bgf 1/6/2010 11:18'!
addContentToInventory: aTextureOrDoc
	"Take the argument (a shared document, or file-texture) 
	 into the inventory as a new card."

	(aTextureOrDoc class includesBehavior: TFileTexture) ifTrue: [
		| cpTxtr |
		"Defeat the careful ways the texture system protects against surplus copies."
		cpTxtr := TFileTexture new.
		TSceneReader new read: cpTxtr from: (aTextureOrDoc persistString).
		^ self addCardToInventory: (QFileCard forTexture: cpTxtr).
	].
	(aTextureOrDoc class includesBehavior: QAbstractDocument) ifTrue: [
		^ self addCardToInventory: (QFileCard forDocument: (aTextureOrDoc copy))
	].
	QLogger logWarning: 'Unknown content type'.
	^ nil
! !

!QAvatarReplica methodsFor: 'useful' stamp: 'bgf 8/31/2008 17:24'!
addNoteToFrame: frame selection: selection
	"Place a sticky-note at the given attachment point on the given frame."
	^ self addAnnotation: QStickyNote
		    toFrame: frame
		    selection: selection! !

!QAvatarReplica methodsFor: 'useful' stamp: 'bgf 2/5/2010 17:12'!
addObjectToInventory: aTFrame

	inventory ifNil: [ ^ self ].

	"File cards aren't wanted as saved objects, they're wanted as document-cards..."
	( aTFrame isKindOf: QFileCard ) ifTrue: [
		^ self addCardToInventory: aTFrame
	].

	self preCapture: { aTFrame }.
	QActivityCapture action: #AddToInventory details: 'Stored' forSession: (self sessionHex).

	"We have to allow duplicates here because the card we add to the rolodex
	 is 'empty'... we have to wait until the copy-stored-object process sends
	 back a document, and in the meantime, the card's content will be nil,
	 making it look like any other stored-object card that hasn't been filled in yet."
	inventory insertCard: (QFileCard forTFrame: aTFrame user: self container: inventory)
			   allowDuplicates: true.
! !

!QAvatarReplica methodsFor: 'materials' stamp: 'bgf 7/20/2010 10:59'!
addPaletteTextures: arrayOfTextures

	| txtrPalette |
	txtrPalette := self texturePalette asOrderedCollection.
	arrayOfTextures do: [ : txtr |
		(txtrPalette noneSatisfy: [ : t | t fileName = txtr fileName ]) ifTrue: [
			txtrPalette add: txtr.
			texturePalette := txtrPalette asArray.
			]].
	
	self signal: #texturePaletteChanged with: txtrPalette asArray
	! !

!QAvatarReplica methodsFor: 'meetings' stamp: 'bgf 8/18/2009 23:52'!
allowedToMove
	"True when I am permitted to move about  in the current meeting.
	 (False only if the leader has taken control of the audience movement.)"

	^ self meetingState ifNotNilDo: [ : meeting |
			self isMeetingLeader 
			 or: [
				meeting audienceStayPut not
				and: [ meeting isAllFollowMe not  ] ]
		] ifNil: [ 
			true
		].
! !

!QAvatarReplica methodsFor: 'meetings' stamp: 'bgf 7/3/2009 13:17'!
allowedToSpeak
	"True when I am permitted to speak in the current meeting.
	 (False only if the leader has muted the audience, and I'm not on the white-list to speak.)"

	^ self meetingState ifNotNilDo: [ : meeting |
			meeting audienceIsMuted not 
			 or: [ self isMeetingLeader ] 
			 or: [ meeting allowedToSpeak includes: (self userLogin) ]
		] ifNil: [ 
			true
		].
! !

!QAvatarReplica methodsFor: 'accessing' stamp: 'bgf 8/29/2009 11:44'!
altitude: aScalar
	"As my altitude changes, adjust where the shadow lies;
	 also adjust the orientation and length of the shadow as I lean."
	aScalar = self altitude ifTrue: [ ^ self ].
	super altitude: aScalar.
	self poseShadow: aScalar.
! !

!QAvatarReplica methodsFor: 'driving' stamp: 'bgf 2/26/2009 23:52'!
apparentPosture: sym

	"User-motion has changed my altitude, 
	 and either put me onto the ground or lifted me off the ground.
	 This only affects my posture if I'm not bound to a Place, or Sitting;
	 it allows transitions from Standing to Hovering and vice-versa.
	 (Sitting is controlled by the defined posture of a seating place.) "

	(self currentPlace isNil and: [ self posture ~= #Sitting ]) ifTrue: [
		super apparentPosture: sym
	]! !

!QAvatarReplica methodsFor: 'goto' stamp: 'bgf 8/12/2009 21:52'!
arrivedAt: placeLabel
	"We've arrived at a place.  If it's not the most recent place we've recorded, forget that."
	super arrivedAt: placeLabel.
	currentPlace ifNil: [
		"If here via transform goto, see if there is a named-place we're at."
		currentPlace := self nearbyPlace: (self currentLandmarks).
		].
	currentPlace ifNotNil: [
		currentPlace label = placeLabel  ifTrue: [
			"Arrival as expected"
			self arrivedAtPlace: currentPlace
		] ifFalse: [
			"My most recent go-to was overridden somewhere." 
			self clearCurrentPlace.
		]].

	"Arrivals also to involve focal transforms becoming active."
	self viewportChanged; doViewportUpdate.

	"If I've gone somewhere when there is a leader-assigned-place in force, snap back."
	self returnToMeetingPlace.
! !

!QAvatarReplica methodsFor: 'goto' stamp: 'bgf 2/16/2010 02:45'!
arrivedAtPlace: curPlace
	"Completion of a goToPlace.
	 If the place we've arrived it is a destination with seating behavior,
	 find a place to sit.   (We didn't find a seat at embarkation time,
	 but someone may have moved while I was in transit.) "
	| seat |

	(curPlace posture = #Sitting) ifFalse: [
		seat := self findSeatIn: curPlace.
		seat ifNotNil: [
			 [ self goToPlace: seat andFollow: false ] ensure: [ self setViaPlace: curPlace ].
			^ self.
		]].

	"How to preserve the currentPlace binding while sitting in a different seat
	 in the other guy's viaPlace ?"

	self checkArrivalFocus: curPlace.
	self posture: (currentPlace posture ifNil: [ #Standing ]).


! !

!QAvatarReplica methodsFor: 'accessing' stamp: 'bgf 3/20/2008 03:11'!
arrowCursor

" my own arrow cursor"

	^ arrowCursor ifNil:[ 
		arrowCursor := QArrowCursor new. 
		arrowCursor color: self avatarColor. 
		arrowCursor userID: self sessionID.
		arrowCursor
	].! !

!QAvatarReplica methodsFor: 'places' stamp: 'bgf 4/5/2008 03:18'!
asPlace

	"When an avatar is to be treated as a place by the default driving short-cuts,
	 the join-me place is used."

	^ self placeForJoiningMe
! !

!QAvatarReplica methodsFor: 'gestures' stamp: 'bgf 8/16/2009 17:07'!
attentionGesture
	"Primal gesture... call attention to yourself.  Default is a quick hop.
	 Bound to the space-bar in driving mode."
	| xform |

	self wantsToSpeak ifFalse: [
		"These are automatic when turning on wants to speak,
	  	 and the 'Hi' is inappropriate in that case."
		self root signal: #infoMessage with: (self publicName, ' says Hi!!').
	] ifTrue: [
		"These are automatic when turning on wants to speak,
	  	 and the 'Hi' is inappropriate in that case."
		self root signal: #infoMessage with: (self publicName, ' wants to speak.').
	].

	(costume respondsTo: #attentionGesture) ifTrue: [
		costume attentionGesture ifNotNil: [^self].
	] .
	"Default implementation: just hop up a bit; fall will take you back down.
	 Fall doesn't work while sitting, so don't jump in your seat."
	self posture = #Sitting ifFalse: [
		xform := self localTransform.
		xform translation: (xform translation + (0@1.5@0)).
		self goto: xform count: 2 sticky: false.
	].
! !

!QAvatarReplica methodsFor: 'accessing' stamp: 'jcg 2/22/2010 17:57'!
audioStreamID
	^speaker ifNotNil: [:spkr | spkr streamID]! !

!QAvatarReplica methodsFor: 'accessing' stamp: 'bgf 4/11/2007 15:56'!
avatarColor
	"Answer either the avatar's preferred color, or if none, the color of its laser;
	 may return nil if there's no laser yet."
	| col |

	userInfo ifNotNil: [
		col := userInfo avatarColor.
		col ifNotNil: [^col].
	].
	laser ifNotNil: [
		col := laser material ambientColor copy.
		col alpha: 1.0.
	].
	^ col
! !

!QAvatarReplica methodsFor: 'accessing' stamp: 'das 3/7/2007 14:32'!
avatarHeight

	^ avatarHeight.! !

!QAvatarReplica methodsFor: 'accessing' stamp: 'bgf 7/25/2009 03:52'!
avatarHeight: aHeight

	"For tools which modify the avatar.
	 Supply the overall vertical extent of the avatar model."

	avatarHeight := aHeight.
	label ifNotNil: [ label update ].! !

!QAvatarReplica methodsFor: 'places' stamp: 'bgf 7/3/2009 18:51'!
avatarIAmWith
	"If I am traveling with another avatar, return it."
	^ self nominalPlace ifNotNilDo: [ :place | place avatarToFollow ]
	! !

!QAvatarReplica methodsFor: 'accessing' stamp: 'bgf 7/25/2009 15:21'!
avatarIconDistance

	"Maximum distance at which to render the 3D avatar;
       beyond this distance, the avatar is replaced by a 2D poster."
	^ (avatarHeight max: 2) * 48.
! !

!QAvatarReplica methodsFor: 'driving' stamp: 'bgf 8/3/2009 13:18'!
avatarStepInterval
	"Should mirror or approximate QAvatarUser minStepInterval"
	"The smallest allowed interval for my avatar to send pulse messages.
	  We adjust upward from croquet basics when meetings get crowded,
	  and especially when a led meeting has movement restricted."

	self root ifNil: [ ^ 500 ].  "Don't care."
	^ QAvatarUser 
		currentStepIntervalForUsers: (self root numUsersPresent)
		favored: (self isMeetingLeader or: [ self hasPermissionToSpeak ]) 
		moving: (self allowedToMove)
! !

!QAvatarReplica methodsFor: 'accessing' stamp: 'bgf 7/25/2009 16:00'!
avatarTop
	"How far above the eye-level origin does the label appear?"
	^ (self avatarHeight max: 5.5) *0.2! !

!QAvatarReplica methodsFor: 'accessing' stamp: 'bgf 3/29/2007 15:47'!
badge
	"If the costume supports badging, return the current badge (a texture)."
	(costume contents respondsTo: #badge) ifTrue: [
		^ costume contents badge
	] ifFalse: [
		^ nil
	].! !

!QAvatarReplica methodsFor: 'accessing' stamp: 'bgf 7/24/2009 23:33'!
badge: aTextureOrNil
	"If the costume supports badging, wear this one."
	(costume respondsTo: #badge:) ifTrue: [
		costume badge: aTextureOrNil.
	] ifFalse: [
	].
	label ifNotNil: [ label update ].! !

!QAvatarReplica methodsFor: 'accessing' stamp: 'bgf 4/3/2007 22:40'!
badgePicture
	"Convenience accessor for userInfo #badgePicture"
	^ userInfo badgePicture! !

!QAvatarReplica methodsFor: 'places' stamp: 'bgf 8/9/2009 14:52'!
besideMeCameraStanceFor: anAvatar
	"Camera-stance for an avatar who is following me.
	 (Allows the follow-me place to channel the cam-xform of a viewing place to followers.)"
	| root |

	"To do - handle having several avatars tied to me at once."

	anAvatar = self ifTrue: [ ^ nil ].
	root := self root.
	root ifNil: [ ^ nil ].	"Don't keep following if I'm leaving.."

	self focalPlace ifNotNil: [ : fp | 
		^ fp cameraStanceFor: anAvatar
	].

	^ nil! !

!QAvatarReplica methodsFor: 'places' stamp: 'bgf 8/9/2009 17:30'!
besideMeStanceFor: anAvatar
	"Stance at which to place an avatar standing beside me."
	| pos root coi |

	"To do - handle having several avatars tied to me at once."

	anAvatar = self ifTrue: [ ^ nil ].
	root := self root.
	root ifNil: [ ^ nil ].	"Don't keep following if I'm leaving.."

	"If I'm seated within a gathering place, send the follower to their own stance in the gp."
	self viaPlace ifNotNil: [ : vp |
		^ vp destinationStanceFor: anAvatar
	].

	pos :=  (self root freePlaceWith: self for: anAvatar).
	root miramarUI ifTrue:[
		coi := self centerOfInterest.
	] ifFalse:[
		coi := (self translation) - (self lookAt * 12).			"I.e. looking ahead of me (not him)"
	].

	^Stance here: pos  there: coi! !

!QAvatarReplica methodsFor: 'accessing' stamp: 'jcg 1/30/2008 18:07'!
birthTime
	^birthTime! !

!QAvatarReplica methodsFor: 'menus' stamp: 'bgf 8/12/2009 20:19'!
bringEveryone: aPlace includingMe: bIncluding
	"Just a raw reminder - this is how to force folks into a gathering."
	self root allUsers do: [ : avatar | 
		(bIncluding or: [ avatar ~= self ]) ifTrue: [
			(avatar isAtPlace: aPlace) ifFalse: [ 
				avatar 
					signal: #alertMessage 
					with: 'Sent to ', (aPlace label ifNil: [ 'gathering']), ' by ', (self publicName).
				avatar goToPlace: aPlace ]
		]].! !

!QAvatarReplica methodsFor: 'accessing' stamp: 'crl 4/22/2010 15:50'!
camera
	^camera! !

!QAvatarReplica methodsFor: 'accessing' stamp: 'crl 4/15/2010 15:21'!
cameraEnabled
	^camera isCapturing! !

!QAvatarReplica methodsFor: 'places' stamp: 'bgf 4/25/2009 12:50'!
cameraStanceForPlace: aPlace viewExtent: ext
	"Return the camera viewing Stance I would assume at the given place,
	 for a view-extent as given.   Harness get:-safe, and must stay that way. "
	| save |
	save := viewExtent.
	[
		viewExtent := ext.
		^ (aPlace cameraStanceFor: self)
	] ensure: [
		viewExtent := save.
	].! !

!QAvatarReplica methodsFor: 'testing' stamp: 'bgf 1/30/2010 00:28'!
canDeleteFrame: frame
	"Is direct delete applicable to this frame?"

	^ (super canDeleteFrame: frame)
	    and: [ self canModifyForum ]! !

!QAvatarReplica methodsFor: 'useful' stamp: 'bgf 2/19/2009 15:51'!
canEditMaterial: aFrame
	"Relay so that the client can preflight-test material editing availability."
	^ aFrame notNil and: [ aFrame component canEditMaterial ]! !

!QAvatarReplica methodsFor: 'useful' stamp: 'bgf 7/3/2009 13:17'!
canModifyForum
	^ (self hasCapability: #modify) 
		and: [ self isMeetingLeader or: [ self meetingState isEditBlocked not ] ]! !

!QAvatarReplica methodsFor: 'useful' stamp: 'bgf 9/1/2009 12:51'!
canUseInventory
	"There are cases where we block Copy-to-Inventory specifially
	 because the receipt of inventory additions requires inventory to be opened,
	 but meetings with editblocked should not allow inventory to appear."

	^ self isMeetingLeader or: [ self meetingState isEditBlocked not ] ! !

!QAvatarReplica methodsFor: 'useful' stamp: 'bgf 6/10/2009 11:40'!
capabilities: capArray features: featureArray
	"Periodic assertion from my client of my enabled features,
	 to guide menu and command access."
	userInfo ifNotNil: [
		userInfo capabilities: capArray;
			      features: featureArray.
	].! !

!QAvatarReplica methodsFor: 'editing' stamp: 'bgf 3/15/2010 08:30'!
captureCopiedFrames: listOfFrames
	"Cut/Copy/Paste support.
	 Capture the given list of tframes as my pasteboard-scrap for subsequent paste."

	listOfFrames isEmptyOrNil ifTrue: [ ^ self ].
	self setCopiedFramesScrap: (self captureFrames: listOfFrames).! !

!QAvatarReplica methodsFor: 'editing' stamp: 'bgf 3/13/2010 08:51'!
captureCreateFrames: listOfFrames description: verb

	| cap |
	listOfFrames isEmptyOrNil ifTrue: [ ^ self ].
	cap  := self captureFrames: listOfFrames.
	self recordAction: #undoForCreation:description:
		class: QCapturedFrameList
		arguments: { cap . verb }.! !

!QAvatarReplica methodsFor: 'editing' stamp: 'bgf 3/13/2010 08:52'!
captureDeletedFrames: listOfFrames
	"Delete-undo support.
	 Capture the given list of tframes as my deleted-frames scrap for subsequent undo."

	listOfFrames isEmptyOrNil ifTrue: [ ^ self ].

	self preCapture: listOfFrames.		"Ensure live documents are up-to-date with current edits."
	"Assign the given capturedFrameList to be the current delete-scrap."
	deletionScrap := self captureFrames: (listOfFrames).

	self recordAction: #undoForDeletion:
		class: QCapturedFrameList
		arguments: { deletionScrap }.
! !

!QAvatarReplica methodsFor: 'editing' stamp: 'bgf 3/12/2010 14:20'!
captureFrames: listOfFrames

	self preCapture: listOfFrames.		"Ensure live documents are up-to-date with current edits."

	^ (QCapturedFrameList new) 
			capture: listOfFrames forAvatar: self;
			yourself 
! !

!QAvatarReplica methodsFor: 'goto' stamp: 'bgf 11/18/2009 09:07'!
checkArrivalFocus: curPlace
	"If we've arrived at a place whose viaPlace has a camera focus,
	 adopt the camera focus for the viaPlace.
	 This happens after we occupy the task seat for a display."

	self signal: #placeArrival with: (self nominalPlace).
	self focalPlace ifNotNilDo: [ : focal  |
		(focal cameraStanceFor: self) ifNotNil: [ : stance | | pov |
			self focalCameraTransform: stance transform.

			"We allow the via-place (e.g. the gathering around seats) to suggest the pov,
	 		 so things like theatre-seating can start from a wider-angle view.
			 If the via place suggests 'Unspecified' (nil), then we use the focal suggestion."

			pov := self viaPlace ifNotNilDo: [ : via | via suggestedPOV ].
			pov ifNil: [ pov :=  focal suggestedPOV ].
			pov ifNotNil: [
				self suggestPOV: ( pov ) forPlace: focal.
			].
			^ true
		]].
	^ false! !

!QAvatarReplica methodsFor: 'places' stamp: 'bgf 8/19/2009 02:29'!
checkRestingPlace

	"Part of routine heartbeat.
	 If we've rested in one spot for a couple of seconds, 
       declare that we're in this stopping place,
       so the Places tool can remember this as my previous place
	 when I move on from here."

	| stance place |

	self nominalPlace ifNotNil: [ ^ self ].		"Already know where I am."
	lastMovedTime ifNil: [ ^ self ].		"Haven't moved since last placed."
	((self island totalSeconds - lastMovedTime) > 2)	
		ifFalse: [ ^ self ].	"Not at rest."
	"Note - stillAlive: happens infrequently when stationary, so the 1-second check
	 effectively makes for a rest-place after 2-3 seconds of not moving."

	"OK, we've come to rest after leaving the last position.  Where am I?"
	stance := Stance here: (self globalPosition)
					  there: (self globalPosition - self lookAt).

	"The place I assign is meant to match what is indicated in the places tool.
	 If I find a census-match place corresponding to where I am, use it; 
	 otherwise make one up.   If the found place would take my camera, don't use it,
	 since that is unexpected and mistimed warping."
	place := self nearbyPlace: (self currentLandmarks).
	(place isNil or: [ place hasCameraStance ]) ifTrue: [
		place := QPlace category: 'Unlisted' 
						label: (self posture)
						stance: stance.
	].

	currentPlace := place.
	self signal: #placeChanged with: place.! !

!QAvatarReplica methodsFor: 'goto' stamp: 'bgf 2/27/2009 00:46'!
clearCurrentPlace

	"Forget my current place assignment - I have moved or unstickied myself.."
	self currentPlace: nil stance: nil andFollow: false.
! !

!QAvatarReplica methodsFor: 'editing' stamp: 'bgf 1/25/2010 12:07'!
clearEditBox
	"If I own a current edit box, dismiss it."
	(editBox notNil and: [ editBox contents notNil ] and: [ editBox root = self root ]) ifTrue: [
		editBox dismiss.
		editBox := nil.
	].
	! !

!QAvatarReplica methodsFor: 'events' stamp: 'bgf 11/7/2009 18:24'!
clearFloatMessageForEvent: event

	"Clear a mouse-over floating-tag label ...
	 unless the event is beyond the warp-to distance for the event target,
	 which suppresses these per-object tags."

	(self warpPlaceUnderMouse: event) ifNil: [ 
		self floatingInfo: nil
	].
! !

!QAvatarReplica methodsFor: 'accessing' stamp: 'bgf 7/8/2009 15:11'!
clientCameraTransform: xformOrNil
	"Cam transform changed by the harness - i.e. by look-mode driving.
	 We do -not- signal these changes, as it is the harness that watches the signals."
	currentFocalTransform = xformOrNil ifFalse: [
		currentFocalTransform := xformOrNil.
		self viewportChanged.
	]
! !

!QAvatarReplica methodsFor: 'initialize' stamp: 'bgf 5/24/2010 10:49'!
completeEntry: clientUTCSeconds
	"Called from the harness at the end of avatar introduction into a new space."
	| pyApps myLogin |
	super completeEntry: clientUTCSeconds.
	userInfo ifNotNil: [
		self root ifNotNilDo: [ :space | 	
			space textMessage: ('', (self publicName ifNil: ['A user']), ' arrived')
				    fromAvatar: nil 		"This is a host-message about me, not from me."
				    atTime: clientUTCSeconds.
			pyApps := self root find:[:frm | frm class = PyPanelApp or:[frm class inheritsFrom: PyPanelApp]].
			myLogin := userInfo userLogin.
			pyApps do:[:pa | 
				pa isRunning ifFalse:[
					pa getStartupOwner = myLogin ifTrue:[
						pa startRunning: self userID.
					].
				].
			].
		].
	].

	"If there's a meeting going on, go where it wants me to go. (I.e. sit in the audience, etc.)"
	self meetingState ifNotNilDo: [ :meeting | 
		meeting meetingPlace ifNotNilDo: [ : plc | 
			self signal: #alertMessage 
				with:  'Joining meeting at ', (plc label ifNil: [ 'assigned landmark' ]).
			self goToPlace: plc ] ].


	TObjectID for: 'Entry-Complete ', (self userLogin).! !

!QAvatarReplica methodsFor: 'menus' stamp: 'bgf 11/17/2009 15:57'!
confirm: shortMsgString yesLabel: yesLabel selector: yes arguments: args noLabel: noLabel  target: targetFrame

	"See the super method.  Our implementation uses a job to run a confirm dialog."
	QConfirmJob
		confirm: shortMsgString 
		yesLabel: yesLabel 
		selector: yes 
		arguments: args 
		noLabel: noLabel  
		target: targetFrame 
		forAvatar: self
! !

!QAvatarReplica methodsFor: 'editing' stamp: 'bgf 2/1/2010 12:09'!
confirmDeleteSelection: evt
	"Get permission to delete the frames I have currently selected.
	 This is a bridging method until undo/cut/copy/paste are implemented."
	
	self canModifyForum ifFalse: [  ^ self ].

	selectedFrames isEmptyOrNil ifTrue: [
		^ self signal: #alertMessage with: 'Delete - No items selected.'
	].
	selectedFrames do: [ : f |
		(self canDeleteFrame: f) ifFalse: [
			f hiliteBriefly.
			^ self signal: #alertMessage with:  'Cannot delete this ', (f genericName).
		]].

	selectedFrames size > 1 ifTrue: [ 
		self confirm: 'Delete ', (selectedFrames size), ' Items?'
			selector: #deleteSelection:
			target: self.
	] ifFalse: [ 
		| frame |
		frame := selectedFrames anyOne.
		self confirm: 'Delete ', (frame deletionLabel), '?'
			selector: #deleteSelection:
			target: self.
	].
! !

!QAvatarReplica methodsFor: 'menus' stamp: 'bgf 5/6/2010 10:07'!
contextMenuEvent: event forFrame: frame
	"A right-button or other event to trigger a context menu post."
	frame ifNil: [ ^ self ].
	event avatar: self.
	"Clicks on other user's avatars go back to the harness for specific handling."
	frame isInAvatar ifTrue: [
		(event avatar = frame avatar) ifFalse: [
			^ self root signal: #avatarClicked with: (frame avatar) with: self.
		]].
	frame resolveEventFrame: event.

	"If we've menu-clicked within  an already-selected TGroup,
	 or within any part of an existing multiple-select,
	 interpret that as operating on what we've already selected.
	 We don't do this for custom group subclasses (e.g. shelves, etc) since 
	 in those cases, the innate rules about what menu appear for content should prevail.
	 This case is to allow access to the containing-group menu just after doing 'Group'  "
	(self selectionIncludesAncestorOf: frame) ifTrue: [
		selectedFrames size > 1 ifTrue: [
			^ self groupContextMenuForFrames: (selectedFrames shallowCopy) event: event
		] ifFalse: [
			| sel |
			sel := selectedFrames first.
			(sel class = TGroup) ifTrue: [
				^ sel showContextMenu: event
			].
		]].

	"A context menu right-mouse click induces selecting that item."
	(self doSelectForClick: event onFrame: frame) ifTrue: [ 
		frame component ifNotNil: [ : c |
			(c fromTemplate not or: [ self isEditingTemplate ]) ifTrue: [
				(self frameSelectionIncludes: c) ifFalse: [
					self selectFrame: c 
				]]].
	].

	"If operating on a group of selected items, divert to the many-item menu."
	((self frameSelectionIncludes: frame component) and: [ selectedFrames size > 1 ]) 
	 ifTrue: [
		^ self groupContextMenuForFrames: (selectedFrames shallowCopy) event: event
	].

	^ frame showContextMenu: event
! !

!QAvatarReplica methodsFor: 'editing' stamp: 'bgf 2/9/2010 23:10'!
copySelection: evt
	"The basic 'Copy' ctl-c operation... whatever frames are selected, 
	 are made the content of the current pasteboard."
	evt avatar: self.
	selectedFrames isEmptyOrNil ifTrue: [ ^ self ].

	selectedFrames size = 1 ifTrue: [
		self signal: #infoMessage with: 'Copied ', (selectedFrames anyOne genericName)
	] ifFalse: [
		self signal: #infoMessage with: 'Copied ', (selectedFrames size), ' items.'
	].
	self captureCopiedFrames: selectedFrames.
! !

!QAvatarReplica methodsFor: 'accessing' stamp: 'bgf 8/23/2009 15:12'!
costume: cost

	super costume: cost.

	shadow extent:3@2; translation: (0@-4.6@0).
	costume addChild: shadow.
	costume addChild: label.
	costume addChild: standin.
	"Give the costume opportunity to adjust replica paramters."
	(costume respondsTo: #configureAvatar:) ifTrue: [
		costume configureAvatar:self.
	].
	"Capture the initial 'standing-on-ground' shadow configuration."
	shadowTranslation := shadow translation copy.	
	shadowExtent := shadow extent copy.

	"Vertical extent of the costume, used to determine the cutoff 
       for rendering a poster-graphic instead of the distant avatar."
	self avatarHeight:( costume ifNotNil:[ costume boundingBox extentY ] ifNil:[5.5]).

	"Initialize from userInfo"
	self onProfileChanged.
	self onAvatarColorChanged.
	self onSmallPictureChanged.
	self onBadgePictureChanged.

	"Let all the geometry settle..."
	label ifNotNil: [ label future update ].! !

!QAvatarReplica methodsFor: 'accessing' stamp: 'bgf 11/21/2007 15:38'!
costumeData

	"The Zipped XML data describing my current costume."
	^ QUserPreferences 
			xmlToZip:(TSceneWriter new storeScene:
							((costume isKindOf: (TAvatarProxy)) 
								ifTrue: [ costume contents ] 
								ifFalse: [costume]))
			unitName: 'avatar'.! !

!QAvatarReplica methodsFor: 'render' stamp: 'bgf 2/18/2008 12:58'!
costumeVisible: ogl
	
	"We substitute the billboard icon beyond #avatarIconDistance,
	 so we suppress the costume in that case."

	^ (self cameraDistanceSquared:ogl asPartOf: self) <  (self avatarIconDistance squared)! !

!QAvatarReplica methodsFor: 'places' stamp: 'bgf 7/4/2009 15:02'!
currentLandmarks
	"Returns an array of QPlaces, 
	 representing the defined landmarks in the current space
 	 that are available to this avatar."

	| lms |
	(self root respondsTo:#landmarksFor:)  ifTrue: [
		lms := (self root landmarksFor: self).
	].
	(lms isNil or: [ lms isEmpty ]) ifTrue: [
		lms := {self homePlace}.
	].

	"If there's a leader-assigned place, list it for the UI.
	 (Some places may not be landmarks in themselves.)" 
	self meetingPlace ifNotNilDo: [ : lap |
		lap avatarToFollow ifNil: [	"This means 'follow-me'; no need for listing it."
			(lms includes: lap) ifFalse: [ 
				lms := lms copyWith: lap.
			]]].
	^lms
! !

!QAvatarReplica methodsFor: 'places' stamp: 'bgf 7/4/2009 16:38'!
currentLandmarksWithCensus
	"Returns the currentLandmarks array, 
	 and a parallel array of ordinal integers representing their current occupancy counts."
	| lms  places place |

	self root ifNil: [ ^ #() ].
	lms := self currentLandmarks.
	lms ifNil: [ ^ #() ].
	places := Dictionary new.
	lms do: [  : lm |
		lm occupants: 0.
		places at: lm put: lm.	
			"So we can use the dictionary lookup rather than linear detect: 
			 to match avatar currentPlace to the same landmark in lms;
			 this is just a weak fight against the n*m order of the basic question."
	].
	self root allUsers do: [ : user |
		| listedPlace |
		place := user currentPlace.

		"The current place result will be the avatar's, find its match in our data list..."
		listedPlace := places 
						at: place 
						ifAbsent: [ places detect: [ : lp | user isAtPlace: lp ] ifNone: [nil] ].
		listedPlace ifNotNil: [
			listedPlace occupants: (listedPlace  occupants + 1 + (user followers size))
		]].
	^ lms! !

!QAvatarReplica methodsFor: 'accessing' stamp: 'bgf 3/15/2008 13:21'!
currentPlace
	"If I currently am positioned as defined by a QPlace, returns it. Otherwise nil."
	^ currentPlace! !

!QAvatarReplica methodsFor: 'goto' stamp: 'bgf 12/23/2009 09:15'!
currentPlace: aQPlace stance: stance andFollow: bFollow
	"Make the given place my current place, and go there.
	 If bFollow is true, arrange keep going to the place if it moves,
	 until further notice (until my current place is changed again.)
	 Signals #placeChanged at any change, 
	 including when the place is cleared to nil."

	currentPlace ~= aQPlace ifTrue: [
		self stopScript: #doFollowPlace:.
		followPlaceTime := nil.
		currentPlace := aQPlace.
		currentPlaceStance := stance.
		self signal: #placeChanged with: currentPlace.
		(bFollow and: [currentPlace notNil]) ifTrue: [
			"A little magic - if I go to a moving destination, I follow it..."
			self startFollowingPlace: currentPlace.
		]
	] ifFalse: [
		(currentPlaceStance = stance) ifFalse: [
			currentPlaceStance := stance.
			self signal: #placeChanged with: currentPlace
		].
	].
	currentPlaceVia := nil.
	currentPlace ifNotNil: [ lastMovedTime := nil ].
	self updateFocalTransform.

	besideMePlace ifNotNil: [
		besideMePlace 
			suggestedPOV: (self currentPlacePOV)
	].
	self refreshLeaderPlace.	"Bring the crew, if I'm leading."

	"The cleaner setup does a stopscript that might stop -me- so it needs to be async."
	self future setCurrentPlaceCleanup: (self destinationFrame).

	^ currentPlace 
! !

!QAvatarReplica methodsFor: 'accessing' stamp: 'bgf 9/22/2008 12:25'!
currentPlaceFrame
	"If I currently am positioned as defined by a QPlace, 
	 return the destination frame; else nil"

	^ currentPlace ifNotNil: [ currentPlace destinationFrame ]! !

!QAvatarReplica methodsFor: 'places' stamp: 'bgf 8/25/2009 00:33'!
currentPlaceOrStance
	"Returns a QPlace representing my current position, never nil.
	 This may be a logical place such as a destination,
	 or it may be a made-up record of where I happen to be standing right now.
	 CF currentPlace, which returns only a named place or nil."
	| stance |

	^ self currentPlace
	   ifNil: [
			stance := Stance here: (self globalPositionReal)
					  		 there: (self globalPositionReal - self lookAtReal).
			QPlace category: 'Unlisted' 
					label: (self posture)
					stance: stance.
	].! !

!QAvatarReplica methodsFor: 'goto' stamp: 'bgf 3/19/2010 19:07'!
currentPlacePOV
	"The viewpoint (viewpointTable entry string)
	 defined by the Place I currently occupy, if any.
	 See also 'checkArrivalFocus' "
	| pov |
	pov := self viaPlace ifNotNilDo: [ : via | via suggestedPOV ].
	pov ifNil: [ 
		pov :=  currentPlace ifNotNil: [ : cp | cp suggestedPOV ]
	].
	^ pov! !

!QAvatarReplica methodsFor: 'editing' stamp: 'bgf 3/13/2010 08:52'!
cutSelection: evt
	"The basic 'Cut'  ctl-x operation for scene-editing... 
	 whatever frames are selected,  are made the content of my pasteboard."

	evt avatar: self.
	selectedFrames isEmptyOrNil ifTrue: [
		^ self signal: #alertMessage with: 'Nothing is selected to cut.'
	].

	selectedFrames do: [ : f |
		(self canDeleteFrame: f) ifFalse: [
			f hiliteBriefly.
			^ self signal: #alertMessage with:  'Cannot remove this ', (f genericName).
		]].

	self captureCopiedFrames: selectedFrames.
	deletionScrap := pasteboardScrap.
	self recordAction: #undoForCut:
		class: QCapturedFrameList
		arguments: { deletionScrap }.

	selectedFrames size = 1 ifTrue: [
		self signal: #infoMessage with: 'Cut ', (selectedFrames anyOne genericName)
	] ifFalse: [
		self signal: #infoMessage with: 'Cut ', (selectedFrames size), ' items.'
	].

	selectedFrames shallowCopy do: [ : f | f reallyDelete: evt ].
	self clearFrameSelection.! !

!QAvatarReplica methodsFor: 'editing' stamp: 'bgf 2/7/2010 08:12'!
deleteFrames: listOfFrames event: evtOrAvatar
	"Common method for carrying out Delete of frames;
	 capture the frame state into the -delete- pasteboard to allow undo."

	self captureDeletedFrames: listOfFrames.
	listOfFrames size = 1 ifTrue: [
		self signal: #infoMessage with: 'Deleted ', (listOfFrames anyOne genericName)
	] ifFalse: [
		self signal: #infoMessage with: 'Deleted ', (listOfFrames size), ' items.'
	].
	super deleteFrames: (listOfFrames) event: evtOrAvatar! !

!QAvatarReplica methodsFor: 'editing' stamp: 'bgf 2/9/2010 23:25'!
deleteSelection: evtOrAvatar
	"Delete the frames I have currently selected."

	self canModifyForum ifFalse: [  ^ self ].

	selectedFrames isEmptyOrNil ifTrue: [
		^ self signal: #alertMessage with: 'Delete - No items selected.'
	].
	selectedFrames do: [ : f |
		(self canDeleteFrame: f) ifFalse: [
			f hiliteBriefly.
			^ self signal: #alertMessage with:  'Cannot delete this ', (f genericName).
		]].
	
	self captureDeletedFrames: selectedFrames.

	selectedFrames size = 1 ifTrue: [
		self signal: #infoMessage with: 'Deleted ', (selectedFrames anyOne genericName)
	] ifFalse: [
		self signal: #infoMessage with: 'Deleted ', (selectedFrames size), ' items.'
	].
	selectedFrames copy do: [ : f  | f reallyDelete: evtOrAvatar ].
! !

!QAvatarReplica methodsFor: 'goto' stamp: 'bgf 8/9/2009 15:02'!
destinationFrame
	"If I am at (or on my way to) a formal destination, return the its defining Frame"
	self nominalPlace ifNotNilDo: [ : place  |
		place isStatic ifFalse: [
			^ place destinationFrame
		]].
	^ nil
! !

!QAvatarReplica methodsFor: 'goto' stamp: 'bgf 8/9/2008 22:53'!
destinationPosition
	"If I am at a formal Place, or on my way to one,
	  return the (imminent) position it told me to go to
	 (which I may not have arrived at yet);
	 otherwise, my current ordinary global position."
	^ currentPlaceStance 
			ifNil: [ self globalTransformReal translation ]
			ifNotNil: [ currentPlaceStance here ]! !

!QAvatarReplica methodsFor: 'initialize' stamp: 'bgf 5/28/2010 10:49'!
destroyAtClientTime: utcSeconds

	"Destroy the replica, logging the event.
	 Note that TAvatarReplica redirects 'destroy' to call this."

	self clearCurrentPlace.
	self clearEditBox.

	userInfo notNil ifTrue: [	"User info is nil if I've already been destroyed..."
		heartbeatState == #superseded ifFalse: [
			self root ifNotNilDo: [ :space|
				space meetingState releaseLeader: (self userLogin).
				space 
					textMessage: ('', (self publicName ifNil: ['A user']), ' left') 
					fromAvatar: nil 	"Sys message, not user-attributed chat"
					atTime:  utcSeconds   
				]].
		self registerVideoStream: nil.
	].

	"Clean up our accessories (inventory, material editor) immediately... 
	 there is special handling that overrides the inherited behavior."
	self carryingDrop.

	self isRecordingSession: false. "inform all users"
	self isBroadcastingSession: false.

	super destroyAtClientTime: utcSeconds.	"This initiates two-phase destroy."

	"There are client-side listeners hooked up to the speaker;
	 it is not a visible item, and not pickable, so can be destroyed immediately."
	speaker ifNotNil: [ speaker obliterateLater. speaker := nil ].
	self stopAllScripts.

	"Clear ivars to aid GC, and so we can spot deleted replicas while debugging."
	speaker := nil.
	arrowCursor := nil.
! !

!QAvatarReplica methodsFor: 'menus' stamp: 'bgf 1/13/2010 16:52'!
doContextMenuFor: tframe
	"Test: should I offer context menu (e.g. on right-click) for that frame right now?
	  Depends on the thing itself, and whether or not we're globally doing the editing thing.
	  Probably could be simply dCM or: editingTemplate, but that it harder to see what dCM is doing."
	^ tframe doContextMenu or: [ tframe fromTemplate and: [self isEditingTemplate ] ]! !

!QAvatarReplica methodsFor: 'private' stamp: 'bgf 7/4/2009 11:22'!
doFollowPlace: aPlace
	"<on: frameChanged in: currentPlace destinationFrame>"
	"When the frame defining my current place moves, I follow it."

	"First check for destruction."
	(aPlace destinationFrame isNil or: [ aPlace destinationFrame root isNil ]) ifTrue: [
		self stopScript: #doFollowPlace:.
		followPlaceTime := nil.
		^ self.	"Documentary... the  stopScript: has terminated me."
	].
	
	"To space out the 'catch-up' gotos a bit to reduce load,
	 we issue the catch-up at 300 ms frequency max
	 (much less than the rate of #frameChange from an avatar we're following.)
	 This is not unreasonable since the following process uses my driver via goto,
	 so I'll let the leader get further and then catch-up more efficiently.

	 We use 'followPlaceTime' saved as an ivar, and passed to the future-send,
	 so that the future send can compare to see if it is the 'most-recent' request
	 (so calls that fire after we've changed the following-configuration can detect
	 that they are unwanted, and simply die quietly.)"

	followPlaceTime ifNil: [
		followPlaceTime := self island time.		"A serial-number to recognize the send."
		(self future: 300) followPlace: aPlace forRequestTime: followPlaceTime.
	].
! !

!QAvatarReplica methodsFor: 'useful' stamp: 'bgf 12/8/2008 07:00'!
doLocalSave: content

	"The given content (a texture, or document) should be saved locally
	 on my user's system."

	(content respondsTo: #doSaveAsForAvatar:) ifTrue: [
		^ content doSaveAsForAvatar: self
	].

	self signal: #alertMessage with: 'Cannot save that content.'

	! !

!QAvatarReplica methodsFor: 'editing' stamp: 'bgf 6/9/2010 15:43'!
doPasteIntoParent: someFrame withDocumentRenames: pairs
	" Followup callback from the harness-side handler for pasteFrames.
	  Resources have been copied.  Put the actual frames into the space."
	pasteboardScrap ifNotNil: [
		pasteboardScrap pasteForAvatar: self intoParent: someFrame withDocumentRenames: pairs.
		self future positionCarriedCards.
	]! !

!QAvatarReplica methodsFor: 'places' stamp: 'jcg 9/16/2008 16:16'!
doPlaceMenuAction: placeMenuActionSelector forPlace: aPlace

	"Trampoline by which the Places context menu invokes
	 a receiver-specific action on the destination-frame of a Place.
	 From the harness, it is not good practice to resolve place destination-Frames;
	 so have the avatar do it for you...."

	^ ([aPlace destinationFrame] on: Error do: [ nil ])
		ifNotNilDo: [ : df | 		
			(df respondsTo: placeMenuActionSelector) ifTrue: [
				df perform: placeMenuActionSelector with: self.
			] ifFalse: [
				QLogger logWarning: 'Unknown selector in place-menu action ', placeMenuActionSelector
			]
		] ifNil: [ 
			QLogger logWarning: 'Dangling place-frame in place-menu action ', placeMenuActionSelector
		].
! !

!QAvatarReplica methodsFor: 'private' stamp: 'bgf 7/14/2010 12:37'!
doViewportUpdate

	"The thirdPerson viewpoint (or the app window size) has changed.
 	 If it's been long enough since the value last changed,
	 adjust  my screenside accessories."

	| timeSinceUpdate |
	lastViewportChange ifNotNil: [
		timeSinceUpdate := (self island millisecondClockValue - lastViewportChange) max: 0.
		timeSinceUpdate < 180 ifTrue: [
			"Too soon (a further change has happened since I was scheduled)"
			(self future: 180-timeSinceUpdate) doViewportUpdate.
		] ifFalse: [
			"OK, we've waited... pull the accessories back to the edge."

			self positionCarriedCards.
			lastViewportChange := nil.	"So a further change will restart me."
		].
	].
	"The problem being worked around here,
       is T-Time future pace versus real-time user-scrolling (or scroll-wheeling).
       Basic avatar driving, for example, has a lot of support for careful interpolation,
 	 but none of that helps us track to the frame-edge for camera-position jumps.
	 Instead of trying to keep up with ongoing camera zoom actions,
 	 we avoid this jitter using the opposite technique: 
	 we wait until the action pauses and then snap into place.

       Not ideal, but far less irksome than stuttery almost-keeping up behavior
 		 which we'd get by default if we rely on harness->space future-sends
 	 to tell us the new camera angle."
! !

!QAvatarReplica methodsFor: 'testing' stamp: 'bgf 11/12/2009 12:55'!
doWarpMoves
	"Policy - is this avatar responsive to direct-click warps (go-to shortcuts like clicking a chair)
	  and long-distance warp-to-me shortcuts  (the finger-cursor over faraway objects) ?"
	^ self allowedToMove and:  [ self isEditingTemplate not ]! !

!QAvatarReplica methodsFor: 'events' stamp: 'bgf 9/1/2008 04:12'!
doorChange: door

	"A door has opened or closed.  If I might be affected,
	 fake a region-change so my attenuation is updated 
       without waiting for me to move."

	speaker region ifNotNilDo: [ :region |
		(region doors includes: door) ifTrue: [
			self signal: #soundRegionChanged with: region.
		]].! !

!QAvatarReplica methodsFor: 'goto' stamp: 'bgf 6/9/2010 10:54'!
drivingStarted
	"The user has engaged driving mode... 
		Forget any place we're currently occupying.
	 	Ff we're seated, get up.
		Note the Driving activity"

	lastMovedTime := (self island totalSeconds).		"To discourage setting the standing-place."
	self clearCurrentPlace.

	"Camera-lag fights awfully with the adjustments we do to keep carried-things 
	 (material editor, inventory) in visible place.  So hide while driving."
	carrying ifNotNil: [ 
		self carryingDo: [ : carry |
			(carry isKindOf: TMenuFrame) ifTrue: [
				carry close ]].
		carrying visible: false
	].
	self posture = #Sitting ifTrue: [
		self posture: #Standing.
	].
	self startActivity: #Driving.

	self refreshLeaderPlace.

! !

!QAvatarReplica methodsFor: 'goto' stamp: 'bgf 2/7/2010 08:26'!
drivingStopped
	"The user has released the driving-controls."
	super drivingStopped.
	self clearCurrentPlace.
	self stopActivity: #Driving.

	"Arrivals also to involve focal transforms becoming active."
	self viewportChanged; doViewportUpdate.

	"Either pull, to or snap back to, the meeting place, if the rules require it."
	self isMeetingLeader 
			ifTrue: [ self refreshLeaderPlace ]
			ifFalse: [ self returnToMeetingPlace ].

	"Brackets the visible: false in startDriving"
	carrying ifNotNil: [ carrying visible: true ].

	pasteboardScrap ifNotNil: [ pasteboardScrap resetStagger ]! !

!QAvatarReplica methodsFor: 'useful' stamp: 'bgf 1/9/2010 13:42'!
dropCard: card event: event rawCardPosition: v3 graspDistance: distance maxDropDistance: maxDistance
	"Worker operation for QViewPane dropCard:position:maxDistance:...
	 the user has dropped a document onto the main view, or done an Import... ,
	 and this method delivers the document into the space.

	 If a drop-target is found within the given maximum range,  drop the payload there.
	 (This is how apps are opened, and how posters are made on walls, and so on.)
	 Otherwise we place the card in view, carried by the avatar;
	 the harness has computed an appropriate global position in v3"
	| selection  |

	self root resolveEventFrame: event.
	event avatar: self.
	selection := event selection.

	"Timestamp the document.  The harness may have set the creation time.
 	 If so, mark this as a change.  Otherwise mark creation (for .c3x-document cards))"
	card creationTime > 0 ifTrue: [
		card docChange: 'Add' by: self time: ((event utcSeconds ifNil: [0]) 
												max: (card creationTime)).
	] ifFalse: [
		card docCreate: 'Add' by: self time: (event utcSeconds).
	].

	"If I'm in a place that forces the destination, just make it so."
	(self dropCardAtMyPlace: card) ifTrue: [ 
		card obliterateLater.
		^ nil ].

	"Check for a natural drop target; if we're close enough,
	 do a drop rather than leaving an intermediate card."
	(selection notNil and: [selection target notNil] 
	 and: [ selection distance < maxDistance ]  and: [ maxDistance > 0 ] 
	 and: [ card dropOnFrame: selection target avatar: self selection: selection ] )
	 ifTrue:  [ 
		card obliterateLater. 
		^ nil. 	"We're done."
		].

	"The card may handle being dropped in open space, e.g. for stored-objects."
	(selection notNil and: [ event cameraTransform notNil ]) ifTrue: [
		(card dropInSpace: event pointer: (event makePointer)) ifTrue: [
			card obliterateLater.
			^ nil
			]].

	"If we haven't consumed the drop yet, we just give the card to the user."
	card extent: 0.6@0.4;
		 translation: v3;
		 orientation: self orientation.
	self root addChild: card.


	"If the pick found a an obstacle, pull the card in front of it."
	(selection notNil and: [ selection target notNil] and: [ selection distance < 6 ]) ifTrue: [
		card dragInFrontOf: selection graspDistance: distance.
	].
	"Carry, so I don't lose it if I move around while the card drop is happening."
	self startCarryingNewFileCard: card.! !

!QAvatarReplica methodsFor: 'useful' stamp: 'bgf 4/11/2010 13:32'!
dropCardAtMyPlace: aFileCard
	"Special handling when I am formally in front of a display panel,
	 and the drop is displayable content... make it so.
	 Returns true if the card is opened in a display panel I am <at>, false in all other cases."
	| pf |

	(QDisplaySheet canAcceptDrop: aFileCard) ifFalse: [ ^ false ].
	
	pf := self currentPlaceFrame.
	(pf isKindOf: QDisplayPanel) ifFalse: [ ^ false ].

	QLogger logInfo: 'Direct placement ', (aFileCard shortSummary).
	pf sheet displayCard: aFileCard fromUser: self withUndoName: 'Drop'.
	^ true

! !

!QAvatarReplica methodsFor: 'useful' stamp: 'bgf 1/25/2010 13:11'!
editClick: target subframeID: frameID forcingDirect: bForceDirect 
	"Handler for QEditTask selection clicks - put the target into an edit box."

	| editFrame c |
	target ifNil: [ ^ nil ].
	target isInAvatar ifTrue: [ 
		self signal: #alertMessage with: 'Cannot resize your avatar elements.'.
		^ nil ].

	editFrame := (target getOwnedObject: frameID) component.
	editFrame ifNil: [ ^ nil ].
	"Disregard if there is an edit box holding me to any degree." 
	c := editFrame.
	[c isNil or: [ c = c root]] whileFalse:[
		c := c parent.
		(c isKindOf: QEditBox) ifTrue: [
			^ nil
		]].

	bForceDirect ifFalse: [
		"Select the topmost container, unless the click is modified, 
		 in which case edit the immediately containing component.	
		 (Further refinement is possible from the selection-box menu.)"
		[editFrame parent = editFrame root ] whileFalse: [ editFrame := editFrame parent ]
	].

	"Is this part of the scene we have edit access to currently? "
	(editFrame fromTemplate and: [ self isEditingTemplate not ])  ifTrue: [ 
		self signal: #alertMessage with: 'This ', (editFrame genericName), ' is a fixed scene element.'.
		^ nil ].

	"Cannot do this with locked-in-place elements (of course)."
	editFrame isPlacementLocked ifTrue: [
		self signal: #alertMessage with: 'This ', (editFrame genericName), ' is locked in place.'.
		^ nil ].
	editFrame canEditInBox ifFalse: [
		self signal: #alertMessage with: 'This ', (editFrame genericName), ' is not resizable.'.
		^ nil
	].

	self editFrame: editFrame.

! !

!QAvatarReplica methodsFor: 'editing' stamp: 'bgf 5/19/2010 14:20'!
editFrame: tframe
	"If possible, wrap the given in a 3D posing / editing tool."
	tframe isInAvatar ifTrue: [ ^ nil ].
	self root ifNil: [ ^ nil ].
	tframe isInEditBox ifTrue: [ ^ nil ].

	TObjectID for: 'Recycle-edit-box ', (self userLogin).
	self clearEditBox.
	editBox := (tframe editClass ifNil: [ ^nil ]) new.
	self root addChild: editBox.

	(self frameSelectionIncludes: tframe) ifTrue: [
		| sel parents |
		sel := self selectedFrames.
		parents := (sel collect: [ : f | f parent ]) asIdentitySet.
		(sel size > 1 and: [ parents size = 1]) ifTrue: [
			"Several distinct frames, children of the same parent.  Edit them as a group."
			| p |
			QLogger logWarning: 'Divert to group editing.'.
			p := parents anyOne.
			sel := sel asArray sort: [ : f1 : f2 | (p indexOfChild: f1) <= (p indexOfChild: f2) ].
			^ editBox editGroup: sel inParent: p forAvatar: self.
		]].

	self selectFrame: tframe.
	editBox edit: tframe forAvatar: self.
! !

!QAvatarReplica methodsFor: 'menus' stamp: 'bgf 7/14/2010 12:37'!
editFrameMaterial: aFrame
	"Menu callback - edit the given frame's material in my material-editor"
	aFrame ifNotNil: [
		(self canEditMaterial: aFrame) ifFalse: [
			self signal: #alertMessage with: 'No editable material.'
		] ifTrue: [
			self signal: #editFrameMaterial with: aFrame.
		].
	] ifNil: [
		self signal: #alertMessage with: 'Cannot edit undefined material.'.
	].! !

!QAvatarReplica methodsFor: 'useful' stamp: 'hrs 8/18/2010 15:59'!
ensureDisplay: cardName
	"If I am at a drop target (e.g., a panel) that contains the specified file card (e.g., on 
	a shelf), ensure it actualy displays the card and answer true. Otherwise false."
	self nearbyDropTarget ifNotNilDo: [:target | 		
		target allFramesDo: [:frame |
			(frame  isKindOf: QFileCard) ifTrue: [ | id | 
				id := frame identifier.
				(QWebServices match: cardName against: frame identifier caseSensitive: false) ifTrue: [
					(target respondsTo: #displayCard:fromUser:) ifFalse: [
						target := self root dropboxFor: self.
						self takeMeToFrame: target.
					].
					target displayCard: frame fromUser: self.
					^true
				].
			]
		].
	].
	^false.! !

!QAvatarReplica methodsFor: 'driving' stamp: 'bgf 4/12/2008 00:20'!
escape
	"Backstop mapping for #escape drive key map entry."
	self stopDriving! !

!QAvatarReplica methodsFor: 'gestures' stamp: 'RJT 5/8/2010 11:41'!
expireAnimation: anAnimation
	anAnimation delete ifTrue: [self costume userIdle].! !

!QAvatarReplica methodsFor: 'places' stamp: 'bgf 8/9/2008 23:34'!
facingMeStanceFor: anAvatar

	| pos root coi |
	anAvatar = self ifTrue: [ ^ nil ].
	root := self root.
	root ifNil:[ ^nil ].

	pos := self root freePlaceFor: anAvatar near: (self translation + (self lookAt * -8)).
	root miramarUI ifTrue:[
		coi := self centerOfInterest.
	] ifFalse:[
		coi := self translation.
	].
	^Stance here: pos there: coi! !

!QAvatarReplica methodsFor: 'useful' stamp: 'bgf 2/14/2010 19:49'!
fakePointerOverEvent: bForGround
	"Synthesize a straight-ahead pointerOver event, for use in mimicking drag-drop."

	^ self fakePointerOverEvent: bForGround 
		  rayTransform: (Matrix4x4 identity).! !

!QAvatarReplica methodsFor: 'useful' stamp: 'bgf 2/14/2010 21:34'!
fakePointerOverEvent: bForGround rayTransform: rayXform

	"Return a fake pointerOver Croquet Event as if my user's mouse were in mid-viewpane.
	 This can be used in conjunction with QDragAdapter dragDrop:inSpace:forEvent: 
	 to place objects in response to software-triggers...  when the object is wanted
	 to appear as if the user had dropped it into the space. 
	 This event is not for dispatch into TFrame doPointerXXX handling;
	 rather it is for code that expects to have been called with such an event supplied.
	 (It is valid to do makePointers for custom picking.) "
	| sel rx |
	[	
		rx := rayXform clone.
		bForGround ifTrue: [ rx addRotationAroundX: (-18) ].
		obstacleRay globalTransform: rx.
		obstacleRay resetTotal; solidOnly: false.
		obstacleRay maxDistance: 10.		"Arbitrary; we just need a pick to initialize everything."
		self root rayTest: obstacleRay ignore: { self . costume }.
		sel := obstacleRay selection copy.
	] ensure: [
		obstacleRay resetTotal
	].
	sel avatar: self.
	sel cameraTransform ifNil:[ 
		sel scale: 1.0.
		sel cameraTransform: (self localTransform).
		rx := rayXform clone.
		bForGround ifTrue: [ rx addRotationAroundX: (-18) ].
		sel rayTransform: rx.
	].

	^ (CroquetEvent new)
		userID: (self sessionID);		
		buttons: 0;
		keyValue: 0;
		selection: sel;
		avatar: self;
		timeStamp: (lastAliveTime);
		utcSeconds: (lastAliveUTC);
	 	eventType: EventPointerOver;
		nrClicks: 0;
		yourself! !

!QAvatarReplica methodsFor: 'menus' stamp: 'bgf 8/24/2010 10:23'!
fillContextMenu: tmenu forFrame: frame event: cEvent
	"Protocol: populate the 'context' (right-button) menu with actions.
	 The avatar argument is the user-principal being offered these commands;
	 it can be used to check for access-rights for destructive operations, etc."
	| mask sep insize |

	insize := tmenu menuItems size.
	sep := false.
	mask := frame contextMenuMaskForAvatar: self.

	(mask includes: #approach) ifTrue: [
		(tmenu addItemSpec: { 'Approach' . #bringAvatarHere: } target: frame)
			enabled: (frame isInAvatar not).
		sep := true].

	(mask includes: #meetHere) ifTrue: [
		(tmenu add: 'Meet Here' target: self action: #setMeetingPlace:  argument: (frame asPlace)).
		sep := true ].


	(mask includes: #attachNote) ifTrue: [
		sep ifTrue: [ tmenu addSeparator ].
		tmenu add: 'Attach Note...' target: self action: #addNoteToFrame:selection: 
									arguments: {frame . cEvent selection}.
		sep := true.
	].

	(mask includes: #material) ifTrue: [
		sep ifTrue: [ tmenu addSeparator. sep := false ].
		(tmenu add: 'Materials...' target: self action: #editFrameMaterial:  argument: frame)
			enabled: (self canEditMaterial: frame).
	].

	(mask includes: #properties) ifTrue: [
		sep ifTrue: [ tmenu addSeparator. sep := false ].
		tmenu addItemSpec: { 'Properties...' . #editProperties: } target: frame.
	].

	(mask includes: #resize) ifTrue: [
		sep ifTrue: [ tmenu addSeparator. sep := false ].
		(tmenu add:  'Resize...' target: self action: #editFrame: argument: frame)
			enabled: (frame isInEditBox not).
	].


	(mask includes: #copyToInventory) ifTrue: [
		tmenu addSeparator.
		tmenu addItemSpec: { 'Copy to Inventory' . #copyToInventory: } target: frame.	
	].

	(mask includes: #delete) ifTrue: [
		tmenu addSeparator.
		(self frameSelectionIncludes: frame) ifTrue: [
			(tmenu addItemSpec: { 'Delete ', (frame deletionLabel) . #confirmDeleteSelection:} target: self)
				enabled: (frame isInAvatar not).
		] ifFalse: [
			(tmenu addItemSpec: { 'Delete ', (frame deletionLabel) . #confirmDelete: } target: frame)
				enabled: (frame isInAvatar not).
		].
	].

	"If we were seeded a menu and added to it, separate the initial part."
	(insize > 0  and: [ tmenu menuItems size > insize ]) ifTrue: [
		(tmenu menuItems at: insize) separator: true.
	].
! !

!QAvatarReplica methodsFor: 'private' stamp: 'bgf 8/9/2009 16:42'!
findCurrentViaPlace
	"Based on my current stance, if I am seated within a seating destination,
	 return that outer destination as a place.   This is a moderately expensive search."
	currentPlace ifNil: [ ^ nil ].
	currentPlaceStance ifNil: [ ^ nil ].
	currentPlace isSeat ifFalse: [ ^ nil ].

	self root findNonNested: [ : f |
		f asPlace ifNotNilDo: [ : place |
			"If my seat is assignable from a volume seating destination, that's the via."
			place seatingVolume ifNotNilDo: [ : box |	
				(box pointInBox: currentPlaceStance here) ifTrue: [  
					^ place
				]].
			"If my seat places me in focal view on a display, the display is the via."
			place hasCameraStance ifTrue: [
				(place cameraStanceFor: self) ifNotNilDo: [ : cstance |
					(self isNearStance: cstance) ifTrue: [
						^ place
					]]].
			true.	"I.e. cull the search ... Via places aren't nested in other place-defining frames."
		] ifNil: [ false "Keep searching"].
	].

	"Fell out of the search with an early return."
	^ nil! !

!QAvatarReplica methodsFor: 'private' stamp: 'bgf 7/24/2009 15:22'!
findSeatIn: aPlace
	"Intercept rule for go-to..... if going to this place implies taking a seat, do that."

	aPlace ifNil: [ ^ nil ].
	self avatarIAmWith ifNotNil: [ : other |
		other viaPlace ifNotNilDo: [ : vp |
				vp ~= aPlace ifTrue: [ ^ self findSeatIn: vp ]
			] ifNil: [
				^ nil	 	"Shortcut... the leader has already failed to find a via."
			]].

	"If the given place does region-based seat assignment,
	 find and return a seating place within it (if any are free.)"
	aPlace seatingVolume ifNotNilDo: [ : box |
		(self nearestSeatWithinBox: box) ifNotNilDo: [ : seat |
			seat = aPlace ifFalse: [
				^ seat
			]].
		^ nil].

	"If the given place is a viewable area with a camera stance, 
	 and there is a matching seat for it, then return that seat.
	 This is so that when you go-to a display on a desktop station,
	 you get seated if there is a matching task chair."
	aPlace hasCameraStance ifTrue: [
		| stance |
		"See also QAvatarReplica isNearStance: , the nearness heuristic should match."
		stance := aPlace cameraStanceFor: self.
		stance ifNil: [ ^ nil ].
		(self nearestSeatWithinDistance: 5 andDegrees: 20 ofStance: stance)
		 ifNotNilDo: [ : seat |
			seat = aPlace ifFalse: [
				^ seat
			]]].

	^ nil
! !

!QAvatarReplica methodsFor: 'accessing' stamp: 'bgf 7/7/2009 16:39'!
focalCameraTransform
	"If I am an override camera (global) xform is in force
	 due to place focus or look-mode, return that transfom."

	self inTransit ifTrue: [ ^ nil ].
	
	^ currentFocalTransform! !

!QAvatarReplica methodsFor: 'accessing' stamp: 'bgf 6/30/2009 10:42'!
focalCameraTransform: xformOrNil

	currentFocalTransform := xformOrNil.
	self signal: #focalTransformChanged with: xformOrNil.
	self viewportChanged.
	
! !

!QAvatarReplica methodsFor: 'places' stamp: 'bgf 8/9/2009 14:41'!
focalPlace
	"If either my via place or my current define a camera stance, return that place."
	self viaPlace ifNotNil: [ : vp | 	
		vp hasCameraStance ifTrue: [ ^ vp ]
	].
	self currentPlace ifNotNil: [ : cp | 
		cp hasCameraStance ifTrue: [ ^ cp ].
	].
	^ nil
! !

!QAvatarReplica methodsFor: 'private' stamp: 'bgf 3/9/2010 16:28'!
followPlace: aPlace forRequestTime: requestTime
	"<on: frameChanged in: currentPlace destinationFrame>"
	"When the frame defining my current place moves, I follow it."
	| alreadyAt |

	"Is this the follow-place I expect to be executing?  
	 If not, then circumstances have changed in the meantime."	
	followPlaceTime = requestTime ifFalse: [ ^ self ].	
	followPlaceTime := nil.
	
	"Check for destruction."
	(aPlace destinationFrame isNil or: [ aPlace destinationFrame root isNil ]) ifTrue: [
		self stopScript: #doFollowPlace:.
		^ self
	].

	(currentPlace = aPlace) ifTrue: [
		alreadyAt := aPlace.
	] ifFalse: [
		"If following a leader who is seated at a table, and we're seated there also,
	 	  then the leader's viaPlace is the table, while ours is the leader. 
	 	  To prevent a spurious top-level go-to the table (with a seat-hop induced),
	 	  detect the case where we are already seated within the leaders' viaPlace."
		aPlace avatarToFollow ifNotNilDo: [ : leader |
			leader viaPlace ifNotNilDo: [ : vp |
				((self isAtPlace: vp) or: [ vp = (self findCurrentViaPlace) ]) 
				ifTrue: [ 
					alreadyAt := vp.
				]]	
	]].

	"We may already be placed appropriately, and just need to ensure we're transformed properly."
	alreadyAt ifNotNil: [
		| stance xform |
		stance := currentPlace destinationStanceFor: self.
		stance ifNil: [ 
			"Whatever we were following is no longer rooted."
			self stopScript: #doFollowPlace:.
			^ self.
		].
		(currentPlaceStance hasEqualElements: stance) ifFalse: [
			xform := stance transform.
			xform ifNotNil: [
				currentPlaceStance := stance.
				self goto: xform
				 	sticky:  (self stickyValueForPlace: aPlace) 
				 	label: (aPlace label).
				self signal: #statusMessage with: (aPlace label).
			]].
		self updateFocalTransform.
	] ifNil:  [
		self goToPlace: aPlace andFollow: true.
	]! !

!QAvatarReplica methodsFor: 'places' stamp: 'bgf 6/28/2009 17:28'!
followers
	"All other avatars in the space who are attached to my beside-me place."
	^ self root ifNotNilDo: [ :space|
		space allUsers select: [ : av | av avatarIAmWith = self ]
	   ] ifNil: [
		^#() 
	   ].! !

!QAvatarReplica methodsFor: 'menus' stamp: 'bgf 3/27/2010 08:45'!
generatePoll: event ballots: ballots
	"Make a poll using my selected ballot-candidate notes, and plop it right here."
	| survey poll |
	ballots size < 2 ifTrue: [
		"Group changed underneath me. "
		event avatar signal: #alertMessage 
					with: 'Need at least 2 choices for a poll.\Choices are notes and topic cards.'.
		^ self ].
	QActivityCapture 
			action: #GeneratePoll 
			details: (ballots size asString)
			forSession: (self sessionHex).

	survey := (QChartData new)
				setDescriptions: ( ballots collect: [ : n | n contents ] );
				setLabels: (ballots collect: [ : n | 
						QStyles shorten: n contents toLabelOfLength: (QPollBarChart labelMax) ]);
				setValues: ( ballots collect: [ :n | 0 ] );
				setColors: ( ballots collect: [ :n | n ballotColor ]);
				minRange: ((self root numUsersPresent // 2) max: 4);
			 yourself.
	poll := QPollBarChart new.
	poll data: (survey).
	poll extent: ((ballots size max: 3) min: 7)*0.8 @ (1.8) @ (0.05).
		"The future below allows the TMenuFrame to get out of the way
	 	  before we start a pick to find where to drop the poll."
	QDragAdapter future  dragDrop: poll intoSpace: (self root) forEvent: event.

	self future captureCreateFrames: { poll } description: 'Add Poll'.

	poll future titleDoubleClick: event.! !

!QAvatarReplica methodsFor: 'places' stamp: 'bgf 3/19/2010 18:39'!
goBeside: anAvatar
	"Go stand by the other avatar and follow them, until separate navigation is done."
	| result pbm |
	anAvatar = self ifTrue: [ ^ 0 ].
	(anAvatar currentPlace = self placeBesideMe) ifTrue: [
		"If I'm following him, he's no longer following me."
		anAvatar clearCurrentPlace
	].
	result := self goToPlace: (pbm := (anAvatar placeBesideMe))
				  andFollow: true.

	self refreshDefaultPOV: pbm.

	self signal: #infoMessage 
		with: ( 'You are with ', anAvatar publicName ).
	anAvatar signal: #infoMessage 
			  with:  (self publicName), ' is following you.'.

	"When someone joins the posse, make sure everyone's in a reasonable spot."
	anAvatar signal: #frameChanged with: self.
	^ result! !

!QAvatarReplica methodsFor: 'places' stamp: 'bgf 6/21/2008 17:25'!
goToBirdsEye

	"Move to the bird's eye landmark.
	 If there is none, go high and wide (it's ok, we're at first and 10.)."

	| lms place |

	lms := self currentLandmarks.
	lms ifNil: [ ^ self ].		"Means the replica isn't ready yet."
	
	"Find a landmark whose label contains 'bird' and 'eye'."
	place := lms detect: [ :lm |
					 (lm label includesSubstring: 'bird' caseSensitive: false) 
					  and: [lm label includesSubstring:'eye' caseSensitive: false]] 
				ifNone: [ nil ].

	QActivityCapture action: #GoToEnd details: 'BirdsEye' forSession: (self sessionHex).

	place ifNotNil: [
		self goToPlace: place.
	] ifNil: [
		"Make one up."
		| xform |
		self clearCurrentPlace.

		xform := Matrix4x4 new.
		xform translation: 120@80@120.
		xform pitchYawRoll: (-30@40@0).
		self goto: xform sticky: true label: 'Bird''s Eye View'.
		self signal: #infoMessage with: 'Bird''s Eye View'.
	].

! !

!QAvatarReplica methodsFor: 'events' stamp: 'bgf 3/22/2010 12:03'!
goToLeaderSiteIfAny: place 
	"If the given meeting-place defines a leader position, go there."
	place destinationFrame ifNotNil: [ : df |
		(df respondsTo: #leaderSite) ifTrue: [
			df leaderSite ifNotNil: [ : ls |
				self goToPlace: (ls asPlace).
				self signal: #infoMessage with: 'You are in the leader position.'
			]]].
! !

!QAvatarReplica methodsFor: 'places' stamp: 'bgf 8/12/2009 22:35'!
goToPlace: aPlace
	"Go the given place, but don't tie myself to its subsequent motion.
	 Returns the number of goto steps entailed by this motion,
 	 as an estimate of timing for the journey."

	"Intercept and honor the special 'go with' place.
	 This allows sending the beside-me place in from a harness-side copy."
	aPlace ifNotNil: [
		aPlace avatarToFollow ifNotNil: [ : avatar |
			avatar = self ifFalse: [
				^ self goBeside: avatar
			]]].
	
	"If we're already seated in this place, we don't need to go again."
	(self viaPlace = aPlace) ifTrue: [ 
		self updateFocalTransform.
		^ self 
	].
	(self findSeatIn: aPlace) ifNotNilDo: [ : seat |
		^ [ self goToPlace: seat andFollow: false ] ensure: [ self setViaPlace: aPlace ].
	].

	^ self goToPlace: aPlace andFollow: false.! !

!QAvatarReplica methodsFor: 'places' stamp: 'bgf 8/19/2009 02:25'!
goToPlace: aPlace andFollow: bFollow
	"Go to the given QPlace; if bFollow is true,
	 attach a listener for motions of the place's frame, and stay with it.
	 Returns the number of goto steps entailed by that motion."
	| stance xform |
	stance := aPlace destinationStanceFor: self.
	stance ifNil: [ 
	   QLogger logInfo: 'No stance: ', aPlace, ' ', aPlace destinationID, ' ', aPlace stanceSelector.
	   ^ 0].

	"Intercept - if the destination defines a focal camera-viewpoint
	 and I am already near it, stay where am I but adopt the camera stance."
	bFollow ifFalse: [
		(aPlace hasCameraStance 
		 and: [ aPlace isFollowMePlace not ] 
		 and: [ currentPlace notNil ])
		 ifTrue: [
			| cstance | 
			cstance := aPlace cameraStanceFor: self.
			(cstance notNil and: [self isNearStance: cstance]) ifTrue: [
				self suggestPOV: ( aPlace suggestedPOV ifNil: [ 'First Person'] ) forPlace: aPlace.
				self setViaPlace: aPlace.
				self updateFocalTransform.
				^ 0 ]]].

	"Normal case - claim this as my place (so occupancy begins now)
	 and have the harness drive me there."
	self currentPlace: aPlace stance: stance andFollow: bFollow.
	xform := stance transform.
	self  goto: xform
		 sticky: (self stickyValueForPlace: aPlace)
		 label: aPlace label.
	self signal: #statusMessage with: (aPlace label).
	^ TAvatarUser stepsToGoTo: xform from: localTransform
	! !

!QAvatarReplica methodsFor: 'places' stamp: 'bgf 6/22/2009 17:58'!
goToUser: anAvatar

	"Go stand beside the other avatar."
	self goToPlace: anAvatar placeForJoiningMe.

	self signal: #infoMessage with: 'You have joined ', anAvatar publicName.
	anAvatar signal: #infoMessage with: (self publicName), ' has joined you.'.! !

!QAvatarReplica methodsFor: 'menus' stamp: 'bgf 2/9/2010 16:26'!
groupContextMenuForFrames: frames event: cEvent
	"Present  the context' (right-button) menu with actions for a group of frames.  
	 (Frames is expected to be a collection of TFrames, and should have  size > 1 .."

	| mask sample tmenu klass |

	self canModifyForum ifFalse: [ ^ self ].
	tmenu := TMenuFrame new.
	sample := frames anyOne.
	mask :=  (frames copyWithout: sample)
			 inject: (sample contextMenuMaskForAvatar: self)
				into: [ : m : frame |
					     m intersection: (frame contextMenuMaskForAvatar: self) ].

	"Some homogenous groups have their own extra menu items."
	klass := QPropertyDialog commonBaseClass: ((frames collect: [ : f | f class ]) asArray).
	(klass includesBehavior: TFrame) ifTrue: [
		klass addGroupMenuItems: tmenu mask: mask frames: frames event: cEvent.
		tmenu addSeparator.
	].

	(mask includes: #properties) ifTrue: [
		tmenu addItemSpec: { 'Properties...' . #editProperties: } target: (sample).
	].

	(mask includes: #resize) ifTrue: [
		(tmenu add:  'Move...' target: self action: #editFrame: argument: (sample))
			enabled: (frames allSatisfy: [ :f | f isInEditBox not ]).
	].

	(mask includes: #delete) ifTrue: [
		tmenu addSeparator.
		(tmenu add: ('Delete ', (frames size), ' Items') 
				target:self 
				action: #confirmDeleteSelection:)
		enabled: (frames allSatisfy: [ :f | f isInAvatar not ]).
	].

	"If we got any commands, show the menu."
	tmenu menuItems size > 0  ifTrue: [
		tmenu menuItems last separator: false.
		self postMenu: tmenu forEvent: cEvent.
	] ifFalse: [
		"No menu, no love."
		tmenu obliterate.
	].! !

!QAvatarReplica methodsFor: 'step' stamp: 'bgf 8/28/2008 23:35'!
hasBeenSuperseded: myState age: hbAge

	"I am superseded if there is another avatar with the same user id in the space,
	 and I am invisible (i.e. not getting heartbeats) and it is not.
	 This happens after reconnect, or after a user hard-crashes and then re-logs in.

	 Note that by extending the match to serviceUserID, we are enforcing
	 a policy of one-account, one-avatar at the implementation level.

	 When there is a hard-crash and relogin,
	 the session id changes, but the user-service-id matches.
	 So comparing service-user-id  is how
	 we clean up zombies from sessions that you crashed out of."

	(myState == #normal) ifTrue: [ ^ false ].

	"Check for other Qwaq avatars that might match the user's login."
	^ (self root avatars anySatisfy: [ : av |
			av ~= self 
			and: [ av timeSinceLastHeartbeat < hbAge ]
			and: [ av respondsTo: #userLogin ] 
			and: [ av userLogin = self userLogin ]
		]) or: [super hasBeenSuperseded: myState age: hbAge].! !

!QAvatarReplica methodsFor: 'useful' stamp: 'bgf 6/10/2009 11:17'!
hasCapability: aSymbol
	^ userInfo notNil and: [ userInfo capabilities includes: aSymbol ]! !

!QAvatarReplica methodsFor: 'useful' stamp: 'bgf 6/10/2009 11:17'!
hasFeatureEntitlement: aSymbol
	^ userInfo notNil and: [
	    	(userInfo capabilities includes: aSymbol)
	    	or: [ userInfo features includes: aSymbol ] ]! !

!QAvatarReplica methodsFor: 'editing' stamp: 'bgf 3/13/2010 08:41'!
hasPastedFrames
	"Do I have frames available for pasting (i.e. cut or copied to pasteboard) ?"
	^ pasteboardScrap notNil! !

!QAvatarReplica methodsFor: 'meetings' stamp: 'bgf 7/25/2009 21:21'!
hasPermissionToSpeak
	"True if the meeting is speech-controlled, and I'm allowed to speak in it now."

	^ self meetingState ifNotNilDo: [ : meeting |
		self isMeetingLeader 	"The meeting-leader is always labeled in this mode."
		   or: [
			meeting audienceIsMuted 
			 and: [ meeting allowedToSpeak includes: (self userLogin) ] ]
		] ifNil: [ 
			false
		].
! !

!QAvatarReplica methodsFor: 'accessing' stamp: 'jcg 2/16/2008 00:10'!
hasVideoStream
	^hasVideoStream! !

!QAvatarReplica methodsFor: 'accessing' stamp: 'bgf 1/23/2009 10:59'!
hasVideoStream: bHasStream
	"Update the hasVideoStream flag, whose changes are observable."
	(bHasStream = hasVideoStream) ifFalse: [
		hasVideoStream := bHasStream.
		self signal: #hasVideoStreamChanged with: bHasStream
	].
! !

!QAvatarReplica methodsFor: 'places' stamp: 'bgf 4/21/2009 13:12'!
heightForPosture: postureSym
	"Return the heightAboveGround value for this avatar when in the given posture
	 (the distance from supporting surface to eye-point). 
	 The standing height is the default result for unknown posture symbols, nil, etc."

	^ postureSym caseOf: {
		[ #Sitting ] -> [ self seatToEyes ] .
	   } otherwise: [ 
		self footToEyes
		].! !

!QAvatarReplica methodsFor: 'render' stamp: 'bgf 3/24/2007 16:09'!
hilite:aColor
	"Disallow this - it causes chaos in some of the custom rendering."! !

!QAvatarReplica methodsFor: 'places' stamp: 'jjl 8/5/2008 12:04'!
homePlace

	"A default 'Home' place mapped to the space home position.
	 Note that the label + category here matters - see goToPlace: "

	^  QPlace 	
			category: 'Named Places'
			label: 'Home' 
			frame: (self root)
			selector: #homeStanceFor:
			sticky: false
! !

!QAvatarReplica methodsFor: 'goto' stamp: 'bgf 3/15/2008 18:37'!
inAStickyPlace
	"Am I in a Place that imposes a sticky viewpoint?"
	^ currentPlace notNil and: [ currentPlace sticky ]! !

!QAvatarReplica methodsFor: 'accessing' stamp: 'bgf 3/19/2010 21:33'!
inTransitChanged: bInTransit
	"The in-transit state has changed to the given value (boolean)"
	| bCheck |
	self root ifNil: [ ^ self ].
	costume ifNotNil: [ costume inTransit: bInTransit ].
	label ifNotNil: [ label update ].
	pasteboardScrap ifNotNil: [ pasteboardScrap resetStagger ].

	"Signal the control system that a seat is or is not available. 
	 The seat-search is expensive so only do it when necessary."
	bInTransit ifFalse: [
		bCheck := self allowedToMove and: [ self avatarIAmWith isNil ].
		self signal: #isNearASeat with: (bCheck and: [ self isNearASeat ] ).
	] ifTrue: [
		"If I've changed to being in transit, and am being led,  then return to my default POV.
		 This gets phone-proxied user POV's back to default automatically."
		self avatarIAmWith ifNotNil: [
			self suggestPOV: #Preferred forPlace: (self currentPlace)
		].
	].! !

!QAvatarReplica methodsFor: 'status' stamp: 'jjl 9/8/2008 13:34'!
informUser: aNameString

	self signal: #informUser: with: aNameString! !

!QAvatarReplica methodsFor: 'initialize' stamp: 'bgf 5/24/2010 10:49'!
initialize
	super initialize.

	avatarHeight := 1.0.
	userInfo := QUserInfo new.
	activities := IdentitySet new.
	"The speaker plays my voice, if my user has voice chat turned on."
	speaker := QSpeaker new.
	speaker rotationAroundY: 180.
	self addChild: speaker.

	camera := QVideoCamera new.
	camera sessionID: sessionID.
	camera visible: false.
	self addChild: camera.

	shadow := QMaterials newRadialShadow.
	shadow translation: (0@-5@0).	"Default, usually overridden in configureAvatar"
	shadow extent: 1@1.
"	self addChild: shadow. add it to the proxy instead"
	
	self startScript: #voiceActivity: when: {speaker . #activity}.

	label := QAvatarLabel new.
	label avatar: self.
	standin := QAvatarStandIn new.
	standin avatar: self.

	"The inventory is just a staging area; for in-space inventory capture."
	self setupInventoryHolder.

	"self addChild: label.  Add these to the proxy instead"

	birthTime := self island time.
	lastMovedTime := nil.
	hasVideoStream := false.
	isRecordingSession := false.
	isBroadcastingSession := false.

	"Connect for subsequent user-info updates."
	self startScript: #onProfileChanged when: {userInfo. #screenNameChanged}.
	self startScript: #onProfileChanged when: {userInfo. #personalSummaryChanged}.
	self startScript: #onProfileChanged when: {userInfo. #publicEmailChanged}.
	self startScript: #onAvatarColorChanged when: {userInfo . #avatarColorChanged}.
	self startScript: #onSmallPictureChanged when: {userInfo . #smallPictureChanged}.
	self startScript: #onBadgePictureChanged when: {userInfo . #badgePictureChanged}.
	
	TObjectID for: 'Replica-Initialize'.! !

!QAvatarReplica methodsFor: 'accessing' stamp: 'bgf 4/1/2008 19:32'!
innerCostume
	"The contents of my costume proxy - the true costume"
	^ costume ifNotNil: [ costume contents ]! !

!QAvatarReplica methodsFor: 'menus' stamp: 'bgf 2/7/2010 19:04'!
inspectorMenuMaskForFrame: frame
	"Returns an array of symbols indicating which inspector-context-menu actions
	 should be available -- coordinated with QFrameTreeItem #treeMenu: "
	| result isUngroupable plainGroup isThumb isMeshTree inBox isBox isReparentable canReorder |

	frame ifNil: [ ^ #() ].
	frame isInAvatar ifTrue: [ ^ #() ].
	
	result := Set new.
	isBox := (frame class = QEditBox).
	inBox := isBox or: [frame isInEditBox].
	isThumb := frame isFrameThumb.
	isReparentable := inBox not and: [ self canDeleteFrame: frame ] and: [ frame cumulativeScale = 1 ].
	canReorder := inBox not and: [ frame canReorder ].
	isMeshTree := ((isThumb not) and: [ frame isMeshTree]).
	isUngroupable := inBox not 
					 	and:[ frame isEditableGroup ]
					 	and:[ frame isSpace not ]
						and:[ frame parent notNil and: [ frame parent isEditableGroup ]].

	plainGroup := (frame class = TGroup).

	(inBox not) & (isReparentable) ifTrue: [ 
		frame canEditInBox ifTrue: [ result add: #resize ] ].
	frame canEditMaterial ifTrue: [ result add: #material ].
	result add: #properties.		"Always, currently."

	canReorder ifTrue: [		"I.e. can shift this child  within its parents child-list"
		| idx size  |
		idx := frame parent indexOfChild: frame.
		size := frame parent children size.

		"We use 'Move Up' rather than 'Move to First' as the only choice if we're at position two."
		idx > 2 ifTrue: [  result add: #reorderFirst ].
		idx > 1  ifTrue: [ result add: #reorderUp ].
		idx <  size ifTrue: [ result add: #reorderDown ].
		idx < (size-1) ifTrue: [ result add: #reorderLast ].
	].

	(isMeshTree & (inBox not)) ifTrue: [ result add: #flatten ].

	isReparentable ifTrue: [ result add: #group ].
	isUngroupable ifTrue: [ result add: #ungroup ].

	"The wrapping-into-behavior options."
	(isReparentable 
	 and: [ frame isGroup or: [ frame isGeometry]] 
	 and: [ frame isLight not ]
	 and: [ frame dragAdapter isNil ])	"Fliters out a lot of prebuilt things."
	 ifTrue: [
		| dest port fs |
		dest := frame ancestorOrSelfThatIsA: QDestination.		"(Seat is kind of destination.)"
		port := frame ancestorOrSelfThatIsA: QPortableObject.
		fs 	 := frame ancestorOrSelfThatIsA: QFixedSurface.
		(fs isNil) & (port isNil) & (dest isNil) ifTrue: [ 
			result add: #makeFixedSurface.
			result add: #makePortable.
			result add: #makeDestination.
			result add: #makeSeat.
		].

		"Avoid making sound regions out of anything complicated... that is usually a misunderstanding.
		 Someday this should allow creating a cube around the bbox of the selection, as the basis of a region."
		(plainGroup or: [ frame isGeometry and: [ frame isProcedural ]]) 
		ifTrue: [
			result add: #makeSoundRegion
		]].

	frame isSpace ifFalse: [
		isBox ifFalse: [ result add: #saveToXML ].
		self canUseInventory ifTrue: [ 
			((frame contextMenuMaskForAvatar: self) includes: #copyToInventory) ifTrue: [
				result add: #copyToInventory ].
			].
		isThumb & (inBox not) ifTrue: [ result add: #convertFromThumb ].
		isMeshTree & (inBox not) ifTrue: [ result add: #convertToThumb ].

		(self canDeleteFrame: frame) ifTrue: [ result add: #delete ].
	].

	^ result asArray

			

	! !

!QAvatarReplica methodsFor: 'menus' stamp: 'bgf 4/11/2010 13:51'!
inspectorMenuMaskForSelection
	"Returns an array of symbols indicating which inspector-context-menu actions
	 should be available -- coordinated with QInspector framesMenu  "
	| sample mask |

	selectedFrames isEmptyOrNil ifTrue: [ ^ #() ].
	self singleSelectedFrame ifNotNilDo: [  : frame  |
		^ self inspectorMenuMaskForFrame: (frame)
	].

	sample := selectedFrames anyOne.
	mask :=  (selectedFrames copyWithout: sample)
				inject: (self inspectorMenuMaskForFrame: sample) asSet
				into: [ : m : frame |
					     m intersection: (self inspectorMenuMaskForFrame: frame) ].
		
	(selectedFrames allSatisfy: [ : f | f parent = (selectedFrames first parent) ]) 
	ifFalse: [
		mask remove: #group ifAbsent: [].
		mask remove: #makeFixedSurface ifAbsent: [].
		mask remove: #makeSoundRegion ifAbsent: [].
	].

	"Elide items not implemented for multiples."
	#(copyToInventory material saveToXML
	   makeSeat makePortable makeDestination 
	   convertToThumb convertFromThumb
	   reorderFirst reorderLast reorderUp reorderDown)
	 do: [ : item |
		mask remove: item ifAbsent: []
		].

	^mask asArray! !

!QAvatarReplica methodsFor: 'render' stamp: 'bgf 8/20/2009 18:26'!
interceptCostumeRender: proxy ogl: ogl
	"Should I render my costume? "

	((ogl inPortal2D not) and: [ ogl inPortal3D not ])
	 ifTrue: [
		"We sometimes omit avatars completely to avoid winking or obstruction."
	 	ogl avatar ifNotNilDo: [ : auser |  
			(self isThisUserAvatar: ogl) 
				ifTrue: [		
					(self shouldRenderSelf: ogl) ifFalse:  [ ^ 0 ]
				] ifFalse: [ 
					 (self shouldRenderOther: ogl) ifFalse: [ ^ 0 ].
				]]].

	"If beyond the visible distance, don't render the costume,  just the standin."
	(self costumeVisible: ogl) ifFalse: [
		"We know the label is a child of the proxy, because we put it there."
		^ standin renderFrame: ogl
	].
	^ -1	  "No intercept if inside the visible distance" ! !

!QAvatarReplica methodsFor: 'events' stamp: 'bgf 11/11/2009 19:17'!
interceptMouseDown: event 
	"Boolean - if true, then the event is not passed on to its target frame.
	 Gives the avatar a chance to handle the mouse event directly
	 before it is dispatched to the target frame. 
	 Here, we use this to provide warp-to when mousing over faraway places. "
	
	self doWarpMoves ifTrue: [
		self signal: #overWarpPlace with: nil.
		(self warpPlaceUnderMouse: event) ifNotNilDo: [ :place |
			self floatingInfo: nil.	"Clear the tag"
			self goToPlace: place.
			^ true
		]].
	^ false

			! !

!QAvatarReplica methodsFor: 'events' stamp: 'bgf 11/9/2008 14:26'!
interceptMouseEnter: event 
	"Boolean - if true, then the event is not passed on to its target frame.
	 Gives the avatar a chance to handle the mouse event directly
	 before it is dispatched to the target frame. 
	 Here, we use this to provide warp-to when mousing over faraway places. "

	(self warpPlaceUnderMouse: event) ifNotNilDo: [ :place |
		self signal: #overWarpPlace with: place.
		^ false
	].
	self signal: #overWarpPlace with: nil.
	^ false

			! !

!QAvatarReplica methodsFor: 'events' stamp: 'bgf 11/9/2008 14:26'!
interceptMouseLeave: event 
	"Boolean - if true, then the event is not passed on to its target frame.
	 Gives the avatar a chance to handle the mouse event directly
	 before it is dispatched to the target frame. 
	 Here, we use this to provide warp-to when mousing over faraway places. "

	self signal: #overWarpPlace with: nil.
	^ false

			! !

!QAvatarReplica methodsFor: 'events' stamp: 'bgf 11/9/2008 14:26'!
interceptMouseOver: event 
	"Boolean - if true, then the event is not passed on to its target frame.
	 Gives the avatar a chance to handle the mouse event directly
	 before it is dispatched to the target frame. 
	 Here, we use this to provide warp-to when mousing over faraway places. "

	(self warpPlaceUnderMouse: event) ifNotNilDo: [ :place |
		self signal: #overWarpPlace with: place.
		^ false
	].
	self signal: #overWarpPlace with: nil.
	^ false

			! !

!QAvatarReplica methodsFor: 'initialize' stamp: 'jjl 8/22/2008 16:24'!
inventoryMessage

	^'Inventory (drag content here.)'
! !

!QAvatarReplica methodsFor: 'private' stamp: 'bgf 8/6/2009 22:07'!
isASeatWithin: distance of: globalPosition
	"Return any (free) sitting place near to my current location,
	 if any are available within the given radius."
	| seats sdist savePlace |
	self root ifNil: [ ^ false ].
	savePlace := currentPlace.
	[	
		currentPlace := nil.
		seats := self root findNonNested: [ : frame | 
			(frame isSeat)
			and:  [ frame isOccupied not ]].
	] ensure: [ currentPlace := savePlace ].

	seats isEmptyOrNil ifTrue: [ ^ false ].
	sdist := distance * distance + (Float epsilon).

	"To save time, we consider the global position of the seat rather than the exact stance;
	 this is imprecise due to seating offsets but fine for our purposes."
	seats do: [ : seat |		
		| d |
			d := (seat globalPositionReal - globalPosition) squaredLength.
			d <= sdist ifTrue: [ ^ true ].
			].
	^ false
! !

!QAvatarReplica methodsFor: 'useful' stamp: 'bgf 6/10/2009 11:29'!
isAdministrator
	^ userInfo notNil and: [ userInfo capabilities includes: #administer ]! !

!QAvatarReplica methodsFor: 'goto' stamp: 'bgf 8/9/2009 15:46'!
isAtDestinationFrame: aFrame
	"If I am at (or on my way to) the destination represented by this frame, return true.
	 This is complicated because we're back-reckoning from a destination TFrame to cases
	 that might possibly drive me to it."

	currentPlace ifNil: [ ^ false ].
	currentPlace isStatic ifFalse: [
		(currentPlace destinationFrame == aFrame) ifTrue: [ ^ true ].
	].
	self viaPlace ifNotNilDo: [ : vp |
		(vp isStatic not and: [ vp destinationFrame == aFrame ]) ifTrue: [ ^ true ].
	].
	currentPlace isFollowMePlace ifTrue: [
		currentPlace avatarToFollow ifNotNilDo: [ : leader| 
			leader viaPlace ifNotNilDo: [ : vp | 
				(vp isStatic not and: [ vp destinationFrame == aFrame ])  ifTrue: [ ^ true ].
			]]].
		
	^ false
! !

!QAvatarReplica methodsFor: 'places' stamp: 'bgf 7/3/2009 18:47'!
isAtPlace: aPlace
	currentPlace = aPlace ifTrue: [ ^ true ].
	self viaPlace = aPlace ifTrue: [ ^ true ].	"E.g. I'm sitting within the place."
	^ false

! !

!QAvatarReplica methodsFor: 'accessing' stamp: 'jcg 9/6/2009 23:58'!
isBroadcastingSession
	"True if the receiver is currently broadcasting the session of the space it is in."
	^isBroadcastingSession! !

!QAvatarReplica methodsFor: 'accessing' stamp: 'jcg 9/6/2009 23:59'!
isBroadcastingSession: aBool
	"The receiver is currently broadcasting the session of the space it is in."
	isBroadcastingSession = aBool ifFalse:[
		isBroadcastingSession := aBool.
		self root signal: #isBroadcastingSessionChanged with: self.
		label ifNotNil: [ label update ].
	].
! !

!QAvatarReplica methodsFor: 'accessing' stamp: 'bgf 1/27/2008 10:15'!
isDelayed
	"True if this user's avatar heartbeat is falling behind enough to warrant warnings."
	^ (heartbeatState == #warning) | (heartbeatState == #invisible)! !

!QAvatarReplica methodsFor: 'testing' stamp: 'bgf 11/12/2009 12:26'!
isEditingTemplate
	"Is my user currently engaged in full scene-editing?"
	^ activities includes: #EditingTemplate! !

!QAvatarReplica methodsFor: 'accessing' stamp: 'bgf 1/27/2008 10:06'!
isIdle
	"Is this person interacting at the moment?"
	^ notice notNil and: [notice beginsWith: 'Idle']! !

!QAvatarReplica methodsFor: 'meetings' stamp: 'bgf 8/19/2009 03:48'!
isInMeetingPlace: meetingPlace
	"The given place is the current leader-imposed place.  
	 Am I meeting the require to be there?"
	| leader | 
	currentPlace = meetingPlace ifTrue: [ ^ true ].
	self viaPlace = meetingPlace ifFalse: [ ^ false ].

	"OK, I claim to be nominally at the meeting place.
	 If not in dynamic follow-the-leader, that's sufficient."
	leader := self meetingLeader.
	(leader notNil and: [leader = meetingPlace avatarToFollow ]) 
		ifFalse: [ ^ true  ].

	"But if the meeting place is follow-the-leader, 
	 my via place is forced to be the leader place (see #viaPlace).   
	 Since my currentPlace is something else... am I really following ?   "
	leader nominalPlace ifNotNil: [ : lp |
		lp = currentPlace ifTrue: [ ^ true ].
		currentPlace ifNotNil: [ 
			(lp seatingRange > 0 and: [currentPlace posture = #Sitting])
				 ifTrue: [ ^ true ].
		].
	].
	^ false! !

!QAvatarReplica methodsFor: 'render' stamp: 'bgf 7/25/2009 21:16'!
isLabeledForSpeaking
	"Whether, under 'While Speaking'  labels preference,
	  the label of this avatar should appear."

	^ self isSpeaking or: [ self wantsToSpeak ] or: [ self hasPermissionToSpeak ]! !

!QAvatarReplica methodsFor: 'events' stamp: 'bgf 3/22/2010 12:01'!
isLeader: bFlag
	"Notification once I have become, or have resigned from, the leader role."
	self meetingPlace ifNotNil: [ : place |
		bFlag ifTrue: [
			self goToLeaderSiteIfAny: place 
		] ifFalse: [
			self goToPlace: place
		]].
! !

!QAvatarReplica methodsFor: 'meetings' stamp: 'bgf 7/3/2009 13:17'!
isMeetingLeader
	"True if I am the designated meeting-leader in this forum."
	^ self root notNil 	
		and: [ self userLogin notNil ]
		and: [ self meetingState leaderLogin = self userLogin ]! !

!QAvatarReplica methodsFor: 'private' stamp: 'bgf 7/24/2009 14:41'!
isNearASeat
	"Return the first-found seat within the avatar's sit-down auto-placement radius,
	 if we're within range of a seat."
	| basis |
	basis := self avatarHeight.
	"Favor seats in front of me where I can see'em."
	^ self isASeatWithin: (basis* 2.5)
		    of: (self globalPosition + (self globalTransform lookAt * basis * -0.7)).! !

!QAvatarReplica methodsFor: 'places' stamp: 'bgf 4/27/2009 21:23'!
isNearStance: stance
	"Return true if my (current-destination, or failing that my live) 
	 position and orientation are close to the given stance.  
	 This is used to count occupancy of places during census. "
	| here yaw |
	currentPlaceStance ifNotNil: [
		here := currentPlaceStance here.
		yaw := currentPlaceStance yaw.
	] ifNil: [
		here := self globalTransformReal translation.
		yaw := self globalTransformReal pitchYawRoll y
	].

	"See also #findSeatIn: , which must use the same nearness heuristics."
	^ (stance here - here) squaredLength < (3.6 * 3.6)
	    and: [ stance yaw within: 18 degreesOf: yaw ]
! !

!QAvatarReplica methodsFor: 'useful' stamp: 'bgf 3/15/2008 21:36'!
isOnGround

	"Rule of thumb... is the avatar currently resting on the ground?"
	^ ( (self altitude - self heightAboveGround) abs < 0.5) ! !

!QAvatarReplica methodsFor: 'accessing' stamp: 'ar 3/24/2009 10:22'!
isRecordingSession
	"True if the receiver is currently recording the session of the space it is in."
	^isRecordingSession! !

!QAvatarReplica methodsFor: 'accessing' stamp: 'bgf 7/24/2009 23:32'!
isRecordingSession: aBool
	"The receiver is currently recording the session of the space it is in."
	isRecordingSession = aBool ifFalse:[
		isRecordingSession := aBool.
		self root signal: #isRecordingSessionChanged with: self.
		label ifNotNil: [ label update ].
	].
! !

!QAvatarReplica methodsFor: 'private' stamp: 'bgf 2/14/2010 14:49'!
isSeatedAtSomething
	"Heuristic - do I appear to be seated in a table-gathering or at a workstation ?"
	posture = #Sitting ifFalse: [ ^ false ].
	self viaPlace ifNotNil: [ : via |
		via destinationFrame ifNotNil: [ : df | 
			((df isKindOf: QDestination)  and: [ df isCircleGathering ]) ifTrue: [ ^ true ].
			via hasCameraStance ifTrue: [ ^ true ].
	]].
	^ false
	! !

!QAvatarReplica methodsFor: 'accessing' stamp: 'bgf 7/25/2009 04:26'!
isSpeaking
	"Is this person speaking at the moment?"
	^ lastSpeech ifNotNil: [
			(Time milliseconds: (self island millisecondClockValue) since: lastSpeech) < 500 "ms"
		] ifNil: [ 	
			false
		]! !

!QAvatarReplica methodsFor: 'accessing' stamp: 'bgf 3/6/2007 20:32'!
islandURL
	"The absolute URL of the island we are in."
	^ islandURL! !

!QAvatarReplica methodsFor: 'accessing' stamp: 'bgf 3/6/2007 20:21'!
islandURL: aURLString
	"Supplied by the harness after space changes,
	 so the avatar has access to an absolute URL for the TIsland it occupies."
	islandURL := aURLString.
! !

!QAvatarReplica methodsFor: 'events' stamp: 'bgf 6/18/2008 00:43'!
keyStroke: event

	self root ifNil: [ ^ self ].

	"Allow some costume switches in a multi-costume (Meredith leftover.)"
	(self costume respondsTo: #selectCostume:) ifTrue: [
		event keyCharacter = $1 ifTrue:[self costume selectCostume:1].
		event keyCharacter = $2 ifTrue:[self costume selectCostume:2].
		event keyCharacter = $3 ifTrue:[self costume selectCostume:3].
		event keyCharacter = $4 ifTrue:[self costume selectCostume:4].
	].! !

!QAvatarReplica methodsFor: 'accessing' stamp: 'bgf 7/25/2009 06:09'!
label
	"My QAvatarLabel "
	^ label! !

!QAvatarReplica methodsFor: 'meetings' stamp: 'bgf 8/19/2009 02:09'!
leaderImposedPlace
	"The QPlace where the meeting-leader requires everyone to be."
	^ self meetingState ifNotNilDo: [ : meeting | 
		| assigned | 
		assigned := meeting meetingPlace.
		(assigned isNil 
		  or: [ meeting audienceStayPut ]
		  or: [ assigned isFollowMePlace ]  ) 
			ifTrue: [  assigned ] 
			ifFalse: [ nil ]]! !

!QAvatarReplica methodsFor: 'editing' stamp: 'bgf 7/22/2010 15:54'!
locateFrame: aFrame
	"Support for various 'Locate' menu commands'.
	 Bring the avatar to look at the given frame, and select it for emphasis."
	aFrame hiliteBriefly.
	self takeMeToFrame: aFrame.
	(self frameSelectionIncludes: aFrame) ifFalse: [
		self selectFrame: aFrame.
	].! !

!QAvatarReplica methodsFor: 'editing' stamp: 'bgf 8/26/2010 08:59'!
locateMaterial: aMaterial
	| mats frames |
	"Support for the 'locate' operation on the material editors."
	
	aMaterial isThemeMaterial ifTrue: [
		mats := self root instancesOfThemeMaterial: aMaterial
	] ifFalse: [
		mats := { aMaterial }.
	].
	mats size < 1 ifTrue: [
		^ self signal: #alertMessage with: 'This material is not used.'.
	].

	frames := self root find: [ : f | ( f myMaterials intersection: mats) notEmpty ].
	frames := ( frames collect: [ : f | f component editableObject ] ).
	frames := frames select: [ : f | f notNil ].
	
	frames ifNotEmpty: [ 
		self locateFrame: (frames first).
		
		self isEditingTemplate ifFalse: [
			self selectFrames: (frames select: [ : f | f isFromTemplate not ] ).
		] ifTrue: [
			self selectFrames: frames.
		].
		
	] ifEmpty: [
		self signal: #alertMessage with: 'Cannot locate any frames that use this material.'
	].
	! !

!QAvatarReplica methodsFor: 'editing' stamp: 'bgf 3/13/2010 08:40'!
locateSelection
	"If I have selected frames, go where I can see them."
	selectedFrames isEmptyOrNil ifFalse: [
		self takeMeToFrame: (selectedFrames anyOne)
	]. ! !

!QAvatarReplica methodsFor: 'places' stamp: 'jcg 9/16/2008 16:17'!
lookupPlaceMenuActions: aPlace

	"This is a trampoline between the client and the destination-frame of a place,
	 allowing the client to fetch place-menu description lists easily."

	^ ([aPlace destinationFrame] on: Error do: [ nil ])  ifNotNilDo: [ : df | 		
			df placeMenuActions: self
			].
! !

!QAvatarReplica methodsFor: 'editing' stamp: 'bgf 2/11/2010 17:40'!
makeGroup: frames event: cEvent
	"Action for the generic make-group menu command."
	self selectFrame: (
		(TGroup new) 
			consistingOfAll: (frames asArray);
			isFromTemplate: false;
			yourself)
! !

!QAvatarReplica methodsFor: 'initialize' stamp: 'bgf 1/18/2007 01:18'!
makeSnow

	| path ptxtr tframe |
	path := 'Content/Qwaq/Textures/Hide/snowflake.png'.
	ptxtr := TMaterial new loadTexture: path extension:#colorKey.
	ptxtr ambientColor: #(1.0 1.0 1.0 0.9).
	ptxtr diffuseColor: #(1.0 1.0 1.0 0.9).
	ptxtr := ptxtr asTFrame.
	ptxtr aspectFixed: false.
	ptxtr extent: 2@2.
	tframe := TParticleTxtr new initializeWithSize: 250.
	tframe rectangle: ptxtr.
	tframe setLifetimeRange: (4000 to: 6000).
	tframe setPositionRangeMin:(-30@0.0@-30) max: (30@10@30).
	tframe setVelocityRangeMin:(-0.4@-1.0@-0.4) max:(0.4@-4.0@0.4).
	tframe setAccelerationRangeMin:(0@0@0) max:(0@0@0).
	tframe startColor: #(1.0 1.0 1.0 0.9).
	tframe endColor: #(1.0 1.0 1.0 0.0).
	tframe translation: 0@10 @ -10.
	self addChild: tframe.
	tframe regenerateAll.! !

!QAvatarReplica methodsFor: 'useful' stamp: 'das 1/4/2007 14:22'!
makeWindow

	| trans win |
	trans := self globalTransform copy.
	trans translation: trans translation + ( trans lookAt * -6.0).
	win := TWindow new.
	win localTransform: trans.
	self root addChild: win.
	^ win.! !

!QAvatarReplica methodsFor: 'meetings' stamp: 'bgf 6/28/2009 10:21'!
meetingLeader
	"If this meeting is currently lead, return the leader avatar."
	self root ifNotNilDo: [ : space |
		space meetingState leaderLogin ifNotNilDo: [ : login |
			^ space allUsers detect: [: av | av userLogin = login ] ifNone: [ nil ]
		]].
	^nil! !

!QAvatarReplica methodsFor: 'meetings' stamp: 'bgf 7/4/2009 15:02'!
meetingPlace
	"The QPlace where the leader wants me to be."
	^ self meetingState ifNotNilDo: [ : ms | ms meetingPlace ]! !

!QAvatarReplica methodsFor: 'meetings' stamp: 'bgf 7/3/2009 13:18'!
meetingState
	"The Meeting State for the space I'm in; a QMeetingState."
	^ self root ifNotNilDo: [ : space | space meetingState ]! !

!QAvatarReplica methodsFor: 'private' stamp: 'hrs 8/19/2010 11:11'!
nearbyDropTarget
	"Answer the display panel drop target that I am heuristically at, or nil.
	Respects seating, workspaces that contain panels"
	| place landmarks |
	self avatarIAmWith ifNotNilDo: [:other | ^other nearbyDropTarget].
	landmarks := self currentLandmarks.
	place := self nearbyPlace: landmarks.
	(place isNil or: [place label = #Standing]) ifTrue: [
		^nil.
		"place :=  landmarks detect: [  : p | p isNear: self ] ifNone: [ ^nil ]."
	].
	"Special case for when we're at a blank target (which won't be among landmarks, but
	could have the same name as other stuff under root)"
	place destinationFrame ifNotNil: [:f |
		(f respondsTo: #displayCard:fromUser:) ifTrue: [^f]
	].
	"Find panels nested within workstations, under signs, etc."
	^self root findDropTargetNamed: place label
		landmarks: landmarks
		avatar: self
		within: 4! !

!QAvatarReplica methodsFor: 'private' stamp: 'bgf 7/3/2009 18:47'!
nearbyPlace: places
	"Report my current placement.
	 If in an assigned QPlace, return that.
	 (Return the viaPlace if any, since that is more likely to be the name-listed place;
	 if there is a viaPlace, then my actual place is just a chair inside it.)
	 Otherwise, find the first candidate place that I'm very near."

	^ self nominalPlace ifNil: [
			(places detect: [  : p | p isNear: self ] ifNone: [ nil ]) 
		].! !

!QAvatarReplica methodsFor: 'private' stamp: 'bgf 7/24/2009 16:05'!
nearbySeat
	"Return the nearest seat within the avatar's sit-down auto-placement radius,
	 I.e. the candidate we'll sit in if given the chance."
	| basis gt gp at |
	basis := self avatarHeight.

	"Favor seats in front of me where I can see'em.   MUST MATCH isNearASeat !!"
	gt := self globalTransformReal.
	at := gt lookAt.
	gp := gt translation  + (at * basis * -0.7).

	^ self nearestSeatWithinDistance: (basis* 2.5)
		  andDegrees: nil
		  ofStance: (Stance here: gp  there: (gp - (at  * basis))).
! !

!QAvatarReplica methodsFor: 'places' stamp: 'jjl 8/5/2008 11:27'!
nearestLandmark: places

	"Return the landmark entry (from the given list of space landmarks) 
	 closest to my current position."

	| distance nearest locale d stance |

	self root ifNil: [ ^nil ].	"Not in space anymore..."

	"Named landmarks are lists (symbol, friendlyName, (x,y,z), (rx, ry, rz))  
	 with an optional fifth element - a boolean sticky flag. "

	distance := Float infinity.
	nearest := nil.
	locale := self globalTransform translation.

	places do: [ : place |
		stance := (place destinationStanceFor: self).
		stance ifNotNil: [
			((d := locale distanceTo: stance here) < distance) ifTrue: [
				distance := d.
				nearest := place.
			]].
	].

	^ nearest! !

!QAvatarReplica methodsFor: 'private' stamp: 'bgf 7/26/2009 13:02'!
nearestSeatWithinBox: tbox
	"Return the nearest (free) sitting place to my current location,
	 if any are available within the given bounding box."
	| savePlace |
	tbox ifNil: [ ^ nil ].

	savePlace := currentPlace.
	[	
		currentPlace := nil.
		^ (self root bestSeatWithin: tbox nearest: (self globalPositionReal))
		    ifNotNilDo: [: seat |
			seat asPlaceAnyway
			].

	] ensure: [ currentPlace := savePlace ].
! !

!QAvatarReplica methodsFor: 'private' stamp: 'bgf 8/6/2009 22:07'!
nearestSeatWithinDistance: distance andDegrees: optionalAngle ofStance: aStance

	"Return the nearest (free) sitting place whose stance
	 is within the given angle and distance to aStance. 
	 I.e this finds seats that are close to (and looking towards) a focal view position."

	| seats  sdist curDist curPlace yaw savePlace |

	savePlace := currentPlace.
	[
		currentPlace := nil.
		seats := self root findNonNested: [ : frame | 
				(frame isSeat) and: [ frame isOccupied not ]].
	] ensure: [ currentPlace := savePlace ].

	seats isEmptyOrNil ifTrue: [ ^ nil ].
	sdist := distance * distance.
	curDist := sdist + Float epsilon.
	curPlace := nil.
	yaw := aStance yaw.

	"To save time, we consider the global position of the seat rather than the exact stance;
	 this is imprecise due to seating offsets but fine for our purposes,
	 since seating-destination seatingRadius is approximate at best due to model shapes."
	seats do: [ : seat |		
		| d | 
		d := (seat globalPositionReal - aStance here) squaredLength.
		d <= curDist ifTrue: [
		(optionalAngle isNil or: [ 
		 (seat globalTransformReal pitchYawRoll y within: optionalAngle degreesOf: yaw) 
		]) ifTrue: [
			curDist := d.
			curPlace := seat .
		]]].

	^ curPlace ifNotNil: [ curPlace asPlaceAnyway ]
! !

!QAvatarReplica methodsFor: 'menus' stamp: 'bgf 3/11/2010 17:30'!
newPlaceMarker
	"Create a place marker at my current position, and offer my user a text-editor to name it."
	| marker |
	marker := (self root placeMarkerClass) new.
	marker  adoptViewPosition: self.

	self captureCreateFrames: { marker} description: 'Landmark'.
	marker editLabel: self.

	^ marker! !

!QAvatarReplica methodsFor: 'places' stamp: 'bgf 7/3/2009 19:38'!
nominalPlace
	"The most-significant place I'm at... my viaPlace, if any,
	 or my currentPlace in any case.
	 If I'm diverted to a seat when going to a panel, the panel is my viaPlace.
	 If I sit at a large table, the table is my via place."
	^ self viaPlace ifNil: [ self currentPlace ]! !

!QAvatarReplica methodsFor: 'status' stamp: 'bgf 7/25/2009 01:30'!
noteActivitiesChange

	label ifNotNil:  [ label update ].

	"Activate this to see each change in the activity-set of each avatar."
	false ifTrue: [
		Transcript cr; show: '', self, ' activities: '.
		activities do: [ : sym | Transcript show: sym, ' ' ].
	].
	! !

!QAvatarReplica methodsFor: 'accessing' stamp: 'bgf 1/15/2008 01:25'!
notice
	"Nil, or a string matching my Org Services presence notice (e.g. 'Idle 3 minutes').
	 Meant to be transient, and to show presence notes that all users want to be aware of."
	^ notice! !

!QAvatarReplica methodsFor: 'accessing' stamp: 'bgf 7/24/2009 23:32'!
notice: aString
	"Nil, or a string matching my Org Services presence notice (e.g. 'Idle 3 minutes').
	 Meant to be transient, and to show presence notes that all users want to be aware of.

 	 If non-nil, will appear as a second line under the avatar name label.
	 This value must be orchestrated from the user's harness,
	 rather than coordinated with org services, as Croquet does not permit inter-island activity."

	notice ~= aString ifTrue: [
		notice := aString.
		self signal: #noticeChanged with: aString.
		label ifNotNil: [ label update ].
	].	
! !

!QAvatarReplica methodsFor: 'accessing' stamp: 'bgf 7/1/2009 20:08'!
notices
	"The array of (comma-separated) items in my presence-notice.
	 Empty array if notice is nil."
	^ notice ifNotNil: [ notice subStrings: ',' ] ifNil: [#() ]! !

!QAvatarReplica methodsFor: 'temp' stamp: 'ar 8/10/2009 14:55'!
on: eventName notify: observer
	"Notify the observer when the given event is triggered.
	Overridden to avoid registering the avatar user multiple times
	due to its automatic forwarding of events."
	self myEventMap ifNotNil:[:emap| emap at: eventName ifPresent:[:list|
		list do:[:each| each == observer ifTrue:[^observer]].
	]].
	^super on: eventName notify: observer.! !

!QAvatarReplica methodsFor: 'events' stamp: 'bgf 1/9/2010 12:18'!
onAvatarColorChanged
	| color |
	"Laser is built early during avatar setup,
	 it's a marker for when this sort of trigger is useful at all."

	laser ifNotNil: [
		color := userInfo avatarColor ifNil: [laser color].
		laser ifNotNil: [ laser color: color ].
		costume ifNotNil: [ costume colorize: color ].
		self arrowCursor color: color.

		self root signal: #avatarColorChanged with: self.
		self root signal: #avatarProfileChanged with: self.
	].
	label ifNotNil: [ label update ].! !

!QAvatarReplica methodsFor: 'events' stamp: 'bgf 4/4/2007 00:17'!
onBadgePictureChanged

	"The user info's badge-picture has changed."

	self badge: (userInfo badgePicture).

	"So upstream tools don't need to deal with the userInfo signal."
	self root signal: #badgePictureChanged with: self.! !

!QAvatarReplica methodsFor: 'events' stamp: 'bgf 1/9/2010 12:18'!
onInventoryChanged
	"A change in the card content of the inventory doc-shelf."
	| cd |

	inventory ifNil: [ ^ self ].	"Safety for space-transition sequences."
	inventory cards isEmptyOrNil ifTrue: [ ^ self ].

	cd := (inventory cards last).
	cd content ifNil: [ ^ self ].	"Probably a new card whose document has not yet called-back.	
							We'll get another shot; the objectCreated: callback signals change."

	"Send the persistent data for the newest card up to the harness."
	self signal: #inventoryCard  
		with: (TSceneWriter new storeSceneAsZip: inventory cards last sceneName: 'Inventory.xml').

	"And discard it... this is not the full inventory, just a blackboard for receiving additions."
	inventory removeCard: cd.
	cd obliterateLater.
! !

!QAvatarReplica methodsFor: 'events' stamp: 'bgf 7/25/2009 00:24'!
onProfileChanged
	"Generic change signal for personal-profile userInfo items 
	 like email, screen name,  personal summary."

	self objectName: self publicName.
	self root signal: #avatarProfileChanged with: self.
	label ifNotNil: [ label update ].! !

!QAvatarReplica methodsFor: 'events' stamp: 'bgf 2/14/2009 17:03'!
onSmallPictureChanged

	"The user info's picture has changed."
	(costume respondsTo: #setFace:) ifTrue: [
		costume setFace: (userInfo smallPicture).
	].! !

!QAvatarReplica methodsFor: 'events' stamp: 'bgf 2/19/2008 16:32'!
onSoundRegionChanged: newRegion

	"Allow the avatar region changes to be observable."
	self signal: #soundRegionChanged with: newRegion! !

!QAvatarReplica methodsFor: 'materials' stamp: 'bgf 8/30/2010 13:28'!
paintFrame: pickedFrame withMaterial: aMaterial
	| frame oldState newState |
	
	"On-island portion of the implementation of dropping a material onto a TFrame."
	
	self canModifyForum ifFalse: [ ^ false ].
	
	"Picking the nearest click-aware container makes the targeting
	 work just like the targeting for popping up a context menu."
	
	(pickedFrame ownerForEventType: EventPointerDown) ifNotNil: [ : owner |
		frame := owner.
	] ifNil: [
		frame := pickedFrame
	].

	frame component ifNotNil: [ : cmpnt |
		| mat target |
		target := cmpnt editableObject.
		(target canEditMaterial not or: [ target myMaterials isEmptyOrNil ]) 
		ifTrue: [ 
			QLogger logInfo: 'Target ', target, ' ', target canEditMaterial, ' ', target myMaterials notEmpty .
			^ self signal: #alertMessage 
				    with: 'That object cannot be painted.'
		]. 
		(target isFromTemplate and: [ self isEditingTemplate not ])  ifTrue: [
			^ self signal: #alertMessage 
				    with: 'You cannot paint this.\You are not editing the fixed scene.'.
		].
		
		mat := target myMaterials first.
		oldState := mat captureState.
		mat from: aMaterial.
		target checkAlpha.
		
		aMaterial isThemeMaterial ifTrue: [
			QActivityCapture action: #ThemeMaterialPaint.
			mat themeMaterial: aMaterial.
		] ifFalse: [
			QActivityCapture action: #MaterialPaint.
			mat themeMaterial: nil.
		].
		newState := mat captureState.
		
		self recordAction: #dropMaterialOntoFrame:oldState:newState:
			class: QAvatarReplica
			arguments: { target . oldState . newState } .
			
		
	].! !

!QAvatarReplica methodsFor: 'editing' stamp: 'bgf 2/9/2010 23:10'!
pasteSelection: evt
	"The basic 'Paste'  ctl-v operation for scene-editing... 
	 if we have a frames-scrap in the pasteboard,  instantiate a copy of it in the current space."
	| target |
	evt avatar: self.
	pasteboardScrap ifNil: [ 
		^ self signal: #alertMessage with: 'You have no copied frames to paste.'
		].

	pasteboardScrap staggerPositions.
	target := self singleSelectedFrame.
	(target notNil and: [target isEditableGroup]) 
	ifFalse: [
		target := (self root)
	].
	
	"Paste does a roundtrip through the harness-side avatar, 
	 which may perform resource manipulations to provide the backing documents."		
	self signal: #pasteFrames with: target! !

!QAvatarReplica methodsFor: 'private' stamp: 'bgf 8/4/2007 14:30'!
phase2Destroy

	super phase2Destroy.

	"User info is an expensive part, but is not a frame child, so it needs its own obliteration."
	userInfo ifNotNil: [userInfo obliterate. userInfo := nil].
! !

!QAvatarReplica methodsFor: 'accessing' stamp: 'bgf 8/29/2009 11:44'!
pitchYawRoll: v3

	"This epsilon-idempotence check helps us keep track 
	 of when goto-driving actually gives us something to do.  
	 By suppressing trivial changes here, we suppress transformChanged, 
	 which is tested in heartbeat:etc:... 
         to see when I've been moved by a step-update."

	super pitchYawRoll: v3.
	self poseShadow: (self altitude).
! !

!QAvatarReplica methodsFor: 'places' stamp: 'bgf 6/26/2009 16:02'!
placeBesideMe
	"The QPlace to take an avatar to bring it  beside me."

	^ besideMePlace ifNil: [

		besideMePlace :=  (QPlace	
			category: 'User'
			label: 'With ', (self publicName)
			frame: self
			selector: #besideMeStanceFor: 
			pov: self currentPlacePOV)
		isSticky: true;
		cameraSelector: #besideMeCameraStanceFor:;
		yourself
	].
! !

!QAvatarReplica methodsFor: 'places' stamp: 'bgf 6/22/2009 17:53'!
placeFacingMe

	"The QPlace to take an avatar to bring it facing me;
	 this is used when joining a person who is not in a predefined destination."

	^ QPlace	
		category: 'User'
		label: 'Facing ', (self publicName)
		frame: self
		selector: #facingMeStanceFor: 
		pov: 'First Person'
		posture: #Standing
	! !

!QAvatarReplica methodsFor: 'places' stamp: 'bgf 8/9/2009 15:49'!
placeForJoiningMe
	"The QPlace to take an avatar to bring it to me.
	 If I'm already a sticky viewing place, the other user also goes to that same place,
	 so he shares the view of what I'm looking at 
	(rather than materializing behind a wall, for instance).
	In other cases, the user is just brought to face me."

	self viaPlace ifNotNilDo: [ :via | ^ via ].

	self currentPlace ifNotNilDo: [ : cp |
		(cp sticky 				"Really a viewing/gathering position"
	 	and: [ cp isStatic not ] 		"I.e not just a resting-spot memory"
	 	and: [ cp isSeat not ]
	 	and: [ cp destinationFrame notNil ]
	 	and: [ cp destinationFrame isInAvatar not ])	
			"Going to the other-guy's gone-to person is confusing... 
			 as if you'd gone to the wrong person."
		ifTrue: [	
			^ cp
		]].
 
	^ self placeFacingMe
	! !

!QAvatarReplica methodsFor: 'places' stamp: 'bgf 3/15/2008 17:22'!
placesNearMe: placesArray

	"Return the sub-array of landmarks that I am currently very near."

	^ (placesArray select: [ : place |
			place isNear: self
		]) asArray

! !

!QAvatarReplica methodsFor: 'gestures' stamp: 'RJT 5/12/2010 20:30'!
playAnimation: name type: aType playType: aPlayType
	"atype = #skeletal or #vertex.  aPlayType = #cycle or #action"

	| |
	self costume contents playAnimation: name type: #skeletal playType: aPlayType.
	" The following code is removed because of the fix from howard to add default animation on new user entry.  I'm leaving this code here because it is a good example of how you could expire animations.
	(self costume contents respondsTo: #currentSkeletalAnimation) ifTrue: [
		activeAnimation := self costume contents currentSkeletalAnimation.
		aPlayType = #action ifTrue: [
			expires := activeAnimation scaledLength.
			(self future: expires * 1000) expireAnimation: activeAnimation.
		].
	]."! !

!QAvatarReplica methodsFor: 'accessing' stamp: 'bgf 8/29/2009 11:43'!
poseShadow: altitude

	| t pitch sin cos trans midHyp |
	(altitude < 25 and: [ altitude > (self heightAboveGround * 0.95) ]) ifTrue: [
		shadow visible: true; pickable: false.
		shadowTranslation ifNil: [ 
			"No hints about shadow placement; just place shadow at the ground plane."
			t := shadow translation.
			shadow translationX: t x y: (altitude-0.03) negated z: t z.
			^ self
		].

		"shadowTranslation is the shadow offset from my origin,
	        when I'm standing straight up on the ground.
		 My pitch is the angle between my body and the vertical;
	       altitude is my current height-above-ground,
	 	 measured from my origin (head) to the ground plane.
		 shadow is a TFrame child of mine, so it is in my transform,
	    	 and so would normally be at my 'virtual feet' when I pitch."

		pitch := self pitchYawRoll x.
		sin := pitch degreeSin.
		cos := pitch degreeCos.

		"Project the shadow down global-y from my origin, rather than along my y axis."
		trans := shadowTranslation x 
			      @ ((altitude *cos) negated + 0.03 + (shadowTranslation z * sin))
				@ ((shadowTranslation z * cos) + (sin * altitude)).

		"Now pull it back in so it is under my body rather than my origin."
		midHyp := (self avatarHeight * 0.33) * (pitch degreesToRadians tan) * cos.
		trans := trans - (0 @ (midHyp*sin) @ (midHyp*cos)).

		shadow translation: trans.
		shadow pitchYawRoll: (-90-pitch)@0@0.
		"Stretch the shadow-z when I'm leaning  (this works for upright avatars;
	        for the balloon-head avatar, not quite so nice."
		shadow extent: (shadowExtent x)@(shadowExtent y * (1+(2 * sin abs))).
		"Shadow fades with altitude, gone at +20"
		shadow material color: (VectorColor r:1 g:1 b:1 a: (0.5 * ((20-altitude/20.0) min: 0.99 max: 0))). 
		
	] ifFalse: [
		shadow visible: false.
	].! !

!QAvatarReplica methodsFor: 'private' stamp: 'bgf 6/9/2010 15:47'!
positionCarriedCards

	"Adjust he placement of any file cards I'm carrying to be front&center and in-sight."
	| z cvp cards y |

	cards := carrying ifNotNil: [ carrying find: [ : fc | fc isKindOf: QFileCard ] ].
	cards isEmptyOrNil ifTrue: [ ^ self ].

	cards doWithIndex: [ : card : index |	

		card visible: true.
		card extent: (QFileCard defaultExtent * 0.35).
		z := -2 * (45.0 / self viewAngle).
		y := z negated * (0.05 * (index+1)).
		cvp := self cameraViewpoint.
		card translation: (cvp transformPoint: (Vector3 x: 0 y: y z:  (z * (1+(index / 20.0))))).
		card orientation: (cvp orientation).
	].
! !

!QAvatarReplica methodsFor: 'menus' stamp: 'bgf 11/14/2009 23:41'!
postMenu: aTMenuFrame forLocation: globalPos
	"We override the inspace behavior and let the harness handle it."
	"Dismiss any old menus."
	self cancelMenu.

	self signal: #menuPosted:itemData: with: aTMenuFrame with: (aTMenuFrame itemData).

	"Store the menu by having the replica take it up as a carry.
	 It is invisible, since we intercept these and use CMenus instead.  
	 We keep it alive to receive callbacks, and store it so it can be deleted later."
	aTMenuFrame visible: false.
	self carrying: aTMenuFrame.
! !

!QAvatarReplica methodsFor: 'accessing' stamp: 'bgf 4/12/2008 20:30'!
posture: aSymbol
	"Official Qwaq Forums policy: 
		only people standing firmly on the ground,
	 	or sprightly flying dynamic gravity-scofflaw daredevils,
	 cast shadows. 
	 Lazy sit-around-on-their-spotty-backsides people, do not.
	 (I.e. shadows look funny in chairs.)"

	super posture: aSymbol.
	costume ifNotNil: [
		aSymbol caseOf: {
			[#Sitting] -> [costume removeChild: shadow].
		} otherwise: [
			costume addChild: shadow
		]]! !

!QAvatarReplica methodsFor: 'private' stamp: 'bgf 3/10/2010 23:02'!
preCapture: listOfFrames
	"Notify anything that needs it in the given frames,
	 that a copy, or delete, state-snapshot event is coming."

	"The edit box steals the frame's geometry,   so dismiss any in progress."
	listOfFrames do: [ : f | 
		f editBoxAroundMe ifNotNilDo: [ : box | box dismiss ].
		f parent ifNotNil: [ : p | p editBoxAroundMe ifNotNilDo: [ : box | box dismiss ]].

		"Attached displays need their documents pre-saved also.."
		f allFramesDo: [ : fc | self preCapture: (fc stickers) ].
	].

	listOfFrames do: [ : f |
		(f find: [ : fc | fc isKindOf: TEmbeddedApp ]) 
		 do: [  : app | 		
			app acceptForUser: (self sessionID). 
			app signal: #saveForUser with: (self sessionID) ].
	].! !

!QAvatarReplica methodsFor: 'places' stamp: 'bgf 5/6/2010 08:57'!
previewStance: viewExtent 
	"Stance for the avatar-chooser preview mirror-camera to see me."
	| pos root coi |
	
	root := self root.
	root ifNil:[ ^nil ].

	"This is specialized for the usually 1x3 tall-thin avatar-chooser preview attachment."
	pos := self translation - (0@1.75@0) + (self lookAt * -9).
	coi := self translation  - (0@1.75@0).

	^Stance here: pos there: coi! !

!QAvatarReplica methodsFor: 'accessing' stamp: 'bgf 6/22/2009 17:53'!
printOn: aStream
	"Include who..."
	aStream nextPutAll: 'AvatarReplica('; 
			 nextPutAll: (self publicName ifNil: ['']);
			 nextPut: $)! !

!QAvatarReplica methodsFor: 'private' stamp: 'bgf 2/9/2010 23:27'!
privateSetSelection: frameSet
	"One edit box per customer, please.
	 If the new selection is not what I'm already editing, then clear the old edit-box."
	(frameSet size = 1 
	 and: [ editBox notNil ] 
	 and: [ editBox contents = frameSet anyOne ]) 
	ifFalse: [
		self clearEditBox.
	] ifTrue: [
		"If the selection is a singleton in an edit box, it may be an implicit group."
		 (frameSet anyOne isKindOf: QTemporaryGroup) ifTrue: [ 
			^ super privateSetSelection: (frameSet anyOne children)
		].
	].
	super privateSetSelection: frameSet! !

!QAvatarReplica methodsFor: 'status' stamp: 'bgf 4/11/2008 16:23'!
privateStartActivity: aSymbol
	(activities includes: aSymbol) ifFalse: [
		activities add: aSymbol.
		self signal: #activityStarted with: aSymbol.
		^ true
	].
	^ false "No change"! !

!QAvatarReplica methodsFor: 'status' stamp: 'bgf 4/11/2008 16:23'!
privateStopActivity: aSymbol
	(activities includes: aSymbol) ifTrue: [
		activities remove: aSymbol.
		self signal: #activityStopped with: aSymbol.
		^ true
	].
	^ false "NO change"! !

!QAvatarReplica methodsFor: 'accessing' stamp: 'bgf 6/22/2009 18:07'!
publicName
	"The public name of the user I represent."
	^ userInfo ifNotNil: [ userInfo publicName ]
			    ifNil: [ 'User' ]! !

!QAvatarReplica methodsFor: 'meetings' stamp: 'bgf 11/20/2009 21:42'!
questionForTally: questionString secretUntilClosed: bSecret
	"Support for the 'Tally a Question' meeeting-menu action.
	 The string is the persistence-string of a QPollBarChart.
	  Instatiate it, place it, and prod everyone to vote on it."

	| poll data  |

 	data := (QChartData new) 
			setValues: (Array with: 0 with: 0);
			setLabels: { 'Yes'.  'No' };
			setDescriptions: { 'Yes, I agree.' . 'No, I disagree' };
			setColors: { (Color r: 0.1 g: 0.75 b: 0.15) asVectorColor. 
					     (Color r: 0.8 g: 0.2 b: 0.1) asVectorColor };
			minRange: 4;	
			yourself.
	poll := (QPollBarChart new)
			titleString: (questionString asString);
			isSecretUntilClosed: bSecret;
			data: data;
			yourself.

	self root placeInstantiatedObject: poll forAvatar: self.
	poll pushPoll: self.
! !

!QAvatarReplica methodsFor: 'editing' stamp: 'bgf 3/13/2010 10:17'!
recordAction: actionMakerSelector class: klass arguments: arguments
	"Relay an undoable acton  to the harness.
 	 The harness will do: 
		class  perform: actionMakerSelector   withArguments: arguments
	 expecting the result to be a new QUndoableAction, 
	 which is added to the user's undo-redo queue.

	 So the selector should be implemented (as a class method) by the given class, 
	 and should return a QUndoableAction based on the given arguments."

	self signal: #recordAction:class:arguments:
		with: actionMakerSelector
		with: klass
		with: arguments
! !

!QAvatarReplica methodsFor: 'editing' stamp: 'bgf 2/24/2010 16:52'!
redoDeleteFrames: frameCap
	"The frame capture has been re-done and now we want to undo that redo...
	  i.e. without recapturing anything, get rid of the frames defined in the capture."

	(self root find: [ : f |  frameCap includesOID: (f objectID) ]) 
		do: [ :  f | f reallyDelete: self ]
! !

!QAvatarReplica methodsFor: 'places' stamp: 'bgf 3/19/2010 18:51'!
refreshDefaultPOV: place
	"If going with another avatar and the place does not mandate a POV, then return to my default.
	 This is necessary to restore 'follower' avatars to their preferred POV after visiting a panel,
	 when a meeting is in all-follow-me mode.  The follower avatars don't have drive-started events
	 to kick them into normal viewing, so this substitutes for that."

	( place suggestedPOV isNil and: [ place avatarToFollow notNil ] ) ifTrue: [
		self suggestPOV: #Preferred forPlace: place.
	].
! !

!QAvatarReplica methodsFor: 'meetings' stamp: 'bgf 8/12/2009 19:52'!
refreshLeaderPlace
	"Re-enforce the binding to be in a leader-assigned place,
	 if leader-assigned placement is in force."
	self isMeetingLeader ifTrue: [
		self leaderImposedPlace ifNotNilDo: [ : place |
			self root allUsers do: [ : avatar | 
				(avatar = self or: [ avatar isMeetingLeader]) ifFalse: [ 
					avatar returnToMeetingPlace 
				] ] ] ].! !

!QAvatarReplica methodsFor: 'accessing' stamp: 'bgf 1/23/2009 07:54'!
registerVideoStream: streamDescriptionOrNil 
	"Register or unregister streamID for this avatar's user."
	self forumVideoStreamRegistry 
					streamDescription: streamDescriptionOrNil 
					forUser: userInfo userLogin 
					birthTime: self birthTime.
	self hasVideoStream: (streamDescriptionOrNil notNil).! !

!QAvatarReplica methodsFor: 'accessing' stamp: 'jcg 2/15/2008 19:50'!
registryStreamDescription
	^self forumVideoStreamRegistry streamDescriptionForUser: self userLogin! !

!QAvatarReplica methodsFor: 'accessing' stamp: 'bgf 4/25/2009 11:47'!
relativeCameraTransform
	"The relative transform between my origin and my camera basis position
	 (i.e. before third-person zoom adjustments are applied.)
	 This is the identity matrix, except when I have an overriding focal transform."

	self focalCameraTransform ifNotNilDo: [ : cft |
		"Must be get:-safe... no updating transforms implicitly, take what we've got."
		^ localTransform inverseByTranspose * cft
	].
	^ Matrix4x4 theIdentity! !

!QAvatarReplica methodsFor: 'carrying' stamp: 'bgf 4/23/2009 15:12'!
releaseCard: card
	"Stop carrying this card if I was (but don't change its transform)"
	carrying ifNil:[ ^card ].

	self forgetCarried: card.
	carrying removeChild: card.
	self framesCarried size = 0 ifTrue: [
		self removeChild: carrying.
		carrying := nil ].

	^card! !

!QAvatarReplica methodsFor: 'useful' stamp: 'hrs 8/18/2010 17:10'!
requestDisplay: cardName
	"True if I am at a drop target (e.g, a panel or poster) that is displaying the specified file card,
	otherwise false, but signal #requestDisplay on myself so that listeners can decide whether to
	ensureDisplay."
	self nearbyDropTarget ifNil: [
		self signal: #alertMessage with: 'Cannot find ', cardName.  "private listeners"
	] ifNotNilDo: [:target | | card |
		(target respondsTo: #card) ifTrue: [card := target card].  "posters"
		(target respondsTo: #contentCard) ifTrue: [card := target contentCard].
		card ifNotNil: [
			(QWebServices match: cardName against: card identifier caseSensitive: false) 
				ifTrue: [^true].  "we're done."
		].
		"Not displayed. Better ask the user before changing things.
		(Consider, e.g., a mobile user who made this request before even entering, and didn't
		know that there were people gathered around a different document at this location.)"
		self signal: #requestDisplay with: cardName.
	].
	^false.! !

!QAvatarReplica methodsFor: 'meetings' stamp: 'bgf 3/19/2010 18:39'!
returnToMeetingPlace
	"Re-enforce the binding to be in a leader-assigned place,
	 if leader-assigned placement is in force."
	self leaderImposedPlace ifNotNilDo: [ : place |
		self isMeetingLeader ifFalse: [
			(self isInMeetingPlace: place)
				ifFalse: [
					self goToPlace: place.
				] ifTrue: [
					self refreshDefaultPOV: place
				]]]! !

!QAvatarReplica methodsFor: 'accessing' stamp: 'bgf 12/4/2009 07:19'!
root: aSpace
	"Extended for script triggers."

	self stopScript: #onSoundRegionChanged:.
	self stopScript: #doorChange:.

	self root ifNotNilDo: [ : oldSpace |
		oldSpace meetingState removeWantsToSpeak: (self userLogin).
		heartbeatState = #superseded ifFalse: [
			oldSpace meetingState releaseLeader: (self userLogin).
		]].

	aSpace ifNotNil: [
		"My speaker tracks/signals region changes; 
		 ride atop that to make my own region changes observable to my harness."
		self startScript: #onSoundRegionChanged: when: { speaker . #regionChanged }. 
		self startScript: #doorChange:  when: { aSpace . #doorOpened }.
		self startScript: #doorChange:  when: { aSpace . #doorClosed }.
	].

	super root: aSpace.

	aSpace ifNotNil: [ 
		"Induce creation-registration before we reach avatar rendering."
		aSpace meetingState.
	].! !

!QAvatarReplica methodsFor: 'gestures' stamp: 'bgf 6/22/2009 17:51'!
sayNoGesture

	"A basic gesture for disagreement (bound to 'n' for no) - shake your head."
	| step t |

	"Status line reinforcement..."
	self root signal: #infoMessage with: (self userInfo publicName, ' does not agree').

	"Delegate the action to the costume if it wants."
	(costume respondsTo: #sayNoGesture) ifTrue: [
		costume sayNoGesture ifNotNil: [^self].
	].

	"Default is to shake your head back and forth..."
	head ifNil: [
		^ self
	].
	step := 100.
	t := 0.
	#(-25 25 25 -25 -25 25 25 -25 -25 25 25 -25) do: [ :angle |
		(head future: t) addRotationAroundY: angle.
		t := t + step.
	].
	(self future: t) restoreHeadOrientation.! !

!QAvatarReplica methodsFor: 'gestures' stamp: 'bgf 6/22/2009 17:51'!
sayYesGesture

	"A basic gesture for agreement (yes, bound to 'y' key) - nod your head."
	| step t |

	"Status line reinforcement..."
	self root signal: #infoMessage with: (self userInfo publicName, ' agrees').

	"Delegate the action to the costume if it wants."
	(costume respondsTo: #sayYesGesture) ifTrue: [
		costume sayYesGesture ifNotNil: [^self].
	].

	"Default is to nod your head a couple of times..."
	head ifNil: [
		^ self
	].
	step := 150.
	t := 0.
	#(-30 30 20 -20 -30 30 20 -20 -30 30) do: [ :angle |
		(head future: t) addRotationAroundX: angle.	
		t := t + step.
	].
	(self future: t) restoreHeadOrientation.! !

!QAvatarReplica methodsFor: 'accessing' stamp: 'bgf 2/6/2010 21:22'!
selectedFrame
	"This is the focal frame for old meredith magic-mapping - not the edit-cut-paste selection!!!!"
	^ selectedFrame.! !

!QAvatarReplica methodsFor: 'accessing' stamp: 'bgf 10/26/2006 13:33'!
selectedFrame: sf

	"Support for the magic map."
	selectedFrame := sf.! !

!QAvatarReplica methodsFor: 'accessing' stamp: 'das 4/18/2006 15:17'!
selectedPoint

	^ selectedPoint.! !

!QAvatarReplica methodsFor: 'accessing' stamp: 'bgf 10/26/2006 13:33'!
selectedPoint: sp

	"Support for the magic map."
	selectedPoint := sp.! !

!QAvatarReplica methodsFor: 'editing' stamp: 'bgf 3/5/2010 14:53'!
selectionSummary

	"One or two words summarizing the selection.
	 Generic name (simplified class name) if single item selected,
	 or countphrase (e.g  ' 3 items selected '  if multiple."

	^ selectedFrames size caseOf: {
		[0] -> [ nil ] .
		[1] -> [ selectedFrames anyOne genericName ] .
	} otherwise: [
		'', (selectedFrames size), ' Items'.
	].! !

!QAvatarReplica methodsFor: 'accessing' stamp: 'bgf 5/19/2008 09:09'!
sessionHex
	"The session ID as a hex string.
 	 (Use as the 'forSession: ' argument of QActivityCapture recording in island-side methods."
	^ sessionID ifNotNil: [sessionID hex] ! !

!QAvatarReplica methodsFor: 'status' stamp: 'bgf 7/24/2009 23:29'!
setActivities: arrayOfSymbols
	"End all activities except the given ones,  
	 and start the given ones if I'm not already doing them.

	Must be set only under the control of the QAvatarUser,
	which maintains a harness-side cache/copy of the set to minimize futures."

	| prev chg |
	chg := false.
	prev := activities copyWithoutAll: arrayOfSymbols.
	prev do: [ :sym | chg := chg | (self privateStopActivity: sym) ].
	arrayOfSymbols do: [ :sym | chg := chg | (self privateStartActivity: sym) ].
	chg ifTrue: [
		self noteActivitiesChange.
	].! !

!QAvatarReplica methodsFor: 'status' stamp: 'bgf 7/24/2009 23:29'!
setActivity: aSymbol
	"End all activities except the given one,  and start the given one if I'm not already doing that.

	Must be set only under the control of the QAvatarUser,
	which maintains a harness-side cache/copy of the set to minimize futures."

	| prev chg |
	chg := false.
	prev := activities copyWithout: aSymbol.
	prev do: [ :sym | chg := chg | (self privateStopActivity: sym) ].
	chg := chg | self privateStartActivity: aSymbol.
	chg ifTrue: [
		self noteActivitiesChange.
	].! !

!QAvatarReplica methodsFor: 'editing' stamp: 'bgf 2/7/2010 08:27'!
setCopiedFramesScrap: capturedFrames
	"Assign the given capturedFrameList to be the current pasteboard-scrap."
	self signal: #copyFrames with: capturedFrames.
	pasteboardScrap := capturedFrames.
	deletionScrap := nil.
! !

!QAvatarReplica methodsFor: 'private' stamp: 'bgf 12/23/2009 09:15'!
setCurrentPlaceCleanup: placeFrame
	"Don't remain logically attached to places that get removed.
	 This is called out into a separate future-sent method because 
	 the methods that call it sometimes arise from clearCurrentPlace,  
	 which is stopped here."

	self stopScript: #clearCurrentPlace.
	(placeFrame notNil and: [ placeFrame = self destinationFrame ] ) 
	 ifTrue: [
		self startScript: #clearCurrentPlace when: {placeFrame . #deleted}.
		self startScript: #clearCurrentPlace when: {placeFrame . #destroyed}.
	].! !

!QAvatarReplica methodsFor: 'initialize' stamp: 'bgf 4/8/2008 17:41'!
setLocalTransform: trans

	"Initializing the local transform after space arrival."
	super setLocalTransform: trans.
	self clearCurrentPlace.
! !

!QAvatarReplica methodsFor: 'menus' stamp: 'bgf 3/22/2010 12:04'!
setMeetingPlace: place
	"Declare the given place to be the meeting place."
	self meetingState ifNotNilDo: [ : meeting | 
		meeting meetingPlace: place.
		(meeting audienceStayPut or: [place notNil and: [place avatarToFollow = self]]) 
		 ifTrue: [
			self refreshLeaderPlace.
		] ifFalse: [			
			"This is the more casual hey-come-see action."
			place ifNotNil: [
				self bringEveryone: place includingMe: (self isMeetingLeader not).
			].
		].

		"If the meeting-place defines a leader position, go there if I'm the leader."
		(place notNil and: [ self isMeetingLeader ]) ifTrue: [
			self goToLeaderSiteIfAny: place
		]].! !

!QAvatarReplica methodsFor: 'meetings' stamp: 'bgf 8/12/2009 21:22'!
setMeetingPlaceWithMe
	"Turn on All-Follow-Me for everyone."
	self isMeetingLeader ifTrue: [	
		self setMeetingPlace: (self placeBesideMe).
		].
! !

!QAvatarReplica methodsFor: 'meetings' stamp: 'bgf 7/4/2009 09:46'!
setMeetingWhereIAm
	"Declare my current place to be the meeting place."
	self nominalPlace ifNotNilDo: [ : place |
		self setMeetingPlace: place
	].! !

!QAvatarReplica methodsFor: 'private' stamp: 'bgf 7/4/2009 16:44'!
setViaPlace: aPlace
	"The via-place is the logical gathering place or viewing destination;
	 it remembers the intent of a goto when we're diverted into a seating place."
	currentPlaceVia := aPlace.
	(currentPlaceVia notNil and: [ currentPlaceVia avatarToFollow notNil ]) 
	 ifTrue: [
		self startFollowingPlace: currentPlaceVia.
	].! !

!QAvatarReplica methodsFor: 'initialize' stamp: 'bgf 1/9/2010 12:18'!
setupInventoryHolder
	"Make the document shelf we use to handle additions to user's inventory from in-space.
	 This is unseen, and contains cards only until the harness detects, copies, and removes them
	 (ie. it is a work-area for the cards representing copy-to-inventories in-progress."

	inventory ifNil: [
		inventory := QDocumentShelf new.
	].
	inventory 
			 objectName: 'Inventory';
			 visible: false; 
			 basicExtent: 0.14@0.105;		"See also positionInventory:"
			 numVisibleSlots: 2;
			 numSlots: nil.	"Infinite"

	self addChild: inventory.
	self runScript: #onInventoryChanged when: { inventory . #contentChanged }.! !

!QAvatarReplica methodsFor: 'accessing' stamp: 'bgf 6/29/2007 16:20'!
shadow
	"The TRectangle we place on the ground.
       Costume configureAvatar: should set the extent and x-z (y is dynamic.)"
	^ shadow! !

!QAvatarReplica methodsFor: 'private' stamp: 'bgf 8/20/2009 18:32'!
shouldRenderOther: ogl
	| viewer |
	"Should I render to users other than my own client?
	 Do not render if I would obstruct the ogl-client's focal view of a place we're both at.
	 This allows gatherings a lot of freedom in how to arrange people.
	 For example, avatars at the same panel will not obstruct each other's view,
	 regardless of camera-distance view-aspect placement details."

	ogl avatar isSticky ifFalse: [ ^ true ].
  	ogl avatar currentThirdPerson > 0.5 ifTrue: [ ^ true ].
	self currentPlace ifNil: [ ^ true ].

 	viewer := self root avatarForSession: (ogl view sessionID).	"This client's replica (not as far-ref)."
	viewer ifNil: [ ^ true ].

	"The dot: below prevents this from applying
	  to people around a circular-gathering when my client-user is in first-person;	
	  only avatars looking roughly my way are filtered."

	(viewer currentPlace = self currentPlace 
	  and: [ (ogl avatar localTransform lookAt dot: (self globalTransformReal lookAt)) > 0.5] )	
	 ifTrue: [
		^ false  "We're together looking at the same focal content, don't need to see each other."
	].

	^ true! !

!QAvatarReplica methodsFor: 'private' stamp: 'bgf 8/20/2009 18:27'!
shouldRenderSelf: ogl
	"Should I render myself to my own user?
	  Do not render me in first-person 
	 (unless the camera is twisted around a lot; I may be trying to see myself.)
	 This affords leniency for placement at viewpoints, 
 	  and allows more use of camera lag without driving-wink."

	ogl avatar focalCameraTransform ifNotNil: [
		ogl avatar relativeCameraTransform pitchYawRoll y abs > 30 ifTrue: [ ^ true ].
		].

	^ ogl avatar currentThirdPerson > 0.8		"I.e at or nearly at first-person."

! !

!QAvatarReplica methodsFor: 'events' stamp: 'bgf 11/13/2008 12:45'!
showFloatMessage: label forEvent: event

	"Show a mouse-over floating-tag label ...
	 unless the event is beyond the warp-to distance for the event target."

	(self warpPlaceUnderMouse: event) ifNil: [ 
		self signal: #floatMessage with: label
	].
! !

!QAvatarReplica methodsFor: 'private' stamp: 'bgf 11/7/2009 18:25'!
sitWhereIAm
	"Support for 't'-key posture toggling.  Sit yer ass down."
	QActivityCapture action: #SitWhereIAm forSession: (self sessionHex).
	posture = #Sitting ifFalse: [
		"force the posture in case a drive/fall intervenes before the goto begins."
	
		self nearbySeat ifNotNilDo: [ : seat |
			(self future:10) posture: #Sitting.
			self goToPlace: seat
		] ifNil: [
			self floatingInfo: 'There is no seat nearby'.
		]].
	! !

!QAvatarReplica methodsFor: 'places' stamp: 'bgf 8/9/2009 14:37'!
sittingPlace
	"If either my via place or my current place is a seat, return it."
	self currentPlace ifNotNil: [ : cp | 
		cp isSeat ifTrue: [ ^ cp ].
	].
	self viaPlace ifNotNil: [ : vp | 	
		vp isSeat ifTrue: [ ^ vp ]
	].
	^ nil
! !

!QAvatarReplica methodsFor: 'gestures' stamp: 'bgf 6/22/2009 17:50'!
smallBowGesture

	"A basic gesture - slow forward head-bow.  Bound to 'b' in driving mode."
	| step t |

	"Status line reinforcement..."
	self root signal: #infoMessage with: (self userInfo publicName, ' bows').

	"Delegate the action to the costume if it wants."
	(costume respondsTo: #smallBowGesture) ifTrue: [
		costume smallBowGesture ifNotNil: [^self].
	].

	"Default is just a Zen head bow..."
	head ifNil: [
		^ self
	].
	step := 120.
	t := 0.
	#(-5 -10 -10 -10 -10 -10 -2 -1 1 2  10 10 10 10 10 5) do: [ :angle |
		(head future: t) addRotationAroundX: angle.
		t := t + step.
	].
	(self future: t) restoreHeadOrientation.! !

!QAvatarReplica methodsFor: 'accessing' stamp: 'bgf 11/9/2006 19:22'!
smallPicture
	"Convenience accessor for userInfo #smallPicture"
	^ userInfo smallPicture! !

!QAvatarReplica methodsFor: 'accessing' stamp: 'bgf 5/24/2007 12:26'!
smallPictureSHA

	"Return the SHA of the current face texture, or nil.
	 Used for fast access from the client."
	| |

	^ self smallPicture ifNil: [ nil ] ifNotNilDo: [:txtr | txtr sha ]! !

!QAvatarReplica methodsFor: 'accessing' stamp: 'bgf 10/27/2006 23:32'!
speaker
	"My vocal annunciator"

	^ speaker! !

!QAvatarReplica methodsFor: 'events' stamp: 'bgf 11/23/2006 11:05'!
spin360
	
	"Pirouette the costume."
	self spin360:360.! !

!QAvatarReplica methodsFor: 'events' stamp: 'jcg 4/16/2007 15:30'!
spin360: angle

	"Pirouette the costume one step, until done."
	angle <= 0 ifTrue:[^self].
	(costume ifNil: [^self]) addRotationAroundY:5.
	(self future:100)spin360:angle-5.! !

!QAvatarReplica methodsFor: 'places' stamp: 'bgf 4/25/2009 12:33'!
stanceForPlace: aPlace viewExtent: ext

	"Return the destination stance I would assume to go to the given place,
	 for an owner view extent as given.    Harness get:-safe, and must stay that way. "
	| save |
	save := viewExtent.
	[
		viewExtent := ext.
		^ (aPlace destinationStanceFor: self)
	] ensure: [
		viewExtent := save.
	].! !

!QAvatarReplica methodsFor: 'private' stamp: 'bgf 4/21/2009 15:11'!
standUpStance

	"This is the destination transform for the synthetic place I go to 
	 when I simply stand up from a seated posture."
	| trans |
	trans := self standUpTransform.
	^Stance here: trans translation there: (trans translation - trans lookAt)
	! !

!QAvatarReplica methodsFor: 'private' stamp: 'bgf 2/17/2010 18:49'!
standUpTransform
	"This is the destination transform for the synthetic place I go to 
	 when I simply stand up from a seated posture (using the 'Stand' shortcut or seat-click.)"

	| newPt ground rayPt gt |
	gt := self globalTransform.
	self isSeatedAtSomething ifFalse:  [
		"See if we can stand up by stepping forward..."
		newPt := gt orientation transformPoint: ((0@0@ (-2.2))).
		rayPt := gt transformPoint: ((0@(0)@(-2.2))).
	
		"But don't do that if there appears to be something in front of us."
		ground := self root groundHeightAt: (rayPt).
		(ground) > (self altitude - (self heightAboveGround) + 2) ifTrue: [
			newPt := rayPt := nil.
		]].

	newPt ifNil: [ 
		"From a circular-gathering (including all tables), stand back and to the right."
		newPt := gt orientation transformPoint: ((2@(seatToEyes-footToEyes)@ (1.2))).
		rayPt := gt transformPoint: ((2@(seatToEyes-footToEyes)@(1.2))).
		ground := self root groundHeightAt: (rayPt).
	].

	"Anticipate where my feet will fall."

	newPt := self translation + newPt.

	^ gt clone translation: (newPt x) @ (ground + self footToEyes) @ (newPt z).

	! !

!QAvatarReplica methodsFor: 'private' stamp: 'bgf 2/27/2009 09:43'!
standWhereIAm
	"Support for 't'-key posture toggling.  Stand up from seated position."

	QActivityCapture action: #StandWhereIAm forSession: (self sessionHex).
	posture = #Sitting ifTrue: [
		(self future:10) posture: #Standing.	"just after this burst"
		self goToPlace: self standingPlaceWhereIAm
	]! !

!QAvatarReplica methodsFor: 'private' stamp: 'jjl 8/5/2008 13:32'!
standingPlaceWhereIAm

	^ QPlace	
		category: 'User'
		label: 'Standing'
		frame: self
		selector: #standUpStance 
		sticky: false
		posture: #Standing
	! !

!QAvatarReplica methodsFor: 'status' stamp: 'bgf 7/24/2009 23:29'!
startActivity: aSymbol

	"Must be set only under the control of the QAvatarUser,
	which maintains a harness-side cache/copy of the set to minimize futures."

	(self privateStartActivity: aSymbol) ifTrue: [
		self noteActivitiesChange.
	].! !

!QAvatarReplica methodsFor: 'private' stamp: 'bgf 3/19/2010 16:12'!
startCarryingNewFileCard: aFileCard
	"A new filecard from the harness is to be placed visibly in front of the avatar
	 and added to Carry, so it stays with me until I dispose of it."
	|  |
	self startCarrying: aFileCard.
	self future positionCarriedCards.

! !

!QAvatarReplica methodsFor: 'private' stamp: 'bgf 7/4/2009 16:42'!
startFollowingPlace: aPlace
	"Move when the place's destination frame moves, to follow it."
	self stopScript: #doFollowPlace:.
	aPlace destinationFrame ifNotNil: [
		self runScript: #doFollowPlace: 
			withArguments: {aPlace} 
			when: {aPlace destinationFrame . #frameChanged}.
	].! !

!QAvatarReplica methodsFor: 'useful' stamp: 'hrs 4/27/2010 09:50'!
startPanel
	"Start the panel that I am heuristically at, answering the place (so clients can choose to formally
	 go there), or nil."
	^self nearbyDropTarget ifNotNilDo: [:f | 
		(f respondsTo: #startRunning:) ifTrue: [
			f startRunning: self.
			f asPlace.
		].
	].
! !

!QAvatarReplica methodsFor: 'useful' stamp: 'hrs 4/27/2010 17:03'!
startPanelLabel
	"startPanel as label"
	^self startPanel ifNotNilDo: [:place | place label].
! !

!QAvatarReplica methodsFor: 'private' stamp: 'bgf 3/4/2009 12:27'!
stickyValueForPlace: aPlace
	"The  goto:sticky:etc:  sticky argument when going to this place.
	 May be a defined point of view.
	 When a place is marked sticky without a POV, use my current thirdPerson value.
	
	 (The Croquet-Level driving system equates 'specifying a POV' with 'sticky',
	 but we want some non-falling places that leave the POV unmodified.)  "

	^ aPlace suggestedPOV ifNil: [ aPlace sticky ifTrue: [ self thirdPerson ]]! !

!QAvatarReplica methodsFor: 'step' stamp: 'bgf 7/24/2009 21:11'!
stillAlive: clientUTC
	"Routine heartbeat... 
	 extended to notice standing still."

	label ifNotNil:  [ label update ].
	super stillAlive: clientUTC.
	self checkRestingPlace.! !

!QAvatarReplica methodsFor: 'status' stamp: 'bgf 7/24/2009 23:29'!
stopActivity: aSymbol

	"Must be set only under the control of the QAvatarUser,
	which maintains a harness-side cache/copy of the set to minimize futures."

	(self privateStopActivity: aSymbol) ifTrue: [
		self noteActivitiesChange.
	].! !

!QAvatarReplica methodsFor: 'status' stamp: 'bgf 7/24/2009 23:29'!
stopAllActivities

	"Must be set only under the control of the QAvatarUser,
	which maintains a harness-side cache/copy of the set to minimize futures."

	| prev |
	prev := activities clone.
	activities := IdentitySet new.
	prev do: [ : sym |
		self signal: #activityStopped with: sym.
	].
	self noteActivitiesChange.! !

!QAvatarReplica methodsFor: 'editing' stamp: 'bgf 2/4/2010 14:27'!
stopEditingTemplate
	"Notice that the fixed-scene editing state has been dropped.
	 Since we cannot select fixed-scene items in this state, we drop those items."

	(self selectedFrames anySatisfy: [ : f | f isFromTemplate ]) ifTrue: [
		self clearFrameSelection
	].! !

!QAvatarReplica methodsFor: 'useful' stamp: 'hrs 4/29/2010 10:37'!
stopPanel
	"Stop the panel that I am heuristically at, answering the place, or nil."
	^self nearbyDropTarget ifNotNilDo: [:f | 
		(f respondsTo: #stopRunning:action:) ifTrue: [
			f stopRunning: self action: #StopAppButton.
			f asPlace.
		].
	].
! !

!QAvatarReplica methodsFor: 'useful' stamp: 'hrs 4/27/2010 17:04'!
stopPanelLabel
	"stopPanel as label"
	^self stopPanel ifNotNilDo: [:place | place label].
! !

!QAvatarReplica methodsFor: 'places' stamp: 'bgf 7/12/2009 11:24'!
suggestPOV: aNameString forPlace: aPlace
	"The string is one of #firstPerson, #near, #normal, #far;
	 use this to trigger a change in the users viewpoint,
	 when (for instance) joining a gathering."

	self signal: #suggestPOV:forPlace: with: aNameString with: aPlace! !

!QAvatarReplica methodsFor: 'places' stamp: 'bgf 7/23/2010 13:28'!
takeMeToFrame: frameRef
	"If the  frame defines a destination for my avatar, go to it."

	frameRef ifNil: [ ^ self ].

	self allowedToMove ifFalse: [ 	
		self signal: #statusMessage with:  'The leader is controlling movement'.
		^ self
	].

	self signal: #statusMessage with: 'Locating ', (frameRef labelForBrowser).
	frameRef asPlace ifNotNilDo: [ :place | 
		place destinationFrame ifNotNilDo: [ : destFrame |
			destFrame visible ifFalse: [ 
				"In case it is in an attaching-manager like a tabbed board."
				destFrame surfaceAttachment ifNotNilDo: [ :sa |
					sa asPlace ifNotNil: [ 
						self signal: #alertMessage 
							with: 'The ', (destFrame genericName), ' is not visible.'.
						^ self takeMeToFrame: sa ] ].
			]].
		^ self goToPlace: place ].

	frameRef visible ifFalse: [ 
		self signal: #alertMessage with: 'The ', (frameRef genericName), ' is not visible.'.
		"In case it is in an attaching-manager like a tabbed board."
		frameRef surfaceAttachment ifNotNilDo: [ :sa |
			sa asPlace ifNotNil: [ 
				self signal: #alertMessage 
					with: 'The ', (frameRef genericName), ' is not visible.'.
				^ self takeMeToFrame: sa ] 
		]].

	"Fallthrough - no QPlace is defined, but there may be a destinationTransformFor:"
	self clearCurrentPlace.

	frameRef ifNil: [ ^ self ].

	(frameRef root = self root) ifFalse: [
		self signal: #alertMessage with: '', (frameRef labelForBrowser), ' is not visible.'.
		^ self
	].

	^ self goToPlace: frameRef locateMePlace.
	
	
! !

!QAvatarReplica methodsFor: 'materials' stamp: 'ar 4/16/2011 18:50'!
texturePalette
	"Returns the array of TFileTextures that describe the textures 
	 to make available for selection in Material editing pane.
	 This island-side method finds the user-provided textures used in this forum
	 that should be offered in the palette."
	
	| textures imageDocs suite |
	texturePalette ifNotNil: [ ^ texturePalette ].

	"Find all the routine resource file-textures in the island, 
	 and include them in the material editor textures list."

	textures := self island textureResources.
	imageDocs := self island imageDocumentResources.
	suite := OrderedCollection new.
	
	textures do: [ : txtr |
		txtr isValidRef ifTrue: [
			| fname |
			fname := txtr fileName.
			(fname notNil 
			 and: [ (fname includesSubString: 'Qwaq') not 
			 and: [ (fname beginsWith: 'Content') not ]					"No builtin textures (Palette gets the ones we want)"
			 and: [ (fname includesSubString: '/screens/') not ]			 "No app-previews"
			 and: [ (fname includesSubString: 'sharedDocuments/') not ]  "Mostly to omit slide-show slides"
			 and: [ (fname beginsWith: '/') not] ] )						 "Nothing from personal-files (avatar-images)"
			ifTrue: [
				(suite noneSatisfy:  [ : t | (t fileName) = fname ]) 
				 ifTrue: [
					suite add: txtr.
				].
			]]].

	imageDocs do: [ : doc |
		doc isValidRef ifTrue: [
			| fname txtr |
			txtr := doc fileTexture.
			fname := txtr fileName.
			(fname notNil 
			 and: [ (fname includesSubString: 'Qwaq') not ]
			 and: [ (fname beginsWith: 'Content') not ] )
			ifTrue: [
				(suite noneSatisfy:  [ : t | (t fileName) = fname ]) 
				 ifTrue: [
					suite add: txtr.
				].
			]]].

	^ texturePalette := (suite asArray)! !

!QAvatarReplica methodsFor: 'accessing' stamp: 'bgf 6/30/2009 12:42'!
thirdPerson: aFloat

	"Change to the relation between camera and avatar...
	 we have to update the screen-side accessory positions to keep them in place;
	 see viewportChanged."

	thirdPerson = aFloat ifFalse: [
		super thirdPerson: aFloat.
		self updateFocalTransform.
		self viewportChanged.
	].! !

!QAvatarReplica methodsFor: 'menus' stamp: 'crl 4/22/2010 15:58'!
toggleCamera
	camera toggle! !

!QAvatarReplica methodsFor: 'gestures' stamp: 'bgf 4/12/2008 21:48'!
togglePosture
	"Switch posture between sitting and standing"
	posture caseOf: {
		[#Standing] -> [ self sitWhereIAm ].
		[#Hovering] -> [ self sitWhereIAm ].
		[#Sitting] -> [ self standWhereIAm ].
	} otherwise: [].! !

!QAvatarReplica methodsFor: 'status' stamp: 'bgf 4/11/2008 17:58'!
transcriptActivities
	"Activate this to see each change in the activity-set of each avatar."
	false ifTrue: [
		Transcript cr; show: '', self, ' activities: '.
		activities do: [ : sym | Transcript show: sym, ' ' ].
	].
	! !

!QAvatarReplica methodsFor: 'places' stamp: 'bgf 9/1/2009 15:37'!
turnTo: aPosition
	"Turn to face in the direction of the given global-position (only in the x-z plane)"
	| myPos bearing xform |
	aPosition ifNil: [ ^ nil ].

	myPos := self translation.
	bearing := (myPos x @ myPos z) bearingToPoint: (aPosition x @ aPosition z).
	
	xform := Matrix4x4 
				translation: myPos
				pitchYawRoll: 0@bearing negated@0.

	self goto: xform sticky: false label: 'Turn'.
! !

!QAvatarReplica methodsFor: 'editing' stamp: 'bgf 3/5/2010 09:49'!
undoDeleteFrames: frameCap

	self selectFrames: (frameCap restoreForAvatar: self ).! !

!QAvatarReplica methodsFor: 'goto' stamp: 'bgf 8/10/2009 20:41'!
updateFocalTransform
	"Refresh my focal transform from my current place.
	 (If the place offers a camera transform, then it is used to make my view
	 provide a direct view on the destination, even if the basic avatar placement is a bit oblique.)"

	| newFocalTransform ct mt dpitch theta |
	newFocalTransform := nil.

	self nominalPlace ifNotNilDo: [ : place |
		place hasCameraStance ifTrue: [
			newFocalTransform :=
				(place cameraStanceFor: self) ifNotNilDo: [ : cstance | cstance transform ].
		]].

	"If we're zooming back from a  camera pitch 
	 (e.g. looking down at a lectern) 
	 recover  avatar-forward (horizontal) camera behavior quickly."
	(newFocalTransform notNil and: [ thirdPerson > 0.4 ]) ifTrue: [
		ct := newFocalTransform pitchYawRoll x.
		mt := self globalTransform pitchYawRoll x.
		dpitch := (ct - mt).
		theta := (dpitch * (thirdPerson min: 1.0)).
		theta  abs > 0.01 ifTrue: [
			newFocalTransform 
				addRotationAroundX: (theta negated min: 90 max: -90).
		]].

	self focalCameraTransform: newFocalTransform.
! !

!QAvatarReplica methodsFor: 'hierarchy' stamp: 'bgf 8/17/2008 10:29'!
updateTransform

	super updateTransform.
	"For place assignment tracking.."
	lastMovedTime := (self island totalSeconds).! !

!QAvatarReplica methodsFor: 'accessing' stamp: 'bgf 10/26/2006 13:37'!
userInfo
	^ userInfo! !

!QAvatarReplica methodsFor: 'accessing' stamp: 'bgf 4/21/2008 12:26'!
userLogin
	"The login (account) name of this user."
	^ userInfo ifNotNil: [ userInfo userLogin ]! !

!QAvatarReplica methodsFor: 'goto' stamp: 'bgf 8/9/2009 17:28'!
viaPlace
	"If I am at a place that one reaches 'via' an outer place,
	 then return that outer more official place (typically a seating volume, or a display)."

	currentPlace ifNil: [ ^ nil ].
	"The search is costly.
	 Only do it once for a given currentPlace, caching the result.
	 So: currentPlaceVia nil means not yet computed, 	
	      currentPlaceVia #None means computed as nil."
	currentPlaceVia ifNotNil: [ 
		^ (currentPlaceVia = #None) ifTrue: [ nil ] ifFalse: [ currentPlaceVia ] ].
	currentPlaceVia := #None.

	"If we're in follow-the-leader, continue to do so."
	self leaderImposedPlace ifNotNilDo: [ : place | 
		(self isMeetingLeader not 
		  and: [ place ~= currentPlace ]
		  and: [ place avatarToFollow notNil ]) ifTrue: [ 
			^ currentPlaceVia := place
		]].

	"The via is the 'greater' place at which we are seated,
	 so if not seated, we have nothing to do."
	currentPlace isSeat ifFalse: [ ^ nil ].
	self findCurrentViaPlace ifNotNilDo: [ : vp |
		^ currentPlaceVia := vp
	].
	^ nil! !

!QAvatarReplica methodsFor: 'private' stamp: 'bgf 3/30/2008 03:58'!
viewAngle: fovy extent: extent

	"Harness notification of the user's fovy and window extent."

	(fovy notNil and: [extent notNil]) ifTrue: [
		super viewAngle: fovy extent: extent.
		self viewportChanged.
	].! !

!QAvatarReplica methodsFor: 'private' stamp: 'bgf 3/30/2008 05:20'!
viewportChanged
	"The screen area or camera position has changed...
	 arrange to reposition my screen-side accesories."

	lastViewportChange ifNil: [	"I.e. no future-train underway."
		(self future: 150) doViewportUpdate.
	].
	lastViewportChange := self island millisecondClockValue.

	"See doViewportUpdate... we don't do intermediate updates while
	 a camera scroll or window resize is underway,
	 instead we wait until things have settled,
	 to avoid jitter-jumpy future-send-paced tracking of camera scroll."
! !

!QAvatarReplica methodsFor: 'accessing' stamp: 'bgf 7/24/2009 23:31'!
visible: aBool
	"Extended so that if the avatar is hidden due to timeout, its voice is suppressed.
	 Helps prevent divergence between voice chat and croquet connectedness.
	 you cannot speak if you become disembodied due to loss of future-flow."
	speaker ifNotNil: [
		aBool ifTrue: [ 
			self addChild: speaker.
		] ifFalse: [
			self removeChild: speaker.
		]].
	aBool ifFalse: [
		self meetingState ifNotNilDo: [ : meeting | 
			meeting removeWantsToSpeak: (self userLogin).
			meeting releaseLeader: (self userLogin)
		]].
	(aBool and: [ label notNil ]) ifTrue: [ label update ].
	super visible: aBool! !

!QAvatarReplica methodsFor: 'events' stamp: 'bgf 7/25/2009 16:44'!
voiceActivity: level
	"This user is speaking.  (This future-message starts from user's client;
	 it is separate from the sound stream channel but emitted in parallel,
	 so we can give independent notice of who is (or should be) sending sound."

	"The internal future is empirical - helps prevent a non-busy client
	  from having the visual blips anticipate the audio heard by overmuch."
	self root ifNotNilDo: [
		:root| root future signal: #voiceActivityFromUser:level: with: (self userLogin) with: level
	].
	lastSpeech := self island millisecondClockValue.
	label ifNotNil: [ label update.  (label future: 510) update. ].! !

!QAvatarReplica methodsFor: 'meetings' stamp: 'bgf 7/3/2009 13:20'!
wantsToSpeak
	"True when I'm on the want-to-speak list in the meeting."
	^ self meetingState 
		ifNotNil: [ : ms | ms wantsToSpeak includes: (self userLogin) ]
		ifNil: [ false ]! !

!QAvatarReplica methodsFor: 'places' stamp: 'bgf 11/4/2009 15:48'!
warpPlaceUnderMouse: event 
	"If the event is over a frame that is a candidate for warp-goto,
	 then return that place.  Otherwise return nil. 
	 Warp-goto is presented for frames that are
		- in a Place-defining construct
		- far away, by a heuristic comparison of bounds to distance
	 If control key is pressed, there is no warping 
	 (so you can do the normal interactions from far away). 
	 ** MUST remain side-effect-free; accessed through get:with: ** "
	| ed frame |

	event ifNil: [ ^ nil ].
	event controlKeyPressed ifTrue: [ ^ nil ].	
	event selection ifNil: [ ^ nil ].
	event selection distance ifNil: [ ^ nil ].		"E.g. direct dispatch from floating app."
	frame := event selection frame.
	frame ifNil: [ ^ nil ].
	frame isInAvatar ifTrue: [ ^ nil ].	"Don't warp to avatars; interferes with click-menu access"
	ed := event selection distance.

	 "Use the avatar size to rule out odd effects from small objects;
	  e.g. you don't want to warp-to a sign you're already in front of."
	(event shiftPressed not 
	 and: [ed < (self avatarHeight*2.5)]) ifTrue: [ ^ nil ].

	frame asPlace ifNotNilDo: [ : place |
		place = currentPlace ifTrue: [ ^  nil ].	  " Never indicate warp once we're at the place."
		place destinationFrame ifNotNilDo: [ : placeFrame |
			(placeFrame distanceIsPastWarpThreshold: ed forViewExtent: self viewExtent)
			ifTrue: [
				^ place
			]]].
	^ nil

			! !

!QAvatarReplica methodsFor: 'accessing' stamp: 'jcg 2/15/2008 19:41'!
webcamStreamCard
	"Answer a file card that can be used to connect to the user's video stream (or nil, if there is no stream)."
	| desc card |
	desc := self registryStreamDescription copy.
	desc
		ifNil: [^nil]
		ifNotNil: [desc streamID: TObjectID nilObjectID].
	card := QFileCard new.
	card kind: #streamingVideoV1 content: desc.
	^card
! !

!TFileTexture class methodsFor: '*Qwaq-Objects-Document' stamp: 'bgf 2/7/2010 12:38'!
copy: txtrRef toActiveForum: aParticipant 

	"Copy-Job support protocol: copy the resource into the shared storage of the active forum.
	 The fileName of the txtrFarRef's referee will be modified
	 (i.e. do this on the copy, not the original).
	 Note that although it is a texture, it goes into sharedDocuments...
	 this is the path for user-inventory transfer of user-level file-cards.
	 When users place pictures directly we handle them as shared documents
 	 which preserves the 'who-dropped-it' information."

	|  newName sha |
	(sha := (txtrRef get: #sha)) ifNil: [
		self error: 'Island-side data not available.'.
		^ self ].

	"If the texture already is known here, no need to copy it again."
	newName := aParticipant session 
					existingResourceFile: sha forIsland: (aParticipant activeSpace island).
	newName ifNotNil: [	
		txtrRef isValidFarRef
			ifTrue: [ txtrRef future serviceURL: (aParticipant currentForumURL).
					txtrRef future fileName: newName ] 
			ifFalse: [ txtrRef serviceURL: (aParticipant currentForumURL);  
					           fileName: newName ].
		^ self ].

	QSharedDocument copy: txtrRef toActiveForum: aParticipant
	! !

!TFileTexture class methodsFor: '*Qwaq-Objects-Document' stamp: 'bgf 3/5/2009 12:18'!
copy: txtrFarRef toUserInventory: aParticipant

	"Copy-Job support protocol.  Copy the resource into this user's private storage.
	 Note that the fileName of the txtrFarRef's referee will be modified
	 (i.e. do this on the copy, not the original.)"

	QSharedDocument copy: txtrFarRef toUserInventory: aParticipant.! !

!TFileTexture class methodsFor: '*Qwaq-Objects-Document' stamp: 'bgf 9/9/2009 01:32'!
existingDocumentFileName: txtrRef forClient: aClient
	"Document protocol - return the name of an existing local file holding this texture."
	| sha |
	sha := txtrRef get: #sha.
	sha ifNotNil: [ 
		(TContentManager default cacheManager findFileName: sha) 
			ifNotNilDo: [ : cached | ^ cached ]].
	QLogger logInfo: 'No cached texture for sha ', sha.
	^ nil! !

!TFileTexture class methodsFor: '*Qwaq-Objects-Document' stamp: 'bgf 3/3/2009 00:40'!
extensionForData: data
	"Document support: figure out what suffix to use when writing this image data as a file."
	ImageReadWriter allSubclassesDo: [ :cls |
		(cls understandsImageFormat: (ReadStream on:data)) ifTrue:[
			cls typicalFileExtensions size > 0 ifTrue: [
				^ cls typicalFileExtensions first
			]]].
	^nil! !

!TFileTexture class methodsFor: '*Qwaq-Objects-Document' stamp: 'bgf 9/9/2009 14:09'!
fetch: document toFileNamed: fileName forClient: owner
	"Simple delegation back to the generic."
	^ QSharedDocument fetch: document toFileNamed: fileName forClient: owner! !

!TFileTexture class methodsFor: '*Qwaq-Objects-Document' stamp: 'bgf 3/5/2009 12:13'!
getContent: txtrRef  forClient: aUI
	"Document protocol - return the raw data bytes (the file content) for this texture.
	 These are the encoded image-format bytes (e.g. jpeg), not the form bits!! "

	"Quick check... maybe directly in the cache."
	
	(TContentManager default cacheManager find: (txtrRef get: #sha)) 
		ifNotNilDo: [ : cached | ^ cached ].

	"Fallthrough - if we failed to open the cached file do the usual route."
	^ QSharedDocument getContent: txtrRef forClient: aUI! !

!TFileTexture class methodsFor: '*Qwaq-Objects-Document' stamp: 'hrs 4/6/2009 09:58'!
mailDocument: docRef forClient: aUI
	"We can't attach docs on Mac yet, so just send a QRL to the doc."
	^aUI mailDocumentLink: docRef! !

!TFileTexture class methodsFor: '*Qwaq-Objects-Document' stamp: 'hrs 4/6/2009 21:28'!
openDocumentWebPage: docRef forClient: aUI
	"Open the Web page for the document."
	^aUI openQrlForDocument: docRef! !

!TFileTexture class methodsFor: '*Qwaq-Objects-Document' stamp: 'bgf 9/9/2009 01:01'!
saveDocument: txtrRef asFile: newName forClient: aUI
	"Document protocol.  See base definition in QAbstractDocument"
	^ QSharedDocument saveDocument: txtrRef asFile: newName forClient: aUI! !

!TFileTexture class methodsFor: '*Qwaq-Objects-Document' stamp: 'bgf 2/12/2009 21:59'!
writeManifestEntry: txtrRef on: xmlWriter checkingDoneSet: doneShas

	"Write the Qwaq Forums manifest entry for the given texture thumb reference."
	| fName sha tag |
 	sha := txtrRef get: #sha.	
	(sha isNil or: [doneShas includes: sha]) ifFalse: [
		fName := txtrRef get:#fileName.
		tag := 'resource'.
		fName ifNotNil: [
			xmlWriter startTag: tag.
			xmlWriter attribute: 'type' value: (self resourceType).
			xmlWriter attribute: 'uri' value: fName.
			xmlWriter attribute: 'sha' value: sha.
			xmlWriter endEmptyTag: tag.
			doneShas add: sha.
		]].! !

!TFileTexture methodsFor: '*Qwaq-Objects-Documents' stamp: 'bgf 3/3/2009 00:43'!
asLocalFileUri
	"Abstract-Document protocol... 
	 Return URI-form of the path on the user's computer
	 where the texture image file is kept.
	 The local file URI for a texture is its cache path.

	 This is in support of open-private.  Not yet wired completely,
	 since the cache file name (going back two years) claims to include the suffix, 
	 but we don't cache textures with the suffix in the cached file name."

	self cacheFileName ifNotNil: [ : nm |
		| cachePath |
		cachePath := (TContentManager default cacheManager directoryForFile: nm) uri path.
		^ cachePath, nm
	].
	^ nil! !

!TFileTexture methodsFor: '*Qwaq-Objects-Documents' stamp: 'bgf 3/3/2009 02:56'!
asServiceManagerPath
	"QServiceManager get/put argument we use to refer to this document ."
	self isBundled ifTrue: [
		"Path support for moving bundled documents to the service provider."
		^ 'system-resources/bundled/', self sharedName
	].
	self isDocumentInForumStorage ifTrue: [
		^ '/services/', (self serviceName), '/', self sharedName.
	] ifFalse: [
		^ self sharedName
	].
	! !

!TFileTexture methodsFor: '*Qwaq-Objects-Documents' stamp: 'bgf 3/8/2009 01:58'!
cacheLocally
	^ true! !

!TFileTexture methodsFor: '*Qwaq-Objects-Documents' stamp: 'bgf 3/3/2009 00:43'!
canBeCopiedToInventory
	"Abstract-Document protocol.
	 True Indicates that the document-inventory transfer operations
	  (isDocumentInInventory isDocumentInForumStorage,
	  and class-side copy:toActiveForum: copy:toUserInventory:)
	 are available for this instance."

	^ fileName notNil! !

!TFileTexture methodsFor: '*Qwaq-Objects-Documents' stamp: 'bgf 3/3/2009 00:55'!
canBeOpenedPrivately
	"Abstract-Document protocol."
	^ true! !

!TFileTexture methodsFor: '*Qwaq-Objects-Documents' stamp: 'bgf 3/3/2009 00:45'!
canBeSavedLocally
	"Abstract-Document protocol.
	 Indicates that doSaveAsForAvatar: is supported and appropriate for this instance."
	^ fileName notNil! !

!TFileTexture methodsFor: '*Qwaq-Objects-Documents' stamp: 'bgf 3/3/2009 02:56'!
doOpenForAvatar: anAvatarReplica
	"Open the texture image privately in an application on the user's computer.
	 Use of this is guarded by  # canBeOpenedPrivately. "
	self canBeOpenedPrivately ifTrue: [
		QOpenDocExternalJob open: self forUser: anAvatarReplica
	] ifFalse: [
		anAvatarReplica signal: #alertMessage with: 'That image cannot be opened privately.'.
	].! !

!TFileTexture methodsFor: '*Qwaq-Objects-Documents' stamp: 'bgf 3/3/2009 02:56'!
doSaveAsForAvatar: anAvatarReplica
	"Save the texture locally in the client of the current avatar.
	 Use of this is guarded by  # canBeSavedLocally  "
	self canBeSavedLocally ifTrue: [
		QSaveDocumentJob save: self forUser: anAvatarReplica
	] ifFalse: [
		anAvatarReplica signal: #alertMessage with: 'That image is not available to be saved.'.
	].! !

!TFileTexture methodsFor: '*Qwaq-Objects-Documents' stamp: 'bgf 3/13/2009 04:17'!
fileCardBacking
	^ 'ImageCard.png'! !

!TFileTexture methodsFor: '*Qwaq-Objects-Documents' stamp: 'bgf 3/3/2009 00:46'!
fileChooserPattern
	"Abstract-Document protocol for local Save-As.
	 When saving this texture, the file-chooser configuration."
	| short ext |
	short := (self shortName).
	ext := (short includes: $.) ifTrue: [ short copyAfterLast: $. ] ifFalse: [ 'png' ].
	^ {{'Images (*.', ext,')' .  '*.',ext}}! !

!TFileTexture methodsFor: '*Qwaq-Objects-Documents' stamp: 'bgf 3/3/2009 00:46'!
fileChooserPatternTweak
	"Abstract-Document protocol for local Save-As.
	 When saving this texture, the file-chooser configuration (Tweak fallback chooser)."
	| short ext |
	short := (self shortName).
	ext := (short includes: $.) ifTrue: [ short copyAfterLast: $. ] ifFalse: [ 'png' ].

	^ { {'Image' . {'*.', ext }}}.! !

!TFileTexture methodsFor: '*Qwaq-Objects-Documents' stamp: 'bgf 3/3/2009 00:46'!
fileSaveLabel
	"Abstract-Document protocol for local Save-As.
	  When saving a texture, the title of the file dialog."
	^ 'Save Image'
! !

!TFileTexture methodsFor: '*Qwaq-Objects-Documents' stamp: 'bgf 3/3/2009 02:57'!
get: aSelector
	"For parity with abstract-document.  
	 Makes far-refs and instances both usable from the same harness-side code."
	^ self perform: aSelector! !

!TFileTexture methodsFor: '*Qwaq-Objects-Documents' stamp: 'hrs 4/4/2009 20:26'!
identifier
	^self fileName copyAfter: $/! !

!TFileTexture methodsFor: '*Qwaq-Objects-Documents' stamp: 'bgf 3/5/2009 21:38'!
isDocumentInForum: forumURL
	"Is the document already available as a resource for this forum (stored in the forum's storage) ?"
	^ self isBundled 
		or: [
			(self isDocumentInInventory not)
			and: [ self serviceURL = forumURL ]].! !

!TFileTexture methodsFor: '*Qwaq-Objects-Documents' stamp: 'bgf 3/3/2009 02:57'!
isDocumentInForumStorage
	"Abstract-Document protocol.
	 True if the resource is stored in the service manager file space for this forum;
	 false (for instance) for bundled textures, or private documents in user inventory.
	 This is part of the filecard/document protocol for resource transfers."
	^ fileName notNil
	   and: [
		(fileName beginsWith: 'sharedDocuments') or: [ fileName beginsWith: 'resources/' ]].
! !

!TFileTexture methodsFor: '*Qwaq-Objects-Documents' stamp: 'bgf 3/3/2009 02:57'!
isDocumentInInventory
	"Abstract-Document protocol.
	 True if the resource is stored in the service manager file space for this forum;
	 false (for instance) for bundled textures, or private documents in user inventory.
	 This is part of the filecard/document protocol for resource transfers."
	^ fileName notNil
	   and: [
		(fileName beginsWith: '/users/') ]! !

!TFileTexture methodsFor: '*Qwaq-Objects-Documents' stamp: 'bgf 3/3/2009 01:30'!
isKeptLocally
	"Abstract-Document protocol.
	 All textures are auto-fetched to the local cache if they have names to fetch,
	 but we nonetheless answer false here,
	 because those cache-files are not available for the user to open.
	 So the document system falls back to getContent:...,
	 which pulls the data from the cache and writes to the privateFileFor"
	^ false! !

!TFileTexture methodsFor: '*Qwaq-Objects-Documents' stamp: 'bgf 2/7/2010 13:06'!
isRenamableContent
	"Document protocol; is this resource a user-modifiable application-document?"
	^ fileName notNil and: [ (fileName beginsWith: 'resources/') not ]

		"Note: content in resources/   is sha-named, so the name itself is authoritivate."! !

!TFileTexture methodsFor: '*Qwaq-Objects-Documents' stamp: 'bgf 5/14/2010 14:05'!
isRepositoryDocument
	^ false! !

!TFileTexture methodsFor: '*Qwaq-Objects-Documents' stamp: 'hrs 4/4/2009 20:42'!
mailFromAvatar: anAvatarReplica
	"Open a mail client for this doc on the machine of the current avatar. "
	QMailDocumentJob mail: self fromUser: anAvatarReplica
! !

!TFileTexture methodsFor: '*Qwaq-Objects-Documents' stamp: 'hrs 4/6/2009 22:10'!
openWebPageForAvatar: anAvatarReplica
	"Open a mail client for this doc on the machine of the current avatar. "
	QOpenDocumentWebPageJob openWebPage: self fromUser: anAvatarReplica
! !

!TFileTexture methodsFor: '*Qwaq-Objects-Documents' stamp: 'bgf 3/3/2009 17:17'!
saveComplete: shortName
	"Callback from save-as (and save for private open) 
	 when the harness side operation completes.  
	 The shortName is that used on the client's machine."

	self signal: #localSaveComplete with: shortName
! !

!TFileTexture methodsFor: '*Qwaq-Objects-Documents' stamp: 'bgf 11/17/2009 18:18'!
serviceName
	"Abstract document protocol.
	 The service (forum) name from the Qwaq/Croquet service URL (//host/realm.../serviceName)
	 under which this item lives.  "
	^ self serviceURL ifNotNilDo:[ : surl | surl copyAfterLast: $/]! !

!TFileTexture methodsFor: '*Qwaq-Objects-Documents' stamp: 'bgf 3/3/2009 17:18'!
sharedName
	"Document protocol.  
	 The service-manager path, relative to forum (or 'absolute' for bundled and user-inventory), 
	 where this item resides on the server."
	^ self fileName! !

!TFileTexture methodsFor: '*Qwaq-Objects-Documents' stamp: 'bgf 8/27/2009 11:56'!
sharedName: newSharedName
	"Document protocol.  
	 The service-manager path, relative to forum (or 'absolute' for bundled and user-inventory), 
	 where this item resides on the server."
	self isBundled ifFalse: [
		self fileName: newSharedName.
	].! !

!TFileTexture methodsFor: '*Qwaq-Objects-Documents' stamp: 'bgf 3/3/2009 00:28'!
shortName
	"Document protocol.  The local name of the texture image file (includes suffix but not path.)"
	| fn | 
	fn := fileName ifNil: [ 'texture' ].
	^ (fn includes: $/) 
		ifTrue: [ fn copyAfterLast: $/ ] 
		ifFalse: [ (fn includes: $\) ifTrue: [ fn copyAfterLast: $\ ] ifFalse: [ fn ]].
		
! !

TGroup subclass: #Q3DSModel
	instanceVariableNames: 'mesh modelScale instanceScale modelFilename boundsDepth optimized optimizedEarly'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Objects-Resources'!

!Q3DSModel commentStamp: '<historical>' prior: 0!
Manages a mesh loaded from a 3DS ASE file.
This class uses the ASE file name (uri) when persisting itself, 
rather than the copying the geometry of the mesh.

It is like TFileTexture in requiring that external resources (images, for TFileTexture,
or ASE files, here)  must be available when the object is constructed from
a persisted form (XML); the model data is not persisted directly by persistOn:, 
only the resource reference.

It differs from TFileTexture in that the ASE resource files are not automatically 
fetched across replicated spaces.    Instead -- for now -- QCache is used to cache 
the first load of a given model in the island.  All construction of instances
of a particular model after the first in an island use the cache fast path
(so replicants don't need to read the ASE model for those later instances.)
!

!Q3DSModel class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!Q3DSModel class methodsFor: 'defaults' stamp: 'bgf 6/1/2009 14:14'!
defaultDoContextMenu
	"Most common use is trees - scenery"
	^ false! !

!Q3DSModel class methodsFor: 'defaults' stamp: 'bgf 6/4/2009 16:11'!
defaultIsFromTemplate
	"Most common use of these objects is trees - scenery"
	^ true! !

!Q3DSModel class methodsFor: 'objects from disk' stamp: 'bgf 2/11/2010 10:03'!
interceptLegacyObjectForThumb: filename

	"Detect old ASE filenames that have been moved
	 in the packaging tree, and return the modified name. 
	 For all other cases, return nil. "

	filename ifNil: [ ^ nil ].

	(filename beginsWith: 'Content/Qwaq/Models/Plants/') ifTrue: [
		"Objects/Plants contains the new client-bundled  c3x's.
	 	 Many forums are full of older Q3DSModel plants that refer to Models/Plants/<file>.ASE "
		^ 'Objects/Qwaq/Plants/', ((filename copyAfterLast: $/)  copyUpTo: $.), '.c3x'
	].
	^ nil! !

!Q3DSModel class methodsFor: 'utilities' stamp: 'bgf 10/26/2008 13:55'!
meshFromASEFile: aseFileName atScale: aScale
	"Return the TMesh described by the given .ASE file, loading
       using the given basic scale."

	^ QCache make:(aseFileName asSymbol) constructor: [
		| loader m fpath fname c3xName |
		(self interceptLegacyObjectForThumb: aseFileName) ifNotNilDo: [ :tfile |
			m := TSceneReader new loadFileNamed: tfile
		].
		m ifNil: [ 
			c3xName := (aseFileName copyUpToLast: $.), '.c3x'.
			m := TSceneReader new loadFileNamed: c3xName.
			m ifNil:[
				fpath := aseFileName copyFrom: 1 to: (aseFileName lastIndexOf: $/).
				fname := aseFileName copyFrom: fpath size+1 to: aseFileName size.
				loader := TLoad3DSMax new.
				loader addSearchPath: fpath.
				[
					m := loader loadFileNamed: fname
						scale: aScale shadeAngle: 90.1 textureMode: GLModulate.
				] on: Error do: [
					QLogger logWarning: 'Absent model: ', aseFileName.
					m  := nil.
				].
				TSceneWriter new storeScene: m onFileNamed: c3xName.
			]].
		m ifNotNil: [ m doUpdateShape ].	"Before caching"
		m
	].! !

!Q3DSModel class methodsFor: 'utilities' stamp: 'bgf 10/26/2008 13:57'!
optimizedMeshFromASEFile: aseFileName atScale: aScale

	"Return the TMesh described by the given .ASE file, loading
       using the given basic scale."

	^ QCache make:(aseFileName asSymbol) constructor: [
		| loader m fpath fname c3xName |

		(self interceptLegacyObjectForThumb: aseFileName) ifNotNilDo: [ :tfile |
			m := TSceneReader new loadFileNamed: tfile
			].
		m ifNil: [
			c3xName := (aseFileName copyUpToLast: $.), '.c3x'.
			m := TSceneReader new loadFileNamed: c3xName.
			m ifNil:[
				fpath := aseFileName copyFrom: 1 to: (aseFileName lastIndexOf: $/).
				fname := aseFileName copyFrom: fpath size+1 to: aseFileName size.
				loader := TLoad3DSMax new.
				loader addSearchPath: fpath.
				[
					m := loader loadFileNamed: fname
						scale: aScale shadeAngle: 90.1 textureMode: GLModulate.
				] on: Error do: [
					QLogger logWarning: 'Absent model: ', aseFileName.
					m  := nil.
				].
				m ifNotNil: [
					m boundsDepth: -1; allUpdateShape.
					m := m flattenMesh.
					TSceneWriter new storeScene: m onFileNamed: c3xName.
				].
			]].
		m ifNotNil: [ m doUpdateShape ].	"Before caching"
		m
	].! !

!Q3DSModel methodsFor: 'accessing' stamp: 'bgf 12/6/2006 16:19'!
boundsDepth
	^ boundsDepth! !

!Q3DSModel methodsFor: 'accessing' stamp: 'bgf 12/7/2006 12:45'!
boundsDepth: aScalar

	boundsDepth := aScalar.
	"We both remember the value, and pass it to children
	 so that the attribute works whether set before or after loading."
	super boundsDepth: aScalar.! !

!Q3DSModel methodsFor: 'persist' stamp: 'bgf 2/12/2010 01:16'!
constructFrom: pStream

	super constructFrom: pStream.

	modelFilename := pStream readString: 'modelFilename' default: modelFilename.
	modelScale := pStream readNumber: 'modelScale' default: modelScale.
	instanceScale := pStream readNumber: 'instanceScale' default: instanceScale.
	boundsDepth := pStream readNumber: 'boundsDepth' default: boundsDepth.
	optimizedEarly := pStream readBool: 'optimizedEarly' default: optimizedEarly.
	optimized := pStream readBool: 'optimized' default: optimized.

	modelFilename ifNotNil: [
		modelFilename := (self class interceptLegacyObjectForThumb: modelFilename)
							ifNil: [ modelFilename ].
		"In future, if we adopt other mesh standards, we can select which kind using the filename,
	  	 or add another property to remember which rebuild path to take."
		self setupASE:modelFilename aseScale: modelScale
			scale: instanceScale optimized: optimized optimizedEarly: optimizedEarly.
	].! !

!Q3DSModel methodsFor: 'initialize' stamp: 'bgf 6/1/2009 14:14'!
defaultIsAtomic
	"By default, use full encapsulation"
	^ true! !

!Q3DSModel methodsFor: 'initialize' stamp: 'bgf 12/6/2006 16:58'!
initialize

	super initialize.
	modelScale := 1.0.
	instanceScale := 1.0.
	modelFilename := nil.
	boundsDepth := 1.
	optimized := false.
	optimizedEarly := false.! !

!Q3DSModel methodsFor: 'initialize' stamp: 'bgf 12/5/2006 15:53'!
installedMesh: msh

	"Optional extension method.  If a subclass needs to operate on the new mesh,
	 this method can be overridden.  It is called at the very end of setup, 
	 after the mesh has been added as my child."! !

!Q3DSModel methodsFor: 'initialize' stamp: 'bgf 12/5/2006 15:49'!
loadedMesh: msh

	"Optional extension method.  If a subclass needs to alter the result of loading the ASE file,
	 it can do so in an override of this method, which is called after ASE file loading, 
	 before any other operations.  The msh parameter is the root mesh returned by the ASE loader."
! !

!Q3DSModel methodsFor: 'accessing' stamp: 'bgf 12/19/2006 16:10'!
mesh
	^ mesh! !

!Q3DSModel methodsFor: 'persist' stamp: 'bgf 12/6/2006 16:14'!
persistOn: pStream

	super persistOn: pStream.
	pStream storeString: 'modelFilename' value: modelFilename.
	pStream storeNumber: 'modelScale' value: modelScale default: 1.0.
	pStream storeNumber: 'instanceScale' value: instanceScale default: 1.0.
	pStream storeNumber: 'boundsDepth' value: boundsDepth default: 1.0.
	pStream storeBool: 'optimized' value: optimized default: false.	
	pStream storeBool: 'optimizedEarly' value: optimizedEarly default: false.! !

!Q3DSModel methodsFor: 'accessing' stamp: 'bgf 2/11/2010 15:54'!
privateScale: aScaleFactor
	super privateScale: aScaleFactor.
	instanceScale := instanceScale * aScaleFactor.! !

!Q3DSModel methodsFor: 'initialize' stamp: 'das 2/7/2007 10:13'!
setupASE: aseFile aseScale: aseScale scale: scale optimized: bSuper

	"Initialize the element by loading the given ASE file (through the caching factory).
	 The aseScale is the scale at which to load the ASE 
	 (must be invariant for all instances of the concrete class,
	 since it applies during ase-loading of the prototype before caching.
	 The scale is applied after loading (or cache cloning) and can vary across instances.
	 If bSuper is true, the mesh is optimized through TFrame#flattenMesh 
	 (after loading and scaling, after calling loadedMesh:, but before
	 adding the mesh to myself as child, and before calling installedMesh:). "

	^ self setupASE: aseFile aseScale: aseScale scale: scale optimized: bSuper optimizedEarly: false.! !

!Q3DSModel methodsFor: 'initialize' stamp: 'bgf 2/11/2010 11:09'!
setupASE: aseFile aseScale: aseScale scale: scale optimized: bSuper optimizedEarly: bCacheSuper

	"Initialize the element by loading the given ASE file (through the caching factory).
	 The aseScale is the scale at which to load the ASE 
	 (must be invariant for all instances of the concrete class,
	 since it applies during ase load and scales the prototype that gets cached.)
	 The instanceScale is applied after loading, and can vary across instances.

	 If bSuper is true, the mesh is optimized through TFrame#flattenMesh 
	 Normall, this is done after loading and scaling, after calling loadedMesh:, 
	 but before adding the mesh to myself as child, and before calling installedMesh:). 

	 IF bCacheSuper is true, the mesh is optimized immediately after loading,
	 before the cache snapshot is taken, and before calling #loadedMesh:."

	bCacheSuper ifTrue: [
		mesh := Q3DSModel optimizedMeshFromASEFile: aseFile atScale: aseScale.
	] ifFalse: [
		mesh := Q3DSModel meshFromASEFile: aseFile atScale: aseScale.
	].
	mesh ifNotNil: [
		self loadedMesh: mesh.
		scale ~= 1.0 ifTrue: [ mesh scaleBy: (scale)].
		(bSuper & (bCacheSuper not)) ifTrue: [
			mesh boundsDepth: -1; allUpdateShape.	"Prepare to flatten."
			mesh := mesh flattenMesh.
		].
		mesh boundsDepth: boundsDepth.
	].

	"Remember for persisting:"
	modelScale := aseScale.
	instanceScale := scale.
	modelFilename := aseFile.
	optimized := bSuper.
	optimizedEarly := bCacheSuper.

	mesh ifNotNil: [
		self addChild: mesh.
		self installedMesh: mesh.
	].! !

TGroup subclass: #QCharacter
	instanceVariableNames: 'geometry geometryList resetList motionDictionary currentMotion isPlaying counter speed updateRate time feet heels lastFoot lastFootPosition lastDelta downRay floor bvhRootHeight rootHeight scaleHeight container lastPlayed'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Objects-Motion'!

!QCharacter commentStamp: 'das 12/3/2006 16:55' prior: 0!
This is a bvh animation player. Can be used as an NPC or as an avatar.!

!QCharacter class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QCharacter methodsFor: 'actions' stamp: 'das 12/14/2006 17:35'!
actionBegin

	self play.! !

!QCharacter methodsFor: 'actions' stamp: 'das 12/14/2006 17:35'!
actionPause

	self stopPlay.! !

!QCharacter methodsFor: 'actions' stamp: 'das 12/14/2006 17:44'!
actionReset

	self stopPlay.
	1 to: geometryList size do:[:index | (geometryList at:index)localTransform:(resetList at:index)].
	time := 0.! !

!QCharacter methodsFor: 'actions' stamp: 'bgf 11/12/2009 13:29'!
actionsList
	^{{ #actionBegin. 'Begin Action'.}. 
	   {#actionPause.  'Pause Action'.}. 
	   {#actionReset. 'Reset Action' }.}.! !

!QCharacter methodsFor: 'initialize' stamp: 'das 12/14/2006 11:54'!
addFoot: footName

"feet are used to track the location of the floor and match the character to the world"
	self addFoot: footName heel: footName.! !

!QCharacter methodsFor: 'initialize' stamp: 'hrs 7/24/2007 13:24'!
addFoot: footName heel: heelName

"feet are used to track the location of the floor and match the character to the world. We use the foot to determine distance to the ground and heel to determine which leg to test."
	| foot heel site |
	foot := self find:[:frame | frame objectName = footName]. 
	foot size>0 ifTrue:[foot := foot at:1] ifFalse:[self error:'Foot not found <', footName,'>'].
	heel := self find:[:frame | frame objectName = heelName.].
	heel size>0 ifTrue:[heel := heel at:1] ifFalse:[self error:'Heel not found <', heelName,'>'].
	feet ifNil:[feet := OrderedCollection new. heels := OrderedCollection new.].
	site := foot find:[:frame | frame objectName = 'Site'].
	site size>0 ifTrue:[foot := site at:1] .
	feet add: foot. heels add: heel.
! !

!QCharacter methodsFor: 'initialize' stamp: 'jdm 8/21/2007 08:15'!
addMotion: motion name: name

"Use this to add an imported BVH motion file to the QCharacter. Give it a name to be able to direct the QCharacter to switch to this motion with #useMotion:"
	| motionList frameList transform channels cName val qMotion xVal yVal zVal translationList |

	geometry ifNil:[self error:'Need geometry to process motion.'].
	motionList := OrderedCollection new.
	translationList := OrderedCollection new.
	1 to: motion frameCount do:[ :index |
		frameList := OrderedCollection new.
		geometryList do:[ :tframe | 
			transform := Matrix4x4 new.
			channels := motion channelsDictionary at: tframe objectName ifAbsent:[].
			channels ifNil:[transform := tframe localTransform copy.]
			ifNotNil:[
				channels do:[: channel |
					cName := channel channelName.
					val := channel values at: index.
					cName caseOf: {
						[ #Xposition ] -> [ xVal := val ].
						[ #Yposition ] -> [ yVal := val ].
						[ #Zposition ] -> [ zVal := val ].
						[ #Xrotation ] -> [ transform addRotationAroundX: val. ].
						[ #Yrotation ] -> [ transform addRotationAroundY: val. ].
						[ #Zrotation ] -> [ transform addRotationAroundZ: val. ].
					}
					otherwise: [].
				].
			].
"			(rotateX notNil) & (rotateY notNil) & (rotateZ notNil) ifTrue:[
				transform addRotationAroundZ: rotateZ.
				transform addRotationAroundX: rotateX.
				transform addRotationAroundY: rotateY.
				transform translation: tframe translation .
			]."
			(xVal notNil) & (yVal notNil) & (zVal notNil) ifTrue:[
				translationList add: (xVal@yVal@zVal).
				xVal := nil. yVal := nil. zVal := nil.
			].
			frameList add: transform asQuaternion.
		].
		motionList add: frameList.
	].
	qMotion := QMotion new motion: motionList translation: translationList frequency: motion frequency*1000.0 size:motion frameCount.
	motionDictionary at: name put:qMotion.! !

!QCharacter methodsFor: 'accessing' stamp: 'das 2/15/2007 08:40'!
bvhRootHeight

	^ bvhRootHeight.! !

!QCharacter methodsFor: 'persist' stamp: 'das 2/19/2007 06:48'!
constructFrom: pStream

	super constructFrom: pStream.
	self speed:(pStream readNumber: 'speed' default: self speed).
	self updateRate:(pStream readNumber: 'updateRate' default: self updateRate).
	self isPlaying: (pStream readBool: 'isPlaying' default: self isPlaying).
	self scaleHeight: (pStream readNumber:'scaleHeight' default: 1).! !

!QCharacter methodsFor: 'accessing' stamp: 'das 2/27/2007 14:24'!
container: con

	container := con.! !

!QCharacter methodsFor: 'animate' stamp: 'jdm 8/16/2007 16:52'!
doPlay

	| matrices cval deltaTrans |

	(isPlaying and: [container root notNil]) ifTrue:[
		time := time + (updateRate*speed).
		deltaTrans := 0@0@0.
		cval := (time*(currentMotion frequency)+1)asInteger. "target counter, should at least be 1"
		[cval > currentMotion size] whileTrue:[
			deltaTrans := deltaTrans + ((currentMotion translation at: currentMotion size) -(currentMotion translation at:counter)).
			cval := cval - currentMotion size. 
			counter := 1.
			time := time - (currentMotion size/currentMotion frequency).
			
		].
		cval ~= counter ifTrue:[
			deltaTrans := deltaTrans + (currentMotion translation at: cval)-(currentMotion translation at: counter).
			counter := cval.
		].

		(counter ~= lastPlayed) ifTrue: [
	 		matrices := currentMotion motion at: counter.
			1 to: geometryList size do:[: index |
				(geometryList at: index)orientation:(matrices at: index) asMatrix4x4.
			].
			lastPlayed := counter.
		].
		container translation: 
				(container orientation transformPoint: deltaTrans*scaleHeight) 
				+ container translation.
		self fall.
	].
	isPlaying ifTrue: [ (self future:updateRate) doPlay.].! !

!QCharacter methodsFor: 'animate' stamp: 'das 2/27/2007 11:02'!
fall
		| min foot dist |
	min := Float infinity.
		feet do:[:ft | 
			"ft parent colorize: (VectorColor r:0 g:0 b:1 a:1). "
			ft globalPosition y<min ifTrue:[foot := ft. min := ft globalPosition y].
		].
		lastFoot = foot ifTrue:[
			lastDelta := lastFootPosition - foot globalPosition.
			floor ifNotNil: [
				lastDelta :=0@0@0.
				lastDelta y: floor-foot globalPosition y.
				container translation: container translation+lastDelta.
			].
			lastFootPosition := foot globalPosition.
		] ifFalse:[
			foot addChild: downRay.
			downRay updateTransform.
			container root rayTest: downRay ignore: {container}.
			downRay selectedFrame ifNotNil:[
				dist := downRay selectedDistance.
				floor := downRay globalPosition y-dist.
				downRay resetTotal.
				lastDelta y: floor-foot globalPosition y.
				container translation: container translation + lastDelta.
			] ifNil: [
				floor := nil.
			].
			lastFoot := foot.
			lastFootPosition := foot globalPosition.
		].
	"	foot parent colorize:  (VectorColor r:1 g:0 b:0 a:1)."
! !

!QCharacter methodsFor: 'initialize' stamp: 'bgf 3/17/2007 16:08'!
geometry
	"The 'Root' of the skeleton group tree."
	^ geometry! !

!QCharacter methodsFor: 'initialize' stamp: 'hrs 7/23/2007 14:46'!
geometry: geo
	| footLoc rootLoc |
	self initializeGeometry: geo.
	footLoc := Float infinity.
	geometry allFramesDo:[:frame |
		frame objectName = 'Root' ifTrue:[rootLoc := frame globalTransform translation y.].
		footLoc > frame globalTransform translation y ifTrue:[
			footLoc := frame globalTransform translation y.
		].
	].
	rootHeight := rootLoc - footLoc.
	scaleHeight := rootHeight/bvhRootHeight.! !

!QCharacter methodsFor: 'initialize' stamp: 'bgf 7/1/2010 12:13'!
initialize

	super initialize.
	motionDictionary := Dictionary new. 
	scaleHeight := 1.
	isPlaying := false.
	counter := 1.
	speed := 1.0.
	time := 0.
	floor := 0.
	lastDelta := 0@0@0.
	updateRate := 75. "future send updates every 75 milliseconds"
	downRay := TRay new.
	downRay isDownRay: true.
	downRay translation:0@4@0.
	container := self.
	self addChild: downRay.! !

!QCharacter methodsFor: 'initialize' stamp: 'hrs 7/27/2007 12:43'!
initializeGeometry: geo

"Geometry is an imported geometry file. The names of the frames that make up the geometry need to match the names in the motion files.  Geometry is not persisted; subclasses are responsible for reconstructing their geometry."
	|  alreadyListed |
	geometry := geo.
	geometry updateTransform.
	geometryList := OrderedCollection new.
	resetList := OrderedCollection new.
	geometry allFramesDo:[:frame |
		"Models can use the same name for skeleton TGroups and meshes...
		 (a side effect of 3DS-Max helper-object naming)
		 We need the skeleton TGroup in the geometry list.  
		 Also take only one of each name, to protect against ill-formed models."

		alreadyListed := geometryList anySatisfy: [ :gl | gl objectName = frame objectName].		
		(alreadyListed not and: [frame isKindOf:TGroup]) ifTrue: [
			geometryList add: frame. 
			resetList add: frame localTransform.
		].
	].
	geometryList := geometryList asArray.
	resetList := resetList asArray.
	self addChild: geometry.
! !

!QCharacter methodsFor: 'accessing' stamp: 'bgf 12/11/2006 11:33'!
isPlaying
	^ isPlaying! !

!QCharacter methodsFor: 'accessing' stamp: 'bgf 12/11/2006 11:34'!
isPlaying: aBool
	
	aBool ifTrue: [ self play ] ifFalse: [self stopPlay ].! !

!QCharacter methodsFor: 'persist' stamp: 'das 2/19/2007 06:47'!
persistOn: pStream
	super persistOn: pStream.

	pStream storeNumber: 'speed' value: self speed default: (1.0).
	pStream storeNumber: 'updateRate' value: self updateRate default: 75.
	pStream storeBool: 'isPlaying' value:self isPlaying default: false.
	pStream storeNumber:'scaleHeight' value: scaleHeight default: 1.
	"Note that motions and geometry are not persisted here; 
	 they are the responsibility of subclass initialization."! !

!QCharacter methodsFor: 'animate' stamp: 'bgf 12/11/2006 14:58'!
play
"This will begin the play of the currentMotion"
	isPlaying := true.
	(self future:(self updateRate)) doPlay. ! !

!QCharacter methodsFor: 'animate' stamp: 'das 2/18/2007 12:50'!
randomize

	self useRandomMotion.
	speed := 1+((10 atRandom-5)/50.0).
speed := 1.0. "for now"! !

!QCharacter methodsFor: 'accessing' stamp: 'das 2/15/2007 00:59'!
rootHeight

	^ rootHeight.! !

!QCharacter methodsFor: 'accessing' stamp: 'das 2/19/2007 06:48'!
scaleHeight: sh

	scaleHeight := sh.! !

!QCharacter methodsFor: 'accessing' stamp: 'bgf 12/11/2006 11:33'!
speed
	^ speed! !

!QCharacter methodsFor: 'accessing' stamp: 'das 12/4/2006 09:49'!
speed: sp
	
	speed := sp.! !

!QCharacter methodsFor: 'animate' stamp: 'das 12/3/2006 23:34'!
stopPlay
"This will terminate the play of the current motion"
	isPlaying := false.
! !

!QCharacter methodsFor: 'accessing' stamp: 'bgf 12/11/2006 11:33'!
updateRate
	^ updateRate! !

!QCharacter methodsFor: 'accessing' stamp: 'das 12/4/2006 09:49'!
updateRate: ur

	updateRate := ur.! !

!QCharacter methodsFor: 'animate' stamp: 'bgf 3/18/2007 23:10'!
useMotion: mName

"This will set the currentMotion to be the one already defined in the dictionary"
	
	| cm |
	cm := motionDictionary at: mName ifAbsent:[].
	cm ~= currentMotion ifTrue:[
		currentMotion := cm.
		time := 0.
		counter := 1.
		lastPlayed := 0.
	].
! !

!QCharacter methodsFor: 'animate' stamp: 'das 2/19/2007 10:42'!
useRandomMotion

"This will set the currentMotion to be the one already defined in the dictionary"
	| newMotion |
	newMotion := motionDictionary atRandom value.
	currentMotion ~= newMotion ifTrue:[
		time := 0.
		counter := 1.
		currentMotion := newMotion.
	].
! !

QCharacter subclass: #QHorse
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Objects-Motion'!

!QHorse commentStamp: 'das 12/5/2006 13:18' prior: 0!
QHorse is a subclass of QCharacter. 

#initialize - generates a pipeline character from a BVH file and loads a number of motions.
#doPlay - method called every updateRate milliseconds. Every so often, we switch the motion, and we are constantly moving in a circle, more or less.  super doPlay actually performs the animation and floor following.!

!QHorse class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QHorse methodsFor: 'animate' stamp: 'das 2/19/2007 09:02'!
doPlay

"the horse walks in a circle (more or less) and every so often, chooses a new motion"

"we can simply choose a random motion for our character"
	"100 atRandom = 1 ifTrue:[ self useRandomMotion.]."
"have him slowly rotate - hardly noticeable unless he is standing still"
	self addRotationAroundY:0.5. "turtle graphics"
"this is where the real animation occurs"
	super doPlay.

! !

!QHorse methodsFor: 'initialize' stamp: 'bgf 9/14/2007 23:23'!
initialize

	| path bvh pHorse |
	super initialize.

"specify which .BVH we will be using for our skeleton"
	path :=FileDirectory pathFrom:{'content'. 'Motion'. 'Horse'. 'HWalk.bvh'}.
	bvh := BVHParser initializeWithFileName: path scale:0.035.
	bvhRootHeight := bvh jointRoot rootHeight.

"construct a placeholder costume for the geometry until we get real content"
	pHorse := bvh jointRoot makePipes.
	self geometry: pHorse.

"need to know where the feet are"
	self addFoot:'LeftToes'.
	self addFoot:'LeftKnuckle'.
	self addFoot:'RightKnuckle'.
	self addFoot:'RightToes'.

	"add the first motion, which also happens to be from the same BVH we got the skeleton from"
	self addMotion: bvh motion name: #Walk.

	"add a few more motions, presuming that their skeletons match up (more or less)"
	path :=FileDirectory pathFrom:{'content'. 'Motion'. 'Horse'. 'HTrot.bvh'}.
	bvh := BVHParser initializeWithFileName: path scale:0.035.
	self addMotion: bvh motion name: #Trot.

	path :=FileDirectory pathFrom:{'content'. 'Motion'. 'Horse'. 'HGallop.bvh'}.
	bvh := BVHParser initializeWithFileName: path scale:0.035.
	self addMotion: bvh motion name: #Gallop.

	"select which motion to use right now"
	self useMotion: #Walk. 

	false ifTrue: [
		path :=FileDirectory pathFrom:{'content'. 'Motion'. 'Horse'. 'HGallopStopRear.bvh'}.
		bvh := BVHParser initializeWithFileName: path scale:0.035.
		self addMotion: bvh motion name: #GallopRear.

		path :=FileDirectory pathFrom:{'content'. 'Motion'. 'Horse'. 'HBow.bvh'}.
		bvh := BVHParser initializeWithFileName: path scale:0.035.
		self addMotion: bvh motion name: #Bow.
	].! !

QCharacter subclass: #QHuman
	instanceVariableNames: 'targetPosition idleThreshold'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Objects-Motion'!

!QHuman class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QHuman methodsFor: 'initialize' stamp: 'das 2/27/2007 12:02'!
colorize: aColor

	"For now, suppress this automatic colorization;
       the laser is the only colorized part.   The FutureFarmer palette
	 is sensitive; avatar colors will come with model variations."
! !

!QHuman methodsFor: 'animate' stamp: 'bgf 9/14/2007 23:23'!
doPlay

"the horse walks in a circle (more or less) and every so often, chooses a new motion"
	| currentPosition  delta |
	currentPosition := self globalPosition.
	targetPosition ifNotNil:[
		delta :=(currentPosition - targetPosition*(1@0@1))length.

		"If threshold is too low, there can be run-on at the end of motion;
	 	 but if too high, then walking becomes very jittery due to excess motion changes."
		(delta < self idleThreshold) ifTrue:[
			self useMotion: #Idle.
		] ifFalse:[
			self useMotion: #Walk.
		].
		container translation: targetPosition.		
	] ifNil:[
		100 atRandom = 1 ifTrue:[ self useRandomMotion.].
		self addRotationAroundY:-0.5.
	].
	super doPlay.! !

!QHuman methodsFor: 'animate' stamp: 'das 2/27/2007 15:51'!
goHere: trgt container: cont

	targetPosition := trgt.! !

!QHuman methodsFor: 'accessing' stamp: 'bgf 3/18/2007 16:19'!
idleThreshold
	"Max dstance from where we are to target
	 at which we use #idle motion."

	^ idleThreshold! !

!QHuman methodsFor: 'accessing' stamp: 'bgf 3/18/2007 16:20'!
idleThreshold: aScalar
	"Max dstance from where we are to target
	 at which we use #idle motion."
	idleThreshold := aScalar! !

!QHuman methodsFor: 'initialize' stamp: 'bgf 3/18/2007 16:25'!
initialize
	super initialize.
	idleThreshold := 0.35! !

!QHuman methodsFor: 'initialize' stamp: 'das 3/4/2009 11:47'!
makeFarmer

	| pHuman |

	pHuman := Q3DSModel 
					meshFromASEFile: 'Content/TreeFort/Avatars/Full-Blond-PaleGreen.ASE'
					atScale: 1.15.
	pHuman := pHuman collapse.
	pHuman allUpdateShape.
	pHuman updateTransform.

	self useGeometry: pHuman.
! !

!QHuman methodsFor: 'initialize' stamp: 'bgf 9/14/2007 23:24'!
makePipes

	| bvh pHuman |
	super initialize.

	bvh := QMotionCaptures manWalk.
	bvhRootHeight := bvh jointRoot rootHeight.

	pHuman := (bvh jointRoot makePipes) childNamed:'Root'.
	self geometry: pHuman.
	self addFoot:'LeftToes' heel: 'LeftHeel'.
	self addFoot:'RightToes' heel: 'RightHeel'.
	self addMotion: bvh motion name: #Walk.

"	path :=FileDirectory pathFrom:{'content'. 'Motion'. 'Male'. 'MWalkBkStand.bvh'}.
	bvh := BVHParser initializeWithFileName: path scale:0.1.
	self addMotion: bvh motion name: #walkBackStand.
"
	self addMotion: QMotionCaptures manIdle motion name: #Idle.

"
	path :=FileDirectory pathFrom:{'content'. 'Motion'. 'Male'. 'stances.bvh'}.
	bvh := BVHParser initializeWithFileName: path scale:0.35.
	self addMotion: bvh motion name: #stances."

"	path :=FileDirectory pathFrom:{'content'. 'Motion'. 'Avatar_Animation.bvh'}.
	bvh := BVHParser initializeWithFileName: path scale:0.35.
	self addMotion: bvh motion name: #testMotion.
"
	self useMotion: #Walk. "select the motion you wish to play here"
	! !

!QHuman methodsFor: 'initialize' stamp: 'das 3/4/2009 11:48'!
useGeometry: avatarMesh

	"Make the QHuman take on the given mesh costume as its appearance;
	 the mesh costume must match our canonical TGroup-as-skeleton structure.
	 Defines #Walk and #Idle motions from bvh human male data."

	| bvh am |
	super initialize.
	bvh := QMotionCaptures manWalk.
	bvhRootHeight := bvh jointRoot rootHeight.

	am := avatarMesh collapse.
	am allUpdateShape.

	self geometry: am.
	self addFoot:'LeftToes'.
	self addFoot:'RightToes'.
	self addMotion: bvh motion name: #Walk.

	self addMotion: QMotionCaptures manIdle motion name: #Idle.

	self useMotion: #Idle. "select the motion you wish to play here"
	targetPosition := nil.! !

TGroup subclass: #QCharacterProxy
	instanceVariableNames: 'contents avatar'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Objects-Motion'!

!QCharacterProxy class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QCharacterProxy methodsFor: 'accessing' stamp: 'das 2/27/2007 15:36'!
avatar: atar

	avatar := atar.! !

!QCharacterProxy methodsFor: 'accessing' stamp: 'das 2/27/2007 15:28'!
contents: cont
	
	contents := cont.	
	contents container: self.
	self addChild: contents.! !

!QCharacterProxy methodsFor: 'proxyMethods' stamp: 'bgf 3/19/2009 22:19'!
goHere: pos pitchYawRoll: pyr

	"Relay driving update to the costume"
	contents goHere: pos pitchYawRoll: pyr! !

!QCharacterProxy methodsFor: 'testing' stamp: 'das 2/27/2007 15:38'!
isAvatarProxy

	^ true.! !

!QCharacterProxy methodsFor: 'proxyMethods' stamp: 'das 2/27/2007 15:24'!
play
	
	contents play.! !

!QCharacterProxy methodsFor: 'render' stamp: 'bgf 11/4/2009 19:47'!
renderFrame: ogl

	| len lastTr rval |

	"If the camera is close to the avatar, fade the avatar to translucence...
	 (but not through portals... so we can see ourself in a mirror.)"
	ogl inPortal2D ifFalse: [
		len := self cameraDistanceSquared: ogl asPartOf: avatar.
		lastTr := ogl realTransparency.
		ogl transparency: (avatar alphaFor:len).
	].

	rval := super renderFrame: ogl.

	lastTr ifNotNil: [
		ogl transparency: lastTr.	
	].

	^ rval.! !

TGroup subclass: #QDestination
	instanceVariableNames: 'button destinationMode isListed leaderSite'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Objects-Places'!

!QDestination commentStamp: '<historical>' prior: 0!
A grouping class that makes its content into a 'destination'
Destinations have a transient mouseover GoTo button above them,
that pulls the avatar to the defined position, or one of the positions in a gathering.

Destinations appear in the places menu, using
the object name of the contained destination frame
(which can be edited in the Inspector.)

There are three destination modes:
	#viewpoint	- the transform of the destination frame is used directly
	#circle		- avatars form a circle around the destination (roundtable style)
	#front		- avatars form within a small arc looking at the front of the destination (panel style)

Usage is simple - if a tframe component is wrapped in a QDestination wrapper,
using the 'consistingOf:' initializer/constructor,
then the origin of the QDestination frame becomes an avatar destination.!

!QDestination class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QDestination class methodsFor: 'inspector' stamp: 'bgf 1/26/2010 12:42'!
addPOVChoicesToProperties: propertyDialog forFrame: destref

	"The Point of View choices for  a destination,
	 match the user-preferences viewpoint table choices."

	| vps values |
	vps := {'First Person' . 'Near' . 'Normal'. 'Far' . 'Unspecified'}.
	values :=  {'First Person' . 'Near' . 'Normal'. 'Far' . nil }.
	propertyDialog addRow: (
		(QChoicesPropertyRow new) 
			titleString: 'Suggested POV';
			setChoiceKeys: vps values: values;
			fetchBlock: [ : f | f get: #suggestedPOV ];
			acceptBlock: [ : v :f | f future suggestedPOV: v ];
			balloonHelp: 'Where the user''s camera is placed on arrival.';
			yourself).
! !

!QDestination class methodsFor: 'inspector' stamp: 'bgf 3/22/2010 11:12'!
addPropertyRows: propertyDialog forFrame: destref

	propertyDialog addRow: (
		(QStringPropertyRow new)
			titleString: 'Go-To Label';
			fetchBlock: [ :f | f get: #buttonLabel ] ;
			acceptBlock: [ :s :f | f future buttonLabel: s ] ;
			balloonHelp: 'Places label for this destination';		
		yourself).

	propertyDialog addRow: (
		(QChoicesPropertyRow new)
			titleString: 'Destination Type';
			setChoices: (self destinationTypeChoices);
			fetchBlock: [ : f | f get: #destinationMode ];
			acceptBlock: [ :t :f | f future destinationMode: t ];
			balloonHelp: 'How the destination arranges visiting avatars.';
		yourself).

	(destref refClass == QDestination) ifTrue: [ 
		propertyDialog addRow: (
			(QBooleanPropertyRow new) 
				titleString: 'Is Listed Place';
				fetchBlock: [ :f | f get: #isListed];
				acceptBlock: [ :b :f| f future isListed: b ];
				balloonHelp: 'Whether the destination can appear in the Places list.';
			yourself).
		self addPOVChoicesToProperties: propertyDialog forFrame: destref 
	].

	propertyDialog addRow: (
		(QChoicesPropertyRow new)
			titleString: 'Places Category';
			setChoices: (self categoryChoices);
			fetchBlock: [ : f | f get: #category ];
			acceptBlock: [ : c : f | f future category: c ];
			balloonHelp: 'Where the entry appears\(if Listed) in Places' withCRs;
			inputEnabled: ((destref get: #isListedLandmark) == true);
		yourself).

	propertyDialog addRow: (
		(QScalarPropertyRow new)
			titleString: 'Seating Range';
			fetchBlock: [  :f | f get: #seatingRange ];
			acceptBlock: [ : val :f | f future seatingRange: val ];
			balloonHelp: 'If non-zero, enables seat-finding.\Value is a margin added to the bounding box of the destination.' withCRs;
		yourself).

	propertyDialog addRow: (
		(QVector3PropertyRow new)
			titleString: 'Go-To Button Position';
			fetchBlock: [  :f | f get:#buttonPosition ];
			acceptBlock: [ :v3 :f  | f future buttonPosition: v3 ];
			balloonHelp: 'Placement of the overhead destination button.';
	 	yourself	).

	propertyDialog addRow: (
		(QTFramePropertyRow new)
			titleString: 'Leader Site';
			fetchBlock: [ :f | (f get: #leaderSite) ];
			acceptBlock: [ :v :f | f future leaderSiteFromProperties: v ] ;
			validateBlock: [ : frameRef : dest : row | self validateLeaderSite: frameRef forRow: row ];
			balloonHelp: 	
				('Destination for the meeting leader\when this is the meeting place.' withCRs);
			yourself).

	super addPropertyRows: propertyDialog forFrame: destref.

		"We equate the button label and object name."
! !

!QDestination class methodsFor: 'inspector' stamp: 'bgf 1/31/2009 09:30'!
categoryChoices
	"Places tool builtininSortOrder is a list of category names...
 	 we just have to double it up into a dictionary where each key has itself as value."
	^ Dictionary newFromPairs: (
		Array streamContents: [ : str | 	
			QPlacesTool builtinSortOrder do: [ : cat | str nextPut: cat; nextPut: cat ]]).! !

!QDestination class methodsFor: 'instance creation' stamp: 'bgf 5/12/2007 15:28'!
consistingOf: aTFrame
	"A new destination frame around the given tframe."
	^ self new consistingOf: aTFrame.
! !

!QDestination class methodsFor: 'defaults' stamp: 'bgf 8/17/2009 23:05'!
defaultButtonPosition
	^ 0@3@0! !

!QDestination class methodsFor: 'defaults' stamp: 'bgf 1/31/2009 10:02'!
defaultCategory
	"Reminder... my value must appear in... "  false ifTrue: [ QPlacesTool builtinSortOrder ].
	^ 'Destinations'! !

!QDestination class methodsFor: 'defaults' stamp: 'bgf 2/8/2008 14:15'!
defaultDestinationMode
	^ #circle! !

!QDestination class methodsFor: 'defaults' stamp: 'bgf 11/10/2009 15:56'!
defaultEventMask
	^ EventPointerOver + EventPointerDown! !

!QDestination class methodsFor: 'defaults' stamp: 'bgf 9/3/2009 13:53'!
defaultIsListed
	"Is this by default a listed landmark?"
	^ true! !

!QDestination class methodsFor: 'defaults' stamp: 'bgf 3/4/2009 11:47'!
defaultIsSticky
	"By default, we let suggested-POV nil/notNil determine stickiness"
	^ nil! !

!QDestination class methodsFor: 'defaults' stamp: 'bgf 4/22/2009 06:55'!
defaultSeatingRange
	"By default, no automatic seating within the destination."
	^ 0! !

!QDestination class methodsFor: 'inspector' stamp: 'bgf 1/31/2009 08:57'!
destinationTypeChoices
	^ Dictionary newFromPairs: #('Gather In Circle' circle 'Look At' front 'Look Out From' viewpoint)! !

!QDestination class methodsFor: 'inspector' stamp: 'bgf 3/22/2010 11:18'!
validateLeaderSite: frame forRow: row
	"Property-row support - validator for the leaderSite frame-type property."

	frame isValidFarRef ifFalse: [ ^ true ].	"Nil leaderSite is OK"
	(frame get: #asPlace) ifNil: [
		self inform: 'Leader Site must be a destination.'.
		^ false
	].
	^ true
! !

!QDestination methodsFor: 'accessing' stamp: 'bgf 5/20/2008 21:05'!
activityDetails
	"QActivityCapture details for the destination; include label for comparison."
	^ (self class name), ': ', (self buttonLabel).! !

!QDestination methodsFor: 'hierarchy' stamp: 'bgf 6/2/2009 09:39'!
addChild: c
	"My destination is my first child, or myself if childless;
	 update my label to be current with that rule."
	self destination = self ifTrue: [ 
		super addChild: c.
		self buttonLabel: (self buttonLabel).
	] ifFalse: [
		super addChild: c.
	].
	self isPlacementLocked: (self isPlacementLocked).! !

!QDestination methodsFor: 'transform' stamp: 'bgf 4/26/2009 12:37'!
asPlace
	"Return the QPlace to bring an avatar to this destination,
	 as if the goto button had been clicked."
	| result |
	button label: (self buttonLabel).
	result := button asPlace.
	(self destinationMode = #front) ifTrue: [
		result cameraSelector: #cameraStanceFor:
	].
	^ result! !

!QDestination methodsFor: 'hierarchy' stamp: 'bgf 7/26/2009 13:54'!
bestSeatWithin: tbox nearest: avatarPosition
	"For viewing things, find the seat closest to it, not the avatar...
	 this helps with best-available behavior in seating in theaters, etc."
	| target |
	target := self isFrontGathering
				ifTrue: [ self destination globalPositionReal ]
				ifFalse: [ avatarPosition ].

	^ super genericBestSeatWithin: tbox nearest: target
! !

!QDestination methodsFor: 'accessing' stamp: 'bgf 4/13/2008 09:00'!
button
	"My goto-button"
	^ button! !

!QDestination methodsFor: 'accessing' stamp: 'bgf 3/16/2009 15:52'!
buttonLabel
	"The identification shown in mouse-over tags, etc."
	^ button ifNotNil: [ button label ] ifNil: [ 'Destination' ]! !

!QDestination methodsFor: 'accessing' stamp: 'bgf 1/31/2009 09:17'!
buttonLabel: aString
	"The mouse-over label for the button."
	button label: aString! !

!QDestination methodsFor: 'accessing' stamp: 'bgf 2/8/2008 16:22'!
buttonPosition
	"The offset of the button within the larger whole;
	 a suitable default is computed when 'consistingOf:' is used to install a destination."

	^ button translation! !

!QDestination methodsFor: 'accessing' stamp: 'bgf 6/27/2009 13:13'!
buttonPosition: aTranslation
	"The offset of the button within the larger whole;
	 a suitable default is computed when 'consistingOf:' is used to install a destination."

	button translation: aTranslation;
			shapeChanged.			"Prod the bounds, they seem to need it sometimes."! !

!QDestination methodsFor: 'transform' stamp: 'bgf 4/26/2009 14:00'!
cameraStanceFor: anAvatar
	"The stance to apply to an avatar (camera) to bring it face-to-face.
	 This is a shortcut to the inner implementation behind normal destination-finding,
	 avoiding all the gathering-circle offset-angle stuff."

	^ QGotoButton frame: (self destination) cameraStanceFor: anAvatar! !

!QDestination methodsFor: 'accessing' stamp: 'bgf 1/31/2009 09:18'!
category
	"The places-menu category for this destination."
	^ button category! !

!QDestination methodsFor: 'accessing' stamp: 'bgf 1/31/2009 09:18'!
category: aString
	"Not persisted... the places-list category for this destination."
	button category: aString! !

!QDestination methodsFor: 'testing' stamp: 'bgf 4/16/2009 17:18'!
childrenForBrowser
	^ self isAtomic ifTrue: [
		#()
	] ifFalse: [
		self children copyWithout: button
	].! !

!QDestination methodsFor: 'testing' stamp: 'bgf 6/22/2009 12:12'!
childrenForPersistence
	"Don't omit things from persistence just because we're atomic."
	^ self children copyWithout: button
! !

!QDestination methodsFor: 'transform' stamp: 'bgf 8/11/2009 13:39'!
circleDestinationStanceFor: anAvatar 

	"The transform to apply to the avatar to bring it to me."

	| distance len ht tan d1 d2 ext |

	ext := self destination boundingBox extent.

	len := (ext x max: ext z) * 0.5.

	"Y basis for distance -to-view determination,  based on viewing posture -
	 we need to see the whole from where we get placed, regardless of where
	 the avatar eye-line (horizon) falls compared to the destination mass.  "
	ht  := (((anAvatar heightForPosture:(self posture)) - ext y) max: (ext y)) * 0.5.

	"Generally we base the view on the 45-degree standard window,
	 so that everyone is placed at the same distance around normal tables and such,
	 which allows occupancy checks to work properly."

	tan := 0.414213562373095.    "((anAvatar viewAngle degreesToRadians/2.0) tan)."
	d1 := ht / tan.
	d2 := len / (tan * 1.25 "Aspect").
	distance := (d1 max: d2) max: 7.0.

	"The distance is heuristic - we want gathered users to be able to see each other
	 from Near POV, so the destination can serve as a meeting-area."

	^ self destinationStanceFor: anAvatar distance: distance angle: 180.0! !

!QDestination methodsFor: 'initialize' stamp: 'bgf 3/4/2009 12:17'!
consistingOf: aTFrame

	"Make this a wrapper around the given frame."
	| ext hover |
	super consistingOf: aTFrame.
	self objectName: 'Destination (', (aTFrame objectName), ')'.

	"Adapt the button offset to stand above the center of the new object."
	ext := aTFrame boundingBox extent.
	hover := ((ext x max: ext z) / 3) min: 2.5 max: 1.2.
	button suggestedPOV: #Normal.
	button destination: self.
	button label: (aTFrame labelForBrowser).
	button translation: (aTFrame localCenter + (0@((ext y /2.0 max: 3) + hover)@0)).! !

!QDestination methodsFor: 'persist' stamp: 'bgf 3/22/2010 11:03'!
constructFrom: pStream

	super constructFrom:pStream.
	self class == QDestination ifTrue:[
		self children: ((pStream readSequence:'children' default: self children) copyWithFirst: button).
	].

	self category: (pStream readString: 'category' default: (self category)).
	self destinationMode: (pStream readString: 'destinationMode' 
							default: self class defaultDestinationMode).
	self suggestedPOV: (pStream readString: 'suggestedPOV' 
						default: (self povForDestinationMode: (self destinationMode))).
	self isListed: (pStream readBool: 'isListedLandmark' default: (self isListed)).	
	button translation: (pStream readVector3: 'buttonPosition' default: (button translation)).
	self buttonLabel: (pStream readString: 'buttonLabel' default: (self destination objectName)).
	self seatingRange: (pStream readNumber: 'seatingRange' default: (self seatingRange)).

	self leaderSite: (pStream readObject: 'leaderSite' default: self leaderSite).! !

!QDestination methodsFor: 'initialize' stamp: 'bgf 4/16/2009 18:36'!
defaultIsAtomic
	^ (self class ~~ QDestination)! !

!QDestination methodsFor: 'accessing' stamp: 'bgf 7/23/2009 02:35'!
destination
	"My destination TFrame is my first child other than the button, if any,
	 or myself if I have no content.  It is the basis for most positioning of incomers."
	frameChildren isEmptyOrNil ifTrue: [ ^ self ].
	^ (frameChildren at: 1) = button 
		ifTrue: [
			frameChildren size < 2 
				ifTrue: [  self ]
				ifFalse: [ frameChildren at: 2 ]
			] ifFalse: [
				frameChildren at: 1
			].
	! !

!QDestination methodsFor: 'transform' stamp: 'bgf 4/26/2009 13:47'!
destinationExtent
	"2-D destination-extent 
	 for computing front-viewing stances."

	^ self destination = self
		 ifTrue: [ super destinationExtent ]
		 ifFalse: [ self destination destinationExtent ]! !

!QDestination methodsFor: 'accessing' stamp: 'bgf 2/8/2008 12:28'!
destinationMode
	^ destinationMode ifNil: [ #front ].! !

!QDestination methodsFor: 'accessing' stamp: 'bgf 7/22/2009 11:14'!
destinationMode: aSym
	"One of #front, #circle, #viewpoint.
	 Determines how avatars gather. "

	destinationMode := aSym asSymbol.
	button suggestedPOV: (self povForDestinationMode: destinationMode).
	button spin: (destinationMode ~= #front).
	self asPlace.  "Trigger inline updates to selector"
! !

!QDestination methodsFor: 'transform' stamp: 'jjl 8/5/2008 22:38'!
destinationStanceFor: anAvatar

	^ self destinationMode caseOf: {
		[#front] -> [ self frontDestinationStanceFor: anAvatar ].
		[#circle] -> [ self circleDestinationStanceFor: anAvatar ].	
		[#viewpoint] -> [ self viewpointDestinationStanceFor: anAvatar ].
	} otherwise: [nil].! !

!QDestination methodsFor: 'transform' stamp: 'bgf 7/27/2009 00:16'!
destinationStanceFor: anAvatar distance: distance angle: degrees

	"Common tail worker for destination-finding in various modes;
	 always floor the destination to the avatar height to prevent teleports into the ground."

	| h stance adjustment |
	stance :=  QGotoButton 
			stanceFor: anAvatar
			toView: (self destination)
			inDestination: self
			inAngle: degrees
			atDistance: distance.

	h := (self root respondsTo: #groundHeightAt:) ifTrue: [
			self root groundHeightAt: (stance here)
		] ifFalse: [
			0.1			" A little slop for typical floors"
		].

	adjustment := (((anAvatar heightForPosture:(self posture)) + h) - stance yHere) max: 0.
	stance yHere:  (stance yHere + adjustment).
	stance yThere: (stance yThere + adjustment).
	^ stance! !

!QDestination methodsFor: 'editing' stamp: 'bgf 2/11/2010 13:42'!
editorBoundingBox
	"If I have a destination mesh, use its bounds rather than mine,
	 so that the edit-box looks properly sized (i.e. omitting the transient hover goto-button)"

	false ifTrue: [
		"Disabled; the appropriate rescaling and translation
	 	 of the child bounding box is tricky to figure
	 	 given that the editor tries to recenter it around the mass of the object.
		(NB This affects only editor boxing, not any other behavior.)"
		^ self destination = self 
			ifTrue: [ self boundingBox ]
			ifFalse: [self destination boundingBox scaled: (self scaleInMatrix) ].
	].
	^ self boundingBox! !

!QDestination methodsFor: 'transform' stamp: 'bgf 7/26/2009 22:24'!
frontDestinationStanceFor: anAvatar 
	"The transform to apply to the avatar to bring it to view me."
	| dest result |
	dest := self destination.

	"Use the standard front-viewing stance-finders."
	self hasCameraStance ifTrue: [
		result := QGotoButton 
					groundedStanceFor: anAvatar toView: dest inDestination: self.
	] ifFalse: [
		result := self destinationStanceFor: anAvatar 
						distance: (QGotoButton distanceToSee: (dest destinationExtent)) 
						angle: 30.
	].
	^result! !

!QDestination methodsFor: 'testing' stamp: 'bgf 4/26/2009 13:13'!
hasCameraStance
	"True if the destination defines a focal camera viewing stance
	 (which is done for most kinds of 2-D display areas like display panels.) "
	^ self destinationMode = #front 
	    or: [ button asPlace hasCameraStance ]! !

!QDestination methodsFor: 'initialize' stamp: 'bgf 11/10/2009 16:31'!
initialize
	super initialize.

	button := QGotoButton new.
	button spin: true.
	button destination: self.
	button disable.
	button translation: (self class defaultButtonPosition).
	self addChild: button.

	self isListed: (self class defaultIsListed).
	self category: (self class defaultCategory).
	self destinationMode: (self class defaultDestinationMode).
	self seatingRange: (self class defaultSeatingRange).! !

!QDestination methodsFor: 'testing' stamp: 'bgf 2/8/2008 13:48'!
isCircleGathering
	^ self destinationMode == #circle! !

!QDestination methodsFor: 'testing' stamp: 'bgf 11/12/2009 10:54'!
isEditableGroup
	"C.F. TGroup; plain destination instances are not atomic so they can be excised with ungroup,
	 but derived components are by default atomic."
	^ (self class == QDestination) and: [ self isAtomic not ]! !

!QDestination methodsFor: 'testing' stamp: 'bgf 2/8/2008 13:47'!
isFrontGathering
	^ self destinationMode == #front! !

!QDestination methodsFor: 'testing' stamp: 'bgf 2/16/2010 09:48'!
isGroundBound
	"Better choice on average for generic tables and such that are wrapped as Destinations
	 and placed in the catalog."
	^ true! !

!QDestination methodsFor: 'accessing' stamp: 'bgf 9/3/2009 13:54'!
isListed
	"Am I a Place-listed landmark (if visible) ?"
	^ isListed! !

!QDestination methodsFor: 'accessing' stamp: 'bgf 9/3/2009 14:07'!
isListed: bListed
	"May I be a place-listed landmark (if visible) ?"
	isListed = bListed ifFalse: [
		isListed := bListed.
	].! !

!QDestination methodsFor: 'accessing' stamp: 'bgf 9/3/2009 14:07'!
isListedLandmark
	"Protocol - should I currently appear in Places ?"
	^ self isListed and: [ self visible ]! !

!QDestination methodsFor: 'accessing' stamp: 'bgf 6/2/2009 09:27'!
isPlacementLocked: bLock
	super isPlacementLocked: bLock.

	"Destinations are often not atomic (the target object is browsable)
	 but locking in place should apply to the whole."
	self destination ifNotNil: [ : d | d = self ifFalse: [ d isPlacementLocked: bLock ]].! !

!QDestination methodsFor: 'accessing' stamp: 'bgf 3/4/2009 12:08'!
isSticky
	"The value of the sticky-but-no-POV override flag."
	^ button isSticky! !

!QDestination methodsFor: 'accessing' stamp: 'bgf 3/4/2009 12:08'!
isSticky: aBool
	"The value of the sticky-but-no-POV override flag."
	button isSticky: aBool! !

!QDestination methodsFor: 'testing' stamp: 'bgf 2/8/2008 16:18'!
isViewpoint
	"True iff this is a 'viewpoint destination - the avatar is pulled to the destination xform,
	 rather than into a viewing position to look at the destination."

	^ self destinationMode == #viewpoint! !

!QDestination methodsFor: 'accessing' stamp: 'bgf 7/9/2009 13:11'!
labelForBrowser
	| tag |
	tag := 	(self destinationMode caseOf: {
			[#viewpoint] -> [ 'Viewpoint'].
			[#front] -> ['Destination'].
		} otherwise: [
			'Gathering' 
		]).

	^ tag, ': ', (self buttonLabel)

! !

!QDestination methodsFor: 'accessing' stamp: 'bgf 3/22/2010 11:01'!
leaderSite
	^ leaderSite! !

!QDestination methodsFor: 'accessing' stamp: 'bgf 3/22/2010 11:02'!
leaderSite: tframe
	leaderSite := tframe! !

!QDestination methodsFor: 'accessing' stamp: 'bgf 3/22/2010 11:16'!
leaderSiteFromProperties: lsDest
	leaderSite := lsDest
! !

!QDestination methodsFor: 'transform' stamp: 'bgf 8/6/2009 16:11'!
localCenter
	^ self destination = self
			ifTrue: [ super localCenter ]
			ifFalse: [ self destination localCenter + (self translation) ]! !

!QDestination methodsFor: 'menu' stamp: 'bgf 7/9/2009 13:21'!
menuTitle
	^ self buttonLabel! !

!QDestination methodsFor: 'accessing' stamp: 'bgf 3/5/2010 11:40'!
objectID: oid
	| redoButton |
	redoButton := (button notNil and: [ button asPlace destinationID = objectID ]).
	super objectID: oid.
	redoButton ifTrue: [ button destination: self ].! !

!QDestination methodsFor: 'accessing' stamp: 'bgf 4/13/2008 11:21'!
objectName: aString
	"If the button label is the default, update it."
	(self buttonLabel = self objectName) ifTrue: [
		self buttonLabel: aString
	].
	super objectName: aString.! !

!QDestination methodsFor: 'persist' stamp: 'bgf 3/22/2010 11:50'!
persistOn: pStream

	super persistOn: pStream.
	self class == QDestination ifTrue:[
		pStream storeSequence: 'children' value: (self childrenForPersistence) default: #().
	].
	pStream storeVector3: 'buttonPosition' 
			 value: (button translation) 
			default: (self class defaultButtonPosition).
	pStream storeString: 'category' value: (self category) default: (self class defaultCategory).

	pStream storeString: 'destinationMode' 
			 value: (self destinationMode) 
			 default: (self class defaultDestinationMode).

	self buttonLabel ifNotNil: [	"Persist-Stream cannot save a nil string"
		pStream storeString: 'buttonLabel' value: (self buttonLabel) 
					default: (self destination objectName).
	].
	pStream storeString: 'suggestedPOV' value: (self suggestedPOV) 
				default: (self povForDestinationMode: (self destinationMode)).
	pStream storeBool: 'sticky' value: (self isSticky) 
				default: (self class defaultIsSticky).
	pStream storeBool: 'isListedLandmark' value: (self isListed) 
				default: (self class defaultIsListed).
	pStream storeNumber: 'seatingRange' value: (self seatingRange) 
				default: (self class defaultSeatingRange).
	
	(leaderSite notNil and: [ pStream sceneRoot isOrContains: leaderSite ] ) 
	 ifTrue: [
		pStream storeObject: 'leaderSite' value: (self leaderSite) default: nil.
	].! !

!QDestination methodsFor: 'accessing' stamp: 'bgf 4/21/2009 12:56'!
posture
	"If indicated, the required posture (Sitting, Standing, Hovering) for this destination.
	 May return nil (meaning no specific posture is mandated by the destination itself.)"
	^ button posture! !

!QDestination methodsFor: 'accessing' stamp: 'bgf 4/13/2008 09:06'!
povForDestinationMode: mode
	"The POV (camera/stickiness) default assignment for a given destination-mode."
	^ mode caseOf: {
		[#front] -> ['First Person'].
		[#circle] -> ['Normal' ].
		[#viewpoint] -> [ nil ].
	} otherwise: [ nil ].
! !

!QDestination methodsFor: 'initialize' stamp: 'bgf 3/13/2010 17:08'!
releaseChildrenAndDestroy
	"Support for user-edit 'Ungroup' to release the destination from this wrapper.
	 The button is not left around, it is destroyed with the QDestination;
	 only the original destination frame is released."
	button ifNotNil: [
		button destroy.
		button := nil.
	].
	^ super releaseChildrenAndDestroy.! !

!QDestination methodsFor: 'hierarchy' stamp: 'bgf 4/13/2008 11:44'!
removeChild: c
	super removeChild: c.

	"My destination is my first child, or myself if childless;
	 update my label to be current with that rule."
	self buttonLabel = (c objectName) ifTrue: [
			self buttonLabel: (self buttonLabel)
		]! !

!QDestination methodsFor: 'accessing' stamp: 'bgf 4/22/2009 07:31'!
seatingRange
	"If zero, no auto-seating within the destination.
	 If > zero, then this is a distance in standard units added to the bounding box,
	 to define the total volume in which seats are detected."
	^ button seatingRange! !

!QDestination methodsFor: 'accessing' stamp: 'bgf 4/22/2009 12:48'!
seatingRange: radius
	"If zero, no auto-seating within the destination.
	 If > zero, then this is a distance in standard units added to the bounding box,
	 to define the total volume in which seats are detected."
	button seatingRange: radius! !

!QDestination methodsFor: 'accessing' stamp: 'bgf 1/31/2009 09:17'!
suggestedPOV
	^ button suggestedPOV
	! !

!QDestination methodsFor: 'accessing' stamp: 'bgf 1/31/2009 09:17'!
suggestedPOV: aString
	"One of the standard goto-button POV strings, or nil for non-sticky destinations.
	 A default for this is supplied whenever the destination mode is set;
	 see povForDestinationMode: "

	button suggestedPOV: aString! !

!QDestination methodsFor: 'transform' stamp: 'bgf 8/7/2009 13:02'!
viewpointDestinationStanceFor: anAvatar

	"Take the avatar to stand atop the destination itself."
	| xform ray boxTop |
	self root ifNil: [ ^ nil ].

	xform := self destination globalTransform clone.
	boxTop := self destination = self ifTrue: [
					0	"If I'm empty,  use my position as the destination point."
				] ifFalse: [
					self destination boundingBox max y.
				].
	ray := self root downRayAt: (self destination globalTransform translation
								 + (Vector3 x: 0 y: boxTop + 1  z: 0)).
	ray pointerPickTree: self destination excluding: { button }.
	ray selection frame ifNotNil: [
		"Picked a place on the geometry of the destination."
		xform translation: (ray selection globalPoint) + 
							 (Vector3 x: 0 y: ((anAvatar heightForPosture:(self posture))) z: 0).
	] ifNil: [
		"Fallback - go atop the bounding box."
		xform translation: xform translation+ 
							(0@(boxTop + (anAvatar heightForPosture:(self posture)))@0).
	].

	^Stance here: (self root freePlaceFor: anAvatar near: (xform translation))
		       there: (xform translation - (xform lookAt * 20)).! !

QDestination subclass: #QSeat
	instanceVariableNames: 'seatOffset'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Objects-Places'!

!QSeat commentStamp: '<historical>' prior: 0!
A destination preconfigured for Sitting as posture and Viewpoint as placement rule.
A seat is wrapper around a frame  (any chairlike object) that provides ... a seat.
Clicking on a seat, or using various other navigation controls (like the Sit control-bar button)
pulls the user into the seat and into a seating posture (and hence into Look-mode.)

The seat offset is the offset from the top of the chair bounding box to where the butt goes.
(Y counts downwards, so most seat-offsets include a Y of somewhere around 1.5 or so,
since chair volumes are usually about that much higher that where the seat surface is.)

Seats participate in complicated conspiracies with the place-finding system,
so that when you go to a Destination that specifies a seating range,
seats within the range are found and the avatar auto-seated.
Similarly there is detection of going to a viewing-destination like a Panel
that is 'right in front' of a seat.   See QAvatarReplica - the places and seating categories -
for most of the interesting Seating code.
!

!QSeat class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QSeat class methodsFor: 'inspector' stamp: 'bgf 1/26/2010 12:34'!
addPropertyRows: propertyDialog forFrame: seatref

	propertyDialog addRow: (
		(QVector3PropertyRow new)
			titleString: 'Seating Position';
			fetchBlock: [ :frame | frame get:#seatOffset ];
			acceptBlock: [ : v3 :frame | frame future seatOffset: v3 ];
			balloonHelp: 'Offset of the seat from mesh origin.';
	 	yourself	).

	self addPOVChoicesToProperties: propertyDialog forFrame: seatref.

	super addPropertyRows: propertyDialog forFrame: seatref.

	propertyDialog removeRow: 'Destination Type'.
	propertyDialog removeRow: 'Places Category'.
	propertyDialog removeRow: 'Go-To Button Position'.
	propertyDialog removeRow: 'Seating Range'.! !

!QSeat class methodsFor: 'defaults' stamp: 'bgf 8/9/2009 01:02'!
defaultAllowWalk
	^ false! !

!QSeat class methodsFor: 'defaults' stamp: 'bgf 11/10/2009 10:39'!
defaultButtonPosition
	"The goto buttons over seat-destinations are suppressed (the seat is clickable itself).
	 So keep them inside where they don't affect bounding boxes.
	 At y=1 rather than zero since most seats are ground-bound, with 0 meaning bottom edge."
	^Vector3 x: 0 y: 1 z: 0! !

!QSeat class methodsFor: 'defaults' stamp: 'bgf 4/11/2008 22:58'!
defaultDestinationMode
	^ #viewpoint! !

!QSeat class methodsFor: 'defaults' stamp: 'bgf 11/10/2009 15:56'!
defaultEventMask
	^ EventPointerOver + EventPointerDown! !

!QSeat class methodsFor: 'defaults' stamp: 'bgf 6/4/2009 16:12'!
defaultIsFromTemplate
	"Seating is almost always part of the template arrangement."
	^ true! !

!QSeat class methodsFor: 'defaults' stamp: 'bgf 9/3/2009 13:57'!
defaultIsListed
	"Overrides base... seats themselves are never listed (their containing seating-ranged place is)."
	^ false! !

!QSeat class methodsFor: 'defaults' stamp: 'bgf 3/14/2009 12:28'!
defaultIsSticky
	"Overrides base... Seats only work if they're sticky, to set the posture."
	^ true! !

!QSeat class methodsFor: 'defaults' stamp: 'bgf 2/27/2009 09:16'!
defaultSeatOffset
	^ Vector3 x: 0 y:0 z:0.5! !

!QSeat class methodsFor: 'instance creation' stamp: 'bgf 6/3/2009 12:49'!
fakeSeat: label 
	"Creates a QSeat wrapping an invisible seat-block.   Aka the InvisibleAssPad (tm).
	 These can be constructed at seating points in furniture meshes.
	 The furniture mesh should be non-solid and possibly non-pickable;
	 the interaction is that the fake seat pad becomes the click-handler for sitting."
	| pillow result |
	pillow := TCube new.
	pillow
		material: (QMaterials clearMaterial);
		extent: (Vector3 x: 2.1 y: 1.4 z: 2.2);
		solid: false.

	result := self consistingOf: pillow.
	result buttonLabel: label.
	result buttonPosition: 0@7.0@0.
	result isAtomic: true.
	result doContextMenu: false.
	result seatOffset: (0@1.5@0.25).
	^ result! !

!QSeat methodsFor: 'accessing' stamp: 'bgf 2/26/2009 22:19'!
addChild: c

	super addChild: c.

	self stopScript: #floorEnter:.
	self startScript: #floorEnter: when: { self destination . #floorEnter }.! !

!QSeat methodsFor: 'transform' stamp: 'bgf 7/22/2009 22:21'!
asPlace
	"Seats are exclusive single-occupancy... 
	 they do not present themselves as places for census or interaction
	 unless they are currently not occupied."
	self isOccupied ifTrue: [ ^ nil ].
	^ super asPlace.
! !

!QSeat methodsFor: 'transform' stamp: 'bgf 7/24/2009 02:06'!
asPlaceAnyway
	"Seats are exclusive single-occupancy... 
	 they do not present themselves as places for census or interaction
	 unless they are currently not occupied.   This gets the place whether occupied or not."
	^ super asPlace.
! !

!QSeat methodsFor: 'seat-finding' stamp: 'bgf 7/26/2009 14:34'!
bestSeatWithin: tbox nearest: avatarPosition
	"I'm the very best seat I know of."
	^ ((tbox pointInBox: self globalPositionReal) and: [ self isOccupied not ])
		ifTrue: [ self  ]
		ifFalse: [ nil ].
! !

!QSeat methodsFor: 'initialize' stamp: 'bgf 4/17/2009 15:21'!
consistingOf: aTFrame

	"Make this a wrapper around the given frame."
	| ext hover |
	super consistingOf: aTFrame.
	self objectName: 'Seat-', (aTFrame objectName).

	"Adapt the button offset to stand above the center of the new object."
	ext := aTFrame boundingBox extent.	
	hover := ((ext x max: ext z) / 3) min: 4.5 max: 3.6.

	button destination: aTFrame.
	button label: (aTFrame labelForBrowser).
	button posture: #Sitting.
	button isSticky: true.
	button suggestedPOV: #Near.
	button translation: (aTFrame localCenter + (0@((ext y /2.0 max: 3) + hover)@0)).! !

!QSeat methodsFor: 'persist' stamp: 'bgf 11/10/2009 04:06'!
constructFrom: pStream
	super constructFrom:pStream.
	self class == QSeat ifTrue:[
		self children: ((pStream readSequence:'children' default: self children) copyWithFirst: button).
	].
	self seatOffset: (pStream readVector3: 'seatOffset' default: (self seatOffset)).
	"Convert legacy  (the trick is that the 'default' will come out of what we just read.)"
	self seatVerticalOffset: (pStream readNumber: 'seatVerticalOffset' default: (self seatVerticalOffset)).

	self buttonPosition: (self class defaultButtonPosition).	 
			"Suppressed, so prevent it from enlarging bounds.  (Forced here to clean up legacy seats.)"! !

!QSeat methodsFor: 'initialize' stamp: 'bgf 4/16/2009 18:36'!
defaultIsAtomic
	^ (self class ~~ QSeat)! !

!QSeat methodsFor: 'transform' stamp: 'bgf 8/9/2009 02:48'!
destinationStanceFor: anAvatar
	"Take the avatar to sit on me."
	| xform  off box trans d |

	d := self destination.
	xform := d globalTransformReal clone.
	off := xform orientation transformVector: (self seatOffset).
	box := (d boundingBox transformedBy: (d orientation)).
	trans := xform translation - off.
	trans y: (trans y + (box max y) + anAvatar seatToEyes).

	"We use my global orientation to determine the direction,
	 as many meshes use a reversed z convention from ours.
	 This way you can rotate the geometry inside me (seat-destination is a group) to get it right."

	^Stance here: trans
		       there: (trans - (self globalTransformReal lookAt)).! !

!QSeat methodsFor: 'initialize' stamp: 'bgf 11/10/2009 16:35'!
initialize
	super initialize.

	button posture: #Sitting.
	button isSticky: true.
	button suppressed: true.
	button suggestedPOV: #Near.

	seatOffset := self class defaultSeatOffset.

	! !

!QSeat methodsFor: 'testing' stamp: 'bgf 4/16/2009 11:34'!
isComponent
	^ true! !

!QSeat methodsFor: 'testing' stamp: 'bgf 6/1/2009 11:43'!
isEditableGroup
	"C.F. TGroup; plain instances are not atomic so they can be excised with ungroup,
	 but derived components are by default atomic."
	^ (self class == QSeat) and: [ self isAtomic not ]! !

!QSeat methodsFor: 'testing' stamp: 'bgf 4/17/2009 10:27'!
isGroundBound
	^ true! !

!QSeat methodsFor: 'testing' stamp: 'bgf 4/12/2008 00:24'!
isListedLandmark
	"Seats do not appear in the places menu."
	^false! !

!QSeat methodsFor: 'testing' stamp: 'bgf 8/9/2009 14:38'!
isOccupied
	"Anyone sitting on me?"
	self root ifNotNilDo: [ : root |
		^ (root allUsers anySatisfy: [ : av | 
			av sittingPlace 
				ifNotNilDo: [ : aplace | aplace destinationID = objectID  ]
				ifNil: [ false ]]).
		].
	^ false! !

!QSeat methodsFor: 'testing' stamp: 'bgf 8/6/2009 22:08'!
isSeat 
	"Faster than isKindOf: ... for seat-finding protocols."
	^ true! !

!QSeat methodsFor: 'accessing' stamp: 'bgf 7/9/2009 13:12'!
labelForBrowser
	"We name the thing explicitly."
	^ 'Seat: ', (self buttonLabel)! !

!QSeat methodsFor: 'persist' stamp: 'bgf 8/9/2009 02:41'!
persistOn: pStream
	super persistOn: pStream.
	self class == QSeat ifTrue:[
		pStream storeSequence: 'children' value: (self childrenForPersistence) default: #().
	].
	pStream storeVector3: 'seatOffset' value: (self seatOffset) default: (self class defaultSeatOffset).
! !

!QSeat methodsFor: 'events' stamp: 'bgf 2/10/2010 17:35'!
pointerDown: event
	"The user can click directly on a chair to sit.
	 If you're sitting in a chair, clicking it stands you up."
	| place occupant |
	self root ifNil: [ ^nil ].
	event avatar doWarpMoves ifFalse: [ ^ nil ].
	event redButtonPressed ifFalse: [ ^ nil ].

	place := super asPlace.
	occupant := self root allUsers 
					detect: [ : av | av currentPlace = place ]
					 ifNone: [ nil ].
	occupant ifNil: [
		button bringToDestination: event
	] ifNotNil: [
		occupant = event avatar ifTrue: [ 
			occupant allowedToMove ifTrue: [ occupant togglePosture ]
		]].

		! !

!QSeat methodsFor: 'events' stamp: 'bgf 2/26/2009 22:32'!
pointerEnter: evt
	evt avatar posture = #Sitting ifFalse: [
		evt avatar signal: #overWarpPlace with: self asPlace.
	].! !

!QSeat methodsFor: 'events' stamp: 'bgf 2/26/2009 22:37'!
pointerOver: evt
	evt avatar posture = #Sitting ifFalse: [
		evt avatar signal: #overWarpPlace with: self asPlace.
	].! !

!QSeat methodsFor: 'render' stamp: 'bgf 2/17/2010 17:54'!
renderSelection: ogl box: box color: color thickness: lineWid
	"Render selection-boxing indicators that this object is selected.
	 For seats, unless we are scene-editing, we outline the waist only,
	 so clicking chairs for go-to doesn't creat annoying busy wireframes."
	
	ogl avatar ifNotNil: [ : avatar | 
		avatar isEditingTemplate ifTrue: [ 
			^ super renderSelection: ogl box: box color: color thickness: lineWid
		]].
		
	ogl glDisable: GLFog.

	"Show the hidden part in a stippled thin outline."

	(OpenGL maxRendererFlags anyMask: 8  "B3D-Antialiasing " ) ifTrue: [
		"If we're not antialiasing, don't do stipples either; they cause trouble with very-lame cards."
		ogl glDisable: GLDepthTest.
		ogl glLineStipple: 1 with: 4r03030303.
		ogl glEnable: GLLineStipple.
		box outlineXZ: ogl color: (color clone alpha: (color alpha * 0.6)) thickness: 1.
		ogl glEnable: GLDepthTest.
		ogl glDisable: GLLineStipple.
	].
	box outlineXZ: ogl color:color thickness: lineWid.
	ogl glEnable: GLFog.

	^ box! !

!QSeat methodsFor: 'accessing' stamp: 'bgf 2/27/2009 08:58'!
seatOffset
	"Adjustment for the seat center, 
	 as compared to what is returned by a vertical pick down from above the destination."
	^ seatOffset! !

!QSeat methodsFor: 'accessing' stamp: 'bgf 2/27/2009 08:59'!
seatOffset: aVector3
	"A vertical adjustment for the seat plane, 
	 as compared to what is returned by a vertical pick down from above the destination."
	seatOffset := aVector3 clone.! !

!QSeat methodsFor: 'accessing' stamp: 'bgf 2/27/2009 09:01'!
seatVerticalOffset
	"A vertical adjustment for the seat plane, 
	 as compared to what is returned by a vertical pick down from above the destination."
	^ seatOffset y! !

!QSeat methodsFor: 'accessing' stamp: 'bgf 2/27/2009 09:02'!
seatVerticalOffset: aScalar
	"A vertical adjustment for the seat plane, 
	 as compared to what is returned by a vertical pick down from above the destination."
	seatOffset y: aScalar! !

TGroup subclass: #QDraggable
	instanceVariableNames: 'adapter'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Objects-Manipulation'!

!QDraggable commentStamp: '<historical>' prior: 0!
A TGroup that is set up as the 'grip' of a QDragAdapter.
Initialized so that itself is both the grip
and luggage of a QDragAdapter.

The luggage (outermost dragged frame)
can be changed during initialization or in subclasses;
the grip must remain the draggable itself.

The luggage must always be an ancestor of the grip (or the grip itself)
but never a child.

The QDragAdapter is available through the #adapter accessor.
!

!QDraggable class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QDraggable methodsFor: 'accessing' stamp: 'bgf 7/18/2006 11:59'!
addChild: aTFrame

	"Draggables should always be disabled when reparented into
	 the scope of another draggable."

	"TODO: walk the ancestor tree..."
	(aTFrame isKindOf: QDraggable) ifTrue: [
		aTFrame disable.
	].
	^ super addChild: aTFrame! !

!QDraggable methodsFor: 'accessing' stamp: 'bgf 9/14/2006 17:22'!
disable
	adapter disable! !

!QDraggable methodsFor: 'accessing' stamp: 'bgf 8/16/2009 01:25'!
dragAdapter
	^ adapter! !

!QDraggable methodsFor: 'accessing' stamp: 'bgf 9/14/2006 17:22'!
enable
	adapter enable! !

!QDraggable methodsFor: 'initialize' stamp: 'bgf 9/14/2006 18:03'!
initialize
	"By default, we are enabled, droppable, and deletable, but not movable."
	super initialize.
	adapter := QDragAdapter forLuggage: self grip: self.! !

!QDraggable methodsFor: 'menu' stamp: 'bgf 9/14/2007 21:25'!
menuStyle 
	^ QStyles menuStyle! !

!QDraggable methodsFor: 'accessing' stamp: 'bgf 12/10/2006 12:55'!
payload
	"Implementing subclasses may override this to return the payload
	 that a drop should make available to the #acceptDrop: recipient.
	 By default, the draggable presents itself as the dropped content."

	^ self
! !

!QDraggable methodsFor: 'persist' stamp: 'bgf 10/19/2006 15:38'!
persistOn:pStream

	super persistOn: pStream.
	self class == QDraggable ifTrue:[
		"See note in TGroup"
		pStream storeSequence: 'children' value: self children default: #().
	].! !

!QDraggable methodsFor: 'drag/drop' stamp: 'bgf 9/13/2006 18:08'!
pick: pointer

	adapter isBeingDragged ifTrue: [
		^false
	].
	^super pick: pointer! !

!QDraggable methodsFor: 'drag/drop' stamp: 'bgf 12/12/2006 13:46'!
snapBack
	"If in the midst of a drag operation, return the luggage to the initial position."
	adapter snapBack! !

QDraggable subclass: #QStorageWrapper
	instanceVariableNames: 'content scaleReduction'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Objects-Manipulation'!

!QStorageWrapper commentStamp: '<historical>' prior: 0!
A wrapper for an object that is "put away" in a container
(such as a toybox, or shelf, or the space-global trash)
and can be taken out again through drag and drop.

Use 
	QStorageWrapper new enclose: anObject within: aVector3 
to initialize a ScaledContainer with its content object;
the object is scaled to fit with the cube,
and is hibernated and reparented into the new wrapper.

QStorageWrapper is an invisible, draggable wrapper around
its content.  When dropped into space, it 'unwraps' the contained object by
restoring its original size and reparenting into the space, and then destroys itself.

Currently requires that the object it is wrapping support the methods
#hibernate and #unhibernate, which are used to "stop" all stepping
and physical processes of the stored object.!

!QStorageWrapper class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QStorageWrapper methodsFor: 'persist' stamp: 'bgf 8/13/2007 12:26'!
constructFrom: pStream

	super constructFrom: pStream.
	
	scaleReduction := pStream readNumber: 'scaleReduction' default: 1.0.
	content := pStream readObject: 'content' default: nil.
	content ifNotNil: [
		(content respondsTo:#hibernate) ifTrue: [content hibernate].
		(content future:(content stepTime +1)) stop. "Defeat TSpace#doStart after reparenting."
		self addChild: content.
	].! !

!QStorageWrapper methodsFor: 'accessing' stamp: 'bgf 12/10/2006 12:56'!
content
	"the wrapped object I contain"
	^ content! !

!QStorageWrapper methodsFor: 'drag' stamp: 'bgf 1/29/2009 13:30'!
dropInSpace: event pointer: pointer

	"Unwrap the stored content and place it back in the space.
 	then discard and destroy this wrapper."

	self unhilite.
	content unhilite.
	self unscaleContent.
	
	QDragAdapter drop: content inSpace: (self root) forEvent:event pointer:pointer.
	QSoundEffects play: #discardObject for: content.

	(content respondsTo:#unhibernate) ifTrue: [content unhibernate].

	"Just in case we came out of the trash..."
	(self root respondsTo: #garbage) ifTrue: [
		self root garbage remove: self ifAbsent: [].
	].
	self parent removeChild:self.
	self destroy.
	^ true			"A drop was carried out."! !

!QStorageWrapper methodsFor: 'drag' stamp: 'jdm 8/21/2007 08:17'!
dropOn: aTarget avatar: avatar selection: aTXSel
	"Drop onto the given drop target."

	| result xform tmp |

	"When a wrapped toy is dropped, it is unwrapped first."

	self unhilite.
	content unhilite.

	"To drop the object, we have to unpack it."
	xform := content localTransform copy.
	self unscaleContent.
	self root addChild: content.
	content localTransform: Matrix4x4 new.
	content translation: self globalTransform translation.
	(content respondsTo:#unhibernate) ifTrue: [content unhibernate].

	"Unlink the reference."
	tmp := content.
	content := nil.
	result := aTarget consumeDrop: (tmp) avatar: avatar selection: aTXSel.

	result ifTrue: [
		"Just in  case we came from the trash... 
		 easier than filtering every removeChild: on the space."
		self root garbage remove: self ifAbsent: [].
		self parent removeChild:self.
		self destroy.

	] ifFalse: [
		content := tmp.
		self scaleContent:scaleReduction.
		content localTransform: xform.
		self addChild: content.
		content hibernate.
		self disable; enable.
	].

	^ result
! !

!QStorageWrapper methodsFor: 'initialize' stamp: 'bgf 8/16/2009 02:12'!
enclose: atframe within: aCube

	"Take the given tframe assembly into this wrapper as its content.
	 Stop all activity in the tframe (stop and hibernate it.)
	 Scale and translate the tframe to fit entirely within the specified cubic volume."

	| bb factor |

	atframe allUpdateShape.
	content := atframe.

	atframe unhilite.
	(atframe respondsTo:#hibernate) ifTrue: [atframe hibernate].
	(atframe future:(content stepTime +1)) stop. "Defeat TSpace#doStart after reparenting."
	self addChild: atframe.
	atframe localTransform: Matrix4x4 new.

	bb := atframe positioningBox.
	factor := 1.0.
	factor := (aCube x / (1.1 * bb extentX max:0.1)) min: factor.
	factor := (aCube y / (1.1 * bb extentY max:0.1)) min: factor.	
	factor := (aCube z / (1.1 * bb extentZ max:0.1)) min: factor.

	self scaleContent: factor.

	"We make toys radically smaller when compressing them."
	atframe allUpdateShape.
	atframe translateIntoContactWith: 0@(aCube y/-2.0)@0 alongNormal: (0@1@0).

	^ self! !

!QStorageWrapper methodsFor: 'drag' stamp: 'bgf 12/10/2006 13:36'!
payload
	"My drop target payload is the wrapped object."
	^ content! !

!QStorageWrapper methodsFor: 'persist' stamp: 'bgf 12/4/2006 13:53'!
persistOn: pStream

	super persistOn: pStream.

	pStream storeNumber: 'scaleReduction' value: scaleReduction default: 1.0.
	pStream storeObject: 'content' value: content default: nil.! !

!QStorageWrapper methodsFor: 'render' stamp: 'bgf 10/21/2006 22:18'!
renderFrame: ogl

	"This is a cheap performance hack - simply ignore wrapped objects through portals.
	 So you can't see what's on shelves when viewing through a portal."

	(ogl inPortal2D | ogl inPortal3D) ifTrue: [
		^ 0
	].
	^ super renderFrame: ogl.! !

!QStorageWrapper methodsFor: 'private' stamp: 'bgf 12/15/2006 17:22'!
scaleContent: factor

	"Only apply the reduction if it makes the object smaller."
	((factor - 1.0)  < Float epsilon) ifTrue: [
		scaleReduction := factor.
		(content respondsTo: #scaleBy:) ifTrue: [
			content scaleBy:  (factor).
		] ifFalse: [
			content scale: (factor).
		].
	].! !

!QStorageWrapper methodsFor: 'private' stamp: 'bgf 8/13/2007 13:00'!
unscaleContent

	(scaleReduction notNil and: [ scaleReduction ~= 1.0 ]) ifTrue: [
		(content respondsTo: #scaleBy:) ifTrue: [
			content scaleBy: (1.0 / scaleReduction).
		] ifFalse: [
			content scale: (1.0 / scaleReduction).
		].
	].! !

TGroup subclass: #QFixedSurface
	instanceVariableNames: 'isNoteCollector noteCollection pendingAttachments clickHandler sweepStart sweepGroup'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Objects-Places'!

!QFixedSurface commentStamp: '<historical>' prior: 0!
A generic class for walls and other large environment frames.

Fixed surfaces accept dropped QFileCards 
and create QDisplayPanels displaying the 2-D content,
and QGates opening onto portals.

Can be set to be a 'Note Collector', in which case clicking on the surface
adds notes without intervening commands.   (Don't do this for large walls,
as it is too annoying; best for smaller surfaces that are visually distinct
and look like note boards.)
!

!QFixedSurface class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QFixedSurface class methodsFor: 'inspector' stamp: 'bgf 5/25/2010 20:27'!
addPropertyRows: propertyDialog forFrame: fsref

	propertyDialog addRow: (
		(QBooleanPropertyRow new)
			titleString: 'Noteboard (Manages Notes and Topics)';
			fetchBlock: [ : f | (f get: #isNoteCollector) = true ] ;
			acceptBlock: [ : b :f | f future 
							setIsNoteCollectorFromEditor: ((b = true) ifTrue: [ true ] ifFalse: [ false])  
						];
			balloonHelp: 'If checked, the surface''s menu includes note and topic options.';
		yourself).

	super addPropertyRows: propertyDialog forFrame: fsref.! !

!QFixedSurface class methodsFor: 'defaults' stamp: 'bgf 11/10/2009 16:59'!
defaultEventMask
	^ EventPointerOver! !

!QFixedSurface class methodsFor: 'undo-redo' stamp: 'bgf 3/14/2010 17:57'!
surface: fsRef  newDisplay: dispref capture: frameCapture cardCapture: cardCap

	"Undoable action representing the creation of a display container in response to a card drop."

	^ QBlockAction
		undoBlock: [ : avatar | dispref future reallyDelete: avatar ]
		redoBlock: [ : avatar |
					fsRef future redoDrop: frameCapture card: cardCap fromUser: avatar ]
		description: 'Drop'
		requiredFrames: { fsRef }! !

!QFixedSurface methodsFor: 'drag-drop' stamp: 'bgf 3/13/2010 08:52'!
acceptDisplayDrop: delivery avatar: avatar selection: aTXSel
	"Can I display this content?"
	| newbie cap cardCap |
	(QDisplaySheet canAcceptDrop: delivery) ifFalse: [
		^ false.
	].

	"If on an upright surface, make a poster if it is poster content,
	 or a display panel for anything else (e.g. apps).
	 On a flat surface, make a display stand."
	(self isHorizontalAtDrop: aTXSel)
		ifTrue: [ newbie := self dropInStand: delivery avatar: avatar selection: aTXSel ]
		ifFalse: [
			(QPoster canAcceptDrop: delivery) 
				ifTrue: [ newbie := self dropInPoster: delivery avatar: avatar selection: aTXSel ] 
				ifFalse: [ newbie := self dropInPanel: delivery avatar: avatar selection: aTXSel ] ].

	cap := avatar captureFrames: { newbie }.
	cardCap := avatar captureFrames: { delivery duplicate }. 
	avatar recordAction: #surface:newDisplay:capture:cardCapture:
		    class: QFixedSurface
		    arguments: { self . newbie . cap . cardCap }.

	^ true! !

!QFixedSurface methodsFor: 'drag-drop' stamp: 'bgf 3/13/2010 08:52'!
acceptLinkDrop: delivery avatar: avatar selection: aTXSel

	| dropPt standoff gate cap cardCap |

	"Can I display this content?"
	((delivery isKindOf: QFileCard) and: [delivery kind = #postcardLink]) ifFalse:[
		^ false.
	].

	dropPt := aTXSel point.
	gate := QGate new.
	gate label: (delivery contentLink shortDisplayString).
	gate isFromTemplate: false.

	(self isVerticalAtDrop: aTXSel) ifTrue: [
		dropPt := aTXSel globalPoint + (Vector3 x: 0 y: 0.1 z: 0).
		gate localTransform: (self orientationForDrop: aTXSel).
	] ifFalse: [
		gate  yaw: (avatar yaw detentBy: 15 atMultiplesOf: 15 snap: true).
		standoff := aTXSel normal * (QGate wallStandoff).
		dropPt := aTXSel frame localToGlobal:((aTXSel point) + standoff).
	].

	"Adjusting to ground won't be correct in all cases, but is necessary in most."
	gate translation: (Vector3 x: dropPt x y: (self root groundHeightAt: dropPt) z: dropPt z).

	self root addChild: gate.
	gate establishLink: delivery contentLink forUser: avatar.

	cap := avatar captureFrames: { gate }.
	cardCap := avatar captureFrames: { delivery duplicate }. 
	avatar recordAction: #surface:newDisplay:capture:cardCapture:
		    class: QFixedSurface
		    arguments: { self . gate . cap . nil }.

	^ true! !

!QFixedSurface methodsFor: 'annotations' stamp: 'bgf 8/18/2009 00:18'!
addNote: aNote
	"Annotation tracking.  I am never an attachment to something else.
	 When a sticker is applied to me, I remember it and the closure of its attachments."
	(self isNoteCollector and: [  aNote isListedTopic ]) ifTrue: [
		(noteCollection add: aNote) ifTrue: [
			aNote root ifNotNilDo: [ : space | 
				(space stickersOn: aNote) do: [ : recursive | self addNote: recursive ]
			]].
	].
	self signal: #notesChanged.! !

!QFixedSurface methodsFor: 'private' stamp: 'bgf 8/2/2009 18:43'!
allNotes
	"Direct access to the full collection of attached note items."
	^ noteCollection items! !

!QFixedSurface methodsFor: 'private' stamp: 'bgf 5/25/2010 19:02'!
ballotsSelectedBy: anAvatar
	"Return all my ballot-candidate attachments that the user has selected."
	^ QPollBarChart sortBallots:   (
		(anAvatar selectedFrames) 
			select: [ : f | f surfaceAttachment = self and: [ QPollBarChart isBallotCandidate: f] ]) ! !

!QFixedSurface methodsFor: 'menu' stamp: 'bgf 4/1/2010 09:03'!
canBeCopiedBy: anAvatar
	^ (super canBeCopiedBy: anAvatar) 
	  and: [ self allNotes allSatisfy: [ : n | n canBeCopiedBy: anAvatar ] ]! !

!QFixedSurface methodsFor: 'private' stamp: 'bgf 8/17/2009 20:56'!
cancelSweep
	sweepGroup ifNotNil: [
		sweepGroup obliterateLater.
		sweepGroup := nil.
	].
	sweepStart := nil.! !

!QFixedSurface methodsFor: 'annotations' stamp: 'bgf 8/15/2009 23:05'!
checkStaleNotes
	"There are many deferred destruction paths when editing spaces;
	 protect from this by discarding stale (unrooted) notes at any opportunity."
	self allNotes clone do: [ : item |	
		(item root = self root and: [ item rootAttachment = self ]) ifFalse: [
			self removeNote: item
		]].
! !

!QFixedSurface methodsFor: 'initialize' stamp: 'bgf 2/8/2008 17:15'!
consistingOf: aTFrame
	"Make this a wrapper around the given frame."

	super consistingOf: aTFrame.
	self objectName: 'Fixed Surface (', (aTFrame objectName), ')'.
! !

!QFixedSurface methodsFor: 'persist' stamp: 'bgf 8/1/2009 14:53'!
constructFrom: pStream
	"Persist the receiver"
	super constructFrom: pStream.
	self class == QFixedSurface ifTrue:[
		self children: (pStream readSequence:'children' default: self children).
	].
	self isNoteCollector: (pStream readBool: 'isNoteCollector' default: self isNoteCollector).
	self isNoteCollector ifTrue: [
		"For recovery of older persistence.  
		 This is the collection of things attached-to me;
	 	 if I'm a standalone (inventory) xml, these won't be rooted yet."
		noteCollection := pStream readObject: 'notes' default: noteCollection.

		"We have to read the noteCollection first because it is the xml subtree
		 where the properties of the collection are first stored;	
		 the pending attachments persist will always be numbered references into the above."
		pendingAttachments := pStream readSequence: 'attachments' default: #()
	].
	! !

!QFixedSurface methodsFor: 'drag-drop' stamp: 'bgf 3/12/2010 16:12'!
consumeDrop: delivery avatar: avatar selection: aTXSel

	"is it a postcard link ? if so, open a gate."
	(self acceptLinkDrop:delivery avatar:avatar selection:aTXSel) ifTrue: [
		^ false "Dont consume the card."
	].

	"Can I display this content?  If so, put it in a display panel."
	(self acceptDisplayDrop: delivery avatar: avatar selection: aTXSel) ifTrue: [
		^ false.  "Not consumed."
	].

	^ false! !

!QFixedSurface methodsFor: 'menu' stamp: 'bgf 10/25/2009 18:01'!
contextMenuMaskForAvatar: avatar
	"Remove the system 'Attach Note' command from the menu,
	 if I'm a note collector, since I put that at the top, along with Topic Card."
	| mask |
	mask := super contextMenuMaskForAvatar: avatar.
	self isNoteCollector ifTrue: [
		mask remove: #attachNote ifAbsent: [].	"This is supplied in the custom part of the menu."
	].
	^ mask! !

!QFixedSurface methodsFor: 'annotations' stamp: 'bgf 8/28/2009 01:31'!
createGroup: event
	"Make an empty worksheet (group-of-notes) at the menu click point."
	event avatar ifNotNil: [
		event avatar addAnnotation: QGroupOfNotes toFrame: self selection: event selection.
	].
! !

!QFixedSurface methodsFor: 'annotations' stamp: 'bgf 8/31/2008 17:14'!
createStickyNote: event

	"Make a sticky note at the event-selection point."
	event avatar ifNotNil: [
		event avatar addAnnotation: QStickyNote toFrame: self selection: event selection
	].! !

!QFixedSurface methodsFor: 'annotations' stamp: 'bgf 8/31/2008 21:23'!
createTopicCard: event

	"Make a topic card (QTopicCard) at the event-selection point."
	event avatar ifNotNil: [
		event avatar addAnnotation: QTopicCard toFrame: self selection: event selection
	].! !

!QFixedSurface methodsFor: 'initialize' stamp: 'bgf 4/16/2009 18:36'!
defaultIsAtomic
	^ (self class ~~ QFixedSurface).
! !

!QFixedSurface methodsFor: 'menu' stamp: 'bgf 11/22/2009 12:29'!
deletionLabel
	^ (self isNoteCollector and: [ noteCollection size > 0 ])
		ifTrue: [ self menuTitle, ' and Contents' ]
		ifFalse: [ self menuTitle ].! !

!QFixedSurface methodsFor: 'initialize' stamp: 'bgf 8/17/2009 22:21'!
destroy	
	self cancelSweep.
	super destroy.
	! !

!QFixedSurface methodsFor: 'drag-drop' stamp: 'bgf 3/12/2010 16:29'!
dropInPanel: delivery avatar: avatar selection: aTXSel
	| panel box size reduce |
	"Accept a drop of 2D content and display it in a display panel;
	 this is used when the drop is against a vertical surface."

	panel := self newPanelFor: delivery.
	panel displayCard: delivery fromUser: avatar.
	panel localTransform: (self orientationForDrop: aTXSel).

	"A bit wrong... but attempt to confine the size of the drop result
	 to my extent.   We don't really have a good way of determining my
	 surface extent on whatever surface was reached... 
	 so just take the max of x and z as the limit.
	 The 0.85 inside the rule are there to make sure the panel is smaller than its wall."
	box := self boundingBox.
	size := (panel class defaultExtent).
	reduce :=  (((box extent x max: (box extent z ))* 0.85 / (size x))  min: (box extent y * 0.85 / (size y)))
				 min: 1.0
				 max: 0.25.
	panel 
		extent: size; 
		scale: reduce;
		isPlacementLocked: false;
		isFromTemplate: false;
		translation: (self translationForPanel: panel droppedAt: aTXSel).	
	self root 
		addChild: panel;
		attach: panel to: self.

	^ panel! !

!QFixedSurface methodsFor: 'drag-drop' stamp: 'bgf 3/12/2010 16:29'!
dropInPoster: delivery avatar: avatar selection: aTXSel

	| panel box size reduce |
	"Accept a drop of 2D content and display it in a poster.
	 This is used when dropping an image against a vertical surface."

	panel := QPoster new.
	panel displayCard: delivery fromUser: avatar.

	"A bit wrong... but attempt to confine the size of the drop result
	 to my extent.   We don't really have a good way of determining my
	 surface extent on whatever surface was reached... 
	 so just take the max of x and z as the limit.
	 The 0.8 inside the rule are there to make sure the panel is smaller than its wall."
	box := self boundingBox.
	size := 5 @ 5.
	reduce :=  (((box extent x max: (box extent z ))* 0.8 / (size x))  min: (box extent y * 0.8 / (size y)))
				 min: 1.0
				 max: 0.25.
	panel extent: (size*reduce).
	panel shapeToFit.
	panel localTransform: (self orientationForDrop: aTXSel).
	panel translation: (self translationForPanel: panel droppedAt: aTXSel).

	self root addChild: panel;
		attach: panel to: self.

	^ panel! !

!QFixedSurface methodsFor: 'drag-drop' stamp: 'bgf 3/12/2010 16:30'!
dropInStand: delivery avatar: avatar selection: aTXSel
	| panel |
	"Accept a drop of 2D content and display it in a display stand."

	panel := QDisplayStand withSimpleFrame.
	panel setupForAvatar: avatar atFloorPoint: (aTXSel globalPoint).
	panel displayCard: delivery fromUser: avatar.
	self root addChild: panel;
	     	 	attach: panel to: self.

	^ panel! !

!QFixedSurface methodsFor: 'annotations' stamp: 'bgf 9/1/2008 19:27'!
editToolClass
	"Only on a reachable flow when I am a 'NoteCollector;
	 the edit action is to open the topics browser."
	^ QTopicsTool! !

!QFixedSurface methodsFor: 'menu' stamp: 'bgf 4/13/2010 14:27'!
fillContextMenu: tmenu forEvent: anEvent
	"This menu is only shown for 'Note Collector' surfaces
	 (see #isNoteCollector:, which sets up the menu trigger."
	| sf |

	anEvent avatar ifNil: [ ^ self ].
	self isNoteCollector ifFalse: [ ^ self ].
	anEvent avatar canModifyForum ifTrue: [
		tmenu add: 'Sticky Note...'
		    target: self
		    action: #createStickyNote:
	    	    arguments: { anEvent}.
	
		tmenu add: 'Topic Card...'
		    target: self
		    action: #createTopicCard:
		    arguments: { anEvent }.

		tmenu addSeparator.
		sf := self ballotsSelectedBy: (anEvent avatar).
		(sf size > 1) ifTrue: [ 
			tmenu add: 'Create a Poll...'
				target: anEvent avatar
				action: #generatePoll:ballots:
				arguments: { anEvent . sf asArray }.	
		]. 
		tmenu add: 'New Group' 
			target:self 
			action: #createGroup:
			arguments: { anEvent }.
	].

	tmenu addSeparator.
	tmenu add: 'Browse Topics...'
		    target: self 	
		    action: #reviewTopics:
		    arguments: { anEvent }.

	tmenu addSeparator.
	(tmenu add: 'Save As Text...'
	   	 	target: self
	    		action: #saveTopicSummary:
	   		arguments: { anEvent avatar })
		enabled: (self notes isEmptyOrNil not).
! !

!QFixedSurface methodsFor: 'menu' stamp: 'bgf 2/9/2010 17:19'!
generateNoteGroup: listOfNotes event: cEvent
	"ListOfNotes is a collection of notes, topics, displays, and similar 2D content objects,
	 all attached to the same fixed-surface.
	 Manufacture  a  GroupOfNotes sheet to surround the items,
	 arrange to have all the list attached to it,  attach 
	 and attach the new group to the original surface."

	| sample fullBox  orientation g normal global invGlobal obox |

	QLogger logInfo: 'Grouping ', (listOfNotes size), ' items on fixed-surface.'.
	sample := listOfNotes anyOne.

	"First figure out the enclosing dimensions needed for the group."
	orientation := sample globalOrientation.
	fullBox := nil.
	listOfNotes do: [ : note |
		| box |
		box := (note positioningBox transformedBy: (note globalTransform)).
		fullBox ifNil: [ fullBox := box ] 
			      ifNotNil: [ fullBox := fullBox unionBox: box ].
	].
	obox := fullBox transformedBy: (orientation inverse).
	self root addChild: (
				g := (QGroupOfNotes new)
						translation: (fullBox center);
						extent: ((obox extent x @ obox extent y) * 1.05);
						orientation: (orientation orientation);
						yourself).
	g allUpdateShape.

	"Attach all the notes to the new group possibly adjusting their offsets."
	normal := g lookAt.
	global := g globalTransform.
	invGlobal := global inverse.
	listOfNotes do: [ : note |
		| pos nm |
		pos := (note globalPosition 
				+ (note globalTransform transformVector:(note positioningBox center))).
		pos := (invGlobal transformPoint: pos).
		pos := global transformPoint: (pos x @ pos y @ 0.001).
		note parent = note root ifFalse: [ 
			pos := note parent globalTransform inverse transformPoint: pos.
			nm := note parent globalOrientation transformVector: normal.
		].
		"Now reinforce the attachment offsets, so they appear in front of me."
		note globalPosition: (pos).
		note allUpdateShape.
		note translateIntoContactWith:  pos alongNormal: (nm ifNil: [ normal ]).
		self root attach: note to: g
	].

	self root attach: g to: self.! !

!QFixedSurface methodsFor: 'menu' stamp: 'das 8/5/2009 13:56'!
getLogs

	"A string comprising a text report of all the topics and notes on this surface."

	| sorted |
	^ String streamContents: [ : stream |
		stream nextPutAll:'<topics>';cr.
		stream nextPutAll:String tab, '<surfaceName>',self objectName,'</sufaceName>';cr.
		stream nextPutAll: String tab, '<time>',DateAndTime now asFileNameString, '</time>';cr.
		"Sort all the notes and topics by date - most recent first - to match the browser"
		sorted := self allNotes 
					sort: [ : n1 : n2 | 
							(n1 lastModifiedTime ifNil: [0]) > (n2 lastModifiedTime ifNil: [0]) ].
		sorted do: [ : note |
			(note respondsTo: #getLog) ifTrue: [
				stream nextPutAll: note getLog;cr.
			].
		].
		stream nextPutAll:'</topics>'.
	].
! !

!QFixedSurface methodsFor: 'initialize' stamp: 'bgf 8/1/2009 21:56'!
initialize
	super initialize.
	isNoteCollector := false.
	noteCollection := QCollection new.
	pendingAttachments := #().
	clickHandler := QClickHandler forFrame: self.! !

!QFixedSurface methodsFor: 'accessing' stamp: 'bgf 7/31/2009 04:06'!
isComponent
	"Fixed surfaces are pretty much always components;
	 they are for organizing receiving and organizing dropped content."
	^ self shouldPersist! !

!QFixedSurface methodsFor: 'drag-drop' stamp: 'bgf 3/25/2007 17:11'!
isDropTarget: payload

	^ (QDisplaySheet canAcceptDrop: payload) ! !

!QFixedSurface methodsFor: 'testing' stamp: 'bgf 6/1/2009 11:43'!
isEditableGroup
	"C.F. TGroup; plain fixed-surface instances are not atomic so they can be excised with ungroup,
	 but derived components are by default atomic."
	^ (self class == QFixedSurface) and: [ self isAtomic not ]! !

!QFixedSurface methodsFor: 'testing' stamp: 'bgf 8/27/2009 19:37'!
isFlatRectangle
	"Subclass-override protocl ... 
	 return true for fixed-surfaces whose extent is x-y rectangular, like noteboards/etc.
	 The default rule just works if I happen to enclose a TRectangle.

	 This test enables edge-clamping when creating panels/posters dropped on me."

	^ (self class = QFixedSurface) 
	    and: [ self children size = 1 ]
	    and: [ self children first isKindOf: TSimpleRectangle ]! !

!QFixedSurface methodsFor: 'testing' stamp: 'bgf 2/9/2010 15:34'!
isFlatSurface
	"Some operations only make sense on designated 2D-planar surfaces..."
	^ (self isFlatRectangle or: [ self isKindOf: QWall ] ) ! !

!QFixedSurface methodsFor: 'drag-drop' stamp: 'bgf 8/6/2009 22:33'!
isHorizontalAtDrop: aTXSel

	"Whether we should treat the drop point as being onto a horizontal surface
	 (where we put up display stands) or a vertical one (where we put sheets).
	 Make display stands only when quite close to flat;
	 for angled surfaces, use posters or panels where possible."

	^ (aTXSel frame globalTransform orientation transformPoint: aTXSel normal) y abs > 0.8.! !

!QFixedSurface methodsFor: 'annotations' stamp: 'bgf 7/20/2008 13:14'!
isNoteCollector
	"If yes, I  actively want annotations, and keep track of them"
	^ isNoteCollector
	! !

!QFixedSurface methodsFor: 'annotations' stamp: 'bgf 10/25/2009 12:55'!
isNoteCollector: aBool

	"A note-collecting surface is 'hot' ... clicks offer note-creation.
 	 This kind of surface also remembers the notes it has attached to it."

	(isNoteCollector ~= aBool) ifTrue: [
		isNoteCollector := aBool.
		isNoteCollector ifTrue: [
			self eventMask: (self eventMask bitOr: (EventKeyboard)).		
							"For escape-canceling sweep-selection."
		] ifFalse: [	
			noteCollection removeAll.
		].
		self cancelSweep.
		self signal: #notesChanged.	"To clear any clients."
	].
! !

!QFixedSurface methodsFor: 'drag-drop' stamp: 'jdm 10/8/2007 17:10'!
isVerticalAtDrop: aTXSel

	"Whether we should treat the drop point as being onto a very-nearly vertical surface."

	^ (aTXSel frame globalTransform orientation transformPoint: aTXSel normal) y abs < 0.1.! !

!QFixedSurface methodsFor: 'events' stamp: 'bgf 8/20/2009 12:59'!
keyStroke: event
	"Escape key cancels sweep-select."
	event keyValue = 27 "escape" ifTrue: [
		sweepGroup ifNotNil: [
			QLogger logInfo: 'Drag-select canceled.'.
			sweepGroup resizeCanceled.
			sweepGroup obliterateLater.
		].
		sweepGroup := nil.
		sweepStart := nil.
	].! !

!QFixedSurface methodsFor: 'private' stamp: 'bgf 8/17/2009 22:20'!
localNormal: selection
	"The normal from the TSelection, in my transform "
	^ (self globalOrientation inverse 
			transformVector: (
				selection frame globalOrientation transformVector: (selection normal)))
		safelyNormalized
	! !

!QFixedSurface methodsFor: 'drag-drop' stamp: 'bgf 8/18/2009 00:35'!
newPanelFor: delivery
	"Instantiate a QDisplayPanel, or some substitute with compatible protocols,
	 to receive dropped content."
	^ QDisplayPanel withSimpleFrame! !

!QFixedSurface methodsFor: 'annotations' stamp: 'bgf 8/1/2009 14:55'!
notes
	"All the annotations that have me in their attachment rooting;
	 nil if I am not a note collector"
	self isNoteCollector ifFalse: [ ^ nil ].
	
	^ noteCollection ifNotNilDo: [ : nc | nc items ] ifNil: [ #() ].
	! !

!QFixedSurface methodsFor: 'initialize' stamp: 'bgf 8/17/2009 22:21'!
obliterateLater
	self cancelSweep.
	super obliterateLater.! !

!QFixedSurface methodsFor: 'drag-drop' stamp: 'jdm 9/27/2007 11:03'!
orientationForDrop: aSelection

	"A Matrix 4x4 oriented to face outwards from the drop point indicated in the selection."

	| normal |
	normal := aSelection frame globalTransform orientation transformVector: aSelection normal.
	(normal x = 0 and:[normal z = 0]) ifTrue:[
		^ Matrix4x4 new at: normal  up: (1@0@0) 
	] ifFalse:[
		^ Matrix4x4 new at: normal up: (0@1@0) 
	].
! !

!QFixedSurface methodsFor: 'persist' stamp: 'bgf 8/15/2009 15:21'!
persistOn: pStream
	"Persist the receiver"
	super persistOn: pStream.

	self class == QFixedSurface ifTrue:[
		"XXXX: allow subclasses to have private children"
		pStream storeSequence: 'children' value: self children default: #().
	].
	pStream storeBool: 'isNoteCollector' value: self isNoteCollector default: false.
	self isNoteCollector ifTrue: [
		"For backwards compatibility with older versions.
		 But since this will often be where the note-bodies are persisted, 
	 	 it will always be the collection we need to read first."
		pStream storeObject: 'notes' 
				value: (noteCollection)
				default: (QCollection empty).	

		"If we're coming out of a rooted context, store the notes-relations as they stand here;	
		 otherwise transcribe what we remember (the transcribe case applies to inventory-import,
	 	 where the harness unpacks and then injects island-bits.)"
		self root ifNotNilDo: [ : space | 
			space attachments 
				persistNotes: (self allNotes) for: self on: pStream as: 'attachments'.
		] ifNil: [
			pStream storeSequence: 'attachments' value: pendingAttachments default: #()
		].			
	].

! !

!QFixedSurface methodsFor: 'events' stamp: 'bgf 8/17/2009 22:01'!
pointerDown: event

	self cancelSweep.
	(event selection notNil and: [ event selection point notNil ]) ifTrue: [
		sweepStart := self globalToLocal: (event selection globalPoint).
	].! !

!QFixedSurface methodsFor: 'events' stamp: 'bgf 8/17/2009 20:57'!
pointerLeave: event
	self cancelSweep.
! !

!QFixedSurface methodsFor: 'events' stamp: 'bgf 2/10/2010 11:39'!
pointerMove: event
	| ext selPoint nm |
	"Check for possible marquis selection of collected notes."
	self isNoteCollector ifFalse: [^ self ].

	"I simply don't trust this gunk to work against general geometry..."
	(self isFlatRectangle or: [ self isKindOf: QWall ] ) ifFalse: [ ^ self ].

	event selection ifNil: [ ^ self ].
	event selection frame ifNil: [ ^ self ].
	event selection point ifNil: [ ^ self ].

	"Ok, we're gonna do a marquis rubber-band rectangle, 
	 out-down-right from the start."
	selPoint := self globalToLocal: (event selection globalPoint).
	sweepGroup ifNil: [
		"First move, so start the marquis by making a GroupOfNotes..."
		sweepStart ifNotNil: [
			ext := (selPoint - sweepStart).
			sweepGroup := Q2DNoteSweeper new.
			sweepGroup extent: (ext x abs)@(ext y abs)..
			self root addChild: sweepGroup.

			nm := (self localNormal: (event selection)).
			sweepGroup 	attachToFrame: (self)
						atPoint: (sweepStart + ((ext x @ (ext y) @ (0.001 * (nm z sign))) *0.5))
						normal: (nm).
			sweepGroup resizeBegin: event forInitialPoint: (self localToGlobal: sweepStart).	
		].
		sweepStart := selPoint.

	] ifNotNil: [
		"Once we've made the  group, its resizer does the marquis work for us.
		 We just need to get the size delta  into its orientation."
		sweepGroup drag: (sweepGroup globalTransform inverse 
							transformPoint: (event selection globalPoint))
					corner: #southeast 
					delta: (sweepGroup globalOrientation inverse transformVector: (
							self globalOrientation transformVector: (selPoint - sweepStart))).		
		sweepStart := selPoint.
	].! !

!QFixedSurface methodsFor: 'events' stamp: 'bgf 8/17/2009 20:57'!
pointerOver: event
	self cancelSweep.

! !

!QFixedSurface methodsFor: 'events' stamp: 'bgf 2/9/2010 15:05'!
pointerUp: event
	"If concluding a marquis sweep...
	 if the sweep captured anything, reify it as a worksheet,
	 otherwise discard it."
	sweepGroup ifNotNil: [
		self pointerMove: event.
		sweepGroup resizeComplete: event.
		sweepGroup obliterateLater.
	].
	sweepGroup := nil.
	sweepStart := nil.
! !

!QFixedSurface methodsFor: 'menu' stamp: 'bgf 1/29/2010 11:42'!
reallyDelete: event
	self cancelSweep.
	self isNoteCollector ifTrue: [
		self allNotes do: [ : n | n ifNotNil: [ n reallyDelete: event ]].
		noteCollection ifNotNil: [ noteCollection removeAll ].
	].
	super reallyDelete: event! !

!QFixedSurface methodsFor: 'annotations' stamp: 'bgf 8/1/2009 16:09'!
recoverPendingAttachments

	"If we're holding attachments waiting to be re-instated into a space, now is the time."
	self root ifNotNil: [ 
		self root attachments recoverNotesFrom: pendingAttachments for: self.
		pendingAttachments := #().
	].! !

!QFixedSurface methodsFor: 'drag-drop' stamp: 'bgf 3/12/2010 18:23'!
redoDrop: frameCapture card: cardCap fromUser: avatar
	"Undo-redo support.   Repeat the creation of a drop-display."
	| disp cd |
	disp := (frameCapture restoreForAvatar: avatar) first.
	cardCap ifNotNil: [
		cd := (cardCap restoreForAvatar: avatar) first.
		disp displayCard: cd fromUser: avatar.
		cd obliterateLater.
	].
	avatar selectFrames: { disp }.

	! !

!QFixedSurface methodsFor: 'annotations' stamp: 'bgf 8/15/2009 23:28'!
removeNote: aNote
	"Annotation tracking."
	self signal: #notesChanged.
	(self isNoteCollector) ifTrue: [
		(noteCollection remove: aNote) ifTrue: [
			aNote root ifNotNilDo: [ : space | 
				(space stickersOn: aNote) do: [ : recursive | self removeNote: recursive ]
			]]].


! !

!QFixedSurface methodsFor: 'annotations' stamp: 'bgf 8/15/2009 14:02'!
reviewTopics: anEvent
	"Open the topic-review browser.  This is only reachable when I am a NoteCollector"

	self checkStaleNotes.
	"Using the text-edit job as a launch relay."
	QTextEditJob
		edit: (self objectName asString)
		title: ('Topic List' )
		forUser: (anEvent avatar)
		recipient: self 
		selector: nil.! !

!QFixedSurface methodsFor: 'accessing' stamp: 'bgf 8/1/2009 16:49'!
root: aSpace

	"After recovery from persistence, recover any internally-persisted notes I carry.
	 (The mechanism will ignore attachments that are already recovered normally,
	 since it only re-instates attachments that are still unrooted at that time.
	 So recovery applies when importing a subscene, but not when loading a forum.)"
	self future recoverPendingAttachments.
	super root: aSpace.! !

!QFixedSurface methodsFor: 'menu' stamp: 'bgf 10/4/2008 01:30'!
saveTopicSummary: anAvatar
	"We do the save-summary by generating a text blob,
 	 and using save-local-copy of a temporary TextDocument"
	| txt doc |

	txt := self topicSummaryText.

	doc :=	(QTextDocument forString: txt).
	doc shortName: 'topics.txt'.

	QSaveDocumentJob save: doc 
					     forUser: anAvatar.! !

!QFixedSurface methodsFor: 'annotations' stamp: 'bgf 8/1/2009 11:59'!
setIsNoteCollectorFromEditor: bNoteCollector
	"An interactive change to being a note collector...
	 if necessary, find all existing notes that are attached to me,
	 and add them to my cached noteCollectioon."

	self isNoteCollector = bNoteCollector ifTrue: [ ^ self ].
	self root ifNil: [ ^ self ].
	bNoteCollector ifTrue: [
		| topics | 
		topics := self root stickersOn: self.
		topics do: [ : topic | topic isListedTopic ifTrue: [ noteCollection add: topic ]].
	] ifFalse: [
		noteCollection removeAll.
	].
	self isNoteCollector: bNoteCollector.


! !

!QFixedSurface methodsFor: 'annotations' stamp: 'bgf 8/1/2009 01:26'!
topicRecords
	"Tool support - gather the QTopicRecords for all my notes&topics
	 as an array for the topic editor."
	^ (self notes collect: [ : note | note asTopicRecord]) asArray
! !

!QFixedSurface methodsFor: 'menu' stamp: 'bgf 2/21/2010 09:00'!
topicSummaryText
	"A string comprising a text report of all the topics and notes on this surface."
	| sorted |
	^ String streamContents: [ : stream |
		stream nextPutAll: self objectName; cr.
		stream nextPutAll: 'Notes and Topics ('.
		DateAndTime now printYMDOn: stream.
		stream space.
		DateAndTime now printHMSOn: stream.
		stream nextPut: $).
		stream cr; cr.
		"Sort by date - most recent first - to match the browser"
		sorted := self allNotes 
					sort: [ : n1 : n2 | 
							(n1 lastModifiedTime ifNil: [0]) > (n2 lastModifiedTime ifNil: [0]) ].
		sorted do: [ : note |
			(note respondsTo: #writeTranscriptOn:) ifTrue: [
				note writeTranscriptOn: stream.
				stream cr.
			]].
	].
! !

!QFixedSurface methodsFor: 'drag-drop' stamp: 'bgf 8/3/2009 11:03'!
translationForPanel: aPanel droppedAt: aTSelection

	"The position for a DisplayPanel created from a drop at the given selection.
	 By default, simply centers the panel at this point."

	| standoff target result |
	standoff := aTSelection normal * 0.1.
	target := aTSelection frame.
	target ifNil: [ target := self ].
	result := target localToGlobal: (aTSelection point + standoff).

	self isFlatRectangle ifTrue: [
		| lim |
		result := self globalToLocal: (result).
		
		"Keep it in view, giving priority to keeping it below the top (so check bottom first)."
		lim :=(self extent y * -0.47).
		(result y - (aPanel height / 2)) < (lim ) ifTrue: [
			result := result x @ (lim + (aPanel height / 2.0)) @ result z.
		].
		lim := (self extent y * 0.47).
		((result y + (aPanel height / 2.0)) > (lim )) ifTrue: [
			result := result x @ (lim - (aPanel height /2.0)) @ result z.
		].

		lim := self extent x * -0.47.
		((result x - (aPanel extent x / 2.0)) < lim) ifTrue: [
			result := (lim + (aPanel extent x  /2.0)) @ result y @ result z.
		].

		lim := self extent x * 0.47.
		((result x + (aPanel extent x / 2.0)) > lim) ifTrue: [
			result := (lim - (aPanel extent x /2.0)) @ result y @ result z.
		].
		result := self localToGlobal: result.
	].

	^ result! !

TGroup subclass: #QHeadAvatar
	instanceVariableNames: 'ribbon lastRibbonOrigin'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Objects-User'!

!QHeadAvatar commentStamp: '<historical>' prior: 0!
Draft costume class for avatar costumes that consist entirely of a head
that floats in space and is bound to the pointer orientation.

Colorizes the TGeometries in its costume by modifying their default (first) material.

Does not accept face pictures or badges.

Currently the costume is a hard-coded (Meredith-style) white head,
so colorization produces the usual range of avatar colors.!

!QHeadAvatar class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QHeadAvatar methodsFor: 'accessing' stamp: 'bgf 6/26/2007 20:05'!
avatar
	"The avatar replica that owns this costume; we have to get it from the TAvatarProxy."
	^ self parent ifNotNilDo: [:p | p get:#avatar]! !

!QHeadAvatar methodsFor: 'initialize' stamp: 'bgf 7/28/2009 22:18'!
balloonBottom
	"This is where the string attaches."
	^ self head localTransform transformVector: (-0.015@-1.16@0.1)! !

!QHeadAvatar methodsFor: 'accessing' stamp: 'bgf 7/3/2007 18:19'!
colorize: laserColor

	"Colorize the geometry parts precisely;
	 the laser color 'raw' is not appropriate so we desaturate and darken it a bit."

	| color |
	color := laserColor asColor.
	color := ((Color gray: 0.05) + (color*0.75)) asColor alpha: 1.

	(self find: [:fc | fc class includesBehavior: TGeometry]) do:[ :part|
		| value opacity | 
		value := 0.95. opacity := 1.0.
		part material 
				diffuseColor: ((color*value)*0.8 alpha:1);
				ambientColor: ((color*value*0.3) alpha: 1);
				emissiveColor: ((color*value*0.65) alpha: 1);
		 		opacity: opacity.
		part checkAlpha.
	].
! !

!QHeadAvatar methodsFor: 'initialize' stamp: 'bgf 7/19/2009 21:00'!
configureAvatar: aReplica

	"These heads are usually large, and can be annoying in the way,
	 so we increase the opaque threshold - the head starts to fade 
	 from further back than the default."

	aReplica opaqueThreshold: 36.0.		"Actually 6.0 units distance."
	"Take control of our coordinate basis."
	aReplica footToEyes: 5.4.
	aReplica seatToEyes: 2.5.

	"The laser emerges from the absent shoulder."
	aReplica laser translationX:0.7 y: -0.4 z: -0.4.
	"The shadow matches the basic shape."
	aReplica shadow extent: 2.4@2.4.
	aReplica shadow translationX: 0 y: -4.98 z: 0.1.

	(self future: 100) step.! !

!QHeadAvatar methodsFor: 'persist' stamp: 'bgf 6/26/2007 21:17'!
constructFrom:pStream
	"For preferences"

	super constructFrom: pStream.
! !

!QHeadAvatar methodsFor: 'accessing' stamp: 'bgf 6/12/2009 16:10'!
head
	"If I have any frame children, the first (my costume body) is my head."
	^ ((self frameChildren ifNil: [ ^ self ]) ifEmpty: [ ^ self ])
		first
! !

!QHeadAvatar methodsFor: 'initialize' stamp: 'bgf 7/28/2009 22:16'!
initialize
	"--- set up the avatar model ---"

	| mesh wrap costumeModel |
	super initialize.
	costumeModel := 'Objects/Qwaq/Avatars/Meredith/head-white.c3x'.
	mesh := (TSceneReader new loadFileNamed: costumeModel).
	mesh addRotationAroundY:180.	"The mesh faces towards Z"
	mesh translationX: 0 y:-0.3 z:0.	"Shift down just a little"
	wrap := TGroup new.
	wrap addChild: mesh.
	self addChild:wrap.
	self head translation: (0@0@0.2).   "So we don't see our nose in first-person."

	ribbon := QRibbon new.
	ribbon wind: (QForce new).
	ribbon wind force: (0.01@-2@1.0).
	ribbon gravity force: (0.001@-5@0.001).
	ribbon extent: 2@8.
	ribbon linkLength: 0.14; linkWidth: 0.033.
	ribbon translation: (self balloonBottom).
	ribbon windOn: true.
	ribbon stepTime: 60.
	self addChild: ribbon.

	ribbon material ambientColor: (VectorColor r: 0.7 g: 0.7 b: 0.55 a: 0.95);
				     diffuseColor: (VectorColor r: 0.85 g: 0.8 b: 0.65 a: 0.95).
	! !

!QHeadAvatar methodsFor: 'persist' stamp: 'bgf 6/26/2007 21:17'!
persistOn:pStream
	"For preferences"

	super persistOn: pStream.
	! !

!QHeadAvatar methodsFor: 'accessing' stamp: 'bgf 6/26/2007 20:05'!
selectCostume: index
	"Backstop, for spaces that assume the avatar is a multicostume;
	 see this method in TMultiCostume."! !

!QHeadAvatar methodsFor: 'private' stamp: 'bgf 6/21/2009 20:37'!
step
	"Wobble a little to show I'm alive."
	| bb dl gt perturb |
	self head translation: 
				((self head translation * 5)
					+ (((-1 to:1)atRandom * 0.05) @ ((-1 to:1)atRandom * 0.05) @ 0)) 
					/ 6.0.
	bb := self balloonBottom.
	gt := self globalTransform.
	perturb := (-0.2 to: 0.2) atRandom.

	lastRibbonOrigin ifNil: [ 
		lastRibbonOrigin := gt transformPoint: bb.
		ribbon wind force: (perturb@-2.0@(perturb*0.5)).
		ribbon translation: bb.
	] ifNotNil: [
		"The ribbon simply gets blown in the opposite direction to my motion;
	 	  so we steer its wind force and set the magnitude.. exagerating short motions
	 	  and dampening long ones (so big gotos don't make it bezerk.)"
		| newOrigin delta  |
		newOrigin := gt transformPoint: bb.
		delta := (lastRibbonOrigin - newOrigin).
		dl := delta length.
		dl > 0.02 ifTrue: [
			| multiplier |
			multiplier :=  (2.0 / dl) atLeast: 0.75 atMost: 12.
			"Below... the wind force vector is in the wind's frame of reference, so localize it."
			ribbon wind force: 
						(gt inverseByTranspose orientation transformVector: (delta * multiplier))
						+ (ribbon wind force) 
						/ 2.0.
			ribbon restart.
		] ifFalse: [
			"Stationary, a light wind down settles the ribbon."
			ribbon wind force: ((perturb@-2.0@(perturb*-0.5)) + (ribbon wind force)) / 2.0.
		].
		ribbon translation: bb.
		lastRibbonOrigin := newOrigin.
	].

	(self future: 60) step.! !

TGroup subclass: #QMaterialHolder
	instanceVariableNames: 'material defaultMaterial'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Objects-Resources'!

!QMaterialHolder commentStamp: '<historical>' prior: 0!
This is nothing but a frame with materials.  It allows the materials to be edited by the materials editor.!

!QMaterialHolder class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QMaterialHolder class methodsFor: 'as yet unclassified' stamp: 'jjl 1/13/2009 15:10'!
parent: parent material: mat

	^(QMaterialHolder new)
		defaultMaterial: mat;
		material: mat;
		parent: parent;
		yourself.! !

!QMaterialHolder methodsFor: 'accessing' stamp: 'jjl 1/13/2009 08:49'!
defaultMaterial
	"Answer the value of defaultMaterial"

	^ defaultMaterial! !

!QMaterialHolder methodsFor: 'accessing' stamp: 'jjl 1/13/2009 08:49'!
defaultMaterial: anObject
	"Set the value of defaultMaterial"

	defaultMaterial := anObject! !

!QMaterialHolder methodsFor: 'accessing' stamp: 'bgf 2/19/2009 15:58'!
isComponent
	"So the material editor opens directly when requested."
	^ true! !

!QMaterialHolder methodsFor: 'accessing' stamp: 'jjl 1/13/2009 08:49'!
material
	"Answer the value of material"

	^ material! !

!QMaterialHolder methodsFor: 'accessing' stamp: 'jjl 1/13/2009 08:49'!
material: anObject
	"Set the value of material"

	material := anObject! !

TGroup subclass: #QPlainAvatar
	instanceVariableNames: 'head face torso backface faceMaterial backfaceMaterial faceTexture videoFaceTexture badgeArea legs videoFace'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Objects-User'!

!QPlainAvatar commentStamp: '<historical>' prior: 0!
A trivial avatar costume that uses a picture (texture) for the head,
and a couple of boxes for the body.

Supports some API's needed by QAvatarUser/Replica,
since those are used as the defaults in the Qwaq UI harness.
!

!QPlainAvatar class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QPlainAvatar class methodsFor: 'defaults' stamp: 'bgf 9/17/2007 02:02'!
chooserDetailsPanelClass

	^ QPlainAvatarDetails! !

!QPlainAvatar class methodsFor: 'defaults' stamp: 'bgf 11/10/2009 16:22'!
defaultEventMask
	^ EventPointerOver 		"For badge drop-receipt"! !

!QPlainAvatar methodsFor: 'accessing' stamp: 'bgf 4/4/2007 01:13'!
avatar
	"The avatar replica that owns this costume; we have to get it from the TAvatarProxy."
	^ self parent ifNotNilDo: [:p | p get:#avatar]! !

!QPlainAvatar methodsFor: 'accessing' stamp: 'bgf 3/29/2007 15:33'!
badge
	^ badgeArea material texture! !

!QPlainAvatar methodsFor: 'accessing' stamp: 'das 12/13/2007 13:53'!
badge: aTexture

	"The badge is the image on the front of my torso"

	aTexture ifNil: [
		"Area remains mapped (but invisible via local opacity) so it is still pickable,
	  	 which allows it to be a drop target for in-space badge changes."
		badgeArea material opacity: Float epsilon.
		badgeArea material cullFace: false.
		badgeArea extent: (self badgeExtent).
	] ifNotNil: [
		badgeArea material opacity: 1.0.

		"Crazy... but to get the aspect/size right, we have to adjust aspect/extent
	 	it in just this order, both before and after changing the texture."
		badgeArea aspect: aTexture aspect.
		badgeArea extent: (self badgeExtent).
		badgeArea material texture: aTexture.
		badgeArea material cullFace: false.
		badgeArea aspect: aTexture aspect.
		badgeArea extent: (self badgeExtent).
		badgeArea normal: nil.	"So others can drop badges."
	].
	badgeArea checkAlpha.
	badgeArea flip:true.
! !

!QPlainAvatar methodsFor: 'initialize' stamp: 'bgf 3/22/2009 00:36'!
badgeExtent
	"For coordination with init-time hardcoded sizing..
	 just a tiny bit above the inner extent of the torso slab."
	^ 1.48@1.76! !

!QPlainAvatar methodsFor: 'accessing' stamp: 'bgf 5/13/2007 03:47'!
colorize: laserColor
	
	"Colorize only the body parts we intend (not the pictures.)"

	| color |
	color := laserColor asColor.
	color := ((Color gray: 0.05) + (color*0.45)) asColor alpha: 1.

	(self find: [:fc | fc objectName beginsWith:'body']) do: [:part|
		| value opacity | 
		part objectName caseOf: {
			['bodyHead'] -> [value := 1.0. opacity := 1.0] .
			['bodyLower'] -> [value := 0.7. opacity := 0.85] .
		} otherwise: [ value := 0.95. opacity := 1.0].

		part material 
				diffuseColor: ((color*value) alpha:1);
				ambientColor: ((color*value*0.9) alpha: 1);
				emissiveColor: ((color*value*0.5) alpha: 1);
		 		opacity: opacity.
		part checkAlpha.
	].
! !

!QPlainAvatar methodsFor: 'initialize' stamp: 'bgf 6/27/2009 13:27'!
configureAvatar: aReplica

	"Take control of our coordinate basis."
	aReplica footToEyes: 5.
	aReplica seatToEyes: 3.06.
	"The laser emerges from the shoulder."
	aReplica laser translationX: 0.8 y: -1.1 z: -0.4.
	aReplica shadow translationX: 0 y: -5 z: 0.3.
	aReplica shadow extent: (3.2@1.6).

	aReplica opaqueThreshold: 36.! !

!QPlainAvatar methodsFor: 'persist' stamp: 'bgf 3/29/2007 17:36'!
constructFrom:pStream
	"For preferences"

	super constructFrom: pStream.
	self badge: (pStream readObject: 'badge' default:nil).

	"The face is handled separately (predates the avatar file, and carries over
	 if we change costumes)."! !

!QPlainAvatar methodsFor: 'drag-drop' stamp: 'bgf 4/29/2009 14:36'!
consumeDrop: delivery avatar: avatar selection: selection

	"Dropping an image onto my badge area puts on that badge."
	(self isDropTarget: delivery) ifFalse:[ ^ false].
	(torso isOrContains: selection frame) ifFalse: [ ^ false ].

	self setBadge: (delivery fileTexture).
	(self parent get:#avatar) signal: #floatMessage with: 'You have been given a new badge.'.
	^ false! !

!QPlainAvatar methodsFor: 'initialize' stamp: 'bgf 5/31/2008 17:39'!
destroy

	"Clean up bits that may not be in the scene graph right now."
	(videoFace notNil and: [ videoFace parent isNil ] ) ifTrue: [
		videoFace destroy
	]. 
	super destroy.
! !

!QPlainAvatar methodsFor: 'events' stamp: 'bgf 5/29/2008 18:15'!
faceClicked: event
	"Handler for pointer clicked on a video rectangle...
	 gives the client a chance to re-enable video,
	 so the disabled rectangle can simply say 'click to enable'"

	"Signal the clicker's avatar, not mine, since that's the guy who wants the change."

	videoFaceTexture ifNotNil: [
		event avatar signal: #videoRectangleClicked.
	].! !

!QPlainAvatar methodsFor: 'accessing' stamp: 'bgf 10/29/2006 22:21'!
facePicture
	"A copy of the TTexture representing the avatar face."
	face ifNotNil: [
	^ face material texture
	].
	^ nil
! !

!QPlainAvatar methodsFor: 'accessing' stamp: 'bgf 3/9/2006 23:17'!
head
	^ head! !

!QPlainAvatar methodsFor: 'initialize' stamp: 'bgf 11/10/2009 16:35'!
initialize
	"--- set up the avatar model ---"

	| back |

	super initialize.

	"Construct body"
	torso := QLozenge new.
	torso extent: 1.45@1.8; radius: 0.16.
	torso translationX: 0.0 y:-1.78 z:0.21.  "body has to be behind origin for pick in direct view"
	torso material: self newMaterial.
	torso visible: true.
	torso objectName:#body.		"Signals #colorize: to paint this part."
	self addChild: torso. 

	badgeArea := TRectangle new.
	badgeArea extent: (self badgeExtent);
			    translationX: 0 y: -1.82 z: (0.205 - (torso radius)).
	badgeArea material: QMaterials fullWhiteAlphaMaterial.
	badgeArea material opacity: Float epsilon.	"So still visible/pickable, but unseen."
	badgeArea rotationAroundY:180.
	badgeArea material cullFace: false.
	badgeArea aspectFixed:true.
	self addChild: badgeArea.

	legs := QLozenge new.
	legs extent: 1.1@2.1; radius: 0.1.
	legs translationX: 0.0 y:-3.9 z:0.21.  "body has to be behind origin for pick in direct view"
	legs material: self newMaterial.
	legs objectName:#bodyLower.		
	legs visible: true.
	self addChild: legs. 

	"Construct head, which encloses forward facing picture
	 (invisible to avatar user) and smaller translucent back,
	 so the user can see head orientation but isn't obstructed by the picture."

	head := TGroup new.
	head translationX: 0.0 y: -0.08 z: 0.11.

	back := QLozenge new.
	back extent: 1.136@1.165;  radius: 0.06;
		translationX: 0.0 y: 0 z:0.09;  "Enough for the face to show."
		objectName: #bodyHead;
		material: self newMaterial.
	head addChild: back.

	self addChild: head.
	self solid: false.
	badgeArea normal: nil.	"So we can set our own badge from behind."

	faceMaterial := QMaterials fullWhiteAlphaMaterial.
	face := TSimpleRectangle new.
	face material: faceMaterial.
	face extent: 1.125@1.15.
	face translationX:0 y:0 z:0.0.
	face rotationAroundY: 180.
	face solid: false.
	head addChild: face.

	backfaceMaterial := QMaterials flatWhiteAlphaMaterial.
	backfaceMaterial opacity: 0.13.
	backfaceMaterial ambientColor: torso material ambientColor.
	backface := TSimpleRectangle new.
	backface material: backfaceMaterial.
	backface extent: 1.13@1.15.
	backface translationX:0 y: 0 z:0.225.
	backface solid: false.
	backface objectName: #head.
	head addChild: backface
! !

!QPlainAvatar methodsFor: 'drag-drop' stamp: 'bgf 3/29/2007 16:13'!
isDropTarget: payload

	"I accept dropped images - and wear one at a time, as my badge (torso front.)"
	^ (payload isKindOf: QFileCard) and: [payload kind = #image].! !

!QPlainAvatar methodsFor: 'initialize' stamp: 'bgf 3/19/2007 09:15'!
newMaterial
	"Make a material for a body part."
	^ QMaterials abstractBodyMaterial
		opacity: 0.8;
		yourself! !

!QPlainAvatar methodsFor: 'persist' stamp: 'bgf 3/29/2007 16:41'!
persistOn:pStream
	"For preferences"

	super persistOn: pStream.
	pStream storeObject: 'badge' value: self badge default: nil.

	"The face is handled separately (predates the avatar file)."! !

!QPlainAvatar methodsFor: 'accessing' stamp: 'bgf 6/5/2009 15:52'!
posture: aPosture

	"For demonstration / testing of posture control.
 	 Use the legs portion of the costume to indicate 
	 	sitting - just out in front and shortened
		hover - just slightly shorter than usual to emphasize lift-off/landing visually"

	aPosture caseOf: {
		[#Standing] -> [ 
			legs pitchYawRoll: (Vector3 x:0 y:0 z:0);
			       extent: 1.2@2.0; 
				 translationX: 0.0 y:-3.9 z: 0.22.
			].
		[#Sitting] -> [ 
			legs pitchYawRoll: (Vector3 x:90 y:0 z:0);
			       extent: 1.2@1.6; 
				 translationX: 0.0 y:(-2.65 - 0.075)  z: (0.22 - 1.0).
			].
		[#Hovering] -> [ 
			legs pitchYawRoll: (Vector3 x:0 y:0 z:0);
			       extent: 1.2@1.8; 
				 translationX: 0.0 y:-3.8 z: 0.22.
			].
	} otherwise: [
		QLogger logWarning: 'Plain avatar - posture unknown: ', aPosture
	].! !

!QPlainAvatar methodsFor: 'accessing' stamp: 'bgf 8/31/2006 02:20'!
selectCostume: index
	"Backstop, for spaces that assume the avatar is a multicostume;
	 see this method in TMultiCostume."! !

!QPlainAvatar methodsFor: 'accessing' stamp: 'bgf 4/4/2007 01:12'!
setBadge: aTexture

	"The badge is the image on the front of my torso.."
	self avatar signal: #badgeDropped with: aTexture! !

!QPlainAvatar methodsFor: 'initialize' stamp: 'bgf 5/31/2008 17:06'!
setFace: texture
	"the supplied texture must be a new, unparented TTexture."

	faceTexture := texture.
	faceMaterial texture: faceTexture.
	backfaceMaterial texture: faceTexture.
	backfaceMaterial ambientColor: (torso material ambientColor).! !

!QPlainAvatar methodsFor: 'initialize' stamp: 'bgf 6/1/2008 09:55'!
setVideoFace: texture
	"The supplied texture must be a new, unparented QVideoTexture (or nil).  
	 The 'video face' overlays the 'non-video face' (i.e. if a video face texture is present, it is seen in preference to the non-video face texture)."

	videoFaceTexture == texture ifTrue: [^self "nothing has changed"].
	videoFaceTexture ifNotNil: [
		videoFace material texture: nil.
		videoFaceTexture destroy].

	videoFaceTexture := texture.

	"The video face sits in front of the default face if it is in use;
	 we only make it when first requested."
	videoFaceTexture ifNotNil: [
		| ratio leftover |
		videoFace ifNil: [
			videoFace := QCullingRectangle new.
			videoFace material: faceMaterial copy;
				 extent: (face extent * 1.03);
				 translationX:0 y:0 z:-0.0125;	"In front of plain face"
				 rotationAroundY: 180;		
				 solid: false.
			videoFace material opacity: 0.9995. "I.e. not quite opaque, so always in alpha pass"

			videoFace eventMask: (EventPointerDown).	"For restart clicking"
			self startScript: #faceClicked: when: { videoFace . #pointerUp }.
		].
		videoFace material texture: videoFaceTexture.

		"Adjust the texture-matrix, since we're displaying a non-square video on square geometry."
		"Assume 4:3 video ratio (i.e. 640@480 or 320@240)."
		ratio := 4.0@3.0.
		leftover := (ratio - (ratio x min: ratio y)) / ratio.
		videoFace uvScale: 1 asPoint - leftover.
		"XXXXX: Working the math tells me that the denominator should be 2, but empirical
		tests show that the result is centered if we divide by 3.  If someone can explain this
		to Josh, he would appreciate it."
		videoFace uvOffset: leftover / 3. 

		head addChild: videoFace.
		videoFace checkAlpha.
	] ifNil: [
		videoFace ifNotNil: [ head removeChild: videoFace ].
	].
! !

TGroup subclass: #QPortableObject
	instanceVariableNames: 'motionGrip rotationGrip blockPlaceLoop'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Objects-Manipulation'!

!QPortableObject commentStamp: '<historical>' prior: 0!
A framework base class for user-manipulated "things."

A manipulation wrapper-group that provides drag-based movement
and rotation to whatever is installed as children inside it,
by configuring grips (parts of the mesh inside the object)
using the property editor.

Intuitively the class is for equipment and toys,
but not walls or other fixtures of the environment.

Uses drag/drop-safe old-fashioned destructive scaling,
but keeps track with scale: instance variable.!

!QPortableObject class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QPortableObject class methodsFor: 'inspector' stamp: 'bgf 1/26/2010 12:29'!
addPropertyRows: propertyDialog forFrame: poref
	"The base version of the QPropertyDialog row-population protocol;
	 add property rows for the user-editable aspects of this frame."

	self addNameRowIfNeeded: propertyDialog forFrame: poref.

	propertyDialog addRow: (
		(QTFramePropertyRow new)
			titleString: 'Grip For Moving';
			fetchBlock: [ : frame | (frame get: #gripForMoving) ];
			acceptBlock: [ : v :frame | frame future gripForMoving: v ] ;
			validateBlock: [ : v :frame | self validateGrip: v forObject: frame inDialog: propertyDialog ];
			balloonHelp: 	
				('An element of the object\to use as the drag-grip for moving.' withCRs);
			yourself).

	propertyDialog addRow: (
		(QTFramePropertyRow new)
			titleString: 'Grip For Turning';
			fetchBlock: [ : frame| (frame get: #gripForTurning) ];
			acceptBlock: [ : v : frame| frame future gripForTurning: v ] ;	
			validateBlock: [ : v :frame | self validateGrip: v forObject: frame inDialog: propertyDialog ];
			balloonHelp: 	
				('An element of the object\to use as the drag-grip for turning.' withCRs);
			yourself).

	super addPropertyRows: propertyDialog forFrame: poref.! !

!QPortableObject class methodsFor: 'instance creation' stamp: 'bgf 5/12/2007 15:28'!
consistingOf: aTFrame
	"A new portableObject frame around the given tframe."
	^ self new consistingOf: aTFrame.
! !

!QPortableObject class methodsFor: 'defaults' stamp: 'bgf 11/10/2009 15:56'!
defaultEventMask
	^ EventPointerOver + EventPointerDown! !

!QPortableObject class methodsFor: 'defaults' stamp: 'bgf 3/28/2009 22:23'!
defaultOpenSpacePolicy
	"See QDragAdapter openSpacePolicy"
	^ #none! !

!QPortableObject class methodsFor: 'defaults' stamp: 'bgf 1/19/2008 10:41'!
defaultSurfacePolicy
	"See QDragAdapter surfacePolicy"
	^ #viewer! !

!QPortableObject class methodsFor: 'inspector' stamp: 'bgf 3/28/2009 22:06'!
validateGrip: grip forObject: portable inDialog: propertyDialog

	grip ifNil: [^ true ].
	(portable get: #isOrContains: with: grip) ifFalse: [
		propertyDialog inform: 'Grips must be a part of the portable object.'.
		^false		
	].
	^ true! !

!QPortableObject methodsFor: 'places' stamp: 'bgf 3/24/2010 15:25'!
asPlace
	| place |
	"This lets 'Approach...' work reasonably for such things as card-holder composites."

	"We cannot just check children for asPlace, since the default TFrame asPlace is a parent-reflection."
	blockPlaceLoop = true ifTrue: [ ^ nil ].
	blockPlaceLoop := true.
	[
		self children do: [ : c | c asPlace ifNotNil: [ : pc | ^ pc  ]].
	] ensure: [ 
		blockPlaceLoop := false.
	].

	place :=	
		QPlace 
			category: 'Notes'
			label: (self objectName)
			frame: self
			selector: #destinationStanceFor: 
			sticky: true.
	place cameraSelector: #cameraStanceFor:.
	^ place
! !

!QPortableObject methodsFor: 'accessing' stamp: 'bgf 3/22/2010 09:01'!
asTopicRecord
	self isListedTopic ifTrue: [ 
		^ self mesh asTopicRecord
	].
	^ nil! !

!QPortableObject methodsFor: 'places' stamp: 'bgf 3/22/2010 09:11'!
cameraStanceFor: anAvatar
	"I offer ground-aware full-on viewing"
	^ QGotoButton frame: self cameraStanceFor: anAvatar! !

!QPortableObject methodsFor: 'initialize' stamp: 'bgf 3/28/2009 22:12'!
consistingOf: aTFrame

	"Make this a wrapper around the given frame."
	super consistingOf: aTFrame.
	self objectName: 'Portable-', (aTFrame objectName).

	self isGroundBound: (self isGroundBound or: [aTFrame isGroundBound]).
	self gripForMoving: (aTFrame).! !

!QPortableObject methodsFor: 'persist' stamp: 'bgf 2/11/2010 21:32'!
constructFrom: pStream

	super constructFrom:pStream.

	self class == QPortableObject ifTrue:[
		"XXXX: See TGroup."
		self children: (pStream readSequence:'children' default: self children).
	].

	self isAttractedToSurfaces: 
		(pStream readBool: 'isAttractedToSurfaces' default: (self isAttractedToSurfaces)).
	self isGroundBound: 
		(pStream readBool: 'isGroundBound' default: (self isGroundBound)).

	motionGrip attractionRange: (pStream readNumber: 'attractionRange' default: nil).
	motionGrip surfacePolicy: (pStream readString: 'surfacePolicy' default: self class defaultSurfacePolicy) asSymbol.
	motionGrip openSpacePolicy: (pStream readString: 'openSpacePolicy' default: self class defaultOpenSpacePolicy) asSymbol.

	(pStream readObject: 'motionGrip' default: (self mesh)) ifNotNilDo: [ : mg | self gripForMoving: mg ].
	(pStream readObject: 'rotationGrip' default: nil) ifNotNilDo: [ : rg | self gripForTurning: rg ].

	"Forward converter. Older ones wrote these properties, and were one-mode-at-atime."
	self class == QPortableObject ifTrue: [
			(pStream readBool: 'isRotationEnabled' default: false) ifTrue: [
				self gripForTurning: (self mesh).
				self gripForMoving: nil.
		]].

	"Subclasses are responsible for applying the scale again; we can only recover it at this point."
	(pStream readNumber: 'scale' default: nil)  ifNotNilDo: [ : scl |
		self defineScale: scl
	].! !

!QPortableObject methodsFor: 'initialize' stamp: 'bgf 4/16/2009 18:36'!
defaultIsAtomic
	^ (self class ~~ QPortableObject)! !

!QPortableObject methodsFor: 'places' stamp: 'bgf 3/22/2010 09:16'!
destinationExtent
	^ self mesh 
		ifNotNil: [ : m  | m destinationExtent ]
		ifNil: [ super destinationExtent ]! !

!QPortableObject methodsFor: 'places' stamp: 'bgf 3/22/2010 09:12'!
destinationStanceFor: anAvatar
	"I offer ground-aware full-on viewing"
	^ QGotoButton groundedStanceFor: anAvatar toView: self inDestination: self! !

!QPortableObject methodsFor: 'accessing' stamp: 'bgf 8/16/2009 00:48'!
dragAdapter
	^ motionGrip! !

!QPortableObject methodsFor: 'accessing' stamp: 'bgf 3/22/2010 09:05'!
editByUser: replica
	"Protocol infill for treating panels as attachments to topic collectors;
	 this is the response to the Topic Tool 'Open' (and topic-list double-click) command."
	
	replica goToPlace: (self asPlace)
	! !

!QPortableObject methodsFor: 'accessing' stamp: 'bgf 3/28/2009 23:26'!
gripForMoving
	^ motionGrip isEnabled ifTrue: [ 
		motionGrip grip ifNotNilDo: [ : grip | 
			"This protects against inadvertent persistence of stale/removed objects."
			(self isOrContains: grip) ifTrue: [ grip ] ifFalse: [ nil ]
		]
	   ] ifFalse: [ nil ]! !

!QPortableObject methodsFor: 'accessing' stamp: 'bgf 3/28/2009 22:52'!
gripForMoving: aChild

	"The descendant tframe that responds to user interaction
	 by moving the whole somewhat freely in space,
	 either following the imaginary sphere formed by the user's laser radius
	 or (more commonly) by following surfaces that the object is dragged against."

	aChild ifNil: [
		^ motionGrip grip: self; disable.
	].
	(self isOrContains: aChild) ifTrue: [
		motionGrip grip: aChild; enable.
	] ifFalse: [
		QLogger logInfo: 'Motion grip must be a part of the portable object.'.
	].! !

!QPortableObject methodsFor: 'accessing' stamp: 'bgf 3/28/2009 23:27'!
gripForTurning
	^ rotationGrip isEnabled ifTrue: [ 
		rotationGrip grip ifNotNilDo: [ : grip | 
			"This protects against inadvertent persistence of stale/removed objects."
			(self isOrContains: grip) ifTrue: [ grip ] ifFalse: [ nil ]
		]
	] ifFalse: [ nil ].! !

!QPortableObject methodsFor: 'accessing' stamp: 'bgf 2/12/2010 09:46'!
gripForTurning: aChild

	"The descendant tframe that responds to user interaction
	 by rotating the whole around its y axis."
	aChild ifNil: [
		motionGrip surfacePolicy: (self class defaultSurfacePolicy).
		^ rotationGrip grip: self; disable.
	].

	(self isOrContains: aChild) ifTrue: [
		rotationGrip grip: aChild; enable.
	] ifFalse: [
		QLogger logInfo: 'Rotation grip must be a part of the portable object.'.
	].
	motionGrip surfacePolicy: #none.! !

!QPortableObject methodsFor: 'initialize' stamp: 'bgf 2/12/2010 14:35'!
initialize
	super initialize.

	(motionGrip := QMotionAdapter forLuggage: self grip: self)
		isGroundBound: true;	"Default to non-flying behavior."
		surfacePolicy: (self class defaultSurfacePolicy);
		openSpacePolicy: (self class defaultOpenSpacePolicy);
		disable.

	(rotationGrip := QPlanePivot forLuggage: self grip: self)
		disable.! !

!QPortableObject methodsFor: 'accessing' stamp: 'bgf 9/14/2006 16:04'!
isAttractedToSurfaces
	^ motionGrip isAttractedToSurfaces! !

!QPortableObject methodsFor: 'accessing' stamp: 'bgf 9/14/2006 16:03'!
isAttractedToSurfaces: aBool
	motionGrip isAttractedToSurfaces: aBool! !

!QPortableObject methodsFor: 'testing' stamp: 'bgf 3/19/2009 13:44'!
isComponent
	^ true! !

!QPortableObject methodsFor: 'testing' stamp: 'bgf 6/1/2009 11:41'!
isEditableGroup
	"C.F. TGroup; plain portable-objects are not atomic so they can be excised with ungroup,
	 but derived components are by default atomic."
	^ (self class == QPortableObject) and: [ self isAtomic not ]! !

!QPortableObject methodsFor: 'accessing' stamp: 'bgf 9/14/2006 16:03'!
isGroundBound
	^ motionGrip isGroundBound! !

!QPortableObject methodsFor: 'accessing' stamp: 'bgf 9/14/2006 16:04'!
isGroundBound: aBool
	motionGrip isGroundBound: aBool! !

!QPortableObject methodsFor: 'testing' stamp: 'bgf 3/22/2010 09:00'!
isListedTopic
	"I play in the note-management schemes if my content wants to."
	^ self mesh 
			ifNotNil: [ : m | m isListedTopic ]
			ifNil: [ false ]! !

!QPortableObject methodsFor: 'accessing' stamp: 'bgf 6/1/2009 23:11'!
isPlacementLocked: bool
	bool~= self isPlacementLocked ifTrue: [
		super isPlacementLocked: bool.
		motionGrip isEnabled: (bool not).
		rotationGrip isEnabled: (bool not).
	].! !

!QPortableObject methodsFor: 'accessing' stamp: 'bgf 1/8/2007 08:53'!
mesh
	"The primary object that I enclose and make portable;
	 to be overidden by objects that provide a real TMesh representation."
	^ frameChildren ifNotNil: [frameChildren first]! !

!QPortableObject methodsFor: 'accessing' stamp: 'bgf 3/28/2009 22:20'!
motionAdapter
	"My motion adapter"
	^ motionGrip! !

!QPortableObject methodsFor: 'accessing' stamp: 'bgf 12/10/2006 12:55'!
payload

	"Drag protocol - I am the payload of my drag adapter."
	^ self! !

!QPortableObject methodsFor: 'persist' stamp: 'bgf 2/11/2010 21:31'!
persistOn: pStream
	super persistOn: pStream.
	self class == QPortableObject ifTrue:[
		"XXXX: See TGroup; similar case here."
		pStream storeSequence: 'children' value: self children default: #().
	].
	pStream storeBool: 'isAttractedToSurfaces' value: (self isAttractedToSurfaces) default: true.
	pStream storeBool: 'isGroundBound' value: (self isGroundBound) default: true.

	"These are arranged so that a default 'Make Portable' new portable object persists compactly;
	 by default the first child is the motion grip and there is no rotation grip."
	
	pStream storeObject: 'motionGrip' value: (self gripForMoving) default: (self mesh).
	pStream storeObject: 'rotationGrip' value: (self gripForTurning) default: nil.

	"Fine control motion-grip policies..."
	pStream storeNumber: 'attractionRange' value: (motionGrip attractionRange) default: nil.
	pStream storeString: 'surfacePolicy' value: (motionGrip surfacePolicy) default: (self class defaultSurfacePolicy).
	pStream storeString: 'openSpacePolicy' value: (motionGrip openSpacePolicy) default: (self class defaultOpenSpacePolicy).


! !

!QPortableObject methodsFor: 'events' stamp: 'bgf 3/28/2009 22:13'!
pointerEnter: event

	event avatar ifNil: [ ^ self ].
	super pointerEnter: event.
	event avatar signal: #statusMessage with: (self labelForBrowser).
! !

!QPortableObject methodsFor: 'accessing' stamp: 'bgf 3/28/2009 22:20'!
rotationAdapter
	"My motion adapter"
	^ rotationGrip! !

!QPortableObject methodsFor: 'places' stamp: 'bgf 3/22/2010 09:18'!
warpThresholdAdjustment
	"We want to drag/arrange them from afar."
	^ 5! !

TGroup subclass: #QSkyDome
	instanceVariableNames: 'dalpha style'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Objects-Space'!

!QSkyDome commentStamp: '<historical>' prior: 0!
A round sky.  By default, is fullBright and so immune to fog.

The sky can be darkened and can be used without fullBright,
so it fades behind fog.      See #brightness:fullBright:

Also the rotation rate can be changed.  #rotationRate: degreesPerSecond
controls the rotation rate of the sky as a whole, which appears as cloud
drift to a viewer in the space.!

!QSkyDome class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QSkyDome class methodsFor: 'defaults' stamp: 'bgf 5/31/2009 02:31'!
defaultDoContextMenu
	^ false! !

!QSkyDome class methodsFor: 'as yet unclassified' stamp: 'bgf 1/19/2010 16:11'!
defaultDoesClickSelect
	^ false! !

!QSkyDome class methodsFor: 'defaults' stamp: 'bgf 6/4/2009 16:13'!
defaultIsFromTemplate
	"Reduces xml volume since these are most always scenery."
	^ true! !

!QSkyDome methodsFor: 'initialize' stamp: 'bgf 6/9/2010 11:56'!
brightness: aFactor fullBright: bFlag

	"Update the sky's material properties to the given brightness factor (0-1)
	 and fullBright choice.    A fullBright sky disregards fog.
	 A non-fullBright sky is subject to fog, and so fog coloring
	 becomes important to the overall effect.
	 Brightess is the linear black-to-white color of the material used
	 to render the sky; 0 is black; 1 is white (so the sky images appear faithfully)"

	|  mat factor base emissive |

	factor := aFactor min: 1.0 max:0.0.

	"If fullBright, the diffuse color dominates, and must be white.
	 If not, the emissive color is the ony one want to have effect,
	 so we clamp diffuse/ambient to black, so that local lighting
	 doesn't shade the skybox faces and reveal corners.
	 (When fullbright, local lighting is ignored in the first place.)"

	emissive := Color white.
	base := (bFlag ifTrue: [  Color white ] ifFalse: [ Color black ]).

	(frameChildren ifNil: [#() ]) do: [ :fc |
		mat := fc material.
		mat ambientColor: base * factor.
		mat diffuseColor: base *factor.
		mat emissiveColor: (emissive*factor).
		mat fullBright: bFlag.
		mat textureMode: (bFlag ifTrue: [GLReplace] ifFalse: [GLModulate]).
	].! !

!QSkyDome methodsFor: 'persist' stamp: 'bgf 2/11/2010 17:11'!
constructFrom: pStream	
	| sc |
	super constructFrom: pStream.
	"Rescale to adjust the sky size, but not the position (whose persistence is already post-scaling)"
	sc := pStream readNumber: 'scale' default: 1.0.
	sc ~= 1.0 ifTrue: [
		self scaleTo: sc		"This moves our old 'scale' to new nominalScale: "
	].
	self style: (pStream readString: 'style' default: self style).
! !

!QSkyDome methodsFor: 'initialize' stamp: 'bgf 2/11/2010 10:08'!
initialize
	"By default, the sky is fully bright, so will render the skybox 
	 accurately without regard to fog, using emissive light.
	 The default rotationRate is 1 degree/second."
	|  loader tframe |

	super initialize.
	style := #normal.

	"We package the .c3x.  The ASE load is for prepackage development use,
	 to update the model."
	tframe := TSceneReader new loadFileNamed: 'Objects/Qwaq/Skydome.c3x'.
	false ifTrue: [
		tframe ifNil:[
			loader := TLoad3DSMax new.
			loader addSearchPath: 'Objects/Qwaq/'.
			tframe := loader loadFileNamed: 'Skydome.ase'
				scale: 0.002 shadeAngle: 90.1 textureMode: GLModulate.
			tframe := tframe collapse.
			TSceneWriter new storeScene: tframe 
				onFileNamed: 'Objects/Qwaq/Skydome.c3x'.
		]].
	tframe translation:0@-15@0.
	tframe boundsDepth: 2.
	tframe solid: false.
	self addChild: tframe.
	self pickable: false.
	self doContextMenu: false.

	self brightness: 1.0 fullBright: true.
	dalpha := 0.015.
! !

!QSkyDome methodsFor: 'persist' stamp: 'bgf 2/11/2010 10:08'!
persistOn: pStream	
	super persistOn: pStream.

	pStream storeString: 'style' value: self style default: #normal.! !

!QSkyDome methodsFor: 'initialize' stamp: 'bgf 2/23/2008 02:40'!
rotationRate: aFloat
	"The background rotation rate, in degrees/second around the Y axis."

	dalpha := (aFloat/25).  "The stepper uses 40 ms increments."! !

!QSkyDome methodsFor: 'stepping' stamp: 'bgf 1/27/2007 13:51'!
start 
	self future step! !

!QSkyDome methodsFor: 'stepping' stamp: 'bgf 2/23/2008 02:39'!
step
	"rotation is dalpha units per 40 milliseconds, and we assume start at zero clock time" 
	self rotationAroundY: (IslandTime millisecondClockValue asFloat/40.0) * dalpha.
	(self future: 40) step! !

!QSkyDome methodsFor: 'accessing' stamp: 'bgf 2/3/2007 23:49'!
style
	"A symbol (#normal, #drawn, #abstract) 
	 describing the rendering style of the sky image."
	^ style.! !

!QSkyDome methodsFor: 'accessing' stamp: 'bgf 10/19/2008 16:20'!
style: aStyle
	"Sky texture style: 
		#normal - a daylight sky;  
		#abstract - a liquified variant
		#drawn - a graphically reduced variant
		#clouds - blurry clouds
	"

	| skyTextureName mat |

	(aStyle = style) ifTrue: [ ^ self].

	skyTextureName := aStyle caseOf: {
		[#abstract] ->['SkyDomeAbstract.jpg'].
		[#drawn]  -> ['SkyDomeDrawn.jpg'] .
		[#clouds] -> ['Clouds.jpg'].
	} otherwise: [ 'SkyDome.jpg'].

	frameChildren do: [ :fc |
		mat := fc material.
		mat texture: (QTextures find: 'Content/Qwaq/Environment/', skyTextureName).
	].
	style := aStyle.! !

!QSkyDome methodsFor: 'hierarchy' stamp: 'bgf 10/12/2008 18:10'!
transformChanged
	"Just do it. Otherwise our rapid-step induces another future-send each time.
	 I.e. don't use lazy-updates here."
	self updateTransform! !

TGroup subclass: #QSoundRegion
	instanceVariableNames: 'attenuation doors'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Objects-Space'!

!QSoundRegion commentStamp: '<historical>' prior: 0!
An invisble volume that acts as a sound attenuation barrier;
avatars together in the same volume hear each other clearly;
avatars outside the volume are attenuated.

The volume described is the bouding box of whatever frames I contain;
currently no smart checking is done, it's just a bounding box check.

Attenuation is applied to the gain when playing sounds in the volume
for listeners that are outside the volume. 
The base number runs [0..1], 1 for full attenuation.

!

!QSoundRegion class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QSoundRegion class methodsFor: 'inspector' stamp: 'bgf 1/26/2010 12:43'!
addPropertyRows: propertyDialog forFrame: srref

	
	propertyDialog addRow: (
		(QScalarPropertyRow new)
			titleString: 'Sound Attenuation';
			fetchBlock: [ :f |f  get:#attenuation ];
			acceptBlock: [ : at : f  | f future attenuation: at ];
			validateBlock: [ : float | 
				(float isNil or: [(float < 0.0) | (float > 1)]) ifTrue: [
					self inform: 'Attenuation must be between 0 and 1.'.
					false
				] ifFalse: [
					true
				]] ;
			balloonHelp: 'Reduction in voice volume across the region boundary.';
	 	yourself	).

	propertyDialog addRow: (self doorsRowForFrame: srref).
	super addPropertyRows: propertyDialog forFrame: srref.! !

!QSoundRegion class methodsFor: 'defaults' stamp: 'bgf 2/19/2008 18:12'!
defaultAttenuation

	"Sounds behave strangely... this is just about enough to mute the background."
	^ 0.8! !

!QSoundRegion class methodsFor: 'defaults' stamp: 'bgf 6/4/2009 16:13'!
defaultIsFromTemplate
	"Reduces xml volume since these are most always scenery."
	^ true! !

!QSoundRegion class methodsFor: 'inspector' stamp: 'bgf 1/26/2010 12:41'!
doorsRowForFrame: srref

	| row |
	row := (QTFrameArrayPropertyRow new).	
	row titleString: 'Connecting Doors';
		fetchBlock: [ : ref | ref  get: #doorArray ];
		acceptBlock: [: v : ref | ref  future doorArray: v ];
		validateBlock: [ :v | (v detect: [ :f | f refClass ~= QDoor ] ifNone: [nil])
							ifNotNilDo: [ : f |
								self inform: ('Only Doors may be used\as connecting doors.\\', 
											(f get: #labelForBrowser), ' is a ', (f get: #genericName), '.'
											) withCRs.
								false
							] ifNil: [ true ]];
		balloonHelp: 'Doors between regions\activate the sound attenuation\when the door is closed.' withCRs.

	^ row! !

!QSoundRegion class methodsFor: 'undo-redo' stamp: 'bgf 3/14/2010 17:54'!
region: rgnRef addDoor: qDoor

	^ QBlockAction
		undoBlock: [ rgnRef future removeDoor: qDoor ]
		redoBlock: [ rgnRef future addDoor: qDoor ]
		description: 'Add Door'
		requiredFrames: { rgnRef . qDoor }! !

!QSoundRegion methodsFor: 'accessing' stamp: 'bgf 7/28/2008 14:07'!
addDoor: aQDoor

	"Add the door to the regions set of checkable doors;
	 if two regions share a door, 
	 then their attenuations contribute only when the door is closed."
	doors add: aQDoor.
! !

!QSoundRegion methodsFor: 'accessing' stamp: 'bgf 3/14/2010 09:09'!
addDoor: aQDoor droppedByUser: anAvatar

	"Add the door to the regions set of checkable doors;
	 if two regions share a door, 
	 then their attenuations contribute only when the door is closed."
	doors add: aQDoor.

	anAvatar recordAction: #region:addDoor:
			 class: QSoundRegion
			 arguments: { self . aQDoor }.
			! !

!QSoundRegion methodsFor: 'accessing' stamp: 'bgf 2/19/2008 14:51'!
attenuation
	"Float [0-1] ... multiply the playback gain on objects in this volume, 
	 when playing for objects outside the volume, and vice-versa;
	 so low values mean high attenuation."

	^ attenuation! !

!QSoundRegion methodsFor: 'accessing' stamp: 'bgf 2/19/2008 17:30'!
attenuation: aFloat
	"Float [0-1] reduction in playback volume when listener and speaker
	 are on opposite sides of my boundary."
	attenuation := (aFloat min: 1.0 max: 0.0).
	self root signal: #soundRegionAttenuationChanged with: self with: attenuation.! !

!QSoundRegion methodsFor: 'accessing' stamp: 'bgf 7/26/2009 13:34'!
bestSeatWithin: tbox nearest: avatarPosition
	"Sound regions will get wrapped around meeting places, so they need seat-assigning."

	^ self genericBestSeatWithin: tbox nearest: avatarPosition
	! !

!QSoundRegion methodsFor: 'persist' stamp: 'bgf 7/28/2008 17:07'!
checkDoors

	"Filter out unrooted (e.g. destroyed) doors... done before persisting to GC them."
	doors := doors select: [ : d | d root notNil ].! !

!QSoundRegion methodsFor: 'testing' stamp: 'bgf 9/1/2008 04:11'!
combinedAttenuation: otherSoundRegion

	"Return the appropriate attenuation between a speaker in this region
	 and a listener in another.  Assumes both regions are non-nil.
	 Recognizes doors -- does not attenuate if there is an open connecting door.

	 Must be side-effect free - queried via get: from the sound system."

	(otherSoundRegion = self) ifTrue: [ ^ 0.0 ].
	(self hasOpenDoorTo: otherSoundRegion) ifTrue: [ ^ 0.0 ].

	"Attenuate..."
	^ 1 - ((1.0 - attenuation )*(1.0 - (otherSoundRegion attenuation)))! !

!QSoundRegion methodsFor: 'persist' stamp: 'bgf 7/28/2008 14:08'!
constructFrom: pStream
	"Construct the receiver"
	super constructFrom: pStream.
	self attenuation: (pStream readNumber: 'attenuation' default: self attenuation).
	self class == QSoundRegion ifTrue: [
		self children: (pStream readSequence:'children' default: self children).
	].
	(pStream readSequence: 'doors' default: #()) do: [ : door |
		self addDoor: door
	].! !

!QSoundRegion methodsFor: 'accessing' stamp: 'bgf 5/27/2009 22:58'!
containsPoint: aGlobalV3

	"Do I enclose the given global point?"
	^ self children anySatisfy: [ : kid |
			(kid boundingBoxUnscaled transformedBy: kid globalTransform) 
		 		pointInBox: aGlobalV3
		].! !

!QSoundRegion methodsFor: 'initialize' stamp: 'bgf 4/16/2009 18:36'!
defaultIsAtomic
	^  (self class ~~ QSoundRegion)! !

!QSoundRegion methodsFor: 'initialize' stamp: 'bgf 2/19/2008 21:31'!
destroy
	super destroy.
	self root ifNotNil: [
		self root removeSoundRegion: self.
	].! !

!QSoundRegion methodsFor: 'accessing' stamp: 'bgf 7/28/2008 14:06'!
doorArray
	"My set of doors, as an array for far-ref get convenience"
	^ doors asArray! !

!QSoundRegion methodsFor: 'accessing' stamp: 'bgf 7/29/2009 21:54'!
doorArray: array
	"Replace the entire set of connecting-doors."
	array ifNil: [
		doors := Set new.
	] ifNotNil: [
		doors := array asSet.
	].
! !

!QSoundRegion methodsFor: 'accessing' stamp: 'bgf 7/28/2008 14:06'!
doors
	"My set of doors."
	^ doors! !

!QSoundRegion methodsFor: 'testing' stamp: 'bgf 9/1/2008 04:04'!
hasOpenDoorTo: soundRegion
	"Is there a connecting door between me and the other region... that is open?"

	| otherDoors |
	otherDoors := soundRegion doors.

	^ doors anySatisfy: [ : door |
				door isOpen 
				and: [ otherDoors includes: door ]].! !

!QSoundRegion methodsFor: 'initialize' stamp: 'bgf 7/28/2008 14:02'!
initialize

	super initialize.
	doors := Set new.
	self attenuation: (self class defaultAttenuation).
! !

!QSoundRegion methodsFor: 'testing' stamp: 'bgf 6/1/2009 11:43'!
isEditableGroup
	^ (self class == QSoundRegion) and: [ self isAtomic not ]! !

!QSoundRegion methodsFor: 'accessing' stamp: 'bgf 7/29/2009 10:44'!
numDoorsConnected
	^ doors size! !

!QSoundRegion methodsFor: 'persist' stamp: 'bgf 7/28/2008 17:06'!
persistOn: pStream
	"Persist the receiver"
	super persistOn: pStream.
	pStream storeNumber: 'attenuation' value: self attenuation default: (self class defaultAttenuation).
	self class == QSoundRegion ifTrue: [
		pStream storeSequence: 'children' value: self children default: #().
	].
	self checkDoors.
	pStream storeSequence: 'doors' value: self doorArray default: #().! !

!QSoundRegion methodsFor: 'accessing' stamp: 'bgf 7/28/2008 14:27'!
removeDoor: aQDoor
	"Remove from my set of edge-controlling doors."
	doors remove: aQDoor ifAbsent: [].! !

!QSoundRegion methodsFor: 'accessing' stamp: 'bgf 2/19/2008 21:31'!
root: newRoot
	frameRoot 
		ifNotNil: [ frameRoot removeSoundRegion: self ].
	super root: newRoot.
	frameRoot 
		ifNotNil: [ frameRoot addSoundRegion: self ].! !

!QSoundRegion methodsFor: 'initialize' stamp: 'bgf 2/19/2008 14:55'!
updateShape
	super updateShape.
	self root signal: #soundRegionShapeChanged with: self with: (self boundingBox)! !

TGroup subclass: #QTFPortableObject
	instanceVariableNames: 'lastPickDirection lastPickPoint rigidBody motionGrip rotationGrip saveEventMask'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Objects-Manipulation'!

!QTFPortableObject commentStamp: '<historical>' prior: 0!
This is the fat tree-fort predecessor of QPortableObject,
retained for debugging while validating the simplified forums replacement.

PortableObjects can be moved, stored, and eventually owned by users.
Intuitively the class includes furniture, toys, and other equipment,
but not walls or other fixtures of the environment.

I provide common services for manipulable objects:

	Scaling with memory... my 'scale' member is the cumulative product
	of all scaling operations performed on me.

	Autoregistration as a physical object, and some simple physical interactions
	(see '#keyStroke:', where keys are interpret as push, pull, stop, etc.)
	(This support is not active at present.)

	A clickable menu to make the object movable, rotatable, or neither,
	and to remove (destroy) the object.!

!QTFPortableObject class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QTFPortableObject class methodsFor: 'instance creation' stamp: 'bgf 5/12/2007 15:28'!
consistingOf: aTFrame
	"A new portableObject frame around the given tframe."
	^ self new consistingOf: aTFrame.
! !

!QTFPortableObject class methodsFor: 'defaults' stamp: 'bgf 11/10/2009 16:23'!
defaultEventMask
	^ EventPointerOver + EventPointerDown + EventKeyboard! !

!QTFPortableObject class methodsFor: 'defaults' stamp: 'bgf 1/19/2008 10:41'!
defaultOpenSpacePolicy
	"See QDragAdapter openSpacePolicy"
	^ #viewer! !

!QTFPortableObject class methodsFor: 'defaults' stamp: 'bgf 1/19/2008 10:41'!
defaultSurfacePolicy
	"See QDragAdapter surfacePolicy"
	^ #viewer! !

!QTFPortableObject methodsFor: 'menu' stamp: 'bgf 9/1/2009 11:18'!
addMenuActions: aTMenu forEvent: event

	motionGrip ifNotNil: [
		(aTMenu addItemSpec: {'Allow Moving'. #toggleMove } target: self)
			checked: motionGrip isEnabled
	].
	rotationGrip ifNotNil: [
		(aTMenu addItemSpec: {'Allow Turning'. #toggleRotate} target: self)
			checked: rotationGrip isEnabled.
	].

	aTMenu addList: {
		#- . 
		{'Bigger'. #enlargeObject} .
		{'Smaller'. #shrinkObject} .
	} target: self.

	aTMenu addList: {	
		#-  .
		{'Remove' . #confirmDelete:}
	} target: self.! !

!QTFPortableObject methodsFor: 'initialize' stamp: 'bgf 7/26/2006 13:16'!
addToEngine: aPhysicsEngine

	| |
	rigidBody := self makeRigidBody.
	rigidBody ifNotNil: [
		aPhysicsEngine add: rigidBody.
	].! !

!QTFPortableObject methodsFor: 'menu' stamp: 'bgf 1/28/2007 15:20'!
checkMenu:aTMenu
	"Set the checkmarks in the menu - uses selector comparison."
	aTMenu menuItems do: [ :item| 
		| selector |
		selector := item action ifNotNilDo:[:a|a selector].
		selector caseOf: {
			[#toggleMove]-> [item checked: (motionGrip isEnabled)] .
			[#toggleRotate]-> [item checked: (rotationGrip isEnabled)] .
		} otherwise:[].
	].
! !

!QTFPortableObject methodsFor: 'initialize' stamp: 'bgf 3/19/2009 14:28'!
consistingOf: aTFrame

	"Make this a wrapper around the given frame."
	super consistingOf: aTFrame.
	self objectName: 'Portable (', (aTFrame objectName), ')'.

	self isGroundBound: (self isGroundBound or: [aTFrame isGroundBound]).
! !

!QTFPortableObject methodsFor: 'persist' stamp: 'bgf 2/11/2010 21:30'!
constructFrom: pStream

	super constructFrom:pStream.

	self class == QPortableObject ifTrue:[
		"XXXX: See TGroup."
		self children: (pStream readSequence:'children' default: self children).
	].

	self isAttractedToSurfaces: 
		(pStream readBool: 'isAttractedToSurfaces' default: (self isAttractedToSurfaces)).
	self isGroundBound: 
		(pStream readBool: 'isGroundBound' default: (self isGroundBound)).

	motionGrip attractionRange: (pStream readNumber: 'attractionRange' default: nil).
	motionGrip surfacePolicy: (pStream readString: 'surfacePolicy' default: self class defaultSurfacePolicy) asSymbol.
	motionGrip openSpacePolicy: (pStream readString: 'openSpacePolicy' default: self class defaultOpenSpacePolicy) asSymbol.

	motionGrip isEnabled:
		(pStream readBool: 'isMotionEnabled' default: (motionGrip isEnabled)).
	rotationGrip isEnabled: 
		(pStream readBool: 'isRotationEnabled' default: (rotationGrip isEnabled)).
	(motionGrip isEnabled or: [rotationGrip isEnabled]) ifFalse:[
		motionGrip enableClickOnly.
	].

	saveEventMask := pStream readNumber: 'saveEventMask'  default: nil.

	"Subclasses are responsible for applying the scale again; we can only recover it at this point."
	(pStream readNumber: 'scale' default: nil) ifNotNil: [ :scl |
		self defineScale: scl
	].! !

!QTFPortableObject methodsFor: 'initialize' stamp: 'bgf 4/16/2009 18:36'!
defaultIsAtomic
	^ (self class ~~ QTFPortableObject)! !

!QTFPortableObject methodsFor: 'events' stamp: 'bgf 8/29/2007 17:19'!
destroy

	self mesh ifNotNil: [
		self mesh rigidBody ifNotNil: [
			self root engine remove: self mesh rigidBody.
		].
	].
	super destroy.
! !

!QTFPortableObject methodsFor: 'accessing' stamp: 'bgf 2/20/2007 00:42'!
disable
	"Disable both grips."
	motionGrip enableClickOnly.	
	rotationGrip disable.
	self eventMask: (EventPointerOver + EventPointerDown + EventKeyboard).
! !

!QTFPortableObject methodsFor: 'accessing' stamp: 'bgf 8/16/2009 00:50'!
dragAdapter
	^ motionGrip! !

!QTFPortableObject methodsFor: 'accessing' stamp: 'bgf 9/24/2006 17:18'!
enable
	motionGrip enable.! !

!QTFPortableObject methodsFor: 'menu' stamp: 'bgf 9/24/2006 20:53'!
enableMove
	rotationGrip disable.
	motionGrip enable.! !

!QTFPortableObject methodsFor: 'menu' stamp: 'bgf 9/24/2006 20:53'!
enableRotate
	motionGrip disable.
	rotationGrip enable.! !

!QTFPortableObject methodsFor: 'menu' stamp: 'bgf 12/14/2006 12:13'!
enlargeObject
	self scaleBy: 1.3.! !

!QTFPortableObject methodsFor: 'step' stamp: 'bgf 11/10/2009 16:44'!
hibernate
	" Stop all stepping and physical interaction; remove from the physics engine."

	self stop.
	lastPickPoint := nil.
	lastPickDirection := nil.

	(self root notNil and: [self rigidBody notNil]) ifTrue: [
		self root engine remove: self rigidBody.
	].
	self stopAllPhysics.
	motionGrip disable.
	rotationGrip disable.

	self unhilite.
	self disable.

	saveEventMask := self eventMask.
	self eventMask: EventNone.
	Processor activeIsland removeMessagesWithReceiver: self.
! !

!QTFPortableObject methodsFor: 'initialize' stamp: 'bgf 2/11/2010 21:31'!
initialize
	super initialize.
	
	(motionGrip := QMotionAdapter forLuggage: self grip: self)
		isDroppable: true;  "So we can put things away."
		isGroundBound: true;	"Default to non-flying behavior."
		surfacePolicy: (self class defaultSurfacePolicy);
		openSpacePolicy: (self class defaultOpenSpacePolicy).

	(rotationGrip := QPlanePivot forLuggage: self grip: self)
		disable.
! !

!QTFPortableObject methodsFor: 'menu' stamp: 'bgf 12/14/2006 11:49'!
inspectObject
	self inspect.! !

!QTFPortableObject methodsFor: 'accessing' stamp: 'bgf 9/14/2006 16:04'!
isAttractedToSurfaces
	^ motionGrip isAttractedToSurfaces! !

!QTFPortableObject methodsFor: 'accessing' stamp: 'bgf 9/14/2006 16:03'!
isAttractedToSurfaces: aBool
	motionGrip isAttractedToSurfaces: aBool! !

!QTFPortableObject methodsFor: 'testing' stamp: 'bgf 3/19/2009 13:44'!
isComponent
	^ true! !

!QTFPortableObject methodsFor: 'testing' stamp: 'bgf 6/1/2009 11:42'!
isEditableGroup
	"C.F. TGroup; plain portable-objects are not atomic so they can be excised with ungroup,
	 but derived components are by default atomic."
	^ (self class == QPortableObject) and: [ self isAtomic not ]! !

!QTFPortableObject methodsFor: 'accessing' stamp: 'bgf 9/14/2006 16:03'!
isGroundBound
	^ motionGrip isGroundBound! !

!QTFPortableObject methodsFor: 'accessing' stamp: 'bgf 9/14/2006 16:04'!
isGroundBound: aBool
	motionGrip isGroundBound: aBool! !

!QTFPortableObject methodsFor: 'events' stamp: 'bgf 1/8/2007 08:22'!
keyStroke: event

	"Toys have some extra operations... +/- to enlarge and shrink,
	 and 's' to start and stop the basic toy animation steps."

	self motionGrip isEnabled ifFalse: [
		^ self
	].

	event keyCharacter caseOf: {
		[$  ]		-> [self push:event.].	"physics"
		[$p]		-> [self pull:event.].
		[$s]		-> [	self stopAllPhysics].
		[$+]	-> [self scaleBy: 1.2].		"Enlarge"
		[$-]		-> [self scaleBy: (1.0/1.2) asFloat].	"Shrink"
		[$t]		-> [self addRotationAroundY: -15]. "turn"
		[$T]		-> [self addRotationAroundY: 15].

	} otherwise: [
		super keyStroke: event.
	].! !

!QTFPortableObject methodsFor: 'initialize' stamp: 'das 2/7/2007 10:22'!
makeRigidBody

	"Returns a TRigidBody based on the mesh of this toy,
       or, if the 'mesh' is actually a group, then its first mesh child."

	|m idx |
	(self mesh isKindOf: TGeometry) ifTrue: [
		m := self mesh.
	] ifFalse: [
		"While experimenting, bind the rigid body to the first mesh child."
		idx := self mesh frameChildren findFirst: [ :mc|mc isKindOf: TGeometry].
		(idx > 0) ifTrue: [
			m := self mesh frameChildren at: idx.
		].
	].
	m ifNotNil: [
		^ TRigidBody new mesh: m density: 1.0
	].
	^ nil.! !

!QTFPortableObject methodsFor: 'menu' stamp: 'bgf 9/14/2007 21:25'!
menuStyle 
	^ QStyles menuStyle! !

!QTFPortableObject methodsFor: 'accessing' stamp: 'bgf 1/8/2007 08:53'!
mesh
	"The primary object that I enclose and make portable;
	 to be overidden by objects that provide a real TMesh representation."
	^ frameChildren ifNotNil: [frameChildren first]! !

!QTFPortableObject methodsFor: 'accessing' stamp: 'bgf 9/14/2006 16:04'!
motionGrip
	"My motion adapter"
	^ motionGrip! !

!QTFPortableObject methodsFor: 'accessing' stamp: 'bgf 12/10/2006 12:55'!
payload

	"Drag protocol - I am the payload of my drag adapter."
	^ self! !

!QTFPortableObject methodsFor: 'persist' stamp: 'bgf 2/11/2010 21:29'!
persistOn: pStream

	super persistOn: pStream.

	self class == QPortableObject ifTrue:[
		"XXXX: See TGroup; similar case here."
		pStream storeSequence: 'children' value: self children default: #().
	].

	pStream storeBool: 'isAttractedToSurfaces' value: (self isAttractedToSurfaces) default: true.
	pStream storeBool: 'isGroundBound' value: (self isGroundBound) default: true.
	pStream storeBool: 'isMotionEnabled' value: (motionGrip isEnabled).
	pStream storeBool: 'isRotationEnabled' value: (rotationGrip isEnabled).

	"Fine control motion-grip policies..."
	pStream storeNumber: 'attractionRange' value: (motionGrip attractionRange) default: nil.
	pStream storeString: 'surfacePolicy' value: (motionGrip surfacePolicy) default: (self class defaultSurfacePolicy).
	pStream storeString: 'openSpacePolicy' value: (motionGrip openSpacePolicy) default: (self class defaultOpenSpacePolicy).

	"In case we're persisted while hibernated."
	pStream storeNumber: 'saveEventMask' value: saveEventMask default: nil.

	"TBD - physics engine state."

! !

!QTFPortableObject methodsFor: 'events' stamp: 'bgf 2/15/2007 11:10'!
pointerClick: event

	event avatar ifNil: [
		"Must have an avatar to post a menu (no cross-space menus)"
		^ self
	].
	self showMenu:event.! !

!QTFPortableObject methodsFor: 'events' stamp: 'bgf 5/5/2007 15:20'!
pointerDown: event

	event avatar ifNil: [ ^ self ].
	self savePoint:event.
	super pointerDown: event.
! !

!QTFPortableObject methodsFor: 'events' stamp: 'bgf 5/5/2007 15:20'!
pointerEnter: event

	event avatar ifNil: [ ^ self ].
	super pointerEnter: event.
	event avatar signal: #statusMessage with: (self objectName).
! !

!QTFPortableObject methodsFor: 'events' stamp: 'bgf 2/15/2007 11:11'!
pointerLeave: event

	lastPickDirection := nil.
	lastPickPoint    := nil.

	super pointerLeave:event.! !

!QTFPortableObject methodsFor: 'events' stamp: 'bgf 5/5/2007 15:20'!
pointerMove: event

	event avatar ifNil: [ ^ self ].
	self savePoint:event.
	super pointerMove:event.
! !

!QTFPortableObject methodsFor: 'events' stamp: 'bgf 5/5/2007 15:20'!
pointerOver: event

	event avatar ifNil: [ ^ self ].
	self savePoint:event.
	super pointerOver:event.

! !

!QTFPortableObject methodsFor: 'events' stamp: 'bgf 2/11/2010 21:30'!
pull: event

	"Emit a motion impulse towards the viewer."

	| direction |
	(lastPickDirection notNil & lastPickPoint notNil & self mesh notNil) ifTrue: [
		self mesh rigidBody ifNotNil: [

			"Until there's a good way to sort out where the push on the body was intended,
	             we simplify pushes to linear pushes (through the center of mass)."

			direction := self localTransform inverse transformVector: lastPickDirection.
			self mesh rigidBody impulseLinear: 
				(direction * self mesh rigidBody mass * (self scale)) / (-500.0).
		].
	]! !

!QTFPortableObject methodsFor: 'events' stamp: 'bgf 2/11/2010 21:30'!
push: event

	"Emit a motion impulse away from the viewer"

	| direction |

	(lastPickDirection notNil & lastPickPoint notNil & self mesh notNil) ifTrue: [
		rigidBody ifNotNil: [

			"Until there's a good way to sort out where the push on the body was intended,
	             we simplify pushes to linear pushes (through the center of mass)."

			"We may be rotated w/respect to the space, and the mesh doesn't know that."
			direction := self localTransform inverse transformVector: lastPickDirection.

			rigidBody impulseLinear: 
					(direction * rigidBody mass * (self scale)) / (500.0).
		].
	]! !

!QTFPortableObject methodsFor: 'menu' stamp: 'bgf 1/29/2010 11:42'!
reallyDelete: event

	"If the space supports discard (for trash recovery) use it,
	 otherwise just go away."

	(self root respondsTo: #discard:) ifTrue: [
		self root discard: self. 
	] ifFalse: [
		super reallyDelete: event
	].! !

!QTFPortableObject methodsFor: 'render' stamp: 'bgf 11/19/2006 21:54'!
renderFrame: ogl

	"This is a cheap performance hack - ignore portable objects through 3D portals.
	 Toys in tree houses are therefore not rendered from the field."		


	(ogl inPortal3D) ifTrue: [
		^ 0
	].
	^ super renderFrame: ogl.! !

!QTFPortableObject methodsFor: 'accessing' stamp: 'das 11/3/2006 15:42'!
rigidBody

	"The rigid body I present to the physics engine."
	^ rigidBody! !

!QTFPortableObject methodsFor: 'events' stamp: 'bgf 8/6/2006 00:24'!
savePoint: event
	| pointer |

	"Save the lastPickPoint (in my frame) and lastPickDirection (a global normal)
	based on the current pointer location and viewer position.
	We do this to remember pointer relationships for key strokes 
	that produce physical impulses."

	pointer := event makePointer.

	(event selection normal notNil and: [event selection point notNil]) ifFalse: [
		lastPickPoint := nil.
		lastPickDirection := nil.
		^ self.
	].
	pointer frame: self pickPlane: event selection point normal: event selection normal.
	lastPickPoint  := pointer selectedPoint.
	lastPickPoint  ifNotNil: [
		"The 2.5 fudge compensates for the viewer's expectation 
		 that his hands are well below his viewpoint."

		lastPickDirection := ((self localToGlobal:lastPickPoint) - 
							(event avatar translation - (0@2.5@0))) normalized.
	].

! !

!QTFPortableObject methodsFor: 'menu' stamp: 'bgf 12/14/2006 12:13'!
shrinkObject
	self scaleBy: (1/1.3)! !

!QTFPortableObject methodsFor: 'events' stamp: 'bgf 7/26/2006 13:50'!
stopAllPhysics

		self mesh ifNotNil: [
			self mesh rigidBody ifNotNil: [
				self mesh rigidBody allStop.
			].
		]
! !

!QTFPortableObject methodsFor: 'menu' stamp: 'bgf 2/20/2007 00:44'!
toggleMove
	rotationGrip disable.
	motionGrip isEnabled: (motionGrip isEnabled not).

	"Don't leave the menu disabled."
	(rotationGrip isEnabled or: [motionGrip isEnabled]) ifFalse: [
		motionGrip enableClickOnly
	].
	self eventMask: (EventPointerOver + EventPointerDown + EventKeyboard).
! !

!QTFPortableObject methodsFor: 'menu' stamp: 'bgf 2/20/2007 00:44'!
toggleRotate
	motionGrip disable.
	rotationGrip isEnabled: (rotationGrip isEnabled not).
	"Don't leave the menu disabled."
	(rotationGrip isEnabled or: [motionGrip isEnabled]) ifFalse: [
		motionGrip enableClickOnly
	].
	self eventMask: (EventPointerOver + EventPointerDown + EventKeyboard).
! !

!QTFPortableObject methodsFor: 'step' stamp: 'bgf 11/8/2006 13:51'!
unhibernate
	"restore normal activity... start stepping, rejoin the physics engine, allow interaction."

	self eventMask: saveEventMask.
	self start.
	self rigidBody ifNotNil: [
		self root engine add: self rigidBody.
	].
	motionGrip enable.
! !

!TGroup methodsFor: '*Qwaq-Objects-Space' stamp: 'bgf 7/26/2009 14:36'!
bestSeatWithin: tbox nearest: avatarPosition
	"Generic seat-finding applies to generic Groups.
	 Returns either nil, or an Unoccupied QSeat that is within the given TBox,
	 and selected to be the right seat to go to for an avatar at the given position.

	 In subclasses:  override the class== check 
	 if the class is a generic container that may hold QSeats.  
	 Culling the search on non-Group subclasses is our main efficiency,
	 reducing the treewalking needed for seat-finding a great deal."

	self class == TGroup ifTrue: [ 
		^ self genericBestSeatWithin: tbox nearest: avatarPosition
	].
	^nil! !

!TGroup methodsFor: '*Qwaq-Objects-Space' stamp: 'bgf 4/1/2010 09:10'!
canBeCopiedBy: anAvatar
	"Make sure my user-relevant children also allow copying to this inventory.
	  (We use childrenForBrowser as an economy, to avoid large scene-trees of irrelevant widgets.
	  It filters for the persistence-relevant things.) "

	^ (super canBeCopiedBy: anAvatar) 
	    and: [ self childrenForBrowser allSatisfy: [ : c | c canBeCopiedBy: anAvatar ]]! !

!TGroup methodsFor: '*Qwaq-Objects-Space' stamp: 'bgf 7/26/2009 13:28'!
genericBestSeatWithin: tbox nearest: avatarPosition
	| seats curDist curSeat |
	"The generic recursive seat-finder.
	 Finds the best candidate among the candidates offered by its children."

	frameChildren ifNil:[ ^ nil ].
	seats :=  Array streamContents: [ :as |
				frameChildren do: [ : child |
					( child bestSeatWithin: tbox nearest: avatarPosition ) ifNotNilDo: [ : seat |
						as nextPut: seat.
					]]].

	seats isEmptyOrNil ifTrue: [ ^ nil ].

	"To save time, we consider the global position of the seat rather than the exact stance;
	 this is imprecise due to seating offsets but fine for our purposes,
	 since seating-destination seatingRadius is approximate at best due to model shapes."
	curDist := Float infinity.
	curSeat := nil.
	seats do: [ : seat |		
				| d | 
				d := (seat globalPositionReal - avatarPosition) squaredLength.
				d <= curDist ifTrue: [
					curDist := d.
					curSeat := seat.
				]].
	^ curSeat 
! !

TGroup subclass: #TJointRoot
	instanceVariableNames: 'myScale'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Objects-Motion'!

!TJointRoot commentStamp: 'das 12/4/2006 10:04' prior: 0!
This is the parent of all of the TJoints.


This was seriously hacked up by DAS. I am just using this as a simple importer and will replace it later with something that imports directly into the form I want.!

!TJointRoot class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!TJointRoot methodsFor: 'initialize' stamp: 'bgf 9/15/2008 09:37'!
initialize
	super initialize.
	self visible: false.! !

!TJointRoot methodsFor: 'accessing' stamp: 'bgf 1/4/2007 02:25'!
makePipes

	| base |
	self setTransformChanged.	"To force recalc here, before I'm rooted."
	base := TGroup new.
	base localTransform: self localTransform.
	frameChildren ifNotNil:[frameChildren do:[:fc | fc makePipes:base.].].
	^ base.! !

!TJointRoot methodsFor: 'accessing' stamp: 'bgf 3/18/2007 12:26'!
makeSkeleton

	"Return a TGroup tree that can be the skeleton of an animated character;
	 each group shares a mocap node name, and has correct Leonardo positioning."

	| result |

	self setTransformChanged.   	"To force recalc of global position (before I'm rooted)"
	result := (TGroup new)
		objectName: 'BVH-Container';
		localTransform: self localTransform;
		yourself.
	(self childNamed: 'Root') makeSkeletonIn: result.
	^ result


! !

!TJointRoot methodsFor: 'accessing' stamp: 'zamp 6/16/2006 09:31'!
motionPlayer
	^self parent child: 4! !

!TJointRoot methodsFor: 'accessing' stamp: 'zamp 5/5/2006 10:38'!
myScale
	^myScale! !

!TJointRoot methodsFor: 'accessing' stamp: 'zamp 6/5/2006 16:07'!
myScale: scale
	myScale := scale! !

!TJointRoot methodsFor: 'events' stamp: 'zamp 5/10/2006 10:20'!
reset
	self frameChildrenDo: [ :child | child reset ].! !

!TJointRoot methodsFor: 'accessing' stamp: 'bgf 3/17/2007 23:50'!
rootHeight

	| rootLoc footLoc |

	self setTransformChanged.
	footLoc := Float infinity.
	self allFramesDo:[:frame |
		frame objectName = 'Root' ifTrue:[rootLoc := frame globalTransform translation y.].
		footLoc > frame globalTransform translation y ifTrue:[
			footLoc := frame globalTransform translation y.
		].
	].
	^ rootLoc - footLoc.

! !

!TJointRoot methodsFor: 'accessing' stamp: 'hrs 7/23/2007 15:09'!
spanHeight
	| rootLoc footLoc thisY|
	"Should we just make TJoint follow boundingBox protocol instead?"

	self setTransformChanged.
	footLoc := Float infinity.
	rootLoc := 0.0 - Float infinity.
	self allFramesDo:[:frame |
		thisY := frame globalTransform translation y.
		rootLoc := rootLoc max: thisY.
		footLoc := footLoc min: thisY.
	].
	^ rootLoc - footLoc.

! !

Heap subclass: #QEventLoopMessageQueue
	instanceVariableNames: 'mutex sequenceNumber'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Objects-Kernel'!

!QEventLoopMessageQueue commentStamp: 'jcg 4/6/2010 01:05' prior: 0!
Note: copied/renamed from Squeak 4.1 "MorphicAlarmQueue".

MorphicAlarmQueue is a specialized Heap.  The main change is to stamp each added MorphicAlarm with a sequence number to ensure that alarms scheduled for the same time are executed in the order that they were added.!

!QEventLoopMessageQueue class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QEventLoopMessageQueue class methodsFor: 'class initialization' stamp: 'jcg 1/9/2010 12:42'!
convertAllAlarms
	"Alarms should be kept in a MorphicAlarmQueue, not a Heap."
	WorldState allSubInstancesDo: [:ws | ws convertAlarms]! !

!QEventLoopMessageQueue class methodsFor: 'class initialization' stamp: 'jcg 1/9/2010 12:37'!
initialize
	self convertAllAlarms.! !

!QEventLoopMessageQueue methodsFor: 'adding' stamp: 'nice 3/20/2010 08:01'!
add: aMorphicAlarm
	(sequenceNumber := sequenceNumber + 1) == 16r3FFFFFFF ifTrue: [
		"Sequence number overflow... reassign sequence numbers starting at 0."
		| alarmList |
		alarmList := self asArray sort: [:msg1 :msg2 |
			 msg1 sequenceNumber < msg2 sequenceNumber
		].
		alarmList withIndexDo: [:msg :ind | msg sequenceNumber: ind-1].
		"The #bitAnd: for the unlikely event that we have > 16r3FFFFFF messages in the queue."
		sequenceNumber := alarmList last sequenceNumber + 1 bitAnd: 16r3FFFFFFF.
	].
	aMorphicAlarm sequenceNumber: sequenceNumber.
	super add: aMorphicAlarm.
	
	"If we doubt our sanity..."
	false ifTrue: [
		self isValidHeap ifFalse: [self error: 'not a valid heap!!!!!!'].
	].
	^aMorphicAlarm! !

!QEventLoopMessageQueue methodsFor: 'initialize' stamp: 'jcg 1/7/2010 10:53'!
initialize
	super initialize.
	sequenceNumber := 0.! !

!QEventLoopMessageQueue methodsFor: 'private' stamp: 'jcg 1/9/2010 12:27'!
isValidHeap
	"Verify the correctness of the heap"
	2 to: tally do:[:i|
		(self sorts: (array at: i // 2) before: (array at: i)) ifFalse:[^false].
	].
	^true! !

!QEventLoopMessageQueue methodsFor: 'accessing' stamp: 'jcg 1/9/2010 14:43'!
mutex
	^mutex ifNil: [mutex := Mutex new]! !

!QEventLoopMessageQueue methodsFor: 'comparing' stamp: 'jcg 1/9/2010 13:30'!
sorts: alarmA before: alarmB
	alarmA scheduledTime = alarmB scheduledTime 
		ifFalse:[^alarmA scheduledTime < alarmB scheduledTime].
	alarmA sequenceNumber = alarmB sequenceNumber
		ifFalse:[^alarmA sequenceNumber < alarmB sequenceNumber].
	^self error: 'These alarms run at the same time'! !

TSimpleRectangle subclass: #QAvatarStandIn
	instanceVariableNames: 'avatar'
	classVariableNames: 'BillboardForm SmallBillboardForm'
	poolDictionaries: ''
	category: 'Qwaq-Objects-User'!

!QAvatarStandIn commentStamp: '<historical>' prior: 0!
This is the little cartoon-figure we render instead of the avatar
when the avatar is too far away to bother rendering as a mesh.
The implementation deals carefully with screen-sizing issues,
so the billboard standin appears at the size of the avatar it is replacing,
and shrinks only slowly as it gets further away.!

!QAvatarStandIn class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QAvatarStandIn class methodsFor: 'class variables' stamp: 'bgf 7/25/2009 16:44'!
defaultExtent
	"Texture is 64x128, and we want to sub in for an avatar, so be that size."
	^ Point x: 2.5 y: 5! !

!QAvatarStandIn class methodsFor: 'class variables' stamp: 'bgf 7/25/2009 13:16'!
defaultMaterial
	"Material for the little substitute-man we use when people are distant."
	^ (TMaterial new)
		texture: (QTextures findAlpha: 'Content/Qwaq/Textures/farAwayAvatar.png');
		diffuseColor: (VectorColor r: 1 g:1 b: 1 a: 0.999);
		opacity: 0.9995;
		uvScale: 0.995@0.995;
		uvOffset: (0.003@-0.003);
		cullFace: false;
		fullBright: true;
		yourself
! !

!QAvatarStandIn class methodsFor: 'utilities' stamp: 'bgf 7/26/2009 11:45'!
winCompensation: ogl
	"A magic number for reverse-scaling rectangles to adjust for the user's viewing situation,
	 when we want to control for net height. "
	| angle |
	angle := [ ogl view viewAngle ] on: Error do: [ 45.0 ].
	^ angle degreeCos! !

!QAvatarStandIn methodsFor: 'rendering' stamp: 'bgf 7/25/2009 13:57'!
alphaFor: ogl distance: distance squared: dSquared 

	"We start faraway distance-fading at the point it appears, linearly until we just wink out."
	| basis alpha |
	basis := avatar avatarIconDistance.
	alpha :=  (( basis / (distance max: basis)) min: 0.9955 max: 0.00).
	^ alpha < 0.15 ifTrue: [ 0 ] ifFalse: [ alpha ].! !

!QAvatarStandIn methodsFor: 'accessing' stamp: 'bgf 7/25/2009 12:40'!
avatar: atar

	avatar := atar.! !

!QAvatarStandIn methodsFor: 'accessing' stamp: 'bgf 7/25/2009 12:40'!
hasAlpha
	^ true! !

!QAvatarStandIn methodsFor: 'accessing' stamp: 'bgf 7/25/2009 12:40'!
hasOpaque
	^ false! !

!QAvatarStandIn methodsFor: 'initialize' stamp: 'bgf 11/28/2009 10:53'!
initialize
	super initialize.
	self pickable: false; solid: false.

	self translation: (Vector3 x: 0 y: (self class defaultExtent y * -0.4) z: 0).

! !

!QAvatarStandIn methodsFor: 'accessing' stamp: 'bgf 7/25/2009 12:40'!
isComponent
	^ false! !

!QAvatarStandIn methodsFor: 'accessing' stamp: 'bgf 7/25/2009 12:40'!
isInAvatar
	^ true! !

!QAvatarStandIn methodsFor: 'rendering' stamp: 'bgf 7/25/2009 12:40'!
pick: ray
	^ false! !

!QAvatarStandIn methodsFor: 'rendering' stamp: 'bgf 7/25/2009 12:40'!
render: ogl
	"Do nothing for the opaque pass."! !

!QAvatarStandIn methodsFor: 'rendering' stamp: 'bgf 7/25/2009 16:06'!
renderPrimitive: ogl alpha: bAlpha
	"Low-level rendering alpha pass ; render the name/icon if needed."
	| dSquared cp gt gp bearing mat factor isMe dForScale saveMc alpha scale | 

	"Lots of reasons to avoid doing this."
	self visible ifFalse: [ ^ 0 ].
	avatar isPresent ifFalse: [ ^ 0 ].
	(avatar costumeVisible: ogl) ifTrue: [ ^ 0 ].

	gt := self globalTransformReal.
	isMe := avatar isThisUserAvatar: ogl.
	isMe ifFalse: [
		"Not me - no local offset."
		dSquared  := (self globalPositionReal - ogl camera globalPosition) squaredLength.
	] ifTrue: [
		"Me - won't be anything in the way.  Honor local anti-wobble offset."
		dSquared := (self cameraDistanceSquared:ogl asPartOf: avatar).
		ogl offsetShift ifNotNilDo: [ : shift | gt := (shift  * gt) ].
	].

	dSquared < 400 ifTrue: [ ^ 0  ].
	dForScale := (dSquared sqrt max: 2).
	alpha := self alphaFor: ogl distance: dForScale squared: dSquared.
	alpha < 0.1 ifTrue: [ ^ 0 ].

	"Very magic numbers coordinated with defaultExtent and avatarIconDistance,
	 to get a constant-height icon that is similar to the height of the original at the point it kicks in,
	 and remains that size as you get further away."
	factor := (2.5 / (avatar avatarIconDistance)).
	factor := factor * (self class winCompensation: ogl).

	cp := ogl camera globalPosition.
	gp := gt translation.
	bearing := (cp x @ cp z) bearingToPoint: (gp x @ gp z).
	(ogl inPortal2D) 
		 ifTrue: [ bearing := (bearing + 180) asSmallPositiveDegrees . 
				factor := factor * 0.9 ].
	(ogl inPortal3D) 
		 ifTrue: [ factor := factor * 0.95 ].
	
	"The magic to face us with the label, but only in the x-z plane; 
	 the skew when angled looks quite fetching."
	scale := factor * dForScale.
	scale := scale * (dForScale ln sqrt) / (avatar avatarIconDistance ln sqrt).

	mat := gt orientation inverseByTranspose 
			* (Matrix4x4 identity 
				scaleX: scale y: scale z: scale; 
		     		addRotationAroundY: (0-bearing);
				yourself).
	
	saveMc := self material diffuseColor.
	self material diffuseColor: (VectorColor r: 1 g: 1 b: 1 a: alpha).
	ogl glPushMatrix.
	ogl glMultTransposeMatrixf: (mat).
	super renderPrimitive: ogl alpha: bAlpha.
	ogl glPopMatrix.
	self material diffuseColor: saveMc.! !

!QAvatarStandIn methodsFor: 'accessing' stamp: 'bgf 7/25/2009 12:55'!
shouldPersist
	^ false! !

TSimpleRectangle subclass: #QLabel
	instanceVariableNames: 'contents margin styleName textColor backgroundColor formChanged wrap textAnchorPoint'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Objects-Widgets'!

!QLabel commentStamp: '<historical>' prior: 0!
A Rectangle presenting styled text on a plain background.

QLabels format their content text into their (inherited) extent,
scaling the text so that it fits the rectangle at its natural font proportions;
the text will fill the label either vertically or horizontally. 
The extent should be set by the client before the QLabel appears.

The texture of the QLabel's material is generated and managed by
the label, and should not be set or extracted by client code
(and especially, it must not be shared.)

A margin is provided (the value is a number of pixels.)  
The margin is in the same dimensions as the pixel size of the texxt style;
it is applied during text formatting into the proportional rectangle,
before the formatted text is scaled into a texture.
The default margin is 4.

Currently, changing the label is inefficient, as each individual property
change results in full regeneration. 

!

QLabel subclass: #QAvatarLabel
	instanceVariableNames: 'avatar'
	classVariableNames: 'ParagraphStyle'
	poolDictionaries: ''
	category: 'Qwaq-Objects-User'!

!QAvatarLabel commentStamp: '<historical>' prior: 0!
This is the QAvatarReplica's nametag (label overhead).!

!QAvatarLabel class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QAvatarLabel class methodsFor: 'class variables' stamp: 'bgf 7/25/2009 16:52'!
defaultBackground
	^ Color black alpha: 0.32! !

!QAvatarLabel class methodsFor: 'class variables' stamp: 'bgf 11/28/2009 10:53'!
defaultExtent
	^ Point x: 2.5 y: 0.5! !

!QAvatarLabel class methodsFor: 'class variables' stamp: 'bgf 7/25/2009 01:57'!
defaultMargin
	^ 1! !

!QAvatarLabel class methodsFor: 'class variables' stamp: 'bgf 11/27/2009 17:49'!
defaultMaterial
	"We want a white fullbright basis for best results from text-painting."
	^ (TMaterial new)
		color: (QColors whiteAlphaColor);
		premultiplied: true;
		opacity: 0.999;
		fullBright: true;
		cullFace: false;
		yourself! !

!QAvatarLabel class methodsFor: 'class initialization' stamp: 'bgf 7/25/2009 17:58'!
initializeStyles
	" QAvatarLabel initializeStyles.
	 The two-font arrangement is assumed by instance method contentsAsText."
	ParagraphStyle := TextStyle fontArray: { 
					    (QStyles avatarFontStyle fontOfPointSize: 14) .
					    (QStyles avatarFontStyle fontOfPointSize: 11)
					  }.
	ParagraphStyle leading: 0.! !

!QAvatarLabel class methodsFor: 'class initialization' stamp: 'bgf 7/25/2009 17:57'!
style
	"On-demand initialization if not yet ready."
	^ ParagraphStyle ifNil: [
		self initializeStyles.
		ParagraphStyle
	].! !

!QAvatarLabel methodsFor: 'rendering' stamp: 'bgf 7/25/2009 13:33'!
alphaFor: ogl distance: distance squared: dSquared 
	"The avatar configuration tells when to start fading for nearness,
	 so big close views of self don't have enormous opaque labels."
	| alpha |
	alpha := 0.995.
	(dSquared-2 > (avatar opaqueThreshold)) 
	ifFalse: [ 
		alpha :=  ((dSquared-6) / (avatar opaqueThreshold*1.5 max: 1)) min: alpha max: 0.0
	 ] ifTrue: [ 
		"We start faraway distance-fading at 40 feet, linearly until we just wink out."
		alpha := (( 28 / (distance max: 28)) min: alpha max: 0.00).
	].

	alpha < 0.15 ifTrue: [ ^ 0.0 ].

	(ogl view owner usersPane isMouseOverUser: (avatar userLogin))
		ifTrue: [ alpha := alpha + 0.1 ].

	avatar isSpeaking ifTrue: [ alpha := alpha + 0.1 ].

	^ alpha! !

!QAvatarLabel methodsFor: 'accessing' stamp: 'das 12/4/2007 17:51'!
avatar: atar

	avatar := atar.! !

!QAvatarLabel methodsFor: 'private' stamp: 'bgf 7/25/2009 17:42'!
contentsAsText
	"The text I'll send to the paragraph builder."
	| txt i |
	txt :=  (self labelString asText)	
			addAttribute: (TextAlignment centered);
			yourself.

	i :=(txt string indexOf: (Character cr)).
	i > 0 ifTrue: [
		txt addAttribute: (TextFontChange fontNumber: 2)
		      from: i		
		      to: txt string size.
	].
	^ txt! !

!QAvatarLabel methodsFor: 'accessing' stamp: 'bgf 1/18/2010 13:44'!
frameBox
	"Overidden since the avatar label's visual extent is variable by viewer and distance
	 as if it were a 2D pixel-sized item."
	^ TBox min: (Vector3 x: -0.1 y: -0.02 z: -0.05) max: (Vector3 x: 0.1 y: 0.02 z: 0.05).! !

!QAvatarLabel methodsFor: 'accessing' stamp: 'bgf 7/24/2009 20:32'!
hasAlpha
	^ true! !

!QAvatarLabel methodsFor: 'accessing' stamp: 'bgf 7/24/2009 20:32'!
hasOpaque
	^ false! !

!QAvatarLabel methodsFor: 'initialize' stamp: 'bgf 7/25/2009 16:52'!
initialize
	super initialize.
	self pickable: false; solid: false.
	self styleName: #AvatarLabelStyle.
	self textColor: Color white.
	self translation: (Vector3 x: 0 y: 0.6 z: 0).! !

!QAvatarLabel methodsFor: 'accessing' stamp: 'bgf 3/6/2009 23:02'!
isComponent
	^ false! !

!QAvatarLabel methodsFor: 'accessing' stamp: 'bgf 3/6/2009 23:03'!
isInAvatar
	^ true! !

!QAvatarLabel methodsFor: 'rendering' stamp: 'bgf 12/2/2009 11:48'!
isNaturalSize
	"We enlarge to compensate for distance so LOD is not applicable."
	^ false! !

!QAvatarLabel methodsFor: 'private' stamp: 'bgf 7/29/2009 18:24'!
labelString
	"The string to use as the avatar label at this moment in time."
	| name filterOut |
	name := avatar userInfo publicName ifNil: [ ^ '' ].

	name := ' ', name, ' '.

	"If there is a presence notice, tack it on after the name.
	 This shows 'Idle', for instance, and 'Mic Off'...
 	 but not we filter out 'Mic Up' (the ptt released indicator)
	 which isn't so useful to see."
	avatar notices do: [ :notice|
		filterOut := #('Mic Up' 'Idle'). 
		(filterOut anySatisfy: [ : f | notice beginsWith: f]) ifFalse: [ 
			name := name, (String cr), notice
		].
	].
	avatar isMeetingLeader ifTrue: [
		name := name, (String cr), 'Leading Meeting'.
	].
	avatar wantsToSpeak ifTrue: [
		name := name, (String cr), 'Wants to Speak'
	].

	^ name! !

!QAvatarLabel methodsFor: 'private' stamp: 'bgf 7/25/2009 17:56'!
paragraph
	"My label as a styled paragraph for the base class to paint.
	 TODO - the lines below the name can be in a smaller font."
	
	^ (Paragraph withText: (self contentsAsText) style: (self class style) )
		perform: #centered! !

!QAvatarLabel methodsFor: 'rendering' stamp: 'bgf 3/6/2009 22:54'!
pick: ray
	^ false! !

!QAvatarLabel methodsFor: 'rendering' stamp: 'bgf 3/7/2009 02:26'!
render: ogl
	"Do nothing for the opaque pass."! !

!QAvatarLabel methodsFor: 'rendering' stamp: 'bgf 9/1/2009 15:55'!
renderPrimitive: ogl alpha: bAlpha
	"Low-level rendering alpha pass ; render the name/icon if needed."
	| dSquared cp gt gp bearing mat factor isMe dForScale saveMc alpha scale | 

	"Lots of reasons to avoid doing this."
	self visible ifFalse: [ ^ 0 ].
	avatar isPresent ifFalse: [ ^ 0 ].
	(avatar costumeVisible: ogl) ifFalse: [ ^ 0 ].
	(ogl view showNameFor: avatar userID whenSpeaking: (avatar isLabeledForSpeaking)) ifFalse: [ ^ 0 ].

	gt := self globalTransformReal.
	isMe := avatar isThisUserAvatar: ogl.
	isMe ifFalse: [
		"Not me - no local offset."
		dSquared  := (gt translation - ogl camera globalPosition) squaredLength.
	] ifTrue: [
		"Me - won't be anything in the way.  Honor local anti-wobble offset."
		dSquared := (self cameraDistanceSquared:ogl asPartOf: avatar).
		ogl offsetShift ifNotNilDo: [ : shift | gt := (shift  * gt) ].
	].
	dSquared < 4 ifTrue: [ ^ 0  ].
	dForScale := (dSquared sqrt max: 2).
	alpha := self alphaFor: ogl distance: dForScale squared: dSquared.
	alpha < 0.1 ifTrue: [ ^ 0 ].

	"Magic numbers to get a label size that is neither space-accurate nor constant-height,
	 but that reduces height smoothly and slowly with increasing distance."
	factor := (101 asFloat / (ogl view bounds height max: 400)).
	cp := ogl camera globalPosition.
	gp := gt translation.
	bearing := (cp x @ cp z) bearingToPoint: (gp x @ gp z).
	ogl isMirror
		 ifTrue: [ bearing := (bearing + 180) asSmallPositiveDegrees ].
	ogl inPortal2D 
		ifTrue: [ factor := factor * 0.9 ].
	ogl inPortal3D
		 ifTrue: [ factor := factor * 0.8 ].
	
	"The magic to face us with the label, but only in the x-z plane; 
	 the skew when angled looks quite fetching."
	scale := factor * dForScale / (dForScale ln sqrt).
	avatar isSpeaking ifTrue: [ scale := scale * 1.03 ].		"Big when I talk."
	mat := gt orientation inverseByTranspose 
			* (Matrix4x4 identity 
				scaleX: scale *0.94 y: scale z: scale*0.94; 
		     		addRotationAroundY: (0-bearing);
				translation: (Vector3 x: 0 y: (scale sqrt  -1) z: 0);
				yourself).
			"The x, z scale adjustment above (both horizontal dims) 
		          lightens the text a bit (narrowing regardless of view angle."

	saveMc := self material diffuseColor.
	self material diffuseColor: (VectorColor r: alpha g: alpha b: alpha a: alpha).		"Premultiplied"
	ogl glPushMatrix.
	ogl glMultTransposeMatrixf: (mat).
	super renderPrimitive: ogl alpha: bAlpha.
	ogl glPopMatrix.
	self material diffuseColor: saveMc.! !

!QAvatarLabel methodsFor: 'accessing' stamp: 'bgf 7/25/2009 12:55'!
shouldPersist
	^ false! !

!QAvatarLabel methodsFor: 'private' stamp: 'bgf 8/11/2009 10:42'!
update
	"Called by the avatar replica whenever any of the important properties change."
	| str lines len |
	self root ifNil: [ ^self ].
	avatar ifNil: [ ^ self ].

	str := self labelString.
	lines := str findTokens: { (Character cr) }.

	"We use len for approximate extent (cheaper than asking the truth.)
	 Discount all but the first line since the later lines are in a smaller font."
	lines size > 0 ifTrue: [
		len := lines allButFirst inject: lines first size into: [:max :each | max max: (each size)-2 ].
	] ifFalse: [
		len := 8	"Not yet set up with label string."
	].

	self margin: 0.
	self translation: (Vector3 x: 0 y: (avatar avatarTop  - (lines size *0.026)) z: 0).
	self rotationAroundY: 180.
	self textColor: (avatar isSpeaking ifTrue: [ Color white ] ifFalse: [ Color gray: 0.9 ]).
	self backgroundColor: (avatar isSpeaking 
							ifTrue: [ (avatar avatarColor *0.5) asColor alpha: 0.9 ] 
							ifFalse: [ (avatar avatarColor *0.32) asColor alpha:0.5 ]).
	self contents: (str).
	self extent: ( ((len*0.105) max: 0.7) @  (0.25 + (lines size - 1  * 0.24))).


! !

!QLabel class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QLabel class methodsFor: 'defaults' stamp: 'bgf 8/4/2009 23:16'!
defaultAnchorPoint

	^ #center! !

!QLabel class methodsFor: 'defaults' stamp: 'bgf 2/10/2007 18:47'!
defaultBackground
	^ Color white! !

!QLabel class methodsFor: 'defaults' stamp: 'bgf 2/10/2007 18:48'!
defaultContents
	^ 'Label'! !

!QLabel class methodsFor: 'defaults' stamp: 'bgf 11/28/2009 10:53'!
defaultExtent
	^ Point x: 3 y: 2! !

!QLabel class methodsFor: 'defaults' stamp: 'bgf 2/10/2007 18:49'!
defaultMargin
	"Margin (in unscaled pixels; the compositing coordinate space)"
	^ 4			! !

!QLabel class methodsFor: 'defaults' stamp: 'bgf 11/27/2009 18:24'!
defaultMaterial
	"The default material allows the texture (the text) 
	 to be rendered as close as possible to its source coloration,
	 with a faint emissive component for legibility in the dark.
	 Labels are cullFace: false so that they can be used freestanding,
	 without the odd behavior of a one-side-only sign."

	^ TMaterial new 
		ambientColor: (QColors gray99); 
		diffuseColor: (QColors blackAlphaColor); 
		emissiveColor: (QColors gray2);
		cullFace: false;
		yourself.! !

!QLabel class methodsFor: 'defaults' stamp: 'bgf 7/24/2009 20:31'!
defaultStyle
	"The default TextStyle to use for rendering me."
	^ QStyles find:#LabelStyle
! !

!QLabel class methodsFor: 'defaults' stamp: 'bgf 2/11/2007 09:49'!
defaultTextColor
	^ Color black! !

!QLabel methodsFor: 'accessing' stamp: 'bgf 2/11/2007 09:44'!
backgroundColor
	^ backgroundColor! !

!QLabel methodsFor: 'accessing' stamp: 'bgf 5/26/2008 13:29'!
backgroundColor: aColor

	| bc |
	bc := aColor asColor.
	bc = backgroundColor ifFalse: [
		backgroundColor := bc.
		self formChanged.
	].! !

!QLabel methodsFor: 'accessing' stamp: 'bgf 2/18/2009 22:22'!
canEditMaterial
	^ false! !

!QLabel methodsFor: 'persist' stamp: 'bgf 8/4/2009 23:17'!
constructFrom: pStream
	super constructFrom: pStream.

	self styleName: (pStream readString: 'styleName' default: styleName).
	self textColor:(pStream readColor: 'textColor' default: textColor) asColor.
	self backgroundColor: 
		(pStream readColor: 'backgroundColor' default: backgroundColor) asColor.
	self margin: (pStream readNumber: 'margin' default:margin).		
	self contents: (pStream readString: 'contents' default: (self contents)).
	self textAnchorPoint: (pStream readString: 'anchor' default: (self class defaultAnchorPoint)) asSymbol.! !

!QLabel methodsFor: 'accessing' stamp: 'bgf 3/1/2007 20:45'!
contents
	"The String displayed in this label."
	^ contents! !

!QLabel methodsFor: 'accessing' stamp: 'bgf 3/19/2010 15:10'!
contents: aTextOrString
	| cts |
	cts := aTextOrString asString.
	cts = contents ifFalse: [
		contents := cts clone.
		self formChanged.
	].! !

!QLabel methodsFor: 'private' stamp: 'bgf 7/24/2009 20:48'!
contentsAsText
	^ (contents ifNil: ['']) asText! !

!QLabel methodsFor: 'events' stamp: 'bgf 3/1/2009 12:50'!
contentsFromEditor: aString

	| tmp |
	self hiliteBriefly.

	"Labels are touchy about leading and trailing linefeeds."
	tmp := aString.
	[tmp isEmpty not and: [tmp last = (Character cr)]] whileTrue: [
		tmp := tmp allButLast.
	].
	[tmp isEmpty not and: [tmp first = (Character cr)]] whileTrue: [
		tmp := tmp allButFirst.
	].
	self contents:tmp.! !

!QLabel methodsFor: 'initializing' stamp: 'bgf 5/7/2007 13:03'!
destroy
	| txtr |
	super destroy.

	"Label textures can be large - make sure we scavenge as much as possible."
	txtr := self material texture.
	self material texture: nil.
	txtr ifNotNil:[ txtr destroy ].! !

!QLabel methodsFor: 'events' stamp: 'bgf 3/1/2009 12:50'!
editByAvatar: anAvatar

	"Open an editor on the label for the given avatar."
	anAvatar ifNil: [ ^ self ].

	QTextEditJob edit: contents 
				title: 'Edit Label' 
				forUser: anAvatar 
				recipient: self 
				selector: #contentsFromEditor:.

	"Flicker the label to reinforce the association."
	self hiliteBriefly.
! !

!QLabel methodsFor: 'accessing' stamp: 'bgf 7/18/2008 10:11'!
extent: ext
	"Regenerate the form when proportions change."
	(self extent ~= ext) ifTrue: [
		super extent: ext.
		self formChanged.
	].! !

!QLabel methodsFor: 'accessing' stamp: 'bgf 9/10/2007 15:16'!
foreground
	^ textColor! !

!QLabel methodsFor: 'accessing' stamp: 'bgf 5/26/2008 13:24'!
foreground: aColor

	| tc |
	tc := aColor asColor.
	tc = textColor ifFalse: [
		textColor := tc.
		self formChanged.
	].! !

!QLabel methodsFor: 'private' stamp: 'bgf 10/14/2007 18:47'!
formChanged

	"The sign appearance has changed.
	 Update the form lazily to avoid multiple rebuilds when attributes are changed."

	formChanged ifFalse: [
		self future updateForm.
		formChanged := true.
	].! !

!QLabel methodsFor: 'accessing' stamp: 'das 9/1/2008 21:20'!
getBackgroundColor
	^ backgroundColor! !

!QLabel methodsFor: 'accessing' stamp: 'das 8/29/2008 17:00'!
getText

	^ self contents
! !

!QLabel methodsFor: 'accessing' stamp: 'das 9/1/2008 21:21'!
getTextColor
	^textColor! !

!QLabel methodsFor: 'accessing' stamp: 'das 2/12/2007 16:39'!
hasAlpha
	^ backgroundColor isTranslucent or: [(textColor isTranslucent) or: 
		[super hasAlpha]].! !

!QLabel methodsFor: 'accessing' stamp: 'das 2/12/2007 16:39'!
hasOpaque

	"No opaque rendering unless necessary, 
      so that the background/foreground translucencies
	 are respected.  Usually I have no material."

	^ ((backgroundColor isTranslucent not) and: [textColor isTranslucent not]) 
		or: [super hasOpaque].
! !

!QLabel methodsFor: 'initializing' stamp: 'bgf 11/28/2009 10:54'!
initialize
	super initialize.

	wrap := nil.
	formChanged := false.
	styleName := #default.
	textAnchorPoint := #center.

	margin := self class defaultMargin.
	textColor := self class defaultTextColor.
	backgroundColor := self class defaultBackground.
	contents := self class defaultContents asString.

	self formChanged.

! !

!QLabel methodsFor: 'accessing' stamp: 'bgf 12/6/2006 17:26'!
margin
	^ margin! !

!QLabel methodsFor: 'accessing' stamp: 'bgf 5/26/2008 13:25'!
margin: anInteger
	(margin = anInteger) ifFalse: [
		margin := anInteger asInteger.
		self formChanged.
	].! !

!QLabel methodsFor: 'accessing' stamp: 'bgf 2/18/2009 22:21'!
material: aMaterial
	"Disregard.  The label manages its material internally,
	 including doing direct updates into the form bits,
	 so we don't want to receive new materials from the material editor,
	 since if they are static we'll clobber the harness-side texture form."
! !

!QLabel methodsFor: 'private' stamp: 'bgf 7/24/2009 20:50'!
paragraph
	| style |
	styleName = #default ifTrue: [
		style := self class defaultStyle
	] ifFalse: [
		style := QStyles find:styleName.
	].
	^ Paragraph withText: (self contentsAsText) style: style.! !

!QLabel methodsFor: 'persist' stamp: 'bgf 8/4/2009 23:16'!
persistOn: pStream
	super persistOn: pStream.
	
	pStream storeString: 'contents' value: contents default: (self class defaultContents).
	pStream storeString: 'styleName' value: styleName default: 'default'.
	pStream storeColor: 'textColor' value: textColor default: (self class defaultTextColor).
	pStream storeColor: 'backgroundColor' 
			value: backgroundColor
			default: (self class defaultBackground).
	pStream storeNumber: 'margin' value: margin default: (self class defaultMargin).
	pStream storeString: 'anchor' value: textAnchorPoint default: (self class defaultAnchorPoint).! !

!QLabel methodsFor: 'fileIn/Out' stamp: 'bgf 10/25/2008 14:15'!
postImportFromIslandCopier: importer
	"If this was taken with a lazy-update pending, reschedule."
	super postImportFromIslandCopier: importer.
	formChanged ifTrue: [
		formChanged := false.
		self formChanged.
	].
! !

!QLabel methodsFor: 'private' stamp: 'bgf 11/19/2009 14:46'!
rebuild
	"Generate the current sign Form from the contents and style,
	 after making sure we have a texture that refers to a unique harness-side form texture."

	| texture sha |

	texture := self material texture.
	texture ifNotNilDo: [ : txtr | 
		sha := txtr sha.
	] ifNil: [ 
		"My texture is a thumbnail for the off-island texture where the form will be kept."
		sha := TContentManager generateSHA: (TObjectID new).
		texture := TTexture new 
			initializeSHA: sha 
			form: (Form extent: 1@1 depth: 32)
			mipmap: false shrinkFit: true extension: nil static: true.
		texture bThumb: true.

		self material texture: texture.
		self material premultiplied: true. "Because Squeak dithers the text without normalizing."
	].
	self checkAlpha.

	"At this point we have an empty TTexture in my material with a unique SHA.
	 We now generate the full texture in the harness-side content manager."
	self rebuildFormTexture: sha.! !

!QLabel methodsFor: 'private' stamp: 'bgf 12/3/2009 20:58'!
rebuildFormTexture: sha

	"Generate the Form containing the composited label text,
	 and deposit it in the harness as a form texture using the SHA of my material texture,
	 which is generated uniquely for just this purpose.

       This routine can be called from render,
	 and so it MUST NOT have side effects inside the space.
 	 The form is composed and then delivered directly to the harness-side TTexture"

	| para marginx2 minXPixels minYPixels rectAspect formHeight formWidth ofs form ext smallForm shrink marge tscale realForm bb |

	para := self paragraph.
	wrap ifNotNil: [
		bb := para boundingBox.
		para wrappingBox: ( 0@0 extent: ( (wrap @4096)  min:(bb extent)) )
			  clippingBox: ( 0@0 extent: ( (wrap @4096) min: (bb extent)) ).
	].

 "We need to allocate a form into which to draw the text,
  from which we generate a TTexture.
  It will be allocated to match my rectangle's aspect ratio,
  while being just large enough in pixel dimensions to accept the rendered text."

	marge := margin max: 1.	"para rendering to the edge gives cut edges - always have some room"
	marginx2 := 2 * marge.

	"The huge sizes are only temporary extremes for very long labels."
	minXPixels := para boundingBox width min: 4096 - marginx2 max: 10.
	minYPixels := para boundingBox height min: 2048 - marginx2 max: 10.
	rectAspect := (self extent y max:0.02) / (self extent x max:0.02).
	rectAspect < (minYPixels / minXPixels) ifTrue: [
		"Use text height as the domininant dimension."
		formHeight := minYPixels + marginx2.
		formWidth := (minYPixels / rectAspect) rounded + marginx2.
		ofs := (((formWidth - minXPixels)/2)) truncated @ (marge - 2).
	] ifFalse: [
		"Use text width as the dominant direction."
		formWidth := minXPixels + marginx2.
		formHeight := (minXPixels * rectAspect) rounded + marginx2.
		ofs := marge @ ((((formHeight - minYPixels)/2) truncated) - 1).
	].
	(textAnchorPoint includesSubstring: 'left' caseSensitive: false) ifTrue: [
		ofs := marge @ ofs y.
	].
	(textAnchorPoint includesSubstring: 'top' caseSensitive: false) ifTrue: [
		ofs := ofs x @ marge -2.
	].

	form := Form extent: formWidth@formHeight depth: 32.
	form fill: form boundingBox rule: (Form blend) fillColor: backgroundColor.
	para foregroundColor: textColor backgroundColor: Color transparent.
	para displayOn: form 
		 at: ofs 
		 clippingBox: (form boundingBox) 
		 rule: (Form over) 
		 fillColor: backgroundColor.

	"Make sure the form we end up with in the texture is capped at 1024 pixels."
	ext := form extent.
	tscale := nil.
	(ext maxDimension > 2047) ifTrue: [
		"If >= 2048, texture shrinkfit won't bring us down to 1024, so do it here."
		ext := ((ext * (1024.0 / (ext maxDimension asFloat))) ceiling) asSmallerPowerOfTwo.
		smallForm := Form extent: ext depth: 32.
		form displayScaledOn: smallForm cellSize: 2.
		form := smallForm.
		shrink := false.	"Already shrunk"
	] ifFalse: [
		"We want the final texture form to be 1024x1024 or smaller."
		shrink := (ext maxDimension > 1024).
		shrink ifFalse: [
			"If not forced to reduce, then use the exact-pixels 
	                 and use texture-scale to render only the needed part of the form
	          	 since that gives truest text-appearance (fewer pixels, but more faithful)"
			realForm := Form extent: (ext asLargerPowerOfTwo) depth: 32.
			form displayOn: realForm at: 0@0.
			tscale := ext / (realForm extent asFloatPoint).
			form := realForm.
		].
	].

	"And finally, send the prepared form to the harness...
	 the connection with my TTexture is only the sha."
	(self material texture) 	
		extent: ext.
	TContentManager default 
		createFormTexture: form sha: sha mipmap: true shrinkFit: shrink scale: tscale disposable: true.
! !

!QLabel methodsFor: 'accessing' stamp: 'jjl 9/26/2008 11:51'!
rectangleExtent: ext

	super extent: ext.! !

!QLabel methodsFor: 'render' stamp: 'bgf 10/14/2007 18:00'!
renderPrimitive: ogl alpha: bAlpha

	"This deals with making the harness-side sign texture after replication.
	 We simply ensure that the content manager has a texture for my texture's sha.
	 Note that rebuildFormTexture: is safe for use during render -  it doesn't generate
	 or change any island state, it just uses the label state to generate the label image off-island.

	 Also note that this is not the path by which changes are propagated.
	 That happens using the formChanged flag of the label.  
	 This check only makes sure that the harness has a form for my bits."

	self material texture ifNotNilDo: [ :txtr | 
		(ogl contentManager lookup: txtr sha) ifNil: [
			self rebuildFormTexture: txtr sha
		].
	].
	super renderPrimitive: ogl alpha: bAlpha! !

!QLabel methodsFor: 'accessing' stamp: 'das 9/1/2008 21:20'!
setBackgroundColor: aColor

	| bc |
	bc := aColor asColor.
	bc = backgroundColor ifFalse: [
		backgroundColor := bc.
		self formChanged.
	].! !

!QLabel methodsFor: 'accessing' stamp: 'das 8/29/2008 16:59'!
setText: aString

	self contents: aString
! !

!QLabel methodsFor: 'accessing' stamp: 'das 9/1/2008 21:21'!
setTextColor: aColor

	| tc |
	tc := aColor asColor.
	tc = textColor ifFalse: [
		textColor := tc.
		self formChanged.
	].! !

!QLabel methodsFor: 'accessing' stamp: 'bgf 12/7/2006 12:11'!
styleName
	"Style names are defined by QwaqStyle.  See QwaqStyle #findTextStyle:"
	^ styleName! !

!QLabel methodsFor: 'accessing' stamp: 'bgf 5/26/2008 13:26'!
styleName: aTextStyleName

	"Style names are defined by QwaqStyle as a small fixed set;
 	 See QwaqStyle #findTextStyle:"

	| sn |
	sn := aTextStyleName ifNil: [ #default ].
	sn = styleName ifFalse: [
		styleName := sn.
		self formChanged.
	].! !

!QLabel methodsFor: 'accessing' stamp: 'bgf 8/4/2009 23:15'!
textAnchorPoint
	^ textAnchorPoint ifNil: [ #center ]! !

!QLabel methodsFor: 'accessing' stamp: 'bgf 8/4/2009 23:32'!
textAnchorPoint: symbol
	"Only variants using top, left, and center are supported."
	textAnchorPoint = symbol ifFalse: [
		textAnchorPoint := symbol.
		self formChanged
	]! !

!QLabel methodsFor: 'accessing' stamp: 'bgf 2/11/2007 09:42'!
textColor
	^textColor! !

!QLabel methodsFor: 'accessing' stamp: 'bgf 5/26/2008 13:26'!
textColor: aColor

	| tc |
	tc := aColor asColor.
	tc = textColor ifFalse: [
		textColor := tc.
		self formChanged.
	].! !

!QLabel methodsFor: 'private' stamp: 'bgf 10/14/2007 18:12'!
updateForm
	"Carry out the reconstruction of the form, and take note that we've done so;
	 this is the worker routine for the lazy update process initiated by 'formChanged'"
	self rebuild.
	formChanged := false.
! !

!QLabel methodsFor: 'accessing' stamp: 'bgf 8/15/2008 17:17'!
wrap
	"Nil (meaning no-autowrap), or the integer width of the text compositing box."
	^ wrap! !

!QLabel methodsFor: 'accessing' stamp: 'bgf 3/19/2010 15:27'!
wrap: scalarOrNil
	"Private for component implementations (not persisted).
	 If non-nil, defines the pixel width of the text compositing box; lines autowrap to fit."
	wrap := scalarOrNil! !

TRectangle subclass: #QPlaneRing
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Objects-Manipulation'!

!QPlaneRing commentStamp: '<historical>' prior: 0!
A very simple planar translucent ring
(a rectangle, painted with a ring texture.)
Used by QPlanePivot to give interactive feedback during rotation.

The ring plane (TRectangle) is self-rotated 90 degrees around X,
so it appears horizontally; the ring is centered at the origin.!

!QPlaneRing class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QPlaneRing methodsFor: 'initialize' stamp: 'bgf 8/30/2007 13:24'!
initialize
	|  |

	super initialize.
	self material: ((TMaterial new)
			transparency: 0.999;
			texture: (QTextures findAlpha: 'rotationRing.png'); 
			emissiveColor: #(1.0 1.0 1.0 0.5);
			diffuseColor: #(1.0 1.0 1.0 0.999); 
			ambientColor: #(1.0 1.0 1.0 0.7);
			shininess: 0.2;
			cullFace: false;
			yourself).

	self rotationAroundX: 90;
		pickable: false.
		! !

!QPlaneRing methodsFor: 'initialize' stamp: 'bgf 9/19/2006 15:49'!
radius: aScalar

	self extent: (aScalar *2.25)@(aScalar*2.25).
! !
QColors initialize!
QDocumentType initialize!
QForms initialize!
QStyles initialize!
QLogger initialize!
QSpace initialize!
QPresenceManager initialize!
QUserPreferences initialize!
QEventLoopMessageQueue initialize!
